/* Seamless3d

Copyright (C) 2001-2017 Graham John Perrett (AKA thyme) thyme@seamless3d.com www.seamless3d.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this
software and associated documentation files (the "Software"), to deal in the Software
without restriction, including without limitation the rights to use, copy, modify,
merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to the following
conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT
OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.
*/

#include <windows.h>
#include <direct.h>
#include <process.h>
#include <shlobj.h>

#include <sys/stat.h>
#include <share.h>
#include <fcntl.h>
#include <io.h>

#include <vector>
#include <map>
#include <algorithm>


#include <iostream>
#include <GL/glew.h>
#include <GL/wglew.h>

#include <string>
#include <vector>

#include <initializer_list>
#include <iostream>
#include <fstream>
using namespace std;


#include "seamless3d.h"
#include "render-ogl.h"
#include "techuelife_triangle_editor.h"
#include "gui.h"
#include "scene_tree.h"
#include "resource.h"
#include "input.h"


#include "seamless3d_gui_render.h"
#include "robot_aid.h"
#include "chat_window.h"


#ifdef WACOM_FILES
#include "WacomMultiTouch.h"
#include "WintabUtils.h"

#define PACKETDATA	(PK_X | PK_Y | PK_BUTTONS | PK_NORMAL_PRESSURE)
#define PACKETMODE	PK_BUTTONS
#include "pktdef.h"
#endif


#define ENGLISH 1
//#define FRENCH 1

#ifdef ENGLISH
#include "global_strings.h"
#elif FRENCH
#include "global_strings.fr.h"
#endif

#ifdef RECORD_SOUNDMOVIE
extern HANDLE g_hNotificationEvent;
#endif

#ifdef PROG_ROB
bool g_controlDown = false;
bool g_altDown = false;
bool g_shiftDown = false;
bool g_tabDown = false;
bool g_zKeyDown = false;
#endif
bool g_useWinTab = false;
Vec2i g_extraWinPix;
void outputScreenShotForDemo();
bool smlsIsSaved();
void saveBeforeClosing(void(*i_yesFunc)(FileCP& fileCP), void(*i_noFunc)(FileCP& fileCP), void(*i_cancelFunc)(FileCP& fileCP));
bool InitRender();
extern bool g_viewPosDragMode;
extern bool g_touchViewPosDragMode;
extern Vec3f g_touchBeginMousePos;
extern Vec3f g_touchBeginViewPos;
extern bool g_viewRotDragMode;
extern bool g_viewZRotDragMode;
extern Rotation g_beginViewRot;
extern Vec3f g_beginViewPos;
extern Vec2i g_beginMousePos;

bool g_winAPI3DMode = false;
void release3dDrag();
void release3dRightDrag();
void mouse2dSync();
void initPen();

HACCEL hAccel;
extern bool g_browseMode;
void renderScene();
void initializeDeveloperScript();
void insertRecentFile(const char* fullFileName);
void openDroppedFile();
bool checkAnimCtrlPoint(const bool perspective, Vec3f sr);
extern String droppedFilename;
extern bool g_recordSoundMovie;
extern bool g_playSoundMovie;
extern bool g_playing;
extern Ptr<PartCageClue> g_partCageClue;
extern BrushPtr g_brush;
extern M_<Ptr<Viewpoint> > g_viewpoints;

bool g_online = false;
bool g_viewer = false;
bool g_minimized = false;
bool g_processComnandLine = true;


double g_previousClickTime = 0;
Vec2i g_previousLeftClickPos;
bool g_leftDoubleClicked = false;
double g_currentClickTime = 0;

extern bool g_windowSizeWindowBool;
extern double g_windowSizeWindowTime;
extern Vec3f g_viewDragMousePos;

//void mouseLeaveCleanUp();

bool g_bActive = true;//false; // Whether the app is active (not minimized)
bool g_bReady = false; // Whether the app is ready to render frames

HINSTANCE  g_hinst;

unsigned int colorToHex(Color3f color);
bool setDirectInput(HWND hWnd);
void openSmlsLastPhase(String fullFilePath, const bool recover = false);
void setupRecentFiles();
bool checkForLockFiles();
void updatePlayBuffer();
void updatePlayBufferRealTimeSynth();
extern bool g_playing;
extern bool g_showCamerasInAnimMode;

HRESULT InitDirectInput(HWND hWnd);
HRESULT FreeDirectInput();
BackUp::Monostable BackUp::monostable;
Rect clientOffest;

long g_crossCursor = 0;
long g_currentType2 = 0;
long g_modellingCursor = 0;

long g_lockedOnPointCursor = 0;
long g_addPointCursor = 0;
long g_insertLineCursor = 0;
long g_deleteLineCursor = 0;
long g_mergePatchesCursor = 0;
long g_addGeometryCursor = 0;
long g_addGeometryMirrorCursor = 0;
long g_ringSelectCursor = 0;
long g_rectangleSelectPointCursor = 0;
long g_rectangleUnselectCursor = 0;
long g_rectangleSelectPolyCursor = 0;
long g_fillCursor = 0;
long g_brushCursor = 0;
long g_eyeDropperCursor = 0;
long g_knifeCursor = 0;
long g_addBoneCursor = 0;
long g_insertBoneCursor = 0;

long g_arrowCursor = 0;
long g_handCursor = 0;
long g_waitCursor = 0;

long g_sizeAllCursor = 0;
long g_sizeNWSECursor = 0;
long g_sizeNESWCursor = 0;
long g_sizeWECursor = 0;
long g_sizeNSCursor = 0;







extern bool g_dontCloseMenu;
bool g_createdINI = false;
bool outPutIndices = false;
extern int	g_mirrorVertex = 0;
//BOOL g_localControl=false;
BackUp::SmlsSaved BackUp::smlsSaved;

double	g_timeBegan;
double g_time;
int g_viewIndex = 0;

extern float g_version;
extern bool initWinINet();
extern int g_numOfVisLines;

//void convertJoinPartToTEPart(IndexedFaceSet* part);
NodePtr g_before;
ArtPtr remPJ;
extern Ptr<Grid> g_grid;



//-----------------------------------------------------------------------------
// Global variables
//-----------------------------------------------------------------------------
//Vec2i g_3dwinPos;
//Vec2i g_3dwinSize=Vec2i(800,572-24*2-2);
//Vec2i g_sceneTreePos;
//Vec2i g_sceneTreeSize=Vec2i(400,0);
//Vec2i g_nodeWinPos=Vec2i(400, 570);


HWND hWndMain = NULL;
HWND g_hDlg = NULL;


//#define GET_WM_COMMAND_ID(wp, lp)               LOWORD(wp)
//#define GET_WM_COMMAND_HWND(wp, lp)             (HWND)(lp)
//#define GET_WM_COMMAND_CMD(wp, lp)              HIWORD(wp)
//BOOL CALLBACK selectedVertexProc(HWND hDlg,UINT iMessage, WPARAM wParam, LPARAM lParam);

bool windowSizeHasChanged();
void convertVisualCluesToRenderFormat();
void resetLooseEnds();
void addNodesForNewFile();
void initVisualClue();
void initRadialTug();



#ifdef WACOM_FILES

#define MAX_LOADSTRING 100

#include <memory>

typedef std::unique_ptr<WacomMTHitRect> WacomMTHitRectPtr;

// Global Variables:
std::map<int, WacomMTHitRectPtr>	g_lastWTHitRect;
std::map<int, WacomMTCapability>	g_caps;
std::vector<int>					g_devices;
extern HCTX g_tabCtx;

TCHAR szTitle[MAX_LOADSTRING];					// The title bar text
TCHAR szWindowClass[MAX_LOADSTRING];			// the main window class name



// Forward declarations of functions included in this code module:
ATOM				MyRegisterClass(HINSTANCE hInstance);
BOOL				InitInstance(HINSTANCE, int);
LRESULT CALLBACK	MainWndProc(HWND, UINT, WPARAM, LPARAM);
INT_PTR CALLBACK	About(HWND, UINT, WPARAM, LPARAM);

bool										g_UseWinHitRect = true;
bool										g_ObserverMode = false;
RECT										g_clientRect = { 0, 0, 0, 0 };

enum EDataType
{
	ENoData,
	EFingerData,
	EBlobData,
	ERawData
}g_DataType = EFingerData;

CRITICAL_SECTION g_graphicsCriticalSection;

//////////////////////////////////////////////////////////////////////////////
// Purpose
//		Dump the MT capabilities of the specified device.
//
void DumpCaps(int deviceID)
{
	WacomTrace("MT Capabilities for deviceID: %i\n", deviceID);
	WacomTrace("\tVersion: %i\n", g_caps[deviceID].Version);
	WacomTrace("\tDeviceID: %i\n", g_caps[deviceID].DeviceID);
	WacomTrace("\tType: %i\n", (int)g_caps[deviceID].Type);
	WacomTrace("\tLogicalOriginX: %f\n", g_caps[deviceID].LogicalOriginX);
	WacomTrace("\tLogicalOriginY: %f\n", g_caps[deviceID].LogicalOriginY);
	WacomTrace("\tLogicalWidth: %f\n", g_caps[deviceID].LogicalWidth);
	WacomTrace("\tLogicalHeight: %f\n", g_caps[deviceID].LogicalHeight);
	WacomTrace("\tPhysicalSizeX: %f\n", g_caps[deviceID].PhysicalSizeX);
	WacomTrace("\tPhysicalSizeY: %f\n", g_caps[deviceID].PhysicalSizeY);
	WacomTrace("\tReportedSizeX: %i\n", g_caps[deviceID].ReportedSizeX);
	WacomTrace("\tReportedSizeY: %i\n", g_caps[deviceID].ReportedSizeY);
	WacomTrace("\tScanSizeX: %i\n", g_caps[deviceID].ScanSizeX);
	WacomTrace("\tScanSizeY: %i\n", g_caps[deviceID].ScanSizeY);
	WacomTrace("\tFingerMax: %i\n", g_caps[deviceID].FingerMax);
	WacomTrace("\tBlobMax: %i\n", g_caps[deviceID].BlobMax);
	WacomTrace("\tBlobPointsMax: %i\n", g_caps[deviceID].BlobPointsMax);
	WacomTrace("\tCapabilityFlags: 0x%X\n", g_caps[deviceID].CapabilityFlags);
}

///////////////////////////////////////////////////////////////////////////////
// Wintab support functions
//

///////////////////////////////////////////////////////////////////////////////
//  FUNCTION: InitWintabAPI()
//
//  PURPOSE: Initialize the Wintab pen API.
//
//  COMMENTS:
//		Loads the Wintab32 DLL and sets up the API function pointers.
//

///////////////////////////////////////////////////////////////////////////////
// Multi-touch API support functions.
WacomMTHitRectPtr GetAppHitRect();
int FingerCallback(WacomMTFingerCollection *fingerData, void *userData);
int BlobCallback(WacomMTBlobAggregate *blobData, void *userData);
int RawCallback(WacomMTRawData *rawData, void *userData);
void AttachCallback(WacomMTCapability deviceInfo, void *userRef);
void DetachCallback(int deviceID, void *userRef);
void DrawFingerData(int count, WacomMTFinger *fingers, int device);
void DrawBlobData(int count, WacomMTBlob *blobs, int device);
void DrawRawData(int count, unsigned short* rawBuf, int device);
void DumpCaps(int deviceID);
bool ClientHitRectChanged(const WacomMTHitRectPtr& wtHitRect_I, int deviceID);
WacomMTError RegisterForData(int deviceID);
WacomMTError MoveCallback(int deviceID);
WacomMTError UnregisterForData(int deviceID);


HCTX InitWintabAPI(HWND hwnd_I)
{
	if (!LoadWintab())
	{
		//ShowError("Wintab not available");
		return 0;
	}

	char TabletName[50] = "";
	gpWTInfoA(WTI_DEVICES, DVC_NAME, TabletName);
	gpWTInfoA(WTI_INTERFACE, IFC_WINTABID, TabletName);

	/* check if WinTab available. */
	if (!gpWTInfoA(0, 0, NULL))
	{
		ShowError("WinTab Services Not Available.");
		return FALSE;
	}

	/* get default region */
	LOGCONTEXTA logContext = { 0 };
	gpWTInfoA(WTI_DEFSYSCTX, 0, &logContext);

	// No need to specify lcInOrg* and lcInExt* as they are the entire
	// physical tablet area by default.

	// Get messages
	logContext.lcOptions |= CXO_MESSAGES;

	// Move the system cursor.
	logContext.lcOptions |= CXO_SYSTEM;

	logContext.lcPktData = PACKETDATA;
	logContext.lcPktMode = PACKETMODE;
	logContext.lcMoveMask = PACKETDATA;
	logContext.lcBtnUpMask = logContext.lcBtnDnMask;

	//// Guarantee the output coordinate space to be in screen coordinates.
	logContext.lcOutOrgX = GetSystemMetrics(SM_XVIRTUALSCREEN);
	logContext.lcOutOrgY = GetSystemMetrics(SM_YVIRTUALSCREEN);
	logContext.lcOutExtX = GetSystemMetrics(SM_CXVIRTUALSCREEN); //SM_CXSCREEN );

	// In Wintab, the tablet origin is lower left.  Move origin to upper left
	// so that it coincides with screen origin.
	logContext.lcOutExtY = -GetSystemMetrics(SM_CYVIRTUALSCREEN);	//SM_CYSCREEN );

	/* open the region */
	return gpWTOpenA(hwnd_I, (LPLOGCONTEXT)&logContext, TRUE);
}



WacomMTHitRectPtr GetAppHitRect()
{
	if (g_UseWinHitRect)
	{
		WINDOWINFO appWindowInfo = { 0 };
		appWindowInfo.cbSize = sizeof(appWindowInfo);
		GetWindowInfo(hWndMain, &appWindowInfo);

		WacomMTHitRect hitRect = {
			(float)appWindowInfo.rcClient.left,
			(float)appWindowInfo.rcClient.top,
			(float)(appWindowInfo.rcClient.right - appWindowInfo.rcClient.left),
			(float)(appWindowInfo.rcClient.bottom - appWindowInfo.rcClient.top)
		};

		return WacomMTHitRectPtr(new WacomMTHitRect(hitRect));
	}
	return WacomMTHitRectPtr();
}

WacomMTProcessingMode CurrentMode(void)
{
	return g_ObserverMode ? WMTProcessingModeObserver : WMTProcessingModeNone;
}


WacomMTError RegisterForData(int deviceID)
{
	WacomMTError res = WMTErrorInvalidParam;

	WacomMTHitRectPtr wtHitRect;
	if (g_caps[deviceID].Type == WMTDeviceTypeIntegrated)
	{
		wtHitRect = GetAppHitRect();
	}

	switch (g_DataType)
	{
	case EFingerData:
		res = WacomMTRegisterFingerReadHWND(deviceID, CurrentMode(), hWndMain, 5);
		break;
	}

	g_lastWTHitRect[deviceID] = std::move(wtHitRect);

	return res;
}


void AttachCallback(WacomMTCapability deviceInfo, void *userRef)
{
	if (!g_caps.count(deviceInfo.DeviceID))
	{
		g_devices.push_back(deviceInfo.DeviceID);
		g_caps[deviceInfo.DeviceID] = deviceInfo;
		WacomMTError res = RegisterForData(deviceInfo.DeviceID);
		if (res != WMTErrorSuccess)
		{
			return;
		}
	}
}

void DetachCallback(int deviceID, void *userRef)
{
	if (g_caps.count(deviceID))
	{
		UnregisterForData(deviceID);
		std::vector<int>::const_iterator iter = std::find(g_devices.begin(), g_devices.end(), deviceID);
		if (iter != g_devices.end())
		{
			g_devices.erase(iter);
		}
		g_caps.erase(deviceID);
	}
}


WacomMTError UnregisterForData(int deviceID) {
	WacomMTError res = WMTErrorInvalidParam;

	switch (g_DataType)
	{
	case EFingerData:
		res = WacomMTUnRegisterFingerReadHWND(hWndMain);
		break;

	}

	g_lastWTHitRect[deviceID].reset();

	return res;
}

enum Gesture {
	FREE = 0,
	MOVE_CURSOR = 1,
	ZOOM_VIEW = 2,
	PAN_VIEW = 3,
	ROTATE_VIEW = 4
};

int g_gestureMode = FREE;

M_Bool g_fingerDown = M_Bool(false, false, false, false);



bool allFingersAreUp() {
	for (int i = g_fingerDown.length; --i >= 0;) {
		if (g_fingerDown[i] == true) {
			return false;
		}
	}
	return true;
}


Vec2i g_panFirstFingerDownPos;




bool setMousePos(int id, int index, WacomMTFinger *fingers, int device) {
	if (fingers[index].FingerID == id && fingers[index].Confidence) {
		double x = fingers[index].X;
		double y = fingers[index].Y;
		if (g_caps[device].Type == WMTDeviceTypeOpaque) {
			// If we're using an opaque tablet, then the X and Y values are not in
			// pixels; they are a percentage of the tablet width (eg: 0.123, 0.37, etc.).
			// We need to convert to client pixels.
			x *= g_clientRect.right - g_clientRect.left;
			x += g_clientRect.left;
			y *= g_clientRect.bottom - g_clientRect.top;
			y += g_clientRect.top;
		}
		POINT pt = { (LONG)x, (LONG)y };
		ScreenToClient(hWndMain, &pt);
		g_newMousePos = Vec2i(pt.x, pt.y);
		return true;
	}
	return false;
}

Vec3f mouse2dSync(Vec2i pixelPos);

void interpretFingers(int fingerCount, WacomMTFinger *fingers, int device) {
	static String str;
	static float beginZoom = 1;

	if (g_devices.size() && fingerCount && fingers) {
		EnterCriticalSection(&g_graphicsCriticalSection);


		for (int index = 0; index < fingerCount; index++) {
			if (fingers[index].FingerID >= g_fingerDown.length) {
				continue;
			}

			if (fingers[index].TouchState == WMTFingerStateUp) {
				g_fingerDown[fingers[index].FingerID] = false;
				if (allFingersAreUp()) {
					g_gestureMode = FREE;
					if (g_winAPI3DMode) {
						release3dDrag();
					}
				}

				continue;
			}

			else if (g_gestureMode == MOVE_CURSOR) {
				continue;
			}
			if (g_gestureMode == ZOOM_VIEW) {
				if (setMousePos(1, index, fingers, device)) {
					float diff = g_newMousePos.y - g_beginMousePos.y;
					diff = diff * .002;

					if (diff > 0) {
						g_scene->setZoom(beginZoom * (1.0 + diff));
					}
					else {
						g_scene->setZoom(beginZoom / (1.0 - diff));
					}
					houseKeepVisualClue();
				}
				continue;
			}
			else if (g_gestureMode == PAN_VIEW) {
				setMousePos(2, index, fingers, device);
				g_touchViewPosDragMode = true;
				continue;
			}
			else if (g_gestureMode == ROTATE_VIEW) {
				setMousePos(3, index, fingers, device);
				continue;
			}


			if (g_gestureMode == FREE) {
				if (fingers[index].TouchState != WMTFingerStateNone) {

					if (fingers[index].FingerID == 0) {
						if (g_fingerDown[0] == false && fingers[index].TouchState == WMTFingerStateDown) {
							g_fingerDown[0] = true;
						}
					}
					if (fingers[index].FingerID == 1) {
						if (fingerCount == 2 && setMousePos(1, index, fingers, device)) {
							if (g_fingerDown[1] == false && fingers[index].TouchState == WMTFingerStateDown) {
								g_panFirstFingerDownPos = g_newMousePos;
								g_fingerDown[1] = true;
							}
							if (g_fingerDown[1]) {

								if (abs(g_newMousePos.y - g_panFirstFingerDownPos.y) > 3) {
									g_gestureMode = ZOOM_VIEW;
									g_beginMousePos = g_newMousePos;
									beginZoom = g_scene->getZoom();


									continue;

								}
							}
						}
					}
					else if (fingers[index].FingerID == 2) {
						// Skip this finger if NOT confident.
						if (fingerCount == 3 && setMousePos(2, index, fingers, device)) {

							if (g_fingerDown[2] == false && fingers[index].TouchState == WMTFingerStateDown) {
								g_panFirstFingerDownPos = g_newMousePos;
								g_fingerDown[2] = true;
								continue;
							}
							if (g_fingerDown[2]) {

								Vec2i diff = g_newMousePos - g_panFirstFingerDownPos;

								if (abs(diff.x) > 3 || abs(diff.y > 3)) {
									g_gestureMode = PAN_VIEW;
									Button::repeatCounter = 0;
									removeTypeFromList<Button>(g_callBackList);
									Window::removeToolTip();
									Window::removeMenu();

									g_prevMousePos = g_newMousePos;
									g_beginMousePos = g_newMousePos;
									g_touchBeginMousePos = mouse2dSync(g_newMousePos);
									g_touchBeginViewPos = g_scene->translation;

									g_winAPI3DMode = true;
									setMouseCapture();
									//Window::checkIfAWindowHasBeenMidClickedAndUpdate(g_newMousePos);
									//if(g_viewPosDragMode){
									g_touchViewPosDragMode = true;
									//}
									g_previousLeftClickPos = g_newMousePos;
								}
							}
						}
					}
					else if (fingers[index].FingerID == 3) {
						if (fingerCount == 4 && setMousePos(3, index, fingers, device)) {

							if (g_fingerDown[3] == false && fingers[index].TouchState == WMTFingerStateDown) {
								g_panFirstFingerDownPos = g_newMousePos;
								g_fingerDown[3] = true;
								continue;
							}
							if (g_fingerDown[3]) {


								Vec2i diff = g_newMousePos - g_panFirstFingerDownPos;

								if (abs(diff.x) > 3 || abs(diff.y > 3)) {
									g_gestureMode = ROTATE_VIEW;
									Button::repeatCounter = 0;
									removeTypeFromList<Button>(g_callBackList);
									Window::removeToolTip();
									Window::removeMenu();

									g_prevMousePos = g_newMousePos;

									g_winAPI3DMode = true;
									setMouseCapture();

									if ((controlKeyState()) || InputEvent::l_ctrlDown) {
										g_viewZRotDragMode = true;
									}
									else {
										g_viewRotDragMode = true;
									}

									g_beginMousePos = g_newMousePos;
									g_beginViewRot = g_scene->rotation;

									mouse2dSync();
								}
							}
						}
					}
				}
			}
		}
		LeaveCriticalSection(&g_graphicsCriticalSection);
	}
}


//end wacom functions////////////////////
#endif

char defaultMsg[] = { "message from thyme's code" };

tagTRACKMOUSEEVENT g_trackMouseEvent = { sizeof(tagTRACKMOUSEEVENT), TME_LEAVE, 0, HOVER_DEFAULT };
//tagTRACKMOUSEEVENT g_trackMouseEvent={sizeof(tagTRACKMOUSEEVENT), 0, 0, HOVER_DEFAULT};

String g_seamlessDir;
const String& createSeamlessConstDirString(String& str) {
	getFolderPath(str, CSIDL_MYDOCUMENTS);
	//getFolderPath(str, CSIDL_PROGRAM_FILES);
	int strPathLen = str.length;
	if (strPathLen > 0 && str[strPathLen - 1] == '\\' || strPathLen > 0 && str[strPathLen - 1] == '/') {
		str += "Seamless3d\\";
	}
	else {
		str += "\\Seamless3d\\";
	}
	return str;
}

const String g_appDataDir = createSeamlessConstDirString(g_seamlessDir) + "Application Data For Seamless3d\\";
//const String g_undoDir = g_appDataDir + "undoCache\\";
const String g_undoDir = g_appDataDir + "Recover\\";
const String g_demoDir = g_appDataDir + "For Robot Demos Only\\";
const String g_demoCacheDir = g_demoDir + "Cache\\";
const String g_setupPath = g_appDataDir + "seamless3dSetup.smls";
const String g_debugLogPath = g_appDataDir + "debugLog.txt";
const String g_screenshotDir = g_seamlessDir + "Screenshot\\";


void showCursor(bool show) {
	static bool remVis = true;
	if (remVis != show) {
		remVis = show;
		ShowCursor(show);
	}

}
void getFolderPath(String& str, unsigned id) {
	TCHAR buf[MAX_PATH];
	SHGetSpecialFolderPath(
		NULL,
		buf,
		id,
		NULL
	);
	//	str="";
	for (int i = 0; buf[i] != 0; i++) {
		str += buf[i];
	}
}
void makeTitleStr(char* str) {
	char tempStr[0x200];
	char drive[0x200];
	char dir[0x200];
	char ext[0x200];

	_splitpath(g_currentFilename.c_str(), drive, dir, str, ext);
	strcat(str, ext);

#ifdef OPENGL
	sprintf(tempStr, " - Seamless3d %.3f OpenGL", g_version);
#elif TTS
	sprintf(tempStr, " - Seamless3d TTS %.3f", g_version);
#else
	sprintf(tempStr, " - Seamless3d %.3f", g_version);
	//	sprintf(tempStr," - Seamless3d %.5f",g_version);
#endif


	if (!BackUp::smlsSaved) {
		strcat(str, " *");
	}
	strcat(str, tempStr);
	strcat(str, " - ");
	strcat(str, g_selNode->name.c_str());
}
void setTitleName() {
	//selectPartFromVertex(g_seamless);
	char TempStr[0x200];
	makeTitleStr(TempStr);
	SetWindowText(hWndMain, TempStr);
	//SendMessage(g_hDlg, WM_USER, 0, 0);
}
void setSmlsSaved() {
	BackUp::smlsSaved.set();
	setTitleName();
}
void BackUp::Monostable::beginCycle() {
	if (beginBool == false) {
		backedUpFileCntr++;
	}
	beginBool = true;
	setTitleName(); //this might as well be done here in case its been forgotten and cant do any significant harm slowing things down
	beginTime = g_time;
	g_selCount.reset();//this might as well be done here in case its been forgotten and cant do any significant harm slowing things down
}
bool BackUp::Monostable::state()const {
	return beginBool && (g_time > beginTime + g_settings->backupDelay);
}
inline void BackUp::checkForBackUpOperation() {
	if (mouse3dMode == false && DownButton::getDownButton() == 0) {
		if (monostable.state()) {
			backupToCache();
			monostable.clear();
		}
	}
}

int msG(String str1, String str2, String str3, unsigned dlgTypeAndIcon) {

	String tempStr = String(str1);
	tempStr += str2;
	tempStr += str3;

	return MessageBox(hWndMain, tempStr.c_str(), defaultMsg, dlgTypeAndIcon);
}

int msG(String str, unsigned dlgTypeAndIcon) {
	return MessageBox(hWndMain, str.c_str(), defaultMsg, dlgTypeAndIcon);
}

extern float g_remVersionLoaded;

void changeExtensionToSmls(String& str) {
	for (int i = str.length; --i >= 0;) {
		if (str[i] == '.') {
			str = str.substr(0, i);
			break;
		}
		else if (str[i] == '\\' || str[i] == '/') {
			break;
		}
	}
	for (int i = 2; ; i++) {
		String newName;
		newName += str;
		newName += i;
		newName += ".smls";
		FILE* file;
		if ((file = fopen(newName.c_str(), "rb")) == 0) {
			str = newName;
			return;
		}
		else {
			fclose(file);
		}
	}
}


void searchTillText(char*& p);

void openSceneTreeInterface() {

	g_stw->newNodeBar.dstPos = Vec2i(220, 25 + 16 + 8);
	g_numOfVisLines = (g_stw->size.y - SceneTree::beginYOffset) / pixSymbolCY;

	g_stw->size = Vec2i(g_settings->sceneTreeSize.x, g_settings->sceneTreeSize.y);
	if (g_stw->size.x < 10) {
		g_stw->size = Vec2i(350, g_stw->size.y);
	}
	if (g_stw->size.y < 10) {
		g_stw->size = Vec2i(g_stw->size.x, 400);
	}
	g_stw->pos = Vec2i(g_settings->sceneTreePos.x, g_settings->sceneTreePos.y);
	g_stw->pixBuf.length = g_stw->size.x*g_stw->size.y;


	//	g_stw->paint();
}
/*
void roundToNearest(){
	ArtPtr part=g_selNode;
	SeamlessPtr smls=part->smls;
	M_Vec3f& point=smls->refPoint();
	Vec3f vec;
	vec=point[smls->ActCtrlPnt].multiply(100);
	vec.x=floor(vec.x+.5);
	vec.y=floor(vec.y+.5);
	vec.z=floor(vec.z+.5);
	point[smls->ActCtrlPnt]=vec.multiply(.01);

	smls->updateAllVertexInBuffers();
	makePointerEqualIndexLocation();
}
*/
void autoAddColor(ArtPtr part) {
	if (part->color.length == 0) {
		M_Art& partList = part->smls->partList;
		for (int k = partList.length; --k >= 0; ) {
			Art* part = partList[k];
			part->color.length = part->coord.length;
		}
	}
}

void workOutVertexSameParts(int& newFrom, int& newTo, M_Int fromVertex, M_Int toVertex) {
	bubbleSort(fromVertex, toVertex);

	for (int i = 0; i < fromVertex.length; i++) {
		if (fromVertex[i] <= newFrom) {
			newFrom++;
		}
		if (fromVertex[i] <= newTo) {
			newTo++;
		}
	}
}
void workOutVertexDiffParts(int& newFrom, M_Int fromVertex, M_Int toVertex) {
	bubbleSort(fromVertex, toVertex);

	for (int i = 0; i < fromVertex.length; i++) {
		if (fromVertex[i] <= newFrom) {
			newFrom++;
		}
	}
}

/*
void loopBackIFS(Seamless* smls){
	M_Build& build=smls->build.refValue();
	for(int k=build.length; --k>=0;){
		if(build[k]->getType()==STEM){
			Stem* stem=dynamic_cast<Stem*>(build[k].getRawPtr());
			M_Node& source=stem->source.refValue();
			M_SArt& part=stem->part.refValue();
			for(int i=source.length; --i>=0; ){
				if(i<part.length){
					Art* prt=dynamic_cast<Art*>(part[i].getRawPtr());
					IndexedFaceSet* ifs=dynamic_cast<IndexedFaceSet*>(source[i].getRawPtr());
					if(ifs!=0){
						int remPointSize=ifs->refPoint().length;
						int remColorSize=ifs->refColor().length;
						int remTriSize=ifs->refCoordIndex().length/4;
						//source[i]=dynamic_cast<Art*>(part[i].getRawPtr())->extractGeometry();
						ifs->copyNotClose(prt->extractGeometry());
						if(remPointSize!=ifs->refPoint().length){
							String str=prt->getName();
							str+=" coord diff ";
							str+=ifs->refPoint().length-remPointSize;
							dLog(str);
						}
						if(remColorSize!=ifs->refColor().length){
							String str=prt->getName();
							str+=" color diff ";
							str+=ifs->refColor().length-remColorSize;
							dLog(str);
						}
						if(remTriSize!=ifs->refCoordIndex().length/4){
							String str=prt->getName();
							str+=" triangle diff ";
							str+=ifs->refCoordIndex().length/4-remTriSize;
							dLog(str);
						}
					}
				}
			}
		}
	}
	smls->doBuild();
	smls->convertToRenderFormat();
	genNodeList();
	SendMessage(hSceneTreeWnd, WM_PAINT, 0, 0);
	makePointerEqualIndexLocation();
}
*/
bool checkVertexBelongsToTriangle(Art* vertexPart, Art* triPart, int vertexIndex, int &corner, int triangleIndex);
/*
void transferSingleVertex(){
	if(remPJ==NULL)return;
	Color3f scolor;
	ArtPtr part=g_selNode;
	SeamlessPtr &smls=part->smls;
	BackUp::monostable.beginCycle();
	int _actCtrlPnt=smls->ActCtrlPnt;
	M_Int &coordIndex=smls->geometry->coordIndex;

	Vec3f svec = smls->geometry->coord->point[_actCtrlPnt];
	svec=part->matrix*svec-remPJ->matrix*Vec3f();

	if(smls->geometry->color!=NULL){
		scolor=smls->geometry->color->color[_actCtrlPnt];
	}

	smls->genVertex(remPJ, svec, scolor);

	if(part->plidx>remPJ->plidx){
		_actCtrlPnt++;
	}

	//replace old vertex reference with the new vertex reference
	for(int i = -1; i<coordIndex.length-1 ; ){
		if(coordIndex[++i]==_actCtrlPnt){
			coordIndex[i]=remPJ->toCoord-1;
		}
		if(coordIndex[++i]==_actCtrlPnt){
			coordIndex[i]=remPJ->toCoord-1;
		}
		if(coordIndex[++i]==_actCtrlPnt){
			coordIndex[i]=remPJ->toCoord-1;
		}
		i++;
	}

	smls->deleteVertex(_actCtrlPnt); //delete old vertex

	smls->orderPolygonsFromVertices();

	smls->convertToRenderFormat();

	smls->ActCtrlPnt=remPJ->toCoord-1;
	makePointerEqualIndexLocation();
}
*/
/*
void joinCloseVertexImmediate(){
	ArtPtr part=g_selNode;
	SeamlessPtr &smls=part->smls;
	BackUp::monostable.beginCycle();

	JoinCloseVertex joinCloseVertex;
	joinCloseVertex.part=part;
	joinCloseVertex.joinRange=g_joinRange;
	joinCloseVertex.build();

	smls->convertToRenderFormat();

	makePointerEqualIndexLocation();
}
*/
/*
void reset3dCursor(){
	g_pointer->translation=Vec3f(0,0,0);
	g_pointer->rotation=Rotation();
	g_pointer->scale=Vec3f(1,1,1);
}
*/
bool _3DWin::click(const Vec2i ptCursor) {
	return false;
}
void Window::checkIfAWindowHasBeenClickedAndUpdate(const Vec2i& mousePos) {
	if (Window::activeWindow->hog) {
		if (false == Window::activeWindow->click(mousePos)) {
			PlaySound("SystemDefault", 0, SND_ASYNC);
		}
	}
	else {
		for (int i = visWinList.length; --i >= 0;) {
			if (visWinList[i]->click(mousePos)) {
				return;
			}
		}
		activeWindow = g_3d.p;
		mouse2dSync();
		//enablePen();
		g_winAPI3DMode = true;
		setMouseCapture();
		g_viewDragMousePos = g_pointer.translation;
		SenseTugGlobals::initTug();
	}
}
inline void Window::checkIfAWindowHasBeenMidClickedAndUpdate(const Vec2i& mousePos) {
	if (Window::activeWindow->hog) {
		if (false == Window::activeWindow->midClick(mousePos)) {
			PlaySound("SystemDefault", 0, SND_ASYNC);
		}
	}
	else {
		for (int i = visWinList.length; --i >= 0;) {
			if (visWinList[i]->midClick(mousePos)) {
				return;
			}
		}
		activeWindow = g_3d.p;
		//enablePen();
		g_winAPI3DMode = true;
		setMouseCapture();
		g_viewPosDragMode = true;
		SenseTugGlobals::initMidTug();
	}
}

void setMouseCapture() {
	g_trackMouseEvent.hwndTrack = hWndMain;
	TrackMouseEvent(&g_trackMouseEvent);
	SetCapture(hWndMain);
}

inline void Window::checkIfAWindowHasBeenRightClickedAndUpdate(const Vec2i& mousePos) {
	if (Window::activeWindow->hog) {
		if (false == Window::activeWindow->rightClick(mousePos)) {
			PlaySound("SystemDefault", 0, SND_ASYNC);
		}
	}
	else {
		for (int i = visWinList.length; --i >= 0;) {
			if (visWinList[i]->rightClick(mousePos)) {
				return;
			}
		}
#ifdef PROG_ROB
		dLog("\nnew Move(" + String(g_mousePos.x) + ',' + String(g_mousePos.y) + "),");
		dLog("\nnew RightButtonDown(),");
#endif

		activeWindow = g_3d.p;
		//enablePen();
		g_winAPI3DMode = true;
		setMouseCapture();
		if ((controlKeyState()) || InputEvent::l_ctrlDown) {
			g_viewZRotDragMode = true;
		}
		else {
			g_viewRotDragMode = true;
		}
		g_beginMousePos = mousePos;
		g_beginViewRot = g_scene->rotation;
		mouse2dSync();
	}
}


inline void Window::checkIfMouseIsHoveringAWindowAndUpdate(const Vec2i& mousePos) {
	if (g_remMenu == 0 && Window::activeWindow->hog) {
		if (Window::activeWindow->hover(mousePos) == false) {
			changeCursor(g_arrowCursor);
		}
	}
	else {
		for (int i = visWinList.length; --i >= 0;) {
			if (visWinList[i]->hover(mousePos)) {
				return;
			}
		}
		if (false == g_panViewMode) {
			//changeCursor(IDC_CROSS);
			//g_modellingCursor = MAKEINTRESOURCE(IDC_CROSS);
			//LPCTSTR lp = MAKEINTRESOURCE(LOCKED_ON_POINT_CURSOR);
			//LONG h = (LONG)LoadCursor(g_hinst, lp);
			//SetClassLong(hWndMain, GCL_HCURSOR, h);
			//(LONG)LoadCursor(0, requestType);
			changeCursor(g_modellingCursor);
		}
	}
}
/*
bool isDoubleLeftClick(){
	if(g_leftDoubleClicked == false && g_previousLeftClickPos == g_mousePos){

		double difference = g_currentClickTime - g_previousClickTime;
		double doubleClickTime = GetDoubleClickTime() * MILISECONDS_TO_SECONDS;

		//if(difference <= doubleClickTime){
		if(difference > doubleClickTime * .1 && difference <= doubleClickTime){
			return true;
		}
	}
	return false;
}
*/

bool isDoubleLeftClick() {
	if (g_leftDoubleClicked == false) {
		if (abs(g_mousePos.x - g_previousLeftClickPos.x) <= 6 && abs(g_mousePos.y - g_previousLeftClickPos.y) <= 6) {

			double difference = g_currentClickTime - g_previousClickTime;
			double doubleClickTime = GetDoubleClickTime() * MILISECONDS_TO_SECONDS;

			if (difference <= doubleClickTime) {
				return true;
			}
		}
	}
	return false;
}


void hoverObject() {
	MultiSelect::resetDistance();
	SenseTugGlobals::node = 0;
	MultiSelect::withinSightsNode = 0;
	if (Viewpoint::withinSitesViewpoint) {
		int a = 0;
	}
	Viewpoint::withinSitesViewpoint = 0;



	if (Viewpoint* selView = dynamic_cast<Viewpoint*>(g_selNode.getRaw())) {
		if (selView->isOverPoint()) {
			MultiSelect::withinSights = true;
			MultiSelect::withinSightsNode = selView;
			changeModellingCursor(g_lockedOnPointCursor);
			return;
		}
	}
	{

		if (MultiSelect::objectList.length && g_transformObj->isOverLever()) {
			return;
		}

		g_scene->senseTug(Matrix(), Matrix(), 0);
		if (SenseTugGlobals::node) {
			MultiSelect::withinSights = true;
			if (Art* part = dynamic_cast<Art*>(SenseTugGlobals::node)) {
				MultiSelect::withinSightsNode = part->smls;
				g_transformObj->mode = TransformObj::SELECT;
			}
			else {
				MultiSelect::withinSightsNode = SenseTugGlobals::node;
			}
			changeModellingCursor(g_handCursor);
		}
		else {
			changeModellingCursor(g_crossCursor);
		}
	}
}


void hoverCameraAnimMode() {
	MultiSelect::resetDistance();
	SenseTugGlobals::node = 0;
	MultiSelect::withinSightsNode = 0;
	if (Viewpoint::withinSitesViewpoint) {
		int a = 0;
	}
	Viewpoint::withinSitesViewpoint = 0;



	{

		if (MultiSelect::objectList.length && g_transformObj->isOverLever()) { //dont think we need this
			return;
		}

		for (int k = g_viewpoints.length; --k >= 0;) {
			g_viewpoints[k]->senseTug(Matrix(), Matrix(), 0);
		}
	}
	if (SenseTugGlobals::node) {
		MultiSelect::withinSights = true;
		MultiSelect::withinSightsNode = SenseTugGlobals::node;
		changeModellingCursor(g_handCursor);
	}
	else {
		changeModellingCursor(g_crossCursor);
	}
}


void hoverWin3d() {
	if (g_rectangleSelectPointReady || g_rectangleUnselectPointReady || g_rectangleSelectPolyReady) {
		return;
	}
	MultiSelect::withinSights = false;
	MultiSelect::withinSightsNode = 0;
	g_partCageClue->mode = 0;
	Window::checkIfMouseIsHoveringAWindowAndUpdate(g_mousePos);

	if (g_addBox.mousePhase == AddBox::READY_FOR_B || g_addBox.mousePhase == AddBox::READY_FOR_C) {
		mouse2dSync();
		g_addBox.updateDimensions();
	}
	else if (g_addSphere.mousePhase == AddSphere::READY_FOR_RADIUS) {
		mouse2dSync();
		g_addSphere.updateRadius();
	}
	else if (g_addLathe->mousePhase == AddLathe::SET_SPINDLE_ORIENTATION) {
		mouse2dSync();
		if (g_addLathe->checkOrientationPoint(g_perspective)) {
		}
	}
	else if (g_addLathe->mousePhase == AddLathe::INITIAL_POINTS) {
		mouse2dSync();
		g_addLathe->checkOverInitPoint(g_perspective);
	}
	else if (g_addNurbsLine->mousePhase == AddNurbsLine::INITIAL_POINTS && g_addNurbsLine->line->point.length == 1) {
		mouse2dSync();
		g_addNurbsLine->checkOverInitPoint(g_perspective);
	}
	else if (g_insertPatch.imousePhase == InsertPatch::NEUTRAL && g_deleteControlLine.mousePhase == DeleteControlLine::NEUTRAL && g_mergePatches.mousePhase == MergePatches::NEUTRAL) {
		if (g_scene->mode & (MODELLING_MODE | ANIMATION_MODE | XRAY_EDIT_MODE | OWNERSHIP_MODE | NURBS_OWNERSHIP_MODE | SELECT_OBJECT_MODE) && tugFlag == false && g_modellingCursor == g_currentType2) {
			if (g_scene->mode & SELECT_OBJECT_MODE) {
				mouse2dSync();
				hoverObject();
				return;
			}
			else if (g_scene->mode & (XRAY_EDIT_MODE | OWNERSHIP_MODE | NURBS_OWNERSHIP_MODE)) {
				if (Art* selPart = dynamic_cast<Art*>(g_selNode.getRaw())) {
					mouse2dSync();
					selPart->smls->initSkeletonTug(g_perspective);
					return;
				}
			}
			else if (MultiSelect::pointList.length) {
				mouse2dSync();
				if (g_transformCP->isOverLever()) {
					//dLog("isOverLever()");
					return;
				}
			}
			if (Build* selBuild = dynamic_cast<Build*>(g_selNode.getRaw())) {
				if (Seamless* smls = getSeamless2()) {
					mouse2dSync();
					smls->initMultiCageTug(g_perspective);
				}
			}
			else if (g_scene->mode & ANIMATION_MODE) {
				if (Art* part = dynamic_cast<Art*>(g_selNode.getRaw())) {
					Seamless* smls = part->smls;
					mouse2dSync();
					g_partCageClue->mode = 0;

					if (g_vcArtAnimBase->initLeverTug(part, g_perspective)) {
						g_partCageClue->mode = PartCageClue::BONE_ROTATION_LEVER;
						return;
					}
					else if (part->niki && part->niki->isOverPoint()) {
						return;
					}
					else if (smls->initSkeletonTug(g_perspective)) {
						return;
					}
					else {
						if (smls->animCP) {
							smls->initMultiCageTug(g_perspective);
							if (dynamic_cast<NurbsPatch*>(MultiSelect::withinSightsNode.p)) {
								g_partCageClue->mode = PartCageClue::MORPH_CP;
								return;
							}
						}
					}
				}
				else if (Viewpoint* selView = dynamic_cast<Viewpoint*>(g_selNode.getRaw())) {
					mouse2dSync();
					if (selView->isOverPoint()) {
						MultiSelect::withinSights = true;
						MultiSelect::withinSightsNode = selView;
						changeModellingCursor(g_lockedOnPointCursor);
						return;
					}
				}
				else if (NurbsTrack* nurbsTrack = dynamic_cast<NurbsTrack*>(g_selNode.getRaw())) {
					mouse2dSync();
					if (nurbsTrack->isOverPoint()) {
						MultiSelect::withinSights = true;
						MultiSelect::withinSightsNode = nurbsTrack;
						changeModellingCursor(g_lockedOnPointCursor);
						return;
					}
				}
				else if (NikInterpolator* niki = dynamic_cast<NikInterpolator*>(g_selNode.getRaw())) {
					mouse2dSync();
					if (niki->isOverPoint()) {
						MultiSelect::withinSights = true;
						MultiSelect::withinSightsNode = niki;
						changeModellingCursor(g_lockedOnPointCursor);
						return;
					}
				}
				if (g_showCamerasInAnimMode) {
					hoverCameraAnimMode();
				}
				else {
					changeModellingCursor(g_crossCursor);
				}
			}
		}
	}
	//dLog("false");
}


BOOL mouse2dHandler(UINT msg, WPARAM wParam, LPARAM lParam) {
	static POINT ptDownMousePos2d;
	switch (msg)
	{
	case WM_RBUTTONDOWN: {
		Button::repeatCounter = 0;
		removeTypeFromList<Button>(g_callBackList);
		Window::removeToolTip();
		Window::removeMenu();

		g_mousePos = Vec2i(LOWORD(lParam), HIWORD(lParam));
		g_prevMousePos = g_mousePos;
		g_newMousePos = g_mousePos;


		Window::checkIfAWindowHasBeenRightClickedAndUpdate(g_mousePos);
		return TRUE;
	}
	case WM_MBUTTONDOWN: {
		Button::repeatCounter = 0;
		removeTypeFromList<Button>(g_callBackList);
		Window::removeToolTip();
		Window::removeMenu();

		g_mousePos = Vec2i(LOWORD(lParam), HIWORD(lParam));
		g_prevMousePos = g_mousePos;
		g_beginMousePos = g_mousePos;
		g_newMousePos = g_mousePos;
		g_beginViewPos = g_scene->translation;

		Window::checkIfAWindowHasBeenMidClickedAndUpdate(g_mousePos);
		return TRUE;
	}
	case WM_LBUTTONDOWN: {
		Button::repeatCounter = 0;
		g_previousClickTime = g_currentClickTime;
		g_currentClickTime = g_time;

		Window::removeToolTip();

		g_mousePos = Vec2i(LOWORD(lParam), HIWORD(lParam));
		g_prevMousePos = g_mousePos;
		g_beginMousePos = g_mousePos;
		g_newMousePos = g_mousePos;
		g_beginViewPos = g_scene->translation;

		//g_leftDoubleClicked = isDoubleLeftClick();
	//	dLog(String("g_leftDoubleClicked = " + String(g_leftDoubleClicked)));

		if (Button* button = Button::getHoverButton()) {

			DownButton::setDownButton(button);

			button->win.onClickButton();
			Window::activeWindow = &button->win;
			button->action();
			//button->win.paint();//do we need this?
			if (g_dontCloseMenu == false) {
				Window::removeMenu();
			}
		}
		else {
			//if(g_insertPatch.imousePhase == InsertPatch::LINE_MODE){
			//	g_insertPatch.imousePhase = InsertPatch::DIVERT_STANDBY;
			//	g_stw->cp->setupNodeWindow(g_insertComplete);
			//}
			//else if(g_deleteControlLine.mousePhase == 2){
			//	g_deleteControlLine.doLastPhase();
			//}
			//else{
			Window::removeMenu();
			Window::checkIfAWindowHasBeenClickedAndUpdate(g_mousePos);
			//}
		}


		g_dontCloseMenu = false;

		if (g_drag) {
			g_trackMouseEvent.hwndTrack = hWndMain;
			TrackMouseEvent(&g_trackMouseEvent);
			SetCapture(hWndMain);
		}

		g_previousLeftClickPos = g_mousePos;

		return TRUE;
	}
						 //	case WM_LBUTTONDBLCLK:{ //why does this not catch double click?
						 //		break;
						 //	}
	case WM_TOUCH: {
		break;
	}
	case WM_GESTURENOTIFY: {
		GESTURECONFIG gc = { 0,GC_ALLGESTURES,0 };
		BOOL bResult = SetGestureConfig(hWndMain, 0, 1, &gc, sizeof(GESTURECONFIG));

		if (!bResult)
		{
			// an error
		}
		return DefWindowProc(hWndMain, WM_GESTURENOTIFY, wParam, lParam);
	}
	case WM_GESTURE: {
		int a = 0;
		break;
	}
	case WM_MBUTTONUP: {
		if (g_winAPI3DMode) {
			release3dDrag();
		}
		break;
	}
	case WM_RBUTTONUP: {

		if (g_winAPI3DMode) {
			release3dRightDrag();
		}
		break;
	}
	case WM_LBUTTONUP: {
		g_mousePos = Vec2i(LOWORD(lParam), HIWORD(lParam));

		if (g_drag) {
			g_drag->endDrag(g_mousePos);
		}
		else if (Button* button = DownButton::getDownButton()) {
			button->onButtonUp();
		}
		else {
			Window::activeWindow->release(g_mousePos);
		}
#ifdef PROG_ROB
		InputEvent::sceneTreeScrolPos = 0;
		if (g_upStr != "") {
			dLog(g_upStr);
		}
		g_upStr = "";

		InputEvent::sceneTreeScrolPos = g_stw->getScrollPosCollapsed();
#endif

		if (g_winAPI3DMode) {
			release3dDrag();
		}



		return TRUE;
	}

					   //case WM_NCMOUSEMOVE:
	case WM_MOUSEMOVE: {
		//if(g_useWinTab && g_penEnabled){
		//	return TRUE;
		//}
		if (g_mergePatches.mousePhase != 0) {
			int a = 0;
		}
		short lw = LOWORD(lParam);
		short hw = HIWORD(lParam);
		//if(lw < 0 || lw >= g_3d->size.x || hw < 0 || hw >= g_3d->size.y){
		//	return TRUE;
		//}
		//if(lw < -100){
		//	int a = 0;
		//}
		int remOldMousePos = g_mousePos.y;
		if (g_mousePos.x == lw && g_mousePos.y == hw) {
			return TRUE;
		}

		if (g_scene->mode & PAINT_MODE) {
			if (g_scene->mode & BRUSH_MODE) {
				g_brush->houseKeep();
			}
		}

		g_mousePos = Vec2i(lw, hw);
		g_newMousePos = g_mousePos;

		if (g_winAPI3DMode) {
			return TRUE;
		}

		//if(g_winAPI3DMode){
			//dLog("WM_MOUSEMOVE");
			//moveMouse3d();
			//break;
		//}

		Button* hoverButton = Button::getHoverButton();
		MultiSelect::resetDistance();
		//if(g_insertPatch.imousePhase == LINE_MODE){
		//	mouse2dSync(hWndMain);
		//	//g_insertPatch.mouseOverPatch->getRatioPhaseForInsert();
		//	g_insertPatch.addInsertVisualClue();
		//}
		if (g_insertPatch.imousePhase == InsertPatch::MONITOR_DIRECTION) {
			mouse2dSync();
			if (g_insertPatch.monitorDirection()) {
				g_insertPatch.m_insert_tempDivertBackup = g_insertPatch.m_insert;
				g_insertPatch.m_insertOD_tempDivertBackup = g_insertPatch.m_insertOD;
				g_insertPatch.remLeftSide_tempDivertBackup = g_insertPatch.remLeftSide;

				g_insertPatch.beginInsertIdx_tempDivertBackup = g_insertPatch.beginInsertIdx;
				g_insertPatch.beginInsertIdxOD_tempDivertBackup = g_insertPatch.beginInsertIdxOD;
				g_insertPatch.imousePhase = InsertPatch::DIVERT_STANDBY;
			}
			return FALSE;
		}
		if (g_insertPatch.imousePhase == InsertPatch::STANDBY || g_insertPatch.imousePhase == InsertPatch::LINE_MODE || g_insertPatch.imousePhase == InsertPatch::DIVERT_STANDBY || g_addBone->mousePhase != AddBone::NEUTRAL) {
			if (hoverButton) {
				hoverButton->onHoverMove(g_mousePos - hoverButton->win.pos);
			}
			else {
				mouse2dSync();
				//SenseTugGlobals::initTug();
				if (NurbsPatch* patch = dynamic_cast<NurbsPatch*>(g_selNode.getRaw())) {
					Seamless* smls = patch->smls;
					if (g_insertPatch.imousePhase == InsertPatch::DIVERT_STANDBY) {
						if (smls->addDiversionMark()) {
							//g_insertPatch.imousePhase = InsertPatch::TEMP_DIVERT;
						}
					}
					else {
						smls->insertControlRowColum();
					}
				}
				else if (g_addBone->mousePhase != AddBone::NEUTRAL) {
					if (Art* part = dynamic_cast<Art*>(g_selNode.getRaw())) {
						mouse2dSync();
						g_addBone->initMouseOverLine(part->smls);
					}
				}
			}
		}
		if (g_mergePatches.mousePhase == MergePatches::STANDBY) {
			if (hoverButton) {
				hoverButton->onHoverMove(g_mousePos - hoverButton->win.pos);
			}
			else {
				mouse2dSync();
				g_mergePatches.smls->mergePatchesStandby();
			}
		}
		if (g_deleteControlLine.mousePhase == 1 || g_deleteControlLine.mousePhase == 2) {
			if (hoverButton) {
				hoverButton->onHoverMove(g_mousePos - hoverButton->win.pos);
			}
			else {
				mouse2dSync();
				if (g_scene->mode & ANIMATION_MODE) {
					if (Art* part = dynamic_cast<Art*>(g_selNode.getRaw())) {
						part->smls->deleteControlLine();
					}
				}
				else if (NurbsPatch* patch = dynamic_cast<NurbsPatch*>(g_selNode.getRaw())) {
					patch->smls->deleteControlLine();
				}
			}
		}
		else if (g_drag) {
			if (g_mousePos.x >= 0) {
				if (g_drag->triggered == false) {
					if (abs(g_mousePos.x - g_drag->beginMousePos.x) > 6 || abs(g_mousePos.y - g_drag->beginMousePos.y) > 6) {
						g_drag->triggered = true;
					}
				}
				if (g_drag->triggered) {
					g_drag->doDrag(g_mousePos);
				}
			}
		}
		else if (hoverButton && hoverButton->onHoverMove(g_mousePos - hoverButton->win.pos)) {
		}
		else {
			hoverWin3d();
		}
		break;
	}
					   /*
				   case WM_MOUSEMOVE:
					   if(MK_LBUTTON&wParam){


						   GetCursorPos( &ptCursor );
						   g_scene->translation=g_scene->rotation.multVec(g_scene->translation);
						   g_scene->translation.x+=.0025*(float)(ptCursor.x-ptDownMousePos2d.x);
						   g_scene->translation=-g_scene->rotation*g_scene->translation;

						   if(controlKeyState()){
							   g_scene->translation=g_scene->rotation.multVec(g_scene->translation);
							   g_scene->translation.z-=.0025*(float)(ptCursor.y-ptDownMousePos2d.y);
							   g_scene->translation=-g_scene->rotation*g_scene->translation;
						   }
						   else{
							   g_scene->translation=g_scene->rotation*g_scene->translation;
							   g_scene->translation.y-=.0025*(float)(ptCursor.y-ptDownMousePos2d.y);
							   g_scene->translation=-g_scene->rotation*g_scene->translation;
						   }

						   GetCursorPos( &ptDownMousePos2d );
						   return TRUE;
					   }
					   if(MK_RBUTTON&wParam){
						   GetCursorPos( &ptCursor );
						   rotateScene(ptCursor.x-ptDownMousePos2d.x, ptCursor.y-ptDownMousePos2d.y);
						   ptDownMousePos2d=ptCursor;
						   return TRUE;
					   }
					   */
	}
	return FALSE;
}

void updateNavKey();

bool keyDown(HWND &hWnd, UINT &message, WPARAM &wParam, LPARAM &lParam) {
	if (Window::activeWindow == g_3d) {
		static String str;
		float tran = (1 / g_scene->scale.x)*.01;
		switch (wParam) {
		case VK_LEFT: {
			if (g_leftKeyDown == false) {
				g_leftKeyTime = g_time;
				g_leftKeyDown = true;
				updateNavKey();
			}
			else {
				g_leftKeyDown = true;
			}
			break;
		}
		case VK_RIGHT: {
			if (g_rightKeyDown == false) {
				g_rightKeyTime = g_time;
				g_rightKeyDown = true;
				updateNavKey();
			}
			else {
				g_rightKeyDown = true;
			}
			break;
		}
		case VK_DOWN: {
			if (g_backwardKeyDown == false) {
				g_backwardKeyTime = g_time;
				g_backwardKeyDown = true;
				updateNavKey();
			}
			else {
				g_backwardKeyDown = true;
			}
			break;
		}
		case VK_UP: {
			if (g_forwardKeyDown == false) {
				g_forwardKeyTime = g_time;
				g_forwardKeyDown = true;
				updateNavKey();
			}
			else {
				g_forwardKeyDown = true;
			}
			break;
		}
		default: {
			return false;
		}
		}
		return true;
	}
	return false;
}

bool getOldColor(Color3f& oldColor) {
	Art* part = dynamic_cast<Art*>(g_selNode.getRaw());
	if (part == 0)return false;
	Seamless* smls = part->smls;
	Art* triPart = smls->actTriPart;
	if (triPart == 0)return false;
	ColorEffect* colorEffect = dynamic_cast<ColorEffect*>(part->workEffect);
	if (colorEffect == 0)return false;
	if (triPart->triangle.length <= smls->ActTri) {
		return false;
	}
	M_Triangle& triangle = triPart->triangle;
	const M_Color3f& color = colorEffect->color;
	oldColor = color[triangle.refTexIndex(smls->ActTri * 3 + smls->ActTriCorner)];
	return true;
}
//-----------------------------------------------------------------------------
// Name: MsgProc()
// Desc: The window's message handler
//-----------------------------------------------------------------------------

void keyboardRadialTug(Vec3f coord, Vec3f pt) {
	Art* part = dynamic_cast<Art*>(g_selNode.getRaw());
	Seamless* smls = part->smls;
	M_Art& partList = smls->partList;

	g_radialTug->displacement.refValue()[0] = pt;

	g_radialTug->slope.refValue() = 1;
	if (g_settings->multiPartTug.refValue()) {
		for (int k = partList.length; --k >= 0;) {
			Art* prt = partList[k];
			g_radialTugStem->part.refValue()[0] = prt;
			g_radialTug->location.refValue()[0] = part->matrix * (prt->imatrix * coord);
			g_radialTugStem->doBuild();
		}
	}
	else {
		g_radialTugStem->part.refValue()[0] = part;
		g_radialTug->location.refValue()[0] = coord;
		g_radialTugStem->doBuild();
	}
}

void openChatLog() {
	String url = g_appDataDir;
	url += "chat.txt";

	PROCESS_INFORMATION pi;
	STARTUPINFO si;
	ZeroMemory(&si, sizeof(si));
	ZeroMemory(&pi, sizeof(pi));
	si.cb = sizeof(si);
	// Start the child process.

	String commandLine = '\"';
	getFolderPath(commandLine, CSIDL_WINDOWS);
	if (commandLine[commandLine.length - 1] != '\\') {
		commandLine += '\\';
	}
	commandLine += "notepad.EXE\" ";
	commandLine += url;

	if (CreateProcess(
		NULL, // No module name (use command line).
		const_cast<char*>(commandLine.c_str()), // Command line. 
		NULL,             // Process handle not inheritable. 
		NULL,             // Thread handle not inheritable. 
		FALSE,            // Set handle inheritance to FALSE. 
		0,                // No creation flags. 
		NULL,             // Use parent's environment block. 
		NULL,             // Use parent's starting directory. 
		&si,              // Pointer to STARTUPINFO structure.
		&pi)) {             // Pointer to PROCESS_INFORMATION structure.    ) 

		//WaitForSingleObject( pi.hProcess, INFINITE );
		// Close process and thread handles.
		CloseHandle(pi.hProcess);
		CloseHandle(pi.hThread);
	}
}


void openWithHtmlBrowser(String url, const String htmlBrowser) {
	PROCESS_INFORMATION pi;
	STARTUPINFO si;
	ZeroMemory(&si, sizeof(si));
	ZeroMemory(&pi, sizeof(pi));
	si.cb = sizeof(si);
	// Start the child process.

	String commandLine = '\"';
	if (htmlBrowser == "") {
		getFolderPath(commandLine, CSIDL_PROGRAM_FILES);
		if (commandLine[commandLine.length - 1] != '\\') {
			commandLine += '\\';
		}
		commandLine += "Internet Explorer\\IEXPLORE.EXE\"";
	}
	else {
		commandLine = htmlBrowser;
	}
	commandLine += ' ';
	commandLine += url;

	if (CreateProcess(
		NULL, // No module name (use command line).
		const_cast<char*>(commandLine.c_str()), // Command line. 
		NULL,             // Process handle not inheritable. 
		NULL,             // Thread handle not inheritable. 
		FALSE,            // Set handle inheritance to FALSE. 
		0,                // No creation flags. 
		NULL,             // Use parent's environment block. 
		NULL,             // Use parent's starting directory. 
		&si,              // Pointer to STARTUPINFO structure.
		&pi)) {             // Pointer to PROCESS_INFORMATION structure.    ) 

		//WaitForSingleObject( pi.hProcess, INFINITE );
		// Close process and thread handles.
		CloseHandle(pi.hProcess);
		CloseHandle(pi.hThread);
	}
}
void getFFmpegPath(String& path) {
	char buffer[MAX_PATH];
	GetModuleFileName(NULL, buffer, MAX_PATH);
	path = buffer;

	String drive;
	String dir;
	String fname;
	String ext;
	splitPath(path, drive, dir, fname, ext);
	path = '\"';
	path += drive;
	path += dir;
	path += "ffmpeg\\ffmpeg.exe\"";
}
void openWithFFmpeg(String url) {
	PROCESS_INFORMATION pi;
	STARTUPINFO si;
	ZeroMemory(&si, sizeof(si));
	ZeroMemory(&pi, sizeof(pi));
	si.cb = sizeof(si);
	// Start the child process.
	String commandLine;
	getFFmpegPath(commandLine);


	//	String commandLine='\"';
	//	getFolderPath(commandLine, CSIDL_PROGRAM_FILES);
	//	if(commandLine[commandLine.length-1]!='\\'){
	//		commandLine+='\\';
	//	}
	//	commandLine+="Seamless3d\\ffmpeg.exe\"";
	commandLine += url;
	//commandLine+='\"';

	if (CreateProcess(
		NULL, // No module name (use command line).
		const_cast<char*>(commandLine.c_str()), // Command line. 
		NULL,             // Process handle not inheritable. 
		NULL,             // Thread handle not inheritable. 
		FALSE,            // Set handle inheritance to FALSE. 
		0,                // No creation flags. 
		NULL,             // Use parent's environment block. 
		NULL,             // Use parent's starting directory. 
		&si,              // Pointer to STARTUPINFO structure.
		&pi)) {             // Pointer to PROCESS_INFORMATION structure.    ) 

		//WaitForSingleObject( pi.hProcess, INFINITE );
		// Close process and thread handles.
		CloseHandle(pi.hProcess);
		CloseHandle(pi.hThread);
	}
}

void openWithWindowsMediaPlayer(String url) {
	PROCESS_INFORMATION pi;
	STARTUPINFO si;
	ZeroMemory(&si, sizeof(si));
	ZeroMemory(&pi, sizeof(pi));
	si.cb = sizeof(si);
	// Start the child process.

	String commandLine = '\"';
	getFolderPath(commandLine, CSIDL_PROGRAM_FILES);
	if (commandLine[commandLine.length - 1] != '\\') {
		commandLine += '\\';
	}
	commandLine += "Windows Media Player\\wmplayer.exe\" \"";
	commandLine += url;
	commandLine += '\"';

	if (CreateProcess(
		NULL, // No module name (use command line).
		const_cast<char*>(commandLine.c_str()), // Command line. 
		NULL,             // Process handle not inheritable. 
		NULL,             // Thread handle not inheritable. 
		FALSE,            // Set handle inheritance to FALSE. 
		0,                // No creation flags. 
		NULL,             // Use parent's environment block. 
		NULL,             // Use parent's starting directory. 
		&si,              // Pointer to STARTUPINFO structure.
		&pi)) {             // Pointer to PROCESS_INFORMATION structure.    ) 

		//WaitForSingleObject( pi.hProcess, INFINITE );
		// Close process and thread handles.
		CloseHandle(pi.hProcess);
		CloseHandle(pi.hThread);
	}
}
void openWithPovRay(String url) {
	PROCESS_INFORMATION pi;
	STARTUPINFO si;
	ZeroMemory(&si, sizeof(si));
	ZeroMemory(&pi, sizeof(pi));
	si.cb = sizeof(si);
	// Start the child process.

	String commandLine = '\"';
	getFolderPath(commandLine, CSIDL_PROGRAM_FILES);
	if (commandLine[commandLine.length - 1] != '\\') {
		commandLine += '\\';
	}
	commandLine += "POV-Ray for Windows v3.62\\bin\\pvengine.exe\" /EDIT \"";
	commandLine += url;
	commandLine += '\"';

	if (CreateProcess(
		NULL, // No module name (use command line).
		const_cast<char*>(commandLine.c_str()), // Command line. 
		NULL,             // Process handle not inheritable. 
		NULL,             // Thread handle not inheritable. 
		FALSE,            // Set handle inheritance to FALSE. 
		0,                // No creation flags. 
		NULL,             // Use parent's environment block. 
		NULL,             // Use parent's starting directory. 
		&si,              // Pointer to STARTUPINFO structure.
		&pi)) {             // Pointer to PROCESS_INFORMATION structure.    ) 

		CloseHandle(pi.hProcess);
		CloseHandle(pi.hThread);
	}
}

void returnToExitSeamless3d() {
	g_settings->writeINIFile();
	BackUp::deleteBackupFiles();
	DestroyWindow(hWndMain);
}

void riProcessMouseMessage(const RAWMOUSE *rmouse)
{
	// Store Mouse Button States
	// Check Mouse Position Relative Motion
	if (MOUSE_MOVE_RELATIVE == rmouse->usFlags)
	{
		int mxr = rmouse->lLastX;
		int myr = rmouse->lLastY;
		g_mousePos.x += mxr;
		g_mousePos.y += myr;
		//dLog(g_mousePos.x);
	}
}

#ifdef WACOM_FILES
extern PACKET* g_wintabPktBuf;
#endif
extern int g_lastPacketIndex;
int g_packetBufLen = 0x200;

LRESULT CALLBACK MsgProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{

	g_before = g_selNode.getRaw();
	Art* part = dynamic_cast<Art*>(g_selNode.getRaw());
	Seamless* smls = 0;
	float tran = (1 / g_scene->getZoom())*.01F;


	Vec3f vec;
	Rotation tmpRot;


	if (TRUE == mouse2dHandler(message, wParam, lParam)) {
		return TRUE;
	}
	//	if(g_activeWindow->input(message, wParam, lParam)){
	//		return 0;
	//	}

#ifdef	OPENGL
	if (mouse3dMode) {
		mouse3dHandler(message, wParam);
	}
#endif

	switch (message) {

	case WM_CREATE: {
#ifdef WACOM_FILES
		WINDOWINFO appWindowInfo = { 0 };
		appWindowInfo.cbSize = sizeof(appWindowInfo);
		GetWindowInfo(hWnd, &appWindowInfo);
		g_clientRect = appWindowInfo.rcClient;
		//should not need above code if we use seamless native screen rectangle
#endif

		DragAcceptFiles(hWnd, TRUE);
		//InitDirectInput(hWnd);
		/*
	 // register raw input devices
		RAWINPUTDEVICE rID[2];

		// Keyboard
		rID[0].usUsagePage = 1;
		rID[0].usUsage = 6;
		rID[0].dwFlags = 0;
		rID[0].hwndTarget=NULL;

		// Mouse
		rID[1].usUsagePage = 1;
		rID[1].usUsage = 2;
		rID[1].dwFlags = 0;
		rID[1].hwndTarget = hWnd;
		if ( !RegisterRawInputDevices( rID, 2, sizeof( RAWINPUTDEVICE ) ) )
		{
			// handle failure
			throw;
		}
		*/
#ifdef WACOM_FILES
		// test for touch
		if (g_settings->wacomPackets) {
			g_tabCtx = InitWintabAPI(hWnd);
			if (g_tabCtx) {
				g_useWinTab = true;
				do {
					g_packetBufLen >>= 1;
					gpWTQueueSizeSet(g_tabCtx, g_packetBufLen);
				} while (gpWTQueueSizeGet(g_tabCtx) == 0);
				dLog("packetBufLen = gpWTQueueSizeSet(g_tabCtx, 0x100);");
				dLog(g_packetBufLen);
				//if(packetBufLen < 0x100){
				//	throw;
				//}
				disablePen();
			}
		}
#endif
		g_bReady = true;


		//InputEvent::setupSequence(hWnd);
		break;
	}
#ifdef WACOM_FILES
	case WM_FINGERDATA:
		interpretFingers(((WacomMTFingerCollection*)lParam)->FingerCount, ((WacomMTFingerCollection*)lParam)->Fingers, ((WacomMTFingerCollection*)lParam)->DeviceID);
		break;

	case WT_PACKET: {


		//int len = gpWTPacketsPeek((HCTX)lParam, 0x100, g_wintabPkt);
		int len = gpWTPacketsGet((HCTX)lParam, g_packetBufLen, g_wintabPktBuf);
		if (len > 50) {
			dLog("if(len > 50){");
			dLog(len);
		}
		g_lastPacketIndex = len - 1;

		//gpWTPacketsGet(g_tabCtx, 0x200, NULL);
		//gpWTPacket((HCTX)lParam, wParam, NULL);

		break;
	}
#endif
					/*
	 case WM_INPUT:
	 {
	RAWINPUT *pRI = NULL;

	 // Determine how big the buffer should be
	 UINT iBuffer;
	 GetRawInputData( ( HRAWINPUT )lParam, RID_INPUT, NULL, &iBuffer, sizeof( RAWINPUTHEADER ) );

	 // Allocate a buffer with enough size to hold the raw input data
	 LPBYTE lpb = new BYTE[ iBuffer ];
	 if( lpb == NULL )
		 return 0;

	 // Get the raw input data
	 UINT readSize = GetRawInputData( ( HRAWINPUT )lParam, RID_INPUT, lpb, &iBuffer, sizeof( RAWINPUTHEADER ) ) ;

	 // Validate that read size is as expected
	 if( readSize != iBuffer )
		 puts( "ERROR: GetRawInputData didn't return correct size!" ) ;
	 pRI = ( RAWINPUT* )lpb;

	 // Process the Mouse Messages
	 if(g_winAPI3DMode){
		 if( pRI->header.dwType== RIM_TYPEMOUSE ){
			 riProcessMouseMessage(&pRI->data.mouse);
		 }
	 }


	 // Destroy the Raw Input Data and Return
	 delete lpb;
	 return 1;
		 break;
	 }
	 */
	case WM_NCMOUSELEAVE: {
		//case WM_MOUSELEAVE:{
		if (g_drag) {
			g_drag->endDrag(g_mousePos);
		}
		break;
	}
	case WM_MOUSEWHEEL: {
		Window::activeOnMouseWheel((short)HIWORD(wParam) / WHEEL_DELTA);
		break;
	}
	case WM_SIZE:
		if (wParam == SIZE_MINIMIZED) {
			g_minimized = true;
			break;
		}
		else if (wParam == SIZE_MAXIMIZED || wParam == SIZE_RESTORED) {
			g_minimized = false;
		}
	case WM_MOVE: {
		RECT rect;
		GetWindowRect(hWndMain, &rect);
		g_settings->win3dPos.refValue().x = rect.left;
		g_settings->win3dPos.refValue().y = rect.top;
		g_settings->win3dSize.refValue().x = rect.right - rect.left - g_extraWinPix.x;
		g_settings->win3dSize.refValue().y = rect.bottom - rect.top - g_extraWinPix.y;
		g_windowSizeWindowBool = true;
		g_windowSizeWindowTime = g_time;
	}
				  break;
	case WM_PAINT:
		PAINTSTRUCT ps;

		BeginPaint(hWnd, &ps);
		renderScene();
		//outputScreenShotForDemo();
		EndPaint(hWnd, &ps);
		break;

	case WM_KEYUP:
		switch (wParam) {
		case 0x5a:	//"Z"
			g_zDragMode = false;
		case VK_LEFT: {
			g_leftKeyDown = false;
			break;
		}
		case VK_RIGHT: {
			g_rightKeyDown = false;
			break;
		}
		case VK_DOWN: {
			g_backwardKeyDown = false;
			break;
		}
		case VK_UP: {
			g_forwardKeyDown = false;
			break;
		}
#ifdef PROG_ROB
		case VK_CONTROL: {
			g_controlDown = false;
			dLog("new VirtKeyUp(VK_CONTROL),");
			break;
		}
		case VK_SHIFT: {
			g_shiftDown = false;
			dLog("new VirtKeyUp(VK_SHIFT),");
			break;
		}
		case VK_TAB: {
			g_tabDown = false;
			dLog("new VirtKeyUp(VK_TAB),");
			break;
		}
#endif

		default:
			return DefWindowProc(hWnd, message, wParam, lParam);
		}
		break;
	case WM_SYSKEYDOWN: {
		//		if(altKeyState()){
		if (keyDownAlt(wParam)) {
			break;
		}
		//		}
		return DefWindowProc(hWnd, message, wParam, lParam);
	}
	case WM_SYSKEYUP: {
#ifdef PROG_ROB
		g_altDown = false;
		dLog("new VirtKeyUp(VK_MENU),");
#endif
		return DefWindowProc(hWnd, message, wParam, lParam);
	}
	case WM_KEYDOWN:
#ifdef PROG_ROB
		if (wParam == VK_TAB) {
			if (g_tabDown == false) {
				g_tabDown = true;
				dLog("new VirtKeyDown(VK_TAB),");
			}
		}
#endif

		if (0x8000 & GetKeyState(VK_LCONTROL)) {
#ifdef PROG_ROB
			if (g_controlDown == false) {
				g_controlDown = true;

				dLog("new VirtKeyDown(VK_CONTROL),");
			}
#endif
			if (keyDownCtrl(wParam)) {
				break;
			}
		}
		else if (shiftKeyState()) {
#ifdef PROG_ROB
			if (g_shiftDown == false) {
				g_shiftDown = true;
				dLog("new VirtKeyDown(VK_SHIFT),");
			}
#endif
			if (keyDownShift(wParam)) {
				break;
			}
		}
		else {
			if (keyDown(hWnd, message, wParam, lParam)) {
				break;
			}
			else if (keyDown(wParam)) {
				break;
			}
		}
		return DefWindowProc(hWnd, message, wParam, lParam);

	case WM_CHAR: {
		for (int i = g_callBackList.length; --i >= 0;) {
			if (CaretWindow* textCell = dynamic_cast<CaretWindow*>(g_callBackList[i])) {
				textCell->charInput(wParam);
				return 0;
			}
		}
		return DefWindowProc(hWnd, message, wParam, lParam);
	}
	case WM_GETMINMAXINFO:
		// Prevent the window from going smaller than some minimum size
		((MINMAXINFO*)lParam)->ptMinTrackSize.x = 100;
		((MINMAXINFO*)lParam)->ptMinTrackSize.y = 100;
		break;
	case WM_MOUSEACTIVATE:   // sent when window changes active state
		return DefWindowProc(hWnd, MA_NOACTIVATE, wParam, lParam);
#ifndef FRAME_TIME_SYNC
	case WM_ACTIVATE:   // sent when window changes active state
		if (WA_INACTIVE == wParam) {
			if (g_drag) {
				g_drag->endDrag(g_mousePos);
			}
			Window::onSeamlessWindowsLoseFocus();
			g_bActive = false;
			return 0;
		}
		else {
			Window::activeWindow->onGainFocus();
			g_bActive = true;
		}
		// Set exclusive mode access to the mouse based on active state
		break;
	case WM_ACTIVATEAPP:
		InputEvent::resumeSequence();
		break;
#endif
	case WM_DROPFILES: {
		HDROP hDrop = (HDROP)wParam;
		char szGile[0x200];
		DragQueryFile(hDrop, 0, szGile, sizeof(szGile));
		DragFinish(hDrop);
		droppedFilename = szGile;
		openDroppedFile();
		break;
	}
	case WM_COMMAND:
		switch LOWORD(wParam) {

			/*
					case MOUSE_3D:
						showCursor(false);
						break;
					case MOUSE_2D:
						mouse3dMode=false;
						UnAcquire();
						showCursor(true);
						break;
			*/
			//		case STRING_FROM_CHAT_WIN_COMMAND:{
			//			doStringFromChat(lParam);
			//			break;
			//		}
					//case CAPSLOCK_ACCEL:
					//	return 0;
					//case HIDE_SELECTED_CAGES:
					//	return 0;
					//case ROTATE_VIEW_ACCEL:
					//	return 0;
		case RESET_3D_CURSOR:
			//reset3dCursor();
			break;
		default:
			return DefWindowProc(hWnd, message, wParam, lParam);
		}
		break;
	case WM_CLOSE:
		if (Window::activeWindow->hog) {
			PlaySound("SystemDefault", 0, SND_ASYNC);
			break;
		}
		saveSmlsFirst(returnToExitSeamless3d);
		break;
	case WM_DESTROY:
		destroyRender();

		FreeDirectInput();
		PostQuitMessage(0);

		ReleaseDC(hWnd, GetDC(hWnd));

		break;

	default:
		return DefWindowProc(hWnd, message, wParam, lParam);
	}
	return 0;
}




void initializeShortCuts();
void getMemberAddresses();
void convert_raw_bitmap_to_cpp();
void convert_raw_alpha_bitmap_to_cpp();

void genButtons();

#include <crtdbg.h>
//static _CrtMemState s1, s2, s3;

//void processWords(int);
//void printOutLines();

const char* g_lpszCmdLine = 0;
bool openFileFromCommandLine() {
	g_processComnandLine = false;
	if (strcmp(g_lpszCmdLine, "") != 0) {
		String file = g_lpszCmdLine;

		if (file[0] == '\"') {
			char longName[0x200];
			file = file.substring(1, file.length - 2);

			GetLongPathName(file.c_str(), longName, sizeof(longName));
			openSmlsLastPhase(longName);
		}
		else {
			openSmlsLastPhase(g_lpszCmdLine);
		}

		return true;
	}
	return false;
}

void saveMemToFiles(unsigned int width, unsigned int height);

//char nibleToHex(char par);


// Test.cpp
#include <iostream>

class Test
{
public:
	template <typename T>
	void function() {
		std::cout << data << std::endl;
	}

	template <> void function<char>();
};



template <>
void Test::function<char>() {
	std::cout << 'Z';
}


extern "C" void CopyLocalToScratchPad_doOp_S_Float();


void ExportMovie::OpenOutputFolder::operator=(const bool& par) {
	openWithHtmlBrowser(g_seamlessDir, "");
}

void ExportVrmlX3d::OpenOutputFolder::operator=(const bool& par) {
	openWithHtmlBrowser(g_seamlessDir, "");
}
/*
inline void convertConcaveIFSIndexToTriIndex(const M_Vec3f& src_point, M_Int& index){
	M_Vec3f point;
	M_Int& tmpIndex=g_tmpIndex;
	tmpIndex.length=0;
	M_Int poly;
	for(int i = 0; i < index.length && index[i] >= 0; i++){
		poly+=index[i];
	}

	Rotation rot;

	for(int s = 0 ;;){
		int idx0 = poly[s++];
		if(s>=poly.length){
			s=0;
		}
		int idx1 = poly[s++];
		if(s>=poly.length){
			s=0;
		}
		int idx2 = poly[s++];
		if(s>=poly.length){
			s=0;
		}

		Vec3f v0 = src_point[idx0];
		Vec3f v1 = src_point[idx1];
		Vec3f v2 = src_point[idx2];

		if(thinTriangle(v0, v1, v2)){
			s -= 2;
			continue;
		}
		Vec3f normal = calcNormal(src_point[index[idx0]], src_point[index[idx1]], src_point[index[idx2]]);
		rot = Rotation(Vec3f(0, 0, 1), normal);
		break;
	}

	point.length = src_point.length;
	for(int i=0; i<point.length; i++){
		point[i] = rot * src_point[i];
	}

	int s=0;
	int c=0;
	while(poly.length>3){
		int idx0 = poly[s++];
		if(s>=poly.length){
			s=0;
			c++;
		}
		int idx1 = poly[s++];
		if(s>=poly.length){
			s=0;
			c++;
		}
		int idx2 = poly[s++];
		if(s>=poly.length){
			s=0;
			c++;
		}

		Vec3f v0 = point[idx0];
		Vec3f v1 = point[idx1];
		Vec3f v2 = point[idx2];

		if(!backFace(v0, v1, v2) || thinTriangle(v0, v1, v2) || vertexInsideTri(point, idx0, idx1, idx2)){
			if(c==3){
				s=0;
				c=0;
				rot=-rot;
				for(int i=0; i<point.length; i++){
					point[i] = rot * src_point[i];
				}
				poly.length=0;
				for(int i = 0; i < index.length && index[i] >= 0; i++){
					poly+=index[i];
				}
				tmpIndex.length=0;
				continue;
			}
			s -= 2;
			if(s < 0){
				s=poly.length + s;
			}
			continue;
		}
		tmpIndex+=idx0;
		tmpIndex+=idx1;
		tmpIndex+=idx2;
		tmpIndex+=-1;

		int g=s-2;
		if(g < 0){
			g=poly.length + g;
		}
		poly.deleteElement(g);
		s=0;
		c=0;
	}
	tmpIndex+=poly;
	index=tmpIndex;
}
*/

void getClientRect(Rect& rect) {
	GetClientRect(hWndMain, reinterpret_cast<RECT*>(&rect));
}

void getCursorPos(Point& point) {
	GetCursorPos(reinterpret_cast<POINT*>(&point));
}

void pivotPointToClipBoard() {
	if (Art* part = dynamic_cast<Art*>(g_selNode.getRaw())) {
		if (OpenClipboard(hWndMain)) {
			EmptyClipboard();

			// Allocate a global memory object for the text. 

			String str;
			part->smls->catIMatrix();
			(part->matrix * part->absolutePivotPoint).toString(str);

			HGLOBAL hglbCopy = GlobalAlloc(GMEM_MOVEABLE, str.length + 1);
			if (hglbCopy == NULL)
			{
				CloseClipboard();
			}
			else {

				// Lock the handle and copy the text to the buffer. 

				LPVOID lptstrCopy = GlobalLock(hglbCopy);
				memcpy(lptstrCopy, str.c_str(), str.length + 1);

				GlobalUnlock(hglbCopy);

				// Place the handle on the clipboard. 

				SetClipboardData(CF_TEXT, hglbCopy);
				CloseClipboard();
			}
		}
	}
}

void changeCursor(long requestType) {
	if (g_currentType2 != requestType) {
		static int g = 0;
		//if(g > 0){
		//	g = g;
		//}
		if (g_currentType2 == g_lockedOnPointCursor) {
			//if(requestType == g_addPointCursor){
			//if(g_currentType2 == g_addPointCursor && requestType != g_lockedOnPointCursor){
			int g = 0;
		}

		SetClassLong(hWndMain, GCL_HCURSOR, (LONG)requestType);
		SetCursor((HCURSOR)requestType);

		g_currentType2 = requestType;
		if (requestType != g_sizeAllCursor) {
			g_panViewMode = false;
		}
	}
}

float InputEvent::localMove(Vec2i endPos) {
	ClientToScreen(hWndMain, reinterpret_cast<POINT*>(&endPos));
	return move(endPos);
}

void CaretWindow::copyToClipboard(bool cut) {
	if (highlightPos < 0)return;
	if (caretPos != highlightPos) {
		if (OpenClipboard(hWndMain)) {
			EmptyClipboard();

			// Allocate a global memory object for the text. 

			int highlightLen = abs(caretPos - highlightPos);
			int index = highlightPos;
			if (caretPos < highlightPos) {
				index = caretPos;
			}

			HGLOBAL hglbCopy = GlobalAlloc(GMEM_MOVEABLE, abs(caretPos - highlightPos) + 1);
			if (hglbCopy == NULL)
			{
				CloseClipboard();
			}
			else {

				// Lock the handle and copy the text to the buffer. 

				LPVOID lptstrCopy = GlobalLock(hglbCopy);
				memcpy(lptstrCopy, &cellStr[index], highlightLen + 1);

				((char*)lptstrCopy)[highlightLen] = 0;    // null character 
				GlobalUnlock(hglbCopy);

				// Place the handle on the clipboard. 

				SetClipboardData(CF_TEXT, hglbCopy);
				CloseClipboard();

				if (cut) {
					String str = cellStr;
					cellStr = str.substr(0, index);
					cellStr += str.substr(index + highlightLen);
					caretPos = highlightPos = index;
					onChange();
					win.paint();
				}
			}
		}
	}
}
void CaretWindow::pasteFromClipboard() {
	if (!IsClipboardFormatAvailable(CF_TEXT))
		return;
	if (!OpenClipboard(hWndMain))
		return;

	HGLOBAL hglb = GetClipboardData(CF_TEXT);
	if (hglb != NULL)
	{
		LPTSTR lptstr = (LPTSTR)GlobalLock(hglb);
		if (lptstr != NULL)
		{
			// Call the application-defined ReplaceSelection 
			// function to insert the text and repaint the 
			// window. 
			deleteHighlight();

			int remLen = cellStr.length;
			cellStr = cellStr.replace(caretPos, 0, lptstr);

			GlobalUnlock(hglb);
			caretPos += cellStr.length - remLen;
			highlightPos = caretPos;

		}
	}
	CloseClipboard();

	updateCaret();
	onChange();
	win.paint();
}

bool tabKeyState() {
	if (0x8000 & GetKeyState(VK_TAB)) {
		return true;
	}
	else return false;
}

bool enterKeyState() {
	if (0x8000 & GetKeyState(VK_RETURN)) {
		return true;
	}
	else return false;
}

bool controlKeyState() {
	if (0x8000 & GetKeyState(VK_CONTROL)) {
		return true;
	}
	else return false;
}
bool altKeyState() {
	if (0x8000 & GetKeyState(VK_MENU)) {
		return true;
	}
	else return false;
}

bool shiftKeyState() {
	if (0x8000 & GetKeyState(VK_SHIFT)) {
		return true;
	}
	else return false;
}

bool zKeyState() {
	if (0x8000 & GetKeyState('Z')) {
		return true;
	}
	else return false;
}

bool aKeyState() {
	if (0x8000 & GetKeyState('A')) {
		return true;
	}
	else return false;
}

bool sKeyState() {
	if (0x8000 & GetKeyState('S')) {
		return true;
	}
	else return false;
}

bool dKeyState() {
	if (0x8000 & GetKeyState('D')) {
		return true;
	}
	else return false;
}

bool r_controlKeyState() {
	if (0x8000 & GetKeyState(VK_RCONTROL)) {
		return true;
	}
	else return false;
}
bool r_altKeyState() {
	if (0x8000 & GetKeyState(VK_RMENU)) {
		return true;
	}
	else return false;
}
bool r_shiftKeyState() {
	if (0x8000 & GetKeyState(VK_RSHIFT)) {
		return true;
	}
	else return false;
}
bool deleteKeyState() {
	if (0x8000 & GetKeyState(VK_DELETE)) {
		return true;
	}
	else return false;
}

void BackUp::deleteBackupFiles() {
	String filename;
	_close(g_lockFile);
	g_lockFile = 0;

	filename = g_undoDir;
	filename += ourBackUpName;
	filename += ".lock";
	DeleteFile(filename.c_str());

	BOOL noError = FALSE;
	int i = 0;
	do {
		filename = g_undoDir;
		filename += ourBackUpName;
		filename += i++;
		filename += ".smls";
		noError = DeleteFile(filename.c_str());
	} while (noError);
	backedUpFileCntr = 0;
}


void BackUp::createUniqueBackupName(bool copy) {
	String filename;
	int i = 0;
	while (1) {
		ourBackUpName = i++;
		ourBackUpName += '_';
		filename = g_undoDir;
		filename += ourBackUpName;
		filename += "0.smls";
		if (FILE* file = fopen(filename.c_str(), "r+b")) {
			fclose(file);
		}
		else {
			String lockName = g_undoDir;
			lockName += ourBackUpName;
			lockName += ".lock";
			if (g_lockFile > 0) {
				_close(g_lockFile);
			}
			g_lockFile = 0;
			SetFileAttributes(lockName.c_str(), FILE_ATTRIBUTE_NORMAL);
			if (_sopen_s(&g_lockFile, lockName.c_str(), _O_RDWR | _O_CREAT | _O_EXCL, _SH_DENYRW, _S_IREAD | _S_IWRITE) == 0) {
				//const String& file0 = g_settings->file.refValue()[0];
				_write(g_lockFile, g_currentFilename.c_str(), g_currentFilename.length);
				backupToCache(copy);
				break;
			}
			else {
				g_lockFile = 0;
			}
		}
	}
}

void BackUp::backupToCache(bool copy) {
	static String filename;

	//changeCursor(IDC_WAIT);

	filename = g_undoDir;
	filename += ourBackUpName;
	filename += backedUpFileCntr;
	filename += ".smls";
	SetFileAttributes(filename.c_str(), FILE_ATTRIBUTE_NORMAL);
	if (copy && g_settings->file.length) {
		CopyFile(g_settings->file.refValue()[0].c_str(), filename.c_str(), FALSE);
	}
	else {
		writeSmlsFile(filename);
	}
	backedUpFileCntrEnd = backedUpFileCntr;

	if (copy) {
		smlsSaved.set();
	}
	//setTitleName();
}
void splitPath(String wholePath, String& drive, String& dir, String& fname, String& ext) {
	char _dir[_MAX_DIR];
	char _fname[_MAX_FNAME];
	char _ext[_MAX_EXT];
	char _drive[_MAX_DRIVE];
	_splitpath(wholePath.c_str(), _drive, _dir, _fname, _ext);
	drive = _drive;
	dir = _dir;
	fname = _fname;
	ext = _ext;
}
String readFile(String filename) {
	static int maxBufSize = 0x4000;
	//dLog(String("read (")+=filename+")");
	if (filename == "") return "";
	int bufSize = 0;
	char* buf;

	gzFile gzfile;


	while (1) {
		buf = new char[maxBufSize + 2];
		if ((gzfile = gzopen(filename.c_str(), "rb")) != NULL) {
			bufSize = gzread(gzfile, buf, maxBufSize);
			gzclose(gzfile);
			if (bufSize == maxBufSize) {
				maxBufSize *= 2;
				delete buf;
			}
			else break;
		}
		else {
			dLog("Failed to open " + filename);
			break;
		}
	}
	buf[bufSize] = 0;
	return buf;
}
bool String::write(String fileName)const {
	if (g_writeGzipped) {
		gzFile gzfile;
		gzfile = gzopen(fileName.c_str(), "w+b");
		if (gzfile == NULL)return false;
		gzwrite(gzfile, const_cast<char*>(c_str()), length);
		gzclose(gzfile);
	}
	else {
		FILE* file;
		file = fopen(fileName.c_str(), "w+b");
		if (file == NULL)return false;
		fwrite(c_str(), length, 1, file);
		fclose(file);
	}
	return true;
}
void setNodeWinTitleBar(Node* node) {
	static String str;

	str = node->getTypeStr();
	/*
	if(node->getType()>=g_userProtoBegin){
		str="ProtoInstance ";
		str+=node->getTypeStr();
	}
	else{
		str=node->getTypeStr();
	}
*/
	str += ' ';
	str += node->name;
	SetWindowText(g_hDlg, str.c_str());

	/*
			String str=nodeLookUpStr[windowNode->getType()];
			str+=' ';
			str+=windowNode->name;
			SetWindowText(hWnd, str.c_str());
	*/

	/*
		str=node->name;
		str+=' ';
		if(node->getType()>=g_userProtoBegin){
			str+="ProtoInstance";
		}
		else{
			str+=node->getTypeStr();
		}
	*/
}

void Drag::endDrag(const Vec2i ptCursor) {
	g_drag = 0;
	ReleaseCapture();
}
void releaseCapture() {
	ReleaseCapture();
}


//new toolbar code

Matrix projMatrix = {
	1, 0, 0, 0,
	0, 1, 0, 0,
	0, 0, 1, 0,
	0, 0, 0, 1
};

int textureWidth = 25 * 8 + 1;
int textureHeight = 25 + 1;

Vec2i toolBarPos;

int selectedButton = 0;
HGLRC tempContext = 0;
HGLRC       hrc = 0;



Matrix identityMatrix = {
	1, 0, 0, 0,
	0, 1, 0, 0,
	0, 0, 1, 0,
	0, 0, 0, 1
};

struct Pixel {
public:
	Pixel() :
		r(255),
		g(255),
		b(255),
		a(255)
	{}
	Pixel(unsigned char r, unsigned char g, unsigned char b) :
		r(r),
		g(g),
		b(b),
		a(0)
	{}
	unsigned char r;
	unsigned char g;
	unsigned char b;
	unsigned char a;
};


float coord[] = {
	-.5, -.5, 0,
	.5, -.5, 0,
	-.5,  .5, 0,
	.5,  .5, 0,
};
float normal[] = {
	0, 0, 1,
	0, 0, 1,
	0, 0, 1,
	0, 0, 1
};
float color[] = {
	.5, 1, 0,
	0, 0, 1,
	1, 1, 0,
	1, 0, 0,
};

float ccColor[] = {
	0, 1, 1,
	0, 0, 1,
	0, 1, 0,
	1, 0, 1,
};


indexInt ccCoordIndex[] = {
	0, 1, 2, 3
};


indexInt coordIndex[] = {
	1, 2, 0,
	1, 3, 2
};


Vec2f t_coord[4];

float t_texCoord[] = {
	0, 0,
	1, 0,
	0, 1,
	1, 1,
};
indexInt t_coordIndex[] = {
	1, 2, 0,
	1, 3, 2
};



CNCShader cnc;
CNTShader cnt;
CCShader cc;



Pixel* pixels = 0;

void generateToolbarTexture(GLuint &texture) {
	Pixel selectedColor = Pixel(255, 0, 0);
	Pixel unselectedColor = Pixel(0, 0, 255);

	Pixel color = unselectedColor;

	for (int b = 0; b < 8; b++) {
		if (b == selectedButton) {
			color = selectedColor;
		}
		else {
			color = unselectedColor;
		}
		for (int w = 1; w < 24; w++) {
			pixels[b * 25 + textureWidth + w] = color;
			pixels[b * 25 + textureWidth * 24 + w] = color;
		}
		for (int h = 0; h < 24; h++) {
			pixels[b * 25 + textureWidth + h * textureWidth + 1] = color;
			pixels[b * 25 + textureWidth + h * textureWidth + 24] = color;
		}
	}
}


void initTexture(GLuint &texture) {
	pixels = new Pixel[textureWidth * textureHeight];


	glGenTextures(1, &texture);
	glBindTexture(GL_TEXTURE_2D, texture);

	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);

	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	generateToolbarTexture(texture);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA8, textureWidth, textureHeight, 0, GL_RGBA, GL_UNSIGNED_BYTE, pixels);
}


void display(HDC hdc)
{
	//glClearColor(.5, .5, .5, 1);
	glClearColor(0, 0, 0, 1);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	float aspect = float(g_3d->size.x) / float(g_3d->size.y);
	float left = -aspect;
	float right = aspect;
	float bottom = -1;
	float top = 1;
	float nearPoint = -3;
	float farPoint = 100;

	projMatrix.m[0][0] = 2.0f / (right - left);
	projMatrix.m[1][1] = 2.0f / (top - bottom);
	projMatrix.m[2][2] = -2.0f / (farPoint - nearPoint);
	projMatrix.m[3][0] = -(right + left) / (right - left);
	projMatrix.m[3][1] = -(top + bottom) / (top - bottom);
	projMatrix.m[3][2] = -(farPoint + nearPoint) / (farPoint - nearPoint);

	glEnable(GL_DEPTH_TEST);

	renderScene();

	cc.renderScene();
	cnc.renderScene();
	glDisable(GL_DEPTH_TEST);

	cnt.renderScene();


	Window::renderVisible();
	for (int i = 0; i < g_callBackList.length; i++) {
		g_callBackList[i]->doCallBack();
	}


	SwapBuffers(hdc);
}


LONG WINAPI WindowProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam) {

	HDC hdc = 0;

	static PAINTSTRUCT ps;


	g_before = g_selNode.getRaw();
	Art* part = dynamic_cast<Art*>(g_selNode.getRaw());
	Seamless* smls = 0;
	float tran = (1 / g_scene->getZoom())*.01F;


	Vec3f vec;
	Rotation tmpRot;


	if (TRUE == mouse2dHandler(message, wParam, lParam)) {
		return TRUE;
	}
	//	if(g_activeWindow->input(message, wParam, lParam)){
	//		return 0;
	//	}

#ifdef	OPENGL
	if (mouse3dMode) {
		mouse3dHandler(message, wParam);
	}
#endif

	switch (message) {

	case WM_CREATE: {
#ifdef WACOM_FILES
		WINDOWINFO appWindowInfo = { 0 };
		appWindowInfo.cbSize = sizeof(appWindowInfo);
		GetWindowInfo(hWnd, &appWindowInfo);
		g_clientRect = appWindowInfo.rcClient;
		//should not need above code if we use seamless native screen rectangle
#endif

		DragAcceptFiles(hWnd, TRUE);
		//InitDirectInput(hWnd);
		/*
	 // register raw input devices
		RAWINPUTDEVICE rID[2];

		// Keyboard
		rID[0].usUsagePage = 1;
		rID[0].usUsage = 6;
		rID[0].dwFlags = 0;
		rID[0].hwndTarget=NULL;

		// Mouse
		rID[1].usUsagePage = 1;
		rID[1].usUsage = 2;
		rID[1].dwFlags = 0;
		rID[1].hwndTarget = hWnd;
		if ( !RegisterRawInputDevices( rID, 2, sizeof( RAWINPUTDEVICE ) ) )
		{
			// handle failure
			throw;
		}
		*/
#ifdef WACOM_FILES
		// test for touch
		if (g_settings->wacomPackets) {
			g_tabCtx = InitWintabAPI(hWnd);
			if (g_tabCtx) {
				g_useWinTab = true;
				do {
					g_packetBufLen >>= 1;
					gpWTQueueSizeSet(g_tabCtx, g_packetBufLen);
				} while (gpWTQueueSizeGet(g_tabCtx) == 0);
				dLog("packetBufLen = gpWTQueueSizeSet(g_tabCtx, 0x100);");
				dLog(g_packetBufLen);
				//if(packetBufLen < 0x100){
				//	throw;
				//}
				disablePen();
			}
		}
#endif
		g_bReady = true;


		//InputEvent::setupSequence(hWnd);
		break;
	}
#ifdef WACOM_FILES
	case WM_FINGERDATA:
		interpretFingers(((WacomMTFingerCollection*)lParam)->FingerCount, ((WacomMTFingerCollection*)lParam)->Fingers, ((WacomMTFingerCollection*)lParam)->DeviceID);
		break;

	case WT_PACKET: {


		//int len = gpWTPacketsPeek((HCTX)lParam, 0x100, g_wintabPkt);
		int len = gpWTPacketsGet((HCTX)lParam, g_packetBufLen, g_wintabPktBuf);
		if (len > 50) {
			dLog("if(len > 50){");
			dLog(len);
		}
		g_lastPacketIndex = len - 1;

		//gpWTPacketsGet(g_tabCtx, 0x200, NULL);
		//gpWTPacket((HCTX)lParam, wParam, NULL);

		break;
	}
#endif
					/*
	 case WM_INPUT:
	 {
	RAWINPUT *pRI = NULL;

	 // Determine how big the buffer should be
	 UINT iBuffer;
	 GetRawInputData( ( HRAWINPUT )lParam, RID_INPUT, NULL, &iBuffer, sizeof( RAWINPUTHEADER ) );

	 // Allocate a buffer with enough size to hold the raw input data
	 LPBYTE lpb = new BYTE[ iBuffer ];
	 if( lpb == NULL )
		 return 0;

	 // Get the raw input data
	 UINT readSize = GetRawInputData( ( HRAWINPUT )lParam, RID_INPUT, lpb, &iBuffer, sizeof( RAWINPUTHEADER ) ) ;

	 // Validate that read size is as expected
	 if( readSize != iBuffer )
		 puts( "ERROR: GetRawInputData didn't return correct size!" ) ;
	 pRI = ( RAWINPUT* )lpb;

	 // Process the Mouse Messages
	 if(g_winAPI3DMode){
		 if( pRI->header.dwType== RIM_TYPEMOUSE ){
			 riProcessMouseMessage(&pRI->data.mouse);
		 }
	 }


	 // Destroy the Raw Input Data and Return
	 delete lpb;
	 return 1;
		 break;
	 }
	 */
	case WM_NCMOUSELEAVE: {
		//case WM_MOUSELEAVE:{
		if (g_drag) {
			g_drag->endDrag(g_mousePos);
		}
		break;
	}
	case WM_MOUSEWHEEL: {
		Window::activeOnMouseWheel((short)HIWORD(wParam) / WHEEL_DELTA);
		break;
	}

	case WM_PAINT: {
		hdc = BeginPaint(hWnd, &ps);
		//wglMakeCurrent(hdc, hrc);
		display(hdc);
		EndPaint(hWnd, &ps);
		return 0;
	}

	case WM_SIZE:
		glViewport(0, 0, LOWORD(lParam), HIWORD(lParam));
		PostMessage(hWnd, WM_PAINT, 0, 0);
		return 0;

	case WM_CHAR:
		switch (wParam) {
		case 27:			/* ESC key */
			PostQuitMessage(0);
			break;
		}
		return 0;

	case WM_CLOSE:
		PostQuitMessage(0);
		return 0;
	}

	return DefWindowProc(hWnd, message, wParam, lParam);
}

HWND CreateOpenGLWindow(char* title, int x, int y) {

	HDC         hDC;
	HWND        hWnd;
	WNDCLASS    wc;
	static HINSTANCE hInstance = 0;

	/* only register the window class once - use hInstance as a flag. */
	if (!hInstance) {
		hInstance = GetModuleHandle(NULL);
		wc.style = CS_OWNDC;
		wc.lpfnWndProc = (WNDPROC)WindowProc;
		wc.cbClsExtra = 0;
		wc.cbWndExtra = 0;
		wc.hInstance = hInstance;
		wc.hIcon = LoadIcon(NULL, IDI_WINLOGO);
		wc.hCursor = LoadCursor(NULL, IDC_ARROW);
		wc.hbrBackground = NULL;
		wc.lpszMenuName = NULL;
		wc.lpszClassName = "OpenGL";

		if (!RegisterClass(&wc)) {
			MessageBox(NULL, "RegisterClass() failed:  "
				"Cannot register window class.", "Error", MB_OK);
			return NULL;
		}
	}

	hWnd = CreateWindow("OpenGL", title, WS_OVERLAPPEDWINDOW |
		WS_CLIPSIBLINGS | WS_CLIPCHILDREN,
		x, y, g_3d->size.x, g_3d->size.y, NULL, NULL, hInstance, NULL);

	if (hWnd == NULL) {
		MessageBox(NULL, "CreateWindow() failed:  Cannot create a window.",
			"Error", MB_OK);
		return NULL;
	}

	hDC = GetDC(hWnd);

	/* there is no guarantee that the contents of the stack that become
	   the pfd are zeroed, therefore _make sure_ to clear these bits. */

	PIXELFORMATDESCRIPTOR pfd;
	memset(&pfd, 0, sizeof(PIXELFORMATDESCRIPTOR));
	pfd.nSize = sizeof(PIXELFORMATDESCRIPTOR);
	pfd.nVersion = 1;
	pfd.dwFlags = PFD_DOUBLEBUFFER | PFD_SUPPORT_OPENGL | PFD_DRAW_TO_WINDOW;
	pfd.iPixelType = PFD_TYPE_RGBA;
	pfd.cColorBits = 32;
	pfd.cDepthBits = 32;
	pfd.iLayerType = PFD_MAIN_PLANE;

	int nPixelFormat = ChoosePixelFormat(hDC, &pfd);

	if (nPixelFormat == 0) return false;

	BOOL bResult = SetPixelFormat(hDC, nPixelFormat, &pfd);

	if (!bResult) return false;

	tempContext = wglCreateContext(hDC);
	wglMakeCurrent(hDC, tempContext);

	GLenum err = glewInit();
	if (GLEW_OK != err)
	{
		throw;
	}

	int attribs[] =
	{
		WGL_CONTEXT_MAJOR_VERSION_ARB, 3,
		WGL_CONTEXT_MINOR_VERSION_ARB, 1,
		WGL_CONTEXT_FLAGS_ARB, 0,
		0
	};

	if (wglewIsSupported("WGL_ARB_create_context") == 1)
	{
		hrc = wglCreateContextAttribsARB(hDC, 0, attribs);
		wglMakeCurrent(NULL, NULL);
		wglDeleteContext(tempContext);
		wglMakeCurrent(hDC, hrc);
	}
	else
	{	//It's not possible to make a GL 3.x context. Use the old style context (GL 2.1 and before)
		hrc = tempContext;
	}

	//Checking GL version
	const GLubyte *GLVersionString = glGetString(GL_VERSION);

	//Or better yet, use the GL3 way to get the version number
	int OpenGLVersion[2];
	glGetIntegerv(GL_MAJOR_VERSION, &OpenGLVersion[0]);
	glGetIntegerv(GL_MINOR_VERSION, &OpenGLVersion[1]);

	if (!hrc) {
		throw;
	}
	return hWnd;
}

int APIENTRY WinMain(HINSTANCE hCurrentInst, HINSTANCE hPreviousInst, LPSTR lpszCmdLine, int nCmdShow) {
	g_hinst = hCurrentInst;
	g_lpszCmdLine = lpszCmdLine;

	g_3d->size = Vec2i(1280, 720);

	t_coord[1] = Vec2f(float(textureWidth) / float(g_3d->size.x) * 2, 0);
	t_coord[2] = Vec2f(0, float(textureHeight) / float(g_3d->size.y) * 2);
	t_coord[3] = Vec2f(float(textureWidth) / float(g_3d->size.x) * 2, float(textureHeight) / float(g_3d->size.y) * 2);


	toolBarPos = Vec2i(g_3d->size.x / 2, g_3d->size.y / 2 - textureHeight);


	setMathHiPresision();
	getMemberAddresses();

	//g_chatLines+=String("NSPE allows the user to hand edit the polygons on NURBS surfaces. This includes being able to drag the vertices anywhere along the NURBS surface as well as join the vertices together, break the vertices apart and color them. NSPE has a significant advantage over simply converting a NURBS surface to a polygon mesh for editing because NSPE lets the user be able continue to modify the NURBS surface for the hand edited polygon structure.");
	//g_chatLines+=String("http://www.seamless3d.com/forum/viewtopic.php?f=1&t=411");
	//g_chatLines+=String("Because NSPE ensures that when a polygon's vertex is dragged it will always be on the NURBS surface, NSPE greatly helps the user to avoid unintentionally changing the shape of the model when optimizing for real time animation.");
	//processWords(8);
	//printOutLines();

	g_extraWinPix.x = GetSystemMetrics(SM_CXSIZEFRAME) * 2;
	g_extraWinPix.y = GetSystemMetrics(SM_CYCAPTION) + GetSystemMetrics(SM_CYSIZEFRAME) * 2;


	g_scene->rotation = Rotation(.831685, .555163, -.009739, .7695);

	try {
		mkdir(g_seamlessDir.c_str());
		mkdir(g_appDataDir.c_str());
		mkdir(g_undoDir.c_str());
		mkdir(g_screenshotDir.c_str());
		mkdir(g_demoDir.c_str());
		mkdir(g_demoCacheDir.c_str());

		SYSTEMTIME ToDaySystemTime;

		String debugStr = "\n\n\nSeamless3d ";
		debugStr += g_version;
#ifdef OPENGL
		debugStr += " OpenGL";
#endif
		dLog(debugStr);
		GetSystemTime(&ToDaySystemTime);
		char tmpChar[0x200];
		sprintf(tmpChar, "GMT: %.2i/%.2i/%i  %.2i:%.2i", ToDaySystemTime.wDay, ToDaySystemTime.wMonth, ToDaySystemTime.wYear, ToDaySystemTime.wHour, ToDaySystemTime.wMinute);
		dLog(tmpChar);

		//sprintf(tmpChar,"OS: - Version: %i.%i - build: %i ",_winmajor,_winminor,_osver);
		dLog(tmpChar);

		initializeShortCuts();

		initLookup();

		Settings::loadINI();
		g_3d->size = Vec2i(g_settings->win3dSize.x, g_settings->win3dSize.y);

		g_mouseSense = 6.0 / g_settings->win3dSize.y;

		addNodesForNewFile();
		BackUp::createUniqueBackupName();

		Window::initializeWindows();
		initWinINet();

		// bumpy hack - recover auto-save file was inserted here

		initVisualClue();
		initRadialTug();



		//		createSting();

#ifdef WACOM_FILES
		InitializeCriticalSection(&g_graphicsCriticalSection);
#endif
		//////////////////////////////////
		// Register the window class
/*
		WNDCLASSEX wcMainWin = { sizeof(WNDCLASSEX), CS_VREDRAW | CS_HREDRAW, MsgProc, 0L, 0L,
			GetModuleHandle(NULL), NULL, NULL, NULL, NULL,
			"wcMainWin", NULL };
		wcMainWin.hIcon = LoadIcon(hInst, MAKEINTRESOURCE(IDI_MAIN_ICON));
		wcMainWin.hCursor = LoadCursor(NULL, IDC_ARROW);

		RegisterClassEx(&wcMainWin);
*/
		char TempStr[0x200];
		makeTitleStr(TempStr);

		RECT rt = RECT{ 0, 0, int(g_settings->win3dSize.x), int(g_settings->win3dSize.y) };

		AdjustWindowRect(
			&rt,
			WS_OVERLAPPEDWINDOW,
			FALSE
		);

		/* Create a window of the previously defined class */
/*
		hWndMain = CreateWindow("wcMainWin", TempStr,
			//WS_SIZEBOX | WS_SYSMENU | WS_CAPTION | WS_MAXIMIZEBOX,
			WS_OVERLAPPEDWINDOW,
			g_settings->win3dPos.x, g_settings->win3dPos.y,
			//g_settings->win3dSize.x + g_extraWinPix.x, g_settings->win3dSize.y + g_extraWinPix.y,
			rt.right - rt.left, rt.bottom - rt.top,
			GetDesktopWindow(), NULL, wcMainWin.hInstance, NULL);
*/



		hWndMain = CreateOpenGLWindow("minimal", 0, 0);
		//ShowWindow(hWndMain, nCmdShow);

		cc.coord = coord;
		cc.color = ccColor;
		cc.coordIndex = ccCoordIndex;
		cc.vertexLen = 4;
		cc.coordIndexLen = 4;

		cc.setupShaders();
		cc.initMesh();

		cnc.coord = coord;
		cnc.normal = normal;
		cnc.color = color;
		cnc.coordIndex = coordIndex;
		cnc.vertexLen = 4;
		cnc.coordIndexLen = 6;

		cnc.setupShaders();
		cnc.initMesh();

		///////
		cnt.coord = t_coord;
		cnt.texCoord = t_texCoord;
		cnt.coordIndex = t_coordIndex;
		cnt.vertexLen = 4;
		cnt.coordIndexLen = 6;


		cnt.setupShaders();

		cnt.initMesh();

		initTexture(cnt.texture);


		initPen();
#ifdef WACOM_FILES
		WacomMTError res = WacomMTInitialize(WACOM_MULTI_TOUCH_API_VERSION);
		if (res == WMTErrorSuccess) {
			int deviceCount = WacomMTGetAttachedDeviceIDs(NULL, 0);
			if (deviceCount) {
				int newCount = 0;
				while (newCount != deviceCount) {
					g_devices.resize(deviceCount, 0);
					newCount = WacomMTGetAttachedDeviceIDs(&g_devices[0], deviceCount * sizeof(int));
				}

				int loopCount = (int)g_devices.size();
				for (int idx = 0; idx < loopCount; idx++) {
					WacomMTCapability cap = { 0 };
					res = WacomMTGetDeviceCapabilities(g_devices[idx], &cap);
					g_caps[g_devices[idx]] = cap;
					DumpCaps(g_devices[idx]);
				}
			}
			//following needed for finger touch
			res = WacomMTRegisterAttachCallback(AttachCallback, NULL);
			res = WacomMTRegisterDetachCallback(DetachCallback, NULL);

			int loopCount = (int)g_devices.size();
			for (int idx = 0; idx < loopCount; idx++)
			{
				res = RegisterForData(g_devices[idx]);
			}
		}

#endif
		getClientRect(clientOffest);
		g_3d->size = Vec2i(clientOffest.right - clientOffest.left, clientOffest.bottom - clientOffest.top);
		ToolBarWin::initToolbar();
		g_acp->initialize();
		g_stw = new SceneTree;
		g_rcp = new ControlPanel;
		//g_acp = new AnimControlPanel;
		openSceneTreeInterface();

		// Initialize Direct3D
		{
			dLog("InitRender SUCCEEDED");
#ifdef BIND_DEVELOPER_SCRIPT
			//initializeDeveloperScript();
#endif
//			convertVisualCluesToRenderFormat();


			// Show the window
			ShowWindow(hWndMain, SW_SHOWDEFAULT);
			UpdateWindow(hWndMain);


			g_arrowCursor = (long)LoadCursor(0, MAKEINTRESOURCE(IDC_ARROW));
			g_handCursor = (long)LoadCursor(0, MAKEINTRESOURCE(IDC_HAND));

			g_sizeAllCursor = (long)LoadCursor(0, MAKEINTRESOURCE(IDC_SIZEALL));
			g_sizeNWSECursor = (long)LoadCursor(0, MAKEINTRESOURCE(IDC_SIZENWSE));
			g_sizeNESWCursor = (long)LoadCursor(0, MAKEINTRESOURCE(IDC_SIZENESW));
			g_sizeWECursor = (long)LoadCursor(0, MAKEINTRESOURCE(IDC_SIZEWE));
			g_sizeNSCursor = (long)LoadCursor(0, MAKEINTRESOURCE(IDC_SIZENS));

			g_currentType2 = g_arrowCursor;


			g_crossCursor = (long)LoadCursor(g_hinst, MAKEINTRESOURCE(CROSS_CURSOR));
			g_lockedOnPointCursor = (long)LoadCursor(g_hinst, MAKEINTRESOURCE(LOCKED_ON_POINT_CURSOR));
			g_addPointCursor = (long)LoadCursor(g_hinst, MAKEINTRESOURCE(ADD_POINT_CURSOR));
			g_insertLineCursor = (long)LoadCursor(g_hinst, MAKEINTRESOURCE(INSERT_LINE_CURSOR));
			g_deleteLineCursor = (long)LoadCursor(g_hinst, MAKEINTRESOURCE(DELETE_LINE_CURSOR));
			g_mergePatchesCursor = (long)LoadCursor(g_hinst, MAKEINTRESOURCE(MERGE_PATCHES_CURSOR));
			g_addGeometryCursor = (long)LoadCursor(g_hinst, MAKEINTRESOURCE(ADD_GEOMETRY_CURSOR));
			g_addGeometryMirrorCursor = (long)LoadCursor(g_hinst, MAKEINTRESOURCE(ADD_GEOMETRY_MIRROR_CURSOR));
			g_ringSelectCursor = (long)LoadCursor(g_hinst, MAKEINTRESOURCE(RING_SELECT_CURSOR));
			g_rectangleSelectPointCursor = (long)LoadCursor(g_hinst, MAKEINTRESOURCE(RECTANGLE_SELECT_POINT_CURSOR));
			g_rectangleUnselectCursor = (long)LoadCursor(g_hinst, MAKEINTRESOURCE(RECTANGLE_UNSELECT_CURSOR));
			g_rectangleSelectPolyCursor = (long)LoadCursor(g_hinst, MAKEINTRESOURCE(RECTANGLE_SELECT_POLY_CURSOR));
			g_fillCursor = (long)LoadCursor(g_hinst, MAKEINTRESOURCE(FILL_CURSOR));
			g_brushCursor = (long)LoadCursor(g_hinst, MAKEINTRESOURCE(BRUSH_CURSOR));
			g_eyeDropperCursor = (long)LoadCursor(g_hinst, MAKEINTRESOURCE(COLOR_PICKER_CURSOR));
			g_knifeCursor = (long)LoadCursor(g_hinst, MAKEINTRESOURCE(CUT_PATCH_CURSOR));
			g_addBoneCursor = (long)LoadCursor(g_hinst, MAKEINTRESOURCE(ADD_BONE_CURSOR));
			g_insertBoneCursor = (long)LoadCursor(g_hinst, MAKEINTRESOURCE(INSERT_BONE_CURSOR));

			g_modellingCursor = g_crossCursor;



			resetLooseEnds();
			Viewpoint::gotoFirstViewpoint();
			g_grid->houseKeep();


			//Window::topWindow=g_stw.p;
			//if(g_settings->showToolBar){
			Window::topWindow = g_tb.p;
			Window::topWindow = g_tb2.p;
			//}
			g_stw->size = Vec2i(330, 660);
			g_stw->pos = Vec2i(g_3d->size.x - 330, 0);
			g_stw->newNodeBar.dstPos = Vec2i(g_stw->size.x - 80, 30);
			g_stw->cp->size = Vec2i(350, g_stw->cp->size.y);
			g_stw->cp->pos = Vec2i(0, 26);
			Window::postLoadProcessWin(*g_scene);
			//if(g_tb->pos.y > g_3d->size.y - g_tb->size.y){
			//	g_tb->pos = Vec2i(0, g_3d->size.y - g_tb->size.y);
			//}

			g_tb->pos = Vec2i(0, 0);
			g_tb2->pos = Vec2i(0, g_3d->size.y - 25);

			Window::activeWindow = g_tb.p;
			Window::activeWindow = g_tb2.p;


			//Window::topWindow=g_pw.p;


			// Load keyboard accelerators
			hAccel = LoadAccelerators(hCurrentInst, MAKEINTRESOURCE(IDR_MAIN_ACCEL));


#ifdef FRAME_TIME_SYNC
			g_timeBegan = g_time = 0;
#else
			g_timeBegan = g_time = timeGetTime() * MILISECONDS_TO_SECONDS;
#endif


			// Enter the message loop
			MSG msg;
			ZeroMemory(&msg, sizeof(msg));

			//if(g_createdINI){
			//}
			BOOL bGotMsg = FALSE;
			g_stw->size = Vec2i(g_stw->size.x, g_stw->size.y); //do we really need this?
			g_stw->paint();

			if (checkForLockFiles() == false) {
				if (openFileFromCommandLine() == false) {
					setupRecentFiles();
				}
			}


			while (WM_QUIT != msg.message) {
				bGotMsg = PeekMessage(&msg, NULL, 0U, 0U, PM_REMOVE);
				if (!g_bActive) {
					//					bGotMsg = GetMessage( &msg, NULL, 0U, 0U );
										//bGotMsg = PeekMessage( &msg, NULL, 0U, 0U, PM_REMOVE );
					Sleep(100);
				}

				if (bGotMsg) {
					// Translate and dispatch the message
					if (0 == TranslateAccelerator(hWndMain, 0, &msg)) {
						TranslateMessage(&msg);
						DispatchMessage(&msg);
					}
				}
				else {
					// Render a frame during idle time (no messages are waiting)
					if (g_bReady) {
						HDC hdc = GetDC(hWndMain);
						display(hdc);
					}

				}
			}


			wglMakeCurrent(NULL, NULL);
			//ReleaseDC(hWnd, hDC);
		   //wglDeleteContext(hrc);
			DestroyWindow(hWndMain);
		}
	}

	catch (Field*e) {
		dLog("error in " + e->getTypeStr());
	}
	catch (M_Float* e) {
		dLog("error in M_Float");
		dLog((int)e);
		dLog((char *)e->showReason());
	}
	catch (M_Vec3f* e) {
		dLog("error in M_Vec3f");
		dLog((int)e);
		dLog((char *)e->showReason());
	}
	catch (M_Vec2f* e) {
		dLog("error in M_Vec2f");
		dLog((char *)e->showReason());
	}
	catch (String* e) {
		dLog("error in String");
		dLog((char *)e->showReason());
	}
	catch (M_Color3f* e) {
		dLog("error in M_Color3f");
		dLog((char *)e->showReason());
	}
	catch (M_Int* e) {
		dLog("error in M_Int");
		dLog((char *)e->showReason());
	}
	catch (M_Node* e) {
		dLog("error in M_Node");
		dLog((char *)e->showReason());
	}

	catch (M_Triangle* e) {
		dLog("error in M_Triangle");
		dLog((int)e);
		dLog((char *)e->showReason());
	}

	catch (Node* e) {
		dLog(String("Exception in ") + e->getTypeStr());
	}
	catch (...) {
		dLog(String("Exception ..."));
	}

	InputEvent::resumeSequence();
	return 0;
}


INT WINAPI WinMainOld(HINSTANCE hCurrentInst, HINSTANCE, LPSTR lpszCmdLine, INT) {//
	g_hinst = hCurrentInst;

	//	CopyLocalToScratchPad_doOp_S_Float();

		//if(FILE* readStream = fopen("C:\\Users\\thyme\\Documents\\winter-2014-vlog\\trench-3b.wav", "r+b")){
		//	unsigned char* buf = new unsigned char[0x2c];
		//	int a = fread(buf, 1, 0x2c, readStream);
		//	fclose(readStream);
		//	String str;

		//	for(int i = 0; i < 0x2c; i++){
		//		unsigned char num = buf[i];
		//		str += "0X";
		//		str += nibleToHex(num >> 4);
		//		str += nibleToHex(num & 0x0f);
		//		str +=",";

		//	}
		//	dLog(str);
		//}

	g_lpszCmdLine = lpszCmdLine;
	//int ss = 69;
	//ss = 252;
	//ss = ss;

//	double a = 2.000001;
//	double b = 1;
//	bool result = roughGE(a, b);
//	result = result;

	// o  = old r = new
	//Vec3f v = Vec3f(0, 1, 0);
	//Rotation o = Rotation(0, 0, 1, PI*.5);
	//Vec3f v2 = o * v;
	//Rotation r = Rotation(0, 1, 0, PI);
	//r = o * r;
	//v = r * v;
	//v = v;

//	int* p = new int;
//	_CrtMemCheckpoint( &s1 );
	//delete p;
//	p = new int;
//	_CrtMemCheckpoint( &s2 );
//	int flag = _CrtMemDifference(&s3, &s1, &s2);

//	dLog(s1.lTotalCount);
//	dLog(s2.lTotalCount);

//	Rotation r = -Rotation(.92497, .268729, .268729, 1.648711);
//	genButtons();

	//convert_raw_alpha_bitmap_to_cpp();

	//	double freq=220;
//	double cent = 1200;
//	double answer = freq * pow((pow(2.0,1.0/1200.0)),cent);

//	drawText2("", 0, Vec2i(0,0), 0, 0, 0);

//	return 0;

//	unsigned char ch = '';
//	Rotation r = Rotation(1, 0, 0, -PI);
//	r*= Rotation(1, 0, 0, PI*.25);
	//return 0;

//	M_Vec3f ma=M_Vec3f(Vec3f(0,1,2));
//	M_Vec3f mb;
//	ma=mb;
//	__asm{
//		push 0xffffff
//		mov	 ecx,0xffffffff 
//	}

//	Vec3f vec = Vec3f(1,0,0) * translationMatrix(Vec3f(1,0,0)) * rotationMatrix(Rotation(0,1,0, PI*.5));


	setMathHiPresision();
	getMemberAddresses();

	//g_chatLines+=String("NSPE allows the user to hand edit the polygons on NURBS surfaces. This includes being able to drag the vertices anywhere along the NURBS surface as well as join the vertices together, break the vertices apart and color them. NSPE has a significant advantage over simply converting a NURBS surface to a polygon mesh for editing because NSPE lets the user be able continue to modify the NURBS surface for the hand edited polygon structure.");
	//g_chatLines+=String("http://www.seamless3d.com/forum/viewtopic.php?f=1&t=411");
	//g_chatLines+=String("Because NSPE ensures that when a polygon's vertex is dragged it will always be on the NURBS surface, NSPE greatly helps the user to avoid unintentionally changing the shape of the model when optimizing for real time animation.");
	//processWords(8);
	//printOutLines();

	g_extraWinPix.x = GetSystemMetrics(SM_CXSIZEFRAME) * 2;
	g_extraWinPix.y = GetSystemMetrics(SM_CYCAPTION) + GetSystemMetrics(SM_CYSIZEFRAME) * 2;


	g_scene->rotation = Rotation(.831685, .555163, -.009739, .7695);

	try {
		mkdir(g_seamlessDir.c_str());
		mkdir(g_appDataDir.c_str());
		mkdir(g_undoDir.c_str());
		mkdir(g_screenshotDir.c_str());
		mkdir(g_demoDir.c_str());
		mkdir(g_demoCacheDir.c_str());

		SYSTEMTIME ToDaySystemTime;

		String debugStr = "\n\n\nSeamless3d ";
		debugStr += g_version;
#ifdef OPENGL
		debugStr += " OpenGL";
#endif
		dLog(debugStr);
		GetSystemTime(&ToDaySystemTime);
		char tmpChar[0x200];
		sprintf(tmpChar, "GMT: %.2i/%.2i/%i  %.2i:%.2i", ToDaySystemTime.wDay, ToDaySystemTime.wMonth, ToDaySystemTime.wYear, ToDaySystemTime.wHour, ToDaySystemTime.wMinute);
		dLog(tmpChar);

		//sprintf(tmpChar,"OS: - Version: %i.%i - build: %i ",_winmajor,_winminor,_osver);
		dLog(tmpChar);

		initializeShortCuts();

		initLookup();

		Settings::loadINI();
		g_3d->size = Vec2i(g_settings->win3dSize.x, g_settings->win3dSize.y);

		g_mouseSense = 6.0 / g_settings->win3dSize.y;

		addNodesForNewFile();
		BackUp::createUniqueBackupName();

		Window::initializeWindows();
		initWinINet();

		// bumpy hack - recover auto-save file was inserted here

		initVisualClue();
		initRadialTug();



		//		createSting();

#ifdef WACOM_FILES
		InitializeCriticalSection(&g_graphicsCriticalSection);
#endif
		//////////////////////////////////
		// Register the window class
/*
		WNDCLASSEX wcMainWin = { sizeof(WNDCLASSEX), CS_VREDRAW | CS_HREDRAW, MsgProc, 0L, 0L,
			GetModuleHandle(NULL), NULL, NULL, NULL, NULL,
			"wcMainWin", NULL };
		wcMainWin.hIcon = LoadIcon(hCurrentInst, MAKEINTRESOURCE(IDI_MAIN_ICON));
		wcMainWin.hCursor = LoadCursor(NULL, IDC_ARROW);

		RegisterClassEx(&wcMainWin);
*/
		char TempStr[0x200];
		makeTitleStr(TempStr);

		RECT rt = RECT{ 0, 0, int(g_settings->win3dSize.x), int(g_settings->win3dSize.y) };

		AdjustWindowRect(
			&rt,
			WS_OVERLAPPEDWINDOW,
			FALSE
		);

		/* Create a window of the previously defined class */
/*
		hWndMain = CreateWindow("wcMainWin", TempStr,
			//WS_SIZEBOX | WS_SYSMENU | WS_CAPTION | WS_MAXIMIZEBOX,
			WS_OVERLAPPEDWINDOW,
			g_settings->win3dPos.x, g_settings->win3dPos.y,
			//g_settings->win3dSize.x + g_extraWinPix.x, g_settings->win3dSize.y + g_extraWinPix.y,
			rt.right - rt.left, rt.bottom - rt.top,
			GetDesktopWindow(), NULL, wcMainWin.hInstance, NULL);
*/

		hWndMain = CreateOpenGLWindow("minimal", 0, 0);
		if (hWndMain == NULL) {
			exit(1);
		}





		initPen();
#ifdef WACOM_FILES
		WacomMTError res = WacomMTInitialize(WACOM_MULTI_TOUCH_API_VERSION);
		if (res == WMTErrorSuccess) {
			int deviceCount = WacomMTGetAttachedDeviceIDs(NULL, 0);
			if (deviceCount) {
				int newCount = 0;
				while (newCount != deviceCount) {
					g_devices.resize(deviceCount, 0);
					newCount = WacomMTGetAttachedDeviceIDs(&g_devices[0], deviceCount * sizeof(int));
				}

				int loopCount = (int)g_devices.size();
				for (int idx = 0; idx < loopCount; idx++) {
					WacomMTCapability cap = { 0 };
					res = WacomMTGetDeviceCapabilities(g_devices[idx], &cap);
					g_caps[g_devices[idx]] = cap;
					DumpCaps(g_devices[idx]);
				}
			}
			//following needed for finger touch
			res = WacomMTRegisterAttachCallback(AttachCallback, NULL);
			res = WacomMTRegisterDetachCallback(DetachCallback, NULL);

			int loopCount = (int)g_devices.size();
			for (int idx = 0; idx < loopCount; idx++)
			{
				res = RegisterForData(g_devices[idx]);
			}
		}

#endif
		getClientRect(clientOffest);
		g_3d->size = Vec2i(clientOffest.right - clientOffest.left, clientOffest.bottom - clientOffest.top);
		ToolBarWin::initToolbar();
		g_acp->initialize();
		g_stw = new SceneTree;
		g_rcp = new ControlPanel;
		//g_acp = new AnimControlPanel;
		openSceneTreeInterface();

		// Initialize Direct3D
		{
			dLog("InitRender SUCCEEDED");
#ifdef BIND_DEVELOPER_SCRIPT
			//initializeDeveloperScript();
#endif
//			convertVisualCluesToRenderFormat();


			// Show the window
			ShowWindow(hWndMain, SW_SHOWDEFAULT);
			UpdateWindow(hWndMain);


			g_arrowCursor = (long)LoadCursor(0, MAKEINTRESOURCE(IDC_ARROW));
			g_handCursor = (long)LoadCursor(0, MAKEINTRESOURCE(IDC_HAND));

			g_sizeAllCursor = (long)LoadCursor(0, MAKEINTRESOURCE(IDC_SIZEALL));
			g_sizeNWSECursor = (long)LoadCursor(0, MAKEINTRESOURCE(IDC_SIZENWSE));
			g_sizeNESWCursor = (long)LoadCursor(0, MAKEINTRESOURCE(IDC_SIZENESW));
			g_sizeWECursor = (long)LoadCursor(0, MAKEINTRESOURCE(IDC_SIZEWE));
			g_sizeNSCursor = (long)LoadCursor(0, MAKEINTRESOURCE(IDC_SIZENS));

			g_currentType2 = g_arrowCursor;


			g_crossCursor = (long)LoadCursor(g_hinst, MAKEINTRESOURCE(CROSS_CURSOR));
			g_lockedOnPointCursor = (long)LoadCursor(g_hinst, MAKEINTRESOURCE(LOCKED_ON_POINT_CURSOR));
			g_addPointCursor = (long)LoadCursor(g_hinst, MAKEINTRESOURCE(ADD_POINT_CURSOR));
			g_insertLineCursor = (long)LoadCursor(g_hinst, MAKEINTRESOURCE(INSERT_LINE_CURSOR));
			g_deleteLineCursor = (long)LoadCursor(g_hinst, MAKEINTRESOURCE(DELETE_LINE_CURSOR));
			g_mergePatchesCursor = (long)LoadCursor(g_hinst, MAKEINTRESOURCE(MERGE_PATCHES_CURSOR));
			g_addGeometryCursor = (long)LoadCursor(g_hinst, MAKEINTRESOURCE(ADD_GEOMETRY_CURSOR));
			g_addGeometryMirrorCursor = (long)LoadCursor(g_hinst, MAKEINTRESOURCE(ADD_GEOMETRY_MIRROR_CURSOR));
			g_ringSelectCursor = (long)LoadCursor(g_hinst, MAKEINTRESOURCE(RING_SELECT_CURSOR));
			g_rectangleSelectPointCursor = (long)LoadCursor(g_hinst, MAKEINTRESOURCE(RECTANGLE_SELECT_POINT_CURSOR));
			g_rectangleUnselectCursor = (long)LoadCursor(g_hinst, MAKEINTRESOURCE(RECTANGLE_UNSELECT_CURSOR));
			g_rectangleSelectPolyCursor = (long)LoadCursor(g_hinst, MAKEINTRESOURCE(RECTANGLE_SELECT_POLY_CURSOR));
			g_fillCursor = (long)LoadCursor(g_hinst, MAKEINTRESOURCE(FILL_CURSOR));
			g_brushCursor = (long)LoadCursor(g_hinst, MAKEINTRESOURCE(BRUSH_CURSOR));
			g_eyeDropperCursor = (long)LoadCursor(g_hinst, MAKEINTRESOURCE(COLOR_PICKER_CURSOR));
			g_knifeCursor = (long)LoadCursor(g_hinst, MAKEINTRESOURCE(CUT_PATCH_CURSOR));
			g_addBoneCursor = (long)LoadCursor(g_hinst, MAKEINTRESOURCE(ADD_BONE_CURSOR));
			g_insertBoneCursor = (long)LoadCursor(g_hinst, MAKEINTRESOURCE(INSERT_BONE_CURSOR));

			g_modellingCursor = g_crossCursor;



			resetLooseEnds();
			Viewpoint::gotoFirstViewpoint();
			g_grid->houseKeep();


			//Window::topWindow=g_stw.p;
			//if(g_settings->showToolBar){
			Window::topWindow = g_tb.p;
			Window::topWindow = g_tb2.p;
			//}
			g_stw->size = Vec2i(330, 660);
			g_stw->pos = Vec2i(g_3d->size.x - 330, 0);
			g_stw->newNodeBar.dstPos = Vec2i(g_stw->size.x - 80, 30);
			g_stw->cp->size = Vec2i(350, g_stw->cp->size.y);
			g_stw->cp->pos = Vec2i(0, 26);
			Window::postLoadProcessWin(*g_scene);
			//if(g_tb->pos.y > g_3d->size.y - g_tb->size.y){
			//	g_tb->pos = Vec2i(0, g_3d->size.y - g_tb->size.y);
			//}

			g_tb->pos = Vec2i(0, 0);
			g_tb2->pos = Vec2i(0, g_3d->size.y - 25);

			Window::activeWindow = g_tb.p;
			Window::activeWindow = g_tb2.p;


			//Window::topWindow=g_pw.p;


			// Load keyboard accelerators
			hAccel = LoadAccelerators(hCurrentInst, MAKEINTRESOURCE(IDR_MAIN_ACCEL));


#ifdef FRAME_TIME_SYNC
			g_timeBegan = g_time = 0;
#else
			g_timeBegan = g_time = timeGetTime() * MILISECONDS_TO_SECONDS;
#endif


			// Enter the message loop
			MSG msg;
			ZeroMemory(&msg, sizeof(msg));

			//if(g_createdINI){
			//}
			BOOL bGotMsg = FALSE;
			g_stw->size = Vec2i(g_stw->size.x, g_stw->size.y); //do we really need this?
			g_stw->paint();

			if (checkForLockFiles() == false) {
				if (openFileFromCommandLine() == false) {
					setupRecentFiles();
				}
			}


			while (WM_QUIT != msg.message) {
				// Use PeekMessage() if the app is active, so we can use idle time to
				// render the scene. Else, use GetMessage() to avoid eating CPU time.

#ifdef RECORD_SOUNDMOVIE
				if (g_playSoundMovie && g_playing) {
					DWORD dwResult = MsgWaitForMultipleObjects(1, &g_hNotificationEvent, FALSE, INFINITE, QS_ALLEVENTS);
					switch (dwResult) {
					case WAIT_OBJECT_0 + 0: {
						if (g_playing) {
							//updatePlayBuffer();
							updatePlayBufferRealTimeSynth();
							//renderScene();
						}
						break;
					}
					case WAIT_OBJECT_0 + 1: {
						// Windows messages are available
						while (PeekMessage(&msg, NULL, 0, 0, PM_REMOVE)) {
							TranslateMessage(&msg);
							DispatchMessage(&msg);
						}
					}
					}
				}
				else
					if (g_recordSoundMovie) {
						DWORD dwResult = MsgWaitForMultipleObjects(1, &g_hNotificationEvent, FALSE, INFINITE, QS_ALLEVENTS);
						switch (dwResult) {
						case WAIT_OBJECT_0 + 0: {
							renderScene();
							HDC hdc = GetDC(hWndMain);
							SwapBuffers(hdc);
							//outputScreenShotForDemo();
							break;
						}
						case WAIT_OBJECT_0 + 1: {
							// Windows messages are available
							while (PeekMessage(&msg, NULL, 0, 0, PM_REMOVE)) {
								TranslateMessage(&msg);
								DispatchMessage(&msg);
							}
						}
						}
					}
					else {
						bGotMsg = PeekMessage(&msg, NULL, 0U, 0U, PM_REMOVE);
						if (!g_bActive) {
							//					bGotMsg = GetMessage( &msg, NULL, 0U, 0U );
												//bGotMsg = PeekMessage( &msg, NULL, 0U, 0U, PM_REMOVE );
							Sleep(100); //sleep with wintab packets enabled causes issues when seamless runs in background
						}

						if (bGotMsg) {
							// Translate and dispatch the message
							if (0 == TranslateAccelerator(hWndMain, hAccel, &msg)) {
								TranslateMessage(&msg);
								DispatchMessage(&msg);
							}
						}
						else {
							// Render a frame during idle time (no messages are waiting)
							if (g_bReady) {
								renderScene();
								HDC hdc = GetDC(hWndMain);
								SwapBuffers(hdc);
							}

							BackUp::checkForBackUpOperation();
						}
					}
#else
				bGotMsg = PeekMessage(&msg, NULL, 0U, 0U, PM_REMOVE);
				if (!g_bActive) {
					//					bGotMsg = GetMessage( &msg, NULL, 0U, 0U );
										//bGotMsg = PeekMessage( &msg, NULL, 0U, 0U, PM_REMOVE );
					Sleep(100);
				}

				if (bGotMsg) {
					// Translate and dispatch the message
					if (0 == TranslateAccelerator(hWndMain, hAccel, &msg)) {
						TranslateMessage(&msg);
						DispatchMessage(&msg);
					}
				}
				else {
					// Render a frame during idle time (no messages are waiting)
					if (g_bReady) {
						renderScene();
					}

					BackUp::checkForBackUpOperation();
				}
#endif
			}
		}

		// Clean up everything and exit the app
		Cleanup();

		UnregisterClass("chatWc", hCurrentInst);
		UnregisterClass("wcMainWin", hCurrentInst);

		GetSystemTime(&ToDaySystemTime);
		sprintf(tmpChar, "GMT: %.2i/%.2i/%i  %.2i:%.2i", ToDaySystemTime.wDay, ToDaySystemTime.wMonth, ToDaySystemTime.wYear, ToDaySystemTime.wHour, ToDaySystemTime.wMinute);
		dLog(tmpChar);
		dLog("end");

	}
	catch (Field*e) {
		dLog("error in " + e->getTypeStr());
	}
	catch (M_Float* e) {
		dLog("error in M_Float");
		dLog((int)e);
		dLog((char *)e->showReason());
	}
	catch (M_Vec3f* e) {
		dLog("error in M_Vec3f");
		dLog((int)e);
		dLog((char *)e->showReason());
	}
	catch (M_Vec2f* e) {
		dLog("error in M_Vec2f");
		dLog((char *)e->showReason());
	}
	catch (String* e) {
		dLog("error in String");
		dLog((char *)e->showReason());
	}
	catch (M_Color3f* e) {
		dLog("error in M_Color3f");
		dLog((char *)e->showReason());
	}
	catch (M_Int* e) {
		dLog("error in M_Int");
		dLog((char *)e->showReason());
	}
	catch (M_Node* e) {
		dLog("error in M_Node");
		dLog((char *)e->showReason());
	}

	catch (M_Triangle* e) {
		dLog("error in M_Triangle");
		dLog((int)e);
		dLog((char *)e->showReason());
	}

	catch (Node* e) {
		dLog(String("Exception in ") + e->getTypeStr());
	}
	catch (...) {
		dLog(String("Exception ..."));
	}

	InputEvent::resumeSequence();
	return 0;
}
