/* Seamless3d

Copyright (C) 2001-2017 Graham John Perrett (AKA thyme) thyme@seamless3d.com www.seamless3d.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this
software and associated documentation files (the "Software"), to deal in the Software
without restriction, including without limitation the rights to use, copy, modify,
merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to the following
conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT
OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.
*/

#include "seamless3d.h"
#include "techuelife_triangle_editor.h"
#include "gui.h"


#include "scene_tree.h"
#include "robot_aid.h"

extern int g_fps;
extern int g_numOfChannels; //1 = mono, 2 = stereo
void releaseCapture();

Ptr<AnimControlPanel> g_acp = new AnimControlPanel;
bool PosCompensator::rightSideDrag = false;
Ptr<TimelineFilter> g_timelineFilter = new TimelineFilter;
extern Ptr<PasteOptions> g_pasteOptions;
extern Ptr<DeleteOptions> g_deleteOptions;
void render();
extern bool g_playSoundMovie;
extern bool g_playing;
extern int g_counter;

void playAnim();

AnimCopy g_animCopy;

bool outOfSyncWithRobot();//if the robot was syncronised with the 3d rendering we would not need this
void sAction(Rotation& v, int command);
void quantizeDec(float& v);
void quantizeInc(float& v);
bool stopDirectSound();
void removeFilename(String& str);

void toggleRadialTug();
void setShowDimFilterVisual();
void rectangleSelect();
void rectangleUnselect();

AnimCurve g_animCurve;

class AnimSliderFakeDrag: public CallBack {
public:
	AnimControlPanel* animCP;
	Anim* anim;
	double timeStart;
	double remDisplayFrom;
	AnimSliderFakeDrag():
		timeStart(0),
		remDisplayFrom(0)
	{}
void doCallBack(){
	AnimTimeLine* timeLine = animCP->timeLine;
	double f = 0;

	const double xPos = g_mousePos.x - animCP->pos.x - timeLine->dstPos.x - timeLine->lineOffset;
	double span = anim->zoom / anim->period;
	
	if(xPos >= timeLine->lineDist &&  anim->zoom < anim->period){
		anim->displayFrom = remDisplayFrom + 1.0 / anim->stepsPerSecond / anim->animZoom * (g_time - timeStart) * 8;
		if(anim->displayFrom >= 1.0 - span){
			anim->displayFrom = 1.0 - span;
		}
		f = anim->displayFrom + span;
	}
	else if(xPos < 0){
		anim->displayFrom = remDisplayFrom - 1.0 / anim->stepsPerSecond / anim->animZoom * (g_time - timeStart) * 8;
		if(anim->displayFrom < 0){
			anim->displayFrom = 0;
		}
		f = anim->displayFrom;
	}
	else{
		timeStart = g_time;
		remDisplayFrom = anim->displayFrom;
		f = xPos / timeLine->lineDist / anim->animZoom + timeLine->anim->displayFrom;
	}
	
	if( f < 0){
		f = 0;
	}
	if( f > 1){
		f = 1;
	}
	animCP->animSlider->bField = f;


	//g_vcArtAnimBase->houseKeep();

/*

	int diff=ptCursor.x-dragPos.x;

	const int beginPos = timeLine.dstPos.x-(bSize.x/2);
	const int endPos = beginPos + timeLine.bSize.x;
	const int relBottom = endPos - beginPos;
	const int relDstPos = dstPos.x - beginPos;


	if( diff < 0 && relDstPos + diff < 0){
		diff = -relDstPos;
	}
	if( diff > 0 && relDstPos + diff >= relBottom){
		diff = relBottom - relDstPos;
	}

	dstPos.x+=diff;
	dragPos.x+=diff;

	float a = static_cast<float>(dstPos.x+(bSize.x/2)-timeLine.dstPos.x);
	float c = a / timeLine.bSize.x;
//	dLog(c);
	bField = c;
//	dLog(bField.refValue());

//	bField = static_cast<float>(dragPos.y-top) / static_cast<float>(bottom-top);
//	g_scene->scroll = static_cast<float>(relDstPos + diff) / relBottom * g_scene->sceneTreeCollapsed.length;
	//win.resizePageUpPageDown();
	//win.keepLeagal();//why is this needed?
//	win.paint();
*/
	};
};

class AnimHighlightFakeDrag: public CallBack {
public:
	AnimControlPanel* animCP;
	Anim* anim;
	double timeStart;
	double remDisplayFrom;
	AnimHighlightFakeDrag():
		timeStart(0),
		remDisplayFrom(0)
	{}
void doCallBack(){
	AnimSelect* animSelect = animCP->animSelect;
	double f = 0;
	
	const double xPos = g_mousePos.x - animCP->pos.x - animSelect->dstPos.x - animSelect->lineOffset;
	double span = anim->zoom / anim->period;

	if(xPos >= animSelect->lineDist &&  anim->zoom < anim->period){
		anim->displayFrom = remDisplayFrom + 1.0 / anim->stepsPerSecond / anim->animZoom * (g_time - timeStart) * 8;
		if(anim->displayFrom >= 1.0 - span){
			anim->displayFrom = 1.0 - span;
		}
		f = anim->displayFrom + span;
	}
	else if(xPos < 0){
		anim->displayFrom = remDisplayFrom - 1.0 / anim->stepsPerSecond / anim->animZoom * (g_time - timeStart) * 8;
		if(anim->displayFrom < 0){
			anim->displayFrom = 0;
		}
		f = anim->displayFrom;
	}
	else{
		timeStart = g_time;
		remDisplayFrom = anim->displayFrom;
		f = xPos / animSelect->lineDist / anim->animZoom + animSelect->anim->displayFrom;
	}



	if( f < 0){
		f = 0;
	}
	if( f > 1){
		f = 1;
	}
	if(animSelect->dragRight){
		if(f < anim->selectFrom){
			animSelect->dragRight = false;
			anim->selectTo = anim->selectFrom;
			anim->selectFrom = f;
		}
		else{
			anim->selectTo = f;
		}
	}
	else{
		if(f > anim->selectTo){
			animSelect->dragRight = true;
			anim->selectFrom = anim->selectTo;
			anim->selectTo = f;
		}
		else{
			anim->selectFrom = f;
		}
	}
}
};

AnimSliderFakeDrag g_animSliderFakeDrag;
AnimHighlightFakeDrag g_animHighlightFakeDrag;

template <class A, class B>void bubbleSort(M_<A>& a, M_<B>& b, int& idx){
	bool flag;
	do{
		flag=false;
		for(int i=0;i<a.length-1;i++){

			if (a[i] > a[i+1]){
				if(idx==i){
					idx++;
				}
				else if(idx==i+1){
					--idx;
				}

				A swapA = a[i+1];
				a[i+1]=a[i];
				a[i]=swapA;

				B swapB = b[i+1];
				b[i+1]=b[i];
				b[i]=swapB;

				flag=true;
			}
		}
	}while(flag);
}



void bubbleSortMorpher(M_Float& a, M_Vec3f& b, int& idx, int numOfChannels){

	bool flag;
	do{
		flag = false;
		for(int i=0;i<a.length-1;i++){

			if (a[i] > a[i + 1]){
				if(idx == i){
					idx++;
				}
				else if(idx == i + 1){
					--idx;
				}

				float swapA = a[i + 1];
				a[i + 1] = a[i];
				a[i] = swapA;

				swapElements((i + 1) * numOfChannels, i * numOfChannels, numOfChannels, b, b);
				flag = true;
			}
		}
	}while(flag);
}

void OrientationInterpolator::setKeyDragPos(int& idx, const float newTime){
	key.refValue()[idx] = newTime;
	bubbleSort(key.refValue(), keyValue.refValue(), idx);
}

void PositionInterpolator::setKeyDragPos(int& idx, const float newTime){
	key.refValue()[idx] = newTime;
	bubbleSort(key.refValue(), keyValue.refValue(), idx);
}

void NurbsTrack::setKeyDragPos(int& idx, const float newTime){
	key.refValue()[idx] = newTime;
	bubbleSort(key.refValue(), keyValue.refValue(), idx);
}

void ScalarInterpolator::setKeyDragPos(int& idx, const float newTime){
	key.refValue()[idx] = newTime;
	bubbleSort(key.refValue(), keyValue.refValue(), idx);
}

void CameraSelector::setKeyDragPos(int& idx, const float newTime){
	key.refValue()[idx] = newTime;
	bubbleSort(key.refValue(), keyValue.refValue(), idx);
}

void CPMorpher::setKeyDragPos(int& idx, const float newTime){
	key.refValue()[idx] = newTime;
	bubbleSortMorpher(key.refValue(), keyValue.refValue(), idx, cpIndex.length);
}

void NikInterpolator::setKeyDragPos(int& idx, const float newTime){
	key.refValue()[idx] = newTime;
	bubbleSort(key.refValue(), keyValue.refValue(), idx);
}

void PosCompensator::setKeyDragPos(int& idx, const float newTime){
	if(rightSideDrag){
		moment.refValue()[idx].y = newTime;
	}
	else{
		moment.refValue()[idx].x = newTime;
	}
	//bubbleSort(key.refValue(), keyValue.refValue(), idx);
}


bool commonSearchKeyFrameInTimeLine(AnimTimeLine* timeLine, AnimSlider* animSlider, M_Float& key, const float clickPos, SamplePose* samplePose){
	Anim* anim = timeLine->anim;
	float error = 0.25 / (anim->stepsPerSecond * anim->period);
	for(int i = 0; i < key.length; i++){
		float a = key[i];
		if(abs(a - clickPos) < error){
			animSlider->bField = key[i];
			timeLine->samplePose = samplePose;
			timeLine->keyIndex = i;
			return true;
		}
	}
	return false;
}

bool CameraSelector::searchKeyFrameInTimeLine(AnimControlPanel& animCP, const float clickPos){
	return commonSearchKeyFrameInTimeLine(animCP.timeLine, animCP.animSlider, key, clickPos, this);
}

bool ScalarInterpolator::searchKeyFrameInTimeLine(AnimControlPanel& animCP, const float clickPos){
	return commonSearchKeyFrameInTimeLine(animCP.timeLine, animCP.animSlider, key, clickPos, this);
}

bool OrientationInterpolator::searchKeyFrameInTimeLine(AnimControlPanel& animCP, const float clickPos){
	return commonSearchKeyFrameInTimeLine(animCP.timeLine, animCP.animSlider, key, clickPos, this);
}

bool PositionInterpolator::searchKeyFrameInTimeLine(AnimControlPanel& animCP, const float clickPos){
	return commonSearchKeyFrameInTimeLine(animCP.timeLine, animCP.animSlider, key, clickPos, this);
}

bool NurbsTrack::searchKeyFrameInTimeLine(AnimControlPanel& animCP, const float clickPos){
	return commonSearchKeyFrameInTimeLine(animCP.timeLine, animCP.animSlider, key, clickPos, this);
}

bool CPMorpher::searchKeyFrameInTimeLine(AnimControlPanel& animCP, const float clickPos){
	return commonSearchKeyFrameInTimeLine(animCP.timeLine, animCP.animSlider, key, clickPos, this);
}

bool NikInterpolator::searchKeyFrameInTimeLine(AnimControlPanel& animCP, const float clickPos){
	return commonSearchKeyFrameInTimeLine(animCP.timeLine, animCP.animSlider, key, clickPos, this);
}


bool PosCompensator::searchKeyFrameInTimeLine(AnimControlPanel& animCP, const float clickPos){
	AnimTimeLine* timeLine = animCP.timeLine;
	AnimSlider* animSlider = animCP.animSlider;
	Anim* anim = timeLine->anim;
	rightSideDrag = false;

	for(int i = 0; i < moment.length; i++){
		float error = .25 / (anim->stepsPerSecond * anim->period);
		float midPoint = (moment[i].x + moment[i].y) * .5;
		if(clickPos >= moment[i].x - error && clickPos <= midPoint){
			animSlider->bField = moment[i].x;
			timeLine->samplePose = this;
			timeLine->keyIndex = i;
			return true;
		}
		if(clickPos >= midPoint && clickPos <= moment[i].y + error){
			animSlider->bField = moment[i].y;
			timeLine->samplePose = this;
			timeLine->keyIndex = i;
			rightSideDrag = true;
			return true;
		}
	}
	return false;
}

void commonPlotRecordNotch(AnimTimeLine* tl, Anim* anim, M_Float& key, unsigned int color){
	for(int i = 0; i < key.length; i++){
		int m = (key[i] - tl->anim->displayFrom) * tl->anim->animZoom * static_cast<float>(tl->lineDist);
		if(m < 0){
			continue;
		}
		if(m <= tl->lineDist){
			m += tl->dstPos.x + tl->lineOffset;
			drawVerticalLine2(color, m-2, tl->dstPos.y, tl->dstPos.y+tl->bSize.y, tl->win.size.x, tl->win.size.y, tl->win.pixBuf.getArray());
			drawVerticalLine2(color, m-1, tl->dstPos.y, tl->dstPos.y+tl->bSize.y, tl->win.size.x, tl->win.size.y, tl->win.pixBuf.getArray());
			drawVerticalLine2(color, m+1, tl->dstPos.y, tl->dstPos.y+tl->bSize.y, tl->win.size.x, tl->win.size.y, tl->win.pixBuf.getArray());
			drawVerticalLine2(color, m,   tl->dstPos.y, tl->dstPos.y+tl->bSize.y, tl->win.size.x, tl->win.size.y, tl->win.pixBuf.getArray());
		}
		else{
			break;
		}
	}
}

void cameraSelectorRecordNotch(AnimTimeLine* tl, Anim* anim, M_Float& key, M_Node& keyValue){
	for(int i = 0; i < key.length; i++){
		Viewpoint* view = dynamic_cast<Viewpoint*>(keyValue[i].p);
		unsigned int color = view->track ? 0xff0000 : 0x0000ff;

		int m = (key[i] - tl->anim->displayFrom) * tl->anim->animZoom * static_cast<float>(tl->lineDist);
		if(m < 0){
			continue;
		}
		if(m <= tl->lineDist){
			m += tl->dstPos.x + tl->lineOffset;
			drawVerticalLine2(color, m-2, tl->dstPos.y, tl->dstPos.y+tl->bSize.y, tl->win.size.x, tl->win.size.y, tl->win.pixBuf.getArray());
			drawVerticalLine2(color, m-1, tl->dstPos.y, tl->dstPos.y+tl->bSize.y, tl->win.size.x, tl->win.size.y, tl->win.pixBuf.getArray());
			drawVerticalLine2(color, m+1, tl->dstPos.y, tl->dstPos.y+tl->bSize.y, tl->win.size.x, tl->win.size.y, tl->win.pixBuf.getArray());
			drawVerticalLine2(color, m,   tl->dstPos.y, tl->dstPos.y+tl->bSize.y, tl->win.size.x, tl->win.size.y, tl->win.pixBuf.getArray());
		}
		else{
			break;
		}
	}
}

void OrientationInterpolator::plotRecordNotch(AnimTimeLine* tl, Anim* anim, Node* node, Field* field){
	if(Art* part = dynamic_cast<Art*>(node)){
		if(g_timelineFilter->rotation){
			if(&part->rotation == field || &part->ikRot == field){
				unsigned int color = part->track ? 0xff0000 : 0x0000ff;
				commonPlotRecordNotch(tl, anim, key, color);
			}
		}
	}
	else if(Viewpoint* view = dynamic_cast<Viewpoint*>(node)){
		if(g_timelineFilter->orientation){
			unsigned int color = view->track ? 0xff0000 : 0x0000ff;
			commonPlotRecordNotch(tl, anim, key, color);
		}
	}
}

void PositionInterpolator::plotRecordNotch(AnimTimeLine* tl, Anim* anim, Node* node, Field* field){
	if(Art* part = dynamic_cast<Art*>(node)){
		if(g_timelineFilter->translation){
			if(&part->translation == field || &part->ikPos == field){
				unsigned int color = part->track ? 0xff0000 : 0x0000ff;
				commonPlotRecordNotch(tl, anim, key, color);
			}
		}
	}
	else if(Viewpoint* view = dynamic_cast<Viewpoint*>(node)){
		if(g_timelineFilter->position){
			unsigned int color = view->track ? 0xff0000 : 0x0000ff;
			commonPlotRecordNotch(tl, anim, key, color);
		}
	}
}

void ScalarInterpolator::plotRecordNotch(AnimTimeLine* tl, Anim* anim, Node* node, Field* field){
	if(Art* part = dynamic_cast<Art*>(node)){
		if(
			g_timelineFilter->fka && &part->fka == field ||
			g_timelineFilter->fkb && &part->fkb == field
			){
			unsigned int color = part->track ? 0xff0000 : 0x0000ff;
			commonPlotRecordNotch(tl, anim, key, color);
		}
	}
	else if(Viewpoint* view = dynamic_cast<Viewpoint*>(node)){
		if(
			g_timelineFilter->zoom && &view->zoom == field ||
			g_timelineFilter->fieldOfView && &view->fieldOfView == field
			){
			unsigned int color = view->track ? 0xff0000 : 0x0000ff;
			commonPlotRecordNotch(tl, anim, key, color);
		}
	}
}

void CameraSelector::plotRecordNotch(AnimTimeLine* tl, Anim* anim, Node* node, Field* field){
	if(g_timelineFilter->live){
		cameraSelectorRecordNotch(tl, anim, key, (dynamic_cast<MFNode*>(field))->refValue());
	}
}

void NurbsTrack::plotRecordNotch(AnimTimeLine* tl, Anim* anim, Node* node, Field* field){
	if(Art* part = dynamic_cast<Art*>(node)){
		if(g_timelineFilter->translation){
			if(&part->translation == field || &part->ikPos == field){
				unsigned int color = part->track ? 0xff0000 : 0x0000ff;
				commonPlotRecordNotch(tl, anim, key, color);
			}
		}
	}
	else if(Viewpoint* view = dynamic_cast<Viewpoint*>(node)){
		if(g_timelineFilter->position){
			unsigned int color = view->track ? 0xff0000 : 0x0000ff;
			commonPlotRecordNotch(tl, anim, key, color);
		}
	}
}
void CPMorpher::plotRecordNotch(AnimTimeLine* tl, Anim* anim, Node* node, Field* field){
	if(Art* part = dynamic_cast<Art*>(node)){
		if(g_timelineFilter->morph){
			unsigned int color = part->track ? 0xff0000 : 0x0000ff;
			commonPlotRecordNotch(tl, anim, key, color);
		}
	}
}

void NikInterpolator::plotRecordNotch(AnimTimeLine* tl, Anim* anim, Node* node, Field* field){
	if(Art* part = dynamic_cast<Art*>(node)){
		unsigned int color = part->track ? 0xff0000 : 0x0000ff;
		commonPlotRecordNotch(tl, anim, key, color);
	}
}

void PosCompensator::plotRecordNotch(AnimTimeLine* tl, Anim* anim, Node* node, Field* field){
	for(int i = 0; i < moment.length; i++){
		int m = tl->dstPos.x + tl->lineOffset + (moment[i].x - tl->anim->displayFrom) * tl->anim->animZoom * static_cast<float>(tl->lineDist);
		int m2 = tl->dstPos.x + tl->lineOffset + (moment[i].y - tl->anim->displayFrom) * tl->anim->animZoom * static_cast<float>(tl->lineDist);
		drawHorizontalLine2(0x0, tl->dstPos.y + 2, m, m2, tl->win.size.x, tl->win.size.y, tl->win.pixBuf.getArray());
		drawHorizontalLine2(0x0, tl->dstPos.y + 3, m, m2, tl->win.size.x, tl->win.size.y, tl->win.pixBuf.getArray());
	}
}

void AnimSlider::paint(){
	setPixelPos();
	if(anim->play){
		animCP->playButton->srcPos.x = 432 + 24;
	}
	else{
		animCP->playButton->srcPos.x = 432;
	}
	if(anim->play == false){
		animCP->pauseButton->srcPos.x = 384 + 24;
	}
	else{
		animCP->pauseButton->srcPos.x = 384;
	}
	if(anim->loop){
		animCP->loopButton->srcPos.x = 384 - 48 + 24;
	}
	else{
		animCP->loopButton->srcPos.x = 384 - 48;
	}

	int srcPosX = srcPos.x;
	int srcPosY = srcPos.y;
	Node* node = g_selNode.getRaw();
	if(Art* part = dynamic_cast<Art*>(node)){
		if(part->track && part->smls->track){
			srcPosX = 3 * 24 + 2;
			srcPosY = 213;
		}
	}

	pasteSprite2(dstPos.x, dstPos.y, srcPosX, srcPosY, win.size.x, win.size.y, iconSrcBufCX, iconSrcBufCY, bSize.x, bSize.y, win.pixBuf.getArray());
	//paintFieldName();
}
void AnimSlider::setPixelPos(){
	AnimTimeLine* timeLine = animCP->timeLine;
	int i = (anim->slider.refValue() - anim->displayFrom) * timeLine->lineDist * anim->animZoom;
	dstPos.x = timeLine->dstPos.x + timeLine->lineOffset + i - (bSize.x / 2);
}
void AnimSlider::action(){
	removeTypeFromList < Button>(g_callBackList);
	removeTypeFromList < DownButton>(g_callBackList);

	g_animSliderFakeDrag.remDisplayFrom = anim->displayFrom;
	g_animSliderFakeDrag.timeStart = g_time;
	g_animSliderFakeDrag.anim = anim;
	g_animSliderFakeDrag.animCP = animCP;
	setDrag(this);
	g_vcArtAnimBase->showLever = false;
	g_callBackList += &g_animSliderFakeDrag;
	anim->selectTo = 0;
	anim->selectFrom = 0;

}

void AnimSlider::doDrag(const Vec2i ptCursor){
}

void updatePlayBuffer(String& filename);

void AnimControlPanel::updateVideo(){
	if(Video* video = dynamic_cast<Video*>(anim->video.p)){
		if(anim->play){
			int a = 0;
		}
		int i = anim->slider * anim->period * anim->stepsPerSecond + .5;
		if(i != video->frame){
			video->frame = i;
			if(g_vcp && g_vcp->isVisible()){
				g_vcp->paint();
//				if(anim->play){
//					updatePlayBuffer(video->pcmFilename);
//				}
			}
		}
	}
}

void AnimSlider::endDrag(const Vec2i ptCursor){
	removeTypeFromList<AnimSliderFakeDrag>(g_callBackList);
	g_drag = 0;
	releaseCapture();
	const Vec2i localCursor = ptCursor-win.pos;

	anim->quantizeSlider();


	dLog(String("new DragAnimSlider(") + String(this->anim->slider.getValue()) + "),");

	if(localCursor >= dstPos && localCursor < dstPos + bSize){
		open = true;//prevent tool tip opening
		g_callBackList += this;
	}

	g_vcArtAnimBase->resetRotationLever();
	g_vcArtAnimBase->showLever = true;
}
//void AnimTimeLine::doCallBack(){
//}
extern M_<Ptr<Viewpoint> > g_viewpoints;
extern int g_viewIndex;
void plotPixel2(unsigned color, int x, int y, int dstWidth, int dstHeight, int* pixBuf);

void Composition::plotRecordNotch(AnimControlPanel* animCP, Anim* anim){
	float* fbuf = this->fbuf;
	if(fbuf == 0){
		return;
	}
	AnimTimeLine* timeLine = animCP->timeLine;

	int dstWidth = animCP->size.x;
	int dstHeight = animCP->size.y;
	int* dstPix = animCP->pixBuf.getArray();
	drawHorizontalLine2(0xc0c0c0, 33, 0, dstWidth, dstWidth, dstHeight, dstPix);
	double pixPeriod = anim->zoom / timeLine->lineDist;
	double time = 0;

/*
	//int prevX = timeLine->lineOffset;
	//int prevY = 33 + (fbuf[0] * 40.0 + .5);
	for(int i = 0; i < timeLine->lineDist; i++){
		int x = i + timeLine->lineOffset;
		int idx = time * 44100;
		int y = 33 + (fbuf[idx] * 40.0 + .5);
		//drawVerticalLine2(0x0000ff, i, y, y - prevY, dstWidth, dstHeight, dstPix);
		//line(0x0000ff, prevX, prevY, x, y, dstWidth, dstHeight, dstPix);
		plotPixel2(0x0000ff, x, y, dstWidth, dstHeight, dstPix);
		//prevY = y;
		//prevX = x;
		time += pixPeriod;
	}
*/
	int from = anim->displayFrom * anim->period * 44100;
	float displayPeriod = anim->zoom;
	if(anim->period < displayPeriod){
		displayPeriod = anim->period;
	}
	for(int i = 0; i < 44100 * displayPeriod; i++){
		int y = 33 + (fbuf[from + i] * 30.0 + .5);
		double time = i / 44100.0;
		int x = time / pixPeriod + timeLine->lineOffset;
		plotPixel2(0x00ff00, x, y, dstWidth, dstHeight, dstPix);
	}
}
extern bool g_realTimeAudioSynth;


void AnimTimeLine::paint(){
	float displayTime = anim->period - (anim->displayFrom * anim->period);
	if(displayTime > anim->zoom){
		displayTime = anim->zoom;
	}

	int numOf = anim->stepsPerSecond * displayTime + .5;
	if(numOf == 0){
		numOf = 1;
	}

	if(anim->soundTrack.length){
		if(Composition* composition = dynamic_cast<Composition*>(anim->soundTrack.refValue()[0].p)){
			composition->plotRecordNotch(animCP, anim);
		}
	}
	float f  = (static_cast<float>(lineDist) / numOf);
	//double d  =  anim.displaySeconds / anim.period / numOf;
	//double step = 1.0 / anim.stepsPerSecond;
	double step = 1.0 / anim->stepsPerSecond / anim->period;

	double qz_displayFrom = quantize(anim->displayFrom, 1.0 / (anim->stepsPerSecond * anim->period));

	//if(anim.period < anim.zoom){
	//	numOf = anim.stepsPerSecond * anim.period + .5;
	//}

	double notchDiv = 1.0 / 5.0;
	if(anim->stepsPerSecond == 10){
		notchDiv = 1.0 / 2.0;
	}
	else if(anim->stepsPerSecond == 15){
		notchDiv = 1.0 / 3.0;
	}
	else if(anim->stepsPerSecond == 20 || anim->stepsPerSecond == 24){
		notchDiv = 1.0 / 4.0;
	}
	else if(anim->stepsPerSecond == 30 || anim->stepsPerSecond == 60){
		notchDiv = 1.0 / 6.0;
	}

	for(int i = 0; i < numOf + 1; i++){
		unsigned int colour = 0x808080;
		int endLine = dstPos.y + bSize.y;
		double a = (i * step + qz_displayFrom);
		double md = abs(fmod(a * anim->period, 1));
		double md2 = abs(fmod(a * anim->period, notchDiv));

		int m = (a - anim->displayFrom) * anim->animZoom * static_cast<float>(lineDist);
		m += dstPos.x + lineOffset;

		if(md2 < .0001 || notchDiv - md2 < .0001){
			endLine += 2;
			colour = 0;
		}
		if(md < .0001 || 1.0 - md < .0001){
			if(anim->zoom <= 32){
				int seconds = a * anim->period + .5;

				int minutes = seconds / 60;

				seconds -= minutes * 60;

				if(seconds <= 9){
					str = '0';
					str += seconds;
				}
				else{
					str = seconds;
				}

				drawText2(str, 0, Vec2i(m - 8, dstPos.y-4), win.size.x, win.size.y, win.pixBuf.getArray());
			}
		}
		drawVerticalLine2(colour, m,     dstPos.y, endLine, win.size.x, win.size.y, win.pixBuf.getArray());
		drawVerticalLine2(colour, m - 1, dstPos.y, endLine, win.size.x, win.size.y, win.pixBuf.getArray());
	}
	//if(Art* art=Window::returnOpenArt()){
	if(Art* part = dynamic_cast<Art*>(g_selNode.getRaw())){
		anim->plotRecordNotch(this, part);
	}
	else if(Viewpoint* viewpoint = dynamic_cast<Viewpoint*>(g_selNode.getRaw())){
		anim->plotRecordNotch(this, viewpoint);
	}

	if(g_realTimeAudioSynth == false){
		animCP->updateVideo();
	}

}

class DeferResetAnimLever : public CallBack{
	void doCallBack(){
		g_vcArtAnimBase->resetRotationLever();
		removeTypeFromList < DeferResetAnimLever>(g_callBackList);
		g_vcArtAnimBase->showLever = true;
	};
};

DeferResetAnimLever g_deferResetAnimLever;

void AnimTimeLine::action(){
	const Vec2i ptCursor = g_mousePos - win.pos;

	removeTypeFromList < Button>(g_callBackList);
	removeTypeFromList < DownButton>(g_callBackList);

	g_vcArtAnimBase->showLever = false;
	g_callBackList += &g_deferResetAnimLever;

	samplePose=0;
	g_animHighlightFakeDrag.remDisplayFrom = anim->displayFrom;
	g_animSliderFakeDrag.timeStart = g_time;

	const float xPos = ptCursor.x - dstPos.x - lineOffset;
	float f = xPos / lineDist / anim->animZoom + anim->displayFrom;

	if(Art* art = dynamic_cast<Art*>(g_selNode.getRaw())){
		if(anim->searchKeyFrameInTimeLine(*animCP, art, f)){
			//animSlider->setPixelPos();
			setDrag(this);
			return;
		}
	}
	f = quantize(f, 1.0 / (anim->stepsPerSecond * anim->period));
	anim->slider = f;
}

void AnimTimeLine::doDrag(const Vec2i ptCursor){
	const float xPos = ptCursor.x - dstPos.x - lineOffset;
	float f = xPos / lineDist / anim->animZoom + anim->displayFrom;

//	dragPos=g_mousePos;
	samplePose->setKeyDragPos(keyIndex, f);
	//win.paint();
}

void AnimTimeLine::endDrag(const Vec2i ptCursor){
	releaseCapture();
	g_drag = 0;
	if(samplePose){
		const float xPos = ptCursor.x - dstPos.x - lineOffset;
		float f = xPos / lineDist / anim->animZoom + anim->displayFrom;
		f = quantize(f, 1.0/(anim->stepsPerSecond * anim->period));
		samplePose->setKeyDragPos(keyIndex, f);
		//anim.slider=f;
	}
}

extern SingleButton* g_animStitchPolyModeButton;
extern SingleButton* g_animToggleBackfaceSelectButton;
extern SingleButton* g_animToggleAccumulateSelectButton;
extern SingleButton* g_togglePolyMorphCaptureButton;
extern SingleButton* g_animToggleZBufferCageButton;
extern SingleButton* g_animToggleHidePolysInHiddenPatchesButton;
extern SingleButton* g_animToggleSingleActiveButton;

void AnimControlPanel::initialize(){
	M_Button& m_button = *(this->m_button);
	g_togglePolyMorphCaptureButton =  new SingleButton(Vec2i(24,24), Vec2i(24 * 1, 208), "toggle morph capture", "togglePolyMorphCapture", togglePolyMorphCapture, *this);
	m_button += g_togglePolyMorphCaptureButton;

	m_button +=  new SingleButton(Vec2i(24,24), Vec2i(24 * 15, 233), "mirror morphed control points", " mirrorMorphedCtrlPoints",  mirrorMorphedCtrlPoints, *this);



	playButton = new SingleButton(Vec2i(24,24), Vec2i(432,176), "play/stop", "playAnim", playAnim, *this);
	m_button += playButton;

	m_button += new SingleButton(Vec2i(24,24), Vec2i(24 * 17, 281), "move selected control point to nearest other (M)", "moveVertex", moveVertex, *this);
	m_button += new SingleButton(Vec2i(24,24), Vec2i(24 * 4, 257), "show all cages (Alt A)", "showAllCages", showAllCages, *this);
	m_button += new SingleButton(Vec2i(24,24), Vec2i(24 * 5, 257), "show selected cages (Alt S)", "showSingleCage", showSingleCage, *this);
	m_button += new SingleButton(Vec2i(24,24), Vec2i(24 * 6, 257), "hide selected cages (Alt H)", "hideSelectedCages", hideSelectedCages, *this);
	m_button += new SingleButton(Vec2i(24,24), Vec2i(24 * 15, 257), "show selected groups (Alt S)", "showSingleCage", showSelectedGroups, *this);
	m_button += new SingleButton(Vec2i(24,24), Vec2i(24 * 14, 257), "hide selected groups", "hideSelectedCages", hideSelectedGroups, *this);
	g_animToggleHidePolysInHiddenPatchesButton = new SingleButton(Vec2i(24,24), Vec2i(24 * 11, 233 + 24), "toggle hide polys in hidden patches", "toggleHidePolysInHiddenPatches", toggleHidePolysInHiddenPatches, *this);
	m_button += g_animToggleHidePolysInHiddenPatchesButton;


	g_animToggleZBufferCageButton = new SingleButton(Vec2i(24,24), Vec2i(24, 305 - 24), "toggle zbuffer control cage mode", "toggleZBufferCageMode", toggleZBufferCageMode, *this);
	m_button += g_animToggleZBufferCageButton;


	g_animToggleSingleActiveButton = new SingleButton(Vec2i(24,24), Vec2i(3 * 24, 233), "unactivate/activate selected points (A)", "setSingleActiveMode", setSingleActiveMode, *this);
	m_button += g_animToggleSingleActiveButton;
	m_button += new SingleButton(Vec2i(24,24), Vec2i(24*7, 209), "open selected control points transform panel (Ctrl T)", "openTransformCP", openTransformCP, *this);

	m_button += new SingleButton(Vec2i(24,24), Vec2i(24*5, 209), "select whole patches mode", "selectWholePatch", selectWholePatch, *this);
	m_button += new SingleButton(Vec2i(24,24), Vec2i(24 * 0, 257), "select ring (R)", "selectRing", selectRing, *this);
	m_button += new SingleButton(Vec2i(24,24), Vec2i(16,16), Vec2i(336,144), "radiul tug", "", toggleRadialTug, *this);

	//m_button += g_animSelectRingButton;
	//m_button += g_animSelectRingsButton;

	g_animToggleBackfaceSelectButton = new SingleButton(Vec2i(24,24), Vec2i(24 * 5, 0), "toggle backface ignore", "toggleBackfaceSelect", toggleBackfaceSelect, *this);
	m_button += g_animToggleBackfaceSelectButton;

	g_animToggleAccumulateSelectButton = new SingleButton(Vec2i(24,24), Vec2i(24 * 7, 24 * 3), "toggle accumulate selection", "toggleAccumulateSelect", toggleAccumulateSelect, *this);
	m_button += g_animToggleAccumulateSelectButton;
	

	m_button += new SingleButton(Vec2i(24,24), Vec2i(24 * 13, 257), "dimension tug filter (Alt Space)", "dimFilter", setShowDimFilterVisual, *this);
	m_button += new SingleButton(Vec2i(24,24), Vec2i(11 * 24, 281), "control point rectangle select", "rectangleSelect", rectangleSelect, *this);
	m_button += new SingleButton(Vec2i(24,24), Vec2i(24 * 16, 257), "control point rectangle unselect", "rectangleUnselect", rectangleUnselect, *this);


	int barY = BORDER_DIST;
	for(int k = 0, h = TOOLBAR_BUTTON_OFFSET_X; k < m_button.length; k++){
		if(k == 0){
			h += 24 * 9;
		}
		if(k == 2){
			h += 24 * 1;
		}
		if(k == 3){
			h += 24 * 20;
		}
		Button* button = m_button[k];
		button->dstPos = Vec2i(h, barY);
		h += button->bSize.x;
	}
}

void AnimControlPanel::setupNodeWindow(Node* node){
	M_Button& m_button = *(this->m_button);

	Window::closeWindow(g_tb2.p);
	g_scene->mode.refValue() &= ~XRAY_EDIT_MODE;
	this->node = node;
	anim = dynamic_cast<Anim*>(node);

	const M_Properties& properties = node->getProperties();

	g_controlPanelOpen = true;
	g_propWinLine.length = 0;
	buttonList.length = 0;
	int yPos = title_bar_height + 1;
	int lastYPos = yPos;

	size = Vec2i(g_3d->size.x, 52);
	pos = Vec2i(g_tb2->pos.x, g_tb2->pos.y - (52 - 25));

	AnimSlider* animSlider = new AnimSlider(Vec2i(20,16), Vec2i(80, 26), Vec2i(316,176), 0, this, this, anim, anim->slider, properties[0], Vec2i(415-24-20, 0), Vec2i(8*8+4, 16));
	this->animSlider = animSlider;
	buttonList += animSlider;
	g_propWinLine += PropWinLine(yPos, animSlider);

	AnimTimeLine* timeLine=new AnimTimeLine(Vec2i(size.x,6), Vec2i(0, 42), 0, this, this, anim, anim->timeLine, properties[1]);
	buttonList += timeLine;
	this->timeLine = timeLine;
	g_propWinLine += PropWinLine(yPos, timeLine);

	animSelect = new AnimSelect(Vec2i(size.x, 17), Vec2i(0, 42 - 17), 0, this, this, anim, anim->animSelect, properties[2]);
	buttonList += animSelect;
	g_propWinLine += PropWinLine(yPos, animSelect);

	//////
	//playButton = new AnimControlPanel::Play(Vec2i(24,24), Vec2i(340-24-20, 0), Vec2i(432,176), 0, this, this, anim, anim->play, properties[3]);

	//g_propWinLine += PropWinLine(yPos, playButton);

	pauseButton = new AnimControlPanel::Pause(Vec2i(24,24), Vec2i(340-20, 0), Vec2i(384,176), 0, this, this, anim, anim->pause, properties[4]);
	buttonList += pauseButton;
	g_propWinLine += PropWinLine(yPos, pauseButton);

	loopButton = dynamic_cast<SFBoolButton*>(anim->loop.setUpControl(*node, anim->loop, properties[5], 0, 340+24, 0, *this));
	loopButton->bSize=Vec2i(24, 24);
	loopButton->dstPos=Vec2i(340+24-20, 0);
	loopButton->srcPos=Vec2i(384-48,176);
	loopButton->iconSize=Vec2i(24, 24);
	buttonList+=loopButton;
	g_propWinLine+=PropWinLine(yPos, loopButton);

	//FieldButton* button = anim->zoom.setUpControl(*node, anim->speed, properties[6], 0, 544+78-24-4-24, 1, *this);
	zoom = new Zoom(Vec2i(24,24), Vec2i(544+78-24-4-24, 1), Vec2i(292,191), 0, this, this, anim, anim->zoom, properties[6]);
	zoom->bSize = Vec2i(24, 24);
	zoom->srcPos = Vec2i(24*3,24);
	zoom->iconSize = Vec2i(24, 24);
	buttonList+=zoom;
	g_propWinLine+=PropWinLine(yPos, zoom);

	animPeriod = new AnimPeriod(Vec2i(24,24), Vec2i(544+78-24-4, 1), Vec2i(292,191), 0, this, this, anim, anim->period, properties[7]);
	animPeriod->iconSize=Vec2i(24, 18);
	buttonList+=animPeriod;
	g_propWinLine+=PropWinLine(yPos, animPeriod);

	FieldButton* button = anim->stepsPerSecond.setUpControl(*node, anim->stepsPerSecond, properties[8], 0, 544+78, 1, *this);
	button->bSize=Vec2i(24, 24);
	button->srcPos=Vec2i(268, 191);
	button->iconSize=Vec2i(24, 18);
	button->valPos.x=button->dstPos.x+24+4;
	button->valSize.x=8*7;
	buttonList+=button;
	g_propWinLine+=PropWinLine(yPos, button);

	button = anim->speed.setUpControl(*node, anim->speed, properties[9], 0, 544+78*2, 1, *this);
	button->bSize=Vec2i(24, 24);
	button->srcPos=Vec2i(24*14, 24);
	button->iconSize=Vec2i(24, 24);
	button->valPos.x=button->dstPos.x+24+4;
	button->valSize.x=8*7;
	buttonList+=button;
	g_propWinLine+=PropWinLine(yPos, button);

//	button = new AnimPlayFrom(*node, anim->speed, properties[10], 0, 400+48+24*1-24-10, 1, *this);
	animPlayFrom = new AnimPlayFrom(Vec2i(24,24), Vec2i(400+48+24*1-24-10, 1), Vec2i(292,191), 0, this, this, anim, anim->playFrom, properties[10]);
	animPlayFrom->bSize=Vec2i(24, 24);
	animPlayFrom->srcPos=Vec2i(96, 209);
	animPlayFrom->iconSize=Vec2i(24, 24);
	buttonList+=animPlayFrom;
	g_propWinLine+=PropWinLine(yPos, animPlayFrom);

	button=anim->copy.setUpControl(*node, anim->copy, properties[11], 0, 400+48+24*1, 0, *this);
	button->dstPos=Vec2i(410+48+24*1, 0);
	button->srcPos=Vec2i(176,176);
	button->bSize=Vec2i(24,24);
	button->iconSize=Vec2i(16,16);
	buttonList+=button;
	g_propWinLine+=PropWinLine(yPos, button);

	button = anim->paste.setUpControl(*node, anim->paste, properties[12], 0, 400+48+24*2, 0, *this);
	button->dstPos = Vec2i(410+48+24*2, 0);
	button->srcPos = Vec2i(16, 192);
	button->bSize = Vec2i(24,24);
	button->iconSize = Vec2i(16,16);
	buttonList += button;
	g_propWinLine += PropWinLine(yPos, button);

	button = anim->remove.setUpControl(*node, anim->remove, properties[13], 0, 400+48+24*3, 0, *this);
	button->dstPos=Vec2i(410+48+24*3, 0);
	button->srcPos=Vec2i(delete_x, delete_y);
	button->bSize=Vec2i(24,24);
	button->iconSize=Vec2i(16,16);
	buttonList+=button;
	g_propWinLine+=PropWinLine(yPos, button);

//	paint();
	houseKeepVisualClue();
}
void AnimControlPanel::setWinSize(Vec2i& raw){
	pixBuf.length = raw.x * raw.y;
	if(timeLine){
		timeLine->bSize = Vec2i(Vec2i(size.x,8));
		timeLine->lineDist = timeLine->bSize.x-timeLine->lineOffset*2;

	}
	if(false == initWindow()){
		throw;
	}
}

void AnimPeriod::action(){
	if(outOfSyncWithRobot()){
		return;
	}
	//double s = anim->period / (anim->period + 1.0);

	float v = anim->period;
	anim->scaleTime(1.0); //add second

	anim->slider = anim->slider * (v / (v + 1.0));
	v++;

	bField = v;
	
	anim->quantizeSlider();
}

void Zoom::action(){
	if(outOfSyncWithRobot()){
		return;
	}
	double& v = anim->zoom.refValue();

	switch(command){
		case 0:
			if(v < 8){
				v++;
			}
			else{
				v = v * 2;
			}
			break;
		case 1:
			if(v <= 8){
				v--;
			}
			else{
				v = v * .5;
			}
			if(v < 1){
				v = 1;
			}
			break;
	}
	anim->zoom = v;
}

void AnimPlayFrom::action(){
	if(outOfSyncWithRobot()){
		return;
	}
	if(anim->slider < .99){
		anim->playFrom = anim->slider;
	}
}

void ArtRotationButton::action(){
	if(outOfSyncWithRobot()){
		return;
	}
	Rotation& v=bField.refValue();
	sAction(v, command);
	bField = v;
	if(Anim* anim=Window::animMode()){
		if(anim->play==false && Anim::autoCapture){
			anim->captureSinglePartPose(&part, &part.rotation);
		}
	}
}

void ArtTranslationButton::action(){
	if(outOfSyncWithRobot()){
		return;
	}
	Vec3f& v = bField.refValue();
	switch(command){
		case 0:
			quantizeDec(v.x);
			break;
		case 1:
			quantizeInc(v.x);
			break;
		case 2:
			quantizeDec(v.y);
			break;
		case 3:
			quantizeInc(v.y);
			break;
		case 4:
			quantizeDec(v.z);
			break;
		case 5:
			quantizeInc(v.z);
			break;
	}
	bField = v;
	if(Anim* anim = Window::animMode()){
		if(anim->play == false && Anim::autoCapture){
			anim->captureSinglePartPose(&part, &part.translation);
		}
	}
}

void ArtIKRotButton::action(){
	if(outOfSyncWithRobot()){
		return;
	}
	Rotation& v=bField.refValue();
	sAction(v, command);
	bField = v;
	if(Anim* anim=Window::animMode()){
		if(anim->play==false && Anim::autoCapture){
			anim->captureSinglePartPose(&part, &part.ikRot);
		}
	}
}

void ArtIKPosButton::action(){
	if(outOfSyncWithRobot()){
		return;
	}
	Vec3f& v = bField.refValue();
	switch(command){
		case 0:
			quantizeDec(v.x);
			break;
		case 1:
			quantizeInc(v.x);
			break;
		case 2:
			quantizeDec(v.y);
			break;
		case 3:
			quantizeInc(v.y);
			break;
		case 4:
			quantizeDec(v.z);
			break;
		case 5:
			quantizeInc(v.z);
			break;
	}
	bField = v;
	if(Anim* anim = Window::animMode()){
		if(anim->play == false && Anim::autoCapture){
			anim->captureSinglePartPose(&part, &part.ikPos);
		}
	}
}

void ArtFkaButton::action(){
	if(outOfSyncWithRobot()){
		return;
	}
	float& v = bField.refValue();

	switch(command){
		case 0:
			quantizeDec(v);
			if(v < 0){
				v = 0;
			}
			break;
		case 1:
			quantizeInc(v);
			if(v > 1){
				v = 1;
			}
			break;
	}
	bField = v;
	if(Anim* anim = Window::animMode()){
		if(anim->play == false && Anim::autoCapture){
			anim->captureSinglePartPose(&part, &part.fka);
		}
	}
}

void ArtFkbButton::action(){
	if(outOfSyncWithRobot()){
		return;
	}
	float& v = bField.refValue();

	switch(command){
		case 0:
			quantizeDec(v);
			if(v < 0){
				v = 0;
			}
			break;
		case 1:
			quantizeInc(v);
			if(v > 1){
				v = 1;
			}
			break;
	}
	bField = v;
	if(Anim* anim = Window::animMode()){
		if(anim->play == false && Anim::autoCapture){
			anim->captureSinglePartPose(&part, &part.fkb);
		}
	}
}

void ArtCoordPointButton::action(){
	if(outOfSyncWithRobot()){
		return;
	}
	float res = getIncDecRes();

	if(part.smls->workAnimCP && (Window::animMode() || g_colorOwnership)){

		M_ControlPoint& controlPoint = part.controlPoint;
		if(controlPoint.length == 0){
			return;
		}
		if(MultiSelect::selectedPoint >= controlPoint.length){
			MultiSelect::selectedPoint = 0;
		}
		ControlPoint& cp = part.controlPoint[MultiSelect::selectedPoint];
		NurbsPatch* patch = cp.p;
		patch->index[0] = cp.i;

		M_Vec3f t = patch->dpoint.refValue();
		Vec3f& e = t[cp.i];

		switch(command){
			case 0:
				quantizeDec(e.x);
				break;
			case 1:
				quantizeInc(e.x);
				break;
			case 2:
				quantizeDec(e.y);
				break;
			case 3:
				quantizeInc(e.y);
				break;
			case 4:
				quantizeDec(e.z);
				break;
			case 5:
				quantizeInc(e.z);
				break;
		}
		patch->dpoint = t;
		updateBuild3dWindow(true);
		return;
	}


	if(bField.length==0){
		return;
	}
	M_Vec3f t = bField;
	Vec3f& e = t[node.getFieldGroupIndex(sprop.groupIndex)-sprop.startAtIndex];

	switch(command){
		case 0:
			quantizeDec(e.x);
			break;
		case 1:
			quantizeInc(e.x);
			break;
		case 2:
			quantizeDec(e.y);
			break;
		case 3:
			quantizeInc(e.y);
			break;
		case 4:
			quantizeDec(e.z);
			break;
		case 5:
			quantizeInc(e.z);
			break;
	}
	bField = t;
	updateBuild3dWindow();
}

bool ArtRotationButton::updateValFromTextCell(String textCellStr, const int &index, const Properties& properties){
	int i = 0;
	Rotation value; textCellStr.toValue(value);
	if(notCloseEnough(bField,value)){
		bField = value;
		if(Anim* anim=Window::animMode()){
			if(anim->play == false && Anim::autoCapture){
				anim->captureSinglePartPose(&part, &part.rotation);
			}
		}
		return true;
	}
	return false;
}

bool ArtTranslationButton::updateValFromTextCell(String textCellStr, const int &index, const Properties& properties){
	int i = 0;
	Vec3f value; textCellStr.toValue(value);
	if(notCloseEnough(bField,value)){
		bField = value;
		if(Anim* anim=Window::animMode()){
			if(anim->play == false && Anim::autoCapture){
				anim->captureSinglePartPose(&part, &part.translation);
			}
		}
		return true;
	}
	return false;
}

bool ArtIKRotButton::updateValFromTextCell(String textCellStr, const int &index, const Properties& properties){
	int i = 0;
	Rotation value; textCellStr.toValue(value);
	if(notCloseEnough(bField,value)){
		bField = value;
		if(Anim* anim=Window::animMode()){
			if(anim->play == false && Anim::autoCapture){
				anim->captureSinglePartPose(&part, &part.ikRot);
			}
		}
		return true;
	}
	return false;
}

bool ArtIKPosButton::updateValFromTextCell(String textCellStr, const int &index, const Properties& properties){
	int i = 0;
	Vec3f value; textCellStr.toValue(value);
	if(notCloseEnough(bField,value)){
		bField = value;
		if(Anim* anim=Window::animMode()){
			if(anim->play == false && Anim::autoCapture){
				anim->captureSinglePartPose(&part, &part.ikPos);
			}
		}
		return true;
	}
	return false;
}

bool ArtFkaButton::updateValFromTextCell(String textCellStr, const int &index, const Properties& properties){
	int i = 0;
	float value; textCellStr.toValue(value);
	if(notCloseEnough(bField,value)){
		bField = value;
		if(Anim* anim=Window::animMode()){
			if(anim->play == false && Anim::autoCapture){
				anim->captureSinglePartPose(&part, &part.fka);
			}
		}
		return true;
	}
	return false;
}

bool ArtFkbButton::updateValFromTextCell(String textCellStr, const int &index, const Properties& properties){
	int i = 0;
	float value; textCellStr.toValue(value);
	if(notCloseEnough(bField,value)){
		bField = value;
		if(Anim* anim=Window::animMode()){
			if(anim->play == false && Anim::autoCapture){
				anim->captureSinglePartPose(&part, &part.fkb);
			}
		}
		return true;
	}
	return false;
}

bool ArtCoordPointButton::updateValFromTextCell(String textCellStr, const int &index, const Properties& properties){
	if(!bField.length)return false;
	Vec3f sval; textCellStr.toValue(sval);
	if(notCloseEnough(bField.refValue()[index],sval)){
		M_<Vec3f> m = bField;
		m[index] = sval;
		bField = m;
		return true;
	}
	return false;
}

bool initDirectSound();
bool freeDirectSound();

extern int g_sampleRate;

int addFrameLengthTotal(){
	int length = 0;
	for(int k = 0; k < g_scene->animBarList.length; k++){
		Anim* anim = g_scene->animBarList[k];
		if(Video* video = dynamic_cast<Video*>(anim->video.p)){
			length += static_cast<int>(anim->period * g_sampleRate + .5);
		}
	}
	return length;
}

void addAudioFile(short int* srcBuf, int srcBufOffset, int srcBufLen, int delay, float* dstBuf, int workBufLen, const M_Vec2f& lPoint, const M_Vec2f& rPoint){
	if(delay + srcBufLen > workBufLen){
		srcBufLen = workBufLen - delay;
	}
	int d = delay * g_numOfChannels;
	int len = srcBufLen * g_numOfChannels;
	int offset = srcBufOffset * g_numOfChannels;

	int envIdx = 0;

	if(lPoint.length){
		float lBeginY = lPoint[envIdx].y;
		if(lPoint.length == 1){
			for(int k = 0; k < len; k++){
				dstBuf[d + k] += lBeginY * srcBuf[offset + k];
				k++;
				dstBuf[d + k] += lBeginY * srcBuf[offset + k];
			}
		}
		else{
			float lTimeDiff = lPoint[envIdx + 1].x - lPoint[envIdx + 0].x;
			float lAmpDiff = lPoint[envIdx + 1].y - lPoint[envIdx + 0].y;


			//float rTimeDiff = rPoint[1].x - rPoint[0].x;
			//float rAmpDiff = rPoint[1].y - rPoint[0].y;
			//float rBeginY = rPoint[0].y;

			for(int k = 0; k < len; k++){
				float time = (k * .5) * (1.0f / g_sampleRate);
				if(time >= lPoint[envIdx + 1].x){
					envIdx++;
					if(envIdx + 1 >= lPoint.length){
						break;
					}

					lTimeDiff = lPoint[envIdx + 1].x - lPoint[envIdx + 0].x;
					lAmpDiff = lPoint[envIdx + 1].y - lPoint[envIdx + 0].y;
					lBeginY = lPoint[envIdx + 0].y;
				}
				float lAmp = lBeginY + (time - lPoint[envIdx].x) / lTimeDiff * lAmpDiff;
				//float rAmp = rBeginY + time / rTimeDiff * rAmpDiff;
				if(lAmp > 1){
					int a = 1;
				}

				dstBuf[d + k] += lAmp * srcBuf[offset + k];
				k++;
				dstBuf[d + k] += lAmp * srcBuf[offset + k];
			}
		}
	}
	else{
		for(int k = 0; k < len; k++){
			dstBuf[d + k] += srcBuf[offset + k];
		}
	}
}
struct AudioStep{
	AudioStep():
		buf(0),
		length(0),
		sampleOffset(0)
	{}
	short int* buf;
	int length;
	int sampleOffset;
};






const int WAVE_HEADER_SIZE = 44;
unsigned char formatS16[WAVE_HEADER_SIZE] = {0X52,0X49,0X46,0X46,0Xce,0Xec,0X5a,0X00,0X57,0X41,0X56,0X45,0X66,0X6d,0X74,0X20,0X10,0X00,0X00,0X00,0X01,0X00,0X02,0X00,0X80,0Xbb,0X00,0X00,0X00,0Xee,0X02,0X00,0X04,0X00,0X10,0X00,0X64,0X61,0X74,0X61,0X40,0Xec,0X5a,0X00};

void saveToWav(short int* iBuf, int workBufLen, int numOfChannels, int sampleRate, const String& filename){
	int sampleSize = 2 * numOfChannels;
	WriteWavHeader* wavHeadePtr = reinterpret_cast<WriteWavHeader*>(&formatS16);
	WriteWavHeader& wavHeader = *wavHeadePtr;
	wavHeader.chunkID[0] = 'R';
	wavHeader.chunkID[1] = 'I';
	wavHeader.chunkID[2] = 'F';
	wavHeader.chunkID[3] = 'F';

	unsigned long TOTAL_FILE_SIZE = workBufLen * sampleSize + sizeof(WriteWavHeader);
	wavHeader.chunkSize = TOTAL_FILE_SIZE - 8;

	wavHeader.format[0] = 'W';
	wavHeader.format[1] = 'A';
	wavHeader.format[2] = 'V';
	wavHeader.format[3] = 'E';

	wavHeader.subchunk1ID[0] = 'f';
	wavHeader.subchunk1ID[1] = 'm';
	wavHeader.subchunk1ID[2] = 't';
	wavHeader.subchunk1ID[3] = ' ';

	wavHeader.subchunk1Size = 16;

	wavHeader.audioFormat = 1; //PCM
	wavHeader.numChannels = numOfChannels;
	wavHeader.sampleRate = sampleRate;
	wavHeader.bitsPerSample = 16;
	wavHeader.byteRate = sampleRate * numOfChannels * wavHeader.bitsPerSample / 8;

	wavHeader.blockAlign = numOfChannels * wavHeader.bitsPerSample / 8;

	wavHeader.data[0] = 'd';
	wavHeader.data[1] = 'a';
	wavHeader.data[2] = 't';
	wavHeader.data[3] = 'a';

	wavHeader.sampleLen = workBufLen * sampleSize; //convert to bytes

	if(FILE* stream = fopen(filename.c_str(), "w+b")){

		fwrite(&wavHeader, 1, sizeof(WriteWavHeader), stream);
		fwrite(iBuf, sampleSize, workBufLen, stream);

		fclose(stream);
	}
}

int loadWavFile(const String& srcFilename, short int*& buf, WavHeader& wavHeader, int sampleSize, bool convertToFloat){ //return bufLen
	
	int sampleLen = 0; 

	FILE* readStream = 0;

	if(readStream = fopen(srcFilename.c_str(), "rb")){
		fread(&wavHeader, 1, sizeof(WavHeader), readStream);
	}
	else{
		return 0;
	}
	char* chunkID = wavHeader.chunkID;
	int fileSize = wavHeader.chunkSize + 8;

	if(wavHeader.audioFormat != 1){//exit if not PCM
		return 0;
	}
	if(wavHeader.sampleRate != g_sampleRate){
		dLog("***warning wavHeader.sampleRate != g_sampleRate");
	}
	if(wavHeader.bitsPerSample != 16){
		return 0;
	}
	if(wavHeader.numChannels != g_numOfChannels){
		dLog("***warning wavHeader.numChannels != g_numOfChannels");
	}

	int subchunk1Size = wavHeader.subchunk1Size;

	short int extraParamSize = 0;
	char subchunk2ID[4];
	int subchunk2Size;

	sampleLen = wavHeader.chunkSize - (sizeof(WavHeader) - 8);

	if(subchunk1Size > 16){
		fread(&extraParamSize, 1, sizeof(short int), readStream);
		sampleLen -= sizeof(short int);
	}
	if(extraParamSize){
		fread(buf, 1, extraParamSize, readStream);
		sampleLen -= extraParamSize;
	}

	fread(subchunk2ID, 1, 4, readStream);
	sampleLen -= 4;
	fread(&subchunk2Size, 1, 4, readStream);
	sampleLen -= 4;

	sampleLen = sampleLen / sampleSize;
	buf = new short int[sampleLen * g_numOfChannels];
	//buf = new short int[2000000];

	if(	subchunk2ID[0] == 'L' &&
		subchunk2ID[1] == 'I' &&
		subchunk2ID[2] == 'S' &&
		subchunk2ID[3] == 'T'){

		fread(buf, 1, 34, readStream);

		sampleLen -= 34 / sampleSize;
	}

	int item = sampleLen * wavHeader.numChannels;
	if(convertToFloat){
		item = item * 2;
	}

	fread(buf, sizeof(short int), item, readStream);

	fclose(readStream);


	if(convertToFloat){
		float* floatBuf = reinterpret_cast<float*>(buf);

		for(int i = sampleLen; --i >= 0;){
			floatBuf[i] = buf[i] * (1.0 / 0x7fff);
		}
	}


	return sampleLen;
}

void convertWorkBufToInt(float* workBuf, short int* iBuf, int workBufLen){
	int len = workBufLen * g_numOfChannels;

	for(int i = 0; i < len; i++){
		iBuf[i] = workBuf[i];
	}
}

bool saveToPcms(short int* iBuf, int sampleSize){
	String cntrStr;
	String filename;
	int samplesPerFrame = g_sampleRate / g_fps;

	for(int k = 0; k < g_scene->animBarList.length; k++){
		Anim* anim = g_scene->animBarList[k];
		if(Video* video = dynamic_cast<Video*>(anim->video.p)){
			int numOfFrames = static_cast<int>(anim->period * g_fps + .5);
			for(int frame = 0; frame < numOfFrames; frame++){
				cntrStr = frame;
				while(cntrStr.length < 5){
					cntrStr = '0' + cntrStr;
				}
				filename = video->path;
				filename += cntrStr;
				filename += ".pcm";
				if(FILE* writeStream = fopen(filename.c_str(), "w+b")){
					fwrite(iBuf, sampleSize, samplesPerFrame, writeStream);
					iBuf += samplesPerFrame * g_numOfChannels;
					fclose(writeStream);
				}
				else{
					return false;
				}
			}
		}
	}
	return true;
}


void createAudioTrack2(short int*& iBuf, int& workBufLen, int sampleSize){
	WavHeader wavHeader;

	workBufLen = addFrameLengthTotal();
	float* workBuf = new float[workBufLen * g_numOfChannels];
	for(int i = workBufLen * g_numOfChannels; --i >= 0;){
		workBuf[i] = 0.0;
	}

	int sampleIndex = 0;
	for(int k = 0; k < g_scene->animBarList.length; k++){
		Anim* anim = g_scene->animBarList[k];
		if(Video* video = dynamic_cast<Video*>(anim->video.p)){
			int vidSampleLen = static_cast<int>(anim->period * g_sampleRate + .5);
			for(int j = 0; j < anim->soundTrack.length; j++){
				if(Audio* audio = dynamic_cast<Audio*>(anim->soundTrack.refValue()[j].p)){
					AudioStep audioStep;
					audioStep.sampleOffset = sampleIndex;
					audioStep.length = loadWavFile(audio->file, audioStep.buf, wavHeader, sampleSize, false);
					int sampleDelay = sampleIndex + audio->delay * g_sampleRate;
					int srcLen = audioStep.length;
					int srcBufOffset = 0;
					if(audio->videoSync && srcLen > vidSampleLen){
						srcLen = vidSampleLen;
						if(audio->videoSync && video->frameRef.length){
							int firstFrame = video->frameRef[0];
							srcBufOffset = g_sampleRate / anim->stepsPerSecond * firstFrame;
						}
					}
					addAudioFile(audioStep.buf, srcBufOffset, srcLen, sampleDelay, workBuf, workBufLen, audio->lPoint, audio->rPoint);
				}
			}
			sampleIndex += vidSampleLen;
		}
	}
	iBuf = reinterpret_cast<short int*>(workBuf);
	convertWorkBufToInt(workBuf, iBuf, workBufLen);
}

void createAudioTrack(){
	int sampleSize = 4; //a sample = 2 short ints for stero
	short int* iBuf = 0;
	int workBufLen = 0;

	createAudioTrack2(iBuf, workBufLen, sampleSize);
	//saveToWav(iBuf, workBufLen, 2, "outFile.wav");
	saveToPcms(iBuf, sampleSize);
}


void outMovieFrames(bool outputSoundTrack, bool outputFrames){

	int frame = 0;

	int sampleSize = 4; //a sample = 2 short ints for stero
	short int* iBuf = 0;
	int workBufLen = 0;

	if(outputSoundTrack){
		String dstPath = g_currentFilename;
		removeFilename(dstPath);
		createAudioTrack2(iBuf, workBufLen, sampleSize);
		dstPath += "\\output\\soundtrack.wav";
		saveToWav(iBuf, workBufLen, 2, g_sampleRate, dstPath);
	}


	if(outputFrames){
		for(int k = 0; k < g_scene->animBarList.length; k++){
			Anim* anim = g_scene->animBarList[k];
			if(Video* video = dynamic_cast<Video*>(anim->video.p)){
				int numOfFrames = static_cast<int>(anim->period * 25 + .5);
				video->outputFrames(frame, numOfFrames);
				frame += numOfFrames;
			}
		}
	}

}


void playAnim(){
	Anim* anim = g_acp->anim;

	if(anim->period < .000001){
		return;
	}
	

	anim->play = !anim->play;

//	if(&anim.play==&bField){
//		int a =0;
//	}
	Video* video = dynamic_cast<Video*>(anim->video.p);

	if(anim->play == false){
		g_callBackList += &g_deferResetAnimLever;
		g_acp->animSlider->setPixelPos();
		if(video){
			freeDirectSound();
			g_playSoundMovie = false;
			g_playing = false;
			video->audioTimeMode = false;
		}
	}
	else{
		g_vcArtAnimBase->showLever = false;
		if(video){
			createAudioTrack();

			g_counter = (anim->playFrom * anim->period / (1.0 / anim->stepsPerSecond) + .5) / (25 / g_fps);

			initDirectSound();
			g_playSoundMovie = true;
			g_playing = true;
			video->audioTimeMode = true;
		}
	}
	houseKeepVisualClue();
}

void AnimControlPanel::Pause::action(){

	if(anim->period < .000001){
		return;
	}
	
	anim->pause = !anim->pause.refValue();

	Video* video = dynamic_cast<Video*>(anim->video.p);

	if(anim->play == false){
		g_callBackList += &g_deferResetAnimLever;
		animCP->animSlider->setPixelPos();
		if(video){
			if(stopDirectSound()){
				g_playSoundMovie = false;
				g_playing = false;
				Video::audioTimeMode = false;
			}
		}
	}
	else{
		g_vcArtAnimBase->showLever = false;
		
		if(video){
			g_counter = video->frame / (25 / g_fps);
			initDirectSound();
			g_playSoundMovie = true;
			g_playing = true;
			video->audioTimeMode = true;
		}
	}
	houseKeepVisualClue();
}

void ArtIKButton::action(){
	bField=!bField.refValue();
	if(bField){
		if(Anim* anim = Window::animMode()){
			Art* part = dynamic_cast<Art*>(g_selNode.getRaw());
			part->ikPos.refValue() = part->absolutePivotPoint;
		}
	}
	removeTypeFromList < Button>(g_callBackList);
	removeTypeFromList < DownButton>(g_callBackList);
	g_stw->cp->setupNodeWindow(g_selNode.getRaw());
	Window::checkIfMouseIsHoveringAWindowAndUpdate(g_mousePos);
}

void AnimSelect::paint(){
	if(anim->selectFrom == anim->selectTo){
		return;
	}
	int numOf = anim->stepsPerSecond * anim->period;
	if(numOf == 0){
		numOf = 1;
	}
	int m = dstPos.x + lineOffset + (anim->selectFrom - anim->displayFrom) * anim->animZoom * static_cast<float>(lineDist);
	int m2 = dstPos.x + lineOffset + (anim->selectTo - anim->displayFrom) * anim->animZoom * static_cast<float>(lineDist);
	paintBlock2(0xFFD060, m, dstPos.y, m2, dstPos.y + bSize.y, win.size.x, win.size.y, win.pixBuf.getArray());
}


void AnimSelect::action(){
	removeTypeFromList<Button>(g_callBackList);
	removeTypeFromList<DownButton>(g_callBackList);

	g_animHighlightFakeDrag.anim = anim;
	g_animHighlightFakeDrag.animCP = animCP;
	setDrag(this);
	g_callBackList += &g_animHighlightFakeDrag;

	const Vec2i ptCursor = g_mousePos;
	const double xPos = g_mousePos.x-win.pos.x - dstPos.x - lineOffset;
	double f = xPos / lineDist / anim->animZoom + anim->displayFrom;
	if( f < 0){
		f = 0;
	}
	if( f > 1){
		f = 1;
	}
	anim->selectFrom = anim->selectTo = f;
}

void AnimSelect::doDrag(const Vec2i ptCursor){
}

void AnimSelect::endDrag(const Vec2i ptCursor){
	removeTypeFromList<AnimHighlightFakeDrag>(g_callBackList);
	g_drag = 0;
	releaseCapture();

	const Vec2i localCursor = ptCursor-win.pos;

	int frames = anim->stepsPerSecond * anim->period + .5;
	anim->selectTo = quantize(anim->selectTo, 1.0 / frames);
	anim->selectFrom = quantize(anim->selectFrom, 1.0 / frames);

	if(localCursor >= dstPos && localCursor < dstPos + bSize){
		open = true;//prevent tool tip opening
		g_callBackList += this;
	}
}

void PasteOptions::Ok::operator=(bool const &){
	//if(home.addRelative){
	//	Window::closeWindow(Window::activeWindow.getActiveWindow());
	//	g_stw->cp->setupNodeWindow(g_addRelative.p);
	//	Window::activeWindow=g_stw->cp.p;
	//}
	Window::closeWindow(Window::activeWindow.getActiveWindow());
	//if(dynamic_cast<Anim*>(g_selNode.getRaw()) == 0){ //only wanted if a part window was previously open, better for now not to have untill extra code added
		//g_stw->cp->setupNodeWindow(g_selNode.getRaw()); 
		//Window::activeWindow = g_stw->cp.p;
	//}
	Window::animMode()->pasteSection();
}

void PasteOptions::Filter::operator=(bool const &){
	Window::closeWindow(Window::activeWindow.getActiveWindow());
	g_timelineFilter->okMode = TimelineFilter::back_to_pasteOptions;
	g_stw->cp->setupNodeWindow(g_timelineFilter);
	Window::activeWindow = g_stw->cp.p;
}

void DeleteOptions::Filter::operator=(bool const &){
	Window::closeWindow(Window::activeWindow.getActiveWindow());
	g_timelineFilter->okMode = TimelineFilter::back_to_deleteOptions;
	g_stw->cp->setupNodeWindow(g_timelineFilter);
	Window::activeWindow = g_stw->cp.p;
}


//void AddRelative::Ok::operator=(bool const &){
//	Window::closeWindow(Window::activeWindow.getActiveWindow());
//	g_stw->cp->setupNodeWindow(g_selNode.getRaw());
//	Window::activeWindow=g_stw->cp.p;
//	Window::animMode()->pasteSection(g_selNode.getRaw());
//}

void openAnimEditOptions(){
	g_timelineFilter->okMode = TimelineFilter::back_to_selected_node;
	g_stw->cp->setupNodeWindow(g_timelineFilter);
	Window::activeWindow = g_stw->cp.p;
}

void TimelineFilter::Ok::operator=(bool const &){
	Window::closeWindow(Window::activeWindow.getActiveWindow());
	switch(g_timelineFilter->okMode){
		case TimelineFilter::back_to_deleteOptions:{
			g_stw->cp->setupNodeWindow(g_deleteOptions);
			Window::activeWindow = g_stw->cp.p;
			break;
		}
		case TimelineFilter::back_to_pasteOptions:{
			g_stw->cp->setupNodeWindow(g_pasteOptions);
			Window::activeWindow = g_stw->cp.p;
			break;
		}
		case TimelineFilter::back_to_selected_node:{
			if(dynamic_cast<Anim*>(g_selNode.getRaw()) == 0){
				g_stw->cp->setupNodeWindow(g_selNode.getRaw());
				Window::activeWindow = g_stw->cp.p;
			}
			break;
		}
	}
}

void DeleteOptions::Ok::operator=(bool const &){
	Window::closeWindow(Window::activeWindow.getActiveWindow());
	if(dynamic_cast<Anim*>(g_selNode.getRaw()) == 0){
		g_stw->cp->setupNodeWindow(g_selNode.getRaw());
		Window::activeWindow = g_stw->cp.p;
	}
	Window::animMode()->deleteSection();
}

void doCurveDistance(bool state){
	g_pasteOptions->curveDistance = 0;
	g_animCurve.processed = false;
	if(state){
		if(NurbsTrack* nurbsTrack = dynamic_cast<NurbsTrack*>(g_selNode.getRaw())){
			g_animCurve.processFields(nurbsTrack, (g_animCopy.selectTo - g_animCopy.selectFrom) * g_animCopy.period);
		}
		else if(NurbsTrack* nurbsTrack = dynamic_cast<NurbsTrack*>(g_copyNode.getRaw())){
			g_animCurve.processFields(nurbsTrack, (g_animCopy.selectTo - g_animCopy.selectFrom) * g_animCopy.period);
		}
	}
}

void doCalcRot(bool state){
	if(Art* selPart = dynamic_cast<Art*>(g_selNode.getRaw())){
		if(state){
			if(selPart->ik){
				g_pasteOptions->rotate = g_animCopy.rotDiff * selPart->ikRot;
				g_pasteOptions->translate = selPart->ikPos - g_pasteOptions->rotate.refValue() * g_animCopy.tranDiff;
			}
			else{
				g_pasteOptions->rotate = g_animCopy.rotDiff * selPart->rotation;
				g_pasteOptions->translate = selPart->translation - g_pasteOptions->rotate.refValue() * g_animCopy.tranDiff;
			}
		}
		else{
			g_pasteOptions->rotate = Rotation();
			if(selPart->ik){
				g_pasteOptions->translate = selPart->ikPos - g_animCopy.tranDiff;
			}
			else{
				g_pasteOptions->translate = selPart->translation - g_animCopy.tranDiff;
			}
		}
	}
}

void PasteOptions::CalcRot::operator=(bool const& par){
	refValue() = par;
	doCalcRot(par);
}

void PasteOptions::Curve::operator=(bool const& par){
	refValue() = par;
	doCurveDistance(par);
}

void Art::scaleSkeleton(Vec3f scale){
	pivotPoint.refValue() = pivotPoint * scale;
	int len = part.length;
	for(int k = 0; k < len; k++){
		part.refValue()[k]->scaleSkeleton(scale);
	}
}

void Seamless::scaleSkeleton(Vec3f pivotPoint, Vec3f scale){
	int len = skeleton.length;
	for(int k = 0; k < len; k++){
		Art* part = skeleton.refValue()[k];
		part->pivotPoint.refValue() = part->pivotPoint - pivotPoint;
		part->scaleSkeleton(scale);
	}
	calcAbsolutePivotPoint();
}

void Anim::PastePose::operator=(const bool& par){
	g_pasteOptions->pivotPoint = Vec3f();
	g_pasteOptions->scale = Vec3f(1, 1, 1);
	g_pasteOptions->rotate = Rotation();
	g_pasteOptions->translate = Vec3f();

	doCalcRot(g_pasteOptions->calcRot);
	doCurveDistance(g_pasteOptions->curve);

	g_pasteOptions->steps = (g_animCopy.selectTo - g_animCopy.selectFrom) * g_animCopy.period / (1.0 / home.stepsPerSecond) + .5;
	g_stw->cp->setupNodeWindow(g_pasteOptions);
	Window::activeWindow = g_stw->cp.p;
		//}
		//else{
		//	float currentSlider = home.slider;
		//	home.slider.refValue() = home.copySlider;
		//	home.update = true;
		//	home.renderNodeFirst();
		//	Seamless* smls = part->smls;
		//	if(smls->animCP && smls->workAnimCP){
		//		smls->doBuild();
		//	}
		//	else{
		//		render();
		//	}
		//	home.slider.refValue() = currentSlider;
		//	home.captureMultiplePartPose(part);
		//	home.update = true;
		//}
}

void Anim::DeletePose::operator=(const bool& par){
	home.copySlider = home.slider;
	g_stw->cp->setupNodeWindow(g_deleteOptions);
	Window::activeWindow = g_stw->cp.p;

	//if(home.selectFrom != home.selectTo){
	//	if(Art* part = dynamic_cast<Art*>(g_selNode.getRaw())){
	//		g_stw->cp->setupNodeWindow(g_deleteSection);
	//	}
	//}
	//else if(MemberFunction* function=dynamic_cast<MemberFunction*>(home.onFrame.p)){

	//	M_Node& m_codeLine=function->codeLine;

	//	M_Word& m_word=function->m_word;
	//	W_Field* w_dstField = 0;
	//	W_Field* w_srcField = 0;
	//	W_Node* w_morpher = 0;
	//	W_Node* w_part = 0;

	//	for(int i=m_codeLine.length; --i>=0;){
	//		if(getEquiveRoute(m_word, w_dstField, w_srcField, m_codeLine, i)){
	//			if(Art* part=dynamic_cast<Art*>(w_dstField->node)){
	//				if(SamplePose* samplePose=dynamic_cast<SamplePose*>(w_srcField->node)){
	//					if((controlKeyState()) && part != dynamic_cast<Art*>(g_selNode.getRaw())){
	//						continue;
	//					}
	//					if((altKeyState())){
	//						continue;
	//					}
	//					samplePose->deleteSample(home.slider);
	//				}
	//			}
	//		}
	//		else if(getMorphFunction(m_word, w_morpher, w_part, m_codeLine, i)){
	//			if(CPMorpher* morpher = dynamic_cast<CPMorpher*>(w_morpher->node)){
	//				if(Art* part = dynamic_cast<Art*>(w_part->node)){
	//					if((controlKeyState()) && part != dynamic_cast<Art*>(g_selNode.getRaw())){
	//						continue;
	//					}
	//					if((shiftKeyState())){
	//						continue;
	//					}
	//					morpher->deleteSample(home.slider);
	//				}
	//			}
	//		}
	//	}
	//	home.update=true;
	//}
}

void Anim::CopyPose::operator=(const bool& par){
	home.copySlider = home.slider;
	if((home.selectFrom - home.selectTo) == 0){
		home.selectFrom = home.selectTo = home.slider;
	}
	home.copySection();
}

void paintTrack(int x, int y, int dstWidth, int dstHeight, int* dstPix);

void AnimControlPanel::paint(){

	anim->animZoom = anim->period / anim->zoom;
	if(anim->play){
		if(anim->slider < anim->displayFrom){
			anim->displayFrom = anim->slider;
		}
		if(anim->slider > anim->displayFrom + anim->zoom / anim->period){
			anim->displayFrom = anim->slider - anim->zoom / anim->period;
		}
	}
	//else if(g_vcArtAnimBase->showLever == false){
	//	if(isInList<CallBack>(g_callBackList, &g_deferResetAnimLever)==false){
	//		g_callBackList += &g_deferResetAnimLever;
	//	}
	//}

	prePaint();

	M_Button& m_button = *(this->m_button);

	for(int k = m_button.length; --k>=0;){
		Button* button = m_button[k];
		button->paint();
	}


	if(g_settings->invertWindowColor){
		invert();
	}
	updateShaderTexture = true;
	if(textCell.cellActive==true){
		textCell.doNoThinkWink();
	}
}

int String::hourMinSecFormatToValue(){
	int i = 0;
	int seconds = 0;
	int minutes = 0;
	int hours = 0;

	stb(i);
	if(i < length){
		hours = getInt(*this, i);
		stb(i);
		if(i < length){
			if(i < length && (*this)[i]==':'){
				i++;
			}
			stb(i);
			if(i < length){
				minutes = getInt(*this, i);
				stb(i);
				if(i < length && (*this)[i]==':'){
					i++;
				}
				stb(i);
				if(i < length){
					seconds = getInt(*this, i);
				}
			}
		}
	}

	return hours * 60 * 60 + minutes * 60 + seconds;
}

bool AnimSlider::updateValFromTextCell(String textCellStr, const int &index, const Properties& properties){
	float seconds = textCellStr.hourMinSecFormatToValue();
	if(seconds != bField * anim->period){
		if(seconds > anim->period){
			seconds = anim->period;
		}
		else if(seconds < 0){
			seconds = 0;
		}
		anim->slider = seconds / anim->period;

		if(anim->slider < anim->displayFrom){
			anim->displayFrom = anim->slider;
		}
		if(anim->slider > anim->displayFrom + anim->zoom / anim->period){
			anim->displayFrom = anim->slider - anim->zoom / anim->period;
		}
		g_callBackList += &g_deferResetAnimLever;
		return true;
	}
	return false;
}

void AnimSlider::toString(String& str)const{
	//bField.refValue() = (60 * 60 * 05 + 60 * 07 + 21) / anim.period;
	if(anim->period == 0){
		int a = 0;
	}
	int seconds = bField.refValue() * anim->period;
	int minutes = seconds / 60;
	int hours = minutes / 60;

	seconds -= minutes * 60;
	minutes -= hours * 60;

	if(hours <= 9){
		str += '0';
	}
	str += hours;
	str += ':';

	if(minutes <= 9){
		str += '0';
	}
	str += minutes;
	str += ':';

	if(seconds <= 9){
		str += '0';
	}
	str += seconds;
}


void AnimPeriod::paint(){
	int* dstPix = win.pixBuf.getArray();

	const int subDx=bSize.x;
	const int subDy=bSize.y;
	const int yLn=dstPos.y;

		
	int offset=0;
	int xLn=dstPos.x;

	int strx = sprop.name.length * 8 - 8;

	if(this==getHoverButton()){
		pasteRectangle2( activeColor, dstPos.x, yLn, win.size.x, win.size.y, subDx, subDy, dstPix);
		drawHorizontalLine2( 0x808080, yLn+subDy-1, xLn+1, xLn+subDx, win.size.x, win.size.y, dstPix);
		drawVerticalLine2( 0x808080, xLn+subDx, yLn+1, yLn+subDy, win.size.x, win.size.y, dstPix);
	}
	if(this==DownButton::getDownButton()){
		pasteRectangle2( activeColor, dstPos.x, yLn, win.size.x, win.size.y, subDx, subDy, dstPix);
		offset++;
		drawHorizontalLine2( 0x808080, yLn, xLn, xLn+subDx, win.size.x, win.size.y, dstPix);
		drawVerticalLine2( 0x808080, xLn, yLn+1, yLn+subDy-2, win.size.x, win.size.y, dstPix);
	}
	int dx = dstPos.x + ((bSize.x - iconSize.x) >> 1) + offset + 4;
	int dy = dstPos.y + ((bSize.y - iconSize.y) >> 1) + offset;

	pasteSprite2(dstPos.x+((bSize.x-iconSize.x)>>1)+offset, dstPos.y+((bSize.y-iconSize.y)>>1)+offset, srcPos.x, srcPos.y, win.size.x, win.size.y, iconSrcBufCX, iconSrcBufCY, iconSize.x, iconSize.y, win.pixBuf.getArray());
}

void AnimPlayFrom::paint(){
	int* dstPix = win.pixBuf.getArray();

	const int subDx=bSize.x;
	const int subDy=bSize.y;
	const int yLn=dstPos.y;

		
	int offset=0;
	int xLn=dstPos.x;

	int strx = sprop.name.length * 8 - 8;

	if(this==getHoverButton()){
		pasteRectangle2( activeColor, dstPos.x, yLn, win.size.x, win.size.y, subDx, subDy, dstPix);
		drawHorizontalLine2( 0x808080, yLn+subDy-1, xLn+1, xLn+subDx, win.size.x, win.size.y, dstPix);
		drawVerticalLine2( 0x808080, xLn+subDx, yLn+1, yLn+subDy, win.size.x, win.size.y, dstPix);
	}
	if(this==DownButton::getDownButton()){
		pasteRectangle2( activeColor, dstPos.x, yLn, win.size.x, win.size.y, subDx, subDy, dstPix);
		offset++;
		drawHorizontalLine2( 0x808080, yLn, xLn, xLn+subDx, win.size.x, win.size.y, dstPix);
		drawVerticalLine2( 0x808080, xLn, yLn+1, yLn+subDy-2, win.size.x, win.size.y, dstPix);
	}
	int dx = dstPos.x + ((bSize.x - iconSize.x) >> 1) + offset + 4;
	int dy = dstPos.y + ((bSize.y - iconSize.y) >> 1) + offset;

	pasteSprite2(dstPos.x+((bSize.x-iconSize.x)>>1)+offset, dstPos.y+((bSize.y-iconSize.y)>>1)+offset, srcPos.x, srcPos.y, win.size.x, win.size.y, iconSrcBufCX, iconSrcBufCY, iconSize.x, iconSize.y, win.pixBuf.getArray());
}

void AnimPeriod::workOutSubIndex(const Vec2i ptCursor){}
void AnimPlayFrom::workOutSubIndex(const Vec2i ptCursor){}


void TimelineFilter::UnTickAllFields::operator=(bool const&){
	home.morph = false;
	home.rotation = false;
	home.translation = false;
	home.nik = false;
	home.posCompensation = false;
	home.fka = false;
	home.fkb = false;
	home.live = false;
	home.zoom = false;
	home.fieldOfView = false;
	home.orientation = false;
	home.lookAt = false;
	home.up = false;
	home.position = false;
}
void TimelineFilter::TickPartFieldsOnly::operator=(bool const&){
	home.morph = true;
	home.rotation = true;
	home.translation = true;
	home.nik = true;
	home.posCompensation = true;
	home.fka = true;
	home.fkb = true;
	home.live = false;
	home.zoom = false;
	home.fieldOfView = false;
	home.orientation = false;
	home.lookAt = false;
	home.up = false;
	home.position = false;
}
void TimelineFilter::TickViewpointFieldsOnly::operator=(bool const&){
	home.morph = false;
	home.rotation = false;
	home.translation = false;
	home.nik = false;
	home.posCompensation = false;
	home.fka = false;
	home.fkb = false;
	home.live = true;
	home.zoom = true;
	home.fieldOfView = true;
	home.orientation = true;
	home.lookAt = true;
	home.up = true;
	home.position = true;
}
void TimelineFilter::SelectAllFields::operator=(bool const&){
	home.morph = true;
	home.rotation = true;
	home.translation = true;
	home.nik = true;
	home.posCompensation = true;
	home.fka = true;
	home.fkb = true;
	home.live = true;
	home.zoom = true;
	home.fieldOfView = true;
	home.lookAt = true;
	home.up = true;
	home.orientation = true;
	home.position = true;
}

/*
void AnimCurve::genCurve(){
	
	line.length=0;
		float step = 1.0 / numOfSteps;
		float t = step;
		
		//line+=Vec3f(point[p].x, point[p].y, point[p].z);
		//pnt.length = nopfs[s];

		for (int k = 1; k < numOfSteps; k++){
		//	line+=genCurve3(numOfSteps, p, s, t, pnt);
			t += step;
		}
	//line+=point[point.length-1];
*/
void CurvePoint (int dim, int degree, const float *knot, const Vec3f *pnt, const float *w, const float t, Vec3f& outPoint);

void NurbsTrack::genCurve(M_Vec3f& line, const int steps){

	if(!keyValue.length){
		return;
	}
	const float inverseSize=(1 / g_scene->getZoom());
	const Vec3f* p = keyValue.getArray();
	const float* w = weight.getArray();;
	
	line.length = 0;

	int lineOffset = 0;
	for(int seg = 0; seg < dim.length; seg++){
		const int dm = dim[seg];

		int const knotOffset = knotRedirect[seg];
		float mint = minT(knotOffset);
		int const endOfKnotSeg = seg < (knotRedirect.length - 1) ? knotRedirect[seg + 1] : knot.length;
		float maxt = maxT(endOfKnotSeg);

		if(dm == 2){
			line.length += 2;
			Vec3f* linePtr = line.getArray();
			linePtr[lineOffset] = p[0];
			linePtr[lineOffset + 1] = p[1];
		}
		else{
			line.length += steps;
			Vec3f* linePtr = line.getArray();
			const float* knotPtr = knot.getArray() + knotOffset;


			for(int i = 0; i < steps; i++){
				float t = static_cast<float>(i) / (steps + 1);
				t = (maxt - mint) * t + mint;
				CurvePoint(dm - 1, degree, knotPtr, p, w, t, *(linePtr + lineOffset + i));
			}
		}
		lineOffset = line.length;
		p += dm - 1;
	}

}

void AnimCurve::genFlatHeight(){
	flatHeight.length = line.length - 1;
	float h = 0;
	for(int i = 0; i < flatHeight.length; i++){
		float xDiff = line[i+1].x - line[i].x;
		float yDiff = line[i+1].y - line[i].y;
		float zDiff = line[i+1].z - line[i].z;
		h += sqrt(xDiff * xDiff + yDiff * yDiff + zDiff * zDiff);
		flatHeight[i] = h;
	}
}

void AnimCurve::genHeightIndex(){
	heightIndex.length=flatHeight.length;

	float bezArcLen=flatHeight[flatHeight.length-1];
	float step=bezArcLen/flatHeight.length;
	for(int i=0, j=0; i<flatHeight.length; i++){
		float h = step*(i+1);
		while(j < line.length - 2 && h > flatHeight[j]){
			j++;
		}
		heightIndex[i]=j;
	}
}

void AnimCurve::generate(float t, int& idx, float& ratio){
	float bezArcLen = flatHeight[flatHeight.length-1];

	float timeDistanceRatio = selectedPeriod / bezArcLen;

	t = t / timeDistanceRatio;
	idx = t / bezArcLen * flatHeight.length;

	if(idx >= heightIndex.length){
		idx = heightIndex.length-1;
	}
	else if(idx < 0){
		idx = 0;
	}
	idx = heightIndex[idx];
		
	while(idx > 0 && t < flatHeight[idx-1]){
		idx--;
	}
	float flatDiff = flatHeight[idx];
	float coordDiff = t;

	if(idx > 0){
		flatDiff -= flatHeight[idx-1];
		coordDiff -= flatHeight[idx-1];
	}
	ratio = coordDiff / flatDiff;
}

void AnimCurve::generate(Vec3f& coord, float t){
	t = t * g_pasteOptions->newPeriod;

	int idx = 0;
	float ratio = 0;

	generate(t, idx, ratio);
	Vec3f s_crd = line[idx] + (line[idx + 1] - line[idx]) * ratio;
	float bezArcLen = flatHeight[flatHeight.length-1];
	float timeDistanceRatio = selectedPeriod / bezArcLen;
	float flatDistrance = t / timeDistanceRatio;



	Vec3f o = line[idx+1] - line[idx];
	o.y = 0;
	Rotation rot = Rotation(Vec3f(0, 0, 1), o);

	Vec3f v = rot * (coord - Vec3f(0, 0, flatDistrance));

	coord = v + s_crd - firstPoint;
}

void AnimCurve::generate(Rotation& rotation, float t){
	t = t * g_pasteOptions->newPeriod;

	int idx = 0;
	float ratio = 0;

	generate(t, idx, ratio);

	Vec3f o = line[idx+1] - line[idx];
	o.y = 0;
	Rotation rot = Rotation(Vec3f(0, 0, 1), o);
	rotation = rotation *  rot;
}

void AnimCurve::processFields(NurbsTrack* nurbsTrack, float selectedPeriod){
	this->selectedPeriod = selectedPeriod;
	flatHeight.length=0;
	heightIndex.length=0;
	line.length=0;

	M_Vec3f& point = nurbsTrack->keyValue.refValue();

	if(point.length==0){
		return;
	}

	firstPoint = point[0];

	nurbsTrack->genCurve(line, numOfSteps);
	genFlatHeight();
	genHeightIndex();

	g_pasteOptions->curveDistance = flatHeight[flatHeight.length-1];
	processed = true;
}

void transformInterpolators2(float selectFrom, OrientationInterpolator* orientationInterpolator, PositionInterpolator* positionInterpolator){
	float rem = orientationInterpolator->set_fraction;
	orientationInterpolator->set_fraction = selectFrom;
	g_animCopy.rotDiff = -orientationInterpolator->value_changed;
	orientationInterpolator->set_fraction = rem;

	rem = positionInterpolator->set_fraction;
	positionInterpolator->set_fraction = selectFrom;
	g_animCopy.tranDiff = positionInterpolator->value_changed;
	positionInterpolator->set_fraction = rem;
}

void Anim::transformInterpolators(const Art* part){
	if(part->ik){
		if(OrientationInterpolator* orientationInterpolator = getOrientationInterpolator(part, &part->ikRot)){
			if(PositionInterpolator* positionInterpolator = getPositionInterpolator(part, &part->ikPos)){
				transformInterpolators2(selectFrom, orientationInterpolator, positionInterpolator);
			}
		}
	}
	else{
		if(OrientationInterpolator* orientationInterpolator = getOrientationInterpolator(part, &part->rotation)){
			if(PositionInterpolator* positionInterpolator = getPositionInterpolator(part, &part->translation)){
				transformInterpolators2(selectFrom, orientationInterpolator, positionInterpolator);
			}
		}
	}
}

void setPasteDefault(){
	g_pasteOptions->insert = false;
	g_pasteOptions->beginEdge = true;
	g_pasteOptions->endEdge = true;
	g_pasteOptions->quantize = true;
	g_pasteOptions->transform = false;
	g_pasteOptions->calcRot = false;
	g_pasteOptions->mirror = false;
	g_pasteOptions->curve = false,

	g_deleteOptions->time = false;
	g_deleteOptions->beginEdge = true;
	g_deleteOptions->endEdge = true;

	g_timelineFilter->morph = true;
	g_timelineFilter->rotation = true;
	g_timelineFilter->translation = true;
	g_timelineFilter->nik = true;
	g_timelineFilter->posCompensation = true;
	g_timelineFilter->fka = true;
	g_timelineFilter->fkb = true;
	g_timelineFilter->zoom = true;
	g_timelineFilter->fieldOfView = true;
	g_timelineFilter->orientation = true;
	g_timelineFilter->position = true;
}
/*
bool AnimControlPanel::hover(Vec2i ptCursor){
	M_Button& m_button = *(this->m_button);


	for(int k = m_button.length; --k>=0;){
		Button* button = m_button[k];
		if(button->initHover(ptCursor - pos)){
			return true;
		}
	}

	return ControlPanel::hover(ptCursor);
}
*/
void AnimControlPanel::buttonHover(Vec2i ptCursor){
	M_Button& m_button = *(this->m_button);

	for(int k = m_button.length; --k>=0;){
		Button* button = m_button[k];
		if(button->initHover(ptCursor)){
			return;
		}
	}

	for(int k = buttonList.length; --k>=0;){
		Button* button = buttonList[k];
		if(button && button->initHover(ptCursor)){
			return;
		}
	}
	if(exitWindow.initHover(ptCursor)){
		return;
	}
}
