/* Seamless3d

Copyright (C) 2001-2017 Graham John Perrett (AKA thyme) thyme@seamless3d.com www.seamless3d.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this
software and associated documentation files (the "Software"), to deal in the Software
without restriction, including without limitation the rights to use, copy, modify,
merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to the following
conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT
OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.
*/

#include "seamless3d.h"

void Art::findSetVertexRange(TextureMapper& textureMapper){
    float negX=0;
    float posX=0;
    float negY=0;
    float posY=0;

    Matrix matrix=rotationMatrix(-textureMapper.orientation);
    for(int i=triangle.length*3;	--i>=0;){
        Vec3f crd=matrix*triangle.refPart(i)->coord[triangle.refIndex(i)];
        if(crd.x<negX){
            negX=crd.x;
        }
        else if(crd.x>posX){
            posX=crd.x;
        }
        if(crd.y<negY){
            negY=crd.y;
        }
        else if(crd.y>posY){
            posY=crd.y;
        }
    }
    textureMapper.position3d=Vec3f(negX,negY,0);
    float xDiff=posX-negX;
    float yDiff=posY-negY;
    textureMapper.size3d = Vec2f(xDiff, yDiff);
}

void Art::mirrorFuncX(){

    if(coord.length==0)return;
    M_Vec3f& coord=this->coord.refValue();
    M_Vec3f& normal=this->normal.refValue();

    for(int i=coord.length;--i>=0;){
        coord[i].x=-coord[i].x;
    }
    if(normal.length){
        for(int i=normal.length;--i>=0;){
            normal[i].x=-normal[i].x;
        }
    }
    insideOut();
}
void Art::mirrorFuncY(){
    if(coord.length==0)return;
    M_Vec3f& coord=this->coord.refValue();
    M_Vec3f& normal=this->normal.refValue();

    for(int i=coord.length;--i>=0;){
        coord[i].y=-coord[i].y;
    }
    if(normal.length){
        for(int i=normal.length;--i>=0;){
            normal[i].y=-normal[i].y;
        }
    }
    insideOut();
}
void Art::mirrorFuncX(int triangleOffest, int coordOffest){
    if(coord.length==0)return;
    M_Vec3f& coord=this->coord.refValue();
    M_Vec3f& normal=this->normal.refValue();

    for(int i=coord.length;--i>=coordOffest;){
        coord[i].x=-coord[i].x;
    }
    if(normal.length){
        for(int i=normal.length;--i>=coordOffest;){
            normal[i].x=-normal[i].x;
        }
    }
    insideOut(triangleOffest/3);
}
void Art::mirrorFuncZ(){
    if(coord.length==0)return;
    M_Vec3f& coord=this->coord.refValue();
    M_Vec3f& normal=this->normal.refValue();

    for(int i=coord.length;--i>=0;){
        coord[i].z=-coord[i].z;
    }
    if(normal.length){
        for(int i=normal.length;--i>=0;){
            normal[i].z=-normal[i].z;
        }
    }
    insideOut();
}
void Art::mirrorFuncY(int triangleOffest, int coordOffest){
    if(coord.length==0)return;
    M_Vec3f& coord=this->coord.refValue();
    M_Vec3f& normal=this->normal.refValue();

    for(int i=coord.length;--i>=coordOffest;){
        coord[i].y=-coord[i].y;
    }
    if(normal.length){
        for(int i=normal.length;--i>=coordOffest;){
            normal[i].y=-normal[i].y;
        }
    }
    insideOut(triangleOffest/3);
}
void Art::mirrorFuncZ(int triangleOffest, int coordOffest){
    if(coord.length==0)return;
    M_Vec3f& coord=this->coord.refValue();
    M_Vec3f& normal=this->normal.refValue();

    for(int i=coord.length;--i>=coordOffest;){
        coord[i].z=-coord[i].z;
    }
    if(normal.length){
        for(int i=normal.length;--i>=coordOffest;){
            normal[i].z=-normal[i].z;
        }
    }
    insideOut(triangleOffest/3);
}
void Art::staticTransform(int coordOffest, Matrix& matrix){

    if(coord.length==0)return;
    M_Vec3f& coord=this->coord.refValue();

    for(int i=coord.length;--i>=coordOffest;){
        coord[i] = matrix * coord[i];
    }
}
