/* Seamless3d

Copyright (C) 2001-2017 Graham John Perrett (AKA thyme) thyme@seamless3d.com www.seamless3d.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this
software and associated documentation files (the "Software"), to deal in the Software
without restriction, including without limitation the rights to use, copy, modify,
merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to the following
conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT
OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.
*/

#include "seamless3d.h"

void commonGetSegAndBandDist(const M_Int& bands, const M_Float& segDist, M_Float& bandDist);

void CCLathe::generate(M_Vec3f& coord, const int h){
    if(!distance.length)return;
    if(g_projectDirection!=PROJECT_FORWARD){
        genCurveFromRadSrc(coord);
        return;
    }

    //double the last 2 values so that the last vertex points are curved
    float storeLastDistance=distance[distance.length-1];
    float storeLastAngle=curve[curve.length-1];
    distance.refValue()[distance.length-1]*=2;
    curve.refValue()[curve.length-1]*=2;



    for(int i=h; i<coord.length; i++){
        float accumHeight=0;
        float finishPosAx=baseRad;
        float finishPosAy=startPosY;
        float accumAng=curve[0];
        float nrad=distance[0]/curve[1];
        float rad=fabs(nrad);

        float AxisBx=-nrad*cos(accumAng)+finishPosAx;
        float AxisBy=-nrad*sin(accumAng)+finishPosAy;

        //float AxisBx=ccd->xzBaseOffset-(ccd->cpoint[1].height-ccd->cpoint[0].height)/fabs(ccd->cpoint[1].angle);
        //	float AxisBy=0;



        //float AxisBx=1;//ccd->xzBaseOffset-(fabs((ccd->cpoint[1].height-ccd->cpoint[0].height)/ccd->cpoint[1].angle));
        //float AxisBy=0;
        float rad0=1;
        float ty;

        float r;
        float zAngle;
        //accumAng=0;
        float angDiff;

    ////////////////////////////////////////////////////////
        int s=1;


        Vec3f &svec=coord[i];
        while(true){
            angDiff=curve[s];

            ty=accumAng*nrad;
            if(svec.y>=accumHeight&&svec.y<accumHeight+distance[s-1]){
                zAngle=(svec.y-accumHeight+ty)/rad;
                r=nrad*cos(zAngle)+AxisBx;
                svec.y=rad*sin(zAngle)+AxisBy;
                svec.z=r*-sin(svec.x/rad0-PI*.5);
                svec.x=r*cos(svec.x/rad0-PI*.5);
                break;
            }

            finishPosAx=nrad*cos(accumAng+angDiff);
            finishPosAy=nrad*sin(accumAng+angDiff);

            accumHeight+=distance[s-1];
            accumAng+=curve[s];
            if(++s>=curve.length)break;

            nrad=distance[s-1]/curve[s];
            rad=fabs(nrad);

            AxisBx+=-nrad*cos(accumAng)+finishPosAx;
            AxisBy+=-nrad*sin(accumAng)+finishPosAy;
        }
    }

    distance.refValue()[distance.length-1]=storeLastDistance;
    curve.refValue()[curve.length-1]=storeLastAngle;
}

void CCLathe::genCurveFromRadSrc(M_Vec3f& coord){
    //double the last 2 values so that the last vertex points are curved
    float storeLastDistance=distance.refValue()[distance.length-1];
    float storeLastAngle=curve.refValue()[curve.length-1];
    distance.refValue()[distance.length-1]*=2;
    curve.refValue()[curve.length-1]*=2;

////////////////////////////////////////////////////////

    for(int i=0;i<coord.length;i++){
        Vec3f &svec=coord[i];
        float accumHeight=0;
        float finishPosAx=baseRad;
        float finishPosAy=startPosY;
        float accumAng=curve[0];
        float nrad=distance[0]/curve[1];
        float rad=fabs(nrad);

        float AxisBx=-nrad*cos(accumAng)+finishPosAx;
        float AxisBy=-nrad*sin(accumAng)+finishPosAy;

        float ty;

        float r;
        float zAngle;
        float angDiff;

        int s=1;
        while(true){
            angDiff=curve[s];

            ty=accumAng*nrad;
            float polarPos=sqrt(svec.x*svec.x+svec.z*svec.z)-baseRad;
            if(polarPos>=accumHeight&&polarPos<accumHeight+distance[s-1]){
                float polarAngle=atan2(svec.z,svec.x);
                zAngle=(polarPos-accumHeight+ty)/rad;
                r=nrad*cos(zAngle)+AxisBx;
                svec.y=rad*sin(zAngle)+AxisBy;
                svec.z=r*sin(polarAngle);
                svec.x=r*cos(polarAngle);
                break;
            }


            finishPosAx=nrad*cos(accumAng+angDiff);
            finishPosAy=nrad*sin(accumAng+angDiff);

            accumHeight+=distance[s-1];
            accumAng+=curve[s];
            if(++s>=curve.length)break;

            nrad=distance[s-1]/curve[s];
            rad=fabs(nrad);

            AxisBx+=-nrad*cos(accumAng)+finishPosAx;
            AxisBy+=-nrad*sin(accumAng)+finishPosAy;
        }
    }

    distance.refValue()[distance.length-1]=storeLastDistance;
    curve.refValue()[curve.length-1]=storeLastAngle;
}

float CCLathe::getPosX(int index)const{
    return _getPosX( index, baseRad.refValue(), distance.refValue(), curve.refValue());
}
float CCLathe::getPosY(int index)const{
    float accumHeight=0;
    float finishPosAx=baseRad.refValue();
    float finishPosAy=startPosY;
    float accumAng=curve.refValue()[0];
    float nrad=distance.refValue()[0]/curve.refValue()[1];

    float AxisBx=-nrad*cos(accumAng)+finishPosAx;
    float AxisBy=-nrad*sin(accumAng)+finishPosAy;

    float angDiff;

////////////////////////////////////////////////////////
    int s=1;


    while(true){
        angDiff=curve.refValue()[s];

        finishPosAx=nrad*cos(accumAng+angDiff);
        finishPosAy=nrad*sin(accumAng+angDiff);

        accumHeight+=distance.refValue()[s-1];
        accumAng+=curve.refValue()[s];
        if(++s>index)break;

        nrad=distance.refValue()[s-1]/curve.refValue()[s];

        AxisBx+=-nrad*cos(accumAng)+finishPosAx;
        AxisBy+=-nrad*sin(accumAng)+finishPosAy;
    }
    return finishPosAy+AxisBy;
}
float CCLathe::getCurve(int index)const{
    if(index>=curve.length){
        index=curve.length-1;
    }
    float curveSum=0;
    int i;
    for(i=0;i<=index;i++){
        curveSum+=curve[i];
    }
    return curveSum;
}

float CCLathe::getDistance(int index)const{
    if(index>=distance.length){
        index=distance.length-1;
    }
    float distanceSum=0;
    int i;
    for(i=0;i<=index;i++){
        distanceSum+=distance.refValue()[i];
    }
    return distanceSum;
}
void CCLathe::calcClose(){
    if(autoClose==false)return;

    float curveSum=0;
    int i;
    for(i=0;i<curve.length-1;i++){
        curveSum+=curve.refValue()[i];
    }
    curve.refValue()[i]=PI*.5-curveSum;
    distance.refValue()[i-1]=getPosX(curve.length-2)/cos(curveSum)*curve[i];

}
void CCLathe::getBandAndYDist(M_Int& bands, M_Float& bandDist, M_Float& yDistance)const{
    commonGetSegAndBandDist(bands, distance, bandDist);
    yDistance=distance;
}
