/* Seamless3d

Copyright (C) 2001-2010 Graham John Perrett (AKA thyme) thyme@seamless3d.com www.seamless3d.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this
software and associated documentation files (the "Software"), to deal in the Software
without restriction, including without limitation the rights to use, copy, modify,
merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to the following
conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT
OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.
*/

#define WIN32_LEAN_AND_MEAN		// Exclude rarely-used stuff from Windows headers

#include <stdio.h>
#include <math.h>
#include <winsock2.h>
#include <wininet.h>
//#include <shellapi.h>

#pragma warning( disable : 4996)


#include "seamless3d.h"
#include "script.h"
#include "techuelife_triangle_editor.h"
#include "gui.h"
#include "ms-gui.h"
#include "chat_window.h"
#include "scene_tree.h"
#include "resource.h"

bool smlsIsSaved();
extern String g_currentFilename;

#ifdef TTS
void textToSpeech(const char* str, bool male, const char* voice);
bool checkVoiceExists(const char* voiceRequest);
void initDefaultVoice();
#endif

enum ChatInstruction{
	CLIENT_REQUEST_TO_JOIN=1,
	SERVER_REPLY_FOR_JOIN,
	SERVER_REPLY_NEED_HIGHER_VERSION,
	NEW_CLIENT_JOINED,
	CLIENT_LEFT,
	CHAT_FROM_CLIENT,
	MESSAGE_FROM_SERVER,
	CLIENT_ORIENTATION_POSITION,
	CLIENT_NAV_KEY,
	CLIENT_GESTURE,
	TOUCH_SENSOR,

	CLIENT_LEFT_KEY=1,
	CLIENT_RIGHT_KEY=2,
	CLIENT_BACKWARD_KEY=4,
	CLIENT_FORWARD_KEY=8,
};

class AvatarSocket : public AGroup, public Download, public PovRay, public MField8 {
typedef bool (AvatarSocket::*MultiTask)();

	MultiTask multiTaskB;

public:
class Orientation : public SFRotation {
public:
	AvatarSocket& home;

	Orientation(const Rotation& par, AvatarSocket& home):
		SFRotation(par),
		home(home)
	{};
	void operator=(const Rotation& par);
};
class Position : public SFVec3f {
public:
	AvatarSocket& home;

	Position(const Vec3f& par, AvatarSocket& home):
		SFVec3f(par),
		home(home)
	{};
	void operator=(const Vec3f& par);
};

	int getType()const{return AVATARSOCKET;}
	int getFLen()const{return size();};
	Field* getField(const int i)const{return (*this)[i];}
	int getFieldGroupIndex(const int &i)const{throw this;}
	void setFieldGroupIndex(const int &i, const int &val){throw this;}
	M_Int* getIndexPtr(){return nullPtr;}

	SFRotation orientation;
	SFVec3f position;
	MFNode children;
	SFString nickname;
	SFString avatarUrl;
	SFBool male;
	SFString voice;
	SFInt32 id;

	AvatarSocket():
		MField8(
			&orientation,
			&position,
			&children,
			&nickname,
			&avatarUrl,
			&male,
			&voice,
			&id
		),

		orientation(Rotation(0,1,0,0)),
		position(Vec3f()),
		children(M_Node()),
		nickname(String()),
		avatarUrl(String()),
		male(false),
		voice(String()),
		id(0),

		AGroup(children),
		/////
		hInternetOpenUrl(0),
		hThread(0),
		dwThreadId(0),
		dwThrdParam(1),
		netSave(false)
	{
	}
	bool download();
	bool openUrl();
	bool netSave;

	String chat;
	int hInternetOpenUrl;
	M_<char> internetBuf;
	unsigned long runThread(); //only runs when g_serverMode == true
	void logOut();
	void sendToServer(M_<char>& buf);
	void getUrlPath(String& url);
	bool decodeChatInstruction();
	void sendStringToChat(String str);
	int getIndexFromId(short int id);
	void doClientHasLeft(int id);

	HANDLE hThread; 	
	DWORD dwThreadId;
	DWORD dwThrdParam;

	M_<char> inBuf;
	M_<char> outBuf;


	void renderNode(Matrix parentMatrix, Matrix catRotMatrix);
	void outPovRay(String& str, M_<Effect*>& effectList, Matrix matrix, Matrix rotMatrix);
	void outObj(String& str, String& mtl, int& vidx, int& vtidx, M_<Effect*>& effectList, Matrix matrix, Matrix rotMatrix);
	void outWebGL(String& str, int& vidx, int& vtidx, M_<Effect*>& effectList, Matrix matrix, Matrix rotMatrix);

	Ptr<GestureGroup> gestureGroup;

	static SOCKET socket;
	static String str;
private:
	static char stackBuf[1024];
	static String str2;
	int limit;
	int targetLen;
};

SOCKET AvatarSocket::socket=0;
char AvatarSocket::stackBuf[1024];
String AvatarSocket::str;
String AvatarSocket::str2;

fd_set g_readfds;
fd_set g_writefds;
fd_set g_exceptfds;

fd_set g_t_readfds;
fd_set g_t_writefds;
fd_set g_t_exceptfds;

void openWithWindowsMediaPlayer(String url);

extern HINTERNET g_hInternetSession;
extern HINSTANCE  g_hinst;
extern HWND hWndMain;
extern int g_wireFrameMode;
extern float g_remVersionLoaded;
extern bool g_forceSaveAs;
extern SettingsPtr g_settings;
extern bool g_bActive;
Gravity g_gravity(0, 0, 100, 50);

//avatarUrl("http://www.seamless3d.com/wrl/test_mu/nurbs_lathe_av.smls"),
//avatarUrl("http://www.seamless3d.com/av/buildColorPerVertexAv.smls"),

//worldUrl("http://www.seamless3d.com/wrl/test_mu/test_mu.smls"),
//worldUrl("http://www.seamless3d.com/wrl/techuelife_island/main.smls"),

//const char* OpenNetFile::init_worldUrl = "http://localhost/mu/server.php";
//const char* OpenNetFile::init_avatarUrl = "http://localhost/test_mu/nurbs_lathe_av.smls";
//const char* OpenNetFile::init_worldUrl = "http://www.seamless3d.com/http://www.seamless3d.com/wrl/my_world/my_world.smls";
const char* OpenNetFile::init_worldUrl = "http://www.seamless3d.com/wrl/my_world/my_world.smls";
const char* OpenNetFile::init_serverUrl = "http://www.seamless3d.com";
const char* OpenNetFile::init_avatarUrl = "http://www.seamless3d.com/av/nurbs_lathe_av.smls";

M_String OpenNetFile::worldListUrls;
M_String OpenNetFile::worldListTitles;
M_String OpenNetFile::worldListOccupants;


Ptr<ChatWin> g_cw = new ChatWin;
bool g_multiUserMode = false;
bool g_loggedIn = false;
bool g_browseMode = false;

M_<String> g_chatLines;
HANDLE hThread;
DWORD dwThreadId = 0, dwThrdParam = 1; 
String g_IPNum;
M_<AvatarSocket*> g_m_avatarSocket;
Ptr<AvatarSocket> g_userAvatarSocket;
OpenNetFilePtr g_openNetFile = new OpenNetFile;
String g_smlsWorldUrl;

bool g_leftKeyDown=false;
bool g_rightKeyDown=false;
bool g_backwardKeyDown=false;
bool g_forwardKeyDown=false;

double g_leftKeyTime=0;
double g_rightKeyTime=0;
double g_backwardKeyTime=0;
double g_forwardKeyTime=0;


//code 1 = ReceiveInitFromServer
//code 2 = NewClientJoined
//code 3 = ClientLeft
//code 4 = ChatFromClientViaServer
//code 5 = MessageFromServer
//code 6 = ClientOrientationPosition

int g_beginYChatOffset=0;
int g_numOfVisChatLines;
//int charWidth=8;
extern bool g_online;

void checkForNameClash(M_Node& importDefList, M_Node& existingDefList);
void finishImportNodesFromFile();
void gmtTime(String& str);
void buildChatWindowMenu();
void openWithHtmlBrowser(String url, const String htmlBrowser);

void AvatarSocket::logOut(){
	closesocket(socket);
	g_loggedIn = false;
	g_m_avatarSocket.length=0;
}
bool showError(char* s){
	char str[0x200];
	int error=WSAGetLastError();
	if(error==0){
		return false;
	}
	sprintf(str, "%s: %d", s, error);
	dLog(str);
	
	MessageBox(NULL,str,NULL,MB_OK|MB_ICONERROR);
	return true;
}
void insertBool(M_<char>& buf, bool data){
	buf+=static_cast<char>(data);
}
void insertShort(M_<char>& buf, short data){
	for(int i=2; --i>=0;){
		buf+=static_cast<char>(data);
		data>>=8;
	}
}
void insertUnsignedShort(M_<char>& buf, int mark, int data){
	if(data>=0x10000){
		dLog("*insertUnsignedShort excedes 16 bits*!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!");
		return;
	}
	for(int i=2; --i>=0;){
		buf[mark++]=static_cast<char>(data);
		data>>=8;
	}
}
void insertInt32(M_<char>& buf, int data){
	for(int i=4; --i>=0;){
		buf+=static_cast<char>(data);
		data>>=8;
	}
}
void insertFloat(M_<char>& buf, float data){
	char* p=reinterpret_cast<char*>(&data);
	for(int i=4; --i>=0;){
		buf+=*p;
		p++;
	}
}

void insertRotation(M_<char>& buf, Rotation data){
	char* p=reinterpret_cast<char*>(&data);
	for(int i=sizeof(Rotation); --i>=0;){
		buf+=*p;
		p++;
	}
}
void insertVec3f(M_<char>& buf, Vec3f data){
	char* p=reinterpret_cast<char*>(&data);
	for(int i=sizeof(Vec3f); --i>=0;){
		buf+=*p;
		p++;
	}
}
const int SIZE_OF_BUFFER_LEN_TYPE = sizeof(unsigned short);

void insertInstruction(M_<char>& buf, char ChatInstruction){
	buf=ChatInstruction;
	buf.length+=SIZE_OF_BUFFER_LEN_TYPE;
}
void insertInstruction(M_<char>& buf, char ChatInstruction, short id){
	insertInstruction(buf, ChatInstruction);
	insertShort(buf, id);
}
void insertStr(M_<char>& buf, const String str){
	for(int i=0; i<str.length; i++){
		buf+=str[i];
	}
	buf+=0;
}
void AvatarSocket::sendToServer(M_<char>& buf){
	insertUnsignedShort(buf, 1, buf.length);
	if(	SOCKET_ERROR==send(socket,	buf.getArray(),	buf.length, 0)){
		showError("send()");
		logOut();
	}
	dLog("sendToServer");
}
bool joinPHP(String filename, String& ip){
	SOCKET itvSocket;
    HOSTENT *hostEntry;

	String domain;
	{	
		int i = filename.indexOf("/") + 2;
		domain = filename.substring(i);
		i = domain.indexOf("/") - 1;
		domain = domain.substring(0, i);
	}

	hostEntry = gethostbyname(domain.c_str());
	if (WSAGetLastError() != 0) {
        return false;
	}
	dLog("gethostbyname ok");
	
	itvSocket = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
	if (itvSocket == INVALID_SOCKET) {
		return false;
	}
	
	SOCKADDR_IN saServer;
	saServer.sin_family = AF_INET;
	saServer.sin_addr = *((LPIN_ADDR)*hostEntry->h_addr_list);
	saServer.sin_port = htons(80);	// html port
	
	// connect to the server
	if(SOCKET_ERROR == connect(itvSocket, (LPSOCKADDR)&saServer, sizeof(saServer))){
		dLog("connect(itvSocket, (LPSOCKADDR)&saServer, sizeof(saServer)) ");
		closesocket(itvSocket);
		return false;
	}
	dLog("connect ok");

	String str=String("GET ");
	str+=filename;
	//str+=" HTTP/1.1\r\nHost: ";
	str+=" HTTP/1.1\r\n\
Host: ";
	str+=domain;
//str += "\r\nUser-Agent: Mozilla/4.0 (compatible; MSIE 7.0)";
str += "\r\nUser-Agent: Seamless3d";
//	str += "\r\nAccept: */*\r\n
//str += "\r\nAccept-Language: en-us,en-au;q=0.5\r\n
//str += "\r\nUA-CPU: x86\r\n
//str += "\r\nAccept-Encoding: gzip, deflate\r\n
//str += "\r\nUser-Agent: Mozilla/4.0 (compatible; MSIE 7.0; Windows NT 5.1; Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; SV1))";
//str += "\r\nUser-Agent: Mozilla/5.0";

//str += "\r\nUser-Agent: Mozilla/5.0 (Windows; U; Windows NT 5.1; en-GB; rv:1.8.1.14) Gecko/20080404 Firefox/2.0.0.14";
//str += "\r\nAccept: text/xml,application/xml,application/xhtml+xml,text/html;q=0.9,text/plain;q=0.8,image/png,*/*;q=0.5";
//str += "\r\nAccept-Language: en-us,en-au;q=0.5";
//str += "\r\nAccept-Encoding: gzip,deflate";
//str += "\r\nAccept-Charset: ISO-8859-1,utf-8;q=0.7,*;q=0.7";
//str += "\r\nKeep-Alive: 300";
//str += "\r\nConnection: keep-alive";
//str += "\r\nCookie: style_cookie=null";
//str += "\r\nCache-Control: max-age=0";

	str+="\r\n\r\n\r\n";
	if( SOCKET_ERROR == send(itvSocket, str.c_str(), str.length, 0)){
		dLog("send(itvSocket, str.c_str(), str.length, 0)");
		closesocket(itvSocket);	
		return false;
	}
	dLog("send GET itv HTTP ok");

	char szBuf[2048];

	int txtLength = recv(itvSocket, szBuf, sizeof(szBuf)-1, 0);

	if (txtLength == INVALID_SOCKET)	{
		dLog("recv() INVALID_SOCKET");
		closesocket(itvSocket);
		return false;
	}
	else if (txtLength == 0)	{
		dLog("recv() 0");
		closesocket(itvSocket);
		return false;
	}
	szBuf[txtLength]=0;
	ip=szBuf;
	dLog("recv ok");

	closesocket(itvSocket);

	bool chunked = false;
	if(ip.indexOf("Transfer-Encoding: chunked") >= 0){
		chunked = true;
	}

	int i = ip.indexOf("\r\n\r\n");
	ip.stb(i);
	if(chunked){
		ip.stw(i);
		ip.stb(i);
	}
	ip=ip.substring(i);
	i = ip.indexOf("\r\n");
	if(i > 0){
		ip = ip.substring(0, i - 1);
	}
	return true;
}

int AvatarSocket::getIndexFromId(short int id){
	int i=g_m_avatarSocket.length;
	for(; --i>=0;){
		if(id==g_m_avatarSocket[i]->id){
			break;
		}
	}
	if(i<0){
		sendStringToChat("**Warning Unknown Identity!!**");
	}
	return i;
}
short int getShort(M_<char>& buf, int& mark){
	short int val = *reinterpret_cast<short int*>(&buf[mark]);
	mark+=sizeof(short);
	return val;
}
unsigned short int getUnsignedShort(const M_<char>& buf, int& mark){
	short int val = *reinterpret_cast<const short int*>(&buf[mark]);
	mark+=2;
	return val;
}
Rotation getRotation(M_<char>& buf, int& mark){
	Rotation val = *reinterpret_cast<Rotation*>(&buf[mark]);
	mark+=sizeof(Rotation);
	return val;
}
Vec3f getVec3f(M_<char>& buf, int& mark){
	Vec3f val = *reinterpret_cast<Vec3f*>(&buf[mark]);
	mark+=sizeof(Vec3f);
	return val;
}
bool getBool(M_<char>& buf, int& mark){
	bool val = *reinterpret_cast<bool*>(&buf[mark]);
	mark+=sizeof(bool);
	return val;
}
void addStr(String& str, M_<char>& buf, int& mark){
	int i=str.length;
	str+=&buf.getArray()[mark];
	mark+=str.length - i + 1;
}

void sendStrToDisplay(String receiveStr){
	// send string to display

	g_chatLines+=receiveStr;
	g_cw->paint();
	if(g_bActive == false){
		FlashWindow(hWndMain, TRUE);
	}

	String chatEditStr="\n";
	chatEditStr+=receiveStr;
	
	String fileName = g_appDataDir;
	fileName += "chat.txt";
	FILE* stream = fopen(fileName.c_str(), "a+t");
	fputs(chatEditStr.c_str(), stream);
	fclose(stream);
}
bool socketToBuf(SOCKET& socket, M_<char>& inBuf){
	int targetLen = 0;
	const int initialLen = 3;
	inBuf.length=0;
	char stackBuf[1024];
	int recvLen = recv(socket, stackBuf, initialLen, 0);

	if(recvLen == INVALID_SOCKET)	{
		dLog("recv() INVALID_SOCKET");
		return false;
	}
	else if(recvLen == 0)	{
		dLog("recv() 0");
		return false;
	}
	for(int i = 0; i < recvLen; i++){
		inBuf+=stackBuf[i];
	}
	if(inBuf[0]>='A'){//html
		recvLen = recv(socket, &stackBuf[3], 200, 0); //offset 3 bytes

		if(recvLen == INVALID_SOCKET){
			dLog("recv() INVALID_SOCKET");
			return false;
		}
		else if(recvLen == 0)	{
			dLog("recv() 0");
			return false;
		}
		recvLen+=3; //add 3 bytes to compensate
		int i = 3;
		for(; i < recvLen; i++){
			inBuf+=stackBuf[i];
		}

		stackBuf[recvLen] = 0;
		String str = stackBuf;

		bool chunked = false;
		if(str.indexOf("Transfer-Encoding: chunked") >= 0){
			chunked = true;
		}

		i = str.indexOf("Content-Length:");
		if(i <= 0){
			return true;
		}
		i += sizeof("Content-Length:");
		str.stb(i);
		int contentLen = getInt(str,i);


		i = str.indexOf("\r\n\r\n");
		str.stb(i);
		if(chunked){
			str.stw(i);
			str.stb(i);
		}
		targetLen = i + contentLen;
	}
	else{
		int mark = 1;
		targetLen = getUnsignedShort(inBuf, mark);
	}

	while(inBuf.length < targetLen){
		int limit = targetLen - inBuf.length;
		if(limit >= sizeof(stackBuf)){
			limit = sizeof(stackBuf);
		}
		recvLen = recv(socket, stackBuf, limit, 0);
		if(recvLen == INVALID_SOCKET)	{
			dLog("recv() INVALID_SOCKET");
			return false;
		}
		else if(recvLen == 0)	{
			dLog("recv() 0");
			return false;
		}

		for(int i = 0; i < recvLen; i++){
			inBuf+=stackBuf[i];
		}
	}
	return true;
}
double g_lastTimeActive = 0;
M_String g_shared_touchSensors;
bool AvatarSocket::decodeChatInstruction(){

	String tmpStr;
	String sendStr;

	if(false == socketToBuf(socket, inBuf)){
		logOut();
		return false;
	}
	g_lastTimeActive = g_time;

	
	outBuf.length=0;
	char instruction = inBuf[0];
	int mark = 1 + SIZE_OF_BUFFER_LEN_TYPE;
	switch(instruction){
		case SERVER_REPLY_NEED_HIGHER_VERSION:{
			msg("You need a higher version of Seamless3d to login", MB_OK|MB_ICONEXCLAMATION);
			dLog("SERVER_REPLY_NEED_HIGHER_VERSION");
			break;
		}
		case SERVER_REPLY_FOR_JOIN:{
			dLog("SERVER_REPLY_FOR_JOIN");

			g_userAvatarSocket->id=getShort(inBuf, mark);
			g_userAvatarSocket->nickname=g_openNetFile->nickname;
			g_userAvatarSocket->avatarUrl=g_openNetFile->avatarUrl;
			g_userAvatarSocket->male=g_openNetFile->male;
			g_userAvatarSocket->voice=g_openNetFile->voice;
#ifdef TTS
			if(checkVoiceExists(g_userAvatarSocket->voice.refValue().c_str())){
				g_userAvatarSocket->voice=g_openNetFile->voice;
			}
			else{
				g_userAvatarSocket->voice="";
			}
#endif
			g_m_avatarSocket += g_userAvatarSocket;
			g_downloadQue += g_userAvatarSocket;

			for(int k=getShort(inBuf, mark); --k>=0;){
				AvatarSocket* avatarSocket = new AvatarSocket;
				avatarSocket->id.refValue()=getShort(inBuf, mark);
				addStr(avatarSocket->nickname.refValue(), inBuf, mark);
				addStr(avatarSocket->avatarUrl.refValue(), inBuf, mark);
				avatarSocket->male.refValue() = getBool(inBuf, mark);
				addStr(avatarSocket->voice.refValue(), inBuf, mark);
#ifdef TTS
				if(checkVoiceExists(avatarSocket->voice.refValue().c_str())==false){
					avatarSocket->voice="";
				}
#endif
				avatarSocket->orientation.refValue() = getRotation(inBuf, mark);
				avatarSocket->position.refValue() = getVec3f(inBuf, mark);
				g_m_avatarSocket += avatarSocket;
				g_downloadQue+=avatarSocket;
			};
			gmtTime(tmpStr="");
			tmpStr+="your existence has been acknowledged by the server ";
			tmpStr+=g_userAvatarSocket->nickname;
			sendStrToDisplay(tmpStr);
#ifdef TTS
			textToSpeech(tmpStr.c_str(), false, "");
#endif
			//if(g_m_avatarSocket.length==1){						
			//}
			break;
		}
		case NEW_CLIENT_JOINED:{
			AvatarSocket* avatarSocket = new AvatarSocket;
			avatarSocket->id.refValue()=getShort(inBuf, mark);
			addStr(avatarSocket->nickname.refValue(), inBuf, mark);
			addStr(avatarSocket->avatarUrl.refValue(), inBuf, mark);
			avatarSocket->male.refValue() = getBool(inBuf, mark);
			addStr(avatarSocket->voice.refValue(), inBuf, mark);
#ifdef TTS
			if(checkVoiceExists(avatarSocket->voice.refValue().c_str())==false){
				avatarSocket->voice="";
			}
#endif
			avatarSocket->orientation.refValue() = getRotation(inBuf, mark);
			avatarSocket->position.refValue() = getVec3f(inBuf, mark);

			g_m_avatarSocket+=avatarSocket;
			g_downloadQue+=avatarSocket;

			gmtTime(tmpStr="");
			tmpStr+=avatarSocket->nickname;
			tmpStr+=" has joined this world ";
			sendStrToDisplay(tmpStr);
			if(g_m_avatarSocket.length == 2 && g_openNetFile->alarmFile != ""){
				openWithWindowsMediaPlayer(g_openNetFile->alarmFile);
			}
			break;
		}
		case CLIENT_LEFT:{
			doClientHasLeft(getShort(inBuf, mark));
			break;
		}
		case CHAT_FROM_CLIENT:{
			int id=getShort(inBuf, mark);
			int idx=getIndexFromId(id);
			if(idx==-1){
				break;
			}
			AvatarSocket* avatarSocket=g_m_avatarSocket[idx];
			avatarSocket->chat="";
			addStr(avatarSocket->chat, inBuf, mark);

			tmpStr=avatarSocket->nickname;
			tmpStr+=" : ";
			tmpStr+=avatarSocket->chat;

			int i=0;
			if(id != g_userAvatarSocket->id){
				if(avatarSocket->chat.cmpCppWord("pos", i)){
					avatarSocket->position.refValue().getField(avatarSocket->chat, i);
				}
				else if(avatarSocket->chat.cmpCppWord("ori", i)){
					avatarSocket->orientation.refValue().getField(avatarSocket->chat, i);
				}
			}
			if(avatarSocket->chat == "version?"){
				sendStrToDisplay(tmpStr);
				sendStr= "version = ";
				sendStr+=g_version;
				sendStringToChat(sendStr);
				break;
			}
			else if(avatarSocket->chat == "client info?"){
				sendStrToDisplay(tmpStr);
				sendStr= "";
				for(int i=g_m_avatarSocket.length; --i>=0;){
					AvatarSocket* avatarSocket=g_m_avatarSocket[i];
					sendStr+=avatarSocket->nickname;
					sendStr+=" > position = ";
					avatarSocket->position.refValue().toString(sendStr);
					sendStr+="; ";
				}
				sendStringToChat(sendStr);
				break;
			}
			else if(avatarSocket->chat.length > 1 && avatarSocket->chat[0] == 'g' && avatarSocket->chat[1] >= '0' && avatarSocket->chat[1] <= '9'){
				int i = 1;
				int gestureIdx = getInt(avatarSocket->chat, i);
				if(Ptr<GestureGroup> gestureGroup = avatarSocket->gestureGroup){
					M_Node& m_anim = gestureGroup->children;
					if(gestureIdx < m_anim.length){
						Anim* anim = dynamic_cast<Anim*>(m_anim[gestureIdx].p);
						anim->play = true;
					}
				}
			}
			sendStrToDisplay(tmpStr);
#ifdef TTS					
			textToSpeech(avatarSocket->chat.c_str(), avatarSocket->male, avatarSocket->voice.refValue().c_str());
#endif

			break;
		}
		case MESSAGE_FROM_SERVER:{
			// send string to display
			tmpStr="";
			addStr(tmpStr, inBuf, mark);
			sendStrToDisplay(tmpStr);
			break;
		}
		case CLIENT_ORIENTATION_POSITION:{
			int idx=getIndexFromId(getShort(inBuf, mark));
			if(idx==-1){
				break;
			}
			AvatarSocket* avatarSocket=g_m_avatarSocket[idx];
			avatarSocket->orientation.refValue()=getRotation(inBuf, mark);
			avatarSocket->position.refValue()=getVec3f(inBuf, mark);
			avatarSocket->position.refValue().toString(tmpStr);
			break;
		}
		case CLIENT_GESTURE:{
			int idx = getIndexFromId(getShort(inBuf, mark));
			if(idx==-1){
				break;
			}
			int gestureIdx = getShort(inBuf, mark);

			AvatarSocket* avatarSocket=g_m_avatarSocket[idx];

			if(Ptr<GestureGroup> gestureGroup = avatarSocket->gestureGroup){
				M_Node& m_anim = gestureGroup->children;
				if(gestureIdx < m_anim.length){
					Anim* anim = dynamic_cast<Anim*>(m_anim[gestureIdx].p);
					anim->play = true;
				}
			}
			break;
		}
		case TOUCH_SENSOR:{
			int idx=getIndexFromId(getShort(inBuf, mark));
			if(idx==-1){
				break;
			}
			AvatarSocket* avatarSocket = g_m_avatarSocket[idx];

			short whoGotClicked = getShort(inBuf, mark);

			tmpStr="";
			addStr(tmpStr, inBuf, mark);
			g_shared_touchSensors+=tmpStr;
			break;
		}
		case CLIENT_NAV_KEY:{
			int idx=getIndexFromId(getShort(inBuf, mark));
			if(idx==-1){
				break;
			}
			AvatarSocket* avatarSocket=g_m_avatarSocket[idx];
			char keys=inBuf[mark++];
			if(keys&CLIENT_LEFT_KEY){
				avatarSocket->orientation.refValue() *= Rotation(0,1,0,PI*.01);
			}
			if(keys&CLIENT_RIGHT_KEY){
				avatarSocket->orientation.refValue() *= Rotation(0,1,0,-PI*.01);
			}
			if(keys&CLIENT_BACKWARD_KEY){
				avatarSocket->position.refValue() += avatarSocket->orientation.refValue() * Vec3f(0,0,-.05);
			}
			if(keys&CLIENT_FORWARD_KEY){
				avatarSocket->position.refValue() += avatarSocket->orientation.refValue() * Vec3f(0,0,.05);
			}
			break;
		}
		default:{
			sendStringToChat("received unknown instruction from server!");
		}
	}
	return 0;
}
void sendToAllExcept(M_<char>& buf, AvatarSocket* block){
		insertUnsignedShort(buf, 1, buf.length);
		for(int i=g_m_avatarSocket.length;--i>=0;){
		AvatarSocket* avatarSocket = g_m_avatarSocket[i];
		if(avatarSocket!= g_userAvatarSocket && avatarSocket!=block){
			if(send(avatarSocket->socket, buf.getArray(), buf.length, 0) == SOCKET_ERROR){
				showError("send() = SOCKET_ERROR");
			}
		}
	}
}
void sendToOne(M_<char>& buf, AvatarSocket* one){
	insertUnsignedShort(buf, 1, buf.length);
	if(send(one->socket, buf.getArray(), buf.length, 0) == SOCKET_ERROR){
		showError("send() = SOCKET_ERROR");
	}
}
void sendToAll(M_<char>& buf){
	insertUnsignedShort(buf, 1, buf.length);
	for(int i=g_m_avatarSocket.length;--i>=0;){
		AvatarSocket* avatarSocket = g_m_avatarSocket[i];
		if(avatarSocket!= g_userAvatarSocket){
			if(send(avatarSocket->socket, buf.getArray(), buf.length, 0) == SOCKET_ERROR){
				showError("send() = SOCKET_ERROR");
				//closesocket(g_m_avatarSocket[i]->socket);	
			}
		}
	}
}
void sendToAll(const String chatStr){
	static M_<char> buf;
	buf.length=0;

	insertInstruction(buf, MESSAGE_FROM_SERVER);

	insertStr(buf, chatStr);

	sendToAll(buf);
}

void AvatarSocket::sendStringToChat(String str){
	static M_<char> buf;

	if(str.length > 1 && str[0] == 'g' && str[1] >= '0' && str[1] <= '9'){
		int i = 1;
		int gestureIdx = getInt(str, i);

		if(Ptr<GestureGroup> gestureGroup = g_userAvatarSocket->gestureGroup){
			M_Node& m_anim = gestureGroup->children;
			if(gestureIdx < m_anim.length){
				Anim* anim = dynamic_cast<Anim*>(m_anim[gestureIdx].p);
				anim->play = true;
			}
		}

		insertInstruction(buf, CLIENT_GESTURE, g_userAvatarSocket->id);
		insertShort(buf, gestureIdx);
		sendToServer(buf);
	}
	else{
		if(Ptr<GestureGroup> gestureGroup = g_userAvatarSocket->gestureGroup){
			M_Node& m_anim = gestureGroup->children;
			for(int k = m_anim.length; --k>=0;){
				Anim* anim = dynamic_cast<Anim*>(m_anim[k].p);
				if(str.indexOf(anim->name)!=-1){
					anim->play = true;
					insertInstruction(buf, CLIENT_GESTURE, g_userAvatarSocket->id);
					insertShort(buf, k);
					sendToServer(buf);
				}
			}
		}
		insertInstruction(buf, CHAT_FROM_CLIENT, g_userAvatarSocket->id);
		insertStr(buf, str);
		sendToServer(buf);
	}
}

void doChat(){
	g_shared_touchSensors.length=0;
	g_t_readfds = g_readfds;
	g_t_writefds = g_writefds;
	g_t_exceptfds = g_exceptfds;
	timeval tv;
	tv.tv_sec = 1;
	tv.tv_usec = 0;

	if(select(FD_SETSIZE, &g_t_readfds, &g_writefds, &g_t_exceptfds, 0) < 1){
		if(showError("select()")){
			return;
		}
	}
	if(FD_ISSET(g_userAvatarSocket->socket, &g_t_readfds)){
		g_userAvatarSocket->decodeChatInstruction();
	}
	if(g_time - g_lastTimeActive > 60*5 && g_userAvatarSocket->id){
		g_lastTimeActive = g_time;
		String& str = g_userAvatarSocket->str;
		gmtTime(str="");
		str+="waiting";
		g_userAvatarSocket->sendStringToChat(str);
	}
}

bool joinSession(){

	//	printf("Client connecting to server: %s on port: %d",
	//				szServer, nPort);
	
	// Find the server
	//in_addr		iaHost;
	
	//iaHost.s_addr = inet_addr(g_IPNum.c_str());
	
//	hostEntry = gethostbyaddr((const char *)&iaHost, sizeof(struct in_addr), AF_INET);
	HOSTENT* hostEntry = gethostbyname(g_IPNum.c_str());
    if (hostEntry == NULL) {
		showError("gethostbyname error: ");
        return false;
    }
	dLog("gethostbyname ok");
	
	// Create a TCP/IP stream socket
	
	g_userAvatarSocket->socket = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
	if (g_userAvatarSocket->socket == INVALID_SOCKET) {
		return false;
	}
	// Fill in the address structure
	SOCKADDR_IN saServer;
	
	saServer.sin_family = AF_INET;
	saServer.sin_addr = *((LPIN_ADDR)*hostEntry->h_addr_list);
	saServer.sin_port = htons(PORT);
	
	// connect to the server
	
	if(SOCKET_ERROR == connect(g_userAvatarSocket->socket, (LPSOCKADDR)&saServer, sizeof(struct sockaddr))){
		g_userAvatarSocket->logOut();
		return false;
	}
	dLog("connect");

#ifdef WIN32
	unsigned long iMode = 1;
	ioctlsocket(g_userAvatarSocket->socket, FIONBIO, &iMode);
#else
	int flags = fcntl(socket,F_GETFL,0);
	fcntl(g_userAvatarSocket->socket, F_SETFL, flags | O_NONBLOCK);
#endif
	FD_ZERO(&g_readfds);
	FD_ZERO(&g_writefds);
	FD_ZERO(&g_exceptfds);
	FD_SET(g_userAvatarSocket->socket, &g_readfds);
	FD_SET(g_userAvatarSocket->socket, &g_writefds);
	FD_SET(g_userAvatarSocket->socket, &g_exceptfds);

	g_loggedIn = true;
	dLog("logged in");
	return true;
}
bool initWinsock(){
	dLog("start ok");
	
	WORD wVersionRequested = MAKEWORD(2,0);
	WSADATA wsaData;
	int nRet;
	
	// Initialize WinSock and check the version
	nRet = WSAStartup(wVersionRequested, &wsaData);
	if (wsaData.wVersion != wVersionRequested) {	
		dLog("\n Wrong version\n");
		return false;
	}
	dLog("nRet = WSAStartup(wVersionRequested, &wsaData)");
	return true;
}
/*
void getDlgItemString(HWND hDlg, int nIDDlgItem, String& value){
	char tmpStr[0x400];
	GetDlgItemText( hDlg, nIDDlgItem, tmpStr, sizeof(tmpStr));
	value=tmpStr;
}
BOOL CALLBACK userFieldsProc(HWND hDlg,UINT iMessage, WPARAM wParam, LPARAM lParam){

//int b;
switch (iMessage){
	case WM_INITDIALOG:


		return(TRUE);


	case WM_COMMAND:
		switch LOWORD(wParam){

			case IDOK:
			    getDlgItemString( hDlg, IDC_CHAT_NAME, g_userNickname);
                EndDialog( hDlg, IDOK );

//				DestroyWindow(hDlg);
				return(FALSE);

			case WM_DESTROY:
                EndDialog( hDlg, IDCANCEL );
	            return TRUE;
			default:
				return(FALSE);
		}
		default:
			return(FALSE);
	}

	return(TRUE);
}
*/
void MultiUser::processFields(){
	if(ip.length){
		g_IPNum=ip.refValue()[0];
		if(initWinsock()){

			dLog("getIPAddress ok " + g_IPNum);
		}	
	}
}
WorldInfo* getWorldInfo(){
	for(int k=0; k<g_scene->sceneTreeList.length; k++){
		if(WorldInfo* worldInfo=dynamic_cast<WorldInfo*>(g_scene->sceneTreeList[k].node.p)){
			return worldInfo;
		}
	}
	return 0;
}
void Gravity::initialize(){
	land = 0;
	part = 0;
	for(int k=0; k<g_scene->sceneTreeList.length; k++){
		if(NurbsPatch* landPatch = dynamic_cast<NurbsPatch*>(g_scene->sceneTreeList[k].node.p)){
			String& name = landPatch->name;
			if(name.length >= 4 && name[0] == 'l' && name[1] == 'a' && name[2] == 'n' && name[3] == 'd'){
				M_Node& m_stem = landPatch->stem;
				if(m_stem.length){
					if(Stem* stem = dynamic_cast<Stem*>(landPatch->stem[0].p)){
						M_SArt& m_part = stem->part.refValue();
						if(m_part.length){
							part = m_part[0];
							land = landPatch;
							int uDim = landPatch->dim.x;
							M_Vec3f& dpoint = landPatch->dpoint;
							float width = dpoint[uDim-1].x - dpoint[0].x;
							g_gravity = Gravity(land, part, width, stem->staves[0]);

							return;
						}
					}
				}
			}
		}
	}
}
float Gravity::getLandHeight(float x, float z){
	float modX = (x + landTrans) / landSquareSize;
	float modZ = (z + landTrans) / landSquareSize;
	float sqX = floor(modX);
	float sqZ = floor(modZ);
	modX = (modX - sqX) * landSquareSize;
	modZ = (modZ - sqZ) * landSquareSize;

	int lx = sqX;
	int lz = landStrips - 1 - sqZ;
	lz = lz * landEdgePntCnt;
	int lz2 = lz + landEdgePntCnt;

	float bl = part->coord[lz + lx].y;
	float br = part->coord[lz + lx + 1].y;
	float tl = part->coord[lz2 + lx].y;
	float tr = part->coord[lz2 + lx + 1].y;

	float tx = modX - halfLSS;
	float tz = modZ - halfLSS;

	float angle = -.25 * PI;
	float newZ = tz * cos(angle) - tx * sin(angle);
	tx = tz * sin(angle) + tx * cos(angle);
	tz = newZ;

	if(tx <= 0){
		tr = br + (tl - bl);
	}
	else {
		bl = tl + (br - tr);
	}

	float by1 = tl + (bl - tl) * (modZ * oneDivByLandSquareSize);
	float by2 = tr + (br - tr) * (modZ * oneDivByLandSquareSize);

	return by1 + (by2 - by1) * (modX * oneDivByLandSquareSize);
}

void initMultiUserMode(){
	if(WorldInfo* worldInfo = getWorldInfo()){
#ifdef TTS
		initDefaultVoice();
#endif
		g_userAvatarSocket->position.refValue().y = g_openNetFile->landOffset;
		g_gravity.initialize();

		g_cw->pos = Vec2i(g_openNetFile->chatPos.x, g_openNetFile->chatPos.y);
		g_cw->size = Vec2i(g_openNetFile->chatSize.x, g_openNetFile->chatSize.y);
		g_cw->checkWinVis();

		Window::activeWindow = g_cw.p;

		g_cw->textCell.cellPos=Vec2i(7, g_cw->size.y - 26);
		g_numOfVisChatLines=(g_cw->size.y-g_beginYChatOffset-title_bar_height-25)/pixSymbolCY;

		//g_serverMode = !g_openNetFile->joinAlreadyExistingSession;


		if(joinSession()){

			g_userAvatarSocket->id = 0; //this is only done to act as a flag to stop the client sendiing a "waiting" message before it has an id
			dLog("start ok");
			M_<char> buf;
			insertInstruction(buf, CLIENT_REQUEST_TO_JOIN);
			insertFloat(buf, g_version);
			insertStr(buf, g_smlsWorldUrl);
			insertStr(buf, worldInfo->title);
			insertStr(buf, g_openNetFile->nickname);
			insertStr(buf, g_openNetFile->avatarUrl);
			insertBool(buf, g_openNetFile->male);
			insertStr(buf, g_openNetFile->voice);
			insertRotation(buf, g_userAvatarSocket->orientation);
			insertVec3f(buf, g_userAvatarSocket->position);
			g_userAvatarSocket->sendToServer(buf);
		}
	}
}
void AvatarSocket::Position::operator=(const Vec3f& par){
	static M_<char> buf;
	refValue()=par;
	if(&home==g_userAvatarSocket){
		buf.length=0;
		insertInstruction(buf, CLIENT_ORIENTATION_POSITION, home.id);
		insertRotation(buf, home.orientation);
		insertVec3f(buf, home.position);
		home.sendToServer(buf);
	}
}
void AvatarSocket::Orientation::operator=(const Rotation& par){
	static M_<char> buf;
	refValue()=par;
	if(&home==g_userAvatarSocket){
		buf.length=0;
		insertInstruction(buf, CLIENT_ORIENTATION_POSITION, home.id);
		insertRotation(buf, home.orientation);
		insertVec3f(buf, home.position);
		home.sendToServer(buf);
	}
}
/*
void leftNavKey(){
	g_userAvatarSocket->orientation.refValue()=g_userAvatarSocket->orientation.refValue()*Rotation(0,1,0,PI*.01);

	static M_<char> buf;
	buf.length=0;
	insertInstruction(buf, CLIENT_NAV_KEY, g_userAvatarSocket->id);
	buf+=CLIENT_LEFT_KEY;
	g_userAvatarSocket->sendToServer(buf);
}

void rightNavKey(){
	g_userAvatarSocket->orientation.refValue()=g_userAvatarSocket->orientation.refValue()*Rotation(0,1,0,-PI*.01);

	static M_<char> buf;
	buf.length=0;
	insertInstruction(buf, CLIENT_NAV_KEY, g_userAvatarSocket->id);
	buf+=CLIENT_RIGHT_KEY;
	g_userAvatarSocket->sendToServer(buf);
}
void backwardNavKey(){
	g_userAvatarSocket->position.refValue() = g_userAvatarSocket->position.refValue() + g_userAvatarSocket->orientation.refValue() * Vec3f(0,0,-.05);

	static M_<char> buf;
	buf.length=0;
	insertInstruction(buf, CLIENT_NAV_KEY, g_userAvatarSocket->id);
	buf+=CLIENT_BACKWARD_KEY;
	g_userAvatarSocket->sendToServer(buf);
}
void forwardNavKey(){
	g_userAvatarSocket->position.refValue() = g_userAvatarSocket->position.refValue() + g_userAvatarSocket->orientation.refValue() * Vec3f(0,0,.05);

	static M_<char> buf;
	buf.length=0;
	insertInstruction(buf, CLIENT_NAV_KEY, g_userAvatarSocket->id);
	buf+=CLIENT_FORWARD_KEY;
	g_userAvatarSocket->sendToServer(buf);
}
void sendOrienPosToServer(){
	static M_<char> buf;
	buf.length=0;
	insertInstruction(buf, CLIENT_ORIENTATION_POSITION, g_userAvatarSocket->id);
	insertRotation(buf, g_userAvatarSocket->orientation);
	insertVec3f(buf, g_userAvatarSocket->position);
	g_userAvatarSocket->sendToServer(buf);
}
*/
void setAvatarView(){
	if(g_gravity.land){
		g_userAvatarSocket->position.refValue().y = g_openNetFile->landOffset + g_gravity.getLandHeight(g_userAvatarSocket->position.refValue().x, g_userAvatarSocket->position.refValue().z);
	}

	g_scene->rotation = -g_userAvatarSocket->orientation * Rotation(0, 1, 0, PI) * Rotation(1, 0, 0, .1 * PI);

	g_scene->translation =  - g_userAvatarSocket->position - Vec3f(0, 1.75, 0);

	g_viewDist=3;
}
double g_prevFrameTime = 0;
void checkKeys(){

	float t = g_time - g_prevFrameTime;

	bool update=false;
	if(g_leftKeyDown){
		g_userAvatarSocket->orientation.refValue()=g_userAvatarSocket->orientation.refValue()*Rotation(0,1,0,PI/8*t);
		setAvatarView();
		if(g_time-g_leftKeyTime>autoRepeatPeriod){
			g_leftKeyTime+=autoRepeatPeriod;
			update=true;
		}
	}
	else if(g_rightKeyDown){
		g_userAvatarSocket->orientation.refValue()=g_userAvatarSocket->orientation.refValue()*Rotation(0,1,0,-PI/8*t);
		setAvatarView();
		if(g_time-g_rightKeyTime>autoRepeatPeriod){
			g_rightKeyTime+=autoRepeatPeriod;
			update=true;
		}
	}
	if(g_backwardKeyDown){
		g_userAvatarSocket->position.refValue() = g_userAvatarSocket->position.refValue() + g_userAvatarSocket->orientation.refValue() * Vec3f(0,0,-2 * t);
		setAvatarView();
		if(g_time-g_backwardKeyTime>autoRepeatPeriod){
			g_backwardKeyTime+=autoRepeatPeriod;
			update=true;
		}
	}
	else if(g_forwardKeyDown){
		g_userAvatarSocket->position.refValue() = g_userAvatarSocket->position.refValue() + g_userAvatarSocket->orientation.refValue() * Vec3f(0,0,2 * t);
		setAvatarView();
		if(g_time-g_forwardKeyTime>autoRepeatPeriod){
			g_forwardKeyTime+=autoRepeatPeriod;
			update=true;
		}
	}
	if(update){
		static M_<char> buf;
		buf.length=0;
		insertInstruction(buf, CLIENT_ORIENTATION_POSITION, g_userAvatarSocket->id);
		insertRotation(buf, g_userAvatarSocket->orientation);
		insertVec3f(buf, g_userAvatarSocket->position);
		g_userAvatarSocket->sendToServer(buf);
	}
	g_prevFrameTime = g_time;
}
void updateNavKey(){
	if(!g_online){
		return;
	}

	if(g_leftKeyDown || g_rightKeyDown || g_backwardKeyDown || g_forwardKeyDown){
		g_prevFrameTime = g_time;
	}
}

void AvatarSocket::doClientHasLeft(int id){
	int idx=getIndexFromId(id);
	if(idx==-1){
		return;
	}
	String str;
	gmtTime(str);
	str+=g_m_avatarSocket[idx]->nickname;
	str+=" has left this world";
	sendStrToDisplay(str);

	AvatarSocket* avatarSocket=g_m_avatarSocket[idx];
	for(int i=idx; i<g_m_avatarSocket.length-1; i++){
		g_m_avatarSocket[i]=g_m_avatarSocket[i+1];
	}
	g_m_avatarSocket.length--;


	int selNodeLine=getNodeLine(avatarSocket);

	int selFieldIdx=0;
	int selElement=0;
	
	int parentSelLine=getParent(selNodeLine);
	
	NodePtr parentSelNode=g_scene->sceneTreeList[parentSelLine].node;
	
	getFNodeElementForParent(parentSelLine, selNodeLine, g_scene->sceneTreeList, selFieldIdx, selElement );
	
	Field* field=parentSelNode->getField(selFieldIdx);
	
	if(field->getType()==MFNODE){
		MFNode& mfnode=*dynamic_cast<MFNode*>(field);
		M_Node& mnode=mfnode.refValue();
		removeNode(selElement, mnode, g_selCount);
		mfnode=mnode;
	}
	else{
		SFNode& snode=*dynamic_cast<SFNode*>(field);
		snode=nullPtr;
	}
	g_stw->genNodeList();
	if(g_stw->selLine>=g_scene->sceneTreeList.length){
		g_stw->selLine=g_scene->sceneTreeList.length-1;
	}
	g_selNode->getSceneDisplayString(g_stw->textCell.cellStr="");
	g_stw->textCell.updateCaret();
	setupSelectedNodeWindow();
	g_stw->paint();
}
bool extractWorldUrl(const String& worldUrl, String& getServerUrl, String& smlsUrl){
	if(worldUrl.length==0){
		return false;
	}
	int worldUrlIndex = worldUrl.indexOf("/http://");
	int serverEnd = worldUrlIndex;
	if(worldUrlIndex==-1){
		return true;	//could be single user
	}
	if(worldUrlIndex==0){
		return false;
	}
	worldUrlIndex++;

	getServerUrl = worldUrl.substring(0, serverEnd);
	getServerUrl+= "smls_server.php";
	smlsUrl = worldUrl.substring(worldUrlIndex);
	return true;
}
int markEndOfUrl(const String& str, int first){
	int last = first;
	while(last < str.length - 1){
		if(str[last] <= ' '){
			return last - 1;
		}
		last++;
	}
	return last;
}
void eraseBeginOfUrlWhiteSpace(String& str){
	int i = 0;
	str.stb(i);
	if(i){
		str = str.substring(i);
	}
}
void eraseEndOfUrlWhiteSpace(String& str){
	int last = markEndOfUrl(str, 0);
	if(last==-1){
		return;
	}
	else if(last+1==str.length){
		return;
	}
	str = str.substring(0, last);
}
void insertHttp(String& str){
	int i = str.indexOf("http://");
	if(i == 0){
		return;
	}
	str = "http://" + str;
}
bool isDecimal(const String& str, int& i){
	int from = i;
	while(i < str.length && str[i] >= 0x30 && str[i] <= 0x39){
		i++;
	}
	return from < i;
}
bool isValidIP(const String& str){
	int i = 0;
	int count = 0;
	while(isDecimal(str, i)){
		if(i < str.length){
			if(str[i] == '.'){
				i++;
				count++;
			}
			else{
				return false;
			}
		}
		else if(count == 3){
			return true;
		}
	}
	return false;
}
void deleteLastForwardSlash(String& str){
	if(str.length){
		if(str[str.length-1]=='/'){
			str = str.substr(0, str.length-1);
		}
	}
}
bool processServerUrl(){
	String& serverUrl = g_openNetFile->serverUrl.refValue();
	eraseBeginOfUrlWhiteSpace(serverUrl);
	eraseEndOfUrlWhiteSpace(serverUrl);
	bool directIP = isValidIP(serverUrl);
	if(directIP){
		g_IPNum = serverUrl;
	}
	else{
		insertHttp(serverUrl);
		deleteLastForwardSlash(serverUrl);
	}
	return directIP;
}


void completeGo(){
	if(g_openNetFile->nickname==""){
		msG("please give yourself a nickname");
	}
	else{
		g_userAvatarSocket=new AvatarSocket;

		g_downloadQue.length = 0;

		g_callBackList.length = 0;

		g_scene->children.refValue().length = 0;

		//g_scene->openUrl(g_openNetFile->worldUrl.refValue());

		if(initWinsock()){
			bool directIP = processServerUrl();

			String& worldUrl = g_openNetFile->worldUrl.refValue();
			eraseBeginOfUrlWhiteSpace(worldUrl);
			eraseEndOfUrlWhiteSpace(worldUrl);
			//insertHttp(serverUrl);
			String getServerUrl;
			if(extractWorldUrl(worldUrl, getServerUrl, g_smlsWorldUrl="")){
				if(directIP){
					g_multiUserMode = true;
					g_smlsWorldUrl = worldUrl;
					g_scene->openUrl(g_smlsWorldUrl);
				}
				else{
					if(getServerUrl.length==0){
						getServerUrl = g_openNetFile->serverUrl;
						getServerUrl += "/smls_server.php";
						g_smlsWorldUrl = worldUrl;
					}
					if(getServerUrl.length && g_smlsWorldUrl.length){
						//"http://localhost/welcome.php?name=thyme&age=43"
						//http://localhost/test_mu/server.php?url=http://localhost/my_world.smls&title=My%World
						String htmlStr;
						String requestUrl = getServerUrl;
						requestUrl += "?session=join";
						joinPHP(requestUrl, g_IPNum);
						g_multiUserMode = true;
						g_scene->openUrl(g_smlsWorldUrl);
					}
					else if(g_openNetFile->worldUrl.refValue().length && g_smlsWorldUrl.length==0){
						g_multiUserMode = false;
						g_scene->openUrl(worldUrl);
					}
				}
				g_browseMode = true;
			}
		}
	}
}


void OpenNetFile::Go::operator=(const bool& par){
	if(g_loggedIn){
		g_userAvatarSocket->logOut();
	}
	saveSmlsFirst(completeGo);
}
void openNetFile(){//Command
	openSceneTree();
	g_openNetFile->worldListRefreshed = false;
	if(g_online==false && g_stw->cp->size.x < 600){
		g_stw->cp->size = Vec2i(600, g_stw->cp->size.y);
	}
	g_stw->cp->setupNodeWindow(g_openNetFile);
	Window::activeWindow=g_stw->cp.p;
}
bool checkUrl(const String& str, int& from, int& last){
	from = str.indexOf("http://");
	if(from >= 0){
		last = markEndOfUrl(str, from);
		return true;
	}
	from = str.indexOf("www");
	if(from >= 0){
		last = markEndOfUrl(str, from);
		return true;
	}
	return false;
}
class ChatLineInfo{
public:
	int line;
	int from;
	int to;
	int urlFrom;
	int urlTo;
	int longUrlFrom;
	int longUrlTo;
	int highlightFrom;
	int highlightTo;
	int nameTo;
	bool message;
	bool user;
	ChatLineInfo():
		line(0),
		from(0),
		to(0),
		urlFrom(-1),
		urlTo(0),
		longUrlFrom(0),
		longUrlTo(0),
		highlightFrom(-1),
		highlightTo(0),
		nameTo(-1),
		message(false),
		user(false)
	{}
inline void setSubSpan2(const int spanCharFrom, const int spanCharTo);
};
M_<ChatLineInfo> g_chatLineInfo;

bool checkForVisibleHighlightedText(){
	for(int k = g_chatLineInfo.length; --k>=0;){
		if(g_chatLineInfo[k].highlightFrom >=0){
			return true;
		}
	}
	return false;
}

void ChatWin::getHighlightedString(String& str){
	if(spanLineFrom == spanLineTo){
		if(spanLineFrom != spanCharTo){
			str += g_chatLines[spanLineFrom].substring(spanCharFrom, spanCharTo-1);
		}
	}
	else{
		str += g_chatLines[spanLineFrom].substring(spanCharFrom);
		int k = spanLineFrom + 1;
		for(; k < spanLineTo; k++){
			str += "\r\n";
			str += g_chatLines[k];
		}
		str += "\r\n";
		str += g_chatLines[k].substring(0, spanCharTo-1);
	}
}
void ChatWin::copyTextCellToClipboard(bool cut){
	if(textCell.highlightPos >= 0 && textCell.highlightPos != textCell.caretPos){
		textCell.copyToClipboard(cut);
	}
	else if(checkForVisibleHighlightedText() && OpenClipboard(hWndMain)){
		EmptyClipboard(); 

		// Allocate a global memory object for the text. 
	
		String str;
		getHighlightedString(str);

		HGLOBAL hglbCopy = GlobalAlloc(GMEM_MOVEABLE, str.length + 1);
		if (hglbCopy == NULL){ 
			CloseClipboard(); 
		}
		else{
			// Lock the handle and copy the text to the buffer. 
		
			LPVOID lptstrCopy = GlobalLock(hglbCopy); 
			memcpy(lptstrCopy, str.c_str(), str.length + 1); 

			((char*)lptstrCopy)[str.length] = 0;    // null character 
			GlobalUnlock(hglbCopy); 
		
			// Place the handle on the clipboard. 
		
			SetClipboardData(CF_TEXT, hglbCopy);
			CloseClipboard();
		}
	} 
}
inline void completeEndOfLine(const int& beginInfoLine){
	int diff = (g_chatLineInfo.length - beginInfoLine) >> 1;
	for(int offset = 0; diff; offset++, diff--){
		ChatLineInfo hold = g_chatLineInfo[beginInfoLine + offset];
		g_chatLineInfo[beginInfoLine + offset] = g_chatLineInfo[g_chatLineInfo.length - 1 - offset];
		g_chatLineInfo[g_chatLineInfo.length - 1 - offset] = hold;
	}
}
inline void ChatLineInfo::setSubSpan2(const int spanCharFrom, const int spanCharTo){
	if(from < spanCharTo){
		if(from > spanCharFrom){
			highlightFrom = from;
		}
		else{
			highlightFrom = spanCharFrom;
		}
		if(to < spanCharTo){
			highlightTo = to;
		}
		else{
			highlightTo = spanCharTo;
		}
	}
}
void setSubSpan(const int spanLineFrom, const int spanLineTo, const int spanCharFrom, const int spanCharTo){
	for(int k = g_chatLineInfo.length; --k>=0;){
		ChatLineInfo& info = g_chatLineInfo[k];
		if(info.line >= spanLineFrom && info.line <= spanLineTo){

			int sCharFrom = spanCharFrom;
			if(info.line == spanLineFrom){
				if(info.to <= sCharFrom){
					continue;
				}
			}
			else{
				sCharFrom = info.from;
			}

			int sCharTo = spanCharTo;
			if(info.line != spanLineTo){
				sCharTo = info.to;
			}
			info.setSubSpan2(sCharFrom, sCharTo);
		}
	}
}
inline void checkUrl(const String& str, const int& beginInfoLine){
	int longUrlFrom = 0;
	int last = 0;
	if(checkUrl(str, longUrlFrom, last)){
		int longUrlTo = last + 1;
		for(int k = beginInfoLine; k < g_chatLineInfo.length; k++){
			ChatLineInfo& info = g_chatLineInfo[k];
			if(longUrlFrom >= info.from){
				info.longUrlFrom = longUrlFrom;
				info.urlFrom = longUrlFrom;
				if(longUrlTo <= info.to){
					info.longUrlTo = longUrlTo;
					info.urlTo = longUrlTo;
					return;
				}
				info.longUrlTo = longUrlTo;
				info.urlTo = info.to;

				for(; ++k < g_chatLineInfo.length; ){
					ChatLineInfo& info = g_chatLineInfo[k];
					info.longUrlFrom = longUrlFrom;
					info.urlFrom = info.from;
					if(longUrlTo <= info.to){
						info.longUrlTo = longUrlTo;
						info.urlTo = longUrlTo;
						return;
					}
					info.longUrlTo = longUrlTo;
					info.urlTo = info.to;
				}
			}
		}
	}
}
void ChatWin::processSpans(){
	spanLineFrom = dragLineFrom;
	spanLineTo = dragLineTo;
	spanCharFrom = dragCharFrom;
	spanCharTo = dragCharTo;

	if(spanLineFrom != spanLineTo || spanCharFrom != spanCharTo){
		if(spanLineFrom > spanLineTo){
			spanLineTo = dragLineFrom;
			spanLineFrom = dragLineTo;
			spanCharTo = dragCharFrom;
			spanCharFrom = dragCharTo;
		}
		if(spanLineFrom == spanLineTo && spanCharFrom > spanCharTo){
			int store = spanCharFrom;
			spanCharFrom = spanCharTo;
			spanCharTo = store;
		}

		setSubSpan(spanLineFrom, spanLineTo, spanCharFrom, spanCharTo);
	}
}
void processWords(int displayNumOfLines){
	g_chatLineInfo.length = 0;
	int charLineWidth = (g_cw->size.x-14)/charWidth;
	//int charLineWidth = 40;
	if(g_chatLines.length>1){
		int a = 0;
	}
	for(int k=g_chatLines.length; --k>=0;){
		String& str = g_chatLines[k];
		int from = 0;
		const int beginInfoLine = g_chatLineInfo.length;

		int marker=str.indexOf(" : ");
		if(marker>=0){
			marker++;
		}
		bool user = false;
		int i = 0;
		if(str.cmpKWord(g_userAvatarSocket->nickname, i)){
			user = true;
		}

		int subLine = 0;
		for(bool moreLines = true; moreLines; subLine++){
			ChatLineInfo info;

			if(subLine == 0){
				if(marker>=0){
					info.nameTo=marker;
				}
			}
			info.message = marker < 0;
			info.user=user;

			info.line = k;
			while(from < str.length && str[from] <= ' '){
				from++;
			}
			info.from = from;
			info.to = info.from + charLineWidth;
			if(info.to >= str.length){
				info.to = str.length;
				moreLines=false;
			}
			else{
				int i = info.to - 1;
				int e = i;
				while(e < str.length && str[e] > ' '){
					e++;
				}
				while(i > 0 && str[i] > ' '){
					--i;
				}
				i++;
				if(e - i < charLineWidth){
					info.to = i;
				}
			}

			g_chatLineInfo+=info;
			from = info.to;

			if(moreLines==false){
				checkUrl(str, beginInfoLine);
				completeEndOfLine(beginInfoLine);
			}
		}
		if(g_chatLineInfo.length > displayNumOfLines){
			g_chatLineInfo.length = displayNumOfLines;
		}
	}	
	g_cw->processSpans();
}
void printOutLines(){
	dLog("--");
	for(int k = g_chatLineInfo.length; --k>=0;){
		String out;
		ChatLineInfo& info = g_chatLineInfo[k];
		String& str = g_chatLines[info.line];
		for(int i = info.from; i<info.to; i++){
			out+=str[i];
		}
		dLog(out);
	}
	dLog("--");
}
void ChatWin::paint(){
	this->clear();

	int dstWidth = g_cw->size.x;
	int dstHeight = g_cw->size.y;

	int* dstPix = g_cw->pixBuf.getArray();
	
	//paintBlock( 0xffffff, 0, 0, dstWidth+1, dstHeight, dstWidth, dstHeight, dstPix);
	//drawHorizontalLine(0x0000ff, dstHeight-20, 0, dstWidth, dstWidth, dstHeight, g_cw.pixBuf.getArray());
	paintBlock2(0xf0f0ff, 5, dstHeight-25, dstWidth+1, dstHeight-4, dstWidth, dstHeight, dstPix);

	paintSideEdge(&iconSrcBuf[(iconSrcBufCY-1-10)*480+444], 8, 0, title_bar_height, dstHeight-10, dstWidth, dstHeight, dstPix);
	paintSideEdge(&iconSrcBuf[(iconSrcBufCY-1-10)*480+452], 9, dstWidth-9, title_bar_height, dstHeight-10, dstWidth, dstHeight, dstPix);

	blitHorizontalColorSweep(Vec2i(10, 0), Vec2i(470, 0), Vec2i(471, 0), Vec2i(dstWidth-9-10, title_bar_height));
	pasteSprite2( 0, 0, 461, 0, dstWidth, dstHeight, iconSrcBufCX, iconSrcBufCY, 10, title_bar_height, dstPix);
	pasteSprite2( dstWidth-9, 0, 471, 0, dstWidth, dstHeight, iconSrcBufCX, iconSrcBufCY, 9, title_bar_height, dstPix);

	//pasteSprite2( 5, 1, openSceneTree_x, openSceneTree_y, dstWidth, dstHeight, iconSrcBufCX, iconSrcBufCY, 24, 24, dstPix);
	drawText2("Chat", 0x0, Vec2i(5, 16), dstWidth, dstHeight, dstPix);

	blitHorizontalColorSweep(Vec2i(8, dstHeight-10), Vec2i(451, 10), Vec2i(452, 10), Vec2i(dstWidth-8-9, 10));
	pasteSprite2( 0, dstHeight-10, 444, 10, dstWidth, dstHeight, iconSrcBufCX, iconSrcBufCY, 8, 10, dstPix);
	pasteSprite2( dstWidth-9, dstHeight-10, 452, 10, dstWidth, dstHeight, iconSrcBufCX, iconSrcBufCY, 9, 10, dstPix);

	processWords(g_numOfVisChatLines);
	//printOutLines();

	bool messageFromCode = true;
	for(int k=0; k < g_chatLineInfo.length; k++){
		ChatLineInfo& info = g_chatLineInfo[k];
		const String& strRef=g_chatLines[info.line];
		int y = dstHeight-34 - k * pixSymbolCY;
		if(info.message){
			drawText2(strRef, info.from, info.to, 0xff00ff, Vec2i(7, y), dstWidth, dstHeight, dstPix);
		}
		else{
			int nameColor=0x008000;
			int lineColor=0x700000;
			int i = 0; //to be discarded
			if(info.user){
				nameColor=0xff0000;
				lineColor=0x000000;
			}
			if(info.nameTo >= 0){
				drawText2(strRef, 0, info.nameTo, nameColor, Vec2i(7, y), dstWidth, dstHeight, dstPix);
			}
			int last = 0;
			if(info.urlFrom > 0){
				int marker = 0;
				int from = info.from;
				if(info.nameTo >= 0){
					marker = from = info.nameTo;
				}
				drawText2(strRef, from, info.urlFrom, lineColor, Vec2i(7 + marker * charWidth, y), dstWidth, dstHeight, dstPix);
				int x = 7 + (info.urlFrom - info.from) * charWidth;
				drawText2(strRef, info.urlFrom, info.urlTo, 0x0000ff, Vec2i(x, y), dstWidth, dstHeight, dstPix);
				int x2 = 7 + (info.urlTo - info.from) * charWidth;
				drawHorizontalLine2(0x0000ff, y+3, x, x2, dstWidth, dstHeight, dstPix);
				drawText2(strRef, info.urlTo, info.to, lineColor, Vec2i(x2 , y), dstWidth, dstHeight, dstPix);
			}
			else{
				int marker = 0;
				int from = info.from;
				if(info.nameTo >= 0){
					marker = from = info.nameTo;
				}
				drawText2(strRef, from, info.to, lineColor, Vec2i(7 + marker * charWidth, y), dstWidth, dstHeight, dstPix);

			}
		}
		if(info.highlightFrom >= 0){
			int x = (info.highlightFrom - info.from) * charWidth + 7;
			int x2 = (info.highlightTo - info.from) * charWidth + 7;

			invertBlock(x, y + 5 - pixSymbolCY, x2, y + 5, dstWidth, dstHeight, dstPix);
		}
	}
	int x = 5;

	Vec2i valPos = textCell.cellPos + Vec2i(0, 12);
	drawText2(textCell.cellStr, textCell.firstVisChar, textCell.cellStr.length, 0, valPos, size.x, size.y, dstPix);
	textCell.highLightBlock();


	g_callBackList+=&textCell;
	if(g_settings->invertWindowColor){
		invert();
	}
	updateShader = true;
	if(textCell.cellActive==true){
		textCell.doNoThinkWink();
	}
}
bool ChatWin::CWTextCell::charInput(unsigned int wParam){
	if(wParam==VK_RETURN){
		g_userAvatarSocket->sendStringToChat(g_cw->textCell.cellStr);
		int i=0;
		if(g_cw->textCell.cellStr.cmpCppWord("pos", i)){
			g_userAvatarSocket->position.refValue().getField(g_cw->textCell.cellStr, i);
		}
		else if(g_cw->textCell.cellStr.cmpCppWord("ori", i)){
			g_userAvatarSocket->orientation.refValue().getField(g_cw->textCell.cellStr, i);
		}

		g_cw->textCell.cellStr="";
		g_cw->textCell.caretPos=0;
		g_cw->textCell.highlightPos=0;
		g_cw->textCell.updateCaret();
		win.paint();
	}
	else if(wParam==VK_BACK){
		if(deleteHighlight()){
			win.paint();
		}
		else if(caretPos>0){
			decCaretPosNH();
			cellStr.deleteChar(caretPos);
			win.paint();
		}
	}
	else if(wParam<' ')return true;
	else {
		deleteHighlight();
		cellStr.insert(wParam, caretPos);
		incCaretPosNH();
		win.paint();
	}
	onChange();
	return true;
}
void ChatWin::CWTextCell::setCellActiveFalse(void){
}
void ChatWin::CWTextCell::onChange(void){
}
void ChatWin::doDrag(Vec2i ptCursor){
	if(dragSentText){
		const Vec2i localCursor=ptCursor-pos;
		int lineLen = g_chatLineInfo.length;
		int k = lineLen - (localCursor.y - (textCell.cellPos.y - lineLen * pixSymbolCY - pixSymbolCY)) / pixSymbolCY;
		if(k >= 0 && k < lineLen){
			ChatLineInfo& info = g_chatLineInfo[k];
			dragLineTo = info.line;
			dragCharTo = info.from + (localCursor.x - 7 + charWidth / 2) / charWidth;

			int strLen = g_chatLines[dragLineTo].length;
			if(dragCharTo > strLen){
				dragCharTo = strLen;
			}
			paint();
		}
	}
	else{
		commonDoDrag(ptCursor, pos, dragPos);
	}
	g_openNetFile->chatPos = Vec2f(pos.x, pos.y);
}
void ChatWin::highLightDoubleClick(int& line, int caretPos){

	String& str = g_chatLines[line];

	if(caretPos==str.length){
		caretPos--;
	}
	int highlightPos=caretPos;	
	while(caretPos < str.length && str[caretPos] > ' '){
		caretPos++;
	}
	while(highlightPos > 0 && str[highlightPos-1] > ' '){
		highlightPos--;
	}
	dragLineFrom = line;
	dragLineTo = line;
	dragCharFrom = highlightPos;
	dragCharTo = caretPos;
	//dragSentText = true;
}
bool isDoubleLeftClick();
void ChatWin::clickUrl(const Vec2i localCursor){
	int lineLen = g_chatLineInfo.length;
	if(localCursor.y < textCell.cellPos.y * pixSymbolCY){
		int k = lineLen - (localCursor.y - (textCell.cellPos.y - lineLen * pixSymbolCY - pixSymbolCY)) / pixSymbolCY;
		if(k >= 0 && k < lineLen){
			ChatLineInfo& info = g_chatLineInfo[k];
			if(info.urlFrom > 0){
				int x = 7 + (info.urlFrom - info.from) * charWidth;
				int x2 = 7 + (info.urlTo - info.from) * charWidth;

				if(localCursor.x >= x && localCursor.x < x2){
					String url = g_chatLines[info.line].substring(info.longUrlFrom, info.longUrlTo - 1);
					openWithHtmlBrowser(url, g_settings->htmlBrowser.refValue());
					return;
				}
			}
			else if(isDoubleLeftClick()){
				highLightDoubleClick(info.line, info.from + (localCursor.x - 7 + charWidth / 2) / charWidth);
			}
			else{
				dragLineFrom = info.line;
				dragLineTo = dragLineFrom;
				dragCharFrom = info.from + (localCursor.x - 7 + charWidth / 2) / charWidth;
				int strLen = g_chatLines[dragLineTo].length;
				if(dragCharFrom > strLen){
					dragCharFrom = strLen;
				}
				dragCharTo = dragCharFrom;
				dragSentText = true;
				setDrag(this);
			}
		}
	}
}
void ChatWin::clearHighLight(){
	dragSentText = false;
	dragLineFrom = 0;
	dragLineTo = 0;
	dragCharFrom = 0;
	dragCharTo = 0;

	spanLineFrom = 0;
	spanLineTo = 0;
	spanCharFrom = 0;
	spanCharTo = 0;
}

bool ChatWin::click(const Vec2i ptCursor){
	if(ptCursor >= pos && ptCursor < pos+size){
		activeWindow = this;
		if(tugBorder){
			tugBorder->window = this;
			setDrag(tugBorder);
			return true;
		}
		const Vec2i localCursor=ptCursor-pos;

		if(	localCursor.x >= textCell.cellPos.x && localCursor.x < textCell.cellPos.x + textCell.cellSize.x &&
			localCursor.y >= textCell.cellPos.y && localCursor.y < textCell.cellPos.y + textCell.cellSize.y
			){
			int newCaretPos = (localCursor.x + charWidth / 2 - (textCell.cellPos.x)) / charWidth;
			{
				textCell.cellActive = false;

				textCell.highlightPos = textCell.caretPos = newCaretPos + textCell.firstVisChar;
				textCell.updateCaret();
				clearHighLight();

				textCell.checkForDoubleLeftClick();

				setDrag(&textCell);
				paint();
				return true;
			}
		}
		textCell.cellActive = false;
		textCell.highlightPos = textCell.caretPos;
		textCell.updateCaret();
		if(localCursor.y < title_bar_height){
			setDrag(this);
			return true;
		}

		clickUrl(localCursor);
		paint();
		return true;
	}
	else{
		return false;
	}
}

bool ChatWin::hover(const Vec2i ptCursor){
	if(
		ptCursor.x>=pos.x &&
		ptCursor.x< pos.x+size.x &&
		ptCursor.y>=pos.y &&
		ptCursor.y< pos.y+size.y
		)
	{

		const Vec2i localCursor=ptCursor-pos;

		if(localCursor.x < 4){
			if(localCursor.y < 4){
				loadCursor(&tB_TopLeft, g_sizeNWSECursor);
			}
			else if(localCursor.y >= size.y - 4){
				loadCursor(&tB_BottomLeft, g_sizeNESWCursor);
			}
			else{
				loadCursor(&tB_Left, g_sizeWECursor);
			}
		}
		else if(localCursor.x >= size.x - 4){
			if(localCursor.y < 4){
				loadCursor(&tB_TopRight, g_sizeNESWCursor);
			}
			else if(localCursor.y >= size.y - 4){
				loadCursor(&tB_BottomRight, g_sizeNWSECursor);
			}
			else{
				loadCursor(&tB_Right, g_sizeWECursor);
			}
		}
		else if(localCursor.y < 4){
			loadCursor(&tB_Top, g_sizeNSCursor);
		}
		else if(localCursor.y >= size.y - 4){
			loadCursor(&tB_Bottom, g_sizeNSCursor);
		}
		else{ 
			int lineLen = g_chatLineInfo.length;
			if(localCursor.y < textCell.cellPos.y * pixSymbolCY){
				int k = lineLen - (localCursor.y - (textCell.cellPos.y - lineLen * pixSymbolCY - pixSymbolCY)) / pixSymbolCY;
				if(k < lineLen && k >= 0){
					ChatLineInfo& info = g_chatLineInfo[k];
					if(info.urlFrom > 0){
						int x = 7 + (info.urlFrom - info.from) * charWidth;
						int x2 = 7 + (info.urlTo - info.from) * charWidth;

						if(localCursor.x >= x && localCursor.x < x2){
							changeCursor(g_handCursor);
							return true;
						}
					}
				}
			}
			tugBorder=0;
			changeCursor(g_arrowCursor);
			buttonHover(ptCursor-pos);
		}
		return true;
	}
	return false;
}

void ChatWin::setWinSize(Vec2i& raw){
	if(raw.x < 400){
		raw.x = 400;
	}
	if(raw.y < 100){
		raw.y = 100;
	}
	pixBuf.length = raw.x * raw.y;

	textCell.cellPos=Vec2i(7, raw.y - 26);
	textCell.cellSize=Vec2i(raw.x - textCell.cellPos.x - 8, 20);
	g_numOfVisChatLines=(raw.y - g_beginYChatOffset - title_bar_height - 25) / pixSymbolCY;
	if(false == initWindow()){
		throw;
	}
	g_openNetFile->chatSize = Vec2f(size.x, size.y);
}
void setGlobalUrlPointers(Node* globalUrlNode, M_<SceneTreeListItem>& sceneTreeList){
	for(int k=sceneTreeList.length; --k>=0;){
		if(Download* download = dynamic_cast<Download*>(sceneTreeList[k].node.p)){
			if(TextureEffect* textureEffect=dynamic_cast<TextureEffect*>(download)){
				textureEffect->urlPathNode=globalUrlNode;
			}
			else if(ImageTexture* imageTexture=dynamic_cast<ImageTexture*>(download)){
				imageTexture->urlPathNode=globalUrlNode;
			}
		}
	}
}
bool Scene::openUrl(String str){
	worldUrl = str;
	dLog("Scene::openUrl");
	HINTERNET& _hInternetOpenUrl=*reinterpret_cast<HINTERNET*>(&hInternetOpenUrl);
	if((_hInternetOpenUrl=InternetOpenUrl( g_hInternetSession, str.c_str(),NULL,0,0,0))!=NULL){
		g_online=true;
		g_viewer=true;
		g_downloadQue.length=0;
		g_downloadQue+=this;
		return true;
	}
	msG("failed to open Scene URL", MB_ICONEXCLAMATION);
	dLog("*Scene::openUrl failed!*");
	return false;
}
bool setNetSave(){
	for(int k = g_scene->sceneTreeList.length; --k>=0;){
		if(WorldInfo* worldInfo=dynamic_cast<WorldInfo*>(g_scene->sceneTreeList[k].node.p)){
			return worldInfo->netSave;
		}
	}
	return false;
}
bool Scene::download(){
	HINTERNET& _hInternetOpenUrl=*reinterpret_cast<HINTERNET*>(&hInternetOpenUrl);
	const int newChunkSize=0x1000;
	DWORD numberOfBytesRead=0;

	int oldlength=internetBuf.length;
	internetBuf.length+=newChunkSize;
	BOOL readFlag=InternetReadFile(_hInternetOpenUrl, internetBuf.getArray()+oldlength, newChunkSize, &numberOfBytesRead);
	internetBuf.length=oldlength+numberOfBytesRead;
	Browser::print("Scene bytes loaded " + String(internetBuf.length));

	if(readFlag==TRUE && numberOfBytesRead==0){
		dLog("downloaded");
		
		InternetCloseHandle(
			_hInternetOpenUrl
		);
		_hInternetOpenUrl=0;
		internetBuf+=0;

		String file=g_undoDir;
		file += "temp.smls";
		internetBuf.write(file);
		internetBuf.freeMem();


		g_lookupProtoInterface.length=g_userProtoBegin;
		nodeLookUpStr.length=g_userProtoBegin;
		containerFieldLookUpStr.length=g_userProtoBegin;

		initCompile();
		Browser::createSceneFromFile(file);
		DeleteFile(file.c_str());
		g_wireFrameMode = g_scene->mode & WIREFRAME_MODE;
		genNodeList();
		netSave=setNetSave();
		setGlobalUrlPointers(this, g_scene->sceneTreeList);

		removeTypeFromList < Button>(g_callBackList); //must before selectFirstPart
//		removeTypeFromList < DownButton>(g_callBackList); //must before selectFirstPart
		if(g_remVersionLoaded < 2.093-.00001){
			g_stw->selectFirstPart();
		}
		decodeFunctions(g_loadStr);
		genNodeList();
		postLoadProcessNodes();

		g_scene->postLoadProcessCollapse();

		resetLooseEnds();
		Viewpoint::gotoFirstViewpoint();
		setTitleName();
		initializeScript();

		g_currentFilename = g_seamlessDir;
		String filename = g_smlsWorldUrl;
		for(int i = filename.length; --i >= 0;){
			if(filename[i]=='/'){
				g_currentFilename += filename.substr(i+1);
				break;
			}
		}
		g_forceSaveAs = true;

		setSmlsSaved();
		
		if(g_stw->isVisible()){
			g_stw->paint();
			Window::paintAllControlPanels();
		}
		else{
			Window::closeWindow(g_stw->cp.p);
		}
		if(g_multiUserMode){
			initMultiUserMode();
		}
		return true;
	}
	return false;
}

bool AvatarSocket::openUrl(){
	dLog("AvatarSocket::openUrl");
	HINTERNET& _hInternetOpenUrl=*reinterpret_cast<HINTERNET*>(&hInternetOpenUrl);
	if((_hInternetOpenUrl=InternetOpenUrl( g_hInternetSession, avatarUrl.refValue().c_str(),NULL,0,0,0))!=NULL){
		g_scene->children += this;
		dLog("g_scene->children += this AvatarSocket");
		g_stw->genNodeList();
		SendMessage(hWndMain, WM_PAINT,	0,	0	);
		//SendMessage(hSceneTreeWnd, WM_PAINT, 0, 0);
		g_stw->paint();

		return true;
	}
	for(int i=0; i<g_downloadQue.length-1; i++){
		g_downloadQue[i]=g_downloadQue[i+1];
	}
	g_downloadQue.length--;

	msg("failed to open AvatarSocket URL", MB_OK|MB_ICONEXCLAMATION);
	dLog("*AvatarSocket::openUrl failed!*");
	return false;
}
bool AvatarSocket::download(){
	if(hInternetOpenUrl==0){
		openUrl();
		return false;
	}
	HINTERNET& _hInternetOpenUrl=*reinterpret_cast<HINTERNET*>(&hInternetOpenUrl);
	const int newChunkSize=0x1000;
	DWORD numberOfBytesRead=0;


	int oldlength=internetBuf.length;
	internetBuf.length+=newChunkSize;
	BOOL readFlag=InternetReadFile(_hInternetOpenUrl, internetBuf.getArray()+oldlength, newChunkSize, &numberOfBytesRead);
	internetBuf.length=oldlength+numberOfBytesRead;
	Browser::print("AvatarSocket bytes loaded " + String(internetBuf.length));

	if(readFlag==TRUE && numberOfBytesRead==0){
		dLog("AvatarSocket downloaded");
		collapsed=true;
		
		InternetCloseHandle(
			_hInternetOpenUrl
		);
		_hInternetOpenUrl=0;
		internetBuf+=0;


		String file=g_undoDir;
		file+="temp.smls";
		internetBuf.write(file);
		g_loadStr = readFile(file);
		DeleteFile(file.c_str());

		internetBuf.freeMem();//**would be good to have this feature since we no longer will ever need this mem again

		ScenePtr importScene = new Scene;
		M_Node& rawLoad=importScene->children;
////////
		Function::selectedLine = 0;//create new code lines
		initCompile();
		createSmlsFromString(rawLoad, g_loadStr);
		for(int k=rawLoad.length; --k>=0;){
			if(GestureGroup* gg = dynamic_cast<GestureGroup*>(rawLoad[k].p)){
				gestureGroup=gg;
			}
		}

		decodeFunctions(g_loadStr);
		importScene->genNodeList();
		netSave=setNetSave();
		setGlobalUrlPointers(this, g_scene->sceneTreeList);
		M_<SceneTreeListItem> importSceneTreeList=g_scene->sceneTreeList;
		importScene->createDefList();
		g_scene->postLoadProcessNodes();

		importScene->postLoadProcessCollapse();
		g_scene->genNodeList();
		{
			M_<SceneTreeListItem>& existingSceneTreeList=g_scene->sceneTreeList;
			M_Node existingDefList;
			g_scene->createDefList();
			checkForNameClash(importScene->defList, g_scene->defList);
			children.refValue()+=rawLoad;
		}
		finishImportNodesFromFile();
///////////


//		createSmlsFromString(children, g_loadStr);
//		postLoadProcessNodes(g_scene->sceneTreeList);

//		g_stw->genNodeList();

		//SendMessage(hSceneTreeWnd, WM_PAINT, 0, 0);
		g_stw->paint();

		return true;
	}
	return false;
}
void AvatarSocket::renderNode(Matrix parentMatrix, Matrix catRotMatrix){

	Matrix thisMatrix=rotationMatrix(orientation)*translationMatrix(position)*parentMatrix;
	catRotMatrix = rotationMatrix(orientation) * catRotMatrix;

	for(int k=0;k<children.refValue().length;k++){
		children.refValue()[k]->renderNode(thisMatrix, catRotMatrix);
	}
}
Node* newAvatarSocket(){
	return new AvatarSocket;
}
void AvatarSocket::outPovRay(String& str, M_<Effect*>& effectList, Matrix matrix, Matrix rotMatrix){
	const M_Node& children = this->children.refValue();
	for(int k=0; k<children.length; k++){
		if(PovRay* povRay=dynamic_cast<PovRay*>(children[k].p)){
			povRay->outPovRay(str, effectList, matrix, rotMatrix);
		}
	}
}
void AvatarSocket::outObj(String& str, String& mtl, int& vidx, int& vtidx, M_<Effect*>& effectList, Matrix matrix, Matrix rotMatrix){
	const M_Node& children = this->children.refValue();
	for(int k=0; k<children.length; k++){
		if(PovRay* povRay=dynamic_cast<PovRay*>(children[k].p)){
			povRay->outObj(str, mtl, vidx, vtidx, effectList, matrix, rotMatrix);
		}
	}
}
void AvatarSocket::outWebGL(String& str, int& vidx, int& vtidx, M_<Effect*>& effectList, Matrix matrix, Matrix rotMatrix){
	const M_Node& children = this->children.refValue();
	for(int k=0; k<children.length; k++){
		if(PovRay* povRay=dynamic_cast<PovRay*>(children[k].p)){
			povRay->outWebGL(str, vidx, vtidx, effectList, matrix, rotMatrix);
		}
	}
}
void AvatarSocket::getUrlPath(String& url){
	url=avatarUrl;
	int j = url.length;
	while(url[--j]!='/');
	url=url.substring(0, j);
}
void Scene::getUrlPath(String& url){
	url=this->worldUrl;
	int j = url.length;
	while(url[--j]!='/');
	url=url.substring(0, j);
}
bool getWorldsOpenPage(String& ip, String& page){
	SOCKET itvSocket;
    HOSTENT *hostEntry;

	hostEntry = gethostbyname(ip.c_str());
	if (WSAGetLastError() != 0) {
        return false;
	}
	dLog("gethostbyname ok");
	
	itvSocket = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
	if (itvSocket == INVALID_SOCKET) {
		return false;
	}
	
	SOCKADDR_IN saServer;
	saServer.sin_family = AF_INET;
	saServer.sin_addr = *((LPIN_ADDR)*hostEntry->h_addr_list);
	saServer.sin_port = htons(PORT);

	// connect to the server
	if(SOCKET_ERROR == connect(itvSocket, (LPSOCKADDR)&saServer, sizeof(saServer))){
		dLog("connect(itvSocket, (LPSOCKADDR)&saServer, sizeof(saServer)) ");
		closesocket(itvSocket);
		return false;
	}
	dLog("connect ok");

	String str=String("GET ");
	str+="http://";
	str+=ip;
	//str+=filename;
	str+=" HTTP/1.1\r\nHost: ";
	str+=ip;
str += "\r\nUser-Agent: Seamless3d";

	str+="\r\n\r\n\r\n";
	if( SOCKET_ERROR == send(itvSocket, str.c_str(), str.length, 0)){
		dLog("send(itvSocket, str.c_str(), str.length, 0)");
		closesocket(itvSocket);	
		return false;
	}
	dLog("send GET HTTP ok");
	M_<char> inBuf;

	bool flag = socketToBuf(itvSocket, inBuf);
	closesocket(itvSocket);
	if(flag == false){
		return false;
	}
	inBuf+=0;
	page+=inBuf.getArray();
	int i = page.indexOf("\r\n\r\n");
	page.stb(i);
	page=page.substring(i);

	dLog("downloaded getWorldsOpenPage");

	return true;
}
/*
	char szBuf[1024];

	//Sleep(1000*3);
	int txtLength = recv(itvSocket, szBuf, sizeof(szBuf)-1, 0);

	if (txtLength == INVALID_SOCKET)	{
		dLog("recv() INVALID_SOCKET");
		closesocket(itvSocket);
		return false;
	}
	else if (txtLength == 0)	{
		dLog("recv() 0");
		closesocket(itvSocket);
		return false;
	}
	szBuf[txtLength]=0;
	page+=szBuf;

	bool chunked = false;
	if(page.indexOf("Transfer-Encoding: chunked") >= 0){
		chunked = true;
	}

	int i = page.indexOf("Content-Length:") + sizeof("Content-Length:");
	page.stb(i);
	int contentLen = getInt(page,i);




	i = page.indexOf("\r\n\r\n");
	page.stb(i);
	if(chunked){
		page.stw(i);
		page.stb(i);
	}
	page=page.substring(i);
	i = page.indexOf("\r\n");
	if(i > 0){
		page = page.substring(0, i - 1);
	}

	while(page.length < contentLen){
		int limit = contentLen - page.length;
		if(limit >= sizeof(szBuf)){
			limit = sizeof(szBuf)-1;
		}
		txtLength = recv(itvSocket, szBuf, sizeof(szBuf)-1, 0);
		szBuf[txtLength]=0;
		page+=szBuf;
	}
*/

void closeBrack(String& html, int& i){
	while(!html.cmpCppWord(">",i)){
		i++;
	}
}
void ignoreIMGTag(String& str, int& i){
	str.stb(i);
	if(str[i] == '<'){
		i++;
		str.stb(i);
		if(str.cmpCppWord("img",i)){
			closeBrack(str, i);
		}
	}
}
void ignoreBTag(String& str, int& i){
	str.stb(i);
	if(str[i] == '<'){
		i++;
		str.stb(i);
		if(str.cmpCppWord("b",i)){
			closeBrack(str, i);
		}
	}
}
void ignoreCloseBTag(String& str, int& i){
	str.stb(i);
	if(str[i] == '<'){
		i++;
		str.stb(i);
		if(str[i] == '/'){
			i++;
			str.stb(i);
			if(str.cmpCppWord("b",i)){
				closeBrack(str, i);
			}
		}
	}
}
void ignoreBRTag(String& str, int& i){
	str.stb(i);
	if(str[i] == '<'){
		i++;
		str.stb(i);
		if(str[i] == '/'){
			i++;
		}
		str.stb(i);
		if(str.cmpCppWord("br",i)){
			closeBrack(str, i);
		}
	}
}
void ignoreCloseATag(String& str, int& i){
	str.stb(i);
	int rem_i = i;
	if(i < str.length && str[i] == '<'){
		i++;
		str.stb(i);
		if(i < str.length && str[i] == '/'){
			i++;
			str.stb(i);
			if(str.cmpCppWord("a",i)){
				closeBrack(str, i);
				return;
			}
		}
	}
	i = rem_i;
	return;
}
bool ignoreATag(String& str, int& i){
	str.stb(i);
	if(i < str.length && str[i] == '<'){
		i++;
		str.stb(i);
		if(str.cmpCppWord("a",i)){
			closeBrack(str, i);
			ignoreIMGTag(str, i);
			ignoreCloseATag(str, i);
			return true;
		}
	}
	return false;
}

void getTitle(String& str, int& i){
	str.stb(i);
	OpenNetFile::worldListTitles+="";
	String& t = OpenNetFile::worldListTitles[OpenNetFile::worldListTitles.length-1];
	while(str[i]!='<'){
		t+=str[i++];
	}
}
void getOccupants(String& str, int& i){
	str.stb(i);
	OpenNetFile::worldListOccupants+="";
	String& o = OpenNetFile::worldListOccupants[OpenNetFile::worldListOccupants.length-1];
	while(str[i]!='<'){
		o+=str[i++];
	}
}
bool decodeATag(String& str, int& i){
	str.stb(i);
	if(str[i] == '<'){
		i++;
		str.stb(i);
		if(str.cmpCppWord("a",i)){
			str.stb(i);
			if(str.cmpCppWord("href",i)){
				str.stb(i);
				if(str[i] == '='){
					str.stb(++i);
					if(str[i] == '\"'){
						OpenNetFile::worldListUrls += str.fromQuotes(i);
						closeBrack(str, i);
						getTitle(str, i);
						closeBrack(str, i);
					}
				}
			}
		}
	}
	return false;
}
int OpenNetFile::getWorldList(String& str){

	worldListUrls.length=0;
	worldListTitles.length=0;
	worldListOccupants.length=0;

	if(processServerUrl() == false){
		String requestUrl = g_openNetFile->serverUrl;
		//requestUrl += "/smls_server.php?session=join";
		requestUrl += "/smls_server.php";
		joinPHP(requestUrl, g_IPNum);
	}

	getWorldsOpenPage(g_IPNum, str);

	int i = 0;
	int k = 0;
	if(str.length > 0){
		while(ignoreATag(str, i)){
			ignoreBTag(str, i);
			decodeATag(str, i);
			ignoreCloseBTag(str, i);
			getOccupants(str, i);
			ignoreBRTag(str, i);
			k++;
		}
		if(k == 0 && str == "The Seamless3d Server is running but no worlds are open."){
			return -2;
		}
		else {
			return k;
		}
	}
	return -1;
}
int OpenNetFile::clickUrl(const Vec2i localCursor){
	if(worldList==false){
		return 0;
	}
	if(worldListRefreshing==true || worldListRefreshed==false){
		return -1;
	}
	int y = 16*4+3;
	if(localCursor.y < y-16){
		return 0;
	}
	for(int k = 0; k<worldListUrls.length; k++){
		int x = 10+worldListTitles[k].length*8;
		if(
			localCursor.y <= y && localCursor.y > y-16 &&
			localCursor.x >= 10 && localCursor.x < x
			){
			worldUrl.refValue() = worldListUrls[k];
			go=true;
			return k;
		}
		y+=16;
	}
	return -1;
}
void OpenNetFile::hoverUrls(const Vec2i localCursor){
	if(worldList == false || worldListRefreshing == true || worldListRefreshed == false){
		return;
	}
	int y = 16 * 4 + 3;
	
	for(int k = 0; k<worldListUrls.length; k++){
		int x = 10+worldListTitles[k].length*8;
		if(
			localCursor.y <= y && localCursor.y > y-16 &&
			localCursor.x >= 10 && localCursor.x < x
			){
			changeCursor(g_handCursor);
			return;
		}
		y+=16;
	}
	changeCursor(g_arrowCursor);
	return;
}
void OpenNetFile::WorldList::operator=(const bool& par){
	refValue()=par;
	home.worldListRefreshed=false;
}
void WorldInfo::NetSave::operator=(const bool& par){
	if(g_browseMode==false){
		refValue()=par;
	}
}
bool getAvatarSocketNetSave(Node* node){
	if(g_browseMode==false){
		return true;
	}
	return dynamic_cast<AvatarSocket*>(node)->netSave;
}

void commonSearchWebFromCell(int k){
	const String& searchEngine = g_settings->searchEngine[k];
	int i = searchEngine.indexOf("|");
	String str = searchEngine.substring(++i);


	int selLen = abs(g_cw->textCell.caretPos-g_cw->textCell.highlightPos);
	int selFrom = g_cw->textCell.highlightPos;
	if(g_cw->textCell.caretPos < selFrom){
		selFrom = g_cw->textCell.caretPos;
	}
	str+=g_cw->textCell.cellStr.substr(selFrom, selLen);

	for(int i = str.length; --i>=0;){
		if(str[i]==' '){
			str[i]='+';
		}
	}
	openWithHtmlBrowser(str, g_settings->htmlBrowser.refValue());
}
void searchWebFromCell0(){
	commonSearchWebFromCell(0);
}
void searchWebFromCell1(){
	commonSearchWebFromCell(1);
}
void searchWebFromCell2(){
	commonSearchWebFromCell(2);
}
void commonSearchWeb(int k){
	const String& searchEngine = g_settings->searchEngine[k];
	int i = searchEngine.indexOf("|");
	String str = searchEngine.substring(++i);

	g_cw->getHighlightedString(str);

	for(int i = str.length; --i>=0;){
		if(str[i]==' '){
			str[i]='+';
		}
	}
	openWithHtmlBrowser(str, g_settings->htmlBrowser.refValue());
}
void searchWeb0(){
	commonSearchWeb(0);
}
void searchWeb1(){
	commonSearchWeb(1);
}
void searchWeb2(){
	commonSearchWeb(2);
}

void cutChat(){
}
void TouchSensor::sendClickToServer(){
	static M_<char> buf;
	if(g_multiUserMode){
		buf.length=0;
		insertInstruction(buf, TOUCH_SENSOR, g_userAvatarSocket->id);
		insertShort(buf, 0);
		insertStr(buf, name);
		g_userAvatarSocket->sendToServer(buf);
	}
}

