/* Seamless3d

Copyright (C) 2001-2017 Graham John Perrett (AKA thyme) thyme@seamless3d.com www.seamless3d.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this
software and associated documentation files (the "Software"), to deal in the Software
without restriction, including without limitation the rights to use, copy, modify,
merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to the following
conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT
OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.
*/


#define activeColor 0xfed887

typedef __int64 int64_t;

class WinShader;
class Window;
class Menu;
class SceneTree;
class ControlPanel;
class AnimSlider;
class AnimPeriod;
class AnimControlPanel;
class VideoControlPanel;
class WaveToPolyControlPanel;
class ChatWin;
class PaletteCP;

extern const char* tip_scroll_up_a_line;
extern const char* tip_scroll_down_a_line;
extern const char* tip_scroll_up_a_page;
extern const char* tip_scroll_down_a_page;
extern const char* tip_drag_to_scroll;
extern const char* tip_close_scene_tree_window;



class Drag{
public:
	virtual void doDrag(const Vec2i ptCursor) = 0;
	virtual void endDrag(const Vec2i ptCursor);
	static bool triggered;
	static void setDrag(Drag* par);
	static Vec2i beginMousePos;
	static Vec2i dragPos;
};
extern Drag* g_drag;

enum buttonSrcPos{
	newSmls_x = 13*24,
	newSmls_y = 0,
	openSmls_x = 14*24,
	openSmls_y = 0,
	openNetFile_x = 54*4+18+24*6,
	openNetFile_y = 0,
	saveSmls_x = 54*4+18+24*5,
	saveSmls_y = 0,
	loadPartFromFile_x = 24*8,
	loadPartFromFile_y = 24,
	savePartAs_x = 24*9,
	savePartAs_y = 24,
	openSceneTree_x = 0,
	openSceneTree_y = 24*2,
	remember_x = 176,
	remember_y = 176,
	undo_x = 54*4+24,
	undo_y = 0,
	redo_x = undo_x+24,
	redo_y = 0,
	resetViewOrientation_x = 24*16,
	resetViewOrientation_y = 24*2,
	resetViewPosition_x = 24*17,
	resetViewPosition_y = 24*2,

	hiddenPart_x=413,
	hiddenPart_y=12,
	hiddenPart_w=13,
	hiddenPart_h=3,

	centerJoint_x = 128,
	centerJoint_y = 176,

	centerViewToControlPoint_x = 168,
	centerViewToControlPoint_y = 48,
	centerViewToPivotPoint_x = 192,
	centerViewToPivotPoint_y = 48,

	delete_x = 0,
	delete_y = 192,

	compileScaleRotation_x = 24*1,
	compileScaleRotation_y = 24*3,
	togglePerspective_x = 24*13,
	togglePerspective_y = 24*3,
	freezeTime_x = 24*15,
	freezeTime_y = 24*2,
	help_x = 24*12,
	help_y = 24*2,

	breakVertexJoin_x = 24*6,
	breakVertexJoin_y = 24*3,
	breakColorJoin_x = 24*7,
	breakColorJoin_y = 24*3,
	deleteTri_x = 24*6,
	deleteTri_y = 24,
	cutDownFromTheCursor_x = 24*2,
	cutDownFromTheCursor_y = 24*2,
	cutToTheLeftOfTheCursor_x = 24*1,
	cutToTheLeftOfTheCursor_y = 24*2,
	cutUpFromTheCursor_x = 24*3,
	cutUpFromTheCursor_y = 24*2,
	copyAndPasteMirroredPart_x = 24*4,
	copyAndPasteMirroredPart_y = 24*2,
	joinVertexKeepDestColor_x = 24*9,
	joinVertexKeepDestColor_y = 24*3,
	joinVertexKeepBothColors_x = 24*8,
	joinVertexKeepBothColors_y = 24*3,
	transferCloseVerticesJoin_x = 24*11,
	transferCloseVerticesJoin_y = 24*3,
	joinCloseVerticesMacro_x = 24*12,
	joinCloseVerticesMacro_y = 0,
	joinMatchingColors_x = 24*2,
	joinMatchingColors_y = 24*3,
	colorVertex_x = 24*13,
	colorVertex_y = 24*2,
	pasteColorToTriangle_x = 288,
	pasteColorToTriangle_y = 72,
	colorPart_x = 24*3,
	colorPart_y = 24*3,
	replaceColorPart_x = 24*4,
	replaceColorPart_y = 24*3,
	replaceColorSkeleton_x = 24*5,
	replaceColorSkeleton_y = 24*3,

	transferSingleVertex_x = 24*11,
	transferSingleVertex_y = 24*2,
	zeroVetexX_x = 24*14,
	zeroVetexX_y = 24*2,
	replaceWithCopyOfRememberedPartMirrorX_x = 24*5,
	replaceWithCopyOfRememberedPartMirrorX_y = 24*2,
	pasteAddCopyOfMirrorX_x = 24*6,
	pasteAddCopyOfMirrorX_y = 24*2,
	swapDivide_x = 320,
	swapDivide_y = 128,
	mirrorTriangleImmediate_x = 24*10,
	mirrorTriangleImmediate_y = 24*3,
	splitTriangleImmediate_x = 24*10,
	splitTriangleImmediate_y = 24*2,
	projectSingleVertexOntoPoly_x = 24*18,
	projectSingleVertexOntoPoly_y = 24*2,

	quickTriangle_x = 24*18,
	quickTriangle_y = 24*1,

	findThinTriangle_x = 24*15,
	findThinTriangle_y = 24,
	resetRotations_x = 24*13,
	resetRotations_y = 24,
	invertBackgroundColor_x = 24*16,
	invertBackgroundColor_y = 24,
	toggleWireframeMode_x = 24*17,
	toggleWireframeMode_y = 24,

	dropSiblingAbove_x = 160,
	dropSiblingAbove_y = 160,
	dropSiblingBelow_x = 352,
	dropSiblingBelow_y = 96,
	dropSiblingInsert_x = 158,
	dropSiblingInsert_y = 176,
	dropSiblingReplace_x = 352,
	dropSiblingReplace_y = 128,
	dropChild_x = 352,
	dropChild_y = 112,
	dropSourceChild_x = 352,
	dropSourceChild_y = 144,
	dropDestChild_x = 192,
	dropDestChild_y = 176,
	branchMenuArrow_x = 403,
	branchMenuArrow_y = 13,

	sphereBuildIcon_x = 16,
	sphereBuildIcon_y = 176,
	cylinderBuildIcon_x = 32,
	cylinderBuildIcon_y = 176,
	coneBuildIcon_x = 48,
	coneBuildIcon_y = 175,
	rectangleBuildIcon_x = 64,
	rectangleBuildIcon_y = 176,
	torusBuildIcon_x = 80,
	torusBuildIcon_y = 176,
	boxBuildIcon_x = 96,
	boxBuildIcon_y = 176,
	bezierLatheBuildIcon_x = 112,
	bezierLatheBuildIcon_y = 176,

	colorSweepIcon_x=16,
	colorSweepIcon_y=94,

	copyPartIcon_x=48,
	copyPartIcon_y=96,

	placeIcon_x=64,
	placeIcon_y=160,

	floorBuild_x = 96,
	floorBuild_y = 24,

	averagePivotPoint_x = 96+24,
	averagePivotPoint_y = 24,

	exitWindow_x = 144,
	exitWindow_y = 128,

	title_bar_height = 25,

	charWidth = 8,
	iconSrcBufCX = 480,
	iconSrcBufCY = 337,

	UPPER_TOOLBAR_CX = (786 + 24 * 3),
	LOWER_TOOLBAR_CX = (786 + 24 * 3),
	SINGLE_BAR_CY = 24,
	TOOLBAR_CY = SINGLE_BAR_CY,
	TOOLBAR_BUTTON_OFFSET_X = 8,
	BORDER_DIST = 1,
	TOOLBAR_CY_ALL = (TOOLBAR_CY+BORDER_DIST*2),
};
extern const unsigned iconSrcBuf[];

class ShortCutKey{
public:
	unsigned keyCode;
	void (*function)();
	ShortCutKey():
		keyCode(0),
		function(0)
	{}
	ShortCutKey(unsigned i_keyCode, void (*i_function)()):
		keyCode(i_keyCode),
		function(i_function)
	{}
};
class CallBack{
public:
	virtual void doCallBack(){};
	virtual ~CallBack()=0;
};
extern M_<CallBack*>g_callBackList;

class Button : public CallBack{
protected:
	static double beginHoverTime;
	static bool open;
	static const float tipTimeOut;
public:

	static Button* getHoverButton();
	static int repeatCounter;


	int refCount;
	Vec2i bSize;
	Vec2i iconSize;
	Vec2i dstPos;
	Vec2i srcPos;
	const int baseCommand;
	int command;
	Window& win;

	const char* tip;

	Button(Window& i_win):
		tip(0),
		baseCommand(0),
		command(0),
		refCount(0),
		win(i_win)
	{}
	Button(Vec2i i_srcPos, Window& i_win):
		srcPos(i_srcPos),
		tip(0),
		baseCommand(0),
		command(0),
		refCount(0),
		win(i_win)
	{}
	Button(Vec2i i_srcPos, int i_command, Window& i_win):
		srcPos(i_srcPos),
		tip(0),
		baseCommand(i_command),
		command(i_command),
		refCount(0),
		win(i_win)
	{}
	Button(Vec2i i_bSize, Vec2i i_srcPos, const char* i_tip, int i_command, Window& i_win, int refCountValue=0):
		bSize(i_bSize),
		iconSize(i_bSize),
		srcPos(i_srcPos),
		tip(i_tip),
		baseCommand(i_command),
		command(i_command),
		refCount(refCountValue),
		win(i_win)
	{}
	Button(Vec2i i_bSize, Vec2i i_dstPos, Vec2i i_srcPos, const char* i_tip, int i_command, Window& i_win, int refCountValue=0):
		bSize(i_bSize),
		iconSize(i_bSize),
		dstPos(i_dstPos),
		srcPos(i_srcPos),
		tip(i_tip),
		baseCommand(i_command),
		command(i_command),
		refCount(refCountValue),
		win(i_win)
	{}
	Button(Vec2i i_bSize, Vec2i iconSize, Vec2i i_dstPos, Vec2i i_srcPos, const char* i_tip, int i_command, Window& i_win, int refCountValue=0):
		bSize(i_bSize),
		iconSize(iconSize),
		dstPos(i_dstPos),
		srcPos(i_srcPos),
		tip(i_tip),
		baseCommand(i_command),
		command(i_command),
		refCount(refCountValue),
		win(i_win)
	{}
	void incRefCount(){
		if(this!=0){ //is this line really needed?
		   ++refCount;
		}
	}
	unsigned decRefCount(){
		--refCount;
		if(refCount<0) throw this; //is this line really needed?
		return refCount;
	}
	virtual const asmPointer getID()const{return 0;};
	virtual void workOutSubIndex(const Vec2i ptCursor){};
	virtual Vec2i getSimClickPos(int subButtonIdx)=0;
	virtual void paint()=0;
	virtual void action()=0;
	static String str;
	bool initHover(const Vec2i ptCursor);
	virtual bool onHoverMove(const Vec2i ptCursor);
	virtual void doCallBack();// while hover;
	void onButtonUp();
};

class DownButton : public CallBack {
	static Button* button;
	static double lastTime;
	static DownButton* downButton;
public:
	void doCallBack();
	void static setDownButton(Button* i_button);
	static Button* getDownButton();
};

class SingleButton : public Button{
public:
	SingleButton(Vec2i size, Vec2i srcPos, const char* tip, const char* name, void (*function)(), Window& win):
		Button(size, srcPos, tip, 0, win),
		iconSize(size),
		name(name),
		function(function),
		offSetIcon(0)
	{}
	SingleButton(Vec2i size, Vec2i iconSize, Vec2i srcPos, const char* tip, const char* name, void (*function)(), Window& win):
		Button(size, srcPos, tip, 0, win),
		iconSize(iconSize),
		name(name),
		function(function),
		offSetIcon(0)
	{}
	const char* name;
	void paint();
	void action();
	Vec2i iconSize;
	int offSetIcon;

	Vec2i getSimClickPos(int subButtonIdx){
		return Vec2i(bSize.x/2, bSize.y/2);
	};
	void (*function)();
	const asmPointer getID()const{return reinterpret_cast<asmPointer>(function);};
};
class ExitButton : public SingleButton{
	void action();
public:
	ExitButton(Vec2i size, Vec2i srcPos, const char* tip, Window& win):
		SingleButton(size, srcPos, tip, 0, 0, win)
	{}
};
class TextButton : public Button{
public:
	TextButton(const char* i_text, const char* i_tip, void (*i_function)(), Window& i_win);
	void paint();
	void action();
	Vec2i getSimClickPos(int subButtonIdx){
		return Vec2i(bSize.x/2, bSize.y/2);
	};
	const char* text;
	void (*function)();
	const asmPointer getID()const{return reinterpret_cast<asmPointer>(function);};
};
class VerticalScrollBarSlider : public Button, public Drag{
public:
	SceneTree& win;

	VerticalScrollBarSlider(Vec2i i_size, Vec2i i_srcPos, const char* i_tip, int i_command, SceneTree& i_win, int refCountValue);

	void doDrag(const Vec2i ptCursor);
	void endDrag(const Vec2i ptCursor);

	void paint();
	void action();
	Vec2i getSimClickPos(int subButtonIdx){
		return Vec2i(bSize.x/2, bSize.y/2);
	};
	bool updateDrag(Vec2i ptCursor);
	void setPixelPos();
};

class VerticalScrollBarTopButton : public Button{
public:
	VerticalScrollBarTopButton(Vec2i i_size, Vec2i i_srcPos, const char* i_tip, int i_command, SceneTree& i_win, int refCountValue);
	void paint();
	void action();
	Vec2i getSimClickPos(int subButtonIdx){
		return Vec2i(bSize.x/2, bSize.y/2);
	};
	SceneTree& win;
};
class VerticalScrollBarBottomButton : public Button{
public:
	VerticalScrollBarBottomButton(Vec2i i_size, Vec2i i_srcPos, const char* i_tip, int i_command, SceneTree& i_win, int refCountValue);
	void paint();
	void action();
	Vec2i getSimClickPos(int subButtonIdx){
		return Vec2i(bSize.x/2, bSize.y/2);
	};
	SceneTree& win;
};
class PageUp : public Button{
public:
	PageUp(Vec2i i_size, Vec2i i_srcPos, const char* i_tip, int i_command, SceneTree& i_win, int refCountValue);
	void paint();
	void action();
	Vec2i getSimClickPos(int subButtonIdx){
		return Vec2i(bSize.x/2, bSize.y/2);
	};
	SceneTree& win;
};
class PageDown : public Button{
public:
	PageDown(Vec2i i_size, Vec2i i_srcPos, const char* i_tip, int i_command, SceneTree& i_win, int refCountValue);
	void paint(){};
	void action();
	Vec2i getSimClickPos(int subButtonIdx){
		return Vec2i(bSize.x/2, bSize.y/2);
	};
	SceneTree& win;
};

//typedef void (*Function0)();
class MenuItem  : public Button{
protected:
	void commonPaintButton();
	void paintHoverBlister();
public:
	MenuItem(String label, Menu& win, bool tick=false);
	MenuItem(Vec2i iconSize, Vec2i srcPos, Menu& win, bool tick=false);
	MenuItem(Vec2i iconSize, Vec2i srcPos, String label, Menu& win, bool tick=false);
	MenuItem(Vec2i iconSize, Vec2i srcPos, String label, int command, Menu& win, bool tick=false);
	MenuItem(Vec2i iconSize, Vec2i srcPos, int command, Menu& win, bool tick=false);
	Menu& win;
	Vec2i iconSize;
	String label;
	bool grey;
	bool tick;
	
	int getIndex(const M_<Ptr<MenuItem> >& list)const;


//	virtual bool hover(const Vec2i ptCursor);
	virtual void paint();
	virtual bool onHoverMove(const Vec2i ptCursor);
	friend Menu;
};
class MenuItem_ : public MenuItem{
public:
	MenuItem_(String label, void (*function)(), Menu& win, bool tick=false):
		MenuItem(label, win, tick),
		function(function)
	{}
	MenuItem_(Vec2i iconSize, Vec2i srcPos, String label, void (*function)(), Menu& win, bool tick=false):
		MenuItem(iconSize, srcPos, label, win, tick),
		function(function)
	{}
	void action();
	Vec2i getSimClickPos(int subButtonIdx){
		return Vec2i(bSize.x/2, bSize.y/2);
	};
	void (*function)();
};
class MenuItem_int : public MenuItem{
public:
	MenuItem_int(String i_label, void (*i_function)(int par), const int& i_par, Menu& i_win):
		MenuItem(i_label, i_win),
		function(i_function),
		par(i_par)
	{}
	void action();
	Vec2i getSimClickPos(int subButtonIdx){
		return Vec2i(bSize.x/2, bSize.y/2);
	};
	const int par;
	void (*function)(int par);
};
class MenuItem_int_int : public MenuItem{
public:
	MenuItem_int_int(String i_label, void (*i_function)(int parA, int parB), int i_parA, int i_parB, Menu& i_win):
		MenuItem(i_label, i_win),
		function(i_function),
		parA(i_parA),
		parB(i_parB)
	{}
	MenuItem_int_int(Vec2i i_iconSize, Vec2i i_srcPos, String i_label, void (*i_function)(int parA, int parB), int i_parA, int i_parB, Menu& i_win):
		MenuItem(i_iconSize, i_srcPos, i_label, i_win),
		function(i_function),
		parA(i_parA),
		parB(i_parB)
	{}
	void action();
	Vec2i getSimClickPos(int subButtonIdx){
		return Vec2i(bSize.x/2, bSize.y/2);
	};
	void (*function)(int parA, int parB);
	int parA;
	int parB;
};
class Seperator : public MenuItem{
public:
	Seperator(Menu& i_win):
		MenuItem(0, i_win)
	{}
	void paint();
	void action();
	Vec2i getSimClickPos(int subButtonIdx){
		return Vec2i();
	};
};

class BranchMenuItem : public MenuItem {
public:
	BranchMenuItem(String i_label, Menu& i_win, bool i_grey=false);
	BranchMenuItem(Vec2i i_iconSize, Vec2i i_srcPos, String i_label, Menu& i_win);

	void doCallBack();//while hover

	Vec2i getSimClickPos(int subButtonIdx){
		return Vec2i(bSize.x/2, bSize.y/2);
	};
	~BranchMenuItem(){
	}

	void paint();

	Ptr<Menu> menuPtr;
	void action();
	bool onHoverMove(const Vec2i ptCursor);
};
inline void twoWorkOutSubIndex(Button& button, const Vec2i& ptCursor){
	button.command=button.baseCommand;
	if(ptCursor.y<(button.bSize.y>>1)){
		button.command++;
	}
};
inline void fourWorkOutSubIndex(Button& button, const Vec2i& ptCursor){
	button.command=button.baseCommand;
	if(ptCursor.y<(button.bSize.y>>1)){
		button.command++;
	}
	if(ptCursor.x>button.bSize.x/2){
		button.command+=2;
	}
};
inline void sixWorkOutSubIndex(Button& button, const Vec2i& ptCursor){
	button.command=button.baseCommand;
	if(ptCursor.y<(button.bSize.y>>1)){
		button.command++;
	}
	if(ptCursor.x>button.bSize.x/3){
		button.command+=2;
		if(ptCursor.x > button.bSize.x/3*2){
			button.command+=2;
		}
	}
};

class SixUpDownButton : public Button{
public:
	SixUpDownButton(Vec2i i_size, Vec2i i_srcPos, const char* i_tip, void (*i_function0)(), void (*i_function1)(), void (*i_function2)(), void (*i_function3)(), void (*i_function4)(), void (*i_function5)(), Window& i_win):
		Button(i_size, i_srcPos, i_tip, 0, i_win),
		function0(i_function0),
		function1(i_function1),
		function2(i_function2),
		function3(i_function3),
		function4(i_function4),
		function5(i_function5)
	{}
	void workOutSubIndex(const Vec2i ptCursor){
		sixWorkOutSubIndex(*this, ptCursor);
	};
	void paint();
	void action();
	Vec2i getSimClickPos(int subButtonIdx){
		return Vec2i(bSize.x/3*(subButtonIdx>>1)+bSize.x/6, bSize.y-(bSize.y/2*(subButtonIdx & 1)+bSize.y/4));
	};
	void (*function0)();
	void (*function1)();
	void (*function2)();
	void (*function3)();
	void (*function4)();
	void (*function5)();
	const asmPointer getID()const{return reinterpret_cast<asmPointer>(function0);};
};

class TwoUpDownButton : public Button{
public:
//	TwoUpDownButton(Vec2i i_size, Vec2i i_dstPos, Vec2i i_srcPos, const char* i_tip, int i_command, Window& i_win):
//		Button(i_size, i_dstPos, i_srcPos, i_tip, i_command, i_win)
//	{}
	TwoUpDownButton(Vec2i i_size, Vec2i i_srcPos, const char* i_tip, void (*i_function0)(), void (*i_function1)(), Window& i_win):
		Button(i_size, i_srcPos, i_tip, 0, i_win),
		function0(i_function0),
		function1(i_function1)
	{}
	void workOutSubIndex(const Vec2i ptCursor){
		twoWorkOutSubIndex(*this, ptCursor);
	};
	Vec2i getSimClickPos(int subButtonIdx){
		return Vec2i(bSize.x/2, bSize.y-(bSize.y/2*(subButtonIdx & 1)+bSize.y/4));
	};
	void paint();
	void action();
	void (*function0)();
	void (*function1)();
	const asmPointer getID()const{return reinterpret_cast<asmPointer>(function0);};
};
typedef Ptr <Button>	ButtonPtr;
typedef Ptr <MenuItem>	MenuItemPtr;


class PropWinLine{
public:
	int pos;
	int backgroundColor;
	ButtonPtr button;
	PropWinLine():
		pos(0),
		backgroundColor(0)
	{}
	PropWinLine(const int i_pos):
		pos(i_pos),
		backgroundColor(0)
	{}
	PropWinLine(const int i_pos, ButtonPtr i_button):
		pos(i_pos),
		backgroundColor(0),
		button(i_button)
	{}
	void operator =(const PropWinLine& par){
		pos=par.pos;
		backgroundColor=par.backgroundColor;
		button=par.button;
	}
 };
extern M_<PropWinLine> g_propWinLine;
enum{
	colum=124
};
class FieldButton  : public Button{
public:
	Node& node;
	Field& bField;
	const Properties& sprop;
	FieldButton(Vec2i i_size, Vec2i i_dstPos, Vec2i i_srcPos, const char* i_tip, Window& i_win, Node& i_node, Field& i_bField, const Properties& i_sprop, const Vec2i valPos, const Vec2i valSize):
		Button(i_size, i_dstPos, i_srcPos, i_tip, 0, i_win),
		node(i_node),
		bField(i_bField),
		sprop(i_sprop),
		valPos(valPos),
		valSize(valSize),
		editMode(false)
	{}
	bool checkClickTextCell(Vec2i ptCursor, int& newCaretPos)const;
	Vec2i valPos;
	Vec2i valSize;
	virtual void toString(String& str)const = 0;
	bool editMode;
	virtual bool updateValFromTextCell(String textCellStr, const int &index, const Properties& properties){return false;};
	virtual void paintFieldName();
};
class MFButton{
public:
};
class IndexButton : public FieldButton{
public:
	IndexButton(Vec2i i_size, Vec2i i_dstPos, Vec2i i_srcPos, const char* i_tip, Window& i_win, Node& i_node, Field& i_bField, const Properties& i_sprop, const Vec2i valPos, const Vec2i valSize):
		FieldButton(i_size, i_dstPos, i_srcPos, i_tip, i_win, i_node, i_bField, i_sprop, valPos, valSize)
	{}
	void workOutSubIndex(const Vec2i ptCursor){
		twoWorkOutSubIndex(*this, ptCursor);
	};
	Vec2i getSimClickPos(int subButtonIdx){
		return Vec2i(bSize.x/2, bSize.y/2);
	};
	void paint();
	void toString(String& str)const;
	void action();
	void paintFieldName();
};

class LengthButton : public FieldButton{
public:
	LengthButton(Vec2i i_size, Vec2i i_dstPos, Vec2i i_srcPos, const char* i_tip, Window& i_win, Node& i_node, Field& i_bField, const Properties& i_sprop, int i_fieldIndex, int i_groupLength, const Vec2i valPos, const Vec2i valSize):
		FieldButton(i_size, i_dstPos, i_srcPos, i_tip, i_win, i_node, i_bField, i_sprop, valPos, valSize),
		fieldIndex(i_fieldIndex),
		groupLength(i_groupLength)
	{}
	void workOutSubIndex(const Vec2i ptCursor){
		twoWorkOutSubIndex(*this, ptCursor);
	};
	Vec2i getSimClickPos(int subButtonIdx){
		return Vec2i(bSize.x/2, bSize.y/2);
	};
	void paint();
	void toString(String& str)const;
	void action();
	void paintFieldName();

	int fieldIndex;
	int groupLength;
};
template <class T, class S=Single> class FB{
};

template<class T> class FB<T,Single> : public FieldButton {
//template<class T, class Single> class FB  : public FieldButton{
public:
	FB(Vec2i i_size, Vec2i i_dstPos, Vec2i i_srcPos, const char* i_tip, int i_command, Window& i_win, Node& i_node, FD<T,Single>& i_bField, const Properties& i_sprop, const Vec2i valPos, const Vec2i valSize):
		FieldButton(i_size, i_dstPos, i_srcPos, i_tip, i_win, i_node, i_bField, i_sprop, valPos, valSize),
		bField(i_bField)
	{}
	void workOutSubIndex(const Vec2i ptCursor);
	void toString(String& str)const{
		str="";
		str += bField.refValue();
	};
	void paint();
	void action();
//template <class T> bool FB<T, Single>::updateValFromTextCell(String textCellStr, const int &index, const Properties& properties){
bool updateValFromTextCell(String textCellStr, const int &index, const Properties& properties);

	Vec2i getSimClickPos(int subButtonIdx){
		return Vec2i(bSize.x/3*(subButtonIdx>>1)+bSize.x/6, bSize.y-(bSize.y/2*(subButtonIdx & 1)+bSize.y/4));
	};
	FD<T,Single>& bField;
};

template <> inline void FB<bool, Single>::toString(String& str)const{
	str="";
};
template <> bool FB<Image, Single>::updateValFromTextCell(String textCellStr, const int &index, const Properties& properties);

template<class T> class FB<T,Multiple> : public FieldButton {
public:
	FB(Vec2i i_size, Vec2i i_dstPos, Vec2i i_srcPos, const char* i_tip, int i_command, Window& i_win, Node& i_node, FD<T,Multiple>& i_bField, const Properties& i_sprop, const Vec2i valPos, const Vec2i valSize):
		FieldButton(i_size, i_dstPos, i_srcPos, i_tip, i_win, i_node, i_bField, i_sprop, valPos, valSize),
		bField(i_bField)
	{}
	void workOutSubIndex(const Vec2i ptCursor);
	void toString(String& str)const{
		if(bField.length){
			int index = node.getFieldGroupIndex(sprop.groupIndex);
			int startAtIndex = sprop.startAtIndex;
			int i = index - startAtIndex;
			if(i >= bField.length){
				node.setFieldGroupIndex(sprop.groupIndex, 0);
				i = startAtIndex;
			}
			if(i >= 0){
				str="";
				str+=bField.refValue()[i];
			}
		}
	};
	void paint();
	void action();
	bool updateValFromTextCell(String textCellStr, const int &index, const Properties& properties);

	Vec2i getSimClickPos(int subButtonIdx){
		return Vec2i(bSize.x/3*(subButtonIdx>>1)+bSize.x/6, bSize.y-(bSize.y/2*(subButtonIdx & 1)+bSize.y/4));
	};
	FD<T,Multiple>& bField;
};
template <> inline void FB<String, Single>::toString(String& str)const{
	str = bField;
};
template <> inline void FB<String, Multiple>::toString(String& str)const{
	if(bField.length){
		int index=node.getFieldGroupIndex(sprop.groupIndex);
		int startAtIndex=sprop.startAtIndex;
		int i=index-startAtIndex;
		if(i>=0 ){
			str=bField.refValue()[i];
		}
	}
};

template <> inline void FB<bool, Multiple>::toString(String& str)const{
	str="";
};
//template <> bool FB<bool, Multiple>::updateValFromTextCell(String textCellStr, const int &index, const Properties& properties);
//template <> bool FB<int, Multiple>::updateValFromTextCell(String textCellStr, const int &index, const Properties& properties);
//template <> bool FB<float, Multiple>::updateValFromTextCell(String textCellStr, const int &index, const Properties& properties);
template <> bool FB<Image, Multiple>::updateValFromTextCell(String textCellStr, const int &index, const Properties& properties);
template <> bool FB<String, Multiple>::updateValFromTextCell(String textCellStr, const int &index, const Properties& properties);

typedef FB<bool,Single> SFBoolButton;
typedef FB<Color3f,Single> SFColorButton;
typedef FB<double,Single> SFDoubleButton;
typedef FB<float,Single> SFFloatButton;
typedef FB<Image,Single> SFImageButton;
typedef FB<int,Single> SFInt32Button;
typedef FB<Rotation,Single> SFRotationButton;
typedef FB<String,Single> SFStringButton;
typedef FB<Vec2f,Single> SFVec2fButton;
typedef FB<Vec3f,Single> SFVec3fButton;

typedef FB<bool,Multiple> MFBoolButton;
typedef FB<Color3f,Multiple> MFColorButton;
typedef FB<double,Multiple> MFDoubleButton;
typedef FB<float,Multiple> MFFloatButton;
typedef FB<Image,Multiple> MFImageButton;
typedef FB<int,Multiple> MFInt32Button;
typedef FB<Rotation,Multiple> MFRotationButton;
typedef FB<String,Multiple> MFStringButton;
typedef FB<Vec2f,Multiple> MFVec2fButton;
typedef FB<Vec3f,Multiple> MFVec3fButton;

class AnimTimeLine : public SFInt32Button, public Drag{
public:
	AnimControlPanel* animCP;
	Anim* anim;
	const int lineOffset;
	int lineDist;

	AnimTimeLine(Vec2i size, Vec2i dstPos, int command, Window* win, AnimControlPanel* animCP, Anim* anim, FD<int,Single>& bField, const Properties& sprop):
	FB(size, dstPos, srcPos, sprop.tip, command, *win, *anim, bField, sprop, Vec2i(), Vec2i()),
		animCP(animCP),
		anim(anim),
		samplePose(0),
		keyIndex(0),
		lineOffset(12),
		lineDist(size.x-12*2)
	{}
	SamplePose* samplePose;
	int keyIndex;

	void doDrag(const Vec2i ptCursor);
	void endDrag(const Vec2i ptCursor);

	void paint();
	void action();
	Vec2i getSimClickPos(int subButtonIdx){
		return Vec2i(bSize.x/2, bSize.y/2);
	};
	bool updateDrag(Vec2i ptCursor);
	void setPixelPos();
	//void doCallBack();// while hover;
};

class AnimSelect : public SFInt32Button, public Drag{
public:
	AnimControlPanel* animCP;
	Anim* anim;
	const int lineOffset;
	int lineDist;

	AnimSelect(Vec2i size, Vec2i dstPos, int command, Window* win, AnimControlPanel* animCP, Anim* anim, FD<int,Single>& bField, const Properties& sprop):
	FB(size, dstPos, srcPos, sprop.tip, command, *win, *anim, bField, sprop, Vec2i(), Vec2i()),
		animCP(animCP),
		anim(anim),
		samplePose(0),
		keyIndex(0),
		lineOffset(12),
		lineDist(size.x-12*2),
		dragRight(true)
	{}
	SamplePose* samplePose;
	int keyIndex;

	void doDrag(const Vec2i ptCursor);
	void endDrag(const Vec2i ptCursor);

	void paint();
	void action();
	Vec2i getSimClickPos(int subButtonIdx){
		return Vec2i(bSize.x/2, bSize.y/2);
	};
	bool updateDrag(Vec2i ptCursor);
	void setPixelPos();
	bool dragRight;
	//void doCallBack();// while hover;
};

class AnimSlider : public SFFloatButton, public Drag{
public:
	AnimControlPanel* animCP;
	Anim* anim;
	AnimSlider(Vec2i size, Vec2i dstPos, Vec2i srcPos, int command, Window* win, AnimControlPanel* animCP, Anim* anim, FD<float,Single>& bField, const Properties& sprop, const Vec2i valPos, const Vec2i valSize):
		FB(size, dstPos, srcPos, sprop.tip, command, *win, *anim, bField, sprop, valPos, valSize),
		anim(anim),
		animCP(animCP)
	{}
	void doDrag(const Vec2i ptCursor);
	void endDrag(const Vec2i ptCursor);

	void paint();
	void action();
	Vec2i getSimClickPos(int subButtonIdx){
		return Vec2i(bSize.x/2, bSize.y/2);
	};
	bool updateDrag(Vec2i ptCursor);
	void setPixelPos();
	bool updateValFromTextCell(String textCellStr, const int &index, const Properties& properties);
	void toString(String& str)const;
};


class AnimPeriod : public SFDoubleButton{
public:
	Anim* anim;
	AnimPeriod(Vec2i size, Vec2i dstPos, Vec2i srcPos, int command, Window* win, AnimControlPanel* animCP, Anim* anim, FD<double,Single>& bField, const Properties& sprop):
		FB(size, dstPos, srcPos, sprop.tip, command, *win, *anim, bField, sprop, Vec2i(), Vec2i()),
		anim(anim)
	{}
	void paint();
	void action();
	void workOutSubIndex(const Vec2i ptCursor);
	Vec2i getSimClickPos(int subButtonIdx){
		return Vec2i(bSize.x/2, bSize.y/2);
	};
};

class Zoom : public SFDoubleButton{
public:
	Anim* anim;
	Zoom(Vec2i size, Vec2i dstPos, Vec2i srcPos, int command, Window* win, AnimControlPanel* animCP, Anim* anim, FD<double,Single>& bField, const Properties& sprop):
		FB(size, dstPos, srcPos, sprop.tip, command, *win, *anim, bField, sprop, Vec2i(), Vec2i()),
		anim(anim)
	{}
	void action();
};

class AnimPlayFrom : public SFFloatButton{
public:
	AnimControlPanel* animCP;
	Anim* anim;
	AnimPlayFrom(Vec2i size, Vec2i dstPos, Vec2i srcPos, int command, Window* win, AnimControlPanel* animCP, Anim* anim, FD<float,Single>& bField, const Properties& sprop):
		FB(size, dstPos, srcPos, sprop.tip, command, *win, *anim, bField, sprop, Vec2i(), Vec2i()),
		anim(anim),
		animCP(animCP)
	{}
	void paint();
	void action();
	void workOutSubIndex(const Vec2i ptCursor);
	Vec2i getSimClickPos(int subButtonIdx){
		return Vec2i(bSize.x/2, bSize.y/2);
	};
};

class ArtRotationButton : public SFRotationButton{
public:
	Art& part;
	ArtRotationButton(Window* win, Art* part, FD<Rotation,Single>& bField, const Properties& sprop):
		FB(Vec2i(54, 24), Vec2i(124, 51), Vec2i(), sprop.tip, 0, *win, *part, bField, sprop, Vec2i(182, 51), Vec2i(168, 16)),
		part(*part)
	{}
	void action();
	bool updateValFromTextCell(String textCellStr, const int &index, const Properties& properties);
};

class ArtTranslationButton : public SFVec3fButton{
public:
	Art& part;
	ArtTranslationButton(Window* win, Art* part, FD<Vec3f,Single>& bField, const Properties& sprop):
		FB(Vec2i(54, 24), Vec2i(124, 51 + 25), Vec2i(), sprop.tip, 0, *win, *part, bField, sprop, Vec2i(182, 51 + 25), Vec2i(168, 16)),
		part(*part)
	{}
	void action();
	bool updateValFromTextCell(String textCellStr, const int &index, const Properties& properties);
};

class ArtIKRotButton : public SFRotationButton{
public:
	Art& part;
	ArtIKRotButton(Window* win, Art* part, FD<Rotation,Single>& bField, const Properties& sprop):
		FB(Vec2i(54, 24), Vec2i(124, 51 + 25 * 3), Vec2i(), sprop.tip, 0, *win, *part, bField, sprop, Vec2i(182, 51 + 25 * 3), Vec2i(168, 16)),
		part(*part)
	{}
	void action();
	bool updateValFromTextCell(String textCellStr, const int &index, const Properties& properties);
};

class ArtIKPosButton : public SFVec3fButton{
public:
	Art& part;
	ArtIKPosButton(Window* win, Art* part, FD<Vec3f,Single>& bField, const Properties& sprop):
		FB(Vec2i(54, 24), Vec2i(124, 51 + 25 * 4), Vec2i(), sprop.tip, 0, *win, *part, bField, sprop, Vec2i(182, 51 + 25 * 4), Vec2i(168, 16)),
		part(*part)
	{}
	void action();
	bool updateValFromTextCell(String textCellStr, const int &index, const Properties& properties);
};

class ArtFkaButton : public SFFloatButton{
public:
	Art& part;
	ArtFkaButton(Window* win, Art* part, FD<float,Single>& bField, const Properties& sprop):
		FB(Vec2i(18, 24), Vec2i(60, 51 + 25 * 5), Vec2i(), sprop.tip, 0, *win, *part, bField, sprop, Vec2i(60 +  18 * 1 + 4, 51 + 25 * 5), Vec2i(168, 16)),
		part(*part)
	{}
	void action();
	bool updateValFromTextCell(String textCellStr, const int &index, const Properties& properties);
};

class ArtFkbButton : public SFFloatButton{
public:
	Art& part;
	ArtFkbButton(Window* win, Art* part, FD<float,Single>& bField, const Properties& sprop):
		FB(Vec2i(18, 24), Vec2i(280, 51 + 25 * 5), Vec2i(), sprop.tip, 0, *win, *part, bField, sprop, Vec2i(280 +  18 * 1 + 4, 51 + 25 * 5), Vec2i(168, 16)),
		part(*part)
	{}
	void action();
	bool updateValFromTextCell(String textCellStr, const int &index, const Properties& properties);
};

class ArtIKButton : public SFBoolButton{
public:
	Art& part;
	ArtIKButton(Window* win, Art* part, FD<bool,Single>& bField, const Properties& sprop):
		FB(Vec2i(12, 12), Vec2i(140, 51 + 25 * 2 + 8), Vec2i(403, 0), sprop.tip, 0, *win, *part, bField, sprop, Vec2i(150+4, 51 + 25 * 2), Vec2i(16,16)),
//return new SFBoolButton(Vec2i(12, 12), Vec2i(xPos, yPos+8), Vec2i(403, 0), properties.tip, 0, win, node, *this, properties, Vec2i(xPos+4, yPos), Vec2i(16,16));
		part(*part)
	{}
	void action();
};

class ArtCoordPointButton : public MFVec3fButton{
public:
	Art& part;
	ArtCoordPointButton(Window* win, Art* part, FD<Vec3f, Multiple>& bField, int yPos, const Properties& sprop):
		FB(Vec2i(54, 24), Vec2i(124, yPos), Vec2i(), sprop.tip, 0, *win, *part, bField, sprop, Vec2i(182, yPos), Vec2i(168, 16)),
		part(*part)
	{}
	void toString(String& str)const;
	void action();
	bool updateValFromTextCell(String textCellStr, const int &index, const Properties& properties);
	void paintFieldName();
};


//why do we have to have this in the h file for the linker?




typedef M_ <ButtonPtr> M_Button;
typedef M_ <M_Button> MM_Button;

//struct ButtonTimer{
//	bool active;
//	double startTime;
//	double lastTime;
//}; 
//extern ButtonTimer g_bt;
//extern HWND g_hToolBarWnd;
/*
class BaseMenuPtr : public Ptr<Menu>{
public:
	Menu* operator = (Menu* par);
};
extern BaseMenuPtr g_remMenu;
*/
extern Ptr<Menu> g_remMenu;

class TextCell{
public:
	TextCell():
		caretPos(0),
		highlightPos(-1),
		active(false)
	{
	}
	void setNHPos(int newPos){
		caretPos=highlightPos=newPos;
	}
	int caretPos;
	int highlightPos;
	bool active;
	String str;
	Vec2i pos;
	Vec2i size;
	void incCaretPos();
	void decCaretPos();
	void incCaretPosNH();
	void decCaretPosNH();
	bool deleteHighlight();
};
class CaretWindow : public Drag, public CallBack{
	void doWinkOff(const Vec2i& pos, Rect& srcRect)const;
	void doWinkOn(const Vec2i& pos, Rect& srcRect)const;

	Window& win;
	void setVisCharLen();
	static Vec2i remPreviousPos;
	static int remFirstChar;
public:
	void doNoThinkWink();
	void doNoThinkWinkOff();

	void onWindowGainFocus(){
		remFirstChar=firstVisChar;
		remPreviousPos=getCaretPos();
		lastWinkTime=0;
		winkOn=false;
	}
	int firstVisChar;
	void doCallBack(); 	//void renderCaret();
	virtual bool charInput(unsigned wParam)=0;

	Vec2i getCaretPos()const{
		return Vec2i(cellPos.x+(caretPos-firstVisChar)*charWidth, cellPos.y);
	}
	void updateCaret();

	double lastWinkTime;
	bool winkOn;

	CaretWindow(Window& i_win):
		win(i_win),
		lastWinkTime(0),
		winkOn(false),
		caretPos(0),
		highlightPos(-1),
		cellActive(*this),
		firstVisChar(0)
	{}
	void incCaretPos();
	void decCaretPos();
	void incCaretPosNH();
	void decCaretPosNH();
	void doHome();
	void doEnd();
	void doHomeShift();
	void doEndShift();
	void doDelete();
	void doDeleteShift();
	void paintCellStr();

	int caretPos;
	int highlightPos;
	class CellActive{
		bool raw;
	public:
		CaretWindow& home;
		CellActive(CaretWindow& i_home):
			home(i_home),
			raw(false)
		{}
		void operator = (const bool& par){
			raw=par;
			if(par==false){
				home.setCellActiveFalse();
			}
		}
		operator const bool ()const{
			return raw;
		}
	};
	CellActive cellActive;
	virtual void setCellActiveFalse()=0;
	String cellStr;
	Vec2i cellPos;
	Vec2i cellSize;
	bool deleteHighlight();
	void highLightBlock();
	virtual void onChange(){};
	void copyToClipboard(bool cut=false);
	void pasteFromClipboard();
	void checkForDoubleLeftClick();
	bool isInCell(const Vec2i localCursor, int& selFrom, int& selLen, int& newCaretPos)const;
	virtual void onRightClick(const Vec2i ptCursor){};
	virtual void doDrag(const Vec2i ptCursor);
};
class Window {
public:
	void releaseWindow();
	static M_<Ptr<Window> >visWinList;
	static MenuItem* getMenuItem(const String str);
	static void initializeWindows();
	static void removeToolTip();
	static void removeMenu();
	static void removeAllAbove(Ptr<Window> par);
	static void closeWindow(Ptr<Window> par);
	static void releaseAllControlPanels();
	//static void removeAll2dWindows();

	static void renderVisible(){
		for(int i=0; i<visWinList.length; i++){
			visWinList[i]->render();
		}
	}
	static void nullAllControlPanels();
	static bool initializeAllControlPanels();
	static void reInitAllVisable();
	static void paintAllControlPanels();
	static void paintAllControlPanelsExceptActive();
	static void reSetupAllControlPanels();
	static void postLoadProcessWin(Scene& scene);
	static void preSaveProcessWin(Scene& scene);
	static void closeAllControlPanelsExceptDemo();
	static void insertUnderActive(Ptr<Window> par);
	static void activeOnMouseWheel(const int par);


	static ControlPanel* getControlPanel(const int layer);
	static ControlPanel* getControlPanel(const Node* node);
	static Window* getWindow(const int layer);
	static Anim* animMode();
	static AnimControlPanel* animCPMode();
	static Anim* getAnimBar(); //dont use this to query animMode
	static ChatWin* chatMode();
	static Art* returnOpenArt();



	static void onSeamlessWindowsLoseFocus(){
		g_callBackList.length = 0;
		activeWindow->onLoseFocus();
		activeWindow->paint();
	}
	bool isVisible(){
		for(int i=0; i<visWinList.length; i++){
			if(this==visWinList[i]){
				return true;
			}
		}
		return false;
	}
	static void checkIfAWindowHasBeenClickedAndUpdate(const Vec2i& mousePos);
	static void checkIfAWindowHasBeenMidClickedAndUpdate(const Vec2i& mousePos);
	static void checkIfAWindowHasBeenRightClickedAndUpdate(const Vec2i& mousePos);
	static void checkIfMouseIsHoveringAWindowAndUpdate(const Vec2i& mousePos);
	class TopWindow{
	public:
		void operator = (Ptr<Window> par);
		operator Window*()const;
		Window* operator ->()const;
	};
	class ActiveWindow{
	public:
		Window* getActiveWindow()const;
		void operator = (Ptr<Window> par);
		Window* operator ->()const;
		bool	operator ==(const Window* par)const{
			return getActiveWindow() == par;
		}
		bool	operator !=(const Window* par)const{
			return getActiveWindow() != par;
		}
	};
	static TopWindow topWindow;
	static ActiveWindow activeWindow;

	int refCount;

	class Size{
		Vec2i raw;
	public:
		Window& home;
		const int& x;
		const int& y;
		operator const Vec2i& () const{
			return raw;
		}
		const Vec2i& operator = (const Vec2i& par){
			if(raw != par){//ignore resize logic
				raw = par;
				home.setWinSize(raw);
			}
			return raw;
		}

		Vec2i operator >>(const int& par)const{
			return Vec2i(x >> par, y >> par);
		}

		Size(Window& home):
		home(home),
			x(raw.x),
			y(raw.y)
		{}
		Size(Window& home, const Vec2i& size):
		home(home),
			raw(size),
			x(raw.x),
			y(raw.y)
		{}
	};
	virtual void setWinSize(Vec2i& raw); //only Size should call this function
	virtual void onGainFocus(){}
	virtual void onLoseFocus(){}
	virtual void onClose(){}
	void commonDoDrag(const Vec2i ptCursor, Vec2i& pos, Vec2i& dragPos);

	Size size;
	Vec2i pos;
	bool visible;
	bool hog;
	M_Int pixBuf;
	Window():
		size(*this),
		refCount(0),
		visible(false),
		hog(false)
	{}
	Window(bool hog):
		size(*this),
		refCount(0),
		visible(false),
		hog(hog)
	{}
	Window(const Vec2i& i_pos, const Vec2i& i_size):
		pos(i_pos),
		size(*this, i_size),
		refCount(0),
		visible(false),
		hog(false)
	{}
	~Window();

	virtual void paint()=0;

	void incRefCount(){
		if(this!=0){ //is this line really needed?
		   ++refCount;
		}
	}
	unsigned decRefCount(){
		--refCount;
		if(refCount<0) throw this; //is this line really needed?
		return refCount;
	}

	void blitBlockFromHorizontalLine(Vec2i dstPos, Vec2i srcPos, Vec2i size);
	void blitVerticalColorSweep(Vec2i srcPosFrom, Vec2i srcPosTo, Vec2i pos, Vec2i size);
	void blitHorizontalColorSweep(Vec2i dstPos, Vec2i srcPosFrom, Vec2i srcPosTo, Vec2i size);

	virtual bool click(const Vec2i ptCursor);
	virtual bool midClick(const Vec2i ptCursor){return false;};
	virtual bool rightClick(const Vec2i ptCursor){return false;};
	virtual void buttonHover(Vec2i ptCursor){}
	virtual bool hover(Vec2i ptCursor);
	virtual void release(Vec2i ptCursor);
	virtual bool keyDown(unsigned key){return false;}
	virtual bool keyDownAlt(unsigned key){return false;}
	virtual bool keyDownCtrl(unsigned key){return false;}
	virtual bool keyDownShift(unsigned key){return false;}
	virtual void onClickButton(){};
	virtual void copyTextCellToClipboard(bool cut = false){};
	void checkWinVis();

	//void housekeepButton();
	virtual void render();
	virtual void onMouseWheel(const int par){};
//	virtual bool input(UINT message, WPARAM wParam, LPARAM lParam){return false;}
	bool initWindow();

	void clear();
	void invert();

	bool updateShaderPosSize = true;
	bool updateShaderTexture = true;
	bool updateShader = true;

	WinShader* shader = 0;
private:
	bool loadSurfaceFromMemory();
};
class ResizeableWindow : public Window{
protected:
	class TugBorder : public Drag{
	public:
		Window* window;
		TugBorder():
			window(0)
		{}
		friend Window;
	};

	ResizeableWindow(const Vec2i& i_pos, const Vec2i& i_size):
		Window(i_pos, i_size),
		tugBorder(0)
	{}
	ResizeableWindow():
		tugBorder(0)
	{}
	TugBorder* tugBorder;

	class TB_Top : public TugBorder{
		void doDrag(const Vec2i ptCursor);
	};
	class TB_Bottom : public TugBorder{
		void doDrag(const Vec2i ptCursor);
	};
	class TB_Left : public TugBorder{
		void doDrag(const Vec2i ptCursor);
	};
	class TB_Right : public TugBorder{
		void doDrag(const Vec2i ptCursor);
	};
	class TB_TopLeft : public TugBorder{
		void doDrag(const Vec2i ptCursor);
	};
	class TB_TopRight : public TugBorder{
		void doDrag(const Vec2i ptCursor);
	};
	class TB_BottomLeft : public TugBorder{
		void doDrag(const Vec2i ptCursor);
	};
	class TB_BottomRight : public TugBorder{
		void doDrag(const Vec2i ptCursor);
	};
	static TB_Top tB_Top;
	static TB_Bottom tB_Bottom;
	static TB_Left tB_Left;
	static TB_Right tB_Right;
	static TB_TopLeft tB_TopLeft;
	static TB_TopRight tB_TopRight;
	static TB_BottomLeft tB_BottomLeft;
	static TB_BottomRight tB_BottomRight;

	bool hover(Vec2i ptCursor);
	void loadCursor(TugBorder* i_tugBorder, int64_t type);
};

class ToolBarWin : public Window, public Drag{
public:
	M_Button* m_button;

	void paint();
	void doDrag(const Vec2i ptCursor);
	bool keyDown(unsigned key);
	bool keyDownAlt(unsigned key);
	bool keyDownCtrl(unsigned key);
	bool keyDownShift(unsigned key);
	void buttonHover(Vec2i ptCursor);
	void onLoseFocus();
	void updateToolBarContext();
	void* decodeVButtonFunc(const String& str, int& i);

	ToolBarWin():
		Window(Vec2i(0,670), Vec2i()),
		m_button(new M_Button())
	{}
	static void initToolbar();

};

class MessageWin : public Window{
public:

	M_Button m_button;
	void paint();
	String str;
	void autoSizeAndStuff(String i_msg);
	void buttonHover(Vec2i ptCursor);
	MessageWin():
		Window(true)
	{}


};

class ToolTipWin : public Window{
public:
	void paint();
	void autoSizeAndStuff(String i_tip);
	String tip;
};
class ChatWin : public ResizeableWindow, public Drag{
public:
	class CWTextCell : public CaretWindow{
	protected:
		friend ChatWin;
	public:
		//Field* tField;
		CWTextCell(ChatWin& win):
			CaretWindow(win),
			win(win)
		{}
		ChatWin& win;
		bool charInput(unsigned wParam);
		void onChange();
		void setCellActiveFalse();
		void onRightClick(const Vec2i ptCursor);
	};
	CWTextCell textCell;

	ChatWin():
		textCell(*this),
		dragSentText(false),
		dragLineFrom(0),
		dragLineTo(0),
		dragCharFrom(0),
		dragCharTo(0),

		spanLineFrom(0),
		spanLineTo(0),
		spanCharFrom(0),
		spanCharTo(0)
	{}
	void paint();
	static M_<ShortCutKey> shortCut;
	static M_<ShortCutKey> shortCutCtrl;
	static M_<ShortCutKey> shortCutShift;

	bool keyDown(unsigned key);
	bool keyDownCtrl(unsigned key);
	bool keyDownShift(unsigned key);
	bool dragSentText;
	int dragLineFrom;
	int dragLineTo;
	int dragCharFrom;
	int dragCharTo;

	int spanLineFrom;
	int spanLineTo;
	int spanCharFrom;
	int spanCharTo;

	void clearHighLight();
	void processSpans();
	void clickUrl(const Vec2i localCursor);
	bool click(const Vec2i ptCursor);
	bool rightClick(const Vec2i ptCursor);
	void highLightDoubleClick(int& line, int caretPos);
	void getHighlightedString(String& str);
	void doDrag(const Vec2i ptCursor);
	void setWinSize(Vec2i& raw);
	static void pasteFromClipboard();
	void copyTextCellToClipboard(bool cut = false);
	bool hover(Vec2i ptCursor);
};
class ControlPanel : public ResizeableWindow, public Drag{
public:
	void setUpIndexControl(Field& field, Node& node, const int fieldIndex, const Properties& properties, const int& index, const int& length, int& yPos);
	enum Constants{
		startFieldLineYpos = title_bar_height + 1
	};
	class CPTextCell : public CaretWindow{
	protected:
		void selectField(FieldButton& bf);
		void setCellActiveFalse();
		friend ControlPanel;
		friend FileCP;
		friend PaletteCP;
		friend WaveToPolyControlPanel;
	public:
		//Field* tField;
		FieldButton* button;
		CPTextCell(ControlPanel& i_win):
			CaretWindow(i_win),
			win(i_win),
			button(0)
		{}
		ControlPanel& win;
		bool charInput(unsigned wParam);
		void onChange();
	};
	ExitButton exitWindow;
	CPTextCell textCell;
	M_<Ptr<FieldButton> > buttonList;

	//void unactivateTextCell();

	ControlPanel():
		ResizeableWindow(Vec2i(0,150), Vec2i(350,100)),
		textCell(*this),
		exitWindow(Vec2i(16,16), Vec2i(exitWindow_x, exitWindow_y), "close control panel", *this)
	{
		pixBuf.length=size.x*size.y;
		if(false == initWindow()){
			throw;
		}
	}
	void prePaint();
	void paint();
	void buttonHover(Vec2i ptCursor);
	NodePtr node;
	void doDrag(const Vec2i ptCursor);
	bool rightClick(const Vec2i ptCursor);
	bool click(const Vec2i ptCursor);
	void onClickButton(){
		textCell.cellActive=false;
	}
	static void pasteFromClipboard();

	void setUpControlPanel(Node* node, const M_Properties &properties);
	virtual void setupNodeWindow(Node* node);
	void onGainFocus();
	void onLoseFocus();
	void copyTextCellToClipboard(bool cut = false){
		textCell.copyToClipboard(cut);
	};

	bool keyDown(unsigned key);
	bool keyDownAlt(unsigned key);
	bool keyDownCtrl(unsigned key);
	bool keyDownShift(unsigned key);
	static M_<ShortCutKey> shortCut;
	static M_<ShortCutKey> shortCutAlt;
	static M_<ShortCutKey> shortCutCtrl;
	static M_<ShortCutKey> shortCutShift;

	void setWinSize(Vec2i& raw);
	bool hover(Vec2i ptCursor);
	void colorBlockColorRgn(int*& dstPix);
};

class AnimControlPanel : public ControlPanel{
public:
//class Play : public SFBoolButton{
//public:
//	AnimControlPanel& animCP;
//	Anim& anim;
//	Play(Vec2i size, Vec2i dstPos, Vec2i srcPos, int command, Window* win, AnimControlPanel* animCP, Anim* anim, FD<bool,Single>& bField, const Properties& sprop):
//		FB(size, dstPos, srcPos, sprop.tip, command, *win, *anim, bField, sprop, Vec2i(), Vec2i()),
//		anim(*anim),
//		animCP(*animCP)
//	{}
//	void action();
//};
class Pause : public SFBoolButton{
public:
	AnimControlPanel* animCP;
	Anim* anim;
	Pause(Vec2i size, Vec2i dstPos, Vec2i srcPos, int command, Window* win, AnimControlPanel* animCP, Anim* anim, FD<bool,Single>& bField, const Properties& sprop):
		FB(size, dstPos, srcPos, sprop.tip, command, *win, *anim, bField, sprop, Vec2i(), Vec2i()),
		anim(anim),
		animCP(animCP)
	{}
	void action();
};
	void updateVideo();
	void paint();

	void setWinSize(Vec2i& raw);
	void initialize();
	void setupNodeWindow(Node* node);
	bool rightClick(const Vec2i ptCursor);
	void buttonHover(Vec2i ptCursor);

	AnimControlPanel():
		anim(0),
		timeLine(0),
		animSlider(0),
		playButton(0),
		pauseButton(0),
		loopButton(0),
		m_button(new M_Button())
	{}
	M_Button* m_button;

	AnimSelect* animSelect;
	AnimTimeLine* timeLine;
	AnimSlider* animSlider;
	//Play* playButton;
	SingleButton* playButton;
	Pause* pauseButton;
	SFBoolButton* loopButton;
	AnimPeriod* animPeriod;
	Zoom* zoom;
	AnimPlayFrom* animPlayFrom;

	Anim* anim;
};

class VideoControlPanel : public ControlPanel{
public:

	void paint();
	void setWinSize(Vec2i& raw);
	void setupNodeWindow(Node* node);
	void getFilename(String& filename);
	void openMovieFile(const char* filename);
	void processFilename();
	void setPeriodFromVideo(Anim* anim);

	VideoControlPanel():
		video(0),
		frameBuffer(0),
		frameWidth(256),
		frameHeight(256),
		openButton(0),
		firstFrame(0)
	{}
	SFBoolButton* openButton;
	Video* video;
	int firstFrame;
	char* frameBuffer;
	int frameWidth;
	int frameHeight;
};


class WaveToPolyControlPanel : public ControlPanel{
public:

class CMark : public SFInt32Button{
public:
	WaveToPolyControlPanel& waveToPolyCP;
	WaveToPoly& waveToPoly;

	CMark(Vec2i size, Vec2i dstPos, int command, Window* win, WaveToPolyControlPanel* waveToPolyCP, WaveToPoly* waveToPoly, FD<int,Single>& bField, const Properties& sprop):
	FB(size, dstPos, srcPos, sprop.tip, command, *win, *waveToPoly, bField, sprop, Vec2i(), Vec2i()),
		waveToPolyCP(*waveToPolyCP),
		waveToPoly(*waveToPoly)
	{}

	void action();
};

class NextCycle : public SFInt32Button{
public:
	WaveToPolyControlPanel& waveToPolyCP;
	WaveToPoly& waveToPoly;

	NextCycle(Vec2i size, Vec2i dstPos, int command, Window* win, WaveToPolyControlPanel* waveToPolyCP, WaveToPoly* waveToPoly, FD<int,Single>& bField, const Properties& sprop):
	FB(size, dstPos, srcPos, sprop.tip, command, *win, *waveToPoly, bField, sprop, Vec2i(), Vec2i()),
		waveToPolyCP(*waveToPolyCP),
		waveToPoly(*waveToPoly)
	{}

	void action();
};

	void paint();
	void setWinSize(Vec2i& raw);
	void setupNodeWindow(Node* node);
	void getFilename(String& filename);
	void openWaveFile(const char* filename);
	void processFilename();
	bool click(const Vec2i ptCursor);
	bool rightClick(const Vec2i ptCursor);
	void clickInsertCycle(const Vec2i ptCursor);
	void clickSelectCycle(const Vec2i ptCursor);
	void clickDeleteCycle(const Vec2i ptCursor);

	WaveToPolyControlPanel():
		waveToPoly(0),
		openButton(0)
	{}
	SFBoolButton* openButton;
	WaveToPoly* waveToPoly;
	String filenameOnly;
};

class PaletteCP : public ControlPanel{
public:
	enum {
		colorTabsBeginX = 10,
		colorTabsYpos = startFieldLineYpos + 25 + 5,
		customColorTabsYpos = 216,
		colorTabsCX = 20,
		colorTabsCY = 16,
		colorTabsXSpace = 5,

		selColorTabYpos = 28,
		selColorTabCX = 40, 
		selColorTabCY = 23,

		tabColums = 15,
		tabRows = 6,

		upperLastRow = (tabRows - 2) * tabColums,
		bottomRow = (tabRows - 1) * tabColums,

		selBlendPosX = colorTabsBeginX + colorTabsCX / 2,
		selBlendPosY = 161,
	};

	PaletteCP():
		selTab(Vec2i(0, 0)),
		selCustom(false)
		{
		pixBuf.length = size.x * size.y;
		if(false == initWindow()){
			throw;
		}
	}
	Vec2i selTab;
	bool selCustom;

	//void setWinSize(Vec2i& raw);
	//void genNodeList(Node* node);
	void setupNodeWindow(Node* node);
	void setupColorSelectorNodeWindow(Node* node);
	void paint();
	void paintColor(String& str, int dstWidth, int dstHeight, int* dstPix, const M_Properties& properties);
	void paintOwnership(Art* part, String& str, int dstWidth, int dstHeight, int* dstPix, const M_Properties& properties);
	bool click(const Vec2i ptCursor);
	void clickColorTabs(const Vec2i ptCursor);
	void clickOwnershipColorTabs(Art* part, const Vec2i ptCursor, int numOfTabs);
	void updateOwnershipFields(Art* part);
	//bool hover(Vec2i ptCursor);

};


template <typename T> class Ptr;
class Menu;

class _3DWin : public Window{
	void paint(){}
	bool click(const Vec2i ptCursor);
	bool keyDown(unsigned key);
	bool keyDownAlt(unsigned key);
	bool keyDownCtrl(unsigned key);
	bool keyDownShift(unsigned key);
	void release(Vec2i ptCursor){};
	void render(){};
	void setWinSize(Vec2i& raw){};
public:
	void onMouseWheel(const int par);
	static M_<ShortCutKey> shortCut;
	static M_<ShortCutKey> shortCutAlt;
	static M_<ShortCutKey> shortCutCtrl;
	static M_<ShortCutKey> shortCutShift;
};
extern ScenePtr g_scene;

class AutoScrollDown : public CallBack{
public:
	float startTime;
	AutoScrollDown():
		startTime(0)
	{}
	void doCallBack();
};
class AutoScrollUp : public CallBack{
public:
	float startTime;
	AutoScrollUp():
		startTime(0)
	{}
	void doCallBack();
};
class DragNode : public Button, public Drag{
public:
	enum{
		none,
		newInstance,
		copyOf,
		move,
		use,
		//topDragDist = 8,
		//baseDragDist = 8,
		yOffset = 12,
	};
	class Item : public Drag{
	public:
		DragNode& home;
		void endDrag(const Vec2i ptCursor);
		void doDrag(const Vec2i ptCursor);
		Item(DragNode& i_home):
			home(i_home)
		{}
	};
	Item item;

	M_<M_<int> >newNodeBarList;
	int dropWhere;
	int potentialDropDestLine;
	int dragMode;
	//Vec2i pos;
	Vec2i size;
	//Vec2i remPos;
	void doDrag(const Vec2i ptCursor);
	void endDrag(const Vec2i ptCursor);
	SceneTree& win;

	Vec2i getSimClickPos(int subButtonIdx){return Vec2i();}
	void paint(){};
	void action();
	bool onHoverMove(const Vec2i ptCursor);
	void doCallBack();// while hover;

	DragNode(SceneTree& i_win);
	AutoScrollDown autoScrollDown;
	AutoScrollUp autoScrollUp;

	void paint(int dstWidth, int dstHeight, int srcWidth, int srcHeight, int* dstPix);
	void paintDragClues(int dstWidth, int dstHeight, int srcWidth, int srcHeight, int* dstPix);
	void dragNodeFromSceneTree(Node& node);
	void doDragDrop(Vec2i ptCursor, const bool dropIt);
	bool dropAsSibling(const Properties& sprop, const M_Int& promptNode, Node*& node, int& nodeLine, int& copyNodeLine, int& clickOffset, int copySibCount, Node*& parent, int& fieldIdx, int& element, const bool dropIt);
	bool dropAsChild  (const Properties& sprop, const M_Int& promptNode,				 int& nodeLine, int& copyNodeLine, int& clickOffset, int copySibCount, Node*& node, int& fieldIdx, const bool dropIt);
	bool dropToReplace(const Properties& sprop, const M_Int& promptNode, Node*& node, int& nodeLine, int& copyNodeLine, int& clickOffset, int copySibCount, Node*& parent, int& fieldIdx, int& element, const bool dropIt);
	void initNewNodeBarList(Node* node);
	void fillNewNodeBarList(const int* dragBarTypes, const int length);

	void workOutSubIndex(const Vec2i ptCursor);
private:
	void positionTipWin();
};

extern Ptr<Settings> g_settings;


class SceneTree : public ResizeableWindow, public Drag{
	void paintNodeBackground(unsigned dstWidth, unsigned dstHeight, int nodeCnt, int lastNode, int* pEnd, int collapsedLine);
	void createActiveLabel(int collapsedLine, bool visible);
	void onLoseFocus();
	void onGainFocus();
	void doDrag(const Vec2i ptCursor);
	void selectNextLine();
	int lastSelNodeLine;
protected:
	int numOfVisLines;
public:
	const int& getScrollPosCollapsed(){
		return g_scene->scroll;
	}
	enum{
		beginYOffset=25,
	};

	void paint();
	void buttonHover(Vec2i ptCursor);

	class STTextCell : public CaretWindow{
		void decodeGlobal(Node* node);
		void replaceVariableType(G_Variable* gv, int i);
	protected:
		void setCellActiveFalse();
	public:
		SceneTree& win;
		STTextCell(SceneTree& i_win):
			CaretWindow(i_win),
			win(i_win),
			cellReadyToEnter(false)
		{}
		bool charInput(unsigned wParam);
		void doDrag(const Vec2i ptCursor);
		void updateText();
		bool cellReadyToEnter;
	};
	static void exitSceneTreeWindow();
	static void decMultipleNodeSelection();
	static void incMultipleNodeSelection();
	static void escape();

	SceneTree():
		ResizeableWindow(Vec2i(500,25), Vec2i(300,480)),
		topScroll( Vec2i(16,15), Vec2i(268,176), tip_scroll_up_a_line, 0, *this, 1),
		bottomScroll( Vec2i(16,15), Vec2i(284,176), tip_scroll_down_a_line, 0, *this, 1),
		pageUp( Vec2i(), Vec2i(407,13), tip_scroll_up_a_page, 0, *this, 1),
		pageDown( Vec2i(), Vec2i(), tip_scroll_down_a_page, 0, *this, 1),
		slider( Vec2i(16,15), Vec2i(300,176), tip_drag_to_scroll, 0, *this, 1),
		exitWindow(Vec2i(16,16), Vec2i(exitWindow_x, exitWindow_y), tip_close_scene_tree_window, 0, exitSceneTreeWindow, *this),
		textCell(*this),
		lastSelNodeLine(0),
		newNodeBar(*this),
		numOfVisLines(0),
		copyLine(0),
		cp( new ControlPanel)
	{
		cp->size=Vec2i(350,200);
		cp->pos=Vec2i(g_settings->nodeWinPos.x, g_settings->nodeWinPos.y);
		cp->pixBuf.length=cp->size.x * cp->size.y;
	}

	VerticalScrollBarTopButton topScroll;
	VerticalScrollBarBottomButton bottomScroll;
	PageUp pageUp;
	PageDown pageDown;
	VerticalScrollBarSlider slider;
	SingleButton exitWindow;
	STTextCell textCell;
	DragNode newNodeBar;

	class SelLine{
	public:
		const int& operator = (const int& par)const{
			g_scene->selLine=par;
			return par;
		}
		operator int& ()const{
			return g_scene->selLine.refValue();
		}
	};
	Ptr<ControlPanel> cp;
	SelLine selLine;
	int copyLine;

	bool click(const Vec2i ptCursor);
	void checkForDoubleLeftClick();
	bool rightClick(const Vec2i ptCursor);

	static M_<ShortCutKey> shortCut;
	static M_<ShortCutKey> shortCutAlt;
	static M_<ShortCutKey> shortCutCtrl;
	static M_<ShortCutKey> shortCutShift;

	Vec2i getCarrotPos();
//	bool input(UINT message, WPARAM wParam, LPARAM lParam);
	void copyTextCellToClipboard(bool cut = false){
		textCell.copyToClipboard(cut);
	};
	static void pasteFromClipboard();

	void selectFirstPart();
	void paintBlueDot(int indent, int selLine);
	void setScrolPos(int offset);
	void resizePageUpPageDown();
	void keepLeagal();
	void reset();
	void selectNode(int newSelNodeLine, Node* node, int newCaretPos, int tmpInt);
	bool keyInput(unsigned wParam);
	void setSelNodeLine(Node*);
	void incEditLine(int incStep);
	void buildRoutePopupMenu(Menu& menu);
	void newNodePopupMenu (Vec2i ptCursor);
	void setWinSize(Vec2i& raw);
	void release(Vec2i ptCursor);
	void senseAllPartsTriangle(SeamlessPtr smls);
	void senseAllPartsTriangle(SeamlessPtr smls, TouchSensor* sensor);
	void genNodeList();
	void setResultsForSelectingNode(Node* node);
	void setResultsForSelectingVertex(Art* part);
	void setResultsForSelectingPoint(Build* bld);
	void selectNodeForExtrude(Node* node);

	void onClickButton(){
		textCell.cellActive=false;
	}
	void commonPostInsertNodeStuff(bool dontUpdateAutoInForClick = false);
	void lesserPostInsertNodeStuff(bool dontUpdateAutoInForClick = false);

	void onMouseWheel(const int par);
	bool keyDown(unsigned key);
	bool keyDownAlt(unsigned key);
	bool keyDownCtrl(unsigned key);
	bool keyDownShift(unsigned key);

	friend VerticalScrollBarTopButton;
	friend VerticalScrollBarBottomButton;
	friend PageUp;
	friend PageDown;
	friend VerticalScrollBarSlider;
	friend SingleButton;
	friend STTextCell;
	friend DragNode;
};
class Menu : public Window{
	void ensurePaintIntegrity();
protected:
	BranchMenuItem* openedBranchItem;
public:
	void buttonHover(Vec2i ptCursor);
	void paint();
	M_<Ptr<MenuItem> > list;
	void autoSizeAndStuff();
	bool click(const Vec2i ptCursor);
	void ensurePaintIntegrity(const Menu* selectedMenu);

	Menu():
		blisterItem(-1),
		openedBranchItem(0)
	{}
	friend BranchMenuItem;

	int blisterItem;
};

extern Ptr<_3DWin> g_3d;
extern Ptr<ToolBarWin> g_tb;
extern Ptr<ToolBarWin> g_tb2;
extern Ptr<ToolTipWin> g_tt;
extern Ptr<SceneTree> g_stw;
extern Ptr<ChatWin> g_cw;
extern Ptr<MessageWin>g_msg;

void pasteSprite( int dx, int dy, int sx, int sy, int dstWidth, int dstHeight, int srcWidth, int srcHeight, int iconWidth, int iconHeight, int* dstPix);
void pasteSprite2( int dx, int dy, int sx, int sy, int dstWidth, int dstHeight, int srcWidth, int srcHeight, int iconWidth, int iconHeight, int* dstPix);
void pasteRectangle( unsigned color, int dx, int dy, int dstWidth, int dstHeight, int iconWidth, int iconHeight, int* dstPix);
void pasteRectangle2( unsigned color, int dx, int dy, int dstWidth, int dstHeight, int iconWidth, int iconHeight, int* dstPix);
void drawHorizontalLine(unsigned color, unsigned y, unsigned from, unsigned to, unsigned dstWidth, unsigned dstHeight, int* pixBuf);
void drawVerticalLine(unsigned color, int x, int from, int to, int dstWidth, int dstHeight, int* pixBuf);

void drawHorizontalLine2(unsigned color, int y, int from, int to, int dstWidth, int dstHeight, int* pixBuf);
void drawVerticalLine2(unsigned color, int x, int from, int to, int dstWidth, int dstHeight, int* pixBuf);
void line(unsigned color, int fromX, int fromY, int toX, int toY, int dstWidth, int dstHeight, int* pixBuf);

void drawSouthEast(unsigned color, int x, int from, int to, int dstWidth, int dstHeight, int* pixBuf);
void drawSouthWest(unsigned color, int x, int from, int to, int dstWidth, int dstHeight, int* pixBuf);
//void plotPixel(unsigned color, int x, int y, int dstWidth, int dstHeight, int* pixBuf);
void plotPixel2(unsigned color, int x, int y, int dstWidth, int dstHeight, int* pixBuf);
void drawText2(const String& str, int fromChar, int toChar, unsigned color, Vec2i dstPos, int dstWidth, int dstHeight, int* dstPix);
void drawText2(const String& str, unsigned color, Vec2i dstPos, int dstWidth, int dstHeight, int* dstPix);
void drawTextR(const String& str, unsigned color, int dx, int dy, int dstWidth, int dstHeight, int* dstPix);
//void drawText(const String& str, unsigned color, int dx, int dy, int dstWidth, int dstHeight, int* dstPix);
//void drawText(const String& str, int fromChar, int toChar, unsigned color, int dx, int dy, int dstWidth, int dstHeight, int* dstPix);
//void plotChar(char charecter, unsigned color, int x, int y, int dstWidth, int dstHeight, int* dstPix);
void invertBlock( int tx, int ty, int bx, int by, int dstWidth, int dstHeight, int* pixBuf);
void changePixelColorInBlock(unsigned fromColor, unsigned toColor, int tx, int ty, int bx, int by, int dstWidth, int dstHeight, int* pixBuf);
void paintBlock( unsigned color, int tx, int ty, int bx, int by, int dstWidth, int dstHeight, int* pixBuf);
void paintBlock2( unsigned color, int tx, int ty, int bx, int by, int dstWidth, int dstHeight, int* pixBuf);
void pasteRectangle(unsigned color, int dx, int dy, int dstWidth, int dstHeight, int iconWidth, int iconHeight, int* dstPix);
void pasteSprite( int dx, int dy, int sx, int sy, int dstWidth, int dstHeight, int srcWidth, int srcHeight, int iconWidth, int iconHeight, int* dstPix);
void pasteSprite( int dx, int dy, int sx, int sy, int dstWidth, int dstHeight, int srcWidth, int srcHeight, int iconWidth, int iconHeight, Pix24* dstPix);
void drawVerticalLine(unsigned color, int x, int from, int to, int dstWidth, int dstHeight, int* pixBuf);
void drawHorizontalLine(unsigned color, unsigned y, unsigned from, unsigned to, unsigned dstWidth, unsigned dstHeight, int* pixBuf);
void paintSideEdge(const unsigned* color, int width, int tx, int ty, int bottom, int dstWidth, int dstHeight, int* dstPix);
void paintColorSweepForCurvedBar(const unsigned* fromColor, const unsigned* toColor, int tx, int ty, Vec2i size, int dstWidth, int dstHeight, int* dstPix);


struct ColorFormat{
	unsigned color;
	int end;
	int x;
	ColorFormat():
		color(0),
		end(0),
		x(0)
	{
	}
	ColorFormat( unsigned i_color, int i_end, int i_x):
		color(i_color),
		end(i_end),
		x(i_x)
	{
	}
};
typedef M_ <ColorFormat> M_ColorFormat;
 
inline BranchMenuItem::BranchMenuItem(String i_label, Menu& i_win, bool i_grey):
	MenuItem(i_label, i_win, i_grey),
	menuPtr(new Menu)
{}
inline BranchMenuItem::BranchMenuItem(Vec2i i_iconSize, Vec2i i_srcPos, String i_label, Menu& i_win):
	MenuItem(i_iconSize, i_srcPos, i_label, i_win),
	menuPtr(new Menu)
{}
/*
inline Menu* BaseMenuPtr::operator = (Menu* par){
	g_drag = 0;
	par->incRefCount();
	if(p!=NULL){
		if(p->decRefCount()==0){
			delete p;
		}
	}
	p=par;
	return *this;
}
*/
extern M_<ShortCutKey> globalShortCut;
extern M_<ShortCutKey> globalShortCutAlt;
extern M_<ShortCutKey> globalShortCutCtrl;
extern M_<ShortCutKey> globalShortCutShift;
extern Vec2i g_mousePos;
extern Vec2i g_newMousePos;
extern Vec2i g_prevMousePos;
extern const float colorIncrement;
extern float floatIncPIFrac;

void setupSelectedNodeWindow();

void changeCursor(int64_t requestType);
void changeModellingCursor(int64_t requestType);

extern int64_t g_currentType2;
extern int64_t g_modellingCursor;

extern int64_t g_crossCursor;
extern int64_t g_lockedOnPointCursor;
extern int64_t g_addPointCursor;
extern int64_t g_insertLineCursor;
extern int64_t g_deleteLineCursor;
extern int64_t g_mergePatchesCursor;
extern int64_t g_addGeometryCursor;
extern int64_t g_addGeometryMirrorCursor;
extern int64_t g_ringSelectCursor;
extern int64_t g_rectangleSelectPointCursor;
extern int64_t g_rectangleUnselectCursor;
extern int64_t g_rectangleSelectPolyCursor;
extern int64_t g_fillCursor;
extern int64_t g_brushCursor;
extern int64_t g_eyeDropperCursor;
extern int64_t g_knifeCursor;
extern int64_t g_addBoneCursor;
extern int64_t g_insertBoneCursor;

extern int64_t g_arrowCursor;
extern int64_t g_handCursor;
extern int64_t g_waitCursor;

extern int64_t g_sizeAllCursor;
extern int64_t g_sizeNWSECursor;
extern int64_t g_sizeNESWCursor;
extern int64_t g_sizeWECursor;
extern int64_t g_sizeNSCursor;

void getClientRect(Rect &rect);
void getCursorPos(Point& point);

enum{
	OK_BUTTON		=	0x00000000L,
	ICONASTERISK	=	0x00000040L,
	ICONINFORMATION	=	ICONASTERISK,
	ICONEXCLAMATION	=	0x00000030L,
	YESNO_BUTTON	=	0x00000004L,
	ICONQUESTION	=	0x00000020L,
	SMLS_IDYES		=	6,
};

int msG(String str1,String str2,String str3="",unsigned dlgTypeAndIcon=OK_BUTTON | ICONINFORMATION);
int msG(String str1,unsigned dlgTypeAndIcon=OK_BUTTON | ICONINFORMATION);
void msg(String strA, String strB, String strC="", unsigned dlgTypeAndIcon=OK_BUTTON | ICONINFORMATION);
void msg(String strA, unsigned dlgTypeAndIcon=OK_BUTTON | ICONINFORMATION);
