/* Seamless3d

Copyright (C) 2001-2017 Graham John Perrett (AKA thyme) thyme@seamless3d.com www.seamless3d.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this
software and associated documentation files (the "Software"), to deal in the Software
without restriction, including without limitation the rights to use, copy, modify,
merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to the following
conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT
OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.
*/


#include <windows.h>

#include "seamless3d.h"
#include "techuelife_triangle_editor.h"
#include "gui.h"
#include "scene_tree.h"
#include "robot_aid.h"

void CurvePoint (int numOf, int degree, const float *knot, const Vec3f* point, const float *weight, const float t, Vec3f& outPoint);

Rotation convRot4dto4f(const Rotation4d& par);
Rotation4d convRot4fto4d(const Rotation& par);

extern M_Rotation4d g_m_accumRot;
extern M_Rotation4d g_m_iaccumRot;


inline bool getPointOnCurve2(const Vec3f* p, const float* w, const M_Float& knot, const int dim, float minT, float maxT, const int degree, const Vec3f prevCrd, Vec3f& crd, double& t, double tInc, const float targetDist, float& prevDistDiff){



	double tPlus = t + tInc;

	if(tPlus <= 1.0f && tPlus >= 0){
		float t2 = (maxT - minT) * tPlus + minT;
		CurvePoint(dim - 1, degree, knot.getArray(), p, w, t2, crd);

		Vec3f diff = crd - prevCrd;

		float dist = diff.length();

		float distDiff = fabs(dist - targetDist);

		if(distDiff < prevDistDiff){
			t = tPlus;
			prevDistDiff = distDiff;
			return true;
		}
	}
	return false;
}


void getPointOnCurve(const M_Vec3f& point, const M_Float& weight, const M_Float& knot, float minT, float maxT, const int degree, Vec3f& prevCrd, double& t, double tInc, float targetDist, Rotation& rot, Vec3f& fromVector){

	const int dim = point.length;


	const Vec3f* p = point.getArray();
	const float* w = weight.getArray();

	Vec3f crd;
	float prevDistDiff = 10000;
	int limitCounter = 1000;


	while(prevDistDiff > .000001 && --limitCounter > 0){
		if(getPointOnCurve2(p, w, knot, dim, minT, maxT, degree, prevCrd, crd, t, tInc, targetDist, prevDistDiff)){
			continue;
		}
		else if(getPointOnCurve2(p, w, knot, dim, minT, maxT, degree, prevCrd, crd, t, -tInc, targetDist, prevDistDiff)){
			continue;
		}
		tInc = tInc * .5;
	};

	Vec3f toVector = crd - prevCrd;
	rot = Rotation(fromVector, toVector);
	fromVector = toVector;

	prevCrd = crd;

}

void NikInterpolator::take(float sliderconnectNode, Node* node, Field* field){
}

void insertSection(M_Vec3f& keyValue, int channels, int sectionIndex);

void NikInterpolator::take(float slider, const Vec3f& diff, const int channel, const int numOfChannels, const int focusOffset){
	M_Float& key = this->key;
	M_Vec3f& keyValue = this->keyValue;

	if(key.length == 0){
		key += slider;
		keyValue.length = key.length * numOfChannels;
		keyValue[channel * 2 + focusOffset] = diff;
	}
	else{
		
		for(int i = 0; i < key.length; i++){
			int kvIdx = numOfChannels * i + channel * 2 + focusOffset;

			if(roughEQ(slider, key[i])){
				keyValue[kvIdx] = diff;
				index[0] = i;
				index[1] = kvIdx;
				break;
			}
			else if(roughGT(key[i], slider)){
				key.insertElement(i, slider);
				insertSection(keyValue, numOfChannels, i);
				keyValue[kvIdx] = diff;
				index[0] = i;
				index[1] = kvIdx;
				break;
			}
			else if(i == key.length - 1){
				i++;
				key.insertElement(i, slider);
				insertSection(keyValue, numOfChannels, i);
				kvIdx += numOfChannels;
				//kvIdx = numOfChannels * i + channel * 2 + focusOffset; //must recalc keyValue index when adding after (instead of inserting before)
				keyValue[kvIdx] = diff;
				index[0] = i;
				index[1] = kvIdx;
				break;
			}
		}
	}
}

void NikInterpolator::setIndex(float slider, const int channel, const int numOfChannels){
	M_Float& key = this->key;

	if(key.length == 0){
		return;
	}
	else{
		
		for(int i = 0; i < key.length; i++){
			int kvIdx = numOfChannels * i + channel * 2 + focusOffset;

			if(roughEQ(slider, key[i])){
				index[0] = i;
				index[1] = kvIdx;
				break;
			}
			else if(roughGT(key[i], slider)){
				index[0] = i;
				index[1] = kvIdx;
				break;
			}
			else if(i == key.length - 1){
				i++;
				index[0] = i;
				index[1] = kvIdx;
				break;
			}
		}
	}
}

void NikInterpolator::deleteSample(float slider){//deleteSample only used for viewpoint?
}


void setNiki(Art* part, NikInterpolator* niki, M_Art& partSequence){
	part->niki = niki;
	partSequence += part;

	if(part->rh){
		return;
	}
	if (part->pin && partSequence.length > 1) {
		return;
	}

	for(int k = 0; k < part->part.length; k++){
		setNiki(part->part.refValue()[k].p, niki, partSequence);
		break;
	}
}

void addCalcFunction(Function* f, NikInterpolator* mo, Art* part);

void NikInterpolator::connectNode(Node* node, Field* field, String& name, Anim* anim, Function* function){
	if(Art* part = dynamic_cast<Art*>(node)){
		M_Float& key = this->key.refValue();
		M_Vec3f& keyValue = this->keyValue.refValue();
		M_Float& weight = this->weight.refValue();

		this->name = name;
		makeExistingNameUnique(this);

		addCalcFunction(function, this, part);
		anim->children.refValue() += this;
		this->calcLongOnly.refValue() = true;

		smls = part->smls;
		partSequence.length = 0;

		Art* parent = 0;
		if (part->ik == true) {
			if (parent = part->getParentPart()) {
				partSequence += parent;
				setNiki(part, this, partSequence);
			}
		}
		if (parent == 0) {
			setNiki(part, this, partSequence);
		}

		Art* endPart = partSequence[partSequence.length - 1];

		keyValue.length = 6;


		keyValue[0] = keyValue[1] = part->absolutePivotPoint;
		keyValue[4] = keyValue[5] = endPart->absolutePivotPoint;
		keyValue[2] = keyValue[3] = (keyValue[0] + keyValue[4]) * .5;

		keyValue += keyValue;

		weight = M_Float(1, 1, 1);

		key.length = 2;
		key[1] = 1;
		knot.refValue() = M_Float(0,0,0,1,1,1);

		processFields();
	}
}

void NikInterpolator::do_set_fraction(const float frac){
	M_Float& key = this->key.refValue();
	M_Vec3f& keyValue = this->keyValue.refValue();
	if(!key.length)return;

	int portionLen = keyValue.length / key.length;
	interpolated.length = portionLen;

	int i = 0;
	if(key.length < 2)goto here;
	for(; i<key.length-1; i++){
		if(key[i] > frac){
			i--;
			break;
		}
	}
	if(i < 0){
here:
		if(key.length > 0){
			//value_changed.sendOutRoute(keyValue[0]);
		}
		return;
	}
	if(i == key.length - 1){
		i = key.length - 2;
	}
	float diff = key[i + 1] - key[i];

	float interp = 1.0 / diff * (frac - key[i]);
	if(interp < 0){
		interp =- interp;
	}


	for(int j = 0; j < portionLen; j++){
		int h = i * portionLen + j;
		interpolated[j] = keyValue[h] + (keyValue[(i + 1) * portionLen + j] - keyValue[h]) * interp;
	}
}



void blendPoints(float amount, const M_Vec3f& longPoint, const M_Vec3f& shortPoint, M_Vec3f& resultPoint){

	float shortAmount = 1.0f - amount;

	for(int k = longPoint.length; --k>=0;){
		Vec3f longVec = longPoint[k];
		Vec3f shortVec = shortPoint[k];

		Vec3f resultVec = longVec * amount + shortVec * shortAmount;

		resultPoint[k] = resultVec;
	}
	//resultLine->processFields();
	//resultLine->tpoint = resultPoint;
}


bool NikInterpolator::calkArmPoints(double& amount, M_Art& m_art, float& prevDistDiff2, double aInc, Vec3f fromVector){
	float aPlus = amount + aInc;
	if(aPlus > 1.0 || aPlus < 0.0){
		return false;
	}

	blendPoints(aPlus, longPoint, shortPoint, resultPoint);

	Vec3f point = resultPoint[0];
	Vec3f prevPoint;
	double t = 0;
	double tInc = .1;
	float targetDist = 0;
	
	Rotation rot;

	String str;

	/*
	for (int k = 0; k < m_art.length; k++) {
		str = k;
		str += ' ';
		g_m_accumRot[k].toString(str);
		dLog(str);
	}
	*/

	for(int k = 0; k < m_art.length - 1; k++){
		Art* part = m_art[k];
		Art* partB = m_art[k + 1];
		targetDist = partB->pivotPoint.refValue().length();
		tInc = .1;
		prevPoint = point;
		getPointOnCurve(resultPoint, weight, knot, uMin(), uMax(), degree, point, t, tInc, targetDist, rot, fromVector);


		part->rotation = rot;
		if(part->pin && part->part.length){
			//part->translation = prevPoint;
			part->ikPos = prevPoint;
			part->translation = part->ikPos - part->absolutePivotPoint;

			//part->translation.toString(str);
			//dLog(str);
		}
	}

	
	Vec3f endPoint = resultPoint[resultPoint.length - 1];

	float distance = (prevPoint - endPoint).length();
	float diff = fabs(targetDist - distance);
	if(diff < prevDistDiff2){
		amount = aPlus;
		prevDistDiff2 = diff;
		return true;
	}
	return false;
}

void completeRotations(Art* pelvis, Art* hip, Art* knee){

	hip->rotation = convRot4dto4f(g_m_accumRot[pelvis->plidx] * convRot4fto4d(hip->rotation) * g_m_iaccumRot[pelvis->plidx]);

	if (pelvis && hip->pin) {
		hip->rotation = hip->ikRot * convRot4dto4f(g_m_iaccumRot[pelvis->plidx]);
	}

	g_m_accumRot[hip->plidx] = convRot4fto4d(hip->rotation) * g_m_accumRot[pelvis->plidx];
	g_m_iaccumRot[hip->plidx] = g_m_iaccumRot[pelvis->plidx] * convRot4fto4d(-hip->rotation);

	M_SArt& m_part = knee->part.refValue();
	pelvis = hip;
	hip = knee;
	for(int k = 0; k < m_part.length; k++){
		knee = m_part[k];
		completeRotations(pelvis, hip, knee);
	}
}


void setRhAnklePos(Vec3f& anklePos, M_Art& partSequence, M_Vec3f& longPoint, M_Vec3f& shortPoint){
	for(int k = 0; k < partSequence.length; k++){
		Art* ankle = partSequence[k];
		if(ankle->rh){
			if(ankle->part.length){
				Art* toes = ankle->part.refValue()[0];

				anklePos = -toes->pivotPoint;
				anklePos = (-ankle->ikRot * toes->ikRot) * anklePos;
				anklePos = anklePos + toes->ikPos;
				longPoint[k] = anklePos;
				shortPoint[k] = anklePos;
			}
		}
	}
}
Art* debugPart = 0;

void NikInterpolator::calcIK(Seamless* smls, Art* selHip){
	if(partSequence.length < 3){
		return;
	}
	if(selHip == 0){
		if(smls->posCompensator && smls->posCompensator->ankle == partSequence[0]){
			return;
		}
	}

	for(int k = interpolated.length / 2; --k>=0;){
		longPoint[k] = interpolated[k * 2];
		shortPoint[k] = interpolated[k * 2 + 1];
	}

	Art* part0 = partSequence[0];
	
	//debugPart = partSequence[8];

	if(part0->pin == false && selHip == 0){
		if(Art* parent = part0->parent){
//			if (parent->ik && parent->pin) {
			//parent->ikRot.refValue() = parent->rotation; //for debugging only
			//parent->ikPos.refValue() = parent->translation; //for debugging only
			Vec3f partPos0 = part0->matrix * part0->absolutePivotPoint;
			longPoint[0] = partPos0;
			shortPoint[0] = partPos0;
			//}
		}
	}

	Vec3f anklePos;

	setRhAnklePos(anklePos, partSequence, longPoint, shortPoint);


	double amount = 1;
	double aInc = .5;

	int numberOfTriesCounter = 1000;
	/*
	{
		Vec3f prevCrd;
		double t = 0;
		double tInc = .1;
		float targetDist = .5;
		Rotation rot;
		Vec3f fromVector = Vec3f(0, -1, 0);

		getPointOnCurve(longPoint, longPoint, longLine->weight, longLine->uKnot, longLine->minU(), longLine->maxU(), longLine->uDegree, prevCrd, t, tInc, targetDist, rot, fromVector);
		seg0->rotation = rot;

		tInc = .1;
		getPointOnCurve(longPoint, longPoint, longLine->weight, longLine->uKnot, longLine->minU(), longLine->maxU(), longLine->uDegree, prevCrd, t, tInc, targetDist, rot, fromVector);
		seg1->rotation = rot;
	}
	*/
	float distDiff = 0;
	float prevDistDiff2 = 10000;

	Vec3f fromVector;

	if(Art* parent = part0->parent){
		Art* part1 = partSequence[1];
		fromVector = convRot4dto4f(g_m_accumRot[parent->plidx]) * part1->pivotPoint.refValue().normalize();
	}
	else{
		Art* part1 = partSequence[1];
		fromVector = part0->ikRot.refValue() * part1->pivotPoint.refValue().normalize();
	}

	if(calcLongOnly){
		calkArmPoints(amount, partSequence, prevDistDiff2, 0, fromVector);
	}
	else{
		while(prevDistDiff2 > .0001 && --numberOfTriesCounter > 0){
			if(calkArmPoints(amount, partSequence, prevDistDiff2, aInc, fromVector)){
				continue;
			}
			if(calkArmPoints(amount, partSequence, prevDistDiff2, -aInc, fromVector)){
				continue;
			}

			aInc = aInc * .5;
		}
	}

	if(Art* parent = part0->parent){
		completeRotations(parent, part0, partSequence[1]);
	}
	else {
		//g_m_accumRot[0] = Rotation4d();
		//g_m_iaccumRot[0] = Rotation4d();
		//part0->rotation = part0->ikRot * convRot4dto4f(g_m_iaccumRot[part0->plidx]);
		completeRotations(part0, partSequence[1], partSequence[2]);
		//partSequence[8]->rotation.refValue() = Rotation(0, 0, 1, .427018); //we want to know how to get this value without hard wiring it here
		//partSequence[1]->rotation.refValue() = part0->rotation * partSequence[1]->rotation;
		//part0->rotation = Rotation();
		//Rotation rot = convRot4dto4f(g_m_iaccumRot[partSequence.length - 2]);
		partSequence[partSequence.length - 1]->rotation.refValue() = partSequence[partSequence.length - 1]->ikRot  * -partSequence[0]->rotation * convRot4dto4f(g_m_iaccumRot[partSequence.length - 2]); //not sure this is the correct order for multiplication
		//partSequence[partSequence.length - 1]->rotation.refValue() = Rotation(0, 0, 1, -.607677) * Rotation(0, 0, 1, 1.03469145);
	}

	if(selHip){
		if(Art* parent = part0->parent){
			Vec3f vec = -part0->pivotPoint.refValue();
			vec = parent->ikRot.refValue() * vec;
			//Vec3f vec = -parent->ikRot.refValue() * -part0->pivotPoint.refValue();

			vec += resultPoint[0];
			vec -= parent->pivotPoint;

			parent->translation.refValue() = vec;

		}
	}
	/*
	String str;
	for (int k = 0; k < partSequence.length; k++) {
		Art* part = partSequence[k];
		Vec3f vec = part->absolutePivotPoint;
		str = k;
		str += " ";
		vec.toString(str);
		dLog(str);
	}
	*/
	for(int k = 0; k < partSequence.length; k++){
		Art* ankle = partSequence[k];
		if(ankle->rh){
			if(ankle->part.length){
				Art* knee = partSequence[k - 1];
				Art* toes = ankle->part.refValue()[0];

				Rotation rot = -convRot4dto4f(g_m_accumRot[knee->plidx]) * (-ankle->ikRot * toes->ikRot);
				ankle->rotation = convRot4dto4f(g_m_accumRot[knee->plidx] * convRot4fto4d(rot) * g_m_iaccumRot[knee->plidx]);
			}
		}
	}

	smls->ctrf = true;
	smls->animCPBuildFlag = true;
	//smls->doBuild();
	if(Art* part = dynamic_cast<Art*>(g_selNode.getRaw())){
		if(part->niki == this && this == g_stw->cp->node){
			this->houseKeep();
		}
	}
}

Art* Art::getParentPart() {
	int px = plidx;
	while (--px >= 0) {
		Art* part = smls->partList[px];
		M_SArt& m_part = part->part.refValue();
		for (int k = m_part.length; --k >= 0;) {
			if(m_part[k] == this)
			return part;
		}
	}
	return 0;
}
void NikInterpolator::calc(const float v, Art& part){
	smls = part.smls;
	smls->nikiList += this;

	partSequence.length = 0;

	Art* parent = 0;
	if (part.ik == true) {
		if (parent = part.getParentPart()) {
			partSequence += parent;
			setNiki(&part, this, partSequence);
		}
	}
	if(parent == 0){
		setNiki(&part, this, partSequence);
	}
	do_set_fraction(v);

	smls->ikProcessFlag = true;
}

void NikInterpolator::processFields(){
	//knot.refValue() = M_Float(0,0,0,1,1,1);
	if(longPoint.length == 0 && keyValue.length){
		int nurbLinelength = getNurbLineLength();

		longPoint.length = nurbLinelength;
		shortPoint.length = nurbLinelength;
		resultPoint.length = nurbLinelength;

		for(int k = 0; k < nurbLinelength; k++){
			longPoint[k] = keyValue[k * 2];
			shortPoint[k] = keyValue[k * 2 + 1];
			resultPoint[k] = keyValue[k * 2];
		}
	}
}

void commonEndForSetupAnim();

void addNiki(){
	if(Art* part=dynamic_cast<Art*>(g_selNode.getRaw())){
		if(part->part.length){
			Art* childPart = part->part.refValue()[0];
			if(childPart->ik){
				if(Anim* anim = Window::animMode()){
					if(anim->play == false){
						if(anim->getNikInterpolator(part)){
						}
						else{
							anim->connectPartToNikInterpolator(part);
						}
						if(NikInterpolator* niki = part->niki){
							niki->houseKeep();
						}
					}
				}
			}
		}
	}
	else{
		needPartMsg();
	}
	
}

void snapTwinFunc(){
	if(Art* part=dynamic_cast<Art*>(g_selNode.getRaw())){
		if(Anim* anim = Window::animMode()){
			if(anim->play == false){
				if(NikInterpolator* niki = part->niki){
					int idx = niki->index[1];
					if(idx & 1){
						niki->keyValue.refValue()[idx] = niki->keyValue.refValue()[idx - 1];
					}
					else{
						niki->keyValue.refValue()[idx + 1] = niki->keyValue.refValue()[idx];
					}
				}
			}
		}
	}
	else{
		needPartMsg();
	}
}

void displayNikiPanel(){
	if(Art* part=dynamic_cast<Art*>(g_selNode.getRaw())){
		if(Anim* anim = Window::animMode()){
			if(anim->play == false){
				if(NikInterpolator* niki = part->niki){
					g_stw->cp->setupNodeWindow(niki);
					Window::activeWindow = g_stw->cp.p;
				}
			}
		}
	}
}
void Art::NikPush::operator=(const bool& par){
	if(Anim* anim = Window::animMode()){
		if(anim->play == false){

			Window::closeWindow(g_stw->cp.p);

			if(home.niki == 0){
				addNiki();
			}
			if(home.niki){
				g_stw->cp->setupNodeWindow(home.niki);
				Window::activeWindow = g_stw->cp.p;
			}
		}
	}
}

void NikInterpolator::SelectShort::operator=(const bool& par){
	refValue() = par;
	if(par){
		home.focusOffset = 1;
	}
	else{
		home.focusOffset = 0;
	}
	home.houseKeep();
}
void selectLastSelectedPart();
extern int g_animBarListIndex;
extern bool g_dontBackup;

void NikInterpolator::CalcLongOnly::operator=(const bool& par){
	refValue() = par;
	if(Anim* anim = Window::getAnimBar()){
		anim->update = true;
	}
	home.houseKeep();
}

void NikInterpolator::SnapTwin::operator=(const bool& par){
	if(Anim* anim = Window::getAnimBar()){
		snapTwinFunc();
		home.smls->animCPBuildFlag = true;
		anim->update = true;
		home.houseKeep();
	}
}

int NikInterpolator::getNurbLineLength() const {
	return (keyValue.length / key.length) / 2;
}
int NikInterpolator::getSampleLength() const {
	return keyValue.length / key.length;
}