/* Seamless3d

Copyright (C) 2001-2019 Graham John Perrett (AKA thyme) thyme@seamless3d.com www.seamless3d.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this
software and associated documentation files (the "Software"), to deal in the Software
without restriction, including without limitation the rights to use, copy, modify,
merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to the following
conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT
OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.
*/


#include <windows.h>

#include "seamless3d.h"
#include "techuelife_triangle_editor.h"
#include "gui.h"
#include "scene_tree.h"
#include "robot_aid.h"

void NikInterpolator::houseKeep() {};

int NikInterpolator::retainLimit = 2;
M_<NikInterpolatorPtr> NikInterpolator::retainView;
bool NikInterpolator::updateVisualClue = false;
float NikInterpolator::insertPointRatio = 0;

void CurvePoint(int numOf, int degree, const float *knot, const Vec3f* point, const float *weight, const float t, Vec3f& outPoint);
bool clickLine(Vec2f click, Vec2f ptA, Vec2f ptB, Vec2f& crossPoint);
float getMidPointRatio(Vec2f ptA, Vec2f ptB, Vec2f crossPoint);
void updateNurbsCurveKnotMid0(int dim, M_Float& knot);

Rotation convRot4dto4f(const Rotation4d& par);
Rotation4d convRot4fto4d(const Rotation& par);

extern M_Rotation4d g_m_accumRot;
extern M_Rotation4d g_m_iaccumRot;
extern bool tugFlag;


/*
inline void getJustbeforeLastPointOnCurve(const Vec3f* p, const float* w, const M_Float& knot, const int dim, float minT, float maxT, const int degree, Vec3f& crd) {

	double t = .99;

	float t2 = (maxT - minT) * t + minT;
	CurvePoint(dim - 1, degree, knot.getArray(), p, w, t2, crd);
}
*/

inline bool getPointOnCurve2(const Vec3f* p, const float* w, const M_Float& knot, const int dim, float minT, float maxT, const int degree, const Vec3f prevCrd, Vec3f& crd, double& prevT, double& t, double tInc, const float targetDist, float& prevDistDiff) {

	double tPlus = t + tInc;

	if (tPlus <= 1.0f && tPlus >= prevT) {
		float t2 = (maxT - minT) * tPlus + minT;
		CurvePoint(dim - 1, degree, knot.getArray(), p, w, t2, crd);

		Vec3f diff = crd - prevCrd;

		float dist = diff.length();

		float distDiff = fabs(dist - targetDist);

		if (distDiff < prevDistDiff) {
			t = tPlus;
			prevDistDiff = distDiff;
			return true;
		}
	}
	return false;
}


void getPointOnCurve(const M_Vec3f& point, const M_Float& weight, const M_Float& knot, float minT, float maxT, const int degree, Vec3f& prevCrd, double& t, double tInc, float targetDist, Rotation& rot, Vec3f& fromVector) {

	const int dim = point.length;


	const Vec3f* p = point.getArray();
	const float* w = weight.getArray();

	Vec3f crd;
	float prevDistDiff = 10000;
	int limitCounter = 1000;

	double prevT = t;

	while (prevDistDiff > .000001 && --limitCounter > 0) {
		if (getPointOnCurve2(p, w, knot, dim, minT, maxT, degree, prevCrd, crd, prevT, t, tInc, targetDist, prevDistDiff)) {
			continue;
		}
		else if (getPointOnCurve2(p, w, knot, dim, minT, maxT, degree, prevCrd, crd, prevT, t, -tInc, targetDist, prevDistDiff)) {
			continue;
		}
		tInc = tInc * .5;
	};

	Vec3f toVector = crd - prevCrd;
	rot = Rotation(fromVector, toVector);
	fromVector = toVector;

	prevCrd = crd;

}

void NikInterpolator::take(float sliderconnectNode, Node* node, Field* field) {
}

void insertSection(M_Vec3f& keyValue, int channels, int sectionIndex);

void NikInterpolator::take(float slider, const Vec3f& diff, const int channel, const int numOfChannels, const int focusOffset) {
 	M_Float& key = this->key;
	M_Vec3f& keyValue = this->keyValue;

	if (key.length == 0) {
		key += slider;
		keyValue.length = key.length * numOfChannels;
		keyValue[channel * 2 + focusOffset] = diff;
	}
	else {

		for (int i = 0; i < key.length; i++) {
			int kvIdx = numOfChannels * i + channel * 2 + focusOffset;

			if (roughEQ(slider, key[i])) {
				keyValue[kvIdx] = diff;
				index[0] = i;
				index[1] = kvIdx;
				break;
			}
			else if (roughGT(key[i], slider)) {
				key.insertElement(i, slider);
				insertSection(keyValue, numOfChannels, i);
				keyValue[kvIdx] = diff;
				index[0] = i;
				index[1] = kvIdx;
				break;
			}
			else if (i == key.length - 1) {
				i++;
				key.insertElement(i, slider);
				insertSection(keyValue, numOfChannels, i);
				kvIdx += numOfChannels;
				//kvIdx = numOfChannels * i + channel * 2 + focusOffset; //must recalc keyValue index when adding after (instead of inserting before)
				keyValue[kvIdx] = diff;
				index[0] = i;
				index[1] = kvIdx;
				break;
			}
		}
	}
}

void NikInterpolator::setIndex(float slider, const int channel, const int numOfChannels) {
	M_Float& key = this->key;

	if (key.length == 0) {
		return;
	}
	else {

		for (int i = 0; i < key.length; i++) {
			int kvIdx = numOfChannels * i + channel * 2 + focusOffset;

			if (roughEQ(slider, key[i])) {
				index[0] = i;
				index[1] = kvIdx;
				break;
			}
			else if (roughGT(key[i], slider)) {
				index[0] = i;
				index[1] = kvIdx;
				break;
			}
			else if (i == key.length - 1) {
				i++;
				index[0] = i;
				index[1] = kvIdx;
				break;
			}
		}
	}
}

void NikInterpolator::deleteSample(float slider) {//deleteSample only used for viewpoint?
}


void setNiki(Art* part, NikInterpolator* niki, M_Art& partSequence) {
	part->niki = niki;
	part->nikOrder = partSequence.length;
	partSequence += part;

	if (part->rh) {
		return;
	}
	if (part->pin && partSequence.length > 1) {
		return;
	}

	for (int k = 0; k < part->part.length; k++) {
		setNiki(part->part.refValue()[k].p, niki, partSequence);
		break;
	}
}

void addCalcFunction(Function* f, NikInterpolator* mo, Art* part);

void NikInterpolator::connectNode(Node* node, Field* field, String& name, Anim* anim, Function* function) {
	if (Art* part = dynamic_cast<Art*>(node)) {
		M_Float& key = this->key.refValue();
		M_Vec3f& keyValue = this->keyValue.refValue();
		M_Float& weight = this->weight.refValue();

		this->name = name;
		makeExistingNameUnique(this);

		addCalcFunction(function, this, part);
		anim->children.refValue() += this;

		smls = part->smls;

		bool brandNew = true;

		if (partSequence.length) {
			brandNew = false;
			partSequence.length = 0;
		}
		
		Art* parent = 0;
		if (part->ik == true) {
			if (parent = part->parent) {
				partSequence += parent;
				setNiki(part, this, partSequence);
			}
		}
		if (parent == 0) {
			setNiki(part, this, partSequence);
		}

		Art* endPart = partSequence[partSequence.length - 1];

		if (brandNew) {
			keyValue.length = 6;
			Vec3f diff = endPart->absolutePivotPoint - part->absolutePivotPoint;
			float forward = diff.length() * 2;
			diff = diff * .9;

			keyValue[0] = keyValue[1] = part->absolutePivotPoint;
			keyValue[4] = keyValue[5] = part->absolutePivotPoint + diff;
			keyValue[2] = keyValue[3] = part->absolutePivotPoint + (diff * .5) + Vec3f(0, 0, forward);

			keyValue += keyValue;

			key.length = 2;
			key[1] = 1;
			processFields();
		}

	}
}

void NikInterpolator::do_set_fraction(const float frac) {
	M_Float& key = this->key.refValue();
	M_Vec3f& keyValue = this->keyValue.refValue();
	if (!key.length)return;

	int portionLen = keyValue.length / key.length;
	interpolated.length = portionLen;

	int i = 0;
	if (key.length < 2)goto here;
	for (; i < key.length - 1; i++) {
		if (key[i] > frac) {
			i--;
			break;
		}
	}
	if (i < 0) {
	here:
		if (key.length > 0) {
			//value_changed.sendOutRoute(keyValue[0]);
		}
		return;
	}
	if (i == key.length - 1) {
		i = key.length - 2;
	}
	float diff = key[i + 1] - key[i];

	float interp = 1.0 / diff * (frac - key[i]);
	if (interp < 0) {
		interp = -interp;
	}


	for (int j = 0; j < portionLen; j++) {
		int h = i * portionLen + j;
		interpolated[j] = keyValue[h] + (keyValue[(i + 1) * portionLen + j] - keyValue[h]) * interp;
	}
}



void blendPoints(float amount, const M_Vec3f& longPoint, const M_Vec3f& shortPoint, M_Vec3f& resultPoint) {

	float shortAmount = 1.0f - amount;

	for (int k = longPoint.length; --k >= 0;) {
		Vec3f longVec = longPoint[k];
		Vec3f shortVec = shortPoint[k];

		Vec3f resultVec = longVec * amount + shortVec * shortAmount;

		resultPoint[k] = resultVec;
	}
	//resultLine->processFields();
	//resultLine->tpoint = resultPoint;
}


bool NikInterpolator::calkArmPoints(double& amount, M_Art& m_art, float& prevDistDiff2, double aInc, Vec3f fromVector) {
	float aPlus = amount + aInc;
	if (aPlus > 1.0 || aPlus < 0.0) {
		return false;
	}

	blendPoints(aPlus, longPoint, shortPoint, resultPoint);

	Vec3f point = resultPoint[0];
	Vec3f prevPoint;
	double t = 0;
	double tInc = .1;
	float targetDist = 0;

	Rotation rot;

	String str;

	/*
	for (int k = 0; k < m_art.length; k++) {
		str = k;
		str += ' ';
		g_m_accumRot[k].toString(str);
		dLog(str);
	}
	*/

	for (int k = 0; k < m_art.length - 1; k++) {
		Art* part = m_art[k];
		Art* partB = m_art[k + 1];
		targetDist = partB->pivotPoint.refValue().length();
		tInc = .01;
		prevPoint = point;
		getPointOnCurve(resultPoint, weight, knot, uMin(), uMax(), degree, point, t, tInc, targetDist, rot, fromVector);


		part->rotation.refValue() = rot;
		if (part->pin && part->part.length) {
			//part->translation = prevPoint;
			part->ikPos = prevPoint;
			part->translation = part->ikPos - part->absolutePivotPoint;

			//part->translation.toString(str);
			//dLog(str);
		}
	}
	/*
	if (relativeEnd){
		const int dim = resultPoint.length;

		Vec3f justBeforeEndPoint;

		//getJustbeforeLastPointOnCurve(resultPoint.getArray(), weight.getArray(), knot, dim, uMin(), uMax(), degree, justBeforeEndPoint);
		//Vec3f toVector = resultPoint[resultPoint.length - 1] - justBeforeEndPoint;

		//prevPoint = resultPoint[resultPoint.length - 2];

		//fromVector = resultPoint[resultPoint.length - 2] - resultPoint[resultPoint.length - 3];
		Vec3f toVector = resultPoint[resultPoint.length - 1] - prevPoint;
		rot = Rotation(fromVector, toVector);
		partSequence[partSequence.length - 1]->rotation.refValue() = rot;
	}
	*/
	Vec3f endPoint = resultPoint[resultPoint.length - 1];

	float distance = (prevPoint - endPoint).length();
	float diff = fabs(targetDist - distance);
	if (diff < prevDistDiff2) {
		amount = aPlus;
		prevDistDiff2 = diff;
		return true;
	}
	return false;
}


void NikInterpolator::completeRotations(Art* parent, int k) {
	int last = partSequence.length - 1;

	for (; k < last; k++) {
		Art* part = partSequence[k];

		part->rotation = convRot4dto4f(g_m_accumRot[parent->plidx] * convRot4fto4d(part->rotation) * g_m_iaccumRot[parent->plidx]);

		g_m_accumRot[part->plidx] = convRot4fto4d(part->rotation) * g_m_accumRot[parent->plidx];
		g_m_iaccumRot[part->plidx] = g_m_iaccumRot[parent->plidx] * convRot4fto4d(-part->rotation);

		parent = part;
	}

	Art* lastPart = partSequence[last];

	if (lastPart->rh) {
		return;
	}
/*
	if (lastPart->rh && last - 1 >= 0 && lastPart->part.length) {//this is the important place the rh part rotation is set
		Art* parent = partSequence[last - 1];
		Art* child = lastPart->part.refValue()[0];

		Rotation rot = -convRot4dto4f(g_m_accumRot[parent->plidx]) * (-lastPart->ikRot * child->ikRot);
		lastPart->rotation = convRot4dto4f(g_m_accumRot[parent->plidx] * convRot4fto4d(rot) * g_m_iaccumRot[parent->plidx]);

		//g_m_accumRot[lastPart->plidx] = convRot4fto4d(lastPart->rotation) * g_m_accumRot[parent->plidx];
		g_m_iaccumRot[lastPart->plidx] = g_m_iaccumRot[parent->plidx] * convRot4fto4d(-lastPart->rotation);


		child->rotation = child->ikRot * convRot4dto4f(g_m_iaccumRot[lastPart->plidx]);
	}
	else */
	if (lastPart->pin) {
		if (relativeEnd) {
			lastPart->rotation = lastPart->ikRot;
		}
		else {
			lastPart->rotation = lastPart->ikRot * convRot4dto4f(g_m_iaccumRot[parent->plidx]);
		}
	}
}


void setRhAnklePos(Vec3f& anklePos, M_Art& partSequence, M_Vec3f& longPoint, M_Vec3f& shortPoint) {
	for (int k = 0; k < partSequence.length; k++) {
		Art* ankle = partSequence[k];
		if (ankle->rh) {
			if (ankle->part.length) {
				Art* toes = ankle->part.refValue()[0];

				anklePos = -toes->pivotPoint;
				anklePos = (-ankle->ikRot * toes->ikRot) * anklePos;
				anklePos = anklePos + toes->ikPos;
				int last = longPoint.length - 1;
				longPoint[last] = anklePos;
				shortPoint[last] = anklePos;
			}
		}
	}
}
Art* debugPart = 0;

void NikInterpolator::calcIK(Seamless* smls, Art* selHip) {
	if (partSequence.length < 3) {
		return;
	}
	//smls->catMatrix(g_m_accumRot); //does not fix anything noticable regarding front feet sliding when using a relative spine
	//smls->catIMatrix(g_m_iaccumRot);

	if (selHip == 0) {
		if (smls->posCompensator && smls->posCompensator->ankle == partSequence[0]) {
			return;
		}
	}

	for (int k = interpolated.length / 2; --k >= 0;) {
		longPoint[k] = interpolated[k * 2];
		shortPoint[k] = interpolated[k * 2 + 1];
	}

	Art* part0 = partSequence[0];

	//debugPart = smls->partList[3];
	//debugPart->rotation = Rotation(-.995523, -.0882, -.033978, 1.165502);
	//debugPart = partSequence[8];

	if (part0->pin == false && selHip == 0) {
		if (Art* parent = part0->parent) {
			//			if (parent->ik && parent->pin) {
						//parent->ikRot.refValue() = parent->rotation; //for debugging only
						//parent->ikPos.refValue() = parent->translation; //for debugging only
			Vec3f partPos0 = part0->matrix * part0->absolutePivotPoint;
			longPoint[0] = partPos0;
			shortPoint[0] = partPos0;
			//}
		}
	}

	Vec3f anklePos;

	setRhAnklePos(anklePos, partSequence, longPoint, shortPoint);


	double amount = 1;
	double aInc = .5;

	int numberOfTriesCounter = 1000;
	/*
	{
		Vec3f prevCrd;
		double t = 0;
		double tInc = .1;
		float targetDist = .5;
		Rotation rot;
		Vec3f fromVector = Vec3f(0, -1, 0);

		getPointOnCurve(longPoint, longPoint, longLine->weight, longLine->uKnot, longLine->minU(), longLine->maxU(), longLine->uDegree, prevCrd, t, tInc, targetDist, rot, fromVector);
		seg0->rotation = rot;

		tInc = .1;
		getPointOnCurve(longPoint, longPoint, longLine->weight, longLine->uKnot, longLine->minU(), longLine->maxU(), longLine->uDegree, prevCrd, t, tInc, targetDist, rot, fromVector);
		seg1->rotation = rot;
	}
	*/
	float distDiff = 0;
	float prevDistDiff2 = 10000;

	Vec3f fromVector;

	if (Art* parent = part0->parent) {
		Art* part1 = partSequence[1];
		fromVector = convRot4dto4f(g_m_accumRot[parent->plidx]) * part1->pivotPoint.refValue().normalize();
	}
	else {
		Art* part1 = partSequence[1];
		fromVector = part0->ikRot.refValue() * part1->pivotPoint.refValue().normalize();
	}

	if (calcLongOnly) {
		calkArmPoints(amount, partSequence, prevDistDiff2, 0, fromVector);
	}
	else {
		while (prevDistDiff2 > .0001 && --numberOfTriesCounter > 0) {
			if (calkArmPoints(amount, partSequence, prevDistDiff2, aInc, fromVector)) {
				continue;
			}
			if (calkArmPoints(amount, partSequence, prevDistDiff2, -aInc, fromVector)) {
				continue;
			}

			aInc = aInc * .5;
		}
	}
	if (Art* parent = part0->parent) {
		completeRotations(parent, 0);
	}
	else {
		//g_m_accumRot[0] = Rotation4d();
		//g_m_iaccumRot[0] = Rotation4d();
		//part0->rotation = part0->ikRot * convRot4dto4f(g_m_iaccumRot[part0->plidx]);
		completeRotations(part0, 1);
		//partSequence[8]->rotation.refValue() = Rotation(0, 0, 1, .427018); //we want to know how to get this value without hard wiring it here
		//partSequence[1]->rotation.refValue() = part0->rotation * partSequence[1]->rotation;
		//part0->rotation = Rotation();
		//Rotation rot = convRot4dto4f(g_m_iaccumRot[partSequence.length - 2]);
		//partSequence[partSequence.length - 1]->rotation.refValue() = Rotation(0, 0, 1, -.607677) * Rotation(0, 0, 1, 1.03469145);
		partSequence[partSequence.length - 1]->rotation.refValue() = partSequence[partSequence.length - 1]->ikRot  * -partSequence[0]->rotation * convRot4dto4f(g_m_iaccumRot[partSequence.length - 2]); //not sure this is the correct order for multiplication
	}
	if (selHip) {
		if (Art* parent = part0->parent) {
			smls->catMatrix(g_m_accumRot);
			Vec3f vec = -part0->pivotPoint.refValue();
			//vec = parent->ikRot.refValue() * vec;
			vec = convRot4dto4f(g_m_accumRot[parent->plidx]) * vec;
			//Vec3f vec = -parent->ikRot.refValue() * -part0->pivotPoint.refValue();

			vec += resultPoint[0];
			vec -= parent->pivotPoint;

			parent->translation.refValue() = vec;

		}
	}


	/*
	String str;
	for (int k = 0; k < partSequence.length; k++) {
		Art* part = partSequence[k];
		Vec3f vec = part->absolutePivotPoint;
		str = k;
		str += " ";
		vec.toString(str);
		dLog(str);
	}
	*/
	
	int last = partSequence.length - 1;
	Art* lastPart = partSequence[last];
	if (lastPart->rh) {//this is the important place the rh part rotation is set
		if (lastPart->part.length) {
			Art* parent = partSequence[last - 1];
			Art* child = lastPart->part.refValue()[0];

			Rotation rot = -convRot4dto4f(g_m_accumRot[parent->plidx]) * (-lastPart->ikRot * child->ikRot);
			lastPart->rotation = convRot4dto4f(g_m_accumRot[parent->plidx] * convRot4fto4d(rot) * g_m_iaccumRot[parent->plidx]);

			g_m_accumRot[lastPart->plidx] = convRot4fto4d(lastPart->rotation) * g_m_accumRot[parent->plidx];
			g_m_iaccumRot[lastPart->plidx] = g_m_iaccumRot[parent->plidx] * convRot4fto4d(-lastPart->rotation);


			child->rotation = child->ikRot * convRot4dto4f(g_m_iaccumRot[lastPart->plidx]);

			//part->rotation = Rotation(0, 0, 0, .1);
			//child->rotation = Rotation();
		}
	}
	
	smls->ctrf = true;
	smls->animCPBuildFlag = true;
	smls->doBuild();
	//if (Art* part = dynamic_cast<Art*>(g_selNode.getRaw())) {
	//	if (part->niki == this) {
			//this->houseKeep();
		updateVisualClue = true;
	//	}
	//}
}

void NikInterpolator::calc(const float v, Art& part) {
	smls = part.smls;
	smls->nikiList += this;

	partSequence.length = 0;

	Art* parent = 0;
	if (part.ik == true) {
		if (parent = part.parent) {
			partSequence += parent;
			setNiki(&part, this, partSequence);
		}
	}
	if (parent == 0) {
		setNiki(&part, this, partSequence);
	}
	do_set_fraction(v);

	smls->ikProcessFlag = true;
}

void NikInterpolator::processFields() {
	int nurbLineLength = longPoint.length;
	if (nurbLineLength == 0) {
		nurbLineLength = keyValue.length / key.length / 2;
	}
		
	

	longPoint.length = nurbLineLength;
	shortPoint.length = nurbLineLength;
	resultPoint.length = nurbLineLength;
}

void commonEndForSetupAnim();

void addNiki() {
	if (Art* part = dynamic_cast<Art*>(g_selNode.getRaw())) {
		if (part->part.length) {
			Art* childPart = part->part.refValue()[0];
			if (childPart->ik) {
				if (Anim* anim = Window::animMode()) {
					if (anim->play == false) {
						if (anim->getNikInterpolator(part)) {
						}
						else {
							anim->connectPartToNikInterpolator(part);
							if (NikInterpolator* niki = part->niki) {
								niki->calcLongOnly.refValue() = true;
							}
						}
						if (part->niki) {
							NikInterpolator::addToRetainView(part->niki);
							NikInterpolator::updateVisualClue = true;
						}
					}
				}
			}
		}
	}
	else {
		needPartMsg();
	}

}

void snapTwinFunc() {
	if (Art* part = dynamic_cast<Art*>(g_selNode.getRaw())) {
		if (Anim* anim = Window::animMode()) {
			if (anim->play == false) {
				if (NikInterpolator* niki = part->niki) {
					int idx = niki->index[1];
					if (idx & 1) {
						niki->keyValue.refValue()[idx] = niki->keyValue.refValue()[idx - 1];
					}
					else {
						niki->keyValue.refValue()[idx + 1] = niki->keyValue.refValue()[idx];
					}
				}
			}
		}
	}
	else {
		needPartMsg();
	}
}

void displayNikiPanel() {
	if (Art* part = dynamic_cast<Art*>(g_selNode.getRaw())) {
		if (Anim* anim = Window::animMode()) {
			if (anim->play == false) {
				if (NikInterpolator* niki = part->niki) {
					g_stw->cp->setupNodeWindow(niki);
					Window::activeWindow = g_stw->cp.p;
				}
			}
		}
	}
}
void Art::NikPush::operator=(const bool& par) {
	if (Anim* anim = Window::animMode()) {
		if (anim->play == false) {

			Window::closeWindow(g_stw->cp.p);

			if (home.niki == 0) {
				addNiki();
			}
			if (home.niki) {
				g_stw->cp->setupNodeWindow(home.niki);
				Window::activeWindow = g_stw->cp.p;
			}
		}
	}
}

void NikInterpolator::SelectShort::operator=(const bool& par) {
	refValue() = par;
	if (par) {
		home.focusOffset = 1;
	}
	else {
		home.focusOffset = 0;
	}
	home.updateVisualClue = true;
}
void selectLastSelectedPart();
extern int g_animBarListIndex;
extern bool g_dontBackup;

void NikInterpolator::CalcLongOnly::operator=(const bool& par) {
	refValue() = par;
	if (Anim* anim = Window::getAnimBar()) {
		anim->update = true;
	}
	home.updateVisualClue = true;
}

void NikInterpolator::SnapTwin::operator=(const bool& par) {
	if (Anim* anim = Window::getAnimBar()) {
		snapTwinFunc();
		home.smls->animCPBuildFlag = true;
		anim->update = true;
		home.updateVisualClue = true;
	}
}

int NikInterpolator::getNurbLineLength() const {
	return longPoint.length;
}
int NikInterpolator::getSampleLength() const {
	return longPoint.length * 2;
}

void NikInterpolator::insertPoint(Vec3f newPoint, int k, float ratio) {
	M_Vec3f& keyValue = this->keyValue.refValue();
	M_Float& weight = this->weight.refValue();
	M_Float& knot = this->knot.refValue();

	int lpLength = longPoint.length;
	int numOfInserts = keyValue.length / (lpLength * 2);


	longPoint.insertElement(k + 1, newPoint);
	shortPoint.insertElement(k + 1, newPoint);
	weight.insertElement(k + 1, 1);
	updateNurbsCurveKnotMid0(longPoint.length, knot);


	for (int ni = numOfInserts; --ni >= 0;) {
		int i = ni * lpLength * 2 + k + 2;
		keyValue.insertElement(i + 1, newPoint);
		keyValue.insertElement(i, newPoint);
	}

	processFields();
}


void NikInterpolator::insertControlPoint() {

	Vec3f dpA = longPoint[MultiSelect::clickedIndex];
	Vec3f dpB = longPoint[MultiSelect::clickedIndex + 1];

	Vec3f newPoint = dpA + (dpB - dpA) * insertPointRatio;

	insertPoint(newPoint, MultiSelect::clickedIndex, insertPointRatio);

	MultiSelect::clickedIndex++;
	selPointIdx = MultiSelect::clickedIndex;

	showCursor(false);
	//g_pointer.translation = g_viewMatrix * newPoint;
	//g_cursorOverPointIlsBase->translation = g_pointer.translation;
	isOverPoint();


	updateVisualClue = true;
}

void NikInterpolator::deleteControlPoint() {
	Anim* anim = Window::animMode();
	if (anim == 0){
		return;
	}

	M_Vec3f& keyValue = this->keyValue.refValue();
	M_Float& weight = this->weight.refValue();
	M_Float& knot = this->knot.refValue();

	int lpLength = longPoint.length;
	int numOfInserts = keyValue.length / (lpLength * 2);

	for (int ni = numOfInserts; --ni >= 0;) {
		int i = ni * lpLength * 2 + selPointIdx + 2;
		keyValue.deleteElement(i);
		keyValue.deleteElement(i);
	}

	weight.deleteElement(selPointIdx);
	updateNurbsCurveKnotMid0(weight.length, knot);

	processFields();

	anim->update = true;
}

bool Art::isNikPinEndRelativeEnd() {
	if (ik && pin) {
		if (niki) {
			if (niki->relativeEnd) {
				M_Art& partSequence = niki->partSequence;
				if (this == partSequence[partSequence.length - 1]) {
					return true;
				}
			}
		}
	}
	return false;
}

bool NikInterpolator::isInRetainView() {
	for (int k = retainView.length; --k >= 0;) {
		if (retainView[k] == this) {
			return true;
		}
	}
	return false;
}

void NikInterpolator::addToRetainView(NikInterpolator* par) {
	if(par->isInRetainView()){
		return;
	}
	
	retainView.insertElement(0, par);
	updateVisualClue = true;

	if (retainView.length > retainLimit) {
		retainView.length = retainLimit;
	}
}

void NikInterpolator::initTug() {
	if (enterKeyState()) {
		insertControlPoint();
	}
	else if (MultiSelect::withinSights && smls) {
		M_Vec3f& focusPoint = focusOffset ? shortPoint : longPoint;

		selPointIdx = MultiSelect::clickedIndex;

		Matrix matrix = g_viewMatrix;
		Vec3f vec3f = focusPoint[MultiSelect::clickedIndex];
		g_pointer.translation = matrix * vec3f;

		if (Anim* anim = Window::getAnimBar()) {
			setIndex(anim->slider, selPointIdx, focusPoint.length * 2);


			Art* part = 0;
			if (MultiSelect::clickedIndex == longPoint.length - 1) {
				part = lastPart();
			}
			else{
				part = firstPart();
			}
			g_stw->setResultsForSelectingVertex(part);
		}

		NikInterpolator::updateVisualClue = true;
		g_stw->cp->paint();
		fastFullBuildForAutoIn(false);
		if (false == controlKeyState()) {
			tugFlag = true;
			showCursor(false);
		}
	}
}

Art* NikInterpolator::firstPart() {
	if (partSequence.length) {
		return partSequence[0];
	}
	return 0;
}

Art* NikInterpolator::lastPart() {
	if (partSequence.length) {
		return partSequence[partSequence.length - 1];
	}
	return 0;
}
