/* Seamless3d

Copyright (C) 2001-2017 Graham John Perrett (AKA thyme) thyme@seamless3d.com www.seamless3d.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this
software and associated documentation files (the "Software"), to deal in the Software
without restriction, including without limitation the rights to use, copy, modify,
merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to the following
conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT
OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.
*/
#include "seamless3d.h"

extern M_Int g_pointLookup;
extern Coordinate* g_nurbsIlsCoord;
extern IndexedLineSet* g_nurbsIls;
extern bool g_perspective;
extern ScenePtr g_scene;
extern ArtPtr g_vcNurbsCtrlPolyPart;
extern Ptr<TransformCP> g_transformCP;
extern bool g_showPatchBorderMode;

void setBackfaceMaskForInnerPoints2(M_Bool& m_pointIsVisible, const M_Vec2f& m_screenPoint, int& i0, int& i1, int& i2, int& i3, Vec2f& centerPoint);
int getCenterBegin(const M_CtrlPolyPoint& ctrlPolyPoint, int poly);
bool backFace(const Vec2f& v0, const Vec2f& v1, const Vec2f& v2);
int FindSpan (int n, int p, float u, const float *U);
void BasisFuns ( int i, float u, int p, const float *U, float *Nf );
bool SurfacePoint(
                  float u, float v, int uOrder, int vOrder, int uDimension, int vDimension,
                  const float *uKnot, const float * vKnot,
                  const Vec3f *P,const float *W,
                  Vec3f *dest // result
                  );
void CurvePoint (int numOf, int degree, const float *knot, const Vec3f* point, const float *weight, const float t, Vec3f& outPoint);

void calcPolyCenterScreenPoints(const M_CtrlPolyPoint& ctrlPolyPoint, M_Vec2f& m_screenPoint, M_Bool& m_pointIsVisible, const Matrix& viewMatrix){
    for(int poly = 0; poly < ctrlPolyPoint.length;){
        int len = ctrlPolyPoint[poly].len;
        const int centerBegin = getCenterBegin(ctrlPolyPoint, poly);
        int centerLen = ctrlPolyPoint[centerBegin].len;

        Vec2f centerPoint;
        if(centerLen > 0){
            Vec3f sr = viewMatrix * ctrlPolyPoint[centerBegin].p->dpoint[ctrlPolyPoint[centerBegin].i];
            centerPoint = Vec2f(sr.x, sr.y);
        }
        else{
            centerLen++;
            for(int k = len; --k >= 0;){
                NurbsPatch* patch = ctrlPolyPoint[poly + k].p;
                int idx = ctrlPolyPoint[poly + k].i;
                centerPoint += m_screenPoint[patch->fromPoint + idx];
            }
            centerPoint = centerPoint / len;
        }
        for(int kA = len; --kA >= 0;){
            int idxA = ctrlPolyPoint[poly + kA].i + ctrlPolyPoint[poly + kA].p->fromPoint;

            int kB = kA + 1;
            if(kB == len){
                kB = 0;
            }
            int idxB = ctrlPolyPoint[poly + kB].i + ctrlPolyPoint[poly + kB].p->fromPoint;

            Vec2f vec2fA = m_screenPoint[idxA];
            Vec2f vec2fB = m_screenPoint[idxB];
            if(backFace(vec2fA, vec2fB, centerPoint)){
                m_pointIsVisible[idxA] = true;
                m_pointIsVisible[idxB] = true;
            }
        }
        poly = centerBegin + centerLen;
    }
}

void NurbsPatch::calcScreenPoints(M_Vec2f& m_screenPoint, const Matrix& viewMatrix){
    //if(cage == false){
    //	return;
    //}

    Vec3f sr;
    int i = 0;

    int uDim = this->dim.x;
    int vDim = this->dim.y;

    if(showAll || g_showPatchBorderMode){
        for(int v=0; v < vDim; v++){
            for(int u = 0; u < uDim; u++){
                int i = v * uDim + u;
                sr = viewMatrix * dpoint[i];
                m_screenPoint[fromPoint + i] = Vec2f(sr.x, sr.y);
            }
        }
    }
    else{
        if(uJoinEnds){
            if(NurbsPatch::MID_1 == checkNurbsType(vDim, vKnot)){
                if(vDim == 1){
                    for(int u = 1; u < uDim; u++){
                        Vec3f pnt = matrix * dpoint[u];
                        sr = viewMatrix * dpoint[i];
                        m_screenPoint[fromPoint + i] = Vec2f(sr.x, sr.y);
                    }
                }
                else{
                    for(int v = 0; v < vDim; v += 2){
                        for(int u = 1; u < uDim; u += 2){
                            int i = v * uDim + u;
                            Vec3f pnt = matrix*dpoint[i];
                            sr = viewMatrix * dpoint[i];
                            m_screenPoint[fromPoint + i] = Vec2f(sr.x, sr.y);
                        }
                        if(v == 0 || v == vDim - 2){
                            v--;
                        }
                    }
                }
            }
            else{
                if(vDim == 1){
                    for(int u = 1; u < uDim; u++){
                        sr = viewMatrix * dpoint[u];
                        m_screenPoint[fromPoint + u] = Vec2f(sr.x, sr.y);
                    }
                }
                else{
                    for(int v = 0; v < vDim; v++){
                        for(int u = 1; u < uDim; u += 2){
                            int i = v * uDim + u;
                            sr = viewMatrix * dpoint[i];
                            m_screenPoint[fromPoint + i] = Vec2f(sr.x, sr.y);
                        }
                    }
                }
            }
        }
        else{
            if(NurbsPatch::MID_1 == checkNurbsType(vDim, vKnot)){
                if(vDim == 1){
                    for(int u = 0; u < uDim; u++){
                        sr = viewMatrix * dpoint[u];
                        m_screenPoint[fromPoint + u] = Vec2f(sr.x, sr.y);
                    }
                }
                else{
                    for(int v = 0; v < vDim; v += 2){
                        for(int u = 0; u < uDim; u += 2){
                            i = v * uDim + u;
                            sr = viewMatrix * dpoint[i];
                            m_screenPoint[fromPoint + i] = Vec2f(sr.x, sr.y);
                            if(u == 0 || u == uDim - 2){
                                u--;
                            }
                        }
                        if(v == 0 || v == vDim - 2){
                            v--;
                        }

                    }
                }
            }
        }
    }
}



void setBackfaceMaskForInnerPoints3(M_Bool& m_pointIsVisible, const M_Vec2f& m_screenPoint, int i0, int i1, int i2, int i3){
	Vec2f vec2f0 = m_screenPoint[i0];
	Vec2f vec2f1 = m_screenPoint[i1];
	Vec2f vec2f2 = m_screenPoint[i2];
	Vec2f vec2f3 = m_screenPoint[i3];

	if(backFace(vec2f0, vec2f1, vec2f3)){
		m_pointIsVisible[i0] = true;
	}
	if(backFace(vec2f3, vec2f0, vec2f2)){
		m_pointIsVisible[i3] = true;
	}

}


void setBackfaceMaskForInnerPoints4(M_Bool& m_pointIsVisible, const M_Vec2f& m_screenPoint, int i0, int i1, int i2){
	Vec2f vec2f0 = m_screenPoint[i0];
	Vec2f vec2f1 = m_screenPoint[i1];
	Vec2f vec2f2 = m_screenPoint[i2];

	if(backFace(vec2f0, vec2f1, vec2f2)){
		m_pointIsVisible[i0] = true;
		m_pointIsVisible[i1] = true;
		m_pointIsVisible[i2] = true;
	}
}


void setBackfaceMaskForInnerPoints5(M_Bool& m_pointIsVisible, const M_Vec2f& m_screenPoint, int i0, int i1, int i2, int i3, int i4){
	Vec2f vec2f0 = m_screenPoint[i0];
	Vec2f vec2f1 = m_screenPoint[i1];
	Vec2f vec2f2 = m_screenPoint[i2];
	Vec2f vec2f3 = m_screenPoint[i3];
	Vec2f vec2f4 = m_screenPoint[i4];

	if(backFace(vec2f0, vec2f1, vec2f2)){
		m_pointIsVisible[i0] = true;
		m_pointIsVisible[i1] = true;
		m_pointIsVisible[i2] = true;
	}
	if(backFace(vec2f2, vec2f3, vec2f4)){
		m_pointIsVisible[i2] = true;
		m_pointIsVisible[i3] = true;
		m_pointIsVisible[i4] = true;
	}

}

void setBackfaceMaskForInnerPoints6(M_Bool& m_pointIsVisible, const M_Vec2f& m_screenPoint, int i0, int i1, int i2, int i3){
	Vec2f vec2f0 = m_screenPoint[i0];
	Vec2f vec2f1 = m_screenPoint[i1];
	Vec2f vec2f2 = m_screenPoint[i2];
	Vec2f vec2f3 = m_screenPoint[i3];

	if(backFace(vec2f0, vec2f1, vec2f2)){
		m_pointIsVisible[i1] = true;
	}
	if(backFace(vec2f1, vec2f2, vec2f3)){
		m_pointIsVisible[i2] = true;
	}
	if(backFace(vec2f2, vec2f3, vec2f0)){
		m_pointIsVisible[i3] = true;
	}
	if(backFace(vec2f3, vec2f0, vec2f1)){
		m_pointIsVisible[i0] = true;
	}

}

void NurbsPatch::setBackfaceMaskForInnerPoints(M_Bool& m_pointIsVisible, const M_Vec2f& m_screenPoint){
    if(cage == false){
        return;
    }


    int uDim = this->dim.x;
    int vDim = this->dim.y;

    Vec2f centerPoint;
    if(isCtrlPolyRef == false){
        if(vDim == 1){
            for(int u = 0; u < uDim; u++){
                m_pointIsVisible[fromPoint + u] = true;
            }
        }
		else{
			if(isInsideOut()){
				if(NurbsPatch::MID_1 == checkNurbsType(vDim, vKnot)){
					for(int v = 0; v < vDim - 1; v += vDim - 2){
						for(int u = 1; u < uDim - 3; u += 2){
							int i0 = fromPoint + v * uDim + u;
							int i1 = i0 + uDim;
							int i2 = i1 + 2;
							int i3 = i2 - uDim;
							setBackfaceMaskForInnerPoints2(m_pointIsVisible, m_screenPoint, i0, i1, i2, i3, centerPoint);
						}
					}
				}
			}
			else{
				if(NurbsPatch::MID_1 == checkNurbsType(vDim, vKnot)){
					for(int v = 0; v < vDim - 1; v += vDim - 2){
						for(int u = 1; u < uDim - 3; u += 2){
							int i0 = fromPoint + v * uDim + u;
							int i1 = i0 + 2;
							int i2 = i1 + uDim;
							int i3 = i2 - 2;
							setBackfaceMaskForInnerPoints2(m_pointIsVisible, m_screenPoint, i0, i1, i2, i3, centerPoint);
						}
					}
				}
			}
		}
		return;
    }
    if(isInsideOut()){
        if(NurbsPatch::MID_1 == checkNurbsType(vDim, vKnot)){
            if(vDim == 1){
                for(int u = 0; u < uDim; u++){
                    m_pointIsVisible[fromPoint + u] = true;
                }
            }
            else{
				int i0 = 0;
				int i1 = 0;
				int i2 = 0;
				int i3 = 0;
				int i4 = 0;
                for(int v = 1; v < vDim - 3; v += 2){
                    for(int u = 1; u < uDim - 3; u += 2){
                        i0 = fromPoint + v * uDim + u;
                        i1 = i0 + uDim * 2;
                        i2 = i1 + 2;
                        i3 = i2 - uDim * 2;
                        setBackfaceMaskForInnerPoints2(m_pointIsVisible, m_screenPoint, i0, i1, i2, i3, centerPoint);
                    }
				}
				if(g_showPatchBorderMode){
					for(int u = 1; u < uDim - 3; u += 2){
						i0 = fromPoint + u + uDim;//south edge
						i1 = i0 - uDim;
						i2 = i1 + 1;
						i3 = i2 + 1;
						i4 = i3 + uDim;
						setBackfaceMaskForInnerPoints5(m_pointIsVisible, m_screenPoint, i4, i3, i2, i1, i0);

						i0 = fromPoint + (vDim - 2) * uDim + u + 2;//north edge
						i1 = i0 + uDim;
						i2 = i1 - 1;
						i3 = i2 - 1;
						i4 = i3 - uDim;
						setBackfaceMaskForInnerPoints5(m_pointIsVisible, m_screenPoint, i4, i3, i2, i1, i0);
					}
					for(int v = 1; v < vDim - 3; v += 2){
						i0 = fromPoint + (v + 2) * uDim + 1;//west edge
						i1 = i0 - 1;
						i2 = i1 - uDim;
						i3 = i2 - uDim;
						i4 = i3 + 1;
						setBackfaceMaskForInnerPoints5(m_pointIsVisible, m_screenPoint, i4, i3, i2, i1, i0);

						i0 = fromPoint + v * uDim + uDim - 2;//east edge
						i1 = i0 + 1;
						i2 = i1 + uDim;
						i3 = i2 + uDim;
						i4 = i3 - 1;
						setBackfaceMaskForInnerPoints5(m_pointIsVisible, m_screenPoint, i4, i3, i2, i1, i0);
					}
					//south west corner
					i0 = fromPoint;
					i1 = i0 + 1;
					i2 = i1 + uDim;
					i3 = i2 - 1;
					setBackfaceMaskForInnerPoints6(m_pointIsVisible, m_screenPoint, i3, i2, i1, i0);
					
					//south east corner
					i0 = fromPoint + uDim - 2;
					i1 = i0 + 1;
					i2 = i1 + uDim;
					i3 = i2 - 1;
					setBackfaceMaskForInnerPoints6(m_pointIsVisible, m_screenPoint, i3, i2, i1, i0);

					//north west corner
					i0 = fromPoint + (vDim - 1) * uDim;
					i1 = i0 - uDim;
					i2 = i1 + 1;
					i3 = i2 + uDim;
					setBackfaceMaskForInnerPoints6(m_pointIsVisible, m_screenPoint, i3, i2, i1, i0);

					//north east corner
					i0 = fromPoint + vDim * uDim - 1;
					i1 = i0 - 1;
					i2 = i1 - uDim;
					i3 = i2 + 1;
					setBackfaceMaskForInnerPoints6(m_pointIsVisible, m_screenPoint, i3, i2, i1, i0);
				}
			}
		}
    }
    else{
        if(NurbsPatch::MID_1 == checkNurbsType(vDim, vKnot)){
            if(vDim == 1){
                for(int u = 0; u < uDim; u++){
                    m_pointIsVisible[fromPoint + u] = true;
                }
            }
            else{
				int i0 = 0;
				int i1 = 0;
				int i2 = 0;
				int i3 = 0;
				int i4 = 0;
                for(int v = 1; v < vDim - 3; v += 2){
                    for(int u = 1; u < uDim - 3; u += 2){
                        i0 = fromPoint + v * uDim + u;
                        i1 = i0 + 2;
                        i2 = i1 + uDim * 2;
                        i3 = i2 - 2;
                        setBackfaceMaskForInnerPoints2(m_pointIsVisible, m_screenPoint, i0, i1, i2, i3, centerPoint);
                    }
				}
				if(g_showPatchBorderMode){
					for(int u = 1; u < uDim - 3; u += 2){
						i0 = fromPoint + u + uDim;//south edge
						i1 = i0 - uDim;
						i2 = i1 + 1;
						i3 = i2 + 1;
						i4 = i3 + uDim;
						setBackfaceMaskForInnerPoints5(m_pointIsVisible, m_screenPoint, i0, i1, i2, i3, i4);

						i0 = fromPoint + (vDim - 2) * uDim + u + 2;//north edge
						i1 = i0 + uDim;
						i2 = i1 - 1;
						i3 = i2 - 1;
						i4 = i3 - uDim;
						setBackfaceMaskForInnerPoints5(m_pointIsVisible, m_screenPoint, i0, i1, i2, i3, i4);
					}
					for(int v = 1; v < vDim - 3; v += 2){
						i0 = fromPoint + (v + 2) * uDim + 1;//west edge
						i1 = i0 - 1;
						i2 = i1 - uDim;
						i3 = i2 - uDim;
						i4 = i3 + 1;
						setBackfaceMaskForInnerPoints5(m_pointIsVisible, m_screenPoint, i0, i1, i2, i3, i4);

						i0 = fromPoint + v * uDim + uDim - 2;//east edge
						i1 = i0 + 1;
						i2 = i1 + uDim;
						i3 = i2 + uDim;
						i4 = i3 - 1;
						setBackfaceMaskForInnerPoints5(m_pointIsVisible, m_screenPoint, i0, i1, i2, i3, i4);
					}
					//south west corner
					i0 = fromPoint;
					i1 = i0 + 1;
					i2 = i1 + uDim;
					i3 = i2 - 1;
					setBackfaceMaskForInnerPoints6(m_pointIsVisible, m_screenPoint, i0, i1, i2, i3);
					
					//south east corner
					i0 = fromPoint + uDim - 2;
					i1 = i0 + 1;
					i2 = i1 + uDim;
					i3 = i2 - 1;
					setBackfaceMaskForInnerPoints6(m_pointIsVisible, m_screenPoint, i0, i1, i2, i3);

					//north west corner
					i0 = fromPoint + (vDim - 1) * uDim;
					i1 = i0 - uDim;
					i2 = i1 + 1;
					i3 = i2 + uDim;
					setBackfaceMaskForInnerPoints6(m_pointIsVisible, m_screenPoint, i0, i1, i2, i3);

					//north east corner
					i0 = fromPoint + vDim * uDim - 1;
					i1 = i0 - 1;
					i2 = i1 - uDim;
					i3 = i2 + 1;
					setBackfaceMaskForInnerPoints6(m_pointIsVisible, m_screenPoint, i0, i1, i2, i3);
				}
            }
        }
    }
}

void NurbsPatch::selectRing(int v, int uDim, int vDim, M_<NurbsPatch*>& contigList){
    int idx = 0;
    if(uJoinEnds){
        if(NurbsPatch::MID_1 == checkNurbsType(vDim, vKnot)){
            if(isCtrlPolyRef){
                for(int u = 1; u < uDim - 1; u += 2){
                    idx = v * uDim + u;
                    contigList[fromPoint + idx ] = this;
                }
            }
            else{
                idx = v * uDim;
                MultiSelect::beginDragPos += dpoint[idx];
                MultiSelect::pointList += ControlPoint(this, idx);
                for(int u = 1; u<uDim; u+=2){
                    idx = v * uDim + u;
                    contigList[fromPoint + idx ] = this;
                }
                idx = v * uDim + uDim - 1;
                contigList[fromPoint + idx ] = this;
            }
        }
        else if(NurbsPatch::MID_0 == checkNurbsType(vDim, vKnot)){
            if(isCtrlPolyRef){
                MultiSelect::uRingDim = uDim / 2;
                MultiSelect::vRingDim = vDim - v - 1;
                for(int u = 1; u < uDim - 1; u += 2){
                    idx = v * uDim+u;
                    contigList[fromPoint + idx ] = this;
                }
            }
            else{
                MultiSelect::uRingDim = uDim / 2;
                MultiSelect::vRingDim = vDim - v;
                idx = v * uDim + uDim - 2;
                for(int u = 1; u < uDim - 1; u+=2){
                    idx = v * uDim + u;
                    contigList[fromPoint + idx ] = this;
                }
            }
        }
    }
    else {
        if(NurbsPatch::MID_1 == checkNurbsType(vDim, vKnot)){
            if(isCtrlPolyRef){
                MultiSelect::uRingDim = uDim / 2;
                MultiSelect::vRingDim = (vDim / 2 - 1) - v / 2;
                for(int u = 1; u < uDim; u += 2){
                    idx = v * uDim + u;

                    contigList[fromPoint + idx ] = this;
                }
            }
            else{
                MultiSelect::uRingDim = uDim / 2;
                MultiSelect::vRingDim = (vDim / 2) - v / 2;
                int idx = v * uDim;
                contigList[fromPoint + idx ] = this;
                for(int u = 1; u < uDim; u += 2){
                    idx = v * uDim + u;
                    contigList[fromPoint + idx ] = this;
                }
                idx = v * uDim + uDim - 1;
                contigList[fromPoint + idx ] = this;
            }
        }
        else{
            if(NurbsPatch::MID_1 == checkNurbsType(uDim, uKnot)){
                if(isCtrlPolyRef){
                    MultiSelect::uRingDim = uDim / 2;
                    MultiSelect::vRingDim = (vDim / 2 - 1) - v / 2;
                    for(int u = 1; u < uDim - 1; u += 2){
                        idx = v * uDim + u;
                        contigList[fromPoint + idx ] = this;
                    }
                }
                else{
                    idx = v * uDim;
                    MultiSelect::beginDragPos += dpoint[idx];
                    MultiSelect::pointList += ControlPoint(this, idx);
                    for(int u = 1; u < uDim; u +=2){
                        idx = v * uDim + u;
                        contigList[fromPoint + idx ] = this;
                    }
                    idx = v * uDim + uDim - 1;
                    contigList[fromPoint + idx ] = this;
                }
            }
            else{
                if(isCtrlPolyRef){
                    for(int u = 1; u < uDim - 1; u++){
                        idx = v * uDim + u;
                        contigList[fromPoint + idx ] = this;
                    }
                }
                else {
                    for(int u = 0; u < uDim; u++){
                        idx = v * uDim + u;
                        contigList[fromPoint + idx ] = this;
                    }
                }
            }
        }
    }
}

void NurbsPatch::selectCol(int u, int uDim, int vDim, M_<NurbsPatch*>& contigList){
    int idx = 0;
    {
        if(NurbsPatch::MID_1 == checkNurbsType(uDim, uKnot)){
            if(isCtrlPolyRef){
                for(int v = 1; v < vDim - 1; v += 2){
                    idx = v * uDim + u;
                    contigList[fromPoint + idx ] = this;
                }
            }
            else{
                int idx = u;
                contigList[fromPoint + idx ] = this;
                for(int v = 1; v < vDim; v += 2){
                    idx = v * uDim + u;
                    contigList[fromPoint + idx ] = this;
                }
				idx = (vDim - 1) * uDim + u;
				contigList[fromPoint + idx ] = this;
			}
		}
        else{
            if(isCtrlPolyRef){
                for(int v = 1; v < vDim - 1; v++){
                    idx = v * uDim + u;
                    contigList[fromPoint + idx ] = this;
                }
            }
            else {
                for(int v = 0; v < vDim; v++){
                    idx = v * uDim + u;
                    contigList[fromPoint + idx ] = this;
                }
            }
        }
	}
}

/*

							patch->selectControlPointsStraightCage(contigList, perspective);

							for(int i = 0; i < ctrlPolyPoint.length;){
								int ctrlPoly = i;
								int len = ctrlPolyPoint[ctrlPoly].len;
								for(int k = len; --k >= 0;){
									NurbsPatch* patch = ctrlPolyPoint[i].p;
									int idx = ctrlPolyPoint[i].i;
									if(patch->cage && MultiSelect::isInBox(patch->dpoint[idx])){
										contigList[patch->fromPoint + idx] = patch;
									}
									i++;
								}
								i = getNextCtrlPoly(ctrlPoly);
							}
						}
					}
				}
			}
		}
	}
	MultiSelect::pointList.length = 0;
	for(int k = 0; k < pointTotal; k++){
		if(NurbsPatch* patch = contigList[k]){
			MultiSelect::pointList += ControlPoint(patch, k - patch->fromPoint);
		}
	}
	g_transformCP->calcPivotPoint();
	g_transformCP->resetRotationLever();
}
*/

void NurbsPatch::selectRings(int v, int uDim, int vDim){
    if(cage == false){
        return;
    }
    int idx = 0;
    if(uJoinEnds){
        if(NurbsPatch::MID_1 == checkNurbsType(vDim, vKnot)){
            if(isCtrlPolyRef){
                for(; v < vDim - 1; v += 2){
                    for(int u = 1; u < uDim - 1; u += 2){
                        idx = v * uDim + u;
                        MultiSelect::beginDragPos += dpoint[idx];
                        MultiSelect::pointList += ControlPoint(this, idx);
                    }
                }
            }
            else{
                for(; v < vDim; v += 2){
                    idx = v * uDim;
                    MultiSelect::beginDragPos += dpoint[idx];
                    MultiSelect::pointList += ControlPoint(this, idx);
                    for(int u = 1; u<uDim; u+=2){
                        idx = v * uDim + u;
                        MultiSelect::beginDragPos += dpoint[idx];
                        MultiSelect::pointList += ControlPoint(this, idx);
                    }
                    idx = v * uDim + uDim - 1;
                    MultiSelect::beginDragPos += dpoint[idx];
                    MultiSelect::pointList += ControlPoint(this, idx);

                    if(v == 0 || v == vDim - 2){
                        v--;
                    }
                }
            }
        }
        else if(NurbsPatch::MID_0 == checkNurbsType(vDim, vKnot)){
            if(isCtrlPolyRef){
                if(vDim > 3){
                    MultiSelect::uRingDim = uDim / 2;
                    MultiSelect::vRingDim = vDim - v - 1;
                    for(; v < vDim - 1; v++){
                        for(int u = 1; u < uDim - 1; u += 2){
                            idx = v * uDim+u;
                            MultiSelect::beginDragPos += dpoint[idx];
                            MultiSelect::pointList += ControlPoint(this, idx);
                        }
                    }
                }
            }
            else{
                MultiSelect::uRingDim = uDim / 2;
                MultiSelect::vRingDim = vDim - v;
                for(; v < vDim; v++){
                    idx = v * uDim + uDim - 2;
                    for(int u = 1; u < uDim - 1; u+=2){
                        idx = v * uDim + u;
                        MultiSelect::beginDragPos += dpoint[idx];
                        MultiSelect::pointList += ControlPoint(this, idx);
                    }
                }
            }
        }
    }
    else {
        if(NurbsPatch::MID_1 == checkNurbsType(vDim, vKnot)){
            if(isCtrlPolyRef){
                if(vDim > 5){
                    MultiSelect::uRingDim = uDim / 2;
                    MultiSelect::vRingDim = (vDim / 2 - 1) - v / 2;
                    for(; v < vDim - 1; v += 2){
                        for(int u = 1; u < uDim - 1; u += 2){
                            idx = v * uDim + u;
                            MultiSelect::beginDragPos += dpoint[idx];
                            MultiSelect::pointList += ControlPoint(this, idx);
                        }
                    }
                }
            }
            else{
                MultiSelect::uRingDim = uDim / 2;
                MultiSelect::vRingDim = (vDim / 2) - v / 2;
                for(; v < vDim; v += 2){
                    int idx = v * uDim;
                    MultiSelect::beginDragPos += dpoint[idx];
                    MultiSelect::pointList += ControlPoint(this, idx);
                    for(int u = 1; u < uDim; u += 2){
                        idx = v * uDim + u;
                        MultiSelect::beginDragPos += dpoint[idx];
                        MultiSelect::pointList += ControlPoint(this, idx);
                    }
                    idx = v * uDim + uDim - 1;
                    MultiSelect::beginDragPos += dpoint[idx];
                    MultiSelect::pointList += ControlPoint(this, idx);

                    if(v == 0 || v == vDim - 2){
                        v--;
                    }
                }
            }
        }
        else{
            if(NurbsPatch::MID_1 == checkNurbsType(uDim, uKnot)){
                if(isCtrlPolyRef){
                    if(vDim > 3){
                        for(; v < vDim - 1; v++){
                            for(int u = 1; u < uDim - 1; u += 2){
                                idx = v * uDim+u;
                                MultiSelect::beginDragPos += dpoint[idx];
                                MultiSelect::pointList += ControlPoint(this, idx);
                            }
                        }
                    }
                }
                else{
                    for(; v < vDim; v++){
                        idx = v * uDim;
                        MultiSelect::beginDragPos += dpoint[idx];
                        MultiSelect::pointList += ControlPoint(this, idx);
                        for(int u = 1; u < uDim; u +=2){
                            idx = v * uDim + u;
                            MultiSelect::beginDragPos += dpoint[idx];
                            MultiSelect::pointList += ControlPoint(this, idx);
                        }
                        idx = v * uDim + uDim - 1;
                        MultiSelect::beginDragPos += dpoint[idx];
                        MultiSelect::pointList += ControlPoint(this, idx);
                    }
                }
            }
            else{
                if(isCtrlPolyRef){
                    for(; v < vDim - 1; v++){
                        for(int u = 1; u < uDim - 1; u++){
                            idx = v * uDim + u;
                            MultiSelect::beginDragPos += dpoint[idx];
                            MultiSelect::pointList += ControlPoint(this, idx);
                        }
                    }
                }
                else {
                    for(; v < vDim; v++){
                        for(int u = 0; u < uDim; u++){
                            idx = v * uDim + u;
                            MultiSelect::beginDragPos += dpoint[idx];
                            MultiSelect::pointList += ControlPoint(this, idx);
                        }
                    }
                }
            }
        }
    }
}

bool NurbsPatch::isInsideOut()const{
    bool insideOut = this->insideOut;
    if(stem.length){
        if(Stem* stem = dynamic_cast<Stem*>(this->stem[0].p)){
            insideOut = stem->insideOut;
        }
    }
    return insideOut;
}

void NurbsPatch::setInsideOut(){
    insideOut.refValue() = true;
    for(int k = stem.length; --k >= 0;){
        if(Stem* stem = dynamic_cast<Stem*>(this->stem.refValue()[k].p)){
            stem->insideOut = true;
        }
    }
}

void NurbsPatch::setInsideOut2(){
    M_Vec3f& point = this->point.refValue();
    M_Float& weight = this->weight.refValue();
	M_Int& cpIndex = this->cpIndex.refValue();
    M_Vec3f storePoint = point;
    M_Float storeWeight = weight;
	M_Int storeCpIndex = cpIndex;
    M_Int lookup;
    lookup.length = point.length;

    const Vec2f& dim = this->dim;
    int uDim = dim.x;
    int vDim = dim.y;
    for(int v = vDim; --v >= 0;){
        for(int u = uDim; --u >= 0;){
            int d = v * uDim + u;
            int s = v * uDim + uDim - 1 - u;
            point[d] = storePoint[s];
            weight[d] = storeWeight[s];
			cpIndex[d] = storeCpIndex[s];
            lookup[d] = s;
        }
    }
    smls->redirectStitchPolyIndices(this, lookup);
}

void NurbsPatch::setInsideOut4(){
	setInsideOut();

    M_Vec3f& point = this->point.refValue();
    M_Float& weight = this->weight.refValue();
	M_Int& cpIndex = this->cpIndex.refValue();
    M_Vec3f storePoint = point;
    M_Float storeWeight = weight;
	M_Int storeCpIndex = cpIndex;

    const Vec2f& dim = this->dim;
    int uDim = dim.x;
    int vDim = dim.y;
    for(int v = vDim; --v >= 0;){
        for(int u = uDim; --u >= 0;){
            int d = v * uDim + u;
            int s = v * uDim + uDim - 1 - u;
            point[d] = storePoint[s];
            weight[d] = storeWeight[s];
			cpIndex[d] = storeCpIndex[s];
        }
    }
}

void NurbsPatch::setInsideOut3(){
    M_Vec3f& point = this->point.refValue();
	//M_Int& cpIndex = this->cpIndex.refValue();
    M_Float& weight = this->weight.refValue();

    static M_Vec3f storePoint;
	//static M_Int storeCpIndex;
    static M_Float storeWeight;

	storePoint = point;
	//storeCpIndex = cpIndex;
	storeWeight = weight;

    const Vec2f& dim = this->dim;
    int uDim = dim.x;
    int vDim = dim.y;
    for(int v = vDim; --v >= 0;){
        for(int u = uDim; --u >= 0;){
            int d = v * uDim + u;
            int s = v * uDim + uDim - 1 - u;
            point[d] = storePoint[s];
			//cpIndex[d] = storeCpIndex[s];
            weight[d] = storeWeight[s];
        }
    }
}

int NurbsPatch::getMirrorStateViaPart(){ //0 == no mirror, 1 == centerMirror, 2 has twin
    if(stem.length){
        if(Stem* stem = dynamic_cast<Stem*>(this->stem[0].p)){
            if(stem->part.length){
                if(Art* part = dynamic_cast<Art*>(stem->part.refValue()[0].p)){
                    if(part->mirrorPart){
                        return 2;
                    }
                    if(part->mirror){
                        return 1;
                    }
                }
            }
        }
    }
    return 0;
}

void NurbsPatch::doBuild(){
    M_Node& m_stem = stem.refValue();
    int len = m_stem.length;
    for(int k = 0; k < len; k++){
        dynamic_cast<Stem*>(m_stem[k].p)->doBuild(0, this);
    }
}

void NurbsPatch::generate2(M_Vec3f& coord, const int h){

    int uDim=dim.x;
    int vDim=dim.y;
    int uOrder=order.x;
    int vOrder=order.y;

    Vec3f* coordPtr = coord.getArray();
    float* uKnotPtr = uKnot.refValue().getArray();
    float* vKnotPtr = vKnot.refValue().getArray();
    Vec3f* pointPtr = point.refValue().getArray();
    float* weightPtr = weight.refValue().getArray();

    Vec3f* topPointPtr = pointPtr+uDim*vDim-uDim;
    float* topWeightPtr = weightPtr+uDim*vDim-uDim;

    Vec3f* r_pointPtr = r_point.getArray();
    float* r_weightPtr = r_weight.getArray();

    for(int i=coord.length; --i>=h;){

        Vec3f t = coord[i];
        if(t.x>1){
            t.x-=1;
        }
        if(t.x<0 || t.y<0 || t.y>1){
            continue;
        }
        float u = (maxU()-minU())*t.x + minU();
        float v = (maxV()-minV())*t.y + minV();


        if(t.x==1){
            CurvePoint(vDim-1, vDegree, vKnotPtr, r_pointPtr, r_weightPtr, v, *(coordPtr+i));
        }
        else if(t.y==1){
            CurvePoint(uDim-1, uDegree, uKnotPtr, topPointPtr, topWeightPtr, u, *(coordPtr+i));
        }
        else{
            SurfacePoint(u, v, uOrder, vOrder, uDim, vDim, uKnotPtr, vKnotPtr, pointPtr, weightPtr, coordPtr+i);
        }
    }
    if(FuseSurface* fuseSurface = dynamic_cast<FuseSurface*>(this->fuseSurface.p)){
        fuseSurface->doFuse(coord, *this);
    }
}

inline void generateWithNormals2(const float tx, const float ty,
                                const float& uMax, const float& vMax, const float& uMin, const float& vMin,
                                const int& uDim, const int& vDim, const int& uOrder, const int& vOrder, const int& uDegree, const int& vDegree,
                                Vec3f& coordDst, const float*& uKnotPtr, const float*& vKnotPtr, const Vec3f*& pointPtr, const float*& weightPtr,
                                const Vec3f*& topPointPtr, const float*& topWeightPtr, const Vec3f*& r_pointPtr, const float*& r_weightPtr){

    float u = (uMax-uMin) * tx + uMin;
    float v = (vMax-vMin) * ty + vMin;

    if(tx == 1){
        CurvePoint(vDim-1, vDegree, vKnotPtr, r_pointPtr, r_weightPtr, v, coordDst);
    }
    else if(ty == 1){
        CurvePoint(uDim-1, uDegree, uKnotPtr, topPointPtr, topWeightPtr, u, coordDst);
    }
    else{
        SurfacePoint(u, v, uOrder, vOrder, uDim, vDim, uKnotPtr, vKnotPtr, pointPtr, weightPtr, &coordDst);
    }
}

void NurbsPatch::generateWithNormals(M_Vec3f& coord, M_Vec3f& normal, const int h){

    int uDim = dim.x;
    int vDim = dim.y;
    int uOrder = order.x;
    int vOrder = order.y;

    Vec3f* coordPtr = coord.getArray();
    const float* uKnotPtr = uKnot.refValue().getArray();
    const float* vKnotPtr = vKnot.refValue().getArray();
    const Vec3f* pointPtr = point.refValue().getArray();
    const float* weightPtr = weight.refValue().getArray();

    const Vec3f* topPointPtr = pointPtr+uDim*vDim-uDim;
    const float* topWeightPtr = weightPtr+uDim*vDim-uDim;

    const Vec3f* r_pointPtr = r_point.getArray();
    const float* r_weightPtr = r_weight.getArray();

    const float uMax = maxU();
    const float vMax = maxV();
    const float uMin = minU();
    const float vMin = minV();

    normal.length = coord.length;

    for(int i = coord.length; --i>=h;){


        Vec3f t = coord[i];
        float remTX = t.x;
        if(t.x > 1){
            t.x -= 1;
        }
        if(t.x < 0 || t.y < 0 || t.y > 1){
            continue;
        }
        Vec3f a = coordPtr[i];
        Vec3f b;
        Vec3f c;
        generateWithNormals2(t.x, t.y, uMax, vMax, uMin, vMin, uDim, vDim, uOrder, vOrder, uDegree, vDegree, a, uKnotPtr, vKnotPtr, pointPtr, weightPtr, topPointPtr, topWeightPtr, r_pointPtr, r_weightPtr);
        bool flip = false;
        t.x += .0001;
        if(t.x > 1){
            t.x -= .0002;
            flip = true;
        }
        t.y += .0001;
        if(t.y > 1){
            t.y -= .0002;
            flip = !flip;
        }
        generateWithNormals2(t.x, t.y, uMax, vMax, uMin, vMin, uDim, vDim, uOrder, vOrder, uDegree, vDegree, b, uKnotPtr, vKnotPtr, pointPtr, weightPtr, topPointPtr, topWeightPtr, r_pointPtr, r_weightPtr);
        t.x = remTX;
        generateWithNormals2(t.x, t.y, uMax, vMax, uMin, vMin, uDim, vDim, uOrder, vOrder, uDegree, vDegree, c, uKnotPtr, vKnotPtr, pointPtr, weightPtr, topPointPtr, topWeightPtr, r_pointPtr, r_weightPtr);

        coordPtr[i] = a;
        if(!flip){
            normal[i] = -calcNormal(a, b, c);
        }
        else{
            normal[i] = calcNormal(a, b, c);
        }
    }
    if(scale != Vec3f(1,1,1) || rotation != Rotation() || translation != Vec3f()){
        for(int i=coord.length; --i>=h;){
            coord[i]*=matrix;
        }
    }
    if(FuseSurface* fuseSurface = dynamic_cast<FuseSurface*>(this->fuseSurface.p)){
        fuseSurface->doFuse(coord, *this);
    }
}


Vec3f NurbsPatch::generateNormal(Vec2f t)const{
    Vec3f normal;

    int uDim = dim.x;
    int vDim = dim.y;
    int uOrder = order.x;
    int vOrder = order.y;

    const float* uKnotPtr = uKnot.refValue().getArray();
    const float* vKnotPtr = vKnot.refValue().getArray();
    const Vec3f* pointPtr = point.refValue().getArray();
    const float* weightPtr = weight.refValue().getArray();

    const Vec3f* topPointPtr = pointPtr+uDim*vDim-uDim;
    const float* topWeightPtr = weightPtr+uDim*vDim-uDim;

    const Vec3f* r_pointPtr = r_point.getArray();
    const float* r_weightPtr = r_weight.getArray();

    const float uMax = maxU();
    const float vMax = maxV();
    const float uMin = minU();
    const float vMin = minV();



    float remTX = t.x;
    if(t.x > 1){
        t.x -= 1;
    }
    if(t.x < 0 || t.y < 0 || t.y > 1){
        return Vec3f(0, 0, 1);
    }
    Vec3f a;
    Vec3f b;
    Vec3f c;
    generateWithNormals2(t.x, t.y, uMax, vMax, uMin, vMin, uDim, vDim, uOrder, vOrder, uDegree, vDegree, a, uKnotPtr, vKnotPtr, pointPtr, weightPtr, topPointPtr, topWeightPtr, r_pointPtr, r_weightPtr);
    bool flip = false;
    t.x += .0001;
    if(t.x > 1){
        t.x -= .0002;
        flip = true;
    }
    t.y += .0001;
    if(t.y > 1){
        t.y -= .0002;
        flip = !flip;
    }
    generateWithNormals2(t.x, t.y, uMax, vMax, uMin, vMin, uDim, vDim, uOrder, vOrder, uDegree, vDegree, b, uKnotPtr, vKnotPtr, pointPtr, weightPtr, topPointPtr, topWeightPtr, r_pointPtr, r_weightPtr);
    t.x = remTX;
    generateWithNormals2(t.x, t.y, uMax, vMax, uMin, vMin, uDim, vDim, uOrder, vOrder, uDegree, vDegree, c, uKnotPtr, vKnotPtr, pointPtr, weightPtr, topPointPtr, topWeightPtr, r_pointPtr, r_weightPtr);

    if(!flip){
        normal = -calcNormal(a, b, c);
    }
    else{
        normal = calcNormal(a, b, c);
    }
    return normal;
}

void NurbsPatch::hairOnSurface(int& coordIdx, M_Vec2f& coord, NurbsPatch* hairPatch){

    int uDim = dim.x;
    int vDim = dim.y;
    int uOrder = order.x;
    int vOrder = order.y;

    const float* uKnotPtr = uKnot.refValue().getArray();
    const float* vKnotPtr = vKnot.refValue().getArray();
    const Vec3f* pointPtr = point.refValue().getArray();
    const float* weightPtr = weight.refValue().getArray();

    const Vec3f* topPointPtr = pointPtr+uDim*vDim-uDim;
    const float* topWeightPtr = weightPtr+uDim*vDim-uDim;

    const Vec3f* r_pointPtr = r_point.getArray();
    const float* r_weightPtr = r_weight.getArray();

    const float uMax = maxU();
    const float vMax = maxV();
    const float uMin = minU();
    const float vMin = minV();

    int r = 0;
    int hairUDim = hairPatch->dim.x;
    int hairVDim = hairPatch->dim.y;
    M_Vec3f& hairDpoint = hairPatch->dpoint.refValue();

    for(r = 0; r < hairVDim && coordIdx < coord.length; r++){

        Vec2f t = coord[coordIdx++];
        float remTX = t.x;
        if(t.x > 1){
            t.x -= 1;
        }
        if(t.x < 0 || t.y < 0 || t.y > 1){
            continue;
        }
        Vec3f a;
        Vec3f b;
        Vec3f c;
        generateWithNormals2(t.x, t.y, uMax, vMax, uMin, vMin, uDim, vDim, uOrder, vOrder, uDegree, vDegree, a, uKnotPtr, vKnotPtr, pointPtr, weightPtr, topPointPtr, topWeightPtr, r_pointPtr, r_weightPtr);
        bool flip = false;
        t.x += .0001;
        if(t.x > 1){
            t.x -= .0002;
            flip = true;
        }
        generateWithNormals2(t.x, t.y, uMax, vMax, uMin, vMin, uDim, vDim, uOrder, vOrder, uDegree, vDegree, b, uKnotPtr, vKnotPtr, pointPtr, weightPtr, topPointPtr, topWeightPtr, r_pointPtr, r_weightPtr);
        t.x = remTX;
        t.y += .0001;
        if(t.y > 1){
            t.y -= .0002;
            flip = !flip;
        }
        generateWithNormals2(t.x, t.y, uMax, vMax, uMin, vMin, uDim, vDim, uOrder, vOrder, uDegree, vDegree, c, uKnotPtr, vKnotPtr, pointPtr, weightPtr, topPointPtr, topWeightPtr, r_pointPtr, r_weightPtr);

        b = b - a;
        c = c - a;

        b = b.normalize();

        Rotation rotB = Rotation(b, Vec3f(0, -1, 0));
        c = rotB * c;
        c = c.normalize();
        Rotation rotC = flip ? Rotation(c, Vec3f(-1, 0, 0)) : Rotation(c, Vec3f(1, 0, 0));

        //Rotation rot = -rotB * -rotC;
        //Rotation rot = -(rotC * rotB);
        Rotation rot = -rotC * -rotB;

        if(r == hairVDim - 1){
            for(int u = 0; u < hairUDim; u++){
                hairDpoint[r * hairUDim + u] = a;
            }
        }
        else{
            float hairRad = .001;
            for(int u = 0; u < hairUDim; u++){
                float angle = PI * 2.0 / (hairUDim - 1) * u;
                float s = hairRad * sin(angle);
                float c = hairRad * cos(angle);
                Vec3f v = rot * Vec3f(c, 0, s) + a;
                hairDpoint[r * hairUDim + u] = v;
            }
        }
    }
    hairPatch->dpointsToPoints();
    hairPatch->processFields(); //do we need this?
}

void NurbsPatch::generate(M_Vec3f& coord, const int h){
    FuseSurface* fuseSurface = dynamic_cast<FuseSurface*>(this->fuseSurface.p);
    if(fuseSurface){
        if(fuseSurface->evolve){
            M_Vec3f& parentFlatCoord = fuseSurface->parentFlatCoord;
            parentFlatCoord.length = fuseSurface->qty;
            for(int i = fuseSurface->qty; --i>=0;){
                parentFlatCoord[i] = coord[i];
            }
        }
    }

    int uDim=dim.x;
    int vDim=dim.y;
    int uOrder=order.x;
    int vOrder=order.y;

    Vec3f* coordPtr = coord.getArray();
    float* uKnotPtr = uKnot.refValue().getArray();
    float* vKnotPtr = vKnot.refValue().getArray();
    Vec3f* pointPtr = point.refValue().getArray();
    float* weightPtr = weight.refValue().getArray();

    Vec3f* topPointPtr = pointPtr+uDim*vDim-uDim;
    float* topWeightPtr = weightPtr+uDim*vDim-uDim;

    static M_Vec3f r_point;
    r_point.length= vDim;
    static M_Float r_weight;
    r_weight.length=vDim;
    Vec3f* r_pointPtr = r_point.getArray();
    float* r_weightPtr = r_weight.getArray();
    for(int i=0; i<vDim; i++){
        r_pointPtr[i] = pointPtr[i*uDim+uDim-1];
        r_weightPtr[i] = weightPtr[i*uDim+uDim-1];
    }


    if(FuseSurface* fuseSurface=dynamic_cast<FuseSurface*>(this->fuseSurface.p)){
        fuseSurface->storeV(coord);
    }
    for(int i=coord.length; --i>=h;){

        Vec3f t = coord[i];
        if(t.x>1){
            t.x-=1;
        }
        if(t.x<0 || t.y<0 || t.y>1){
            continue;
        }
        float u = (maxU()-minU())*t.x + minU();
        float v = (maxV()-minV())*t.y + minV();


        if(t.x==1){
            CurvePoint(vDim-1, vDegree, vKnotPtr, r_pointPtr, r_weightPtr, v, *(coordPtr+i));
            continue;
        }
        else if(t.y==1){
            CurvePoint(uDim-1, uDegree, uKnotPtr, topPointPtr, topWeightPtr, u, *(coordPtr+i));
            continue;
        }
        SurfacePoint(u, v, uOrder, vOrder, uDim, vDim, uKnotPtr, vKnotPtr, pointPtr, weightPtr, coordPtr+i);
    }
    if(scale != Vec3f(1,1,1) || rotation != Rotation() || translation != Vec3f()){
        for(int i=coord.length; --i>=h;){
            coord[i]*=matrix;
        }
    }
    if(fuseSurface){
        fuseSurface->doFuse(coord, *this);
    }
}
void NurbsPatch::DPoint::operator=(const M_Vec3f& m_par){
    int index = home.index[0];
    home.point.refValue()[index] = m_par[index] * home.weight[index];
	if(home.smls && home.cpIndex.length){
		home.smls->cp.refValue()[home.cpIndex[index]] = home.point[index];
	}
    home.processFields();
}
void NurbsPatch::Weight::operator=(const M_Float& m_par){
    refValue() = m_par;
    int index = home.index[0];
    home.point.refValue()[index] = home.dpoint[index] * home.weight[index];
    if(Seamless* smls = home.smls){
        smls->shareWeightInStitchPoly(&home, index);
		if(home.cpIndex.length){
			smls->cw.refValue()[home.cpIndex[index]] = home.weight[index];
		}
    }
}

void NurbsPatch::selectControlPointsNoBoxTest(M_<NurbsPatch*>& contigList){
	object = true;

	int uDim = this->dim.x;
	int vDim = this->dim.y;

	if(cage == false){
		return;
	}

	if(showAll){
		M_Vec3f& dpoint = this->dpoint.refValue();
		int len = dpoint.length;
		for(int i = 0; i < len; i++){
			contigList[fromPoint + i] = this;
		}
	}
	else{
		if(uJoinEnds){
			if(isCtrlPolyRef){
				for(int u = 1; u < uDim - 1; u += 2){
					for(int v = 1; v < vDim - 1; v++){
						int i = v * uDim + u;
						contigList[fromPoint + i] = this;
					}
				}
			}
			else{
				if(vDim == 1){
					for(int u = 1; u < uDim - 1; u++){
						contigList[fromPoint + u] = this;
					}
				}
				else{
					for(int u = 1; u < uDim; u += 2){
						for(int v = 0; v < vDim; v++){
							int i = v * uDim + u;
							contigList[fromPoint + i] = this;
						}
					}
				}
			}
		}
		else{
			if(NurbsPatch::MID_1 == checkNurbsType(vDim, vKnot)){
				if(isCtrlPolyRef){
					for(int v = 1; v < vDim - 1; v += 2){
						for(int u = 1; u < uDim - 1; u += 2){
							int i = v * uDim + u;
							contigList[fromPoint + i] = this;
						}
					}
				}
				else{
					if(vDim == 1){
						for(int u = 0; u < uDim; u++){
							contigList[fromPoint + u] = this;
						}
					}
					else{
						for(int v = 0; v < vDim; v += 2){
							for(int u = 0; u < uDim; u += 2){
								int i = v * uDim + u;
								contigList[fromPoint + i] = this;
								if(u == 0 || u == uDim - 2){
									u--;
								}
							}
							if(v == 0 || v == vDim - 2){
								v--;
							}
						}
					}
				}
			}
			else{
				if(NurbsPatch::MID_1 == checkNurbsType(uDim, uKnot)){
					if(isCtrlPolyRef){
						for(int v = 1; v < vDim - 1; v++){
							for(int u = 1; u < uDim - 1; u += 2){
								int i = v * uDim+u;
								contigList[fromPoint + i] = this;
							}
						}
					}
					else{
						for(int v = 0; v < vDim; v++){
							for(int u = 1; u < uDim; u +=2){
								int i = v * uDim + u;
								contigList[fromPoint + i] = this;
							}
						}
					}
				}
				else{
					if(isCtrlPolyRef){
						for(int v = 1; v < vDim - 1; v++){
							for(int u = 1; u < uDim - 1; u++){
								int i = v * uDim + u;
								contigList[fromPoint + i] = this;
							}
						}
					}
					else {
						for(int v = 0; v < vDim; v++){
							for(int u = 0; u < uDim; u++){
								int i = v * uDim + u;
								contigList[fromPoint + i] = this;
							}
						}
					}
				}
			}
		}
	}
}

void NurbsPatch::selectAllPoints(M_<NurbsPatch*>& contigList){
	int uDim = this->dim.x;
	int vDim = this->dim.y;

	if(cage == false){
		return;
	}

	if(showAll){
		M_Vec3f& dpoint = this->dpoint.refValue();
		int len = dpoint.length;
		for(int i = 0; i < len; i++){
			contigList[fromPoint + i] = this;
		}
	}
	else{
		if(uJoinEnds){
			if(isCtrlPolyRef){
				for(int u = 1; u < uDim - 1; u += 2){
					for(int v = 1; v < vDim - 1; v++){
						int i = v * uDim + u;
						contigList[fromPoint + i] = this;
					}
				}
			}
			else{
				if(vDim == 1){
					for(int u = 1; u < uDim - 1; u++){
						contigList[fromPoint + u] = this;
					}
				}
				else{
					for(int u = 1; u < uDim; u += 2){
						for(int v = 0; v < vDim; v++){
							int i = v * uDim + u;
							contigList[fromPoint + i] = this;
						}
					}
				}
			}
		}
		else{
			if(NurbsPatch::MID_1 == checkNurbsType(vDim, vKnot)){
				if(vDim == 1){
					for(int u = 0; u < uDim; u++){
						contigList[fromPoint + u] = this;
					}
				}
				else{
					for(int v = 0; v < vDim; v += 2){
						for(int u = 0; u < uDim; u += 2){
							int i = v * uDim + u;
							contigList[fromPoint + i] = this;
							if(u == 0 || u == uDim - 2){
								u--;
							}
						}
						if(v == 0 || v == vDim - 2){
							v--;
						}
					}
				}
			}
			else{
				if(NurbsPatch::MID_1 == checkNurbsType(uDim, uKnot)){
					if(isCtrlPolyRef){
						for(int v = 1; v < vDim - 1; v++){
							for(int u = 1; u < uDim - 1; u += 2){
								int i = v * uDim + u;
								contigList[fromPoint + i] = this;
							}
						}
					}
					else{
						if(uDim == 1){
							for(int v = 0; v < vDim; v++){
								contigList[fromPoint + v] = this;
							}
						}
						else{
							for(int v = 0; v < vDim; v++){
								for(int u = 1; u < uDim; u +=2){
									int i = v * uDim + u;
									contigList[fromPoint + i] = this;
								}
							}
						}
					}
				}
				else{
					if(isCtrlPolyRef){
						for(int v = 1; v < vDim - 1; v++){
							for(int u = 1; u < uDim - 1; u++){
								int i = v * uDim + u;
								contigList[fromPoint + i] = this;
							}
						}
					}
					else {
						for(int v = 0; v < vDim; v++){
							for(int u = 0; u < uDim; u++){
								int i = v * uDim + u;
								contigList[fromPoint + i] = this;
							}
						}
					}
				}
			}
		}
	}
}

void NurbsPatch::selectControlPoints(M_<NurbsPatch*>& contigList, const bool perspective){
	if(cage == false){
		return;
	}

	const int wireResolution = 4;

	tugRingPivotPoint = false;

	//int j = index[0];
	const int uDim = dim.x;
	const int vDim = dim.y;

	int uCells = uDim / 2;
	int vCells = vDim / 2;

	M_Vec3f& point = dynamic_cast<Coordinate*>(g_nurbsIlsCoord)->point;
	int lookUpIndex = g_pointLookup[plidx];

	for(int vc = vCells; --vc >= 2;){
		for(int uc = uCells; --uc >= 2;){
			int i = vc * (uCells + 2) + uc;
			Vec3f pnt = point[lookUpIndex + i];
			if(MultiSelect::isInBox(pnt)){
				int idx = convertVisualToActual(i, uDim, vDim);
				contigList[fromPoint + idx] = this;
			}
		}
	}
}

void NurbsPatch::selectControlPointsStraightCage(M_<NurbsPatch*>& contigList, const bool perspective){
	int uDim = this->dim.x;
	int vDim = this->dim.y;

	if(cage == false){
		return;
	}

	if(showAll){
		M_Vec3f& dpoint = this->dpoint.refValue();
		int len = dpoint.length;
		for(int i = 0; i < len; i++){
			if(MultiSelect::isInBox(dpoint[i])){
				contigList[fromPoint + i] = this;
			}
		}
	}
	else{
		if(uJoinEnds){
			if(isCtrlPolyRef){
				for(int u = 1; u < uDim - 1; u += 2){
					for(int v = 1; v < vDim - 1; v++){
						int i = v * uDim + u;
						if(MultiSelect::isInBox(dpoint[i])){
							contigList[fromPoint + i] = this;
						}
					}
				}
			}
			else{
				if(vDim == 1){
					for(int u = 1; u < uDim - 1; u++){
						if(MultiSelect::isInBox(dpoint[u])){
							contigList[fromPoint + u] = this;
						}
					}
				}
				else{
					for(int u = 1; u < uDim; u += 2){
						for(int v = 0; v < vDim; v++){
							int i = v * uDim + u;
							if(MultiSelect::isInBox(dpoint[i])){
								contigList[fromPoint + i] = this;
							}
						}
					}
				}
			}
		}
		else{
			if(NurbsPatch::MID_1 == checkNurbsType(vDim, vKnot)){
				if(isCtrlPolyRef){
					for(int v = 1; v < vDim - 1; v += 2){
						for(int u = 1; u < uDim - 1; u += 2){
							int i = v * uDim + u;
							if(MultiSelect::isInBox(dpoint[i])){
								contigList[fromPoint + i] = this;
							}
						}
					}
				}
				else{
					if(vDim == 1){
						for(int u = 0; u < uDim; u++){
							if(MultiSelect::isInBox(dpoint[u])){
								contigList[fromPoint + u] = this;
							}
						}
					}
					else{
						for(int v = 0; v < vDim; v += 2){
							for(int u = 0; u < uDim; u += 2){
								int i = v * uDim + u;
								if(MultiSelect::isInBox(dpoint[i])){
									contigList[fromPoint + i] = this;
								}
								if(u == 0 || u == uDim - 2){
									u--;
								}
							}
							if(v == 0 || v == vDim - 2){
								v--;
							}
						}
					}
				}
			}
			else{
				if(NurbsPatch::MID_1 == checkNurbsType(uDim, uKnot)){
					if(isCtrlPolyRef){
						for(int v = 1; v < vDim - 1; v++){
							for(int u = 1; u < uDim - 1; u += 2){
								int i = v * uDim + u;
								if(MultiSelect::isInBox(dpoint[i])){
									contigList[fromPoint + i] = this;
								}
							}
						}
					}
					else{
						if(uDim == 1){
							for(int v = 0; v < vDim; v++){
								if(MultiSelect::isInBox(dpoint[v])){
									contigList[fromPoint + v] = this;
								}
							}
						}
						else{
							for(int v = 0; v < vDim; v++){
								for(int u = 1; u < uDim; u +=2){
									int i = v * uDim + u;
									if(MultiSelect::isInBox(dpoint[i])){
										contigList[fromPoint + i] = this;
									}
								}
							}
						}
					}
				}
				else{
					if(isCtrlPolyRef){
						for(int v = 1; v < vDim - 1; v++){
							for(int u = 1; u < uDim - 1; u++){
								int i = v * uDim + u;
								if(MultiSelect::isInBox(dpoint[i])){
									contigList[fromPoint + i] = this;
								}
							}
						}
					}
					else {
						for(int v = 0; v < vDim; v++){
							for(int u = 0; u < uDim; u++){
								int i = v * uDim + u;
								if(MultiSelect::isInBox(dpoint[i])){
									contigList[fromPoint + i] = this;
								}
							}
						}
					}
				}
			}
		}
	}
}

void NurbsPatch::genCage(M_Color3f& color){
		
	//patch->matrix = scaleMatrix(patch->scale) * rotationMatrix(patch->rotation) * translationMatrix(patch->translation);
	if(g_scene->mode.refValue() & BOX_PATCH_MODELLING_MODE){
		addNurbsEdgeCage(g_nurbsIls);
		if(MultiSelect::boxPolySelList.length){
			M_Vec3f& point=dynamic_cast<Coordinate*>(g_nurbsIlsCoord)->point;
			smls->genSelectedBoxPolys(g_vcNurbsCtrlPolyPart, point);
		}
	}
	else{
		if(showAll){
			addCageShowMid(g_nurbsIls);
		}
		else{
			Color3f c = (g_transformCP->selectObject && object) ? Color3f(1, 0, 0) : Color3f(0, 1, 1);
			//for(int i = patch->dpoint.length; --i>=0;){
			//	color[fromPoint + i] = c;
			//}
			//if(g_scene->mode & BACKFACE_IGNORE_MODE){
			//	patch->addCage(g_nurbsIls, m_pointIsVisible);
			//}
			//else{
			addCage(g_nurbsIls);
		}
	}
}

void NurbsPatch::uAddControlPolyLine(int uc, int vc, IndexedLineSet* polyLine)const{
	const int wireResolution = 4;

	M_Vec3f& coord = dynamic_cast<Coordinate*>(polyLine->coord.p)->point;
	M_Int& coordIndex = polyLine->coordIndex.refValue();


	Vec3f pnt;
	int uOrder = order.x;
	int vOrder = order.y;

	int uDegree = this->uDegree;
	int vDegree = this->vDegree;

	int uDim = dim.x;
	int vDim = dim.y;

	float maxU = this->maxU();
	float maxV = this->maxV();
	float minU = this->minU();
	float minV = this->minV();

    const float* uKnotPtr = uKnot.refValue().getArray();
    const float* vKnotPtr = vKnot.refValue().getArray();
    const Vec3f* pointPtr = point.refValue().getArray();
    const float* weightPtr = weight.refValue().getArray();

	const Vec3f* r_pointPtr = r_point.getArray();
    const float* r_weightPtr = r_weight.getArray();
	const Vec3f* topPointPtr = pointPtr + uDim * vDim - uDim;
	const float* topWeightPtr = weightPtr + uDim * vDim - uDim;

	const int uCells = (uDim / 2);
	const int vCells = (vDim / 2);

	const int uSteps = uCells * wireResolution;
	const int vSteps = vCells * wireResolution;

	float u = 0;
	float v = 0;

	int cpBegin = coord.length;
	int idx = cpBegin;
	int cpidx = g_pointLookup[plidx] + vc * (uCells + 2) + uc;

	if(vc > 0){
		if(vc == vCells + 1){
			v = 1;
		}
		else{
			v = static_cast<float>(vc - 1) / vCells + 0.5 * (1.0 / vCells);
		}
	}

	int uResCounter = 1;

	int uStart = 0;
	if(uc > 0){
		uStart = uc * wireResolution - wireResolution / 2;
	}

	int uEnd = wireResolution;
	if(uc == 0 || uc == uCells){
		uEnd /= 2;
	}
	uEnd += uStart + 1;

	for(int ui = uStart; ui < uEnd; ui++){
		if(--uResCounter == 0){
			if(ui == 0 || ui == uSteps - 2){
				uResCounter = wireResolution / 2;
			}
			else{
				uResCounter = wireResolution;
			}
			coordIndex += cpidx;
			cpidx++;
		}
		else{
			u = static_cast<float>(ui) / uSteps;
			if(u == 1){
				CurvePoint(vDim-1, vDegree, vKnotPtr, r_pointPtr, r_weightPtr, v, pnt);
			}
			else if(v == 1){
				CurvePoint(uDim-1, uDegree, uKnotPtr, topPointPtr, topWeightPtr, u, pnt);
			}
			else{
				SurfacePoint(u, v, uOrder, vOrder, uDim, vDim, uKnotPtr, vKnotPtr, pointPtr, weightPtr, &pnt);
			}
			coord += pnt;
			coordIndex += idx;
			idx++;
		}
	}
	coordIndex += -1;
}
void NurbsPatch::vAddControlPolyLine(int uc, int vc, IndexedLineSet* polyLine)const{
	const int wireResolution = 4;

	M_Vec3f& coord = dynamic_cast<Coordinate*>(polyLine->coord.p)->point;
	M_Int& coordIndex = polyLine->coordIndex.refValue();


	Vec3f pnt;
	int uOrder = order.x;
	int vOrder = order.y;

	int uDegree = this->uDegree;
	int vDegree = this->vDegree;

	int uDim = dim.x;
	int vDim = dim.y;

	float maxU = this->maxU();
	float maxV = this->maxV();
	float minU = this->minU();
	float minV = this->minV();

    const float* uKnotPtr = uKnot.refValue().getArray();
    const float* vKnotPtr = vKnot.refValue().getArray();
    const Vec3f* pointPtr = point.refValue().getArray();
    const float* weightPtr = weight.refValue().getArray();

	const Vec3f* r_pointPtr = r_point.getArray();
    const float* r_weightPtr = r_weight.getArray();
	const Vec3f* topPointPtr = pointPtr + uDim * vDim - uDim;
	const float* topWeightPtr = weightPtr + uDim * vDim - uDim;

	const int uCells = (uDim / 2);
	const int vCells = (vDim / 2);

	const int uSteps = uCells * wireResolution;
	const int vSteps = vCells * wireResolution;

	float u = 0;
	float v = 0;

	int cpBegin = coord.length;
	int idx = cpBegin;
	int cpidx = g_pointLookup[plidx] + vc * (uCells + 2) + uc;

	if(uc > 0){
		if(uc == uCells + 1){
			u = 1;
		}
		else{
			u = static_cast<float>(uc - 1) / uCells + 0.5 * (1.0 / uCells);
		}
	}

	int vResCounter = 1;

	int vStart = 0;
	if(vc > 0){
		vStart = vc * wireResolution - wireResolution / 2;
	}

	int vEnd = wireResolution;
	if(vc == 0 || vc == vCells){
		vEnd /= 2;
	}
	vEnd += vStart + 1;

	for(int vi = vStart; vi < vEnd; vi++){
		if(--vResCounter == 0){
			if(vi == 0 || vi == vSteps - 2){
				vResCounter = wireResolution / 2;
			}
			else{
				vResCounter = wireResolution;
			}
			coordIndex += cpidx;
			cpidx += uCells + 2;
		}
		else{
			v = static_cast<float>(vi) / vSteps;
			if(u == 1){
				CurvePoint(vDim-1, vDegree, vKnotPtr, r_pointPtr, r_weightPtr, v, pnt);
			}
			else if(v == 1){
				CurvePoint(uDim-1, uDegree, uKnotPtr, topPointPtr, topWeightPtr, u, pnt);
			}
			else{
				SurfacePoint(u, v, uOrder, vOrder, uDim, vDim, uKnotPtr, vKnotPtr, pointPtr, weightPtr, &pnt);
			}
			coord += pnt;
			coordIndex += idx;
			idx++;
		}
	}
	coordIndex += -1;
}

void NurbsPatch::genStraightCage(M_Color3f& color, const M_Bool& m_pointIsVisible){
		
	matrix = scaleMatrix(scale) * rotationMatrix(rotation) * translationMatrix(translation);
	if(g_scene->mode.refValue() & BOX_PATCH_MODELLING_MODE){
		addNurbsEdgeCage(g_nurbsIls);
		if(MultiSelect::boxPolySelList.length){
			M_Vec3f& point=dynamic_cast<Coordinate*>(g_nurbsIlsCoord)->point;
			smls->genSelectedBoxPolys(g_vcNurbsCtrlPolyPart, point);
		}
	}
	else{
		if(showAll){
			addCageShowMid(g_nurbsIls);
		}
		else{
			Color3f c = (g_transformCP->selectObject && object) ? Color3f(1, 0, 0) : Color3f(0, 1, 1);
			for(int i = dpoint.length; --i>=0;){
				color[fromPoint + i] = c;
			}
			if(g_scene->mode & BACKFACE_IGNORE_MODE){
				if(g_showPatchBorderMode){
					addStraightCageForShowBorder(g_nurbsIls, m_pointIsVisible);
				}
				else{
					addStraightCage(g_nurbsIls, m_pointIsVisible);
				}
			}
			else{
				if(g_showPatchBorderMode){
					addStraightCageForShowBorder(g_nurbsIls);
				}
				else{
					addStraightCage(g_nurbsIls);
				}
			}
			//if(g_showPatchBorderMode){
			//	addBorder(g_nurbsIls);
			//}
			//colorColBlue(patch, color, 11);
		}
	}
}

//hide backface
void NurbsPatch::addStraightCage(IndexedLineSet* polyLine, const M_Bool& m_pointIsVisible){

	int uDim = this->dim.x;
	int vDim = this->dim.y;

	M_Vec3f& coord = dynamic_cast<Coordinate*>(polyLine->coord.p)->point;
	M_Int& coordIndex = polyLine->coordIndex.refValue();

	for(int i = dpoint.length; --i>=0;){
		coord[fromPoint + i] = matrix * dpoint[i];
	}
	if(cage == false){
		return;
	}

	if(uJoinEnds){
		if(NurbsPatch::MID_1 == checkNurbsType(vDim, vKnot)){
			if(isCtrlPolyRef){
				if(vDim > 3){
					for(int v = 1; v < vDim - 1; v += 2){
						for(int u = 1; u < uDim - 1; u += 2){
							coordIndex += fromPoint + v * uDim+u;
						}
						coordIndex += -1;
					}
					for(int u = 1; u < uDim - 1; u += 2){
						for(int v = 1; v < vDim - 1; v++){
							coordIndex += fromPoint + v * uDim + u;
						}
						coordIndex += -1;
					}
				}
			}
			else{
				if(vDim == 1){
					coordIndex += fromPoint;
					for(int u = 1; u < uDim; u++){
						coordIndex += fromPoint + u;
					}
					coordIndex += fromPoint + uDim - 1;
					coordIndex+=-1;
				}
				else{
					for(int v = 0; v < vDim; v += 2){
						coordIndex += fromPoint + v * uDim;
						for(int u = 1; u < uDim; u += 2){
							coordIndex += fromPoint + v * uDim + u;
						}
						coordIndex += fromPoint + v * uDim + uDim - 1;
						coordIndex+=-1;

						if(v == 0 || v == vDim - 2){
							v--;
						}
					}
					for(int u=1; u<uDim; u += 2){
						for(int v=0; v<vDim; v++){
							coordIndex += fromPoint + v * uDim + u;
						}
						coordIndex += -1;
					}
				}
			}
		}
		else{
			if(isCtrlPolyRef){
				if(vDim > 3){
					for(int v = 1; v < vDim - 1; v++){
						for(int u = 1; u < uDim - 1; u += 2){
							coordIndex += fromPoint + v * uDim+u;
						}
						coordIndex += -1;
					}
					for(int u = 1; u < uDim - 1; u += 2){
						for(int v = 1; v < vDim - 1; v++){
							coordIndex += fromPoint + v * uDim + u;
						}
						coordIndex += -1;
					}
				}
			}
			else{
				for(int v=0; v < vDim; v++){
					coordIndex += fromPoint + v * uDim + uDim - 2;
					for(int u=1; u<uDim; u+=2){
						coordIndex += fromPoint + v * uDim + u;
					}
					coordIndex += -1;
				}
				for(int u=1; u<uDim; u+=2){
					for(int v=0; v<vDim; v++){
						coordIndex += fromPoint + v * uDim + u;
					}
					coordIndex += -1;
				}
			}
		}
	}
	else {
		if(NurbsPatch::MID_1 == checkNurbsType(vDim, vKnot)){
			if(isCtrlPolyRef){
				if(uDim > 3 && vDim > 3){
					for(int v = 1; v < vDim - 1; v += 2){
						for(int u = 1; u < uDim - 2; u += 2){
							int i = fromPoint + v * uDim + u;
							if(m_pointIsVisible[i]){
								int i2 = i + 2;
								if(u == uDim - 3 || m_pointIsVisible[i2] == false){
									continue;
								}
								coordIndex += i;
								coordIndex += i2;
								u += 4;
								for(; u < uDim - 1; u += 2){
									i = fromPoint + v * uDim + u;
									if(m_pointIsVisible[i] == false){
										break;
									}
									coordIndex += i;
								}
								coordIndex += -1;
							}
						}
					}
					for(int u = 1; u < uDim - 1; u += 2){
						for(int v = 1; v < vDim - 2; v += 2){
							int i = fromPoint + v * uDim + u;
							if(m_pointIsVisible[i]){
								int i2 = i + uDim * 2;
								if(v == vDim - 3 || m_pointIsVisible[i2] == false){
									continue;
								}
								coordIndex += i;
								coordIndex += i2;
								v += 4;
								for(; v < vDim - 1; v += 2){
									i = fromPoint + v * uDim + u;
									if(m_pointIsVisible[i] == false){
										break;
									}
									coordIndex += i;
								}
								coordIndex += -1;
							}
						}
					}
				}
			}
			else{
				if(vDim == 1){
					for(int u = 0; u < uDim; u++){
						coordIndex += fromPoint + u;
					}
					coordIndex += -1;
				}
				else{
					for(int v = 0; v < vDim; v += 2){
						coordIndex += fromPoint + v * uDim;
						for(int u = 1; u<uDim; u += 2){
							coordIndex += fromPoint + v * uDim + u;
						}
						coordIndex += fromPoint + v * uDim + uDim - 1;
						coordIndex+=-1;

						if(v == 0 || v == vDim - 2){
							v--;
						}
					}
					for(int u=0; u<uDim; u += 2){
						coordIndex += fromPoint + u;
						for(int v=1; v<vDim; v+=2){
							coordIndex += fromPoint + v * uDim + u;
						}
						coordIndex += fromPoint + uDim * vDim - uDim + u;
						coordIndex += -1;
						if(u == 0 || u == uDim - 2){
							u--;
						}
					}
				}
			}
		}
		else{
			if(NurbsPatch::MID_1 == checkNurbsType(uDim, uKnot)){
				if(isCtrlPolyRef){
					if(vDim > 3){
						for(int v = 1; v < vDim - 1; v++){
							for(int u = 1; u < uDim - 1; u += 2){
								coordIndex += fromPoint + v * uDim+u;
							}
							coordIndex += -1;
						}
						for(int u = 1; u < uDim - 1; u += 2){
							for(int v = 1; v < vDim - 1; v++){
								coordIndex += fromPoint + v * uDim + u;
							}
							coordIndex += -1;
						}
					}
				}
				else{
					for(int v = 0; v < vDim; v++){
						coordIndex += fromPoint + v * uDim;
						for(int u = 1; u < uDim; u +=2){
							coordIndex += fromPoint + v * uDim + u;
						}
						coordIndex += fromPoint + v * uDim + uDim - 1;
						coordIndex += -1;
					}
					for(int u = 0; u < uDim; u += 2){
						coordIndex += fromPoint + u;
						for(int v = 0; v < vDim; v++){
							coordIndex += fromPoint + v * uDim + u;
						}
						coordIndex += fromPoint + uDim * vDim - uDim + u;
						coordIndex += -1;
						if(u == 0 || u == uDim - 2){
							u--;
						}
					}
				}
			}
			else{
				if(isCtrlPolyRef){
					if(vDim > 3 && uDim > 3){
						for(int v = 1; v < vDim - 1; v++){
							for(int u = 1; u < uDim - 1; u++){
								coordIndex += fromPoint + v * uDim + u;
							}
							coordIndex += -1;
						}
						for(int u = 1; u < uDim - 1; u++){
							for(int v = 1; v < vDim - 1; v++){
								coordIndex += fromPoint + v * uDim + u;
							}
							coordIndex += -1;
						}
					}
				}
				else {
					for(int v = 0; v < vDim; v++){
						for(int u = 0; u < uDim; u++){
							coordIndex += fromPoint + v * uDim + u;
						}
						coordIndex += -1;
					}
					for(int u = 0; u < uDim; u++){
						for(int v = 0; v < vDim; v++){
							coordIndex += fromPoint + v * uDim + u;
						}
						coordIndex += -1;
					}
				}
			}
		}
	}
}
//hide backface
void NurbsPatch::addStraightCageForShowBorder(IndexedLineSet* polyLine, const M_Bool& m_pointIsVisible){

	int uDim = this->dim.x;
	int vDim = this->dim.y;

	M_Vec3f& coord = dynamic_cast<Coordinate*>(polyLine->coord.p)->point;
	M_Color3f& color = dynamic_cast<Color*>(polyLine->color.p)->color;
	M_Int& coordIndex = polyLine->coordIndex.refValue();

	for(int i = dpoint.length; --i>=0;){
		coord[fromPoint + i] = matrix * dpoint[i];
	}
	if(cage == false){
		return;
	}

	if(NurbsPatch::MID_1 == checkNurbsType(vDim, vKnot)){
		if(vDim == 1){
			for(int u = 0; u < uDim; u++){
				coordIndex += fromPoint + u;
			}
			coordIndex += -1;
		}
		else{
			for(int v = 1; v < vDim - 1; v += 2){
				for(int u = 0; u < uDim - 1; u += 2){
					int i = fromPoint + v * uDim + u;
					if(m_pointIsVisible[i]){
						int i2 = i + 2;
						if(u == 0 || u == uDim - 2){
							i2--;
						}

						if(m_pointIsVisible[i2] == false){
							continue;
						}
						coordIndex += i;
						coordIndex += i2;
						if(u == uDim - 3){
							u += 2;
						}
						else if(u == 0 || u == uDim - 4){
							u += 3;
						}
						else{
							u += 4;
						}
						for(; u < uDim; u += 2){
							i = fromPoint + v * uDim + u;
							if(m_pointIsVisible[i] == false){
								break;
							}
							coordIndex += i;
							if(u == uDim - 2){
								u--;
							}
						}
						coordIndex += -1;
					}
					if(u == 0 || u == uDim - 2){
						u--;
					}
				}
			}
			for(int u = 1; u < uDim - 1; u += 2){
				for(int v = 0; v < vDim - 1; v += 2){
					int i = fromPoint + v * uDim + u;
					if(m_pointIsVisible[i]){
						int i2 = i + uDim * 2;
						if(v == 0 || v == vDim - 2){
							i2 -= uDim;
						}
						if(m_pointIsVisible[i2] == false){
							continue;
						}
						coordIndex += i;
						coordIndex += i2;
						if(v == vDim - 3){
							v += 2;
						}
						else if(v == 0 || v == vDim - 4){
							v += 3;
						}
						else{
							v += 4;
						}
						for(; v < vDim; v += 2){
							i = fromPoint + v * uDim + u;
							if(m_pointIsVisible[i] == false){
								break;
							}
							coordIndex += i;
							if(v == vDim - 2){
								v--;
							}
						}
						coordIndex += -1;
					}
					if(v == 0 || v == vDim - 2){
						v--;
					}
				}
			}
		}
	}
	else{
		if(NurbsPatch::MID_1 == checkNurbsType(uDim, uKnot)){
			if(isCtrlPolyRef){
				if(vDim > 3){
					for(int v = 1; v < vDim - 1; v++){
						for(int u = 1; u < uDim - 1; u += 2){
							coordIndex += fromPoint + v * uDim+u;
						}
						coordIndex += -1;
					}
					for(int u = 1; u < uDim - 1; u += 2){
						for(int v = 1; v < vDim - 1; v++){
							coordIndex += fromPoint + v * uDim + u;
						}
						coordIndex += -1;
					}
				}
			}
			else{
				for(int v = 0; v < vDim; v++){
					coordIndex += fromPoint + v * uDim;
					for(int u = 1; u < uDim; u +=2){
						coordIndex += fromPoint + v * uDim + u;
					}
					coordIndex += fromPoint + v * uDim + uDim - 1;
					coordIndex += -1;
				}
				for(int u = 0; u < uDim; u += 2){
					coordIndex += fromPoint + u;
					for(int v = 0; v < vDim; v++){
						coordIndex += fromPoint + v * uDim + u;
					}
					coordIndex += fromPoint + uDim * vDim - uDim + u;
					coordIndex += -1;
					if(u == 0 || u == uDim - 2){
						u--;
					}
				}
			}
		}
		else{
			if(isCtrlPolyRef){
				if(vDim > 3 && uDim > 3){
					for(int v = 1; v < vDim - 1; v++){
						for(int u = 1; u < uDim - 1; u++){
							coordIndex += fromPoint + v * uDim + u;
						}
						coordIndex += -1;
					}
					for(int u = 1; u < uDim - 1; u++){
						for(int v = 1; v < vDim - 1; v++){
							coordIndex += fromPoint + v * uDim + u;
						}
						coordIndex += -1;
					}
				}
			}
			else {
				for(int v = 0; v < vDim; v++){
					for(int u = 0; u < uDim; u++){
						coordIndex += fromPoint + v * uDim + u;
					}
					coordIndex += -1;
				}
				for(int u = 0; u < uDim; u++){
					for(int v = 0; v < vDim; v++){
						coordIndex += fromPoint + v * uDim + u;
					}
					coordIndex += -1;
				}
			}
		}
	}
//add border
	Color3f midColor = Color3f(0, 1, 0);
	Color3f visColor = Color3f(0, 1, 1);

	if(g_transformCP->selectObject && object){
		midColor = Color3f(1, 0, 0);
		visColor = Color3f(1, 0, 0);
	}

	for(int i = dpoint.length; --i>=0;){
		color[fromPoint + i] = midColor;
	}

	for(int u = 0; u < uDim - 1; u++){//south edge
		int i = fromPoint + u;
		if(m_pointIsVisible[i] && m_pointIsVisible[i + 1]){
			coordIndex += i;
			coordIndex += i + 1;
			u += 2;
			for(; u < uDim; u++){
				i = fromPoint + u;
				if(m_pointIsVisible[i] == false){
					break;
				}
				coordIndex += i;
			}
			coordIndex += -1;
		}
	}

	int offset = uDim * (vDim - 1);
	for(int u = 0; u < uDim - 1; u++){ //north edge
		int i = fromPoint + u + offset;
		if(m_pointIsVisible[i] && m_pointIsVisible[i + 1]){
			coordIndex += i;
			coordIndex += i + 1;
			u += 2;
			for(; u < uDim; u++){
				i = fromPoint + u + offset;
				if(m_pointIsVisible[i] == false){
					break;
				}
				coordIndex += i;
			}
			coordIndex += -1;
		}
	}

	for(int v = 0; v < vDim - 1; v++){//west edge
		int i = fromPoint + v * uDim;
		if(m_pointIsVisible[i] && m_pointIsVisible[i + uDim]){
			coordIndex += i;
			coordIndex += i + uDim;
			v += 2;
			for(; v < vDim; v++){
				i = fromPoint + v * uDim;
				if(m_pointIsVisible[i] == false){
					break;
				}
				coordIndex += i;
			}
			coordIndex += -1;
		}
	}

	for(int v = 0; v < vDim - 1; v++){//east edge
		int i = fromPoint + v * uDim + uDim - 1;
		if(m_pointIsVisible[i] && m_pointIsVisible[i + uDim]){
			coordIndex += i;
			coordIndex += i + uDim;
			v += 2;
			for(; v < vDim; v++){
				i = fromPoint + v * uDim + uDim - 1;
				if(m_pointIsVisible[i] == false){
					break;
				}
				coordIndex += i;
			}
			coordIndex += -1;
		}
	}

	if(NurbsPatch::MID_1 == checkNurbsType(vDim, vKnot)){
		if(NurbsPatch::MID_1 == checkNurbsType(uDim, uKnot)){
			for(int v = 1; v < vDim - 1; v += 2){
				for(int u = 1; u < uDim; u += 2){
					color[fromPoint + v * uDim + u] = visColor;
				}
				if(v == 0 || v == vDim - 2){
					v--;
				}
			}
		}
		else{
			for(int v = 1; v < vDim - 1; v += 2){
				for(int u = 1; u < uDim - 1; u += 1){
					color[fromPoint + v * uDim + u] = visColor;
				}
				if(v == 0 || v == vDim - 2){
					v--;
				}
			}
		}
	}
	else{
		if(NurbsPatch::MID_1 == checkNurbsType(uDim, uKnot)){
			for(int v = 1; v < vDim - 1; v++){
				for(int u = 1; u < uDim; u += 2){
					color[fromPoint + v * uDim + u] = visColor;
				}
			}
		}
		else{
			for(int v = 1; v < vDim - 1; v++){
				for(int u = 1; u < uDim -  1; u++){
					color[fromPoint + v * uDim + u] = visColor;
				}
			}
		}
	}
}

// show backface
void NurbsPatch::addStraightCage(IndexedLineSet* polyLine){

	int uDim = this->dim.x;
	int vDim = this->dim.y;

	M_Vec3f& coord = dynamic_cast<Coordinate*>(polyLine->coord.p)->point;
	M_Int& coordIndex = polyLine->coordIndex.refValue();

	for(int i = dpoint.length; --i>=0;){
		coord[fromPoint + i] = matrix * dpoint[i];
	}
	if(cage == false){
		return;
	}

	if(uJoinEnds){
		if(NurbsPatch::MID_1 == checkNurbsType(vDim, vKnot)){
			if(isCtrlPolyRef){
				if(vDim > 3){
					for(int v = 1; v < vDim - 1; v += 2){
						for(int u = 1; u < uDim - 1; u += 2){
							coordIndex += fromPoint + v * uDim+u;
						}
						coordIndex += -1;
					}
					for(int u = 1; u < uDim - 1; u += 2){
						for(int v = 1; v < vDim - 1; v++){
							coordIndex += fromPoint + v * uDim + u;
						}
						coordIndex += -1;
					}
				}
			}
			else{
				if(vDim == 1){
					coordIndex += fromPoint;
					for(int u = 1; u < uDim; u++){
						coordIndex += fromPoint + u;
					}
					coordIndex += fromPoint + uDim - 1;
					coordIndex+=-1;
				}
				else{
					for(int v = 0; v < vDim; v += 2){
						coordIndex += fromPoint + v * uDim;
						for(int u = 1; u < uDim; u += 2){
							coordIndex += fromPoint + v * uDim + u;
						}
						coordIndex += fromPoint + v * uDim + uDim - 1;
						coordIndex+=-1;

						if(v == 0 || v == vDim - 2){
							v--;
						}
					}
					for(int u=1; u<uDim; u += 2){
						for(int v=0; v<vDim; v++){
							coordIndex += fromPoint + v * uDim + u;
						}
						coordIndex += -1;
					}
				}
			}
		}
		else{
			if(isCtrlPolyRef){
				if(vDim > 3){
					for(int v = 1; v < vDim - 1; v++){
						for(int u = 1; u < uDim - 1; u += 2){
							coordIndex += fromPoint + v * uDim+u;
						}
						coordIndex += -1;
					}
					for(int u = 1; u < uDim - 1; u += 2){
						for(int v = 1; v < vDim - 1; v++){
							coordIndex += fromPoint + v * uDim + u;
						}
						coordIndex += -1;
					}
				}
			}
			else{
				for(int v=0; v < vDim; v++){
					coordIndex += fromPoint + v * uDim + uDim - 2;
					for(int u=1; u<uDim; u+=2){
						coordIndex += fromPoint + v * uDim + u;
					}
					coordIndex += -1;
				}
				for(int u=1; u<uDim; u+=2){
					for(int v=0; v<vDim; v++){
						coordIndex += fromPoint + v * uDim + u;
					}
					coordIndex += -1;
				}
			}
		}
	}
	else {
		if(NurbsPatch::MID_1 == checkNurbsType(vDim, vKnot)){
			if(isCtrlPolyRef){
				if(uDim > 3 && vDim > 3){
					for(int v = 1; v < vDim - 1; v += 2){
						for(int u = 1; u < uDim - 1; u += 2){
							coordIndex += fromPoint + v * uDim+u;
						}
						coordIndex += -1;
					}
					for(int u = 1; u < uDim - 1; u += 2){
						for(int v = 1; v < vDim - 1; v += 2){
							coordIndex += fromPoint + v * uDim + u;
						}
						coordIndex += -1;
					}
				}
			}
			else{
				if(vDim == 1){
					for(int u = 0; u < uDim; u++){
						coordIndex += fromPoint + u;
					}
					coordIndex += -1;
				}
				else{
					for(int v = 0; v < vDim; v += 2){
						coordIndex += fromPoint + v * uDim;
						for(int u = 1; u<uDim; u += 2){
							coordIndex += fromPoint + v * uDim + u;
						}
						coordIndex += fromPoint + v * uDim + uDim - 1;
						coordIndex+=-1;

						if(v == 0 || v == vDim - 2){
							v--;
						}
					}
					for(int u=0; u<uDim; u += 2){
						coordIndex += fromPoint + u;
						for(int v=1; v<vDim; v+=2){
							coordIndex += fromPoint + v * uDim + u;
						}
						coordIndex += fromPoint + uDim * vDim - uDim + u;
						coordIndex += -1;
						if(u == 0 || u == uDim - 2){
							u--;
						}
					}
				}
			}
		}
		else{
			if(NurbsPatch::MID_1 == checkNurbsType(uDim, uKnot)){
				if(isCtrlPolyRef){
					if(vDim > 3){
						for(int v = 1; v < vDim - 1; v++){
							for(int u = 1; u < uDim - 1; u += 2){
								coordIndex += fromPoint + v * uDim+u;
							}
							coordIndex += -1;
						}
						for(int u = 1; u < uDim - 1; u += 2){
							for(int v = 1; v < vDim - 1; v++){
								coordIndex += fromPoint + v * uDim + u;
							}
							coordIndex += -1;
						}
					}
				}
				else{
					for(int v = 0; v < vDim; v++){
						coordIndex += fromPoint + v * uDim;
						for(int u = 1; u < uDim; u +=2){
							coordIndex += fromPoint + v * uDim + u;
						}
						coordIndex += fromPoint + v * uDim + uDim - 1;
						coordIndex += -1;
					}
					for(int u = 0; u < uDim; u += 2){
						coordIndex += fromPoint + u;
						for(int v = 0; v < vDim; v++){
							coordIndex += fromPoint + v * uDim + u;
						}
						coordIndex += fromPoint + uDim * vDim - uDim + u;
						coordIndex += -1;
						if(u == 0 || u == uDim - 2){
							u--;
						}
					}
				}
			}
			else{
				if(isCtrlPolyRef){
					if(vDim > 3 && uDim > 3){
						for(int v = 1; v < vDim - 1; v++){
							for(int u = 1; u < uDim - 1; u++){
								coordIndex += fromPoint + v * uDim + u;
							}
							coordIndex += -1;
						}
						for(int u = 1; u < uDim - 1; u++){
							for(int v = 1; v < vDim - 1; v++){
								coordIndex += fromPoint + v * uDim + u;
							}
							coordIndex += -1;
						}
					}
				}
				else {
					for(int v = 0; v < vDim; v++){
						for(int u = 0; u < uDim; u++){
							coordIndex += fromPoint + v * uDim + u;
						}
						coordIndex += -1;
					}
					for(int u = 0; u < uDim; u++){
						for(int v = 0; v < vDim; v++){
							coordIndex += fromPoint + v * uDim + u;
						}
						coordIndex += -1;
					}
				}
			}
		}
	}
}

// show backface
void NurbsPatch::addStraightCageForShowBorder(IndexedLineSet* polyLine){

	int uDim = this->dim.x;
	int vDim = this->dim.y;

	M_Vec3f& coord = dynamic_cast<Coordinate*>(polyLine->coord.p)->point;
	M_Color3f& color = dynamic_cast<Color*>(polyLine->color.p)->color;
	M_Int& coordIndex = polyLine->coordIndex.refValue();

	for(int i = dpoint.length; --i>=0;){
		coord[fromPoint + i] = matrix * dpoint[i];
	}
	if(cage == false){
		return;
	}

	if(uJoinEnds){
		if(NurbsPatch::MID_1 == checkNurbsType(vDim, vKnot)){
			if(isCtrlPolyRef){
				if(vDim > 3){
					for(int v = 1; v < vDim - 1; v += 2){
						for(int u = 1; u < uDim - 1; u += 2){
							coordIndex += fromPoint + v * uDim+u;
						}
						coordIndex += -1;
					}
					for(int u = 1; u < uDim - 1; u += 2){
						for(int v = 1; v < vDim - 1; v++){
							coordIndex += fromPoint + v * uDim + u;
						}
						coordIndex += -1;
					}
				}
			}
			else{
				if(vDim == 1){
					coordIndex += fromPoint;
					for(int u = 1; u < uDim; u++){
						coordIndex += fromPoint + u;
					}
					coordIndex += fromPoint + uDim - 1;
					coordIndex+=-1;
				}
				else{
					for(int v = 0; v < vDim; v += 2){
						coordIndex += fromPoint + v * uDim;
						for(int u = 1; u < uDim; u += 2){
							coordIndex += fromPoint + v * uDim + u;
						}
						coordIndex += fromPoint + v * uDim + uDim - 1;
						coordIndex+=-1;

						if(v == 0 || v == vDim - 2){
							v--;
						}
					}
					for(int u=1; u<uDim; u += 2){
						for(int v=0; v<vDim; v++){
							coordIndex += fromPoint + v * uDim + u;
						}
						coordIndex += -1;
					}
				}
			}
		}
		else{
			if(isCtrlPolyRef){
				if(vDim > 3){
					for(int v = 1; v < vDim - 1; v++){
						for(int u = 0; u < uDim; u += 2){
							coordIndex += fromPoint + v * uDim+u;
							if(u == 0 || u == uDim - 2){
								u--;
							}

						}
						coordIndex += -1;
					}
					for(int u = 1; u < uDim - 1; u += 2){
						for(int v = 1; v < vDim - 1; v++){
							coordIndex += fromPoint + v * uDim + u;
						}
						coordIndex += -1;
					}
				}
			}
			else{
				for(int v = 0; v < vDim; v++){
					coordIndex += fromPoint + v * uDim + uDim - 2;
					for(int u = 1; u < uDim; u+=2){
						coordIndex += fromPoint + v * uDim + u;
					}
					coordIndex += -1;
				}
				for(int u = 1; u < uDim; u+=2){
					for(int v=0; v < vDim; v++){
						coordIndex += fromPoint + v * uDim + u;
					}
					coordIndex += -1;
				}
			}
		}
	}
	else {
		if(NurbsPatch::MID_1 == checkNurbsType(vDim, vKnot)){
			if(isCtrlPolyRef){
				for(int v = 1; v < vDim - 1; v += 2){
					for(int u = 0; u < uDim; u += 2){
						coordIndex += fromPoint + v * uDim+u;
						if(u == 0 || u == uDim - 2){
							u--;
						}
					}
					coordIndex += -1;
				}
				for(int u = 1; u < uDim - 1; u += 2){
					for(int v = 0; v < vDim; v += 2){
						coordIndex += fromPoint + v * uDim + u;
						if(v == 0 || v == vDim - 2){
							v--;
						}
					}
					coordIndex += -1;
				}
			}
			else{
				if(vDim == 1){
//					for(int u = 0; u < uDim; u++){
//						coordIndex += fromPoint + u;
//					}
//					coordIndex += -1;
				}
				else{
					for(int v = 0; v < vDim; v += 2){
						coordIndex += fromPoint + v * uDim;
						for(int u = 1; u<uDim; u += 2){
							coordIndex += fromPoint + v * uDim + u;
						}
						coordIndex += fromPoint + v * uDim + uDim - 1;
						coordIndex+=-1;

						if(v == 0 || v == vDim - 2){
							v--;
						}
					}
					for(int u=0; u<uDim; u += 2){
						coordIndex += fromPoint + u;
						for(int v=1; v<vDim; v+=2){
							coordIndex += fromPoint + v * uDim + u;
						}
						coordIndex += fromPoint + uDim * vDim - uDim + u;
						coordIndex += -1;
						if(u == 0 || u == uDim - 2){
							u--;
						}
					}
				}
			}
		}
		else{
			if(NurbsPatch::MID_1 == checkNurbsType(uDim, uKnot)){
				if(isCtrlPolyRef){
					if(vDim > 3){
						for(int v = 1; v < vDim - 1; v++){
							for(int u = 1; u < uDim - 1; u += 2){
								coordIndex += fromPoint + v * uDim+u;
							}
							coordIndex += -1;
						}
						for(int u = 1; u < uDim - 1; u += 2){
							for(int v = 1; v < vDim - 1; v++){
								coordIndex += fromPoint + v * uDim + u;
							}
							coordIndex += -1;
						}
					}
				}
				else{
					for(int v = 0; v < vDim; v++){
						coordIndex += fromPoint + v * uDim;
						for(int u = 1; u < uDim; u +=2){
							coordIndex += fromPoint + v * uDim + u;
						}
						coordIndex += fromPoint + v * uDim + uDim - 1;
						coordIndex += -1;
					}
					for(int u = 0; u < uDim; u += 2){
						coordIndex += fromPoint + u;
						for(int v = 0; v < vDim; v++){
							coordIndex += fromPoint + v * uDim + u;
						}
						coordIndex += fromPoint + uDim * vDim - uDim + u;
						coordIndex += -1;
						if(u == 0 || u == uDim - 2){
							u--;
						}
					}
				}
			}
			else{
				if(isCtrlPolyRef){
					if(vDim > 3 && uDim > 3){
						for(int v = 1; v < vDim - 1; v++){
							for(int u = 1; u < uDim - 1; u++){
								coordIndex += fromPoint + v * uDim + u;
							}
							coordIndex += -1;
						}
						for(int u = 1; u < uDim - 1; u++){
							for(int v = 1; v < vDim - 1; v++){
								coordIndex += fromPoint + v * uDim + u;
							}
							coordIndex += -1;
						}
					}
				}
				else {
					for(int v = 0; v < vDim; v++){
						for(int u = 0; u < uDim; u++){
							coordIndex += fromPoint + v * uDim + u;
						}
						coordIndex += -1;
					}
					for(int u = 0; u < uDim; u++){
						for(int v = 0; v < vDim; v++){
							coordIndex += fromPoint + v * uDim + u;
						}
						coordIndex += -1;
					}
				}
			}
		}
	}
//add border
	Color3f midColor = Color3f(0, 1, 0);
	Color3f visColor = Color3f(0, 1, 1);

	if(g_transformCP->selectObject && object){
		midColor = Color3f(1, 0, 0);
		visColor = Color3f(1, 0, 0);
	}

	for(int i = dpoint.length; --i>=0;){
		coord[fromPoint + i] = matrix * dpoint[i];
		color[fromPoint + i] = midColor;
	}
	if(cage == false){
		return;
	}

	for(int u = 0; u < uDim; u++){
		coordIndex += fromPoint + u;
	}
	coordIndex += -1;

	if(vDim > 1){
		int offset = uDim * (vDim - 1);
		for(int u = 0; u < uDim; u++){
			coordIndex += fromPoint + u + offset;
		}
		coordIndex += -1;

		for(int v = 0; v < vDim; v++){
			coordIndex += fromPoint + v * uDim;
		}
		coordIndex += -1;

		for(int v = 0; v < vDim; v++){
			coordIndex += fromPoint + v * uDim + uDim - 1;
		}
		coordIndex += -1;
	}

	if(NurbsPatch::MID_1 == checkNurbsType(vDim, vKnot)){
		if(NurbsPatch::MID_1 == checkNurbsType(uDim, uKnot)){
			for(int v = 1; v < vDim - 1; v += 2){
				for(int u = 1; u < uDim; u += 2){
					color[fromPoint + v * uDim + u] = visColor;
				}
				if(v == 0 || v == vDim - 2){
					v--;
				}
			}
		}
		else{
			for(int v = 1; v < vDim - 1; v += 2){
				for(int u = 1; u < uDim - 1; u += 1){
					color[fromPoint + v * uDim + u] = visColor;
				}
				if(v == 0 || v == vDim - 2){
					v--;
				}
			}
		}
	}
	else{
		if(NurbsPatch::MID_1 == checkNurbsType(uDim, uKnot)){
			for(int v = 1; v < vDim - 1; v++){
				for(int u = 1; u < uDim; u += 2){
					color[fromPoint + v * uDim + u] = visColor;
				}
			}
		}
		else{
			for(int v = 1; v < vDim - 1; v++){
				for(int u = 1; u < uDim -  1; u++){
					color[fromPoint + v * uDim + u] = visColor;
				}
			}
		}
	}
}

void NurbsPatch::pointsToCps(){
	M_Vec3f& cp = smls->cp.refValue();

	for(int k = point.length; --k >= 0;){
		cp[cpIndex[k]] = point[k];
	}
}
void NurbsPatch::pointsToCpws(){
	M_Vec3f& cp = smls->cp.refValue();
	M_Float& cw = smls->cw.refValue();

	for(int k = point.length; --k >= 0;){
		int cpIdx = cpIndex[k];
		cp[cpIdx] = point[k];
		cw[cpIdx] = weight[k];
	}
}

int NurbsPatch::getMirrorTwinlessIdx(int idx)const{
	int uDim = dim.x;
	int u = idx % uDim;
	int v = idx / uDim;
	return v * uDim + uDim - 1 - u;
}

int NurbsPatch::getMirrorIdx(int idx)const{
	if(mirrorPatch){
		return idx;
	}
	return getMirrorTwinlessIdx(idx);
}

void NurbsPatch::joinEastToWest(){
	int uDim = dim.x;
	int vDim = dim.y;

	M_Int& cpIndex = this->cpIndex.refValue();

	for(int v = vDim; --v>=0;){
		cpIndex[v * uDim] = cpIndex[v * uDim + uDim - 1];
	}
}


void NurbsPatch::redirectCol(int col){
	int uDim = dim.x;
	int vDim = dim.y;

	M_Int& cpIndex = this->cpIndex.refValue();

	for(int v = vDim; --v>=0;){
		smls->redirectSingleCpIndex(cpIndex[v * uDim + col - 1], cpIndex[v * uDim + col + 1]);
	}
}

void NurbsPatch::redirectRow(int row){
	int uDim = dim.x;
	int vDim = dim.y;

	M_Int& cpIndex = this->cpIndex.refValue();

	int offset = row * uDim;

	for(int u = uDim; --u>=0;){
		smls->redirectSingleCpIndex(cpIndex[offset - uDim + u], cpIndex[offset + uDim + u]);
	}
}


void NurbsPatch::redirectEastToWest(){
	int uDim = dim.x;
	int vDim = dim.y;

	M_Int& cpIndex = this->cpIndex.refValue();

	for(int v = vDim; --v>=0;){
		smls->redirectSingleCpIndex(cpIndex[v * uDim], cpIndex[v * uDim + uDim - 1]);
	}
}

void NurbsPatch::redirectSouthToNorth(){
	int uDim = dim.x;
	int vDim = dim.y;

	M_Int& cpIndex = this->cpIndex.refValue();

	int offset = (vDim - 1) * uDim;

	for(int u = uDim; --u>=0;){
		smls->redirectSingleCpIndex(cpIndex[offset + u], cpIndex[u]);
	}
}

void NurbsPatch::joinExtraCenterShaftPoints(NurbsPatch* extra){
	int uDim = dim.x;
	int vDim = dim.y;

	int euDim = extra->dim.x;

	M_Int& cpIndex = this->cpIndex.refValue();
	M_Int& ecpIndex = extra->cpIndex.refValue();

	int v2 = vDim;

	for(int v = vDim; --v>=0;){
		--v2;
		cpIndex[v * uDim] = ecpIndex[v2 * euDim + euDim - 1];
		cpIndex[v * uDim + uDim - 1] = ecpIndex[v2 * euDim];
	}
}

void NurbsPatch::scrunchNorth(){
	int uDim = dim.x;
	int vDim = dim.y;
	M_Int& cpIndex = this->cpIndex.refValue();
	int offset = (vDim - 1) * uDim;

	for(int k = uDim; --k>=0;){
		cpIndex[offset + k] = cpIndex[offset];
	}
}
void NurbsPatch::scrunchSouth(){
	int uDim = dim.x;
	int vDim = dim.y;
	M_Int& cpIndex = this->cpIndex.refValue();

	for(int k = uDim; --k>=0;){
		cpIndex[k] = cpIndex[0];
	}
}

