//typedef unsigned short int RenderInt;
//#define INDEX_TYPE GL_UNSIGNED_SHORT

typedef unsigned int RenderInt;
#define RENDER_INT GL_UNSIGNED_INT

class Shader {
public:

	virtual void setupShaders() = 0;
	virtual void initMesh() = 0;
	virtual void convertToRenderFormat(Mesh& mesh) = 0;
	virtual void render() = 0;
	virtual void deleteProgram() = 0;

	GLuint program = 0;
	int vertexLen = 0;
	int coordIndexLen = 0;

	const RenderInt* coordIndex = 0;
	Matrix viewMatrix;
	Matrix rotationMat;
	GLuint viewMatrixLoc = 0;
	GLuint projMatrixLoc = 0;
	GLuint rotationMatrixLoc = 0;
	GLuint fogBeginLoc = 0;
	GLuint fogScaleLoc = 0;
	GLuint fogColorLoc = 0;
	static float fogBegin;
	static float fogScale;
	static Color3f fogColor;
	GLuint vertexArrayObject[1];
	GLuint texture = 0;

	float transparency = 0;
	GLuint transparencyLoc = 0;
};


class CMonoCLineShader : public Shader {
public:
	static const int arrayBufferLen = 2;
	GLuint arrayBuffer[arrayBufferLen];

	GLuint program = 0;
	const float* coord = 0;
	GLuint coordLoc = 0;

	GLuint scaleMatrixLoc = 0;
	GLuint translationMatrixLoc = 0;

	Color3f f_color;
	GLuint f_colorLoc = 0;

	GLuint v = 0;
	GLuint f = 0;

	void setupShaders();
	void initMesh();
	void convertToRenderFormat(Mesh& mesh);
	void render();
	void deleteProgram();

	CMonoCLineShader()
	{}

	bool firstTime = true;
};

class CCLineShader : public Shader {
public:
	static const int arrayBufferLen = 3;
	GLuint arrayBuffer[arrayBufferLen];

	GLuint program = 0;
	const float* coord = 0;
	GLuint coordLoc = 0;

	const float* color = 0;
	GLuint colorLoc = 0;
	GLuint scaleMatrixLoc = 0;
	GLuint translationMatrixLoc = 0;

	GLuint v = 0;
	GLuint f = 0;

	void setupShaders();
	void initMesh();
	void convertToRenderFormat(Mesh& mesh);
	void render();
	void deleteProgram();

	void convertToRenderFormat(Line& mesh);
	CCLineShader()
	{}

	bool firstTime = true;
};


class CNMonoCShader : public Shader {
public:
	static const int arrayBufferLen = 3;
	GLuint arrayBuffer[arrayBufferLen];

	const float* coord = 0;
	GLuint coordLoc = 0;

	const float* normal = 0;
	GLuint normalLoc = 0;

	GLuint scaleMatrixLoc = 0;
	GLuint translationMatrixLoc = 0;

	Color3f f_color;
	GLuint f_colorLoc = 0;

	GLuint v = 0;
	GLuint f = 0;

	void setupShaders();
	void initMesh();
	void convertToRenderFormat(Mesh& mesh);
	void render();
	void deleteProgram();

	void dumpMesh() {}

	CNMonoCShader()
	{}

	bool firstTime = true;
};


class CNCShader : public Shader {
public:
	static const int arrayBufferLen = 4;
	GLuint arrayBuffer[arrayBufferLen];

	const float* coord = 0;
	GLuint coordLoc = 0;

	const float* normal = 0;
	GLuint normalLoc = 0;

	const float* color = 0;
	GLuint colorLoc = 0;

	GLuint scaleMatrixLoc = 0;
	GLuint translationMatrixLoc = 0;

	GLuint v = 0;
	GLuint f = 0;

	void setupShaders();
	void initMesh();
	void convertToRenderFormat(Mesh& mesh);
	void render();
	void deleteProgram();

	int getIndex(const int& ci, const int& meshCoordLen, const M_Int& rc) {	//get redirected index
		if (ci < meshCoordLen) {
			return ci;
		}
		else {
			return rc[ci - meshCoordLen];
		}
	}

	void dumpMesh();


	CNCShader()
	{}

	bool firstTime = true;

};

class CNTShader : public Shader {
public:
	static const int arrayBufferLen = 4;
	GLuint arrayBuffer[arrayBufferLen];

	const float* coord = 0;
	GLuint coordLoc = 0;

	const float* normal = 0;
	GLuint normalLoc = 0;

	const float* texCoord = 0;
	GLuint texCoordLoc = 0;

	GLuint scaleMatrixLoc = 0;
	GLuint translationMatrixLoc = 0;

	GLuint v = 0;
	GLuint f = 0;

	void setupShaders();
	void initMesh();
	void convertToRenderFormat(Mesh& mesh);
	void render();
	void deleteProgram();

	int getIndex(const int& ci, const int& meshCoordLen, const M_Int& rc) {	//get redirected index
		if (ci < meshCoordLen) {
			return ci;
		}
		else {
			return rc[ci - meshCoordLen];
		}
	}

	void initTexture(Vec2i textureSize, unsigned char* pixels);


	CNTShader()
	{}

	bool firstTime = true;

	M_<Pix32> pixels;
};

class CNC0WeightBoneShader : public Shader {
public:
	static const int arrayBufferLen = 5;
	GLuint arrayBuffer[arrayBufferLen];


	const float* coord = 0;
	GLuint coordLoc = 0;

	const float* normal = 0;
	GLuint normalLoc = 0;

	const float* color = 0;
	GLuint colorLoc = 0;

	M_Int boneIndex;
	GLuint boneIndexLoc = 0;

	M_Matrix boneMatrix;
	GLuint boneMatrixLoc = 0;

	M_Matrix rotBoneMatrix;
	GLuint rotBoneMatrixLoc = 0;

	GLuint translationMatrixLoc = 0;

	GLuint lookDirLoc = 0;

	GLuint v = 0;
	GLuint f = 0;

	void setupShaders();
	void initMesh();
	void convertToRenderFormat(Mesh& mesh);
	void render();
	void deleteProgram();

	int getIndex(const int& ci, const int& meshCoordLen, const M_Int& rc) {	//get redirected index
		if (ci < meshCoordLen) {
			return ci;
		}
		else {
			return rc[ci - meshCoordLen];
		}
	}


	CNC0WeightBoneShader()
	{}

	bool firstTime = true;
};

class CNT0WeightBoneShader : public Shader {
public:
	static const int arrayBufferLen = 5;
	GLuint arrayBuffer[arrayBufferLen];


	const float* coord = 0;
	GLuint coordLoc = 0;

	const float* normal = 0;
	GLuint normalLoc = 0;

	const float* texCoord = 0;
	GLuint texCoordLoc = 0;

	M_Int boneIndex;
	GLuint boneIndexLoc = 0;

	M_Matrix rotBoneMatrix;
	GLuint rotBoneMatrixLoc = 0;

	M_Matrix boneMatrix;
	GLuint boneMatrixLoc = 0;

	GLuint translationMatrixLoc = 0;

	GLuint lookDirLoc = -1;

	GLuint v = 0;
	GLuint f = 0;

	void setupShaders();
	void initMesh();
	void convertToRenderFormat(Mesh& mesh);
	void render();
	void deleteProgram();

	int getIndex(const int& ci, const int& meshCoordLen, const M_Int& rc) {	//get redirected index
		if (ci < meshCoordLen) {
			return ci;
		}
		else {
			return rc[ci - meshCoordLen];
		}
	}
	void initTexture(Vec2i textureSize, unsigned char* pixels);

	CNT0WeightBoneShader()
	{}

	bool firstTime = true;

	M_<Pix32> pixels;
};

class WinShader : public Shader {
public:
	static const int arrayBufferLen = 4;
	GLuint arrayBuffer[arrayBufferLen];

	GLuint program = 0;
	int vertexLen = 4;
	int coordIndexLen = 6;
	Vec2f* coord = 0;
	GLuint coordLoc = 0;

	GLuint texCoordLoc = 0;
	float* texCoord = 0;

	GLuint v = 0;
	GLuint f = 0;

	void setupShaders();
	void initMesh();
	void convertToRenderFormat(Mesh& mesh) {};
	void render();
	void deleteProgram();
	void dumpMesh(){}
	void initTexture(Vec2i textureSize, unsigned char* pixels);
	void setCoord(Vec2i pos, Vec2i size);
	void updatePosSize(Vec2i pos, Vec2i size);

	WinShader()
	{}

};
