/* Seamless3d

Copyright (C) 2001-2013 Graham John Perrett (AKA thyme) thyme@seamless3d.com www.seamless3d.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this
software and associated documentation files (the "Software"), to deal in the Software
without restriction, including without limitation the rights to use, copy, modify,
merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to the following
conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT
OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.
*/


#include "seamless3d.h"

#include "techuelife_triangle_editor.h"
#include "gui.h"
#include "scene_tree.h"
#include "input.h"
#include "render_dx.h"
#include "seamless3d_gui_render.h"
#include "robot_aid.h"


extern double g_time;

//bool g_alphaEnable = false;
//bool g_alphaTexture = false;
extern bool g_backfaceRender;
extern int g_wireFrameMode;
bool g_windowSizeWindowBool = false;
extern bool g_minimized;
extern bool g_runAnim;
extern Group* g_vcSelectRectGroup;
double g_windowSizeWindowTime=0;

extern bool g_multiUserMode;
extern bool g_browseMode;
extern bool g_screenshotEarly;
extern bool g_screenshotLate;

extern Pix24* g_frameBuffer;

extern TransformPtr g_vcTransferBCylBase;
extern bool g_winAPI3DMode;
void processPackets();
void moveMouse3d();
void updatePatchesForMidi();


Render* g_redRenderPoint = 0;
Render* g_lastRenderPoint = 0;
Render* g_nextRenderPoint = 0;
Render* g_yellowRenderPoint = 0;

Render* g_redConeRenderPoint = 0;
Render* g_greenConeRenderPoint = 0;
Render* g_blueConeRenderPoint = 0;


extern IndexedLineSet* g_triClue;
extern Coordinate* g_triClueCoord;
extern Color* g_triClueColor;
extern Shape* g_triClueShape;
extern bool g_curvedCageMode;
//extern Ptr<ControlPanel> g_rcp;



Matrix matProj;
Matrix matView;

extern M_<Ptr<Viewpoint> > g_viewpoints;
extern int g_viewIndex;
bool g_headlight = false;

struct RenderLast{
	Seamless* smls;
	Matrix matrix;
	Vec3f parentColScale;
	RenderLast(Seamless* smls, Matrix matrix, Vec3f parentColScale):
		smls(smls),
		matrix(matrix),
		parentColScale(parentColScale)
	{}
	RenderLast():
		smls(0)
	{}

};

struct TransparencyList{
	Node* node;
	Matrix matrix;
	Vec3f parentColScale;
	TransparencyList(Node* node, Matrix matrix, Vec3f parentColScale):
		node(node),
		matrix(matrix),
		parentColScale(parentColScale)
	{}
	TransparencyList():
		node(0)
	{}

};

struct AlphaChanList{
	Node* node;
	Matrix matrix;
	Vec3f parentColScale;
	AlphaChanList(Node* node, Matrix matrix, Vec3f parentColScale):
		node(node),
		matrix(matrix),
		parentColScale(parentColScale)
	{}
	AlphaChanList():
		node(0)
	{}

};


extern M_<TransparencyList> g_transparentList;
extern M_<AlphaChanList> g_alphaChanList;
extern bool g_renderLater;

extern M_Node renderProcessedQue;

extern Ptr<PartClue> g_vcArtTriBase;
extern TransformPtr g_vcOscBase;
CommonSpec* getCommonSpec(const Art* part);
extern bool g_recordSoundMovie;

extern ShapePtr g_gridShape;
//extern IndexedLineSet* g_nurbsIls;
//extern Coordinate* g_nurbsIlsCoord;
//extern Color* g_nurbsIlsColor;
//extern Shape* g_nurbsIlsShape;
extern TransformPtr g_nurbsIlsBase;

bool backFace(const Vec3f& v0, const Vec3f& v1, const Vec3f& v2);
extern M_Matrix m_matrix;
extern M_Int m_smi;
extern M_Vec3f g_m_coord;

//-----------------------------------------------------------------------------
// Global variables
//-----------------------------------------------------------------------------


void resetImageTextures(){
}


bool getBackDirectX(){
	return false;
}

//-----------------------------------------------------------------------------
// Name: Cleanup()
// Desc: Releases all previously initialized objects
//-----------------------------------------------------------------------------
void cleanupForRefreshSize(){
	g_remMenu=0;

	Window::releaseAllControlPanels();

	g_stw->cp->releaseWindow();
	g_stw->releaseWindow();
	g_tb->releaseWindow();
	g_tt->releaseWindow();
	g_msg->releaseWindow();
	g_cw->releaseWindow();

}
//-----------------------------------------------------------------------------
// Name: Cleanup()
// Desc: Releases all previously initialized objects
//-----------------------------------------------------------------------------
void Cleanup(){
	Window::releaseAllControlPanels();

	g_remMenu=0;
	g_stw=0;
	g_tb=0;
	g_tt=0;
	g_msg=0;
	g_cw=0;
//	g_rcp=0;

}

void SetupMatrices(){
	
}



extern unsigned g_state;


void callBindScript();

inline void manageDownload(){
	if(g_downloadQue.length){
		if(g_downloadQue.length){
			if(g_downloadQue[0]->download()){
				for(int i=0; i<g_downloadQue.length-1; i++){
					g_downloadQue[i]=g_downloadQue[i+1];
				}
				g_downloadQue.length--;
			}
		}
	}
}
void checkKeys();

/*
inline HRESULT checkDXDeviceIsValid(){
	HRESULT  result=g_pd3dDevice->TestCooperativeLevel();
	if(FAILED(result)){
		dLog("TestCooperativeLevel() failed");
		cleanupForRefreshSize();

		InitD3D();
		resetImageTextures();
	}
	return S_OK;
}
http://www.seamless3d.com/wrl/techuelife_island/gm_tree_s9b.smls
*/

extern int g_waitMiliseconds;
void Settings::FreeCPU::operator=(const int& par){
	raw = par;
	if(raw>100){
		raw=100;
	}
	if(raw<0){
		raw = 0;
	}
	g_waitMiliseconds = raw;
}

void scriptBindAnimation();
void doChat();
extern bool g_loggedIn;
bool tugFlag2=false;
extern bool g_debugFlag;
extern bool g_fogExist;

void saveScreenshotToMem(unsigned int* pBits, unsigned int screenWidth, unsigned int width, unsigned int height, Pix24* lpDest);
void saveFrameToMem(unsigned int* pBits, unsigned int screenWidth, unsigned int width, unsigned int height);
void updateBuffer();

void outputScreenShotForDemo1();
void outputScreenShotForDemo2();




Node* Shape::AppearanceField::operator = (Node* par){
	refValue()=par;
	home.mesh.shader = 0;
	home.mesh.reset();
	return par;
}
Node* Shape::GeometryField::operator = (Node* par){
	refValue()=par;
	home.mesh.shader = 0;
	home.mesh.reset();
	return par;
}

Matrix dxWindMatrix(Matrix& parentMatrix, Vec3f& scale, Rotation& rotation, Vec3f& translation);

inline Matrix Scene::dxWindMatrix(Matrix& parentMatrix){

	Matrix sclMatrix = scaleMatrix(scale);
	Matrix rotMatrix = rotationMatrix(rotation);
	Matrix trnMatrix = translationMatrix(translation);
	Matrix outputMatrix = sclMatrix * rotMatrix * trnMatrix * parentMatrix;

	return outputMatrix;
}
Matrix Transform::dxWindMatrix(Matrix& parentMatrix){
//   P' = T  C  R  SR  S  -SR  -C  P

	return translationMatrix(-center) * rotationMatrix(-scaleOrientation) * scaleMatrix(scale) * rotationMatrix(scaleOrientation) *
		rotationMatrix(rotation) * translationMatrix(center) * translationMatrix(translation) * parentMatrix;
}
void Inline::renderNode(Matrix parentMatrix, Matrix catRotMatrix){
	Matrix thisMatrix=parentMatrix;
	for(int k=0;k<children.length;k++){
		children[k]->renderNode(thisMatrix, catRotMatrix);
	}
}
void Scene::renderNode(Matrix parentMatrix, Matrix catRotMatrix){
	catRotMatrix = rotationMatrix(rotation) * catRotMatrix;

	Matrix thisMatrix=dxWindMatrix(parentMatrix);
	M_Node& children=this->children.refValue();
	for(int k=0;k<children.length;k++){
		children[k]->renderNode(thisMatrix, catRotMatrix);
	}
}
void Transform::renderNode(Matrix parentMatrix, Matrix catRotMatrix){
	catRotMatrix = rotationMatrix(rotation) * catRotMatrix;

	Matrix thisMatrix=dxWindMatrix(parentMatrix);
	for(int k=0;k<children.refValue().length;k++){
		children.refValue()[k]->renderNode(thisMatrix, catRotMatrix);
	}
}
void ProtoDeclare::renderNode(Matrix parentMatrix, Matrix catRotMatrix){
	M_Node& children=protoBody->children.refValue();
	Matrix thisMatrix=parentMatrix;
	for(int k=0;k<children.length;k++){
		children[k]->renderNode(thisMatrix, catRotMatrix);
	}
}
void AGroup::renderNode(Matrix parentMatrix, Matrix catRotMatrix){
	Matrix thisMatrix=parentMatrix;
	M_Node& children=this->children.refValue();
	for(int k=0;k<children.length;k++){
		children[k]->renderNode(thisMatrix, catRotMatrix);
	}
}

void NoOutputGroup::renderNode(Matrix parentMatrix, Matrix catRotMatrix){
	if(g_screenshotEarly){
		return;
	}
	M_Node& children = this->children.refValue();
	for(int k=0;k<children.length;k++){
		children[k]->renderNode(parentMatrix, catRotMatrix);
	}
}

void Mesh::transformRenderTexture(Render*& render, TextureTransform* textureTransform){
}



void resetCoordLen(M_Art& partList);
void setUpInternalEffectPtrsTmpForRenderingTransparancy(Seamless* smls);
void setUpInternalEffectPtrs2(Seamless* smls);

void RingBuild::resetCo(Art* part, int& co){
	int i = index[1];
	if(this->part.length && i < this->part.length){
		if(part == this->part[i].p){
			co = 0;
		}
	}
}
void TransferOwnership::resetCo(Art* part, int& co){
	int i = index[0];
	if(dest.length && i < dest.length){
		if(part == dest[i].p){
			co = 0;
		}
	}
}

void Seamless::convertToOwnershipAidRenderFormat(){
}
extern SeamlessPtr g_vcTransferBCyl;

void Seamless::convertToSimpleRenderFormat(){
}

void HAnimJoint::setJointIndices(int& smi, int midx, COLORVERTEX_1WEIGHT* pVertices, M_Int& maskIndex){
}
void HAnimJoint::setJointIndices(int& smi, int midx, TEXTUREVERTEX_1WEIGHT* pVertices, M_Int& maskIndex){
}
void HAnimDisplacer::Weight::operator =(const float& interp){
}


void Morpher::Weight::operator =(const float& interp){
}


template<typename T> void dragPartForMorph2(Render*& render, Art*& part, Vec3f newPt, Vec3f diff){
	T* pVertices;
#ifdef DX8
	if( FAILED( render->pVB->Lock( 0, 0, (BYTE**)&pVertices, 0 ) ) ){
		throw;
	}
#else
	if( FAILED( render->pVB->Lock( 0, 0, (void**)&pVertices, 0 ) ) ){
		throw;
	}
#endif
	for(int i = MultiSelect::coordList.length; --i >= 0;){
		Art* part = MultiSelect::coordList[i].p;
		int idx = MultiSelect::coordList[i].i;
		M_Vec3f& coord = part->coord.refValue();
		int renderFromCoord = part->renderFromCoord;
		(*(Vec3f*)&pVertices[renderFromCoord + idx].position) = (g_m_coord[part->fromCoord + idx] = MultiSelect::beginDragPos[i] + diff) - part->absolutePivotPoint;

		int mark = renderFromCoord + part->coord.length; //if need be we could make this more efficiant by flagging each part and doing this after each element in the coordList has been processed
		const M_Int& rc = part->rc;
		for(int i = rc.length; --i>=0; ){
			pVertices[mark + i].position = pVertices[renderFromCoord + rc[i]].position;
		}
	}

	//smls->morphFlag = true;
	render->pVB->Unlock();
}


void AnimLever::dragPartForMorph(Vec3f pt){
	if(Art* part = dynamic_cast<Art*>(g_selNode.getRaw())){
		Seamless* smls = part->smls;
		g_dragActuallyHappened = true;
		const M_Int& coordGroup = part->coordGroup;
		const M_Float& weight = part->weight;

		smls->catIMatrix();
		Matrix imatrix = part->plidx > 0 ? smls->partList[part->plidx - 1]->imatrix : smls->imatrix;

		Vec3f translation = part->translation.refValue();
		Rotation rotation = part->rotation.refValue();
		Vec3f scale = part->scale.refValue();
		Vec3f absolutePivotPoint = part->absolutePivotPoint;
		
		int idx = part->index[0];
		int coordsInGroups = 0;
		float w = 1;

		for(int cg = 0; cg < coordGroup.length; cg++){
			coordsInGroups += coordGroup[cg];
			if(idx < coordsInGroups){
				w = weight[cg];
				break;
			}
		}

		rotation.angle = rotation.angle * w;

		imatrix = imatrix * translationMatrix(-translation - absolutePivotPoint) * rotationMatrix( -rotation) * scaleMatrix(Vec3f(1./scale.x, 1./scale.y, 1./scale.z)) * translationMatrix(absolutePivotPoint);

		Vec3f newPt = g_scene->getIViewMatrix() * g_catchIMatrix * imatrix * pt;
		int i = part->index[0];
		Vec3f diff = newPt - MultiSelect::beginDragPos[MultiSelect::clickedIndex];

		throw;
	}
}

void renderMorphPart(const Art& part, const M_Vec3f& srcCoordA, const M_Vec3f& srcCoordB, float interp){
	throw;
}


void Seamless::updateAllVertexInBuffers(){
	ctrf = true;
	return;
}


void checkFieldsMatch(IndexedFaceSet &geometry){
	Coordinate* coord=dynamic_cast<Coordinate*>(geometry.coord.getRawPtr());
	Normal* normal=dynamic_cast<Normal*>(geometry.normal.getRawPtr());
	Color* color=dynamic_cast<Color*>(geometry.color.getRawPtr());
	TextureCoordinate* texCoord=dynamic_cast<TextureCoordinate*>(geometry.texCoord.getRawPtr());
	if(coord!=NULL){
		if(normal!=NULL && coord->point.refValue().length != normal->vector.refValue().length){
			dLog("coord to normal mismatch");
			normal->vector.resize(coord->point.refValue().length);
			geometry.genNormals();
			//throw geometry;
		}
		if(color!=NULL && coord->point.refValue().length != color->color.refValue().length){
			dLog("**error!** coord to color mismatch");
			color->color.resize(coord->point.refValue().length);
			//throw geometry;
		}
		if(texCoord!=NULL && coord->point.refValue().length != texCoord->point.refValue().length){
			dLog("**error!** coord to texCoord mismatch");
			texCoord->point.resize(coord->point.refValue().length);
			//throw geometry;
		}

	}
	else{
		if(normal!=NULL){
			dLog("**error!** normal != NULL");
			normal=NULL;
//			throw geometry;
		}
		if(color!=NULL){
			dLog("**error!** color != NULL");
			color=NULL;
//			throw geometry;
		}
		if(texCoord!=NULL){
			dLog("**error!** texCoord != NULL");
			texCoord=NULL;
//			throw geometry;
		}
	}
}
/*
void Art::genDynamicTransforms(Matrix parentMatrix, int& smi){
	
	pivotPointMatrix = translationMatrix(absolutePivotPoint) * translationMatrix(translation)*parentMatrix;

	for(int cg=0;cg<coordGroup.length; cg++){
		g_pd3dDevice->SetTransform( D3DTS_WORLDMATRIX(smi), (D3DMATRIX*)&(translationMatrix(-absolutePivotPoint)*scaleMatrix(scale)*rotationMatrix(Rotation(rotation.x,rotation.y,rotation.z, rotation.angle*weight[cg]))*pivotPointMatrix) );
		smi++;
	}

	actualPivotPointMatrix=scaleMatrix(scale)*rotationMatrix(rotation)*pivotPointMatrix;
	pivotPointMatrix=translationMatrix(-absolutePivotPoint)*actualPivotPointMatrix;


	g_pd3dDevice->SetTransform( D3DTS_WORLDMATRIX(smi), (D3DMATRIX*)&pivotPointMatrix );
	smi++;
	for(int k=0;k<part.length;k++){
		part.refValue()[k]->genDynamicTransforms(pivotPointMatrix, smi);
	}
}
*/


void HAnimJoint::genDynamicTransforms(Matrix parentMatrix, int& smi){
}
extern ShapePtr g_vcEllipticalPolyline2D;
extern ShapePtr g_nurbsIlsShape;
extern ShapePtr g_transformLeverClue;
extern ShapePtr g_translateDimensionClue;
extern ShapePtr g_vcFuseSurfaceFlatStageAidShape;


extern Ptr<PartCageClue> g_partCageClue;
extern ShapePtr g_vcSelectRectShape;



extern SeamlessPtr g_cameraObject;
extern bool g_showCamerasInAnimMode;

extern TransformPtr g_redConeBase;
extern TransformPtr g_greenConeBase;
extern TransformPtr g_blueConeBase;





extern Coordinate* g_vcSkeletonLinesCoordinate;

void renderPivotPoint(Art* part, Matrix& tranMatrix, Matrix catRotMatrix);

void CenterJointClue::renderNode(Matrix parentMatrix, Matrix catRotMatrix){
	throw;
}
void TransferOwnershipClue::renderNode(Matrix parentMatrix, Matrix catRotMatrix){
	throw;
}
void TransferBCylClue::renderNode(Matrix parentMatrix, Matrix catRotMatrix){
	throw;
}





template<class T> inline void  do_polyFrameCommon(T pVertices, const int& renderFromCoord, const M_Vec3f& coord, const M_Int& rc, const Vec3f& absolutePivotPoint, const int& i, const int& portionLen, const M_Vec3f& keyValue, const float& interp, const M_Int& cpIndex){
	int toCoord = renderFromCoord + coord.length;
	int coordLen = coord.length;

	for(int j = 0; j < portionLen; j++){
		int h = i * portionLen + j;
		Vec3f result = keyValue[h] + (keyValue[(i + 1) * portionLen + j] - keyValue[h]) * interp;
		int idx = cpIndex[j];
		if(idx < coordLen){
			pVertices[renderFromCoord + idx].position = (D3DXVECTOR3&) (coord[idx] - absolutePivotPoint + result);
		}
	}

	int mark = renderFromCoord + coord.length;
	for(int i = rc.length; --i>=0; ){
		pVertices[mark + i].position = pVertices[renderFromCoord + rc[i]].position;
	}
}

void CPMorpher::do_polyFrame(const Art* part){
	throw;
}


Seamless* getModelMorpher(ModelMorpher* modelMorpher){
	int selLine = g_stw->selLine;
	if(Seamless* selSmls = getSeamless(selLine)){
		selLine = getParent(selLine);
		if(dynamic_cast<ModelSequence*>(g_scene->sceneTreeList[selLine].node.p)){
			if(modelMorpher == g_scene->sceneTreeList[getParent(selLine)].node){
				return selSmls;
			}
		}
	}
	return 0;
}

void ModelMorpher::renderNode(Matrix matrix, Matrix catRotMatrix){
	
	if(Seamless* selSmls = getModelMorpher(this)){
		selSmls->renderNode(matrix, catRotMatrix);
	}
	else{
		if(smls){
			smls->renderNode(matrix, catRotMatrix);
		}
	}
}
void HAnimHumanoid::renderNode(Matrix matrix, Matrix catRotMatrix){
}
void Background::renderNode(Matrix parentMatrix, Matrix catRotMatrix){
	if(skyColor.refValue().length){
		g_scene->backgroundColor=skyColor.refValue()[0];
	}
}
int g_lightStack;
void DirectionalLight::renderNode(Matrix parentMatrix, Matrix catRotMatrix){
	return;
}

void NavigationInfo::renderNode(Matrix parentMatrix, Matrix catRotMatrix){
	g_headlight = headlight.refValue();
}

Vec2i get2dMousePosFrom3d(Vec3f par){
	Vec2i pixPos;
	Vec2f size = Vec2f(g_3d->size.x, g_3d->size.y);

	pixPos.y = -(par.y - 1.0) * .5 * size.y + .5;

	float ratioBig = size.x / size.y;
	float ratioSmall = size.y / size.x;
	pixPos.x = ((par.x + ratioBig) * ratioSmall) * .5 * size.x + .5;
	
	return pixPos;
}

Vec2i get2dMousePosFrom2d(Vec2f par){
	return get2dMousePosFrom3d(Vec3f(par.x, par.y, 0));
}


 
// calculate the cofactor of element (row,col)
int GetMinor(float **src, float **dest, int row, int col, int order)
{
    // indicate which col and row is being copied to dest
    int colCount=0,rowCount=0;
 
    for(int i = 0; i < order; i++ )
    {
        if( i != row )
        {
            colCount = 0;
            for(int j = 0; j < order; j++ )
            {
                // when j is not the element
                if( j != col )
                {
                    dest[rowCount][colCount] = src[i][j];
                    colCount++;
                }
            }
            rowCount++;
        }
    }
 
    return 1;
}
 
// Calculate the determinant recursively.
double CalcDeterminant( float **mat, int order)
{
    // order must be >= 0
    // stop the recursion when matrix is a single element
    if( order == 1 )
        return mat[0][0];
 
    // the determinant value
    float det = 0;
 
    // allocate the cofactor matrix
    float **minor;
    minor = new float*[order-1];
    for(int i=0;i<order-1;i++)
        minor[i] = new float[order-1];
 
    for(int i = 0; i < order; i++ )
    {
        // get minor of element (0,i)
        GetMinor( mat, minor, 0, i , order);
        // the recusion is here!
 
        det += (i%2==1?-1.0:1.0) * mat[0][i] * CalcDeterminant(minor,order-1);
        //det += pow( -1.0, i ) * mat[0][i] * CalcDeterminant( minor,order-1 );
    }
 
    // release memory
    for(int i=0;i<order-1;i++)
        delete [] minor[i];
    delete [] minor;
 
    return det;
}
//Jason Yu-Tseh Chis Notes
//https://chi3x10.wordpress.com/2008/05/28/calculate-matrix-inversion-in-c/
// matrix inversioon
// the result is put in Y
void MatrixInversion(float **A, int order, float **Y)
{
    // get the determinant of a
    double det = 1.0/CalcDeterminant(A,order);
 
    // memory allocation
    float *temp = new float[(order-1)*(order-1)];
    float **minor = new float*[order-1];
    for(int i=0;i<order-1;i++)
        minor[i] = temp+(i*(order-1));
 
    for(int j=0;j<order;j++)
    {
        for(int i=0;i<order;i++)
        {
            // get the co-factor (matrix) of A(j,i)
            GetMinor(A,minor,j,i,order);
            Y[i][j] = det*CalcDeterminant(minor,order-1);
            if( (i+j)%2 == 1)
                Y[i][j] = -Y[i][j];
        }
    }
 
    // release memory
    //delete [] minor[0];
    delete [] temp;
    delete [] minor;
}

void destroyRender(){
}



Effect* workOutEffectType(const Appearance*& appearance);

//void Cone::renderNode(Matrix parentMatrix, Vec3f parentColScale, Mesh*& mesh, Render*& render, Appearance* appearance);
void Script::renderNode(Matrix parentMatrix, Matrix catRotMatrix){
}
void SceneTree::resizePageUpPageDown(){
	pageUp.dstPos=topScroll.dstPos+Vec2i(5, slider.bSize.y);
	pageUp.bSize=Vec2i(6, slider.dstPos.y - topScroll.dstPos.y - topScroll.bSize.y);
	pageDown.dstPos=slider.dstPos+Vec2i(5, slider.bSize.y);
	pageDown.bSize=Vec2i(6, bottomScroll.dstPos.y - pageDown.dstPos.y);
}

void Window::releaseAllControlPanels(){
}


void convertBMPToPng(const char* filename, int width, int height, M_<char>& header, Pix24* pixels){
}

unsigned int TextureEffect::readPixelFromTexture(){
	return 0;
}
extern BrushPtr g_brush;
extern FillPtr g_fill;

unsigned pix24ToInt(Pix24 pix){
	unsigned r = pix.r;
	unsigned g = pix.g;
	unsigned b = pix.b;

	r <<= 16;
	g <<= 8;

	return r + g + b;
}

Pix24 intToPix24(unsigned int i){
	return Pix24((i >> 16) & 0x0ff, (i >> 8) & 0x0ff, i & 0x0ff);
}

Color3f pix24ToColor(Pix24 pix){
	return Color3f(pix.r / 255., pix.g / 255., pix.b / 255.);
}

Color3f intToColor3f(unsigned int i){
	return pix24ToColor(intToPix24(i));
}
bool compareColorTolerance(Color3f color, Color3f targetColor, float tolerance);

void TextureEffect::colourWithFill(unsigned paletteColor){
}

void TextureEffect::copyPixelsToStore(){
}


void TextureEffect::writePixelsToTexture(const Color3f color){
}
