/* Seamless3d

Copyright (C) 2001-2017 Graham John Perrett (AKA thyme) thyme@seamless3d.com www.seamless3d.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this
software and associated documentation files (the "Software"), to deal in the Software
without restriction, including without limitation the rights to use, copy, modify,
merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to the following
conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT
OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.
*/

extern const String g_demoDir;
extern const String g_demoCacheDir;
extern Ptr<Demo> g_robotDemo;
extern Ptr<ControlPanel> g_rcp;
extern Ptr<AnimControlPanel> g_acp;
extern Ptr<VideoControlPanel> g_vcp;
extern Ptr<WaveToPolyControlPanel> g_wtpcp;

struct TugData{
	int index;
	Vec2f coord;
};
struct GenPolyData{
	int selLine;
	int index;
	Vec2f coord;
};
struct SwapTriData{
	int selLine;
	int index;
};

void robotDemoStart(void (* i_queFunction)());
void setupSequence();
void setupSequence2();

class InputEvent{
public:
	//enum {
	//	TIMEID=75
	//};
	static float period;
	virtual bool doEvent()= 0;
	static Vec2i pos;
	static Vec2i startPos;
	static Vec3f startPos3d;
	static const float downPeriod;
	static const float upPeriod;
	static const float prePeriod;
	static const float postPeriod;
	static int sceneTreeScrolPos;

	static int renderCounter;
	static int syncNumOfActions;
	static int prevRenderCounter;

	static M_<InputEvent*> list;
	static void emptyList();
	static int sequencerIdx;

	static bool loadPrevFromCache(int operation);
	static InputEvent* readDemoEvent(const String& str, int& i);
	static void fillList(const String& str, int& i);
	static void getInstruction(const String& str, int& i);
	static void decodeDemo(const String& str, int& i);
	static void* getFirstButtonParam(const String& str, int& i);
	static int get2ndButtonParam(const String& str, int& i);
	static bool doDoubleClick(int& phase);
	static bool doSingleClick(int& phase);
	static bool doMoveToNewNode(const int id);
	static bool doMoveToNode(const int line);
	static bool doLeftButtonDown();
	static bool doLeftButtonUp(bool& released);
	static bool doRightButtonDown();
	static bool doRightButtonUp(bool& released);
	static bool doMoveToNodeAsSiblingAbove(const int nodeIdx);
	static bool doMoveToNodeAsSiblingBelow(const int nodeIdx);
	static bool doMoveToNodeAsChild(int& phase, const int nodeIdx);//pase must be initialised with 100 before calling this function for the first time
	static bool doMoveToNodeAsSource(int& phase, const int nodeIdx);
	static bool doMoveToNodeAsDest(int& phase, const int nodeIdx);
	static bool doMoveToNodeAsReplacement(int& phase, const int nodeIdx);
	static bool doDecIncField(const char* fieldName, const int xOffset, const int yOffset, const int layer); //16 for dec, 6 for inc 
	static bool doClickAndAutoRepeat(int& phase, const int numOfActions=1);
	static bool doVirtKeyDown(const unsigned char vk);
	static bool doVirtKeyUp(const unsigned char vk);
	static bool doFieldCtrl(int& phase, const char* fieldName, const int xOffset, const int yOffset, const int numOf, const int layer);
	static bool doTypeString(const String str, int& charIdx);
	static bool doMoveToFieldValue(const char* fieldName, int& phase, const int layer);
	static bool doTypeVirtKey(char vk);

	static void preYouTubeTut();
	static void youTubeTut();

	static void preBeelineAnimDemo();
	static void beelineAnimDemo();

	static void preSwitchAnimIn3dDemo();
	static void switchAnimIn3dDemo();

	static void preImmediateRadialTugDemo();
	static void immediateRadialTugDemo();

	static void preBuildTexturedObjectDemo();
	static void buildTexturedObjectDemo();
	static void preWormAnimationDemo();
	static void wormAnimationDemo();
	static void preNurbsLatheAvDemo();
	static void nurbsLatheAvDemo();

	static void preMultiCtrlPointEditDemo();
	static void multiCtrlPointEditDemo();

	static void preNurbsStitchedHeadDemo();
	static void nurbsStitchedHeadDemo();

	static void preLimbAnimDemo();
	static void limbAnimDemo();

	static void preNurbsPatchAvDemo();
	static void nurbsPatchAvDemo();

	static void preBringToLifeDemo();
	static void bringToLifeDemo();

	static void preImmediateAvDemo();
	static void immediateAvDemo();

	static void buildSeamedAvDemoGenTextureMap();
	static void buildSeamedAvDemoGenAnimate();

	static void preYouTubeTutDemo();
	static void showArtDemo();

	static void copyNodeAsChild(int nodeLineFrom, int nodeLineTo);
	static void copyNodeAsSiblingAbove(int nodeLineFrom, int nodeLineTo);
	static void copyNodeAsReplacement(int nodeLineFrom, int nodeLineTo);
	static void copyNodeAsSiblingBelow(int nodeLineFrom, int nodeLineTo);
	static void copyNode(int nodeLine);



	static void rightClickNode(int nodeLine);
	static void newNodeAsChild(int nodeID, int nodeLine);
	static void newNodeAsReplacement(int nodeID, int nodeLine);
	static void newNodeAsSiblingBelow(int nodeID, int nodeLine);
	static void nameNode(String str, int col, int line);
	static void useNodeAsSource(int nodeLineFrom, int nodeLineTo);
	static void useNodeAsDest(int nodeLineFrom, int nodeLineTo);
	static void useNodeAsSiblingBelow(int nodeLineFrom, int nodeLineTo);
	static void useNodeAsSiblingAbove(int nodeLineFrom, int nodeLineTo);
	static void useNodeAsReplacement(int nodeLineFrom, int nodeLineTo);
	static void setFieldValue(const char* field, const char* value);
	static void setFieldValueNoShift(const char* field, const char* value);

	static void incField(const char* field, int num, const int buttonOffset=0, const int layer=0);
	static void incFieldCtrl(const char* field, int num, const int buttonOffset=0);
	static void decField(const char* field, int num, const int buttonOffset=0);
	static void decFieldCtrl(const char* field, int num, const int buttonOffset=0);
	static void collapse(int nodeLine);
	static void textureTile(Vec2i point);
	static void hide(int nodeLine);
	static void deleteMinus(const char* field);
	static void addMinus(const char* field);
	static void invertRotZ(const char* field);

	static void incGroupCount(const char* field);
	static void decGroupCount(const char* field);

	static void decIndex(const char* field, int numOf=1);
	static void incIndex(const char* field, int numOf=1);
	static void selNode(int nodeLine);
	static void selNodeByName(const char* name);
	static void nameNode(int nodeLine, const char* newName);
	static void renameNode(int nodeLine, const char* newName);
	static void typeToNode(int nodeLine, const char* text, int charIndex=0);

	static void renameLtoR(int nodeLine);

	static void clickButton(void* id, int numOfTimes=1, const int offset=0);
	static void autoRepeat(void* id, int numOfTimes=1, const int offset=0);
	static void autoRepeatCtrl(void* id, int numOfTimes=1, const int offset=0);
	static void newCenterJoint(int nodeLine, int numOf);
	static void joinParts(int copy, int sel);
	static void tugVertices(const GenPolyData* tugData, int numOf);
	static void tugVertices(const TugData* tugData, int numOf, int selLine);
	static void tugRings(const TugData* tugData, int numOf);
	static void joinVertices(const int* coordIndex, int numOf);
	static void joinVertices2(const int* coordIndex, int numOf);
	static void projectVertices(const int* coordIndex, int numOf, int selLine);
	static void selVertices(const int* tugData, int numOf);
	static void patchTriangles(const int line, const int* tugData, int numOf);
	static void patchTriangles2(const int* tugData, int numOf);
	static void splitMove(const GenPolyData* tugData, int numOf);
	static void swapTriangles(const SwapTriData* tugData, int numOf);
	static void deleteTriangles(const int* tugData, int numOf, int line=-1);
	static void paintTriangles(const int* triIndex, int numOf, const int line);




	static float InputEvent::move(Vec2i endPos);
	static float InputEvent::localMove(Vec2i endPos);
	static float InputEvent::moveTo3dPoint(Vec2f endPos);

	static Vec2i InputEvent::getSysMouseCoord();
	static void runRobotSequence();
	//static UINT_PTR timeID;
	static unsigned int timeID;
	static float fraction_changed;
	static double beginRobotTime;
	static bool robotActive;
	static void _killTimer();
	static bool resumeSequence();
	static bool spaceDown;
	static bool l_altDown;
	static bool l_ctrlDown;
	static bool l_shiftDown;
	static bool r_altDown;
	static bool r_ctrlDown;
	static bool r_shiftDown;
	static bool arrowDown;
	static bool leftClick;
	static bool doubleClick;
	static bool rightClick;
	static bool leftDrag;
	static bool rightDrag;
	static void outputStatus();
	static M_<char> dynamChar;
	static void nextIdxForDynamChar(int& i);
};
class Section : public InputEvent {
	float period;
	int phase;
public:
	const char* title;
	Section(const char* title):
		title(title),
		period(2),
		phase(0)
	{}
	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
};
class Wait : public InputEvent {
	float period;
	float moveSpeed;
public:
	Wait(const float& period):
		period(period),
		moveSpeed(0)
	{}
	Wait(const float& period, const float moveSpeed):
		period(period),
		moveSpeed(moveSpeed)
	{}
	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
};
class Move3d : public InputEvent {
	const Vec2f to;
public:
	Move3d(const Vec3f to):
	  to(to.x, to.y)
	{}
	Move3d(const Vec2f to):
	  to(to)
	{}
	Move3d(const float x, float y):
	  to(x,y)
	{}
	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
};

class DemoMove : public InputEvent {
	static Vec3f from;
	float mouseFromX;
	float mouseFromY;
	int phase;
public:
	DemoMove(float mouseFromX, float mouseFromY):
		mouseFromX(mouseFromX),
		mouseFromY(mouseFromY),
		phase(0)
	{}
	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
};

class ShowFile : public InputEvent {
	String file;
public:
	ShowFile(String file):
		file(file)
	{}
	bool doEvent();
};

class Move : public InputEvent {
	Vec2i to;
	float moveSpeed;
public:
	Move(unsigned x,unsigned y):
		to(x,y),
		moveSpeed(0)
	{}
	Move(const Vec2i& to):
		to(to),
		moveSpeed(0)
	{}
	Move(unsigned x,unsigned y, float moveSpeed):
		to(x,y),
		moveSpeed(moveSpeed)
	{}
	bool doEvent();
};

class RelMove : public InputEvent {
	Vec2i to;
	int phase;
public:
	RelMove(int x, int y):
		to(x,y),
		phase(0)
	{}
	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
};
class MoveRelWindow : public InputEvent {
	Vec2i to;
	int phase;
	const int layer;
public:
	MoveRelWindow(int x, int y, const int layer=0):
		to(x,y),
		layer(layer),
		phase(0)
	{}
	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
};
class ResizeSceneTreeWin : public InputEvent {
public:
	int width;
	int height;
	ResizeSceneTreeWin(int width, int height):
		width(width),
		height(height)
	{}
	bool doEvent(){
		g_stw->size=Vec2i(width, height); 
		g_stw->pos = Vec2i(g_3d->size.x - (width), 0);
		g_stw->newNodeBar.dstPos = Vec2i(g_stw->size.x - 80, 30);
		return true;
	}
};
class ClickAndAutoRepeat : public InputEvent {
	int phase;
	int numOfActions;
public:
	ClickAndAutoRepeat(const int& numOfActions):
		phase(0),
		numOfActions(numOfActions)
	{}
	bool doEvent(){
		return doClickAndAutoRepeat(phase, numOfActions);
	}
};

class RightClick : public InputEvent {
	int phase;
	float downPeriod;
	float upPeriod;
public:
	RightClick(const float& downPeriod=.25):
		phase(0),
		downPeriod(downPeriod),
		upPeriod(.125)
	{}
	bool doEvent();
};
class DemoSound : public InputEvent {
public:
	const char* filename;
	DemoSound(const char* filename):
		filename(filename)
	{}
	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
};

class LeftButtonDown : public InputEvent {
public:
	LeftButtonDown()
	{}
	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
};
class LeftButtonUp : public InputEvent {
	bool released;
public:
	LeftButtonUp():
		released(false)
	{}
	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
};
 
class RightButtonDown : public InputEvent {
	float period;
public:
	RightButtonDown():
		period(.25)
	{}
	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
};
class RightButtonUp : public InputEvent {
	bool released;
public:
	RightButtonUp():
		released(false)
	{}
	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
};


class Click : public InputEvent {
	int phase;
public:
	Click():
		phase(0)
	{}
	bool doEvent(){
		return doSingleClick(phase);
	}
	static InputEvent* decode(const String& str, int& i);
};
class DoubleClick : public InputEvent {
	int phase;
public:
	DoubleClick():
		phase(0)
	{}
	bool doEvent(){
		return doDoubleClick(phase);
	}
	static InputEvent* decode(const String& str, int& i);
};
class MoveToExitWindow : public InputEvent {
	const int layer;
public:
	MoveToExitWindow(const int layer=0):
		layer(layer)
	{}
	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
};
class MoveToExitSceneTree : public InputEvent {
public:
	MoveToExitSceneTree()
	{}
	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
};
class MoveToTitleBar : public InputEvent {
	const int layer;
public:
	MoveToTitleBar(const int layer=0):
		layer(layer)
	{}
	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
};
class ToggleField : public InputEvent {
	const char* fieldName;
	const int layer;
	int phase;
public:
	ToggleField(const char* fieldName, const int layer=0):
		fieldName(fieldName),
		layer(layer),
		phase(0)
	{}
	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
};

class ClickColorCell : public InputEvent {
	const Vec2i cellIndex;
	int phase;
public:
	ClickColorCell(int row, int colum):
		cellIndex(row, colum),
		phase(0)
	{}
	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
};

class ClickYes : public InputEvent {
	int phase;
public:
	ClickYes():
		phase(0)
	{}
	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
};

class MoveToField : public InputEvent {
	const char* fieldName;
	const int layer;
	int phase;
public:
	MoveToField(const char* fieldName, const int layer=0):
		fieldName(fieldName),
		layer(layer),
		phase(0)
	{}
	bool doEvent(){
		return doMoveToFieldValue(fieldName, phase, layer);
	}
	static InputEvent* decode(const String& str, int& i);
};
class MoveToF10 : public InputEvent {
public:
	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
};
class MoveToTextureTile : public InputEvent {
	const Vec2i tile;
	const int layer;
public:
	MoveToTextureTile(const Vec2i& tile, const int layer=0):
		tile(tile),
		layer(layer)
	{}
	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
};
class MoveToFieldInc : public InputEvent {
	const char* fieldName;
	const int buttonOffset;
	const int layer;
public:
	MoveToFieldInc(const char* fieldName, const int& buttonOffset=0, const int layer=0):
		fieldName(fieldName),
		buttonOffset(buttonOffset),
		layer(layer)
	{}
	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
};
class MoveToFieldDec : public InputEvent {
	const char* fieldName;
	const int buttonOffset;
	const int layer;
public:
	MoveToFieldDec(const char* fieldName, const int& buttonOffset=0, const int layer=0):
		fieldName(fieldName),
		buttonOffset(buttonOffset),
		layer(layer)
	{}
	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
};
class DecFieldX : public InputEvent {
	const char* fieldName;
	const int numOf;
	const int layer;
	int phase;
public:
	DecFieldX(const char* fieldName, const int numOf=1, const int layer=0):
		fieldName(fieldName),
		numOf(numOf),
		layer(layer),
		phase(0)
	{}
	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
};
class DecFieldY : public InputEvent {
	const char* fieldName;
	const int numOf;
	const int layer;
	int phase;
public:
	DecFieldY(const char* fieldName, const int numOf=1, const int layer=0):
		fieldName(fieldName),
		numOf(numOf),
		layer(layer),
		phase(0)
	{}
	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
};
class DecFieldZ : public InputEvent {
	const char* fieldName;
	const int numOf;
	const int layer;
	int phase;
public:
	DecFieldZ(const char* fieldName, const int numOf=1, const int layer=0):
		fieldName(fieldName),
		numOf(numOf),
		layer(layer),
		phase(0)
	{}
	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
};
class IncFieldX : public InputEvent {
	const char* fieldName;
	const int numOf;
	const int layer;
	int phase;
public:
	IncFieldX(const char* fieldName, const int numOf=1, const int layer=0):
		fieldName(fieldName),
		numOf(numOf),
		layer(layer),
		phase(0)
	{}
	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
};
class IncFieldY : public InputEvent {
	const char* fieldName;
	const int numOf;
	const int layer;
	int phase;
public:
	IncFieldY(const char* fieldName, const int numOf=1, const int layer=0):
		fieldName(fieldName),
		numOf(numOf),
		layer(layer),
		phase(0)
	{}
	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
};
class IncFieldZ : public InputEvent {
	const char* fieldName;
	const int numOf;
	const int layer;
	int phase;
public:
	IncFieldZ(const char* fieldName, const int numOf=1, const int layer=0):
		fieldName(fieldName),
		numOf(numOf),
		layer(layer),
		phase(0)
	{}
	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
};
class DecFieldCtrlX : public InputEvent {
	const char* fieldName;
	const int numOf;
	const int layer;
	int phase;
public:
	DecFieldCtrlX(const char* fieldName, const int numOf=1, const int layer=0):
		fieldName(fieldName),
		numOf(numOf),
		layer(layer),
		phase(0)
	{}
	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
};
class DecFieldCtrlY : public InputEvent {
	const char* fieldName;
	const int numOf;
	const int layer;
	int phase;
public:
	DecFieldCtrlY(const char* fieldName, const int numOf=1, const int layer=0):
		fieldName(fieldName),
		numOf(numOf),
		layer(layer),
		phase(0)
	{}
	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
};
class DecFieldCtrlZ : public InputEvent {
	const char* fieldName;
	const int numOf;
	const int layer;
	int phase;
public:
	DecFieldCtrlZ(const char* fieldName, const int numOf=1, const int layer=0):
		fieldName(fieldName),
		numOf(numOf),
		layer(layer),
		phase(0)
	{}
	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
};
class IncFieldCtrlX : public InputEvent {
	const char* fieldName;
	const int numOf;
	const int layer;
	int phase;
public:
	IncFieldCtrlX(const char* fieldName, const int numOf=1, const int layer=0):
		fieldName(fieldName),
		numOf(numOf),
		layer(layer),
		phase(0)
	{}
	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
};
class IncFieldCtrlY : public InputEvent {
	const char* fieldName;
	const int numOf;
	const int layer;
	int phase;
public:
	IncFieldCtrlY(const char* fieldName, const int numOf=1, const int layer=0):
		fieldName(fieldName),
		numOf(numOf),
		layer(layer),
		phase(0)
	{}
	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
};
class IncFieldCtrlZ : public InputEvent {
	const char* fieldName;
	const int numOf;
	const int layer;
	int phase;
public:
	IncFieldCtrlZ(const char* fieldName, const int numOf=1, const int layer=0):
		fieldName(fieldName),
		numOf(numOf),
		layer(layer),
		phase(0)
	{}
	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
};
class IncFieldGroup : public InputEvent {
public:
	const char* fieldName;
	const int layer;
	int phase;
	IncFieldGroup(const char* fieldName, const int layer=0):
		fieldName(fieldName),
		layer(layer),
		phase(0)
	{}
	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
};
class DecFieldGroup : public InputEvent {
public:
	const char* fieldName;
	const int layer;
	int phase;
	DecFieldGroup(const char* fieldName, const int layer=0):
		fieldName(fieldName),
		layer(layer),
		phase(0)
	{}
	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
};
class SetIndex : public InputEvent {
	const char* fieldName;
	const int layer;
	int phase;
	int target;
public:
	SetIndex(const char* fieldName, const int target, const int layer=0):
		fieldName(fieldName),
		target(target),
		layer(layer),
		phase(0)
	{}
	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
};
class DecIndex : public InputEvent {
	const char* fieldName;
	const int layer;
	int phase;
public:
	DecIndex(const char* fieldName, const int layer=0):
		fieldName(fieldName),
		layer(layer),
		phase(0)
	{}
	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
};
class IncIndex : public InputEvent {
	const char* fieldName;
	const int layer;
	int phase;
public:
	IncIndex(const char* fieldName, const int layer=0):
		fieldName(fieldName),
		layer(layer),
		phase(0)
	{}
	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
};
class DragAnimSlider : public InputEvent {
	int phase;
	const float value;
public:
	DragAnimSlider(const float value):
		value(value),
		phase(0)
	{}
	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
};
class DragAnimSelect : public InputEvent {
	int phase;
	const float firstValue;
	const float secondValue;
public:
	DragAnimSelect(const float firstValue, const float secondValue):
		firstValue(firstValue),
		secondValue(secondValue),
		phase(0)
	{}
	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
};
class MoveToDecGroupIndex : public InputEvent {
	const char* fieldName;
	const float period;
	const int layer;
public:
	MoveToDecGroupIndex(const char* fieldName, const float& period=1, const int layer=0):
		fieldName(fieldName),
		period(period),
		layer(layer)
	{}
	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
};
class MoveToIncGroupIndex : public InputEvent {
	const char* fieldName;
	const float period;
	const int layer;
public:
	MoveToIncGroupIndex(const char* fieldName, const float& period=1, const int layer=0):
		fieldName(fieldName),
		period(period),
		layer(layer)
	{}
	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
};
class MoveToNodeLable : public InputEvent {
	const int nodeIdx;
	const int charIdx;
public:
	MoveToNodeLable(const int nodeIdx, const int charIdx=0):
		nodeIdx(nodeIdx),
		charIdx(charIdx)
	{}
	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
};
class MoveToNodeAsChild : public InputEvent {
	const int nodeIdx;
	int phase;
public:
	MoveToNodeAsChild(const int nodeIdx):
		nodeIdx(nodeIdx),
		phase(0)
	{}
	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
};
class MoveToNodeAsDest : public InputEvent {
	const int nodeIdx;
public:
	MoveToNodeAsDest(const int nodeIdx):
		nodeIdx(nodeIdx)
	{}
	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
};
class MoveToNodeAsSource : public InputEvent {
	const int nodeIdx;
public:
	MoveToNodeAsSource(const int nodeIdx):
		nodeIdx(nodeIdx)
	{}
	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
};
class MoveToNodeAsSiblingAbove : public InputEvent {
	const int nodeIdx;
public:
	MoveToNodeAsSiblingAbove(const int nodeIdx):
		nodeIdx(nodeIdx)
	{}
	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
};
class MoveToNodeAsReplacement : public InputEvent {
	const int nodeIdx;
	int phase;
public:
	MoveToNodeAsReplacement(const int nodeIdx):
		nodeIdx(nodeIdx),
		phase(0)
	{}
	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
};
class MoveToNodeAsSiblingBelow : public InputEvent {
	const int nodeIdx;
public:
	MoveToNodeAsSiblingBelow(const int nodeIdx):
		nodeIdx(nodeIdx)
	{}
	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
};
class MoveToNode : public InputEvent {
	const int nodeIdx;
public:
	MoveToNode(const int nodeIdx):
		nodeIdx(nodeIdx)
	{}
	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
};
class UseNodeAsReplacement : public InputEvent {
	const int nodeLineFrom;
	const int nodeLineTo;
	int phase;
public:
	UseNodeAsReplacement(int nodeLineFrom, int nodeLineTo):
		nodeLineFrom(nodeLineFrom),
		nodeLineTo(nodeLineTo),
		phase(0)
	{}
	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
};
class CopyNodeAsChild : public InputEvent {
	const int nodeLineFrom;
	const int nodeLineTo;
	int phase;
public:
	CopyNodeAsChild(int nodeLineFrom, int nodeLineTo):
		nodeLineFrom(nodeLineFrom),
		nodeLineTo(nodeLineTo),
		phase(0)
	{}
	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
};
class CopyNodeAsSiblingBelow : public InputEvent {
	const int nodeLineFrom;
	const int nodeLineTo;
	int phase;
public:
	CopyNodeAsSiblingBelow(int nodeLineFrom, int nodeLineTo):
		nodeLineFrom(nodeLineFrom),
		nodeLineTo(nodeLineTo),
		phase(0)
	{}
	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
};
class CopyNodeAsSiblingAbove : public InputEvent {
	const int nodeLineFrom;
	const int nodeLineTo;
	int phase;
public:
	CopyNodeAsSiblingAbove(int nodeLineFrom, int nodeLineTo):
		nodeLineFrom(nodeLineFrom),
		nodeLineTo(nodeLineTo),
		phase(0)
	{}
	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
};
class Hide : public InputEvent {
	const int nodeIdx;
	int phase;
public:
	Hide(const int nodeIdx):
		nodeIdx(nodeIdx),
		phase(0)
	{}
	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
};
class UseNodeAsSource : public InputEvent {
	const int nodeLineFrom;
	const int nodeLineTo;
	int phase;
public:
	UseNodeAsSource(int nodeLineFrom, int nodeLineTo):
		nodeLineFrom(nodeLineFrom),
		nodeLineTo(nodeLineTo),
		phase(0)
	{}
	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
};
class UseNodeAsDest : public InputEvent {
	const int nodeLineFrom;
	const int nodeLineTo;
	int phase;
public:
	UseNodeAsDest(int nodeLineFrom, int nodeLineTo):
		nodeLineFrom(nodeLineFrom),
		nodeLineTo(nodeLineTo),
		phase(0)
	{}
	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
};
class MoveNodeAsDest : public InputEvent {
	const int nodeLineFrom;
	const int nodeLineTo;
	int phase;
public:
	MoveNodeAsDest(int nodeLineFrom, int nodeLineTo):
		nodeLineFrom(nodeLineFrom),
		nodeLineTo(nodeLineTo),
		phase(0)
	{}
	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
};
class ReplacePartialCode : public InputEvent {
	const int line;
	const int offset;
	const int numOf;
	const String str;
	int phase;
	int charIdx;
public:
	ReplacePartialCode(const int line, const int offset, const int numOf, const String str):
		line(line),
		offset(offset),
		numOf(numOf),
		str(str),
		phase(0),
		charIdx(0)
	{}
	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
};
class SelectNode : public InputEvent {
	const int nodeIdx;
	int phase;
public:
	SelectNode(const int nodeIdx):
		nodeIdx(nodeIdx),
		phase(0)
	{}
	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
};

class SelectNodeByName : public InputEvent {
	const char* name;
	int phase;
public:
	SelectNodeByName(const char* name):
		name(name),
		phase(0)
	{}
	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
};

class NameNode : public InputEvent {
	const String name;
	const int nodeIdx;
	int phase;
	int charIdx;
public:
	NameNode(const int nodeIdx, const String& name):
		nodeIdx(nodeIdx),
		name(name),
		phase(0),
		charIdx(0)
	{}
	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
};
class Collapse : public InputEvent {
	const int nodeIdx;
	int phase;
public:
	Collapse(const int nodeIdx):
		nodeIdx(nodeIdx),
		phase(0)
	{}
	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
};
class MoveToNodeNameAsChild : public InputEvent {
	const char* name;
public:
	MoveToNodeNameAsChild(const char* name):
		name(name)
	{}
	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
};
class ClickControlPoint : public InputEvent {
	const int pointIndex;
public:
	ClickControlPoint(const int pointIndex):
		pointIndex(pointIndex)
	{}
	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
};
class MoveToControlPoint : public InputEvent {
	const int pointIndex;
public:
	MoveToControlPoint(const int pointIndex):
		pointIndex(pointIndex)
	{}
	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
};
class TugVertex : public InputEvent {
	const int coordIndex;
	const int nodeLine;
	int phase;
	const Vec2f to;
public:
	TugVertex(const int coordIndex, const Vec2f to, const int nodeLine=-1):
		coordIndex(coordIndex),
		nodeLine(nodeLine),
		to(to),
		phase(0)
	{}
	TugVertex(const int coordIndex, const float x, const float y, const int nodeLine=-1):
		coordIndex(coordIndex),
		nodeLine(nodeLine),
		to(x,y),
		phase(0)
	{}
	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
};

class TugCP : public InputEvent {
	const int pointIndex;
	int phase;
	const Vec2f to;
public:
	TugCP(const int pointIndex, const float x, const float y):
		pointIndex(pointIndex),
		to(x,y),
		phase(0)
	{}
	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
};

class TugPivotPoint : public InputEvent {
	int phase;
	const Vec2f to;
public:
	TugPivotPoint(const float x, const float y):
		to(x,y),
		phase(0)
	{}
	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
};

class Neutral : public InputEvent {
	const int leverIdx;
	int phase;
	const Vec2f to;
public:
	Neutral(const float x, const float y):
		leverIdx(0),
		to(x,y),
		phase(0)
	{}
	Neutral(const int leverIdx, const float x, const float y):
		leverIdx(leverIdx),
		to(x,y),
		phase(0)
	{}
	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
};

//class NeutralSnap : public Neutral {
//public:
//	NeutralSnap(const int leverIdx, const float x, const float y):
//	  Neutral(leverIdx, x, y)
//	{}
//};

class Rotate : public InputEvent {
	const int leverIdx;
	int phase;
	const Vec2f to;
public:
	Rotate(const float x, const float y):
		leverIdx(0),
		to(x,y),
		phase(0)
	{}
	Rotate(const int leverIdx, const float x, const float y):
		leverIdx(leverIdx),
		to(x,y),
		phase(0)
	{}
	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
};

class Stretch : public InputEvent {
	const int leverIdx;
	int phase;
	const Vec2f to;
public:
	Stretch(const float x, const float y):
		leverIdx(0),
		to(x,y),
		phase(0)
	{}
	Stretch(const int leverIdx, const float x, const float y):
		leverIdx(leverIdx),
		to(x,y),
		phase(0)
	{}
	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
};

class Scale1d : public InputEvent {
	const int leverIdx;
	int phase;
	const Vec2f to;
public:
	Scale1d(const float x, const float y):
		leverIdx(0),
		to(x,y),
		phase(0)
	{}
	Scale1d(const int leverIdx,const float x, const float y):
		leverIdx(leverIdx),
		to(x,y),
		phase(0)
	{}
	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
};

class Scale2d : public InputEvent {
	const int leverIdx;
	int phase;
	const Vec2f to;
public:
	Scale2d(const float x, const float y):
		leverIdx(0),
		to(x,y),
		phase(0)
	{}
	Scale2d(const int leverIdx,const float x, const float y):
		leverIdx(leverIdx),
		to(x,y),
		phase(0)
	{}
	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
};

class Scale3d : public InputEvent {
	const int leverIdx;
	int phase;
	const Vec2f to;
public:
	Scale3d(const float x, const float y):
		leverIdx(0),
		to(x,y),
		phase(0)
	{}
	Scale3d(const int leverIdx,const float x, const float y):
		leverIdx(leverIdx),
		to(x,y),
		phase(0)
	{}
	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
};

class TugVertex2 : public InputEvent {//uses coordIndexIndex instead of coordIndex
	const int coordIndexIndex;
	const int nodeLine;
	int phase;
	const Vec2f to;
public:
	TugVertex2(const int coordIndexIndex, const Vec2f to, const int nodeLine=-1):
		coordIndexIndex(coordIndexIndex),
		nodeLine(nodeLine),
		to(to),
		phase(0)
	{}
	TugVertex2(const int coordIndexIndex, const float x, const float y, const int nodeLine=-1):
		coordIndexIndex(coordIndexIndex),
		nodeLine(nodeLine),
		to(x,y),
		phase(0)
	{}
	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
};

class LeverTran : public InputEvent {
	const int leverIdx;
	int phase;
	const Vec2f to;
public:
	LeverTran(const float x, const float y):
		leverIdx(0),
		to(x,y),
		phase(0)
	{}
	LeverTran(const int leverIdx,const float x, const float y):
		leverIdx(leverIdx),
		to(x,y),
		phase(0)
	{}
	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
};

class SelectVertex : public InputEvent {
	const int coordIndex;
	const int nodeLine;
	float clickTime;
public:
	SelectVertex(const int coordIndex, const int nodeLine=-1):
		coordIndex(coordIndex),
		nodeLine(nodeLine),
		clickTime(0)
	{}
	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
};
class SelectBone : public InputEvent {
	const int nodeIdx;
	float clickTime;
	int phase;
public:
	SelectBone(const int nodeIdx):
		nodeIdx(nodeIdx),
		clickTime(0),
		phase(0)
	{}
	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
};
class SelectTriangle : public InputEvent {
	const int coordIdxIdx;
	const int nodeLine;
	float clickTime;
public:
	SelectTriangle(const int coordIdxIdx, const int nodeLine=-1):
		coordIdxIdx(coordIdxIdx),
		nodeLine(nodeLine),
		clickTime(0)
	{}
	static InputEvent* decode(const String& str, int& i);
	bool doEvent();
};
class Click3d : public InputEvent {
	float clickTime;
public:
	Click3d():
		clickTime(0)
	{}
	static InputEvent* decode(const String& str, int& i);
	bool doEvent();
};
class SelectTriCornerAnd2d : public InputEvent {
	const int coordIdxIdx;
	const int nodeLine;
	float clickTime;
public:
	SelectTriCornerAnd2d(const int coordIdxIdx, const int nodeLine=-1):
		coordIdxIdx(coordIdxIdx),
		nodeLine(nodeLine),
		clickTime(0)
	{}
	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
};

class MoveToAnimRotPoint : public InputEvent{
	const int leverIdx;
public:
	MoveToAnimRotPoint(const int leverIdx):
		leverIdx(leverIdx)
	{}


	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
};

class MoveToMultiSelectRotPoint : public InputEvent{
	const int leverIdx;
public:
	MoveToMultiSelectRotPoint(const int leverIdx):
		leverIdx(leverIdx)
	{}


	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
};

class MoveToMultiSelectPivotPoint : public InputEvent {
public:
	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
};

class MoveToPivotPoint : public InputEvent {
	int nodeIdx;
public:
	MoveToPivotPoint(int nodeIdx):
	  nodeIdx(nodeIdx)
	{}

	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
};

class TugRing : public InputEvent {
	const int ringPointIndex;
	int phase;
	const Vec2f to;
public:
	TugRing(const int ringPointIndex, const Vec2f to):
		ringPointIndex(ringPointIndex),
		to(to),
		phase(0)
	{}
	TugRing(const int ringPointIndex, const float x, const float y):
		ringPointIndex(ringPointIndex),
		to(x,y),
		phase(0)
	{}
	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
};
class DragViewRot : public InputEvent {
	const Rotation to;
	static Rotation from;
	float mouseFromX;
	float mouseFromY;
	int phase;
public:
	DragViewRot(float x, float y, float z, float angle, float mouseFromX = 0, float mouseFromY = 0):
		to(x, y, z, angle),
		mouseFromX(mouseFromX),
		mouseFromY(mouseFromY),
		phase(0)
	{}
	DragViewRot(const Rotation& r, float mouseFromX = 0, float mouseFromY = 0):
		to(r),
		mouseFromX(mouseFromX),
		mouseFromY(mouseFromY),
		phase(0)
	{}
	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
};
class DragViewPos : public InputEvent {
	const Vec3f to;
	static Vec3f from;
	float mouseFromX;
	float mouseFromY;
	int phase;
public:
	DragViewPos(float x, float y, float z, float mouseFromX = 0, float mouseFromY = 0):
		to(x, y, z),
		mouseFromX(mouseFromX),
		mouseFromY(mouseFromY),
		phase(0)
	{}
	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
};

class SelNurbsLine : public InputEvent {
	int idxA;
	int idxB;
	const float ratio;
	int phase;
public:
	SelNurbsLine(int idxA, int idxB, const float ratio):
		idxA(idxA),
		idxB(idxB),
		ratio(ratio),
		phase(0)
	{}
	bool doEvent();
};

class AddCtrlPnt : public InputEvent {
	Vec2f to;
	int phase;
public:
	AddCtrlPnt(float toX, float toY):
		to(toX, toY),
		phase(0)
	{}
	bool doEvent();
};

class AddDiversion : public InputEvent {
	Vec2f to;
	int phase;
public:
	AddDiversion(float toX, float toY):
		to(toX, toY),
		phase(0)
	{}
	bool doEvent();
};

class RectVert : public InputEvent {
	Vec2f from;
	Vec2f to;
	int phase;
public:
	RectVert(float fromX, float fromY, float toX, float toY):
		from(fromX, fromY),
		to(toX, toY),
		phase(0)
	{}
	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
};

class UnRectVert : public InputEvent {
	Vec2f from;
	Vec2f to;
	int phase;
public:
	UnRectVert(float fromX, float fromY, float toX, float toY):
		from(fromX, fromY),
		to(toX, toY),
		phase(0)
	{}
	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
};

class RectPoly : public InputEvent {
	Vec2f from;
	Vec2f to;
	int phase;
public:
	RectPoly(float fromX, float fromY, float toX, float toY):
		from(fromX, fromY),
		to(toX, toY),
		phase(0)
	{}
	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
};

class MouseZoom : public InputEvent {
	const float to;
	static float from;
	float mouseFromX;
	float mouseFromY;
	int phase;
public:
	MouseZoom(float par, float mouseFromX = 0, float mouseFromY = 0):
		to(par),
		mouseFromX(mouseFromX),
		mouseFromY(mouseFromY),
		phase(0)
	{}
	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
};

class MovePointer3d : public InputEvent {
	float mouseFromX;
	float mouseFromY;
	int phase;
public:
	MovePointer3d(float mouseFromX, float mouseFromY):
		mouseFromX(mouseFromX),
		mouseFromY(mouseFromY),
		phase(0)
	{}
	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
};

class MouseWheel : public InputEvent {
	int clickSum;
	unsigned int direction;
	int click;
	const float clickPeriod;
public:
	MouseWheel(const int& i_clickSum):
		clickSum(abs(i_clickSum)),
		click(0),
		clickPeriod(.125),
		direction(120)
	{
		if(i_clickSum<0){
			direction=-120;
		}
		

	}
	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
};
class SetFieldValue : public InputEvent {
	int charIdx;
	const char* field;
	const char* value;
	int phase;
	const int layer;
public:
	SetFieldValue(const char* field, const char* value, const int layer=0):
		field(field),
		value(value),
		charIdx(0),
		layer(layer),
		phase(0)
	{}
	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
};
class TypeString : public InputEvent {
	const float charPeriod;
	int charIdx;
	String str;
	int phase;
public:
	TypeString(String str):
		str(str),
		charIdx(0),
		charPeriod(.125),
		phase(0)
	{}
	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
};
class TypeVirtKey : public InputEvent {
	const unsigned char vk;
public:
	TypeVirtKey(const unsigned char vk):
		vk(vk)
	{
	}
	bool doEvent(){
		return doTypeVirtKey(vk);
	}
};

class VirtKeyDown : public InputEvent {
	const unsigned char vk;
public:
	VirtKeyDown(const unsigned char vk):
		vk(vk)
	{
	}
	bool doEvent(){
		return doVirtKeyDown(vk);
	}
};
class VirtKeyUp : public InputEvent {
	const unsigned char vk;
public:
	VirtKeyUp(const unsigned char vk):
		vk(vk)
	{
	}
	bool doEvent(){
		return doVirtKeyUp(vk);
	}
};


class Move_CreateNewFile_Save : public InputEvent {
	const float period;
	int phase;
public:
	Move_CreateNewFile_Save(const float& period=1):
		period(period),
		phase(0)
	{
	}
	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
};
class MoveToButton : public InputEvent {
	const int id;
	const int subButtonIdx;
	const float period;
public:
	/*
	MoveToButton(const int& id, const int& subButtonIdx=0, const float& period=1):
		id(id),
		subButtonIdx(subButtonIdx),
		period(period)
	{}
	*/
	MoveToButton(void* id, const int& subButtonIdx=0, const float& period=1):
		id(reinterpret_cast<__int64>(id)),
		subButtonIdx(subButtonIdx),
		period(period)
	{}
	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
};
class ClickButton : public InputEvent {
public:
	int phase;
	int numOfActions;
	const __int64 id;
	const int subButtonIdx;
	ClickButton(void* id, const int& numOfActions=1, const int& subButtonIdx=0):
		id(reinterpret_cast<__int64>(id)),
		subButtonIdx(subButtonIdx),
		phase(0),
		numOfActions(numOfActions)
	{}
	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
	bool doClickButton(int& phase, const int& numOfActions);
};
class DecButtonX : public ClickButton {
public:
	DecButtonX(void* id, const int& numOfActions=1):
		ClickButton(id, numOfActions, 0)
	{}
};
class IncButtonX : public ClickButton {
public:
	IncButtonX(void* id, const int& numOfActions=1):
		ClickButton(id, numOfActions, 1)
	{}
};
class DecButtonY : public ClickButton {
public:
	DecButtonY(void* id, const int& numOfActions=1):
		ClickButton(id, numOfActions, 2)
	{}
};
class IncButtonY : public ClickButton {
public:
	IncButtonY(void* id, const int& numOfActions=1):
		ClickButton(id, numOfActions, 3)
	{}
};
class DecButtonZ : public ClickButton{
public:
	DecButtonZ(void* id, const int& numOfActions=1):
		ClickButton(id, numOfActions, 4)
	{}
};
class IncButtonZ : public ClickButton {
public:
	IncButtonZ(void* id, const int& numOfActions=1):
		ClickButton(id, numOfActions, 5)
	{}
};
/////
class ButtonCtrl : public ClickButton {
public:
	ButtonCtrl(void* id, const int& numOfActions, const int index):
		ClickButton(id, numOfActions, index)
	{}
	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
};
class DecButtonCtrlX : public ButtonCtrl {
public:
	DecButtonCtrlX(void* id, const int& numOfActions=1):
		ButtonCtrl(id, numOfActions, 0)
	{}
};
class IncButtonCtrlX : public ButtonCtrl {
public:
	IncButtonCtrlX(void* id, const int& numOfActions=1):
		ButtonCtrl(id, numOfActions, 1)
	{}
};
class DecButtonCtrlY : public ButtonCtrl {
public:
	DecButtonCtrlY(void* id, const int& numOfActions=1):
		ButtonCtrl(id, numOfActions, 2)
	{}
};
class IncButtonCtrlY : public ButtonCtrl {
public:
	IncButtonCtrlY(void* id, const int& numOfActions=1):
		ButtonCtrl(id, numOfActions, 3)
	{}
};
class DecButtonCtrlZ : public ButtonCtrl{
public:
	DecButtonCtrlZ(void* id, const int& numOfActions=1):
		ButtonCtrl(id, numOfActions, 4)
	{}
};
class IncButtonCtrlZ : public ButtonCtrl {
public:
	IncButtonCtrlZ(void* id, const int& numOfActions=1):
		ButtonCtrl(id, numOfActions, 5)
	{}
};

class Move_MenuItem : public InputEvent {
	const char* str;
	const float pausePeriod;
public:
	Move_MenuItem(const char* str, const float& pausePeriod=.3):
		str(str),
		pausePeriod(pausePeriod)
	{}
	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
};
class MoveToNewNode : public InputEvent{
	const int id;
public:
	MoveToNewNode(const int& id):
		id(id)
	{}
	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
};
class DragNewNodeBar : public InputEvent{
	const Vec2f position;
	int phase;
public:
	DragNewNodeBar(int x, int y):
		position(x, y),
		phase(0)
	{}
	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
};
class NewNodeAsSiblingAbove : public InputEvent{
	const int id;
	const int nodeLine;
	int phase;
public:
	NewNodeAsSiblingAbove(const int& id, const int& nodeLine):
		id(id),
		nodeLine(nodeLine),
		phase(0)
	{}
	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
};
class NewNodeAsSiblingBelow : public InputEvent{
	const int id;
	const int nodeLine;
	int phase;
public:
	NewNodeAsSiblingBelow(const int& id, const int& nodeLine):
		id(id),
		nodeLine(nodeLine),
		phase(0)
	{}
	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
};
class NewNodeAsChild : public InputEvent{
	const int id;
	const int nodeLine;
	int phase;
public:
	NewNodeAsChild(const int& id, const int& nodeLine):
		id(id),
		nodeLine(nodeLine),
		phase(0)
	{}
	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
};
class MoveNodeAsChild : public InputEvent {
	const int fromLine;
	const int toLine;
	int phase;
public:
	MoveNodeAsChild(const int fromLine, const int toLine):
		fromLine(fromLine),
		toLine(toLine),
		phase(0)
	{}
	bool doEvent();
	static InputEvent* decode(const String& str, int& i);
};

#define TugVertices(X) tugVertices(X, sizeof(X)/sizeof(GenPolyData))
#define TugVertices2(X, Y) tugVertices(X, sizeof(X)/sizeof(TugData), Y)
#define TugRings(X) tugRings(X, sizeof(X)/sizeof(TugData))
#define PatchTriangles(line, X) patchTriangles(line, X, sizeof(X)/sizeof(int))
#define PatchTriangles2(X) patchTriangles2(X, sizeof(X)/sizeof(int))
#define SelVertices(X) selVertices(X, sizeof(X)/(sizeof(int)*2))
#define SplitMove(X) splitMove(X, sizeof(X)/sizeof(GenPolyData))
#define SwapTriangles(X) swapTriangles(X, sizeof(X)/sizeof(SwapTriData))
#define DeleteTriangles(X) deleteTriangles(X, sizeof(X)/sizeof(int))
#define DeleteTriangles2(X, line) deleteTriangles(X, sizeof(X)/sizeof(int), line)
#define JoinVertices(X) joinVertices(X, sizeof(X)/sizeof(int))
#define JoinVertices2(X) joinVertices2(X, sizeof(X)/sizeof(int))
#define PaintTriangles(X, line) paintTriangles(X, sizeof(X)/sizeof(int), line)

#define ProjectVertices(X, line) projectVertices(X, sizeof(X)/sizeof(int), line)
void nameOrLineNumber(const int line, String& str);
