/* Seamless3d

Copyright (C) 2001-2008 Graham John Perrett (AKA thyme) thyme@seamless3d.com www.seamless3d.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this
software and associated documentation files (the "Software"), to deal in the Software
without restriction, including without limitation the rights to use, copy, modify,
merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to the following
conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT
OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.
*/

#define _WIN32_WINNT 0x0410


#include <windows.h>
//#include <process.h>
//#include <shlobj.h>

#include "seamless3d.h"
#include "resource.h"
#include "gui.h"
#include "scene_tree.h"
#include "techuelife_triangle_editor.h"

#include"robot_aid.h"

int convertLabelToInt(const String label);


InputEvent* InputEvent::readDemoEvent(const String& str, int& i){
	str.stb(i);
	if(!str.cmpCppWord("new",i)){
		throw;
	}
	str.stb(i);
	switch(str[i++]){
		case 'C':{
			if(str.cmpCppWord("lick",i)){
				return Click::decode(str, i);
			}
			else if(str.cmpCppWord("lickButton",i)){
				return ClickButton::decode(str, i);
			}
			else if(str.cmpCppWord("lickControlPoint",i)){
				return ClickControlPoint::decode(str, i);
			}
		}	break;
		case 'D':{
			if(str.cmpCppWord("ecButtonX",i)){
				void * id=getFirstButtonParam(str, i);
				int numOfActions=get2ndButtonParam(str, i);
				return new DecButtonX(id, numOfActions);
			}
			else if(str.cmpCppWord("ecButtonY",i)){
				void * id=getFirstButtonParam(str, i);
				int numOfActions=get2ndButtonParam(str, i);
				return new DecButtonY(id, numOfActions);
			}
			else if(str.cmpCppWord("ecButtonZ",i)){
				void * id=getFirstButtonParam(str, i);
				int numOfActions=get2ndButtonParam(str, i);
				return new DecButtonZ(id, numOfActions);
			}
			else if(str.cmpCppWord("emoSound",i)){
				return DemoSound::decode(str, i);
			}
			else if(str.cmpCppWord("ragViewRot",i)){
				return DragViewRot::decode(str, i);
			}
			else if(str.cmpCppWord("ragViewPos",i)){
				return DragViewPos::decode(str, i);
			}
			else if(str.cmpCppWord("oubleClick",i)){
				return DoubleClick::decode(str, i);
			}
			else if(str.cmpCppWord("ecFieldGroup",i)){
				return DecFieldGroup::decode(str, i);
			}
			break;
		}	
		case 'I':{
			if(str.cmpCppWord("ncButtonX",i)){
				void * id=getFirstButtonParam(str, i);
				int numOfActions=get2ndButtonParam(str, i);
				return new IncButtonX(id, numOfActions);
			}
			else if(str.cmpCppWord("ncButtonY",i)){
				void * id=getFirstButtonParam(str, i);
				int numOfActions=get2ndButtonParam(str, i);
				return new IncButtonY(id, numOfActions);
			}
			else if(str.cmpCppWord("ncButtonZ",i)){
				void * id=getFirstButtonParam(str, i);
				int numOfActions=get2ndButtonParam(str, i);
				return new IncButtonZ(id, numOfActions);
			}
			else if(str.cmpCppWord("ncFieldGroup",i)){
				return IncFieldGroup::decode(str, i);
			}
		}
 		case 'M':{
			if(str.cmpCppWord("oveRelWindow",i)){
				return MoveRelWindow::decode(str, i);
			}
			else if(str.cmpCppWord("oveToButton",i)){
				return MoveToButton::decode(str, i);
			}
			else if(str.cmpCppWord("oveToControlPoint",i)){
				return MoveToControlPoint::decode(str, i);
			}
			else if(str.cmpCppWord("ove_CreateNewFile_Save",i)){
				return Move_CreateNewFile_Save::decode(str, i);
			}
			else if(str.cmpCppWord("ove3d",i)){
				return Move3d::decode(str, i);
			}
			else if(str.cmpCppWord("oveToNode",i)){
				return MoveToNode::decode(str, i);
			}
			else if(str.cmpCppWord("ouseZoom",i)){
				return MouseZoom::decode(str, i);
			}
			break;
		}
		case 'S':{
			if(str.cmpCppWord("electNode",i)){
				return SelectNode::decode(str, i);
			}
			else if(str.cmpCppWord("electTriangle",i)){
				return SelectTriangle::decode(str, i);
			}
			else if(str.cmpCppWord("electVertex",i)){
				return SelectVertex::decode(str, i);
			}
			else if(str.cmpCppWord("ection",i)){
				return Section::decode(str, i);
			}
			break;
		}
		case 'T':{
			if(str.cmpCppWord("ugVertex",i)){
				return TugVertex::decode(str, i);
			}
			else if(str.cmpCppWord("ypeString",i)){
				return TypeString::decode(str, i);
			}
			break;
		}
 		case 'W':{
			if(str.cmpCppWord("ait",i)){
				return Wait::decode(str, i);
			}
			break;
		}
	}
	throw;	
}
InputEvent* Wait::decode(const String& str, int& i){
	if(str[i++]!='('){
		throw;
	}
	str.stb(i);
	return new Wait(getFloat(str, i));
}

void* ToolBarWin::decodeVButtonFunc(const String& str, int& i){
	M_Button& m_button = *(this->m_button);
	for(int k=0; k<m_button.length; k++){
		Button* button = m_button[k];
		if(SingleButton* singleButton = dynamic_cast<SingleButton*>(button)){
			if(str.cmpCppWord(singleButton->name,i)){
				return singleButton->function;
			}
		}
	}
	return 0;
}

InputEvent* MoveToButton::decode(const String& str, int& i){
	if(str[i++]!='('){
		throw;
	}
	str.stb(i);
	void* buttonFunc = g_tb->decodeVButtonFunc(str, i);
	if(buttonFunc == 0){
		buttonFunc = g_tb2->decodeVButtonFunc(str, i);
	}
	return new MoveToButton(buttonFunc);
}

InputEvent* ClickButton::decode(const String& str, int& i){
	if(str[i++]!='('){
		throw;
	}
	str.stb(i);
	void* buttonFunc = g_tb->decodeVButtonFunc(str, i);
	if(buttonFunc == 0){
		buttonFunc = g_tb2->decodeVButtonFunc(str, i);
	}
	return new ClickButton(buttonFunc);
}

InputEvent* TypeString::decode(const String& str, int& i){
	if(str[i++]!='('){
		throw;
	}
	str.stb(i);
	String str2;
	str2.getField(str, i);
	return new TypeString(str2);
}
InputEvent* IncFieldGroup::decode(const String& str, int& i){
	if(str[i++]!='('){
		throw;
	}
	String str2;
	str2.getField(str, i);
	for(int i = 0; i < str2.length; i++){
		dynamChar += str2[i];
	}
	dynamChar += 0;
	return new IncFieldGroup(0);
}
InputEvent* DecFieldGroup::decode(const String& str, int& i){
	if(str[i++]!='('){
		throw;
	}
	String str2;
	str2.getField(str, i);
	for(int i = 0; i < str2.length; i++){
		dynamChar += str2[i];
	}
	dynamChar += 0;
	return new DecFieldGroup(0);
}
InputEvent* Section::decode(const String& str, int& i){
	if(str[i++]!='('){
		throw;
	}
	String str2;
	str2.getField(str, i);
	for(int i = 0; i < str2.length; i++){
		dynamChar += str2[i];
	}
	dynamChar += 0;
	return new Section(0);
}
InputEvent* DemoSound::decode(const String& str, int& i){
	if(str[i++]!='('){
		throw;
	}
	String str2;
	str2.getField(str, i);
	for(int i = 0; i < str2.length; i++){
		dynamChar += str2[i];
	}
	dynamChar += 0;
	return new DemoSound(0);
}
InputEvent* MoveToControlPoint::decode(const String& str, int& i){
	if(str[i++]!='('){
		throw;
	}
	int coordIndex=getInt(str,i);
	str.stb(i);
	return new MoveToControlPoint(coordIndex);
}
InputEvent* ClickControlPoint::decode(const String& str, int& i){
	if(str[i++]!='('){
		throw;
	}
	int coordIndex=getInt(str,i);
	str.stb(i);
	return new ClickControlPoint(coordIndex);
}
InputEvent* TugVertex::decode(const String& str, int& i){
	if(str[i++]!='('){
		throw;
	}
	int coordIndex=getInt(str,i);
	str.stb(i);
	float x=getFloat(str,i);
	str.stb(i);
	float y=getFloat(str,i);
	str.stb(i);
	return new TugVertex(coordIndex, x, y, convertLabelToInt(str.extractCppLabel(i)));
}
InputEvent* SelectVertex::decode(const String& str, int& i){
	if(str[i++]!='('){
		throw;
	}
	int coordIndex=getInt(str,i);
	str.stb(i);
	return new SelectVertex(coordIndex, convertLabelToInt(str.extractCppLabel(i)));
}
InputEvent* MoveToNode::decode(const String& str, int& i){
	if(str[i++]!='('){
		throw;
	}
	str.stb(i);
	return new MoveToNode(getInt(str,i));
}
InputEvent* SelectNode::decode(const String& str, int& i){
	if(str[i++]!='('){
		throw;
	}
	str.stb(i);
	return new SelectNode(getInt(str,i));
}
InputEvent* SelectTriangle::decode(const String& str, int& i){
	if(str[i++]!='('){
		throw;
	}
	int coordIndex=getInt(str,i);
	str.stb(i);
	return new SelectTriangle(coordIndex, convertLabelToInt(str.extractCppLabel(i)));
}
InputEvent* MoveRelWindow::decode(const String& str, int& i){
		if(str[i++]!='('){
		throw;
	}
	str.stb(i);
	int x = getInt(str,i);
	str.stb(i);
	int y = getInt(str,i);
	return new MoveRelWindow(x,y);
}
InputEvent* MouseZoom::decode(const String& str, int& i){
	str.stb(i);
	if(str[i++]!='('){
		throw;
	}
	str.stb(i);
	float par = getFloat(str, i);
	float mouseFromX = 0;
	float mouseFromY = 0;

	str.stb(i);
	if(str[i] != ')'){
		mouseFromX = getFloat(str, i);
		str.stb(i);
		mouseFromY = getFloat(str, i);
	}
	return new MouseZoom(par, mouseFromX, mouseFromY);
}
InputEvent* DragViewRot::decode(const String& str, int& i){
	if(str[i++]!='('){
		throw;
	}
	str.stb(i);
	float x = getFloat(str,i);
	str.stb(i);
	float y = getFloat(str,i);
	str.stb(i);
	float z = getFloat(str,i);
	str.stb(i);
	float angle = getFloat(str,i);
	float mouseFromX = 0;
	float mouseFromY = 0;
	str.stb(i);
	if(str[i] != ')'){
		mouseFromX = getFloat(str, i);
		str.stb(i);
		mouseFromY = getFloat(str, i);
	}
	return new DragViewRot(x, y, z, angle, mouseFromX, mouseFromY);
}
InputEvent* DragViewPos::decode(const String& str, int& i){
	if(str[i++]!='('){
		throw;
	}
	str.stb(i);
	float x = getFloat(str,i);
	str.stb(i);
	float y = getFloat(str,i);
	str.stb(i);
	float z = getFloat(str,i);
	float mouseFromX = 0;
	float mouseFromY = 0;
	str.stb(i);
	if(str[i] != ')'){
		mouseFromX = getFloat(str, i);
		str.stb(i);
		mouseFromY = getFloat(str, i);
	}
	return new DragViewPos(x, y, z, mouseFromX, mouseFromY);
}
InputEvent* Move3d::decode(const String& str, int& i){
	if(str[i++]!='('){
		throw;
	}
	str.stb(i);
	float x = getFloat(str,i);
	str.stb(i);
	float y = getFloat(str,i);
	str.stb(i);
	return new Move3d(x, y);
}
InputEvent* Click::decode(const String& str, int& i){
	if(str[i++]!='('){
		throw;
	}
	return new Click();
};
InputEvent* DoubleClick::decode(const String& str, int& i){
	if(str[i++]!='('){
		throw;
	}
	return new DoubleClick();
};
void InputEvent::fillList(const String& str, int& i){
	while(str[i]!='}'){
		list+=readDemoEvent(str, i);
		str.stb(i);
		if(str[i++]!=')'){
			throw;
		};
		str.stb(i);
	}
	i++;
}

int InputEvent::get2ndButtonParam(const String& str, int& i){
	str.stb(i);
	if(str[i]==')'){
		return 1;
	}
	return getInt(str,i);
}
void* InputEvent::getFirstButtonParam(const String& str, int& i){
	str.stb(i);
	if(str[i++]!='('){
		throw;
	}
	if(str.cmpCppWord("joinVertexKeepBothColors",i)){
		return joinVertexKeepBothColors;
	}
	if(str.cmpCppWord("remember",i)){
		return remember;
	}
	if(str.cmpCppWord("invertBackgroundColor",i)){
		return invertBackgroundColor;
	}
	if(str.cmpCppWord("quickTriangle",i)){
		return quickTriangle;
	}
	if(str.cmpCppWord("swapDivide",i)){
		return swapDivide;
	}
	if(str.cmpCppWord("toggleWireFrame",i)){
		return toggleWireFrame;
	}
	if(str.cmpCppWord("viewOrientation",i)){
		return viewOrientation;
	}
	if(str.cmpCppWord("viewOrientation90",i)){
		return viewOrientation90;
	}
	throw;
}
InputEvent* Move_CreateNewFile_Save::decode(const String& str, int& i){
	str.stb(i);
	if(str[i++]!='('){
		throw;
	}
	return new Move_CreateNewFile_Save();
}
