/* Seamless3d

Copyright (C) 2001-2017 Graham John Perrett (AKA thyme) thyme@seamless3d.com www.seamless3d.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this
software and associated documentation files (the "Software"), to deal in the Software
without restriction, including without limitation the rights to use, copy, modify,
merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to the following
conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT
OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.
*/

#include "seamless3d.h"
M_Float Stem::bandDist;


Seamless* getSeamless2();
void addRelevantPart(M_Art& relevant, Art*& part);
bool commonIllegalImmediateRef(const M_SArt& m_art, const Art* part, M_Art& accumOthers);

void Stem::convertCoordIndexToTriangle(Art* part, int h, int c, int s){
	M_Triangle& triangle = part->triangle;
	const M_Int& rc = this->rc;
	int ia = 0;
	int ib = 0;
	int ic = 0;

	int ta = 0;
	int tb = 0;
	int tc = 0;
	for(int ci=0; ci < coordIndex.length; ){
		
		ia = rcCoord(ci++, ta);
		ib = rcCoord(ci++, tb);
		ic = rcCoord(ci++, tc);

		triangle[s++] = S_Triangle(part,ia+h, ta+c, part,ib+h,tb+c, part,ic+h,tc+c);
	}
}


void sweepColorY(const int s, const M_Triangle& triangle, M_Vec3f& coord, M_Color3f& color, const Color3f beginColor,const Color3f endColor, float beginOpacity, float endOpacity, float beginHeight, float endHeight){
	float opacityDiff=endOpacity-beginOpacity;
	float rColorDiff=endColor.r-beginColor.r;
	float gColorDiff=endColor.g-beginColor.g;
	float bColorDiff=endColor.b-beginColor.b;
	float heightDiff=endHeight-beginHeight;
	for(int i = triangle.getTriangleLength(); --i >= s;){
		//if(triangle.refPart(i)==part){
		int ci=triangle.refTexIndex(i);
		Vec3f vec = coord[triangle.refIndex(i)];
		float length = vec.y;
		if(length >= beginHeight && length <= endHeight){
			
			float frac = (length-beginHeight)/heightDiff;
			float opacity = (beginOpacity+frac*opacityDiff);
			float clarity = 1.0 - opacity;

			color[ci]=Color3f(
				color[ci].r*clarity+opacity*(beginColor.r+frac*rColorDiff),
				color[ci].g*clarity+opacity*(beginColor.g+frac*gColorDiff),
				color[ci].b*clarity+opacity*(beginColor.b+frac*bColorDiff)
				);
		}
	}
}

void sweepColorSpherical(const int s, const M_Triangle& triangle, M_Vec3f& coord, M_Color3f& color, const Color3f beginColor,const Color3f endColor, float beginOpacity, float endOpacity, float beginHeight, float endHeight){

	float opacityDiff=endOpacity-beginOpacity;
	float rColorDiff=endColor.r-beginColor.r;
	float gColorDiff=endColor.g-beginColor.g;
	float bColorDiff=endColor.b-beginColor.b;
	float heightDiff=endHeight-beginHeight;
	for(int i=triangle.length*3;--i>=0;){
		int ci=triangle.refTexIndex(i);
		Art* prt = triangle.refPart(i);
		Vec3f vec=prt->matrix * prt->coord[triangle.refIndex(i)];

		float length = vec.length();
		if(length >= beginHeight && length <= endHeight){
			float frac = (length-beginHeight)/heightDiff;
			float opacity = (beginOpacity+frac*opacityDiff);
			float clarity = 1.0-opacity;
			color[ci] = Color3f(
				color[ci].r*clarity+opacity*(beginColor.r+frac*rColorDiff),
				color[ci].g*clarity+opacity*(beginColor.g+frac*gColorDiff),
				color[ci].b*clarity+opacity*(beginColor.b+frac*bColorDiff)
			);
		}
	}
}

void ColorSweep::doColorForStem(M_Triangle& m_triangle, M_Vec3f& coord, M_Color3f& eColor, const int h, const int c, const int s){
	if(color.length == 0){
		return;
	}
	float posFrom=0;
	float posTo=distance.refValue()[0];
	int s3 = s * 3;

	for(int i=0;;){
		if(straight){
			sweepColorY(s3, m_triangle, coord, eColor, color[i], color[i+1], opacity[i], opacity[i+1], posFrom, posTo);
		}
		else if(spherical){
			sweepColorSpherical(s3, m_triangle, coord, eColor, color[i], color[i+1], opacity[i], opacity[i+1], posFrom, posTo);
		}

		if(i>=distance.length-1)break;


		posFrom = posTo;
		i++;
		posTo += distance[i];

	}
}

void doSlide(const Vec2f& slide, M_Vec3f& coord, const int& h){
    if(slide != Vec2f(0,0)){
        for(int i = coord.length; --i>=h;){
            Vec3f v = coord[i];
            v.x += slide.x;
            v.y += slide.y;
            coord[i] = v;
        }
    }
}

inline void genIndices(Art*& part, M_Triangle& triangle, const int& tiOffset, const int& coordLen, int& numStaves, int& numRings, int bottomIsPoint, int topIsPoint, const int& h, const int& s){
    int i,m,j;
    triangle.length = s + (numStaves-1)*(numRings-1)*2+(numStaves-1)*(bottomIsPoint + topIsPoint);

    int b = s;
    // Generate triangles for bottom cap
    if(bottomIsPoint){
        for(i = 1; i < numStaves; i++,b++ ){
            triangle[b].ta=tiOffset+(triangle[b].ia = h);
            triangle[b].tb=tiOffset+(triangle[b].ib = h + i + 1);
            triangle[b].tc=tiOffset+(triangle[b].ic = h + i);

            triangle[b].pa = part;
            triangle[b].pb = part;
            triangle[b].pc = part;
        }
    }
    // Generate triangles for top cap.
    if(topIsPoint){
        int t = triangle.length - (numStaves-1);
        for(i = coordLen-numStaves-topIsPoint; i < coordLen-1-topIsPoint; i++, t++ ){
            triangle[t].ta=tiOffset+(triangle[t].ia = coordLen-1);
            triangle[t].tb=tiOffset+(triangle[t].ib = i);
            triangle[t].tc=tiOffset+(triangle[t].ic = i + 1);

            triangle[t].pa = part;
            triangle[t].pb = part;
            triangle[t].pc = part;
        }
    }
    // Generate triangles for the rings
    for (m = h + bottomIsPoint; b<triangle.length-(topIsPoint*(numStaves-1)); m += numStaves){
        for(j = 0;j < numStaves-1;j++){
            triangle[b].ta=tiOffset+(triangle[b].ia = m + j + 1);
            triangle[b].tb=tiOffset+(triangle[b].ib = m + j + numStaves);
            triangle[b].tc=tiOffset+(triangle[b].ic = m + j);
            triangle[b].pa = part;
            triangle[b].pb = part;
            triangle[b].pc = part;
            b++;
            triangle[b].ta=tiOffset+(triangle[b].ia = m + j + 1);
            triangle[b].tb=tiOffset+(triangle[b].ib = m + j + numStaves + 1);
            triangle[b].tc=tiOffset+(triangle[b].ic = m + j + numStaves);
            triangle[b].pa = part;
            triangle[b].pb = part;
            triangle[b].pc = part;
            b++;
        }
    }
}
inline void genIndices(Art*& part, M_Triangle& triangle, const int& coordLen, int& numStaves, int& numRings, int bottomIsPoint, int topIsPoint, const int& h, const int& s){
    int i,m,j;
    triangle.length = s + (numStaves-1)*(numRings-1)*2+(numStaves-1)*(bottomIsPoint + topIsPoint);

    int b = s;
    // Generate triangles for bottom cap
    if(bottomIsPoint){
        for(i = 1; i < numStaves; i++,b++ ){
            (triangle[b].ia = h);
            (triangle[b].ib = h + i + 1);
            (triangle[b].ic = h + i);

            triangle[b].pa = part;
            triangle[b].pb = part;
            triangle[b].pc = part;
        }
    }
    // Generate triangles for top cap.
    if(topIsPoint){
        int t = triangle.length - (numStaves-1);
        for(i=coordLen-numStaves-topIsPoint; i < coordLen-1-topIsPoint; i++, t++ ){
            (triangle[t].ia = coordLen-1);
            (triangle[t].ib = i);
            (triangle[t].ic = i + 1);

            triangle[t].pa = part;
            triangle[t].pb = part;
            triangle[t].pc = part;
        }
    }
    // Generate triangles for the rings
    for (m = h + bottomIsPoint; b<triangle.length-(topIsPoint*(numStaves-1)); m += numStaves){
        for(j = 0;j < numStaves-1;j++){
            (triangle[b].ia = m + j + 1);
            (triangle[b].ib = m + j + numStaves);
            (triangle[b].ic = m + j);
            triangle[b].pa = part;
            triangle[b].pb = part;
            triangle[b].pc = part;
            b++;
            (triangle[b].ia = m + j + 1);
            (triangle[b].ib = m + j + numStaves + 1);
            (triangle[b].ic = m + j + numStaves);
            triangle[b].pa = part;
            triangle[b].pb = part;
            triangle[b].pc = part;
            b++;
        }
    }
}
void genIndicesJoinEnds(Art*& part, M_Triangle& triangle, const int& coordLen, int& numStaves, int& numRings, int bottomIsPoint, int topIsPoint, const int& h, const int& s){
    int i,m,j;
    triangle.length = s + (numStaves)*(numRings-1)*2+(numStaves)*(bottomIsPoint + topIsPoint);

    int b = s;
    // Generate triangles for bottom cap
    if(bottomIsPoint){
        for(i = 1; i < numStaves; i++,b++ ){
            triangle[b].ia = h;
            triangle[b].ib = h + i + 1;
            triangle[b].ic = h + i;

            triangle[b].pa = part;
            triangle[b].pb = part;
            triangle[b].pc = part;
        }
        triangle[b].ia = h;
        triangle[b].ib = h + i + 1 - numStaves;
        triangle[b].ic = h + i;

        triangle[b].pa = part;
        triangle[b].pb = part;
        triangle[b].pc = part;
        b++;
    }
    // Generate triangles for top cap.
    if(topIsPoint){
        int tl=triangle.length;
        int t = tl - (numStaves);
        tl--;
        for(i=coordLen-numStaves-topIsPoint; t < tl; i++, t++ ){
            triangle[t].ia = coordLen-1;
            triangle[t].ib = i;
            triangle[t].ic = i + 1;

            triangle[t].pa = part;
            triangle[t].pb = part;
            triangle[t].pc = part;
        }
        triangle[t].ia = coordLen-1;
        triangle[t].ib = i;
        triangle[t].ic = i + 1 - numStaves;

        triangle[t].pa = part;
        triangle[t].pb = part;
        triangle[t].pc = part;
    }
    // Generate triangles for the rings
    for (m = h + bottomIsPoint; b<triangle.length-(topIsPoint*(numStaves)); m += numStaves){
        for(j = 0;j < numStaves-1;j++){
            triangle[b].ia = m + j + 1;
            triangle[b].ib = m + j + numStaves;
            triangle[b].ic = m + j;
            triangle[b].pa = part;
            triangle[b].pb = part;
            triangle[b].pc = part;
            b++;
            triangle[b].ia = m + j + 1;
            triangle[b].ib = m + j + numStaves + 1;
            triangle[b].ic = m + j + numStaves;
            triangle[b].pa = part;
            triangle[b].pb = part;
            triangle[b].pc = part;
            b++;
        }
        triangle[b].ia = m + j + 1 - numStaves;
        triangle[b].ib = m + j + numStaves;
        triangle[b].ic = m + j;
        triangle[b].pa = part;
        triangle[b].pb = part;
        triangle[b].pc = part;
        b++;
        triangle[b].ia = m + j + 1 - numStaves;
        triangle[b].ib = m + j + numStaves + 1 - numStaves;
        triangle[b].ic = m + j + numStaves;
        triangle[b].pa = part;
        triangle[b].pb = part;
        triangle[b].pc = part;
        b++;
    }
}
void genIndicesJoinEnds(Art*& partA, Art*& partB, M_Triangle& triangle, const int& staves, const int& c){
    int m = partA->coord.length - staves;
    int o = partB->coord.length - staves;
    if(partA==partB){
        m -= staves;
    }

    int b = triangle.length;
    triangle.length+=staves*2;

    int j = 0;
    for(; j < staves-1; j++){
        triangle[b].ia = m + j + 1;
        triangle[b].ib = o + j;
        triangle[b].ic = m + j;

        triangle[b].ta = c + j + 1 - staves;
        triangle[b].tb = c + j;
        triangle[b].tc = c + j - staves;

        triangle[b].pa = partA;
        triangle[b].pb = partB;
        triangle[b].pc = partA;
        b++;
        triangle[b].ia = m + j + 1;
        triangle[b].ib = o + j + 1;
        triangle[b].ic = o + j;

        triangle[b].ta = c + j + 1 - staves;
        triangle[b].tb = c + j + 1;
        triangle[b].tc = c + j;

        triangle[b].pa = partA;
        triangle[b].pb = partB;
        triangle[b].pc = partB;
        b++;
    }
    triangle[b].ia = m;
    triangle[b].ib = o + j;
    triangle[b].ic = m + j;

    triangle[b].ta = c - staves;
    triangle[b].tb = c + j;
    triangle[b].tc = c + j - staves;

    triangle[b].pa = partA;
    triangle[b].pb = partB;
    triangle[b].pc = partA;
    b++;
    triangle[b].ia = m;
    triangle[b].ib = o;
    triangle[b].ic = o + j;

    triangle[b].ta = c - staves;
    triangle[b].tb = c;
    triangle[b].tc = c + j;

    triangle[b].pa = partA;
    triangle[b].pb = partB;
    triangle[b].pc = partB;
}
void genIndicesJoinEndsBeginSP(Art*& partA, Art*& partB, M_Triangle& triangle, const int& staves, const int& c){
    int m = partA->coord.length - staves;
    int o = partB->coord.length - staves;
    if(partA==partB){
        m -= staves;
    }

    int b = triangle.length;
    triangle.length+=staves;

    int j = 0;
    for(; j < staves-1; j++){
        triangle[b].ia = o - 1;
        triangle[b].ib = o + j + 1;
        triangle[b].ic = o + j;

        triangle[b].ta = c - 1;
        triangle[b].tb = c + j + 1;
        triangle[b].tc = c + j;

        triangle[b].pa = partA;
        triangle[b].pb = partB;
        triangle[b].pc = partB;
        b++;
    }
    triangle[b].ia = o - 1;
    triangle[b].ib = o;
    triangle[b].ic = o + j;

    triangle[b].ta = c - 1;
    triangle[b].tb = c;
    triangle[b].tc = c + j;

    triangle[b].pa = partA;
    triangle[b].pb = partB;
    triangle[b].pc = partB;
}
void genIndicesJoinEndsEndSP(Art*& partA, Art*& partB, M_Triangle& triangle, const int& staves, const int& c){
    int m = partA->coord.length - staves;
    int o = partB->coord.length - 1;
    if(partA==partB){
        m -= 1;
    }

    int b = triangle.length;
    triangle.length+=staves;

    int j = 0;
    for(; j < staves-1; j++){
        triangle[b].ia = m + j + 1;
        triangle[b].ib = o;
        triangle[b].ic = m + j;

        triangle[b].ta = c + j + 1 - staves;
        triangle[b].tb = c;
        triangle[b].tc = c + j - staves;

        triangle[b].pa = partA;
        triangle[b].pb = partB;
        triangle[b].pc = partA;
        b++;
    }
    triangle[b].ia = m;
    triangle[b].ib = o ;
    triangle[b].ic = m + j;

    triangle[b].ta = c - staves;
    triangle[b].tb = c;
    triangle[b].tc = c + j - staves;

    triangle[b].pa = partA;
    triangle[b].pb = partB;
    triangle[b].pc = partA;
}

inline void genIndicesTexJoinEnds(Art*& part, M_Triangle& triangle, const int& coordLen, int& numStaves, int& numRings, int bottomIsPoint, int topIsPoint, const int&, const int&){
    int i,m,j;
    triangle.length=(numStaves)*(numRings-1)*2+(numStaves)*(bottomIsPoint + topIsPoint);
    const M_Vec3f& coord=part->coord;
    M_Int& rc=part->rc;
    rc.length=numRings;
    int cl=coord.length;

    int b = 0;
    // Generate triangles for bottom cap
    if(bottomIsPoint){
        for(i = 1; i < numStaves; i++,b++ ){
            triangle[b].ia = 0;
            triangle[b].ib = i + 1;
            triangle[b].ic = i;

            triangle[b].pa = part;
            triangle[b].pb = part;
            triangle[b].pc = part;
        }
        triangle[b].ia = 0;
        triangle[b].ib = cl;
        triangle[b].ic = i;

        triangle[b].pa = part;
        triangle[b].pb = part;
        triangle[b].pc = part;
        b++;
    }
    // Generate triangles for top cap.
    if(topIsPoint){
        int tl=triangle.length;
        int t = tl - (numStaves);
        tl--;
        for(i=coordLen-numStaves-topIsPoint; t < tl; i++, t++ ){
            triangle[t].ia = coordLen-1;
            triangle[t].ib = i;
            triangle[t].ic = i + 1;

            triangle[t].pa = part;
            triangle[t].pb = part;
            triangle[t].pc = part;
        }
        triangle[t].ia = coordLen-1;
        triangle[t].ib = i;
        triangle[t].ic = coord.length+rc.length-1;

        triangle[t].pa = part;
        triangle[t].pb = part;
        triangle[t].pc = part;
    }
    // Generate triangles for the rings
    for (m = bottomIsPoint; b<triangle.length-(topIsPoint*(numStaves)); m += numStaves){
        for(j = 0;j < numStaves-1;j++){
            triangle[b].ia = m + j + 1;
            triangle[b].ib = m + j + numStaves;
            triangle[b].ic = m + j;
            triangle[b].pa = part;
            triangle[b].pb = part;
            triangle[b].pc = part;
            b++;
            triangle[b].ia = m + j + 1;
            triangle[b].ib = m + j + numStaves + 1;
            triangle[b].ic = m + j + numStaves;
            triangle[b].pa = part;
            triangle[b].pb = part;
            triangle[b].pc = part;
            b++;
        }
        triangle[b].ia = cl;
        triangle[b].ib = m + j + numStaves;
        triangle[b].ic = m + j;
        triangle[b].pa = part;
        triangle[b].pb = part;
        triangle[b].pc = part;
        b++;
        triangle[b].ia = cl;
        triangle[b].ib = ++cl;
        triangle[b].ic = m + j + numStaves;
        triangle[b].pa = part;
        triangle[b].pb = part;
        triangle[b].pc = part;
        b++;
    }
    for (int g=0; g<rc.length; g++ ){
        rc[g]= g*numStaves+bottomIsPoint;
    }
}

inline void addVerticesToGrid(M_Vec3f& coord, const bool& xCentred, const float& xDistance, const M_Float& bandDist, const int& staves, int index, const int& lastRing, const int& h){

    float xTileDistance = xDistance/(staves-1);
    float xTrans = 0;

    if(xCentred){
        xTrans = xDistance / 2.0;
    }

    for(int r = index; r < lastRing; r++){
        for(int s = 0; s < staves; s++, index++){
            coord[h + index] = Vec3f(xTileDistance * s - xTrans, bandDist[r], 0);
        }
    }
}

inline void addVerticesToGridJoinEnds(M_Vec3f& coord, const bool& xCentred, const float& xDistance, const M_Float& bandDist, const int& staves, int index, const int& lastRing, const int& h){

    float xTileDistance=xDistance/staves;//only line different to addVerticesToGrid
    float xTrans=0;

    if(xCentred){
        xTrans=xDistance/2.0;
    }

    for(int r=index; r<lastRing; r++){
        for(int s=0; s<staves; s++, index++){
            coord[h+index] = Vec3f(xTileDistance*s-xTrans, bandDist[r], 0);
        }
    }
}


void genFlat(
        Art* part, M_Vec3f& coord, M_Triangle& triangle, const int& tiOffset,
        const bool& xCentred, const float xDistance, const M_Float &bandDist, int staves, bool beginSP, bool endSP, const int h, const int s)
{
    int numRings=bandDist.length;
    int lastRing=numRings;
    int index=0;

    int coordLen=h+staves*(numRings);
    if(beginSP){
        coordLen+=1-staves;
    }
    if(endSP){
        coordLen+=1-staves;
    }
    coord.length=coordLen;

    if(beginSP){
        coord[h]=Vec3f(0, bandDist[0], 0);
        numRings--;
        index++;
    }
    if(endSP){
        coord[coordLen-1]=Vec3f(0, bandDist[bandDist.length-1], 0);
        numRings--;
        lastRing--;
    }
    addVerticesToGrid(coord, xCentred, xDistance, bandDist, staves, index, lastRing, h);
    genIndices(part, triangle, tiOffset, coord.length, staves, numRings, (int)beginSP, (int)endSP, h, s);
}




float Stem::getYDistance()const{
    const M_Float& yDistance=this->yDistance;
    float distanceSum=0;
    int i;
    for(i=0;i<yDistance.length;i++){
        distanceSum+=yDistance[i];
    }
    return distanceSum;
}
void Stem::convertYDistanceToHeight(){
    if(staves.length > 1){
        return;
    }
    M_Float& height = this->height;
    height.length=0;
    if(yDistance.length){
        height+=0;
        float h = 0;
        for(int i = 0; i < yDistance.length; i++){
            h+=yDistance[i];
            height+=h;
        }
    }
}
void Stem::YDistance::operator=(const M_Float& par){
    refValue()=par;
    home.convertYDistanceToHeight();
}
void Stem::YDistance::resize(int newLen){
    refValue().length=newLen;
    home.convertYDistanceToHeight();
}
void Stem::processFields(){
    M_Float& yDistance = this->yDistance;
    M_Float& height = this->height;

    if(yDistance.length>height.length){
        convertYDistanceToHeight();
    }
    if(yDistance.length==0&&bands.length>0){
        sync.refValue()=true;
    }
    if(staves.length == 1){
        yDistance.length=0;
        for(int i = 0; i < height.length-1; i++){
            yDistance+=height[i+1]-height[i];
        }
    }
}

void Stem::Cells::operator=(const Vec2f& par){
    refValue() = par;
    int uCells = par.x;
    int vCells = par.y;

    M_Int& bands = home.bands.refValue();
    M_Int& staves = home.staves.refValue();
    M_Float& xDistance = home.xDistance.refValue();
    M_Float& yDistance = home.yDistance.refValue();


    int tesselate = 4;

    if(Seamless* smls = getSeamless2()){
        tesselate = smls->tesselate;
    }

    bands.length = vCells * uCells * 2 + vCells;
    for(int i = bands.length; --i >= 0;){
        bands[i] = tesselate;
    }
    staves.length = vCells * uCells * 2 + uCells;
    for(int i = staves.length; --i >= 0;){
        staves[i] = tesselate;
    }
    xDistance.length = uCells + 1;
    for(int i = uCells + 1; --i >= 0;){
        xDistance[i] = i / par.x;
    }
    yDistance.length = vCells + 1;
    for(int i = vCells + 1; --i >= 0;){
        yDistance[i] = i / par.y;
    }


}

inline bool isClosest(const Vec3f& ba3f, const Vec3f& bb3f, const Vec3f& ta3f, const Vec3f& tb3f){
    Vec2f ba = Vec2f(ba3f.x, ba3f.y);
    Vec2f bb = Vec2f(bb3f.x, bb3f.y);

    Vec2f ta = Vec2f(ta3f.x, ta3f.y);
    Vec2f tb = Vec2f(tb3f.x, tb3f.y);

    Vec2f diff1 = ba - tb;
    Vec2f diff2 = bb - ta;

    return diff1.length() >= diff2.length();
}

inline void genWestIndices(Art*& part, M_Triangle& triangle, const int& tiOffset, const int staves1, const M_Vec3f& coord, int& b, int& bRow, int& tRowA, int& tRowB, int bEnd, const int tEnd){

    goto beginHere;

    while(1){
        //if(coord[tRowA].y + tHalfStep >= coord[bRow].y + bHalfStep){
        //if(coord[tRowB].y - coord[bRow].y >= coord[bRow + 1].y - coord[tRowA].y){
        if(isClosest(coord[bRow], coord[bRow + 1], coord[tRowA], coord[tRowB])){

            triangle[b].ta = tiOffset + (triangle[b].ia = bRow);//add bottom tri
            triangle[b].tb = tiOffset + (triangle[b].ib = tRowA);
            triangle[b].tc = tiOffset + (triangle[b].ic = bRow + 1);
            triangle[b].pa = part;
            triangle[b].pb = part;
            triangle[b].pc = part;
            b++;
            bRow++;
beginHere:
            if(bRow >= bEnd){
                while(tRowB <= tEnd){
                    triangle[b].ta = tiOffset+(triangle[b].ia = bRow);
                    triangle[b].tb = tiOffset+(triangle[b].ib = tRowA);
                    triangle[b].tc = tiOffset+(triangle[b].ic = tRowB);
                    triangle[b].pa = part;
                    triangle[b].pb = part;
                    triangle[b].pc = part;
                    b++;
                    tRowA = tRowB;
                    tRowB += staves1 - 1;
                }
                break;
            }
        }
        else{
            triangle[b].ta = tiOffset+(triangle[b].ia = bRow);//add top tri
            triangle[b].tb = tiOffset+(triangle[b].ib = tRowA);
            triangle[b].tc = tiOffset+(triangle[b].ic = tRowB);
            triangle[b].pa = part;
            triangle[b].pb = part;
            triangle[b].pc = part;
            b++;
            tRowA = tRowB;
            tRowB += staves1 - 1;
            if(tRowB > tEnd){
                while(bRow < bEnd){
                    triangle[b].ta = tiOffset + (triangle[b].ia = bRow);
                    triangle[b].tb = tiOffset + (triangle[b].ib = tRowA);
                    triangle[b].tc = tiOffset + (triangle[b].ic = bRow + 1);
                    triangle[b].pa = part;
                    triangle[b].pb = part;
                    triangle[b].pc = part;
                    b++;
                    bRow++;
                }
                break;
            }
        }
    }
}

inline void genEastIndices(Art*& part, M_Triangle& triangle, const int& tiOffset, const int staves1, const M_Vec3f& coord, int& b, int bRowA, int bRowB, int tRow, const int bEnd, const int tEnd){
    if(bRowA == bEnd){
        goto fanOut;
    }
    goto beginLoop;

    while(1){

        //if(coord[tRow].y + tHalfStep >= coord[bRowA].y + bHalfStep){
        //if(coord[tRow + 1].x - coord[bRowA].x < coord[bRowB].x - coord[tRow].x){
        if(isClosest(coord[bRowA], coord[bRowB], coord[tRow], coord[tRow + 1])){


            triangle[b].ta = tiOffset + (triangle[b].ia = bRowA);//add bottom tri
            triangle[b].tb = tiOffset + (triangle[b].ib = tRow);
            triangle[b].tc = tiOffset + (triangle[b].ic = bRowB);
            triangle[b].pa = part;
            triangle[b].pb = part;
            triangle[b].pc = part;
            b++;
            bRowA = bRowB;
            bRowB += staves1 - 1;
            if(bRowA >= bEnd){
fanOut:
                while(tRow < tEnd){
                    triangle[b].ta = tiOffset+(triangle[b].ia = bRowA);//add top tri
                    triangle[b].tb = tiOffset+(triangle[b].ib = tRow);
                    triangle[b].tc = tiOffset+(triangle[b].ic = tRow + 1);
                    triangle[b].pa = part;
                    triangle[b].pb = part;
                    triangle[b].pc = part;
                    b++;
                    tRow++;
                }
                break;
            }
        }
        else{
            triangle[b].ta = tiOffset+(triangle[b].ia = bRowA);//add top tri
            triangle[b].tb = tiOffset+(triangle[b].ib = tRow);
            triangle[b].tc = tiOffset+(triangle[b].ic = tRow + 1);
            triangle[b].pa = part;
            triangle[b].pb = part;
            triangle[b].pc = part;
            b++;
            tRow++;
beginLoop:
            if(tRow >= tEnd){
                while(bRowB <= bEnd){
                    triangle[b].ta = tiOffset + (triangle[b].ia = bRowA);//add bottom tri
                    triangle[b].tb = tiOffset + (triangle[b].ib = tRow);
                    triangle[b].tc = tiOffset + (triangle[b].ic = bRowB);
                    triangle[b].pa = part;
                    triangle[b].pb = part;
                    triangle[b].pc = part;
                    b++;
                    bRowA = bRowB;
                    bRowB += staves1 - 1;
                }
                break;
            }
        }
    }
}

inline void genNorthSPIndices(Art*& part, M_Triangle& triangle, const int& tiOffset, int& b, int bRowA, int bRowB, int tRow, int firstBEnd, int secondBEnd){

    do{
        triangle[b].ta = tiOffset + (triangle[b].ia = bRowA);//bottom tri
        triangle[b].tb = tiOffset + (triangle[b].ib = bRowB);
        triangle[b].tc = tiOffset + (triangle[b].ic = tRow);
        triangle[b].pa = part;
        triangle[b].pb = part;
        triangle[b].pc = part;
        b++;
        bRowA = bRowB;
        bRowB++;
    }while(bRowB <= firstBEnd);

    triangle[b].ta = tiOffset + (triangle[b].ia = bRowA);//bottom tri
    triangle[b].tb = tiOffset + (triangle[b].ib = secondBEnd);
    triangle[b].tc = tiOffset + (triangle[b].ic = tRow);
    triangle[b].pa = part;
    triangle[b].pb = part;
    triangle[b].pc = part;
    b++;
}

inline void genSouthFanIndices(Art*& part, M_Triangle& triangle, const int& tiOffset, int& b, int bRow, int tRowA, int tRowB, int tEnd){

    do{
        triangle[b].ta = tiOffset + (triangle[b].ia = bRow);//add top tri
        triangle[b].tb = tiOffset + (triangle[b].ib = tRowB);
        triangle[b].tc = tiOffset + (triangle[b].ic = tRowA);
        triangle[b].pa = part;
        triangle[b].pb = part;
        triangle[b].pc = part;
        b++;
        tRowA = tRowB;
        tRowB++;
    }while(tRowB <= tEnd);
}

inline void genSouthFanIndices(Art*& part, M_Triangle& triangle, const int& tiOffset, int& b, int bRow, int tRowA, int tRowB, int firstTEnd, int secondTEnd){

    do{
        triangle[b].ta = tiOffset + (triangle[b].ia = bRow);//add top tri
        triangle[b].tb = tiOffset + (triangle[b].ib = tRowB);
        triangle[b].tc = tiOffset + (triangle[b].ic = tRowA);
        triangle[b].pa = part;
        triangle[b].pb = part;
        triangle[b].pc = part;
        b++;
        tRowA = tRowB;
        tRowB++;
    }while(tRowB <= firstTEnd);

    triangle[b].ta = tiOffset + (triangle[b].ia = bRow);//add top tri
    triangle[b].tb = tiOffset + (triangle[b].ib = secondTEnd);
    triangle[b].tc = tiOffset + (triangle[b].ic = tRowA);
    triangle[b].pa = part;
    triangle[b].pb = part;
    triangle[b].pc = part;
    b++;
}

inline void genWestFanIndices(Art*& part, M_Triangle& triangle, const int& tiOffset, int& b, int bRow, int tRowA, int tRowB, int incAmount, int firstTEnd, int secondTEnd){

    do{
        triangle[b].ta = tiOffset + (triangle[b].ia = bRow);//add top tri
        triangle[b].tb = tiOffset + (triangle[b].ib = tRowA);
        triangle[b].tc = tiOffset + (triangle[b].ic = tRowB);
        triangle[b].pa = part;
        triangle[b].pb = part;
        triangle[b].pc = part;
        b++;
        tRowA = tRowB;
        tRowB += incAmount;
    }while(tRowB <= firstTEnd);

    triangle[b].ta = tiOffset + (triangle[b].ia = bRow);//add top tri
    triangle[b].tb = tiOffset + (triangle[b].ib = tRowA);
    triangle[b].tc = tiOffset + (triangle[b].ic = secondTEnd);
    triangle[b].pa = part;
    triangle[b].pb = part;
    triangle[b].pc = part;
    b++;
}

inline void genEastFanIndices(Art*& part, M_Triangle& triangle, const int& tiOffset, int& b, int bRowA, int bRowB, int tRow, int incAmount, int firstBEnd, int secondBEnd){

    do{
        triangle[b].ta = tiOffset + (triangle[b].ia = tRow);//add bottom tri
        triangle[b].tb = tiOffset + (triangle[b].ib = bRowB);
        triangle[b].tc = tiOffset + (triangle[b].ic = bRowA);
        triangle[b].pa = part;
        triangle[b].pb = part;
        triangle[b].pc = part;
        b++;
        bRowA = bRowB;
        bRowB += incAmount;
    }while(bRowB <= firstBEnd);

    triangle[b].ta = tiOffset + (triangle[b].ia = tRow);//add bottom tri
    triangle[b].tb = tiOffset + (triangle[b].ib = secondBEnd);
    triangle[b].tc = tiOffset + (triangle[b].ic = bRowA);
    triangle[b].pa = part;
    triangle[b].pb = part;
    triangle[b].pc = part;
    b++;
}

inline void genSouthIndices(Art*& part, M_Triangle& triangle, const int& tiOffset, const M_Vec3f& coord, int& b, int& bRow, int& tRow, const int bEnd, const int tEnd){
    if(tRow >= tEnd){
        goto fanIn;
    }
    goto beginLoop;
    while(1){
        //if(coord[tRow].x + tHalfStep >= coord[bRow].x + bHalfStep){
        if(isClosest(coord[bRow], coord[bRow + 1], coord[tRow], coord[tRow + 1])){

            triangle[b].ta = tiOffset + (triangle[b].ia = bRow + 1);//add bottom tri
            triangle[b].tb = tiOffset + (triangle[b].ib = tRow);
            triangle[b].tc = tiOffset + (triangle[b].ic = bRow);
            triangle[b].pa = part;
            triangle[b].pb = part;
            triangle[b].pc = part;
            b++;
            bRow++;
beginLoop:
            if(bRow >= bEnd){
                while(tRow < tEnd){
                    triangle[b].ta = tiOffset+(triangle[b].ia = bRow);
                    triangle[b].tb = tiOffset+(triangle[b].ib = tRow + 1);
                    triangle[b].tc = tiOffset+(triangle[b].ic = tRow);
                    triangle[b].pa = part;
                    triangle[b].pb = part;
                    triangle[b].pc = part;
                    b++;
                    tRow++;
                }
                break;
            }
        }
        else{
            triangle[b].ta = tiOffset+(triangle[b].ia = bRow);//add top tri
            triangle[b].tb = tiOffset+(triangle[b].ib = tRow + 1);
            triangle[b].tc = tiOffset+(triangle[b].ic = tRow);
            triangle[b].pa = part;
            triangle[b].pb = part;
            triangle[b].pc = part;
            b++;
            tRow++;
            if(tRow >= tEnd){
fanIn:
                while(bRow < bEnd){
                    triangle[b].ta = tiOffset + (triangle[b].ia = bRow + 1);
                    triangle[b].tb = tiOffset + (triangle[b].ib = tRow);
                    triangle[b].tc = tiOffset + (triangle[b].ic = bRow);
                    triangle[b].pa = part;
                    triangle[b].pb = part;
                    triangle[b].pc = part;
                    b++;
                    bRow++;
                }
                break;
            }
        }
    }
}

inline void genCornerIndices(Art*& part, M_Triangle& triangle, const int& tiOffset, const M_Vec3f& coord, int& b, int bRowA, int bRowB, int tRowA, int tRowB){

    if(isClosest(coord[bRowA], coord[bRowB], coord[tRowA], coord[tRowB])){
        triangle[b].ta = tiOffset + (triangle[b].ia = bRowA);//bottom tri
        triangle[b].tb = tiOffset + (triangle[b].ib = bRowB);
        triangle[b].tc = tiOffset + (triangle[b].ic = tRowA);
        triangle[b].pa = part;
        triangle[b].pb = part;
        triangle[b].pc = part;
        b++;

        triangle[b].ta = tiOffset+(triangle[b].ia = tRowB);//top tri
        triangle[b].tb = tiOffset+(triangle[b].ib = tRowA);
        triangle[b].tc = tiOffset+(triangle[b].ic = bRowB);
        triangle[b].pa = part;
        triangle[b].pb = part;
        triangle[b].pc = part;
        b++;
    }
    else{
        triangle[b].ta = tiOffset + (triangle[b].ia = bRowA);//bottom tri
        triangle[b].tb = tiOffset + (triangle[b].ib = bRowB);
        triangle[b].tc = tiOffset + (triangle[b].ic = tRowB);
        triangle[b].pa = part;
        triangle[b].pb = part;
        triangle[b].pc = part;
        b++;

        triangle[b].ta = tiOffset+(triangle[b].ia = tRowB);//top tri
        triangle[b].tb = tiOffset+(triangle[b].ib = tRowA);
        triangle[b].tc = tiOffset+(triangle[b].ic = bRowA);
        triangle[b].pa = part;
        triangle[b].pb = part;
        triangle[b].pc = part;
        b++;


    }
}

inline void genMidIndices(Art*& part, M_Triangle& triangle, const int& tiOffset, int& bRow, int& tRow, const int staves0, const int staves1, const M_Vec3f& coord, int& b, const int tEnd){
    int end = staves1 - 2;
    for(; tRow <= tEnd; bRow++, tRow++){
        for(int j = 0; j < end; j++){
            triangle[b].ta = tiOffset+(triangle[b].ia = bRow + 1);
            triangle[b].tb = tiOffset+(triangle[b].ib = tRow);
            triangle[b].tc = tiOffset+(triangle[b].ic = bRow);
            triangle[b].pa = part;
            triangle[b].pb = part;
            triangle[b].pc = part;
            b++;
            triangle[b].ta = tiOffset+(triangle[b].ia = bRow + 1);
            triangle[b].tb = tiOffset+(triangle[b].ib = tRow + 1);
            triangle[b].tc = tiOffset+(triangle[b].ic = tRow);
            triangle[b].pa = part;
            triangle[b].pb = part;
            triangle[b].pc = part;
            bRow++;
            tRow++;
            b++;
        }
    }
}

inline void genNorthIndices(Art*& part, M_Triangle& triangle, const int& tiOffset, const int staves1, const M_Vec3f& coord, int& b, int& bRow, int& tRow, const int bEnd, const int tEnd){
    if(bRow >= bEnd){
        goto fanOut;
    }
    goto beginLoop;
    while(1){
        //if(coord[tRow].x + tHalfStep >= coord[bRow].x + bHalfStep){
        //if(coord[tRow + 1].x - coord[bRow].x >= coord[bRow + 1].x - coord[tRow].x){
        if(isClosest(coord[bRow], coord[bRow + 1], coord[tRow], coord[tRow + 1])){

            triangle[b].ta = tiOffset + (triangle[b].ia = bRow + 1);//add bottom tri
            triangle[b].tb = tiOffset + (triangle[b].ib = tRow);
            triangle[b].tc = tiOffset + (triangle[b].ic = bRow);
            triangle[b].pa = part;
            triangle[b].pb = part;
            triangle[b].pc = part;
            b++;
            bRow++;
            if(bRow >= bEnd){
fanOut:
                while(tRow < tEnd){
                    triangle[b].ta = tiOffset+(triangle[b].ia = bRow);
                    triangle[b].tb = tiOffset+(triangle[b].ib = tRow + 1);
                    triangle[b].tc = tiOffset+(triangle[b].ic = tRow);
                    triangle[b].pa = part;
                    triangle[b].pb = part;
                    triangle[b].pc = part;
                    b++;
                    tRow++;
                }
                break;
            }
        }
        else{
            triangle[b].ta = tiOffset+(triangle[b].ia = bRow);//add top tri
            triangle[b].tb = tiOffset+(triangle[b].ib = tRow + 1);
            triangle[b].tc = tiOffset+(triangle[b].ic = tRow);
            triangle[b].pa = part;
            triangle[b].pb = part;
            triangle[b].pc = part;
            b++;
            tRow++;
beginLoop:
            if(tRow >= tEnd){
                while(bRow < bEnd){
                    triangle[b].ta = tiOffset + (triangle[b].ia = bRow + 1);
                    triangle[b].tb = tiOffset + (triangle[b].ib = tRow);
                    triangle[b].tc = tiOffset + (triangle[b].ic = bRow);
                    triangle[b].pa = part;
                    triangle[b].pb = part;
                    triangle[b].pc = part;
                    b++;
                    bRow++;
                }
                break;
            }
        }
    }
}

class StemCellItem{
public:
    int westCoordOffset;
    int southCoordOffset;
    int midCoordOffset;
    int northCoordOffset;
    int eastCoordOffset;
    int endCellCoord;



    int northWestCoordOffset;
    int southEastCoordOffset;
    int northEastCoordOffset;
    int southWestCoordOffset;
    //int northCoordEnd;

    StemCellItem():
        westCoordOffset(0),
        southCoordOffset(0),
        midCoordOffset(0),
        northCoordOffset(0),
        eastCoordOffset(0),
        endCellCoord(0),


        northWestCoordOffset(0),
        southEastCoordOffset(0),
        northEastCoordOffset(0),
        southWestCoordOffset(0)
        //northCoordEnd(0),
    {}
/*
    StemCellItem(int westCoordOffset, int southCoordOffset, int midCoordOffset, int northCoordOffset, int eastCoordOffset, int endCellCoord):
        westCoordOffset(westCoordOffset),
        southCoordOffset(southCoordOffset),
        midCoordOffset(midCoordOffset),
        northCoordOffset(northCoordOffset),
        eastCoordOffset(eastCoordOffset),
        endCellCoord(endCellCoord)
    {}
*/
};

inline void genTriangleIndices(Art*& part, M_Triangle& triangle, const int& tiOffset, int& b, int ai, int bi, int ci){
    triangle[b].ta = tiOffset + (triangle[b].ia = ai);
    triangle[b].tb = tiOffset + (triangle[b].ib = bi);
    triangle[b].tc = tiOffset + (triangle[b].ic = ci);
    triangle[b].pa = part;
    triangle[b].pb = part;
    triangle[b].pc = part;
    b++;
}

static M_<StemCellItem> g_stemCellArray;

inline void genIndices(Art* part, M_Vec3f& coord, M_Triangle& triangle, const int& tiOffset,
    const M_Float& xDistance, const M_Float& yDistance, const M_Int staves, const M_Int bands, const int uCells, const int vCells, const int s){
    int b = s;
    int bandsIdx = 0;
    int stavesIdx = 0;
    for(int v = 0; v < vCells; v++){
        stavesIdx = v * uCells * 2;
        bandsIdx = v;
        for(int u = 0; u < uCells; u++){
            StemCellItem& sci = g_stemCellArray[v * uCells + u];
            const int westCoordOffset = sci.westCoordOffset;
            const int southCoordOffset = sci.southCoordOffset;
            const int midCoordOffset = sci.midCoordOffset;
            const int northCoordOffset = sci.northCoordOffset;
            const int eastCoordOffset = sci.eastCoordOffset;
            const int endCellCoord = sci.endCellCoord;

            const int bands0 = bands[bandsIdx];
            bandsIdx += vCells;
            const int bands1 = bands[bandsIdx];
            bandsIdx += vCells;
            const int bands2 = bands[bandsIdx];

            const int staves0 = staves[stavesIdx];
            const int staves1 = staves[stavesIdx + uCells];
            const int staves2 = staves[stavesIdx + uCells * 2];
            stavesIdx++;

            int southTriLen = staves0 == 0 ? staves1 : staves0 - 2 + staves1 - 2;
            int northTriLen = staves2 == 0 ? staves1 : staves1 - 2 + staves2 - 2;
            int midTriLen = (staves1 - 2) * (bands1 - 2) * 2;
            int westTriLen = bands0 == 0 ? bands1 : bands0 - 2 + bands1 - 2;
            int eastTriLen = bands1 - 2 + bands2 - 2;

            //if(uCell == 0){
                //triangle.length += westTriLen + southTriLen + midTriLen + northTriLen + eastTriLen;
            //}
            //else{
            //triangle.length += southTriLen + westTriLen + midTriLen;
            int triLen = westTriLen + southTriLen + midTriLen + northTriLen + eastTriLen;

            if(bands0 > 0 && staves0 > 0){
                triLen += 4;
            }
            if(staves2 > 0){
                triLen += 4;
            }
            triangle.length += triLen;

            if(v == 1){
                triangle.length = triangle.length;
            }
            bool allSouthComplete = false;
            if(bands0 == 1){
                if(staves0 == 0){
                    int bRow = sci.southWestCoordOffset;
                    int tRowA = sci.northWestCoordOffset;
                    int tRowB = sci.northWestCoordOffset + 1;
                    int tEnd = tRowA + staves2;
                    genSouthFanIndices(part, triangle, tiOffset, b, bRow, tRowA, tRowB, tEnd);
                    continue;
                }
                if(staves0 == 1){
                    int tRow = sci.northWestCoordOffset;
                    int bRow = sci.southWestCoordOffset;
                    int bEnd = bRow + 1;
                    int tEnd = tRow + staves2;
                    genNorthIndices(part, triangle, tiOffset, staves1, coord, b, bRow, tRow, bEnd, tEnd);
                    continue;
                }
            }
            if(bands0 > 0 && staves0 > 0){
                int bRowA = sci.southWestCoordOffset;
                int bRowB = southCoordOffset;
                int tRowA = westCoordOffset;
                int tRowB = midCoordOffset;
                genCornerIndices(part, triangle, tiOffset, coord, b, bRowA, bRowB, tRowA, tRowB); //south west
                if(southTriLen <= 0 && staves1 > 1){
                    if(bands2 == 0){
                        if(staves1 > 2){
                            genTriangleIndices(part, triangle, tiOffset, b, sci.southEastCoordOffset, midCoordOffset + 1, midCoordOffset);
                        }
                    }
                    else{
                        int bRow = sci.southEastCoordOffset;
                        int tRowA = midCoordOffset;
                        int tRowB = midCoordOffset + 1;
                        int firstTEnd = tRowA + staves1 - 2;
                        int secondTEnd = eastCoordOffset;
                        if(tRowA == firstTEnd){
                            if(staves0 == 1){
                                genTriangleIndices(part, triangle, tiOffset, b, bRow, secondTEnd, firstTEnd);
                                allSouthComplete = true;
                            }
                        }
                        else{
                            genSouthFanIndices(part, triangle, tiOffset, b, bRow, tRowA, tRowB, firstTEnd, secondTEnd);
                            allSouthComplete = true;
                        }
                    }
                }
            }
            if(bands0 == 0){
                int bRow = sci.southWestCoordOffset;
                int tRowA = southCoordOffset;
                int tRowB = midCoordOffset;
                int incAmount = staves1 - 1;
                int firstTEnd = midCoordOffset + (staves1 - 1) * (bands1 - 2);
                int secondTEnd = northCoordOffset;
                if(staves0 == 1){
                    tRowA = midCoordOffset;
                    tRowB += incAmount;
                }
                genWestFanIndices(part, triangle, tiOffset, b, bRow, tRowA, tRowB, incAmount, firstTEnd, secondTEnd);
            }
            else {
                int bRow = westCoordOffset;
                int tRowA = midCoordOffset;
                int tRowB = midCoordOffset + staves1 - 1;
                int bEnd = midCoordOffset - 1;
                int tEnd = midCoordOffset  + (staves1 - 1) * (bands1 - 2);
                genWestIndices(part, triangle, tiOffset, staves1, coord, b, bRow, tRowA, tRowB, bEnd, tEnd);
            }

            if(staves0 == 0){
                int bRow = southCoordOffset;
                int tRowA = westCoordOffset;
                int tRowB = midCoordOffset;
                int firstTEnd = tRowB + staves1 - 2;
                int secondTEnd = eastCoordOffset;
                genSouthFanIndices(part, triangle, tiOffset, b, bRow, tRowA, tRowB, firstTEnd, secondTEnd);
            }
            else{
                if(southTriLen > 0){
                    int bRow = southCoordOffset;
                    int tRow = midCoordOffset;
                    int bEnd = bRow + staves0 - 2;
                    int tEnd = tRow + staves1 - 2;
                    genSouthIndices(part, triangle, tiOffset, coord, b, bRow, tRow, bEnd, tEnd);

                    if(staves0 == 1 && bands0 != 0 && bands2 != 0){
                        genTriangleIndices(part, triangle, tiOffset, b, sci.southEastCoordOffset, eastCoordOffset, midCoordOffset + staves1 - 2);
                    }
                }


                if(bands2 == 0){
                    int tRow = sci.southEastCoordOffset;
                    int bRowA = southCoordOffset + staves0 - 2;
                    int bRowB = midCoordOffset + staves1 - 2;
                    int incAmount = staves1 - 1;
                    int firstBEnd = midCoordOffset + (staves1 - 1) * (bands1 - 2) + staves1 - 2;
                    int secondBEnd = northCoordOffset + staves2 - 2;
                    if(staves0 == 1){
                        bRowA = midCoordOffset + staves1 - 2;
                        bRowB += incAmount;
                    }
                    genEastFanIndices(part, triangle, tiOffset, b, bRowA, bRowB, tRow, incAmount, firstBEnd, secondBEnd);
                }
                else if(allSouthComplete == false && bands2 > 0 && staves0 > 0 && false == (southTriLen > 0 && staves0 == 1)){
                    int bRowA = southCoordOffset + staves0 - 2;
                    int bRowB = sci.southEastCoordOffset;
                    int tRowA = midCoordOffset + staves1 - 2;
                    int tRowB = eastCoordOffset;
                    genCornerIndices(part, triangle, tiOffset, coord, b, bRowA, bRowB, tRowA, tRowB);//south east
                    if(southTriLen <= 0 && staves1 > 1){
                        if(bands0 == 0){
                            if(staves1 > 2){
                                genTriangleIndices(part, triangle, tiOffset, b, sci.southEastCoordOffset, midCoordOffset + 1, midCoordOffset);
                            }
                        }
                        else{
                            int bRow = sci.southEastCoordOffset;
                            int tRowA = midCoordOffset;
                            int tRowB = midCoordOffset + 1;
                            int firstTEnd = tRowA + staves1 - 2;
                            int secondTEnd = eastCoordOffset;
                            if(tRowA == firstTEnd){
                                if(staves0 == 1){
                                    genTriangleIndices(part, triangle, tiOffset, b, bRow, secondTEnd, firstTEnd);
                                }
                            }
                            else{
                                genSouthFanIndices(part, triangle, tiOffset, b, bRow, tRowA, tRowB, firstTEnd, secondTEnd);
                            }
                        }
                    }
                }
            }
            if(midTriLen){
                int bRow = midCoordOffset;
                int tRow = midCoordOffset + staves1 - 1;
                int end = midCoordOffset + (staves1 - 1) * (bands1 - 2) + staves1 - 2;
                genMidIndices(part, triangle, tiOffset, bRow, tRow, staves0, staves1, coord, b, end);
            }

            if(staves2 == 0){
                int tRow = northCoordOffset;
                int bRowA = westCoordOffset + bands0 - 2;
                int bRowB = midCoordOffset + (staves1 - 1) * (bands1 - 2);
                int firstBEnd = bRowB + staves1 - 2;
                int secondBEnd = eastCoordOffset + bands2 - 2;
                genNorthSPIndices(part, triangle, tiOffset, b, bRowA, bRowB, tRow, firstBEnd, secondBEnd);
            }
            else{
                if(bands0 > 0){
                    int bRowA = westCoordOffset + bands0 - 2;
                    int bRowB = midCoordOffset + (staves1 - 1) * (bands1 - 2);
                    //int keep_bRowB = bRowB;
                    int tRowA = sci.northWestCoordOffset;
                    int tRowB = tRowA + 1;

                    genCornerIndices(part, triangle, tiOffset, coord, b, bRowA, bRowB, tRowA, tRowB); //north west
                }
                if(northTriLen){
                    int tRow = northCoordOffset;
                    int bRow = midCoordOffset + (staves1 - 1) * (bands1 - 2); //do we need this?
                    int bEnd = midCoordOffset + (staves1 - 1) * (bands1 - 1) - 1;
                    int tEnd = tRow + staves2 - 2;
                    genNorthIndices(part, triangle, tiOffset, staves1, coord, b, bRow, tRow, bEnd, tEnd);
                }
                //if(v == 1){
                //	int i = 0;
                //}
                if(bands2 > 0){
                    int bRowA = midCoordOffset + (staves1 - 1) * (bands1 - 2) + staves1 - 2;
                    int bRowB = eastCoordOffset + bands2 - 2;
                    int tRowA = sci.northCoordOffset + staves2 - 2;
                    int tRowB = sci.northEastCoordOffset;
                    genCornerIndices(part, triangle, tiOffset, coord, b, bRowA, bRowB, tRowA, tRowB); //north east
                }
            }

            if(bands2 > 0){

                int bRowA = midCoordOffset + staves1 - 2;
                int bRowB = bRowA + staves1 - 1;
                int tRow = eastCoordOffset;
                int bEnd = eastCoordOffset - 1;
                int tEnd = eastCoordOffset + bands2 - 2;

                genEastIndices(part, triangle, tiOffset, staves1, coord, b, bRowA, bRowB, tRow, bEnd, tEnd);
            }
            if(staves0 == 1 && bands0 == 0 && bands2 == 0){
                genTriangleIndices(part, triangle, tiOffset, b, sci.southEastCoordOffset, midCoordOffset, sci.southWestCoordOffset);
                if(staves1 == 3){
                    genTriangleIndices(part, triangle, tiOffset, b, sci.southEastCoordOffset, eastCoordOffset, midCoordOffset + staves1 - 2);
                }
            }
            if(v == 0){
            //	triangle.length = b;
                //return;
            }

        }
        bandsIdx++;
    }
}
//staves field increments to the next cell right then starts on the next row of cells above
//bands field increments to the next cell above then starts on the next collum to the east

//bands formula for cell's u v index:
//west/bands0: u * vCells * 2 + v
//mid/bands1: u * vCells * 2 + vCells + v
//east/bands2: (u + 1) * vCells * 2 + v

void genFlatStem2(
        Art* part, M_Vec3f& coord, M_Triangle& triangle, const int& tiOffset,
        const bool xCentred, const bool uWrap, const bool vWrap, const M_Float& xDistance, const M_Float& yDistance, const M_Int staves, const M_Int bands, const int uCells, const int vCells, const int h, const int s)
{
    g_stemCellArray.length = uCells * vCells;

    int uCells_1 = uCells + 1;
    int vCells_1 = vCells + 1;

    int coordLen = h;
    int bandsIdx = 0;
    int bands0 = 0;
    int bands1 = 0;
    int staves0 = 0;
    int staves1 = 0;
    int stavesIdx = 0;
    if(vCells * uCells * 2 + vCells != bands.length){
        throw;
    }
    if(vCells * uCells * 2 + uCells != staves.length){
        throw;
    }

    int u = 0;
    int v = 0;
    for(v = 0; v < vCells; v++){
        stavesIdx = v * uCells * 2;
        for(u = 0; u < uCells; u++){
            StemCellItem& sci = g_stemCellArray[v * uCells + u];

            sci.southWestCoordOffset = coordLen;
            sci.southCoordOffset = coordLen + 1;

            staves0 = staves[stavesIdx++];
            sci.southEastCoordOffset = coordLen + staves0;

            if(staves0 == 0){
                sci.southCoordOffset = coordLen;
            }
            if(v > 0){
                if(bands[u * vCells * 2 + v - 1] == 0){
                    sci.southCoordOffset = coordLen;
                    sci.southWestCoordOffset = g_stemCellArray[(v - 1) * uCells + u].southWestCoordOffset;
                    staves0--;//we are not changing the number of staves just the value to add to coordLen (will be discarded after end of loop)
                    sci.southEastCoordOffset = coordLen + staves0;
                }
                if(bands[(u + 1) * vCells * 2 + v - 1] == 0){
                    sci.southEastCoordOffset = g_stemCellArray[(v - 1) * uCells + u].southEastCoordOffset;
                    staves0--;//we are not changing the number of staves just the value to add to coordLen (will be discarded after end of loop)
                }
            }

            coordLen += staves0;
        }
        if(uWrap == false){
            coordLen++;
        }
        stavesIdx = v * uCells * 2;
        bandsIdx = v;
        for(u = 0; u < uCells; u++){
            bands0 = bands[bandsIdx];
            bandsIdx += vCells;
            bands1 = bands[bandsIdx];
            bandsIdx += vCells;

            staves0 = staves[stavesIdx];
            staves1 = staves[stavesIdx + uCells];
            stavesIdx++;

            int westCoordLen = bands0 == 0 ? 0 : bands0 - 1;
            int midCoordLen = (staves1 - 1) * (bands1 - 1);

            StemCellItem& sci = g_stemCellArray[v * uCells + u];

            sci.westCoordOffset = coordLen;
            coordLen += westCoordLen;
            sci.midCoordOffset = coordLen;
            coordLen += midCoordLen;
        }
        --u;

        if(uWrap){
            g_stemCellArray[v * uCells + u].southEastCoordOffset = g_stemCellArray[v * uCells].southWestCoordOffset;
            g_stemCellArray[v * uCells + u].eastCoordOffset = g_stemCellArray[v * uCells].westCoordOffset;
        }
        else{
            if(v > 0 && bands[bandsIdx - 1] == 0){
                g_stemCellArray[v * uCells + u].southEastCoordOffset = g_stemCellArray[(v - 1) * uCells + u].southEastCoordOffset;
            }
            if(bands[bandsIdx] != 0){//need more work for if previous was 0 bands?
                g_stemCellArray[v * uCells + u].eastCoordOffset = coordLen;
                coordLen += bands[bandsIdx] - 1;
            }
        }
    }
    stavesIdx = v * uCells * 2;
    v--;
    if(vWrap == false){

        for(u = 0; u < uCells; u++){
            StemCellItem& sci = g_stemCellArray[v * uCells + u];

            sci.northWestCoordOffset = coordLen;
            sci.northCoordOffset = coordLen + 1;


            int staves2 = staves[stavesIdx++];
            sci.northEastCoordOffset = coordLen + staves2;

            if(staves2 == 0){
                sci.northCoordOffset = coordLen;
            }
            if(bands[u * vCells * 2 + v] == 0){//west
                sci.northCoordOffset = coordLen;
                sci.northWestCoordOffset = g_stemCellArray[v * uCells + u].southWestCoordOffset;
                staves2--;//we are not changing the number of staves just the value to add to coordLen (will be discarded after end of loop)
                sci.northEastCoordOffset = coordLen + staves2;
            }
            if(bands[(u + 1) * vCells * 2 + v] == 0){//east
                sci.northEastCoordOffset = g_stemCellArray[v * uCells + u].southEastCoordOffset;
                staves2--;//we are not changing the number of staves just the value to add to coordLen (will be discarded after end of loop)
            }

            coordLen += staves2;
        }
    }

    if(vWrap){
        for(u = 0; u < uCells; u++){
            StemCellItem& dst = g_stemCellArray[v * uCells + u];
            StemCellItem& src = g_stemCellArray[u];

            dst.northCoordOffset = src.southCoordOffset;
            dst.northWestCoordOffset = src.southWestCoordOffset;
            dst.northEastCoordOffset = src.southEastCoordOffset;
        }


    }
    if(uWrap == false){
        if(vWrap == false){
            coordLen++;
        }
    }
    else{
        for(v = 0; v < vCells; v++){
            StemCellItem& dst = g_stemCellArray[v * uCells + uCells - 1];
            StemCellItem& src = g_stemCellArray[v * uCells];

            dst.eastCoordOffset = src.westCoordOffset;
            dst.southEastCoordOffset = src.southWestCoordOffset;
            dst.northEastCoordOffset = src.northWestCoordOffset;
        }
    }



    coord.length = coordLen;
    for(v = 0; v < vCells; v++){
        for(u = uCells - 1; --u >= 0;){
            StemCellItem& dst = g_stemCellArray[v * uCells + u];
            StemCellItem& src = g_stemCellArray[v * uCells + u + 1];

            dst.eastCoordOffset = src.westCoordOffset;
            dst.southEastCoordOffset = src.southWestCoordOffset;
            dst.northEastCoordOffset = src.northWestCoordOffset;
        }

    }
    //for(v = 0; v < vCells; v++){
        //g_stemCellArray[v * uCells + vCells - 1].northCoordEnd = g_stemCellArray[v * uCells + u + 1].westCoordOffset;
    //}
    //g_stemCellArray[uCells * vCells - 1].northCoordEnd = coordLen;

    for(v = vCells - 1; --v >= 0;){
        for(int u = 0; u < uCells; u++){
            StemCellItem& dst = g_stemCellArray[v * uCells + u];
            StemCellItem& src = g_stemCellArray[(v + 1) * uCells + u];

            dst.northCoordOffset = src.southCoordOffset;
            dst.northEastCoordOffset = src.southEastCoordOffset;
            dst.northWestCoordOffset = src.southWestCoordOffset;
        }
    }


    float xDiff = 0;
    float yDiff = 0;
    float bandsStep = 0;
    float stavesStep = 0;
    int lastCoord = 0;

    int index = h;
    float x = 0;
    float y = 0;
    bool previousSouthSP = false;
    for(v = 0; v < vCells; v++){
        y = yDistance[v];
        stavesIdx = v * uCells * 2;
        previousSouthSP = false;
        for(u = 0; u < uCells; u++){
            x = xDistance[u];
            staves0 = staves[stavesIdx++];
            if(staves0 == 0){
                if(previousSouthSP == false){
                    coord[index++] = Vec3f(x + xDiff * .5, y, 0);
                    previousSouthSP = true;
                }
            }
            else{
                xDiff = xDistance[u + 1] - x;
                stavesStep = xDiff / staves0;
                lastCoord = index + staves0;
                if(previousSouthSP){
                    lastCoord--;
                    x += stavesStep;
                }
                previousSouthSP = false;

                if(v > 0){
                    if(bands[u * vCells * 2 + v - 1] == 0){
                        lastCoord--;
                        x += stavesStep;
                    }
//					else if(bands[(u + 1) * vCells * 2 + v - 1] == 0){
//						lastCoord--;
//					}
                }

                while(index < lastCoord){
                    coord[index++] = Vec3f(x, y, 0);
                    x += stavesStep;
                }
            }
//			g_stemCellArray[v * uCells + u].southCoordOffset = coordLen;
//			coordLen += staves[stavesIdx];
        }
        if(uWrap == false && previousSouthSP == false){
            if(g_stemCellArray[v * uCells + u - 1].southEastCoordOffset == index){
                coord[index++] = Vec3f(xDistance[u], y, 0);
            }
        }

        stavesIdx = v * uCells * 2;
        bandsIdx = v;
        for(u = 0; u < uCells; u++){
            bands0 = bands[bandsIdx];
            bandsIdx += vCells;
            bands1 = bands[bandsIdx];
            bandsIdx += vCells;

            staves0 = staves[stavesIdx];
            staves1 = staves[stavesIdx + uCells];
            stavesIdx++;

            const int eastCoordLen = staves0 - 1;
            const int westCoordLen = bands0 - 1;
            const int midCoordLen = (staves1 - 1) * (bands1 - 1);

            //StemCellItem& sci = g_stemCellArray[v * uCells + u];
            x = xDistance[u];
            y = yDistance[v];
            yDiff = yDistance[v + 1] - y;
            bandsStep = yDiff / bands0;
            y += bandsStep;

            lastCoord = index + bands0 - 1;
            while(index < lastCoord){
                coord[index++] = Vec3f(x, y, 0);
                y += bandsStep;
            }

            //sci.westCoordOffset = coordLen;
            //coordLen += westCoordLen;
            //sci.midCoordOffset = coordLen;
            //coordLen += midCoordLen;

            x = xDistance[u];
            y = yDistance[v];
            float xDiff = xDistance[u + 1] - x;
            yDiff = yDistance[v + 1] - y;
            float stavesStep = xDiff / staves1;
            bandsStep = yDiff / bands1;
            x += stavesStep;
            y += bandsStep;
            float beginX = x;
            for(int b = bands1 - 1; --b >= 0;){
                x = beginX;
                lastCoord = index + staves1 - 1;
                while(index < lastCoord){
                    coord[index++] = Vec3f(x, y, 0);
                    x += stavesStep;
                }
                y += bandsStep;

            }
        }
        if(uWrap == false){
            x = xDistance[u];
            y = yDistance[v];
            yDiff = yDistance[v + 1] - y;
            int bands2 = bands[bandsIdx];
            bandsIdx += vCells;
            bandsStep = yDiff / bands2;
            y += bandsStep;

            lastCoord = index + bands2 - 1;
            while(index < lastCoord){
                coord[index++] = Vec3f(x, y, 0);
                y += bandsStep;
            }
        }
        --u;

        //g_stemCellArray[v * uCells + u].eastCoordOffset = coordLen;
        //coordLen += bands[bandsIdx++] - 1;
    }
    if(vWrap == false){

        y = yDistance[v];
        stavesIdx = v * uCells * 2;
        v--;
        bool previousNorthSP = false;
        for(u = 0; u < uCells; u++){
            x = xDistance[u];
            xDiff = xDistance[u + 1] - x;
            staves0 = staves[stavesIdx];
            stavesIdx++;
            if(staves0 == 0){
                if(previousNorthSP == false){
                    coord[index++] = Vec3f(x + xDiff * .5, y, 0);
                    previousNorthSP = true;
                }
            }
            else{
                stavesStep = xDiff / staves0;
                lastCoord = index + staves0;
                if(previousNorthSP){
                    lastCoord--;
                    x += stavesStep;
                }
                previousNorthSP = false;

                if(v >= 0){
                    if(bands[u * vCells * 2 + v] == 0){//west
                        lastCoord--;
                        x += stavesStep;
                    }
                    if(bands[(u + 1) * vCells * 2 + v] == 0){//east
                        lastCoord--;
                    }
                }

                if(coord.length <= lastCoord){//should not need to fix length but for some reason our calculations are wrong when u edges ends with a single point
                    coord.length = lastCoord + 1;
                }

                while(index < lastCoord){
                    coord[index++] = Vec3f(x, y, 0);
                    x += stavesStep;
                }
            }
        }
        if(uWrap == false && previousNorthSP == false){
            coord[index++] = Vec3f(x, y, 0);
        }
    }
    genIndices(part, coord, triangle, tiOffset, xDistance, yDistance, staves, bands, uCells, vCells, s);
}

void genFlat(
        Art* part, M_Vec3f& coord, M_Triangle& triangle,
        const bool xCentred, const float xDistance, const M_Float &bandDist, int staves, const M_Int bands, bool beginSP, bool endSP, const int h, const int s)
{
    int numRings=bandDist.length;
    int lastRing=numRings;
    int index=0;

    int coordLen = h + staves*(numRings);
    if(beginSP){
        coordLen+=1-staves;
    }
    if(endSP){
        coordLen+=1-staves;
    }
    coord.length=coordLen;


    if(beginSP){
        coord[0].x=0;
        coord[0].y=0;
        coord[0].z=0;
        numRings--;
        index++;
    }
    if(endSP){
        coord[coordLen-1].x=0;
        coord[coordLen-1].y=bandDist[bandDist.length-1];
        coord[coordLen-1].z=0;
        numRings--;
        lastRing--;
    }
    addVerticesToGrid(coord, xCentred, xDistance, bandDist, staves, index, lastRing, h);
    genIndices(part, triangle, coord.length, staves, numRings, (int)beginSP, (int)endSP, h, s);
}

//**** genFlatJoinEnds is exactly the same as genFlat except for addVerticesToGridJoinEnds and genIndicesJoinEnds functions
void genFlatJoinEnds(
        Art* part, M_Vec3f& coord, M_Triangle& triangle,
        const bool& xCentred, float xDistance, const M_Float &bandDist, int staves, bool beginSP, bool endSP, const int h, const int s)
{
    int numRings=bandDist.length;
    int lastRing=numRings;
    int index=0;

    int coordLen = h + staves*(numRings);
    if(beginSP){
        coordLen+=1-staves;
    }
    if(endSP){
        coordLen+=1-staves;
    }
    coord.length=coordLen;


    if(beginSP){
        coord[h] = Vec3f();
        numRings--;
        index++;
    }
    if(endSP){
        coord[coordLen-1] = Vec3f(0, bandDist[bandDist.length-1], 0);
        numRings--;
        lastRing--;
    }
    addVerticesToGridJoinEnds(coord, xCentred, xDistance, bandDist, staves, index, lastRing, h);
    genIndicesJoinEnds(part, triangle, coord.length, staves, numRings, (int)beginSP, (int)endSP, h, s);
}

void genFlatTexJoinEnds(
        Art* part, M_Vec3f& coord, M_Triangle& triangle,
        float xDistance, M_Float &yDistance, int staves, M_Int bands, bool beginSP, bool endSP, const int h, const int s)
{
    int bandsTotal=0;
    int i;
    for (i=0;i<yDistance.length;i++){
        bandsTotal+=bands[i];
    }

    int coordLen=staves*(bandsTotal+1);
    if(beginSP){
        coordLen+=1-staves;
    }
    if(endSP){
        coordLen+=1-staves;
    }
    coord.length=coordLen;

    float lastDistance=0;
    float yTrans=0;
    int index=0;
    int numRings=bandsTotal+1;
    for (i=0;i<yDistance.length;i++){
        int ypc=bands[i];
        float yd=yDistance[i];
        lastDistance+=yd;
        float bandDist=yd/(float)ypc;

        if(i==0){
            if(beginSP){
                coord[h]=Vec3f();
                numRings--;
                yTrans=bandDist;
                yd-=bandDist;
                index++;
            }
            else{
                ypc++;
            }
        }
        else {
            yd-=bandDist;
            yTrans+=bandDist;
        }
        if( i == yDistance.length-1){
            if(endSP){
                coord[coordLen-1].x=0;
                coord[coordLen-1].y=lastDistance;
                coord[coordLen-1].z=0;
                numRings--;
                ypc--;
                yd-=bandDist;
                lastDistance-=bandDist;
            }
        }
        //addVerticesToGridJoinEnds(part, coord, xDistance, yd, lastDistance, yTrans, staves, ypc, index);
        throw;
        if(i==0){
            yTrans=yDistance[i];
        }
        else {
            yTrans+=yd;
        }
        index+=ypc*staves;
    }
    genIndicesTexJoinEnds(part, triangle, coord.length, staves, numRings, (int)beginSP, (int)endSP, h, s);
}

void MonoColorEffect::stemGenerateTri(Stem& stem, const ALathe& lathe, const int& k, Art*& part, M_Vec3f& coord, const M_Node& textureMapper, const int h, const int s){
    lathe.getBandAndYDist(stem.bands, Stem::bandDist, stem.yDistance);
    stem.convertYDistanceToHeight();
    if(stem.uWrap){
        genFlatJoinEnds( part, coord, part->triangle, stem.xCentred, stem.xDistance[0], Stem::bandDist, stem.staves[0], stem.beginSP, stem.endSP, h, s);
    }
    else {
        genFlat( part, coord, part->triangle, stem.xCentred, stem.xDistance[0], Stem::bandDist, stem.staves[0] + 1, stem.bands, stem.beginSP, stem.endSP, h, s);
    }
    doSlide(stem.slide, coord, h);
}

void ColorEffect::stemGenerateTri(Stem& stem, const ALathe& lathe, const int& k, Art*& part, M_Vec3f& coord, const M_Node& textureMapper, const int h, const int s){
    M_Color3f& color=this->color.refValue();
    int texCoordBase= part->fromEffectCoord=color.length;

    lathe.getBandAndYDist(stem.bands, Stem::bandDist, stem.yDistance);
    stem.convertYDistanceToHeight();
    if(stem.uWrap){
        genFlatJoinEnds( part, coord, part->triangle, stem.xCentred, stem.xDistance[0], Stem::bandDist, stem.staves[0], stem.beginSP, stem.endSP, h, s);
        M_Triangle& triangle=part->triangle;
        int idxDiff = texCoordBase - h;
        for(int i=triangle.length*3; --i>=0;){
            triangle.refTexIndex(i)=triangle.refIndex(i)+idxDiff;
        }
    }
    else {
        genFlat( part, coord, part->triangle, texCoordBase, stem.xCentred, stem.xDistance[0], Stem::bandDist, stem.staves[0] + 1, stem.beginSP, stem.endSP, h, s);
    }
    color.length+=coord.length-h;
    part->toEffectCoord=color.length;
    Vec2f slide = stem.slide;
    if(stem.height.length){
        slide.y+=stem.height[0];
    }
    doSlide(slide, coord, h);
}

void addTexture( const M_Vec3f& coord, const int& h, const int& tiOffset, M_Vec2f& texture, const Vec2f& srcSize, const Vec3f& srcPosition, const Vec2f& dstSize, const Vec2f& dstPosition, Matrix matrix){
    float scaleX=(1./srcSize.x)*dstSize.x;
    float scaleY=(1./srcSize.y)*dstSize.y;
    {
        for(int i=coord.length;	--i>=h;){
            int g=i+tiOffset;
            Vec3f crd=matrix*coord[i];
            texture[g].x=(crd.x-srcPosition.x)*scaleX+dstPosition.x;
            texture[g].y=(crd.y-srcPosition.y)*scaleY+dstPosition.y;
        }
    }
}

void TextureEffect::stemGenerateTri(Stem& stem, const ALathe& lathe, const int& k, Art*& part, M_Vec3f& coord, const M_Node& textureMapper, const int h, const int s){
    M_Vec2f& texCoord=this->texCoord.refValue();
    int texCoordBase= part->fromEffectCoord=texCoord.length;
    lathe.getBandAndYDist(stem.bands, Stem::bandDist, stem.yDistance);
    stem.convertYDistanceToHeight();

    genFlat( part, coord, part->triangle, texCoordBase, stem.xCentred, stem.xDistance[0], Stem::bandDist, stem.staves[0]+1, stem.beginSP, stem.endSP, h, s);
    texCoord.length+=coord.length-h;
    part->toEffectCoord=texCoord.length;
    //if(stem.mirror){
    //	stem.duplicateMirror(part);
    //}

    doSlide(stem.slide, coord, h);
    if(k<textureMapper.length){
        TextureMapper* texMapper=dynamic_cast<TextureMapper*>(textureMapper[k].p);
        Vec2f& size3d=texMapper->size3d.refValue();
        Vec3f& position3d=texMapper->position3d.refValue();
        if(texMapper->autoSize3d==true){
            size3d.x=stem.xDistance[0];
            size3d.y=stem.getYDistance();
            //position3d=Vec3f(stem.xDistance*-.5-stem.slide.x,-stem.slide.y,0);
            position3d=Vec3f(stem.xDistance[0]*-.5,0,0);
        }
        addTexture(coord, h, texCoordBase-h, texCoord, size3d, position3d, texMapper->size2d, texMapper->position2d, rotationMatrix(-texMapper->orientation));
    }
    if(stem.uWrap){
        genFlatJoinEnds(part, coord, part->triangle, stem.xCentred, stem.xDistance[0], Stem::bandDist, stem.staves[0], stem.beginSP, stem.endSP, h, s);
    }
}
void commonGetSegAndBandDist(const M_Int& bands, const M_Float& segDist, M_Float& bandDist){
    if(bands.length==segDist.length){
        bandDist.length=0;
        float yd=0;
        for(int k=0; k<segDist.length; k++){
            float step=segDist[k]/bands[k];
            bandDist+=yd;
            float h=yd;
            for(int j=1; j<bands[k]; j++){
                h+=step;
                bandDist+=h;
            }
            yd+=segDist[k];
        }
        bandDist+=yd;
    }
}

void MonoColorEffect::stemGenerateTri(NurbsPatch* patch, const Stem& stem, const int& k, Art*& part, M_Vec3f& coord, const M_Node& textureMapper, const int h, const int s){
    commonGetSegAndBandDist(stem.bands, stem.yDistance, Stem::bandDist);

    genFlat( part, coord, part->triangle, stem.xCentred, stem.xDistance[0], Stem::bandDist, stem.staves[0]+1, stem.bands, stem.beginSP, stem.endSP, h, s);

    if(stem.uWrap){
        genFlatJoinEnds(part, coord, part->triangle, stem.xCentred, stem.xDistance[0], Stem::bandDist, stem.staves[0], stem.beginSP, stem.endSP, h, s);
    }
    doSlide(stem.slide, coord, h);
}

void Stem::duplicateMirror(Art* part)const{ //this function should be replaced by how mirror dup is done in void Stem::doBuild(FuseSurface* parentFuse, NurbsPatch* patch);
    static CopyPart cp;
    cp.dest.refValue().length = 0;
    cp.dest.refValue() += part;
    cp.add.refValue() = true;
    if(xCentred == false){
        cp.stemNurbsMirror = true;
        cp.mirrorX = false;
    }
    else{
        cp.stemNurbsMirror = false;
        cp.mirrorX = true;
    }
    cp.doBuild();
}
void TextureEffect::stemGenerateTri(NurbsPatch* patch, const Stem& stem, const int& k, Art*& part, M_Vec3f& coord, const M_Node& textureMapper, const int h, const int s){
    M_Vec2f& texCoord=this->texCoord.refValue();
    commonGetSegAndBandDist(stem.bands, stem.yDistance, Stem::bandDist);
    Vec2f slide = stem.slide;
    if(stem.height.length){
        slide.y+=stem.height[0];
    }
    if(k<textureMapper.length){
        if(TextureMapper* texMapper=dynamic_cast<TextureMapper*>(textureMapper[k].p)){
            int texCoordBase= part->fromEffectCoord=texCoord.length;

            genFlat( part, coord, part->triangle, texCoordBase-h, stem.xCentred, stem.xDistance[0], Stem::bandDist, stem.staves[0]+1, stem.beginSP, stem.endSP, h, s);
            doSlide(slide, coord, h);

            texCoord.length+=coord.length-h;
            part->toEffectCoord=texCoord.length;

            Vec2f& size3d=texMapper->size3d.refValue();
            Vec3f& position3d=texMapper->position3d.refValue();
            if(texMapper->autoSize3d==true){
                size3d.x=stem.xDistance[0];
                size3d.y=stem.getYDistance();
    //			position3d=Vec3f(stem.xDistance*-.5-stem.slide.x,-stem.slide.y,0);
                if(stem.xCentred){
                    position3d=Vec3f(stem.xDistance[0]*-.5,0,0);
                }
            }
            if(stem.mirror){
                stem.duplicateMirror(part);
            }
            addTexture(coord, h, texCoordBase-h, texCoord, size3d, texMapper->position3d.refValue(), texMapper->size2d.refValue(), texMapper->position2d.refValue(),rotationMatrix(-texMapper->orientation));

            if(stem.uWrap){
                genFlatJoinEnds(part, coord, part->triangle, stem.xCentred, stem.xDistance[0], Stem::bandDist, stem.staves[0], stem.beginSP, stem.endSP, h, s);
                doSlide(slide, coord, h);
            }
        }
        else goto here;
    }
    else {
here:
        if(stem.uWrap){
            genFlatJoinEnds(part, coord, part->triangle, stem.xCentred, stem.xDistance[0], Stem::bandDist, stem.staves[0], stem.beginSP, stem.endSP, h, s);
        }
        else {
            genFlat( part, coord, part->triangle, stem.xCentred, stem.xDistance[0], Stem::bandDist, stem.staves[0]+1, stem.bands, stem.beginSP, stem.endSP, h, s);
        }
        texCoord.length+=coord.length-h;
        doSlide(slide, coord, h);
    }
}
void ColorEffect::stemGenerateTri(NurbsPatch* patch, const Stem& stem, const int& k, Art*& part, M_Vec3f& coord, const M_Node& textureMapper, const int h, const int s){
    M_Color3f& color=this->color.refValue();
    int texCoordBase = part->fromEffectCoord=color.length;

    commonGetSegAndBandDist(stem.bands, stem.yDistance, Stem::bandDist);
    M_Triangle& triangle = part->triangle;

    if(stem.uWrap){
        int toTri = triangle.length*3;
        genFlatJoinEnds(part, coord, part->triangle, stem.xCentred, stem.xDistance[0], Stem::bandDist, stem.staves[0], stem.beginSP, stem.endSP, h, s);
        int idxDiff = texCoordBase - h;
        for(int i=triangle.length*3; --i>=toTri;){
            triangle.refTexIndex(i)=triangle.refIndex(i)+idxDiff;
        }
    }
    else{
        genFlat( part, coord, part->triangle, texCoordBase-h, stem.xCentred, stem.xDistance[0], Stem::bandDist, stem.staves[0]+1, stem.beginSP, stem.endSP, h, s);
    }
    color.length+=coord.length-h;
    part->toEffectCoord=color.length;
    Vec2f slide = stem.slide;
    if(stem.height.length){
        slide.y+=stem.height[0];
    }
    doSlide(slide, coord, h);

    if(stem.mirror && stem.mirrorStem == 0){
        stemDupMirrorTri(part, coord, triangle, h, texCoordBase, s, coord.length - h, color.length - texCoordBase, triangle.length - s);
    }

    //patch->doColorForStem(part, coord, h, texCoordBase); why did we want this instead of the code below?
    if(k < textureMapper.length){
        if(ColorSweep* colorSweep=dynamic_cast<ColorSweep*>(textureMapper[k].p)){
            colorSweep->doColorForStem(part, coord, h, texCoordBase);
        }
    }
}

void TextureEffect::stem2GenerateTri(NurbsPatch* patch, const Stem& stem, const int& k, Art*& part, M_Vec3f& coord, const M_Node& textureMapper, const int h, const int s){
    M_Vec2f& texCoord = this->texCoord.refValue();
    int texCoordBase = part->fromEffectCoord = texCoord.length;

    M_Triangle& triangle = part->triangle;

    genFlatStem2(part, coord, part->triangle, texCoordBase-h, false, stem.uWrap, stem.vWrap, stem.xDistance, stem.yDistance, stem.staves, stem.bands, stem.cells.x,  stem.cells.y, h, s);
    doSlide(stem.slide, coord, h);

    texCoord.length += coord.length - h;
    part->toEffectCoord = texCoord.length;

    doSlide(stem.slide, coord, h);

    if(k < textureMapper.length){
        if(TextureMapper* texMapper=dynamic_cast<TextureMapper*>(textureMapper[k].p)){

            Vec2f& size3d=texMapper->size3d.refValue();
            Vec3f& position3d=texMapper->position3d.refValue();
            if(texMapper->autoSize3d==true){
                size3d.x=stem.xDistance[0];
                size3d.y=stem.getYDistance();
    //			position3d=Vec3f(stem.xDistance*-.5-stem.slide.x,-stem.slide.y,0);
                if(stem.xCentred){
                    position3d=Vec3f(stem.xDistance[0]*-.5,0,0);
                }
            }
            if(stem.mirror){
                stem.duplicateMirror(part);
            }
            addTexture(coord, h, texCoordBase-h, texCoord, size3d, texMapper->position3d.refValue(), texMapper->size2d.refValue(), texMapper->position2d.refValue(),rotationMatrix(-texMapper->orientation));
        }
    }
}

void MonoColorEffect::stem2GenerateTri(NurbsPatch* patch, const Stem& stem, const int& k, Art*& part, M_Vec3f& coord, const M_Node& textureMapper, const int h, const int s){
}

void ColorEffect::stem2GenerateTri(NurbsPatch* patch, const Stem& stem, const int& k, Art*& part, M_Vec3f& coord, const M_Node& textureMapper, const int h, const int s){
    M_Color3f& color = this->color.refValue();
    int texCoordBase = part->fromEffectCoord=color.length;

    //commonGetSegAndBandDist(stem.bands, stem.yDistance, Stem::bandDist);
    M_Triangle& triangle = part->triangle;

    genFlatStem2(part, coord, part->triangle, texCoordBase-h, false, stem.uWrap, stem.vWrap, stem.xDistance, stem.yDistance, stem.staves, stem.bands, stem.cells.x,  stem.cells.y, h, s);

    color.length += coord.length - h;
    part->toEffectCoord = color.length;

    doSlide(stem.slide, coord, h);
    //stemDupMirrorTri(part, coord, triangle, h, texCoordBase, s, coord.length - h, color.length - texCoordBase, triangle.length - s);
	//patch->doColorForStem(part, coord, h, texCoordBase);

    if(k < textureMapper.length){
        if(ColorSweep* colorSweep=dynamic_cast<ColorSweep*>(textureMapper[k].p)){
            colorSweep->doColorForStem(part, coord, h, texCoordBase);
        }
    }

}

void inline doBnspe(Seamless*& smls, Art*& prt, M_Vec3f& dstCoord, int& h, NurbsPatch*& patch){
    if(smls->bnspe){
        M_NurbsPatch& patchOwner = prt->patchOwner;
        M_Vec2f& uvCoord = prt->uvCoord.refValue();
        uvCoord.length = dstCoord.length;
        patchOwner.length = dstCoord.length;
        for(int k = uvCoord.length; --k >= h;){
            uvCoord[k] = Vec2f(dstCoord[k].x, dstCoord[k].y);
            patchOwner[k] = patch;
        }
    }
}

void convertPoleToPole(M_Vec3f& coord){
    for(int i = coord.length; --i>=0;){
        coord[i] = Vec3f(1-coord[i].x, 1-coord[i].y, 0);
    }
}
void convertPoleToPole(M_Vec2f& coord){
    for(int i = coord.length; --i>=0;){
        coord[i] = Vec2f(1-coord[i].x, 1-coord[i].y);
    }
}
void convertMercatorToNorthPole(M_Vec3f& coord){
    for(int i = coord.length; --i>=0;){
        float ang = coord[i].x * (PI * 2);
        float dist = 1 - coord[i].y;
        coord[i] = Vec3f(-dist*sin(ang), dist*cos(ang), 0);
    }
}
void convertMercatorToSouthPole(M_Vec3f& coord){
    for(int i = coord.length; --i>=0;){
        float ang = coord[i].x * (PI * 2);
        float dist = -coord[i].y;
        coord[i] = Vec3f(dist*sin(ang), dist*cos(ang), 0);
    }
}
void convertMercatorToSouthPole(M_Vec2f& coord){
    for(int i = coord.length; --i>=0;){
        float ang = coord[i].x * (PI * 2);
        float dist = -coord[i].y;
        coord[i] = Vec2f(dist*sin(ang), dist*cos(ang));
    }
}
void convertNorthPoleToMercator(M_Vec3f& coord){
    for(int i = coord.length; --i>=0;){
        float x = coord[i].x;
        float y = coord[i].y;
        float hyp=sqrt(x*x + y*y);
        if(hyp==0){
            coord[i]=Vec3f(0,1,0);
            continue;
        }
        if(hyp>=1){
            coord[i]=Vec3f();
            continue;
        }
        float angle = acos(y/hyp);

        if(x > 0){
            angle = PI*2 - angle;
        }
        if(angle < 0){
            angle=0;
        }
        coord[i] = Vec3f(angle/(PI*2), 1-hyp, 0);
    }
}
void convertSouthPoleToMercator(M_Vec3f& coord){
    for(int i = coord.length; --i>=0;){
        float x = coord[i].x;
        float y = coord[i].y;
        float hyp=sqrt(x*x + y*y);
        if(hyp==0){
            coord[i]=Vec3f();
            continue;
        }
        if(hyp>=1){
            coord[i]=Vec3f(0,1,0);
            continue;
        }
        float angle = asin(y/hyp) + PI*.5;

        if(x > 0){
            angle = PI*2 - angle;
        }
        if(angle < 0){
            angle=0;
        }
        coord[i] = Vec3f(angle/(PI*2), hyp, 0);
    }
}
void convertSouthPoleToMercator(M_Vec2f& coord){
    for(int i = coord.length; --i>=0;){
        float x = coord[i].x;
        float y = coord[i].y;
        float hyp=sqrt(x*x + y*y);
        if(hyp==0){
            coord[i]=Vec2f();
            continue;
        }
        if(hyp>=1){
            coord[i]=Vec2f(0,1);
            continue;
        }
        float angle = asin(y/hyp) + PI*.5;

        if(x > 0){
            angle = PI*2 - angle;
        }
        if(angle < 0){
            angle=0;
        }
        coord[i] = Vec2f(angle/(PI*2), hyp);
    }
}



void convertMercatorToNorthPole(M_Vec2f& coord){
    for(int i = coord.length; --i>=0;){
        float ang = coord[i].x * (PI * 2);
        float dist = 1 - coord[i].y;
        coord[i] = Vec2f(-dist*sin(ang), dist*cos(ang));
    }
}
void convertNorthPoleToMercator(M_Vec2f& coord){
    for(int i = coord.length; --i>=0;){
        float x = coord[i].x;
        float y = coord[i].y;
        float hyp=sqrt(x*x + y*y);
        if(hyp==0){
            coord[i]=Vec2f(0,1);
            continue;
        }
        if(hyp>=1){
            coord[i]=Vec2f();
            continue;
        }
        float angle = acos(y/hyp);

        if(x > 0){
            angle = PI*2 - angle;
        }
        if(angle < 0){
            angle=0;
        }
        coord[i] = Vec2f(angle/(PI*2), 1-hyp);
    }
}
Vec2f convertNorthPoleToMercator(const Vec2f& coord){
    float x = coord.x;
    float y = coord.y;
    float hyp=sqrt(x*x + y*y);
    if(hyp==0){
        return Vec2f(0,1);
    }
    if(hyp>=1){
        return Vec2f();
    }
    float angle = acos(y/hyp);

    if(x > 0){
        angle = PI*2 - angle;
    }
    if(angle < 0){
        angle=0;
    }
    return Vec2f(angle/(PI*2), 1-hyp);
}
Vec2f convertSouthPoleToMercator(const Vec2f& coord){
    float x = coord.x;
    float y = coord.y;
    float hyp=sqrt(x*x + y*y);
    if(hyp==0){
        return Vec2f();
    }
    if(hyp>=1){
        return Vec2f(0,1);
    }
    float angle = asin(y/hyp) + PI*.5;

    if(x > 0){
        angle = PI*2 - angle;
    }
    if(angle < 0){
        angle=0;
    }
    return Vec2f(angle/(PI*2), hyp);
}

void textureMapTriIndex(TextureMapper* texMapper, M_Triangle& triangle, const int triOffset, M_Vec2f& texCoord, const Vec2f& srcSize, const Vec3f& srcPosition, const Vec2f& dstSize, const Vec2f& dstPosition, Matrix matrix){
    float scaleX=(1. / srcSize.x) * dstSize.x;
    float scaleY=(1. / srcSize.y) * dstSize.y;

    for(int i = triangle.getTriangleLength(); --i >= triOffset;){
        Art* part = triangle.refPart(i);
        Vec3f crd = matrix * part->coord.refValue()[triangle.refIndex(i)];

        texCoord[triangle.refTexIndex(i)] = Vec2f((crd.x - srcPosition.x) * scaleX + dstPosition.x, (crd.y - srcPosition.y) * scaleY + dstPosition.y);
    }
}

void Stem::doBuild(FuseSurface* parentFuse, NurbsPatch* patch){
    Seamless* smls = patch->smls;
    if(store == false && staves.length > 1){
        for(int k = 0; k < part.length; k++){
            Art* prt=part.refValue()[k];
            M_Vec3f& dstCoord=prt->coord.refValue();
            M_Triangle& triangle = prt->triangle;
            if(add == false && (bands.length || store)){
                dstCoord.length=0;
                triangle.length=0;
            }
            int h = dstCoord.length;
            beginCoordOffset = h;
            beginTriOffset = triangle.length;

            int triOffset = triangle.length;

            Effect* effect = prt->workEffect;
            effect->stem2GenerateTri(patch, *this, k, prt, dstCoord, textureMapper, h, triangle.length);
            doBnspe(smls, prt, dstCoord, h, patch);

            static CopyPart cp;
            int c = 0;

            if(mould){
                if(prt->smls->surfaceNormals){
                    patch->generateWithNormals(dstCoord, prt->normal.refValue(), h);
                }
                else {
                    patch->generate2(dstCoord, h);
                }
            }
            else if(northPole && k>=source.length){
                convertMercatorToNorthPole(dstCoord);
            }
            else if(southPole && k>=source.length){
                convertMercatorToSouthPole(dstCoord);
            }
            if(insideOut){
                prt->insideOut(beginTriOffset);
            }
            if(prt->smls->buildNormals){
                prt->genNormals(h, beginTriOffset);
            }
            addRelevantPart(prt->relevantPart, prt);
        }
        return;
    }


    int xPntCnt=staves.refValue()[0];
    if(xPntCnt==0) return;
    for(int i=0; i<bands.length; i++){
        if(bands[i]==0) return;
    }
    bool insideOut=this->insideOut;
    Rotation orientation = this->orientation;
    const PartChild& source=this->source;
    const M_Node& textureMapper=this->textureMapper;

//	if(ownership.length){
//		buildMultiOwnedRings(surface);
//		return;
//	}

    for(int k = 0; k < part.length; k++){
        Art* prt=part.refValue()[k];
        M_Vec3f& dstCoord=prt->coord.refValue();
        M_Triangle& triangle = prt->triangle;
        if(add == false && (bands.length || store)){
            dstCoord.length=0;
            triangle.length=0;
        }
        int h = dstCoord.length;
        beginCoordOffset = h;
        beginTriOffset = triangle.length;

        int triOffset = triangle.length;
        Effect* effect=prt->workEffect;
        static CopyPart cp;
        int c = 0;

        if(store){
            if(k<source.length){		//triangles come from GeomSource node
                cp.source.refValue().length = 0;
                cp.dest.refValue().length = 0;

                cp.source.refValue()+=source[k];
                cp.dest.refValue() += prt;
                cp.add.refValue() = add;

                if(TextureEffect* textureEffect = dynamic_cast<TextureEffect*>(effect)){
                    c = textureEffect->texCoord.length;
                    cp.doBuild();
                    doSlide(slide, dstCoord, h);
                    int cLen = textureEffect->texCoord.length;
                    if(mirror && mirrorStem == 0){
                        textureEffect->stemDupMirrorTri(prt, dstCoord, triangle, h, c, beginTriOffset, dstCoord.length - h, cLen - c, triangle.length - beginTriOffset);
                    }
                }
                else if(ColorEffect* colorEffect = dynamic_cast<ColorEffect*>(effect)){
                    c = colorEffect->color.length;
                    cp.doBuild();
                    doSlide(slide, dstCoord, h);
                    int cLen = colorEffect->color.length;
                    if(mirror && mirrorStem == 0){
                        colorEffect->stemDupMirrorTri(prt, dstCoord, triangle, h, c, beginTriOffset, dstCoord.length - h, cLen - c, triangle.length - beginTriOffset);
                    }
                }
            }
            else{//use local store
                c = effect->stemCopyLocalTri(this, prt, dstCoord, h, beginTriOffset);
                doBnspe(smls, prt, dstCoord, h, patch);
            }
            if(mould){
                if(northPole){
                    convertNorthPoleToMercator(dstCoord);
                }
                else if(southPole){
                    convertSouthPoleToMercator(dstCoord);
                }
            }
            //effect->processGeomSource(k, prt, dstCoord, *dynamic_cast<GeomSource*>(this->source.refValue()[k].getRawPtr()), textureMapper);
            goto here;
        }
        if(bands.length==0){	//triangles come from existing part
            if(mirror && mirrorStem){
                h = dstCoord.length;
                effect->stemGenerateTri(patch, *mirrorStem, k, prt, dstCoord, textureMapper, h, triangle.length);
                doBnspe(smls, prt, dstCoord, h, patch);
                goto here2;
            }
            doSlide(slide, dstCoord, h);
here:
            if(k < textureMapper.length){
                if(TextureEffect* textureEffect = dynamic_cast<TextureEffect*>(effect)){
                    if(TextureMapper* texMapper = dynamic_cast<TextureMapper*>(textureMapper[k].p)){
                        if(texMapper->autoSize3d){
                            prt->findSetVertexRange(*texMapper);
                        }
                        textureMapTriIndex(texMapper, triangle, triOffset * 3, textureEffect->texCoord.refValue(), texMapper->size3d, texMapper->position3d.refValue(), texMapper->size2d.refValue(), texMapper->position2d.refValue(),rotationMatrix(-texMapper->orientation));
                    }
                }
                else if(ColorEffect* colorEffect = dynamic_cast<ColorEffect*>(effect)){
                    if(ColorSweep* colorSweep = dynamic_cast<ColorSweep*>(textureMapper[k].p)){
                        colorSweep->doColorForStem(triangle, dstCoord, colorEffect->color.refValue(), h, c, triOffset);
                    }
                }
            }


            //if(effect->getType()==TEXTUREEFFECT && k<textureMapper.length){
            //	M_Vec2f& texCoord=dynamic_cast<TextureEffect*>(effect)->texCoord.refValue();
            //	TextureMapper* texMapper=dynamic_cast<TextureMapper*>(textureMapper[k].p);
            //	int texCoordBase= prt->fromEffectCoord=texCoord.length;

            //	texCoord.length+=dstCoord.length-h;
            //	prt->toEffectCoord=texCoord.length;

            //	const Vec2f& size3d=texMapper->size3d;

            //	if(texMapper->autoSize3d){
            //		prt->findSetVertexRange(*texMapper);
            //	}
            //	doSlide(slide, dstCoord, h);
            //	addTexture(dstCoord, h, texCoordBase-h, texCoord, size3d, texMapper->position3d.refValue(), texMapper->size2d.refValue(), texMapper->position2d.refValue(),rotationMatrix(-texMapper->orientation));
            //	M_Triangle& triangle=prt->triangle;
            //	for(int i=triangle.length*3; --i>=0; ){
            //		triangle.refTexIndex(i)=texCoordBase+triangle.refIndex(i);
            //	}
            //}
            //else{
            //	doSlide(slide, dstCoord, h);
            //}
        }
        else{//triangles are generated using Stem yDistance
            h = dstCoord.length;
            effect->stemGenerateTri(patch, *this, k, prt, dstCoord, textureMapper, h, triangle.length);
            doBnspe(prt->smls, prt, dstCoord, h, patch);		}
here2:
/*
        if(slide != Vec2f(0,0)){
            for(int i = dstCoord.length; --i>=h;){
                Vec3f v = dstCoord[i];
                v.x += slide.x;
                v.y += slide.y;
                dstCoord[i] = v;
            }
        }
*/
        if(orientation!=Rotation()){
            //if(orientation.refValue()==Rotation(1, 0, 0, -1.570796)){
            if(orientation.getAxis()==Vec3f(1, 0, 0)&&fabs(orientation.angle-(-PI*.5))<.000001){
                for(int i=dstCoord.length; --i>=h;){
                    dstCoord[i].z = -dstCoord[i].y;
                    dstCoord[i].y=0;
                }
            }
            else {
                Rotation orientation=this->orientation.refValue();
                for(int i=dstCoord.length; --i>=h;){
                    dstCoord[i] = orientation * dstCoord[i];
                }
            }
        }
        if(parentFuse && k == 0){
            parentFuse->storeV(dstCoord);
            if(parentFuse->evolve){
                M_Vec3f& parentFlatCoord = parentFuse->parentFlatCoord;
                parentFlatCoord.length = parentFuse->qty;
                for(int i = parentFuse->qty; --i>=0;){
                    parentFlatCoord[i] = dstCoord[i];
                }
            }
        }
        if(mould){
            if(prt->smls->surfaceNormals){
                patch->generateWithNormals(dstCoord, prt->normal.refValue(), h);
            }
            else {
                patch->generate2(dstCoord, h);
            }
        }
        else if(northPole && k>=source.length){
            convertMercatorToNorthPole(dstCoord);
        }
        else if(southPole && k>=source.length){
            convertMercatorToSouthPole(dstCoord);
        }
        if(insideOut){
            prt->insideOut(beginTriOffset);
        }
        if(prt->smls->buildNormals){
            prt->genNormals(h, beginTriOffset);
        }
        addRelevantPart(prt->relevantPart, prt);
    }

    M_Node& m_fuse = fuse.refValue();
    int len = m_fuse.length;
    for(int k = 0; k < len; k++){
        dynamic_cast<FuseSurface*>(m_fuse[k].p)->doFuse(*patch, *this);
    }
}

void Stem::doBuild(){
    M_Node* generator=0;

    Generator* surface=dynamic_cast<Generator*>(this->surface.p);
    if(SurfaceGenerator* surfaceGenerator=dynamic_cast<SurfaceGenerator*>(surface)){
        generator = &surfaceGenerator->generator.refValue();
    }

    if(store == false && staves.length > 1){
        for(int k = 0; k < part.length; k++){
            Art* prt=part.refValue()[k];
            M_Vec3f& dstCoord = prt->coord.refValue();
            M_Triangle& triangle = prt->triangle;
            int h = 0;
            h = dstCoord.length;
            Effect* effect = prt->workEffect;
			NurbsPatch* patch = 0;
            effect->stem2GenerateTri(patch, *this, k, prt, dstCoord, textureMapper, h, triangle.length);

            addRelevantPart(prt->relevantPart, prt);
            if(surface && mould){
                surface->generate(dstCoord, h);
            }
            else if(northPole){
                convertMercatorToNorthPole(dstCoord);
            }
            else if(southPole){
                convertMercatorToSouthPole(dstCoord);
            }
            if(insideOut){
                prt->insideOut(beginTriOffset);
            }
        }
        return;
    }


    int xPntCnt=staves.refValue()[0];
    if(xPntCnt==0) return;
    for(int i=0; i<bands.length; i++){
        if(bands[i]==0) return;
    }
    bool insideOut=this->insideOut;
    Rotation orientation = this->orientation;
    const PartChild& source=this->source;
    const M_Node& textureMapper=this->textureMapper;
    ALathe* lathe;

//	if(ownership.length){
//		buildMultiOwnedRings(surface);
//		return;
//	}

    for(int k = 0; k < part.length; k++){
        Art* prt=part.refValue()[k];
        M_Vec3f& dstCoord=prt->coord.refValue();
        M_Triangle& triangle = prt->triangle;
        int h = 0;
        if(add == false && bands.length){
            dstCoord.length = 0;
            triangle.length = 0;
        }
        bool genSource = true;
        Effect* effect = prt->workEffect;
        if(k < source.length){		//triangles come from GeomSource node
            static CopyPart cp;
            cp.source.refValue().length=0;
            cp.dest.refValue().length=0;

            cp.source.refValue()+=source[k];
            cp.dest.refValue()+=prt;
            cp.add.refValue()=add;
            cp.doBuild();
            //effect->processGeomSource(k, prt, dstCoord, *dynamic_cast<GeomSource*>(this->source.refValue()[k].getRawPtr()), textureMapper);
            goto here;
        }
        if(bands.length == 0){	//triangles come from existing part
here:
            if(effect->getType()==TEXTUREEFFECT && k<textureMapper.length){
                M_Vec2f& texCoord=dynamic_cast<TextureEffect*>(effect)->texCoord.refValue();
                TextureMapper* texMapper=dynamic_cast<TextureMapper*>(textureMapper[k].p);
                int texCoordBase= prt->fromEffectCoord=texCoord.length;

                texCoord.length+=dstCoord.length-h;
                prt->toEffectCoord=texCoord.length;

                const Vec2f& size3d=texMapper->size3d;

                if(texMapper->autoSize3d){
                    prt->findSetVertexRange(*texMapper);
                }
                doSlide(slide, dstCoord, h);
                addTexture(dstCoord, h, texCoordBase-h, texCoord, size3d, texMapper->position3d.refValue(), texMapper->size2d.refValue(), texMapper->position2d.refValue(),rotationMatrix(-texMapper->orientation));
                for(int i=triangle.length*3; --i>=0; ){
                    triangle.refTexIndex(i)=texCoordBase+triangle.refIndex(i);
                }
            }
            else{
                doSlide(slide, dstCoord, h);
            }
        }
        else if( sync && generator && generator->length &&	(lathe=dynamic_cast<ALathe*>((*generator)[0].getRawPtr()))){ //triangles are generated using CCLathe distance
            h = dstCoord.length;
            lathe->calcClose();
            effect->stemGenerateTri(*this, *lathe, k, prt, dstCoord, textureMapper, h, triangle.length);
        }
        else{//triangles are generated using Stem yDistance
            h = dstCoord.length;
			NurbsPatch* patch = 0;
            effect->stemGenerateTri(patch, *this, k, prt, dstCoord, textureMapper, h, triangle.length);
        }
/*
        if(slide != Vec2f(0,0)){
            for(int i = dstCoord.length; --i>=h;){
                Vec3f v = dstCoord[i];
                v.x += slide.x;
                v.y += slide.y;
                dstCoord[i] = v;
            }
        }
*/
        if(orientation!=Rotation()){
            //if(orientation.refValue()==Rotation(1, 0, 0, -1.570796)){
            if(orientation.getAxis()==Vec3f(1, 0, 0)&&fabs(orientation.angle-(-PI*.5))<.000001){
                for(int i=dstCoord.length; --i>=h;){
                    dstCoord[i].z = -dstCoord[i].y;
                    dstCoord[i].y=0;
                }
            }
            else {
                Rotation orientation=this->orientation.refValue();
                for(int i=dstCoord.length; --i>=h;){
                    dstCoord[i] = orientation * dstCoord[i];
                }
            }
        }
        if(surface && mould){
            surface->generate(dstCoord, h);
        }
        else if(northPole){
            convertMercatorToNorthPole(dstCoord);
        }
        else if(southPole){
            convertMercatorToSouthPole(dstCoord);
        }
        if(insideOut){
            prt->insideOut(beginTriOffset);
        }
        addRelevantPart(prt->relevantPart, prt);
    }
}

void Stem::NorthPole::operator=(bool const& par){
    M_Node& m_fuse = home.fuse.refValue();
    int len = m_fuse.length;
    for(int k = 0; k < len; k++){
        FuseSurface* fuseSurface = dynamic_cast<FuseSurface*>(m_fuse[k].p);
        M_Vec2f& baseCoord = fuseSurface->baseCoord;
        if(home.southPole == false && par == true){
            convertMercatorToNorthPole(baseCoord);
        }
        else if(home.northPole == true && par == false){
            convertNorthPoleToMercator(baseCoord);
        }
        else if(home.southPole == true && par == true){
            convertPoleToPole(baseCoord);
            home.southPole.refValue()=false;
        }
    }
    refValue()=par;
}
void Stem::SouthPole::operator=(bool const& par){
    M_Node& m_fuse = home.fuse.refValue();
    int len = m_fuse.length;
    for(int k = 0; k < len; k++){
        FuseSurface* fuseSurface = dynamic_cast<FuseSurface*>(m_fuse[k].p);
        M_Vec2f& baseCoord = fuseSurface->baseCoord;
        if(home.northPole == false && par == true){
            convertMercatorToSouthPole(baseCoord);
        }
        else if(home.southPole == true && par == false){
            convertSouthPoleToMercator(baseCoord);
        }
        else if(home.northPole == true && par == true){
            convertPoleToPole(baseCoord);
            home.northPole.refValue()=false;
        }
    }
    refValue()=par;
}

bool Stem::illegalImmediateRef(const Art* p, M_Art& accumOthers)const{
    return commonIllegalImmediateRef(part, p, accumOthers);
}

void Stem::Mirror::operator=(const bool& par){
    refValue() = par;
    if(home.part.length){
        home.part[0]->smls->populateMirrorNodes();
    }
    else if(home.source.length){
        home.source[0]->smls->populateMirrorNodes();
    }
}
