/* Seamless3d

Copyright (C) 2001-2017 Graham John Perrett (AKA thyme) thyme@seamless3d.com www.seamless3d.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this
software and associated documentation files (the "Software"), to deal in the Software
without restriction, including without limitation the rights to use, copy, modify,
merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to the following
conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT
OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.
*/

#include "seamless3d.h"
float g_version = 3.007;
float g_file_version = 0;
float g_remVersionLoaded = 0;

bool g_outputMode = false;
bool g_writeX3DXML = false;
bool g_writeX3DVRMLClassic = false;
bool g_outputPovRay = false;

bool g_genHAnimNodes = false;
bool g_genIdxTriSetNodes = false;
bool g_genContactSBVT = false;
bool g_genNormals = false;
bool g_genColorIndex = false;
bool g_joinAndTransferActive = true;
extern bool g_writeObj;
extern Ptr<TransformCP> g_transformCP;
int g_userProtoBegin = 0;
extern Node* g_parent;
Matrix getViewMatrix();
extern float g_viewDist;

float screenRatio();
M_Int g_tmpIndex;
M_Vec3f g_tmpVec3f;
M_<Download*> g_downloadQue;


#ifdef SEAMLESS3DGUI
extern int g_sceneTreeNodeIndex;
void showBuildError(String str);
Node* newFunction();
Node* newMemberFunction();
Node* newAvatarSocket();
void outputUserScripts(String& str, bool& newList);
ScenePtr g_scene = new Scene;
ScenePtr g_load_scene;

const char* mFieldErrorMsg = "Exception in MField class";
const char* protoInitilaiseStr2=0;
void initCompile();
void decodeFunctions(const String& str);
int GestureGroup::numOfGestures=0;
bool GestureGroup::addProto=false;
int RingBuild::quadrant=0;
BezierLathePtr ConeBuild::lathe = new BezierLathe;


Vec3f FuseSurface::baseVisPoint;
Vec3f FuseSurface::thisVisPoint;
bool Art::tugRadialTugSize = false;

void setAllIns(Stem* stem, const bool inState);
void setAllIns(NurbsPatch* patch, const bool inState);
void setAllIns(M_Build& m_build, const bool inState);
void processAllPatches(M_<NurbsPatch*>& cpNodeList);

#endif
//M_Art g_changePart;
//M_Node g_changeNode;
//ScriptPtr g_glue;

char floatFormat[]={"%.6f"};
char floatFormat2[]={" %.6f"};
char g_vrmlSquareOpenBracket[]={"["};
char g_vrmlSquareCloseBracket[]={"]"};
char g_vrmlQuote[]={"\""};

const char g_exposedField[]="exposedField";
const char g_eventIn[]="eventIn";
const char g_eventOut[]="eventOut";
const char g_field[]="field";

const char g_inputOutput[]="inputOutput";
const char g_inputOnly[]="inputOnly";
const char g_outputOnly[]="outputOnly";
const char g_initializeOnly[]="initializeOnly";

const char* g_exposedFieldPtr=g_exposedField;
const char* g_eventInPtr=g_eventIn;
const char* g_eventOutPtr=g_eventOut;
const char* g_fieldPtr=g_field;

void changeFloatFormat(int numOfDigits){
	char chr='0'+numOfDigits;
	floatFormat[2]=floatFormat2[3]=chr;
}

M_String nodeLookUpStr;
M_String containerFieldLookUpStr;
char g_exceptioninNode[]={"exception in node"};

extern Ptr<DrawNurbsLine> g_drawNurbsLine;

bool g_exitBuildFlag=false;
extern int numOfBuiltInNodes;
extern SettingsPtr g_settings; //legacy
void houseKeepVisualClue();
void swapCPPointStorage(NurbsPatch* patch);
void commonGetSegAndBandDist(const M_Int& bands, const M_Float& segDist, M_Float& bandDist);
void genIndicesJoinEndsBeginSP(Art*& partA, Art*& partB, M_Triangle& triangle, const int& staves, const int& c);
void genIndicesJoinEndsEndSP(Art*& partA, Art*& partB, M_Triangle& triangle, const int& staves, const int& c);
void genIndicesJoinEnds(Art*& part, M_Triangle& triangle, const int& coordLen, int& numStaves, int& numRings, int bottomIsPoint, int topIsPoint, const int& h, const int& s);
void genIndicesJoinEnds(Art*& partA, Art*& partB, M_Triangle& triangle, const int& staves, const int& c);
void addRelevantPart(M_Art& relevant, Art*& part);
void removeNonRelevantParts(Art* source);
bool verifyPartRelivanceToTriangles(Art* source, M_Triangle& triangle);

const char fieldTypeLookUpStr[FIELDLOOKUPLEN][11]={
	"SFBool",
	"SFColor",
	"SFFloat",
	"SFImage",
	"SFInt32",
	"SFNode",
	"SFRotation",
	"SFString",
	"SFTime",
	"SFVec2f",
	"SFVec3f",
	"MFBool",
	"MFColor",
	"MFFloat",
	"MFImage",
	"MFInt32",
	"MFNode",
	"MFRotation",
	"MFString",
	"MFTime",
	"MFVec2f",
	"MFVec3f"
};


bool isLegalNameChar(const char& ch){
	if( ch>='a'&&ch<='z')return true;
	if( ch>='/'&&ch<='Z')return true;
	if(	ch<=' '||
		ch==0x7f||
		ch=='\''||
		ch=='\"'||
		ch=='#'||
		ch==','||
		ch=='.'||
		ch=='['||
		ch==']'||
		ch=='\\'||
		ch=='{'||
		ch=='}'
		){
		return false;
	}
	return true;
}

template<> void SFBool::resize(int newLen){throw this;}
void SFColor::resize(int newLen){throw this;}
template<> void SFFloat::resize(int newLen){throw this;}
template<> void SFImage::resize(int newLen){throw this;}
template<> void SFInt32::resize(int newLen){throw this;}
void SFNode::resize(int newLen){throw this;}
void SFRotation::resize(int newLen){throw this;}
void SFString::resize(int newLen){throw this;}
template<> void SFTime::resize(int newLen){throw this;}
void SFVec2f::resize(int newLen){throw this;}
void SFVec3f::resize(int newLen){throw this;}

template<> void MFBool::resize(int newLen){refValue().length=newLen;}
template<> void MFColor::resize(int newLen){refValue().length=newLen;}
template<> void MFFloat::resize(int newLen){refValue().length=newLen;}
template<> void MFImage::resize(int newLen){refValue().length=newLen;}
template<> void MFInt32::resize(int newLen){refValue().length=newLen;}
template<> void MFNode::resize(int newLen){refValue().length=newLen;}
template<> void MFRotation::resize(int newLen){refValue().length=newLen;}
template<> void MFString::resize(int newLen){refValue().length=newLen;}
template<> void MFTime::resize(int newLen){refValue().length=newLen;}
template<> void MFVec2f::resize(int newLen){refValue().length=newLen;}
template<> void MFVec3f::resize(int newLen){refValue().length=newLen;}

String Field::getTypeStr()const{
	return fieldTypeLookUpStr[getType()];
}
M_<ScriptPtr> g_outputScriptStack;
M_Bool g_outputScriptStackExist;

extern Script* g_outputScript;
Script* getOutputScriptPtr(){
	return g_outputScript;
}

inline bool compareName(const M_Properties& properties, const String str, int& i){
	for(i=properties.length; --i>=0; ){
		if(properties[i].name==str){
			return true;
		}
	}
	return false;
}

template<>
int M_Properties::getIndex( String str)const{
	int i;
	if(compareName(*this, str, i)){
		return i;
	}
	if(str.length>4&&str.substr(0,4)=="set_"){
		if(compareName(*this, str.substr(4), i)){
			return i;
		}
	}
	if(str.length>8&&str.substr(str.length-8)=="_changed"){
		if(compareName(*this, str.substr(0,str.length-8), i)){
			return i;
		}
	}
	throw this;
}

template <class T> void String::cat(const M_<T>& par){
	requestMoreIfNeeded(amountToRequest<T>()*par.length + 5);

	if(par.length==1){
		(*this)+=par[0];
	}
	else {
		(*this)+=g_vrmlSquareOpenBracket;
		if(par.length > 0){
			(*this)+=par[0];
			for(int i=1;i<par.length;i++){
//				(*this)+=',';
				(*this)+=' ';
				(*this)+=par[i];
			}
		}
		(*this)+=g_vrmlSquareCloseBracket;
	}
}
template <> void String::cat<S_Triangle>(const M_<S_Triangle>& par){
	if(par.length==1){
		(*this)+=par[0];
	}
	else {
		(*this)+=g_vrmlSquareOpenBracket;
		if(par.length > 0){
			(*this)+=par[0];
			for(int i=1;i<par.length;i++){
				(*this)+=' ';
				(*this)+=par[i];
			}
		}
		(*this)+=g_vrmlSquareCloseBracket;
	}
}
template <> void String::cat<String>(const M_<String>& par){
	if(par.length==1){
		(*this)+=par[0];
	}
	else {
		(*this)+=g_vrmlSquareOpenBracket;
		if(par.length > 0){
			(*this)+=par[0];
			for(int i=1;i<par.length;i++){
				(*this)+=' ';
				(*this)+=par[i];
			}
		}
		(*this)+=g_vrmlSquareCloseBracket;
	}
}

template<class T>
void M_<T>::toString(String& str)const{
	str.cat(*this);
}

template void M_Triangle::toString(String& str)const;
template void M_Int::toString(String& str)const;


template<>
void M_Bool::insertElement(int targetIndex, bool value){iinsertElement(*this, targetIndex, value);}
template<>
void M_Color3f::insertElement(int targetIndex, Color3f value){iinsertElement(*this, targetIndex, value);}
template<>
void M_Float::insertElement(int targetIndex, float value){iinsertElement(*this, targetIndex, value);}
template<>
void M_Image::insertElement(int targetIndex, Image value){iinsertElement(*this, targetIndex, value);}
template<>
void M_Int::insertElement(int targetIndex, int value){iinsertElement(*this, targetIndex, value);}
template<>
void M_Rotation::insertElement(int targetIndex, Rotation value){iinsertElement(*this, targetIndex, value);}
template<>
void M_String::insertElement(int targetIndex, String value){iinsertElement(*this, targetIndex, value);}
template<>
void M_Double::insertElement(int targetIndex, double value){iinsertElement(*this, targetIndex, value);}
template<>
void M_Vec2f::insertElement(int targetIndex, Vec2f value){iinsertElement(*this, targetIndex, value);}
template<>
void M_Vec3f::insertElement(int targetIndex, Vec3f value){iinsertElement(*this, targetIndex, value);}
template<>
void M_Node::insertElement(int targetIndex, NodePtr value){iinsertElement(*this, targetIndex, value);}
template<>
void M_<Ptr<Viewpoint> >::insertElement(int targetIndex, Ptr<Viewpoint> value){iinsertElement(*this, targetIndex, value);}
template<>
void M_<Ptr<Stem> >::insertElement(int targetIndex, Ptr<Stem> value){iinsertElement(*this, targetIndex, value);}
template<>
void M_<Ptr<NikInterpolator> >::insertElement(int targetIndex, Ptr<NikInterpolator> value) { iinsertElement(*this, targetIndex, value); }
template<>
void M_CtrlPolyPoint::insertElement(int targetIndex, CtrlPolyPoint value){iinsertElement(*this, targetIndex, value);}
template<>
void M_<Strand*>::insertElement(int targetIndex, Strand* value){iinsertElement(*this, targetIndex, value);}
template<>
void M_<Insert>::insertElement(int targetIndex, Insert value){iinsertElement(*this, targetIndex, value);}

template <class T> void M_<T>::deleteElement(int i){
	for(; i < length-1; i++){
		(*this)[i]=(*this)[i+1];
	}
	length--;
}
template void M_Bool::deleteElement(int targetIndex);
template void M_Color3f::deleteElement(int targetIndex);
template void M_Float::deleteElement(int targetIndex);
template void M_Image::deleteElement(int targetIndex);
template void M_Int::deleteElement(int targetIndex);
template void M_Rotation::deleteElement(int targetIndex);
template void M_String::deleteElement(int targetIndex);
template void M_Double::deleteElement(int targetIndex);
template void M_Vec2f::deleteElement(int targetIndex);
template void M_Vec3f::deleteElement(int targetIndex);
template void M_Route::deleteElement(int targetIndex);
template void M_Node::deleteElement(int targetIndex);
template void M_Art::deleteElement(int targetIndex);
template void M_SArt::deleteElement(int targetIndex);
template void M_<Ptr<Viewpoint> >::deleteElement(int targetIndex);
template void M_<MidiNoteOnItem>::deleteElement(int targetIndex);
template void M_<StemPtr>::deleteElement(int targetIndex);
template void M_ControlPoint::deleteElement(int targetIndex);
template void M_<S_VertexIndex>::deleteElement(int targetIndex);
template void M_<Link>::deleteElement(int targetIndex);
template void M_<Insert>::deleteElement(int targetIndex);
template void M_<Strand*>::deleteElement(int targetIndex);
template void M_<DelCtrlLn>::deleteElement(int targetIndex);

template <class T> void M_<T>::deleteElement(int i, int numOf){
	for(; i<this->length-numOf; i++){
		(*this)[i]=(*this)[i+numOf];
	}
	this->length-=numOf;
}
template void M_Bool::deleteElement(int from, int numOf);
template void M_Color3f::deleteElement(int from, int numOf);
template void M_Float::deleteElement(int from, int numOf);
template void M_Image::deleteElement(int from, int numOf);
template void M_Int::deleteElement(int from, int numOf);
template void M_Rotation::deleteElement(int from, int numOf);
template void M_String::deleteElement(int from, int numOf);
template void M_Double::deleteElement(int from, int numOf);
template void M_Vec2f::deleteElement(int from, int numOf);
template void M_Vec3f::deleteElement(int from, int numOf);
template void M_Route::deleteElement(int from, int numOf);
template void M_Node::deleteElement(int from, int numOf);
template void M_Art::deleteElement(int from, int numOf);
template void M_CtrlPolyPoint::deleteElement(int from, int numOf);
template void M_<Strand*>::deleteElement(int from, int numOf);
template void M_<Insert>::deleteElement(int from, int numOf);

bool isTriangleHit(const Vec2f& pt, const Vec2f* v);
bool backFace(const Vec3f& v0, const Vec3f& v1, const Vec3f& v2);
bool backFace(Vec2f* v);
const int PolyToNurbs::initPolyToNurbs_random[]={0x15, 7, 3, 8, 0x27, 5, 5, 4, 0x25, 5, 0x16, 4, 0x13, 3, 4, 6, 7, 0x16, 5, 8, 0x23, 1, 5, 0x18, 0x28, 3, 5, 0x15, 8, 4, 6, 0x14, 5, 1, 0x26, 4, 1, 1, 0x22, 3, 5, 0x18, 5, 0x14, 6, 4, 8, 1, 0x15, 7, 3, 6, 5, 5, 2, 6, 0x22, 6, 3, 2, 8, 0x26, 7, 1, 2, 3, 6, 0x18, 2, 1, 2, 6, 3, 3, 0x24, 2, 4, 4, 0x21, 8, 5, 1, 5, 3, 0x14, 3, 8, 8, 1, 6, 3, 6, 0x23, 1, 4, 3, 2, 8, 8, 8, 0x16, 7, 1, 8, 8, 7, 8, 5, 7, 5, 0x25, 4, 6, 1, 4, 2, 5, 5, 0x12, 6, 0x14, 7, 0x11, 1, 7, 8, 0x23};
void setValueFieldValueOnly(Field* dest, Field* src);

void Node::copyFields2(Node* from){
	for(int i=0; i < getFLen(); i++){
		Field* fromField = from->getField(i);
		setValueFieldValueOnly(getField(i), fromField);
	}
	collapsed = from->collapsed;
}

void Node::copyFields(Node* from){
	for(int i=0; i < getFLen(); i++){
		Field* fromField = from->getField(i);
		getField(i)->set(fromField);
		//setValueFieldValueOnly(getField(i), fromField);
	}
	collapsed = from->collapsed;
	processFields();
}

bool g_writeGzipped = true;
ProtoDeclare* g_protoDeclare;
M_ProtoDeclare g_protoList;
short ControlWord;
short ModifiedControlWord;

//lowLevelCode
float atof2(const char* str){
//	return atof(str);
#ifdef DIRECT3D
#endif
	long double num=0.0;
	int i=0;
	long double negFlag=1;
	if(str[i]=='-'){
		negFlag=-1;
		i++;
	}
	while(1){
		if(str[i]=='.'){
			i++;
			float counter=.1;
			while(str[i]>='0'&&str[i]<='9'){
				num+=counter*(float)(str[i]-'0');
				counter*=.1;
				i++;
			}
			if(str[i]=='E'||str[i]=='e'){
				num=0;
				negFlag=0;
			}
			break;
		}
		else if (str[i]>='0'&&str[i]<='9'){
			num*=10.;
			num+=(float)(str[i]-'0');
		}
		else{
			if(str[i]=='E'||str[i]=='e'){
				num=0;
				negFlag=0;
			}
			break;
		}
		i++;
	}
#ifdef DIRECT3D
#endif
	return num*negFlag;
}
double atod(const char* str){
#ifdef DIRECT3D
#endif
	long double num=0;
	int i=0;
	long double negFlag=1;
	if(str[i]=='-'){
		negFlag=-1;
		i++;
	}
	while(1){
		if(str[i]=='.'){
			i++;
			double counter=.1;
			while(str[i]>='0'&&str[i]<='9'){
				num+=counter*(double)(str[i]-'0');
				counter*=.1;
				i++;
			}
			if(str[i]=='E'||str[i]=='e'){
				num=0;
				negFlag=0;
			}
			break;
		}
		else if (str[i]>='0'&&str[i]<='9'){
			num*=10.;
			num+=(double)(str[i]-'0');
		}
		else{
			if(str[i]=='E'||str[i]=='e'){
				num=0;
				negFlag=0;
			}
			break;
		}
		i++;
	}
	return num*negFlag;
}

void searchTillNextWord(char*& p){
	while(*p>' '&&*p!=','&&*p!=']')p++;
	if(*p==0)return;
	searchTillText(p);
}
void searchTillNextVec3f(char*& p){
	for(int i=0;i<3;i++){
		searchTillNextWord(p);
	}
}
bool cmpStr(char* str ,char*& p){
	char* p2=p;
	while(*str!='\0'){
		if(*str++!=*p2++){
			return false;
		};
	}
	if(*p2>' '&&*p2!='{'&&*p2!='}'&&*p2!='['&&*p2!=']'){
		return false;
	}
	p=p2;
	return true;
}

int CmpRest(const char* match ,char*& from,int StrLen){
	for(int i=0;i<StrLen&&*match!=0;i++){
		if (*match++!=*from++)return 0;
	}
	return 1;
}
int countUpSingleIntegers(char* p){
	int i=0;
	searchTillText(p);
	p++;
	searchTillText(p);
	while(*p!=']'){
		searchTillNextWord(p);
		i++;
	}
	p++;
	return i;
}

int search(const char* match ,char*& from,char* BufEnd){
	int StrLen=usStrlen(match);
	while(from<BufEnd){
	if(*from++==*match)
		if(CmpRest(&match[1], from,StrLen-1)){
			return 1;
		}
	}
	return 0;
}
void searchTillText(char*& p){
	while(*p<=' '||*p==','){
		if(*p==0){
			return;
		}
		p++;
	}
	while(*p=='#'){
		search("\n",p, p+0x1000000);
		searchTillText(p);
	}
}
void searchTillText(char*& p, char* charBuffEnd){
	while(*p<=' '||*p==','){
		if(p>=charBuffEnd)return;
		p++;
	}
	while(*p=='#'){
		search("\n",p, p+0x1000000);
		searchTillText(p);
	}
}

bool g_removeFirstZero = true;

void removeZeros(char * str){
	int k;
	if(g_removeFirstZero && str[0] == '0' && str[1]=='.'){
		for(k = 0; str[k] != 0; k++){
			str[k] = str[k + 1];
		}
		k-=2;
		while(k > 0 && str[k] == '0'){
			str[k--] = 0;
			if(str[k] == '.'){
				str[k] = '0';
				break;
			}
		}
	}
	else if(g_removeFirstZero && str[0] == '-' && str[1] == '0' && str[2] == '.'){
		for(k = 1; str[k] != 0; k++){
			str[k] = str[k + 1];
		}
		k-=2;
		while(k > 0 && str[k] == '0'){
			str[k--]=0;
			if(str[k] == '.'){
				str[0] = '0';
				str[1] = 0;
				break;
			}
		}
	}
	else{
		k = usStrlen(str) - 1;
		while(k > 0 && str[k] == '0'){
			str[k--] = 0;
			if(str[k] == '.'){
				str[k] = 0;
				break;
			}
		}
	}
	k = usStrlen(str);
	if(k == 2 && str[0] == '-' && str[1] == '0'){
		str[0] = '0';
		str[1] = 0;
	}
}

int countUpVec3f(char* p){
	searchTillText(p);
	if('['!=*p)return -1;
	p++;
	int i=0;
	while(1){
		searchTillText(p);
		if(*p==']')break;
		searchTillNextVec3f(p);
		i++;
	}
	return i;
}
Vec2f getField(char*& p){
	Vec2f v;
	searchTillText(p);
	v.x=atof2(p);
	searchTillNextWord(p);
	v.y=atof2(p);
	searchTillNextWord(p);
	return v;
}


void searchTillNextVec2f(char*& p){
	for(int i=0;i<2;i++){
		searchTillNextWord(p);
	}
}

int countUpVec2f(char* p){
	searchTillText(p);
	if('['!=*p)return -1;
	p++;
	int i=0;
	while(1){
		searchTillText(p);
		if(*p==']')break;
		searchTillNextVec2f(p);
		i++;
	}
	return i;
}
unsigned Color3f::colorToHex()const{
	unsigned integer=0xffffff00;
	integer|=(unsigned)(r*255);
	integer<<=8;
	integer|=(unsigned)(g*255);
	integer<<=8;
	integer|=(unsigned)(b*255);
	return integer;
}
unsigned Color3f::colorToHex(const float a)const{
	unsigned integer=0x0;
	integer|=(unsigned)(a*255);
	integer<<=8;
	integer|=(unsigned)(r*255);
	integer<<=8;
	integer|=(unsigned)(g*255);
	integer<<=8;
	integer|=(unsigned)(b*255);
	return integer;
}
Color3f hexToColor(unsigned hex){
	Color3f color;
	color.b=(hex&0xff)/255.;
	hex>>=8;
	color.g=(hex&0xff)/255.;
	hex>>=8;
	color.r=(hex&0xff)/255.;
	return color;
}
int countWords(char* p){
	int i=0;
	searchTillText(p);
	while(*p!=0){
		searchTillNextWord(p);
		i++;
	}
	return i;
}

float getFloat(const String& str, int& i){
	str.stb(i);
	float value=atof2(str.c_str()+i);
	str.stwo(i);
	return value;
}
float getFloat2(const String& str, int& i){
	str.stb(i);
	float value=atof2(str.c_str()+i);
	str.endOfNum(i);
	return value;
}
double getDouble(const String& str, int& i){
	str.stb(i);
	float value=atod(str.c_str()+i);
	str.endOfNum(i);
	return value;
}
int hexStrToInt(const String& str, int& i){
	int value=0;
	while(true) {
		char ch=str[i];
		if(ch>='0'&&ch<='9'){
			value<<=4;
			value|=str[i]-'0';
		}
		else if(ch>='a'&&ch<='f'){
			value<<=4;
			value|=str[i]-'a'+10;
		}
		else if(ch>='A'&&ch<='F'){
			value<<=4;
			value|=str[i]-'A'+10;
		}
		else {
			break;
		}
		i++;
	}
	return value;
}
int getInt(const String& str, int& i){
	str.stb(i);
	if(str[i]=='0'){
		if(++i<str.length){
			if(str[i]=='x'||str[i]=='X'){
				return hexStrToInt(str, ++i);
			}
		}
		i--;
	}
	int value=atoi(str.c_str()+i);
	str.stwo(i);
	return value;
}
int getInt2(const String& str, int& i){
	str.stb(i);
	if(str[i]=='0'){
		if(++i<str.length){
			if(str[i]=='x'||str[i]=='X'){
				return hexStrToInt(str, ++i);
			}
		}
		i--;
	}
	int value=atoi(str.c_str()+i);
	if(str[i]=='-'){
		i++;
	}
	str.stwo2(i);
	return value;
}
bool getBool(const String& str, int& i){
	str.stb(i);
	if(str.cmpKWord("TRUE",i))return true;
	else if(str.cmpKWord("FALSE",i))return false;
	else throw 0;
}
bool getCppBool(const String& str, int& i){
	str.stb(i);
	if(str.cmpCppWord("true",i))return true;
	else if(str.cmpCppWord("false",i))return false;
	else throw 0;
}

void String::getField(const String& str, int& i){
	*this=str.fromQuotes(i);
}
template<>
void SFBool::read(const String& str, int& i){
	str.stb(i);
	if		(str.cmpKWord("TRUE",i)) refValue()=true;
	else if	(str.cmpKWord("FALSE",i))refValue()=false;
	else if	(str.cmpCppWord("true",i)) refValue()=true;
	else if	(str.cmpCppWord("false",i))refValue()=false;
	else throw 0;
}
void SFColor::read(const String& str, int& i){
	refValue().r=getFloat(str, i);
	str.stb(i);
	refValue().g=getFloat(str, i);
	str.stb(i);
	refValue().b=getFloat(str, i);
}
template<>
void SFFloat::read(const String& str, int& i){
	refValue()=getFloat(str, i);
}
template<>
void SFInt32::read(const String& str, int& i){
	refValue()=getInt(str, i);
}
template<>
void SFImage::read(const String&, int&){
	throw;
}
void SFRotation::read(const String& str, int& i){
	Vec3f axis;
	axis.x=getFloat(str, i);
	str.stb(i);
	axis.y=getFloat(str, i);
	str.stb(i);
	axis.z=getFloat(str, i);
	str.stb(i);
	refValue()=Rotation(axis.normalize(),getFloat(str, i));
}
void SFString::read(const String& str, int& i){
	refValue()=str.fromQuotes(i);
}
template<>
void SFTime::read(const String& str, int& i){
	refValue()=getFloat(str, i);
}
void SFVec2f::read(const String& str, int& i){
	refValue().x=getFloat(str, i);
	str.stb(i);
	refValue().y=getFloat(str, i);
}
void SFVec3f::read(const String& str, int& i){
	refValue().x=getFloat(str, i);
	str.stb(i);
	refValue().y=getFloat(str, i);
	str.stb(i);
	refValue().z=getFloat(str, i);
}
void SFNode::read(const String &str, int &i){
	refValue()=getNode2( str, i);
}
Color3f g_bColor;
bool g_collapse=false;

void checkBackgroundColor(const String& str, int& i){
	str.stbc(i);
	while(i < str.length && str[i]=='#'){
		i++;
		if(str.cmpKWord("bColor", i)){
			str.stb(i);
			g_bColor.getField(str,i);
		}
//		str.stb(i);
		if(str[i]==' ')i++;
		if(str.cmpKWord("collapse", i)){
			g_collapse=true;
		}

		str.stnl(i);
		str.stbc(i);
	}
}
template<>
void MFNode::read(const String &str, int &i){
	NodePtr node;

	bool brackFlag=false;
	str.stb(i);
	if(str[i]=='['){
		brackFlag=true;
		i++;
	}
	checkBackgroundColor(str, i);
//	str.stb(i);
	while(str[i]!=']'){
		if(str.cmpKWord("ROUTE",i)){
			S_Route(str,i);
		}
		else {
			refValue()+=getNode2(str, i);
		}
		str.stb(i);
		if(!brackFlag) break;
	}
	if(brackFlag) i++;
}
template<>
void MFString::read(const String& str,int& i){
	str.stb(i);
	String dummy;
	if(str[i]=='['){
		i++;
		str.stb(i);
		int k,ctr;


		for(k=i,ctr=0;str[k]!=']';ctr++){
			dummy.getField(str, k);
			str.stb(k);
		}
		refValue().length=ctr;
		for(ctr=0;str[i]!=']';ctr++){
			refValue()[ctr].getField(str, i);
			str.stb(i);
		}
		i++;
	}
	else{
		dummy.getField(str, i);
		refValue()=dummy;
	}
}
template<>
void MFColor::read(const String& str,int& i){
	str.stb(i);
	Color3f dummy;
	if(str[i]=='['){
		i++;
		str.stb(i);
		int k,ctr;


		for(k=i,ctr=0;str[k]!=']';ctr++){
			dummy.getField(str, k);
			str.stb(k);
		}
		M_Color3f tmpThis; tmpThis.length=ctr;

		for(ctr=0;str[i]!=']';ctr++){
			tmpThis[ctr].getField(str, i);
			str.stb(i);
		}
		refValue()=tmpThis;
		i++;
	}
	else{
		dummy.getField(str, i);
		refValue()=dummy;
	}
}
template<>
void MFBool::read(const String& str,int& i){
	str.stb(i);
	if(str[i]=='['){
		i++;
		str.stb(i);
		int k,ctr;


		for(k=i,ctr=0;str[k]!=']';ctr++){
			getFloat(str, k);
			str.stb(k);
		}
		M_Bool tmpThis; tmpThis.length=ctr;

		for(ctr=0;str[i]!=']';ctr++){
			tmpThis[ctr]=getBool(str, i);
			str.stb(i);
		}
		refValue()=tmpThis;
		i++;
	}
	else{
		refValue()=getBool(str, i);
	}
}
template<>
void MFFloat::read(const String& str,int& i){
	str.stb(i);
	if(str[i]=='['){
		i++;
		str.stb(i);
		int k,ctr;


		for(k=i,ctr=0;str[k]!=']';ctr++){
			getFloat(str, k);
			str.stb(k);
		}
		M_Float tmpThis; tmpThis.length=ctr;

		for(ctr=0;str[i]!=']';ctr++){
			tmpThis[ctr]=getFloat(str, i);
			str.stb(i);
		}
		refValue()=tmpThis;
		i++;
	}
	else{
		refValue()=getFloat(str, i);
	}
}
template<>
void MFInt32::read(const String& str, int& i){
	str.stb(i);
	refValue().length=0;
	if(str[i]=='['){
		i++;
		str.stb(i);
		int k,ctr;

		for(k=i,ctr=0;str[k]!=']';ctr++){
			getInt(str, k);
			str.stb(k);
		}
		M_Int tmpThis; tmpThis.length=ctr;

		for(ctr=0;str[i] != ']'; ctr++){
			tmpThis[ctr] = getInt(str, i);
			str.stb(i);
		}
		refValue()=tmpThis;
		i++;
	}
	else{
		refValue()=getInt(str,i);
	}
}

template<>
void MFImage::read(const String&,int&){
	throw;
}
template<>
void MFRotation::read(const String& str,int& i){
	str.stb(i);
	Rotation tmp;
	if(str[i]=='['){
		i++;
		str.stb(i);
		for(;str[i]!=']';){
			tmp.getField(str, i);
			refValue()+=(tmp);
			str.stb(i);
		}
		i++;
	}
	else{
		tmp.getField(str, i);
		refValue()=tmp;
	}
}
template<>
void MFTime::read(const String& str,int& i){
	str.stb(i);
	if(str[i]=='['){
		i++;
		str.stb(i);
		int k,ctr;


		for(k=i,ctr=0;str[k]!=']';ctr++){
			getFloat(str, k);
			str.stb(k);
		}
		M_Double tmpThis; tmpThis.length=ctr;

		for(ctr=0;str[i]!=']';ctr++){
			tmpThis[ctr]=getFloat(str, i);
			str.stb(i);
		}
		refValue()=tmpThis;
		i++;
	}
	else{
		refValue()=getFloat(str, i);
	}
}
template<>
void MFVec2f::read(const String& str,int& i){
	str.stb(i);
	Vec2f dummy;
	if(str[i]=='['){
		i++;
		str.stb(i);
		int k,ctr;


		for(k=i,ctr=0;str[k]!=']';ctr++){
			dummy.getField(str, k);
			str.stb(k);
		}
		M_Vec2f tmpThis; tmpThis.length=ctr;

		for(ctr=0;str[i]!=']';ctr++){
			tmpThis[ctr].getField(str, i);
			str.stb(i);
		}
		refValue()=tmpThis;
		i++;
	}
	else{
		dummy.getField(str, i);
		refValue()=dummy;
	}
}
template<>
void MFVec3f::read(const String& str,int& i){
	str.stb(i);
	Vec3f dummy;
	if(str[i]=='['){
		i++;
		str.stb(i);
		int k,ctr;


		for(k=i,ctr=0;str[k]!=']';ctr++){
			dummy.getField(str, k);
			str.stb(k);
		}
		M_Vec3f tmpThis; tmpThis.length=ctr;

		for(ctr=0;str[i]!=']';ctr++){
			tmpThis[ctr].getField(str, i);
			str.stb(i);
		}
		refValue()=tmpThis;
		i++;
	}
	else{
		dummy.getField(str, i);
		refValue()=dummy;
	}
}
template<>
void M_String::getField(const String& str,int& i){// assumed length=0
	str.stb(i);
	if(str[i]=='['){
		String dummy;
		i++;
		str.stb(i);

		while(str[i]!=']'){
			dummy.getField(str, i);
			(*this)+=String(dummy.length+1,0); //economize on buf size
			(*this)[length-1]+=dummy;
			str.stb(i);
		}
		i++;
	}
	else{
		length=1;
		(*this)[0].getField(str, i);
	}
}
template<>
void M_Color3f::getField(const String& str,int& i){
	str.stb(i);
	Color3f dummy;
	if(str[i]=='['){
		i++;
		str.stb(i);
		int k,ctr;


		for(k=i,ctr=0;str[k]!=']';ctr++){
			dummy.getField(str, k);
			str.stb(k);
		}
		M_Color3f tmpThis; tmpThis.length=ctr;

		for(ctr=0;str[i]!=']';ctr++){
			tmpThis[ctr].getField(str, i);
			str.stb(i);
		}
		*this=tmpThis;
		i++;
	}
	else{
		dummy.getField(str, i);
		*this=dummy;
	}
}
template<>
void M_Vec3f::getField(const String& str,int& i){
	str.stb(i);
	Vec3f dummy;
	if(str[i]=='['){
		i++;
		str.stb(i);
		int k,ctr;


		for(k=i,ctr=0;str[k]!=']';ctr++){
			dummy.getField(str, k);
			str.stb(k);
		}
		M_Vec3f tmpThis; tmpThis.length=ctr;

		for(ctr=0;str[i]!=']';ctr++){
			tmpThis[ctr].getField(str, i);
			str.stb(i);
		}
		*this=tmpThis;
		i++;
	}
	else{
		dummy.getField(str, i);
		*this=dummy;
	}
}
template<>
void M_Rotation::getField(const String& str,int& i){
	str.stb(i);
	Rotation dummy;
	if(str[i]=='['){
		i++;
		str.stb(i);
		int k,ctr;


		for(k=i,ctr=0;str[k]!=']';ctr++){
			dummy.getField(str, k);
			str.stb(k);
		}
		M_Rotation tmpThis; tmpThis.length=ctr;

		for(ctr=0;str[i]!=']';ctr++){
			tmpThis[ctr].getField(str, i);
			str.stb(i);
		}
		*this=tmpThis;
		i++;
	}
	else{
		dummy.getField(str, i);
		*this=dummy;
	}
}
template<>
void M_Vec2f::getField(const String& str,int& i){
	str.stb(i);
	Vec2f dummy;
	if(str[i]=='['){
		i++;
		str.stb(i);
		int k,ctr;


		for(k=i,ctr=0;str[k]!=']';ctr++){
			dummy.getField(str, k);
			str.stb(k);
		}
		M_Vec2f tmpThis; tmpThis.length=ctr;

		for(ctr=0;str[i]!=']';ctr++){
			tmpThis[ctr].getField(str, i);
			str.stb(i);
		}
		*this=tmpThis;
		i++;
	}
	else{
		dummy.getField(str, i);
		*this=dummy;
	}
}
template <class T> void Vec3<T>::getField(const String& str, int& i){
	x=getFloat(str, i);
	str.stb(i);
	y=getFloat(str, i);
	str.stb(i);
	z=getFloat(str, i);
}

template <class T> void Vec2<T>::getField(const String& str, int& i){
	x=getFloat(str, i);
	str.stb(i);
	y=getFloat(str, i);
}

void Rotation::getField(const String& str, int& i){
	Vec3f axis;
	axis.x=getFloat(str, i);
	str.stb(i);
	axis.y=getFloat(str, i);
	str.stb(i);
	axis.z=getFloat(str, i);
	str.stb(i);
	*this=Rotation(axis.normalize(),getFloat(str, i));
}
void Color3f::toString(String& str)const{
	str.cat(*this);
}
template<>
void Vec2f::toString(String& str)const{
	str+=x;
	str+=' ';
	str+=y;
}
template<>
void Vec3f::toString(String& str)const{
	str+=x;
	str+=' ';
	str+=y;
	str+=' ';
	str+=z;
}
template<>
void Vec3d::toString(String& str)const{
	str+=x;
	str+=' ';
	str+=y;
	str+=' ';
	str+=z;
}
void Rotation::toString(String& str)const{
	str+=x;
	str+=' ';
	str+=y;
	str+=' ';
	str+=z;
	str+=' ';
	str+=angle;
}
void Rotation4d::toString(String& str)const {
	str += x;
	str += ' ';
	str += y;
	str += ' ';
	str += z;
	str += ' ';
	str += angle;
}

void S_Triangle::toString(String& str)const{
	str.cat(*this);
}

inline void SFBoolToString(bool par, String& str){
	if(g_writeX3DXML){
		if(par==true){
			str+="true";
		}
		else{
			str+="false";
		}
	}
	else{
		if(par==true){
			str+="TRUE";
		}
		else{
			str+="FALSE";
		}
	}
}
template <> void String::cat<Image>(const M_<Image>&){
	throw;
}
template <> void String::cat<bool>(const M_<bool>& par){
	if(par.length==1){
		SFBoolToString(par[0], *this);
	}
	else {
		(*this)+=g_vrmlSquareOpenBracket;
		if(par.length > 0){
			SFBoolToString(par[0], *this);
			for(int i=1;i<par.length;i++){
				(*this)+=' ';
				SFBoolToString(par[i], *this);
			}
		}
		(*this)+=g_vrmlSquareCloseBracket;
	}
}
template<>
void M_Int::toCppString(String& str){
	for(int i=0;;){
		str+=length.pointer[i++];
		if(i>=length){
			break;
		}
		str+=',';
	}
}
template<>
void M_Float::toCppString(String& str){
	for(int i=0;;){
		str+=length.pointer[i++];
		if(i>=length){
			break;
		}
		str+=',';
	}
}
template<>
void M_Double::toCppString(String& str) {
	for (int i = 0;;) {
		str += length.pointer[i++];
		if (i >= length) {
			break;
		}
		str += ',';
	}
}
int Color3f::fromString(const String& str, int i){
	getField(str,i);
	return i;
}
void Color3f::getField(const String& str, int& i){
	r=getFloat(str, i);
	str.stb(i);
	g=getFloat(str, i);
	str.stb(i);
	b=getFloat(str, i);
}
template <class T> Vec2<T> Vec2<T>::normalize()const{
	float dist = sqrt(x * x + y * y);
	if(dist == 0)	return Vec2f(0, 0);
	float sc = 1. / dist;

	return (*this) * sc;
}

template Vec2f Vec2f::normalize()const;

template <class T> Vec3<T> Vec3<T>::normalize()const{
	float dist=sqrt(x * x + y * y + z * z);
	if(dist==0)	return Vec3<T>(0, 0, 0);
	float sc = 1./dist;

	return (*this) * sc;
}
template<>
void M_Vec3f::translate( float x, float y, float z, int offset){
	for (unsigned i=offset;i<length.rawLen;i++){
		length.pointer[i].x+=x;length.pointer[i].y+=y;length.pointer[i].z+=z;
	}
}
template<>
void M_Vec3f::translate(Vec3f tran,int offset){
	translate(tran.x, tran.y, tran.z, offset);
}
template<>
void M_Vec3f::scale(float xsc,float ysc,float zsc){
int i;
	for (i=0;i<length;i++){
		length.pointer[i].x*=xsc;
		length.pointer[i].y*=ysc;
		length.pointer[i].z*=zsc;
	}
}
template<>
void M_Vec3f::rotate( float x, float y, float z, float angle, int offset){
	for (unsigned i=offset;i<length.rawLen;i++){
		length.pointer[i]=Rotation(x,y,z,angle)*length.pointer[i];
	}
}
template<>
void M_Vec3f::rotate(Rotation rot,int offset){
	rotate( rot.x, rot.y, rot.z, rot.angle, offset);
}
//http://www.3dkingdoms.com/weekly/quat.h
//http://forum.onlineconversion.com/showthread.php?t=5408
//assumes Quaternion is normalised
Vec3d Quaternion::eulerYZX(){ //adpated from http://www.euclideanspace.com/maths/geometry/rotations/conversions/quaternionToEuler/index.htm
	double x2 = x * x;
	double y2 = y * y;
	double z2 = z * z;
	double test= x * y + z * w;
	if(test > 0.49999999) {
		return Vec3d (atan2(x, w), PI*.5, 0);
	}
	else if(test < -0.49999999) {
		return Vec3d(-atan2(x, w), -PI*.5, 0);
	}
	return Vec3d(
		atan2(2 * y * w - 2 * x * z, 1 - 2 * y2 - 2 * z2),
		asin(2 * x * y + 2 * z * w),
		atan2(2 * x * w - 2 * y * z, 1 - 2 * x2 - 2 * z2)
	);
}



//Some quaternion/rotation maths C code that has been adapted was originally written by: Gavin Bell

Quaternion invert(const Quaternion &srcQuat){
	Quaternion  destQuat;
	float srcQuatNorm;

	srcQuatNorm = 1.0 / (srcQuat.x*srcQuat.x + srcQuat.y*srcQuat.y + srcQuat.z*srcQuat.z + srcQuat.w*srcQuat.w);

	destQuat.x = -srcQuat.x * srcQuatNorm;
	destQuat.y = -srcQuat.y * srcQuatNorm;
	destQuat.z = -srcQuat.z * srcQuatNorm;
	destQuat.w =  srcQuat.w * srcQuatNorm;
	return(destQuat);

}

void rotationToQuaternion(const Rotation &rot, Quaternion &result){
	result.x = rot.x*sin(rot.angle / 2.0);
	result.y = rot.y*sin(rot.angle / 2.0);
	result.z = rot.z*sin(rot.angle / 2.0);
	result.w = cos(rot.angle / 2.0);
}

void rotationToQuaternion(const Rotation4d &rot, Quaternion &result){
	result.x = rot.x*sin(rot.angle / 2.0);
	result.y = rot.y*sin(rot.angle / 2.0);
	result.z = rot.z*sin(rot.angle / 2.0);
	result.w = cos(rot.angle / 2.0);
}

void quaternionToRotation(Quaternion q, Rotation& rot){

	if(q.w>1.0L){
		q.w=1;
	}
	rot.angle = acos(q[3]) * 2.0;
	if(rot.angle==0){
		rot=Rotation();
	}
	else{
		rot.x = q[0] / sin(rot.angle / 2.0);
		rot.y = q[1] / sin(rot.angle / 2.0);
		rot.z = q[2] / sin(rot.angle / 2.0);
	}
}

void quaternionToRotation(Quaternion q, Rotation4d& rot){

	if(q.w>1.0L){
		q.w=1;
	}
	rot.angle = acos(q[3]) * 2.0;
	if(rot.angle==0){
		rot=Rotation4d();
	}
	else{
		rot.x = q[0] / sin(rot.angle / 2.0);
		rot.y = q[1] / sin(rot.angle / 2.0);
		rot.z = q[2] / sin(rot.angle / 2.0);
	}
}

void multiplyQuaternion(Quaternion& q1, Quaternion& q2, Quaternion& result){
	float tmp[4];
	tmp[0] = q2[3] * q1[0] + q2[0] * q1[3] + q2[1] * q1[2] - q2[2] * q1[1];
	tmp[1] = q2[3] * q1[1] + q2[1] * q1[3] + q2[2] * q1[0] - q2[0] * q1[2];
	tmp[2] = q2[3] * q1[2] + q2[2] * q1[3] + q2[0] * q1[1] - q2[1] * q1[0];
	tmp[3] = q2[3] * q1[3] - q2[0] * q1[0] - q2[1] * q1[1] - q2[2] * q1[2];
	result[0] = tmp[0];
	result[1] = tmp[1];
	result[2] = tmp[2];
	result[3] = tmp[3];
}

Rotation::Rotation (Vec3f fromVector, Vec3f toVector){
//
////////////////////////////////////////////////////////////////////////
	Quaternion quat;
	Vec3f	from = fromVector.normalize();
	Vec3f	to = toVector.normalize();
	Vec3f	axis;
	float	cost;

	cost = from.dot(to);

	// check for degeneracies
	if (cost > 0.99999f) {		// vectors are parallel
		quat[0] = quat[1] = quat[2] = 0.0f;
		quat[3] = 1.0f;
		quaternionToRotation(quat, *this);
		return;
	}
	else if (cost < -0.99999f) {		// vectors are opposite
		// find an axis to rotate around, which should be
		// perpendicular to the original axis
		// Try cross product with (1,0,0) first, if that's one of our
		// original vectors then try  (0,1,0).
		Vec3f tmp = from.cross(Vec3f(1.0f, 0.0f, 0.0f));
		if (tmp.length() < 0.00001f){
			tmp = from.cross(Vec3f(0.0f, 1.0f, 0.0f));
		}

		tmp=tmp.normalize();
		quaternionToRotation(Quaternion(tmp.x, tmp.y, tmp.z, 0.0f), *this);
		return;
	}

	axis = fromVector.cross(toVector);
	axis=axis.normalize();

	// use half-angle formulae
	// sin^2 t = ( 1 - cos (2t) ) / 2
	axis *= sqrt(0.5f * (1.0f - cost));

	// scale the axis by the sine of half the rotation angle to get
	// the normalized quaternion
	quat[0] = axis.x;
	quat[1] = axis.y;
	quat[2] = axis.z;

	// cos^2 t = ( 1 + cos (2t) ) / 2
	// w part is cosine of half the rotation angle
	quat[3] = sqrt(0.5f * (1.0f + cost));

	quaternionToRotation(quat, *this);
}

Rotation4d::Rotation4d (Vec3d fromVector, Vec3d toVector){
//
////////////////////////////////////////////////////////////////////////
	Quaternion quat;
	Vec3d	from = fromVector.normalize();
	Vec3d	to = toVector.normalize();
	Vec3d	axis;
	double	cost;

	cost = from.dot(to);

	// check for degeneracies
	if (cost > 0.99999999999) {		// vectors are parallel
		quat[0] = quat[1] = quat[2] = 0.0f;
		quat[3] = 1.0f;
		quaternionToRotation(quat, *this);
		return;
	}
	else if (cost < -0.99999999999) {		// vectors are opposite
		// find an axis to rotate around, which should be
		// perpendicular to the original axis
		// Try cross product with (1,0,0) first, if that's one of our
		// original vectors then try  (0,1,0).
		Vec3d tmp = from.cross(Vec3d(1.0f, 0.0f, 0.0f));
		if (tmp.length() < 0.00001f){
			tmp = from.cross(Vec3d(0.0, 1.0, 0.0));
		}

		tmp=tmp.normalize();
		quaternionToRotation(Quaternion(tmp.x, tmp.y, tmp.z, 0.0f), *this);
		return;
	}

	axis = fromVector.cross(toVector);
	axis=axis.normalize();

	// use half-angle formulae
	// sin^2 t = ( 1 - cos (2t) ) / 2
	axis *= sqrt(0.5 * (1.0 - cost));

	// scale the axis by the sine of half the rotation angle to get
	// the normalized quaternion
	quat[0] = axis.x;
	quat[1] = axis.y;
	quat[2] = axis.z;

	// cos^2 t = ( 1 + cos (2t) ) / 2
	// w part is cosine of half the rotation angle
	quat[3] = sqrt(0.5f * (1.0f + cost));

	quaternionToRotation(quat, *this);
}

/*
Rotation::Rotation (Vec3f fromVector, Vec3f toVector){

	float zAngle=atan2( fromVector.y, fromVector.x);

	fromVector = Rotation(0,0,1,-zAngle)*fromVector;
	//toVector = Rotation(0,0,1,-zAngle)*toVector;

	float yAngle=-atan2( fromVector.z, fromVector.x);

	fromVector = Rotation(0,1,0,-yAngle)*fromVector;
	//toVector =	 Rotation(0,1,0,-yAngle)*toVector;

	Rotation rot = Rotation(0,1,0,yAngle)*Rotation(0,0,1,zAngle);

	//toVector=Rotation(rot.x,rot.y,rot.z,-rot.angle)*toVector;

	zAngle=atan2( toVector.y, toVector.x);

	toVector = Rotation(0,0,1,-zAngle)*toVector;

	yAngle=-atan2( toVector.z, toVector.x);
	toVector =	 Rotation(0,1,0,-yAngle)*toVector;

	Rotation rotB=Rotation(0,1,0,yAngle)*Rotation(0,0,1,zAngle);

	*this = rotB*-rot;
}
*/

Rotation Rotation::operator -()const{
	Rotation result;
	Quaternion quat;
	rotationToQuaternion(*this, quat);
	quat=invert(quat);
	quaternionToRotation(quat, result);
	result.setAxis(result.getAxis().normalize());

	return result;
}

Rotation4d Rotation4d::operator -()const{
	Rotation4d result;
	Quaternion quat;
	rotationToQuaternion(*this, quat);
	quat=invert(quat);
	quaternionToRotation(quat, result);
	result.setAxis(result.getAxis().normalize());

	return result;
}

Rotation Rotation::operator *=(const Rotation &par1){
	Quaternion q1;
	Quaternion q2;
	Quaternion q3;

	rotationToQuaternion(*this, q1);
	rotationToQuaternion(par1, q2);
	multiplyQuaternion(q1, q2, q3);
	quaternionToRotation(q3, *this);
	setAxis(getAxis().normalize());

	return *this;
}

Rotation Rotation::operator *(const Rotation &par1)const{
	Quaternion q1;
	Quaternion q2;
	Quaternion q3;
	Rotation result;

	rotationToQuaternion(*this, q1);
	rotationToQuaternion(par1, q2);
	multiplyQuaternion(q1, q2, q3);
	quaternionToRotation(q3, result);
	result.setAxis(result.getAxis().normalize());

	return result;
}

Rotation4d Rotation4d::operator *(const Rotation4d &par1)const{
	Quaternion q1;
	Quaternion q2;
	Quaternion q3;
	Rotation4d result;

	rotationToQuaternion(*this, q1);
	rotationToQuaternion(par1, q2);
	multiplyQuaternion(q1, q2, q3);
	quaternionToRotation(q3, result);
	result.setAxis(result.getAxis().normalize());

	return result;
}

Rotation Rotation::multiply(const Rotation &par1)const{
	Quaternion q1;
	Quaternion q2;
	Quaternion q3;
	Rotation result;

	rotationToQuaternion(*this, q1);
	rotationToQuaternion(par1, q2);
	multiplyQuaternion(q1, q2, q3);
	quaternionToRotation(q3, result);
	result.setAxis(result.getAxis().normalize());

	return result;
}

Vec3f Rotation::getAxis()const{
	return Vec3f(x, y, z);
}

Vec3d Rotation4d::getAxis()const{
	return Vec3d(x, y, z);
}

void rotateVectorX(Vec3f& v,float angle){
	float newY;
	newY = (float)(v.y * cos(angle) - v.z * sin(angle));
	v.z = (float)(v.y * sin(angle) + v.z * cos(angle));
	v.y=newY;
}

void rotateVectorX(Vec3d& v,float angle){
	double newY;
	newY = v.y * cos(angle) - v.z * sin(angle);
	v.z = v.y * sin(angle) + v.z * cos(angle);
	v.y=newY;
}

void rotateVectorY(Vec3f& v,float angle){
	float newZ;
	newZ = (float)(v.z * cos(angle) - v.x * sin(angle));
	v.x = (float)(v.z * sin(angle) + v.x * cos(angle));
	v.z=newZ;
}

void rotateVectorY(Vec3d& v, double angle){
	double newZ;
	newZ = v.z * cos(angle) - v.x * sin(angle);
	v.x = v.z * sin(angle) + v.x * cos(angle);
	v.z = newZ;
}

void rotateVectorZ(Vec3f& v,float angle){
	float newX;
	newX=(float)(v.x * cos(angle) - v.y * sin(angle));
	v.y=(float)(v.x * sin(angle) + v.y * cos(angle));
	v.x = newX;
}

void rotateVectorZ(Vec3d& v, double angle){
	double newX;
	newX = v.x * cos(angle) - v.y * sin(angle);
	v.y = v.x * sin(angle) + v.y * cos(angle);
	v.x = newX;
}

Vec3f Rotation::operator *(const Vec3f& vec)const{
		Vec3f axis=getAxis(); //put the axis rotation into the vector axis

		//unwind the axis to 1 0 0
		float zAngle=atan2(axis.y,axis.x);	  //get the z angle necessary for unrotating the axis
		rotateVectorZ(axis,-zAngle);

		float yAngle=-atan2( axis.z, axis.x);	//get the y angle necessary for unrotating the axis
		rotateVectorY( axis, -yAngle);
		//axis should allways = 1 0 0 here because we have unwound the vector so that it points to the right, that is the axis the x axis or (1 0 0)

		//rotate the second VisibleObject only on the x y and x rotations to simulate rotating an object on any given axis to any given angle
		Vec3f retvec=vec;
		rotateVectorZ(retvec,-zAngle);//unwind vec to the x axis
		rotateVectorY(retvec,-yAngle);

		rotateVectorX(retvec, angle);//rotate the vector on the x axis with the angle component exactly the same as if it was rotated on any axis

		rotateVectorY(retvec, yAngle);// wind vec back to the specified axis
		rotateVectorZ(retvec, zAngle);
		return retvec;
}

Vec3d Rotation4d::operator *(const Vec3d& vec)const{
		Vec3d axis=getAxis(); //put the axis rotation into the vector axis

		//unwind the axis to 1 0 0
		double zAngle=atan2(axis.y,axis.x);	  //get the z angle necessary for unrotating the axis
		rotateVectorZ(axis,-zAngle);

		double yAngle=-atan2( axis.z, axis.x);	//get the y angle necessary for unrotating the axis
		rotateVectorY(axis, -yAngle);
		//axis should allways = 1 0 0 here because we have unwound the vector so that it points to the right, that is the axis the x axis or (1 0 0)

		//rotate the second VisibleObject only on the x y and x rotations to simulate rotating an object on any given axis to any given angle
		Vec3d retvec=vec;
		rotateVectorZ(retvec,-zAngle);//unwind vec to the x axis
		rotateVectorY(retvec,-yAngle);

		rotateVectorX(retvec, angle);//rotate the vector on the x axis with the angle component exactly the same as if it was rotated on any axis

		rotateVectorY(retvec, yAngle);// wind vec back to the specified axis
		rotateVectorZ(retvec, zAngle);
		return retvec;
}


// Construct a rotation of a given angle about a given axis.
// Derived from Eric Haines's SPD (Standard Procedural
// Database).
Matrix rotationMatrix(Rotation rot){
	float angle=rot.angle;
	Vec3f axis=rot.getAxis();
	Matrix M;

	if (angle == 0.0) {
		return Matrix();
	}
	axis.normalize();

	float cosine = cos(angle);
	float sine = sin(angle);
	float one_minus_cosine = 1 - cosine;

	M.m[0][0] = axis.x * axis.x + (1.0 - axis.x * axis.x) * cosine;
	M.m[0][1] = axis.x * axis.y * one_minus_cosine + axis.z * sine;
	M.m[0][2] = axis.x * axis.z * one_minus_cosine - axis.y * sine;
	M.m[0][3] = 0;

	M.m[1][0] = axis.x * axis.y * one_minus_cosine - axis.z * sine;
	M.m[1][1] = axis.y * axis.y + (1.0 - axis.y * axis.y) * cosine;
	M.m[1][2] = axis.y * axis.z * one_minus_cosine + axis.x * sine;
	M.m[1][3] = 0;

	M.m[2][0] = axis.x * axis.z * one_minus_cosine + axis.y * sine;
	M.m[2][1] = axis.y * axis.z * one_minus_cosine - axis.x * sine;
	M.m[2][2] = axis.z * axis.z + (1.0 - axis.z * axis.z) * cosine;
	M.m[2][3] = 0;

	M.m[3][0] = 0;
	M.m[3][1] = 0;
	M.m[3][2] = 0;
	M.m[3][3] = 1;

	return M;
}

String String::fromQuotes(int& position)const{
	stb(position);
	if((*this)[position++]!='\"')throw this;

	int begin=position;

	while((*this)[position++]!='\"');
	return substr( begin, position-begin-1);
}

template<>
void M_Bool::getField(const String& str,int& i){
	str.stb(i);
	if(str[i]=='['){
		i++;
		str.stb(i);
		int k,ctr;

		for(k=i,ctr=0;str[k]!=']';ctr++){
			getBool(str, k);
			str.stb(k);
		}
		length=ctr;

		for(ctr=0;str[i]!=']';ctr++){
			length.pointer[ctr]=getBool(str, i);
			str.stb(i);
		}
		i++;
	}
	else{
		*this+=getBool(str,i);
	}
}
template<>
void M_Int::getField(const String& str,int& i){
	str.stb(i);
	if(str[i]=='['){
		i++;
		str.stb(i);
		int k,ctr;

		for(k=i,ctr=0;str[k]!=']';ctr++){
			getInt(str, k);
			str.stb(k);
		}
		length=ctr;

		for(ctr=0;str[i]!=']';ctr++){
			length.pointer[ctr]=getInt(str, i);
			str.stb(i);
		}
		i++;
	}
	else{
		*this+=getInt(str,i);
	}
}

/*
void M_RedirectCoord::getField(const String& str,int& i){
	str.stb(i);
	if(str[i]=='['){
		i++;
		str.stb(i);
		int k,ctr;

		for(k=i,ctr=0;str[k]!=']';ctr++){
			getInt(str, k);
			str.stb(k);
		}
		length=ctr;

		for(ctr=0;str[i]!=']';ctr++){
			length.pointer[ctr].i=getInt(str, i);
			str.stb(i);
		}
		i++;
	}
	else{
		length=1;
		length.pointer[0].i=getInt(str, i);
	}
}
*/
template<>
void M_Bool::toString(String& str)const{
	if(length==1){
		str+=length.pointer[0];
	}
	else {
		str+=g_vrmlSquareOpenBracket;
		for(int i=0;i<length;i++){
			str+=length.pointer[i];
			str+=" ";
		}
		str+=g_vrmlSquareCloseBracket;
	}
}
template<>
void M_Image::toString(String&)const{
	throw this;
}
/*
String& String::operator +=(const M_Vec3f& par1){
	requestMoreIfNeeded(10*3*par1.length+5);
	(*this)+=g_vrmlSquareOpenBracket;
	char* p=pointer->rawChar+pointer->rawLen;

	for(int i=0;i<par1.length;){
		sprintf(p,floatFormat,par1[i].x);
		removeZeros(++p);
		p+=usStrlen(p);
		sprintf(p,floatFormat2,par1[i].y);
		removeZeros(++p);
		p+=usStrlen(p);
		sprintf(p,floatFormat2,par1[i].z);
		removeZeros(++p);
		p+=usStrlen(p);
		i++;
		if(i<<par1.length){
		}
	}
	pointer->rawLen=p-pointer->rawChar;
	(*this)+=g_vrmlSquareCloseBracket;
	return *this;
}
*/
void String::toString(String& str){
	str+='\"';
	str+=pointer->rawChar;
	str+='\"';
}
template<>
void M_String::toString(String& str)const{
	if(length==1){
		str+='\"';
		str+=length.pointer[0];
		str+='\"';
	}
	else {
		str+=g_vrmlSquareOpenBracket;
		if(length){
			int i=0;
			while(1){
				str+='\"';
				str+=length.pointer[i];
				str+='\"';
				i++;
				if(i<length){
					str+=" ";
				}
				else break;
			}
		}
		str+=g_vrmlSquareCloseBracket;
	}
}
Art* S_Triangle::getOwner(int& idx)const{//get lowest bider
	Art* part=pa;
	idx=ia;
	if(pb->plidx<part->plidx){
		part=pb;
		idx=ib;
	}
	if(pc->plidx<part->plidx){
		part=pc;
		idx=ic;
	}
	return part;
}

void String::cat(const S_Triangle& par){//we should be using lower level codeing here if this is used significantly
	(*this)+='(';
	for(int i=3; --i>=0;){
		(*this)+=' ';
		(*this)+=par.refPart(i)->name;
		(*this)+=' ';
		(*this)+=par.refIndex(i);
		(*this)+=' ';
		(*this)+=par.refTexIndex(i);
	}
	(*this)+=')';
}
String& String::operator= (const char& par){
	unsigned lEn=1;

	if(pointer->refCount>1||lEn>pointer->bufLen){
		if(--pointer->refCount==0){
			delete pointer;
		}
		pointer=createData(lEn+defaultBufLen);
	}
	pointer->rawLen=lEn;
	pointer->rawChar[0]=par;
	return *this;
}
String& String::operator= (float par1){
	char tmp[50];
	sprintf(tmp,"%.6f",par1);
	removeZeros(tmp);
	unsigned lEn=usStrlen(tmp);

	if(pointer->refCount>1||lEn>pointer->bufLen){
		if(--pointer->refCount==0){
			delete pointer;
		}
		pointer=createData(lEn+defaultBufLen);
	}
	pointer->rawLen=lEn;
	for(unsigned i=0;i<pointer->rawLen+1;i++)pointer->rawChar[i]=tmp[i];
	return *this;
}
String& String::operator= (int par1){
	char tmp[50];
	sprintf(tmp,"%i",par1);
	unsigned lEn=usStrlen(tmp);

	if(pointer->refCount>1||lEn>pointer->bufLen){
		if(--pointer->refCount==0){
			delete pointer;
		}
		pointer=createData(lEn+defaultBufLen);
	}
	pointer->rawLen=lEn;
	for(unsigned i=0;i<pointer->rawLen+1;i++)pointer->rawChar[i]=tmp[i];
	return *this;
}
void String::cat(const String &par1){
	unsigned i;
	unsigned newLength=pointer->rawLen+par1.pointer->rawLen;
	if(pointer->refCount>1||newLength>pointer->bufLen){

		Data* newPointer=createData( newLength+defaultBufLen);
		newPointer->rawLen=pointer->rawLen;
		for(i=0;i<pointer->rawLen+1;i++){
			newPointer->rawChar[i]=pointer->rawChar[i];
		}

		if(--pointer->refCount==0){
			delete pointer;
		}
		pointer=newPointer;
	}

	for(i=0;i<par1.pointer->rawLen+1;i++){
		pointer->rawChar[pointer->rawLen+i]=par1.pointer->rawChar[i];
	}
	pointer->rawLen=newLength;
}
void String::cat(const char& par1){
	Data* newPointer;
	unsigned i;
	if(pointer->refCount>1){
		--pointer->refCount;
		newPointer=createData( pointer->rawLen+defaultBufLen);
		newPointer->rawLen=pointer->rawLen;
		for(i=0; i<pointer->rawLen+1; i++){
			newPointer->rawChar[i]=pointer->rawChar[i];
		}
		pointer=newPointer;
	}
	else if(pointer->rawLen==pointer->bufLen){
		newPointer=createData( pointer->rawLen+defaultBufLen);
		newPointer->rawLen=pointer->rawLen;
		for(i=0; i<pointer->rawLen+1; i++){
			newPointer->rawChar[i]=pointer->rawChar[i];
		}
		if(--pointer->refCount==0){
			delete pointer;
		}
		pointer=newPointer;
	}
	pointer->rawChar[pointer->rawLen++]=par1;
	pointer->rawChar[pointer->rawLen]=0;
}
void String::cat(const unsigned char& par){
	cat(static_cast<const char>(par));
}
//needed gzip 0
extern bool g_winAPI3DMode;

void dLog(String str){
#ifndef PROG_ROB
	//return;
#endif
	FILE* stream = fopen(g_debugLogPath.c_str(),"a+t");
	if(stream){
		fputs((str+="\n").c_str(), stream);
		fclose(stream);
	}
}

int String::indexOf( String searchString, int position)const{
	char* match=searchString.pointer->rawChar;
	char* from=&pointer->rawChar[position];
	int newLength=pointer->rawLen-position;
	for(int i=0;i<newLength;i++){
		if(*from++==*match){
		char* match2=match+1;
		if(*match2==0)return position+i;
		char* from2=from;
			for(int k=i;k<newLength;k++){
				if (*match2++!=*from2++)break;
				if(*match2==0)return position+i;
			}

		}
	}
	return -1;
}
bool String::findWord(const String str, int& i)const{
	do{
		i=indexOf(str,i);
		int j=i;
		if(cmpKWord(str,j))return true;
		i++;
	}while(i<length);
	return false;
}

void String::stnl(int& position)const {
	unsigned char* p = (unsigned char*)&pointer->rawChar[position];
	unsigned char* endP = (unsigned char*)&pointer->rawChar[pointer->rawLen];
	while (p < endP && *p != '\n') {
		p++;
	}
	while (p < endP && *p < ' ') {
		p++;
	}
	position = p - (unsigned char*)pointer->rawChar;
}


void String::stb(int& position)const{
	unsigned char* p=(unsigned char*)&pointer->rawChar[position];
	int newLength=pointer->rawLen-position-1;
	for(int i=0;i<newLength;i++){
		if(*p>' '&&*p!=','){
			if(*p=='#'){
				while(*p>=' '||*p=='\t')p++;//search for new line
			}
			else{
				break;
			}
		}
		p++;
	}
	position=p-(unsigned char*)pointer->rawChar;
}
void String::stbCpp(int& position)const{
	unsigned char* p=(unsigned char*)&pointer->rawChar[position];
	int newLength=pointer->rawLen-position-1;
	for(int i=0;i<newLength;i++){
		if(*p>' '){
			if(*p=='/'&&p[1]=='/'){
				p++;
				while(*p>=' '||*p=='\t')p++;//seerch for new line
			}
			else{
				break;
			}
		}
		p++;
	}
	position=p-(unsigned char*)pointer->rawChar;
}
void String::stbnlCpp(int& position)const{
	unsigned char* p=(unsigned char*)&pointer->rawChar[position];
	int newLength=pointer->rawLen-position-1;
	for(int i=0;i<newLength;i++){
		if( *p=='\n'){
			break;
		}
		else if(*p>' '){
			if(*p=='/'&&p[1]=='/'){
				p++;
				while(*p>=' '||*p=='\t')p++;//search for new line
			}
			else{
				break;
			}
		}
		p++;
	}
	position=p-(unsigned char*)pointer->rawChar;
}
void String::stbc(int& position)const{
	unsigned char* p=(unsigned char*)&pointer->rawChar[position];
	int newLength=pointer->rawLen-position-1;
	for(int i=0;i<newLength;i++){
		if(*p>' '&&*p!=','){
			break;
		}
		p++;
	}
	position=p-(unsigned char*)pointer->rawChar;
}

void String::stw(int& position)const{
	char* p=&pointer->rawChar[position];
	int newLength=pointer->rawLen-position;
	for(int i=0;i<newLength;i++){
		if(*p<=' '||*p==','){
			position=p-pointer->rawChar;
			return;
		}
		p++;
	}
	throw this;
}

void String::endOfName(int& index)const{
	char* p=&pointer->rawChar[index];
	int newLength=pointer->rawLen-index;
	int i=0;
	while(i < newLength	&& isLegalNameChar(*p) ){
		p++;
	}
	index=p-pointer->rawChar;
}
String String::extractName(int& i)const{
	stb(i);
	int start = i;
	endOfName(i);
	return substr(start,i-start);
}
void String::stwo(int& position)const{
	char* rawChar=pointer->rawChar;
	char* p=rawChar+position;
	char* endAddress=rawChar+pointer->rawLen;
	while(
		(p  <	endAddress	&& *p == '_') ||
		(*p == '-' || *p	== '+'|| *p	== '.')|| 	//+ - . needed for legacy reasons ???????
		(*p >= '0' && *p	<= '9')||
		(*p >= 'A' && *p	<= 'Z')||
		(*p >= 'a' && *p	<= 'z')
		){
		p++;
	}
	position=p-rawChar;
}
void String::stwo2(int& position)const{
	char* rawChar=pointer->rawChar;
	char* p=rawChar+position;
	char* endAddress=rawChar+pointer->rawLen;
	while(
		(p  <	endAddress	&& *p == '_')||
		(*p >= '0' && *p	<= '9')||
		(*p >= 'A' && *p	<= 'Z')||
		(*p >= 'a' && *p	<= 'z')
		){
		p++;
	}
	position=p-rawChar;
}
void String::endOfNum(int& position)const{
	char* rawChar=pointer->rawChar;
	char* p=rawChar+position;
	char* endAddress=rawChar+pointer->rawLen;
	int dotCounter=0;
	if(*p == '-'){
		p++;
		while (p<endAddress && *p<=' '){
			p++;
		}
	}
	if(*p == '.'){
		dotCounter++;
		p++;
	}
	while(p < endAddress){
here:
		if(*p=='E'||*p=='e'){
			p++;
			p++;
			while(p < endAddress && *p >= '0' && *p	<= '9'){
				p++;
			}
			break;
		}
		if(*p < '0' || *p	> '9')break;
		p++;
	}
	if(p < endAddress && *p=='.' && dotCounter==0){
		p++;
		goto here;
	}
	position=p-rawChar;
}
/*
void String::stNonCppOperator(int& position)const{//C++
	char* p=&pointer->rawChar[position];
	int newLength=pointer->rawLen-position;
	int i=0;
	while(
		i  <	newLength	&&
		*p != '_'&&
		(*p < '0' || *p	> '9')&&
		(*p < 'A' || *p	> 'Z')&&
		(*p < 'a' || *p	> 'z')
		){
		p++;
	}
	position=p-pointer->rawChar;
}
*/
void String::stw_stb(int& position)const{
	stw(position);
	stb(position);
}
void String::stwo_stb(int& position)const{
	stwo(position);
	stb(position);
}
bool String::cmpKWord(const char* str,int& index)const{
	const char* p2=c_str()+index;
	while(*str!='\0'){
		if(*str++!=*p2++){
			return false;
		};
	}
	if(*p2>' '&&*p2!=','&&*p2!='{'&&*p2!='}'&&*p2!='['&&*p2!=']'&&*p2!='#'&&*p2!='\"'){
		return false;
	}
	index= p2-c_str();
	return true;
}
bool String::cmpKWord(const String& str,int& index)const{
	return cmpKWord(str.pointer->rawChar, index);
}
String String::getWord(int& i)const{
	stb(i);
	int start=i;
	stwo(i);
	return substring(start,i-1);
}
String String::getWord2(int& i)const{
	stb(i);
	int start=i;
	stwo2(i);
	return substring(start,i-1);
}


int countWords(char* p);
template<>
void M_Float::getField(const String& str,int& i){
	str.stb(i);
	if(str[i]=='['){
		i++;
		str.stb(i);
		int k,ctr;


		for(k=i,ctr=0;str[k]!=']';ctr++){
			getFloat(str, k);
			str.stb(k);
		}
		M_Float tmpThis; tmpThis.length=ctr;

		for(ctr=0;str[i]!=']';ctr++){
			tmpThis[ctr]=getFloat(str, i);
			str.stb(i);
		}
		*this=tmpThis;
		i++;
	}
	else{
		*this+=getFloat(str, i);
	}
}

bool Node::setDEFgetUSE( String &str, bool& newList){
	if(newList){
		g_scene->defList.length = 0;
		newList = false;
		g_protoList = M_ProtoDeclare();
#ifdef SEAMLESS3DGUI
		g_sceneTreeNodeIndex = 0;
#endif
	}
	if(name!=""){
		for(int k=0; k<g_scene->defList.length; k++){
			if(g_scene->defList[k]->name == name){
				if(g_writeX3DXML){
					str+=" USE='";
					str+=name;
					str+="'/>";
				}
				else{
					str+="USE ";
					str+=name;
					str+=" ";
				}
				return true;
			}
		}
		if(g_writeX3DXML){
			str+=" DEF='";
			str+=name;
			str+='\'';
			g_scene->defList += this;
		}
		else{
			str+="DEF ";
			str+=name;
			str+=' ';
			g_scene->defList += this;
		}
	}
	return false;
}

inline bool protoAlreadyAdded(const String& name){
	for(int i=g_userProtoBegin; i<nodeLookUpStr.length; i++){
		if( nodeLookUpStr[i]==name ){
			return true;
		}
	}
	return false;
}
void addNewProtoDeclareToList(ProtoDeclare* protoDeclare){

	g_protoList+=protoDeclare;//should this be inside the following if statement?

	if(protoAlreadyAdded(protoDeclare->name)==false){
		protoDeclare->lookupDeclarationsIndex=g_lookupProtoInterface.length;
		g_lookupProtoInterface+=protoDeclare->protoInterface;
		nodeLookUpStr+=protoDeclare->name;
		containerFieldLookUpStr.length++;
	}
}

Node* getNode2(const String& str, int& i){
	static Seamless* smlsLegacy = 0;
	checkBackgroundColor(str, i);
	int bracketCntr=0;

	Node* node=0;

//	str.stb(i);
	String strDefName;
	if(str.substring(i, i + 2)=="DEF"){
		str.stw_stb(i);
		strDefName=str.extractName(i);
		str.stb(i);
	}
	if(str.cmpKWord("USE",i)){
		node = g_load_scene->getUSE(str, i);
		return node;
	}
	else switch(str[i++]){
		case 'A':{
			if(str.cmpKWord("nim",i)){
				Anim* anim=new Anim;
				if(g_load_scene==0 || g_load_scene->version < 2.161-.0005){
					anim->stepsPerSecond.refValue() = 25;
				}
				node = anim;
			}
			else if(str.cmpKWord("nimLever",i)){
				node = new AnimLever;
			}
			else if(str.cmpKWord("mplitudeModulator",i)){
				node = new AmplitudeModulator;
			}
			else if(str.cmpKWord("ppearance",i)){
				node=new Appearance;
			}
			else if(str.cmpKWord("nchor",i)){
				node=new Anchor;
			}
			else if(str.cmpKWord("udioClip",i)){
				node=new AudioClip;
			}
			else if(str.cmpKWord("nimByAnkle",i)){
				node=new AnimByAnkle;
			}
			else if(str.cmpKWord("nimByWrist",i)){
				node = new AnimByWrist;
			}
			else if(str.cmpKWord("udio",i)){
				node = new Audio;
			}
			else if(str.cmpKWord("vatarSocket",i)){
				node = newAvatarSocket();
			}
			else goto protoPerhaps;
			break;
		}
		case 'B':{
			if(str.cmpKWord("endyLathe",i)){
				node = new BendyLathe;
			}
			else if(str.cmpKWord("ezierLathe",i)){
				node = new BezierLathe;
			}
			else if(str.cmpKWord("ezierLatheBuild",i)){
				node = new BezierLatheBuild;
			}
			else if(str.cmpKWord("ezierScale",i)){
				node = new BezierScale;
			}
			else if(str.cmpKWord("ezierCircumScale",i)){
				node = new BezierCircumScale;
			}
			else if(str.cmpKWord("ezierPatch",i)){
				node = new BezierPatch;
			}
			else if(str.cmpKWord("ezierShift",i)){
				node = new BezierShift;
			}
			else if(str.cmpKWord("ezierTwist",i)){
				node = new BezierTwist;
			}
			else if(str.cmpKWord("ezierTape",i)){
				node = new BezierTape;
			}
			else if(str.cmpKWord("ezierBend",i)){
				node = new BezierBend;
			}
			else if(str.cmpKWord("ooleanFilter",i)){
				node = new BooleanFilter;
			}
			else if(str.cmpKWord("ooleanSequencer",i)){
				node = new CameraSelector;
			}
			else if(str.cmpKWord("ooleanToggle",i)){
				node = new BooleanToggle;
			}
			else if(str.cmpKWord("ooleanTrigger",i)){
				node = new BooleanTrigger;
			}
			else if(str.cmpKWord("ranchJoinNegX",i)){
				node = new BranchJoinNegX;
			}
			else if(str.cmpKWord("ranchJoinPosX",i)){
				node = new BranchJoinPosX;
			}
			else if(str.cmpKWord("ranchJoinNegY",i)){
				node = new BranchJoinNegY;
			}
			else if(str.cmpKWord("ranchJoinNegZ",i)){
				node = new BranchJoinNegZ;
			}
			else if(str.cmpKWord("ranchJoinPosZ",i)){
				node = new BranchJoinPosZ;
			}
			else if(str.cmpKWord("ranchJoinPosY",i)){
				node = new BranchJoinPosY;
			}
			else if(str.cmpKWord("rush",i)){
				node = new Brush;
			}
			else if(str.cmpKWord("spTree",i)){
				node = new BspTree;
			}
			else if(str.cmpKWord("illboard",i)){
				node = new Billboard;
			}
			else if(str.cmpKWord("ox",i)){
				node = new Box;
			}
			else if(str.cmpKWord("oxBuild",i)){
				node = new BoxBuild;
			}
			else if(str.cmpKWord("ackground",i)){
				node = new Background;
			}
			else goto protoPerhaps;
			break;
		}
		case 'C':{
			if(str.cmpKWord("ameraSelector",i)){
				node = new CameraSelector;
			}
			else if(str.cmpKWord("opyToMorph",i)){
				node = new CopyToMorph;
			}
			else if(str.cmpKWord("hords",i)){
				node = new Chords;
			}
			else if(str.cmpKWord("olor3f",i)){
				return decodeInitializedGlobal<Color3f,Single>(str,i, bracketCntr);
			}
			else if(str.cmpKWord("olorEffect",i)){
				node = new ColorEffect;
			}
			else if(str.cmpKWord("ommonSpec",i)){
				node = new CommonSpec;
			}
			else if(str.cmpKWord("ylinderSensor",i)){
				node = new CylinderSensor;
			}
			else if(str.cmpKWord("omposition",i)){
				node = new Composition;
			}
			else if(str.cmpKWord("CPath",i)){
				node = new CCPath;
			}
			else if(str.cmpKWord("olorInterpolator",i)){
				node = new ColorInterpolator;
			}
			else if(str.cmpKWord("oordinateInterpolator",i)){
				node = new CoordinateInterpolator;
			}
			else if(str.cmpKWord("olorMapper",i)){
				node = new ColorMapper;
			}
			else if(str.cmpKWord("olorPatch",i)){
				//ColorPatch* patch = new ColorPatch;
				NurbsPatch* patch = new NurbsPatch;
				if(g_load_scene == 0 || g_load_scene->version < 2.1725){
					patch->dim.refValue() = Vec2f(2, 2);
				}
				node = patch;
			}
			else if(str.cmpKWord("olorSweep",i)){
				node = new ColorSweep;
			}
			else if(str.cmpKWord("opyPart",i)){
				node = new CopyPart;
			}
			else if(str.cmpKWord("enterJoint",i)){
				node = new CenterJoint;
			}
			else if(str.cmpKWord("CLathe",i)){
				node = new CCLathe;
			}
			else if(str.cmpKWord("CBend",i)){
				node = new CCBend;
			}
			else if(str.cmpKWord("PMorpher",i)){
				node = new CPMorpher;
			}
			else if(str.cmpKWord("CScale",i)){
				node = new CCScale;
			}
			else if(str.cmpKWord("CRingScale",i)){
				node = new CCRingScale;
			}
			else if(str.cmpKWord("CTranslate",i)){
				node = new CCTranslate;
			}
			else if(str.cmpKWord("olor",i)){
				node = new Color;
			}
			else if(str.cmpKWord("ollision",i)){
				node = new Collision;
			}
			else if(str.cmpKWord("oordinate",i)){
				node = new Coordinate;
			}
			else if(str.cmpKWord("one",i)){
				node = new Cone;
			}
			else if(str.cmpKWord("oneBuild",i)){
				node = new ConeBuild;
			}
			else if(str.cmpKWord("ylinder",i)){
				node = new Cylinder;
			}
			else if(str.cmpKWord("ylinderBuild",i)){
				node = new CylinderBuild;
			}
			else if(str.cmpKWord("odeLine",i)){
				node = new CodeLine;
			}
			else if(str.cmpKWord("onvertToPatch",i)){
				node = new ConvertToPatch;
			}
			else goto protoPerhaps;
			break;
		}
		case 'D':{
			if(str.cmpKWord("elSphere",i)){
				node = new DelSphere;
			}
			else if(str.cmpKWord("elPlane",i)){
				node = new DelPlane;
			}
			else if(str.cmpKWord("eleteOptions",i)){
				node = new DeleteOptions;
			}
			else if(str.cmpKWord("irectionalLight",i)){
				node = new DirectionalLight;
			}
			else if(str.cmpKWord("elPoly",i)){
				node = new DelPoly;
			}
			else if(str.cmpKWord("isplacer",i)){
				node = new HAnimDisplacer;
			}
			else goto protoPerhaps;
			break;
		}
		case 'E':{
			if(str.cmpKWord("nvelope",i)){
				node = new Envelope;
			}
			else if(str.cmpKWord("CBend",i)){
				node = new ECBend;
			}
			else if(str.cmpKWord("CLathe",i)){
				node = new ECLathe;
			}
			else if(str.cmpKWord("CScale",i)){
				node = new ECScale;
			}
			else if(str.cmpKWord("CShift",i)){
				node = new ECShift;
			}
			else if(str.cmpKWord("CSyncBend",i)){
				node = new ECSyncBend;
			}
			else if(str.cmpKWord("CTwist",i)){
				node = new ECTwist;
			}
			else if(str.cmpKWord("volveCoord",i)){
				node = new EvolveCoord;
			}
			else if(str.cmpKWord("levationGrid",i)){
				node = new ElevationGrid;
			}
			else if(str.cmpKWord("xtrusion",i)){
				node = new Extrusion;
			}
			else if(str.cmpKWord("xitBuild",i)){
				node = new ExitBuild;
			}
			else if(str.cmpKWord("xportMovie",i)){
				node = new ExportMovie;
			}
			else if(str.cmpKWord("xportVrmlX3d",i)){
				node = new ExportVrmlX3d;
			}
			else goto protoPerhaps;
			break;
		}
		case 'F':{
			if(str.cmpKWord("ill",i)){
				node = new Fill;
			}
			else if(str.cmpKWord("ilter",i)){
				node = new Filter;
			}
			else if(str.cmpKWord("og",i)){
				node = new Fog;
			}
			else if(str.cmpKWord("og2",i)){
				node = new Fog2;
			}
			else if(str.cmpKWord("ormant",i)){
				node = new Formant;
			}
			else if(str.cmpKWord("requencyModulator",i)){
				node = new FrequencyModulator;
			}
			else if(str.cmpKWord("ontStyle",i)){
				node = new FontStyle;
			}
			else if(str.cmpKWord("loatAccelerator",i)){
				node = new FloatAccelerator;
			}
			else if(str.cmpKWord("unction",i)){
				node = newFunction();
			}
			else if(str.cmpKWord("eedback",i)){
				node = new Feedback;
			}
			else if(str.cmpKWord("useSurface",i)){
				node = new FuseSurface;
			}
			else goto protoPerhaps;
			break;
		}
		case'G':{
			if(str.cmpKWord("enPoly",i)){
				node = new GenPoly;
			}
			else if(str.cmpKWord("estureGroup",i)){
				node = new GestureGroup;
			}
			else if(str.cmpKWord("roup",i)){
				node = new Group;
			}
			else if(str.cmpKWord("rid",i)){
				node = new Grid;
			}
			else if(str.cmpKWord("roupBuild",i)){
				node = new GroupBuild;
			}
			else if(str.cmpKWord("_S_Bool",i)){
				node= new V_<S_Bool,Single>;
			}
			else if(str.cmpKWord("_S_Color",i)){
				node= new V_<Color3f,Single>;
			}
			else if(str.cmpKWord("_S_Float",i)){
				node= new V_<S_Float,Single>;
			}
			else if(str.cmpKWord("_S_Image",i)){
				node= new V_<Image,Single>;
			}
			else if(str.cmpKWord("_S_Int32",i)){
				node= new V_<int,Single>;
			}
			else if(str.cmpKWord("_S_NodePtr",i)){
				node= new V_<NodePtr,Single>;
			}
			else if(str.cmpKWord("_S_Rotation",i)){
				node= new V_<Rotation,Single>;
			}
			else if(str.cmpKWord("_S_String",i)){
				node= new V_<String,Single>;
			}
			else if(str.cmpKWord("_S_Time",i)){
				node= new V_<double,Single>;
			}
			else if(str.cmpKWord("_S_Vec2f",i)){
				node= new V_<Vec2f,Single>;
			}
			else if(str.cmpKWord("_S_Vec3f",i)){
				node= new V_<Vec3f,Single>;
			}
			else if(str.cmpKWord("_M_Bool",i)){
				node= new V_<S_Bool,Multiple>;
			}
			else if(str.cmpKWord("_M_Color",i)){
				node= new V_<Color3f,Multiple>;
			}
			else if(str.cmpKWord("_M_Float",i)){
				node= new V_<S_Float,Multiple>;
			}
			else if(str.cmpKWord("_M_Image",i)){
				node= new V_<Image,Multiple>;
			}
			else if(str.cmpKWord("_M_Int32",i)){
				node= new V_<int,Multiple>;
			}
			else if(str.cmpKWord("_M_Node",i)){
				node= new V_<NodePtr,Multiple>;
			}
			else if(str.cmpKWord("_M_Rotation",i)){
				node= new V_<Rotation,Multiple>;
			}
			else if(str.cmpKWord("_M_String",i)){
				node= new V_<String,Multiple>;
			}
			else if(str.cmpKWord("_M_Time",i)){
				node= new V_<double,Multiple>;
			}
			else if(str.cmpKWord("_M_Vec2f",i)){
				node= new V_<Vec2f,Multiple>;
			}
			else if(str.cmpKWord("_M_Vec3f",i)){
				node= new V_<Vec3f,Multiple>;
			}
			else goto protoPerhaps;
			break;
		}
		case 'H':{
			if(str.cmpKWord("armonic",i)){
				node = new Harmonic;
			}
			else if(str.cmpKWord("air",i)){
				node = new Hair;
			}
			else if(str.cmpKWord("armonicPatch",i)){
				node = new HarmonicPatch;
			}
			else if(str.cmpKWord("AnimHumanoid",i)){
				node = new HAnimHumanoid;
			}
			else if(str.cmpKWord("umanoid",i)){
				node = new HAnimHumanoid;
			}
			else if(str.cmpKWord("AnimJoint",i)){
				node = new HAnimJoint;
			}
			else if(str.cmpKWord("AnimDisplacer",i)){
				node = new HAnimDisplacer;
			}
			else if(str.cmpKWord("AnimSegment",i)){
				node = new HAnimSegment;
			}
			else if(str.cmpKWord("AnimSite",i)){
				node = new HAnimSite;
			}
			else goto protoPerhaps;
			break;
		}
		case 'I':{
			if(str.cmpKWord("nline",i)){
				node = new Inline;
			}
			else if(str.cmpKWord("ntegerSequencer",i)){
				node = new IntegerSequencer;
			}
			else if(str.cmpKWord("ntegerTrigger",i)){
				node = new IntegerTrigger;
			}
			if(str.cmpKWord("mmediatePart",i)){
				ImmediatePart* part=new ImmediatePart;
				part->smls=smlsLegacy;
				node=part;
			}
			else if(str.cmpKWord("mageTexture",i)){
				node = new ImageTexture;
			}
			else if(str.cmpKWord("ndexedFaceSet",i)){
				node = new IndexedFaceSet;
			}
			else if(str.cmpKWord("ndexedLineSet",i)){
				node = new IndexedLineSet;
			}
			else if(str.cmpKWord("ndexedTriangleSet",i)){
				node = new IndexedTriangleSet;
			}
			else goto protoPerhaps;
			break;
		}
		case 'J':{
			if(str.cmpKWord("oinVertex",i)){
				node = new JoinVertex;
			}
			else if(str.cmpKWord("oinCloseVertex",i)){
				node = new JoinCloseVertex;
			}
			else if(str.cmpKWord("oint",i)){
				node = new HAnimJoint;
			}
			else goto protoPerhaps;
			break;
		}
		case 'L':{
			if(str.cmpKWord("arynx",i)){
				node = new Larynx;
			}
			else if (str.cmpKWord("ine", i)) {
				node = new Line;
			}
			else if(str.cmpKWord("owPass",i)){
				node = new LowPass;
			}
			else if(str.cmpKWord("ock",i)){
				node = new Lock;
			}
			else if(str.cmpKWord("OD",i)){
				node = new LOD;
			}
			else goto protoPerhaps;
			break;
		}
		case 'M':{
			if(str.cmpKWord("aterial",i)){
				node = new Material;
			}
			else if(str.cmpKWord("emberFunction",i)){
				node = newMemberFunction();
			}
			else if(str.cmpKWord("idiFormant",i)){
				node = new MidiFormant;
			}
			else if(str.cmpKWord("obility",i)){
				node = new Mobility;
			}
			else if(str.cmpKWord("odelMorpher",i)){
				node = new ModelMorpher;
			}
			else if(str.cmpKWord("odelSequence",i)){
				node = new ModelSequence;
			}
			else if(str.cmpKWord("onoColorEffect",i)){
				node = new MonoColorEffect;
			}
			else if(str.cmpKWord("orphPart",i)){
				node = new MorphPart;
			}
			else if(str.cmpKWord("orpher",i)){
				node = new Morpher;
			}
			else if(str.cmpKWord("ovie",i)){
				node = new Movie;
			}
			else if(str.cmpKWord("ovieTexture",i)){
				node = new MovieTexture;
			}
			else if(str.cmpKWord("ultiFormant",i)){
				node = new MultiFormant;
			}
			else if(str.cmpKWord("_Bool",i)){
				return decodeInitializedGlobal<S_Bool,Multiple>(str,i, bracketCntr);
			}
			else if(str.cmpKWord("_Float",i)){
				return decodeInitializedGlobal<S_Float,Multiple>(str,i, bracketCntr);
			}
			else if(str.cmpKWord("_Image",i)){
				return decodeInitializedGlobal<Image,Multiple>(str,i, bracketCntr);
			}
			else if(str.cmpKWord("_Int",i)||str.cmpKWord("_Int32",i)){
				return decodeInitializedGlobal<int,Multiple>(str,i, bracketCntr);
			}
			else if(str.cmpKWord("_Node",i)){
				return decodeInitializedGlobal<NodePtr,Multiple>(str,i, bracketCntr);
			}
			else if(str.cmpKWord("_Color3f",i)){
				return decodeInitializedGlobal<Color3f,Multiple>(str,i, bracketCntr);
			}
			else if(str.cmpKWord("_Rotation",i)){
				return decodeInitializedGlobal<Rotation,Multiple>(str,i, bracketCntr);
			}
			else if(str.cmpKWord("_String",i)){
				return decodeInitializedGlobal<String,Multiple>(str,i, bracketCntr);
			}
			else if(str.cmpKWord("_Double",i)){
				return decodeInitializedGlobal<double,Multiple>(str,i, bracketCntr);
			}
			else if(str.cmpKWord("_Vec2f",i)){
				return decodeInitializedGlobal<Vec2f,Multiple>(str,i, bracketCntr);
			}
			else if(str.cmpKWord("_Vec3f",i)){
				return decodeInitializedGlobal<Vec3f,Multiple>(str,i, bracketCntr);
			}
			else if(str.cmpKWord("irrorEdit",i)){
				node = new MirrorEdit;
			}
			else if(str.cmpKWord("irrorGroup",i)){
				node = new MirrorGroup;
			}
			else if(str.cmpKWord("ultiUser",i)){
				node = new MultiUser;
			}
			else if(str.cmpKWord("esh",i)){
				node = new Mesh;
			}
			else goto protoPerhaps;
			break;
		}
		case 'N':{
			if(str.cmpKWord("avigationInfo",i)){
				node = new NavigationInfo;
			}
			else if(str.cmpKWord("urbsPatch",i)){
				NurbsPatch* patch = new NurbsPatch;
				if(g_load_scene == 0 || g_load_scene->version < 2.1725){
					patch->dim.refValue() = Vec2f(2, 2);
				}
				node = patch;
			}
			else if(str.cmpKWord("ikInterpolator",i)){
				node = new NikInterpolator;
			}
			else if(str.cmpKWord("urbsTrack",i)){
				node = new NurbsTrack;
			}
			else if(str.cmpKWord("urbsLathe",i)){
				node = new NurbsLathe;
			}
			else if(str.cmpKWord("urbsLine",i)){
				node = new NurbsLine;
			}
			else if(str.cmpKWord("urbsSurface",i)){
				node = new NurbsSurface;
			}
			else if(str.cmpKWord("ormalInterpolator",i)){
				node = new NormalInterpolator;
			}
			else if(str.cmpKWord("ormal",i)){
				node = new Normal;
			}
			else if(str.cmpKWord("ULL",i)){
				return node;
			}
			else if(str.cmpKWord("oOutputGroup",i)){
				node = new NoOutputGroup;
			}
			else goto protoPerhaps;
			break;
		}
		case 'O':{
			if(str.cmpKWord("rientationInterpolator",i)){
				node = new OrientationInterpolator;
			}
			else if(str.cmpKWord("scillator",i)){
				node = new Oscillator;
			}
			else if(str.cmpKWord("penNetFile",i)){
				node = new OpenNetFile;
			}
			else goto protoPerhaps;
			break;
		}
		case 'P':{
			if(str.cmpKWord("art",i)||str.cmpKWord("artJoiner",i)){
				Part* part = new Part;
				part->smls = smlsLegacy;
				node=part;
			}
			else if(str.cmpKWord("lace",i)){
				node = new Place;
			}
			else if(str.cmpKWord("olymap",i)){
				node = new Polymap;
			}
			else if(str.cmpKWord("hrase",i)){
				node = new Phrase;
			}
			else if(str.cmpKWord("ositionInterpolator",i)){
				node = new PositionInterpolator;
			}
			else if(str.cmpKWord("osCompensator",i)){
				node = new PosCompensator;
			}
			else if(str.cmpKWord("asteOptions",i)){
				node = new PasteOptions;
			}
			else if(str.cmpKWord("ovLight",i)){
				node = new PovLight;
			}
			else if(str.cmpKWord("ovFog",i)){
				node = new PovFog;
			}
			else if(str.cmpKWord("ixelTexture",i)){
				node = new PixelTexture;
			}
			else if(str.cmpKWord("laneSensor",i)){
				node = new PlaneSensor;
			}
			else if(str.cmpKWord("ointLight",i)){
				node = new PointLight;
			}
			else if(str.cmpKWord("ointSet",i)){
				node = new PointSet;
			}
			else if(str.cmpKWord("laceTexture",i)){
				node = new PlaceTexture;
			}
			else if(str.cmpKWord("rojectOntoTri",i)){
				node = new ProjectOntoTri;
			}
			else if(str.cmpKWord("roximitySensor",i)){
				node = new ProximitySensor;
			}
			else if(str.cmpKWord("osAccelerator",i)){
				node = new PosAccelerator;
			}
			else if(str.cmpKWord("olyline2D",i)){
				node = new Polyline2D;
			}
			else if(str.cmpKWord("olyToNurbs",i)){
				node = new PolyToNurbs;
			}
			else if(str.cmpKWord("rotoDeclare",i)){
				node = new ProtoDeclare;
			}
			else if(str.cmpKWord("ROTO",i)){
				ProtoDeclare* protoDeclare=new ProtoDeclare;
				if(g_collapse){
					protoDeclare->collapsed=true;
					g_collapse=false;
				}
				ProtoInterface* protoInterface=new ProtoInterface;
				ProtoBody* protoBody=new ProtoBody;

				protoDeclare->protoInterface=protoInterface;
				protoDeclare->protoBody=protoBody;
				node=protoDeclare;
				g_protoDeclare=protoDeclare;
				str.stb(i);
				protoDeclare->name=str.getWord(i);
				str.stb(i);
				protoDeclare->getFields(str, i);
				addNewProtoDeclareToList(protoDeclare);
				return node;
			}
			else if(str.cmpKWord("rotoBody",i)){
				node = new ProtoBody;
			}
			else if(str.cmpKWord("alette",i)){
				node = new Palette;
			}

			else goto protoPerhaps;
			break;
		}
		case 'R':{
			if(str.cmpKWord("adialTug",i)){
				node = new RadialTug;
			}
			else if(str.cmpKWord("ectangleBuild",i)){
				node = new RectangleBuild;
			}
			else if(str.cmpKWord("obotSave",i)){
				node = new RobotSave;
			}
			else if(str.cmpKWord("otAccelerator",i)){
				node = new RotAccelerator;
			}
			else if(str.cmpKWord("otation",i)){
				return decodeInitializedGlobal<Rotation,Single>(str,i, bracketCntr);
			}
			else if(str.cmpKWord("otFilter",i)){
				node = new RotFilter;
			}
			else if(str.cmpKWord("otMult",i)){
				node = new RotMult;
			}
			else if(str.cmpKWord("elateMorpher",i)){
				node = new RelateMorpher;
			}
			else if(str.cmpKWord("ingBuild",i)){
				node = new RingBuild;
			}
			else goto protoPerhaps;
			break;
		}
		case 'S':{
			if(str.cmpKWord("trand",i)){
				node = new Strand;
			}
			else if(str.cmpKWord("eamless",i)||str.cmpKWord("CompileParts",i)){ //"Seamless"
				//legacy
				smlsLegacy = new Seamless;
				smlsLegacy->geometry = new IndexedFaceSet;
				if(g_load_scene->version == 0 || g_load_scene->version >= 2.173){
					smlsLegacy->bnspe.refValue() = true;
				}
				node = smlsLegacy;
			}
			else if(str.cmpKWord("hape",i)){ //Shape
				node = new Shape;
			}
			else if(str.cmpKWord("cene",i)){
				//static ScenePtr oldScene=g_scene; //dont know why this is needed but it stops it crashing when it makes a flat list for some reason
				g_load_scene = new Scene;
				node = g_load_scene;
			}
			else if(str.cmpKWord("imulSineTug",i)){
				node = new SimulSineTug;
			}
			else if(str.cmpKWord("ineTug",i)){
				node = new SineTug;
			}
			else if(str.cmpKWord("tem",i)){
				Stem* stem = new Stem;
				node=stem;
				if(g_load_scene==0 || g_load_scene->version<2.1225){
					stem->staves.refValue()=8;
				}
				if(g_load_scene==0 || g_load_scene->version<2.1275){
					stem->add.refValue()=false;
				}
				if(g_load_scene==0 || g_load_scene->version < 2.1725){
					stem->uWrap.refValue() = true;
				}
			}
			else if(str.cmpKWord("tem2",i)){
				node = new Stem;
			}
			else if(str.cmpKWord("ubdivision",i)){
				node = new Subdivision;
			}
			else if(str.cmpKWord("urfaceGenerator",i)){
				node = new SurfaceGenerator;
			}
			else if(str.cmpKWord("urfaceTransform",i)){
				node = new SurfaceTransform;
			}
			else if(str.cmpKWord("CGroupBuild",i)){
				node = new SCGroupBuild;
			}
			else if(str.cmpKWord("cript",i)){
				node = new Script;
			}
			else if(str.cmpKWord("taticPart",i)){
				Part* part=new Part;
				part->smls=smlsLegacy;
				node = part;
			}
			else if(str.cmpKWord("taticGroup",i)){
				node = new StaticGroup;
			}
			else if(str.cmpKWord("calarInterpolator",i)){
				node = new ScalarInterpolator;
			}
			else if(str.cmpKWord("ound",i)){
				node = new Sound;
			}
			else if(str.cmpKWord("ample",i)){
				node = new Sample;
			}
			else if(str.cmpKWord("phere",i)){
				node = new Sphere;
			}
			else if(str.cmpKWord("phereBuild",i)){
				SphereBuild* sphere = new SphereBuild;
				if(g_load_scene == 0 || g_load_scene->version < 2.161 - .0005){
					sphere->yCentred.refValue() = false;
				}
				node = sphere;
			}
			else if(str.cmpKWord("phereSensor",i)){
				node = new SphereSensor;
			}
			else if(str.cmpKWord("potLight",i)){
				node = new SpotLight;
			}
			else if(str.cmpKWord("witch",i)){
				node = new Switch;
			}
			else if(str.cmpKWord("plitTri",i)){
				node = new SplitTri;
			}
			else if(str.cmpKWord("wapTriDivide",i)){
				node = new SwapTriDivide;
			}
			else if(str.cmpKWord("ettings",i)){
				node = new Settings;
			}
			else if(str.cmpKWord("_Bool",i)){
				return decodeInitializedGlobal<S_Bool,Single>(str,i, bracketCntr);
			}
			else if(str.cmpKWord("_Color3f",i)){
				return decodeInitializedGlobal<Color3f,Single>(str,i, bracketCntr);
			}
			else if(str.cmpKWord("_Float",i)){
				return decodeInitializedGlobal<S_Float,Single>(str,i, bracketCntr);
			}
			else if(str.cmpKWord("_Image",i)){
				return decodeInitializedGlobal<Image,Single>(str,i, bracketCntr);
			}
			else if(str.cmpKWord("_Int",i)){
				return decodeInitializedGlobal<int,Single>(str,i, bracketCntr);
			}
			else if(str.cmpKWord("_NodePtr",i)){
				return decodeInitializedGlobal<NodePtr,Single>(str,i, bracketCntr);
			}
			else if(str.cmpKWord("_Rotation",i)){
				return decodeInitializedGlobal<Rotation,Single>(str,i, bracketCntr);
			}
			else if(str.cmpKWord("_String",i)){
				return decodeInitializedGlobal<String,Single>(str,i, bracketCntr);
			}
			else if(str.cmpKWord("_Double",i)){
				return decodeInitializedGlobal<double,Single>(str,i, bracketCntr);
			}
			else if(str.cmpKWord("_Vec2f",i)){
				return decodeInitializedGlobal<Vec2f,Single>(str,i, bracketCntr);
			}
			else if(str.cmpKWord("_Vec3f",i)){
				return decodeInitializedGlobal<Vec3f,Single>(str,i, bracketCntr);
			}
			else if(str.cmpKWord("egment",i)){
				node = new HAnimSegment;
			}
			else if(str.cmpKWord("ite",i)){
				node = new HAnimSite;
			}
			else goto protoPerhaps;
			break;
		}
		case 'T':{
			if(str.cmpKWord("rack",i)){
				node = new Track;
			}
			else if(str.cmpKWord("imelineFilter",i)){
				node = new TimelineFilter;
			}
			else if(str.cmpKWord("imeSensor",i)){
				node = new TimeSensor;
			}
			else if(str.cmpKWord("imeTrigger",i)){
				node = new TimeTrigger;
			}
			else if(str.cmpKWord("ransform",i)){
				node = new Transform;
			}
			else if(str.cmpKWord("ransformCP",i)){
				node = new TransformCP;
			}
			else if(str.cmpKWord("extureCoordinate",i)){
				node = new TextureCoordinate;
			}
			else if(str.cmpKWord("ransferBCyl",i)){
				node = new TransferBCyl;
			}
			else if(str.cmpKWord("apeMeasure",i)){
				node = new TapeMeasure;
			}
			else if(str.cmpKWord("ouchSensor",i)){
				node = new TouchSensor;
			}
			else if(str.cmpKWord("ransferOwnership",i)){
				node = new TransferOwnership;
			}
			else if(str.cmpKWord("ransferVertex",i)){
				node = new TransferVertex;
			}
			else if(str.cmpKWord("extureMapper",i)){
				TextureMapper* textureMapper = new TextureMapper;
				node=textureMapper;
				if(g_load_scene == 0 || g_load_scene->version < 2.130-.0005){
					textureMapper->autoSize3d.refValue()=true;
				}
			}
			else if(str.cmpKWord("extureEffect",i)){
				node = new TextureEffect;
			}
			else if(str.cmpKWord("ext",i)){
				node = new Text;
			}
			else if(str.cmpKWord("orusBuild",i)){
				node = new TorusBuild;
			}
			else if(str.cmpKWord("extureTransform",i)){
				node = new TextureTransform;
			}
			else goto protoPerhaps;
			break;
		}
		case 'V':{
			if(str.cmpKWord("ec2f",i)){
				return decodeInitializedGlobal<Vec2f,Single>(str,i, bracketCntr);
			}
			else if(str.cmpKWord("ec3f",i)){
				return decodeInitializedGlobal<Vec3f,Single>(str,i, bracketCntr);
			}
			else if(str.cmpKWord("ertexBank",i)){
				Morpher* morpher=new Morpher;
				morpher->smls=smlsLegacy;
				node=morpher;
			}
			else if(str.cmpKWord("iewpoint",i)){
				node = new Viewpoint;
			}
			else if(str.cmpKWord("ideo",i)){
				node = new Video;
			}
			else if(str.cmpKWord("isibilitySensor",i)){
				node = new VisibilitySensor;
			}
			else goto protoPerhaps;
			break;
		}
		case 'W':{
			if(str.cmpKWord("aveToPoly",i)){
				node = new WaveToPoly;
			}
			else if(str.cmpKWord("orldInfo",i)){
				node = new WorldInfo;
			}
			else goto protoPerhaps;
			break;
		}
		case 'b':{
			if(str.cmpKWord("ool",i)){
				return decodeInitializedGlobal<bool,Single>(str,i, bracketCntr);
			}
			else goto protoPerhaps;
			break;
		}
		case 'd':{
			if(str.cmpKWord("ouble",i)){
				return decodeInitializedGlobal<double,Single>(str,i, bracketCntr);
			}
			else goto protoPerhaps;
			break;
		}
		case 'f':{
			if(str.cmpKWord("loat",i)){
				return decodeInitializedGlobal<float,Single>(str,i, bracketCntr);
			}
			else goto protoPerhaps;
			break;
		}
		case 'i':{
			if(str.cmpKWord("nt",i)){
				return decodeInitializedGlobal<int,Single>(str,i, bracketCntr);
			}
			else goto protoPerhaps;
			break;
		}
		case 'v':{
			if(str.cmpKWord("oid",i)){
				return decodeInitializedGlobal<void,Single>(str,i, bracketCntr);
			}
			else goto protoPerhaps;
			break;
		}
		case '*':{
			if(str.cmpKWord("PROTO",i)){
				ProtoInterface* protoInterface=new ProtoInterface;
				node=protoInterface;
				str.stb(i);
				str.stwo(i);//nodeLookUpStr+=str.getWord(i);
				str.stb(i);
				if(str[i]!='['){
					containerFieldLookUpStr+=str.getWord(i);
					str.stb(i);
				}
				else{
					containerFieldLookUpStr.length++;
				}

				protoInterface->getFields(str, i);
				if(str[i]!='{'){
					throw node;
				}
				str.stb(++i);
				i++;
				return node;
			}
			else throw;
		}
		default: {
protoPerhaps:
			--i;
			String protoName=str.getWord(i);
			for(int k=g_userProtoBegin; k<nodeLookUpStr.length; k++){
				if(nodeLookUpStr[k]==protoName){;
					ProtoInstance* protoInstance=new ProtoInstance(k);
					node=protoInstance;
					protoInstance->userNodeType=k;

					node->name=strDefName;
					g_load_scene->defList += node;
					//node->name=protoName;
					if(g_collapse){
						node->collapsed=true;
						g_collapse=false;
					}
					node->getNode(str, i);
					return node;
				}
			}
			dLog(protoName+" unknown node!!!!");
			throw 0;
		}
	}
	if(node==0){
		throw;
	}
	node->name=strDefName;
	g_load_scene->defList += node;
	if(g_collapse){
		node->collapsed=true;
		g_collapse=false;
	}
	node->getNode(str, i);

	if(Stem* stem = dynamic_cast<Stem*>(node)){
		if(g_load_scene == 0 || g_load_scene->version < 2.130 - .0005){
			if(stem->textureMapper.length > 0 && stem->slide!=Vec2f()){
				TextureMapper* textureMapper = dynamic_cast<TextureMapper*>(stem->textureMapper[0].p);
				textureMapper->autoSize3d.refValue() = false;
				textureMapper->position3d.refValue().x+=stem->slide.x;
				textureMapper->position3d.refValue().y+=stem->slide.y;
			}
		}
		if(g_load_scene == 0 || g_load_scene->version < 2.164 - .0005){
			if(stem->source.length){
				stem->store.refValue() = true;
			}
		}
	}
	else if(NurbsLathe* lathe = dynamic_cast<NurbsLathe*>(node)){
		if(g_load_scene == 0 || g_load_scene->version < 2.162 - .0005){
			lathe->endSP.refValue() = false;
			lathe->beginSP.refValue() = false;
			if(lathe->point[0].x == 0){
				lathe->beginSP.refValue() = true;
			}
			if(lathe->point[lathe->point.length - 1].x == 0){
				lathe->endSP.refValue() = true;
			}
		}
	}
	else if(NurbsPatch* patch = dynamic_cast<NurbsPatch*>(node)){
		if(patch->scale != Vec3f(1,1,1) || patch->rotation != Rotation() || patch->translation != Vec3f()){

			patch->transformPoints(patch->scale, patch->rotation, patch->translation);

			patch->scale.refValue() = Vec3f(1, 1, 1);
			patch->rotation.refValue() = Rotation();
			patch->translation.refValue() = Vec3f();
		}
	}
	return node;
}

void NurbsPatch::bnspeBuildPolies(){
	if(smls->bnspe){
		if(stem.length){
			if(dynamic_cast<Stem*>(stem[0].p)){
				smls->deletePoliesOwnedByPatch(this);
				doBuild();
				smls->ctrf = true;
			}
		}
	}
}

void NurbsPatch::bnspeBuildSyncPolies(){//exactly same as above except we do not delete anything before we build
	if(smls->bnspe && sync){
		if(stem.length){
			if(dynamic_cast<Stem*>(stem[0].p)){
				doBuild();
				smls->ctrf = true;
			}
		}
	}
}

void NurbsPatch::transformPoints(Vec3f scale, Rotation rotation, Vec3f translation){
	Matrix matrix = scaleMatrix(scale) * rotationMatrix(rotation) * translationMatrix(translation);
	M_Vec3f& point = this->point.refValue();

	for(int i = point.length; --i>=0;){
		float w = weight[i];
		point[i] = matrix * (point[i] / w) * w;
	}
}

void IndexedFaceSet::triangle(
	float ax, float ay, float az,
	float bx, float by, float bz,
	float cx, float cy, float cz
	){
	Coordinate* coordinate = new Coordinate();
	coord.p = coordinate;
	M_Vec3f& point = coordinate->point.refValue();
	M_Int& coordIndex = this->coordIndex.refValue();
	point.length=3;

	point[0]=Vec3f(ax,ay,az);
	point[1]=Vec3f(bx,by,bz);
	point[2]=Vec3f(cx,cy,cz);
	coordIndex.length=4;
	coordIndex[0]=0;
	coordIndex[1]=1;
	coordIndex[2]=2;
	coordIndex[3]=-1;
}
/*
void IndexedFaceSet::getFields( String& str,int& i){

	while(str[i]!='}'){
		if(		str.cmpKWord("coord",i))	coord=getNode2(str,i);
		else if(str.cmpKWord("normal",i))	normal=getNode2(str,i);
		else if(str.cmpKWord("color",i))	color=getNode2(str,i);
		else if(str.cmpKWord("texCoord",i))	texCoord=getNode2(str,i);
		else if(str.cmpKWord("coordIndex",i)) coordIndex.getField(str,i);
		else if(str.cmpKWord("texCoordIndex",i)) texCoordIndex.getField(str,i);
		else if(str.cmpKWord("creaseAngle",i))creaseAngle=getFloat(str, i);
		else if(str.cmpKWord("ccw",i))			ccw=getBool(str,i);
		else if(str.cmpKWord("colorIndex",i))	colorIndex.getField(str,i);
		else if(str.cmpKWord("colorPerVertex",i))colorPerVertex=getBool(str,i);
		else if(str.cmpKWord("convex",i))convex=getBool(str,i);
		else if(str.cmpKWord("normalIndex",i))normalIndex.getField(str,i);
		else if(str.cmpKWord("normalPerVertex",i))normalPerVertex=getBool(str,i);
		else if(str.cmpKWord("solid",i))		solid=getBool(str,i);
		else if(str.cmpKWord("texCoordIndex",i))texCoordIndex.getField(str,i);
		else throw this;

		str.stb(i);
	}
	if(coordIndex.length&&coordIndex[coordIndex.length-1]!=-1){
		coordIndex+=-1;
	}
	for(int j=3; j<coordIndex.length; j+=4){
		if(coordIndex[j]!=-1){
			M_Int triangles;
			triangles.resize(coordIndex.length);
			for(int h=0,j=0; j<coordIndex.length; ){
				int g=j;
				triangles[h++]=coordIndex[j++];
				triangles[h++]=coordIndex[j++];
				triangles[h++]=coordIndex[j++];
				triangles[h++]=-1;
				if(coordIndex[j++]==-1){
					continue;
				}
				else {
					for(; j<coordIndex.length; ){
						triangles.addLength(3);
						j-=2;
						triangles[h++]=coordIndex[g];
						triangles[h++]=coordIndex[j++];
						triangles[h++]=coordIndex[j++];
						triangles[h++]=-1;
						if(coordIndex[j++]==-1){
							break;
						}
					}
				}
			}
			if(triangles.length){
				coordIndex=triangles;
			}
		}
	}

	str.stb(i);
	return;

}
*/


void IndexedFaceSet::genNormals(){
	Coordinate* coordinate = dynamic_cast<Coordinate*>(coord.p);
	Normal* normal = dynamic_cast<Normal*>(this->normal.p);
	if(coordinate == 0 || normal == 0){
		return;
	}
	M_Vec3f& vector = normal->vector.refValue();
	M_Vec3f& point = coordinate->point.refValue();

	M_Int& coordIndex = this->coordIndex.refValue();
	int TriSum = coordIndex.length / 4;
	M_Vec3f flatNormals;
	flatNormals.length = TriSum;
	int normalLen = point.length;
	vector.length = point.length;
	int* fnormal = (int*)vector.getArray();//clearMemory
	for(int i = 0; i < normalLen * 3; i++){
		fnormal[i] = 0;
	}
	for(int i = 0;i<TriSum;i++){
		calcNormal(
			point[coordIndex[i*4]],
			point[coordIndex[i*4+2]],
			point[coordIndex[i*4+1]],
			flatNormals[i]
		);
	}
	for(int ti = 0,i = 0;ti < TriSum; ti++){
		vector[coordIndex[i]]+=flatNormals[ti];
		i++;
		vector[coordIndex[i]]+=flatNormals[ti];
		i++;
		vector[coordIndex[i]]+=flatNormals[ti];
		i++;
		i++;
	}
	for(int i = 0; i < vector.length; i++){
		vector[i] = vector[i].normalize();
	}
}
void IndexedTriangleSet::genNormals(){
	Coordinate* coordinate = dynamic_cast<Coordinate*>(coord.p);
	Normal* normal = dynamic_cast<Normal*>(this->normal.p);
	if(coordinate == 0 || normal == 0){
		return;
	}
	M_Vec3f& vector = normal->vector.refValue();
	M_Vec3f& point = coordinate->point.refValue();

	M_Int& index=this->index.refValue();
	int TriSum = index.length / 3;
	M_Vec3f flatNormals;
	flatNormals.length = TriSum;
	int normalLen = point.length;
	vector.length = normalLen;
	int* fnormal=(int*)vector.getArray();//clearMemory
	for(int i = 0; i < normalLen * 3; i++){
		fnormal[i] = 0;
	}
	for(int i = 0; i < TriSum; i++){
		calcNormal(
			point[index[i*3]],
			point[index[i*3+2]],
			point[index[i*3+1]],
			flatNormals[i]
		);
	}
	for(int ti = 0,i = 0;ti < TriSum; ti++){
		vector[index[i]] += flatNormals[ti];
		i++;
		vector[index[i]] += flatNormals[ti];
		i++;
		vector[index[i]] += flatNormals[ti];
		i++;
	}
	for(int i = 0; i < vector.length; i++){
		vector[i] = vector[i].normalize();
	}
}

void Mesh::genNormals(){
	M_Vec3f& normal=this->normal.refValue();
	int i,ti;
	int TriSum=coordIndex.length/3;
	M_Vec3f flatNormals;
	flatNormals.length=TriSum;
	int coordLen=coord.length;
	normal.length=coordLen;
	int* fnormal=(int*)normal.getArray();//clearMemory
	for(i=coordLen*3;--i>=0;){
		fnormal[i]=0;
	}
	for(i=0;i<TriSum;i++){

		Vec3f coord0=coord[redirectCoord(coordIndex[i*3],coordLen)];
		Vec3f coord1=coord[redirectCoord(coordIndex[i*3+1],coordLen)];
		Vec3f coord2=coord[redirectCoord(coordIndex[i*3+2],coordLen)];

		calcNormal(
			coord0,
			coord2,
			coord1,

			flatNormals[i]
		);
	}
	for(ti=0,i=0;ti<TriSum;ti++){
		normal[redirectCoord(coordIndex[i],coordLen)]+=flatNormals[ti];
		i++;
		normal[redirectCoord(coordIndex[i],coordLen)]+=flatNormals[ti];
		i++;
		normal[redirectCoord(coordIndex[i],coordLen)]+=flatNormals[ti];
		i++;
	}
	for(i=normal.length;--i>=0;){
		normal[i]=normal[i].normalize();
	}
}
void Art::genNormals(){
	int i;
	int triLen=triangle.length;
	M_Vec3f flatNormals; flatNormals.length=triLen;
	int normalLen=coord.length;
	M_Vec3f& normal=this->normal;
	normal.length=normalLen;
	int* fnormal=(int*)normal.getArray();//clearMemory
	for(i=0;i<normalLen*3;i++){
		fnormal[i]=0;
	}
	for(i=triLen;--i>=0;){
		calcNormal(
			coord[triangle[i].ic],
			coord[triangle[i].ib],
			coord[triangle[i].ia],
			flatNormals[i]
		);
	}
	for(i=triLen;--i>=0;){
		normal[triangle[i].ia]+=flatNormals[i];
		normal[triangle[i].ib]+=flatNormals[i];
		normal[triangle[i].ic]+=flatNormals[i];
	}
	for(i=normal.length;--i>=0;){
		normal[i]=normal[i].normalize();
	}
}

void Art::genNormals(int h, int beginTriOffset){
	int i;
	int triLen=triangle.length;
	M_Vec3f flatNormals; flatNormals.length=triLen;
	int normalLen=coord.length;
	M_Vec3f& normal=this->normal;
	normal.length=normalLen;
	int* fnormal=(int*)normal.getArray();//clearMemory
	for(i = normalLen*3; --i >= h*3;){
		fnormal[i]=0;
	}
	for(i=triLen; --i>=beginTriOffset;){
		calcNormal(
			coord[triangle[i].ic],
			coord[triangle[i].ib],
			coord[triangle[i].ia],
			flatNormals[i]
		);
	}
	for(i=triLen; --i>=beginTriOffset;){
		normal[triangle[i].ia]+=flatNormals[i];
		normal[triangle[i].ib]+=flatNormals[i];
		normal[triangle[i].ic]+=flatNormals[i];
	}
	for(i = normal.length; --i >= h;){
		normal[i]=normal[i].normalize();
	}
}

#define CalcoordLen numRings * numStaves + bottomIsPoint + topIsPoint
#define CalNumIndices (numStaves-1)*(numRings-1)*2*3+(numStaves-1)*(bottomIsPoint + topIsPoint)*3

extern int maxBufSize;

void IndexedFaceSet::write(const String filename){
	static String str;
	str="";

	str+="#VRML V2.0 utf8\n";
	str+="#Generated by Seamless3d Code ";
	//str+=g_version;
	str+=" Copyright � 2002-2007 Graham John Perrett thyme@seamless3d.com www.seamless3d.com\n";
	str+="NavigationInfo {type \"examine\"}\n";

	str+="Shape {\nappearance Appearance{material Material {}}\n";
	if(this!=NULL){
		str+="\ngeometry ";
		this->toString(str, true);
	}
	str+="}\n";
	str.write(filename);
}


void rotXMVec3f(Vec3f* point,int AngleCnt,float angle){
float newZ,newY;
int i;
	float CosAngle=(float)cos(angle);
	float SinAngle=(float)sin(angle);
	for (i=0;i<AngleCnt;i++,point++){
		newY=point->y*CosAngle-point->z*SinAngle;
		newZ=point->y*SinAngle+point->z*CosAngle;
		point->y=newY;
		point->z=newZ;
	}
}
void rotYMVec3f(Vec3f* point,int AngleCnt,float angle){
float newZ,newX;
int i;
	float CosAngle=(float)cos(angle);
	float SinAngle=(float)sin(angle);
	for (i=0;i<AngleCnt;i++,point++){
		newZ=point->z*CosAngle-point->x*SinAngle;
		newX=point->z*SinAngle+point->x*CosAngle;
		point->z=newZ;
		point->x=newX;
	}
}
void rotZMVec3f(Vec3f* point,int AngleCnt,float angle){
float newY,newX;
int i;
	float CosAngle=(float)cos(angle);
	float SinAngle=(float)sin(angle);
	for (i=0;i<AngleCnt;i++,point++){
		newX=point->x*CosAngle-point->y*SinAngle;
		newY=point->x*SinAngle+point->y*CosAngle;
		point->x=newX;
		point->y=newY;
	}
}

void scaleMVec3f(Vec3f* point,int count,float sx,float sy,float sz){
	for (int i=0;i<count;i++){
		point[i].x*=sx;
		point[i].y*=sy;
		point[i].z*=sz;
	}
}

void lineIntersection(float x1, float y1,
					float x2, float y2,
					float x3, float y3,
					float x4, float y4,
					float& xi,float& yi){
	float ua=((x4-x3)*(y1-y3)-(y4-y3)*(x1-x3))/((y4-y3)*(x2-x1)-(x4-x3)*(y2-y1));
	//float ub=((x2-x1)*(y1-y3)-(y2-y1)*(x1-x3))/((y4-y3)*(x2-x1)-(x4-x3)*(y2-y1));

	xi = (float)(x1 + ua *(x2 - x1));
	yi = (float)(y1 + ua *(y2 - y1));//should this be ub?????????
}

/*
void IndexedFaceSet::negateNormals(){
	NormalPtr normalPtr=normal;
	for(int i=0;i<normalPtr->vector.length;i++){
		normalPtr->vector[i]=normalPtr->vector[i].negate();
	}
}
*/
/*
void IndexedFaceSet::genIndices(){
	int i,m,j;
	M_Int& coordIndex=refCoordIndex();
	coordIndex.resize((numStaves-1)*(numRings-1)*2*4+(numStaves-1)*(bottomIsPoint + topIsPoint)*4);

	int b = 0;
	// Generate triangles for bottom cap
	if(bottomIsPoint){
		for(i = 1; i < numStaves; i++ ){
			coordIndex[b++] = 0;
			coordIndex[b++] = i + 1;
			coordIndex[b++] = i;
			coordIndex[b++] = -1;
		}
	}
	// Generate triangles for top cap.
	if(topIsPoint){
		int t = coordIndex.length - (numStaves-1) * 4;
		CoordinatePtr pointPtr=coord;
		for(i=pointPtr->point.length-numStaves-topIsPoint; i < pointPtr->point.length-1-topIsPoint; i++ ){
			coordIndex[t++] = pointPtr->point.length-1;
			coordIndex[t++] = i;
			coordIndex[t++] = i + 1;
			coordIndex[t++] = -1;
		}
	}
	// Generate triangles for the rings
	for (m = bottomIsPoint; b<coordIndex.length-(topIsPoint*(numStaves-1)*4); m += numStaves){
		for(j = 0;j < numStaves-1;j++){
			coordIndex[b++] = m + j + 1;
			coordIndex[b++] = m + j + numStaves;
			coordIndex[b++] = m + j;
			coordIndex[b++] = -1;

			coordIndex[b++] = m + j + 1;
			coordIndex[b++] = m + j + numStaves + 1;
			coordIndex[b++] = m + j + numStaves;
			coordIndex[b++] = -1;
		}
	}
}

void IndexedFaceSet::GenIndices(){
	genIndices();
}
*/
/*
IndexedFaceSetPtr IndexedFaceSet::copy(){
	M_Vec3f& point=refPoint();
	M_Vec3f& vector=refVector();
	M_Color3f& color=refColor();
	M_Vec2f& texPoint=refTexPoint();
	M_Int& coordIndex=refCoordIndex();

	IndexedFaceSetPtr sec = new IndexedFaceSet;
	M_Int& secCoordIndex=sec->refCoordIndex();

	sec->creaseAngle=creaseAngle;

	if(coord!=NULL){
		CoordinatePtr coordPtr=new Coordinate;
		coordPtr->point=refPoint();
		sec->coord=coordPtr;
	}
	if(&vector!=0){
		NormalPtr normalPtr=new Normal;
		normalPtr->vector=refVector();
		sec->normal=normalPtr;
	}
	if(&color!=NULL){
		ColorPtr colorPtr=new Color;
		colorPtr->color=refColor();
		sec->color=colorPtr;
	}
	if(&texPoint!=NULL){
		TextureCoordinatePtr texPointPtr=new TextureCoordinate;
		texPointPtr->point=refTexPoint();
		sec->texCoord=texPointPtr;
	}
	sec->texCoordIndex=texCoordIndex;
	sec->coordIndex=coordIndex;
	return sec;
}
*/
/*
void IndexedFaceSet::copyNotClose(IndexedFaceSetPtr sec){
	if(sec==this)return;
	creaseAngle=sec->creaseAngle;
	numStaves=sec->numStaves;
	numRings=sec->numRings;
	bottomIsPoint=sec->bottomIsPoint;
	topIsPoint=sec->topIsPoint;

	M_Vec3f& secPoint=sec->refPoint();
	M_Vec3f& secVector=sec->refVector();
	M_Color3f& secColor=sec->refColor();
	M_Vec2f& secTexPoint=sec->refTexPoint();
	int change=0;


	if(sec->coord!=0){
		if(coord==0){
			coord=new Coordinate;
		}
		M_Vec3f& point=refPoint();
		refPoint().length=secPoint.length;

		for(int i=refPoint().length; --i>=0;){
			if(notCloseEnough(point[i],secPoint[i])){
				point[i]=secPoint[i];
				change++;
			}
		}
	}
	else coord=0;
	if(sec->normal!=0){
		if(normal==0){
			normal=new Normal;
		}
		M_Vec3f& vector=refVector();
		vector.length=secVector.length;

		for(int i=refVector().length; --i>=0;){
			if(notCloseEnough(vector[i],secVector[i])){
				vector[i]=secVector[i];
				change++;
			}
		}
	}
	else normal=0;
	if(sec->color!=0){
		if(color==0){
			color=new Color;
		}
		M_Color3f& fColor=refColor();
		fColor.length=secColor.length;

		for(int i=fColor.length; --i>=0;){
			if(notCloseEnough(fColor[i],secColor[i])){
				fColor[i]=secColor[i];
				change++;
			}
		}
	}
	else color=0;
	if(sec->texCoord!=0){
		if(texCoord==0){
			texCoord=new TextureCoordinate;
		}
		M_Vec2f& texPoint=refTexPoint();
		texPoint.length=secTexPoint.length;

		for(int i=texPoint.length; --i>=0;){
			if(notCloseEnough(texPoint[i],secTexPoint[i])){
				texPoint[i]=secTexPoint[i];
				change++;
			}
		}
	}
	else texCoord=0;
	coordIndex=sec->coordIndex;
	normalIndex=sec->normalIndex;
	colorIndex=sec->colorIndex;
	texCoordIndex=sec->texCoordIndex;
	if(change){
		dLog(String("change ")+change);
	}

}
*/

void IndexedFaceSet::scale(float sc){
	if(Coordinate* coord = dynamic_cast<Coordinate*>(this->coord.p)){
		M_Vec3f &point = coord->point.refValue();
		for (int i = 0; i < point.length; i++){
			point[i].x *= sc;
			point[i].y *= sc;
			point[i].z *= sc;
		}
	}
}

void IndexedFaceSet::scale(float xsc,float ysc,float zsc){
	if(Coordinate* coord = dynamic_cast<Coordinate*>(this->coord.p)){
		M_Vec3f &point = coord->point.refValue();
		point.scale( xsc, ysc, zsc);
	}
}
void IndexedFaceSet::scale(Vec3f sc){
	scale(sc.x, sc.y, sc.z);
}
void rotX(float angle, M_Vec3f& coord, M_Vec3f& normal){
float newZ,newY;
int i;

	for (i=coord.length;--i>=0;){
		newY=(float)(coord[i].y*cos(angle)-coord[i].z*sin(angle));
		newZ=(float)(coord[i].y*sin(angle)+coord[i].z*cos(angle));
		coord[i].y=newY;
		coord[i].z=newZ;
	}
	for (i=normal.length; --i>=0;){
		newY=(float)(normal[i].y*cos(angle)-normal[i].z*sin(angle));
		newZ=(float)(normal[i].y*sin(angle)+normal[i].z*cos(angle));
		normal[i].y=newY;
		normal[i].z=newZ;
	}
}
void rotY(float angle, M_Vec3f& coord, M_Vec3f& normal){
float newZ,newX;
int i;
	for (i=coord.length;--i>=0;){
		newZ=(float)(coord[i].z*cos(angle)-coord[i].x*sin(angle));
		newX=(float)(coord[i].z*sin(angle)+coord[i].x*cos(angle));
		coord[i].z=newZ;
		coord[i].x=newX;
	}
	for (i=normal.length; --i>=0;){
		newZ=(float)(normal[i].z*cos(angle)-normal[i].x*sin(angle));
		newX=(float)(normal[i].z*sin(angle)+normal[i].x*cos(angle));
		normal[i].z=newZ;
		normal[i].x=newX;
	}
}
void rotY(float angle, M_Vec3f& coord){
float newZ,newX;
int i;
	for (i=coord.length;--i>=0;){
		newZ=(float)(coord[i].z*cos(angle)-coord[i].x*sin(angle));
		newX=(float)(coord[i].z*sin(angle)+coord[i].x*cos(angle));
		coord[i].z=newZ;
		coord[i].x=newX;
	}
}
void rotZ(float angle, M_Vec3f& coord, M_Vec3f& normal){
float newY,newX;
int i;

	float CosAngle=(float)cos(angle);
	float SinAngle=(float)sin(angle);
	for (i=coord.length;--i>=0;){
		newX=(float)(coord[i].x*CosAngle-coord[i].y*SinAngle);
		newY=(float)(coord[i].x*SinAngle+coord[i].y*CosAngle);
		coord[i].x=newX;
		coord[i].y=newY;

	}
	for (i=normal.length; --i>=0;){
		newX=(float)(normal[i].x*CosAngle-normal[i].y*SinAngle);
		newY=(float)(normal[i].x*SinAngle+normal[i].y*CosAngle);
		normal[i].x=newX;
		normal[i].y=newY;
	}
}
void rotZ(float angle, M_Vec3f& coord){
float newY,newX;
int i;

	float CosAngle=(float)cos(angle);
	float SinAngle=(float)sin(angle);
	for (i=coord.length;--i>=0;){
		newX=(float)(coord[i].x*CosAngle-coord[i].y*SinAngle);
		newY=(float)(coord[i].x*SinAngle+coord[i].y*CosAngle);
		coord[i].x=newX;
		coord[i].y=newY;
	}
}

void IndexedFaceSet::swapTriangles(int poly, int search){
	M_Int& coordIndex=refCoordIndex();
	M_Int& texCoordIndex=refTexCoordIndex();
	int temp;
	int pdx=poly*4;
	int sdx=search*4;
	temp=coordIndex[pdx+0]; coordIndex[pdx+0]=coordIndex[sdx+0]; coordIndex[sdx+0]=temp;
	temp=coordIndex[pdx+1]; coordIndex[pdx+1]=coordIndex[sdx+1]; coordIndex[sdx+1]=temp;
	temp=coordIndex[pdx+2]; coordIndex[pdx+2]=coordIndex[sdx+2]; coordIndex[sdx+2]=temp;

	if(texCoordIndex.length){
		temp=texCoordIndex[pdx+0]; texCoordIndex[pdx+0]=texCoordIndex[sdx+0]; texCoordIndex[sdx+0]=temp;
		temp=texCoordIndex[pdx+1]; texCoordIndex[pdx+1]=texCoordIndex[sdx+1]; texCoordIndex[sdx+1]=temp;
		temp=texCoordIndex[pdx+2]; texCoordIndex[pdx+2]=texCoordIndex[sdx+2]; texCoordIndex[sdx+2]=temp;
	}
}
void Art::fastDeleteVertex(int vertex){

	int length=coord.length-1;

	for(int i=vertex; i<length; i++){
		coord.refValue()[i]=coord.refValue()[i+1];
	}
	coord.resize(length);
	if(normal.length){
		for(int i=vertex; i<length; i++){
			normal.refValue()[i]=normal.refValue()[i+1];
		}
		normal.resize(length);
	}
	if(color.length){
		for(int i=vertex; i<length; i++){
			color[i]=color[i+1];
		}
		color.length=length;
	}
	if(texCoord.length){
		for(int i=vertex; i<length; i++){
			texCoord[i]=texCoord[i+1];
		}
		texCoord.length=length;
	}
}
/*
inline bool transferTriToRelevant(S_Triangle& sourceTri, Art*& source){
	M_Art& relevantPart=source->relevantPart;
	for(int k=relevantPart.length;--k>=0;){
		Art* relevant=relevantPart[k];
		if(source==relevant)continue;
		if(relevant->polyBelongs(sourceTri)){
			relevant->triangle+=sourceTri;
			return true;
		}
	}
	return false;
}
*/
/*
inline void transferTriToOwner(S_Triangle& tri){// must not be transfered to the poly it came from
	Art* part=tri.pa;
	if(tri.pb->plidx>part->plidx){
		part=tri.pb;
	}
	if(tri.pc->plidx>part->plidx){
		part=tri.pc;
	}
	part->triangle+=tri;

}
*/
/*
inline bool verifyPartRelivanceToTriangles2(Art* source, M_Triangle& triangle){
	for(int i=triangle.length; --i>=0; ){
		S_Triangle tri=triangle[i];
		if(	tri.pa==source ){
			dLog(tri.pa->name);
			return true;
		}
		if(	tri.pb==source ){
			dLog(tri.pb->name);
			return true;
		}
		if(	tri.pc==source ){
			dLog(tri.pc->name);
			return true;
		}
	}
	return false;
}
*/

void transferOneVertex(Art* source, Art* dest, int vertex){
	Vec3f& svec = source->coord.refValue()[vertex];
//	svec=source->matrix*svec-dest->matrix*Vec3f();
	svec*=(source->matrix*dest->imatrix);

	TransferVertex transferVertex;
	transferVertex.fromVertex=vertex;
	transferVertex.numOf=1;
	transferVertex.source=source;
	transferVertex.dest=dest;
	transferVertex.doBuild();

}
/*
inline void movePartToEnd(Art* target, M_Art& relevantPart){
	int len=relevantPart.length;
	for(int i=len; --i>=0;){
		if(relevantPart[i]==target){
			Art* store=relevantPart[i];
			relevantPart[i]=relevantPart[len-1];
			relevantPart[len-1]=store;
		}
	}
}*/

inline int getTriLen(int fromCoord, int i, M_Int& coordIndex){
	int coordAccumLen=0;
	while(i<coordIndex.length){
		if(coordIndex[i++]<fromCoord){
			break;
		}
		if(coordIndex[i++]<fromCoord){
			break;
		}
		if(coordIndex[i++]<fromCoord){
			break;
		}
		coordAccumLen++;
	}
	return coordAccumLen;
}
inline Art* findLaterPart(M_Art& partList, Art* part, int& i){

	while(i>=part->coord.length){
		i-=part->coord.length;
		part=partList[part->plidx+1];
	}
	return part;
}


inline bool findVertexToTransfer(M_Triangle& triangle, Art* thisPart, Art*& rel){

	for(int i=0; i<triangle.length; i++ ){
		S_Triangle& tri=triangle[i];
		if(tri.pa==thisPart){
			if(	(tri.pb->plidx>=rel->plidx || tri.pb==thisPart) &&
				(tri.pc->plidx>=rel->plidx || tri.pc==thisPart)){

				transferOneVertex(thisPart, rel, tri.ia);
				return true;
			}
		}
		if(tri.pb==thisPart){
			if(	(tri.pa->plidx>=rel->plidx || tri.pa==thisPart) &&
				(tri.pc->plidx>=rel->plidx || tri.pc==thisPart)){

				transferOneVertex(thisPart, rel, tri.ib);
				return true;
			}
		}
		if(tri.pc==thisPart){
			if(	(tri.pa->plidx>=rel->plidx || tri.pa==thisPart) &&
				(tri.pb->plidx>=rel->plidx || tri.pb==thisPart)){

				transferOneVertex(thisPart, rel, tri.ic);
				return true;
			}
		}
	}
	return false;
}
void Art::deleteGeometry(){
	static DelPoly delPoly;
	delPoly.part = this;
	M_Int& poly = delPoly.poly.refValue();
	poly.length = triangle.length;
	for(int i=0; i<triangle.length; i++){
		poly[i]=i;
	}
	delPoly.doBuild();

	M_Art relevantPartCopy=relevantPart;//make a copy to avoid problem of us getting deleted from list later

	for(int k=0; k<relevantPartCopy.length; k++){
//		verifyAllTrianglesDontRefUp(smls);
		Art* rel=relevantPartCopy[k];
		if(rel==this)continue;

		if(findVertexToTransfer(rel->triangle, this, rel)){
			k=-1;
		}
	}
	coord.refValue().length=0;
	normal.refValue().length=0;
	color.length=0;
	texCoord.length=0;

	smls->resetRelevantParts();
	smls->process(); //can crash without this
//	smls->verifyIntegrity();
}

void Art::deleteGeometryInBranch(){
	M_SArt& children=part.refValue();
	for(int k=0;k<children.length; k++){
		children[k]->deleteGeometryInBranch();
	}
	deleteGeometry();
}
int getIndexRelativeToPart(Art* part, Art* relitivePart, int index){
	M_Art& partList=part->smls->partList;

	if(relitivePart->plidx<part->plidx){
		while(part!=relitivePart){
			part=partList[part->plidx-1];
			index-=part->triangle.length*3;
		}
	}
	else{
		while(part!=relitivePart){
			index+=part->triangle.length*3;
			part=partList[part->plidx+1];
		}
	}
	return index;
}
void convertToCoordIndexTriIndex(Art* part, M_Int& vertex){
	M_Int& lookUp=part->partTree;
	{
		lookUp.length=part->coord.length;
		for(int i=lookUp.length; --i>=0; ){
			lookUp[i]=-1;
		}
	}
	{
		for(int i=vertex.length; --i>=0; ){
			lookUp[vertex[i]]=0;
		}
	}
	{
		M_Art& relevantPart=part->relevantPart;
		for(int k=relevantPart.length; --k>=0;){
			Art* pt=relevantPart[k];
			M_Triangle& triangle=pt->triangle;
			for(int i=triangle.length*3; --i>=0;){
				//if(i<737){
				//	_asm{int 3}
				//}
				if(triangle.refPart(i)==part&&lookUp[triangle.refIndex(i)]==0)
				//lookUp[triangle.refIndex(i)]=i;
				lookUp[triangle.refIndex(i)]=getIndexRelativeToPart(part, pt, i);
			}
		}
	}
	{
		for(int i=vertex.length;--i>=0; ){
			vertex[i]=lookUp[vertex[i]];
		}
	}
}
void Art::joinCloseVerticesFrom(Art* source, float joinRange, bool srcColor, bool dstColor){//this=source
	JoinVertex jv;
	jv.srcColor = srcColor;
	jv.dstColor = dstColor;

	jv.source = this;
	jv.part = source;

	M_Int& fromVertex = jv.srcCoord.refValue();
	M_Int& toVertex = jv.dstCoord.refValue();

	const M_Vec3f& srcCoord = source->coord;

	int j,i;
	for(i=0;i<coord.length;i++){
		for(j=0;j<srcCoord.length;j++){
			Vec3f dstVec=matrix*coord.refValue()[i];
			Vec3f srcVec=source->matrix*srcCoord[j];
			if( i != j &&
				fabs(dstVec.x-srcVec.x)<joinRange&&
				fabs(dstVec.y-srcVec.y)<joinRange&&
				fabs(dstVec.z-srcVec.z)<joinRange
				){
				fromVertex+=i;
				toVertex+=j;
			}
		}
	}
	for(i=fromVertex.length; --i>=0;){
		for(j=toVertex.length; --j>=0; ){
			if(fromVertex[i]==toVertex[j]){
				for(int d=j; d<toVertex.length-1; d++){
					fromVertex[d]=fromVertex[d+1];
					toVertex[d]=toVertex[d+1];

				}
				fromVertex.length--;
				toVertex.length--;
				if(i>=fromVertex.length){
					i--;
				}
				if(j>=toVertex.length){
					j--;
				}
			}
		}
	}
	/*
	String str="fromVertex";
	str+=fromVertex.toString();
	str+="\ntoVertex";
	str+=toVertex.toString();
*/

	convertToCoordIndexTriIndex(jv.source.refValue(), fromVertex);
	convertToCoordIndexTriIndex(jv.part.refValue(), toVertex);
/*
	str+="\n\nfromVertex";
	str+=fromVertex.toString();
	str+="\ntoVertex";
	str+=toVertex.toString();
	str.write("show.txt");
*/
	jv.doBuild();
}
void Art::transferCloseVertices(Art* source, float joinRange){//this=source

	M_Vec3f& srcCoord=source->coord.refValue();

	int i,j;
	for(i=0;i<coord.length;i++){
		for(j=0;j<srcCoord.length;j++){
			Vec3f dstVec=matrix*coord.refValue()[i];
			Vec3f srcVec=source->matrix*srcCoord[j];
			if( i != j &&
				fabs(dstVec.x-srcVec.x)<joinRange&&
				fabs(dstVec.y-srcVec.y)<joinRange&&
				fabs(dstVec.z-srcVec.z)<joinRange
				){
				transferOneVertex(source,this, j);
				i=-1;
				break;

			}
		}
	}
}


/*
void Art::joinCloseVerticesFromAllPartsInSkeleton(float joinRange){//this=destination
	for(int k=0; k<smls->partList.length; k++){
		joinCloseVerticesFrom(smls->partList[k], joinRange);
	}
}
*/
void Art::colorize(Color3f newColor){
	if(workEffect==0||workEffect->getType()!=COLOREFFECT)return;

	for(int i=triangle.length*3; --i>=0; ){
		Art* part = triangle.refPart(i);
		if(ColorEffect* colorEffect=dynamic_cast<ColorEffect*>(part->workEffect)){
			M_Color3f& color = colorEffect->color.refValue();
			color[triangle.refTexIndex(i)]=newColor;
		}
	}



	//for(int k=relevantPart.length; --k>=0; ){
	//	M_Triangle& triangle=relevantPart[k]->triangle;
	//	for (int i=triangle.length*3; --i>=0; ){
	//		if(triangle.refPart(i)==this){
	//			color[triangle.refTexIndex(i)]=newColor;
	//		}
	//	}
	//}
}

bool Art::findNextPolyThatBelongs(int &poly){
	int polyLength=geometry->coordIndex.length/4;

	for(; poly<polyLength; poly++){
		if(polyBelongs(poly)){
			return true;
		}
	}
	return false;
}

int Art::swapVertex(int oldIndex, int newIndex){

	Vec3f store;

	if(coord.length==0)return 1;

	M_Vec3f& coord=this->coord.refValue();
	M_Vec3f& normal=this->normal.refValue();

	store=coord[oldIndex];
	coord[oldIndex]=coord[newIndex];
	coord[newIndex]=store;

	if(normal.length!=0){
		store=normal[oldIndex];
		normal[oldIndex]=normal[newIndex];
		normal[newIndex]=store;
	}
	for(int k=relevantPart.length; --k>=0;){
		Art* part=relevantPart[k];
		M_Triangle& triangle=part->triangle;

		for(int i=triangle.length*3;--i>=0;){
			if(triangle.refPart(i)==this){
				if(triangle.refIndex(i)==oldIndex){
					triangle.refIndex(i)=newIndex;
				}
				else if(triangle.refIndex(i)==newIndex){
					triangle.refIndex(i)=oldIndex;
				}
			}
		}
	}
	return 0;
}

int Art::uvSwapVertex(int oldIndex, int newIndex){

	if(uvCoord.length==0)return 1;

	M_Vec2f& uvCoord = this->uvCoord.refValue();
	M_Vec3f& coord = this->coord.refValue();

	Vec2f store = uvCoord[oldIndex];
	uvCoord[oldIndex] = uvCoord[newIndex];
	uvCoord[newIndex] = store;

	Vec3f store3f = coord[oldIndex];
	coord[oldIndex] = coord[newIndex];
	coord[newIndex] = store3f;

	NurbsPatch* storePatchOwner = patchOwner[oldIndex];
	patchOwner[oldIndex] = patchOwner[newIndex];
	patchOwner[newIndex] = storePatchOwner;

	for(int k=relevantPart.length; --k>=0;){
		Art* part=relevantPart[k];
		M_Triangle& triangle=part->triangle;

		for(int i=triangle.length*3;--i>=0;){
			if(triangle.refPart(i)==this){
				if(triangle.refIndex(i)==oldIndex){
					triangle.refIndex(i)=newIndex;
				}
				else if(triangle.refIndex(i)==newIndex){
					triangle.refIndex(i)=oldIndex;
				}
			}
		}
	}
	return 0;
}


inline void transferTrianglesToLeft(Art*& source, Art*&){

	static M_Art ownerList;	// any vertex that gets transfered may have the source in it's part's relevantList so make a list of any thing transfered to be checked for removal later
	ownerList.length=0;

//	source->relevantPart+=dest;
	for(int k=source->relevantPart.length;--k>=0;){
		Art* part=source->relevantPart[k];

		M_Triangle& triangle=part->triangle;
		int d=0;
		int numTransfered=0;
		for(int i=0; i<triangle.length; i++){
			S_Triangle& tri=triangle[i];

			Art* owner=tri.getOwner();

			if(part->polyBelongs(tri)){
				triangle[d++]=tri;
			}
			else {
				owner->triangle+=tri;
				//transferTriToOwner(tri);
				numTransfered++;
			}
			addRelevantPart(tri.pa->relevantPart, owner);
			addRelevantPart(tri.pb->relevantPart, owner);
			addRelevantPart(tri.pc->relevantPart, owner);

			addRelevantPart(ownerList, tri.pa);
			addRelevantPart(ownerList, tri.pb);
			addRelevantPart(ownerList, tri.pc);
		}
		triangle.length-=numTransfered;
	}
//	removeNonRelevantParts(dest);
//	removeNonRelevantParts(source);


/*
	removeNonRelevantParts(source);

	M_Art& relevantPart = source->relevantPart;
	for(int k=relevantPart.length; --k>=0; ){
		Art* part=relevantPart[k];
		if(verifyPartRelivanceToTriangles(dest, part->triangle)){
			addRelevantPart(part->relevantPart, dest);
		}
	}

	if(verifyPartRelivanceToTriangles(dest, source->triangle)){
		addRelevantPart(source->relevantPart, dest);
	}
*/
}
inline int getIndexFromRelIndex(Art* part, int index, M_Art& partList){
	if(index>=part->triangle.length*3){
		do{
			index-=part->triangle.length*3;
			part=partList[part->plidx+1];
		}while(index>=part->triangle.length*3);
	}
	else if(index<0){
		do{
			part=partList[part->plidx-1];
			index+=part->triangle.length*3;
		}while(index<0);
	}
	return part->triangle.refIndex(index);
}
inline int getColorIndexFromRelIndex(Art* part, int index, M_Art& partList){
	if(index>=part->triangle.length*3){
		do{
			index-=part->triangle.length*3;
			part=partList[part->plidx+1];
		}while(index>=part->triangle.length*3);
	}
	else if(index<0){
		do{
			part=partList[part->plidx-1];
			index+=part->triangle.length*3;
		}while(index<0);
	}
	return part->triangle.refTexIndex(index);
}

inline void changeVerticesFromToDstColor(Art*& source, Art*& dest, M_Int& fromVertexC, M_Int& toVertex, M_Int& colorIndex, M_Art& partList){

	const M_Vec3f& coord = source->coord;
	M_Int& mask = source->partTree;
	mask.length = 0;
	mask.length = coord.length;
	colorIndex.length = coord.length;

	{
		for(int i = coord.length; --i>=0; ){
			mask[i] = -1;
		}
	}
	{
		for(int i = fromVertexC.length; --i>=0; ){
			mask[fromVertexC[i]] = getIndexFromRelIndex( dest, toVertex[i], partList);
			colorIndex[fromVertexC[i]] = getColorIndexFromRelIndex( dest, toVertex[i], partList);
		}
	}
	M_Art& relevantPart=source->relevantPart;
	for(int k=relevantPart.length; --k>=0; ){
		M_Triangle& triangle=relevantPart[k]->triangle;
		for(int i=triangle.length*3; --i>=0; ){
			int& trrIx = triangle.refIndex(i);
			if(triangle.refPart(i) == source && mask[trrIx]!= -1){
				triangle.refPart(i) = dest;
				triangle.refTexIndex(i) = colorIndex[trrIx];
				trrIx = mask[trrIx];
			}
		}
	}
}

inline void changeVerticesFromToSrcColor(Art*& source, Art*& dest, M_Int& fromVertex, M_Int& fromVertexC, M_Int& toVertex, M_Art& partList){
	{
		const M_Vec3f& coord = dest->coord;

		M_Int& mask = dest->partTree;
		mask.length = 0;
		mask.length = coord.length;
		{
			for(int i = coord.length; --i>=0; ){
				mask[i]=-1;
			}
		}
		{
			for(int i = toVertex.length; --i>=0; ){
				int h=getIndexFromRelIndex( dest, toVertex[i], partList);
				mask[h] = getColorIndexFromRelIndex(source, fromVertex[i], partList);
			}
		}
		M_Art& relevantPart=dest->relevantPart;
		for(int k=relevantPart.length; --k>=0; ){
			M_Triangle& triangle=relevantPart[k]->triangle;
			for(int i = triangle.length*3; --i>=0; ){
				int& trrIx = triangle.refIndex(i);
				if(triangle.refPart(i) == dest && mask[trrIx]!=-1){
					triangle.refTexIndex(i) = mask[trrIx];
				}
			}
		}
	}

	M_Int& mask=source->partTree;
	mask.length=0;
	mask.length=source->coord.length;
	const M_Vec3f& coord=source->coord;

	{
		for(int i=coord.length; --i>=0; ){
			mask[i]=-1;
		}
	}
	{
		for(int i=fromVertexC.length; --i>=0; ){
			mask[fromVertexC[i]]=getIndexFromRelIndex( dest, toVertex[i], partList);
		}
	}
	M_Art& relevantPart=source->relevantPart;
	for(int k=relevantPart.length; --k>=0; ){
		M_Triangle& triangle=relevantPart[k]->triangle;
		for(int i=triangle.length*3; --i>=0; ){
			int& trrIx=triangle.refIndex(i);
			if(triangle.refPart(i)==source && mask[trrIx]!=-1){
				triangle.refPart(i)=dest;
				trrIx=mask[trrIx];
			}
		}
	}
}
inline void changeVerticesFromTo(Art*& source, Art*& dest, M_Int& fromVertexC, M_Int& toVertex, M_Art& partList){
	M_Int& mask=source->partTree;
	mask.length=0;
	mask.length=source->coord.length;
	const M_Vec3f& coord=source->coord;

	{
		for(int i=coord.length; --i>=0; ){
			mask[i]=-1;
		}
	}
	{
		for(int i=fromVertexC.length; --i>=0; ){
			mask[fromVertexC[i]]=getIndexFromRelIndex( dest, toVertex[i], partList);
		}
	}
	M_Art& relevantPart=source->relevantPart;
	for(int k=relevantPart.length; --k>=0; ){
		M_Triangle& triangle=relevantPart[k]->triangle;
		for(int i=triangle.length*3; --i>=0; ){
			int& trrIx=triangle.refIndex(i);
			if(triangle.refPart(i)==source && mask[trrIx]!=-1){
				triangle.refPart(i)=dest;
				trrIx=mask[trrIx];
			}
		}
	}
}

/*
inline void changeVerticesFromTo(Art*& source, Art*& dest, M_Int& fromVertex, M_Int& toVertex, bool srcColor, bool dstColor){
	M_Int& mask=source->partTree;
	mask.length=0;
	mask.resize(source->coord.length);
	M_Vec3f& coord=source->coord;
	static M_Int colorIndex;
	{
		for(int i=coord.length; --i>=0; ){
			mask[i]=-1;
		}
	}
	{
		for(int i=fromVertex.length; --i>=0; ){
			mask[fromVertex[i]]=toVertex[i];
		}
	}
	if(srcColor==false){
		colorIndex.resize(dest->coord.length);
		M_Art& relevantPart=dest->relevantPart;
		for(int k=relevantPart.length; --k>=0; ){
			M_Triangle& triangle=relevantPart[k]->triangle;
			for(int i=triangle.length*3; --i>=0; ){
				if(triangle.refPart(i)==dest){
					colorIndex[triangle.refIndex(i)]=triangle.refTexIndex(i);
				}
			}
		}
		{
			M_Art& relevantPart=source->relevantPart;
			for(int k=relevantPart.length; --k>=0; ){
				M_Triangle& triangle=relevantPart[k]->triangle;
				for(int i=triangle.length*3; --i>=0; ){
					int& trrIx=triangle.refIndex(i);
					if(triangle.refPart(i)==source && mask[trrIx]!=-1){
						triangle.refPart(i)=dest;
						trrIx=mask[trrIx];
						triangle.refTexIndex(i)=colorIndex[trrIx];
					}
				}
			}
		}
	}
	else if(dstColor==false){
		colorIndex.resize(source->coord.length);
		M_Art& relevantPart=source->relevantPart;
		for(int k=relevantPart.length; --k>=0; ){
			M_Triangle& triangle=relevantPart[k]->triangle;
			for(int i=triangle.length*3; --i>=0; ){
				if(triangle.refPart(i)==source){
					colorIndex[triangle.refIndex(i)]=triangle.refTexIndex(i);
				}
			}
		}
		{
			M_Art& relevantPart=source->relevantPart;
			for(int k=relevantPart.length; --k>=0; ){
				M_Triangle& triangle=relevantPart[k]->triangle;
				for(int i=triangle.length*3; --i>=0; ){
					int& trrIx=triangle.refIndex(i);
					if(triangle.refPart(i)==source && mask[trrIx]!=-1){
						triangle.refPart(i)=dest;
						trrIx=mask[trrIx];
						dest->triangle.refTexIndex(i)=colorIndex[trrIx];
					}
				}
			}
		}
	}
	else{//dstColor and srcColor=true
		M_Art& relevantPart=source->relevantPart;
		for(int k=relevantPart.length; --k>=0; ){
			M_Triangle& triangle=relevantPart[k]->triangle;
			for(int i=triangle.length*3; --i>=0; ){
				int& trrIx=triangle.refIndex(i);
				if(triangle.refPart(i)==source && mask[trrIx]!=-1){
					triangle.refPart(i)=dest;
					trrIx=mask[trrIx];
				}
			}
		}
	}
}
*/

inline void createMaskAndDeleteDeadVertertices(M_Int& fromVertex, M_Int& mask, Art*& part){
	M_Vec3f& coord = part->coord.refValue();
	M_Vec3f& normal = part->normal.refValue();

	mask.length = 0;
	mask.length = coord.length;

	{
		for(int i = fromVertex.length; --i>=0; ){
			mask[fromVertex[i]]=1;
		}
	}
	{
		if(part->smls->bnspe){
			M_Vec2f& uvCoord = part->uvCoord.refValue();
			M_NurbsPatch& patchOwner = part->patchOwner;
			for(int accum = 0, i = 0; i < mask.length; i++){
				int& mi = mask[i];
				int h = mi;
				accum += h;
				mi = i - accum;
				if(h == 0){
					coord[mi] = coord[i];
					uvCoord[mi] = uvCoord[i];
					patchOwner[mi] = patchOwner[i];
				}
			}
			uvCoord.length -= fromVertex.length;
			patchOwner.length -= fromVertex.length;
		}
		else{
			for(int accum = 0, i = 0; i < mask.length; i++){
				int& mi = mask[i];
				int h = mi;
				accum += h;
				mi = i - accum;
				if(h == 0){
					coord[mi] = coord[i];
				}
			}
		}
		coord.length -= fromVertex.length;

		if(normal.length){
			normal.length -= fromVertex.length;
		}
	}
}
inline void referenceNoDeadVertices(Art*& source, M_Int& mask){
	M_Art& relevantPart=source->relevantPart;
	for(int k = relevantPart.length; --k>=0; ){
		Art* part = relevantPart[k];
		M_Triangle& triangle=part->triangle;
		for(int i = triangle.length * 3; --i >= 0; ){
			if(triangle.refPart(i) == source){
				triangle.refIndex(i) = mask[triangle.refIndex(i)];
			}
		}
	}
}
/*
void joinVertex(Art* source, Art* dest, int fromCoord, int toCoord){

	source->fastDeleteVertex(fromCoord);

	M_Art& relevantPart=source->relevantPart;

	for(int k=relevantPart.length; --k>=0; ){
		Art* part=relevantPart[k];
		M_Triangle& triangle=part->triangle;
		for(int i=triangle.length;--i>=0;){
			S_Triangle& tri=triangle[i];
			if(tri.pa==source){
				if(tri.ia == fromCoord){
					tri.ia=toCoord;
					tri.pa=dest;
				}
				else if(tri.ia > fromCoord){
					tri.ia--;
				}
			}
			if(tri.pb==source){
				if(tri.ib == fromCoord){
					tri.ib=toCoord;
					tri.pb=dest;
				}
				else if(tri.ib > fromCoord){
					tri.ib--;
				}
			}
			if(tri.pc==source){
				if(tri.ic == fromCoord){
					tri.ic=toCoord;
					tri.pc=dest;
				}
				else if(tri.ic > fromCoord){
					tri.ic--;
				}
			}
		}
	}
	if(source!=dest){
		transferTrianglesToLeft(source, fromCoord, int(fromCoord+1), dest);
	}
}
*/
void printOutRelevants(Art* part){
	dLog("---------");
	dLog(part->name);
	dLog(":");

	M_Art& relevantPart=part->relevantPart;

	for(int r=0; r<relevantPart.length; r++){
		dLog(relevantPart[r]->name);
	}
}

void printOutRelevants(Seamless& smls){
	M_Art& partList=smls.partList;
	for(int k=0; k<partList.length; k++){
		Art& part=*partList[k];
		if(part.relevantPart.length){

			dLog(":");
			dLog(part.name);
			dLog(":");

			M_Art& relevantPart=partList[k]->relevantPart;

			for(int r=0; r<relevantPart.length; r++){
				dLog(relevantPart[r]->name);
			}
		}
	}
}

extern bool g_dragActuallyHappened;

void JoinVertex::localIn(Art* part, int h, M_Int& fromVertex, M_Int& toVertex){
	const M_Vec3f& coord=part->coord;
	//Matrix& matrix=part->matrix;

	for(int i = h; i < coord.length; i++){
		for(int j = h; j < coord.length;j++){
			Vec3f dstVec=coord[i];
			Vec3f srcVec=coord[j];
			if( i != j &&
				fabs(dstVec.x-srcVec.x)<range&&
				fabs(dstVec.y-srcVec.y)<range&&
				fabs(dstVec.z-srcVec.z)<range
				){
				fromVertex+=i;
				toVertex+=j;
			}
		}
	}
	for(int i = fromVertex.length; --i>=0;){
		for(int j = toVertex.length; --j>=0; ){
			if(fromVertex[i]==toVertex[j]){
				for(int d=j; d<toVertex.length-1; d++){
					fromVertex[d]=fromVertex[d+1];
					toVertex[d]=toVertex[d+1];

				}
				fromVertex.length--;
				toVertex.length--;
				if(i>=fromVertex.length){
					i--;
				}
				if(j>=toVertex.length){
					j--;
				}
			}
		}
	}

	convertToCoordIndexTriIndex(part, fromVertex);
	convertToCoordIndexTriIndex(part, toVertex);
}

void JoinVertex::doBuild(){

	if(part == 0 || (g_joinAndTransferActive == false && part->smls->bnspe == false)){
		return;
	}

	Art* part = this->part.refValue();
	Art* source = 0;

	if(this->source.refValue()==0){
		source = part;
	}
	else {
		source = this->source.refValue();
	}
	if(in){
		M_Int& fromVertex = srcCoord.refValue();
		M_Int& toVertex = dstCoord.refValue();

		static M_Int redirectMask;

		fromVertex.length = 0;
		toVertex.length = 0;

		if(source == part){
			localIn(part, 0, fromVertex, toVertex);
		}
		else{
			M_Vec3f& srcCoord = source->coord.refValue();
			M_Vec3f& dstCoord = part->coord.refValue();

			//src->partTree.resize(src->coord.length);
			M_Int& srcMask=source->partTree;
			srcMask.length=source->coord.length;
			redirectMask.length=srcMask.length;
			M_Int& dstMask=part->partTree;
			dstMask.length=part->coord.length;
			{
				for(int i=srcMask.length; --i>=0;){
					srcMask[i]=-1;
				}
			}
			{
				for(int i=dstMask.length; --i>=0;){
					dstMask[i]=-1;
				}
			}
			for(int d=dstCoord.length;--d>=0;){
				for(int s=srcCoord.length;--s>=0;){
					Vec3f dstVec=dstCoord[d];
					Vec3f srcVec=srcCoord[s];
					if( fabs(dstVec.x-srcVec.x)<range&&
						fabs(dstVec.y-srcVec.y)<range&&
						fabs(dstVec.z-srcVec.z)<range
						){
						srcMask[s]=1;
						redirectMask[s]=d;
						dstMask[d]=1;
					}
				}
			}
			{
				M_Art& srcRelevantPart=source->relevantPart;	//convert srcMask to crdIdxIdx
				for( int k=srcRelevantPart.length; --k>=0; ){
					Art* p=srcRelevantPart[k];
					M_Triangle& triangle=p->triangle;
					for(int i=triangle.length*3; --i>=0;){
						if(triangle.refPart(i)==source&&srcMask[triangle.refIndex(i)]!=-1){
							//srcMask[triangle.refIndex(i)]=i;
							srcMask[triangle.refIndex(i)]=getIndexRelativeToPart(source, p, i);
						}
					}
				}
			}
			{
				M_Art& dstRelevantPart=part->relevantPart;	//convert dstMask to crdIdxIdx
				for( int k=dstRelevantPart.length; --k>=0; ){
					Art* p=dstRelevantPart[k];
					M_Triangle& triangle=p->triangle;


					for(int i=triangle.length*3; --i>=0;){
						if(triangle.refPart(i)==part&&dstMask[triangle.refIndex(i)]!=-1){
							dstMask[triangle.refIndex(i)]=getIndexRelativeToPart(part, p, i);
						}
					}
				}
			}
			{
				for(int i=0; i<srcCoord.length; i++){	//order dst to  src
					if(srcMask[i]!=-1){
						fromVertex+=srcMask[i];
						toVertex+=dstMask[redirectMask[i]];
					}
				}
			}
		}
	}
	if( !srcCoord.length || !dstCoord.length){
		return;
	}

	M_Int& fromVertex = this->srcCoord.refValue();

	M_Art& partList=source->smls->partList;
	static M_Int colorIndex;
	static M_Int fromVertexC;

	fromVertexC.length=fromVertex.length;
	{
		for(int i=fromVertex.length; --i>=0;){
			fromVertexC[i]=getIndexFromRelIndex(source, fromVertex[i], partList);
		}
	}
	Seamless* smls = part->smls;
	if(move){
		M_Int& toVertex = this->dstCoord.refValue();

		if(smls->bnspe){
			M_Vec2f& srcUvCoord = source->uvCoord.refValue();
			M_Vec2f& dstUvCoord = part->uvCoord.refValue();
			for(int i = fromVertex.length; --i >= 0;){
				srcUvCoord[fromVertexC[i]] = dstUvCoord[getIndexFromRelIndex(part,toVertex[i],partList)];
			}
		}
		else{
			M_Vec3f& srcCoord = source->coord.refValue();
			M_Vec3f& dstCoord = part->coord.refValue();
			for(int i = fromVertex.length; --i >= 0;){
				srcCoord[fromVertexC[i]] = dstCoord[getIndexFromRelIndex(part,toVertex[i],partList)];
			}
		}
	}
	else if(smls->buildNormals || smls->surfaceNormals){
		M_Int& toVertex = this->dstCoord.refValue();
		M_Vec3f& srcCoord = source->coord.refValue();
		M_Vec3f& dstCoord = part->coord.refValue();

		M_Vec3f& srcNormal = source->normal.refValue();
		M_Vec3f& dstNormal = part->normal.refValue();
		for(int i = fromVertex.length; --i >= 0;){
			int si = fromVertexC[i];
			int di = getIndexFromRelIndex(part,toVertex[i],partList);
			srcCoord[si] = dstCoord[di];
			srcNormal[si] = dstNormal[di] = (dstNormal[di] + srcNormal[si]).normalize();
		}
	}
	else{
		if(srcColor == false){
			changeVerticesFromToDstColor(source, part, fromVertexC, dstCoord.refValue(), colorIndex, partList);
		}
		else if(dstColor == false){
			changeVerticesFromToSrcColor(source, part, fromVertex, fromVertexC, dstCoord.refValue(), partList);
		}
		else {
			changeVerticesFromTo(source, part, fromVertexC, dstCoord.refValue(), partList);
		}

		M_Int& mask = source->partTree;

		createMaskAndDeleteDeadVertertices(fromVertexC, mask, source);
		referenceNoDeadVertices(source, mask);

		transferTrianglesToLeft(source, part);
	}
}

void joinVertex(Art* source, Art* dest, int fromVertex, int toVertex){
	JoinVertex jv;
	jv.source=source;
	jv.part=dest;
	jv.srcCoord=fromVertex;
	jv.dstCoord=toVertex;
	jv.doBuild();
}


void JoinCloseVertex::doBuild(){
	//both parts must not be from same part;
	if(!part.length)return;

	static JoinVertex jv;
	jv.srcColor=srcColor;
	jv.dstColor=dstColor;

	M_Int& fromVertex=jv.srcCoord.refValue();
	M_Int& toVertex=jv.dstCoord.refValue();
	const M_Node& part=this->part;
	const M_Node& source=this->source;
	dynamic_cast<Art*>(part[0].getRawPtr())->smls->catIMatrix();
	Art* prt;// checked these are not needed to be set to 0
	Art* src;
	static M_Int redirectMask;

	float joinRange=this->joinRange;

	for(int k=0; k < part.length; k++){
		fromVertex.length=0;
		toVertex.length=0;
		prt=dynamic_cast<Art*>(part[k].getRawPtr());
		if(k<source.length){
			src=dynamic_cast<Art*>(source[k].getRawPtr());
		}
		else {
			src=prt;
		}
		if(src==prt){
			prt->joinCloseVerticesFrom(prt, joinRange, srcColor, dstColor);
		}
		else{

			jv.part=prt;
			jv.source=src;
			M_Vec3f& srcCoord=src->coord.refValue();
			M_Vec3f& dstCoord=prt->coord.refValue();

			//src->partTree.resize(src->coord.length);
			M_Int& srcMask=src->partTree;
			srcMask.length=src->coord.length;
			redirectMask.length=srcMask.length;
			M_Int& dstMask=prt->partTree;
			dstMask.length=prt->coord.length;
			{
				for(int i=srcMask.length; --i>=0;){
					srcMask[i]=-1;
				}
			}
			{
				for(int i=dstMask.length; --i>=0;){
					dstMask[i]=-1;
				}
			}
			for(int d=dstCoord.length;--d>=0;){
				for(int s=srcCoord.length;--s>=0;){
					Vec3f dstVec=prt->matrix*dstCoord[d];
					Vec3f srcVec=src->matrix*srcCoord[s];
					if( fabs(dstVec.x-srcVec.x)<joinRange&&
						fabs(dstVec.y-srcVec.y)<joinRange&&
						fabs(dstVec.z-srcVec.z)<joinRange
						){
						srcMask[s]=1;
						redirectMask[s]=d;
						dstMask[d]=1;
					}
				}
			}
			{
				M_Art& srcRelevantPart=src->relevantPart;	//convert srcMask to crdIdxIdx
				for( int k=srcRelevantPart.length; --k>=0; ){
					Art* p=srcRelevantPart[k];
					M_Triangle& triangle=p->triangle;
					for(int i=triangle.length*3; --i>=0;){
						if(triangle.refPart(i)==src&&srcMask[triangle.refIndex(i)]!=-1){
							//srcMask[triangle.refIndex(i)]=i;
							srcMask[triangle.refIndex(i)]=getIndexRelativeToPart(src, p, i);
						}
					}
				}
			}
			{
				M_Art& dstRelevantPart=prt->relevantPart;	//convert dstMask to crdIdxIdx
				for( int k=dstRelevantPart.length; --k>=0; ){
					Art* p=dstRelevantPart[k];
					M_Triangle& triangle=p->triangle;


					for(int i=triangle.length*3; --i>=0;){
						if(triangle.refPart(i)==prt&&dstMask[triangle.refIndex(i)]!=-1){
							dstMask[triangle.refIndex(i)]=getIndexRelativeToPart(prt, p, i);
							//dstMask[triangle.refIndex(i)]=getIndexRelativeToPart( prt,p, i);
//							if(p!=prt){
//								dLog("joinCloseVertex code not complete!");
//								throw this;
//							}
							//dstMask[triangle.refIndex(i)]=i;
						}
					}
				}
			}
			{
				for(int i=0; i<srcCoord.length; i++){	//order dst to  src
					if(srcMask[i]!=-1){
						fromVertex+=srcMask[i];
						toVertex+=dstMask[redirectMask[i]];
					}
				}
			}
			jv.doBuild();
		}
	}
}



inline void bend( float beginHeight,float endHeight,float angle,float axis, M_Vec3f& coord, M_Float& yLookUp){
	rotY(axis,coord);
	int i;
	float diff=endHeight-beginHeight;
	if(diff==0){
		coord.translate(0,-beginHeight,0);
		rotZ(angle,coord);
		coord.translate(0,beginHeight,0);
		rotY(-axis,coord);
		return;
	}
	Vec3f reletiveDis;
	Vec3f v;
	Vec3f n;
	Vec3f c;
	float rad=diff/angle;
	for(i=0;i<coord.length;i++){
		float sy=yLookUp[i];
		if(sy>=beginHeight&&sy<=endHeight){
			v=coord[i];
			reletiveDis.y=v.y-beginHeight;
			float stepAngle=reletiveDis.y/diff*angle;

			v.y=0;
			v.x+=rad;
			rotateVectorZ(v,stepAngle);
			v.y+=beginHeight;
			v.x-=rad;
			coord[i]=v;
		}
		else if(sy>endHeight){
			v=coord[i];
			reletiveDis.y=v.y-beginHeight;
			v.y=(rad+v.x)*sin(angle)+beginHeight+	(coord[i].y-endHeight)*(cos(angle));
			v.x=(rad+v.x)*cos(angle)-rad	+	(coord[i].y-endHeight)*-(sin(angle));

			coord[i]=v;
		}
	}

	rotY(-axis,coord);
}
void Art::insideOut(){
	for(int i=triangle.length; --i>=0;){
		int is=triangle[i].ib;
		int ts=triangle[i].tb;
		Art* ps=triangle[i].pb;
		triangle[i].ib=triangle[i].ic;
		triangle[i].tb=triangle[i].tc;
		triangle[i].pb=triangle[i].pc;
		triangle[i].ic=is;
		triangle[i].tc=ts;
		triangle[i].pc=ps;
	}
}
void Art::insideOut(int offset){
	for(int i=triangle.length; --i>=offset;){
		int is=triangle[i].ib;
		int ts=triangle[i].tb;
		Art* ps=triangle[i].pb;
		triangle[i].ib=triangle[i].ic;
		triangle[i].tb=triangle[i].tc;
		triangle[i].pb=triangle[i].pc;
		triangle[i].ic=is;
		triangle[i].tc=ts;
		triangle[i].pc=ps;
	}
}

////////////////////////////////////////////////////////////
int getFileSize(FILE* Stream){
	fseek(Stream, 0, SEEK_END);
	int Size=ftell(Stream);
	fseek(Stream, 0, SEEK_SET);
	return Size;
}


int countUpIntegers(char* p){
	searchTillText(p);
	if('['!=*p)return -1;
	p++;

	int i=0,idx2;
	searchTillText(p);
	while(*p!=']'){
		searchTillNextWord(p);
		searchTillNextWord(p);
		do{
			i+=3;
			searchTillNextWord(p);
			if(*p==']')return i;// in case last poly does not end with -1
			idx2=atoi(p);
		}while(idx2!=-1);
		searchTillNextWord(p);
	}
	return i;
}
int getCoordIndex(M_Int& coordIndex, char*& p){
	searchTillText(p);

	if('['!=*p)return -1;
	p++;

	int i=0,idx0,idx1,idx2;
	searchTillText(p);
	while(*p!=']'){
		idx0=atoi(p);
		searchTillNextWord(p);
		idx1=atoi(p);
		searchTillNextWord(p);
		idx2=atoi(p);
		do{
			coordIndex[i++]=idx0;
			coordIndex[i++]=idx1;
			coordIndex[i++]=idx2;

			idx1=idx2;
			searchTillNextWord(p);
			if(*p==']')return i;// in case last poly does not end with -1
			idx2=atoi(p);
		}while(idx2!=-1);
		searchTillNextWord(p);
	}
	p++;
	return i;
}
int getNonJoiningVertices( Art* partA, int indexA, Art* partB, int indexB, int& notConectedA, int& notConectedB){
	if(indexA>=partA->triangle.length||indexB>=partB->triangle.length)return 1; //triangle does not exist
	if(partA==partB && indexA==indexB)return 2; //both triangles are the same triangle
	int a,b,i,k;
	int triangleACorners[2];
	int triangleBCorners[2];
	S_Triangle& triA=partA->triangle[indexA];
	S_Triangle& triB=partB->triangle[indexB];
	for(a=0,k=0;a<3;a++){
		for(b=0;b<3;b++){
			if(	triA.refPart(a)==triB.refPart(b) && triA.refIndex(a)==triB.refIndex(b)){
				triangleACorners[k]=a;
				triangleBCorners[k]=b;
				if(++k==2)break;
			}
		}
	}
	if(k==3)return 3; //all 3 points are joined
	if(k!=2)return 4; //2 points are not shared
	for(i=0;i<3;i++){
		if(i!=triangleACorners[0]&&i!=triangleACorners[1]){
			notConectedA=i;
			break;
		}
	}
	for(i=0;i<3;i++){
		if(i!=triangleBCorners[0]&&i!=triangleBCorners[1]){
			notConectedB=i;
			break;
		}
	}
	return 0;
}
int swapJoiningVertices( Art* partA, int indexA, Art* partB, int indexB){

	int notConectedA;
	int notConectedB;
	int k=getNonJoiningVertices( partA, indexA, partB, indexB, notConectedA, notConectedB);
	if(!k)return k;
/*
	S_Triangle& triA=partA->triangle[indexA];
	S_Triangle& triB=partB->triangle[indexB];

	triA.refPart(triangleACorners[0])=triB.refPart(notConectedB);
	triA.refIndex(triangleACorners[0])=triB.refIndex(notConectedB);
	triA.refTexIndex(triangleACorners[0])=triB.refTexIndex(notConectedB);

	triB.refPart(triangleBCorners[1])=triA.refPart(notConectedA);
	triB.refIndex(triangleBCorners[1])=triA.refIndex(notConectedA);
	triB.refTexIndex(triangleBCorners[1])=triA.refTexIndex(notConectedA);
*/
	return 0;
}
void Mesh::genColorList(Color3f c){
	for(int i=color.length; --i>=0;){
		color.refValue()[i]=c;
	}
}

void VertexGeometry::genColorList(Color3f c){
	if(Coordinate* coord = dynamic_cast<Coordinate*>(this->coord.p)){
		M_Vec3f &point = coord->point.refValue();
		if(Color* color = dynamic_cast<Color*>(this->color.p)){
			M_Color3f &colorColor = color->color.refValue();
			colorColor.length=point.length;
			for(int i=0;i<colorColor.length;i++){
				colorColor[i]=c;
			}
		}
	}
}

void VertexGeometry::genColorList(float r,float g,float b){
	genColorList(Color3f(r,g,b));
}
/*
void IndexedFaceSet::sweepColorX(Color3f beginColor,Color3f endColor, float beginHeight, float endHeight){
	M_Vec3f& point=refPoint();
	M_Color3f& color=refColor();
	int i;
	float rColorDiff=endColor.r-beginColor.r;
	float gColorDiff=endColor.g-beginColor.g;
	float bColorDiff=endColor.b-beginColor.b;
	float height=endHeight-beginHeight;
	for(i=0;i<point.length;i++){
		if(point[i].x>=beginHeight&&point[i].x<=endHeight){
			color[i].r=beginColor.r+(((point[i].x-beginHeight)/height)*rColorDiff);
			color[i].g=beginColor.g+(((point[i].x-beginHeight)/height)*gColorDiff);
			color[i].b=beginColor.b+(((point[i].x-beginHeight)/height)*bColorDiff);
		}
	}
}

void IndexedFaceSet::sweepColorZ(Color3f beginColor,Color3f endColor, float beginHeight, float endHeight){
	M_Vec3f& point=refPoint();
	M_Color3f& color=refColor();
	int i;
	float rColorDiff=endColor.r-beginColor.r;
	float gColorDiff=endColor.g-beginColor.g;
	float bColorDiff=endColor.b-beginColor.b;
	float height=endHeight-beginHeight;
	for(i=0;i<point.length;i++){
		if(point[i].z>=beginHeight&&point[i].z<=endHeight){
			color[i].r=beginColor.r+(((point[i].z-beginHeight)/height)*rColorDiff);
			color[i].g=beginColor.g+(((point[i].z-beginHeight)/height)*gColorDiff);
			color[i].b=beginColor.b+(((point[i].z-beginHeight)/height)*bColorDiff);
		}
	}
}

void IndexedFaceSet::sweepColorY(Color3f beginColor,Color3f endColor, float beginHeight, float endHeight){
	M_Vec3f& point=refPoint();
	M_Color3f& color=refColor();
	if(0==color.length)
		genColorList();
	int i;
	float rColorDiff=endColor.r-beginColor.r;
	float gColorDiff=endColor.g-beginColor.g;
	float bColorDiff=endColor.b-beginColor.b;
	float height=endHeight-beginHeight;
	for(i=0;i<point.length;i++){
		if(point[i].y>=beginHeight&&point[i].y<=endHeight){
			color[i].r=beginColor.r+(((point[i].y-beginHeight)/height)*rColorDiff);
			color[i].g=beginColor.g+(((point[i].y-beginHeight)/height)*gColorDiff);
			color[i].b=beginColor.b+(((point[i].y-beginHeight)/height)*bColorDiff);
		}
	}
}
void IndexedFaceSet::sweepColorCircularZ(float beginHeight, float endHeight, Color3f beginColor,Color3f endColor){
	M_Vec3f& point=refPoint();
	M_Color3f& color=refColor();
	if(0==color.length)
		genColorList();
	int i;
	float rColorDiff=endColor.r-beginColor.r;
	float gColorDiff=endColor.g-beginColor.g;
	float bColorDiff=endColor.b-beginColor.b;
	float heightDiff=endHeight-beginHeight;
	for(i=0;i<point.length;i++){
		float length=point[i].length();
		if(length>=beginHeight&&length<=endHeight){
			float frac=(length-beginHeight)/heightDiff;
			color[i].r=beginColor.r+frac*rColorDiff;
			color[i].g=beginColor.g+frac*gColorDiff;
			color[i].b=beginColor.b+frac*bColorDiff;
		}
	}
}
void IndexedFaceSet::sweepColorCircularZ(float beginHeight, float endHeight, Color3f beginColor,Color3f endColor, float beginOpacity, float endOpacity){
	M_Vec3f& point=refPoint();
	M_Color3f& color=refColor();
	if(beginOpacity==1&&endOpacity==1){
		sweepColorCircularZ(beginHeight, endHeight, beginColor, endColor);
		return;
	}

	if(0==color.length)
		genColorList();
	int i;
	float opacityDiff=endOpacity-beginOpacity;
	float rColorDiff=endColor.r-beginColor.r;
	float gColorDiff=endColor.g-beginColor.g;
	float bColorDiff=endColor.b-beginColor.b;
	float heightDiff=endHeight-beginHeight;
	for(i=0;i<point.length;i++){
		float length=point[i].length();
		if(length>=beginHeight&&length<=endHeight){
			float frac=(length-beginHeight)/heightDiff;
			float opacity=(beginOpacity+frac*opacityDiff);
			float clarity=1.0-opacity;
			color[i].r=color[i].r*clarity+opacity*(beginColor.r+frac*rColorDiff);
			color[i].g=color[i].g*clarity+opacity*(beginColor.g+frac*gColorDiff);
			color[i].b=color[i].b*clarity+opacity*(beginColor.b+frac*bColorDiff);
		}
	}
}
void IndexedFaceSet::sweepColorY(float rb,float gb,float bb,float re,float ge,float be, float beginHeight, float endHeight){
	sweepColorY(Color3f(rb,gb,bb),Color3f(re,ge,be), beginHeight, endHeight);
}
void IndexedFaceSet::genEndRingNormalsY(){
	M_Vec3f& point=refPoint();
	M_Vec3f& vector=refVector();
	int i;
	for(i=0;i<numStaves;i++){
		float angle=((float)i)/((float)(numStaves-1))*PI*2+PI*.5;
		vector[point.length-numStaves+i].x=cos(angle);
		vector[point.length-numStaves+i].y=0;
		vector[point.length-numStaves+i].z=-sin(angle);
	}
}
void IndexedFaceSet::genRingNormals(int count, int offset, int endMeetStart, Rotation rot, float startAngle, float endAngle){
	M_Vec3f& vector=refVector();
	float angDiff=endAngle-startAngle;
	float step=angDiff/(count-endMeetStart);
	int i;
	for(i=0;i<count;i++){
		float angle=step*((float)i)+startAngle;
		vector[offset+i].x=cos(angle);
		vector[offset+i].y=0;
		vector[offset+i].z=-sin(angle);
		vector[offset+i]=rot.multVec(vector[offset+i]);
	}
}
*/


void joinCoordinate(SFNode& par1, Coordinate* par2){
	if(par2!=NULL){
		if(par1==NULL){
			Coordinate* coord = new Coordinate();
			par1 = coord;
			coord->point = par2->point;
		}
		else{
			Coordinate* coord = dynamic_cast<Coordinate*>(par1.p);
			coord->point.refValue() += par2->point;
		}
	}
}
void joinNormal(SFNode& par1, Normal* par2){
	if(par2!=NULL){
		if(par1==NULL){
			Normal* normal=new Normal();
			par1=normal;
			normal->vector = par2->vector;
		}
		else{
			Normal* normal=dynamic_cast<Normal*>(par1.p);
			normal->vector.refValue()+=par2->vector;
		}
	}
}
void joinColor(SFNode& par1, Color* par2){
	if(par2!=NULL){
		if(par1==NULL){
			Color* color = new Color();
			par1 = color;
			color->color = par2->color;
		}
		else{
			Color* color = dynamic_cast<Color*>(par1.p);
			color->color.refValue()+=par2->color;
		}
	}
}
void joinTextureCoordinate(SFNode& par1, TextureCoordinatePtr par2){
	if(par2!=NULL){
		if(par1==NULL){
			TextureCoordinate* texCoord=new TextureCoordinate();
			par1=texCoord;
			texCoord->point=par2->point;
		}
		else{
			TextureCoordinate* texCoord=dynamic_cast<TextureCoordinate*>(par1.p);
			texCoord->point.refValue()+=par2->point;
		}
	}
}
/*
void IndexedFaceSet::joinParts(IndexedFaceSetPtr sec){
	return;
	M_Vec3f& point=refPoint();
	M_Color3f& color=refColor();
	M_Int& coordIndex=refCoordIndex();

	M_Vec3f& secPoint=sec->refPoint();
	M_Color3f& secColor=sec->refColor();
	M_Int& secCoordIndex=sec->refCoordIndex();
	int i;
	if(&color!=NULL&&&secColor!=NULL){
		if(secPoint.length&&color.length&&secColor.length!=secPoint.length)
			secColor.length=secPoint.length;
		if(color.length!=point.length&&point.length&&secColor.length)
			color.length=point.length;
	}


	M_Int I; I.length=coordIndex.length+sec->coordIndex.length;

	for(i=0;i<coordIndex.length;i++){
		I[i]=coordIndex[i];
	}
	int coordPointLength=0;
	if(coord!=NULL)coordPointLength=point.length;
	for(i=0;i<sec->coordIndex.length;){
		I[coordIndex.length+i]=secCoordIndex[i]+coordPointLength;
		i++;
		I[coordIndex.length+i]=secCoordIndex[i]+coordPointLength;
		i++;
		I[coordIndex.length+i]=secCoordIndex[i]+coordPointLength;
		i++;
		I[coordIndex.length+i]=-1;
		i++;
	}
	coordIndex=I;

	joinCoordinate(coord, dynamic_cast<Coordinate*>(sec->coord.p));
	joinNormal(normal, dynamic_cast<Normal*>(sec->normal.p));
	joinColor(this->color, dynamic_cast<Color*>(sec->color.p));
	joinTextureCoordinate(texCoord, dynamic_cast<TextureCoordinate*>(sec->texCoord.p));



//	topIsPoint=sec->topIsPoint;///DONT REMOVE THIS(bottomIsPoint stays the same)
//	numRings+=sec->numRings;

}
*/
void copyTextFromFile(FILE* stream, char* fileName2){
	FILE* stream2=fopen(fileName2,"rb");
	int f2size=getFileSize(stream2);
	char* p=new char[f2size];
	fread(p,1,f2size,stream2);
	fwrite(p,1,f2size,stream);
	fclose(stream2);
}
void Art::cutSphere(const Vec3f location, const float radius){
	static DelPoly dp;
	M_Int& poly=dp.poly.refValue();
	poly.length=0;
	dp.part=this;

	for(int i=0; i<triangle.getTriangleLength(); i++ ){
		Vec3f vec=triangle.refPart(i)->coord[triangle.refIndex(i)];
		float vecX=vec.x-location.x;
		float vecY=vec.y-location.y;
		float vecZ=vec.z-location.z;
		float HYP=sqrt(vecX*vecX + vecY*vecY + vecZ*vecZ);

		if(radius>HYP){
			int q = i / 3;
			poly += q;
			q++;
			i = q * 3 - 1;

		}
	}

	dp.doBuild();
}
void Art::cutLeft(float pos){
	static DelPoly dp;
	M_Int& poly=dp.poly.refValue();
	poly.length=0;
	dp.part=this;

	for(int i=0; i<triangle.length; i++ ){
		Art* pa=triangle.refPart(i*3+0);
		Art* pb=triangle.refPart(i*3+1);
		Art* pc=triangle.refPart(i*3+2);
		if(	(pa->matrix*pa->coord.refValue()[triangle.refIndex(i*3+0)]).x<=pos&&
			(pb->matrix*pb->coord.refValue()[triangle.refIndex(i*3+1)]).x<=pos&&
			(pc->matrix*pc->coord.refValue()[triangle.refIndex(i*3+2)]).x<=pos){
			poly+=i;
		}
	}
	dp.doBuild();
}
void Art::cutRight(float pos){
	static DelPoly dp;
	M_Int& poly=dp.poly.refValue();
	poly.length=0;
	dp.part=this;

	for(int i=0; i<triangle.length; i++ ){
		Art* pa=triangle.refPart(i*3+0);
		Art* pb=triangle.refPart(i*3+1);
		Art* pc=triangle.refPart(i*3+2);
		if(	(pa->matrix*pa->coord.refValue()[triangle.refIndex(i*3+0)]).x>=pos&&
			(pb->matrix*pb->coord.refValue()[triangle.refIndex(i*3+1)]).x>=pos&&
			(pc->matrix*pc->coord.refValue()[triangle.refIndex(i*3+2)]).x>=pos){
			poly+=i;
		}
	}
	dp.doBuild();
}
void Art::cutDown(float pos){
	static DelPoly dp;
	M_Int& poly=dp.poly.refValue();
	poly.length=0;
	dp.part=this;

	for(int i=0; i<triangle.length; i++ ){
		Art* pa=triangle.refPart(i*3+0);
		Art* pb=triangle.refPart(i*3+1);
		Art* pc=triangle.refPart(i*3+2);
		if(	(pa->matrix*pa->coord.refValue()[triangle.refIndex(i*3+0)]).y<=pos&&
			(pb->matrix*pb->coord.refValue()[triangle.refIndex(i*3+1)]).y<=pos&&
			(pc->matrix*pc->coord.refValue()[triangle.refIndex(i*3+2)]).y<=pos){
			poly+=i;
		}
	}
	dp.doBuild();
}
void Art::cutUp(float pos){
	static DelPoly dp;
	M_Int& poly=dp.poly.refValue();
	poly.length=0;
	dp.part=this;

	for(int i=0; i<triangle.length; i++ ){
		Art* pa=triangle.refPart(i*3+0);
		Art* pb=triangle.refPart(i*3+1);
		Art* pc=triangle.refPart(i*3+2);
		if(	(pa->matrix*pa->coord.refValue()[triangle.refIndex(i*3+0)]).y>=pos&&
			(pb->matrix*pb->coord.refValue()[triangle.refIndex(i*3+1)]).y>=pos&&
			(pc->matrix*pc->coord.refValue()[triangle.refIndex(i*3+2)]).y>=pos){
			poly+=i;
		}
	}
	dp.doBuild();
}
void Art::cutForward(float pos){
	static DelPoly dp;
	M_Int& poly=dp.poly.refValue();
	poly.length=0;
	dp.part=this;

	for(int i=0; i<triangle.length; i++ ){
		Art* pa=triangle.refPart(i*3+0);
		Art* pb=triangle.refPart(i*3+1);
		Art* pc=triangle.refPart(i*3+2);
		if(	(pa->imatrix*pa->coord.refValue()[triangle.refIndex(i*3+0)]).z<=pos&&
			(pb->imatrix*pb->coord.refValue()[triangle.refIndex(i*3+1)]).z<=pos&&
			(pc->imatrix*pc->coord.refValue()[triangle.refIndex(i*3+2)]).z<=pos){
			poly+=i;
		}
	}
	dp.doBuild();
}
void Art::cutBackward(float pos){
	static DelPoly dp;
	M_Int& poly=dp.poly.refValue();
	poly.length=0;
	dp.part=this;

	for(int i=0; i<triangle.length; i++ ){
		Art* pa=triangle.refPart(i*3+0);
		Art* pb=triangle.refPart(i*3+1);
		Art* pc=triangle.refPart(i*3+2);
		if(	(pa->imatrix*pa->coord.refValue()[triangle.refIndex(i*3+0)]).z>=pos&&
			(pb->imatrix*pb->coord.refValue()[triangle.refIndex(i*3+1)]).z>=pos&&
			(pc->imatrix*pc->coord.refValue()[triangle.refIndex(i*3+2)]).z>=pos){
			poly+=i;
		}
	}
	dp.doBuild();
}
void Art::cutLeftFull(float pos){
	static DelPoly dp;
	M_Int& poly=dp.poly.refValue();
	poly.length=0;
	dp.part=this;

	for(int i=0; i<triangle.length; i++ ){
		Art* pa=triangle.refPart(i*3+0);
		Art* pb=triangle.refPart(i*3+1);
		Art* pc=triangle.refPart(i*3+2);
		if(	(pa->imatrix*pa->coord.refValue()[triangle.refIndex(i*3+0)]).x<pos||
			(pb->imatrix*pb->coord.refValue()[triangle.refIndex(i*3+1)]).x<pos||
			(pc->imatrix*pc->coord.refValue()[triangle.refIndex(i*3+2)]).x<pos){
			poly+=i;
		}
	}
	dp.doBuild();
}

/*
void IndexedFaceSet::genFlatTextureMapZ( float left,float right, float down, float up,float xOffset, float yOffset, float xScale, float yScale){
	M_Vec3f& point=refPoint();
	if(texCoord==NULL){
		texCoord=new TextureCoordinate;
	}
	float xdis=right-left;
	float ydis=up-down;

	M_Vec2f& texPoint=refTexPoint();
	texPoint.resize(point.length);
	for(int i=0;i<texPoint.length;i++){
		texPoint[i].x=xOffset+xScale*((point[i].x-left)/xdis);
		texPoint[i].y=yOffset+yScale*((point[i].y-down)/ydis);
	}
}
*/
template<>
void M_Vec3f::translateToZero(int offset,int numOf, Vec3f& center){
	if((unsigned)(offset+numOf)>length.rawLen)throw 0;
	for(int i=0;i<numOf;i++){
		center+=length.pointer[offset+i];
	}
	center/=(float)numOf;
	translate( -center);
}
/*
void IndexedFaceSet::rotate(Rotation rot){
	M_Vec3f& point=refPoint();
	M_Vec3f& vector=refVector();
	Matrix m;
	m=rotationMatrix(rot);
	int i;
	if(&point!=NULL){
		for(i=0;i<point.length;i++){
			point[i]=m.multVecMatrix(point[i]);
		}
	}
	if(&vector!=0){
		for(i=0;i<vector.length;i++){
			vector[i]=m.multVecMatrix(vector[i]);
		}
	}
}
void IndexedFaceSet::rotate(float x, float y, float z, float angle){
	rotate(Rotation(x,y,z,angle));
}
*/
void Matrix::scale(Vec3f& sc){
	f11 = f22 = f33 = f44 = 1.0f;
	f12 = f13 = f14 = f41 = 0.0f;
	f21 = f23 = f24 = f42 = 0.0f;
	f31 = f32 = f34 = f43 = 0.0f;
	f11= sc.x;
	f22= sc.y;
	f33= sc.z;
}
Matrix scaleMatrix(Vec3f sc){
	Matrix mat;
	mat.f11 = mat.f22 = mat.f33 = mat.f44 = 1.0f;
	mat.f12 = mat.f13 = mat.f14 = mat.f41 = 0.0f;
	mat.f21 = mat.f23 = mat.f24 = mat.f42 = 0.0f;
	mat.f31 = mat.f32 = mat.f34 = mat.f43 = 0.0f;
	mat.f11= sc.x;
	mat.f22= sc.y;
	mat.f33= sc.z;
	return mat;
}

Matrix translationMatrix(Vec3f tran){
	Matrix mat;
	mat.f41= tran.x;
	mat.f42= tran.y;
	mat.f43= tran.z;
/*
	Matrix mat;
	mat.f11 = mat.f22 = mat.f33 = mat.f44 = 1;
	mat.f12 = mat.f13 = mat.f14 = 0;
	mat.f21 = mat.f23 = mat.f24 = 0;
	mat.f31 = mat.f32 = mat.f34 = 0;
	mat.f41= tran.x;
	mat.f42= tran.y;
	mat.f43= tran.z;
*/
	return mat;
}

Vec3f Matrix::operator *(const Vec3f vec)const{
	Vec3f tmp;
	tmp.x = vec.x*m[0][0]+vec.y*m[1][0]+vec.z*m[2][0]+m[3][0];
	tmp.y = vec.x*m[0][1]+vec.y*m[1][1]+vec.z*m[2][1]+m[3][1];
	tmp.z = vec.x*m[0][2]+vec.y*m[1][2]+vec.z*m[2][2]+m[3][2];
	return tmp;
}
template <class T> Vec3<T> Vec3<T>::operator *=(const Matrix matrix){
	Vec3<T> tmp;
	tmp.x = x*matrix.m[0][0]+y*matrix.m[1][0]+z*matrix.m[2][0]+matrix.m[3][0];
	tmp.y = x*matrix.m[0][1]+y*matrix.m[1][1]+z*matrix.m[2][1]+matrix.m[3][1];
	tmp.z = x*matrix.m[0][2]+y*matrix.m[1][2]+z*matrix.m[2][2]+matrix.m[3][2];
	*this=tmp;
	return *this;
}

Vec3f Matrix::multVecMatrix(const Vec3f vec)const{
	Vec3f tmp;
	tmp.x = vec.x*m[0][0]+vec.y*m[1][0]+vec.z*m[2][0]+m[3][0];
	tmp.y = vec.x*m[0][1]+vec.y*m[1][1]+vec.z*m[2][1]+m[3][1];
	tmp.z = vec.x*m[0][2]+vec.y*m[1][2]+vec.z*m[2][2]+m[3][2];
	return tmp;
}

template<>
void M_Vec3f::multMatrixLeft(Matrix& matrix){
	Vec3f tmp;
	int i;
	M_Vec3f& l=*this;

	for(i=0;i<length;i++){
		tmp.x = l[i].x*matrix.m[0][0]+l[i].y*matrix.m[1][0]+l[i].z*matrix.m[2][0]+matrix.m[3][0];
		tmp.y = l[i].x*matrix.m[0][1]+l[i].y*matrix.m[1][1]+l[i].z*matrix.m[2][1]+matrix.m[3][1];
		tmp.z = l[i].x*matrix.m[0][2]+l[i].y*matrix.m[1][2]+l[i].z*matrix.m[2][2]+matrix.m[3][2];
		l[i].x=tmp.x;
		l[i].y=tmp.y;
		l[i].z=tmp.z;
	}
}
Matrix Matrix::multRight(Matrix mat){

// Multiply matrix MAT1 by matrix MAT2,
//  returning the result in RESULT
	Matrix tmp;

	for (int i=0; i<4; i++){
		for (int j=0; j<4; j++){
			tmp.m[i][j]=0;
			for (int k=0; k<4; k++){
				tmp.m[i][j]+=mat.m[i][k] * this->m[k][j];
			}
		}
	}
	return tmp;
}
Matrix Matrix::operator *(const Matrix mat)const{
	Matrix tmp;

	for (int i=0; i<4; i++){
		for (int j=0; j<4; j++){
			tmp.m[i][j]=0;
			for (int k=0; k<4; k++){
				tmp.m[i][j]+=this->m[i][k] * mat.m[k][j];
			}
		}
	}
	return tmp;
}
Matrix Matrix::operator *=(Matrix mat){
	Matrix tmp;

	for (int i=0; i<4; i++){
		for (int j=0; j<4; j++){
			tmp.m[i][j]=0;
			for (int k=0; k<4; k++){
				tmp.m[i][j]+=this->m[i][k] * mat.m[k][j];
			}
		}
	}
	*this=tmp;
	return *this;
}

/*
Matrix Matrix::multLeft(Matrix mat){

// Multiply matrix MAT1 by matrix MAT2,
//  returning the result in RESULT
	Matrix tmp;

	for (int i=0; i<4; i++){
		for (int j=0; j<4; j++){
			tmp.m[i][j]=0;
			for (int k=0; k<4; k++){
				tmp.m[i][j]+=this->m[i][k] * mat.m[k][j];
			}
		}
	}
	return tmp;
}
*/
void Matrix::setTransform(Vec3f translation, Rotation rotation, Vec3f scaleFactor){
	Matrix tran;
	tran.f41= translation.x;
	tran.f42= translation.y;
	tran.f43= translation.z;

	f44 = 1.0f;
	f12 = f13 = f14 = f41 = 0.0f;
	f21 = f23 = f24 = f42 = 0.0f;
	f31 = f32 = f34 = f43 = 0.0f;
	f11= scaleFactor.x;
	f22= scaleFactor.y;
	f33= scaleFactor.z;


	if (rotation.angle != 0.0f) {
		*this=*this*rotationMatrix(rotation);
	}
	*this=*this*tran;
}

void addNodeFieldToScript(Node* node){
	Script& script=*getOutputScriptPtr();

	for(int k=script.size(); --k>=0;){
		if(SFNode* sfnode=dynamic_cast<SFNode*>(script[k])){
			if(sfnode->p==node){
				return;
			}
		}
	}
	M_Properties& properties=const_cast<M_Properties&>(script.getProperties());

	script+=new SFNode(node);
	properties+=Properties(node->name);

}
void Script::useSFNode(Node* node){

	for(int k=size(); --k>=0;){
		if(SFNode* sfnode=dynamic_cast<SFNode*>((*this)[k])){
			if(sfnode->p==node){
				return;
			}
		}
	}
	const M_Properties& properties=getProperties();

	*this+=new SFNode(node);
	const_cast<M_Properties&>(properties)+=Properties(node->name);
}
void replaceNodeReferenceIfSameName(Transform* transform, Node*& node){
	if(transform->name == node->name){
		node = transform;
		return;
	}
	M_Node& children = transform->children.refValue();
	for(int k = 0; k < children.length; k++){
		if(Transform* transform = dynamic_cast<Transform*>(children[k].p)){
			replaceNodeReferenceIfSameName(transform, node);
		}
	}
}
void addNodeNameForScript(String& str, Node* node, int fieldIndex){
	const M_Properties& properties=node->getProperties();
	Art* part=dynamic_cast<Art*>(node);
	bool outSegmented = (part && part->smls->outSegmented);
	Morpher* morpher;

	if(g_outputMode&&g_genHAnimNodes==false && outSegmented==false){
		if(part){
			Seamless* smls=part->smls;
			Script* script=dynamic_cast<Script*>(smls->outputNode.p);

			str+=script->name;
			str+='.';
			str+=part->name;
			str+=fieldIndex;
			addNodeFieldToScript(script);
			return;
		}
		else if((morpher=dynamic_cast<Morpher*>(node))){
			Seamless* smls=morpher->smls;
			Script* script=dynamic_cast<Script*>(smls->outputNode.p);

			str+=script->name;
			str+='.';
			str+="w";
			makeNodeNameFromAddr(str, node);
			addNodeFieldToScript(script);
			return;
		}
	}
	if(outSegmented){
		if((part=dynamic_cast<Art*>(node))){
			if(Transform* transform=dynamic_cast<Transform*>(part->smls->outputNode.p)){
				replaceNodeReferenceIfSameName(transform, node);
			}
		}
	}
	else if(g_genHAnimNodes){
		if((part=dynamic_cast<Art*>(node))){
			if(HAnimHumanoid* humanoid=dynamic_cast<HAnimHumanoid*>(part->smls->outputNode.p)){
				M_Node& joints=humanoid->joints;
				for(int i=joints.length; --i>=0;){
					if(joints[i]->name==part->name){
						node=joints[i];
						break;
					}
				}
			}
		}
	}
	str+=node->name;
	str+='.';
	str+=properties[fieldIndex].name;

	addNodeFieldToScript(node);

}

void addNodeNameForRoute(String& str, Node* node, int fieldIndex){
	ExportVrmlX3d* exportVrmlX3d = dynamic_cast<ExportVrmlX3d*>(g_scene->exportVrmlX3d.p);
	const M_Properties& properties=node->getProperties();
	Art* part;
	Morpher* morpher;
	if(g_outputMode && g_genHAnimNodes == false && exportVrmlX3d->segmentedParts == false){
		if((part=dynamic_cast<Art*>(node))!=0){
			Seamless* smls=part->smls;
			Script* script=dynamic_cast<Script*>(smls->outputNode.p);

			str+=script->name;
			str+='.';
			str+=part->name;
			str+=fieldIndex;
			return;
		}
		else if((morpher=dynamic_cast<Morpher*>(node))!=0){
			Seamless* smls=morpher->smls;
			Script* script=dynamic_cast<Script*>(smls->outputNode.p);

			str+=script->name;
			str+='.';
			str+="w";
			makeNodeNameFromAddr(str, node);
			return;
		}
	}
	str+=node->name;
	str+='.';
	str+=properties[fieldIndex].name;
}
void addNodeNameForRouteXML(String& str, Node* node, int fieldIndex, String fromTo){
	const M_Properties& properties=node->getProperties();
	if(g_outputMode&&g_genHAnimNodes==false){
		if(Art* part=dynamic_cast<Art*>(node)){
			Seamless* smls=part->smls;
			Script* script=dynamic_cast<Script*>(smls->outputNode.p);

			str+=script->name;
			str+=fromTo;
			str+=part->name;
			str+=fieldIndex;
			return;
		}
		else if(Morpher* morpher=dynamic_cast<Morpher*>(node)){
			Seamless* smls=morpher->smls;
			Script* script=dynamic_cast<Script*>(smls->outputNode.p);

			str+=script->name;
			str+=fromTo;
			str+="w";
			makeNodeNameFromAddr(str, node);
			return;
		}
	}
	str+=node->name;
	str+=fromTo;
	str+=properties[fieldIndex].name;
}
void listRoutes(String& str){
	for(int k=0; k<g_scene->defList.length; k++){
		Node* fromNode=g_scene->defList[k];
		for(int i=0; i<fromNode->getFLen(); i++){
			Field* field=fromNode->getField(i);
			const M_Route &route=field->getRoute();
			for(int r=0; r<route.length;r++){

				Node* toNode=route[r].toNode;

				if(g_writeX3DXML){
					str+="\n<ROUTE fromNode='";
					addNodeNameForRouteXML(str, fromNode, i, "' fromField='");
					str+="' toNode='";

					addNodeNameForRouteXML(str, toNode, route[r].toFieldIdx, "' toField='");
					str+="'/>";
				}
				else{
					str+="\nROUTE ";
					addNodeNameForRoute(str, fromNode, i);
					str+=" TO ";

					addNodeNameForRoute(str, toNode, route[r].toFieldIdx);
				}
			}
		}
	}
	g_scene->defList.length=0;
}
template<>
void M_Node::toMNodeString(String& str, bool newList){
	if(length==1){
		(*this)[0]->toString(str,newList);
	}
	else {
		str+=g_vrmlSquareOpenBracket;
		for(int i=0;i<length;i++){
			str+='\n';
			(*this)[i]->toString(str,newList);
		}
		if(length > 0){
			str+='\n';
		}
		str+=g_vrmlSquareCloseBracket;
	}
}
void getNodesFromStr(M_Node& m_node, const String &str, int &i, bool multipleNodes=true){
	str.stb(i);
	while(i < str.length-1 && str[i]!=']'){
		/*
		if(str.cmpKWord("PROTO",i)){
			i=str.indexOf("}",i)+1;
			str.stb(i);
			continue;
		}
		*/
		if(str.cmpKWord("ROUTE",i)){
			S_Route(str,i);
		}
		else if(str.cmpKWord("PROFILE",i)){
			str.stb(i);
			str.stw_stb(i);
			continue;
		}
		else if(str.cmpKWord("META",i)){
			str.stnl(i);
			continue;
		}
		else if(str.cmpKWord("COMPONENT",i)){
			str.stnl(i);
			continue;
		}
		else{
			m_node += getNode2(str, i);
		}
		checkBackgroundColor(str, i);
		if(!multipleNodes) break;
	}
	if(multipleNodes) i++;
}
template<>
void M_Node::getField(const String& str, int &i){
	bool brackFlag=false;
	str.stb(i);
	if(str[i]=='['){
		brackFlag=true;
		i++;
	}
	str.stb(i);
	getNodesFromStr(*this, str, i, brackFlag);
}
template<>
void M_SArt::getField(class String const &,int &){
	throw this;
}

void SurfaceGenerator::removeGenerator(M_Node& par1){
	M_Node& gnrtr=generator.refValue();
	int i,j,k;
	for(j=0;j<par1.length;j++){
		for(i=0;i<gnrtr.length;i++){
			if(gnrtr[i]==par1[j]){
				/*
				Node* tmp=&*children[i];
				if(--tmp->refCount==0){
					delete tmp;
				}
*/
				for(k=i;k<generator.length-1;k++){
					gnrtr[k]=gnrtr[k+1];
				}
				gnrtr.length--;
			}
		}
	}
}
void Art::addPart(Art* parA){
	part+=parA;
}
template<>
void M_Node::removeNode(NodePtr par1){
	for(int i=0;i<length;i++){
		if(length.pointer[i]==par1){
			for(int k=i;k<length-1;k++){
				length.pointer[k]=length.pointer[k+1];
			}
			this->length--;
		}
	}
}
template<>
void M_Node::removeNode(M_Node &par1){
	for(int j=0;j<par1.length;j++){
		removeNode(par1[j]);
	}
}
void Scene::removeChildren(M_Node& par1){
	children.refValue().removeNode( par1);
}
void Art::removePart(M_SArt& par){
	((M_Node*)&(part.refValue()))->removeNode( *(M_Node*)&par);
}


template<>
void SFBool::toString(String& str, Field *protoField, bool){
	if(refValue()!=dynamic_cast<SFBool*>(protoField)->refValue()){
		if(g_writeX3DXML){
			if(refValue()){
				str+="true";
			}
			else {
				str+="false";
			}
		}
		else{
			if(refValue()){
				str+="TRUE";
			}
			else {
				str+="FALSE";
			}
		}
	}
}
void SFColor::toString(String& str, Field *protoField, bool){
	if(refValue()!=dynamic_cast<SFColor*>(protoField)->refValue()){
		str+=refValue();
	}
}
template<>
void SFFloat::toString(String& str, Field *protoField, bool){
	if(fabs(refValue()-dynamic_cast<SFFloat*>(protoField)->refValue())>.0000001){
		str+=refValue();
	}
}
template<>
void SFImage::toString(String&, Field*, bool){
	throw this;
}
template<>
void SFInt32::toString(String& str, Field *protoField, bool){
	if(refValue()!=dynamic_cast<SFInt32*>(protoField)->refValue()){
		str+=refValue();
	}
}
void SFNode::toString(String& str, Field*, bool newList){
//	NodePtr node=dynamic_cast<SFNode*>(field)->refValue();
	if(refValue()!=0){
		refValue()->toString(str,newList);
	}
}
void SFRotation::toString(String& str, Field *protoField, bool){
	if(refValue()!=dynamic_cast<SFRotation*>(protoField)->refValue()){
		str+=refValue();
	}
}
void SFString::toString(String& str, Field *protoField, bool){
	if(refValue()!=dynamic_cast<SFString*>(protoField)->refValue()){
		str+=g_vrmlQuote;
		str+=refValue();
		str+=g_vrmlQuote;
	}
}
template<>
void SFTime::toString(String& str, Field *protoField, bool){
	if(refValue()!=dynamic_cast<SFTime*>(protoField)->refValue()){
		str+=refValue();
	}
}
void SFVec2f::toString(String& str, Field *protoField, bool){
	if(refValue()!=dynamic_cast<SFVec2f*>(protoField)->refValue()){
		str+=refValue();
	}
}
void SFVec3f::toString(String& str, Field *protoField, bool){
	if(refValue()!=dynamic_cast<SFVec3f*>(protoField)->refValue()){
		str+=refValue();
	}
}
template<>
void MFBool::toString(String& str, Field *protoField, bool){
	if(length!=0||dynamic_cast<MFBool*>(protoField)->length!=0){
		str+=refValue();
	}
}
template<>
void MFColor::toString(String& str, Field *protoField, bool){
	if(length!=0||dynamic_cast<MFColor*>(protoField)->length!=0){
		str+=refValue();
	}
}
template<>
void MFFloat::toString(String& str, Field *protoField, bool){
	if(getValue()!=dynamic_cast<MFFloat*>(protoField)->getValue()){
		str+=refValue();
	}
}
template<>
void MFImage::toString(String& str, Field *protoField, bool){
	if(length!=0||dynamic_cast<MFImage*>(protoField)->length!=0){
		str+=refValue();
	}
}
template<>
void MFInt32::toString(String& str, Field *protoField, bool){
	if(length!=0||dynamic_cast<MFInt32*>(protoField)->length!=0){
		str+=refValue();
	}
}
template<>
void MFNode::toString(String& str, Field *protoField, bool newList){
	if(g_outputMode&&length==1&&refValue()[0]->getType()==SEAMLESS&&g_writeX3DXML==false){
		str+="[";
		if(length!=0||dynamic_cast<MFNode*>(protoField)->length!=0){
			refValue().toMNodeString(str, newList);
		}
		str+=']';
	}
	else if(length!=0||dynamic_cast<MFNode*>(protoField)->length!=0){
		refValue().toMNodeString(str, newList);
	}
}
template<>
void MFRotation::toString(String& str, Field *protoField, bool){
	if(length!=0||dynamic_cast<MFRotation*>(protoField)->length!=0){
		str+=refValue();
	}
}
template<>
void MFString::toString(String& str, Field *protoField, bool){
	if(length!=0||dynamic_cast<MFString*>(protoField)->length!=0){
		if(length==1){
			str+='\"'; //contact does not like no double quote for xml (g_vrmlQuote). Is no double quote for a single string in a MFString legal?
			str+=refValue()[0];
			str+='\"';
		}
		else {
			str+=g_vrmlSquareOpenBracket;
			if(length > 0){
				str+='\"';
				str+=refValue()[0];
				str+='\"';
				for(int i=1;i<length;i++){
					str+=' ';
					str+='\"';
					str+=refValue()[i];
					str+='\"';
				}
			}
			str+=g_vrmlSquareCloseBracket;
		}
	}
}
template<>
void MFTime::toString(String& str, Field *protoField, bool){
	if(length!=0||dynamic_cast<MFTime*>(protoField)->length!=0){
		str+=refValue();
	}
}
template<>
void MFVec2f::toString(String& str, Field *protoField, bool){
	if(length!=0||dynamic_cast<MFVec2f*>(protoField)->length!=0){
		str+=refValue();
	}
}
template<>
void MFVec3f::toString(String& str, Field *protoField, bool){
	if(length!=0||dynamic_cast<MFVec3f*>(protoField)->length!=0){
		str+=refValue();
	}
}






void toStringRoundDown(float number, int numOfDigits, String& str){
	if(numOfDigits>9)numOfDigits=9;
	char format[6]="%.6f";
	format[2]='0'+numOfDigits;

	char p[20];

	sprintf(p,format,number);
	removeZeros(p+1);
	str+=p;
}

void toStringRoundDown(float* number, int len, int numOfDigits, String& str){
	if(numOfDigits>9)numOfDigits=9;
	char format[6]=" %.6f";
	format[3]='0'+numOfDigits;

	char p[20];

	str+='[';
	for(int i=0; i<len; i++){
		sprintf(p,format,number[i]);
		removeZeros(p+1);
		str+=p;
	}
	str+=']';
}
int g_formatCoords=10;
const int g_formatColor=2;
inline void extraPartToString(String& str, Art* part){

	if(part->show==false){
		str+="\nshow FALSE";
	}
	Seamless* smls=part->smls;

	if(dynamic_cast<ImmediatePart*>(part) || smls->bnspe){
/*
		if(part->coord.length){
			str+="\ncoord ";
			toStringRoundDown((float*)&part->coord[0], part->coord.length*3, g_formatCoords, str);
		}
*/
		if(part->rc.length){
			str+="\nrc [";
			M_Int& rc=part->rc;

			for (int i=0; i<rc.length; i++){
				str+=' ';
				str+=part->fromCoord + rc[i];
			}

			str+="]\n";
		}
/*
		if(part->normal.length){
			str+="\nnormal ";
			toStringRoundDown((float*)&part->normal[0], part->normal.length*3, g_formatCoords, str);
			//str+=part->normal.toString();
		}
*/
		if(part->color.length){
			str+="\ncolor ";
			toStringRoundDown((float*)&part->color[0], part->color.length*3, g_formatColor, str);
//			str+=part->color.toString();
		}
		if(part->texCoord.length){
			str+="\ntexCoord ";
			toStringRoundDown((float*)&part->texCoord[0], part->texCoord.length*2, g_formatCoords, str);
			//str+=part->texCoord.toString();
		}
	}
	str+="\n";
}
String xmlAddRoute(Node* fromNode, String fromField, Node* toNode, String toField){
	String str;
	str+="\n<ROUTE fromNode='";
	str+=fromNode->name;
	str+="' fromField='";
	str+=fromField;
	str+="' toNode='";
	str+=toNode->name;
	str+="' toField='";
	str+=toField;
	str+="'/>";
	return str;
}
String vrmlAddRoute(Node* fromNode, String fromField, Node* toNode, String toField){
	String str;
	str+="\nROUTE ";
	str+=fromNode->name;
	str+='.';
	str+=fromField;
	str+=" TO ";
	str+=toNode->name;
	str+=".";
	str+=toField;
	return str;
}
void AnimByAnkle::outputToString(String& str){
	bool newList=false;
	Art* pelvis=this->pelvis.refValue();
	Art* hip=this->hip.refValue();
	Art* knee=this->knee.refValue();
	Art* ankle=this->ankle.refValue();
	if(ankle==0){
		return;
	}
	int rotIndex=hip->getFieldIndex(&hip->rotation);
	Node* destScript=pelvis->smls->outputNode;
	if(g_writeX3DXML){
		str+="<AnimByAnkle ";
		if(setDEFgetUSE( str, newList)){
			return;
		}
		if(pelvis!=0){
			str+=" pelvisIdx='";
			str+=pelvis->plidx;
			str+='\'';
		}
		if(hip!=0){
			str+=" hipIdx='";
			str+=hip->plidx;

			str+="' hipTran='";
			hip->translation.toString(str);
			str+='\'';
		}
		if(knee!=0){
			str+=" kneeIdx='";
			str+=knee->plidx;

			str+="' kneeTran='";
			knee->translation.toString(str);
			str+='\'';
		}
		if(ankle!=0){
			str+=" ankleIdx='";
			str+=ankle->plidx;

			str+="' ankleTran='";
			ankle->translation.toString(str);
			str+='\'';
		}
		str+=">\n<Script containerField='script' USE='";
		str+=destScript->name;
		str+="'/>\n";


		str+=xmlAddRoute(this, "hipOut", destScript, hip->name+rotIndex);
		str+=xmlAddRoute(this, "kneeOut", destScript, knee->name+rotIndex);
		str+=xmlAddRoute(this, "ankleOut", destScript, ankle->name+rotIndex);


		str+="</AnimByAnkle>\n";

	}
	else{
		if(setDEFgetUSE( str, newList)){
			return;
		}
		str+=getTypeStr();
		str+=" {\n";

		str+="\nscript USE ";
		str+=destScript->name;

		if(pelvis!=0){
			str+="\npelvisIdx ";
			str+=pelvis->plidx;
		}
		if(hip!=0){
			str+="\nhipIdx ";
			str+=hip->plidx;

			str+="\nhipTran ";
			hip->translation.toString(str);
		}
		if(knee!=0){
			str+="\nkneeIdx ";
			str+=knee->plidx;

			str+="\nkneeTran ";
			knee->translation.toString(str);
		}
		if(ankle!=0){
			str+="\nankleIdx ";
			str+=ankle->plidx;

			str+="\nankleTran ";
			ankle->translation.toString(str);

		}
		str+='\n';

		str+=vrmlAddRoute(this, "hipOut", destScript, hip->name+rotIndex);
		str+=vrmlAddRoute(this, "kneeOut", destScript, knee->name+rotIndex);
		str+=vrmlAddRoute(this, "ankleOut", destScript, ankle->name+rotIndex);

		str+="\n}";
	}
}
void ProtoInstance::toString(String& str, bool newList){
	if(setDEFgetUSE( str, newList)){
		return;
	}
	str+=nodeLookUpStr[getType()];
	str+='{';
	for(int i = 0; i < getFLen(); i++){
		Field* field=getField(i);
		const Properties& sprop=getProp(i);
		if(field->isRef()){
			str+='\n';
			str+=sprop.name;
			str+=" IS ";
			ProtoDeclare* protoDeclare=0;
			int j=0;
			findMatchingIs(field, protoDeclare, j);
			str+=protoDeclare->protoInterface->properties[j].name;
		}
		else{
			if(	(sprop.isReadable()^sprop.isWritable())==false){

				Field* protoField=g_lookupProtoInterface[getType()]->getField(i);

				String tmpStr;
				field->toString(tmpStr, protoField, newList);
				if(tmpStr!=""){
					str+='\n';
					str+=sprop.name;
					str+=' ';
					str+=tmpStr;
				}
			}
		}
	}
	str+="\n}";
}

template<>
void MFNode::toXMLString(String& str, bool newList, const char* containerField){
	if(refValue().length){
		refValue()[0]->toXMLString(str, newList, containerField);
		for(int i=1;i<length;i++){
			str+='\n';
			refValue()[i]->toXMLString(str, newList, containerField);
		}
	}
}

void ProtoInstance::toXMLString(String& str, bool newList, const char*){
	str+="<ProtoInstance name='";
	//str+=name;
	str+=nodeLookUpStr[getType()];
	str+='\'';
	if(setDEFgetUSE( str, newList)){
		return;
	}
	str+='>';
	int i;
	static String strIS;
	strIS="\n<IS>";
	ProtoInterface* protoInterface=g_lookupProtoInterface[getType()];

	for(i=0; i < getFLen(); i++){
		Field* field=getField(i);
		const Properties& sprop=getProp(i);
		if(field->isRef()){
			strIS+="\n<connect nodeField='";
			strIS+=sprop.name;
			strIS+="' protoField='";
			ProtoDeclare* protoDeclare=0;
			int j=0;
			findMatchingIs(field, protoDeclare, j);
			strIS+=protoDeclare->protoInterface->properties[j].name;
			strIS+="'/>";
		}
		else if((sprop.isReadable()^sprop.isWritable())==false){
			MFNode* mfnode=dynamic_cast<MFNode*>(field);
			SFNode* sfnode=dynamic_cast<SFNode*>(field);
			if(mfnode || sfnode){
				str+="\n<fieldValue name='";
				str+=sprop.name;
				str+="'>\n";
				if(mfnode){
					mfnode->toXMLString(str, newList, 0);
				}
				else if(sfnode){
					if(Node* node=sfnode->p){
						node->toXMLString(str, newList, 0);
					}
				}
				str+="\n</fieldValue>";
			}
			else{
				Field* protoField=protoInterface->getField(i);
				String tmpStr;
				field->toString(tmpStr, protoField, newList);
				if(tmpStr!=""){
					str+="\n<fieldValue name='";
					str+=sprop.name;

					str+="' value='";

					str+=tmpStr;

					str+="'/>";
				}
			}
		}
	}
	if(strIS.length!=5){
		str+=strIS;
		str+="\n</IS>";
	}
	str+="\n</ProtoInstance>";
}

inline void protoDeclareFieldtoString(const ProtoInterface* protoInterface, const int& i, String& str){
	Field* field=protoInterface->getField(i);
	const Properties& sprop=protoInterface->properties[i];
	if(sprop.isWritable()){
		if(sprop.isReadable()){
			str+=g_exposedFieldPtr;
			str+=' ';
			str+=field->getTypeStr();
			str+=' ';
			str+=sprop.name;
			str+=' ';
			field->toString(str);
		}
		else{
			str+=g_eventInPtr;
			str+=' ';
			str+=field->getTypeStr();
			str+=' ';
			str+=sprop.name;
		}
	}
	else if(sprop.isReadable()){
		str+=g_eventOutPtr;
		str+=' ';
		str+=field->getTypeStr();
		str+=' ';
		str+=sprop.name;
	}
	else {
		str+=g_fieldPtr;
		str+=' ';
		str+=field->getTypeStr();
		str+=' ';
		str+=sprop.name;
		str+=' ';
		field->toString(str);
	}
}
void ProtoDeclare::toString(String& str, bool newList){
	newList=false; //should be true but we cant do this yet
	if(newList){
		g_scene->defList=M_Node();
#ifdef SEAMLESS3DGUI
		g_sceneTreeNodeIndex=0;
#endif
	}
	g_protoList+=this;

	ProtoInterfacePtr protoInterface=this->protoInterface;
	ProtoBody* protoBody=dynamic_cast<ProtoBody*>(this->protoBody.p);
	const M_Node& children=protoBody->children;
	const M_Properties& properties=protoInterface->properties;

	if(g_writeX3DXML){
		str+="<ProtoDeclare name='";
		str+=name;
		str+="'>";

		if(getFLen()){
			str+="\n<ProtoInterface>";
			for(int i=0; i<protoInterface->getFLen(); i++){
				Field* field=protoInterface->getField(i);
				const Properties& sprop=properties[i];
				str+="\n<field name='";
				str+=sprop.name;
				str+="' type='";
				str+=field->getTypeStr();
				if(sprop.isWritable()){
					if(sprop.isReadable()){
						str+="' accessType='inputOutput' value='";
						field->toString(str);
					}
					else{
						str+="' accessType='inputOnly";
					}
				}
				else if(sprop.isReadable()){
					str+="' accessType='outputOnly";
				}
				else {
					str+="' accessType='initializeOnly' value='";
					field->toString(str);
				}
				str+="'/>";
			}
		}
		str+="\n</ProtoInterface>\n<ProtoBody>";
		{
			for(int i=0; i<children.length; i++){
				str+='\n';
				children[i]->toXMLString(str, newList, 0);
			}
		}
	}
	else{
		str+="PROTO ";
		str+=name;
		str+="[";

		{
			if(protoInterface->getFLen() > 0){
				for(int i=0;i<protoInterface->getFLen();i++){
					str+='\n';
					protoDeclareFieldtoString(protoInterface, i, str);
				}
				str+='\n';
			}
		}
		str+="]\n{";
		{
			for(int i=0; i<children.length; i++){
				str+='\n';
				children[i]->toString(str, newList);
			}
		}
	}
	outputUserScripts(str, newList);

	listRoutes(str);

	if(g_writeX3DXML){
		str+="\n</ProtoBody>\n</ProtoDeclare>";
	}
	else{
		str+="\n}";
	}

/*
	if(g_outputMode&&name=="Avatar"){
		if(g_writeX3DXML){
			str+="<ProtoInstance name='Avatar'/>\n";
		}
		else{
			str+="Avatar{}\n";
		}
	}
	*/
}
/*
inline bool isPartRef(Field* field){
	if(g_outputMode && field->getType()==SFNODE){
		NodePtr node = dynamic_cast<SFNode*>(field)->refValue();
		for(int k=0; k<g_changePart.length; k++){
			if(g_changePart[k]==node){
				return true;
			}
		}
	}
	return false;
}
*/
inline void addUrl(M_String& dstUrl, M_String& srcUrl){
	dstUrl[0]+=srcUrl[0].substring(String("javascript:").length);
}
void Node::outputXMLFields(int startField, int fieldLen, const M_Properties& properties, ProtoInterface& protoInterface, bool newList, String& str, int& i){
	String strIS;
	strIS="\n<IS>";
	for(i=startField; i<fieldLen; i++){
		const Properties& sprop=properties[i];
		if(sprop.doNotOutput()){
			continue;
		}
		Field* field=getField(i);
		if(field->isRef()){
			strIS+="\n<connect nodeField='";
			strIS+=properties[i].name;
			strIS+="' protoField='";
			ProtoDeclare* protoDeclare=0;
			int j=0;
			findMatchingIs(field, protoDeclare, j);
			strIS+=protoDeclare->protoInterface->properties[j].name;
			strIS+="'/>";
		}
		else if(field->getType()!=SFNODE && field->getType()!=MFNODE){
			if(	(sprop.isReadable()^sprop.isWritable())==false){

				Field* protoField=0;
				protoField=protoInterface.getField(i);

				String tmpStr;
				field->toString(tmpStr, protoField, newList);
				if(tmpStr!=""){
					str+=' ';
					str+=properties[i].name;
					str+="='";
					str+=tmpStr;
					str+='\'';
				}
			}
		}
	}
	str+=">";
	for(i=startField; i<fieldLen; i++){
		const Properties& sprop=properties[i];
		if(sprop.doNotOutput()){
			continue;
		}
		Field* field=getField(i);
		MFNode* mfnode=dynamic_cast<MFNode*>(field);
		SFNode* sfnode=dynamic_cast<SFNode*>(field);
		if(mfnode || sfnode){
			if(	(sprop.isReadable()^sprop.isWritable())==false){
				if(sfnode){
					Node* node=sfnode->p;
					if(node!=0){
						str+='\n';
						node->toXMLString(str, newList, sprop.name.c_str());
					}
				}
				else{
					if(mfnode->refValue().length){
						str+='\n';
						mfnode->toXMLString(str, newList, sprop.name.c_str());
					}
				}
			}
		}
	}
	if(strIS.length!=5){
		str+=strIS;
		str+="\n</IS>";
	}
	str+="\n</";
	str+=getTypeStr();
	str+='>';
}

bool getAvatarSocketNetSave(Node* node);

void Node::toString(String& str, bool newList){
	M_Vec3f storeCoord;
	M_Vec3f storeNormal;
	M_Color3f storeColor;
	M_Vec2f storeTexCoord;

	if(g_outputMode==true){
		if(G_Variable* v=dynamic_cast<G_Variable*>(this)){
			v->addToOutputScript();
			return;
		}
		switch(getType()){
			case NOOUTPUTGROUP:
				return;
			case AVATARSOCKET:
				if(getAvatarSocketNetSave(this) == false){
					return;
				}
			case SEAMLESS:
				dynamic_cast<Seamless*>(this)->outputToString(str);
				return;
			case ANIMBYANKLE:
				if(g_genHAnimNodes==false){
					dynamic_cast<AnimByAnkle*>(this)->outputToString(str);
					return;
				}
		}
	}
	int i;
	if(setDEFgetUSE( str, newList)){
		return;
	}
	if(ColorEffect* colorEffect = dynamic_cast<ColorEffect*>(this)){
		M_Color3f& color = colorEffect->color.refValue();
		storeColor = color;
		color.length=0;
	}
	else if(TextureEffect* textureEffect = dynamic_cast<TextureEffect*>(this)){
		M_Vec2f& texCoord = textureEffect->texCoord.refValue();
		storeTexCoord = texCoord;
		texCoord.length = 0;
	}
	else if(NurbsPatch* patch = dynamic_cast<NurbsPatch*>(this)){
		swapCPPointStorage(patch);
	}
	else if(Hair* hair = dynamic_cast<Hair*>(this)){
		hair->processToFileFormat();
	}

	if(g_outputMode){
		switch(getType()){
		case NORMAL:
		case TEXTURECOORDINATE:
		case COORDINATE:
			changeFloatFormat(g_formatCoords);
			break;
		case COLOR:
			changeFloatFormat(g_formatColor);
			break;
		default:
			changeFloatFormat(6);
		}
	}

	if(g_genHAnimNodes&&g_outputMode&&getType()==MORPHER){
		Morpher* vb=dynamic_cast<Morpher*>(this);
		M_Vec3f& displacement=vb->displacement.refValue();
		storeCoord=displacement;
		displacement.multMatrixLeft(vb->parent->matrix);
	}
	const M_Properties &properties=getProperties();
	str+=getTypeStr();
	str+='{';

	if(Seamless* smls = dynamic_cast<Seamless*>(this)){
		smls->extraSeamlessToString(str);
	}
	else if(Art* part=dynamic_cast<Art*>(this)){
		storeNormal = part->normal;
		part->normal.refValue().length=0;
		extraPartToString(str, part);
		Seamless* smls=part->smls;
		if(smls->build.length && dynamic_cast<ImmediatePart*>(part) == 0 && smls->bnspe == false){
			storeCoord = part->coord;
			part->coord.refValue().length = 0;
		}
	}
	for(i = 0; i < getFLen(); i++){
		const Properties& sprop = properties[i];
		if(sprop.doNotSave() || (g_outputMode && sprop.doNotOutput())){
			continue;
		}
		Field* field=getField(i);
		if(field->isRef()){
			str+='\n';
			str+=sprop.name;
			str+=" IS ";
			ProtoDeclare* protoDeclare=0;
			int j=0;
			findMatchingIs(field, protoDeclare, j);
			str+=protoDeclare->protoInterface->properties[j].name;
		}
		else{
			if(	(sprop.isReadable()^sprop.isWritable())==false){

				Field* protoField=g_lookupProtoInterface[getType()]->getField(i);

	/*
				if(field->getType()==SFNODE){//dont use toString for sfnode coz of newList parameter
					NodePtr node=dynamic_cast<SFNode*>(field)->refValue();
					if(node!=0){
						str+=properties[i].name;
						str+=' ';
						str+=node->toString(newList);
					}
				}
				else{
	*/
				String tmpStr;
				field->toString(tmpStr, protoField, newList);

				if(tmpStr!=""){
					str+='\n';
					if(Anim* anim=dynamic_cast<Anim*>(this)){
						if(field!=&anim->onFrame){
							str+=properties[i].name;
						}
					}
					else if(TimeSensor* ts=dynamic_cast<TimeSensor*>(this)){
						if(field!=&ts->on_fraction_changed){
							str+=properties[i].name;
						}
					}
					else if(TouchSensor* ts=dynamic_cast<TouchSensor*>(this)){
						if(field!=&ts->onTouchTime){
							str+=properties[i].name;
						}
					}
					else if(Mobility* mobility=dynamic_cast<Mobility*>(this)){
						if(field!=&mobility->onRot && field!=&mobility->onTran){
							str+=properties[i].name;
						}
					}
					else if(Build* build=dynamic_cast<Build*>(this)){
						if(field!=&build->setup){
							str+=properties[i].name;
						}
					}
					else{
						str+=properties[i].name;
					}
					str+=' ';
					str+=tmpStr;
				}
			}
		}
	}
	if(g_genHAnimNodes && g_outputMode && getType() == MORPHER){
		Morpher* vb = dynamic_cast<Morpher*>(this);
		M_Vec3f& displacement=vb->displacement.refValue();
		displacement=storeCoord;
	}
	if(ColorEffect* colorEffect = dynamic_cast<ColorEffect*>(this)){
		M_Color3f& color = colorEffect->color.refValue();
		color=storeColor;
	}
	else if(TextureEffect* textureEffect = dynamic_cast<TextureEffect*>(this)){
		M_Vec2f& texCoord = textureEffect->texCoord.refValue();
		texCoord = storeTexCoord;
	}
	else if(NurbsPatch* patch = dynamic_cast<NurbsPatch*>(this)){
		swapCPPointStorage(patch);
	}
	else if(Hair* hair = dynamic_cast<Hair*>(this)){
		hair->processToWorkFormat();
	}
	else if(Seamless* smls=dynamic_cast<Seamless*>(this)){
		if(g_colorOwnership){
			smls->ctrf = true; //to restore rc values
		}
	}
	else if(Art* part=dynamic_cast<Art*>(this)){
		Seamless* smls=part->smls;
		part->normal.refValue()=storeNormal;
		if(smls->build.length && dynamic_cast<ImmediatePart*>(part) == 0 && smls->bnspe == false){
			part->coord.refValue() = storeCoord;
		}
	}


	str+="\n}";
}

void Node::toXMLString(String& str, bool newList, const char* containerField){
	M_Vec3f storeCoord;
	M_Color3f storeColor;
	M_Vec2f storeTexCoord;

	if(getType()==PROTODECLARE){
		toString(str, newList);
		return;
	}
	if(g_outputMode){
		if(G_Variable* v=dynamic_cast<G_Variable*>(this)){
			v->addToOutputScript();
			return;
		}
		switch(getType()){
			case NOOUTPUTGROUP:
				return;
			case SEAMLESS:
				dynamic_cast<Seamless*>(this)->outputToString(str);
				return;
			case ANIMBYANKLE:
				if(g_genHAnimNodes==false){
					dynamic_cast<AnimByAnkle*>(this)->outputToString(str);
					return;
				}
		}
	}
	if(getType()==COLOREFFECT){
		M_Color3f& color=dynamic_cast<ColorEffect*>(this)->color.refValue();
		storeColor=color;
		color.length=0;
	}
	else if(getType()==TEXTUREEFFECT){
		M_Vec2f& texCoord=dynamic_cast<TextureEffect*>(this)->texCoord.refValue();
		storeTexCoord=texCoord;
		texCoord.length=0;
	}

	int i;
	const M_Properties &properties=getProperties();
	str+="<";
	str+=getTypeStr();
	if(containerField && (containerFieldLookUpStr[getType()]!=containerField) ){
		str+=" containerField='";
		str+=containerField;
		str+='\'';
	}
	if(setDEFgetUSE( str, newList)){
		return;
	}
	if(g_genHAnimNodes&&g_outputMode&&getType()==MORPHER){
		Morpher* vb=dynamic_cast<Morpher*>(this);
		M_Vec3f& displacement=vb->displacement.refValue();
		storeCoord=displacement;
		displacement.multMatrixLeft(vb->parent->matrix);
	}
	if(g_outputMode){
		switch(getType()){
		case NORMAL:
		case TEXTURECOORDINATE:
		case COORDINATE:
			changeFloatFormat(g_formatCoords);
			break;
		case COLOR:
			changeFloatFormat(g_formatColor);
			break;
		default:
			changeFloatFormat(6);
		}
	}

	if(getType()==SEAMLESS){
		dynamic_cast<Seamless*>(this)->extraSeamlessToString(str);
	}
	else if(dynamic_cast<Art*>(this)!=0){
		Art* part=dynamic_cast<Art*>(this);
		part->normal.refValue().length=0;
		extraPartToString(str, part);
		Seamless* smls=part->smls;
		if(dynamic_cast<ImmediatePart*>(part)==0&&smls->build.length){
			storeCoord=part->coord;
			part->coord.refValue().length=0;
		}
	}
	outputXMLFields(0, getFLen(), properties, *(g_lookupProtoInterface[getType()]), newList, str,i);

	if(g_genHAnimNodes&&g_outputMode&&getType()==MORPHER){
		Morpher* vb=dynamic_cast<Morpher*>(this);
		M_Vec3f& displacement=vb->displacement.refValue();
		displacement=storeCoord;
	}
	if(getType()==COLOREFFECT){
		dynamic_cast<ColorEffect*>(this)->color.refValue()=storeColor;
	}
	else if(getType()==TEXTUREEFFECT){
		dynamic_cast<TextureEffect*>(this)->texCoord.refValue()=storeTexCoord;
	}
	else if(dynamic_cast<Art*>(this)!=0){
		Art* part=dynamic_cast<Art*>(this);
		Seamless* smls=part->smls;
		if(dynamic_cast<ImmediatePart*>(part)==0&&smls->build.length){
			part->coord.refValue()=storeCoord;
		}
	}
}
static const char* singleFieldType[]={
	"bool",
	"Color3f",
	"float",
	"Image",
	"int",
	"NodePtr",
	"Rotation",
	"String",
	"double",
	"Vec2f",
	"Vec3f",
	"bool,Multiple",
	"Color3f,Multiple",
	"float,Multiple",
	"Image,Multiple",
	"int,Multiple",
	"NodePtr,Multiple",
	"Rotation,Multiple",
	"String,Multiple",
	"double,Multiple",
	"Vec2f,Multiple",
	"Vec3f,Multiple"
};
void addAccess(String& str, const Properties& sprop){
	str+=", ";
	if		(!sprop.isWritable()&&!sprop.isReadable()){
		str+="initializeOnly)";
	}
	else if	(!sprop.isWritable()&&sprop.isReadable()){
		str+="outputOnly)";
	}
	else if	(sprop.isWritable()&&!sprop.isReadable()){
		str+="inputOnly)";
	}
	else if	(sprop.isWritable()&&sprop.isReadable()){
		str+="inputOutput)";
	}
}
String g_protoInterfaceListStr;
String g_protoInterfaceNodePromptStr;
String g_protoInterfaceTips;
void ifExtra(String& str, bool& extra){
	if(extra==false){
		extra=true;
		str+='{';
	}
	else{
		str+=' ';
	}
}
void ProtoInterface::toCPPString(String& str, bool newList){
	int i = 0;
	String nodeName;
	if(name!=""){
		nodeName=name;
	}
	else {
		nodeName+='n';
		makeNodeNameFromAddr(nodeName, this);
	}
	if(setDEFgetUSE( str, newList)){
		return;
	}
	const M_Properties &properties=getProperties();

	/*
	str+=getTypeStr();
	str+="* ";
	str+=nodeName;
	str+=" = new ";
	str+=getTypeStr();
	str+=";\n";
	String nodeFieldName;
	*/
	static String node_field;
	int groupIndex=-1;
	for(i = 0; i < getFLen(); i++){
		Field* field = getField(i);
		const Properties& sprop = properties[i];
		node_field = nodeName;
		node_field+=sprop.name;

		str+="\tBIField(";
		str+=field->getTypeStr().toUpperCase();
		str+=",\t\"";


		str+=sprop.name;
//		if((sprop.flags&UNWRITEABLE && sprop.flags&UNREADABLE) || (!(sprop.flags&UNWRITEABLE) && !(sprop.flags&UNREADABLE))){
		if(((!sprop.isReadable()) && (!sprop.isWritable())) || (sprop.isReadable() && sprop.isWritable())){
			str+=' ';
			field->toString(str);
		}
		bool extra=false;
		if(sprop.groupIndex!=-1 && sprop.groupIndex!= groupIndex){
			groupIndex=sprop.groupIndex;
			if(i < getFLen()-1 && properties[i+1].groupIndex == groupIndex){
				ifExtra(str, extra);
				str+="beginGroup TRUE";
			}
		}
		if(sprop.groupIndex==-1 && sprop.groupIndex!= -1){
			ifExtra(str, extra);
			str+="endGroup TRUE";
			groupIndex=-1;
		}
		if(sprop.startAtIndex){
			ifExtra(str, extra);
			str+="startAtIndex ";
			str+=sprop.startAtIndex;
		}
		/*
		if(sprop.initElLen!=1){
			ifExtra(str, extra);
			str+="initElLen ";
			str+=sprop.groupIndex;
		}
		if(sprop.incElLen!=1){
			ifExtra(str, extra);
			str+="incElLen ";
			str+=sprop.groupIndex;
		}
		*/
		if(sprop.rangeFrom!=-2147483647){
			ifExtra(str, extra);
			str+="rangeFrom ";
			str+=sprop.rangeFrom;
		}
		if(sprop.rangeTo!=2147483647){
			ifExtra(str, extra);
			str+="rangeTo ";
			str+=sprop.rangeTo;
		}
		if(sprop.initElVal!=""){
			ifExtra(str, extra);
			str+="initElVal \\\"";
			str+=sprop.initElVal;
			str+="\\\"";
		}
		if(sprop.incElVal!=""){
			ifExtra(str, extra);
			str+="incElVal \\\"";
			str+=sprop.incElVal;
			str+="\\\"";
		}
		if(extra){
			str+="\n}";
		}
		str+="\", ";
		bool firstTime=true;

		if(sprop.isReadable()==false){
			if(firstTime==false){
				str+=" | ";
			}
			firstTime=false;
			str+="UNREADABLE";
		}
		if(sprop.isWritable()==false){
			if(firstTime==false){
				str+=" | ";
			}
			firstTime=false;
			str+="UNWRITEABLE";
		}
		if(sprop.flags & DO_NOT_OUTPUT){
			if(firstTime==false){
				str+=" | ";
			}
			firstTime=false;
			str+="DO_NOT_OUTPUT";
		}
		if(sprop.flags & DO_NOT_SAVE){
			if(firstTime==false){
				str+=" | ";
			}
			firstTime=false;
			str+="DO_NOT_SAVE";
		}
		if(sprop.flags & BLOCK_USE){
			if(firstTime==false){
				str+=" | ";
			}
			firstTime=false;
			str+="BLOCK_USE";
		}
		if(sprop.flags & BLOCK_NEW){
			if(firstTime==false){
				str+=" | ";
			}
			firstTime=false;
			str+="BLOCK_NEW";
		}
		if(sprop.flags & SAME_AS_LAST){
			if(firstTime==false){
				str+=" | ";
			}
			firstTime=false;
			str+="SAME_AS_LAST";
		}
		if(sprop.flags & PUSH){
			if(firstTime==false){
				str+=" | ";
			}
			firstTime=false;
			str+="PUSH";
		}
		if(sprop.flags & DISABLE){
			if(firstTime==false){
				str+=" | ";
			}
			firstTime=false;
			str+="DISABLE";
		}
		if(sprop.flags & DISABLE_LEN){
			if(firstTime==false){
				str+=" | ";
			}
			firstTime=false;
			str+="DISABLE_LEN";
		}
		if(sprop.flags & AXIS3F){
			if(firstTime==false){
				str+=" | ";
			}
			firstTime=false;
			str+="AXIS3F";
		}
		if(firstTime){
			str+=0;
		}
		if(sprop.promptNode.length){
			g_protoInterfaceNodePromptStr+="static const int promptNode_";
			g_protoInterfaceNodePromptStr+=node_field;
			g_protoInterfaceNodePromptStr+="[] = {";
			String tmp;
			for(int i=0; i<sprop.promptNode.length; i++){
				if(sprop.promptNode[i]==_GROUP){
					g_protoInterfaceNodePromptStr+="_GROUP";
				}
				else if(sprop.promptNode[i]==LOD_TYPE){
					g_protoInterfaceNodePromptStr+="LOD_TYPE";
				}
				else{
					tmp=nodeLookUpStr[sprop.promptNode[i]];
					g_protoInterfaceNodePromptStr+=tmp.toUpperCase();
				}
				if(i<sprop.promptNode.length-1){
					g_protoInterfaceNodePromptStr+=", ";
				}
			}
			g_protoInterfaceNodePromptStr+="};\n";

			str+=",\tINIT_PROMPT_NODE(";
			str+="promptNode_";
			str+=node_field;
			str+=")";
		}
		if(sprop.tip){
			g_protoInterfaceTips+="const char* ";
			g_protoInterfaceTips+="tip_";
			g_protoInterfaceTips+=node_field;
			g_protoInterfaceTips+=" = \"";
			g_protoInterfaceTips+=sprop.tip;
			g_protoInterfaceTips+="\";\n";

			str+=",\ttip_";
			str+=node_field;
		}
		str+=')';
		if(i < getFLen()-1){
			str+=',';
		}
		str+='\n';

/*
		str+=nodeName;
		str+="->properties+=Properties(\"";
		str+=sprop.name;
		str+="\", ";
		str+=sprop.groupIndex;
		str+=", ";
		str+=sprop.startAtIndex;
		str+=", ";
		str+=sprop.flags;
		str+=", ";
		str+=sprop.initElLen;
		str+=", ";
		str+=sprop.incElLen;
		str+=", ";
		str+=sprop.rangeFrom;
		str+=", ";
		str+=sprop.rangeTo;
		if(sprop.tip){
			str+=", \"";
			str+=sprop.tip;
			str+="\"";
		}
		str+=");\n";

		Field* field=getField(i);

		str+="(*";
		str+=nodeName;
		str+=")+= new ";
		field->toProtoCppStr(str);
		str+=";\n";
*/
	}
}
void Node::toCPPString(String& str, bool newList){
	M_Vec3f storeCoord;
	M_Color3f storeColor;
	M_Vec2f storeTexCoord;

	if(getType()==COLOREFFECT){
		M_Color3f& color=dynamic_cast<ColorEffect*>(this)->color;
		storeColor = color;
		color.length = 0;
	}
	else if(getType()==TEXTUREEFFECT){
		M_Vec2f& texCoord=dynamic_cast<TextureEffect*>(this)->texCoord.refValue();
		storeTexCoord=texCoord;
		texCoord.length=0;
	}

	int i = 0;
	String nodeName;
	if(name!=""){
		nodeName=name;
	}
	else {
		nodeName+='n';
		makeNodeNameFromAddr(nodeName, this);
	}
	if(setDEFgetUSE( str, newList)){
		return;
	}
	const M_Properties &properties=getProperties();

	str+=getTypeStr();
	str+="* ";
	str+=nodeName;
	str+=" = new ";
	str+=getTypeStr();
	str+=";\n";

	if(dynamic_cast<Art*>(this)!=0){
		Art* part=dynamic_cast<Art*>(this);
		part->normal.refValue().length=0;
//		extraPartToString(str, part);
		Seamless* smls=part->smls;
		if(dynamic_cast<ImmediatePart*>(part)==0&&smls->build.length){
			storeCoord=part->coord;
			part->coord.refValue().length=0;
		}
	}
	String nodeFieldName;
	for(i=0; i < getFLen(); i++){
		Field* field=getField(i);
		const Properties& sprop=getProp(i);
		if(	(sprop.isReadable()^sprop.isWritable())==false){

			Field* protoField=g_lookupProtoInterface[getType()]->getField(i);

			nodeFieldName=nodeName;
			nodeFieldName+="->";
			nodeFieldName+=properties[i].name;
			String tmpStr;
			field->toCPPString(tmpStr, nodeFieldName, protoField, newList);
			if(tmpStr!=""){
//				str+=nodeName;
//				str+="->";
//				str+=properties[i].name;
				str+=tmpStr;
				str+=";\n";
			}
		}
	}
	if(getType()==COLOREFFECT){
		dynamic_cast<ColorEffect*>(this)->color.refValue() = storeColor;
	}
	else if(getType()==TEXTUREEFFECT){
		dynamic_cast<TextureEffect*>(this)->texCoord.refValue() = storeTexCoord;
	}
	else if(dynamic_cast<Art*>(this)!=0){
		Art* part=dynamic_cast<Art*>(this);
		Seamless* smls=part->smls;
		if(dynamic_cast<ImmediatePart*>(part)==0&&smls->build.length){
			part->coord.refValue()=storeCoord;
		}
	}
}


void staticTransformParts(Art* part, Matrix matrix){
/*
	{
		M_Morpher& m_morpher=part->morpher.refValue();
		for(int k=m_morpher.length; --k>=0;){
			M_Vec3f& displacement=m_morpher[k]->displacement;
			for( int i=displacement.length; --i>=0; ){
				displacement[i] *= matrix;
			}
		}
	}
	{
	*/
	if(dynamic_cast<ImmediatePart*>(part)==0){
		M_Vec3f& coord=part->coord.refValue();

		for( int i=coord.length; --i>=0; ){
			coord[i] *= matrix;
		}
	}
	M_SArt& children=part->part.refValue();
	for( int k=children.length; --k>=0; ){
		staticTransformParts( children[k], matrix);
	}
}
void staticTransformDecendentParts(Art* part, Matrix matrix){
	M_SArt& m_part = part->part.refValue();
	for(int k = 0; k < m_part.length; k++){
		staticTransformParts(m_part[k], matrix);
	}
}
void staticTransformPartsNoDecendents(Art* part, Matrix matrix){

	if(dynamic_cast<ImmediatePart*>(part)==0){
		M_Vec3f& coord=part->coord.refValue();

		for( int i=coord.length; --i>=0; ){
			coord[i] *= matrix;
		}
	}
}


int g_projectDirection=PROJECT_FORWARD;

inline float interpolate(float par1, float par2, float pan){
	/*
	if(pan>1.0001){
		pan=pan;
	}
	if(pan<0.0001){
		pan=pan;
	}
*/
	return par1*(1.0f-pan)+par2*pan;
}
void morphToSawnEndY(M_Vec3f &dest, M_Vec3f &nonFlat, int length, float peak, Matrix &matrix, Matrix &imatrix, float tempTranslate ){
	for(int i=0;i<length;i++){
		Vec3f v=imatrix.multVecMatrix(dest[i]);
//		Vec3f nf=Rotation(-1,0,0,PI*.5).multVec(nonFlat[i]);
//		Vec3f v=dest[i];
		Vec3f nf=nonFlat[i];

		v.y=interpolate(v.y, nf.y, fabs((nf.y-tempTranslate)/peak));
		dest[i]=matrix.multVecMatrix(v);
//		dest[i]=v;
	}
}

void morphToSawnEnd(M_Vec3f &dest, M_Vec3f &nonFlat, int length, float peak, Matrix &matrix, Matrix &imatrix){
	for(int i=0;i<length;i++){
		Vec3f v=imatrix.multVecMatrix(dest[i]);
//		Vec3f nf=Rotation(-1,0,0,PI*.5).multVec(nonFlat[i]);
//		Vec3f v=dest[i];
		Vec3f nf=nonFlat[i];

		v.z=interpolate(v.z, nf.z, fabs(nf.z/peak));
		dest[i]=matrix.multVecMatrix(v);
//		dest[i]=v;
	}
}

void BranchJoinNegY::genJoinCurve(){
	static M_Vec3f oneCoord=Vec3f();
	static M_Vec3f project;
	static M_Vec3f joinNonFlat;
	int i;

	Rotation orientation=this->orientation;
	Rotation ibrot=-orientation;
	Vec3f location=this->location;

	Art* prt=part.refValue()[0];
	M_Vec3f& coord=prt->coord.refValue();

	Vec3f &svec=oneCoord[0];

	project=coord;
	joinNonFlat=project;

	Matrix matrix=rotationMatrix(orientation)*translationMatrix(location);
	Matrix imatrix=translationMatrix(-location)*rotationMatrix(ibrot);


	int generatorLength = surface->generator.length;

	for(i=fromCoord;i<toCoord;i++){
		Vec3f& destVec=coord[i];
		Vec3f &projVec=project[i];
		int maxLoop=0x10000;
		do{
			svec=destVec;
			//parent->genShape1(temp);

			genSurface(generatorLength, oneCoord);

			svec=imatrix.multVecMatrix(svec);
			float xDiff=svec.x-projVec.x;
			destVec.x-=xDiff*.5;
			float zDiff=svec.z-projVec.z;
			destVec.z-=zDiff*.5;
			if(fabs(xDiff)<.0001&&fabs(zDiff)<.0001)break;

		}while(--maxLoop);
		if(!maxLoop)
			return;

		svec=destVec;
		genSurface(generatorLength, oneCoord);
		coord[i]=svec;
	}
	for( i=coord.length; --i>=0; ){
		if(i<fromCoord||i>=toCoord){
			coord[i]=matrix*coord[i];
		}
	}

	if( sawn){
		for( i=fromCoord;i<toCoord;i++){
			Vec3f v=imatrix.multVecMatrix(coord[i]);
			Vec3f nf=joinNonFlat[i];

			v.y=interpolate(v.y, nf.y, fabs((nf.y-tempTranslate)/peak));
			coord[i]=matrix.multVecMatrix(v);
		}
	}
	else if(decay > 0){
		for(i = fromCoord; i < toCoord; i++){
			Vec3f nf = joinNonFlat[i];
			Vec3f ivec = imatrix * coord[i];

			if(nf.y < decay){
				float sineCurve = 1.0 - sin((abs(nf.y) / decay) * PI *.5);
				nf += Vec3f(0, ivec.y * sineCurve, 0);
			}
			coord[i] = matrix * nf;
		}
	}
	else{
		for( i=fromCoord;i<toCoord;i++){
			Vec3f v = coord[i];
			v += orientation * Vec3f(0,joinNonFlat[i].y,0);
			coord[i] = v;
		}
	}

	if(descendants){
		M_SArt& children=prt->part.refValue();
		for(int k=children.length; --k>=0;){
			staticTransformParts( children[k], matrix);
		}
	}
}

float findPeakPoint(M_Vec3f& point, int length, float& peak){
	peak=0;
	for(int i=0;i<length;i++){
		if(fabs(point[i].y)>fabs(peak)){
			peak=point[i].y;
		}

	}
	return peak;
}
inline void ECLathe::inverseSurface(M_Node& generator, BranchJoinNegY& branchJoinNegY){
	int i,j;
	static ECLathePtr inverse=new ECLathe();


	M_Vec2f& inversePoint=inverse->point;
	M_Float& inverseTangent=inverse->tangent;

	inverseTangent.length=inversePoint.length=point.length;


	for( i=0,j=point.length-1; i<point.length;i++,j--){
		inversePoint[i]=point[j];
		inverseTangent[i]=tangent[j]+PI;
	}
	inverse->processFields();


	ECLathePtr store=this;	//store pointer is only for stopping this from getting deleted in the next line
	generator[0]=inverse;
	branchJoinNegY.genJoinCurve();
	generator[0]=this;
}
inline void ECSyncBend::inverseSurface(M_Node& generator, BranchJoinNegY& branchJoinNegY){
	int i,j;
	static ECSyncBendPtr inverse=new ECSyncBend();


	M_Vec2f& inversePoint=inverse->point;
	M_Float& inverseTangent=inverse->tangent;

	inverseTangent.length=inversePoint.length=point.length;


	for( i=0,j=point.length-1; i<point.length;i++,j--){
		inversePoint[i]=point[j];
		inverseTangent[i]=tangent[j]+PI;
	}
	inverse->processFields();


	ECSyncBendPtr store=this;	//store pointer is only for stopping this from getting deleted in the next line
	generator[0]=inverse;
	branchJoinNegY.genJoinCurve();
	generator[0]=this;
}
void CCLathe::inverseSurface(M_Node& generator, BranchJoinNegY& branchJoinNegY){
	int i,j;
	static CCLathePtr inverse=new CCLathe();

	inverse->curve.refValue()=curve;
	inverse->distance.refValue()=distance;
	inverse->autoClose.refValue()=autoClose;

	inverse->curve.refValue()[0]=getCurve()-PI;
	inverse->startPosY=getPosY(curve.length-1);
	inverse->baseRad=getPosX(curve.length-1);

	for( i=1,j=curve.length-1; i<curve.length;i++,j--){
		inverse->curve.refValue()[i]=-curve[j];
	}
	for( i=0,j=distance.length-1; i<distance.length;i++,j--){
		inverse->distance.refValue()[i]=distance[j];
	}

	CCLathePtr store=this;	//store pointer is only for stopping this from getting deleted in the next line
	generator[0]=inverse;
	branchJoinNegY.genJoinCurve();
	generator[0]=this;
}

void BranchJoinNegY::doBuild(){
	int i;
	int len = numOf;
	if(!part.length||surface==0||len==0){
		return;
	}
	Art* prt=part.refValue()[0];
	M_Vec3f& coord=prt->coord.refValue();
	if(coord.length==0)return;
	M_Node& generator=surface->generator.refValue();
	if( &generator==NULL || !generator.length){
		Matrix mat=rotationMatrix(orientation) * translationMatrix(location);
		for(i=coord.length; --i>=0; ){
			coord[i]*=mat;
		}
		return;
	}
	if(static_cast<unsigned>(len)>static_cast<unsigned>(coord.length)){
		len=coord.length;
	}
	if(backward){
		toCoord=coord.length;
		fromCoord=toCoord-len;
		tempTranslate=coord[toCoord-1].y;
	}
	else {
		fromCoord=0;
		toCoord=len;
		tempTranslate=coord[fromCoord].y;
	}
	peak=fabs(coord[fromCoord].y-coord[toCoord-1].y);

	g_projectDirection=PROJECT_DOWN;

	ALathe* lathe=dynamic_cast<ALathe*>(&*generator[0]);
	if(lathe){
		lathe->inverseSurface(generator, *this);
	}
}
void BranchJoinPosY::genJoinCurve(){
	static M_Vec3f oneCoord=Vec3f();
	static M_Vec3f project;
	static M_Vec3f joinNonFlat;
	int i;

	Rotation orientation=this->orientation;
	Rotation ibrot=-orientation;
	Vec3f location=this->location;

	Art* prt=part.refValue()[0];
	M_Vec3f &coord=prt->coord.refValue();

	Vec3f &svec=oneCoord[0];

	project=coord;
	joinNonFlat=project;

	Matrix matrix=rotationMatrix(orientation)*translationMatrix(location);
	Matrix imatrix=translationMatrix(-location)*rotationMatrix(ibrot);


	int generatorLength = surface->generator.length;

	for(i=fromCoord;i<toCoord;i++){
		Vec3f &destVec=coord[i];
		Vec3f &projVec=project[i];
		int maxLoop=0x10000;
		do{
			svec=destVec;
			//parent->genShape1(oneCoord);

			genSurface(generatorLength, oneCoord);

			svec=imatrix.multVecMatrix(svec);
			float xDiff=svec.x-projVec.x;
			destVec.x-=xDiff*.5;
			float zDiff=svec.z-projVec.z;
			destVec.z-=zDiff*.5;
			if(fabs(xDiff)<.0001&&fabs(zDiff)<.0001)break;

		}while(--maxLoop);
		if(!maxLoop)
			return;

		svec=destVec;
		genSurface(generatorLength, oneCoord);
		coord[i]=svec;
	}
	for( i=coord.length; --i>=0; ){
		if(i<fromCoord||i>=toCoord){
			coord[i]=matrix*coord[i];
		}
	}

	if( sawn){
		for(i=fromCoord;i<toCoord;i++){
			Vec3f v=imatrix * coord[i];
			Vec3f nf=joinNonFlat[i];
			v.y=interpolate(v.y, nf.y, fabs((nf.y-tempTranslate)/peak));
			coord[i]=matrix * v;
		}
	}
	else if(decay > 0){
		for(i=fromCoord;i<toCoord;i++){
			Vec3f nf=joinNonFlat[i];
			Vec3f ivec=imatrix*coord[i];
			if(nf.y > -decay){
				float height = ivec.y;
				//float sineCurve=cos((abs(nf.y)/decay)*PI)*.5+.5;
				//float sineCurve=cos((abs(nf.y)/decay)*PI*.5);
				//float sineCurve=1.0-cos((abs(nf.y)/decay)*PI*.5-PI*.5);
				float sineCurve=1.0-sin((abs(nf.y)/decay)*PI*.5);
				//nf.y+=height*sineCurve;

				nf += Vec3f(0, height*sineCurve, 0);
			}
			coord[i]=matrix*nf;
		}
	}
	else{
		for( i=fromCoord;i<toCoord;i++){
			Vec3f v = coord[i];
			v += orientation * Vec3f(0,joinNonFlat[i].y,0);
			coord[i] = v;
		}
	}
	if(descendants){
		M_SArt& children=prt->part.refValue();
		for(int k=children.length; --k>=0;){
			staticTransformParts( children[k], matrix);
		}
	}
}

void BranchJoinPosY::doBuild(){
	int i;
	int len=numOf;
	if(!part.length||surface==0||len==0){
		return;
	}
	Art* prt=part.refValue()[0];
	M_Vec3f& coord=prt->coord.refValue();
	if(coord.length==0)return;
	M_Node& generator=surface->generator.refValue();
	if( &generator==NULL || !generator.length){
		Matrix mat=rotationMatrix(orientation) * translationMatrix(location);
		for(i=coord.length; --i>=0; ){
			coord[i]*=mat;
		}
		return;
	}
	if(static_cast<unsigned>(len)>static_cast<unsigned>(coord.length)){
		len=coord.length;
	}
	if(backward){
		toCoord=coord.length;
		fromCoord=toCoord-len;
		tempTranslate=coord[toCoord-1].y;
	}
	else {
		fromCoord=0;
		toCoord=len;
		tempTranslate=coord[fromCoord].y;
	}
	peak=fabs(coord[fromCoord].y-coord[toCoord-1].y);

	g_projectDirection=PROJECT_UP;
	genJoinCurve();
}

void BranchJoinPosX::genJoinCurve(){
	static M_Vec3f oneCoord=Vec3f();
	static M_Vec3f project;
	static M_Vec3f joinNonFlat;
	int i;

	Rotation orientation=this->orientation;
	Rotation ibrot=-orientation;
	Vec3f location=this->location;

	Art* prt=part.refValue()[0];
	M_Vec3f &coord=prt->coord.refValue();

	Vec3f &svec=oneCoord[0];

	project=coord;
	joinNonFlat=project;

	Matrix matrix=rotationMatrix(orientation)*translationMatrix(location);
	Matrix imatrix=translationMatrix(-location)*rotationMatrix(ibrot);

	int generatorLength=surface->generator.length;

	for(i=fromCoord;i<toCoord;i++){
		Vec3f &destVec=coord[i];
		Vec3f &projVec=project[i];
		int maxLoop=0x10000;
		do{
			svec=destVec;

			genSurface(generatorLength, oneCoord);

			svec=imatrix.multVecMatrix(svec);
			float zDiff=svec.z-projVec.z;
			destVec.x-=zDiff;
			float yDiff=svec.y-projVec.y;
			destVec.y-=yDiff;
			if(fabs(zDiff)<.0001&&fabs(yDiff)<.0001)break;

		}while(--maxLoop);
		if(!maxLoop)return;

		svec=destVec;
		genSurface(generatorLength, oneCoord);
		coord[i]=svec;
	}
	for( i=coord.length; --i>=0; ){
		if(i<fromCoord||i>=toCoord){
			coord[i]=matrix*coord[i];
		}
	}

	if(sawn){
		for(i = fromCoord;i<toCoord;i++){
			Vec3f v=imatrix.multVecMatrix(coord[i]);
			Vec3f nf=joinNonFlat[i];

			v.x=interpolate(v.x, nf.x, fabs((nf.x-tempTranslate)/peak));
			coord[i]=matrix.multVecMatrix(v);
		}
	}
	else if(decay > 0){
		for(i = fromCoord; i < toCoord; i++){
			Vec3f nf = joinNonFlat[i];
			Vec3f ivec = imatrix * coord[i];

			if(nf.x > -decay){
				float sineCurve = 1.0 - sin((abs(nf.x) / decay) * PI *.5);
				nf += Vec3f(ivec.x * sineCurve, 0, 0);
			}
			coord[i] = matrix * nf;
		}
	}
	else{
		for(i = fromCoord;i<toCoord;i++){
			Vec3f v = coord[i];
			v += orientation * Vec3f(joinNonFlat[i].x,0,0);
			coord[i] = v;
		}
	}
	if(descendants){
		M_SArt& children=prt->part.refValue();
		for(int k=children.length; --k>=0;){
			staticTransformParts( children[k], matrix);
		}
	}
}
void BranchJoinPosX::doBuild(){
	int i;
	int len=numOf;
	if(!part.length||surface==0||len==0){
		return;
	}
	Art* prt=part.refValue()[0];
	M_Vec3f& coord=prt->coord.refValue();
	if(coord.length==0)return;
	M_Node& generator=surface->generator.refValue();
	if( &generator==NULL || !generator.length){
		Matrix mat=rotationMatrix(orientation) * translationMatrix(location);
		for(i=coord.length; --i>=0; ){
			coord[i]*=mat;
		}
		return;
	}
	if(static_cast<unsigned>(len)>static_cast<unsigned>(coord.length)){
		len=coord.length;
	}
	if(backward){
		toCoord=coord.length;
		fromCoord=toCoord-len;
		tempTranslate=coord[toCoord-1].x;
	}
	else {
		fromCoord=0;
		toCoord=len;
		tempTranslate=coord[fromCoord].x;
	}
	peak=fabs(coord[fromCoord].x-coord[toCoord-1].x);

	g_projectDirection=PROJECT_FORWARD;
	genJoinCurve();
}
void BranchJoinNegX::genJoinCurve(){
	static M_Vec3f oneCoord=Vec3f();
	static M_Vec3f project;
	static M_Vec3f joinNonFlat;
	int i;

	Rotation orientation=this->orientation;
	Rotation ibrot=-orientation;
	Vec3f location=this->location;

	Art* prt=part.refValue()[0];
	M_Vec3f &coord=prt->coord.refValue();

	Vec3f &svec=oneCoord[0];

	project=coord;
	joinNonFlat=project;

	Matrix matrix=rotationMatrix(orientation)*translationMatrix(location);
	Matrix imatrix=translationMatrix(-location)*rotationMatrix(ibrot);


	int generatorLength=surface->generator.length;

	for(i=fromCoord;i<toCoord;i++){
		Vec3f &destVec=coord[i];
		Vec3f &projVec=project[i];
		int maxLoop=0x10000;
		do{
			svec=destVec;

			genSurface(generatorLength, oneCoord);

			svec=imatrix.multVecMatrix(svec);
			float zDiff=svec.z-projVec.z;
			destVec.x+=zDiff;
			float yDiff=svec.y-projVec.y;
			destVec.y-=yDiff;
			if(fabs(zDiff)<.0001&&fabs(yDiff)<.0001)break;

		}while(--maxLoop);
		if(!maxLoop)return;

		svec=destVec;
		genSurface(generatorLength, oneCoord);
		coord[i]=svec;
	}
	for(i=coord.length; --i>=0; ){
		if(i<fromCoord||i>=toCoord){
			coord[i]=matrix*coord[i];
		}
	}

	if(sawn){
		for(i = fromCoord; i < toCoord; i++){
			Vec3f v = imatrix * coord[i];
			Vec3f nf = joinNonFlat[i];

			v.x=interpolate(v.x, nf.x, fabs((nf.x-tempTranslate)/peak));
			coord[i] = matrix * v;
		}
	}
	else if(decay > 0){
		for(i = fromCoord; i < toCoord; i++){
			Vec3f nf = joinNonFlat[i];
			Vec3f ivec = imatrix * coord[i];

			if(nf.x < decay){
				float sineCurve = 1.0 - sin((abs(nf.x) / decay) * PI *.5);
				nf += Vec3f(ivec.x * sineCurve, 0, 0);
			}
			coord[i] = matrix * nf;
		}
	}
	else{
		for(i = fromCoord;i < toCoord;i++){
			Vec3f v = coord[i];
			v += orientation * Vec3f(joinNonFlat[i].x,0,0);
			coord[i] = v;
		}
	}
	if(descendants){
		M_SArt& children=prt->part.refValue();
		for(int k = children.length; --k >= 0;){
			staticTransformParts( children[k], matrix);
		}
	}
}

void BranchJoinNegX::doBuild(){
	int i;
	int len=numOf;
	if(!part.length||surface==0||len==0){
		return;
	}
	Art* prt=part.refValue()[0];
	M_Vec3f& coord=prt->coord.refValue();
	if(coord.length==0)return;
	M_Node& generator=surface->generator.refValue();
	if( &generator==NULL || !generator.length){
		Matrix mat=rotationMatrix(orientation) * translationMatrix(location);
		for(i=coord.length; --i>=0; ){
			coord[i]*=mat;
		}
		return;
	}
	if(static_cast<unsigned>(len)>static_cast<unsigned>(coord.length)){
		len=coord.length;
	}
	if(backward){
		toCoord=coord.length;
		fromCoord=toCoord-len;
		tempTranslate=coord[toCoord-1].x;
	}
	else {
		fromCoord=0;
		toCoord=len;
		tempTranslate=coord[fromCoord].x;
	}
	peak=fabs(coord[fromCoord].x-coord[toCoord-1].x);

	g_projectDirection=PROJECT_FORWARD;
	genJoinCurve();
}

void BranchJoinNegZ::genJoinCurve(){
	static M_Vec3f oneCoord=Vec3f();
	static M_Vec3f project;
	static M_Vec3f joinNonFlat;
	int i;

	Rotation orientation=this->orientation;
	Rotation ibrot=-orientation;
	Vec3f location=this->location;

	Art* prt=part.refValue()[0];
	M_Vec3f &coord=prt->coord.refValue();

	Vec3f &svec=oneCoord[0];

	project=coord;
	joinNonFlat=project;

	Matrix matrix=rotationMatrix(orientation)*translationMatrix(location);
	Matrix imatrix=translationMatrix(-location)*rotationMatrix(ibrot);


	int generatorLength=surface->generator.length;

	for(i=fromCoord;i<toCoord;i++){
		Vec3f &destVec=coord[i];
		Vec3f &projVec=project[i];
		int maxLoop=0x10000;
		do{
			svec=destVec;

			genSurface(generatorLength, oneCoord);

			svec=imatrix.multVecMatrix(svec);
			float xDiff=svec.x-projVec.x;
			destVec.x-=xDiff;
			float yDiff=svec.y-projVec.y;
			destVec.y-=yDiff;
			if(fabs(xDiff)<.0001&&fabs(yDiff)<.0001)break;

		}while(--maxLoop);
		if(!maxLoop)return;

		svec=destVec;
		genSurface(generatorLength, oneCoord);
		coord[i]=svec;
	}
	for(i = coord.length; --i>=0; ){
		if(i<fromCoord||i>=toCoord){
			coord[i]=matrix*coord[i];
		}
	}

	if(sawn){
		for( i=fromCoord;i<toCoord;i++){
			Vec3f v=imatrix * coord[i];
			Vec3f nf=joinNonFlat[i];

			v.z=interpolate(v.z, nf.z, fabs((nf.z-tempTranslate)/peak));
			coord[i]=matrix * v;
		}
	}
	else if(decay > 0){
		for(i = fromCoord; i < toCoord; i++){
			Vec3f nf = joinNonFlat[i];
			Vec3f ivec = imatrix * coord[i];

			if(nf.z < decay){
				float sineCurve = 1.0 - sin((abs(nf.z) / decay) * PI *.5);
				nf += Vec3f(0, 0, ivec.z * sineCurve);
			}
			coord[i] = matrix * nf;
		}
	}
	else{
		for( i=fromCoord;i<toCoord;i++){
			Vec3f v = coord[i];
			v += orientation * Vec3f(0,0,joinNonFlat[i].z);
			coord[i] = v;
		}
	}
	if(descendants){
		M_SArt& children=prt->part.refValue();
		for(int k=children.length; --k>=0;){
			staticTransformParts( children[k], matrix);
		}
	}
}
void BranchJoinNegZ::doBuild(){
	int i;
	int len=numOf;
	if(!part.length||surface==0||len==0){
		return;
	}
	Art* prt=part.refValue()[0];
	M_Vec3f& coord=prt->coord.refValue();
	if(coord.length==0)return;
	M_Node& generator=surface->generator.refValue();
	if( &generator==NULL || !generator.length){
		Matrix mat=rotationMatrix(orientation) * translationMatrix(location);
		for(i=coord.length; --i>=0; ){
			coord[i]*=mat;
		}
		return;
	}
	if(static_cast<unsigned>(len)>static_cast<unsigned>(coord.length)){
		len=coord.length;
	}
	if(backward){
		toCoord=coord.length;
		fromCoord=toCoord-len;
		tempTranslate=coord[toCoord-1].z;
	}
	else {
		fromCoord=0;
		toCoord=len;
		tempTranslate=coord[fromCoord].z;
	}

	peak=fabs(coord[fromCoord].z-coord[toCoord-1].z);
	g_projectDirection=PROJECT_FORWARD;
	genJoinCurve();
}

void BranchJoinPosZ::genJoinCurve(){
	static M_Vec3f oneCoord=Vec3f();
	static M_Vec3f project;
	static M_Vec3f joinNonFlat;
	int i;

	Rotation orientation=this->orientation;
	Rotation ibrot=-orientation;
	Vec3f location=this->location;

	Art* prt=part.refValue()[0];
	M_Vec3f &coord=prt->coord.refValue();

	Vec3f &svec=oneCoord[0];

	project=coord;
	joinNonFlat=project;

	Matrix matrix=rotationMatrix(orientation)*translationMatrix(location);
	Matrix imatrix=translationMatrix(-location)*rotationMatrix(ibrot);


	int generatorLength=surface->generator.length;

	for(i=fromCoord;i<toCoord;i++){
		Vec3f &destVec=coord[i];
		Vec3f &projVec=project[i];
		int maxLoop=0x10000;
		do{
			svec=destVec;

			genSurface(generatorLength, oneCoord);

			svec=imatrix.multVecMatrix(svec);
			float xDiff=svec.x-projVec.x;
			destVec.x+=xDiff;
			float yDiff=svec.y-projVec.y;
			destVec.y-=yDiff;
			if(fabs(xDiff)<.0001&&fabs(yDiff)<.0001)break;

		}while(--maxLoop);
		if(!maxLoop)return;

		svec=destVec;
		genSurface(generatorLength, oneCoord);
		coord[i]=svec;
	}
	for( i=coord.length; --i>=0; ){
		if(i<fromCoord||i>=toCoord){
			coord[i]=matrix*coord[i];
		}
	}
	if(sawn){
		for( i=fromCoord;i<toCoord;i++){
			Vec3f v=imatrix * coord[i];
			Vec3f nf=joinNonFlat[i];

			v.z=interpolate(v.z, nf.z, fabs((nf.z-tempTranslate)/peak));
			coord[i]=matrix * v;
		}
	}
	else if(decay > 0){
		for(i = fromCoord; i < toCoord; i++){
			Vec3f nf = joinNonFlat[i];
			Vec3f ivec = imatrix * coord[i];

			if(nf.z > -decay){
				float sineCurve = 1.0 - sin((abs(nf.z) / decay) * PI *.5);
				nf += Vec3f(0, 0, ivec.z * sineCurve);
			}
			coord[i] = matrix * nf;
		}
	}
	else{
		for( i=fromCoord;i<toCoord;i++){
			Vec3f v = coord[i];
			v += orientation * Vec3f(0,0,joinNonFlat[i].z);
			coord[i] = v;
		}
	}
	if(descendants){
		M_SArt& children=prt->part.refValue();
		for(int k=children.length; --k>=0;){
			staticTransformParts( children[k], matrix);
		}
	}
}

void BranchJoinPosZ::doBuild(){
	int i;
	int len=numOf;
	if(!part.length||surface==0||len==0){
		return;
	}
	Art* prt=part.refValue()[0];
	M_Vec3f& coord=prt->coord.refValue();
	if(coord.length==0)return;
	M_Node& generator=surface->generator.refValue();
	if( &generator==NULL || !generator.length){
		Matrix mat=rotationMatrix(orientation) * translationMatrix(location);
		for(i=coord.length; --i>=0; ){
			coord[i]*=mat;
		}
		return;
	}
	if(static_cast<unsigned>(len)>static_cast<unsigned>(coord.length)){
		len=coord.length;
	}
	if(backward){
		toCoord=coord.length;
		fromCoord=toCoord-len;
		tempTranslate=coord[toCoord-1].z;
	}
	else {
		fromCoord=0;
		toCoord=len;
		tempTranslate=coord[fromCoord].z;
	}
	peak=fabs(coord[fromCoord].z-coord[toCoord-1].z);

	g_projectDirection=PROJECT_FORWARD;
	genJoinCurve();
}
inline bool Art::checkPointInTriangleForProjectOntoTri(const Vec3f& projectVertex, const Matrix& matrix, const int& triIdx)const{
	Art* part[3] = {
		triangle.refPart(triIdx),
		triangle.refPart(triIdx+1),
		triangle.refPart(triIdx+2)
	};

	int crdIdx[3] = {
		triangle.refIndex(triIdx),
		triangle.refIndex(triIdx+1),
		triangle.refIndex(triIdx+2)
	};

	Vec3f sr[3] = {
		matrix * part[0]->coord.refValue()[crdIdx[0]],
		matrix * part[1]->coord.refValue()[crdIdx[1]],
		matrix * part[2]->coord.refValue()[crdIdx[2]]
	};

	Vec2f v[3];
	Vec2f mv;

	v[0].x=sr[0].x;
	v[0].y=-sr[0].z;

	v[1].x=sr[1].x;
	v[1].y=-sr[1].z;

	v[2].x=sr[2].x;
	v[2].y=-sr[2].z;

	mv.x=projectVertex.x;
	mv.y=-projectVertex.z;


	if(isTriangleHit(mv, v)){
		return backFace(v);
	}
	return false;
}
bool getTriangleHitPos(Vec3f& clickPos, const Vec2f& pt, const Vec3f* v);

void swapYZ(Vec3f& v){
	float s = -v.y;
	v.y = -v.z;
	v.z = s;
}
void ProjectOntoTri::doBuild(){
	static M_Vec3f joinNonFlat;

	int len=numOf;
	Art* surface=dynamic_cast<Art*>(this->surface.refValue().p);
	if(!part.length||surface==0||len==0){
		return;
	}
	Art* prt=part.refValue()[0];
	M_Vec3f& coord=prt->coord.refValue();
	if(coord.length==0)return;

	joinNonFlat=coord;

	if(static_cast<unsigned>(len)>static_cast<unsigned>(coord.length)){
		len=coord.length;
	}
	int fromCoord=0;
	int toCoord=0;
	float tempTranslate=0;

	if(backward){
		toCoord=coord.length;
		fromCoord=toCoord-len;
		tempTranslate=coord[toCoord-1].y;
	}
	else {
		fromCoord=0;
		toCoord=len;
		tempTranslate=coord[fromCoord].y;
	}
	float peak=fabs(coord[fromCoord].y-coord[toCoord-1].y);

	M_Triangle& triangle=surface->triangle;

	Rotation orientation=this->orientation;
	Rotation ibrot=-orientation;
	Vec3f location=this->location;

	Matrix matrix=rotationMatrix(orientation)*translationMatrix(location);
	Matrix imatrix=translationMatrix(-location)*rotationMatrix(ibrot);

	if(Feedback* feedback=dynamic_cast<Feedback*>(this->feedback.p)){
		if(Art* source=dynamic_cast<Art*>(feedback->source.p)){
			const Vec2f multiply = feedback->multiply;

			M_Vec3f& src = source->coord.refValue();
			M_Vec3f& surfaceCoord = surface->coord.refValue();

			for(int i = fromCoord, j=0; i < toCoord; i++, j++){
				Vec3f diff = coord[i] - imatrix * surfaceCoord[j];
				src[j].x += diff.x * multiply.x;
				src[j].y -= diff.z * multiply.y;
				//float diffY = surfaceCoord[j].y - matrix * coord[i].y;
			}
		}
	}
	for(int i=fromCoord; i<toCoord; i++){
		const Vec3f projectVertex = coord[i];
		for(int triIdx=triangle.getTriangleLength(); (triIdx-=3)>=0; ){
			if(surface->checkPointInTriangleForProjectOntoTri(projectVertex, imatrix, triIdx)){

				int a = triIdx;
				int b = triIdx+1;
				int c = triIdx+2;

				int ia=triangle.refIndex(a);
				int ib=triangle.refIndex(b);
				int ic=triangle.refIndex(c);

				Art* pa=triangle.refPart(a);
				Art* pb=triangle.refPart(b);
				Art* pc=triangle.refPart(c);

				Vec3f vecA = imatrix * pa->coord.refValue()[ia];
				Vec3f vecB = imatrix * pb->coord.refValue()[ib];
				Vec3f vecC = imatrix * pc->coord.refValue()[ic];

				swapYZ(vecA);
				swapYZ(vecB);
				swapYZ(vecC);

				Vec3f v[3]={vecA, vecB, vecC};

				Vec3f clickPos;
				if(getTriangleHitPos(clickPos, Vec2f(projectVertex.x, -projectVertex.z), v)){
					coord[i]=matrix*Vec3f(projectVertex.x, -clickPos.z, projectVertex.z);
				}
			}
		}
	}
	for(int i=coord.length; --i>=0; ){
		if(i<fromCoord||i>=toCoord){
			coord[i]=matrix*coord[i];
		}
	}
	if(sawn){
		for(int i=fromCoord;i<toCoord;i++){
			Vec3f v=imatrix.multVecMatrix(coord[i]);
			Vec3f nf=joinNonFlat[i];

			v.y=interpolate(v.y, nf.y, fabs((nf.y-tempTranslate)/peak));
			coord[i]=matrix.multVecMatrix(v);
		}
	}
	else if(decay > 0){
		for(int i = fromCoord; i < toCoord; i++){
			Vec3f nf = joinNonFlat[i];
			Vec3f ivec = imatrix * coord[i];

			if(nf.y < decay){
				float sineCurve = 1.0 - sin((abs(nf.y) / decay) * PI *.5);
				nf += Vec3f(0, ivec.y * sineCurve, 0);
			}
			coord[i] = matrix * nf;
		}
	}
	else{
		for(int i=fromCoord;i<toCoord;i++){
			Vec3f v = coord[i];
			v += orientation * Vec3f(0,joinNonFlat[i].y,0);
			coord[i] = v;
		}
	}

	if(descendants){
		M_SArt& children=prt->part.refValue();
		for(int k=children.length; --k>=0;){
			staticTransformParts( children[k], matrix);
		}
	}
}

bool Browser::createSceneFromFile(const String filename){
	g_scene->wOpen.refValue().length = 0;
	g_scene->wPos.refValue().length = 0;
	g_scene->wSize.refValue().length = 0;
	g_colorOwnership = false;
	g_paletteMode = false;
	g_scene->version = 0;
	M_Node children;
	if(createVrmlFromFile(filename, children)){
		if(children.length){
			if(Scene* scene = dynamic_cast<Scene*>(children[0].p)){
				g_scene = scene;
				g_scene->mode.refValue() = g_scene->mode & ~(STITCH_POLY_MODE | HIDE_MESH_MODE);
			}
			else{
				g_scene->processFields();
			}
		}
		return true;
	}
	return false;
}

Scene* Browser::getWorld(){
	return g_scene;
}

void Browser::print(char * str){
	dLog(str);
}
void Browser::print(const String str){
	dLog(str.c_str());
}
void Browser::print(int i){
	dLog(String(i).c_str());
}
void loadInlines(M_Node& children){
	for(int k=0; k<children.length; k++){
		Progenitor* childrenNode=dynamic_cast<Progenitor*>(children[k].p);
		if(childrenNode){
			loadInlines(childrenNode->children);
		}
		else{
			Inline* inlineNode=dynamic_cast<Inline*>(children[k].p);
			if(inlineNode){
				M_String& url=inlineNode->url;
				M_Node& children=inlineNode->children;
				if(children.length==0){ //if it already contains nodes it must be a USED Inline
					for(int i=0; i<url.length; i++){
						if(Browser::createVrmlFromFile(url[i], children)){
							break;
						}
					}
				}
			}
		}
	}
}
void createSmlsFromString(M_Node& nodes, const String& str){
	g_bColor = Color3f();
	g_protoDeclare = 0;

	String& str2 = const_cast<String&>(str);
	str2.requestMoreIfNeeded(2);
	char* p = const_cast<char*>(str2.c_str());
	p[str2.length + 1] = ']'; //insert after end of str charecter

	int i = 0;
	getNodesFromStr(nodes, str2, i);

}
ScenePtr createSmlsFromString(const String& str){
	g_bColor = Color3f();
	g_protoDeclare = 0;

	String& str2 = const_cast<String&>(str);
	str2.requestMoreIfNeeded(2);
	char* p = const_cast<char*>(str2.c_str());
	p[str2.length + 1] = ']'; //insert after end of str charecter

	int i = 0;
	M_Node nodes;
	getNodesFromStr(nodes, str2, i);

	loadInlines(nodes);

	if(nodes.length){
		if(Scene* inScene=dynamic_cast<Scene*>(nodes[0].p)){
			return inScene;
		}
	}
	return g_load_scene;
}
bool decodeSmlsVrmlHeader(String& str, int& i){
	if(	str.substr(0,5)=="#SMLS"){
		int i=5;
		str.stb(i);
		if(str[i++]!='V'){
			return false;
		}
		g_scene->version = getFloat(str,i);
		g_load_scene=g_scene;
	}
	else if(	str.substr(0,15)!="#VRML V2.0 utf8"&&
				str.substr(0,14)!="#X3D V3.0 utf8"){
		return false;
	}
	i=str.indexOf("\n",0);
	return true;
}

bool Browser::createVrmlFromFile(const String filename, M_Node& m_node){
	m_node.length = 0;
	g_bColor = Color3f();
	g_protoList = M_ProtoDeclare();
	g_protoDeclare = NULL;
	String& str = g_loadStr;
	str = "";
	str = readFile(filename);

	if(str.length){
		int i=0;
		if(decodeSmlsVrmlHeader(str, i)==false){
			return false;
		}
		checkBackgroundColor(str, i);
		i--;
		if(i<str.length){
			str[i]='[';
			str+=']';
			m_node.getField(str, i);
			loadInlines(m_node);
			return true;
		}
	}
	dLog(filename + " is empty!");
	return false;
}
/*
void Browser::createSceneFromString(){
	String& str=g_loadStr;

	g_bColor=Color3f();
	g_scene->defList.length=0;
	g_protoList=M_ProtoDeclare();
	g_protoDeclare=NULL;
	M_Node& m_node=g_scene->children;

	if(str.length){

		initCompile();

		if(	str.substr(0,15)!="#VRML V2.0 utf8"&&
			str.substr(0,14)!="#X3D V3.0 utf8"){
			throw;
		}
		int i=str.indexOf("\n",0);

		checkBackgroundColor(str, i);
		i--;
		if(i<str.length){
			str[i]='[';
			str+=']';
			int ti=i+1;
			m_node.getField(str, i);
			loadInlines(m_node);
		}
		else{
			return;
		}
		if(m_node.length){
			if(Scene* scene=dynamic_cast<Scene*>(m_node[0].p)){
				g_scene=scene;
			}
			else{
				g_scene->processFields();
			}
		}
		decodeFunctions(g_loadStr);
	}
}
*/

Browser Browser;


float CCScale::getDistance(int index)const{
	const M_Float& distance=this->distance;
	float distanceSum=0;
	int i;
	for(i=0;i<=index;i++){
		distanceSum+=distance[i];
	}
	return distanceSum;
}

float CCBend::getDistance(int index)const{
	const M_Float& distance=this->distance;
	float distanceSum=0;
	int i;
	for(i=0;i<=index;i++){
		distanceSum+=distance[i];
	}
	return distanceSum;
}
float CCBend::getYangleSum(int index){
	const M_Float& yAngle=this->yAngle;
	float yAngleSum=0;
	int i;
	for(i=0;i<=index;i++){
		yAngleSum+=yAngle[i];
	}
	return yAngleSum;
}
void CCBend::generate(M_Vec3f& coord, const int h){

	M_Vec3f normal; //dummy
	coord.multMatrixLeft(imatrix);
	M_Float& curve=this->curve.refValue();

	int i;
	static M_Float yLookUp;
	yLookUp.length=coord.length;
	for(i=yLookUp.length; --i>=h;){
		yLookUp[i]=coord[i].y;
	}
	for(i=curve.length;--i>1;){
		bend( getDistance(i-2), getDistance(i-1), curve[i], getYangleSum(i-1),coord,yLookUp);
	}
	if(i==1){
		bend( 0, distance.refValue()[0], curve[1], yAngle.refValue()[0],coord,yLookUp);
		int height=location.refValue().y;
		float curve0=curve[0];
		for(i=coord.length; --i>=0;){
			if(yLookUp[i]>=height){
				coord[i]=Rotation(0,0,1,curve0)*coord[i];
			}
		}
	}
	coord.multMatrixLeft(matrix);
}
void CCPath::bend(Vec3f& position, bool& firstCalcDone, Rotation& accumCurveAngle, float beginHeight,float endHeight,float angle,float axis){
	Rotation axisRot=Rotation (0,1,0, axis);
	position = axisRot*position;
	float diff=endHeight-beginHeight;
	if(diff==0){
		position.y-=beginHeight;
		position = Rotation (0,0,1, angle)*position;
		position.y+=beginHeight;
		position = Rotation (0,-1,0, axis)*position;
		return;
	}
	Vec3f reletiveDis;
	Vec3f v;
	float rad=diff/angle;

	v=position;
	if(!firstCalcDone&&v.y>=beginHeight){
		firstCalcDone=true;
		reletiveDis.y=v.y-beginHeight;
		float stepAngle=reletiveDis.y/diff*angle;
		accumCurveAngle.angle=stepAngle;
		v.y=0;
		v.x+=rad;
		rotateVectorZ(v,stepAngle);
		v.y+=beginHeight;
		v.x-=rad;
		position=v;
	}
	else if(firstCalcDone){
		reletiveDis.y=v.y-beginHeight;
		v.y=(rad+v.x)*sin(angle)+beginHeight+	(position.y-endHeight)*(cos(angle));
		v.x=(rad+v.x)*cos(angle)-rad	+	(position.y-endHeight)*-(sin(angle));
		position=v;
		accumCurveAngle*=Rotation(0,-1,0,angle);
	}
	axisRot.y=-1;
	position = axisRot*position;
}
float CCPath::getDistance(int index)const{
	const M_Float& distance=this->distance;
	float distanceSum=0;
	int i;
	for(i=0;i<=index;i++){
		distanceSum+=distance[i];
	}
	return distanceSum;
}
float CCPath::getZangleSum(int index){
	float zAngleSum=0;
	int i;
	for(i=0;i<=index;i++){
		zAngleSum+=zAngle.refValue()[i];
	}
	return zAngleSum;
}

void CCPath::do_set_fraction(){
	Vec3f position;
	Rotation accumCurveAngle=Rotation(0,-1,0,0);

	position.y=getDistance(distance.length-1)*set_fraction;
	M_Float& curve=this->curve.refValue();

	int i=0;
	bool firstCalcDone=false;
	for(i=curve.length;--i>1;){
		bend( position, firstCalcDone, accumCurveAngle, getDistance(i-2), getDistance(i-1), curve[i], getZangleSum(i-1));
//		if(firstCalcDone){
//		}
	}
	if(i==1){
		bend( position, firstCalcDone, accumCurveAngle, 0, distance.refValue()[0], curve[1], zAngle.refValue()[0]);
		//ifs->rotate(Rotation(0,0,1,curve[0]));
	}
	position=Rotation(1,0,0,PI*.5)*position;
	position=orientation.refValue()*position;
	position+=location;

	accumCurveAngle*=orientation;

	rot_changed.sendOutRoute(accumCurveAngle);
	pos_changed.sendOutRoute(position);
	fraction_changed.sendOutRoute(set_fraction);
}


/*
void CCBend::getFields( String& str,int& i){
	while(str[i]!='}'){
		if(str.cmpKWord("curve",i)) {
			str.stb(i);
			curve.getField(str,i);
		}
		else if(str.cmpKWord("distance",i)) {
			str.stb(i);
			distance.getField(str,i);
		}
		else if(str.cmpKWord("autoClose",i))autoClose=getBool(str,i);

		else if(str.cmpKWord("orientation",i))orientation.getField(str, i);
		else if(str.cmpKWord("location",i))	location.getField(str, i);
		else if(str.cmpKWord("yAngle",i))	yAngle.getField(str, i);

		else throw this;

		str.stb(i);
	}

	matrix=rotationMatrix(orientation)*translationMatrix(location);
	imatrix=translationMatrix(-location)*rotationMatrix(-orientation);
	str.stb(i);
	return;
}
*/
float ECLathe::getDistance(int index)const{
	if(index>=distance.length){
		index=distance.length-1;
	}
	return height[index];
}
float ECSyncBend::getDistance(int index)const{
	if(index>=distance.length){
		index=distance.length-1;
	}
	return height[index];
}

float CCTranslate::getDistance(int index)const{
	const M_Float& distance=this->distance;
	if(index>=distance.length){
		index=distance.length-1;
	}
	float distanceSum=0;
	int i;
	for(i=0;i<=index;i++){
		distanceSum+=distance[i];
	}
	return distanceSum;
}

void CCScale::calcClose(){
	float curveSum=0;
	int i;
	for(i=0;i<curve.length-1;i++){
		curveSum+=curve.refValue()[i];
	}
	curve.refValue()[i]=PI*.5-curveSum;
	distance.refValue()[i-1]=getPosX(curve.length-2)/cos(curveSum)*curve[i];
}


void CCScale::doPosX(M_Vec3f& coord){
	float fPosX=1/posX;

	const M_Float& curve=this->curve.refValue();
	const M_Float& distance=this->distance.refValue();

	float accumHeight=0;//startPosY;
	float finishPosAx=baseRad.refValue();
	float finishPosAy=accumHeight;
	float posY=accumHeight;
	float prevPosY=accumHeight;
	float accumAng=curve[0];
	float nrad=distance[0]/curve[1];
	float rad=fabs(nrad);

	float AxisBx=-nrad*cos(accumAng)+finishPosAx;
	float AxisBy=-nrad*sin(accumAng)+finishPosAy;

	float ty;

	int i;
	float r;
	float zAngle;
	float angDiff;

////////////////////////////////////////////////////////
	int s=1;


	while(true){
		angDiff=curve[s];

		ty=accumAng*nrad;
		finishPosAx=nrad*cos(accumAng+angDiff);
		finishPosAy=nrad*sin(accumAng+angDiff);
		prevPosY=posY;
		posY=finishPosAy+AxisBy;
		float height=posY-prevPosY;
		for(i=0;i<coord.length;i++){
			if(coord[i].x<=0)continue;
			if(coord[i].y<accumHeight||coord[i].y>=accumHeight+height)continue;

			zAngle=asin(((coord[i].y-accumHeight)/rad+sin(ty/rad)));
			r=nrad*cos(zAngle)+AxisBx;
			coord[i].x*=r*fPosX;
		}

		accumHeight+=height;
		accumAng+=curve[s];
		if(++s>=curve.length)break;

		nrad=distance[s-1]/curve[s];
		rad=fabs(nrad);

		AxisBx+=-nrad*cos(accumAng)+finishPosAx;
		AxisBy+=-nrad*sin(accumAng)+finishPosAy;
	}
}
void CCScale::doNegX(M_Vec3f& coord){
	float fNegX=1/negX;
	const M_Float& curve=this->curve.refValue();
	const M_Float& distance=this->distance.refValue();

	float accumHeight=0;//startPosY;
	float finishPosAx=baseRad.refValue();
	float finishPosAy=accumHeight;
	float posY=accumHeight;
	float prevPosY=accumHeight;
	float accumAng=curve[0];
	float nrad=distance[0]/curve[1];
	float rad=fabs(nrad);

	float AxisBx=-nrad*cos(accumAng)+finishPosAx;
	float AxisBy=-nrad*sin(accumAng)+finishPosAy;

	float ty;

	int i;
	float r;
	float zAngle;
	float angDiff;

////////////////////////////////////////////////////////
	int s=1;


	while(true){
		angDiff=curve[s];

		ty=accumAng*nrad;
		finishPosAx=nrad*cos(accumAng+angDiff);
		finishPosAy=nrad*sin(accumAng+angDiff);
		prevPosY=posY;
		posY=finishPosAy+AxisBy;
		float height=posY-prevPosY;
		for(i=0;i<coord.length;i++){
			if(coord[i].x>=0)continue;
			if(coord[i].y<accumHeight||coord[i].y>=accumHeight+height)continue;

			zAngle=asin(((coord[i].y-accumHeight)/rad+sin(ty/rad)));
			r=nrad*cos(zAngle)+AxisBx;
			coord[i].x*=r*fNegX;
		}

		accumHeight+=height;
		accumAng+=curve[s];
		if(++s>=curve.length)break;

		nrad=distance[s-1]/curve[s];
		rad=fabs(nrad);

		AxisBx+=-nrad*cos(accumAng)+finishPosAx;
		AxisBy+=-nrad*sin(accumAng)+finishPosAy;
	}
}
void CCScale::doPosZ(M_Vec3f& coord){
	float fPosZ=1/posZ;
	const M_Float& curve=this->curve.refValue();
	const M_Float& distance=this->distance.refValue();

	float accumHeight=0;	//startPosY;
	float finishPosAx=baseRad.refValue();
	float finishPosAy=accumHeight;
	float posY=accumHeight;
	float prevPosY=accumHeight;
	float accumAng=curve[0];
	float nrad=distance[0]/curve[1];
	float rad=fabs(nrad);

	float AxisBx=-nrad*cos(accumAng)+finishPosAx;
	float AxisBy=-nrad*sin(accumAng)+finishPosAy;

	float ty;

	int i;
	float r;
	float zAngle;
	float angDiff;

////////////////////////////////////////////////////////
	int s=1;


	while(true){
		angDiff=curve[s];

		ty=accumAng*nrad;
		finishPosAx=nrad*cos(accumAng+angDiff);
		finishPosAy=nrad*sin(accumAng+angDiff);
		prevPosY=posY;
		posY=finishPosAy+AxisBy;
		float height=posY-prevPosY;
		for(i=0;i<coord.length;i++){
			if(coord[i].z<=0)continue;
			if(coord[i].y<accumHeight||coord[i].y>=accumHeight+height)continue;

			zAngle=asin(((coord[i].y-accumHeight)/rad+sin(ty/rad)));
			r=nrad*cos(zAngle)+AxisBx;
			coord[i].z*=r*fPosZ;
		}

		accumHeight+=height;
		accumAng+=curve[s];
		if(++s>=curve.length)break;

		nrad=distance[s-1]/curve[s];
		rad=fabs(nrad);

		AxisBx+=-nrad*cos(accumAng)+finishPosAx;
		AxisBy+=-nrad*sin(accumAng)+finishPosAy;
	}
}
void CCScale::doNegZ(M_Vec3f& coord){
	float fNegZ=1/negZ;
	const M_Float& curve=this->curve.refValue();
	const M_Float& distance=this->distance.refValue();

	float accumHeight=0;//startPosY;
	float finishPosAx=baseRad.refValue();
	float finishPosAy=accumHeight;
	float posY=accumHeight;
	float prevPosY=accumHeight;
	float accumAng=curve[0];
	float nrad=distance[0]/curve[1];
	float rad=fabs(nrad);

	float AxisBx=-nrad*cos(accumAng)+finishPosAx;
	float AxisBy=-nrad*sin(accumAng)+finishPosAy;

	float ty;

	int i;
	float r;
	float zAngle;
	float angDiff;

////////////////////////////////////////////////////////
	int s=1;

	while(true){
		angDiff=curve[s];

		ty=accumAng*nrad;
		finishPosAx=nrad*cos(accumAng+angDiff);
		finishPosAy=nrad*sin(accumAng+angDiff);
		prevPosY=posY;
		posY=finishPosAy+AxisBy;
		float height=posY-prevPosY;
		for(i=0;i<coord.length;i++){
			if(coord[i].z>=0)continue;
			if(coord[i].y<accumHeight||coord[i].y>=accumHeight+height)continue;

			zAngle=asin(((coord[i].y-accumHeight)/rad+sin(ty/rad)));
			r=nrad*cos(zAngle)+AxisBx;
			coord[i].z*=r*fNegZ;
		}

		accumHeight+=height;
		accumAng+=curve[s];
		if(++s>=curve.length)break;

		nrad=distance[s-1]/curve[s];
		rad=fabs(nrad);

		AxisBx+=-nrad*cos(accumAng)+finishPosAx;
		AxisBy+=-nrad*sin(accumAng)+finishPosAy;
	}
}
void CCScale::generate(M_Vec3f& coord, const int){
	if(curve.length==0)return;

	coord.multMatrixLeft(imatrix);
	if(bNegX==true){
		doNegX(coord);
	}
	if(bPosX==true){
		doPosX(coord);
	}
	if(bNegZ==true){
		doNegZ(coord);
	}
	if(bPosZ==true){
		doPosZ(coord);
	}
	coord.multMatrixLeft(matrix);
}


float CCRingScale::getScaleVal(float scaleVal){
	const M_Float& curve=this->curve;
	const M_Float& distance=this->distance;

	float accumHeight=0;//startPosY;
	float finishPosAx=baseRad.refValue();
	float finishPosAy=accumHeight;
	float posY=accumHeight;
	float prevPosY=accumHeight;
	float accumAng=curve[0];
	float nrad=distance[0]/curve[1];
	float rad=fabs(nrad);

	float AxisBx=-nrad*cos(accumAng)+finishPosAx;
	float AxisBy=-nrad*sin(accumAng)+finishPosAy;

	float ty;

	float r;
	float zAngle;
	float angDiff;

////////////////////////////////////////////////////////
	int s=1;


	while(true){
		angDiff=curve[s];

		ty=accumAng*nrad;
		finishPosAx=nrad*cos(accumAng+angDiff);
		finishPosAy=nrad*sin(accumAng+angDiff);
		prevPosY=posY;
		posY=finishPosAy+AxisBy;
		float height=posY-prevPosY;
		if(scaleVal>=accumHeight&&scaleVal<=accumHeight+height){;

			zAngle=asin(((scaleVal-accumHeight)/rad+sin(ty/rad)));
			r=nrad*cos(zAngle)+AxisBx;
			return r;
		}

		accumHeight+=height;
		accumAng+=curve[s];
		if(++s>=curve.length)break;

		nrad=distance[s-1]/curve[s];
		rad=fabs(nrad);

		AxisBx+=-nrad*cos(accumAng)+finishPosAx;
		AxisBy+=-nrad*sin(accumAng)+finishPosAy;
	}
	return getPosX(curve.length-1);
}


void CCRingScale::generate(M_Vec3f& coord, const int){
	if(!curve.length)return;
	float fNegX=1/negX;
	float fPosX=1/posX;
	float fNegZ=1/negZ;
	float fPosZ=1/posZ;

	M_Float& curve=this->curve.refValue();
	M_Float& distance=this->distance.refValue();

	float ccHeight=(getPosY(curve.length-1))/(PI*2);
	coord.multMatrixLeft(imatrix);

	//double the last 2 values so that the last vertex points are curved
	float storeLastDistance=distance[distance.length-1];
	float storeLastAngle=curve[curve.length-1];

////////////////////////////////////////////////////////

	for(int i=0;i<coord.length;i++){
		Vec3f &svec=coord[i];
		if(svec.y<0)continue;
		float polarPos=sqrt(svec.x*svec.x+svec.z*svec.z);
		float polarAngle=atan2(svec.z,svec.x);

		if(polarAngle<=-PI*.5){
			polarAngle+=PI*2.5;
		}
		else {
			polarAngle+=PI*.5;
		}


		float r=getScaleVal(polarAngle*ccHeight);
		//float r=1;

		if(svec.z>0){
			svec.z=((fPosZ*r)*(polarPos-centerOffset))*-sin(polarAngle+PI*.5);
			svec.z+=centerOffset*-sin(polarAngle+PI*.5);
		}
		else{
			svec.z=((fNegZ*r)*(polarPos-centerOffset))*-sin(polarAngle+PI*.5);
			svec.z+=centerOffset*-sin(polarAngle+PI*.5);
		}
		if(svec.x>0){
			svec.x=((fPosX*r)*(polarPos-centerOffset))*-cos(polarAngle+PI*.5);
			svec.x+=centerOffset*-cos(polarAngle+PI*.5);
		}
		else{
			svec.x=((fNegX*r)*(polarPos-centerOffset))*-cos(polarAngle+PI*.5);
			svec.x+=centerOffset*-cos(polarAngle+PI*.5);
		}
	}

	distance[distance.length-1]=storeLastDistance;
	curve[curve.length-1]=storeLastAngle;


	coord.multMatrixLeft(matrix);
}


void CCTranslate::generate(M_Vec3f& coord, const int){
	if(!curve.length)return;
	const M_Float& curve=this->curve.refValue();
	const M_Float& distance=this->distance.refValue();

	coord.multMatrixLeft(imatrix);

	float accumHeight=0;//startPosY;
	float finishPosAx=baseRad.refValue();
	float finishPosAy=accumHeight;
	float posY=accumHeight;
	float prevPosY=accumHeight;
	float accumAng=curve[0];
	float nrad=distance[0]/curve[1];
	float rad=fabs(nrad);

	float AxisBx=-nrad*cos(accumAng)+finishPosAx;
	float AxisBy=-nrad*sin(accumAng)+finishPosAy;

	float ty;

	int i;
	float r;
	float zAngle;
	float angDiff;

////////////////////////////////////////////////////////
	int s=1;

	while(true){
		angDiff=curve[s];

		ty=accumAng*nrad;
		finishPosAx=nrad*cos(accumAng+angDiff);
		finishPosAy=nrad*sin(accumAng+angDiff);
		prevPosY=posY;
		posY=finishPosAy+AxisBy;
		float height=posY-prevPosY;
		for(i=0;i<coord.length;i++){
			if(coord[i].y<accumHeight||coord[i].y>=accumHeight+height)continue;

			zAngle=asin(((coord[i].y-accumHeight)/rad+sin(ty/rad)));
			r=nrad*cos(zAngle)+AxisBx;
			coord[i].x+=r;
		}

		accumHeight+=height;
		accumAng+=curve[s];
		if(++s>=curve.length)break;

		nrad=distance[s-1]/curve[s];
		rad=fabs(nrad);

		AxisBx+=-nrad*cos(accumAng)+finishPosAx;
		AxisBy+=-nrad*sin(accumAng)+finishPosAy;
	}
	coord.multMatrixLeft(matrix);
}

inline void curveVertex(long double  pt1x, long double  pt1y,
		long double  ang1z,
		long double  pt2x,long double  pt2y,
		long double& a, long double& b,
		float t,
		Vec3f& coord
		){

	long double x = pt2x-pt1x;
	long double y = pt2y-pt1y;

	rotateZ(x, y, ang1z);

	long double j = y / b;
	if(j > 1){
		j = 1;
	}

	Matrix3 matrix=rotationMatrix(ang1z-PI*.5)*translationMatrix(Vec2f(pt1x,pt1y));
/////////

	Vec2f crd2f=matrix*Vec2f(a*cosl(t)-a,b*sinl(t));

	coord.y=crd2f.y;
	coord.z=crd2f.x*cos(coord.x);
	coord.x=crd2f.x*sin(coord.x);
}

inline void curveVertexRadSource(	long double  pt1x,long double  pt1y,
		long double  ang1z,
		long double  pt2x,long double  pt2y,
		long double& a, long double& b,
		float t,
		Vec3f& coord
		){

	long double x = pt2x - pt1x;
	long double y = pt2y - pt1y;

	rotateZ(x, y, ang1z);

	long double j = y / b;
	if(j > 1){
		j = 1;
	}

	Matrix3 matrix=rotationMatrix(ang1z-PI*.5)*translationMatrix(Vec2f(pt1x,pt1y));
/////////

	Vec2f crd2f=matrix*Vec2f(a*cosl(t)-a,b*sinl(t));

	coord.y=crd2f.y;
	float polarAngle=atan2(coord.z,coord.x);
	coord.z=crd2f.x*sin(polarAngle);
	coord.x=crd2f.x*cos(polarAngle);
}
void ECLathe::generate(M_Vec3f& coord, const int){
	if(!tangent.length)return;
	if(g_projectDirection!=PROJECT_FORWARD){
		for(int i=coord.length; --i>=0;){
			float x=coord[i].x;
			float z=coord[i].z;
			float polarPos=sqrt(x*x+z*z)-point[0].x; //point[0].x is same as a CCLathe's baseRad

			int idx=findSegment(polarPos);
			if(idx>=0){
				curveVertexRadSource(point[idx].x, point[idx].y, tangent[idx], point[idx+1].x, point[idx+1].y, a[idx], b[idx],	(polarPos-height[idx])*scaleDis[idx], coord[i]);
			}
		}
	}
	else{
		for(int i=coord.length; --i>=0;){
			float h=coord[i].y;
			int idx=findSegment(h);
			if(idx>=0){
				curveVertex(point[idx].x, point[idx].y, tangent[idx], point[idx+1].x, point[idx+1].y, a[idx], b[idx],	(h-height[idx])*scaleDis[idx], coord[i]);
			}
		}
	}
}
inline void curveVertexSyncBend(	long double  pt1x,long double  pt1y,
		long double  ang1z,
		long double  pt2x,long double  pt2y,
		long double& a, long double& b,
		float t,
		Vec3f& coord
		){

	long double x = pt2x - pt1x;
	long double y = pt2y - pt1y;

	rotateZ(x, y, ang1z);

	long double j = y / b;
	if(j > 1){
		j=1;
	}

	Matrix3 matrix=rotationMatrix(ang1z-PI*.5)*translationMatrix(Vec2f(pt1x,pt1y));
/////////

	Vec2f crd2f=matrix*Vec2f(a*cosl(t)-a,b*sinl(t));

	coord.y=crd2f.y;
	coord.z=crd2f.x;
//	coord.x=crd2f.x*sin(coord.x);
}
inline void curveVertexRadSourceSyncBend(	long double  pt1x,long double  pt1y,
		long double  ang1z,
		long double  pt2x,long double  pt2y,
		long double& a, long double& b,
		float t,
		Vec3f& coord
		){

	long double x = pt2x - pt1x;
	long double y = pt2y - pt1y;

	rotateZ(x, y, ang1z);

	long double j = y/b;
	if(j > 1){
		j=1;
	}

	Matrix3 matrix=rotationMatrix(ang1z-PI*.5)*translationMatrix(Vec2f(pt1x,pt1y));
/////////

	Vec2f crd2f=matrix*Vec2f(a*cosl(t)-a,b*sinl(t));

	coord.y=crd2f.y;
	coord.z=crd2f.x;
	//coord.x=crd2f.x*cos(polarAngle);
}
/*
void setAB(const M_Vec3f& relPoint, const M_Float& relTangent, M_LongDouble& _a, M_LongDouble& _b){
	_a.length=_b.length=relPoint.length-1;
	for(int i=0; i<relPoint.length-1; i++){
		long double x,y,angle,xPlus,yPlus;
		long double ang1z=relTangent[i];
		long double& a=_a[i];
		long double& b=_b[i];

		x=relPoint[i+1].x;
		y=relPoint[i+1].y;

		angle=ang1z+PI*.5;
		xPlus =x+cosl(angle)*TINY_AMOUNT;
		yPlus =y+sinl(angle)*TINY_AMOUNT;
		a=(powl(y,2)*powl(xPlus,2)-powl(yPlus,2)*powl(x,2))/(2*powl(yPlus,2)*x-2*powl(y,2)*xPlus);
		b=sqrtl(fabsl(powl(a,2)*powl(y,2)/(powl(x,2)+2*a*x)));
	}
}
*/
void ECSyncBend::generate(M_Vec3f& coord, const int){
	if(!tangent.length)return;
	if(g_projectDirection!=PROJECT_FORWARD){
		for(int i=coord.length; --i>=0;){
			float h=coord[i].z;
			int idx=findSegment(h);
			if(idx>=0){
				curveVertexRadSourceSyncBend(point[idx].x, point[idx].y, tangent[idx], point[idx+1].x, point[idx+1].y, a[idx], b[idx],	(h-height[idx])*scaleDis[idx], coord[i]);
			}
		}
	}
	else{
		for(int i=coord.length; --i>=0;){
			float h=coord[i].y;
			int idx=findSegment(h);
			if(idx>=0){
				curveVertexSyncBend(point[idx].x, point[idx].y, tangent[idx], point[idx+1].x, point[idx+1].y, a[idx], b[idx],	(h-height[idx])*scaleDis[idx], coord[i]);
			}
		}
	}
}
inline Vec3f curveVertexECBend(
		Matrix& pointMatrix,
		long double& a, long double& b,
		float t,
		Vec3f& s_crd
		){

/////////

	Vec2f crd2fA=Vec2f(a*cosl(t)-a,b*sinl(t));
	Vec2f crd2fB=Vec2f(a*cosl(t+.00001)-a,b*sinl(t+.00001));
	long double angle=atan2l(crd2fB.y-crd2fA.y, crd2fB.x-crd2fA.x)-PI*.5;

	Vec3f vec3f=Vec3f(crd2fA.x+cosl(angle)*s_crd.x, crd2fA.y+sinl(angle)*s_crd.x, s_crd.z);

	//rotateVectorY(vec3f,yAngle);

	return pointMatrix*vec3f;

}
void ECBend::generate(M_Vec3f& coord, const int){
	if(!tangent.length)return;
	for(int i=coord.length; --i>=0;){
		Vec3f s_crd=imatrix*coord[i];
		float h=s_crd.y;
		int idx=findSegment(h);
		if(idx>=0){
			s_crd=m_yMatrix[idx]*s_crd;
			coord[i]=curveVertexECBend(m_matrix[idx], a[idx], b[idx],	(h-height[idx])*scaleDis[idx], s_crd);
		}
	}
}
inline void curveVertexECBendLine(
		Matrix& pointMatrix,
		float	 yAngle,
		long double& a, long double& b,
		float t,
		Vec3f& coord
		){


/////////

	Vec3f vec3f=Vec3f(a*cosl(t)-a,b*sinl(t),0);

	rotateVectorY(vec3f,yAngle);

	coord=pointMatrix*vec3f;
}

float getEllipticalSegmentDistance(
		long double  y,
		long double a, long double b
		){

	int step=32;
	long double xyt,t,j;


	j=y/b;
	if(j>1)j=1;
	xyt=asinl(j);

	float dist=0;
	Vec3f prevVec3f=Vec3f(0,0,0);
	for(int i=1;i<step+1;i++){
		t=xyt/step*i;

		Vec3f vec3f=Vec3f(a*cosl(t)-a,b*sinl(t),0);

		float xDiff=vec3f.x-prevVec3f.x;
		float yDiff=vec3f.y-prevVec3f.y;
		dist+=sqrt(xDiff*xDiff + yDiff*yDiff);
		prevVec3f=vec3f;
	}
	return dist;
}

void processFieldsECBend(M_Vec3f& relPoint, M_Float& relTangent, M_LongDouble& ma, M_LongDouble& mb, M_Float& scaleDis, M_Float& distance, M_Float& height){
	ma.length=mb.length=scaleDis.length=distance.length=(height.length=relPoint.length)-1;

#ifdef DIRECT3D
#endif
	height[0]=0;
	for(int i=0; i<relPoint.length-1; i++){
		long double x,y,angle,xPlus,yPlus,xyt,j;
		long double angz=relTangent[i+1];
		long double& a=ma[i];
		long double& b=mb[i];

		x=relPoint[i+1].x;
		y=relPoint[i+1].y;

		angle=angz+PI*.5;
		xPlus =x+cosl(angle)*TINY_AMOUNT;
		yPlus =y+sinl(angle)*TINY_AMOUNT;
		a=(powl(y,2)*powl(xPlus,2)-powl(yPlus,2)*powl(x,2))/(2*powl(yPlus,2)*x-2*powl(y,2)*xPlus);
		b=sqrtl(fabsl(powl(a,2)*powl(y,2)/(powl(x,2)+2*a*x)));
		j=y/b;
		if(j>1)j=1;
		xyt=fabsl(asinl(j));

/*
		float bigRad;
		if(a>b){
			bigRad=a;
		}
		else{
			bigRad=b;
		}
		distance[i]=xyt*bigRad;
*/
		distance[i]=getEllipticalSegmentDistance(relPoint[i+1].y, a, b);

		height[i+1]=height[i]+distance[i];
		scaleDis[i]=xyt/distance[i];
	}
}

float getYangle(const float& x, const float& z){
	return -atan2(z,x);
}
inline void ECBend::copyPointTangentToXZ(){
	if(point.length!=tangent.length||point.length-1!=yAngle.length)return;
	relPoint.length=relTangent.length=point.length;

	point.refValue()[0]=Vec2f(0,0);

	{
		for( int i=point.length; --i>0;){
			relTangent[i]=tangent[i]-tangent[i-1];
			relPoint[i]=Rotation(0,0,1,-(tangent[i-1]-PI*.5))*Vec3f(point[i].x-point[i-1].x, point[i].y-point[i-1].y, 0);

		}
//		relPoint[0]=Rotation(0,0,1,-(tangent[0]))*Vec3f(point[0].x, point[0].y, 0);
		relTangent[0]=(tangent[0]-PI*.5);
	}
	{
//		Vec3f testEnd=point[3];
		m_yMatrix.length=m_matrix.length=point.length;

		m_matrix[0]=Matrix();
//		m_matrix[1]=Matrix();
//		m_matrix[2]=Matrix();

		m_yMatrix[0]=Matrix();
		Matrix prevYmatrix = Matrix();
//		m_matrix[0]=rotationMatrix(Rotation(0,0,1,relTangent[0]));


		m_matrix[0]=rotationMatrix(Rotation(0,0,1,relTangent[0]));
		//m_matrix[0]=m_matrix[0]*translationMatrix(Vec3f(point[0].x, point[0].y, 0));

		for( int i=1; i<point.length; i++){
//			relPoint[i]=Vec3f(point[i].x, point[i].y, 0);
//			relPoint[i]=matrix*Vec3f(point[i].x, point[i].y, 0);
//			yAngle[i-1]=getYangle(relPoint[i].x, relPoint[i].z);
//			if(relPoint[i].x<0){
//				yAngle[i-1]+=PI;
//			}
			//relPoint[i]=Rotation(0,1,0,-yAngle[i-1])*relPoint[i];
//			matrix=matrix*translationMatrix(-relPoint[i])*rotationMatrix(Rotation(0,0,1,-relTangent[i]));

//			testEnd=Rotation(0,1,0,-yAngle[i-1])*testEnd;
//			testEnd=testEnd-relPoint[i];
//			testEnd=Rotation(0,0,1,-relTangent[i])*testEnd;

			m_matrix[i]=rotationMatrix(Rotation(0,0,1,relTangent[i]));
			m_matrix[i]=m_matrix[i]*translationMatrix(relPoint[i]);
			m_matrix[i]=m_matrix[i]*rotationMatrix(Rotation(0,1,0,yAngle[i-1]));
			m_matrix[i]=m_matrix[i]*m_matrix[i-1];
			m_yMatrix[i-1]=prevYmatrix=prevYmatrix*rotationMatrix(Rotation(0,1,0,-yAngle[i-1]));
			//testEnd=matrix*testEnd;
		}
		//relPoint[0]=Rotation(0,0,1,-tangent[0])*Vec3f(point[0].x, point[0].y, 0);
	}
	{
		for( int i=point.length-1; --i>=0;){
			m_matrix[i]=rotationMatrix(Rotation(0,1,0,yAngle[i]))*m_matrix[i]*this->matrix;
		}
	}

	processFieldsECBend(relPoint, relTangent, a, b, scaleDis, distance, height);
}

void rotateZ(Vec2f *p,int numberOf,float angle){
float newY,newX;
int i;
	for (i=0;i<numberOf;i++,p++){
		newX=p->x*cos(angle)-p->y*sin(angle);
		newY=p->x*sin(angle)+p->y*cos(angle);
		p->x=newX;
		p->y=newY;
	}
}
void setAB(M_ECurve& m_ecurve, const M_Vec2f& point, const M_Float& tangent, M_LongDouble& _a, M_LongDouble& _b){
	for(int i=0; i<point.length-1; i++){
		long double ang1z = tangent[i];
		long double ang2z = tangent[i+1];
		long double& a=_a[i];
		long double& b=_b[i];
		Vec2f& centerPos=m_ecurve[i].centerPos;

		long double x = point[i+1].x-point[i].x;
		long double y = point[i+1].y-point[i].y;

		rotateZ(x, y, ang1z);

		long double angle = ang2z - ang1z + PI*.5;
		long double xPlus = x + cosl(angle)*TINY_AMOUNT;
		long double yPlus = y + sinl(angle)*TINY_AMOUNT;
		a=(powl(y,2)*powl(xPlus,2)-powl(yPlus,2)*powl(x,2))/(2*powl(yPlus,2)*x-2*powl(y,2)*xPlus);
		b=sqrtl(fabsl(powl(a,2)*powl(y,2)/(powl(x,2)+2*a*x)));
		long double j = y / b;
		if(j > 1){
			j = 1;
		}
		centerPos.x=-a;
		centerPos.y=0;
		rotateZ(centerPos,tangent[i]-PI*.5);
	}
}
void setUpMatrices(M_ECurve& m_ecurve, const M_Vec2f& point, const M_Float& tangent, M_LongDouble& _a, M_LongDouble& _b){
	if(point.length!=tangent.length)return;
	_a.length=_b.length=(m_ecurve.length=point.length)-1;
#ifdef DIRECT3D
#endif
	setAB(m_ecurve, point, tangent, _a, _b);

	for(int i=point.length-1; --i>=0;){

		long double a=fabsl(_a[i]);
		long double b=fabsl(_b[i]);
		float tang=tangent[i];
		Vec2f centerPos=m_ecurve[i].centerPos;


		/////////////////////////////
		//	make line center of universe
		Vec2f g=point[i]+centerPos;
		Matrix3 matrix=translationMatrix(-g);

		//rotate inverse
		matrix*=rotationMatrix(PI*.5-tang);

		//convert to circle
		Vec2f abScale=Vec2f(1/a,1/b);
		matrix*=scaleMatrix(abScale);

		Vec2f lineTo=matrix*point[i+1];
		Vec2f lineFrom=matrix*Vec2f(point[i+1].x-1, point[i+1].y);

		//rotate to make line straight
		float xDiff=lineTo.x-lineFrom.x;
		float yDiff=lineTo.y-lineFrom.y;
		float hyp=sqrt(xDiff*xDiff+yDiff*yDiff);
		float straightAngle=-asin(yDiff/hyp);
		if(m_ecurve[i].centerPos.x>0){
			straightAngle+=PI;
		}

		matrix*=rotationMatrix(straightAngle);
		m_ecurve[i].matrixA=matrix;
		lineTo=matrix*point[i+1];




	//reverse rotations and scale
		matrix=rotationMatrix(-straightAngle);
		matrix*=scaleMatrix(Vec2f(a,b));
		matrix*=rotationMatrix(-(PI*.5-tang));
		matrix*=translationMatrix(g);

		m_ecurve[i].matrixB=matrix;
	}
}
/*
void setUpMatrices(M_ECurve& m_ecurve, const M_Vec3f& point, const M_Vec2f& tangent, M_LongDouble& _a, M_LongDouble& _b){
	if(point.length!=tangent.length)return;
	_a.length=_b.length=(m_ecurve.length=point.length)-1;

	setAB(m_ecurve, point, tangent, _a, _b);

	for(int i=point.length-1; --i>=0;){

		long double a=fabsl(_a[i]);
		long double b=fabsl(_b[i]);
		float tang=tangent[i].x;
		Vec2f centerPos=m_ecurve[i].centerPos;


		/////////////////////////////
		//	make line center of universe
		Vec2f g=Vec2f(point[i].x, point[i].y)+centerPos;
		Matrix3 matrix=translationMatrix(-g);

		//rotate inverse
		matrix*=rotationMatrix(PI*.5-tang);

		//convert to circle
		Vec2f abScale=Vec2f(1/a,1/b);
		matrix*=scaleMatrix(abScale);

		Vec2f lineTo=matrix*Vec2f(point[i+1].x,point[i+1].y);
		Vec2f lineFrom=matrix*Vec2f(point[i+1].x-1, point[i+1].y);

		//rotate to make line straight
		float xDiff=lineTo.x-lineFrom.x;
		float yDiff=lineTo.y-lineFrom.y;
		float hyp=sqrt(xDiff*xDiff+yDiff*yDiff);
		float straightAngle=-asin(yDiff/hyp);
		if(m_ecurve[i].centerPos.x>0){
			straightAngle+=PI;
		}

		matrix*=rotationMatrix(straightAngle);
		m_ecurve[i].matrixA=matrix;
		lineTo=matrix*Vec2f(point[i+1].x, point[i+1].y);




	//reverse rotations and scale
		matrix=rotationMatrix(-straightAngle);
		matrix*=scaleMatrix(Vec2f(a,b));
		matrix*=rotationMatrix(-(PI*.5-tang));
		matrix*=translationMatrix(g);

		m_ecurve[i].matrixB=matrix;
	}
}
*/
void ECScale::generate(M_Vec3f& coord, const int){
	if(!point.length)return;
	float fNegX=1/negX;
	float fPosX=1/posX;
	float fNegZ=1/negZ;
	float fPosZ=1/posZ;

	for(int i=0; i<coord.length; i++){
		Vec3f s_crd=imatrix*coord[i];
		int segment = findSegment(s_crd.y);
		if(segment!=-1){
			Matrix3& matrixA=m_ecurve[segment].matrixA;
			Matrix3& matrixB=m_ecurve[segment].matrixB;

			Vec2f lineTo=matrixA*Vec2f(0,s_crd.y);
			float as=asin(lineTo.y);
			lineTo=matrixB*Vec2f(cos(as),sin(as));

			if(bNegX==true && s_crd.x<0){
				s_crd.x*=lineTo.x*fNegX;
				if(bPosX==true && s_crd.x>0){
jumpPosX:
					s_crd.x*=lineTo.x*fPosX;
				}
				if(bNegZ==true && s_crd.z<0){
jumpNegZ:
					s_crd.z*=lineTo.x*fNegZ;
				}
				if(bNegZ==true && s_crd.z>0){
jumpPosZ:
					s_crd.z*=lineTo.x*fPosZ;
				}
				coord[i]=matrix*s_crd;
			}
			else if(bPosX==true && s_crd.x>0){
				goto jumpPosX;
			}
			else if(bNegZ==true && s_crd.z<0){
				goto jumpNegZ;
			}
			else if(bPosZ==true && s_crd.z>0){
				goto jumpPosZ;
			}
		}
	}
}

void ECShift::generate(M_Vec3f& coord, const int){
	if(!point.length)return;

	float negDecay=this->negDecay;
	float posDecay=this->posDecay;

	for(int i=0; i<coord.length; i++){
		Vec3f s_crd=imatrix*coord[i];
		int segment = findSegment(s_crd.y);
		if(segment!=-1){
			Matrix3& matrixA=m_ecurve[segment].matrixA;
			Matrix3& matrixB=m_ecurve[segment].matrixB;

			Vec2f lineTo=matrixA*Vec2f(0,s_crd.y);
			float as=asin(lineTo.y);
			lineTo=matrixB*Vec2f(cos(as),sin(as));

			if(s_crd.x<0){
				if(s_crd.x<-negDecay)continue;
				float sineCurve=cos((fabs(s_crd.x)/negDecay)*PI)*.5+.5;
				s_crd.x+=lineTo.x*sineCurve;
			}
			else{
				if(s_crd.x>posDecay)continue;
				float sineCurve=cos((fabs(s_crd.x)/posDecay)*PI)*.5+.5;
				s_crd.x+=lineTo.x*sineCurve;
			}
			coord[i]=matrix*s_crd;
		}
	}
}
void ECTwist::generate(M_Vec3f& coord, const int){
	if(!point.length)return;

	for(int i=0; i<coord.length; i++){
		Vec3f s_crd=imatrix*coord[i];
		int segment = findSegment(s_crd.y);
		if(segment!=-1){
			Matrix3& matrixA=m_ecurve[segment].matrixA;
			Matrix3& matrixB=m_ecurve[segment].matrixB;

			Vec2f lineTo=matrixA*Vec2f(0,s_crd.y);
			float as=asin(lineTo.y);
			lineTo=matrixB*Vec2f(cos(as),sin(as));

			float cosAngle=cos(PI*2*lineTo.x);
			float sinAngle=sin(PI*2*lineTo.x);
			float newZ=s_crd.z*cosAngle-s_crd.x*sinAngle;
			float newX=s_crd.z*sinAngle+s_crd.x*cosAngle;
			s_crd.z=newZ;
			s_crd.x=newX;

			coord[i]=matrix*s_crd;
		}
	}
}

void EvolveCoord::doBuild(){
	if(!point.length)return;
	M_SArt& source=this->source.refValue();
	M_SArt& part=this->part.refValue();
	int length=source.length;
	if(length>part.length){
		length=part.length;
	}
	if(!length)return;

	location=location.refValue();
	for(int k=0; k<length;k++){
		M_Vec3f& srcCoord=source[k]->coord;
		M_Vec3f& prtCoord=part[k]->coord;
		if(srcCoord.length!=prtCoord.length)continue;

		for(int i=0; i<prtCoord.length; i++){
			Vec3f p_crd=imatrix*prtCoord[i];
			int segment = findSegment(p_crd.y);
			if(segment!=-1){
				Matrix3& matrixA=m_ecurve[segment].matrixA;
				Matrix3& matrixB=m_ecurve[segment].matrixB;

				Vec2f lineTo=matrixA*Vec2f(0,p_crd.y);
				float as=asin(lineTo.y);
				lineTo=matrixB*Vec2f(cos(as),sin(as));
				float m=lineTo.x;
				float im=1-m;

				prtCoord[i]=prtCoord[i]*im + srcCoord[i]*m;
			}
		}
	}
}

void addEffect(M_<Effect*>& effectList, Effect* effect){
	for(int k = effectList.length; --k>=0;){
		if(effectList[k] == effect)return;
	}
	effectList += effect;
	effect->triLen = 0;
	effect->partList.length=0;
}

void Art::setUpInternalEffectPtrsTmpForRenderingTransparancy(Effect* lastEffect){
	if(effect.p){//will be all types of effects pters when working how it should be
		lastEffect = dynamic_cast<Effect*>(effect.refValue().getRawPtr()); //***will be uncommented when working how it should be
	}
	workEffect=lastEffect;
	M_SArt& mpart = part.refValue();
	for(int k=0; k<mpart.length;k++){
		mpart[k]->setUpInternalEffectPtrsTmpForRenderingTransparancy(lastEffect);
	}
}
void setUpInternalEffectPtrsTmpForRenderingTransparancy(Seamless* smls){
	Effect* lastEffect=dynamic_cast<Effect*>(smls->effect.refValue().p);
	const M_SArt& skeleton=smls->skeleton;
	for(int k=0; k<skeleton.length;k++){
		skeleton[k]->setUpInternalEffectPtrsTmpForRenderingTransparancy(lastEffect);
	}
}

void Art::setUpInternalEffectPtrs2(Effect* lastEffect){ //this (2) wont be needed when setUpInternalEffectPtrs is fixed
	if(effect != 0){
		lastEffect = dynamic_cast<Effect*>(effect.refValue().getRawPtr());
		lastEffect = dynamic_cast<Effect*>(effect.refValue().getRawPtr());
		addEffect(smls->effectList, effect);
	}
	lastEffect->partList += this;
	if(show){
		lastEffect->triLen += triangle.length;
	}
	workEffect = lastEffect;

	M_SArt& mpart = part.refValue();
	for(int k=0; k < mpart.length;k++){
		mpart[k]->setUpInternalEffectPtrs2(lastEffect);
	}
}
void Art::setUpInternalEffectPtrs(Effect* lastEffect){
	if(effect.p && dynamic_cast<MonoColorEffect*>(lastEffect)){//will be all types of effects pters when working how it should be
		lastEffect = dynamic_cast<Effect*>(effect.refValue().getRawPtr());
	}
	workEffect = lastEffect;
	M_SArt& mpart = part.refValue();
	for(int k=0; k<mpart.length;k++){
		mpart[k]->setUpInternalEffectPtrs(lastEffect);
	}
}
/*
inline Node* Seamless::EffectField::operator = (NodePtr& par){
	refValue()=par->getRawPtr();
}
*/
inline Node* Art::EffectField::operator = (Node* par){
	refValue()=par;
/*
	setUpInternalEffectPtrs(home->smls);
	if(par==0)return 0;
	Seamless* smls=home->smls;
	smls->process();
	M_Art& partList=smls->partList;
	if(refValue()->getType()==COLOREFFECT){
		for(int k=0; k<partList.length; k++){
			Art* part=partList[k];
			M_Color3f& color=dynamic_cast<ColorEffect*>(part->workEffect)->color.refValue();
			color.length+=part->coord.length;
			M_Triangle& triangle=part->triangle;
			for(int i=triangle.length*3; --i>=0;){
				triangle.refTexIndex(i)=triangle.refIndex(i)+triangle.refPart(i)->fromCoord;
			}
		}
	}
	else if(refValue()->getType()==TEXTUREEFFECT){
		for(int k=0; k<partList.length; k++){
			Art* part=partList[k];
			M_Vec2f& texCoord=dynamic_cast<TextureEffect*>(part->workEffect)->texCoord.refValue();
			texCoord.length+=part->texCoord.length;
			M_Triangle& triangle=part->triangle;
			for(int i=triangle.length*3; --i>=0;){
				triangle.refTexIndex(i)=triangle.refIndex(i)+triangle.refPart(i)->fromCoord;
			}
		}
	}
	else throw this;
	*/
	return getRawPtr();
}


bool searchForField(String& str, int& i, String field, String nodeType, String name){
	while(i<str.length){
		i=str.indexOf(field,i);
		if(i==-1)return 0;
		i+=field.length;
		if (str[i]!=' '){
			continue;
		}
		str.stw_stb(i);
		if (!str.cmpKWord(nodeType,i)){
			continue;
		}
		str.stb(i);
		if (!str.cmpKWord(name,i)){
			continue;
		}
		str.stb(i);
		i--;
		return true;
	}
	return false;
}

void Art::catMatrix(Matrix catMatrix){
	matrix = translationMatrix(-absolutePivotPoint) * scaleMatrix(scale) * rotationMatrix(rotation) * translationMatrix(absolutePivotPoint + translation) * catMatrix;

	for(int k = 0; k < part.length; k++){
//		part->part[k]->parent=part;
		part.refValue()[k]->catMatrix(matrix);
	}
}


Vec3d conv3fto3d(const Vec3f& par){
	return Vec3d(par.x, par.y, par.z);
}
Vec3f conv3dto3f(const Vec3d& par){
	return Vec3f(par.x, par.y, par.z);
}
Rotation convRot4dto4f(const Rotation4d& par){
	return Rotation(par.x, par.y, par.z, par.angle);
}

Rotation4d convRot4fto4d(const Rotation& par){
	return Rotation4d(par.x, par.y, par.z, par.angle);
}

M_Rotation4d g_m_accumRot;
M_Rotation4d g_m_iaccumRot;


void Art::buildSceneTranslationMatrices(Matrix parentMatrix){

	Matrix thisMatrix = translationMatrix(translation) * parentMatrix;
	matrix = thisMatrix;

	for(int k=0; k < part.length; k++){
		part.refValue()[k]->buildSceneTranslationMatrices(thisMatrix);
	}
}
void Art::buildSceneScaleTranslationMatrices(Matrix parentMatrix){

	matrix = scaleMatrix(scale) * translationMatrix(translation) * parentMatrix;

	for(int k=0; k < part.length; k++){
		part.refValue()[k]->buildSceneTranslationMatrices(matrix);
	}
}
void Art::buildSceneITranslationMatrices(Matrix parentMatrix){

	Matrix thisMatrix = translationMatrix( -translation) * parentMatrix;
	matrix = thisMatrix;

	for(int k=0; k < part.length; k++){
		part.refValue()[k]->buildSceneTranslationMatrices(thisMatrix);
	}
}

void Art::catIMatrix(Matrix catMatrix){
	imatrix = catMatrix*translationMatrix(-translation - absolutePivotPoint) * rotationMatrix( -rotation) * scaleMatrix(Vec3f(1./scale.x, 1./scale.y, 1./scale.z)) * translationMatrix(absolutePivotPoint);
	M_SArt& part=this->part.refValue();
	for(int k=0; k<part.length; k++){
		part[k]->catIMatrix(imatrix);
	}
}



void Art::catMatrix(Matrix catMatrix, Rotation4d catRot, M_Rotation4d& m_accumRot){
	if(ik && pin){
		rotation.refValue() = ikRot;
		catRot = Rotation4d();
	}

	matrix = translationMatrix(-absolutePivotPoint) * scaleMatrix(scale) * rotationMatrix(rotation) * translationMatrix(absolutePivotPoint + translation) * catMatrix;
	catRot = convRot4fto4d(rotation) * catRot;
	m_accumRot[plidx] = catRot;

	for(int k = 0; k < part.length; k++){
//		part->part[k]->parent = part;
		part.refValue()[k]->catMatrix(matrix, catRot, m_accumRot);
	}
}


void Art::catIMatrix(Matrix catMatrix, Rotation4d catRot, M_Rotation4d& m_accumRot){
	if(ik && pin){
		rotation.refValue() = ikRot;
		catRot = Rotation4d();
	}

	imatrix = catMatrix*translationMatrix(-translation - absolutePivotPoint) * rotationMatrix(-rotation) * scaleMatrix(Vec3f(1./scale.x, 1./scale.y, 1./scale.z)) * translationMatrix(absolutePivotPoint);
	catRot = catRot * convRot4fto4d(-rotation);
	m_accumRot[plidx] = catRot;

	M_SArt& part = this->part.refValue();
	for(int k=0; k<part.length; k++){
		part[k]->catIMatrix(imatrix, catRot, m_accumRot);
	}
}


void Art::buildPartListNoShowLoop(int& _plidx){
	if(show == false && relevantPart.length==1)return;
	smls->partList += this;
	plidx = _plidx++;
	M_SArt& m_part = part.refValue();
	for(int k = 0; k < part.length; k++){
		m_part[k]->buildPartListNoShowLoop(_plidx);
	}
}

void Art::buildPartListLoop(int &_plidx){
	smls->partList += this;
	plidx = _plidx++;
	M_SArt& m_part = part.refValue();
	for(int k = 0; k < part.length; k++){
		m_part[k]->buildPartListLoop(_plidx);
	}
}



bool singleColor(M_Color3f& color3f){
	if(color3f.length){
		Color3f firstColor = color3f[0];
		for(int i = color3f.length; --i>=1;){
			if(firstColor!=color3f[i]){
				return false;
			}
		}
	}
	return true;
}

void Art::processSegmentedProduct(M_Node& children){
	if(show==false){
		return;
	}
	Shape* shape = new Shape;
	workEffect->copyFieldsData(shape);
	IndexedFaceSetPtr indexedFaceSet = extractGeometry();
	if(Color* color = dynamic_cast<Color*>(indexedFaceSet->color.p)){
		M_Color3f& color3f = color->color.refValue();
		if(color3f.length){
			Color3f firstColor = color3f[0];
			if(singleColor(color3f)){
				indexedFaceSet->color = 0;
				indexedFaceSet->colorIndex.refValue().length = 0;
				dynamic_cast<Material*>(dynamic_cast<Appearance*>(shape->appearance.p)->material.p)->diffuseColor.refValue()=firstColor;
			}
		}
		else{
			indexedFaceSet->color = 0;
			indexedFaceSet->colorIndex.refValue().length = 0;
		}
	}
	shape->geometry = indexedFaceSet;
	M_SArt& m_part = part.refValue();
	if(name != ""){
		Transform* transform = new Transform;
		transform->name = name;

		transform->translation = translation;
		transform->rotation = rotation;
		transform->scale = scale;
		transform->center = absolutePivotPoint;

		transform->children.refValue() += shape;
		children += transform;

		for(int k=0; k<m_part.length; k++){
			m_part[k]->processSegmentedProduct(transform->children.refValue());
		}
	}
	else {
		children += shape;
		M_SArt& m_part = part.refValue();
		for(int k=0; k<m_part.length; k++){
			m_part[k]->processSegmentedProduct(children);
		}
	}
}
inline void genCoordIndexForIFS(Seamless* smls, M_Int& coordIndex){
	int d=0,k;
	int coordIndexlen=0;
	M_Art& partList=smls->partList;
	int accumFromCoord=0;

	for(k=0; k<partList.length;k++){
		Art* part=partList[k];
		coordIndexlen+=part->triangle.length;
		part->fromCoord=accumFromCoord;
		accumFromCoord+=part->coord.length;
	}
	coordIndexlen*=4;
	coordIndex.length=coordIndexlen;

	for(k=0; k<partList.length;k++){
		Art* part=partList[k];
		M_Triangle& triangle=part->triangle;

		for(int i=0; i<triangle.length; i++){
			S_Triangle& tri=triangle[i];
			coordIndex[d++]=tri.pa->fromCoord+tri.ia;
			coordIndex[d++]=tri.pb->fromCoord+tri.ib;
			coordIndex[d++]=tri.pc->fromCoord+tri.ic;
			coordIndex[d++]=-1;
		}
	}
}
inline void genTexCoordIndexForIFS(Seamless* smls, M_Int& texCoordIndex){
	int d=0,k;
	int texCoordIndexLen=0;
	M_Art& partList=smls->partList;

	for(k=0; k<partList.length;k++){
		Art* part=partList[k];
		texCoordIndexLen+=part->triangle.length;
	}
	texCoordIndexLen*=4;
	texCoordIndex.length=texCoordIndexLen;

	for(k=0; k<partList.length;k++){
		Art* part=partList[k];
		M_Triangle& triangle=part->triangle;

		for(int i=0; i<triangle.length; i++){
			S_Triangle& tri=triangle[i];
			texCoordIndex[d++]=tri.ta;
			texCoordIndex[d++]=tri.tb;
			texCoordIndex[d++]=tri.tc;
			texCoordIndex[d++]=-1;
		}
	}
}
/*
String Coordinate::toStringRounded(bool newList){
	String str;

	if(setDEFgetUSE( str, newList)){
		return str;
	}
	str+="Coordinate {";
	if(point.length){
		str+="point";
		M_Vec3f& point=this->point.refValue();
		toStringRoundDown((float*)&point[0], point.length*3, g_formatCoords, str);
	}
	str+="\n}";

	return str;
}
String Color::toStringRounded(bool newList){
	String str;

	if(setDEFgetUSE( str, newList)){
		return str;
	}
	str+="Color {";
	if(color.length){
		str+="color";
		M_Color3f& color=this->color.refValue();
		toStringRoundDown((float*)&color[0], color.length*3, g_formatColor, str);
	}
	str+="\n}";

	return str;
}
String Normal::toStringRounded(bool newList){
	String str;

	if(setDEFgetUSE( str, newList)){
		return str;
	}
	str+="Normal {";
	if(vector.length){
		str+="vector";
		M_Vec3f& vector=this->vector.refValue();
		toStringRoundDown((float*)&vector[0], vector.length*3, g_formatCoords, str);
	}
	str+="\n}";

	return str;
}
String TextureCoordinate::toStringRounded(bool newList){
	String str;
	if(setDEFgetUSE( str, newList)){
		return str;
	}
	str+="TextureCoordinate {";
	if(point.length){
		str+="point";
		M_Vec2f& point=this->point.refValue();
		toStringRoundDown((float*)&point[0], point.length*2, g_formatCoords, str);
	}
	str+="\n}";
	return str;
}
*/
inline bool checkExtraCoord( M_Int& coordLookUp, int& newTexPoint, int& ci , int& tci ){
	for(int i=newTexPoint; i<coordLookUp.length; i++){
		if(coordLookUp[i]==ci){
			tci=i;
			return true;
		}
	}
	return false;
}
void IndexedFaceSet::addExtraCoordAndNormalAndTexCoord(){
	creaseAngle=0;
	M_Int& coordIndex=refCoordIndex();
	M_Int& texCoordIndex=refTexCoordIndex();
	M_Vec3f& point=refPoint();
	M_Vec3f& vector=refVector();
	M_Vec2f& texPoint=refTexPoint();
	static M_Int coordLookUp;
	static M_Vec3f newPoint;
	static M_Vec3f newVector;
	coordLookUp.length=texPoint.length;
	newPoint.length=texPoint.length;
	newVector.length=texPoint.length;
	int newTexPoint=texPoint.length;
	{
		for(int i=newTexPoint; --i>=0;){
			coordLookUp[i]=-1;
		}
	}
	{
		for(int i=0; i<coordIndex.length; i++){
			int ci=coordIndex[i];
			if(ci==-1)continue;
			int tci=texCoordIndex[i];
			if(coordLookUp[tci]==-1){
				coordLookUp[tci]=ci;
				newPoint[tci]=point[ci];
				newVector[tci]=vector[ci];
			}
			else if(coordLookUp[tci]==ci){
				//good its already done
			}
			else if(checkExtraCoord( coordLookUp, newTexPoint, ci, texCoordIndex[i] )){
			}
			else { //add another texPoint
				texCoordIndex[i]=texPoint.length;
				coordLookUp+=ci;
				newPoint+=point[ci];
				newVector+=vector[ci];
				texPoint+=texPoint[tci];
			}
		}
	}
	point=newPoint;
	vector=newVector;
	texCoordIndex.length=0;
}

void IndexedFaceSet::addExtraTexCoord(){
	M_Int& coordIndex=refCoordIndex();
	M_Int& texCoordIndex=refTexCoordIndex();
	M_Vec2f& texPoint=refTexPoint();
	static M_Int coordLookUp;
	coordLookUp.length=texPoint.length;
	int newTexPoint=texPoint.length;
	{
		for(int i=newTexPoint; --i>=0;){
			coordLookUp[i]=-1;
		}
	}
	{
		for(int i=0; i<coordIndex.length; i++){
			int ci=coordIndex[i];
			if(ci==-1)continue;
			int tci=texCoordIndex[i];
			if(coordLookUp[tci]==-1){
				coordLookUp[tci]=ci;
			}
			else if(coordLookUp[tci]==ci){
				//good its already done
			}
			else if(checkExtraCoord( coordLookUp, newTexPoint, ci, texCoordIndex[i] )){
			}
			else { //add another texPoint
				texCoordIndex[i]=texPoint.length;
				coordLookUp+=ci;
				texPoint+=texPoint[tci];
			}
		}
	}
}
void IndexedFaceSet::addExtraColor(){
	M_Int& coordIndex=refCoordIndex();
	M_Int& colorIndex=refColorIndex();
	M_Color3f& color=refColor();
	static M_Int coordLookUp;
	coordLookUp.length=color.length;
	int newColor=color.length;
	{
		for(int i=newColor; --i>=0;){
			coordLookUp[i]=-1;
		}
	}
	{
		for(int i=0; i<coordIndex.length; i++){
			int ci=coordIndex[i];
			if(ci==-1)continue;
			int tci=colorIndex[i];
			if(coordLookUp[tci]==-1){
				coordLookUp[tci]=ci;
			}
			else if(coordLookUp[tci]==ci){
				//good its already done
			}
			else if(checkExtraCoord( coordLookUp, newColor, ci, colorIndex[i] )){
			}
			else { //add another color
				colorIndex[i]=color.length;
				coordLookUp+=ci;
				color+=color[tci];
			}
		}
	}
}

/*
String IndexedFaceSet::toStringRounded(bool newList){
	String str;
	if(setDEFgetUSE( str, newList)){
		return str;
	}
	str+="IndexedFaceSet {";
	if(creaseAngle!=0){
		str+="\ncreaseAngle ";
		str+=creaseAngle.refValue();
	}
	if(coord!=0){
		str+="\ncoord ";
		str+=CoordinatePtr(coord)->toStringRounded(newList);
	}
	if(normal!=0){
		str+="\nnormal ";
		str+=NormalPtr(normal)->toStringRounded(newList);
	}
	if(color!=0){
		str+="\ncolor ";
		str+=ColorPtr(color)->toStringRounded(newList);
	}
	if(colorIndex.size!=0){
		str+="\ncolorIndex ";
		str+=colorIndex.refValue().toString();
	}
	if(texCoord!=0){
		str+="\ntexCoord ";
		str+=TextureCoordinatePtr(texCoord)->toStringRounded(newList);
	}
	if(texCoordIndex.size!=0){
		str+="\ntexCoordIndex ";
		str+=texCoordIndex.refValue().toString();
	}
	if(coordIndex.size!=0){
		str+="\ncoordIndex ";
		str+=coordIndex.refValue().toString();
	}

	str+="\n}";
	return str;
}

String Shape::toStringRounded(bool newList){
	String str;
	if(setDEFgetUSE( str, newList)){
		return str;
	}
	str+="Shape {";
	if(appearance!=0){
		str+="\nappearance ";
		str+=AppearancePtr(appearance)->toString(newList);
	}
	if(geometry!=0){
		str+="\ngeometry ";
		str+=IndexedFaceSetPtr(geometry)->toStringRounded(newList);
	}
	str+="\n}";
	return str;
}
*/

void IndexedFaceSet::genTransformedNormals(M_Art& partList){
	M_Vec3f& point = refPoint();
	M_Vec3f& vector = refVector();
	M_Int& coordIndex=this->coordIndex.refValue();

	{
		int coordLen=0;
		int coordIndexLen=0;
		for(int k=0; k<partList.length; k++){
			Art* part=partList[k];
			int clen=part->coord.length;
			part->texCoord.length=clen;
			part->color.length=clen;
			part->normal.refValue().length=clen;
			part->fromCoord=coordLen;
			M_Int& localCoordLookUp=part->partTree;
			localCoordLookUp.length=clen;
			coordLen+=clen;
			part->toCoord=clen;
			coordIndexLen+=part->triangle.length;
			for(int i=clen; --i>=0; ){
				localCoordLookUp[i]=-1;
			}
		}
		coordIndexLen*=4;
		coordIndex.length=coordIndexLen;
		point.length=coordLen;
	}
	{
		for(int j = 0, c = 0, k = 0; k < partList.length; k++){
			Art* part = partList[k];
			M_Triangle& triangle = part->triangle;
			int triLen = triangle.length * 3;
			Matrix matrix = part->matrix;
			M_Vec3f& coord = part->coord.refValue();
			int length = part->coord.length;
			{
				for(int i=0; i<length; i++){
					point[c++] = matrix * coord[i];
				}
			}
			{
				for(int i = 0; i < triLen; ){
					coordIndex[j++] = triangle.refPart(i)->fromCoord + triangle.refIndex(i);
					i++;
					coordIndex[j++] = triangle.refPart(i)->fromCoord + triangle.refIndex(i);
					i++;
					coordIndex[j++] = triangle.refPart(i)->fromCoord + triangle.refIndex(i);
					i++;
					coordIndex[j++] = -1;
				}
			}
		}
	}

	Seamless* smls = partList[0]->smls;
	if(smls->buildNormals || smls->surfaceNormals){
		vector.length=point.length;
		for(int n=0, k=0; k<partList.length; k++){
			Art* part = partList[k];
			M_Vec3f& normal = part->normal.refValue();
			if(dynamic_cast<ImmediatePart*>(part)){
				part->genNormals();
			}
			n += normal.length;
		}
	}
	else{
		genNormals();
		for(int n=0, k=0; k<partList.length; k++){
			Art* part = partList[k];
			M_Vec3f& normal = part->normal.refValue();
			for(int i = 0, j = n; i<normal.length; ){
				normal[i++] = vector[j++];
			}
			n += normal.length;
		}
	}
}

void IndexedTriangleSet::genTransformedNormals(M_Art& partList){
	M_Vec3f& point = refPoint();
	M_Vec3f& vector = refVector();
	M_Int& index=this->index.refValue();

	{
		int coordLen=0;
		int coordIndexLen=0;
		for(int k=0; k<partList.length; k++){
			Art* part=partList[k];
			int clen=part->coord.length;
			part->texCoord.length=clen;
			part->color.length=clen;
			part->normal.refValue().length=clen;
			part->fromCoord=coordLen;
			M_Int& localCoordLookUp=part->partTree;
			localCoordLookUp.length=clen;
			coordLen+=clen;
			part->toCoord=clen;
			coordIndexLen+=part->triangle.length;
			for(int i=clen; --i>=0; ){
				localCoordLookUp[i]=-1;
			}
		}
		coordIndexLen*=3;
		index.length=coordIndexLen;
		point.length=coordLen;
	}
	{
		for(int j=0, c=0, k=0; k<partList.length; k++){
			Art* part=partList[k];
			M_Triangle& triangle=part->triangle;
			int triLen=triangle.length*3;
			Matrix matrix=part->matrix;
			M_Vec3f& coord=part->coord.refValue();
			int length=part->coord.length;
			{
				for(int i=0; i<length; i++){
					point[c++]=matrix*coord[i];
				}
			}
			{
				for(int i=0; i<triLen; ){
					index[j++]=triangle.refPart(i)->fromCoord+triangle.refIndex(i);
					i++;
					index[j++]=triangle.refPart(i)->fromCoord+triangle.refIndex(i);
					i++;
					index[j++]=triangle.refPart(i)->fromCoord+triangle.refIndex(i);
					i++;
				}
			}
		}
	}
	genNormals();
	{
		for(int n=0, k=0; k<partList.length; k++){
			M_Vec3f& normal=partList[k]->normal.refValue();
			for(int i=0; i<normal.length; ){
				normal[i++]=vector[n++];
			}
		}
	}
}
void Mesh::genTransformedNormals(M_Art& partList){
	M_Vec3f& point = coord.refValue();
	M_Vec3f& vector = normal.refValue();
	M_Int& coordIndex=this->coordIndex.refValue();

	{
		int coordLen=0;
		int coordIndexLen=0;
		for(int k=0; k<partList.length; k++){
			Art* part=partList[k];
			int clen=part->coord.length;
			part->texCoord.length=clen;
			part->color.length=clen;
			part->normal.refValue().length=clen;
			part->fromCoord=coordLen;
			M_Int& localCoordLookUp=part->partTree;
			localCoordLookUp.length=clen;
			coordLen+=clen;
			part->toCoord=clen;
			coordIndexLen+=part->triangle.length;
			for(int i=clen; --i>=0; ){
				localCoordLookUp[i]=-1;
			}
		}
		coordIndexLen*=3;
		coordIndex.length=coordIndexLen;
		point.length=coordLen;
	}
	{
		for(int j=0, c=0, k=0; k<partList.length; k++){
			Art* part=partList[k];
			M_Triangle& triangle=part->triangle;
			int triLen=triangle.length*3;
			Matrix matrix=part->matrix;
			M_Vec3f& coord=part->coord.refValue();
			int length=part->coord.length;
			{
				for(int i=0; i<length; i++){
					point[c++]=matrix*coord[i];
				}
			}
			{
				for(int i=0; i<triLen; ){
					coordIndex[j++]=triangle.refPart(i)->fromCoord+triangle.refIndex(i);
					i++;
					coordIndex[j++]=triangle.refPart(i)->fromCoord+triangle.refIndex(i);
					i++;
					coordIndex[j++]=triangle.refPart(i)->fromCoord+triangle.refIndex(i);
					i++;
				}
			}
		}
	}
	genNormals();
	{
		for(int n=0, k=0; k<partList.length; k++){
			M_Vec3f& normal=partList[k]->normal.refValue();
			for(int i=0; i<normal.length; ){
				normal[i++]=vector[n++];
			}
		}
	}
}

void Art::workoutVerticesForGroupCoords(){
	static MM_Int32 rvcg;
	M_Int vcg;

	vcg = legalCoordGroup;
	{
		int c = 0;
		for(int i=vcg.length; --i>=0; ){
			c += vcg[i];
		}
		vcg += toCoord-c;
	}

	//at this stage vCoordGroup = the extra vertex (that were made for normals) coordinate's sub index value. vcg is filling in for vCoordGroup for now

	rvcg.length = vcg.length;

	int j=0;
	for(int r=0; r<vcg.length; r++){
		M_Int& m=rvcg[r];
		m.length=vcg[r];
		for(int i=0; i<m.length; i++){
			m[i]=j++;
		}
	}
	{
		for(int i=0; i<vCoordGroup.length; i++){
			rvcg[vCoordGroup[i]] += toCoord + i;
			vcg[vCoordGroup[i]]++;
		}
	}
	vCoordGroup=vcg;
	{
		scratchC.length=0;
		for(int i=0; i<rvcg.length; i++){
			scratchC+=rvcg[i];		//scratch redirects the coordinates for the coordinates
		}
	}
	{
		partTree.length=scratchC.length;
		for(int i=0; i<partTree.length; i++){
			partTree[scratchC[i]]=i;		//partTree redirects the coordinates for the triangles
		}
	}
	if(scratchC.length != coord.length || partTree.length != coord.length){
		throw;
	}
}

void IndexedFaceSet::addExtraCoords(Seamless* smls){
	static M_Int lookUpCoord;

	M_Art& partList = smls->partList;
	M_Int& coordIndex=this->coordIndex.refValue();
	if(partList.length==0) return;
	Effect* workEffect=partList[0]->workEffect;
	if(workEffect->getType()==COLOREFFECT){
		lookUpCoord.length=dynamic_cast<ColorEffect*>(workEffect)->color.length;
	}
	else if(workEffect->getType()==MONOCOLOREFFECT){
		return;
	}
	else if(workEffect->getType()==TEXTUREEFFECT){
		lookUpCoord.length=dynamic_cast<TextureEffect*>(workEffect)->texCoord.length;
	}
	else throw partList[0]->workEffect;
	{
		for(int i=lookUpCoord.length; --i>=0; ){
			lookUpCoord[i]=-1;
		}
	}
	{
		for(int k = 0; k<partList.length; k++){
			Art* part=partList[k];
			M_Int& scratchC=part->scratchC;
			scratchC.length=part->coord.length;
			const M_Int& coordGroup=part->coordGroup;
			part->vCoordGroup.length=0;
			part->rc.length=0;
			{
				int j=0,g=0;
				for(; g<coordGroup.length; g++ ){
					for(int i=0; i<coordGroup[g]&&j<scratchC.length; i++){
						scratchC[j++]=g;
					}
				}
				while(j<scratchC.length){
					scratchC[j++]=g;
				}

			}
		}
	}
	/*
	{
		int accumLength=0;
		int coordIndexLen=0;
		for(int k=0; k<partList.length; k++){
			Art* part=partList[k];
			part->fromCoord=accumLength;
			accumLength+=part->coord.length;
			part->toCoord=accumLength;
			coordIndexLen+=part->triangle.length;
			M_Int& localCoordLookUp=part->partTree;
			localCoordLookUp.resize(part->coord.length);
			for(int i=localCoordLookUp.length; --i>=0; ){
				localCoordLookUp[i]=-1;
			}
		}
		coordIndexLen*=4;
		coordIndex.resize(coordIndexLen);
	}
	*/
	if(workEffect->getType()==COLOREFFECT){
		const M_Color3f& color=dynamic_cast<ColorEffect*>(workEffect)->color;
		for(int j=0, k=0; k<partList.length; k++){
			Art* part=partList[k];
			M_Triangle& triangle=part->triangle;
			int triLen=triangle.length*3;
			for(int o=3, i=0; i<triLen; i++){
				int linTci=triangle.refTexIndex(i);
				int ci=triangle.refIndex(i);
				Art* cornerPart=triangle.refPart(i);
				M_Int& localCoordLookUp=cornerPart->partTree;
				if( lookUpCoord[linTci]==-1){
					if( localCoordLookUp[ci]==-1){
						localCoordLookUp[ci]=linTci;
						cornerPart->color[ci]=color[linTci];
						lookUpCoord[linTci]=ci;
						coordIndex[j++]=ci;
					}
					else {
						coordIndex[j++]=lookUpCoord[linTci]=cornerPart->coord.length;
						cornerPart->color+=color[linTci];
						cornerPart->rc+=ci;
						cornerPart->coord.refValue()+=cornerPart->coord[ci];
						cornerPart->normal.refValue()+=cornerPart->normal[ci];
						cornerPart->vCoordGroup+=cornerPart->scratchC[ci];
					}
				}
				else {
					coordIndex[j++]=lookUpCoord[linTci];
				}
				if(--o==0){
					o=3;
					j++;//coordIndex[j++]=-1;
				}
			}
		}
	}
	else if(workEffect->getType()==TEXTUREEFFECT){
		const M_Vec2f& texCoord=dynamic_cast<TextureEffect*>(workEffect)->texCoord;
		for(int j=0, k=0; k<partList.length; k++){
			Art* part=partList[k];
			M_Triangle& triangle=part->triangle;
			int triLen=triangle.length*3;
			for(int o=3, i=0; i<triLen; i++){
				int linTci=triangle.refTexIndex(i);
				int ci=triangle.refIndex(i);
				Art* cornerPart=triangle.refPart(i);
				M_Int& localCoordLookUp=cornerPart->partTree;
				if( lookUpCoord[linTci]==-1){
					if( localCoordLookUp[ci]==-1){
						localCoordLookUp[ci]=linTci;
						cornerPart->texCoord[ci]=texCoord[linTci];
						lookUpCoord[linTci]=ci;
						coordIndex[j++]=ci;
					}
					else {
						coordIndex[j++]=lookUpCoord[linTci]=cornerPart->coord.length;
						cornerPart->texCoord+=texCoord[linTci];
						cornerPart->rc+=ci;
						cornerPart->coord.refValue()+=cornerPart->coord[ci];
						cornerPart->normal.refValue()+=cornerPart->normal[ci];
						cornerPart->vCoordGroup+=cornerPart->scratchC[ci];
					}
				}
				else {
					coordIndex[j++]=lookUpCoord[linTci];
				}
				if(--o==0){
					o=3;
					j++;//coordIndex[j++]=-1;
				}
			}
		}
	}
	else throw workEffect;
	smls->resetFromCoordAndWorkoutVerticesForGroupCoords();
	{
		for(int j=0, k=0; k<partList.length; k++){
			Art* part=partList[k];
			M_Triangle& triangle=part->triangle;
			int triLen=triangle.length*3;
			for(int i=0; i<triLen; ){
				Art* triPart;
				triPart=triangle.refPart(i++);
				coordIndex[j]=triPart->partTree[coordIndex[j]]+triPart->fromCoord;
				j++;
				triPart=triangle.refPart(i++);
				coordIndex[j]=triPart->partTree[coordIndex[j]]+triPart->fromCoord;
				j++;
				triPart=triangle.refPart(i++);
				coordIndex[j]=triPart->partTree[coordIndex[j]]+triPart->fromCoord;
				j++;

				//coordIndex[j++]+=triangle.refPart(i++)->fromCoord;
				//coordIndex[j++]+=triangle.refPart(i++)->fromCoord;
				//coordIndex[j++]+=triangle.refPart(i++)->fromCoord;
				j++;
			}
		}
	}
}
void IndexedTriangleSet::addExtraCoords(Seamless* smls){
	M_Art& partList = smls->partList;
	static M_Int lookUpCoord;
	M_Int& index=this->index.refValue();
	if(partList.length==0) return;
	Effect* workEffect=partList[0]->workEffect;
	if(workEffect->getType()==COLOREFFECT){
		lookUpCoord.length=dynamic_cast<ColorEffect*>(workEffect)->color.length;
	}
	else if(workEffect->getType()==TEXTUREEFFECT){
		lookUpCoord.length=dynamic_cast<TextureEffect*>(workEffect)->texCoord.length;
	}
	else throw partList[0]->workEffect;
	{
		for(int i=lookUpCoord.length; --i>=0; ){
			lookUpCoord[i]=-1;
		}
	}
	{
		for(int k = 0; k < partList.length; k++){
			Art* part=partList[k];
			M_Int& scratchC=part->scratchC;
			scratchC.length=part->coord.length;
			const M_Int& coordGroup=part->coordGroup;
			part->vCoordGroup.length=0;
			part->rc.length=0;
			{
				int j=0,g=0;
				for(; g<coordGroup.length; g++ ){
					for(int i=0; i<coordGroup[g]&&j<scratchC.length; i++){
						scratchC[j++]=g;
					}
				}
				while(j<scratchC.length){
					scratchC[j++]=g;
				}

			}
		}
	}
	if(workEffect->getType()==COLOREFFECT){
		const M_Color3f& color=dynamic_cast<ColorEffect*>(workEffect)->color;
		for(int j=0, k=0; k<partList.length; k++){
			Art* part=partList[k];
			M_Triangle& triangle=part->triangle;
			int triLen=triangle.length*3;
			for(int o=3, i=0; i<triLen; i++){
				int linTci=triangle.refTexIndex(i);
				int ci=triangle.refIndex(i);
				Art* cornerPart=triangle.refPart(i);
				M_Int& localCoordLookUp=cornerPart->partTree;
				if( lookUpCoord[linTci]==-1){
					if( localCoordLookUp[ci]==-1){
						localCoordLookUp[ci]=linTci;
						cornerPart->color[ci]=color[linTci];
						lookUpCoord[linTci]=ci;
						index[j++]=ci;
					}
					else {
						index[j++]=lookUpCoord[linTci]=cornerPart->coord.length;
						cornerPart->color+=color[linTci];
						cornerPart->rc+=ci;
						cornerPart->coord.refValue()+=cornerPart->coord[ci];
						cornerPart->normal.refValue()+=cornerPart->normal[ci];
						cornerPart->vCoordGroup+=cornerPart->scratchC[ci];
					}
				}
				else {
					index[j++]=lookUpCoord[linTci];
				}
				if(--o==0){
					o=3;
				}
			}
		}
	}
	else if(workEffect->getType()==TEXTUREEFFECT){
		const M_Vec2f& texCoord=dynamic_cast<TextureEffect*>(workEffect)->texCoord;
		for(int j=0, k=0; k<partList.length; k++){
			Art* part=partList[k];
			M_Triangle& triangle=part->triangle;
			int triLen=triangle.length*3;
			for(int o=3, i=0; i<triLen; i++){
				int linTci=triangle.refTexIndex(i);
				int ci=triangle.refIndex(i);
				Art* cornerPart=triangle.refPart(i);
				M_Int& localCoordLookUp=cornerPart->partTree;
				if( lookUpCoord[linTci]==-1){
					if( localCoordLookUp[ci]==-1){
						localCoordLookUp[ci]=linTci;
						cornerPart->texCoord[ci]=texCoord[linTci];
						lookUpCoord[linTci]=ci;
						index[j++]=ci;
					}
					else {
						index[j++]=lookUpCoord[linTci]=cornerPart->coord.length;
						cornerPart->texCoord+=texCoord[linTci];
						cornerPart->rc+=ci;
						cornerPart->coord.refValue()+=cornerPart->coord[ci];
						cornerPart->normal.refValue()+=cornerPart->normal[ci];
						cornerPart->vCoordGroup+=cornerPart->scratchC[ci];
					}
				}
				else {
					index[j++]=lookUpCoord[linTci];
				}
				if(--o==0){
					o=3;
				}
			}
		}
	}
	else throw workEffect;
	smls->resetFromCoordAndWorkoutVerticesForGroupCoords();
	{
		for(int j=0, k=0; k<partList.length; k++){
			Art* part=partList[k];
			M_Triangle& triangle=part->triangle;
			int triLen=triangle.length*3;
			for(int i=0; i<triLen; ){
				Art* triPart;
				triPart=triangle.refPart(i++);
				index[j]=triPart->partTree[index[j]]+triPart->fromCoord;
				j++;
				triPart=triangle.refPart(i++);
				index[j]=triPart->partTree[index[j]]+triPart->fromCoord;
				j++;
				triPart=triangle.refPart(i++);
				index[j]=triPart->partTree[index[j]]+triPart->fromCoord;
				j++;

			}
		}
	}
}

void genHAnimSkinCoordWeight(Art* parentPart, HAnimJointPtr parentJoint, int& skinCoordIndexCntr, const bool usingPivotPoints){
	M_SArt& partField=parentPart->part.refValue();
	M_Node& children=parentJoint->children.refValue();
	for(int k=0; k<partField.length;k++){
		Art* part=partField[k];
		if(part->show==false&&part->relevantPart.length==1)continue;
		HAnimJoint* joint=dynamic_cast<HAnimJoint*>(children[k].p);
		if(usingPivotPoints){
			joint->center=part->absolutePivotPoint;
		}
		else{
			joint->center=part->matrix*Vec3f();
		}

		M_Int& parentSkinCoordIndex=parentJoint->skinCoordIndex.refValue();
		M_Float& parentSkinCoordWeight=parentJoint->skinCoordWeight.refValue();
		M_Int& skinCoordIndex=joint->skinCoordIndex.refValue();
		M_Float& skinCoordWeight=joint->skinCoordWeight.refValue();
		skinCoordWeight.length=part->coord.length;
		skinCoordIndex.length=part->coord.length;
		M_Float& weight=part->weight.refValue();
		M_Int& vCoordGroup=part->vCoordGroup;
		int i=0;
		int h=0;
		float wt;
		for(int  j=0; j<vCoordGroup.length; j++){
			h+=vCoordGroup[j];
			if(j<weight.length){
				wt=weight[j];
			}
			else{
				wt=1;
			}
			for( ; i<h; i++){
				skinCoordWeight[i]=wt;
				skinCoordIndex[i]=skinCoordIndexCntr;
				if(wt!=1){
					parentSkinCoordIndex+=skinCoordIndexCntr;
					parentSkinCoordWeight+=1-wt;
				}
				skinCoordIndexCntr++;
			}
		}
		{
			const M_Morpher& m_morpher=part->morpher;
			for(int i=0; i<m_morpher.length; i++){
				const Morpher* morpher=m_morpher[i];
				HAnimDisplacer* hAnimDisplacer=new HAnimDisplacer;
				hAnimDisplacer->name=dynamic_cast<Node*>(hAnimDisplacer)->name=morpher->name;
				hAnimDisplacer->displacements.refValue()=morpher->displacement;
				hAnimDisplacer->coordIndex.refValue()=joint->skinCoordIndex;
				joint->displacers.refValue()+=hAnimDisplacer;
			}
		}
		genHAnimSkinCoordWeight(part, joint, skinCoordIndexCntr, usingPivotPoints);
	}
}
String generateMatrixScript(M_Int& partTree){
	String str;
	int greatist=0;
	{
		for(int i=0; i<partTree.length; i+=3){
			if(partTree[i]>greatist){
				greatist=partTree[i];
			}
		}
	}
	{
		for(int i=greatist+1; --i>=0;){
			str+='c';
			str+=i;
			str+="=new VrmlMatrix();\n";
		}
	}
	if(g_genNormals){
		for(int i=greatist+1; --i>=0;){
			str+='n';
			str+=i;
			str+="=new VrmlMatrix();\n";
		}
		str+="q=new SFVec3f(0,0,0);\n";
	}

	for(int i=0; i<partTree.length; i+=3){
		int matrixStack=partTree[i];
		str+='c';
		str+=matrixStack;
		str+=".setTransform(translatePart[";
		str+=i/3;
		str+="],o[";
		str+=i/3;
		str+="]);\n";
		if(g_genNormals){
			str+='n';
			str+=matrixStack;
			str+=".setTransform(q,o[";
			str+=i/3;
			str+="]);\n";
		}
		if(matrixStack>0){
			str+='c';
			str+=matrixStack;
			str+="=c";
			str+=matrixStack-1;
			str+=".multLeft(c";
			str+=matrixStack;
			str+=");\n";
			if(g_genNormals){
				str+='n';
				str+=matrixStack;
				str+="=n";
				str+=matrixStack-1;
				str+=".multLeft(n";
				str+=matrixStack;
				str+=");\n";
			}
		}
		str+="for(i=";
		str+=partTree[i+1];
		str+=";i<";
		str+=partTree[i+2];
		if(g_genNormals){
			str+=";i++){dcrd[i]=c";
			str+=matrixStack;
			str+=".multVecMatrix(scrd[i]);";
			str+="dnrm[i]=n";
			str+=matrixStack;
			str+=".multVecMatrix(snrm[i]);}\n";
		}
		else{
			str+=";i++)dcrd[i]=c";
			str+=matrixStack;
			str+=".multVecMatrix(scrd[i]);\n";
		}
	}
	str+="crd.point=dcrd;";
	if(g_genNormals){
		str+="\nnrm.vector=dnrm;";
	}
	return str;
}

String addTransformScript(Seamless* smls, M_Int& partTree){
	String url;
	url+="\nfunction eventsProcessed(){\n";
	if(g_genContactSBVT){
		if(smls->addScaleField){
			url+="crd.point.setByVertexTransform(dcrd, partTree, translatePart, o, scalePart);";
		}
		else{
			url+="crd.point.setByVertexTransform(dcrd, partTree, translatePart, o);";
		}
		if(g_genNormals){
			url+="\nnrm.vector.setByVertexTransform(snrm, partTree, blankTranslate, o);";
		}
	}
	else{
		url+=generateMatrixScript(partTree);
	}

	url+="\n}\nfunction initialize(){\n\
	if(scrd.length==0){\n\
		scrd=crd.point;\n\
		dcrd=scrd;";

	if(g_genNormals){
		url+="\tsnrm=nrm.vector;\n\tdnrm=snrm;";
	}
	url+="\n\t}\n\tblankTranslate.length=translatePart.length;\n\
	o.length=translatePart.length;\n\
	eventsProcessed();\n\
}\n";
	return url;
}
void addMorpherToScript(Script*& script, M_Properties& properties, const Art*& part, Morpher* morpher, String& str){
	M_Vec3f btsDisplacement=morpher->displacement;

	const M_Int& rc=part->rc;
	for(int i=0; i<rc.length; i++){
		btsDisplacement+=btsDisplacement[rc[i]];
	}

	String displacementStr="d";
	makeNodeNameFromAddr(displacementStr, morpher);

	String weightStr="w";
	makeNodeNameFromAddr(weightStr, morpher);

	*script+=new MFVec3f(btsDisplacement);
	properties+=Properties(displacementStr);
	*script+=new SFFloat(0);
	properties+=Properties(weightStr);

	str+="function ";
str+=weightStr;
str+="(v){\n\
	for(j=";
str+=part->fromCoord+btsDisplacement.length;
str+=", i=";
str+=btsDisplacement.length;
str+="; --i>=0; ){\n\
		--j;\n\
		dcrd[j]=scrd[j].add(";
str+=displacementStr;
str+="[i].multiply(v));\n\
	}\n\
}\n";
}
void IndexedFaceSet::groupTrianglesByEffect(Art* art, M_Node& m_shape, int& triPos, Effect* lastEffect){
	if(art->show==false&&art->relevantPart.length==1)return;

	Effect* effect=art->effect;
	if( effect!=nullPtr){
		lastEffect=effect;
	}
	Shape* outputShape=lastEffect->outputShape;

	if( effect && effect->outputShape == nullPtr ){
		outputShape=new Shape;
		m_shape+=outputShape;
		effect->copyFieldsData(outputShape);
		IndexedFaceSet* geometry=new IndexedFaceSet;
		if(g_genNormals==false){
			geometry->creaseAngle=3.14;
		}

		outputShape->geometry=geometry;
		geometry->coord=coord;
		geometry->normal=normal;
		effect->addColorNode(geometry, this);
	}

	int ciLen=art->triangle.length*4;

	IndexedFaceSet* outputGeometry=dynamic_cast<IndexedFaceSet*>(outputShape->geometry.getRawPtr());

	if(outputGeometry!=this){

		M_Int& oCoordIndex=outputGeometry->coordIndex;

		int oOldLen=oCoordIndex.length;
		oCoordIndex.length+=ciLen;

		for(int i=0; i<ciLen; i++){
			oCoordIndex[oOldLen+i]=coordIndex[triPos+i];
		}
		if(colorIndex.length){
			M_Int& oColorIndex=outputGeometry->colorIndex;
			oColorIndex.length+=ciLen;
			for(int i=0; i<ciLen; i++){
				oColorIndex[oOldLen+i]=colorIndex[triPos+i];
			}
		}
		if(texCoordIndex.length){
			M_Int& oTexCoordIndex=outputGeometry->texCoordIndex;
			oTexCoordIndex.length+=ciLen;
			for(int i=0; i<ciLen; i++){
				oTexCoordIndex[oOldLen+i]=texCoordIndex[triPos+i];
			}
		}
		if(ciLen){
			coordIndex.refValue()[triPos]=-ciLen;
		}
	}

	triPos+=ciLen;

	M_SArt& part=art->part.refValue();
	for(int k=0; k<part.length; k++){
		groupTrianglesByEffect(part[k], m_shape, triPos, lastEffect);
	}
}
void IndexedTriangleSet::groupTrianglesByEffect(Art* art, M_Node& m_shape, int& triPos, Effect* lastEffect){
	if(art->show==false&&art->relevantPart.length==1)return;

	Effect* effect=art->effect;
	if( effect!=nullPtr){
		lastEffect=effect;
	}
	Shape* outputShape=lastEffect->outputShape;

	if( effect!=nullPtr && effect->outputShape == nullPtr ){
		outputShape=new Shape;
		m_shape+=outputShape;
		effect->copyFieldsData(outputShape);
		IndexedTriangleSet* geometry=new IndexedTriangleSet;
		outputShape->geometry=geometry;
		geometry->coord=coord;
		geometry->normal=normal;
		effect->addColorNode(geometry, this);
	}

	int ciLen=art->triangle.length*3;

	IndexedTriangleSet* outputGeometry=dynamic_cast<IndexedTriangleSet*>(outputShape->geometry.getRawPtr());

	if(outputGeometry!=this){

		M_Int& oIndex=outputGeometry->index;

		int oOldLen=oIndex.length;
		oIndex.length+=ciLen;

		for(int i=0; i<ciLen; i++){
			oIndex[oOldLen+i]=index[triPos+i];
		}
		if(ciLen){
			index.refValue()[triPos]=-ciLen;
		}
	}

	triPos+=ciLen;

	M_SArt& part=art->part.refValue();
	for(int k=0; k<part.length; k++){
		groupTrianglesByEffect(part[k], m_shape, triPos, lastEffect);
	}
}
void IndexedFaceSet::deleteUnwantedTriFromSource(){

	int subLen=0;
	for(int i=0, j=0; i<coordIndex.length && j<coordIndex.length; ){
		int b=coordIndex[j];
		if(b<0){
			subLen-=b;
			j-=b;
			continue;
		}
		for(int h=4; --h>=0; ){
			if(colorIndex.length){
				colorIndex.refValue()[i]=colorIndex[j];
			}
			if(texCoordIndex.length){
				texCoordIndex.refValue()[i]=texCoordIndex[j];
			}
			coordIndex.refValue()[i++]=coordIndex[j++];
		}
	}
	coordIndex.refValue().length-=subLen;
	if(colorIndex.length){
		colorIndex.refValue().length=coordIndex.length;
	}
	if(texCoordIndex.length){
		texCoordIndex.refValue().length=coordIndex.length;
	}
}
void IndexedTriangleSet::deleteUnwantedTriFromSource(){

	int subLen=0;
	for(int i=0, j=0; i<index.length && j<index.length; ){
		int b=index[j];
		if(b<0){
			subLen-=b;
			j-=b;
			continue;
		}
		for(int h=3; --h>=0; ){
			index.refValue()[i++]=index[j++];
		}
	}
	index.refValue().length-=subLen;
}
float sumUpColor(const Color3f& color){
	return (color.r+color.g+color.b)/3;
}
void copyMultiUrlWithSingleUrl(M_String& m_url, const String& url){
	if(url.length){
		m_url.length = 0;
		m_url += url;
	}
}
void copySingleUrlWithMultiUrl(String& url, const M_String& m_url){
	if(m_url.length){
		url = m_url[0];
	}
}
void TextureEffect::copyFieldsData(Shape* shape){
	outputShape=shape;

	Appearance* appearance=new Appearance;
	shape->appearance=appearance;

	ImageTexture* texture=new ImageTexture;
	appearance->texture=texture;
	copyMultiUrlWithSingleUrl(texture->url.refValue(), this->url);

	Material* material=new Material;
	material->name=this->name;
	g_scene->removeFromDefList(this); //needed to stop USE but why this way?
	appearance->material=material;

	material->ambientIntensity=sumUpColor(this->ambient);
	material->emissiveColor=this->emissive;
	material->shininess=this->shininess;
	material->specularColor=this->specular;
	material->transparency=this->transparency;
}
void ColorEffect::copyFieldsData(Shape* shape){
	outputShape=shape;

	Appearance* appearance=new Appearance;
	shape->appearance=appearance;

	Material* material=new Material;
	material->name=this->name;
	g_scene->removeFromDefList(this); //needed to stop USE but why this way?
	appearance->material=material;

	material->ambientIntensity=sumUpColor(this->ambient);
	material->diffuseColor=this->diffuse;
	material->emissiveColor=this->emissive;
	material->shininess=this->shininess;
	material->specularColor=this->specular;
	material->transparency=this->transparency;
}
void MonoColorEffect::copyFieldsData(Shape* shape){
	outputShape=shape;

	Appearance* appearance=new Appearance;
	shape->appearance=appearance;

	Material* material=new Material;
	material->name=this->name;
	g_scene->removeFromDefList(this); //needed to stop USE but why this way?
	appearance->material=material;

	material->diffuseColor=this->diffuse;
	material->ambientIntensity=sumUpColor(this->ambient);
	material->emissiveColor=this->emissive;
	material->shininess=this->shininess;
	material->specularColor=this->specular;
	material->transparency=this->transparency;
}
void Art::checkUsingPivotPoints(bool& state){
	if(pivotPoint != Vec3f()){
		state = true;
		return;
	}
	M_SArt& m_part = part.refValue();
	for(int k = 0; k < m_part.length && state == false; k++){
		m_part[k]->checkUsingPivotPoints(state);
	}
}

void Art::convertPivotPointsToTranslationsNoChangePivotPoint(Vec3f accumTran){
	accumTran += pivotPoint;
	translation.refValue() += pivotPoint;
	M_Vec3f& coord = this->coord;
	coord.translate(-accumTran);
//	part->pivotPoint.refValue()=Vec3f();
//	part->absolutePivotPoint=Vec3f();
	M_SArt& m_part = part.refValue();
	for(int k = 0; k < m_part.length; k++){
		m_part[k]->convertPivotPointsToTranslationsNoChangePivotPoint(accumTran);
	}
}
void outO3D(String& str, VertexGeometry* geometry, Color* color){
	M_Color3f& m_color = color->color.refValue();
	int len = m_color.length;
	str+="\nmeshPrimitive.numberVertices = ";
	str+=len;
	IndexedTriangleSet* its = dynamic_cast<IndexedTriangleSet*>(geometry);
	if(its == 0){
		return;
	}
	Coordinate* coord = dynamic_cast<Coordinate*>(its->coord.p);
	if(coord == 0){
		return;
	}
	M_Vec3f& point = coord->point;

	Normal* normal = dynamic_cast<Normal*>(its->normal.p);
	if(normal == 0){
		return;
	}
	M_Vec3f& vector = normal->vector;


	M_Int& m_index = its->index.refValue();
	int indexLen = m_index.length;

	str+=";\nmeshPrimitive.numberPrimitives = ";
	str+=indexLen/3;
	str+=";\nmeshPrimitive.createDrawElement(g_pack, null);\n";

	str+="var positionArray = [";
	if(len){
		for(int i = 0;;){
			str.cat(point[i]);
			if(++i >= len){
				break;
			}
			str+=',';
		}
	}
	str+="];\nvar normalArray = [";
	if(len){
		for(int i = 0;;){
			str.cat(vector[i]);
			if(++i >= len){
				break;
			}
			str+=',';
		}
	}
	str+="];\nvar colorArray = [";
	if(len){
		for(int i = 0;;){
			str.catColor4f(m_color[i]);
			if(++i >= len){
				break;
			}
			str+=',';
		}
	}

	str+="];\nvar indicesArray = [";
	if(indexLen){
		for(int i = 0;;){
			str+= m_index[i];
			if(++i >= indexLen){
				break;
			}
			str+=',';
		}
	}
	str+="];\n";
}


/*
template<> int& M_LocalTriangle::refIndex(int i)const{
	if((unsigned)i>=length.rawLen*3)throw this;
	return ((LocalVertexIndex*)length.pointer)[i].i;
}
template<> int& M_LocalTriangle::refTexIndex(int i)const{
	if((unsigned)i>=length.rawLen*3)throw this;
	return ((LocalVertexIndex*)length.pointer)[i].t;
}
*/
inline bool moreThanOneReferenceToVertex(Art* vertexPart, int vertexIdx){
	int count=0;
	M_Art& relevantPart=vertexPart->relevantPart;
	for(int k=relevantPart.length; --k>=0;){
		Art* part=relevantPart[k];
		M_Triangle& triangle=part->triangle;
		for(int i=triangle.length*3; --i>=0; ){
			if(triangle.refPart(i)==vertexPart && triangle.refIndex(i)==vertexIdx){
				if(++count>1)return true;
			}
		}
	}
	return false;
}
void Art::fastDeleteTriangle(int poly){
	if(poly<0 || poly >= triangle.length) throw this;

	for(int i=poly;i<triangle.length-1;i++){
		triangle[i]=triangle[i+1];
	}
	triangle.length--;
}

void Art::deleteTriangle(int poly){
	if(poly<0 || poly >= triangle.length) return;

	for(int c=3; --c>=0; ){
		Art* part=triangle.refPart(poly*3+c);
		int index=triangle.refIndex(poly*3+c);
		if(moreThanOneReferenceToVertex(part, index)==false){
			part->deleteVertex(index);
		}
	}
	fastDeleteTriangle(poly);
}

void Art::deleteVertex(int vertexIndex){

	fastDeleteVertex(vertexIndex);

	for( int k=relevantPart.length; --k>=0; ){
		M_Triangle& triangle=relevantPart[k]->triangle;
		for(int i=triangle.length*3; --i>=0;){
			if(triangle.refPart(i)==this && triangle.refIndex(i)>=vertexIndex){
				triangle.refIndex(i)--;
			}
		}
	}
	removeNonRelevantParts(this);
}



void swapCPPointStorage(NurbsPatch* patch){
	M_Vec3f& tpoint = patch->tpoint;
	M_Vec3f& point = patch->point.refValue();

	for(int k = tpoint.length; --k >= 0;){
		Vec3f v = tpoint[k];
		tpoint[k] = point[k];
		point[k] = v;
	}
}

void storePoints(M_<NurbsPatch*>& cpNodeList){
	for(int k = 0; k < cpNodeList.length; k++){
		NurbsPatch* patch = cpNodeList[k];
		patch->tpoint = patch->point;
	}
}


//bool g_cpNodeListStrandCages = false;  //false = add strand geometry to cpNodeList //true = add strand cages to cpNodeList

void Hair::cpNodeListAdd(Seamless* smls){
	M_Strand& m_strand = this->strand.refValue();
	for(int i = 0; i < m_strand.length; i++){
		Strand* strand = m_strand[i];
		strand->cpNodeListAdd(smls);
		strand->geometry->cpNodeListAdd(smls);
	}
	for(int i = 0; i < lockList.length; i++){
		NurbsPatch* patch = lockList[i];
		patch->cpNodeListAdd(smls);
	}
}
void BendyLathe::cpNodeListAdd(Seamless* smls){
	smls->cpNodeList += patch;
}

void RectangleBuild::cpNodeListAdd(Seamless* smls){
	smls->cpNodeList += patch;
}

void BoxBuild::cpNodeListAdd(Seamless* smls){
	int len = m_patch.length;
	for(int k = 0; k < len; k++){
		smls->cpNodeList += m_patch[k];
	}
}

void BezierLatheBuild::cpNodeListAdd(Seamless* smls){
	for(int k = m_patch.length; --k>=0;){
		smls->cpNodeList += m_patch[k];
	}
}

void TorusBuild::cpNodeListAdd(Seamless* smls){
	smls->cpNodeList += patch;
}

void ConeBuild::cpNodeListAdd(Seamless* smls){
	int len = m_patch.length;
	for(int k = 0; k < len; k++){
		smls->cpNodeList += m_patch[k];
	}
}

void CylinderBuild::cpNodeListAdd(Seamless* smls){
	int len = m_patch.length;
	for(int k = 0; k < len; k++){
		smls->cpNodeList += m_patch[k];
	}
}

void SphereBuild::cpNodeListAdd(Seamless* smls){
	smls->cpNodeList += patch;
}

void NurbsLathe::cpNodeListAdd(Seamless* smls){
	//patch->insideOut.refValue() = insideOut;
	//patch->mirror.refValue() = mirror;
	//if(insideOut){
	//	patch->setInsideOut();
	//	patch->setInsideOut3();
	//}
	smls->cpNodeList += patch;
}

void NurbsLine::cpNodeListAdd(Seamless* smls){
	smls->nurbsLine += this;
	smls->cpNodeList += this;
}

void NurbsPatch::cpNodeListAdd(Seamless* smls){
	smls->cpNodeList += this;
	smls->cpNodeListAddForStem(stem.refValue());
}

void Polymap::cpNodeListAdd(Seamless* smls){
	M_Node& surface = this->surface.refValue();
	for(int i = 0; i < surface.length; i++){
		if(NurbsPatch* patch = dynamic_cast<NurbsPatch*>(surface[i].p)){
			patch->plidx = smls->cpNodeList.length;
			smls->cpNodeList += patch;
		}
	}
}

void GroupBuild::cpNodeListAdd(Seamless* smls){
	for(int k = 0; k < build.length; k++){
		build.refValue()[k]->cpNodeListAdd(smls);
	}
}

void Art::transformPoints(Matrix matrix){

	int k = 0;
	for(int cg=0; cg < coordGroup.length; cg++){
		Matrix fracMatrix = translationMatrix(-absolutePivotPoint) * scaleMatrix(scale) * rotationMatrix(Rotation(rotation.x,rotation.y,rotation.z, rotation.angle*weight[cg])) * translationMatrix(absolutePivotPoint + translation) * matrix;

		for(int c = coordGroup[cg]; --c >= 0 && k < controlPoint.length; k++){
			NurbsPatch* patch = controlPoint[k].p;
			int idx = controlPoint[k].i;
			float weight = patch->weight[idx];
			patch->point.refValue()[idx] = fracMatrix * (patch->point[idx] / weight) * weight;
		}
	}
	matrix = translationMatrix(-absolutePivotPoint) * scaleMatrix(scale) * rotationMatrix(rotation) * translationMatrix(absolutePivotPoint + translation) * matrix;
	for(; k < controlPoint.length; k++){
		NurbsPatch* patch = controlPoint[k].p;
		int idx = controlPoint[k].i;
		float weight = patch->weight[idx];
		patch->point.refValue()[idx] = matrix * (patch->point[idx] / weight) * weight;
	}

	for(int k = 0; k < part.length; k++){
		part.refValue()[k]->transformPoints(matrix);
	}
}

void Art::transformPoints(Matrix matrix, NurbsPatch* targetPatch)const{

	int k = 0;
	for(int cg=0; cg < coordGroup.length; cg++){
		Matrix fracMatrix = translationMatrix(-absolutePivotPoint) * scaleMatrix(scale) * rotationMatrix(Rotation(rotation.x,rotation.y,rotation.z, rotation.angle*weight[cg])) * translationMatrix(absolutePivotPoint + translation) * matrix;

		for(int c = coordGroup[cg]; --c >= 0 && k < controlPoint.length; k++){
			NurbsPatch* patch = controlPoint[k].p;
			if(patch == targetPatch){
				int idx = controlPoint[k].i;
				float weight = patch->weight[idx];
				patch->point.refValue()[idx] = fracMatrix * (patch->point[idx] / weight) * weight;
			}
		}
	}
	matrix = translationMatrix(-absolutePivotPoint) * scaleMatrix(scale) * rotationMatrix(rotation) * translationMatrix(absolutePivotPoint + translation) * matrix;
	for(; k < controlPoint.length; k++){
		NurbsPatch* patch = controlPoint[k].p;
		if(patch == targetPatch){
			int idx = controlPoint[k].i;
			float weight = patch->weight[idx];
			patch->point.refValue()[idx] = matrix * (patch->point[idx] / weight) * weight;
		}
	}
	for(int k = 0; k < part.length; k++){
		part.refValue()[k]->transformPoints(matrix, targetPatch);
	}
}


void copySourceCP(M_<NurbsPatch*>& cpNodeList){
	for(int k = 0; k < cpNodeList.length; k++){
		NurbsPatch* patch = cpNodeList[k];
		patch->point = patch->tpoint;
	}
}

void fastFullBuildForAutoIn(bool inState, Seamless* smls);
void setAllIns(M_Build& m_build, const bool inState);

void Seamless::AnimCP::operator = (const bool&){

	M_<NurbsPatch*>& cpNodeList = home.cpNodeList;
	if(home.workAnimCP == false){
		setAllIns(home.build.refValue(), false);
		refValue() = true;
		home.workAnimCP = true;
		storePoints(cpNodeList);
	}
	else{
		refValue() = false;
		home.workAnimCP = false;
		for(int k = 0; k < cpNodeList.length; k++){
			NurbsPatch* patch = cpNodeList[k];
			patch->point = patch->tpoint;
			patch->tpoint.length = 0;
		}
		home.calcAllPolyMidPoints();
	}
	home.doBuild();
	home.ctrf = true;
}

void processAllPatches(M_<NurbsPatch*>& cpNodeList){
	for(int k = 0; k < cpNodeList.length; k++){
		NurbsPatch* patch = cpNodeList[k];
		patch->processFields();
	}
}


void Seamless::deleteSyncPolies(){
	static DelPoly delPoly;
	M_Int& poly = delPoly.poly.refValue();

	for(int k = partList.length; --k >= 0;){
		Art* part = partList[k];
		delPoly.part = part;
		poly.length = 0;

		M_Triangle& m_triangle = part->triangle;
		for(int k = 0; k < m_triangle.length; k++){
			for(int j = 0; j < 3; j++){
				Art* prt = m_triangle[k].refPart(j); //if poly contains a coord owned by a synced patch add to delete list;
				int idx = m_triangle[k].refIndex(j);
				if(idx < prt->patchOwner.length){
					poly += k;
					break;
				}
			}
		}
		if(poly.length){
			delPoly.doBuild();
		}
	}
}

//works but not needed
void Seamless::transferCoordOverPatchBorder(Art* part, int crdIdx){
	M_NurbsPatch& partPatchOwner = part->patchOwner;
	NurbsPatch* patch = partPatchOwner[crdIdx];
	M_Vec2f& uvCoord = part->uvCoord.refValue();
	Vec2f crd = uvCoord[crdIdx];


	if(crd.x < .0001 && transferVertexOverBorderWest(patch, crdIdx, partPatchOwner, uvCoord)){
	}
	else if(crd.x > 0.9999f && transferVertexOverBorderEast(patch, crdIdx, partPatchOwner, uvCoord)){
	}
	else if(crd.y < 0.0001f && transferVertexOverBorderSouth(patch, crdIdx, partPatchOwner, uvCoord)){
	}
	else if(crd.y > 0.9999f && transferVertexOverBorderNorth(patch, crdIdx, partPatchOwner, uvCoord)){
	}
}


bool Seamless::getTransferedCoordOverPatchBorder(const Art* part, int crdIdx, Vec2f& returnCoord, NurbsPatch*& returnPatch)const {
	const M_NurbsPatch& partPatchOwner = part->patchOwner;
	NurbsPatch* patch = partPatchOwner[crdIdx];
	const M_Vec2f& uvCoord = part->uvCoord.refValue();
	Vec2f crd = uvCoord[crdIdx];
	if(crd.x < .0001 && getTransferedCoordOverPatchBorderWest(patch, crdIdx, partPatchOwner, uvCoord, returnCoord, returnPatch)){
		return true;
	}
	else if(crd.x > 0.9999f && getTransferedCoordOverPatchBorderEast(patch, crdIdx, partPatchOwner, uvCoord, returnCoord, returnPatch)){
		return true;
	}
	else if(crd.y < 0.0001f && getTransferedCoordOverPatchBorderSouth(patch, crdIdx, partPatchOwner, uvCoord, returnCoord, returnPatch)){
		return true;
	}
	else if(crd.y > 0.9999f && getTransferedCoordOverPatchBorderNorth(patch, crdIdx, partPatchOwner, uvCoord, returnCoord, returnPatch)){
		return true;
	}
	return false;
}


void Seamless::transferBorderVertices(NurbsPatch* patch){
	for(int k = partList.length; --k >= 0;){
		Art* part = partList[k];
		M_Vec2f& uvCoord = part->uvCoord.refValue();
		M_NurbsPatch& partPatchOwner = part->patchOwner;
		for(int i = partPatchOwner.length; --i >= 0;){
			if(partPatchOwner[i] == patch){
				Vec2f crd = uvCoord[i];
				if(crd.x < 0.0001f && transferVertexOverBorderWest(patch, i, partPatchOwner, uvCoord)){
				}
				else if(crd.x > 0.9999f && transferVertexOverBorderEast(patch, i, partPatchOwner, uvCoord)){
				}
				else if(crd.y < 0.0001f && transferVertexOverBorderSouth(patch, i, partPatchOwner, uvCoord)){
				}
				else if(crd.y > 0.9999f && transferVertexOverBorderNorth(patch, i, partPatchOwner, uvCoord)){
				}
			}
		}
	}
}


void Seamless::joinBorderVertices(NurbsPatch* patch){
	for(int k = partList.length; --k >= 0;){
		M_Triangle& m_triangle = partList[k]->triangle;
		for(int j = m_triangle.getTriangleLength(); --j >= 0;){
			Art* part = m_triangle.refPart(j);
			int i = m_triangle.refIndex(j);

			M_NurbsPatch& partPatchOwner = part->patchOwner;
			if(partPatchOwner[i] == patch){
				M_Vec2f& uvCoord = part->uvCoord.refValue();
				Vec2f crd = uvCoord[i];
				if(crd.x < 0.0001f && joinBorderVerticesWest(patch, i, partPatchOwner, uvCoord, part, j)){//west
				}
				else if(crd.x > 0.9999f && joinBorderVerticesEast(patch, i, partPatchOwner, uvCoord, part, j)){//east
				}
				else if(crd.y < 0.0001f && joinBorderVerticesSouth(patch, i, partPatchOwner, uvCoord, part, j)){//south
				}
				else if(crd.y > 0.9999f && joinBorderVerticesNorth(patch, i, partPatchOwner, uvCoord, part, j)){//north
				}
			}
		}
	}
}

bool Seamless::isCloseCoord(const NurbsPatch* patch, const Vec2f coord, Art* dstPart, int dstIdx){
	for(int k = partList.length; --k >= 0;){
		M_Triangle& m_triangle = partList[k]->triangle;
		for(int j = m_triangle.getTriangleLength(); --j >= 0;){
			Art* part = m_triangle.refPart(j);
			int i = m_triangle.refIndex(j);

			M_NurbsPatch& partPatchOwner = part->patchOwner;
			if(partPatchOwner[i] == patch){
				M_Vec2f& uvCoord = part->uvCoord.refValue();
				Vec2f coord2 = uvCoord[i];
				if(abs(coord.x - coord2.x) < .0001 && abs(coord.y - coord2.y) < .0001){
					if(dstPart != part || dstIdx != j){
						joinVertex(part, dstPart, j, dstIdx);
						return true;
					}
				}
			}
		}
	}
	return false;
}

bool triangleIsOnPatchCorner(const Vec2f& crdA, const Vec2f& crdB, const Vec2f& crdC){
	if(crdA.y < 0.0001f && crdA.x < 0.0001f){ //SW
		if(crdB.y < 0.0001f && crdC.x < 0.0001f){ //S W
			return true;
		}
		if(crdC.y < 0.0001f && crdB.x < 0.0001f){ //S W
			return true;
		}
	}
	if(crdA.y < 0.0001f && crdA.x > 0.9999f){ //SE
		if(crdB.y < 0.0001f && crdC.x > 0.9999f){ //S E
			return true;
		}
		if(crdC.y < 0.0001f && crdB.x > 0.9999f){ //S E
			return true;
		}
	}
	if(crdA.y > 0.9999f && crdA.x > 0.9999f){ //NE
		if(crdB.y > 0.9999f && crdC.x > 0.9999f){ //N E
			return true;
		}
		if(crdC.y > 0.9999f && crdB.x > 0.9999f){ //N E
			return true;
		}
	}
	if(crdA.y > 0.9999f && crdA.x < 0.0001f){ //NW
		if(crdB.y > 0.9999f && crdC.x < 0.0001f){ //N W
			return true;
		}
		if(crdC.y > 0.9999f && crdB.x < 0.0001f){ //N W
			return true;
		}
	}
	return false;
}

/*
bool triangleIsOnPatchCorner(const S_Triangle& triangle){
	int i = triangle.refIndex(0);
	Art* part = triangle.refPart(0);
	Vec2f crdA = part->uvCoord[i];

	i = triangle.refIndex(1);
	part = triangle.refPart(1);
	Vec2f crdB = part->uvCoord[i];

	i = triangle.refIndex(2);
	part = triangle.refPart(2);
	Vec2f crdC = part->uvCoord[i];

	if(triangleIsOnPatchCorner(crdA, crdB, crdC)){
		return true;
	}
	if(triangleIsOnPatchCorner(crdB, crdA, crdC)){
		return true;
	}
	if(triangleIsOnPatchCorner(crdC, crdA, crdB)){
		return true;
	}
	return false;
}
*/

bool triangleIsOnPatchCorner(Vec2f coord[3]){
	if(triangleIsOnPatchCorner(coord[0], coord[1], coord[2])){
		return true;
	}
	if(triangleIsOnPatchCorner(coord[1], coord[0], coord[2])){
		return true;
	}
	if(triangleIsOnPatchCorner(coord[2], coord[0], coord[1])){
		return true;
	}
	return false;
}

bool Seamless::triangleIsTransferable(const S_Triangle& triangle, NurbsPatch* delPatch){
	NurbsPatch* prevPatch = 0;
	int sameCntr = 0;
	for(int j = 3; --j>=0;){
		int i = triangle.refIndex(j);
		Art* part = triangle.refPart(j);

		M_Vec2f& uvCoord = part->uvCoord.refValue();
		M_NurbsPatch& partPatchOwner = part->patchOwner;
		NurbsPatch* patch = partPatchOwner[i];

		if(prevPatch == patch){
			sameCntr++;
		}

		prevPatch = patch;

		Vec2f crd = uvCoord[i];

		//int uDim = patch->dim.x;
		//int vDim = patch->dim.y;

		//if(crd.x < .0001 && crd.y < .0001 && isPatchJoinedByStitchPolyCorner(patch, 0, delPatch)){//SW
		//}
		//else if(crd.x > .9999 && crd.y < .0001 && isPatchJoinedByStitchPolyCorner(patch, uDim - 1, delPatch)){//SE
		//}
		//else if(crd.x > .9999 && crd.y > .9999 && isPatchJoinedByStitchPolyCorner(patch, uDim * vDim - 1, delPatch)){//NE
		//}
		//else if(crd.x < .0001 && crd.y > .9999 && isPatchJoinedByStitchPolyCorner(patch, uDim * (vDim - 1), delPatch)){//NW
		//}

		if(crd.x < 0.0001f && transferableVertexOverBorderWest(patch, i, uvCoord, delPatch)){
		}
		else if(crd.x > 0.9999f && transferableVertexOverBorderEast(patch, i, uvCoord, delPatch)){
		}
		else if(crd.y < 0.0001f && transferableVertexOverBorderSouth(patch, i, uvCoord, delPatch)){
		}
		else if(crd.y > 0.9999f && transferableVertexOverBorderNorth(patch, i, uvCoord, delPatch)){
		}
		else{
			return false;
		}
	}
	if(sameCntr == 2){//all are transferable to same patch
		S_Triangle newTri;
		Vec2f coord[3];
		for(int j = 3; --j>=0;){
			int i = triangle.refIndex(j);
			Art* part = triangle.refPart(j);

			NurbsPatch* returnPatch = 0;
			getTransferedCoordOverPatchBorder(part, i, coord[j], returnPatch);
			//transferCoordOverPatchBorder(part, i);
		}
		//return false;
		return triangleIsOnPatchCorner(coord);
	}
	return true;
}

void Seamless::deleteAwkwardPoliesLeftBehind(NurbsPatch* patch, DelPoly& delPoly){
	M_Int& poly = delPoly.poly.refValue();

	for(int k = partList.length; --k >= 0;){
		Art* part = partList[k];
		delPoly.part = part;
		poly.length = 0;

		M_Triangle& m_triangle = part->triangle;
		int len = m_triangle.length;
		for(int k = 0; k < len; k++){
			S_Triangle& triangle = m_triangle[k];
			if(triangleIsTransferable(triangle, patch)){
				poly += k;
			}
		}
		if(poly.length){
			delPoly.doBuild();
		}
	}
}

static DelPoly g_delPoly;

void Seamless::deletePoliesOwnedByPatch(NurbsPatch* patch){
	transferBorderVertices(patch);
	fastDeletePoliesOwnedByPatch(patch);

	deleteAwkwardPoliesLeftBehind(patch, g_delPoly);
}

void Seamless::fastDeletePoliesOwnedByPatch(NurbsPatch* patch){
	//int plidx = patch->plidx;
	M_Int& poly = g_delPoly.poly.refValue();

	for(int k = partList.length; --k >= 0;){
		Art* part = partList[k];
		g_delPoly.part = part;
		poly.length = 0;

		M_Triangle& m_triangle = part->triangle;
		for(int k = 0; k < m_triangle.length; k++){
			for(int j = 0; j < 3; j++){
				Art* prt = m_triangle[k].refPart(j);
				int idx = m_triangle[k].refIndex(j);
				if(patch == prt->patchOwner[idx]){
					poly += k;
					break;
				}
			}
		}
		if(poly.length){
			g_delPoly.doBuild();
		}
	}
}

void Seamless::deleteAllSyncParts(){
	for(int k = partList.length;--k >= 0;){
		Art* part = partList[k];
		M_NurbsPatch& patchOwner = part->patchOwner;
		if(patchOwner.length){
			if(patchOwner[0]->sync){
				part->deleteGeometry();
			}
		}
	}
	//resetRelevantParts();
}

void Seamless::deleteAllPartsNonBnspe(){
	for(int k = cpNodeList.length; --k >= 0;){
		NurbsPatch* patch = cpNodeList[k];
		M_Node& m_stem = patch->stem.refValue();
		for(int k = m_stem.length; --k >= 0;){
			dynamic_cast<Stem*>(m_stem[k].p)->beginCoordOffset = 0x7fffffff;
		}
	}
	for(int k = partList.length;--k >= 0;){
		Art* part = partList[k];
		part->doClearCycle();
	}
	resetRelevantParts();
}

void Seamless::deletePoliesOwnedByInvalidPatchs(){
	//int plidx = patch->plidx;
	M_Int& poly = g_delPoly.poly.refValue();

	for(int k = partList.length; --k >= 0;){
		Art* part = partList[k];
		g_delPoly.part = part;
		poly.length = 0;

		M_Triangle& m_triangle = part->triangle;
		for(int k = 0; k < m_triangle.length; k++){
			for(int j = 0; j < 3; j++){
				Art* prt = m_triangle[k].refPart(j);
				int idx = m_triangle[k].refIndex(j);
				if(idx >= prt->patchOwner.length){//do we ever need this?
					poly += k;
					break;
				}
				else{
					NurbsPatch* patch = prt->patchOwner[idx];
					int plidx = patch->plidx;
					if(plidx >= cpNodeList.length || patch != cpNodeList[plidx]){
						poly += k;
						break;
					}
				}
			}
		}
		if(poly.length){
			g_delPoly.doBuild();
		}
	}
}

void Seamless::doBuild(){
#ifdef CPINDEX
	centralCPsToPatches();
#endif
	if(nurbsLineShape){
		if(IndexedLineSet* ils = dynamic_cast<IndexedLineSet*>(nurbsLineShape->geometry.p)){
			dynamic_cast<Coordinate*>(ils->coord.p)->point.refValue().length = 0;
			dynamic_cast<Color*>(ils->color.p)->color.refValue().length = 0;
			ils->coordIndex.refValue().length = 0;
		}
	}

	M_Build& build=this->build.refValue();

	if(workAnimCP){
		catMatrix();
		ctrf = true;
		if(bnspe == false){
			for(int k = partList.length; --k >= 0;){
				Art* part = partList[k];
				part->doClearCycleAnimCP();
			}
		}
		//cpNodeList.length = 0;
		//cpNodeListAdd(build, cpNodeList);
		if(animCPBuildFlag){

			copySourceCP(cpNodeList);


			//int nodeIdx = 0;
			//int cpOffset = 0;
			for(int k = cpMorpherPart.length; --k >= 0;){
				CPMorpherPart& cpmp = cpMorpherPart[k];
				cpmp.morpher->do_frame(cpmp.part);
			}

			for(int k = skeleton.length; --k >= 0;){
				Art* part = skeleton.refValue()[k].p;
				part->transformPoints(Matrix());
			}
			calcAllPolyMidPoints();
		}
	}
	else if(bnspe == false){
		deleteAllPartsNonBnspe();
	}
	defragEffect();
	g_exitBuildFlag=false;

	String str;
	if(bnspe){
		ctrf = true;
		doBorderlessBuild();
		if(workAnimCP){
			updateCage();//after the hair doBuild
		}
		//dont forget the line nodes still need to be built in bnspe mode
		for(int k = 0; k < nurbsLine.length; k++){
			nurbsLine[k]->doBuild();
		}
		if(g_drawNurbsLine->mousePhase == DrawNurbsLine::READY_TO_DRAG){
			g_drawNurbsLine->displayDrawnLine();
		}
	}
	else{
		try{
			deleteSyncPolies();
			for(int k=0; k<build.length;k++){
				Build* bld=build[k];
				bld->doSetup();
				bld->doBuild();
				if(g_exitBuildFlag)break;
			}
			if(workAnimCP){
				updateCage();//after the hair doBuild
			}
		}
		catch(NodePtr*){
			str+="NodePtr";
			showBuildError(str);
		}
		catch(M_Triangle*){
			str+="M_Triangle";
			showBuildError(str);
		}
		catch(M_SArt*){
			str+="M_SArt";
			showBuildError(str);
		}
		catch(Node* e){
			str+=e->getTypeStr();
			showBuildError(str);
		}

		catch(Field* e){
			str+=e->getTypeStr();
			showBuildError(str);
		}
		catch(M_Bool*){
			str+="M_Bool";
			showBuildError(str);
		}
		catch(M_Color3f*){
			str+="M_Color3f";
			showBuildError(str);
		}
		catch(M_Float*){
			str+="M_Float";
			showBuildError(str);
		}
		catch(M_Image*){
			str+="M_Image";
			showBuildError(str);
		}
		catch(M_Int*){
			str+="M_Int";
			showBuildError(str);
		}
		catch(M_Node*){
			str+="M_Node";
			showBuildError(str);
		}
		catch(M_Rotation*){
			str+="M_Rotation";
			showBuildError(str);
		}
		catch(M_String*){
			str+="M_String";
			showBuildError(str);
		}
		catch(M_Double*){
			str+="M_Double";
			showBuildError(str);
		}
		catch(M_Vec2f*){
			str+="M_Vec2f";
			showBuildError(str);
		}
		catch(M_Vec3f*){
			str+="M_Vec3f";
			showBuildError(str);
		}
		catch(M_ProtoDeclare){
			str+="M_ProtoDeclare";
			showBuildError(str);
		}
		catch(M_ProtoInterface){
			str+="M_ProtoInterface";
			showBuildError(str);
		}

		catch(int){
			str+="Exception in non class function";
			showBuildError(str);
		}
		catch(void*){
			str+="void";
			showBuildError(str);
		}
		catch(...){
			str+=String("...");
			showBuildError(str);
		}
	}

	animCPBuildFlag = false;
	cpMorpherPart.length = 0;
	catMatrix();
}
void GroupBuild::doBuild(){
	M_Build& build=this->build.refValue();

	for( int k=0; k<build.length;k++){
		Build* bld=build[k];
		bld->doSetup();
		bld->doBuild();
		if(g_exitBuildFlag)break;
	}
}
void ExitBuild::doBuild(){
	if(enabled==true){
		g_exitBuildFlag=true;
	}
}
void PositionInterpolator::processFields(){
	if(key.length < keyValue.length){
		keyValue.refValue().length=key.length;
	}
	else if(keyValue.length < key.length){
		key.refValue().length=keyValue.length;
	}
}
void OrientationInterpolator::processFields(){
	if(key.length < keyValue.length){
		keyValue.refValue().length=key.length;
	}
	else if(keyValue.length < key.length){
		key.refValue().length=keyValue.length;
	}
}
void ColorInterpolator::processFields(){
	if(key.length < keyValue.length){
		keyValue.refValue().length=key.length;
	}
	else if(keyValue.length < key.length){
		key.refValue().length=keyValue.length;
	}
}
void ScalarInterpolator::processFields(){
	if(key.length < keyValue.length){
		keyValue.refValue().length=key.length;
	}
	else if(keyValue.length < key.length){
		key.refValue().length=keyValue.length;
	}
}
void CameraSelector::processFields(){
	if(key.length < keyValue.length){
		keyValue.refValue().length=key.length;
	}
	else if(keyValue.length < key.length){
		key.refValue().length=keyValue.length;
	}
}
void Group::processFieldsLoop(HAnimHumanoid* humanoid){
	for(int k=children.length; --k>=0;){
		dynamic_cast<AGroup*>(children[k].getRawPtr())->processFieldsLoop(humanoid);
	}
}
void HAnimSegment::processFieldsLoop(HAnimHumanoid*){
}
void HAnimJoint::processFieldsLoop(HAnimHumanoid* humanoid){
	this->humanoid=humanoid;
	humanoid->joints.refValue()+=this;
	for(int i=displacers.length; --i>=0;){
		dynamic_cast<HAnimDisplacer*>(displacers[i].getRawPtr())->humanoid=humanoid;
	}
	for(int k=children.length; --k>=0;){
		if(AGroup* agroup = dynamic_cast<AGroup*>(children[k].getRawPtr())){
			agroup->processFieldsLoop(humanoid);
		}
	}
}
void HAnimHumanoid::processFields(){
	coordMask.length=0;
	maskIndex.length=0;

	joints.refValue().length=0;
	for(int k=skeleton.length; --k>=0;){
		dynamic_cast<AGroup*>(skeleton[k].getRawPtr())->processFieldsLoop(this);
	}
}
static const char povEffectNames[14][16]={
	"brilliance",
	"phong",
	"phong_size",
	"roughness",
	"crand",
	"conserve_energy",
	"reflection",
	"fresnel",
	"falloff",
	"exponent",
	"metallic",
	"iridescence",
	"thickness",
	"turbulence",
};
static const char vrmlEffectNames[1][16]={
	"shininess",
};
void setSceneFieldsForPov(M_Properties& properties, bool outputPovRay){
	for(int k=properties.length; --k>=0;){
		for(int i=14; --i>=0;){
			if(properties[k].name==povEffectNames[i]){
				properties[k].set_disable(!outputPovRay);
			}
		}
		for(int i=1; --i>=0;){
			if(properties[k].name==vrmlEffectNames[i]){
				properties[k].set_disable(outputPovRay);
			}
		}
	}
}
void updateNurbsCurveKnotBezier(int dim, M_Float& knot){
	knot.length=dim*2;
	//Create Bezier knots
	for(int i = 0; i < dim; i++) {
		knot[i] = 0;
		knot[i + dim] = 1;
	}
}

void updateNurbsCurveKnotMid2(int dim, M_Float& knot){
	knot.length = dim + 4;

	for(int n = 0; n < 4; n++){
		knot[n] = 0;
	}
	double g = 1;
	int m = (dim - 1 ) / 3;
	for(int n = 1 ; n < m; n++){
		knot[n * 3 + 3] = knot[n * 3 + 2] = knot[n * 3 + 1] = g / m * n;
	}
	for(int n=knot.length - 4; n < knot.length; n++){
		knot[n] = 1;
	}
}

void updateNurbsCurveKnotMid1(int dim, M_Float& knot){
	knot.length = dim+3;

	for(int n = 0; n < 3; n++){
		knot[n] = 0;
	}
	double g = 1;
	int m = (dim-1)/2;
	for(int n = 1 ; n < m; n++){
		knot[n*2+2] = knot[n*2+1] = g/m*n;
	}
	for(int n=knot.length-3; n < knot.length; n++){
		knot[n] = 1;
	}
}

void NurbsPatch::uUpdateKnotMid1(){
	updateNurbsCurveKnotMid1(dim.x, uKnot.refValue());
	if(mirrorPatch){
		updateNurbsCurveKnotMid1(mirrorPatch->dim.x, mirrorPatch->uKnot.refValue());
	}
}
void NurbsPatch::vUpdateKnotMid1(){
	updateNurbsCurveKnotMid1(dim.y, vKnot.refValue());
	if(mirrorPatch){
		updateNurbsCurveKnotMid1(mirrorPatch->dim.y, mirrorPatch->vKnot.refValue());
	}
}


void updateNurbsCurveKnotMid0(int dim, M_Float& knot){
	knot.length = dim + 3;
	knot[0] = 0;
	knot[1] = 0;
	int i = 0;
	for(; i < knot.length - 4; i++){
		knot[i + 2] = i;
	}
	knot[i + 2] = i - 1;
	knot[i + 3] = i - 1;
}
void updateNurbsCurveKnotMid0(int dim, M_Float& knot, int offset){
	knot.length += dim + 3;
	knot[offset + 0] = 0;
	knot[offset + 1] = 0;
	int i = 0;
	for(; i < dim - 1; i++){
		knot[offset + i + 2] = i;
	}
	knot[offset + i + 2] = i - 1;
	knot[offset + i + 3] = i - 1;
}
bool checkBezier(const M_Float& knot){
	if(knot.length < 6){
		return false;
	}
	for(int i = knot.length; --i>=0;){
		if(knot[i] != 0 && knot[i] != 1){
			return false;
		}
	}
	return true;
}
NurbsPatch::CurveType checkNurbsType(int dim, const M_Float& knot){
	if(knot.length==0){
		return NurbsPatch::MID_0;
	}
	if(knot[knot.length - 1] == 1){
		if(dim > 3 && checkBezier(knot)){
			return NurbsPatch::BEZIER;
		}
		else{
			if(knot[3] == 0){
				return NurbsPatch::MID_2;
			}
			else{
				return NurbsPatch::MID_1;
			}
		}
	}
	return NurbsPatch::MID_0;
}
void NurbsPatch::updateKnot(int dim, M_Float& knot, float& order){
	switch(checkNurbsType(dim, knot)){
		case NurbsPatch::MID_1:
			updateNurbsCurveKnotMid1(dim, knot);
			break;
		case NurbsPatch::MID_0:
			updateNurbsCurveKnotMid0(dim, knot);
			break;
		case NurbsPatch::MID_2:
			updateNurbsCurveKnotMid2(dim, knot);
			break;
		case NurbsPatch::BEZIER:
			updateNurbsCurveKnotBezier(dim, knot);
			order = dim;
			break;
		default: throw;
	}
}

void NurbsPatch::updateKnot(int oldDim, int newDim, M_Float& knot, float& order){
	switch(checkNurbsType(oldDim, knot)){
		case NurbsPatch::MID_1:
			updateNurbsCurveKnotMid1(newDim, knot);
			break;
		case NurbsPatch::MID_0:
			updateNurbsCurveKnotMid0(newDim, knot);
			break;
		case NurbsPatch::MID_2:
			updateNurbsCurveKnotMid2(newDim, knot);
			break;
		case NurbsPatch::BEZIER:
			updateNurbsCurveKnotBezier(newDim, knot);
			order = newDim;
			break;
		default: throw;
	}
}

void ExportMovie::processFields(){
	setSceneFieldsForPov(g_lookupProtoInterface[COLOREFFECT]->properties, povRay);
	setSceneFieldsForPov(g_lookupProtoInterface[TEXTUREEFFECT]->properties, povRay);
	if(!(avi || mpg || mp4 || flv)){
		avi.refValue() = true;
	}
}
void BranchJoin::processFields(){
	if(part.length>1){
		descendants=true;
		part.refValue().length=1;
	}
	return;
}
void Script::processFields(){
}
void Place::processFields(){
	alignX=multiply.refValue().x<0;
	alignY=multiply.refValue().y<0;
	alignZ=multiply.refValue().z<0;
}
void SurfaceTransform::processFields(){
	Matrix mat;
	mat=rotationMatrix(rotation);
	matrix=scaleMatrix(scale)*rotationMatrix(rotation)*translationMatrix(translation);
	return;
}
void CCTranslate::processFields(){
	matrix=rotationMatrix(orientation)*translationMatrix(location);
	imatrix=translationMatrix(-location)*rotationMatrix(-orientation);
	return;
}
void CCRingScale::processFields(){
	matrix=rotationMatrix(orientation)*translationMatrix(location);
	imatrix=translationMatrix(-location)*rotationMatrix(-orientation);
	return;
}
void CCScale::processFields(){
	matrix=rotationMatrix(orientation)*translationMatrix(location);
	imatrix=translationMatrix(-location)*rotationMatrix(-orientation);
	return;
}
void CCBend::processFields(){
	matrix=rotationMatrix(orientation)*translationMatrix(location);
	imatrix=translationMatrix(-location)*rotationMatrix(-orientation);
	return;
}
void processFieldsECLatheECSyncBend(M_Vec2f& point, M_Float& tangent, M_LongDouble& ma, M_LongDouble& mb, M_Float& scaleDis, M_Float& distance, M_Float& height){
	ma.length=mb.length=scaleDis.length=distance.length=(height.length=point.length)-1;

	height[0]=0;
	for(int i=0; i<point.length-1; i++){
		long double x,y,angle,xPlus,yPlus,xyt,j;
		long double ang1z=tangent[i];
		long double ang2z=tangent[i+1];
		long double& a=ma[i];
		long double& b=mb[i];

		x=point[i+1].x-point[i].x;
		y=point[i+1].y-point[i].y;

		rotateZ(x, y, ang1z);

		angle=ang2z-ang1z+PI*.5;
		xPlus =x+cosl(angle)*TINY_AMOUNT;
		yPlus =y+sinl(angle)*TINY_AMOUNT;
		a=(powl(y,2)*powl(xPlus,2)-powl(yPlus,2)*powl(x,2))/(2*powl(yPlus,2)*x-2*powl(y,2)*xPlus);
		b=sqrtl(fabsl(powl(a,2)*powl(y,2)/(powl(x,2)+2*a*x)));
		j=y/b;
		if(j>1)j=1;
		xyt=asinl(j);


		float bigRad;
		if(a>b){
			bigRad=a;
		}
		else{
			bigRad=b;
		}
		distance[i]=xyt*bigRad;

		height[i+1]=height[i]+distance[i];
		scaleDis[i]=xyt/distance[i];
	}
}
void Bezier::addControlPoint(int& newLen){
	if(point.length==0){
		point+=M_Vec2f(Vec2f(.25, 0), Vec2f(.5, .5));
		fix+=M_Bool(true, true);
	}
	else{
		int idx=index[0];
		if(idx<point.length-1){
			idx++;
		}
		point.insertElement(idx, (point[idx]-point[idx-1])*Vec2f(.5,.5)+point[idx-1]);
		fix.insertElement(idx, false);
		index=idx;
	}
	newLen=point.length;
}
void Bezier3f::addControlPoint(int& newLen){
	if(point.length==0){
		point+=M_Vec3f(Vec3f(), Vec3f(.5, .5, 0));
		fix+=M_Bool(true, true);
	}
	else{
		int idx=index[0];
		if(idx<point.length-1){
			idx++;
		}
		point.insertElement(idx, (point[idx]-point[idx-1])*Vec3f(.5,.5,.5)+point[idx-1]);
		fix.insertElement(idx, false);
		index[0]=idx;
	}
	newLen=point.length;
}
void BezierLatheBuild::addControlPoint(int& newLen){
	if(point.length==0){
		Bezier::point+=M_Vec2f(Vec2f(.25, 0), Vec2f(.5, .5));
		Bezier::fix+=M_Bool(true, true);
		bands.refValue()+=M_Int(16,16);
		NurbsPatch* patch = new NurbsPatch;
		Stem* stem = new Stem;
		stem->bands.refValue() = 16;
		patch->stem+=stem;
		m_patch=M_<NurbsPatchPtr>(patch);
	}
	else{
		int idx=index[0];
		if(idx<point.length-1){
			idx++;
		}
		Bezier::point.insertElement(idx, (point[idx]-point[idx-1])*Vec2f(.5,.5)+point[idx-1]);
		Bezier::fix.insertElement(idx, false);
		index[0]=idx;
	}
	newLen=point.length;
}

void BezierBend::addControlPoint(int& newLen){
	if(point.length==0){
		Bezier3f::point+=M_Vec3f(Vec3f(), Vec3f(.5, .5, 0));
		Bezier3f::fix+=M_Bool(true, true);
	}
	else{
		int idx=index[0];
		if(idx<point.length-1){
			idx++;
		}
		Bezier3f::point.insertElement(idx, (point[idx]-point[idx-1])*Vec3f(.5,.5,.5)+point[idx-1]);
		Bezier3f::fix.insertElement(idx, false);
		index=idx;
	}
	newLen=point.length;
}

void NurbsLathe::removeControlPoint(int idx){
	M_Vec2f& point = this->point;
	M_Float& weight = this->weight;
	M_Float& knot = this->knot;

	if(point.length == 3){
		return;
	}
	int ia = idx;
	int ib = idx;
	int ic = idx;

	if(ia == 0){
		ia += 2;
	}
	else if(ia == point.length - 1){
		ib -= 2;
		ic = ib;
	}
	else if(ia == point.length - 2){
		ib--;
		ic = ib - 1;
	}
	point.deleteElement(ia);
	weight.deleteElement(ia);
	point.deleteElement(ib);
	weight.deleteElement(ib);

	index[0] = ic;

	updateNurbsCurveKnotMid1(point.length, knot);
	bands.refValue() = point.length / 2 * patch->smls->tesselate;
}

void BendyLathe::addControlPoint(int& newLen){
	M_Vec3f& point = this->point;
	M_Float& weight = this->weight;
	M_Float& knot = this->knot;

	if(point.length==2){
		weight=M_Float(1,1);
	}
	int idx=index[0];
	if(idx<point.length-1){
		idx++;
	}
	point.insertElement(idx, (point[idx]+point[idx-1])*.5);
	weight.insertElement(idx, 1);

	updateNurbsCurveKnotMid0(point.length, knot);

	index[0]=idx;
	newLen=point.length;
}

void BendyLathe::addRadius(int& newLen){
	M_Float& radius = this->radius.refValue();
	M_Float& at = this->at.refValue();

	if(radius.length==2){
		at=M_Float(0,1);
	}
	int idx=index[2];
	if(idx<radius.length-1){
		idx++;
	}
	radius.insertElement(idx, (radius[idx]+radius[idx-1])*.5);
	at.insertElement(idx, (at[idx]+at[idx-1])*.5);

	index[2]=idx;
	newLen=radius.length;
}
void BendyLathe::removeControlPoint(int& newLen){
	M_Vec3f& point = this->point;
	M_Float& weight = this->weight;
	M_Float& knot = this->knot;

	if(point.length==2){
		return;
	}
	int idx=index[0];

	if(idx==0){
		return;
	}
	else if(idx==point.length-1){
		point.deleteElement(idx);
		weight.deleteElement(idx);
		idx--;
		index[0]=idx;
	}
	else {
		point.deleteElement(idx);
		weight.deleteElement(idx);
	}
	updateNurbsCurveKnotMid0(point.length, knot);
	newLen=point.length;
}
void BendyLathe::removeRadius(int& newLen){
	M_Float& radius = this->radius.refValue();
	M_Float& at = this->at.refValue();

	if(radius.length==2){
		return;
	}
	int idx=index[2];

	if(idx==0){
		return;
	}
	else if(idx==radius.length-1){
		radius.deleteElement(idx);
		at.deleteElement(idx);
		idx--;
		index[2]=idx;
	}
	else {
		radius.deleteElement(idx);
		at.deleteElement(idx);
	}
	newLen=point.length;
}
float Bezier::getFirstTanCP(){
	float xDiff=point[1].x-point[0].x;
	float yDiff=point[1].y-point[0].y;
	float hyp=sqrt(xDiff*xDiff+yDiff*yDiff);
	return -asin(yDiff/hyp);
}
float Bezier::getLastTanCP(){
	float xDiff=point[point.length-1].x-point[point.length-2].x;
	float yDiff=point[point.length-1].y-point[point.length-2].y;
	float hyp=sqrt(xDiff*xDiff+yDiff*yDiff);
	return -asin(yDiff/hyp);
}
void Bezier::genFlatHeight(){
	flatHeight.length=line.length-1;
	float h=0;
	for(int i=0; i<flatHeight.length; i++){
		float xDiff=line[i+1].x-line[i].x;
		float yDiff=line[i+1].y-line[i].y;
		h+=sqrt(xDiff*xDiff+yDiff*yDiff);
		flatHeight[i]=h;
	}
}
void Bezier::genCircumHeight(){
	flatHeight.length=line.length-1;
	for(int i=0; i<flatHeight.length; i++){
		float polarAngle=line[i+1].getAngle3();
		flatHeight[i]=polarAngle;
	}
	flatHeight[flatHeight.length-1]=PI*2;
}
void Bezier3f::genFlatHeight(){
	flatHeight.length=line.length-1;
	float h=0;
	for(int i=0; i<flatHeight.length; i++){
		float xDiff=line[i+1].x-line[i].x;
		float yDiff=line[i+1].y-line[i].y;
		float zDiff=line[i+1].z-line[i].z;
		h+=sqrt(xDiff*xDiff+yDiff*yDiff+zDiff*zDiff);
		flatHeight[i]=h;
	}
}
void Bezier::genHeightForScaleOrShift(){
	flatHeight.length=line.length-1;
	float h=0;
	for(int i=0; i<flatHeight.length; i++){
		h+=line[i+1].y-line[i].y;// is this any different to: flatHeight[i]=line[i+1].y;
		flatHeight[i]=h;
	}
}
void Bezier::genSegDistance(){
	segDistance.length=0;
	segDistance+=flatHeight[numOfSteps-1];
	for(int i=1; i<nopfs.length; i++){
		segDistance+=flatHeight[(i+1)*numOfSteps-1]-flatHeight[i*numOfSteps-1];
	}
}
void Bezier3f::genSegDistance(){
	segDistance.length=0;
	segDistance+=flatHeight[numOfSteps-1];
	for(int i=1; i<nopfs.length; i++){
		segDistance+=flatHeight[(i+1)*numOfSteps-1]-flatHeight[i*numOfSteps-1];
	}
}
void Bezier::genHeightIndex(){

	heightIndex.length=flatHeight.length;

	float bezArcLen=flatHeight[flatHeight.length-1];
	float step=bezArcLen/flatHeight.length;
	for(int i=0, j=0; i<flatHeight.length; i++){
		float h = step*(i+1);
		while(j < line.length - 2 && h > flatHeight[j]){
			j++;
		}
		heightIndex[i]=j;
	}
}
void Bezier3f::genHeightIndex(){

	heightIndex.length=flatHeight.length;

	float bezArcLen=flatHeight[flatHeight.length-1];
	float step=bezArcLen/flatHeight.length;
	for(int i=0, j=0; i<flatHeight.length; i++){
		float h = step*(i+1);
		while(j < line.length - 2 && h > flatHeight[j]){
			j++;
		}
		heightIndex[i]=j;
	}
}
void Bezier::getHypes(int i, float& xDiffA, float& yDiffA, float& hypA, float& xDiffB, float& yDiffB, float& hypB){
	xDiffA=point[i-1].x-point[i].x;
	yDiffA=point[i-1].y-point[i].y;
	hypA=sqrt(xDiffA*xDiffA+yDiffA*yDiffA);

	xDiffB=point[i+1].x-point[i].x;
	yDiffB=point[i+1].y-point[i].y;
	hypB=sqrt(xDiffB*xDiffB+yDiffB*yDiffB);
}
void Bezier3f::getHypes(int i, float& xDiffA, float& yDiffA, float& zDiffA, float& hypA, float& xDiffB, float& yDiffB, float& zDiffB, float& hypB){
	xDiffA=point[i-1].x-point[i].x;
	yDiffA=point[i-1].y-point[i].y;
	zDiffA=point[i-1].z-point[i].z;
	hypA=sqrt(xDiffA*xDiffA + yDiffA*yDiffA + zDiffA*zDiffA);

	xDiffB=point[i+1].x-point[i].x;
	yDiffB=point[i+1].y-point[i].y;
	zDiffB=point[i+1].z-point[i].z;
	hypB=sqrt(xDiffB*xDiffB + yDiffB*yDiffB + zDiffB*zDiffB);
}
void Bezier::convertBorderToGroup(){
	fix[0]=true;
	fix[fix.length-1]=true;
	nopfs.length=0;
	for(int i=0; i<fix.length-1; i++){
		if(fix[i]){
			nopfs.length++;
			nopfs[nopfs.length-1]++;
		}
		nopfs[nopfs.length-1]++;
	}
}
void Bezier3f::convertBorderToGroup(){
	fix[0]=true;
	fix[fix.length-1]=true;
	nopfs.length=0;
	for(int i=0; i<fix.length-1; i++){
		if(fix[i]){
			nopfs.length++;
			nopfs[nopfs.length-1]++;
		}
		nopfs[nopfs.length-1]++;
	}
}
void Bezier::convertGroupToFix(){
	if(point.length){
		fix.length=point.length;
		for(int k=0, h=0; k<nopfs.length; k++){
			fix[h++]=true;
			for(int i=1; i<nopfs[k]-1; i++){
				fix[h++]=false;
			}
		}
		fix[fix.length-1]=true;
	}
}
void Bezier3f::convertGroupToFix(){
	if(point.length){
		fix.length=point.length;
		for(int k=0, h=0; k<nopfs.length; k++){
			fix[h++]=true;
			for(int i=1; i<nopfs[k]-1; i++){
				fix[h++]=false;
			}
		}
		fix[fix.length-1]=true;
	}
}
inline Vec2f Bezier::genCurve3(const int&, const int& p, const int& s, const float& t, M_Vec2f& pnt)const{
	for(int i=0; i<nopfs[s]; i++){
		pnt[i] = point[p+i];
	}
	for (int j = pnt.length; --j >= 0;){  //  calc inbetweens for line
		for (int i = 0; i < j; i++){
			pnt[i] = Vec2f(
				(1.0 - t) * pnt[i].x + t * pnt[i + 1].x,
				(1.0 - t) * pnt[i].y + t * pnt[i + 1].y
			);
		}
	}
	return pnt[0];
}
inline Vec3f Bezier3f::genCurve3(const int&, const int& p, const int& s, const float& t, M_Vec3f& pnt)const{
	for(int i=0; i<nopfs[s]; i++){
		pnt[i] = point[p+i];
	}
	for (int j = pnt.length; --j >= 0;){  //  calc inbetweens for line
		for (int i = 0; i < j; i++){
			pnt[i] = Vec3f(
				(1.0 - t) * pnt[i].x + t * pnt[i + 1].x,
				(1.0 - t) * pnt[i].y + t * pnt[i + 1].y,
				(1.0 - t) * pnt[i].z + t * pnt[i + 1].z
			);
		}
	}
	return pnt[0];
}
void Bezier::genCurve(){
	line.length=0;
	for(int p = 0, s = 0; s < nopfs.length; p+=nopfs[s++]-1){
		float step = 1.0 / numOfSteps;
		float t = step;

		line+=Vec2f(point[p].x, point[p].y);
		pnt.length = nopfs[s];

		for (int k = 1; k < numOfSteps; k++){
			line+=genCurve3(numOfSteps, p, s, t, pnt);
			t += step;
		}
	}
	line+=point[point.length-1];
}
void Bezier3f::genCurve(){
	line.length=0;
	for(int p = 0, s = 0; s < nopfs.length; p+=nopfs[s++]-1){
		float step = 1.0 / numOfSteps;
		float t = step;

		line+=Vec3f(point[p].x, point[p].y, point[p].z);
		pnt.length = nopfs[s];

		for (int k = 1; k < numOfSteps; k++){
			line+=genCurve3(numOfSteps, p, s, t, pnt);
			t += step;
		}
	}
	line+=point[point.length-1];
}
float getAngleDiff(Vec2f a, Vec2f b){
	float angA=a.getAngle();
	float angB=b.getAngle();
	float angDiff=fabsf(angB-angA);
	if(angDiff>PI){
		angDiff = PI*2 - fabs(angA) - fabs(angB);
	}
	return angDiff;
}
/*
void BezierLathe::genTangentDiff(){
	tangentDiff.length=flatHeight.length;

	for(int g=0, t=0; g<group.length; g++){
		float accum = 0;
		tangentDiff[t] = accum;
		t++;
		for(int i=1; i<numOfSteps; i++, t++){
			accum += getAngleDiff(line[t]-line[t-1], line[t+1]-line[t]);
			tangentDiff[t] = accum;
		}
	}
}
*/
void BezierLathe::processFields(){
	segDistance.length=0;
	flatHeight.length=0;
	heightIndex.length=0;
	line.length=0;
	if(point.length==0){
		return;
	}
	if(point.length!=fix.length){
		convertGroupToFix();
	}
	convertBorderToGroup();
	m_bp.length = nopfs.length;

	for(int i=0, accum = 0; i<nopfs.length; i++){
		m_bp[i] = accum;
		accum += nopfs[i]-1;
	}

//	genCurve();
//	genFlatHeight();
//	genHeightIndex();
//	genSegDistance();
//	genTangentDiff();
}
void BezierScale::processFields(){
	matrix=rotationMatrix(orientation)*translationMatrix(location);
	imatrix=translationMatrix(-location)*rotationMatrix(-orientation);

	Vec3f firstPoint = Vec3f(0, point.refValue()[0].y,0);

	gen_matrix=translationMatrix(firstPoint)*rotationMatrix(orientation)*translationMatrix(location);
	gen_imatrix=translationMatrix(-location)*rotationMatrix(-orientation)*translationMatrix(-firstPoint);

	segDistance.length=0;
	flatHeight.length=0;
	heightIndex.length=0;
	line.length=0;
	if(point.length==0){
		return;
	}
	if(point.length!=fix.length){
		convertGroupToFix();
	}
	convertBorderToGroup();

	genCurve();
	//genFlatHeight();
	genHeightForScaleOrShift();
	genHeightIndex();
	genSegDistance();
}
void BezierCircumScale::processFields(){
	matrix=rotationMatrix(orientation)*translationMatrix(location);
	imatrix=translationMatrix(-location)*rotationMatrix(-orientation);

	Vec3f firstPoint = Vec3f(0, point.refValue()[0].y,0);

	gen_matrix=translationMatrix(firstPoint)*rotationMatrix(orientation)*translationMatrix(location);
	gen_imatrix=translationMatrix(-location)*rotationMatrix(-orientation)*translationMatrix(-firstPoint);

	segDistance.length=0;
	flatHeight.length=0;
	heightIndex.length=0;
	line.length=0;
	if(point.length==0){
		return;
	}
	if(point.length!=fix.length){
		convertGroupToFix();
	}
	convertBorderToGroup();

	genCurve();
	//genFlatHeight();
	genCircumHeight();
	genHeightIndex();
	genSegDistance();
}
void BezierShift::processFields(){
	matrix=rotationMatrix(orientation)*translationMatrix(location);
	imatrix=translationMatrix(-location)*rotationMatrix(-orientation);

	Vec3f firstPoint = Vec3f(0, point.refValue()[0].y,0);

	gen_matrix=translationMatrix(firstPoint)*rotationMatrix(orientation)*translationMatrix(location);
	gen_imatrix=translationMatrix(-location)*rotationMatrix(-orientation)*translationMatrix(-firstPoint);

	segDistance.length=0;
	flatHeight.length=0;
	heightIndex.length=0;
	line.length=0;
	if(point.length==0){
		return;
	}
	if(point.length!=fix.length){
		convertGroupToFix();
	}
	convertBorderToGroup();

	genCurve();
	//genFlatHeight();
	genHeightForScaleOrShift();
	genHeightIndex();
	genSegDistance();
}
void BezierTwist::processFields(){
	matrix=rotationMatrix(orientation)*translationMatrix(location);
	imatrix=translationMatrix(-location)*rotationMatrix(-orientation);

	Vec3f firstPoint = Vec3f(0, point.refValue()[0].y,0);

	gen_matrix=translationMatrix(firstPoint)*rotationMatrix(orientation)*translationMatrix(location);
	gen_imatrix=translationMatrix(-location)*rotationMatrix(-orientation)*translationMatrix(-firstPoint);

	segDistance.length=0;
	flatHeight.length=0;
	heightIndex.length=0;
	line.length=0;
	if(point.length==0){
		return;
	}
	if(point.length!=fix.length){
		convertGroupToFix();
	}
	convertBorderToGroup();

	genCurve();
	//genFlatHeight();
	genHeightForScaleOrShift();
	genHeightIndex();
	genSegDistance();
}
void BezierTape::processFields(){
	segDistance.length=0;
	flatHeight.length=0;
	heightIndex.length=0;
	line.length=0;
	if(point.length==0){
		return;
	}
	if(point.length!=fix.length){
		convertGroupToFix();
	}
	convertBorderToGroup();

	genCurve();
	genFlatHeight();
	genHeightIndex();
	genSegDistance();
}
void BezierBend::processFields(){
	segDistance.length=0;
	flatHeight.length=0;
	heightIndex.length=0;
	line.length=0;
	if(point.length==0){
		return;
	}

	matrix=rotationMatrix(orientation)*translationMatrix(location);
	imatrix=translationMatrix(-location)*rotationMatrix(-orientation);

	Vec3f firstPoint = point[0];

	gen_matrix=rotationMatrix(orientation)*translationMatrix(location-firstPoint);
	gen_imatrix=translationMatrix(-location)*rotationMatrix(-orientation)*translationMatrix(-firstPoint);

	if(point.length!=fix.length){
		convertGroupToFix();
	}
	convertBorderToGroup();

	genCurve();
	genFlatHeight();
	genHeightIndex();
	genSegDistance();
}
int BezierPatch::getPatchIndex(int point)const{

	int k = 0;

	for(; k < m_bp.length; k++){

		int bp = m_bp[k];

		if(bp > point){
			break;
		}
	}
	return k - 1;
}
void BezierPatch::processFields(){

	const M_Vec2f& dim = this->dim;
	const M_Int& link = this->link;
	const M_Vec3f& point = this->point;
	M_Float& weight = this->weight.refValue();

	if(weight.length!=point.length){
		weight.length=point.length;
		for(int i=weight.length; --i>=0;){
			weight[i]=1;
		}
	}

	zoneDim=Vec2i(0,0);
	for(int k=dim.length; --k>=0;){
		if(zone[k].x>zoneDim.x){
			zoneDim.x = zone[k].x;
		}
		if(zone[k].y>zoneDim.y){
			zoneDim.y = zone[k].y;
		}
	}
	zoneDim += Vec2i(1,1);
	zoneLookUp.length = zoneDim.x*zoneDim.y;
	m_bp.length = dim.length;

	for(int k = zoneLookUp.length; --k>=0;){
		zoneLookUp[k]=-1;
	}
	for(int k = 0, bp=0; k<zone.length; k++){
		zoneLookUp[zone[k].y*zoneDim.x+zone[k].x] = k;
		m_bp[k] = bp;
		bp+=dim[k].x*dim[k].y;
	}
//	zoneLookUp = M_<int>(0, 1, 2, 3, 4, 5);
//	m_bp = M_Int(0, 16, 32, 48, 64, 80);

	linkToPatchIdx.length=link.length;
	for(int i = link.length; --i>=0;){
		linkToPatchIdx[i] = getPatchIndex(link[i]);
	}
	//linkToPatchIdx = M_Int(0, 1, 3, 2, 2, 3, 5, 4);

//	order_u=degree_u+1;
//	order_v=degree_v+1;


/*
	M_Vec3f& point = this->point.refValue();
	M_Int& numOfU = this->numOfU.refValue();
	M_Int& numOfV = this->numOfV.refValue();


	for(int k=0; k<numOfU.length; k++){

		int shouldBeLen=numOfU[k]*numOfV[k];

		if(shouldBeLen!=point.length){
			point.length=shouldBeLen;
			for(int v = 0; v < numOfV[k]; v++){
				for(int u = 0; u < numOfU[k]; u++){
					point[v*numOfU[k]+u]=Vec3f(1.0/(numOfU[k]-1) * u - .5, 1.0/(numOfV[k]-1) * v, 0);
				}
			}
		}
	}
*/
}
void BezierPatch::Dim::operator=(const M_Vec2f& m_par){
	M_Vec3f& point = home.point.refValue();
	M_Float& weight = home.weight.refValue();

	if(point.length==0){
		refValue()=m_par;
		home.processFields();
		return;
	}

	int numOfUGroupIdx = home.index[home.numOfUGroup];

	Vec2f s_dim=home.dim.refValue()[numOfUGroupIdx];
	int numOfU=s_dim.x;
	int numOfV=s_dim.y;
	Vec2f par=m_par[numOfUGroupIdx];
	int newU=numOfU;
	int newV=numOfV;

	M_Vec3f oldPoints=point;
	M_Float oldWeights=weight;

	int idx=home.index[0];
	int col=idx%numOfU;
	int row=idx/numOfU;



	if(par.x>s_dim.x){//insert colum
		newU++;
		int newLen=newU*newV;
		weight.length=point.length=newLen;
		for(int v = 0; v < numOfV; v++){//row
			for(int u = 0, nu=0; u<numOfU; u++, nu++){//collum
				point[v*(numOfU+1)+nu]=oldPoints[v*numOfU+u];
				weight[v*(numOfU+1)+nu]=oldWeights[v*numOfU+u];
				if(u==col){
					if(u!=numOfU-1){
						Vec3f va = oldPoints[v*numOfU+u];
						Vec3f vb = oldPoints[v*numOfU+u+1];

						float wa = oldWeights[v*numOfU+u];
						float wb = oldWeights[v*numOfU+u+1];
						nu++;
						point[v*(numOfU+1)+nu] = va + (vb-va)*.5;
						weight[v*(numOfU+1)+nu] = wa + (wb-wa)*.5;

					}
					else{
						Vec3f va = oldPoints[v*numOfU+u-1];
						Vec3f vb = oldPoints[v*numOfU+u];

						float wa = oldWeights[v*numOfU+u-1];
						float wb = oldWeights[v*numOfU+u];

						point[v*(numOfU+1)+nu+1]=point[v*(numOfU+1)+nu];
						point[v*(numOfU+1)+nu] = va + (vb-va)*.5;

						weight[v*(numOfU+1)+nu+1]=weight[v*(numOfU+1)+nu];
						weight[v*(numOfU+1)+nu] = wa + (wb-wa)*.5;
					}
				}
			}
		}
	}
	else if(par.x<s_dim.x){
		newU--;
		int newLen=newU*newV;
		weight.length=point.length=newLen;
		for(int v = 0; v < numOfV; v++){//row
			for(int u = 0, nu=0; u<numOfU; u++){//collum
				if(u==col){
					continue;
				}
				point[v*(numOfU-1)+nu]=oldPoints[v*numOfU+u];
				weight[v*(numOfU-1)+nu]=oldWeights[v*numOfU+u];
				nu++;
			}
		}
	}
	else if(par.y>s_dim.y){
		newV++;
		int newLen=newU*newV;
		weight.length=point.length=newLen;
		for(int v = 0, nv=0; v < numOfV; v++, nv++){//row
			if(v!=row){
				for(int u = 0; u<numOfU; u++){//collum
					point[nv*numOfU+u]=oldPoints[v*numOfU+u];
					weight[nv*numOfU+u]=oldWeights[v*numOfU+u];
				}
			}
			else{
				if(v < numOfV-1){
					nv++;
					for(int u = 0; u<numOfU; u++){//collum
						Vec3f va = oldPoints[v*numOfU+u];
						Vec3f vb = oldPoints[(v+1)*numOfU+u];

						float wa = oldWeights[v*numOfU+u];
						float wb = oldWeights[(v+1)*numOfU+u];
						point[nv*numOfU+u] = va + (vb-va)*.5;
						weight[nv*numOfU+u] = wa + (wb-wa)*.5;
					}
				}
				else{
					for(int u = 0; u<numOfU; u++){//collum
						point[(nv+1)*numOfU+u] = point[nv*numOfU+u];
						Vec3f vb = oldPoints[v*numOfU+u];
						Vec3f va = oldPoints[(v-1)*numOfU+u];

						float wb = oldWeights[v*numOfU+u];
						float wa = oldWeights[(v-1)*numOfU+u];

						point[nv*numOfU+u] = va + (vb-va)*.5;
						weight[nv*numOfU+u] = wa + (wb-wa)*.5;
					}
				}
			}
		}
	}
	else if(par.y<s_dim.y){
		int newLen=newU*newV;
		newV--;
		point.length=newLen;
		for(int v = 0, nv=0; v < numOfV; v++){//row
			if(v==row){
				continue;
			}
			for(int u = 0; u<numOfU; u++){//collum
				point[nv*numOfU+u]=oldPoints[v*numOfU+u];
				weight[nv*numOfU+u]=oldWeights[v*numOfU+u];
			}
			nv++;
		}
	}

	refValue()[numOfUGroupIdx]=Vec2f(newU, newV);
	if(col>=numOfU){
		col--;
	}
	home.index[0]=row*newU+col;
}

void NurbsPatch::insertCol2(const int uDim, const int vDim, M_Vec3f& dpoint, M_Float& weight, M_Int& cpIndex, M_Vec3f& oldPoints, M_Float& oldWeights, M_Int& oldCpIndex, const int col, const float ratio){

	//lookup.length = dpoint.length;
	oldPoints = dpoint;
	oldWeights = weight;
	oldCpIndex = cpIndex;
	M_Vec3f& cp = smls->cp.refValue();
	M_Float& cw = smls->cw.refValue();
	int cpMark = cp.length;
	cp.length += vDim;
	cw.length += vDim;

	int newU = uDim + 1;
	int newLen = vDim * newU;
	cpIndex.length = weight.length = dpoint.length = newLen;


	for(int v = 0; v < vDim; v++){//row
		for(int u = 0, nu = 0; u < uDim; u++, nu++){//collum
			int nIdx = v * (uDim + 1) + nu;
			int oIdx = v * uDim + u;
			//lookup[oIdx] = nIdx;
			dpoint[nIdx] = oldPoints[oIdx];
			weight[nIdx] = oldWeights[oIdx];
			cpIndex[nIdx] = oldCpIndex[oIdx];
			if(u == col){
				if(u != uDim - 1){
					Vec3f va = oldPoints[v * uDim + u];
					Vec3f vb = oldPoints[v * uDim + u + 1];

					float wa = oldWeights[v * uDim + u];
					float wb = oldWeights[v * uDim + u + 1];

					nu++;
					dpoint[v * (uDim + 1) + nu] = va + (vb - va) * ratio;
					weight[v * (uDim + 1) + nu] = wa + (wb - wa) * .5;
					cpIndex[v * (uDim + 1) + nu] = cpMark + v;
				}
				else{
					Vec3f va = oldPoints[v * uDim + u - 1];
					Vec3f vb = oldPoints[v * uDim + u];

					float wa = oldWeights[v * uDim + u - 1];
					float wb = oldWeights[v * uDim + u];

					dpoint[v * (uDim + 1) + nu + 1] = dpoint[v * (uDim + 1) + nu];
					dpoint[v * (uDim + 1) + nu] = va + (vb - va) * ratio;

					weight[v * (uDim + 1) + nu + 1] = weight[v * (uDim + 1) + nu];
					weight[v * (uDim + 1) + nu] = wa + (wb - wa) *.5;
					cpIndex[v * (uDim + 1) + nu] = cpMark + v;

				}
			}
		}
	}
	dim.refValue().x = newU;
	//smls->redirectStitchPolyIndices(this, lookup);
	if(ratio == 0){
		int idx = newU * (vDim - 1) + col;
		idx = idx;
	}



	int idxA = col;
	int idxB = idxA + 2;
	if(cpIndex[idxA] == cpIndex[idxB]){
		cpIndex[idxA + 1] = cpIndex[idxA];
	}
	idxA = (vDim - 1) * newU + col;
	idxB = idxA + 2;

	if(cpIndex[idxA] == cpIndex[idxB]){
		cpIndex[idxA + 1] = cpIndex[idxA];
	}
}
void NurbsPatch::insertCol(M_Vec3f& oldPoints, M_Float& oldWeights, M_Int& oldCpIndex, const int col, const float ratio, int& uDim, const bool centerMirrorLine){
	if(mirror && centerMirrorLine == false){
		if(mirrorPatch){
			insertCol2(dim.x, dim.y, dpoint.refValue(), weight, cpIndex.refValue(), oldPoints, oldWeights, oldCpIndex, col, ratio);
			mirrorPatch->insertCol2(mirrorPatch->dim.x, mirrorPatch->dim.y, mirrorPatch->dpoint.refValue(), mirrorPatch->weight.refValue(), mirrorPatch->cpIndex.refValue(), oldPoints, oldWeights, oldCpIndex, col, ratio);
		}
		else{
			int mirrorCol = getMirrorTwinlessIdx(col);
			int colA = col;
			int colB = mirrorCol;
			if(mirrorCol > col){
				colB = col;
				colA = mirrorCol;
			}
			insertCol2(dim.x, dim.y, dpoint.refValue(), weight, cpIndex.refValue(), oldPoints, oldWeights, oldCpIndex, colB, ratio);
			insertCol2(dim.x, dim.y, dpoint.refValue(), weight, cpIndex.refValue(), oldPoints, oldWeights, oldCpIndex, colA, ratio);
		}
	}
	else{
		insertCol2(dim.x, dim.y, dpoint.refValue(), weight, cpIndex.refValue(), oldPoints, oldWeights, oldCpIndex, col, ratio);
	}
	uDim = dim.x;
}

void NurbsPatch::deleteCol(const int uDim, const int vDim, M_Int& cpIndex, M_Vec3f& dpoint, M_Float& weight, M_Int& oldCpIndex, M_Vec3f& oldPoints, M_Float& oldWeights, const int col){
	//lookup.length = dpoint.length;
	oldCpIndex = cpIndex;
	oldPoints = dpoint;
	oldWeights = weight;

	int newU = uDim - 1;
	int newLen = newU * vDim;

	cpIndex.length = newLen;
	dpoint.length = newLen;
	weight.length = newLen;
	
	for(int v = 0; v < vDim; v++){//row
		for(int u = 0, nu = 0; u < uDim; u++){//collum
			int nIdx = v * (uDim - 1) + nu;
			int oIdx = v * uDim + u;
			//lookup[oIdx] = nIdx;

			if(u == col){
				continue;
			}
			cpIndex[nIdx] = oldCpIndex[oIdx];
			dpoint[nIdx] = oldPoints[oIdx];
			weight[nIdx] = oldWeights[oIdx];
			nu++;
		}
	}
	dim.refValue().y = newU;
	//smls->redirectStitchPolyIndices(this, lookup);
}

void NurbsPatch::insertRow2(const int uDim, const int vDim, M_Vec3f& dpoint, M_Float& weight, M_Int& cpIndex, M_Vec3f& oldPoints, M_Float& oldWeights, M_Int& oldCpIndex, const int row, const float ratio){
	//lookup.length = dpoint.length;
	oldPoints = dpoint;
	oldWeights = weight;
	oldCpIndex = cpIndex;
	M_Vec3f& cp = smls->cp.refValue();
	M_Float& cw = smls->cw.refValue();
	int cpMark = cp.length;
	cp.length += uDim;
	cw.length += uDim;


	int newV = vDim + 1;
	int newLen = uDim * newV;


	cpIndex.length = weight.length = dpoint.length = newLen;
	for(int v = 0, nv = 0; v < vDim; v++, nv++){//row
		for(int u = 0; u < uDim; u++){//column
			int nIdx = nv * uDim + u;
			int oIdx = v * uDim + u;
			//lookup[oIdx] = nIdx;
			dpoint[nIdx] = oldPoints[oIdx];
			weight[nIdx] = oldWeights[oIdx];
			cpIndex[nIdx] = oldCpIndex[oIdx];
		}
		if(v == row){
			if(v < vDim - 1){
				nv++;
				for(int u = 0; u < uDim; u++){//column
					Vec3f va = oldPoints[v * uDim + u];
					Vec3f vb = oldPoints[(v + 1) * uDim + u];

					float wa = oldWeights[v*uDim+u];
					float wb = oldWeights[(v+1)*uDim+u];

					dpoint[nv * uDim + u] = va + (vb-va) * ratio;
					weight[nv * uDim + u] = wa + (wb-wa) * .5;
					cpIndex[nv * uDim + u] = cpMark + u;
				}
			}
			else{
				for(int u = 0; u < uDim; u++){//collum
					int nIdx = (nv + 1) * uDim + u;
					int oIdx = nv * uDim + u;
					dpoint[nIdx] = dpoint[oIdx];
					weight[nIdx] = weight[oIdx];

					Vec3f vb = oldPoints[v*uDim+u];
					Vec3f va = oldPoints[(v-1)*uDim+u];

					float wb = oldWeights[v*uDim+u];
					float wa = oldWeights[(v-1)*uDim+u];

					dpoint[nv * uDim + u] = va + (vb-va) * ratio;
					weight[nv * uDim + u] = wa + (wb-wa) * .5;
					cpIndex[nv * uDim + u] = cpMark + u;
				}
			}
		}
	}
	dim.refValue().y = newV;
	//smls->redirectStitchPolyIndices(this, lookup);


	int idxA = row * uDim;
	int idxB = idxA + 2 * uDim;
	if(cpIndex[idxA] == cpIndex[idxB]){
		cpIndex[idxA + 1 * uDim] = cpIndex[idxA];
	}
	idxA += uDim - 1;
	idxB += uDim - 1;

	if(cpIndex[idxA] == cpIndex[idxB]){
		cpIndex[idxA + 1 * uDim] = cpIndex[idxA];
	}

}

void NurbsPatch::insertRow(M_Vec3f& oldPoints, M_Float& oldWeights, M_Int& oldCpIndex, const int row, const float ratio, int& vDim){
	insertRow2(dim.x, dim.y, dpoint.refValue(), weight.refValue(), cpIndex.refValue(), oldPoints, oldWeights, oldCpIndex, row, ratio);
	if(mirrorPatch){
		mirrorPatch->insertRow2(mirrorPatch->dim.x, mirrorPatch->dim.y, mirrorPatch->dpoint.refValue(), mirrorPatch->weight.refValue(), mirrorPatch->cpIndex.refValue(), oldPoints, oldWeights, oldCpIndex, row, ratio);
	}
	vDim = dim.y;
}

void NurbsPatch::deleteRow(const int uDim, const int vDim, M_Int& cpIndex, M_Vec3f& dpoint, M_Float& weight, M_Int& oldCpIndex, M_Vec3f& oldPoints, M_Float& oldWeights, const int row){
	//lookup.length = dpoint.length;
	oldCpIndex = cpIndex;
	oldPoints = dpoint;
	oldWeights = weight;

	int newV = vDim - 1;
	int newLen = uDim * newV;

	cpIndex.length = newLen;
	dpoint.length = newLen;
	weight.length = newLen;

	for(int v = 0, nv=0; v < vDim; v++){//row
		if(v == row){
			continue;
		}
		for(int u = 0; u<uDim; u++){//collum
			int nIdx = nv * uDim + u;
			int oIdx = v * uDim + u;
			//lookup[oIdx] = nIdx;
			cpIndex[nIdx] = oldCpIndex[oIdx];
			dpoint[nIdx] = oldPoints[oIdx];
			weight[nIdx] = oldWeights[oIdx];
		}
		nv++;
	}
	dim.refValue().y = newV;
	//smls->redirectStitchPolyIndices(this, lookup);
}

void NurbsPatch::doDim(const Vec2f par, const float ratio, M_Vec3f& dpoint, M_Float& weight, Vec2f&, Vec2f& dim, int& idx, int uDim, int vDim, M_Int& lookup){
	lookup.length = dpoint.length;

	M_Vec3f oldPoints = dpoint;
	M_Float oldWeights = weight;
	M_Int oldCpIndex = cpIndex;

	int col = idx % uDim;
	int row = idx / uDim;

	if(par.x > dim.x){
		insertCol(oldPoints, oldWeights, oldCpIndex, col, ratio, uDim, false);
		col++;
	}
	else if(par.x < dim.x){
		deleteCol(uDim, vDim, cpIndex.refValue(), dpoint, weight, oldCpIndex, oldPoints, oldWeights, col);
		uDim--;
	}
	else if(par.y > dim.y){
		insertRow(oldPoints, oldWeights, oldCpIndex,  row, ratio, vDim);
		row++;
	}
	else if(par.y < dim.y){
		deleteRow(uDim, vDim, cpIndex.refValue(), dpoint, weight, oldCpIndex, oldPoints, oldWeights,  row);
		vDim--;
	}

	dim = Vec2f(uDim, vDim);
	if(col >= uDim){
		col--;
	}
	idx = row * vDim + col;

	smls->redirectStitchPolyIndices(this, lookup);
}


void NurbsPatch::Dim::operator=(const Vec2f&){
	//M_Int lookup;

	//M_Vec3f& dpoint = home.dpoint.refValue();
	//M_Float& weight = home.weight.refValue();
	//Vec2f& order = home.order.refValue();
	//M_Float& uKnot = home.uKnot.refValue();
	//M_Float& vKnot = home.vKnot.refValue();

	//if(dpoint.length==0){
	//	refValue()=par;
	//	home.processFields();
	//	MultiSelect::pointList.length = 0;
	//	return;
	//}

	//int& idx = home.index[0];
	//int remIdx = idx;

	//Vec2f& dim=home.dim.refValue();
	//int uDim = dim.x;
	//int vDim = dim.y;
	//int uOldDim = uDim;
	//int vOldDim = vDim;

	//home.doDim(par, .5, dpoint, weight, order, dim, idx, uDim, vDim, lookup);
	//if(home.showAll == false && uDim > 1){
	//	Vec2f par2 = dim;
	//	if(uDim < par.x){
	//		if(idx % uDim == 0){
	//		}
	//		else{
	//			idx += 2;
	//		}

	//		uDim++;
	//		par2.x = uDim + 1;
	//	}
	//	else if(uDim > par.x){
	//		uDim--;
	//		par2.x = uDim - 1;
	//	}

	//	if(vDim < par.y){
	//		vDim++;
	//		par2.y = vDim + 1;
	//	}
	//	else if(vDim > par.y){
	//		vDim--;
	//		par2.y = vDim - 1;
	//	}

	//	uDim = dim.x;
	//	vDim = dim.y;
	//	home.doDim(par2, .5, dpoint, weight, order, dim, idx, uDim, vDim, lookup);
	//}

	//updateKnot(uOldDim, dim.x, uKnot, order.x);
	//updateKnot(vOldDim, dim.y, vKnot, order.y);
	//
	//home.dpointsToPoints();
	//home.synchronizeAllStemCells();

	//if(idx >= dpoint.length){
	//	idx = dpoint.length-1;
	//}

	//home.processFields();
	//home.smls->populateMirrorNodes();
	//MultiSelect::pointList.length = 0;

}
void NurbsPatch::dpointsToPoints(){
	M_Vec3f& point = this->point.refValue();
	M_Vec3f& dpoint = this->dpoint.refValue();

	point.length = dpoint.length;

	for(int i = dpoint.length; --i>=0;){
		point[i] = dpoint[i] * weight[i];
	}
}


void NurbsPatch::pointsToDpoints(){
	M_Vec3f& point = this->point.refValue();
	M_Vec3f& dpoint = this->dpoint.refValue();

	dpoint.length = point.length;

	for(int i = dpoint.length; --i>=0;){
		dpoint[i] = point[i] / weight[i];
	}
}

/*
void BezierPatch::NumOfV::operator=(const M_Int& m_par){
	M_Vec3f& point = home.point.refValue();

	if(point.length==0){
		refValue()=m_par;
		home.processFields();
		return;
	}
	int numOfUGroupIdx = home.index[home.numOfUGroup];

	int numOfU=home.numOfU.refValue()[numOfUGroupIdx];
	int numOfV=home.numOfV.refValue()[numOfUGroupIdx];
	int par=m_par[numOfUGroupIdx];
	int newValue=numOfV;
	if(par>numOfV){
		newValue++;
	}
	if(par<numOfV){
		newValue--;
	}
	M_Vec3f oldPoints=point;


	int newLen=newValue*numOfU;
	int idx=home.index[0];
	int col=idx%numOfU;
	int row=idx/numOfU;



	if(newLen>point.length){//insert row
		point.length=newLen;
		for(int v = 0, nv=0; v < numOfV; v++, nv++){//row
			if(v!=row){
				for(int u = 0; u<numOfU; u++){//collum
					point[nv*numOfU+u]=oldPoints[v*numOfU+u];
				}
			}
			else{
				if(v < numOfV-1){
					nv++;
					for(int u = 0; u<numOfU; u++){//collum
						Vec3f va = oldPoints[v*numOfU+u];
						Vec3f vb = oldPoints[(v+1)*numOfU+u];
						point[nv*numOfU+u] = va + (vb-va)*.5;
					}
				}
				else{
					for(int u = 0; u<numOfU; u++){//collum
						point[(nv+1)*numOfU+u] = point[nv*numOfU+u];
						Vec3f vb = oldPoints[v*numOfU+u];
						Vec3f va = oldPoints[(v-1)*numOfU+u];
						point[nv*numOfU+u] = va + (vb-va)*.5;
					}
				}
			}
		}
	}
	else{
		point.length=newLen;
		for(int v = 0, nv=0; v < numOfV; v++){//row
			if(v==row){
				continue;
			}
			for(int u = 0; u<numOfU; u++){//collum
				point[nv*numOfU+u]=oldPoints[v*numOfU+u];
			}
			nv++;
		}
	}
	refValue()=newValue;
}
*/
void NurbsPatch::processFields(){
	const int uDim=dim.x;
	const int vDim=dim.y;

	uDegree = order.x-1;
	vDegree = order.y-1;

	M_Vec3f& point = this->point.refValue();
	M_Vec3f& dpoint = this->dpoint.refValue();
	if(point.length == 0){
		return;
	}

	dpoint.length = point.length;

	if(uJoinEnds){
		for(int v = 0; v < vDim; v++){
			point[v*uDim+uDim-1]=point[v*uDim];
		}
	}
	if(showAll){
		if(beginSP){
			for(int u = 1; u < uDim; u++){
				point[u] = point[0] * weight[u];
			}
		}
		if(endSP){
			int o = uDim * (vDim-1);
			for(int u = 1; u < uDim; u++){
				point[o+u] = point[o] * weight[o+u];
			}
		}
	}
	else{
		if(beginSP){
			const int o = 0;
			const int o2 = uJoinEnds ? o + 1 : o;
			if(index[0]<uDim){
				index[0] = o2;
			}
			Vec3f cp = point[o2]/weight[o2];
			for(int u = 1; u < uDim; u++){
				point[o+u] = cp * weight[o+u];
			}
		}
		if(endSP){
			const int o = uDim * (vDim-1);
			const int o2 = uJoinEnds ? o + 1 : o;
			if(index[0]>=o){
				index[0] = o2;
			}
			Vec3f cp = point[o2]/weight[o2];
			for(int u = 1; u < uDim; u++){
				point[o+u] = cp * weight[o+u];
			}
		}
	}


	for(int i = dpoint.length; --i>=0;){
		dpoint[i] = point[i] / weight[i];
	}

	int shouldBeLen=uDim*vDim;

	if(shouldBeLen != point.length){
		point.length = shouldBeLen;
		for(int v = 0; v < vDim; v++){
			for(int u = 0; u < uDim; u++){
				point[v*uDim+u]=Vec3f(1.0/(uDim-1) * u - .5, 1.0/(vDim-1) * v, 0);
			}
		}
	}
	if(showAll == false){
		if(uJoinEnds){
			if(isCtrlPolyRef){
				for(int v = 1; v < vDim - 1; v++){
					int i = v * uDim;
					dpoint[i] = (dpoint[i + 1] + dpoint[i + uDim - 2]) * .5;
					point[i] = dpoint[i] * weight[i];

					i = v * uDim + uDim - 1;
					dpoint[i] = (dpoint[i - 1] + dpoint[i - uDim + 2]) * .5;
					point[i] = dpoint[i] * weight[i];
				}
			}
			else{
				for(int v = 0; v < vDim; v++){
					int i = v * uDim;
					dpoint[i] = (dpoint[i + 1] + dpoint[i + uDim - 2]) * .5;
					point[i] = dpoint[i] * weight[i];

					i = v * uDim + uDim - 1;
					dpoint[i] = (dpoint[i - 1] + dpoint[i - uDim + 2]) * .5;
					point[i] = dpoint[i] * weight[i];
				}
			}
		}
		if(MID_1 == checkNurbsType(uDim, uKnot)){
			if(isCtrlPolyRef){
				for(int v = 1; v < vDim - 1; v++){
					for(int u = 2; u < uDim - 1; u += 2){
						int i = v * uDim + u;
						dpoint[i] = (dpoint[i - 1] + dpoint[i + 1]) * .5;
						point[i] = dpoint[i] * weight[i];
					}
				}
			}
			else{
				for(int v = 0; v < vDim; v++){
					for(int u = 2; u < uDim - 1; u += 2){
						int i = v * uDim + u;
						dpoint[i] = (dpoint[i - 1] + dpoint[i + 1]) * .5;
						point[i] = dpoint[i] * weight[i];
					}
				}
			}
		}
		if(MID_1 == checkNurbsType(vDim, vKnot)){
			if(isCtrlPolyRef){
				for(int v = 2; v < vDim - 1; v += 2){
					for(int u = 1; u < uDim - 1; u++){
						int i = v * uDim+u;
						dpoint[i] = (dpoint[i - uDim] + dpoint[i + uDim]) * .5;
						point[i] = dpoint[i] * weight[i];
					}
				}
			}
			else{
				for(int v = 2; v < vDim - 1; v += 2){
					for(int u = 0; u < uDim; u++){
						int i = v * uDim+u;
						dpoint[i] = (dpoint[i - uDim] + dpoint[i + uDim]) * .5;
						point[i] = dpoint[i] * weight[i];
					}
				}
			}
		}
	}

	r_point.length = vDim;
	r_weight.length = vDim;
	for(int i=0; i<vDim; i++){
		r_point[i] = point[i*uDim+uDim-1];
		r_weight[i] = weight[i*uDim+uDim-1];
	}
}

void ColorPatch::processFields(){
	NurbsPatch::processFields();
	color.refValue().length = point.length;
}

void Harmonic::processFields(){
	degree = order-1;

	M_Vec2f& point = this->point.refValue();
	M_Vec2f& dpoint = this->dpoint.refValue();

	dpoint.length=point.length;

	if(weight.length==0){
		dpoint = point;
	}
	else{
		for(int i=dpoint.length; --i>=0;){
			dpoint[i] = point[i] / weight[i];
		}
	}
}

void Oscillator::processFields(){
	fDegree = fOrder - 1;

	M_Vec2f& fPoint = this->note.refValue();
	M_Vec2f& fdpoint = this->fdpoint.refValue();

	fdpoint.length = fPoint.length;

	if(fWeight.length==0){
		fdpoint = fPoint;
	}
	else{
		for(int i=fdpoint.length; --i>=0;){
			fdpoint[i] = fPoint[i] / fWeight[i];
		}
	}

	if(Larynx* larynx = dynamic_cast<Larynx*>(this->larynx.p)){
		M_Node& m_chords = larynx->chords.refValue();
		int limitIdx = 0;
		for(int k = 0; k < m_chords.length; k++){
			if(Chords* chords = dynamic_cast<Chords*>(m_chords[k].p)){
				if(limitIdx >= limit.length){
					limit += chords->multiple.length;
				}
				if(limitIdx >= chordsBufLen.length){
					chordsBufLen += 0;
					chordsBuf += 0;
				}
				limitIdx++;
			}
			else if(Envelope* envelope = dynamic_cast<Envelope*>(m_chords[k].p)){
				M_Node& m_chords = envelope->chords.refValue();
				for(int k = 0; k < m_chords.length; k++){
					if(Chords* chords = dynamic_cast<Chords*>(m_chords[k].p)){
						if(limitIdx >= limit.length){
							limit += chords->multiple.length;
						}
						if(limitIdx >= chordsBufLen.length){
							chordsBufLen += 0;
							chordsBuf += 0;
						}
						limitIdx++;
					}
				}
			}
		}
	}

	for(int k = filter.length; --k>=0;){
		if(NurbsPatch* patch = dynamic_cast<NurbsPatch*>(filter.refValue()[k].p)){
			patch->processFields();
			patch->updateVCoord(*this);
		}
	}
/*
	M_Float& phase = this->phase.refValue();
	M_Float& multiple = this->multiple;
	M_Float& amplitude = this->amplitude;

	if(multiple.length == 0){
		//multiple.length = 92;
		multiple.length = 2000;
		amplitude.length = multiple.length;
		for(int k = 0; k < multiple.length; k++){
			multiple[k] = k * .03 + 1;
			//multiple[k] = k + 1;
			amplitude[k] = 1.0 / (k + 1);
		}
	}
	if(multiple.length == 0 || phase.length != multiple.length){
	//if(amplitude.length != multiple.length){
		phase.length = multiple.length;
		amplitude.length = multiple.length;

		//if(phase.length > 400){
		if(1){
			double range_minf = 0;
			double range_maxf = 1;
			for(int k = 0; k < phase.length; k++){
				phase[k] = ((double)rand() / (RAND_MAX + 1) * (range_maxf - range_minf) + range_minf);
				amplitude[k] = 1;
			}
		}
		else{
			for(int k = 0; k < phase.length; k++){
				if(k & 1){
					phase[k] = .5;
				}
				else{
					phase[k] = 0;
					amplitude[k] = 0;
				}
			}
		}
	}
*/
}


void Envelope::processFields(){
	M_Vec2f& point = this->point.refValue();
	if(point.length){
		weight.length = point.length;

		for(int i = point.length; --i>=0;){
			weight[i] = 1;
		}
		updateNurbsCurveKnotMid0(point.length, knot);
		coord.length = resolution;

		updateCoord();
	}
}

void AmplitudeModulator::processFields(){
	M_Vec2f& point = this->wet.refValue();
	if(point.length){
		weight.length = point.length;

		for(int i = point.length; --i>=0;){
			weight[i] = 1;
		}
		updateNurbsCurveKnotMid0(point.length, knot);
		coord.length = resolution;

		updateWetCoord();
	}
}

void FrequencyModulator::processFields(){
	M_Vec2f& point = this->wet.refValue();
	if(point.length){
		weight.length = point.length;

		for(int i = point.length; --i>=0;){
			weight[i] = 1;
		}
		updateNurbsCurveKnotMid0(point.length, knot);
		coord.length = resolution;

		updateWetCoord();
	}
}

void Chords::processFields(){
	M_Vec2f& point = this->spectrum.refValue();
	if(point.length){
//		point[0].x = .022;
		double freq = .022 * pow((pow(2.0, 1.0 / 12.0)), double(interval));
		point[0].x = freq;

		weight.length = point.length;

		for(int i = point.length; --i>=0;){
			weight[i] = 1;
		}
		updateNurbsCurveKnotMid0(point.length, knot);
		coord.length = harmonics;

		if(patials.refValue()){
			updateSpectrumPatch();
		}
		else{
			updateSpectrum();
		}
	}
}

void ECLathe::processFields(){
	if(point.length==0 || point.length!=tangent.length)return;
	processFieldsECLatheECSyncBend(point, tangent, a, b, scaleDis, distance, height);
	return;
}

void ECBend::processFields(){
	if(point.length==0)return;

	matrix=rotationMatrix(orientation)*translationMatrix(location);
	imatrix=translationMatrix(-location)*rotationMatrix(-orientation);

	copyPointTangentToXZ();
/*
	dLog("----x----");
	dLog(pointX[0].x);
	dLog(pointX[0].y);
	dLog(pointX[1].x);
	dLog(pointX[1].y);
	dLog("----z----");
	dLog(pointZ[0].x);
	dLog(pointZ[0].y);
	dLog(pointZ[1].x);
	dLog(pointZ[1].y);
*/
//	processFieldsECLatheECSyncBend(pointX, immediateTangent, a, b, scaleDis, distance, height);

	return;
}
void ECSyncBend::processFields(){
	if(point.length==0 || point.length!=tangent.length)return;
	processFieldsECLatheECSyncBend(point, tangent, a, b, scaleDis, distance, height);
	matrix=rotationMatrix(orientation)*translationMatrix(location);
	imatrix=translationMatrix(-location)*rotationMatrix(-orientation);
	return;
}

void EvolveCoord::processFields(){
	matrix=rotationMatrix(orientation)*translationMatrix(location);
	imatrix=translationMatrix(-location)*rotationMatrix(-orientation);
	m_ecurve.length=point.length;
	setUpMatrices(m_ecurve, point, tangent,a,b);
	return;
}
void ECScale::processFields(){
	matrix=rotationMatrix(orientation)*translationMatrix(location);
	imatrix=translationMatrix(-location)*rotationMatrix(-orientation);
	m_ecurve.length=point.length;
	setUpMatrices(m_ecurve, point, tangent,a,b);
	return;
}
void ECShift::processFields(){
	matrix=rotationMatrix(orientation)*translationMatrix(location);
	imatrix=translationMatrix(-location)*rotationMatrix(-orientation);
	m_ecurve.length=point.length;
	setUpMatrices(m_ecurve, point, tangent,a,b);
	return;
}
void ECTwist::processFields(){
	matrix=rotationMatrix(orientation)*translationMatrix(location);
	imatrix=translationMatrix(-location)*rotationMatrix(-orientation);
	m_ecurve.length=point.length;
	setUpMatrices(m_ecurve, point, tangent,a,b);
	return;
}
/*
void Progressor::processFields(){
	matrix=rotationMatrix(orientation)*translationMatrix(location);
	imatrix=translationMatrix(-location)*rotationMatrix(-orientation);
	m_ecurve.length=point.length;
	setUpMatrices(m_ecurve, point, tangent,a,b);
	return;
}
*/
void SineTug::processFields(){
	M_Float& decay=this->decay.refValue();
	M_Float& xScale=this->xScale.refValue();

	int oldSize=decay.length;
	decay.length=distance.length;
	int i;
	for(i=oldSize; i<decay.length; i++){
		decay[i]=1;
	}
	oldSize=xScale.length;
	xScale.length=distance.length;
	for(i=oldSize; i<xScale.length; i++){
		xScale[i]=1;
	}
	return;
}

void RadialTug::processFields(){
	M_Float& slope=this->slope.refValue();

	int oldSize=slope.length;
	slope.length=displacement.length;
	int i;
	for(i=oldSize; i<slope.length; i++){
		slope[i]=1;
	}
	return;
}

void SphereBuild::processFields(){


	if(part==0){
		return;
	}
	Matrix matrix = scaleMatrix(scale) * rotationMatrix(rotation) * translationMatrix(translation);


	int uDim = ctrlStaves * 2 + 1;
	int vDim = ctrlBands * 2 + 1;

	stem->textureMapper.refValue().length = 0;
	if(textureMapper){
		stem->textureMapper.refValue()+=textureMapper.refValue();
	}
	else{
		colorSweep->color.refValue().length = 0;
		colorSweep->color.refValue() += Color3f(color);
		colorSweep->color.refValue() += Color3f(color);
		colorSweep->straight.refValue() = true;
		colorSweep->distance.refValue().length = 0;
		colorSweep->distance.refValue() += 1;
		stem->textureMapper.refValue() += colorSweep.p;
	}



	M_Vec3f& patchDPoint = patch->dpoint.refValue();
	M_Float& patchWeight = patch->weight.refValue();

	patch->dim.refValue() = Vec2f(uDim, vDim);

	patchDPoint.length = vDim * uDim;
	patchWeight.length = vDim * uDim;

	int rings = vDim-1;
	double sinWeight = cos(PI / ctrlStaves);

	for(int r = 0; r < vDim; r++){

		double vRatio = 1;
		double vSinWeight = 1;
		if(r&1){
			vSinWeight = cos(PI/rings);
			vRatio = 1.0 / vSinWeight;
		}
		Vec2d point = Vec2d(vRatio*cos(-PI*.5+PI/(vDim-1)*r), vRatio*(sin(-PI*.5+PI/(vDim-1)*r))+!yCentred)*radius;

		double a = PI*2/(ctrlStaves*2);
		for(int c = 0; c < uDim; c+=2){
			double b = a*c;
			patchDPoint[r*uDim + c] = Vec3f(-sin(b)*point.x, point.y, -cos(b)*point.x)*vSinWeight;
			patchWeight[r*uDim + c] = vSinWeight*1;
		}
		double uDiv = cos(a);
		double uRatio = 0;
		if(uDiv!=0){
			uRatio = 1.0 / uDiv;
		}

		for(int c = 1; c < uDim; c+=2){
			double b = a*c;
			patchDPoint[r*uDim + c] = Vec3f(-sin(b)*point.x*uRatio, point.y, -cos(b)*point.x*uRatio);
			patchWeight[r*uDim + c] = vSinWeight*sinWeight;
		}
	}
	for(int i = patchDPoint.length; --i>=0;){
		patchDPoint[i] = matrix * patchDPoint[i];
	}
	patch->dpointsToPoints();

	patch->insideOut.refValue() = insideOut;
	patch->mirror.refValue() = mirror;
	if(insideOut){
		patch->setInsideOut();
		patch->setInsideOut3();
	}

	updateNurbsCurveKnotMid1(uDim, patch->uKnot.refValue());
	updateNurbsCurveKnotMid1(vDim, patch->vKnot.refValue());

	stem->staves.refValue().length = 0;
	stem->bands.refValue().length = 0;
	stem->staves.refValue() += staves;
	stem->bands.refValue() += bands;
	stem->endSP.refValue() = true;
	stem->beginSP.refValue() = true;

	stem->part.refValue().length = 0;
	if(part){
		stem->part+=part.refValue();
	}

	colorSweep->processFields();
	stem->processFields();
	patch->processFields();
	//patch->patchToCentralCPs(); //why is this not needed if the sphere was created (as opposed to a sphere existing in an opened file) ?
}

void doCap(const int begin, const int end, const int m, const int _rings, const double yOffset, const double xRadius, const double yRadius, const int ctrlStaves, const int dimU, M_Vec3f& patchPoint, M_Float& patchWeight, double beginAngle){
	double sinWeight = cos(PI / ctrlStaves);

	int rings = _rings - 1;
	for(int r = begin; r <= end; r++){

		double vRatio = 1;
		double vSinWeight = 1;
		double xRad = xRadius;
		if(r & 1){
			vSinWeight = cos(PI/rings);
			vRatio = 1.0 / vSinWeight;
			if(yRadius == 0){
				xRad = xRadius * .5;
			}
		}

		Vec2d point = Vec2d(xRad * (vRatio * cos(beginAngle + PI / (_rings - 1) * r)), yOffset + yRadius * (vRatio * sin(beginAngle + PI / (_rings-1) * r)));

		double a = PI*2 / (ctrlStaves * 2);
		for(int c = 0; c < dimU; c+=2){
			double b = a*c;
			patchPoint[m + r * dimU + c] = Vec3f(-sin(b) * point.x, point.y, -cos(b) * point.x) * vSinWeight;
			patchWeight[m + r * dimU+c] = vSinWeight * 1;
		}
		double uDiv = cos(a);
		double uRatio = 0;
		if(uDiv != 0){
			uRatio = 1.0 / uDiv;
		}

		for(int c = 1; c < dimU; c+=2){
			double b = a * c;
			patchPoint[m + r * dimU + c] = Vec3f(-sin(b) * point.x * uRatio, point.y, -cos(b) * point.x * uRatio);
			patchWeight[m + r * dimU + c] = vSinWeight * sinWeight;
			patchPoint[m + r * dimU + c] *= vSinWeight * sinWeight;
		}
	}
}


void CylinderBuild::processBaseCap(NurbsPatch* patch, int uDim, Matrix& matrix){
	int vDim = 3;
	int len = uDim * vDim;
	M_Vec3f& point = patch->point.refValue();
	M_Float& weight = patch->weight.refValue();
	point.length = len;
	weight.length = len;
	patch->dim.refValue() = Vec2f(uDim, vDim);

	double yOffset = baseHeight;

	if(yCentred){
		yOffset = -height*.5;
	}

	doCap(0, 2, 0, 5, yOffset, radius, baseHeight, ctrlStaves, uDim, point, weight, -PI * .5);

	for(int i = point.length; --i>=0;){
		point[i] = matrix * (point[i] / weight[i]) * weight[i];
	}

	updateNurbsCurveKnotMid1(uDim, patch->uKnot.refValue());
	updateNurbsCurveKnotMid1(vDim, patch->vKnot.refValue());

	if(patch->stem.refValue().length){
		Stem* stem = dynamic_cast<Stem*>(patch->stem[0].p);
		M_Float& yDistance = stem->yDistance.refValue();
		M_Float& xDistance = stem->xDistance.refValue();
		M_Int& staves = stem->staves.refValue();
		M_Int& bands = stem->bands.refValue();

		yDistance.length = 1;
		yDistance[0] = 1;
		xDistance.length = 1;
		xDistance[0] = 1;
		staves.length = 1;
		staves[0] = this->staves;
		bands.length = 1;
		bands[0] = baseBands;

		stem->part.refValue().length = 0;
		stem->part += part.refValue();

		stem->processFields();
	}

	patch->processFields();

}

void CylinderBuild::processShaft(NurbsPatch* patch, int uDim, Matrix& matrix){
	int _rings = ctrlBands * 2 + 1;

	double sinWeight = cos(PI / ctrlStaves);


	M_Vec3f& point = patch->point.refValue();
	M_Float& weight = patch->weight.refValue();

	int patchLen = _rings;
	double yOffset = baseHeight;
	patchLen *= uDim;

	point.length = patchLen;
	weight.length = patchLen;

	int rings = _rings-1;
	if(yCentred){
		yOffset = -height*.5;
	}
	for(int r = 0; r < _rings; r++){
		double a = PI / ctrlStaves;
		double y = height / rings * r + yOffset;
		for(int c = 0; c < uDim; c+=2){
			double b = a*c;
			point[r * uDim + c] = Vec3f(-sin(b) * radius, y, -cos(b) * radius);
			weight[r * uDim + c] = 1;
		}
		double div = cos(a) * radius;
		double ratio = 0;
		if(div != 0){
			ratio = -point[r * uDim].z / div;
		}

		for(int c = 1; c < uDim; c+=2){
			double b = a * c;
			point[r * uDim+c] = Vec3f(-sin(b) * radius * ratio, y, -cos(b) * radius * ratio);
			weight[r * uDim+c] = sinWeight;
			point[r * uDim+c] *= sinWeight;
		}
	}

	for(int i = point.length; --i>=0;){
		point[i] = matrix * (point[i] / weight[i]) * weight[i];
	}
	updateNurbsCurveKnotMid1(uDim, patch->uKnot.refValue());
	updateNurbsCurveKnotMid1(_rings, patch->vKnot.refValue());
	patch->dim.refValue() = Vec2f(uDim, _rings);

	patch->processFields();

	if(patch->stem.refValue().length){
		Stem* stem = dynamic_cast<Stem*>(patch->stem[0].p);
		M_Float& yDistance = stem->yDistance.refValue();
		M_Float& xDistance = stem->xDistance.refValue();
		M_Int& staves = stem->staves.refValue();
		M_Int& bands = stem->bands.refValue();

		yDistance.length = 1;
		yDistance[0] = 1;
		xDistance.length = 1;
		xDistance[0] = 1;
		staves.length = 1;
		staves[0] = this->staves;
		bands.length = 1;
		bands[0] = this->bands;

		stem->part.refValue().length = 0;
		stem->part += part.refValue();

		stem->processFields();
	}
}

void CylinderBuild::processTopCap(NurbsPatch* patch, int uDim, Matrix& matrix){
	int vDim = 3;
	int len = uDim * vDim;
	M_Vec3f& point = patch->point.refValue();
	M_Float& weight = patch->weight.refValue();
	point.length = len;
	weight.length = len;
	patch->dim.refValue() = Vec2f(uDim, vDim);

	double yOffset = baseHeight;

	if(yCentred){
		yOffset = -height*.5;
	}

	doCap(0, 2, 0, 5, yOffset + height, radius, topHeight, ctrlStaves, uDim, point, weight, 0);

	for(int i = point.length; --i>=0;){
		point[i] = matrix * (point[i] / weight[i]) * weight[i];
	}

	updateNurbsCurveKnotMid1(uDim, patch->uKnot.refValue());
	updateNurbsCurveKnotMid1(vDim, patch->vKnot.refValue());

	if(patch->stem.refValue().length){
		Stem* stem = dynamic_cast<Stem*>(patch->stem[0].p);
		M_Float& yDistance = stem->yDistance.refValue();
		M_Float& xDistance = stem->xDistance.refValue();
		M_Int& staves = stem->staves.refValue();
		M_Int& bands = stem->bands.refValue();

		yDistance.length = 1;
		yDistance[0] = 1;
		xDistance.length = 1;
		xDistance[0] = 1;
		staves.length = 1;
		staves[0] = this->staves;
		bands.length = 1;
		bands[0] = topBands;

		stem->part.refValue().length = 0;
		stem->part += part.refValue();

		stem->processFields();
	}

	patch->processFields();
}
void setColorSweeps(M_<NurbsPatchPtr>& m_patch, M_Node& textureMapper, ColorSweep* colorSweep, Color3f& color, Art* part){
	Node* latstTM = 0;

	for(int k = 0; k < m_patch.length; k++){
		NurbsPatch* patch = m_patch[k];
		Stem* stem = 0;
		if(patch->stem.refValue().length){
			stem = dynamic_cast<Stem*>(patch->stem.refValue()[0].p);
			if(k < textureMapper.length){
				latstTM = textureMapper[k];
			}
			stem->textureMapper.refValue().length = 0;
			if(latstTM){
				stem->textureMapper += latstTM;
			}
			if(latstTM == 0){
				colorSweep->color.refValue().length = 0;
				colorSweep->color.refValue() += Color3f(color);
				colorSweep->color.refValue() += Color3f(color);
				colorSweep->straight.refValue() = true;
				//colorSweep->distance.refValue().length = 0; //why does this cause a crash in release mode?????????????????
				colorSweep->distance.refValue() = 1;
				stem->textureMapper.refValue() += colorSweep;
			}

			stem->part.refValue().length = 0;
			stem->part += part;
		}
	}
}


void CylinderBuild::processFields(){

	setColorSweeps(m_patch, textureMapper.refValue(), colorSweep, color, part.refValue());

	Matrix matrix = scaleMatrix(scale) * rotationMatrix(rotation) * translationMatrix(translation);

	int uDim = ctrlStaves * 2 + 1;

	processShaft(patch, uDim, matrix);
	//patch->patchToCentralCPs(); //why is this not needed if the cylinder was created (as opposed to a cylinder existing in an opened file) ?

	if(baseCap){
		processBaseCap(basePatch, uDim, matrix);
		//basePatch->patchToCentralCPs(); //why is this not needed if the cylinder was created (as opposed to a cylinder existing in an opened file) ?
	}
	if(topCap){
		processTopCap(topPatch, uDim, matrix);
		//topPatch->patchToCentralCPs(); //why is this not needed if the cylinder was created (as opposed to a cylinder existing in an opened file) ?
	}

	colorSweep->processFields();
}

void ConeBuild::processBaseCap(NurbsPatch* patch, int uDim, Matrix& matrix){
	int vDim = 3;
	int len = uDim * vDim;
	M_Vec3f& point = patch->point.refValue();
	M_Float& weight = patch->weight.refValue();
	point.length = len;
	weight.length = len;
	patch->dim.refValue() = Vec2f(uDim, vDim);

	double yOffset = 0;

	if(yCentred){
		yOffset = -height*.5;
	}

	//doCap(0, 2, 0, 5, yOffset, baseRad, height, ctrlStaves, uDim, point, weight, -PI * .5);
	doCap(0, 2, 0, 5, yOffset, baseRad, 0, ctrlStaves, uDim, point, weight, -PI * .5);

	for(int i = point.length; --i>=0;){
		point[i] = matrix * (point[i] / weight[i]) * weight[i];
	}

	updateNurbsCurveKnotMid1(uDim, patch->uKnot.refValue());
	updateNurbsCurveKnotMid1(vDim, patch->vKnot.refValue());

	if(patch->stem.refValue().length){
		Stem* stem = dynamic_cast<Stem*>(patch->stem[0].p);
		M_Float& yDistance = stem->yDistance.refValue();
		M_Float& xDistance = stem->xDistance.refValue();
		M_Int& staves = stem->staves.refValue();
		M_Int& bands = stem->bands.refValue();

		yDistance.length = 1;
		yDistance[0] = 1;
		xDistance.length = 1;
		xDistance[0] = 1;
		staves.length = 1;
		staves[0] = this->staves;
		bands.length = 1;
		bands[0] = baseBands;

		stem->part.refValue().length = 0;
		stem->part += part.refValue();

		stem->processFields();
	}

	patch->processFields();
}

void ConeBuild::processShaft(NurbsPatch* patch, int uDim, Matrix& matrix){
	lathe->point.refValue().length=0;
	lathe->point.refValue()+=Vec2f(baseRad, 0);

	if(baseRad>topRad){
		float radDiff = baseRad - topRad;
		float ratio = topRad < .0001 ? .0001 : topRad/baseRad;
		ratio*=.5;
		lathe->point.refValue()+=Vec2f(topRad+radDiff*ratio, height*(1.0-ratio));
		//lathe->point.refValue()+=Vec2f(topRad+radDiff*ratio, height*(1.0-ratio));
	}
	else{
		float radDiff = topRad - baseRad;
		float ratio = baseRad < .0001 ? .0001 : baseRad/topRad;
		ratio*=.5;
		lathe->point.refValue()+=Vec2f(topRad-radDiff*(1.0 - ratio), height * ratio);
		//lathe->point.refValue()+=Vec2f(topRad-radDiff*(1.0-ratio)+ratio*.005f, height*ratio);
	}
	lathe->point.refValue() += Vec2f(topRad, height);
	lathe->fix.refValue().length = 0;
	lathe->fix.refValue()+= true;
	lathe->fix.refValue()+= false;
	//lathe->fix.refValue()+=false;
	lathe->fix.refValue()+= true;
	lathe->processFields();



	int _rings = ctrlBands * 2 + 1;

	double sinWeight = cos(PI / ctrlStaves);

	const M_Vec2f& lathePoint = lathe->point;
	const M_Int& nopfs = lathe->nopfs;
	M_Vec2f& line = lathe->line;

	M_Vec3f& patchPoint = patch->point.refValue();
	M_Float& patchWeight = patch->weight.refValue();

	int patchLen = _rings;
	double yOffset = 0;;
	patchLen *= uDim;

	patchPoint.length = patchLen;
	patchWeight.length = patchLen;

	int rings = _rings - 1;
	float step = 1.0 / rings;
	float t = 0;

	line.length = 0;
	M_Vec2f pnt;
	pnt.length = nopfs[0];
	for(int r = 0; r < _rings; r++){
	//for (int k = 1; k < numOfSteps; k++){
		line += lathe->genCurve3(lathePoint.length, 0, 0, t, pnt);
		t += step;
	}

	if(yCentred){
		yOffset -= height*.5;
	}
	for(int r = 0; r < _rings; r++){

		double a = PI / ctrlStaves;
		Vec2f v = line[r];
		v.y += yOffset;
		for(int c = 0; c < uDim; c+=2){
			double b = a * c;
			patchPoint[r * uDim + c] = Vec3f(-sin(b) * v.x, v.y, -cos(b) * v.x);
			patchWeight[r * uDim + c] = 1;
		}
		double div = cos(a) * v.x;
		double ratio = 0;
		if(div != 0){
			ratio = -patchPoint[r * uDim].z / div;
		}

		for(int c = 1; c < uDim; c+=2){
			double b = a * c;
			patchPoint[r * uDim + c] = Vec3f(-sin(b)*v.x*ratio, v.y, -cos(b)*v.x*ratio);
			patchWeight[r * uDim + c] = sinWeight;
			patchPoint[r * uDim + c] *= sinWeight;
		}
	}
	for(int i = patchPoint.length; --i>=0;){
		patchPoint[i]=matrix * (patchPoint[i] / patchWeight[i]) * patchWeight[i];
	}
	updateNurbsCurveKnotMid1(uDim, patch->uKnot.refValue());
	updateNurbsCurveKnotMid1(_rings, patch->vKnot.refValue());
	patch->dim.refValue()=Vec2f(uDim, _rings);

	patch->processFields();
	if(patch->stem.refValue().length){
		Stem* stem = dynamic_cast<Stem*>(patch->stem[0].p);
		M_Float& yDistance = stem->yDistance.refValue();
		M_Float& xDistance = stem->xDistance.refValue();
		M_Int& staves = stem->staves.refValue();
		M_Int& bands = stem->bands.refValue();

		yDistance.length = 1;
		yDistance[0] = 1;
		xDistance.length = 1;
		xDistance[0] = 1;
		staves.length = 1;
		staves[0] = this->staves;
		bands.length = 1;
		bands[0] = this->bands;

		stem->beginSP.refValue() = false;
		stem->endSP.refValue() = false;
		if(baseRad == 0){
			stem->beginSP.refValue() = true;
		}
		if(topRad == 0){
			stem->endSP.refValue() = true;
		}

		stem->part.refValue().length = 0;
		stem->part += part.refValue();

		stem->processFields();
	}
}

void ConeBuild::processTopCap(NurbsPatch* patch, int uDim, Matrix& matrix){
	int vDim = 3;
	int len = uDim * vDim;
	M_Vec3f& point = patch->point.refValue();
	M_Float& weight = patch->weight.refValue();
	point.length = len;
	weight.length = len;
	patch->dim.refValue() = Vec2f(uDim, vDim);

	double yOffset = 0;

	if(yCentred){
		yOffset = -height*.5;
	}

	doCap(0, 2, 0, 5, yOffset + height, topRad, 0, ctrlStaves, uDim, point, weight, 0);
	//doCap(0, 2, 0, 5, yOffset + height, topRad, 0, ctrlStaves, uDim, point, weight, -PI * .5);

	for(int i = point.length; --i>=0;){
		point[i] = matrix * (point[i] / weight[i]) * weight[i];
	}

	updateNurbsCurveKnotMid1(uDim, patch->uKnot.refValue());
	updateNurbsCurveKnotMid1(vDim, patch->vKnot.refValue());

	if(patch->stem.refValue().length){
		Stem* stem = dynamic_cast<Stem*>(patch->stem[0].p);
		M_Float& yDistance = stem->yDistance.refValue();
		M_Float& xDistance = stem->xDistance.refValue();
		M_Int& staves = stem->staves.refValue();
		M_Int& bands = stem->bands.refValue();

		yDistance.length = 1;
		yDistance[0] = 1;
		xDistance.length = 1;
		xDistance[0] = 1;
		staves.length = 1;
		staves[0] = this->staves;
		bands.length = 1;
		bands[0] = topBands;

		stem->part.refValue().length = 0;
		stem->part += part.refValue();

		stem->processFields();
	}

	patch->processFields();
}

void ConeBuild::processFields(){
	m_patch.length = 0;
	if(baseCap && baseRad > 0){
		m_patch += basePatch;
	}
	m_patch += patch;
	if(topCap && topRad > 0){
		m_patch += topPatch;
	}
/*
	Stem* stem = dynamic_cast<Stem*>(patch->stem.refValue()[0].p);
	stem->beginSP.refValue() = false;
	stem->endSP.refValue() = false;
	if(baseCap == false || baseRad != 0){
		stem->beginSP.refValue() = true;
	}
	if(topCap == false || topRad != 0){
		stem->endSP.refValue() = true;
	}
*/
	setColorSweeps(m_patch, textureMapper.refValue(), colorSweep, color, part.refValue());

	Matrix matrix = scaleMatrix(scale) * rotationMatrix(rotation) * translationMatrix(translation);

	int uDim = ctrlStaves * 2 + 1;

	processShaft(patch, uDim, matrix);
	//patch->patchToCentralCPs(); //why is this not needed if the cone was created (as opposed to a cone existing in an opened file) ?

	if(baseCap && baseRad > 0){
		processBaseCap(basePatch, uDim, matrix);
		//basePatch->patchToCentralCPs(); //why is this not needed if the cone was created (as opposed to a cone existing in an opened file) ?
	}
	if(topCap && topRad > 0){
		processTopCap(topPatch, uDim, matrix);
		//topPatch->patchToCentralCPs(); //why is this not needed if the cone was created (as opposed to a cone existing in an opened file) ?
	}

	//if(baseCap && baseRad > 0){
	//	doCap(0, 2, 0, 5, yOffset, baseRad, 0, ctrlStaves, uDim, patchPoint, patchWeight, -PI * .5);
	//	_rings += baseRings;
	//}
	//if(topCap && topRad > 0){
	//	doCap(3, 4, uDim*(baseRings+rings-2), 5, yOffset+height, topRad, 0, ctrlStaves, uDim, patchPoint, patchWeight, -PI * .5);
	//	_rings += topRings;
	//}
	colorSweep->processFields();
}

void TorusBuild::setStemValues(Stem* stem){
	stem->sync.refValue() = true;
	stem->staves.refValue()[0] = staves;
	stem->bands.refValue().length = 0;
	stem->bands.refValue() += bands;
	stem->sync.refValue() = true;
	stem->processFields();
}

void TorusBuild::processFields(){
	M_Node& externalNode = textureMapper.refValue();
	if(externalNode.length){
		if(dynamic_cast<Stem*>(externalNode[0].p)){
			part.refValue() = 0;
			patch->stem.refValue().length = 0;
			patch->stem.refValue() += externalNode;
		}
		else{
			Stem* stem = dynamic_cast<Stem*>(patch->stem[0].p);
			stem->textureMapper.refValue().length=0;
			stem->textureMapper.refValue()+=externalNode;
			setStemValues(stem);
		}
	}
	else{
		if(patch->stem.length == 0){
			Stem* stem = new Stem;
			stem->xDistance.refValue() = 1;
			stem->xCentred.refValue() = false;
			stem->yDistance.refValue() = 1;
			patch->stem.refValue() += stem;
		}
		Stem* stem = dynamic_cast<Stem*>(patch->stem[0].p);
		ColorSweep* colorSweep = 0;
		if(stem->textureMapper.length == 0){
			colorSweep = new ColorSweep;
			stem->textureMapper.refValue() += colorSweep;
		}
		else{
			colorSweep = dynamic_cast<ColorSweep*>(stem->textureMapper[0].p);
		}

		colorSweep->color.refValue().length = 0;
		colorSweep->color.refValue() += Color3f(color);
		colorSweep->color.refValue() += Color3f(color);
		colorSweep->straight.refValue() = true;
		colorSweep->distance.refValue().length = 0;
		colorSweep->distance.refValue() += 1;
		colorSweep->processFields();

		stem->part.refValue().length = 0;
		if(part){
			stem->part += part.refValue();
		}
		setStemValues(stem);

	}




/////////////////////
	Matrix matrix = scaleMatrix(scale) * rotationMatrix(rotation) * translationMatrix(translation);

	int _rings = (ctrlBands+1)*2-1;

	double sinWeight=cos(PI/ctrlStaves);
	int uDim = ctrlStaves * 2 + 1;

	M_Vec3f& patchPoint = patch->point.refValue();
	M_Float& patchWeight = patch->weight.refValue();

	patchPoint.length = _rings*uDim;
	patchWeight.length = _rings*uDim;
	int rings = _rings-1;
	for(int r = 0; r < _rings; r++){

		double vRatio = 1;
		double vSinWeight = 1;
		if(r & 1){
			vSinWeight = cos(PI*2/rings);
			vRatio = 1.0 / vSinWeight;
		}
		Vec2d point = Vec2d(vRatio*cos(-PI+PI*2/(_rings-1)*r), vRatio*(sin(-PI+PI*2/(_rings-1)*r))+!yCentred)*tubeRad;
		point.x += tubeRad + holeRad;

		double a = PI*2/(ctrlStaves*2);
		for(int c = 0; c < uDim; c+=2){
			double b = a*c;
			patchPoint[r*uDim+c] = Vec3f(-sin(b)*point.x, point.y, -cos(b)*point.x)*vSinWeight;
			patchWeight[r*uDim+c] = vSinWeight*1;
		}
		double uDiv = cos(a);
		double uRatio = 0;
		if(uDiv!=0){
			uRatio = 1.0 / uDiv;
		}

		for(int c = 1; c < uDim; c+=2){
			double b = a*c;
			patchPoint[r*uDim+c] = Vec3f(-sin(b)*point.x*uRatio, point.y, -cos(b)*point.x*uRatio);
			patchWeight[r*uDim+c] = vSinWeight*sinWeight;
			patchPoint[r*uDim+c] *= vSinWeight*sinWeight;
		}
	}
	for(int i = patchPoint.length; --i>=0;){
		patchPoint[i]=matrix * (patchPoint[i] / patchWeight[i]) * patchWeight[i];
	}
	updateNurbsCurveKnotMid1(uDim, patch->uKnot.refValue());
	updateNurbsCurveKnotMid1(_rings, patch->vKnot.refValue());
	patch->dim.refValue()=Vec2f(uDim, _rings);

	patch->processFields();
	//patch->patchToCentralCPs(); //why is this not needed if the torus was created (as opposed to a torus existing in an opened file) ?
}

void commonBoxBuild(NurbsPatch* patch, int uDim, int vDim){

	patch->dim.refValue() = Vec2f(uDim, vDim);
	patch->point.refValue().length = uDim * vDim;
	patch->weight.refValue().length = uDim * vDim;
	updateNurbsCurveKnotMid1(uDim, patch->uKnot.refValue());
	updateNurbsCurveKnotMid1(vDim, patch->vKnot.refValue());
}

void BoxBuild::processFields(){
	if(mirror){
		xCentred.refValue() = true;
	}
	setColorSweeps(m_patch, textureMapper.refValue(), colorSweep, color, part.refValue());

	Matrix matrix = scaleMatrix(scale) * rotationMatrix(rotation) * translationMatrix(translation);
	float xPos = 0;
	float yPos = 0;
	float zPos = 0;
	if(xCentred){
		xPos = -1;
	}
	if(yCentred){
		yPos = -1;
	}
	if(zCentred){
		zPos = -1;
	}

	int dimU = ctrlStaves * 2 + 1;
	int dimV = ctrlBands * 2 + 1;
	int dimZ = ctrlZStaves * 2 + 1;

	for(int k = 0; k < 6; k++){
		NurbsPatch* patch = m_patch[k];

		M_Vec3f& patchPoint = patch->point.refValue();
		M_Float& patchWeight = patch->weight.refValue();


		switch(k){
			case 0:{//front
				commonBoxBuild(patch, dimU, dimV);
				for(int v = 0; v < dimV; v++){
					for(int u = 0; u < dimU; u++){
						patchPoint[v*dimU+u] = matrix * Vec3f(2.0 / (dimU-1) * u + xPos, 2.0 / (dimV - 1) * v + yPos, 2.0 + zPos);
						patchWeight[v*dimU+u] = 1;
					}
				}
				break;
			}
			case 1:{//west
				commonBoxBuild(patch, dimZ, dimV);
				//patch->name = "r_side";
				if(insideOut){
					for(int v = 0; v < dimV; v++){
						for(int u = 0; u < dimZ; u++){
							patchPoint[v*dimZ + u] = matrix * Vec3f(1, 2.0 / (dimV-1) * v + yPos, -2.0 / (dimZ - 1) * u + 2.0 + zPos);
							patchWeight[v*dimZ + u] = 1;
						}
					}
				}
				else{
					for(int v = 0; v < dimV; v++){
						for(int u = 0; u < dimZ; u++){
							patchPoint[v*dimZ + u] = matrix * Vec3f(-1 + xPos + 1, 2.0 / (dimV-1) * v + yPos, 2.0 / (dimZ-1) * u + zPos);
							patchWeight[v*dimZ + u] = 1;
						}
					}
				}
				break;
			}
			case 2:{//east
				commonBoxBuild(patch, dimZ, dimV);
				if(mirror && mirrorContainer == 0){

					if(patch->stem.length){
						if(Stem* stem = dynamic_cast<Stem*>(patch->stem[0].p)){
							stem->insideOut.refValue() = true;
						}
					}
					//patch->name = "l_side";
					patch->insideOut.refValue() = true;
					patch->mirrorPatch = m_patch[k - 1];
					m_patch[k - 1]->mirrorPatch = patch;

					for(int j = patchWeight.length; --j >=0;){
						patchWeight[j] = 1;
					}

					m_patch[k - 1]->mirrorTwinPatch2(patch);

				}
				else{
					if(patch->stem.length){
						if(Stem* stem = dynamic_cast<Stem*>(patch->stem[0].p)){
							stem->insideOut.refValue() = false;
						}
					}
					patch->insideOut.refValue() = insideOut;
					patch->mirrorPatch = 0;
					m_patch[k - 1]->mirrorPatch = 0;

					if(insideOut){
						for(int v = 0; v < dimV; v++){
							for(int u = 0; u < dimZ; u++){
								patchPoint[v*dimZ+u] = matrix * Vec3f(-1, 2.0 / (dimV-1) * v + yPos, 2.0 / (dimZ-1) * u + zPos);
								patchWeight[v*dimZ+u] = 1;
							}
						}
					}
					else{
						for(int v = 0; v < dimV; v++){
							for(int u = 0; u < dimZ; u++){
								patchPoint[v*dimZ+u] = matrix * Vec3f(1 + xPos + 1, 2.0 / (dimV-1) * v + yPos, 1.0 - 2.0 / (dimZ - 1) * u + zPos + 1);
								patchWeight[v*dimZ+u] = 1;
							}
						}
					}
				}
				break;
			}

			case 3:{//back
				commonBoxBuild(patch, dimU, dimV);
				for(int v = 0; v < dimV; v++){
					for(int u = 0; u < dimU; u++){
						patchPoint[v*dimU+u] = matrix * Vec3f(1.0 - 2.0 / (dimU-1) * u + xPos + 1, 2.0 / (dimV - 1) * v + yPos,  zPos);
						patchWeight[v*dimU+u] = 1;
					}
				}
				break;
			}
			case 4:{//bottom
				commonBoxBuild(patch, dimU, dimZ);
				for(int v = 0; v < dimZ; v++){
					for(int u = 0; u < dimU; u++){
						patchPoint[v*dimU+u] = matrix * Vec3f(2.0 / (dimU-1) * u + xPos, 0 + yPos, 2.0 / (dimZ-1) * v + zPos);
						patchWeight[v*dimU+u] = 1;
					}
				}
				break;
			}
			case 5:{//top
				commonBoxBuild(patch, dimU, dimZ);
				for(int v = 0; v < dimZ; v++){
					for(int u = 0; u < dimU; u++){
						patchPoint[v*dimU+u] = matrix * Vec3f(1.0 - 2.0 / (dimU-1) * u + xPos + 1, 2.0 + yPos, 2.0 / (dimZ - 1) * v + zPos);
						patchWeight[v*dimU+u] = 1;
					}
				}
				break;
			}
		}
		patch->mirror.refValue() = mirror;

		if(mirror && mirrorContainer){
			patch->insideOut.refValue() = insideOut;
			if(insideOut){
				patch->setInsideOut();
				patch->setInsideOut3();
			}
		}


		patch->processFields();
		//patch->patchToCentralCPs(); //why is this not needed if the box was created (as opposed to a box existing in an opened file) ?
	}

	colorSweep->processFields();
}

void RectangleBuild::processFields(){
	stem->textureMapper.refValue().length = 0;

	if(textureMapper){
		stem->textureMapper.refValue()+=textureMapper.refValue();
	}
	else{
		colorSweep->color.refValue().length=0;
		colorSweep->color.refValue()+=Color3f(color);
		colorSweep->color.refValue()+=Color3f(color);
		colorSweep->straight.refValue()=true;
		colorSweep->distance.refValue().length=0;
		colorSweep->distance.refValue()+=1;
		stem->textureMapper.refValue()+=colorSweep.p;
	}
////
	patch->beginSP.refValue() = beginSP;
	patch->endSP.refValue() = endSP;

	Matrix matrix = scaleMatrix(scale) * rotationMatrix(rotation) * translationMatrix(translation);

	int dimU=ctrlStaves * 2 + 1;
	int dimV=ctrlBands * 2 + 1;

	M_Vec3f& patchPoint = patch->point.refValue();
	M_Float& patchWeight = patch->weight.refValue();

	patchPoint.length = dimU * dimV;
	patchWeight.length = dimU * dimV;

	Vec2f slide;
	if(yCentred){
		slide.y=-yDistance*.5;
	}
	if(xCentred){
		slide.x=-xDistance*.5;
	}
	if(xNegSide){
		slide.x=-xDistance;
	}
	if(yNegSide){
		slide.y=-yDistance;
	}

	for(int v = 0; v < dimV; v++){
		for(int u = 0; u < dimU; u++){
			patchPoint[v*dimU+u] = matrix * Vec3f(xDistance/(dimU-1)*u+slide.x, yDistance/(dimV-1)*v+slide.y, 0);
			patchWeight[v*dimU+u] = 1;
		}
	}
	if(beginSP){
		for(int u = 0; u < dimU; u++){
			patchPoint[u] = matrix * Vec3f(0, slide.y, 0);
			patchWeight[u] = 1;
		}
	}
	if(endSP){
		for(int u = 0; u < dimU; u++){
			patchPoint[(dimV-1)*dimU+u] = matrix * Vec3f(0, yDistance+slide.y, 0);
			patchWeight[(dimV-1)*dimU+u] = 1;
		}
	}
	updateNurbsCurveKnotMid1(dimU, patch->uKnot.refValue());
	updateNurbsCurveKnotMid1(dimV, patch->vKnot.refValue());
	patch->dim.refValue()=Vec2f(dimU, dimV);

	stem->staves.refValue()[0] = staves;
	stem->bands.refValue()[0] = bands;
	stem->part.refValue().length = 0;
	if(part){
		stem->part+=part.refValue();
	}
//	if(part.refValue()->isRightMirror()){
//		stem->insideOut.refValue() = true;
//	}


	colorSweep->processFields();
	stem->processFields();
	patch->processFields();
	//patch->patchToCentralCPs(); //why is this not needed if the rectangle was created (as opposed to a rectangle existing in an opened file) ?
}
float getHighestPoint(M_Vec2f& point){
	float highest=-1000000000;
	for(int i=point.length; --i>=0;){
		if(point[i].y>highest){
			highest=point[i].y;
		}
	}
	return highest;
}
float getHighestPoint(M_Vec3f& point){
	float highest=-1000000000;
	for(int i=point.length; --i>=0;){
		if(point[i].y>highest){
			highest=point[i].y;
		}
	}
	return highest;
}
float getLowestPoint(M_Vec2f& point){
	float lowest=1000000000;
	for(int i=point.length; --i>=0;){
		if(point[i].y<lowest){
			lowest=point[i].y;
		}
	}
	return lowest;
}
float getLowestPoint(M_Vec3f& point){
	float lowest=1000000000;
	for(int i=point.length; --i>=0;){
		if(point[i].y<lowest){
			lowest=point[i].y;
		}
	}
	return lowest;
}
void BezierLatheBuild::processFields(){
	if(part==0 || point.length==0){
		return;
	}

	if(m_patch.length != bands.length){
		if(m_patch.length){
			convertBorderToGroup();
		}
		m_patch.length=0;
		for(int i = 0;  i < bands.length; i++){
			Stem* stem = new Stem;
			stem->xDistance.refValue() = 1;
			stem->yDistance.refValue() += 1;
			stem->xCentred.refValue() = false;
			stem->bands.refValue().length = 1;
			NurbsPatch* patch = new NurbsPatch;
			patch->showAll.refValue()=false;
			patch->uJoinEnds.refValue()=true;
			patch->stem.refValue()+=stem;
			m_patch += patch;
		}
		convertGroupToFix();
	}

	convertBorderToGroup();


	dbands.refValue().length=point.length;
	for(int i=0, j=-1; i<fix.length; i++){
		if(fix[i] && i<fix.length-1){
			j++;
		}
		dbands.refValue()[i]=bands[j];
	}

	matrix=scaleMatrix(scale)*rotationMatrix(rotation)*translationMatrix(translation);
	imatrix=translationMatrix(-translation)*rotationMatrix(-Rotation(rotation))*scaleMatrix(Vec3f(1,1,1)/scale);

	Node* latstTM = 0;

	for(int s = 0, rg=0; s < m_patch.length; s++, --rg){
		NurbsPatch* patch = m_patch[s];
		Stem* stem = dynamic_cast<Stem*>(patch->stem[0].p);
		if(s < textureMapper.refValue().length){
			latstTM=textureMapper[s];
		}
		stem->textureMapper.refValue().length=0;
		if(latstTM){
			stem->textureMapper += latstTM;
		}
		if(latstTM==0){
			colorSweep->color.refValue().length=0;
			colorSweep->color.refValue()+=Color3f(color);
			colorSweep->color.refValue()+=Color3f(color);
			colorSweep->straight.refValue()=true;
			colorSweep->distance.refValue().length=0;
			colorSweep->distance.refValue()+=1;
			stem->textureMapper.refValue()+=colorSweep.p;
		}

		stem->bands.refValue()[0] = bands[s];
		stem->part.refValue().length=0;
		stem->part+=part.refValue();

		stem->staves.refValue()[0] = staves;


		M_Vec2f& point = this->point.refValue();

		double sinWeight=cos(PI/ctrlStaves);
		int dimU=ctrlStaves*2+1;

		M_Vec3f& patchPoint = patch->point.refValue();
		M_Float& patchWeight = patch->weight.refValue();

		patchPoint.length = nopfs[s]*dimU;
		patchWeight.length = nopfs[s]*dimU;
		for(int r = 0; r < nopfs[s]; r++, rg++){
			double a = PI*2/(ctrlStaves*2);
			for(int c = 0; c < dimU; c+=2){
				double b = a*c;
				patchPoint[r*dimU+c] = Vec3f(-sin(b)*point[rg].x, point[rg].y, -cos(b)*point[rg].x);
				patchWeight[r*dimU+c] = 1;
			}
			double div = cos(a)*point[rg].x;
			double ratio = 0;
			if(div!=0){
				ratio = -patchPoint[r*dimU].z / div;
			}

			for(int c = 1; c < dimU; c+=2){
				double b = a*c;
				patchPoint[r*dimU+c] = Vec3f(-sin(b)*point[rg].x*ratio, point[rg].y, -cos(b)*point[rg].x*ratio);
				patchWeight[r*dimU+c] = sinWeight;
				patchPoint[r*dimU+c] *= sinWeight;
			}
		}
		for(int i = patchPoint.length; --i>=0;){
			patchPoint[i]=matrix * (patchPoint[i] / patchWeight[i]) * patchWeight[i];
		}
		M_Float& uKnot = patch->uKnot.refValue();
		M_Float& vKnot = patch->vKnot.refValue();
		int dimV = nopfs[s];
		patch->order.refValue().y = dimV;

		updateNurbsCurveKnotBezier(dimV, vKnot);
		updateNurbsCurveKnotMid1(dimU, uKnot);
		patch->dim.refValue()=Vec2f(dimU, dimV);

		patch->processFields();

		if(s == 0){
			patch->beginSP.refValue() = stem->beginSP.refValue() = point[0].x==0;
		}
		if(s == bands.length-1){
			patch->endSP.refValue() = stem->endSP.refValue() = point[dimV-1].x==0;
		}
		stem->processFields();
	}

	if(bands.length > 1){
		if(joinVertex == 0){
			joinVertex = new JoinVertex;
			joinVertex->part.refValue()=part.refValue();
		}
		M_Int& srcCoord = joinVertex->srcCoord.refValue();
		M_Int& dstCoord = joinVertex->dstCoord.refValue();

		srcCoord.length = 0;
		dstCoord.length = 0;

		int accumBands = bands[0];
		int beginSPOffset = 0;
		NurbsPatch* patch = m_patch[0];
		Stem* stem = dynamic_cast<Stem*>(patch->stem[0].p);
		if(stem->beginSP){
			beginSPOffset = staves * 3;
		}
		int soFarDst = staves * accumBands * 6 - 2 - beginSPOffset;
		int soFarSrc = soFarDst + staves * 6 - 1;
		//src 87 68
		//dst 58 41
		for(int k = 1; k < bands.length; k++){

			srcCoord += soFarSrc;
			dstCoord += soFarDst;
			soFarSrc-=(staves-2)*6+1;
			soFarDst-=(staves-2)*6-1;

			for(int i = 1; i < staves; i++){
				srcCoord += soFarSrc;
				dstCoord += soFarDst;
				soFarDst += 6;
				soFarSrc += 6;
			}
			accumBands += bands[k];
			soFarDst = staves * accumBands * 6 - 2 - beginSPOffset;
			soFarSrc = soFarDst + staves * 6 - 1;
		}
	}
}
/*
inline void workOutFromToCoord(M_Art& partList){
	int coordAccumLen=0;
	for(int k=0; k<partList.length; k++){
		Art* part=partList[k];
		part->fromCoord=coordAccumLen;
		coordAccumLen+=part->coord.length;
		part->toCoord=coordAccumLen;
	}
}
*/
bool allRelevantsAreReferenced(Art* target){
	M_Art& relevantPart=target->relevantPart;
	for(int k=relevantPart.length; --k>=0; ){
		if(verifyPartRelivanceToTriangles( target, relevantPart[k]->triangle)==false){
			dLog("****");
			dLog(relevantPart[k]->name);
			return false;
		}
	}
	return true;
}
void getIndex(const Art*& part, int& index){
	if(index>=part->coord.length){
		int ei=index-part->coord.length;
		index=part->rc[ei];
	}
}
void verifyAllVerticesAreReferenced2(M_Art& partList){
	int k;
	for(k=partList.length; --k>=0; ){
		Art* part=partList[k];
		part->partTree.length=0;	//set all elements to 0;
		//part->partTree.resize(part->coord.length+part->rc.length);
		part->partTree.length=part->coord.length;
	}
	for(k=partList.length; --k>=0; ){
		Art* part=partList[k];
		M_Triangle& triangle=part->triangle;
		for(int i=triangle.length*3; --i>=0;){
			const Art* part=triangle.refPart(i);
			int index=triangle.refIndex(i);
			getIndex(part, index);
			const_cast<Art*>(part)->partTree[index]=1;
		}
	}
	for(k=partList.length; --k>=0; ){
		Art* part=partList[k];
//		for(int i=part->coord.length+part->rc.length; --i>=0;){
		for(int i=part->coord.length; --i>=0;){
			if(part->partTree[i]==0){
				throw;
				part->deleteVertex(i);
			}
		}
	}
}

void creatPartLookUp(M_Art& partLookUp, M_Art& partList){

	{// work out length
		int coordTotalLen=0;
		for(int k=0; k<partList.length; k++){
			Art* part=partList[k];
			part->fromCoord=coordTotalLen;
			coordTotalLen+=part->coord.length;
		}
		partLookUp.length=coordTotalLen;
	}
	{// make look up table

		int j=0;
		for(int k=0; k<partList.length; k++){
			Art* part=partList[k];
			for(int i=part->coord.length; --i>=0; ){
				partLookUp[j++]=part;
			}
		}
	}
}

//void SFBool::connect(Node* node, Node* protoInterface, Field *protoField){
//    S_Route( protoInterface, protoInterface->getFieldIndex(protoField), node, node->getFieldIndex(this));
//    is=true;
//}
void SFColor::connect(Node* node, Node* protoInterface, Field *protoField){
	S_Route( protoInterface, protoInterface->getFieldIndex(protoField), node, node->getFieldIndex(this));
	is=true;
}
//void SFFloat::connect(Node* node, Node* protoInterface, Field *protoField){
//    S_Route( protoInterface, protoInterface->getFieldIndex(protoField), node, node->getFieldIndex(this));
//    is=true;
//}
//void SFInt32::connect(Node* node, Node* protoInterface, Field *protoField){
//    S_Route( protoInterface, protoInterface->getFieldIndex(protoField), node, node->getFieldIndex(this));
//    is=true;
//}
//void SFImage::connect(Node* node, Node* protoInterface, Field *protoField){
//    S_Route( protoInterface, protoInterface->getFieldIndex(protoField), node, node->getFieldIndex(this));
//    is=true;
//}
void SFString::connect(Node* node, Node* protoInterface, Field *protoField){
	S_Route( protoInterface, protoInterface->getFieldIndex(protoField), node, node->getFieldIndex(this));
	is=true;
}
//void SFTime::connect(Node* node, Node* protoInterface, Field *protoField){
//    S_Route( protoInterface, protoInterface->getFieldIndex(protoField), node, node->getFieldIndex(this));
//    is=true;
//}
void SFVec2f::connect(Node* node, Node* protoInterface, Field *protoField){
	S_Route( protoInterface, protoInterface->getFieldIndex(protoField), node, node->getFieldIndex(this));
	is=true;
}
void SFVec3f::connect(Node* node, Node* protoInterface, Field *protoField){
	S_Route( protoInterface, protoInterface->getFieldIndex(protoField), node, node->getFieldIndex(this));
	is=true;
}
void SFRotation::connect(Node* node, Node* protoInterface, Field *protoField){
	S_Route( protoInterface, protoInterface->getFieldIndex(protoField), node, node->getFieldIndex(this));
	is=true;
}
void SFNode::connect(Node* node, Node* protoInterface, Field *protoField){
	S_Route( protoInterface, protoInterface->getFieldIndex(protoField), node, node->getFieldIndex(this));
	is=true;
}

template<> void MFVec3f::connect(Node* node, Node* protoInterface, Field *protoField){
	S_Route( protoInterface, protoInterface->getFieldIndex(protoField), node, node->getFieldIndex(this));
}


/*
void Seamless::addSeamlessRefToSubdiv(M_Build& build){
	for(int k=0; k<build.length; k++){
		if(GroupBuild* group=dynamic_cast<GroupBuild*>(build[k].p)){
			addSeamlessRefToSubdiv(group->build.refValue());
		}
		else if(Subdivision* subDiv=dynamic_cast<Subdivision*>(build[k].p)){
			subDiv->smls=this;
		}
	}
}
*/
void ColorEffect::smlsProcEffects(M_Art& partList){
// create extra vertices
	M_Color3f& color=this->color.refValue();
	color.length=0;

	int linEffectCoordLen=0;
	for(int k=0; k<partList.length; k++){
		Art* part=partList[k];
		M_Vec3f& coord=part->coord.refValue();
		M_Int& rc=part->rc;
		coord.length+=rc.length;
		part->fromEffectCoord=linEffectCoordLen;
		linEffectCoordLen+=part->color.length;
		color+=part->color;
	}
}
void MonoColorEffect::smlsProcEffects(M_Art& partList){
// create extra vertices
	M_Color3f& color=this->color.refValue();
	color.length=0;

	int linEffectCoordLen=0;
	for(int k=0; k<partList.length; k++){
		Art* part=partList[k];
		M_Vec3f& coord=part->coord.refValue();
		M_Int& rc=part->rc;
		coord.length+=rc.length;
		part->fromEffectCoord=linEffectCoordLen;
		linEffectCoordLen+=part->color.length;
		color+=part->color;
	}
}
void TextureEffect::smlsProcEffects(M_Art& partList){
	M_Vec2f& texCoord=this->texCoord.refValue();
	texCoord.length=0;

	int linEffectCoordLen=0;
	for(int k=0; k<partList.length; k++){
		Art* part=partList[k];
		M_Vec3f& coord=part->coord.refValue();
		M_Int& rc=part->rc;
		coord.length+=rc.length;
		part->fromEffectCoord=linEffectCoordLen;
		linEffectCoordLen+=part->texCoord.length;
		texCoord+=part->texCoord;
	}
}

void Seamless::decodePatchOwner(){
	int po = 0;
	for(int k = 0; k < partList.length; k++){
		Art* part = partList[k];
		const int coordLen = part->uvCoord.length;
		if(coordLen){
			M_NurbsPatch& partPatchOwner = part->patchOwner;
			do{
				int len = patchOwner[po++];
				NurbsPatch* patch = cpNodeList[patchOwner[po++]];

				int offset = partPatchOwner.length;
				partPatchOwner.length += len;
				while(--len >= 0){
					partPatchOwner[offset + len] = patch;
				}
			}while(partPatchOwner.length < coordLen);
		}
	}
}

void Seamless::processFields(){
	g_remVersionLoaded = version;
	process();

	setUpInternalEffectPtrs();
	if(version >= 2.027999){
		if(Effect* effect=dynamic_cast<Effect*>(this->effect.p)){
			effect->smlsProcEffects(partList);
		}

		static M_Art partLookUp;

		creatPartLookUp(partLookUp, partList);

		int ia = 0;
		int ib = 0;
		int ic = 0;

		int ta = 0;
		int tb = 0;
		int tc = 0;
		int i = 0;
		Art* pa = 0;
		Art* pb = 0;
		Art* pc = 0;
		int g = 0;
		int rci = 0;

		for(int ci = 0; ci<coordIndex.length; ){
			{

				i=coordIndex[ci++];
				pa=partLookUp[i];
				ta=i;
				ia=i-pa->fromCoord;
				M_Int& rc=pa->rc;
				g=ia-(pa->coord.length-rc.length);
				if(g>=0){
					rci=rc[g];
					pa=partLookUp[rci];
					ia=rci-pa->fromCoord;
				}
			}
			{
				i=coordIndex[ci++];
				pb=partLookUp[i];
				tb=i;
				ib=i-pb->fromCoord;
				M_Int& rc=pb->rc;
				g=ib-(pb->coord.length-rc.length);
				if(g>=0){
					rci=rc[g];
					pb=partLookUp[rci];
					ib=rci-pb->fromCoord;
				}
			}
			{
				i=coordIndex[ci++];
				pc=partLookUp[i];
				tc=i;
				ic=i-pc->fromCoord;
				M_Int& rc=pc->rc;
				g=ic-(pc->coord.length-rc.length);
				if(g>=0){
					rci=rc[g];
					pc=partLookUp[rci];
					ic=rci-pc->fromCoord;
				}
			}
			S_Triangle tri=S_Triangle(pa,ia,ta, pb,ib,tb, pc,ic,tc);

			Art* owner=tri.getOwner();
			owner->triangle+=tri;
			addRelevantPart(pa->relevantPart, owner);
			addRelevantPart(pb->relevantPart, owner);
			addRelevantPart(pc->relevantPart, owner);
		}
		/*
		{// convert linear rc to part and local index
			for(int k=0; k<partList.length; k++){
				M_RedirectCoord& rc=partList[k]->rc;
				for(int i=rc.length; --i>=0; ){
					Art* part=partLookUp[rc[i].i];
					rc[i].p=part;
					rc[i].i-=part->fromCoord;
				}
			}
		}
		*/
		{// delete extra vertices
			for(int k=0; k<partList.length; k++){
				Art* part=partList[k];
				part->coord.refValue().length-=(part->rc.length);
			}
		}
		populateMirrorNodes();
		//centralCPsToPatches();
		if(bnspe){
			decodePatchOwner();
		}

		//if(animCP){
		//	workAnimCP = true;
		//	storePoints(cpNodeList);
		//}
		decodeAllot();
		flagCtrlPolyPointRefs();
#ifdef CPINDEX
//		if(cp.length == 0){
			createCpFromPatchPoints();
			if(stitchPolysAreCorrupt()){
				ctrlPolyPoint.length = 0;
				generateRawStitchPolys();
			}
			else{
				createSharedPointsFromStitchPolys();
				cleanCps();
			}
//		}
#endif
	}
	else{
		orderPolygonsFromVertices();
		convertTo028();
		//convert color to 29
		if(effect==0){
			effect=new ColorEffect;
			setUpInternalEffectPtrs();
		}
		ColorEffect* effect=dynamic_cast<ColorEffect*>(this->effect.refValue().getRawPtr());
		M_Color3f& color=effect->color.refValue();
		color.length=0;
		for(int k=0; k<partList.length; k++){
			Art* part=partList[k];
			color+=part->color;
			M_Triangle& triangle=part->triangle;
			for(int i=0; i<triangle.length*3; i++){
				triangle.refTexIndex(i)=triangle.refIndex(i)+triangle.refPart(i)->fromCoord;
			}
		}
	}
//	addSeamlessRefToSubdiv(build.refValue());
}

inline Art* getPartFromRelIndex(Art* part, int& index, M_Art& partList){
	if(index>=part->triangle.length*3){
		do{
			index-=part->triangle.length*3;
			part=partList[part->plidx+1];
		}while(index>=part->triangle.length*3);
	}
	else if(index<0){
		do{
			part=partList[part->plidx-1];
			index+=part->triangle.length*3;
		}while(index<0);
	}
	return part;
}
void GenPoly::doBuild(){
	Art* part=this->part.refValue();
	if(part==NULL || !vertex.length)return;

	M_Art& partList=part->smls->partList;

	M_Int& vertex=this->vertex.refValue();


	if(independentColor){
		for(int i=0 ; i<vertex.length;){
			int ia=vertex[i++];
			int ib=vertex[i++];
			int ic=vertex[i++];

			Art* pa=getPartFromRelIndex(part, ia, partList);
			Art* pb=getPartFromRelIndex(part, ib, partList);
			Art* pc=getPartFromRelIndex(part, ic, partList);

			int nia=pa->triangle.refIndex(ia);
			int nib=pb->triangle.refIndex(ib);
			int nic=pc->triangle.refIndex(ic);

			int ta=0;
			int tb=0;
			int tc=0;
			if(ColorEffect* colorEffectA = dynamic_cast<ColorEffect*>(pa->workEffect)){
				M_Color3f& color = colorEffectA->color.refValue();
				ta = color.length;
				color+=color[pa->triangle.refTexIndex(ia)];
			}
			if(ColorEffect* colorEffectB = dynamic_cast<ColorEffect*>(pb->workEffect)){
				M_Color3f& color = colorEffectB->color.refValue();
				tb = color.length;
				color+=color[pb->triangle.refTexIndex(ib)];
			}
			if(ColorEffect* colorEffectC = dynamic_cast<ColorEffect*>(pc->workEffect)){
				M_Color3f& color = colorEffectC->color.refValue();
				tc = color.length;
				color+=color[pc->triangle.refTexIndex(ic)];
			}

			pa=pa->triangle.refPart(ia);
			pb=pb->triangle.refPart(ib);
			pc=pc->triangle.refPart(ic);

			Art* owner=pa;
			if(pb->plidx<owner->plidx){
				owner=pb;
			}
			if(pc->plidx<owner->plidx){
				owner=pc;
			}
			owner->triangle+=S_Triangle(pa,nia,ta,pb,nib,tb,pc,nic,tc);

			addRelevantPart(pa->relevantPart, owner);
			addRelevantPart(pb->relevantPart, owner);
			addRelevantPart(pc->relevantPart, owner);
		}
	}
	else{
		for(int i=0 ; i<vertex.length;){
			int ia=vertex[i++];
			int ib=vertex[i++];
			int ic=vertex[i++];

			Art* pa=getPartFromRelIndex(part, ia, partList);
			Art* pb=getPartFromRelIndex(part, ib, partList);
			Art* pc=getPartFromRelIndex(part, ic, partList);

			int nia=pa->triangle.refIndex(ia);
			int nib=pb->triangle.refIndex(ib);
			int nic=pc->triangle.refIndex(ic);

			int ta=pa->triangle.refTexIndex(ia);
			int tb=pb->triangle.refTexIndex(ib);
			int tc=pc->triangle.refTexIndex(ic);

			pa=pa->triangle.refPart(ia);
			pb=pb->triangle.refPart(ib);
			pc=pc->triangle.refPart(ic);

			Art* owner=pa;
			if(pb->plidx<owner->plidx){
				owner=pb;
			}
			if(pc->plidx<owner->plidx){
				owner=pc;
			}
			owner->triangle+=S_Triangle(pa,nia,ta,pb,nib,tb,pc,nic,tc);

			addRelevantPart(pa->relevantPart, owner);
			addRelevantPart(pb->relevantPart, owner);
			addRelevantPart(pc->relevantPart, owner);
		}
	}
}
void SineTug::generate(M_Vec3f& coord, const int){
	static M_Vec3f accum;

	const M_Vec3f& location=this->location.refValue();
	const M_Rotation& orientation=this->orientation.refValue();

	if(location.length==1){
		for(int j=0; j<location.length; j++){
			Matrix matrix=rotationMatrix(orientation[j])*translationMatrix(location[j]);
			Matrix imatrix=translationMatrix(-location[j])*rotationMatrix(-orientation[j]);
			float sdecay=this->decay.refValue()[j];

			float xsc=1.0/xScale.refValue()[j];
			float rad=radius.refValue()[j];
			float height=distance.refValue()[j];

			for(int i=0;i<coord.length;i++){
				Vec3f ivec=imatrix*coord[i];
				if(ivec.y<-sdecay)continue;
				float ADJ=(float)(fabs(ivec.x)*xsc);
				float OPP=(float)(fabs(ivec.z));
				float HYP=(float)(sqrt(ADJ*ADJ + OPP*OPP));

				//if(rad>HYP && abs(ivec.y)<sdecay){//forward decay
				//	float sineCurve=cos((fabs(ivec.y)/sdecay)*PI)*.5+.5;
				//	ivec.y+=((1+cos(HYP/rad*PI))/2.*height)*sineCurve;
				//	coord[i]=matrix*ivec;
				//}
				if(rad>HYP){
					if(ivec.y>0){
						ivec.y+=((1+cos(HYP/rad*PI))/2.*height);
					}
					else{
						float sineCurve=cos((fabs(ivec.y)/sdecay)*PI)*.5+.5;
						ivec.y+=((1+cos(HYP/rad*PI))/2.*height)*sineCurve;
					}
					coord[i]=matrix*ivec;
				}
			}
		}
	}
	else{
		accum.length=0; //reset all values to 0 0 0
		accum.length=coord.length;
		float f=0;
		for(int j=0; j<location.length; j++){
			Matrix matrix=rotationMatrix(orientation[j]);
			Matrix imatrix=translationMatrix(-location[j])*rotationMatrix(-orientation[j]);
			float sdecay=this->decay.refValue()[j];

			float xsc=1.0/xScale.refValue()[j];
			float rad=radius.refValue()[j];
			float height=distance.refValue()[j];

			for (int i=0;i<coord.length;i++){
				Vec3f ivec=imatrix*coord[i];
				if(ivec.y<-sdecay)continue;
				float ADJ=(float)(fabs(ivec.x)*xsc);
				float OPP=(float)(fabs(ivec.z));
				float HYP=(float)(sqrt(ADJ*ADJ + OPP*OPP));

				if(rad>HYP){
					if(ivec.y>0){
						f=((1+cos(HYP/rad*PI))/2.*height);
					}
					else{
						float sineCurve=cos((fabs(ivec.y)/sdecay)*PI)*.5+.5;
						f=((1+cos(HYP/rad*PI))/2.*height)*sineCurve;
					}
					accum[i]+=matrix*Vec3f(0,f,0);
				}
			}
		}
		for(int i=coord.length; --i>=0;){
			coord[i]+=accum[i];
		}
	}
}
void SimulSineTug::generate(M_Vec3f& coord, const int){
	static M_Vec3f accum;

	accum.length=0; //reset all values to 0 0 0
	accum.length=coord.length;

	const Vec3f& location=this->location.refValue();
	const M_Rotation& orientation=this->orientation.refValue();
	float rad=radius.refValue();


	for(int j=orientation.length; --j>=0;){
		Matrix matrix=rotationMatrix(orientation[j]);
		Matrix imatrix=translationMatrix(-location)*rotationMatrix(-orientation[j]);

		float height=distance.refValue()[j];

		for(int i=coord.length; --i>=0;){
			Vec3f ivec=imatrix*coord[i];
			if(ivec.y<0)continue;
			float ADJ=abs(ivec.x);
			float OPP=abs(ivec.z);
			float HYP=sqrt(ADJ*ADJ + OPP*OPP);

			if(rad>HYP){
				Vec3f v = Vec3f(0, ((1.0+cos(HYP/rad*PI))*.5*height), 0);
				accum[i]+=matrix*v;
			}
		}
	}
	for(int i=coord.length; --i>=0;){
		coord[i]+=accum[i];
	}
}
void RadialTug::generate(M_Vec3f& coord, const int){
	const M_Vec3f& location=this->location;
	const M_Vec3f& displacement=this->displacement;
	const M_Float& radius=this->radius;
	for(int j=0; j<location.length; j++){
		float pp=slope[j]*PI;
		float ipp=PI-pp;
		float h=1/(cos(ipp)+1);
		Vec3f loc=location[j];
		Vec3f disp=displacement[j];
		float rad=radius[j];
		for (int i=0;i<coord.length;i++){
			Vec3f vec=coord[i];
			float vecX=vec.x-loc.x;
			float vecY=vec.y-loc.y;
			float vecZ=vec.z-loc.z;
			float HYP=(float)(sqrt(vecX*vecX + vecY*vecY + vecZ*vecZ));

			if(rad>HYP){
				float sineCurve=(cos((fabs(HYP)/rad)*pp+ipp)+1)*h;
				vec.x+=disp.x*sineCurve;
				vec.y+=disp.y*sineCurve;
				vec.z+=disp.z*sineCurve;
				coord[i]=vec;
			}
		}
	}
}

void expandColorIndexIFS(const M_Int& coordIndex, M_Int& colorIndex, ColorPtr colorPtr){
	M_Color3f& color=colorPtr->color.refValue();
	M_Int& colorLookUp=g_tmpIndex;
	colorLookUp.length=color.length;
	{
		for(int i=colorLookUp.length; --i>=0;){
			colorLookUp[i]=-1;
		}
	}

	for(int i=coordIndex.length; --i>=0;){
		int coordIdx=coordIndex[i];
		if(coordIdx==-1)continue;
		int colorIdx=colorIndex[i];
		int colorLU=colorLookUp[colorIdx];
		if(colorLU==-1){
			colorLookUp[colorIdx]=coordIdx;
		}
		else if(colorLU!=coordIdx){
			int newIdx=color.length;
			color+=color[colorIdx];
			colorLookUp+=coordIdx;
			{
				for(int i=colorIndex.length; --i>=0;){ //-1 wont cause a problem
					if(colorIndex[i]==colorIdx&&coordIndex[i]==coordIdx){
						colorIndex[i]=newIdx;
					}
				}
			}
		}
	}
}
void expandTexCoordIndexIFS(const M_Int& coordIndex, M_Int& texCoordIndex, TextureCoordinatePtr texCoordPtr){
	M_Vec2f& texCoord=texCoordPtr->point.refValue();
	M_Int& texCoordLookUp=g_tmpIndex;
	texCoordLookUp.length=texCoord.length;
	{
		for(int i=texCoordLookUp.length; --i>=0;){
			texCoordLookUp[i]=-1;
		}
	}

	for(int i=coordIndex.length; --i>=0;){
		int coordIdx=coordIndex[i];
		if(coordIdx==-1)continue;
		int texCoordIdx=texCoordIndex[i];
		int texCoordLU=texCoordLookUp[texCoordIdx];
		if(texCoordLU==-1){
			texCoordLookUp[texCoordIdx]=coordIdx;
		}
		else if(texCoordLU!=coordIdx){
			int newIdx=texCoord.length;
			texCoord+=texCoord[texCoordIdx];
			texCoordLookUp+=coordIdx;
			{
				for(int i=texCoordIndex.length; --i>=0;){ //-1 wont cause a problem
					if(texCoordIndex[i]==texCoordIdx&&coordIndex[i]==coordIdx){
						texCoordIndex[i]=newIdx;
					}
				}
			}
		}
	}
}
/// polygons must be triangles. texCoordIndex get messed up
inline void Mesh::convertFromTexCoordIndex( M_Int& coordIndex, M_Int& texCoordIndex, const M_Vec2f& ifsTexCoord){
	M_Int& rc=this->rc.refValue();
	M_Vec2f& texCoord=this->texCoord.refValue();
	M_Int& coordLookUp=g_tmpIndex;
	const int coordLen=coord.getValue().length;
	const int coordIdxLen=coordIndex.length;
	coordLookUp.length=coordLen;
	texCoord.length=coordLen;
	{
		for(int i=0, j=0; j<coordIdxLen; i++){
			if(texCoordIndex[i]!=-1){
				texCoordIndex[j++]=texCoordIndex[i];
			}
		}
		texCoordIndex.length=coordIdxLen;
	}
	{
		;
		for(int i=coordLookUp.length; --i>=0;){
			coordLookUp[i]=-1;
		}
	}
	for(int i=0; i<coordIdxLen; i++){
		int coordIdx=coordIndex[i];
		int texCoordIdx=texCoordIndex[i];
		int coordLU=coordLookUp[coordIdx];
		if(coordLU==-1){
			texCoord[coordIdx]=ifsTexCoord[texCoordIdx];
			coordLookUp[coordIdx]=texCoordIdx;
		}
		else if(coordLU!=texCoordIdx){
			int newIdx=coordLen+rc.length;
			coordIndex[i]=newIdx;
			texCoord+=ifsTexCoord[texCoordIdx];
			coordLookUp+=texCoordIdx;
			rc+=coordIdx;
			{
				for(int i=coordIdxLen; --i>=0;){
					if(texCoordIndex[i]==texCoordIdx&&coordIndex[i]==coordIdx){
						coordIndex[i]=newIdx;
					}
				}
			}
		}
	}
}
// polygons must be triangles. ifsColorIndex gets messed up
inline void Mesh::convertFromColorIndex( M_Int& coordIndex, M_Int& ifsColorIndex, const M_Color3f& ifsColor){
	M_Int& rc=this->rc.refValue();
	M_Color3f& color=this->color.refValue();
	M_Int& coordLookUp=g_tmpIndex;
	const int coordLen=coord.getValue().length;
	const int coordIdxLen=coordIndex.length;
	coordLookUp.length=coordLen;
	color.length=coordLen;

	{
		for(int i=0, j=0; j<coordIdxLen; i++){
			if(ifsColorIndex[i]!=-1){
				ifsColorIndex[j++]=ifsColorIndex[i];
			}
		}
		ifsColorIndex.length=coordIdxLen;
	}
	{
		for(int i=coordLookUp.length; --i>=0;){
			coordLookUp[i]=-1;
		}
	}
	for(int i=0; i<coordIdxLen; i++){
		int coordIdx=coordIndex[i];
		int colorIdx=ifsColorIndex[i];
		int coordLU=coordLookUp[coordIdx];
		if(coordLU==-1){
			color[coordIdx]=ifsColor[colorIdx];
			coordLookUp[coordIdx]=colorIdx;
		}
		else if(coordLU!=colorIdx){
			int newIdx=coordLen+rc.length;
			coordIndex[i]=newIdx;
			color+=ifsColor[colorIdx];
			coordLookUp+=colorIdx;
			rc+=coordIdx;
			{
				for(int i=coordIdxLen; --i>=0;){
					if(ifsColorIndex[i]==colorIdx&&coordIndex[i]==coordIdx){
						coordIndex[i]=newIdx;
					}
				}
			}
		}
	}
}
/*
inline void convertConvexIFSIndexToTriIndex(M_Int& index){
	M_Int& tmpIndex=g_tmpIndex;
	tmpIndex.length=0;
	int s=0;
	int d=0;
	while(s<index.length){
		int firstIdx=index[s++];
		tmpIndex+=firstIdx;
		tmpIndex+=index[s++];
		tmpIndex+=index[s++];
		tmpIndex+=-1;
		while (s<index.length&&index[s]!=-1){
			tmpIndex+=firstIdx;
			tmpIndex+=index[s-1];
			tmpIndex+=index[s++];
			tmpIndex+=-1;
		}
		s++;
	}
	index=tmpIndex;
}
void convertIFSIndexToTriIndex(M_Int& index, bool convex){
	if(convex){
		convertConvexIFSIndexToTriIndex(index);
	}
	else{
		convertConcaveIFSIndexToTriIndex(index);
	}
}
*/
bool vertexInsideTri(M_Vec3f& point, int idx0, int idx1, int idx2){
	const Vec2f v[3] = {Vec2f(point[idx0].x, point[idx0].y), Vec2f(point[idx1].x, point[idx1].y), Vec2f(point[idx2].x, point[idx2].y)};

	for(int i=0; i<point.length; i++){
		Vec2f pt = Vec2f(point[i].x, point[i].y);
		if(	pt != v[0] &&
			pt != v[1] &&
			pt != v[2]){

			if(isTriangleHit(pt, v)){
				return true;
			}
		}
	}
	return false;
}
inline bool thinTriangle(const Vec3f& a, const Vec3f& b, const Vec3f& c){
	float triThinAmount=.0005;
	if(
		fabs(a.x-b.x)<triThinAmount&&
		fabs(a.y-b.y)<triThinAmount&&

		fabs(a.x-c.x)<triThinAmount&&
		fabs(a.y-c.y)<triThinAmount
	)
	{
		return true;
	}
	if(
		fabs(a.x-b.x)<triThinAmount&&
		fabs(a.z-b.z)<triThinAmount&&

		fabs(a.x-c.x)<triThinAmount&&
		fabs(a.z-c.z)<triThinAmount
	)
	{
		return true;
	}
	if(
		fabs(a.y-b.y)<triThinAmount&&
		fabs(a.z-b.z)<triThinAmount&&

		fabs(a.y-c.y)<triThinAmount&&
		fabs(a.z-c.z)<triThinAmount
	)
	{
		return true;
	}
	/////////////////////
	if(
		fabs(a.x-b.x)<triThinAmount&&
		fabs(a.y-b.y)<triThinAmount&&

		fabs(b.x-c.x)<triThinAmount&&
		fabs(b.y-c.y)<triThinAmount
	)
	{
		return true;
	}
	if(
		fabs(a.x-b.x)<triThinAmount&&
		fabs(a.z-b.z)<triThinAmount&&

		fabs(b.x-c.x)<triThinAmount&&
		fabs(b.z-c.z)<triThinAmount
	)
	{
		return true;
	}
	if(
		fabs(a.y-b.y)<triThinAmount&&
		fabs(a.z-b.z)<triThinAmount&&

		fabs(b.y-c.y)<triThinAmount&&
		fabs(b.z-c.z)<triThinAmount
	)
	{
		return true;
	}
///////////////////
		if(
		fabs(a.x-c.x)<triThinAmount&&
		fabs(a.y-c.y)<triThinAmount&&

		fabs(b.x-c.x)<triThinAmount&&
		fabs(b.y-c.y)<triThinAmount
	)
	{
		return true;
	}
	if(
		fabs(a.x-c.x)<triThinAmount&&
		fabs(a.z-c.z)<triThinAmount&&

		fabs(b.x-c.x)<triThinAmount&&
		fabs(b.z-c.z)<triThinAmount
	)
	{
		return true;
	}
	if(
		fabs(a.y-c.y)<triThinAmount&&
		fabs(a.z-c.z)<triThinAmount&&

		fabs(b.y-c.y)<triThinAmount&&
		fabs(b.z-c.z)<triThinAmount
	)
	{
		return true;
	}
///////////
	if(
		(fabs(a.x-b.x)<triThinAmount&&
		 fabs(a.y-b.y)<triThinAmount&&
		 fabs(a.z-b.z)<triThinAmount)||

		(fabs(a.x-c.x)<triThinAmount&&
		 fabs(a.y-c.y)<triThinAmount&&
		fabs(a.z-c.z)<triThinAmount)||

		(fabs(b.x-c.x)<triThinAmount&&
		 fabs(b.y-c.y)<triThinAmount&&
		fabs(b.z-c.z)<triThinAmount)
	)
	{
		return true;
	}
	return false;
}
inline void convertConcaveIFSIndexToTriIndex(const M_Vec3f& src_point, M_Int& index){
	M_Vec3f point;
	M_Int& tmpIndex=g_tmpIndex;
	tmpIndex.length=0;
	M_Int poly;
	for(int i = 0; i < index.length && index[i] >= 0; i++){
		poly+=index[i];
	}

	Rotation rot;

	for(int s = 0 ;;){
		int idx0 = poly[s++];
		if(s>=poly.length){
			s=0;
		}
		int idx1 = poly[s++];
		if(s>=poly.length){
			s=0;
		}
		int idx2 = poly[s++];
		if(s>=poly.length){
			s=0;
		}

		Vec3f v0 = src_point[idx0];
		Vec3f v1 = src_point[idx1];
		Vec3f v2 = src_point[idx2];

		if(thinTriangle(v0, v1, v2)){
			s -= 2;
			continue;
		}
		Vec3f normal = calcNormal(src_point[index[idx0]], src_point[index[idx1]], src_point[index[idx2]]);
		rot = Rotation(Vec3f(0, 0, 1), normal);
		break;
	}

	point.length = src_point.length;
	for(int i=0; i<point.length; i++){
		point[i] = rot * src_point[i];
	}

	int s=0;
	int c=0;
	while(poly.length>3){
		int idx0 = poly[s++];
		if(s>=poly.length){
			s=0;
			c++;
		}
		int idx1 = poly[s++];
		if(s>=poly.length){
			s=0;
			c++;
		}
		int idx2 = poly[s++];
		if(s>=poly.length){
			s=0;
			c++;
		}

		Vec3f v0 = point[idx0];
		Vec3f v1 = point[idx1];
		Vec3f v2 = point[idx2];

		if(!backFace(v0, v1, v2) || thinTriangle(v0, v1, v2) || vertexInsideTri(point, idx0, idx1, idx2)){
			if(c==3){
				s=0;
				c=0;
				rot = rot * Rotation(0,-1,0,PI);
				for(int i=0; i<point.length; i++){
					point[i] = rot * src_point[i];
				}
				poly.length=0;
				for(int i = 0; i < index.length && index[i] >= 0; i++){
					poly+=index[i];
				}
				tmpIndex.length=0;
				continue;
			}
			s -= 2;
			if(s < 0){
				s=poly.length + s;
			}
			continue;
		}
		tmpIndex+=idx0;
		tmpIndex+=idx1;
		tmpIndex+=idx2;
		tmpIndex+=-1;

		int g=s-2;
		if(g < 0){
			g=poly.length + g;
		}
		poly.deleteElement(g);
		s=0;
		c=0;
	}
	tmpIndex+=poly;
	index=tmpIndex;
}

void convertIFSIndexToTriIndex(M_Int& index){
	M_Int& tmpIndex=g_tmpIndex;
	tmpIndex.length=0;
	int s=0;
	while(s<index.length){
		int firstIdx=index[s++];
		tmpIndex+=firstIdx;
		tmpIndex+=index[s++];
		tmpIndex+=index[s++];
		tmpIndex+=-1;
		while (s<index.length&&index[s]!=-1){
			tmpIndex+=firstIdx;
			tmpIndex+=index[s-1];
			tmpIndex+=index[s++];
			tmpIndex+=-1;
		}
		s++;
	}
	index=tmpIndex;
}
void IndexedFaceSet::convertToTriangles(){
	convertIFSIndexToTriIndex(this->coordIndex.refValue());
	convertIFSIndexToTriIndex(this->colorIndex.refValue());
	convertIFSIndexToTriIndex(this->texCoordIndex.refValue());

}
void IndexedTriangleSet::convertToTriangles(const IndexedFaceSet* ifs){
	const M_Int& coordIndex=ifs->coordIndex;
	M_Int& index=this->index.refValue();

	M_Int& tmpIndex=g_tmpIndex;
	tmpIndex.length=0;
	int s=0;
	while(s<coordIndex.length){
		int firstIdx=coordIndex[s++];
		tmpIndex+=firstIdx;
		tmpIndex+=coordIndex[s++];
		tmpIndex+=coordIndex[s++];
		while (s<coordIndex.length&&coordIndex[s]!=-1){
			tmpIndex+=firstIdx;
			tmpIndex+=coordIndex[s-1];
			tmpIndex+=coordIndex[s++];
		}
		s++;
	}
	index=tmpIndex;
}

Effect* workOutEffectType(const Appearance* appearance){
	if(appearance==nullPtr){
		return nullPtr;
	}
	if(ImageTexture* imageTexture=dynamic_cast<ImageTexture*>(appearance->texture.p)){
		TextureEffect* textureEffect= new TextureEffect;

		copySingleUrlWithMultiUrl(textureEffect->url.refValue(), imageTexture->url);
		return textureEffect;
	}
	else{
		MonoColorEffect* monoColorEffect = new MonoColorEffect;

		if (Material* material = dynamic_cast<Material*>(appearance->material.p)) {
			monoColorEffect->diffuse.refValue() = material->diffuseColor;
		}
		return monoColorEffect;
	}
}
void findShapeNodeLoop(Node* node, Shape*& shape){
	if(shape)return;
	shape=dynamic_cast<Shape*>(node);
	if(shape && shape->geometry==0){
		shape=0;
	}
	if(Progenitor* progenitor=dynamic_cast<Progenitor*>(node)){
		M_Node& children=progenitor->children;
		for(int k=0; k<children.length; k++){
			findShapeNodeLoop(children[k], shape);
		}
	}
}

void findNurbsSurfaceNodeLoop(Node* node, NurbsSurface*& nurbsSurface){
	if(nurbsSurface)return;
	if(Shape* shape=dynamic_cast<Shape*>(node)){
		if(nurbsSurface = dynamic_cast<NurbsSurface*>(shape->geometry.p)){
			return;
		}
	}
	Progenitor* progenitor=dynamic_cast<Progenitor*>(node);
	if(progenitor){
		M_Node& children=progenitor->children;
		for(int k=0; k<children.length; k++){
			findNurbsSurfaceNodeLoop(children[k], nurbsSurface);
		}
	}
}

void addGeometry(Art* part, const Shape* shape){

	const Geometry* geometry=dynamic_cast<Geometry*>(shape->geometry.p);
	const Appearance* appearance=dynamic_cast<Appearance*>(shape->appearance.p);
	if(geometry){
		GeomRender mesh;
		geometry->convertToGeomSource(mesh, g_tmpIndex);
		if(dynamic_cast<ColorEffect*>(part->workEffect)&&mesh.color.length==0){
			if(appearance){
				Material* material=dynamic_cast<Material*>(appearance->material.p);
				if(material){
					M_Color3f& color=mesh.color;
					color.length=mesh.coord.length;
					Color3f diffuseColor=material->diffuseColor;
					for(int i=color.length; --i>=0;){
						color[i]=diffuseColor;
					}
				}
			}
		}
		part->workEffect->convertGeomSourceToPart(part, mesh);
	}
}
bool Art::createPartFromString(const String str){
	deleteGeometry();
	M_Node nodes;
	createSmlsFromString(nodes, str);
	Shape* shape=0;
	for(int k=0; k<nodes.length; k++){
		findShapeNodeLoop(nodes[k], shape);
	}
	if(shape){
		addGeometry(this, shape);
		return true;
	}
	return false;
}

bool NurbsPatch::loadDataFromString(const String str){
	M_Node nodes;
	createSmlsFromString(nodes, str);
	NurbsSurface* nurbsSurface = 0;
	for(int k=0; k < nodes.length && nurbsSurface == 0; k++){
		findNurbsSurfaceNodeLoop(nodes[k], nurbsSurface);
	}
	if(nurbsSurface){
		point.refValue()	=	nurbsSurface->controlPoint;
		weight.refValue()	=	nurbsSurface->weight;
		uKnot.refValue()	=	nurbsSurface->uKnot;
		vKnot.refValue()	=	nurbsSurface->vKnot;
		dim.refValue()		=	Vec2f(nurbsSurface->uDimension, nurbsSurface->vDimension);
		order.refValue()	=	Vec2f(nurbsSurface->uOrder, nurbsSurface->vOrder);
		processFields();
		patchToCentralCPs();
		return true;
	}
	return false;
}

bool Seamless::loadGZippedPart(Art* part, String filename){
	String debugStr="Seamless::loadGZippedPart(";
	debugStr+=filename;
	debugStr+=")";
	dLog(debugStr);
	if(filename=="") return false;

	if(filename.length==0){
		return false;
	}
	String str =readFile(filename);
	if(str=="")return false;
	return part->createPartFromString(str);
}
void commonSaveToWrl(String& str, Effect* workEffect){
	str="#VRML V2.0 utf8\n";
	str+="#Generated by Seamless3d ";
	str+=" Copyright � 2001-2010 Graham John Perrett (AKA thyme) thyme@seamless3d.com www.seamless3d.com\n";
	str+="NavigationInfo {type \"examine\"}\n";

	str+="Shape {\n";
	str+="appearance Appearance {material Material {}";

	TextureEffect* textureEffect=dynamic_cast<TextureEffect*>(workEffect);
	if(textureEffect!=0){
		ImageTexturePtr texture = new ImageTexture;
		copyMultiUrlWithSingleUrl(texture->url, textureEffect->url);
		str+=" texture ";
		texture->toString(str);
	}

	str+="}\n\ngeometry ";
}
void Art::saveToWrl(String filename){

	String str;

	commonSaveToWrl(str, workEffect);

	extractGeometry()->toString(str,true);

	str+="\n}";

	str.write(filename);
}
void NurbsPatch::saveToWrl(String filename){

	String str;
	NurbsSurface* nurbsSurface = new NurbsSurface;
	Effect* effect = 0;

	M_Node& m_stem = stem.refValue();
	if(m_stem.length){
		Stem* stem = dynamic_cast<Stem*>(m_stem[0].p);
		if(stem->bands.length==1){
			nurbsSurface->uTessellation = stem->staves[0];
			nurbsSurface->vTessellation = stem->bands[0];
		}
		const M_SArt& m_part = stem->part;
		if(m_part.length){
			effect = m_part[0]->smls->effect;
		}
	}
	commonSaveToWrl(str, effect);

	nurbsSurface->controlPoint=point;
	nurbsSurface->weight=weight;
	nurbsSurface->uKnot=uKnot;
	nurbsSurface->vKnot=vKnot;
	nurbsSurface->uDimension=dim.x;
	nurbsSurface->vDimension=dim.y;
	nurbsSurface->uOrder=order.x,
	nurbsSurface->vOrder=order.y;

	nurbsSurface->toString(str, true);

	str+="\n}";

	str.write(filename);
}


extern int g_projectDirection;



void Seamless::adjustDescendents(int plidx, int coordDiff, int polyDiff){
	int i;

	Art* part = partList[plidx];
	part->toCoord += coordDiff;
	part->toPoly += polyDiff;

	for(i=plidx+1; i<partList.length; i++){

		part = partList[i];

		part->fromCoord += coordDiff;
		part->toCoord += coordDiff;

		part->fromPoly += polyDiff;
		part->toPoly += polyDiff;

		part->coordLen = part->toCoord - part->fromCoord;
	}
}
/*
void Art::resizePart(int newCoordLen, int newPolyLen){
	int i;

	M_Vec3f& point=refPoint();
	M_Vec3f& vector=refVector();
	M_Color3f& color=refColor();
	M_Vec2f& texPoint=refTexPoint();
	M_Int& coordIndex=refCoordIndex();


	int oldCoordLen = toCoord - fromCoord;
	int oldPolyLen = toPoly - fromPoly;
	int coordDiff = newCoordLen - oldCoordLen;
	int polyDiff = newPolyLen - oldPolyLen;

	if(coordDiff > 0){
		point.addLength(coordDiff);

		for( i = point.length; --i-coordDiff >= toCoord;){
			point[i] = point[i - coordDiff];
		}
		if(&color!=NULL){
			color.resize(point.length);
			for( i = point.length; --i-coordDiff >= toCoord;){
				color[i] = color[i - coordDiff];
			}
		}
		if(&texPoint!=NULL){
			texPoint.resize(point.length);
			for( i = point.length; --i-coordDiff >= toCoord;){
				texPoint[i] = texPoint[i - coordDiff];
			}
		}
	}
	else{
		for( i = toCoord; i < point.length; i++){
			point[i + coordDiff] = point[i];
		}
		if(&color!=NULL){
			for( i = toCoord; i < point.length; i++){
				color[i + coordDiff] = color[i];
			}
			color.addLength(coordDiff);
		}
		if(&texPoint!=NULL){
			for( i = toCoord; i < point.length; i++){
				texPoint[i + coordDiff] = texPoint[i];
			}
			texPoint.addLength(coordDiff);
		}
		point.addLength(coordDiff);
	}
	if(polyDiff > 0){
		coordIndex.addLength(polyDiff*4);

		for( i = coordIndex.length/4; --i-polyDiff >= toPoly;){
			coordIndex[i*4+0] = coordIndex[(i-polyDiff)*4+0];
			coordIndex[i*4+1] = coordIndex[(i-polyDiff)*4+1];
			coordIndex[i*4+2] = coordIndex[(i-polyDiff)*4+2];
			coordIndex[i*4+3] = -1;
		}
	}
	else{
		for( i = toPoly; i < coordIndex.length/4; i++){
			coordIndex[(i+polyDiff)*4+0] = coordIndex[i*4+0];
			coordIndex[(i+polyDiff)*4+1] = coordIndex[i*4+1];
			coordIndex[(i+polyDiff)*4+2] = coordIndex[i*4+2];
			coordIndex[(i+polyDiff)*4+3] = -1;
		}
		coordIndex.addLength(polyDiff*4);
	}
	for( i = -1; i<coordIndex.length-1 ; ){
		if(coordIndex[++i]>=fromCoord){
			coordIndex[i] += coordDiff;
		}
		if(coordIndex[++i]>=fromCoord){
			coordIndex[i] += coordDiff;
		}
		if(coordIndex[++i]>=fromCoord){
			coordIndex[i] += coordDiff;
		}
		i++;
	}
	smls->adjustDescendents(plidx, coordDiff, polyDiff);
}
*/
/*
void copyIfsToPart(Art* &part, IndexedFaceSetPtr &ifs){
	Stem stem;
	stem.part=part;
	stem.source=ifs;
	stem.doBuild();
}
*/
inline bool allReadyColorAdded(int& index, M_Int& appendIndex, int& foundIndex){
	for(foundIndex=appendIndex.length; --foundIndex>=0; ){
		if(appendIndex[foundIndex]==index){
			return true;
		}
	}
	return false;
}
void bubbleSortRise(M_Art& part){
	bool flag;
	do{
		flag=false;
		for(int i=0;i<part.length-1;i++){

			if (part[i]->plidx>part[i+1]->plidx){

				Art* swap = part[i+1];
				part[i+1]=part[i];
				part[i]=swap;

				flag=true;
			}
		}
	}while(flag);
}

//M_Vec3f& ifsCoord, M_Color3f& ifsColor, M_Int& ifsCoordIndex, M_Int& ifsColorIndex,
//const M_Vec3f& coord, const M_Triangle& triangle)

inline void ColorEffect::convertPartToIFS( Art* thisPart, IndexedFaceSet* ifs, M_Vec3f& ifsCoord, M_Int& ifsCoordIndex, const M_Vec3f& coord, const M_Triangle& triangle)
{
	Color* colorNode=new Color;
	ifs->color=colorNode;
	M_Color3f& ifsColor=colorNode->color;
	M_Int& ifsColorIndex=ifs->colorIndex;


	M_Int& appendColorIndex=g_tmpIndex;	appendColorIndex.length=0;

	for(int i=0,c=3; i<triangle.length*3; ){
		int colorIdx=triangle.refTexIndex(i);

		int foundIndex;

		if( allReadyColorAdded(colorIdx, appendColorIndex, foundIndex)){
			ifsColorIndex+=foundIndex;
		}
		else {
			appendColorIndex+=colorIdx;

			ifsColorIndex+=ifsColor.length;

			ifsColor+=color[colorIdx];
		}
		i++;
		if(--c==0){
			c=3;
			ifsColorIndex+=-1;
		}
	}
}
inline void MonoColorEffect::convertPartToIFS( Art* thisPart, IndexedFaceSet* ifs, M_Vec3f& ifsCoord, M_Int& ifsCoordIndex, const M_Vec3f& coord, const M_Triangle& triangle)
{
	Color* colorNode=new Color;
	ifs->color=colorNode;
	M_Color3f& ifsColor=colorNode->color;
	M_Int& ifsColorIndex=ifs->colorIndex;


	M_Int& appendColorIndex=g_tmpIndex;	appendColorIndex.length=0;

	for(int i=0,c=3; i<triangle.length*3; ){
		int colorIdx=triangle.refTexIndex(i);

		int foundIndex;

		if( allReadyColorAdded(colorIdx, appendColorIndex, foundIndex)){
			ifsColorIndex+=foundIndex;
		}
		else {
			appendColorIndex+=colorIdx;

			ifsColorIndex+=ifsColor.length;

			ifsColor+=color[colorIdx];
		}
		i++;
		if(--c==0){
			c=3;
			ifsColorIndex+=-1;
		}
	}
}
inline void TextureEffect::convertPartToIFS( Art* thisPart, IndexedFaceSet* ifs, M_Vec3f& ifsCoord, M_Int& ifsCoordIndex, const M_Vec3f& coord, const M_Triangle& triangle)
{
	TextureCoordinate* texCoordNode=new TextureCoordinate;
	ifs->texCoord=texCoordNode;
	M_Vec2f& ifsTexCoord=texCoordNode->point;
	M_Int& ifsTexCoordIndex=ifs->texCoordIndex;


	M_Int& appendTextureIndex=g_tmpIndex;	appendTextureIndex.length=0;

	for(int i=0,c=3; i<triangle.length*3; ){
		int texCoordIdx=triangle.refTexIndex(i);

		int foundIndex;

		if( allReadyColorAdded(texCoordIdx, appendTextureIndex, foundIndex)){
			ifsTexCoordIndex+=foundIndex;
		}
		else {
			appendTextureIndex+=texCoordIdx;

			ifsTexCoordIndex+=ifsTexCoord.length;

			ifsTexCoord+=texCoord[texCoordIdx];
		}
		i++;
		if(--c==0){
			c=3;
			ifsTexCoordIndex+=-1;
		}
	}
}
//	workEffect->sortOutIndex(gs,	gsCoord, gsCoordIndex, coord, triangle);
void TextureEffect::sortOutIndex( Mesh& gs, M_Vec3f& gsCoord, M_Int& gsCoordIndex, const Art* thisPart, const M_Vec3f& coord, const M_Triangle& triangle, Matrix& imatrix, M_Int& maskIndex) const
{
	gsCoordIndex.length=triangle.length*3;
	gsCoord=coord;
	M_Int texCoordIndex;
	texCoordIndex.length=gsCoordIndex.length;
	//M_Int& maskIndex=g_tmpIndex;
	maskIndex.length=gsCoordIndex;


	{
		for(int i=gsCoordIndex.length; --i>=0;){
			maskIndex[i]=-1;
		}
	}
	{
		for(int i=0; i<triangle.length*3; i++){

			texCoordIndex[i]=triangle.refTexIndex(i);

			if(triangle.refPart(i)!=thisPart){
				if(maskIndex[i]!=-1)continue;
				Art* pi=triangle.refPart(i);
				int ti=triangle.refIndex(i);
				for(int j=gsCoordIndex.length; --j>=0;){
					if(triangle.refIndex(j)==ti && triangle.refPart(j)==pi){
						gsCoordIndex[j]=maskIndex[j]=gsCoord.length;
					}
				}
				gsCoord+=imatrix*(pi->matrix*pi->coord[ti]);
			}
			else{
				gsCoordIndex[i]=triangle.refIndex(i);
			}
		}
	}

//inline void Mesh::convertFromTexCoordIndex( M_Int& coordIndex, M_Int& texCoordIndex, const M_TexCoord& ifsTexCoord){
	M_Int& rc=gs.rc.refValue();
	M_Vec2f& gsTexCoord=gs.texCoord.refValue();
	const M_Vec2f& eTexCoord=texCoord.refValue();
	M_Int& coordLookUp=g_tmpIndex;
	const int coordLen=gsCoord.length;
	const int coordIdxLen=gsCoordIndex.length;
	coordLookUp.length=coordLen;
	gsTexCoord.length=coordLen;

	{
		for(int i=coordLookUp.length; --i>=0;){
			coordLookUp[i]=-1;
		}
	}
	for(int i=0; i<coordIdxLen; i++){
		int coordIdx=gsCoordIndex[i];
		int texCoordIdx=texCoordIndex[i];
		int coordLU=coordLookUp[coordIdx];
		if(coordLU==-1){
			gsTexCoord[coordIdx]=eTexCoord[texCoordIdx];
			coordLookUp[coordIdx]=texCoordIdx;
		}
		else if(coordLU!=texCoordIdx){
			int newIdx=coordLen+rc.length;
			gsCoordIndex[i]=newIdx;
			gsTexCoord+=eTexCoord[texCoordIdx];
			coordLookUp+=texCoordIdx;
			rc+=coordIdx;
			{
				for(int i=coordIdxLen; --i>=0;){
					if(texCoordIndex[i]==texCoordIdx&&gsCoordIndex[i]==coordIdx){
						gsCoordIndex[i]=newIdx;
					}
				}
			}
		}
	}
	//delete dead vertices
	maskIndex.length=0;
	maskIndex.length=gsTexCoord.length;
	{
		for(int i=gsCoordIndex.length; --i>=0;){
			maskIndex[gsCoordIndex[i]]=1;
		}
	}
	int reduction;
	{
		int i=0, j=0;
		for(; i<maskIndex.length; i++){
			int store=maskIndex[i];
			maskIndex[i]=j;
			j+=store;
		}
		reduction=i-j;
	}
	{
		for(int i=gsCoordIndex.length; --i>=0;){
			gsCoordIndex[i]=maskIndex[gsCoordIndex[i]];
		}
	}
	{
		for(int i=rc.length; --i>=0;){
			rc[i]=maskIndex[rc[i]];
		}
	}
	{
		for(int i=0; i<gsTexCoord.length; i++){
			gsTexCoord[maskIndex[i]]=gsTexCoord[i];
		}
	}
	{
		for(int i=0; i<gsCoord.length; i++){
			gsCoord[maskIndex[i]]=gsCoord[i];
		}
	}
	gsCoord.length-=reduction;
	gsTexCoord.length-=reduction;
}

void TextureEffect::sortOutIndex(Mesh& gs, M_Vec3f& gsCoord, M_Vec2f& gsUvCoord, M_<NurbsPatch*>& gsPatchOwner, M_Int& gsCoordIndex, const Art* thisPart, const M_Vec3f& coord, const M_Vec2f& uvCoord, const M_<NurbsPatch*>& patchOwner, const M_Triangle& triangle, Matrix& imatrix, M_Int& maskIndex) const
{
}

void TextureEffect::sortOutIndex( Mesh& gs, M_Vec3f& gsCoord, M_Vec3f& gsNormal, M_Int& gsCoordIndex, const Art* thisPart, const M_Vec3f& coord, const M_Vec3f& normal, const M_Triangle& triangle, Matrix& imatrix, M_Int& maskIndex) const
{
	gsCoordIndex.length=triangle.length*3;
	gsCoord=coord;
	gsNormal=normal;
	M_Int texCoordIndex;
	texCoordIndex.length=gsCoordIndex.length;
	//=g_tmpIndex;
	maskIndex.length=gsCoordIndex;


	{
		for(int i=gsCoordIndex.length; --i>=0;){
			maskIndex[i]=-1;
		}
	}
	{
		for(int i=0; i<triangle.length*3; i++){

			texCoordIndex[i]=triangle.refTexIndex(i);

			if(triangle.refPart(i)!=thisPart){
				if(maskIndex[i]!=-1)continue;
				Art* pi=triangle.refPart(i);
				int ti=triangle.refIndex(i);
				for(int j=gsCoordIndex.length; --j>=0;){
					if(triangle.refIndex(j)==ti && triangle.refPart(j)==pi){
						gsCoordIndex[j]=maskIndex[j]=gsCoord.length;
					}
				}
				gsCoord+=imatrix*(pi->matrix*pi->coord[ti]);
				gsNormal+=pi->normal[ti];
			}
			else{
				gsCoordIndex[i]=triangle.refIndex(i);
			}
		}
	}

//inline void Mesh::convertFromTexCoordIndex( M_Int& coordIndex, M_Int& texCoordIndex, const M_TexCoord& ifsTexCoord){
	M_Int& rc=gs.rc.refValue();
	M_Vec2f& gsTexCoord=gs.texCoord.refValue();
	const M_Vec2f& eTexCoord=texCoord.refValue();
	M_Int& coordLookUp=g_tmpIndex;
	const int coordLen=gsCoord.length;
	const int coordIdxLen=gsCoordIndex.length;
	coordLookUp.length=coordLen;
	gsTexCoord.length=coordLen;

	{
		for(int i=coordLookUp.length; --i>=0;){
			coordLookUp[i]=-1;
		}
	}
	for(int i=0; i<coordIdxLen; i++){
		int coordIdx=gsCoordIndex[i];
		int texCoordIdx=texCoordIndex[i];
		int coordLU=coordLookUp[coordIdx];
		if(coordLU==-1){
			gsTexCoord[coordIdx]=eTexCoord[texCoordIdx];
			coordLookUp[coordIdx]=texCoordIdx;
		}
		else if(coordLU!=texCoordIdx){
			int newIdx=coordLen+rc.length;
			gsCoordIndex[i]=newIdx;
			gsTexCoord+=eTexCoord[texCoordIdx];
			coordLookUp+=texCoordIdx;
			rc+=coordIdx;
			{
				for(int i=coordIdxLen; --i>=0;){
					if(texCoordIndex[i]==texCoordIdx&&gsCoordIndex[i]==coordIdx){
						gsCoordIndex[i]=newIdx;
					}
				}
			}
		}
	}
	//delete dead vertices
	maskIndex.length=0;
	maskIndex.length=gsTexCoord.length;
	{
		for(int i=gsCoordIndex.length; --i>=0;){
			maskIndex[gsCoordIndex[i]]=1;
		}
	}
	int reduction;
	{
		int i=0, j=0;
		for(; i<maskIndex.length; i++){
			int store=maskIndex[i];
			maskIndex[i]=j;
			j+=store;
		}
		reduction=i-j;
	}
	{
		for(int i=gsCoordIndex.length; --i>=0;){
			gsCoordIndex[i]=maskIndex[gsCoordIndex[i]];
		}
	}
	{
		for(int i=rc.length; --i>=0;){
			rc[i]=maskIndex[rc[i]];
		}
	}
	{
		for(int i=0; i<gsTexCoord.length; i++){
			gsTexCoord[maskIndex[i]]=gsTexCoord[i];
		}
	}
	{
		for(int i=0; i<gsCoord.length; i++){
			gsCoord[maskIndex[i]]=gsCoord[i];
			gsNormal[maskIndex[i]]=gsNormal[i];
		}
	}
	gsCoord.length-=reduction;
	gsNormal.length-=reduction;
	gsTexCoord.length-=reduction;
}

void ColorEffect::sortOutIndex(Mesh& gs, M_Vec3f& gsCoord, M_Vec2f& gsUvCoord, M_<NurbsPatch*>& gsPatchOwner, M_Int& gsCoordIndex, const Art* thisPart, const M_Vec3f& coord, const M_Vec2f& uvCoord, const M_<NurbsPatch*>& patchOwner, const M_Triangle& triangle, Matrix& imatrix, M_Int& maskIndex) const
{
	gsCoordIndex.length = triangle.length*3;
	gsCoord = coord;
	gsUvCoord = uvCoord;
	gsPatchOwner = patchOwner;
	M_Int colorIndex;
	colorIndex.length = gsCoordIndex.length;
	//M_Int& maskIndex=g_tmpIndex;
	maskIndex.length = gsCoordIndex;


	{
		for(int i=gsCoordIndex.length; --i>=0;){
			maskIndex[i]=-1;
		}
	}
	{
		for(int i=0; i<triangle.length*3; i++){

			colorIndex[i]=triangle.refTexIndex(i);

			if(triangle.refPart(i)!=thisPart){
				if(maskIndex[i]!=-1)continue;
				Art* pi=triangle.refPart(i);
				int ti=triangle.refIndex(i);
				for(int j=gsCoordIndex.length; --j>=0;){
					if(triangle.refIndex(j)==ti && triangle.refPart(j)==pi){
						gsCoordIndex[j]=maskIndex[j]=gsCoord.length;
					}
				}
				gsCoord += imatrix * (pi->matrix * pi->coord[ti]);
				gsUvCoord += pi->uvCoord[ti];
				gsPatchOwner += patchOwner[ti];
			}
			else{
				gsCoordIndex[i] = triangle.refIndex(i);
			}
		}
	}

//inline void Mesh::convertFromColorIndex( M_Int& coordIndex, M_Int& colorIndex, const M_Color3f& ifsColor){
	M_Int& rc = gs.rc.refValue();
	M_Color3f& gsColor = gs.color.refValue();
	const M_Color3f& eColor = color.refValue();
	M_Int& coordLookUp = g_tmpIndex;
	const int coordLen = gsCoord.length;
	const int coordIdxLen = gsCoordIndex.length;
	coordLookUp.length = coordLen;
	gsColor.length = coordLen;

	{
		for(int i=coordLookUp.length; --i>=0;){
			coordLookUp[i]=-1;
		}
	}
	for(int i = 0; i<coordIdxLen; i++){
		int coordIdx = gsCoordIndex[i];
		int colorIdx = colorIndex[i];
		int coordLU = coordLookUp[coordIdx];
		if(coordLU==-1){
			gsColor[coordIdx] = eColor[colorIdx];
			coordLookUp[coordIdx] = colorIdx;
		}
		else if(coordLU != colorIdx){
			int newIdx=coordLen + rc.length;
			gsCoordIndex[i] = newIdx;
			gsColor += eColor[colorIdx];
			coordLookUp += colorIdx;
			rc += coordIdx;
			{
				for(int i = coordIdxLen; --i>=0;){
					if(colorIndex[i] == colorIdx&&gsCoordIndex[i] == coordIdx){
						gsCoordIndex[i] = newIdx;
					}
				}
			}
		}
	}
	//delete dead vertices
	maskIndex.length = 0;
	maskIndex.length = gsColor.length;
	{
		for(int i = gsCoordIndex.length; --i>=0;){
			maskIndex[gsCoordIndex[i]] = 1;
		}
	}
	int reduction;
	{
		int i=0, j=0;
		for(; i<maskIndex.length; i++){
			int store = maskIndex[i];
			maskIndex[i] = j;
			j+=store;
		}
		reduction = i-j;
	}
	{
		for(int i = gsCoordIndex.length; --i>=0;){
			gsCoordIndex[i] = maskIndex[gsCoordIndex[i]];
		}
	}
	{
		for(int i=rc.length; --i>=0;){
			rc[i] = maskIndex[rc[i]];
		}
	}
	{
		for(int i = 0; i<gsColor.length; i++){
			gsColor[maskIndex[i]] = gsColor[i];
		}
	}
	{
		for(int i = 0; i < gsCoord.length; i++){
			gsCoord[maskIndex[i]] = gsCoord[i];
			gsUvCoord[maskIndex[i]] = gsUvCoord[i];
		}
	}


	gsCoord.length -= reduction;
	gsColor.length -= reduction;
}
void MonoColorEffect::sortOutIndex(Mesh& gs, M_Vec3f& gsCoord, M_Vec2f& gsUvCoord, M_<NurbsPatch*>& gsPatchOwner, M_Int& gsCoordIndex, const Art* thisPart, const M_Vec3f& coord, const M_Vec2f& uvCoord, const M_<NurbsPatch*>& patchOwner, const M_Triangle& triangle, Matrix& imatrix, M_Int& maskIndex) const
{
}

void ColorEffect::sortOutIndex( Mesh& gs, M_Vec3f& gsCoord, M_Int& gsCoordIndex, const Art* thisPart, const M_Vec3f& coord, const M_Triangle& triangle, Matrix& imatrix, M_Int& maskIndex) const
{
	gsCoordIndex.length=triangle.length*3;
	gsCoord=coord;
	M_Int colorIndex;
	colorIndex.length=gsCoordIndex.length;
	//M_Int& maskIndex=g_tmpIndex;
	maskIndex.length=gsCoordIndex;


	{
		for(int i=gsCoordIndex.length; --i>=0;){
			maskIndex[i]=-1;
		}
	}
	{
		for(int i=0; i<triangle.length*3; i++){

			colorIndex[i]=triangle.refTexIndex(i);

			if(triangle.refPart(i)!=thisPart){
				if(maskIndex[i]!=-1)continue;
				Art* pi=triangle.refPart(i);
				int ti=triangle.refIndex(i);
				for(int j=gsCoordIndex.length; --j>=0;){
					if(triangle.refIndex(j)==ti && triangle.refPart(j)==pi){
						gsCoordIndex[j]=maskIndex[j]=gsCoord.length;
					}
				}
				gsCoord+=imatrix*(pi->matrix*pi->coord[ti]);
			}
			else{
				gsCoordIndex[i]=triangle.refIndex(i);
			}
		}
	}

//inline void Mesh::convertFromColorIndex( M_Int& coordIndex, M_Int& colorIndex, const M_Color3f& ifsColor){
	M_Int& rc=gs.rc.refValue();
	M_Color3f& gsColor=gs.color.refValue();
	const M_Color3f& eColor=color.refValue();
	M_Int& coordLookUp=g_tmpIndex;
	const int coordLen=gsCoord.length;
	const int coordIdxLen=gsCoordIndex.length;
	coordLookUp.length=coordLen;
	gsColor.length=coordLen;

	{
		for(int i=coordLookUp.length; --i>=0;){
			coordLookUp[i]=-1;
		}
	}
	for(int i=0; i<coordIdxLen; i++){
		int coordIdx=gsCoordIndex[i];
		int colorIdx=colorIndex[i];
		int coordLU=coordLookUp[coordIdx];
		if(coordLU==-1){
			gsColor[coordIdx]=eColor[colorIdx];
			coordLookUp[coordIdx]=colorIdx;
		}
		else if(coordLU!=colorIdx){
			int newIdx=coordLen+rc.length;
			gsCoordIndex[i]=newIdx;
			gsColor+=eColor[colorIdx];
			coordLookUp+=colorIdx;
			rc+=coordIdx;
			{
				for(int i=coordIdxLen; --i>=0;){
					if(colorIndex[i]==colorIdx&&gsCoordIndex[i]==coordIdx){
						gsCoordIndex[i]=newIdx;
					}
				}
			}
		}
	}
	//delete dead vertices
	maskIndex.length=0;
	maskIndex.length=gsColor.length;
	{
		for(int i=gsCoordIndex.length; --i>=0;){
			maskIndex[gsCoordIndex[i]]=1;
		}
	}
	int reduction;
	{
		int i=0, j=0;
		for(; i<maskIndex.length; i++){
			int store=maskIndex[i];
			maskIndex[i]=j;
			j+=store;
		}
		reduction=i-j;
	}
	{
		for(int i=gsCoordIndex.length; --i>=0;){
			gsCoordIndex[i]=maskIndex[gsCoordIndex[i]];
		}
	}
	{
		for(int i=rc.length; --i>=0;){
			rc[i]=maskIndex[rc[i]];
		}
	}
	{
		for(int i=0; i<gsColor.length; i++){
			gsColor[maskIndex[i]]=gsColor[i];
		}
	}
	{
		for(int i=0; i<gsCoord.length; i++){
			gsCoord[maskIndex[i]]=gsCoord[i];
		}
	}
	gsCoord.length-=reduction;
	gsColor.length-=reduction;
}
void ColorEffect::sortOutIndex( Mesh& gs, M_Vec3f& gsCoord, M_Vec3f& gsNormal, M_Int& gsCoordIndex, const Art* thisPart, const M_Vec3f& coord, const M_Vec3f& normal, const M_Triangle& triangle, Matrix& imatrix, M_Int& maskIndex) const
{
	gsCoordIndex.length=triangle.length*3;
	gsCoord=coord;
	gsNormal=normal;
	M_Int colorIndex;
	colorIndex.length=gsCoordIndex.length;
	//M_Int& maskIndex=g_tmpIndex;
	maskIndex.length=gsCoordIndex;


	{
		for(int i=gsCoordIndex.length; --i>=0;){
			maskIndex[i]=-1;
		}
	}
	{
		for(int i=0; i<triangle.length*3; i++){

			colorIndex[i]=triangle.refTexIndex(i);

			if(triangle.refPart(i)!=thisPart){
				if(maskIndex[i]!=-1)continue;
				Art* pi=triangle.refPart(i);
				int ti=triangle.refIndex(i);
				for(int j=gsCoordIndex.length; --j>=0;){
					if(triangle.refIndex(j)==ti && triangle.refPart(j)==pi){
						gsCoordIndex[j]=maskIndex[j]=gsCoord.length;
					}
				}
				gsCoord+=imatrix*(pi->matrix*pi->coord[ti]);
				gsNormal+=pi->normal[ti];
			}
			else{
				gsCoordIndex[i]=triangle.refIndex(i);
			}
		}
	}

//inline void Mesh::convertFromColorIndex( M_Int& coordIndex, M_Int& colorIndex, const M_Color3f& ifsColor){
	M_Int& rc=gs.rc.refValue();
	M_Color3f& gsColor=gs.color.refValue();
	const M_Color3f& eColor=color.refValue();
	M_Int& coordLookUp=g_tmpIndex;
	const int coordLen=gsCoord.length;
	const int coordIdxLen=gsCoordIndex.length;
	coordLookUp.length=coordLen;
	gsColor.length=coordLen;

	{
		for(int i=coordLookUp.length; --i>=0;){
			coordLookUp[i]=-1;
		}
	}
	for(int i=0; i<coordIdxLen; i++){
		int coordIdx=gsCoordIndex[i];
		int colorIdx=colorIndex[i];
		int coordLU=coordLookUp[coordIdx];
		if(coordLU==-1){
			gsColor[coordIdx]=eColor[colorIdx];
			coordLookUp[coordIdx]=colorIdx;
		}
		else if(coordLU!=colorIdx){
			int newIdx=coordLen+rc.length;
			gsCoordIndex[i]=newIdx;
			gsColor+=eColor[colorIdx];
			coordLookUp+=colorIdx;
			rc+=coordIdx;
			{
				for(int i=coordIdxLen; --i>=0;){
					if(colorIndex[i]==colorIdx&&gsCoordIndex[i]==coordIdx){
						gsCoordIndex[i]=newIdx;
					}
				}
			}
		}
	}
	//delete dead vertices
	maskIndex.length=0;
	maskIndex.length=gsColor.length;
	{
		for(int i=gsCoordIndex.length; --i>=0;){
			maskIndex[gsCoordIndex[i]]=1;
		}
	}
	int reduction;
	{
		int i=0, j=0;
		for(; i<maskIndex.length; i++){
			int store=maskIndex[i];
			maskIndex[i]=j;
			j+=store;
		}
		reduction=i-j;
	}
	{
		for(int i=gsCoordIndex.length; --i>=0;){
			gsCoordIndex[i]=maskIndex[gsCoordIndex[i]];
		}
	}
	{
		for(int i=rc.length; --i>=0;){
			rc[i]=maskIndex[rc[i]];
		}
	}
	{
		for(int i=0; i<gsColor.length; i++){
			gsColor[maskIndex[i]]=gsColor[i];
		}
	}
	{
		for(int i=0; i<gsCoord.length; i++){
			gsCoord[maskIndex[i]]=gsCoord[i];
			gsNormal[maskIndex[i]]=gsNormal[i];
		}
	}
	gsCoord.length-=reduction;
	gsNormal.length-=reduction;
	gsColor.length-=reduction;
}

void MonoColorEffect::sortOutIndex( Mesh& gs, M_Vec3f& gsCoord, M_Int& gsCoordIndex, const Art* thisPart, const M_Vec3f& coord, const M_Triangle& triangle, Matrix& imatrix, M_Int& maskIndex) const
{
	gs.coord=coord;

	int i=gsCoordIndex.length=triangle.length*3;
	while(--i>=0){
		gsCoordIndex[i]=triangle.refIndex(i);
	}
}
void MonoColorEffect::sortOutIndex( Mesh& gs, M_Vec3f& gsCoord, M_Vec3f& gsNormal, M_Int& gsCoordIndex, const Art* thisPart, const M_Vec3f& coord, const M_Vec3f& normal, const M_Triangle& triangle, Matrix& imatrix, M_Int& maskIndex) const
{
	gs.coord=coord;
	gs.normal=normal;

	int i=gsCoordIndex.length=triangle.length*3;
	while(--i>=0){
		gsCoordIndex[i]=triangle.refIndex(i);
	}
}

IndexedFaceSetPtr Art::extractGeometry(){
	smls->catMatrix();
	smls->catIMatrix();
	IndexedFaceSetPtr ifs = new IndexedFaceSet;
	//convertToRcForExtractGeometry(smls->partList);
	ifs->creaseAngle=3.14;
	Coordinate* coordNode = new Coordinate;
	ifs->coord=coordNode;
	M_Vec3f& ifsCoord=coordNode->point;
	M_Int& ifsCoordIndex=ifs->coordIndex;


	ifsCoordIndex.length=triangle.length*4;
	ifsCoord.length=0;

	static M_Art refPartList;// parts refrenced by triangles not to be confused with relevantParts
	refPartList.length=0;

	{
		for(int i=triangle.length*3; --i>=0; ){
			addRelevantPart(refPartList, triangle.refPart(i));
		}
	}
	bubbleSortRise(refPartList);
	{
		for(int k=refPartList.length; --k>=0;){
			Art* part=refPartList[k];
			part->partTree.length=0;
			part->partTree.length=part->coord.length;
		}
	}
	{
		for(int i=triangle.length*3; --i>=0;){
			triangle.refPart(i)->partTree[triangle.refIndex(i)]=1;
		}
	}
	{
		for(int k=0;k<refPartList.length; k++){
			Art* part=refPartList[k];
			M_Int& mask=part->partTree;
			for(int i=0; i<mask.length; i++){
				if(mask[i]){
					ifsCoord+=imatrix*(part->matrix*part->coord.refValue()[i]);
				}
				mask[i]=ifsCoord.length-1;

			}
		}
	}
	{
		for(int i=0,c=3,t=0; i<triangle.length*3; i++){
			M_Int& mask=triangle.refPart(i)->partTree;
			ifsCoordIndex[t++]=mask[triangle.refIndex(i)];
			if(--c==0){
				ifsCoordIndex[t++]=-1;
				c=3;
			}
		}
	}

	workEffect->convertPartToIFS( this, ifs, ifsCoord, ifsCoordIndex, coord, triangle);

	return ifs;
}
void Art::extractGeometryGS(Mesh& gs, M_Int& maskIndex){
	gs.reset();

	smls->catMatrix();
	smls->catIMatrix();

	M_Vec3f &gsCoord = gs.coord;
	M_Vec3f &gsNormal = gs.normal;
	M_Int &gsCoordIndex=gs.coordIndex.refValue();

	if(coord.length == normal.length){
		workEffect->sortOutIndex(gs,	gsCoord, gsNormal, gsCoordIndex, this, coord, normal, triangle, imatrix, maskIndex);
	}
	else{
		workEffect->sortOutIndex(gs,	gsCoord, gsCoordIndex, this, coord, triangle, imatrix, maskIndex);
	}

}

void Art::extractGeometryGS(Mesh& gs, M_Vec2f& gsUvCoord, M_<NurbsPatch*>& gsPatchOwner,  M_Int& maskIndex){
	gs.reset();

	smls->catMatrix();
	smls->catIMatrix();

	M_Vec3f& gsCoord = gs.coord;
	M_Vec3f& gsNormal = gs.normal;
	M_Int& gsCoordIndex = gs.coordIndex.refValue();

	workEffect->sortOutIndex(gs, gsCoord, gsUvCoord, gsPatchOwner, gsCoordIndex, this, coord, uvCoord, patchOwner, triangle, imatrix, maskIndex);
}

inline void convertCoordIndexToTri(Art*& part, M_Int& coordIndex, int texCoordBase){
	M_Triangle& triangle=part->triangle;
	triangle.length=coordIndex.length/4;
	for(int ci=coordIndex.length, i=triangle.length; --i>=0;){
		ci-=2;
		triangle[i].tc=texCoordBase+(triangle[i].ic=coordIndex[ci]);
		triangle[i].tb=texCoordBase+(triangle[i].ib=coordIndex[--ci]);
		triangle[i].ta=texCoordBase+(triangle[i].ia=coordIndex[--ci]);

		triangle[i].pc=part;
		triangle[i].pb=part;
		triangle[i].pa=part;
	}
}
inline void convertCoordColorIndexToTri(Art*& part, M_Int& coordIndex, M_Int& colorIndex, int texCoordBase){
	M_Triangle& triangle=part->triangle;
	triangle.length=coordIndex.length/4;
	if(colorIndex.length==coordIndex.length){
		for(int ci=coordIndex.length, i=triangle.length; --i>=0;){
			ci-=2;
			triangle[i].ic=coordIndex[ci];
			triangle[i].tc=texCoordBase+colorIndex[ci];
			triangle[i].ib=coordIndex[--ci];
			triangle[i].tb=texCoordBase+colorIndex[ci];
			triangle[i].ia=coordIndex[--ci];
			triangle[i].ta=texCoordBase+colorIndex[ci];


			triangle[i].pc=part;
			triangle[i].pb=part;
			triangle[i].pa=part;
		}
	}
	else {
		for(int ci=coordIndex.length, i=triangle.length; --i>=0;){
			ci-=2;
			triangle[i].ic=coordIndex[ci];
			triangle[i].tc=texCoordBase+coordIndex[ci];
			triangle[i].ib=coordIndex[--ci];
			triangle[i].tb=texCoordBase+coordIndex[ci];
			triangle[i].ia=coordIndex[--ci];
			triangle[i].ta=texCoordBase+coordIndex[ci];


			triangle[i].pc=part;
			triangle[i].pb=part;
			triangle[i].pa=part;
		}
	}
}
void TextureEffect::convertGeomSourceToPart(Art* part, const Mesh& source){
	M_Vec2f& texCoord=this->texCoord.refValue();
	const int texCoordBase= part->fromEffectCoord=texCoord.length;
	const M_Vec3f& srcCoord=source.coord;
	const M_Int& coordIndex=source.coordIndex;
	const M_Int& rc=source.rc;
	const int coordLen=srcCoord.length;
	part->coord=srcCoord;
	part->normal.refValue()=source.normal;
	M_Triangle& triangle=part->triangle;
	triangle.length=coordIndex.length/3;

	for(int i=coordIndex.length; --i>=0;){
		int ci=coordIndex[i];
		if(ci<coordLen){
			triangle.refIndex(i)=coordIndex[i];
		}
		else{
			triangle.refIndex(i)=rc[ci-coordLen];
		}
		triangle.refTexIndex(i)=texCoordBase+ci;
		triangle.refPart(i)=part;
	}
/*
	if(textureMapper.length>k){
		texCoord.length+=coord.length;
		part->toEffectCoord=texCoord.length;

		TextureMapperPtr texMapper=textureMapper[k];
		const Vec2f& select=texMapper->select;
		addTexture(coord, texCoordBase, texCoord, select, texMapper->position3d.refValue(), texMapper->size2d.refValue(), texMapper->position2d.refValue());
	}
	else {
*/
	const M_Vec2f& srcTexCoord=source.texCoord;
	if(srcTexCoord.length<srcCoord.length){
		texCoord.length+=srcCoord.length;
	}
	else {
		texCoord+=srcTexCoord;
	}
	part->toEffectCoord=texCoord.length;
//	}
	addRelevantPart(part->relevantPart, part);
}

void TextureEffect::convertGeomSourceToPart(Art* part, const Mesh& source, const M_Vec2f& gsUvCoord, const M_<NurbsPatch*>& gsPatchOwner){
}

void ColorEffect::convertGeomSourceToPart(Art* part, const Mesh& source){
	M_Vec3f& coord=part->coord.refValue();
	part->normal.refValue()=source.normal;
	M_Color3f& color=this->color.refValue();
	const int texCoordBase= part->fromEffectCoord=color.length;
	const M_Vec3f& srcCoord=source.coord;
	const M_Int& coordIndex=source.coordIndex;
	const M_Int& rc=source.rc;
	const int coordLen=srcCoord.length;
	part->coord.refValue()=srcCoord;
	M_Triangle& triangle=part->triangle;
	triangle.length=coordIndex.length/3;

	for(int i=coordIndex.length; --i>=0;){
		int ci=coordIndex[i];
		if(ci<coordLen){
			triangle.refIndex(i)=ci;
		}
		else{
			triangle.refIndex(i)=rc[ci-coordLen];
		}
		triangle.refTexIndex(i)=texCoordBase+ci;
		triangle.refPart(i)=part;
	}
	const M_Color3f& srcColor=source.color;
	if(srcColor.length<coord.length){
		color.length+=coord.length;
	}
	else {
		color+=srcColor;
	}
	part->toEffectCoord=color.length;
	addRelevantPart(part->relevantPart, part);
}
void ColorEffect::convertGeomSourceToPart(Art* part, const Mesh& source, const M_Vec2f& gsUvCoord, const M_<NurbsPatch*>& gsPatchOwner){
	M_Color3f& color = this->color.refValue();
	const int texCoordBase= part->fromEffectCoord=color.length;
	const M_Vec3f& srcCoord = source.coord;
	const M_Int& coordIndex = source.coordIndex;
	const M_Int& rc = source.rc;
	const int coordLen = srcCoord.length;

	part->coord.refValue() = srcCoord;
	part->uvCoord.refValue() = gsUvCoord;
	part->patchOwner = gsPatchOwner;

	M_Triangle& triangle = part->triangle;
	triangle.length = coordIndex.length/3;

	for(int i=coordIndex.length; --i>=0;){
		int ci=coordIndex[i];
		if(ci<coordLen){
			triangle.refIndex(i)=ci;
		}
		else{
			triangle.refIndex(i)=rc[ci-coordLen];
		}
		triangle.refTexIndex(i) = texCoordBase+ci;
		triangle.refPart(i) = part;
	}

	const M_Color3f& srcColor = source.color;

	if(srcColor.length < srcCoord.length){
		color.length += srcCoord.length;
	}
	else {
		color += srcColor;
	}
	part->toEffectCoord = color.length;
	addRelevantPart(part->relevantPart, part);
}

void MonoColorEffect::convertGeomSourceToPart(Art* part, const Mesh& source){
	part->coord.refValue()=source.coord;

	const M_Int& coordIndex=source.coordIndex;

	M_Triangle& triangle=part->triangle;
	triangle.length=coordIndex.length/3;

	for(int i=coordIndex.length; --i>=0;){
		triangle.refPart(i)=part;
		triangle.refIndex(i)=coordIndex[i];
	}
	addRelevantPart(part->relevantPart, part);
}

void MonoColorEffect::convertGeomSourceToPart(Art* part, const Mesh& source, const M_Vec2f& gsUvCoord, const M_<NurbsPatch*>& gsPatchOwner){
}

void IndexedFaceSet::convertFromGeomSource(const Mesh& source){
	const M_Vec3f& srcCoord=source.coord;
	const M_Color3f& srcColor=source.color;
	const M_Vec2f& srcTexCoord=source.texCoord;
	const M_Int& srcCoordIndex=source.coordIndex;
	const M_Int& rc=source.rc;
	const int coordLen=srcCoord.length;
	M_Int& coordIndex=this->coordIndex;

	int triLen=srcCoordIndex.length/3;
	coordIndex.length=triLen*4;
	if(srcCoord.length){
		if(coord==nullPtr){
			coord=new Coordinate;
		}
		refPoint()=srcCoord;
	}
	if(srcColor.length){
		if(color==nullPtr){
			color=new Color;
		}
		refColor()=srcColor;

		M_Int& colorIndex=this->colorIndex;
		colorIndex.length=coordIndex.length;

		for(int j=0, i=0, h=3; i<srcCoordIndex.length; ){
			int ci=colorIndex[j]=srcCoordIndex[i];
			if(ci<coordLen){
				coordIndex[j]=srcCoordIndex[i];
			}
			else{
				coordIndex[j]=rc[ci-coordLen];
			}
			j++;
			i++;
			if(--h<=0){
				h=3;
				coordIndex[j]=-1;
				colorIndex[j]=-1;
				j++;
			}
		}
	}
	if(srcTexCoord.length){
		if(texCoord == nullPtr){
			texCoord = new TextureCoordinate;
		}
		TextureCoordinate* texCoord = dynamic_cast<TextureCoordinate*>(this->texCoord.p);
		texCoord->point.refValue() = srcTexCoord;

		M_Int& texCoordIndex=this->texCoordIndex;
		texCoordIndex.length=coordIndex.length;

		for(int j=0, i=0, h=3; i<srcCoordIndex.length; ){
			int ci=texCoordIndex[j]=srcCoordIndex[i];
			if(ci<coordLen){
				coordIndex[j]=srcCoordIndex[i];
			}
			else{
				coordIndex[j]=rc[ci-coordLen];
			}
			j++;
			i++;
			if(--h<=0){
				h=3;
				coordIndex[j]=-1;
				texCoordIndex[j]=-1;
				j++;
			}
		}
	}
}



void Art::setDefaultVrmlTextureMap(){
	Stem stem;
	stem.part+=this;
	TextureMapper& textureMapper=*new TextureMapper;
	textureMapper.autoSize3d=false;

	float negX=0;
	float posX=0;
	float negY=0;
	float posY=0;
	float negZ=0;
	float posZ=0;

	for(int i=triangle.length*3;	--i>=0;){
		Vec3f crd=triangle.refPart(i)->coord[triangle.refIndex(i)];
		if(crd.x<negX){
			negX=crd.x;
		}
		else if(crd.x>posX){
			posX=crd.x;
		}
		if(crd.y<negY){
			negY=crd.y;
		}
		else if(crd.y>posY){
			posY=crd.y;
		}
		if(crd.z<negZ){
			negZ=crd.z;
		}
		else if(crd.z>posZ){
			posZ=crd.z;
		}
	}
	float xDiff=posX-negX;
	float yDiff=posY-negY;
	float zDiff=posZ-negZ;
	bool swapUV = false;
	float x2d=1;
	float y2d=1;

	if(yDiff>xDiff && zDiff>xDiff){		//project on X dimension
		textureMapper.orientation=Rotation(0,1,0,-PI*.5);
		textureMapper.position3d=Vec3f(negY, negZ,0);
		textureMapper.size3d=Vec2f(yDiff, zDiff);

		if(yDiff>=zDiff){
			swapUV=true;
			y2d=zDiff/yDiff;
		}
		else{
			x2d=yDiff/zDiff;
		}
	}
	else if(zDiff>yDiff && xDiff>yDiff){//project on Y dimension
		textureMapper.orientation=Rotation(1,0,0,PI*.5);
		textureMapper.position3d=Vec3f(negX, negZ,0);
		textureMapper.size3d=Vec2f(xDiff, zDiff);

		if(xDiff>=zDiff){
			y2d=zDiff/xDiff;
		}
		else{
			x2d=xDiff/zDiff;
			swapUV=true;
		}
	}
	else if(xDiff>zDiff && yDiff>zDiff){//project on Z dimension
		textureMapper.position3d=Vec3f(negX,negY,0);
		textureMapper.size3d=Vec2f(xDiff, yDiff);

		if(xDiff>=yDiff){
			y2d=yDiff/xDiff;
		}
		else{
			x2d=xDiff/yDiff;
			swapUV=true;
		}
	}
	textureMapper.size2d=Vec2f(x2d, y2d);
	stem.textureMapper.refValue()+=&textureMapper;
	stem.doBuild();

	if(swapUV){
		stem.textureMapper.refValue()+=&textureMapper;
		stem.doBuild();
		M_Vec2f& texCoord=dynamic_cast<TextureEffect*>(smls->effect.p)->texCoord;
		{
			for(int i=texCoord.length; --i>=0;){
				float x=texCoord[i].x;
				texCoord[i].x=texCoord[i].y;
				texCoord[i].y=x;
			}
		}
	}
/*
	PlaceTexture placeTexture;
	placeTexture.part+=this;;
	placeTexture.rotation=PI*.5;

	placeTexture.doBuild();
*/
}

void SurfaceGenerator::generate(M_Vec3f& coord, const int h){
	g_projectDirection=PROJECT_FORWARD;
	M_Node& generator=this->generator.refValue();

	for(int i=0; i<generator.length; i++){
		dynamic_cast<Generator*>(generator[i].p)->generate(coord, h);
	}
}
int TextureEffect::stemCopyLocalTri(Stem* stem, Art* part, M_Vec3f& dstCoord, int h, int s){
	return 0;
}
int TextureEffect::stemCopyLocalTriNonMirror(Stem* stem, Art* part, M_Vec3f& dstCoord, int h, int s){
	return 0;
}

int ColorEffect::stemCopyLocalTri(Stem* stem, Art* part, M_Vec3f& dstCoord, int h, int s){
	Stem* copyStem = stem;
	if(stem->coord.length == 0 && stem->mirrorStem){
		copyStem = stem->mirrorStem;
	}
	const M_Color3f& srcColor = copyStem->color;
	int srcColorLen = srcColor.length;
	const M_Vec2f& srcCoord = copyStem->coord;
	int srcCoordLen = srcCoord.length;

	M_Color3f& dstColor = color.refValue();
	int c = dstColor.length;

	dstColor += srcColor;
	dstCoord.length += srcCoordLen;
	for(int i = 0; i < srcCoordLen; i++){
		Vec2f crd2f = srcCoord[i];
		dstCoord[h + i] = Vec3f(crd2f.x, crd2f.y, 0);
	}
	part->triangle.length += copyStem->coordIndex.length / 3;
	copyStem->convertCoordIndexToTriangle(part, h, c, s);

	if(stem->mirror && stem->mirrorStem == 0){
		int h2 = dstCoord.length;
		int c2 = dstColor.length;
		int s2 = part->triangle.length;

		dstColor += srcColor;
		dstCoord.length += srcCoordLen;
		for(int i = 0; i < srcCoordLen; i++){
			Vec2f crd2f = srcCoord[i];
			dstCoord[h2 + i] = Vec3f(-(crd2f.x - .5) + .5, crd2f.y, 0);
		}
		part->triangle.length += copyStem->coordIndex.length / 3;
		copyStem->convertCoordIndexToTriangle(part, h2, c2, s2);
		part->insideOut(s2);
	}
	return c;
}

int ColorEffect::stemCopyLocalTriNonMirror(Stem* stem, Art* part, M_Vec3f& dstCoord, int h, int s){
	Stem* copyStem = stem;
	if(stem->coord.length == 0 && stem->mirrorStem){
		copyStem = stem->mirrorStem;
	}
	const M_Color3f& srcColor = copyStem->color;
	int srcColorLen = srcColor.length;
	const M_Vec2f& srcCoord = copyStem->coord;
	int srcCoordLen = srcCoord.length;

	M_Color3f& dstColor = color.refValue();
	int c = dstColor.length;

	dstColor += srcColor;
	dstCoord.length += srcCoordLen;
	for(int i = 0; i < srcCoordLen; i++){
		Vec2f crd2f = srcCoord[i];
		dstCoord[h + i] = Vec3f(crd2f.x, crd2f.y, 0);
	}
	part->triangle.length += copyStem->coordIndex.length / 3;
	copyStem->convertCoordIndexToTriangle(part, h, c, s);

	return c;
}

int MonoColorEffect::stemCopyLocalTri(Stem* stem, Art* part, M_Vec3f& dstCoord, int h, int s){
	return 0;
}
int MonoColorEffect::stemCopyLocalTriNonMirror(Stem* stem, Art* part, M_Vec3f& dstCoord, int h, int s){
	return 0;
}

void TextureEffect::stemDupMirrorTri(Art* part, M_Vec3f& coord, M_Triangle& triangle, const int h, const int c, const int s, const int coordLen, const int colorLen, const int triLen){
	M_Vec2f& texCoord = this->texCoord.refValue();

	int h2 = h + coordLen;
	coord.length += coordLen;
	for(int i = coordLen; --i>=0;){
		Vec3f crd = coord[h + i];
		crd.x = -(crd.x - .5) + .5;
		coord[h2 + i] = crd;
	}

	int c2 = c + colorLen;
	texCoord.length += colorLen;
	for(int i = colorLen; --i>=0;){
		texCoord[c2 + i] = texCoord[c + i];
	}

	int s2 = s + triLen;
	triangle.length += triLen;
	for(int i = triLen; --i>=0;){
		S_Triangle tri = triangle[s + i];
		tri.ia += coordLen;
		tri.ta += colorLen;

		int is = tri.ib + coordLen;
		int ts = tri.tb + colorLen;
		tri.ib = tri.ic + coordLen;
		tri.tb = tri.tc + colorLen;
		tri.ic = is;
		tri.tc = ts;

		triangle[s2 + i] = tri;
	}
}

void ColorEffect::stemDupMirrorTri(Art* part, M_Vec3f& coord, M_Triangle& triangle, const int h, const int c, const int s, const int coordLen, const int colorLen, const int triLen){
	M_Color3f& color = this->color.refValue();

	int h2 = h + coordLen;
	coord.length += coordLen;
	for(int i = coordLen; --i>=0;){
		Vec3f crd = coord[h + i];
		crd.x = -(crd.x - .5) + .5;
		coord[h2 + i] = crd;
	}

	int c2 = c + colorLen;
	color.length += colorLen;
	for(int i = colorLen; --i>=0;){
		color[c2 + i] = color[c + i];
	}

	int s2 = s + triLen;
	triangle.length += triLen;
	for(int i = triLen; --i>=0;){
		S_Triangle tri = triangle[s + i];
		tri.ia += coordLen;
		tri.ta += colorLen;

		int is = tri.ib + coordLen;
		int ts = tri.tb + colorLen;
		tri.ib = tri.ic + coordLen;
		tri.tb = tri.tc + colorLen;
		tri.ic = is;
		tri.tc = ts;

		triangle[s2 + i] = tri;
	}
}

void MonoColorEffect::stemDupMirrorTri(Art* part, M_Vec3f& coord, M_Triangle& triangle, const int h, const int c, const int s, const int coordLen, const int colorLen, const int triLen){
}



int ColorEffect::polymapCopyLocalTri(Polymap* polymap, Art* part, M_Vec3f& dstCoord, int h, int s){
	const M_Color3f& srcColor = polymap->color;
	int srcColorLen = srcColor.length;
	const M_Vec2f& srcCoord = polymap->coord;
	int srcCoordLen = srcCoord.length;

	M_Color3f& dstColor = color.refValue();
	int c = dstColor.length;

	dstColor += srcColor;
	dstCoord.length += srcCoordLen;
	for(int i = 0; i < srcCoordLen; i++){
		Vec2f crd2f = srcCoord[i];
		dstCoord[h + i] = Vec3f(crd2f.x, crd2f.y, 0);
	}
	part->triangle.length += polymap->coordIndex.length / 3;
	polymap->convertCoordIndexToTriangle(part, h, c, s);

	return c;
}

void Polymap::processFields(){
	patchOwner.refValue().length = coord.length;
}

void Seamless::doBorderlessBuild(){
	for(int p = 0; p < partList.length; p++){
		Art* part = partList[p];
		M_Vec3f& coord = part->coord;
		M_Vec2f& uvCoord = part->uvCoord;
		M_NurbsPatch& patchOwner = part->patchOwner;
		coord.length = uvCoord.length;
		if(uvCoord.length){
			for(int k = 0; k < uvCoord.length; k++){
				NurbsPatch* patch = patchOwner[k];
				//coord[k] = Vec3f(uvCoord[k].x, uvCoord[k].y, 0);
				patch->generate3(coord, k, uvCoord[k]);
			}
		}
	}
}

void Polymap::doBuild(){

	if(Art* prt = this->part.refValue()){
		M_Vec3f& dstCoord=prt->coord.refValue();
		M_Triangle& triangle = prt->triangle;
		if(add == false){
			dstCoord.length=0;
			triangle.length=0;
		}
		int h = dstCoord.length;
		beginCoordOffset = h;
		beginTriOffset = triangle.length;

		int triOffset = triangle.length;
		ColorEffect* effect = dynamic_cast<ColorEffect*>(prt->workEffect);
		static CopyPart cp;
		int c = 0;

		c = effect->polymapCopyLocalTri(this, prt, dstCoord, h, beginTriOffset);

		static M_Vec3f tmpCoord;
		tmpCoord.length = 1;

		for(int k = 0; k < dstCoord.length; k++){
			int po = patchOwner[k];
			NurbsPatch* patch = dynamic_cast<NurbsPatch*>(surface.refValue()[po].p);
			tmpCoord[0] = dstCoord[k];
			patch->generate2(tmpCoord, h);
			dstCoord[k] = tmpCoord[0];
		}

		addRelevantPart(prt->relevantPart, prt);
	}
}


void SurfaceTransform::generate(M_Vec3f& coord, const int h){
	for(int i=coord.length;--i>=h;){
		coord[i]=matrix.multVecMatrix(coord[i]);
	}
}
bool Art::sharedVertices(){
	return false;
	if(relevantPart.length>1)return true;
	M_Art& partList=smls->partList;
	for(int k=plidx+1; k<partList.length; k++){
		M_Art& relevantPart=partList[k]->relevantPart;
		for(int k=relevantPart.length; --k>=0;){
			if(relevantPart[k]==this){
				return true;
			}
		}
	}
	return false;
}
void ColorEffect::copyAddColor(Art* src, Art* dst, const M_Triangle& srcTriangle, M_Triangle& dstTriangle,int& markTriangle){
	static M_Int redirect;
	redirect.length=0;
	ColorEffect* effect=dynamic_cast<ColorEffect*>(src->workEffect);
	M_Color3f& color=effect->color.refValue();
	M_Int& lookup=effect->lookup;
	if(src==dst){
		for(int i=0;i<markTriangle;i++){
			int cli=dstTriangle.refTexIndex(i);
			int lup=lookup[cli];
			if(lup==-1){
				dstTriangle.refTexIndex(i+markTriangle)=lookup[cli]=color.length;
				color+=color[cli];
				redirect+=cli;
			}
			else{
				dstTriangle.refTexIndex(i+markTriangle)=lup;
			}
		}
	}
	else{
		int srcTriLen=srcTriangle.length*3;
		for(int i=0;i<srcTriLen;i++){
			int cli=srcTriangle.refTexIndex(i);
			int lup=lookup[cli];
			if(lup==-1){
				dstTriangle.refTexIndex(i+markTriangle)=lookup[cli]=color.length;
				color+=color[cli];
				redirect+=cli;
			}
			else{
				dstTriangle.refTexIndex(i+markTriangle)=lup;
			}
		}
	}
	{
		for(int i=redirect.length; --i>=0;){
			lookup[redirect[i]]=-1;
		}
	}
}
void MonoColorEffect::copyAddColor(Art* src, Art* dst, const M_Triangle& srcTriangle, M_Triangle& dstTriangle,int& markTriangle){
}
void TextureEffect::copyAddColor(Art* src, Art* dst, const M_Triangle& srcTriangle, M_Triangle& dstTriangle,int& markTriangle){
	static M_Int redirect;
	redirect.length=0;
	TextureEffect* effect=dynamic_cast<TextureEffect*>(src->workEffect);
	M_Vec2f& texCoord=effect->texCoord.refValue();
	M_Int& lookup=effect->lookup;
	if(src==dst){
		for(int i=0;i<markTriangle;i++){
			int cli=dstTriangle.refTexIndex(i);
			int lup=lookup[cli];
			if(lup==-1){
				dstTriangle.refTexIndex(i+markTriangle)=lookup[cli]=texCoord.length;
				texCoord+=texCoord[cli];
				redirect+=cli;
			}
			else{
				dstTriangle.refTexIndex(i+markTriangle)=lup;
			}
		}
	}
	else{
		int srcTriLen=srcTriangle.length*3;
		for(int i=0;i<srcTriLen;i++){
			int cli=srcTriangle.refTexIndex(i);
			int lup=lookup[cli];
			if(lup==-1){
				dstTriangle.refTexIndex(i+markTriangle)=lookup[cli]=texCoord.length;
				texCoord+=texCoord[cli];
				redirect+=cli;
			}
			else{
				dstTriangle.refTexIndex(i+markTriangle)=lup;
			}
		}
	}
	{
		for(int i=redirect.length; --i>=0;){
			lookup[redirect[i]]=-1;
		}
	}
}


void ColorEffect::copyColor(Art* src, Art* dst, const M_Triangle& srcTriangle, M_Triangle& dstTriangle){
	static M_Int redirect;
	redirect.length=0;
	ColorEffect* effect=dynamic_cast<ColorEffect*>(src->workEffect);
	M_Color3f& color=effect->color.refValue();
	M_Int& lookup=effect->lookup;
	{
		for(int i=srcTriangle.length*3; --i>=0;){
			int cli=srcTriangle.refTexIndex(i);
			int lup=lookup[cli];
			if(lup==-1){
				dstTriangle.refTexIndex(i)=lookup[cli]=color.length;
				color+=color[cli];
				redirect+=cli;
			}
			else{
				dstTriangle.refTexIndex(i)=lup;
			}
		}
	}
	{
		for(int i=redirect.length; --i>=0;){
			lookup[redirect[i]]=-1;
		}
	}
}
void MonoColorEffect::copyColor(Art* src, Art* dst, const M_Triangle& srcTriangle, M_Triangle& dstTriangle){
}
void TextureEffect::copyColor(Art* src, Art* dst, const M_Triangle& srcTriangle, M_Triangle& dstTriangle){
	static M_Int redirect;
	redirect.length=0;
	TextureEffect* effect=dynamic_cast<TextureEffect*>(src->workEffect);
	M_Vec2f& texCoord=effect->texCoord.refValue();
	dst->fromEffectCoord=texCoord.length;
	M_Int& lookup=effect->lookup;
	{
		for(int i=srcTriangle.length*3; --i>=0;){
			int cli=srcTriangle.refTexIndex(i);
			int lup=lookup[cli];
			if(lup==-1){
				dstTriangle.refTexIndex(i)=lookup[cli]=texCoord.length;
				texCoord+=texCoord[cli];
				redirect+=cli;
			}
			else{
				dstTriangle.refTexIndex(i)=lup;
			}
		}
	}
	{
		for(int i=redirect.length; --i>=0;){
			lookup[redirect[i]]=-1;
		}
	}
	dst->toEffectCoord=texCoord.length;;
}


void Node::undoIS(int fieldIdx){// route needs to be removed probably
	this->getField(fieldIdx)->isRef()=false;
}
void findMatchingIs(Field* field, ProtoDeclare*& protoDeclare, int& i){
	for(int k=g_protoList.length; --k>=0;){
		protoDeclare=g_protoList[k];
		ProtoInterfacePtr protoInterface=protoDeclare->protoInterface;
		for(i=0; i<protoInterface->getFLen(); i++){
			if(protoInterface->getField(i)->testRouteTo(field)){
				return;
			}
		}
	}
	i=-1;
}
void findMatchingIs(Field* field, ProtoInterface *&protoInterface, int& i){
	ProtoDeclare* protoDeclare=0;
	findMatchingIs(field, protoDeclare, i);
	protoInterface=protoDeclare->protoInterface;
}

ProtoInstance::ProtoInstance(const int k){
	userNodeType=k;
	const ProtoInterface* protoInterface=g_lookupProtoInterface[k];
#ifdef SEAMLESS3DGUI
		index.length=50;
#endif
	for(int i=0; i<protoInterface->size(); i++){
		Field* field=protoInterface->getField(i);
		switch(field->getType()){
		case MFBOOL:{
			MFBool* f=dynamic_cast<MFBool*>(field);
			*this+=new MFBool(M_Bool(f->getValue()));
		}
		break;
		case MFCOLOR:{
			MFColor* f=dynamic_cast<MFColor*>(field);
			*this+=new MFColor(M_Color3f(f->getValue()));
		}
		break;
		case MFFLOAT:{
			MFFloat* f=dynamic_cast<MFFloat*>(field);
			*this+=new MFFloat(M_Float(f->getValue()));
		}
		break;
		case MFINT32:{
			MFInt32* f=dynamic_cast<MFInt32*>(field);
			*this+=new MFInt32(M_Int(f->getValue()));
		}
		break;
		case MFNODE:{
			MFNode* f=dynamic_cast<MFNode*>(field);
			*this+=new MFNode(M_Node(f->getValue()));
		}
		break;
		case MFROTATION:{
			MFRotation* f=dynamic_cast<MFRotation*>(field);
			*this+=new MFRotation(M_Rotation(f->getValue()));
		}
		break;
		case MFSTRING:{
			MFString* f=dynamic_cast<MFString*>(field);
			*this+=new MFString(M_String(f->getValue()));
		}
		break;
		case MFDOUBLE:{
			MFTime* f=dynamic_cast<MFTime*>(field);
			*this+=new MFTime(M_Double(f->getValue()));
		}
		break;
		case MFVEC2F:{
			MFVec2f* f=dynamic_cast<MFVec2f*>(field);
			*this+=new MFVec2f(M_Vec2f(f->getValue()));
		}
		break;
		case MFVEC3F:{
			MFVec3f* f=dynamic_cast<MFVec3f*>(field);
			*this+=new MFVec3f(M_Vec3f(f->getValue()));
		}
		break;
		case SFBOOL:{
			SFBool* f=dynamic_cast<SFBool*>(field);
			*this+=new SFBool(bool(f->getValue()));
		}
		break;
		case SFCOLOR:{
			SFColor* f=dynamic_cast<SFColor*>(field);
			*this+=new SFColor(Color3f(f->getValue()));
		}
		break;
		case SFFLOAT:{
			SFFloat* f=dynamic_cast<SFFloat*>(field);
			*this+=new SFFloat(float(f->getValue()));
		}
		break;
		case SFINT32:{
			SFInt32* f=dynamic_cast<SFInt32*>(field);
			*this+=new SFInt32(int(f->getValue()));
		}
		break;
		case SFNODE:{
			SFNode* f=dynamic_cast<SFNode*>(field);
			*this+=new SFNode(NodePtr(f->getValue()));
		}
		break;
		case SFROTATION:{
			SFRotation* f=dynamic_cast<SFRotation*>(field);
			*this+=new SFRotation(Rotation(f->getValue()));
		}
		break;
		case SFSTRING:{
			SFString* f=dynamic_cast<SFString*>(field);
			*this+=new SFString(String(f->getValue()));
		}
		break;
		case SFTIME:{
			SFTime* f=dynamic_cast<SFTime*>(field);
			*this+=new SFTime(double(f->getValue()));
		}
		break;
		case SFVEC2F:{
			SFVec2f* f=dynamic_cast<SFVec2f*>(field);
			*this+=new SFVec2f(Vec2f(f->getValue()));
		}
		break;
		case SFVEC3F:{
			SFVec3f* f=dynamic_cast<SFVec3f*>(field);
			*this+=new SFVec3f(Vec3f(f->getValue()));
		}
		break;
		default: throw;
		}
	}
}

void DynamicMField::read2(Field *&field, Properties& sprop, bool& wasMF, const String& str,int& i){
	str.stb(i);
	if(str.cmpKWord("MFBool",i)){
		wasMF=true;
		sprop.name=str.getWord(i);
		field=new MFBool(*new M_Bool());
	}
	else if(str.cmpKWord("MFColor",i)){
		wasMF=true;
		sprop.name=str.getWord(i);
		field=new MFColor(*new M_Color3f());
	}
	else if(str.cmpKWord("MFFloat",i)){
		wasMF=true;
		sprop.name=str.getWord(i);
		field=new MFFloat(*new M_Float());
	}
	else if(str.cmpKWord("MFInt32",i)){
		wasMF=true;
		sprop.name=str.getWord(i);
		field=new MFInt32(*new M_Int());
	}
	else if(str.cmpKWord("MFNode",i)){
		wasMF=true;
		sprop.name=str.getWord(i);
		field=new MFNode(*new M_Node());
	}
	else if(str.cmpKWord("MFRotation",i)){
		wasMF=true;
		sprop.name=str.getWord(i);
		field=new MFRotation(*new M_Rotation());
	}
	else if(str.cmpKWord("MFString",i)){
		wasMF=true;
		sprop.name=str.getWord(i);
		field=new MFString(*new M_String());
	}
	else if(str.cmpKWord("MFTime",i)){
		wasMF=true;
		sprop.name=str.getWord(i);
		field=new MFTime(*new M_Double());
	}
	else if(str.cmpKWord("MFVec2f",i)){
		wasMF=true;
		sprop.name=str.getWord(i);
		field=new MFVec2f(*new M_Vec2f());
	}
	else if(str.cmpKWord("MFVec3f",i)){
		wasMF=true;
		sprop.name=str.getWord(i);
		field=new MFVec3f(*new M_Vec3f());
	}
	else if(str.cmpKWord("SFBool",i)){
		sprop.name=str.getWord(i);
		field=new SFBool(*new bool(false));
	}
	else if(str.cmpKWord("SFColor",i)){
		sprop.name=str.getWord(i);
		field=new SFColor(*new Color3f());
	}
	else if(str.cmpKWord("SFFloat",i)){
		sprop.name=str.getWord(i);
		field=new SFFloat(*new float(0));
	}
	else if(str.cmpKWord("SFInt32",i)){
		sprop.name=str.getWord(i);
		field=new SFInt32(*new int(0));
	}
	else if(str.cmpKWord("SFNode",i)){
		sprop.name=str.getWord(i);
		field=new SFNode(*new NodePtr(0));
	}
	else if(str.cmpKWord("SFRotation",i)){
		sprop.name=str.getWord(i);
		field=new SFRotation(*new Rotation());
	}
	else if(str.cmpKWord("SFString",i)){
		sprop.name=str.getWord(i);
		field=new SFString(*new String());
	}
	else if(str.cmpKWord("SFTime",i)){
		sprop.name=str.getWord(i);
		field=new SFTime(*new double(0));
	}
	else if(str.cmpKWord("SFVec2f",i)){
		sprop.name=str.getWord(i);
		field=new SFVec2f(*new Vec2f());
	}
	else if(str.cmpKWord("SFVec3f",i)){
		sprop.name=str.getWord(i);
		field=new SFVec3f(*new Vec3f());
	}
	else throw;

}
void Properties::read(Field& field, const String& str,int& i, bool& beginGroup, bool& endGroup){
	str.stb(i);
	if(str[i]!='{'){
		throw this;
	}
	str.stb(++i);
	while(str[i]!='}'){
		if(str.cmpKWord("promptNode",i)){
			str.stb(i);
			i++;
			int j=0;
			while(str[i]!='\"'){
				int j;
				str.stb(i);
				for(j=0; j < numOfBuiltInNodes; j++){
					if(str.cmpKWord(nodeLookUpStr[j],i)){
						promptNode+=j;
						break;
					}
				}
				if(j==numOfBuiltInNodes) throw this;
			}
			i++;
		}
		else if (str.cmpKWord("tip",i)){
			str.stb(i);
			i++;
			tip=protoInitilaiseStr2+i;
			while(str[i]!='\"')i++;
			*const_cast<char*>(protoInitilaiseStr2+i)=0;
			i++;
		}
		else if (str.cmpKWord("push",i))set_push(getBool(str,i));
		else if (str.cmpKWord("doNotOutput",i))set_doNotOutput(getBool(str,i));
		else if (str.cmpKWord("doNotSave",i))set_doNotSave(getBool(str,i));
		else if (str.cmpKWord("blockUse",i))set_blockUse(getBool(str,i));
		else if (str.cmpKWord("blockNew",i)) set_blockNew(getBool(str,i));
		else if (str.cmpKWord("beginGroup",i)) beginGroup=getBool(str,i);
		else if (str.cmpKWord("sameAsLast",i)) set_sameAsLast(getBool(str,i));
		else if (str.cmpKWord("axis",i)) set_axis(getBool(str,i));
		else if (str.cmpKWord("startAtIndex",i)){startAtIndex=getInt(str,i);}
		else if (str.cmpKWord("endGroup",i)) endGroup=getBool(str,i);
		else if (str.cmpKWord("disable",i)) set_disable(getBool(str,i));
		else if (str.cmpKWord("disableLen",i)) set_disableLen(getBool(str,i));

		else if (str.cmpKWord("initElLen",i)){initElLen=getInt(str,i);}
		else if (str.cmpKWord("incElLen",i)){incElLen=getInt(str,i);}

		else if (str.cmpKWord("rangeFrom",i)){rangeFrom=getFloat(str,i);}
		else if (str.cmpKWord("rangeTo",i)){rangeTo=getFloat(str,i);}

		else if	(str.cmpKWord("initElVal",i)){
			initElLen=field.countItems(str,i);
			initElVal.getField(str,i);
		}
		else if	(str.cmpKWord("incElVal",i)){
			str.stb(i);
			incElLen=field.countItems(str,i);
			incElVal.getField(str,i);
		}


		else throw;
		str.stb(i);
	}
//	else if (str.cmpKWord("cycleInterval",i)) getFieldB(this, 0, str, i);
}
void DynamicMField::read(M_Properties &properties,const String& str,int& i){
	bool groupHasBegan=false;
	bool wasMF;
	int groupCntr=-1;
	if(str[i]!='[')throw this;
	i++;
	str.stb(i);
	Field *field;
	while(str[i]!=']'){
		Properties sprop;
		wasMF=false;
		if(str.cmpKWord(g_initializeOnly,i)||str.cmpKWord(g_field,i)){
			sprop.flags = UNREADABLE | UNWRITEABLE;
			read2(field, sprop, wasMF, str,i);
			field->read(str,i);
		}
		else if(str.cmpKWord(g_inputOnly,i)||str.cmpKWord(g_eventIn,i)){
			sprop.flags = UNREADABLE;
			read2(field, sprop, wasMF, str,i);
		}
		else if(str.cmpKWord(g_outputOnly,i)||str.cmpKWord(g_eventOut,i)){
			sprop.flags = UNWRITEABLE;
			read2(field, sprop, wasMF, str,i);
		}
		else if(str.cmpKWord(g_inputOutput,i)||str.cmpKWord(g_exposedField,i)){
			read2(field, sprop, wasMF, str,i);
			field->read(str,i);
		}
		else throw this;
/*
		if(wasMF&&(field->isReadable()^field->isWritable())==false&&field->getType()!=MFNODE){
			sprop.groupIndex=groupCntr;
			groupCntr++;
		}
		else{
			sprop.groupIndex=-1;
		}
*/
		*this+=field;
		str.stbc(i);

		bool beginGroup=false;
		bool endGroup=false;
		if(str[i]=='#'){
			i++;
			if(str.cmpKWord("Properties",i)){
				sprop.read(*field, str, i, beginGroup, endGroup);
			}
			i=str.indexOf('\n', i);
		}
		if( wasMF&&(sprop.isReadable()^sprop.isWritable())==false&&field->getType()!=MFNODE){
			if(groupHasBegan==false&&beginGroup==false){
				sprop.groupIndex=++groupCntr;
			}
			else if(beginGroup){
				sprop.groupIndex=++groupCntr;
				groupHasBegan=true;
			}
			else if(groupHasBegan){
				sprop.groupIndex=groupCntr;
			}
			if(endGroup){
				groupHasBegan=false;
			}
		}

		properties+=sprop;

		str.stb(i);
	}
	i++;
}

void ProtoInterface::getFields(const String& str,int& i){
	str.stb(i);

	DynamicMField::read(properties, str,i);

	str.stb(i);
}
void ProtoDeclare::getFields(const String& str,int& i){

	protoInterface->getFields(str,i);

	if(str[i]!='{'){
		throw this;
	}
	str.stb(++i);
	while(str[i]!='}'){
		if(str.cmpKWord("ROUTE",i)){
			S_Route(str,i);
			str.stb(i);
			continue;
		}
		protoBody->children.refValue()+=getNode2(str, i);
		str.stb(i);
	}
	i++;
}
void readIS(Node* node, Field* field, const String& str, int& i ){
	String name=str.getWord(i);
	ProtoInterface* protoInterface=g_protoDeclare->protoInterface;
	M_Properties& properties=protoInterface->properties;
	for(int j=0;j<protoInterface->size();j++){
		if(properties[j].name==name){
			Field* protoField=protoInterface->getField(j);
			field->connect(node, protoInterface, protoField); //field = previous
			break;
		}
	}
	str.stb(i);
}
void DynamicMField::readForScript(Node* node, M_Properties &properties, Script* script, const String& str,int& i){
	bool groupHasBegan=false;
	bool wasMF;
	int groupCntr=0;
	Field *field;

	while(str[i]!='}'){
		Properties sprop;
		wasMF=false;
		bool foundIs=false;
		if(str.cmpKWord(g_initializeOnly,i)||str.cmpKWord(g_field,i)){
			sprop.flags = UNREADABLE | UNWRITEABLE;
			read2(field, sprop, wasMF, str,i);
			str.stb(i);
			if(str.cmpKWord("IS",i)){
				foundIs=true;
			}
			else{
				field->read(str,i);
			}
		}
		else if(str.cmpKWord(g_inputOnly,i)||str.cmpKWord(g_eventIn,i)){
			sprop.flags = UNREADABLE;
			read2(field, sprop, wasMF, str,i);
			str.stb(i);
			if(str.cmpKWord("IS",i)){
				foundIs=true;
			}
		}
		else if(str.cmpKWord(g_outputOnly,i)||str.cmpKWord(g_eventOut,i)){
			sprop.flags = UNWRITEABLE;
			read2(field, sprop, wasMF, str,i);
			str.stb(i);
			if(str.cmpKWord("IS",i)){
				foundIs=true;
			}
		}
		else if(str.cmpKWord(g_inputOutput,i)||str.cmpKWord(g_exposedField,i)){
			read2(field, sprop, wasMF, str,i);
			str.stb(i);
			if(str.cmpKWord("IS",i)){
				foundIs=true;
			}
			else{
				field->read(str,i);
			}
		}
		else if(str.cmpKWord("directOutput",i)){
			script->directOutput.refValue()=getBool(str,i);
			str.stb(i);
			continue;
		}
		else if(str.cmpKWord("mustEvaluate",i)){
			script->mustEvaluate.refValue()=getBool(str,i);
			str.stb(i);
			continue;
		}
		else if(str.cmpKWord("url",i)){
			script->url.refValue().getField(str,i);
			str.stb(i);
			continue;
		}
/*
		else if(str.cmpKWord("IS",i)){
			String name=str.getWord(i);
			ProtoInterface* protoInterface=g_protoDeclare->protoInterface;
			for(int j=0;j<protoInterface->length;j++){
				if(protoInterface->properties[j].name==name){
					Field* protoField=protoInterface->getField(j);
					field->connect(node, protoInterface, protoField); //field = previous
					break;
					//field.reseat(fieldIdx, g_protoDeclare->field[j]);
				}
			}
			str.stb(i);
			continue;
		}
*/
		else throw this;
/*
		if(wasMF&&(field->isReadable()^field->isWritable())==false&&field->getType()!=MFNODE){
			sprop.groupIndex=groupCntr;
			groupCntr++;
		}
		else{
			sprop.groupIndex=-1;
		}
*/
		*this+=field;
		if(foundIs){
			readIS(node, field, str, i );
		}
		str.stbc(i);

		bool beginGroup=false;
		bool endGroup=false;
		if(str[i]=='#'){
			i++;
			if(str.cmpKWord("Properties",i)){
				sprop.read(*field, str, i, beginGroup, endGroup);
			}
			i=str.indexOf('\n', i);
		}
		if( wasMF&&(sprop.isReadable()^sprop.isWritable())==false&&field->getType()!=MFNODE){
			if(groupHasBegan==false&&beginGroup==false){
				sprop.groupIndex=++groupCntr;
			}
			else if(beginGroup){
				sprop.groupIndex=++groupCntr;
				groupHasBegan=true;
			}
			else if(groupHasBegan){
				sprop.groupIndex=groupCntr;
			}
			if(endGroup){
				groupHasBegan=false;
			}
		}

		properties+=sprop;

		str.stb(i);
	}
}

void Script::getFields(const String& str,int& i){
	str.stb(i);

	DynamicMField::readForScript(this,properties,this, str,i);

}


inline void Node::getField(const int& fieldIdx, const String& str,int& i){
	checkBackgroundColor(str, i);
//	str.stb(i);
	if(str.cmpKWord("IS",i)){
		String name=str.getWord(i);
		ProtoInterfacePtr protoInterface=g_protoDeclare->protoInterface;
		for(int j=0;j<protoInterface->size();j++){
			if(protoInterface->properties[j].name==name){
				Field* protoField=protoInterface->getField(j);
				Field* thisField=getField(fieldIdx);
				thisField->connect(this, protoInterface, protoField);
				//field.reseat(fieldIdx, g_protoDeclare->field[j]);
				break;
			}
		}
	}
	else {
		getField(fieldIdx)->read(str,i);
	}
}
/*
Route& getRoute(String &str,int &i){
	Route &route=*new Route;

	route.fromNode=&*getUSE(str, i);
	return route;
}
*/

//SFFloat *g_field;
///MField *g_toNode;
//M_Route* g_route;

S_Route::S_Route(const String &str,int &i){
	Node* fromNode = g_load_scene->getUSE(str, i);//should ref count perhaps
	if(str[i]!='.')throw fromNode;
	str.stb(++i);
	int fromFieldIdx=fromNode->getProperties().getIndex(str.getWord(i));
	str.stb(i);
	if(!str.cmpKWord("TO",i))throw fromNode;
	str.stb(i);

	toNode = g_load_scene->getUSE(str, i);
	str.stb(i);
	if(str[i]!='.')throw this;
	str.stb(++i);
	toFieldIdx=toNode->getProperties().getIndex(str.getWord(i));
	toFieldPtr=toNode->getField(toFieldIdx);

	*(fromNode->getField(fromFieldIdx))+=(*this);
}
S_Route::S_Route( Node* fromNode, const int fromFieldIdx, Node* i_toNode, const int i_toFieldIdx):
	toNode(i_toNode),
	toFieldIdx(i_toFieldIdx),
	toFieldPtr(i_toNode->getField(i_toFieldIdx))
{
	*(fromNode->getField(fromFieldIdx))+=(*this);
}
Color3f floatToColor(float f){
	return Color3f(f,f,f);
}

void lookForLegacyFields(Node* node, const String &str,int &i){
	switch (node->getType()){
	case SCENE:{
		Scene* scene=dynamic_cast<Scene*>(node);
		if(str.cmpKWord("outputX3DVRMLClasic",i)) getBool(str,i);
		else if(str.cmpKWord("colorOwnership",i)) getBool(str,i);
		else if(str.cmpKWord("sameScreenPosition",i)){g_transformCP->sameScreenPosition.refValue() = getBool(str,i);}
		else if(str.cmpKWord("wireframe",i)){
			if(getBool(str,i) == false){
				scene->mode.refValue() &= ~WIREFRAME_MODE;
			}
		}
		else if(str.cmpKWord("xRayEdit",i)){
			if(getBool(str,i)){
				scene->mode.refValue() |= XRAY_EDIT_MODE;
			}
		}
		else if(str.cmpKWord("openOutputFolder",i))getBool(str,i);
		else if(str.cmpKWord("outputGZipped",i)) getBool(str,i);
		else if(str.cmpKWord("outputVRML",i)) getBool(str,i);
		else if(str.cmpKWord("outputX3DVRMLClassic",i)) getBool(str,i);
		else if(str.cmpKWord("outputX3DXML",i)) getBool(str,i);
		else if(str.cmpKWord("outputObj",i)) getBool(str,i);
		else if(str.cmpKWord("outputSculpties",i)) getBool(str,i);
		else if(str.cmpKWord("outputPovRay",i)) getBool(str,i);
		else if(str.cmpKWord("outputPovRayMovie",i)) getBool(str,i);

		else if(str.cmpKWord("movieAVI",i)) getBool(str,i);
		else if(str.cmpKWord("movieMPG",i)) getBool(str,i);
		else if(str.cmpKWord("movieMP4",i)) getBool(str,i);
		else if(str.cmpKWord("movieFLV",i)) getBool(str,i);

		else if(str.cmpKWord("genHAnimNodes",i)) getBool(str,i);
		else if(str.cmpKWord("genIdxTriSetNodes",i)) getBool(str,i);
		else if(str.cmpKWord("genContactSBVT",i)) getBool(str,i);
		else if(str.cmpKWord("genNormals",i)) getBool(str,i);
		else if(str.cmpKWord("genColorIndex",i)) getBool(str,i);
		else if(str.cmpKWord("autoOutSegmentedParts",i)) getBool(str,i);
		else if(str.cmpKWord("roundCoordsOutput",i)) getInt(str,i);
		else if(str.cmpKWord("outScriptAsRoutes",i)) getBool(str,i);


		else throw node;
		return;
	}
	case SEAMLESS:{
		Seamless* smls=dynamic_cast<Seamless*>(node);
		if(str.cmpKWord("coordIndex",i))smls->coordIndex.getField(str,i);
		else if(str.cmpKWord("patchOwner",i))smls->patchOwner.getField(str,i);
		else if(str.cmpKWord("link",i)){
			smls->readLink(str,i);
			smls->link.length = 0;
		}
		else if(str.cmpKWord("templateOutput",i))String().getField(str,i);
		else if(str.cmpKWord("partTreeFieldName",i))String().getField(str,i);
		else if(str.cmpKWord("translatePartFieldName",i))String().getField(str,i);
		else if(str.cmpKWord("outputFile",i))String().getField(str,i);
		else if(str.cmpKWord("children",i))smls->skeleton.refValue().getField(str,i);
		else if(str.cmpKWord("write",i))getBool(str,i);
		else if(str.cmpKWord("multiCage",i)) getBool(str,i);
		else if(str.cmpKWord("outputSingleIfs",i))getBool(str,i);
		else if(str.cmpKWord("clearCycle",i))getBool(str,i);
		else if(str.cmpKWord("geometry",i))smls->geometry = dynamic_cast<IndexedFaceSet*>(getNode2(str,i));
		else if(str.cmpKWord("appearance",i)){
			getNode2(str,i);
			smls->effect=new ColorEffect;
		}
		else if(str.cmpKWord("mirrorGroup",i)){
			getNode2(str,i);
		}
		else throw node;
		return;
	}
	case IMMEDIATEPART:
	case PART: {
		Art* part=dynamic_cast<Art*>(node);
		if(str.cmpKWord("children",i))part->part.refValue().getField(str,i);
		else if(str.cmpKWord("isJoint",i))part->toPoly=getBool(str,i);//toPoly used for isJoint flag for static part conversion
		else if(str.cmpKWord("rc",i))part->rc.getField(str,i);
		else if(str.cmpKWord("color",i))part->color.getField(str,i);
		else if(str.cmpKWord("texCoord",i))part->texCoord.getField(str,i);
		else if(str.cmpKWord("show",i))part->show=getBool(str,i);
		else if(str.cmpKWord("coordLen",i))part->coordLen=getInt(str,i);
		else if(str.cmpKWord("isJoint",i))getInt(str,i);
		else if(str.cmpKWord("localDrag",i))Vec3f().getField(str,i);
		else if(str.cmpKWord("clearCycle",i))getBool(str,i);
		else throw node;
		return;
	}
	case NURBSPATCH:{
		NurbsPatch* patch=dynamic_cast<NurbsPatch*>(node);
		if(str.cmpKWord("mirrorTug",i)) patch->mirror.refValue()=getBool(str,i);
		else if(str.cmpKWord("backfaceTug",i)) getBool(str,i);
		else if(str.cmpKWord("sameViewPositionTug",i)) getBool(str,i);
		else if(str.cmpKWord("ringTug",i)) getBool(str,i);
		else if(str.cmpKWord("numOfRings",i)) getInt(str,i);
		else if(str.cmpKWord("xTug",i)) getBool(str,i);
		else if(str.cmpKWord("yTug",i)) getBool(str,i);
		else if(str.cmpKWord("zTug",i)) getBool(str,i);
		else if(str.cmpKWord("color",i)){
			M_Color3f color;
			color.getField(str,i);
		}
		else throw node;
		return;
	}
	case HAIR: {
		Hair* hair = dynamic_cast<Hair*>(node);
		if(str.cmpKWord("bendyLathe",i)){
			M_Node& m_strand = hair->strand.MFNode::refValue();
			m_strand.getField(str,i);
			for(int k = m_strand.length; --k>=0;){
				NurbsPatch* ctrlLine = dynamic_cast<NurbsPatch*>(m_strand[k].p);
				Strand* strand = new Strand;
				strand->point.refValue() = ctrlLine->point;
				m_strand[k] = strand;
			}
		}
		else if(str.cmpKWord("nurbsPatch",i)){
			M_Node& m_lock = hair->lock.refValue();
			m_lock.getField(str,i);
			for(int k = m_lock.length; --k>=0;){
				NurbsPatch* patch = dynamic_cast<NurbsPatch*>(m_lock[k].p);
				Lock* lock = new Lock;
				lock->point.refValue() = patch->point;
				lock->weight.refValue() = patch->weight;
				lock->uKnot.refValue() = patch->uKnot;
				lock->vKnot.refValue() = patch->vKnot;
				lock->order.refValue() = patch->order;
				lock->endSP.refValue() = patch->endSP;
				lock->uJoinEnds.refValue() = patch->uJoinEnds;
				lock->beginSP.refValue() = patch->beginSP;
				lock->dim.refValue() = patch->dim;
				m_lock[k] = lock;
			}
		}
		else throw node;
		return;
	}
	case CCLATHE:{
		CCLathe* cclathe=dynamic_cast<CCLathe*>(node);
		if(str.cmpKWord("master",i)) getBool(str,i);
		else throw node;
		return;
	}
	case BRANCHJOINNEGX:
	case BRANCHJOINPOSX:
	case BRANCHJOINNEGY:
	case BRANCHJOINPOSY:
	case BRANCHJOINNEGZ:
	case BRANCHJOINPOSZ:{
		BranchJoin* branchJoin=dynamic_cast<BranchJoin*>(node);
		if(str.cmpKWord("vertexLen",i)) branchJoin->numOf.refValue()=getInt(str,i);
		else throw node;
		return;
	}
	case ECSCALE:{
		ECScale* ecscale=dynamic_cast<ECScale*>(node);
		if(str.cmpKWord("fNegX",i)) ecscale->negX=1/getFloat(str,i);
		else if(str.cmpKWord("fPosX",i)) ecscale->posX=1/getFloat(str,i);
		else if(str.cmpKWord("fNegZ",i)) ecscale->negZ=1/getFloat(str,i);
		else if(str.cmpKWord("fPosZ",i)) ecscale->posZ=1/getFloat(str,i);
		else throw node;
		return;
	}
	case CCSCALE:{
		CCScale* ccscale=dynamic_cast<CCScale*>(node);
		if(str.cmpKWord("fNegX",i)) ccscale->negX=1/getFloat(str,i);
		else if(str.cmpKWord("fPosX",i)) ccscale->posX=1/getFloat(str,i);
		else if(str.cmpKWord("fNegZ",i)) ccscale->negZ=1/getFloat(str,i);
		else if(str.cmpKWord("fPosZ",i)) ccscale->posZ=1/getFloat(str,i);
		else throw node;
		return;
	}
	case CCRINGSCALE:{
		CCRingScale* ccRingScale=dynamic_cast<CCRingScale*>(node);
		if(str.cmpKWord("fNegX",i)) ccRingScale->negX=1/getFloat(str,i);
		else if(str.cmpKWord("fPosX",i)) ccRingScale->posX=1/getFloat(str,i);
		else if(str.cmpKWord("fNegZ",i)) ccRingScale->negZ=1/getFloat(str,i);
		else if(str.cmpKWord("fPosZ",i)) ccRingScale->posZ=1/getFloat(str,i);
		else throw node;
		return;
	}
	case STEM:{
		Stem* stem=dynamic_cast<Stem*>(node);
		if(str.cmpKWord("xPointCnt",i))stem->staves.refValue()=getInt(str,i)-1;
		else if(str.cmpKWord("yPointCnt",i))stem->bands.refValue().getField(str,i);
		else if(str.cmpKWord("joinEnds",i) || str.cmpKWord("xJoinEnds",i))stem->uWrap.refValue()=getBool(str,i);
		else throw node;
		return;
	}
	case LOWPASS:{
		LowPass* lowPass = dynamic_cast<LowPass*>(node);
		if(str.cmpKWord("amp",i)){
			lowPass->amp.refValue().getField(str,i);
			lowPass->pass.refValue() = lowPass->amp;
		}
		else throw node;
		return;
	}
	case SURFACEGENERATOR:{
		SurfaceGenerator* surfaceGenerator=dynamic_cast<SurfaceGenerator*>(node);
		if(str.cmpKWord("generators",i))surfaceGenerator->generator.refValue().getField(str,i);
		else throw node;
		return;
	}
	case JOINVERTEX:{
		JoinVertex* joinVertex=dynamic_cast<JoinVertex*>(node);
		if(str.cmpKWord("fromVertex",i))joinVertex->srcCoord.refValue().getField(str,i);
		else if(str.cmpKWord("toVertex",i))joinVertex->dstCoord.refValue().getField(str,i);
		else if(str.cmpKWord("breakPoly",i))getInt(str,i);
		else throw node;
		return;
	}
	case TRANSFERVERTEX:{
		TransferVertex* transferVertex=dynamic_cast<TransferVertex*>(node);
		if(str.cmpKWord("toVertex",i))transferVertex->numOf.refValue()=getFloat(str,i)-transferVertex->fromVertex;
		else throw node;
		return;
	}
	case TRANSFERBCYL:{
		TransferBCyl* transferBCyl=dynamic_cast<TransferBCyl*>(node);
		if(str.cmpKWord("distance",i))transferBCyl->height.refValue()=getFloat(str,i);
		else if(str.cmpKWord("baseRad",i))transferBCyl->radius.refValue()=getFloat(str,i);
		else throw node;
		return;
	}
	case CENTERJOINT:{
		CenterJoint* centerJoint=dynamic_cast<CenterJoint*>(node);
		if(str.cmpKWord("baseVertex",i))centerJoint->fromVertex.refValue()=getFloat(str,i)-centerJoint->fromVertex;
		else throw node;
		return;
	}
	case TAPEMEASURE:{
		TransferVertex* transferVertex=dynamic_cast<TransferVertex*>(node);
		if(str.cmpKWord("baseVertex",i))transferVertex->fromVertex.refValue()=getFloat(str,i)-transferVertex->fromVertex;
		else if(str.cmpKWord("staves",i))transferVertex->numOf.refValue()=getFloat(str,i)-transferVertex->fromVertex;
		else throw node;
		return;
	}
	case SWITCH:{
		Switch* _switch=dynamic_cast<Switch*>(node);
		if(str.cmpKWord("choice",i))_switch->children.refValue().getField(str,i);
		else throw node;
		return;
	}
	case LOD_TYPE:{
		LOD* lod=dynamic_cast<LOD*>(node);
		if(str.cmpKWord("level",i))lod->children.refValue().getField(str,i);
		else throw node;
		return;
	}
	case TIMESENSOR:{
		TimeSensor* timeSensor=dynamic_cast<TimeSensor*>(node);
		if(str.cmpKWord("freezeControlled",i)) getBool(str,i);
		else if(str.cmpKWord("samplePoseEnabled",i))getBool(str,i);
		else throw node;
		return;
	}
	case MORPHER:{
		Morpher* morpher=dynamic_cast<Morpher*>(node);
		if(str.cmpKWord("translation",i))Vec3f().getField(str,i);
		else throw node;
		return;
	}
	case SETTINGS:{
		if(str.cmpKWord("tugVertex0X",i)){
			getBool(str,i);
		}
		else if(str.cmpKWord("windowSize",i)){
			g_settings->win3dSize.refValue().getField(str,i);
		}
		else if(str.cmpKWord("backgroundColor",i))Color3f().getField(str,i);
		else if(str.cmpKWord("scene3DFontColor",i))Color3f().getField(str,i);
		else if(str.cmpKWord("perspective",i))getBool(str,i);
		else if(str.cmpKWord("shiftDragTranslate",i))getBool(str,i);
		else throw node;
		return;
	}
	case COLOREFFECT:{
		ColorEffect* colorEffect=dynamic_cast<ColorEffect*>(node);
		if(str.cmpKWord("ambientIntensity",i))colorEffect->ambient.refValue()=floatToColor(getFloat(str,i));
		else if(str.cmpKWord("diffuseColor",i))colorEffect->diffuse.refValue().getField(str,i);
		else if(str.cmpKWord("emissiveColor",i))colorEffect->emissive.refValue().getField(str,i);
		else if(str.cmpKWord("specularColor",i))colorEffect->specular.refValue().getField(str,i);
		else throw node;
		return;
	}
	case TEXTUREEFFECT:{
		TextureEffect* textureEffect=dynamic_cast<TextureEffect*>(node);
		if(str.cmpKWord("ambientIntensity",i))textureEffect->ambient.refValue()=floatToColor(getFloat(str,i));
		else if(str.cmpKWord("emissiveColor",i))textureEffect->emissive.refValue().getField(str,i);
		else if(str.cmpKWord("specularColor",i))textureEffect->specular.refValue().getField(str,i);
		else throw node;
		return;
	}
	case FUSESURFACE:{
		FuseSurface* fuseSurface=dynamic_cast<FuseSurface*>(node);
		if(str.cmpKWord("basePoint",i))fuseSurface->baseCoord.refValue().getField(str,i);
		else if(str.cmpKWord("multiply",i))fuseSurface->coordChange.refValue()=getFloat(str,i);
		else if(str.cmpKWord("enabled",i))fuseSurface->bond.refValue()=getBool(str,i);
		else if(str.cmpKWord("polyMapAid",i))fuseSurface->polyMapAid.refValue()=getBool(str,i);
		else if(str.cmpKWord("polyMapAidMirror",i))fuseSurface->polyMapAidMirror.refValue()=getBool(str,i);
		else throw node;
		return;
	}
	case POLYTONURBS:{
		PolyToNurbs* polyToNurbs=dynamic_cast<PolyToNurbs*>(node);
		if(str.cmpKWord("update",i))getBool(str,i);
		else if(str.cmpKWord("random",i)){
			M_Int temp;
			temp.getField(str,i);
		}
		else throw node;
		return;
	}
	case OSCILLATOR:{
		Oscillator* osc=dynamic_cast<Oscillator*>(node);
		if(str.cmpKWord("point",i))osc->note.refValue().getField(str,i);
		else if(str.cmpKWord("fPoint",i))osc->note.refValue().getField(str,i);
		else if(str.cmpKWord("weight",i))osc->fWeight.refValue().getField(str,i);
		else if(str.cmpKWord("knot",i))osc->fKnot.refValue().getField(str,i);
		else if(str.cmpKWord("timbre",i))osc->filter.refValue().getField(str,i);
		else if(str.cmpKWord("multiple",i)){
			M_Float multiple;
			multiple.getField(str,i);
		}
		else if(str.cmpKWord("amplitude",i)){
			M_Float amplitude;
			amplitude.getField(str,i);
		}
		else if(str.cmpKWord("phase",i)){
			M_Float phase;
			phase.getField(str,i);
		}
		else if(str.cmpKWord("harmonic",i)){
			M_Float multiple;
			multiple.getField(str,i);
		}
		else if(str.cmpKWord("envelope",i)){
			M_Float amplitude;
			amplitude.getField(str,i);
		}
		else throw node;
		return;
	}
	case LARYNX:{
		Larynx* larynx=dynamic_cast<Larynx*>(node);
		if(str.cmpKWord("amplitude",i)){
			M_Float amplitude;
			amplitude.getField(str,i);
		}
		else if(str.cmpKWord("phase",i)){
			M_Float phase;
			phase.getField(str,i);
		}
		else if(str.cmpKWord("multiple",i)){
			M_Float multiple;
			multiple.getField(str,i);
		}
		else throw node;
		return;
	}
	}
	throw node;
}


void Node::getFields(const String &str,int &i){
	str.stb(i);

	M_Properties &properties=g_lookupProtoInterface[getType()]->properties;

	while(str[i]!='}'){
		if(str.cmpKWord("ROUTE",i)){
			S_Route(str,i);
		}
		else {
			int fieldIdx = 0;
			for(; fieldIdx<getFLen(); fieldIdx++){
				if(str.cmpKWord(properties[fieldIdx].name,i)){
					getField(fieldIdx, str, i);
					break;
				}
			}
			if(fieldIdx == getFLen()){
				if(str.cmpKWord("void",i)){
					int bracketCntr = 0;
					str.stbCpp(i);
					decodeVoid(str, i, bracketCntr);
				}
				else{
					lookForLegacyFields(this, str, i);
				}
			}
		}
		str.stb(i);
	}
}
void Art::getFields(const String &str,int &i){// only needed to catch legacy VertexBank
	M_Node tempPart;
	//str.stb(i);
	checkBackgroundColor(str, i);

	M_Properties& properties=g_lookupProtoInterface[getType()]->properties;

	while(str[i]!='}'){
		if(str.cmpKWord("ROUTE",i)){
			S_Route(str,i);
		}
		else {
			int fieldIdx=0;
			for(; fieldIdx < getFLen(); fieldIdx++){
				if(smls->version<2.044&&str.cmpKWord("part",i)){
					tempPart.getField(str, i);
					for(int k=0; k<tempPart.length; k++){
						Node* node=tempPart[k];
						if(node->getType()==MORPHER){
							part.refValue()[part.length-1]->morpher.refValue()+=dynamic_cast<Morpher*>(node);
						}
						else{
							part.refValue()+=dynamic_cast<Art*>(node);
						}
					}
					break;
				}
				else if(str.cmpKWord(properties[fieldIdx].name,i)){
					Node::getField(fieldIdx, str, i);
					break;
				}
			}
			if(fieldIdx == getFLen()){
				lookForLegacyFields(this, str, i);
			}
		}
		//str.stb(i);
		checkBackgroundColor(str, i);
	}
}

void bubbleSortRise(M_Int& mint, M_Art& part){
	bool flag;
	do{
		flag=false;
		for(int i=0;i<mint.length-1;i++){

			if (mint[i]>mint[i+1]){

				int swap = mint[i+1];
				mint[i+1]=mint[i];
				mint[i]=swap;

				Art* swapPart = part[i+1];
				part[i+1]=part[i];
				part[i]=swapPart;

				flag=true;
			}
		}
	}while(flag);
}
inline void createMaskAndDeleteDeadVerterticesDP(M_Int& mask, Art*& part){
	M_Vec3f& coord=part->coord.refValue();
	M_Vec3f& normal=part->normal.refValue();

	mask.length=coord.length;
	{
		for(int i=mask.length; --i>=0; ){
			mask[i]=1;
		}
	}
	{
		M_Art& relevantPart=part->relevantPart;
		for(int k=relevantPart.length; --k>=0; ){
			M_Triangle& triangle=relevantPart[k]->triangle;
			for(int i=triangle.length*3; --i>=0;){
				if(triangle.refPart(i)==part){
					mask[triangle.refIndex(i)]=0;
				}
			}
		}
	}
	{
		int count=0;
		for(int accum=0, i=0; i<mask.length; i++ ){
			int& mi=mask[i];
			int h=mi;
			accum+=h;
			mi=i-accum;
			if(h==0){
				coord[mi]=coord[i];
				if(normal.length){
					normal[mi]=normal[i];
				}
				count++;
			}
		}
		coord.length=count;
		if(normal.length){
			normal.length=count;
		}
	}
}

inline void uvCreateMaskAndDeleteDeadVerterticesDP(M_Int& mask, Art*& part){
	M_Vec2f& uvCoord = part->uvCoord.refValue();
	M_Vec3f& coord = part->coord.refValue();
	M_NurbsPatch& patchOwner = part->patchOwner;

	mask.length = uvCoord.length;
	{
		for(int i = mask.length; --i>=0; ){
			mask[i] = 1;
		}
	}
	{
		M_Art& relevantPart = part->relevantPart;
		for(int k = relevantPart.length; --k>=0; ){
			M_Triangle& triangle = relevantPart[k]->triangle;
			for(int i = triangle.length * 3; --i >= 0;){
				if(triangle.refPart(i) == part){
					mask[triangle.refIndex(i)] = 0;
				}
			}
		}
	}
	{
		int count=0;
		for(int accum=0, i=0; i<mask.length; i++ ){
			int& mi=mask[i];
			int h=mi;
			accum+=h;
			mi=i-accum;
			if(h==0){
				coord[mi] = coord[i];
				uvCoord[mi] = uvCoord[i];
				patchOwner[mi] = patchOwner[i];
				count++;
			}
		}
		uvCoord.length = count;
		coord.length = count;
		patchOwner.length = count;
	}
}

void DelPoly::doBuild(){
	Art* part = this->part.refValue();
	const M_Int& poly = this->poly;
	if(poly.length==0 || part==0){
		return;
	}
	M_Triangle& triangle = part->triangle;
	int trilen = triangle.length;

//	part->write(part->name+String(".wrl"));

	static M_Art superior;
	superior.length = 0;
	for(int d = 0,s = 0,p = 0; s < trilen; ){
		if(poly[p] == s){
			S_Triangle& tri = triangle[s];
			addRelevantPart(superior, tri.pa);
			addRelevantPart(superior, tri.pb);
			addRelevantPart(superior, tri.pc);
			p++;
			s++;
			if(p>=poly.length){
				while(s<trilen){
					S_Triangle& tri=triangle[s];
					addRelevantPart(superior, tri.pa);
					addRelevantPart(superior, tri.pb);
					addRelevantPart(superior, tri.pc);
					triangle[d] = tri;
					s++;
					d++;
				}
				break;
			}
		}
		else{
			triangle[d++] = triangle[s++];
		}
	}
	triangle.length -= poly.length;

//	part->write(part->name+String(".wrl"));
	if(part->smls->bnspe){
		for(int k = superior.length; --k >= 0; ){
			Art* part = superior[k];
			M_Int& mask = part->partTree;
			uvCreateMaskAndDeleteDeadVerterticesDP(mask, part);
			referenceNoDeadVertices(part, mask);
		}
	}
	else{
		for(int k = superior.length; --k >= 0; ){
			Art* part = superior[k];
			M_Int& mask = part->partTree;
			createMaskAndDeleteDeadVerterticesDP(mask, part);
			referenceNoDeadVertices(part, mask);
	//		removeNonRelevantParts(part);

	//		part->write(part->name+String(".wrl"));

		}
	}



	removeNonRelevantParts(part);


}
void moveToAlignNegX(Art* part){
	M_Triangle& triangle=part->triangle;
	for(int i=triangle.length*3; --i>=0;){
		Art*p=triangle.refPart(i);
		M_Vec3f& coord=p->coord.refValue();
		int index=triangle.refIndex(i);
		Vec3f vec=p->matrix*coord[index];
		if(vec.x<0){
			vec.x=0;
			coord[index]=p->imatrix*vec;
		}
	}
}
void moveToAlignPosX(Art* part){
	M_Triangle& triangle=part->triangle;
	for(int i=triangle.length*3; --i>=0;){
		if(triangle.refPart(i)==part){
			M_Vec3f& coord=part->coord.refValue();
			int index=triangle.refIndex(i);
			if(coord[index].x>0){
				coord[index].x=0;
			}
		}
	}
}
void moveToAlignNegY(Art* part){
	M_Vec3f& coord=part->coord.refValue();
	for(int i=coord.length; --i>=0;){
		if(coord[i].y<0)coord[i].y=0;
	}
}
void moveToAlignPosY(Art* part){
	M_Vec3f& coord=part->coord.refValue();
	for(int i=coord.length; --i>=0;){
		if(coord[i].y>0)coord[i].y=0;
	}
}
void moveToAlignNegZ(Art* part){
	M_Vec3f& coord=part->coord.refValue();
	for(int i=coord.length; --i>=0;){
		if(coord[i].z<0)coord[i].z=0;
	}
}
void moveToAlignPosZ(Art* part){
	M_Vec3f& coord=part->coord.refValue();
	for(int i=coord.length; --i>=0;){
		if(coord[i].z>0)coord[i].z=0;
	}
}
void DelPlane::doBuild(){
	M_SArt& part=this->part.refValue();
	if(part.length==0)return;
	Seamless* smls=dynamic_cast<Art*>(part[0].getRawPtr())->smls;

/*	{
		Matrix parentMatrix=translationMatrix(location);
		for(int k=0; k<smls->skeleton.length;k++){
			buildSceneTranslationMatrices( smls->skeleton.refValue()[k], parentMatrix);
		}
	}
*/
	smls->catIMatrix(translationMatrix(-location)*rotationMatrix(-orientation));
	if(full){
		for(int k=part.length;--k>=0;){
			Art* prt=dynamic_cast<Art*>(part[k].getRawPtr());
			if(negX){
				prt->cutLeftFull(0);
			}
			else if(posX){
				prt->cutRight(0);
			}
			else if(negY){
				prt->cutDown(0);
			}
			else if(posY){
				prt->cutUp(0);
			}
			else if(negZ){
				prt->cutForward(0);
			}
			else if(posZ){
				prt->cutBackward(0);
			}
		}
	}
	else
	{
		{
			Matrix parentMatrix=translationMatrix(-location);
			for(int k=0; k<smls->skeleton.length;k++){
				smls->skeleton.refValue()[k]->buildSceneITranslationMatrices(parentMatrix);
			}
		}
		for(int k=part.length;--k>=0;){
			Art* prt=dynamic_cast<Art*>(part[k].getRawPtr());
			if(negX){
				prt->cutLeft(0);
				if(align){
					moveToAlignNegX(prt);
				}
			}
			else if(posX){
				prt->cutRight(0);
				if(align){
					moveToAlignPosX(prt);
				}
			}
			else if(negY){
				prt->cutDown(0);
			}
			else if(posY){
				prt->cutUp(0);
			}
			else if(negZ){
				prt->cutForward(0);
			}
			else if(posZ){
				prt->cutBackward(0);
			}
		}
	}
	smls->catIMatrix();
}

void DelSphere::doBuild(){
	M_SArt& part=this->part.refValue();
	if(part.length==0)return;
	Seamless* smls=dynamic_cast<Art*>(part[0].getRawPtr())->smls;

	for(int k=part.length;--k>=0;){
		Art* prt=dynamic_cast<Art*>(part[k].getRawPtr());
		prt->cutSphere(location, radius);
	}
}

void Place::doBuild(){
	multiply.refValue()=Vec3f(alignX*-1, alignY*-1, alignZ*-1);

	M_SArt& part=this->part.refValue();
	if(!part.length)return;
	Art* prt=part[0];
	Seamless* smls=prt->smls;

	int index;
	Vec3f svec;
	if(staticTransform){
		if(prt->coord.length){
			const M_Vec3f& coord=prt->coord;
			Matrix matrix=scaleMatrix(scale)*rotationMatrix(rotation);
			index=tranVertex;
			if(static_cast<unsigned>(index)>=static_cast<unsigned>(coord.length)){
				index=coord.length-1;
			}
			svec = matrix * coord[index] * multiply.refValue();
		}
		Matrix matrix=scaleMatrix(scale)*rotationMatrix(rotation)*translationMatrix(translation+svec);
		for(int k=0;k<part.length;k++){
			Art* prt=part[k];
			if(descendants){
				staticTransformParts( prt, matrix);
			}
			else{
				staticTransformPartsNoDecendents(prt, matrix);
			}
		}
	}
	else{
		prt->translation=translation.refValue()+svec;
		prt->rotation=rotation.refValue();
		prt->scale=scale.refValue();
	}
}
void CenterJoint::doBuild(){
	if(part==0)return;

	Vec3f center;
	int i;

	Art* part=this->part.refValue();
	Seamless* smls=part->smls;

	const int numOf=this->numOf;
	int fromVertex=this->fromVertex;

	M_Vec3f& coord=part->coord.refValue();
	M_Vec3f* calcCoord=&coord;
	while(fromVertex<0){
		calcCoord=&smls->partList[part->plidx-1]->coord.refValue();
		fromVertex+=calcCoord->length;
	}
	if(fromVertex+numOf>calcCoord->length){
		dLog(String("Warning, bad values in CenterJoint for part")+part->name);
		return;
	}

	for(i=0;i<numOf;i++){
		center+=(*calcCoord)[fromVertex+i];
	}
	if(numOf){
		center/=(float)numOf;
	}
	center+=translation;

	for(i=coord.length; --i>=0; ){
		coord[i]-=center;
	}

	part->translation.refValue()+=center;

	M_SArt& children=part->part.refValue();
	if(dynamic_cast<ImmediatePart*>(part)){
		for(int k=children.length;--k>=0;){
			Art* bPart = children[k];
			bPart->translation.refValue()-=center;
		}
	}
	else{
		for(int k=children.length;--k>=0;){
			Art* bPart = children[k];
			if(dynamic_cast<Part*>(bPart)){
				bPart->translation.refValue()-=center;
			}
		}
	}
}
Art* Seamless::decToFindPart (int partIndex, int targetVertex) {
	Art* part;
	for(int k=partIndex; --k >= 0; ){
		part=partList[k];
		if( targetVertex >= part->fromCoord && targetVertex < part->toCoord){
			return part;
		}
	}
	throw;
}
Art* Seamless::incToFindPart (int partIndex, int targetVertex) {
	Art* part;
	for(int k=partIndex; ++k < partList.length; ){
		part=partList[k];
		if( targetVertex >= part->fromCoord && targetVertex < part->toCoord){
			return part;
		}
	}
	throw;
}
Art* Seamless::decToFindPartFromPoly (int partIndex, int targetPoly) {
	Art* part;
	for(int k=partIndex; --k >= 0; ){
		part=partList[k];
		if( targetPoly >= part->fromPoly && targetPoly < part->toPoly){
			return part;
		}
	}
	throw;
}
Art* Seamless::incToFindPartFromPoly (int partIndex, int targetPoly) {
	Art* part;
	for(int k=partIndex; ++k < partList.length; ){
		part=partList[k];
		if( targetPoly >= part->fromPoly && targetPoly < part->toPoly){
			return part;
		}
	}
	throw;
}

Art* Seamless::getPartFromVertex( int targetVertex, Art* part){

	if( targetVertex >= part->fromCoord && targetVertex < part->toCoord){
		return part;
	}
	else {
		if( targetVertex < part->fromCoord){
			return decToFindPart(part->plidx, targetVertex);
		}
		else{
			return incToFindPart(part->plidx, targetVertex);
		}
	}
}
Art* Seamless::getPartFromPoly( int targetPoly, Art* part){

	if( targetPoly >= part->fromPoly && targetPoly < part->toPoly){
		return part;
	}
	else {
		if( targetPoly < part->fromPoly){
			return decToFindPartFromPoly(part->plidx, targetPoly);
		}
		else{
			return incToFindPartFromPoly(part->plidx, targetPoly);
		}
	}
}

void Art::mirrorTriangle(int triIdx, int corner){
	S_Triangle& tri=triangle[triIdx];
	if(tri.getOwner()!=this){
		throw this;
	}

	coord.refValue()+=Vec3f();
	Vec3f& newCoord=coord.refValue()[coord.length-1];


	int a=corner;  //abc=triangle we r mirroring;
	int b=(a+1)%3;
	int c=(b+1)%3;

	int ia=tri.refIndex(a);
	int ib=tri.refIndex(b);
	int ic=tri.refIndex(c);

	Art* pa=tri.refPart(a);
	Art* pb=tri.refPart(b);
	Art* pc=tri.refPart(c);

	if(color.length){
		color+=pa->color[ia];
	}

	Vec3f vecA=pa->matrix*pa->coord.refValue()[ia];
	Vec3f vecB=pb->matrix*pb->coord.refValue()[ib];
	Vec3f vecC=pc->matrix*pc->coord.refValue()[ic];

	newCoord = vecA + (vecB - vecC);

	newCoord=imatrix*newCoord;

	int newColIndex=0;
	if(workEffect->getType()==COLOREFFECT){
		ColorEffect* effect=dynamic_cast<ColorEffect*>(workEffect);
		M_Color3f& color=effect->color.refValue();
		newColIndex=color.length;

		Color3f colA=color[tri.refTexIndex(a)];
		Color3f colB=color[tri.refTexIndex(b)];

		Color3f newColor;
		newColor.r = colA.r - (colA.r - colB.r)/2.;
		newColor.g = colA.g - (colA.g - colB.g)/2.;
		newColor.b = colA.b - (colA.b - colB.b)/2.;
		color+=newColor;
	}
	else if(workEffect->getType()==TEXTUREEFFECT){
		TextureEffect* effect=dynamic_cast<TextureEffect*>(workEffect);
		M_Vec2f& texCoord=effect->texCoord.refValue();
		newColIndex=texCoord.length;
		Vec2f texA=texCoord[tri.refTexIndex(a)];;
		Vec2f texB=texCoord[tri.refTexIndex(b)];;

		Vec2f newTexCoord;
		newTexCoord.x = texA.x - (texA.x - texB.x)/2.;
		newTexCoord.y = texA.y - (texA.y - texB.y)/2.;
		texCoord+=newTexCoord;
	}
	else throw this;
	if(normal.length){
		normal.refValue()+=Vec3f();
	}

	triangle+=S_Triangle(pa, ia, tri.refTexIndex(a), this, coord.length-1, newColIndex, pb, ib, tri.refTexIndex(b));

}

void Art::splitTriangle(int triIdx, int corner){
	S_Triangle& tri=triangle[triIdx];
	if(tri.getOwner()!=this){
		throw this;
	}

	Vec3f newCoord;


	int a=corner;  //abc=triangle we r splitting;
	int b=(a+1)%3;
	int c=(b+1)%3;

	int ia=tri.refIndex(a);
	int ib=tri.refIndex(b);
	int ic=tri.refIndex(c);

	Art* pa=tri.refPart(a);
	Art* pb=tri.refPart(b);
	Art* pc=tri.refPart(c);

	int newCoordIndex=pb->coord.length;

	Vec3f vecA=pa->coord.refValue()[ia];
	Vec3f vecB=pb->coord.refValue()[ib];
	vecA=pa->matrix*vecA;
	vecB=pb->matrix*vecB;

	newCoord.x = vecA.x - (vecA.x - vecB.x)/2.;
	newCoord.y = vecA.y - (vecA.y - vecB.y)/2.;
	newCoord.z = vecA.z - (vecA.z - vecB.z)/2.;

	newCoord=pb->imatrix*newCoord;

	int newColIndex=0;
	if(workEffect->getType()==COLOREFFECT){
		ColorEffect* effect=dynamic_cast<ColorEffect*>(workEffect);
		M_Color3f& color=effect->color.refValue();
		newColIndex=color.length;

		Color3f colA=color[tri.refTexIndex(a)];
		Color3f colB=color[tri.refTexIndex(b)];

		Color3f newColor;
		newColor.r = colA.r - (colA.r - colB.r)/2.;
		newColor.g = colA.g - (colA.g - colB.g)/2.;
		newColor.b = colA.b - (colA.b - colB.b)/2.;
		color+=newColor;
	}
	else if(workEffect->getType()==TEXTUREEFFECT){
		TextureEffect* effect=dynamic_cast<TextureEffect*>(workEffect);
		M_Vec2f& texCoord=effect->texCoord.refValue();
		newColIndex=texCoord.length;
		Vec2f texA=texCoord[tri.refTexIndex(a)];;
		Vec2f texB=texCoord[tri.refTexIndex(b)];;

		Vec2f newTexCoord;
		newTexCoord.x = texA.x - (texA.x - texB.x)/2.;
		newTexCoord.y = texA.y - (texA.y - texB.y)/2.;
		texCoord+=newTexCoord;
	}
	else throw this;

	pb->coord.refValue()+=newCoord;

	if(normal.length){
		pb->normal.refValue()+=Vec3f();
	}

	int tb=tri.refTexIndex(b);
	tri.refPart(b)=pb;
	tri.refIndex(b)=newCoordIndex;
	tri.refTexIndex(b)=newColIndex;

	if(tri.getOwner()!=this){
		throw this;
	}
	S_Triangle newTri = S_Triangle(
		pb, newCoordIndex,newColIndex,
		pb, ib, tb,
		pc, ic, tri.refTexIndex(c)
	);

	Art* newTriOwner = newTri.getOwner();

	if(newTriOwner!=this){
		int i=0;
	}
	newTriOwner->triangle+=newTri; // tri can loose its reference to the memory after this instruction

	addRelevantPart(newTriOwner->relevantPart, newTriOwner);
}

void Art::uvSplitTriangle(int triIdx, int corner){
	S_Triangle& tri=triangle[triIdx];
	if(tri.getOwner()!=this){
		throw this;
	}

	Vec2f newCoord;


	int a=corner;  //abc=triangle we r splitting;
	int b=(a+1)%3;
	int c=(b+1)%3;

	int ia=tri.refIndex(a);
	int ib=tri.refIndex(b);
	int ic=tri.refIndex(c);

	Art* pa=tri.refPart(a);
	Art* pb=tri.refPart(b);
	Art* pc=tri.refPart(c);

	int newCoordIndex = pb->uvCoord.length;

	Vec2f vecA = pa->uvCoord.refValue()[ia];
	Vec2f vecB = pb->uvCoord.refValue()[ib];
	NurbsPatch* newPatchOwner = pb->patchOwner[ib];

	newCoord.x = vecA.x - (vecA.x - vecB.x)/2.;
	newCoord.y = vecA.y - (vecA.y - vecB.y)/2.;

	int newColIndex=0;
	if(workEffect->getType()==COLOREFFECT){
		ColorEffect* effect=dynamic_cast<ColorEffect*>(workEffect);
		M_Color3f& color=effect->color.refValue();
		newColIndex=color.length;

		Color3f colA=color[tri.refTexIndex(a)];
		Color3f colB=color[tri.refTexIndex(b)];

		Color3f newColor;
		newColor.r = colA.r - (colA.r - colB.r)/2.;
		newColor.g = colA.g - (colA.g - colB.g)/2.;
		newColor.b = colA.b - (colA.b - colB.b)/2.;
		color+=newColor;
	}
	else if(workEffect->getType()==TEXTUREEFFECT){
		TextureEffect* effect=dynamic_cast<TextureEffect*>(workEffect);
		M_Vec2f& texCoord=effect->texCoord.refValue();
		newColIndex=texCoord.length;
		Vec2f texA=texCoord[tri.refTexIndex(a)];;
		Vec2f texB=texCoord[tri.refTexIndex(b)];;

		Vec2f newTexCoord;
		newTexCoord.x = texA.x - (texA.x - texB.x)/2.;
		newTexCoord.y = texA.y - (texA.y - texB.y)/2.;
		texCoord+=newTexCoord;
	}
	else throw this;

	int dstIdx = pb->coord.refValue().length;
	pb->coord.refValue().length++;
	pb->uvCoord.refValue() += newCoord;
	pb->patchOwner += newPatchOwner;
	newPatchOwner->generate3(pb->coord.refValue(), dstIdx, newCoord);

	int tb=tri.refTexIndex(b);
	tri.refPart(b)=pb;
	tri.refIndex(b)=newCoordIndex;
	tri.refTexIndex(b)=newColIndex;

	if(tri.getOwner()!=this){
		throw this;
	}
	S_Triangle newTri = S_Triangle(
		pb, newCoordIndex,newColIndex,
		pb, ib, tb,
		pc, ic, tri.refTexIndex(c)
	);

	Art* newTriOwner = newTri.getOwner();

//	if(newTriOwner!=this){
//		throw this;
//	}
	newTriOwner->triangle+=newTri; // tri can loose its reference to the memory after this instruction

	addRelevantPart(newTriOwner->relevantPart, newTriOwner);
}



void Art::split2Triangles(int triIdx, int corner, Art* otherPart, int otherTriIdx, int otherCorner){
	S_Triangle& tri=triangle[triIdx];
	S_Triangle& otherTri=otherPart->triangle[otherTriIdx];
	if(tri.getOwner()!=this){
		throw this;
	}

	Vec3f newCoord;


	const int a=corner;  //abc=triangle we r splitting;
	const int b=(a+1)%3;
	const int c=(b+1)%3;
	const int a2=otherCorner;
	const int b2=(a2+1)%3;
	const int c2=(b2+1)%3;

	int ia=tri.refIndex(a);
	int ib=tri.refIndex(b);
	int ic=tri.refIndex(c);
	int ia2=otherTri.refIndex(a2);
	int ib2=otherTri.refIndex(b2);
	int ic2=otherTri.refIndex(c2);

	Art* pa=tri.refPart(a);
	Art* pb=tri.refPart(b);
	Art* pc=tri.refPart(c);
	Art* pa2=otherTri.refPart(a2);
	Art* pb2=otherTri.refPart(b2);
	Art* pc2=otherTri.refPart(c2);

	int newCoordIndex=pb->coord.length;

	Vec3f vecA=pa->coord.refValue()[ia];
	Vec3f vecB=pb->coord.refValue()[ib];
	vecA=pa->matrix*vecA;
	vecB=pb->matrix*vecB;

	newCoord.x = vecA.x - (vecA.x - vecB.x)/2.;
	newCoord.y = vecA.y - (vecA.y - vecB.y)/2.;
	newCoord.z = vecA.z - (vecA.z - vecB.z)/2.;

	newCoord=pb->imatrix*newCoord;

	int newColIndex=0;
	if(workEffect->getType()==COLOREFFECT){
		ColorEffect* effect=dynamic_cast<ColorEffect*>(workEffect);
		M_Color3f& color=effect->color.refValue();
		newColIndex=color.length;

		Color3f colA=color[tri.refTexIndex(a)];
		Color3f colB=color[tri.refTexIndex(b)];

		Color3f newColor;
		newColor.r = colA.r - (colA.r - colB.r)/2.;
		newColor.g = colA.g - (colA.g - colB.g)/2.;
		newColor.b = colA.b - (colA.b - colB.b)/2.;
		color+=newColor;
	}
	else if(workEffect->getType()==TEXTUREEFFECT){
		TextureEffect* effect=dynamic_cast<TextureEffect*>(workEffect);
		M_Vec2f& texCoord=effect->texCoord.refValue();
		newColIndex=texCoord.length;
		Vec2f texA=texCoord[tri.refTexIndex(a)];;
		Vec2f texB=texCoord[tri.refTexIndex(b)];;

		Vec2f newTexCoord;
		newTexCoord.x = texA.x - (texA.x - texB.x)/2.;
		newTexCoord.y = texA.y - (texA.y - texB.y)/2.;
		texCoord+=newTexCoord;
	}
	else throw this;

	pb->coord.refValue()+=newCoord;

	if(normal.length){
		pb->normal.refValue()+=Vec3f();
	}

	int tb=tri.refTexIndex(b);
	tri.refPart(b)=pb;
	tri.refIndex(b)=newCoordIndex;
	tri.refTexIndex(b)=newColIndex;

//	int tb2=otherTri.refTexIndex(b2);
	int tc2=otherTri.refTexIndex(c2);
	otherTri.refPart(c2)=pb;
	otherTri.refIndex(c2)=newCoordIndex;
	otherTri.refTexIndex(c2)=newColIndex;

	if(tri.getOwner()!=this || otherTri.getOwner()!=this){
//		throw this;
	}

	S_Triangle newTri=S_Triangle(
		pb, newCoordIndex, newColIndex,
		pb, ib, tb,
		pc, ic, tri.refTexIndex(c)
		);

	S_Triangle otherNewTri=S_Triangle(
		pb, newCoordIndex, newColIndex,
		pb2, ib2, otherTri.refTexIndex(b2),
		pc2, ic2, tc2
		);

	Art* newTriOwner = newTri.getOwner();
	Art* otherNewTriOwner = otherNewTri.getOwner();

	newTriOwner->triangle+=newTri;		// tri and otherTri can loose their reference to the memory after this instruction
	otherNewTriOwner->triangle+=otherNewTri;

	addRelevantPart(newTriOwner->relevantPart, newTriOwner);
	addRelevantPart(otherNewTriOwner->relevantPart, otherNewTriOwner);
}

void Art::uvSplit2Triangles(int triIdx, int corner, Art* otherPart, int otherTriIdx, int otherCorner){
	S_Triangle& tri=triangle[triIdx];
	S_Triangle& otherTri=otherPart->triangle[otherTriIdx];
	if(tri.getOwner()!=this){
		throw this;
	}

	Vec2f newCoord;


	const int a=corner;  //abc=triangle we r splitting;
	const int b=(a+1)%3;
	const int c=(b+1)%3;
	const int a2=otherCorner;
	const int b2=(a2+1)%3;
	const int c2=(b2+1)%3;

	int ia=tri.refIndex(a);
	int ib=tri.refIndex(b);
	int ic=tri.refIndex(c);
	int ia2=otherTri.refIndex(a2);
	int ib2=otherTri.refIndex(b2);
	int ic2=otherTri.refIndex(c2);

	Art* pa=tri.refPart(a);
	Art* pb=tri.refPart(b);
	Art* pc=tri.refPart(c);
	Art* pa2=otherTri.refPart(a2);
	Art* pb2=otherTri.refPart(b2);
	Art* pc2=otherTri.refPart(c2);

	int newCoordIndex=pb->uvCoord.length;

	Vec2f vecA = pa->uvCoord.refValue()[ia];
	Vec2f vecB = pb->uvCoord.refValue()[ib];
	NurbsPatch* newPatchOwner = pb->patchOwner[ib];

	newCoord.x = vecA.x - (vecA.x - vecB.x)/2.;
	newCoord.y = vecA.y - (vecA.y - vecB.y)/2.;


	int newColIndex=0;
	if(workEffect->getType()==COLOREFFECT){
		ColorEffect* effect=dynamic_cast<ColorEffect*>(workEffect);
		M_Color3f& color=effect->color.refValue();
		newColIndex=color.length;

		Color3f colA=color[tri.refTexIndex(a)];
		Color3f colB=color[tri.refTexIndex(b)];

		Color3f newColor;
		newColor.r = colA.r - (colA.r - colB.r)/2.;
		newColor.g = colA.g - (colA.g - colB.g)/2.;
		newColor.b = colA.b - (colA.b - colB.b)/2.;
		color+=newColor;
	}
	else if(workEffect->getType()==TEXTUREEFFECT){
		TextureEffect* effect=dynamic_cast<TextureEffect*>(workEffect);
		M_Vec2f& texCoord = effect->texCoord.refValue();
		newColIndex=texCoord.length;
		Vec2f texA=texCoord[tri.refTexIndex(a)];;
		Vec2f texB=texCoord[tri.refTexIndex(b)];;

		Vec2f newTexCoord;
		newTexCoord.x = texA.x - (texA.x - texB.x)/2.;
		newTexCoord.y = texA.y - (texA.y - texB.y)/2.;
		texCoord+=newTexCoord;
	}
	else throw this;

	int dstIdx = pb->coord.refValue().length;
	pb->coord.refValue().length++;
	pb->uvCoord.refValue() += newCoord;
	pb->patchOwner += newPatchOwner;
	newPatchOwner->generate3(pb->coord.refValue(), dstIdx, newCoord);

	if(normal.length){
		pb->normal.refValue()+=Vec3f();
	}

	int tb = tri.refTexIndex(b);
	tri.refPart(b) = pb;
	tri.refIndex(b) = newCoordIndex;
	tri.refTexIndex(b) = newColIndex;

//	int tb2=otherTri.refTexIndex(b2);
	int tc2=otherTri.refTexIndex(c2);
	otherTri.refPart(c2) = pb;
	otherTri.refIndex(c2) = newCoordIndex;
	otherTri.refTexIndex(c2) = newColIndex;

	if(tri.getOwner()!=this || otherTri.getOwner()!=this){
//		throw this;
	}

	S_Triangle newTri = S_Triangle(
		pb, newCoordIndex, newColIndex,
		pb, ib, tb,
		pc, ic, tri.refTexIndex(c)
		);

	S_Triangle otherNewTri=S_Triangle(
		pb, newCoordIndex, newColIndex,
		pb2, ib2, otherTri.refTexIndex(b2),
		pc2, ic2, tc2
		);

	Art* newTriOwner = newTri.getOwner();
	Art* otherNewTriOwner = otherNewTri.getOwner();

	newTriOwner->triangle+=newTri;		// tri and otherTri can loose their reference to the memory after this instruction
	otherNewTriOwner->triangle+=otherNewTri;

	addRelevantPart(newTriOwner->relevantPart, newTriOwner);
	addRelevantPart(otherNewTriOwner->relevantPart, otherNewTriOwner);
}


template<> inline void SFBool::toString(String& str){
	SFBoolToString(refValue(), str);
}
inline void SFColor::toString(String& str){refValue().toString(str);}
template<> inline void SFFloat::toString(String& str){str+=refValue();}
template<> inline void SFImage::toString(String&){throw this;}
template<> inline void SFInt32::toString(String& str){str+=refValue();}
inline void SFNode::toString(String& str){
	if(refValue()==0){
		str+="NULL";
	}
	else{
		refValue()->toString(str,false);
	}
}
inline void SFRotation::toString(String& str){refValue().toString(str);}
inline void SFString::toString(String& str){refValue().toString(str);}
template<> inline void SFTime::toString(String& str){str+=refValue();}
inline void SFVec2f::toString(String& str){refValue().toString(str);}
inline void SFVec3f::toString(String& str){refValue().toString(str);}

template<> inline void MFBool::toString(String& str){
	if(length==1){
		SFBoolToString((*this)[0], str);
	}
	else {
		str+=g_vrmlSquareOpenBracket;
		if(length > 0){
			SFBoolToString((*this)[0], str);
			for(int i=1;i<length;i++){
				str+=' ';
				SFBoolToString((*this)[i], str);
			}
		}
		str+=g_vrmlSquareCloseBracket;
	}
}

template<> inline void MFColor::toString(String& str){refValue().toString(str);}
template<> inline void MFFloat::toString(String& str){refValue().toString(str);}
template<> inline void MFImage::toString(String& str){refValue().toString(str);}
template<> inline void MFInt32::toString(String& str){refValue().toString(str);}
template<> inline void MFNode::toString(String& str){refValue().toMNodeString(str);}
template<> inline void MFRotation::toString(String& str){refValue().toString(str);}
template<> inline void MFString::toString(String& str){refValue().toString(str);}
template<> inline void MFTime::toString(String& str){refValue().toString(str);}
template<> inline void MFVec2f::toString(String& str){refValue().toString(str);}
template<> inline void MFVec3f::toString(String& str){refValue().toString(str);}



template<> inline int SFBool::size(){throw this;}
inline int SFColor::size(){throw this;}
template<> inline int SFFloat::size(){throw this;}
template<> inline int SFImage::size(){throw this;}
template<> inline int SFInt32::size(){throw this;}
inline int SFNode::size(){throw this;}
inline int SFRotation::size(){throw this;}
inline int SFString::size(){throw this;}
template<> inline int SFTime::size(){throw this;}
inline int SFVec2f::size(){throw this;}
inline int SFVec3f::size(){throw this;}

template<> inline int MFBool::size(){return length;}
template<> inline int MFColor::size(){return length;}
template<> inline int MFFloat::size(){return length;}
template<> inline int MFImage::size(){return length;}
template<> inline int MFInt32::size(){return length;}
template<> inline int MFNode::size(){return length;}
template<> inline int MFRotation::size(){return length;}
template<> inline int MFString::size(){return length;}
template<> inline int MFTime::size(){return length;}
template<> inline int MFVec2f::size(){return length;}
template<> inline int MFVec3f::size(){return length;}

template<> inline void SFBool::addValFromStr(const String& str){throw this;}
inline void SFColor::addValFromStr(const String& str){throw this;}
template<> inline void SFFloat::addValFromStr(const String& str){throw this;}
template<> inline void SFImage::addValFromStr(const String& str){throw this;}
template<> inline void SFInt32::addValFromStr(const String& str){throw this;}
inline void SFNode::addValFromStr(const String& str){throw this;}
inline void SFRotation::addValFromStr(const String& str){throw this;}
inline void SFString::addValFromStr(const String& str){throw this;}
template<> inline void SFTime::addValFromStr(const String& str){throw this;}
inline void SFVec2f::addValFromStr(const String& str){throw this;}
inline void SFVec3f::addValFromStr(const String& str){throw this;}

template<> inline void MFBool::addValFromStr(const String& str){M_Bool value;	int i=0;value.getField(str, i);	refValue()+=value;*this=this->refValue();}
template<> inline void MFColor::addValFromStr(const String& str){M_Color3f value;	int i=0;value.getField(str, i);	refValue()+=value;*this=this->refValue();}
template<> inline void MFFloat::addValFromStr(const String& str){M_Float value;	int i=0;value.getField(str, i);	refValue()+=value;*this=this->refValue();}
template<> inline void MFImage::addValFromStr(const String& str){throw this;}
template<> inline void MFInt32::addValFromStr(const String& str){M_Int value;	int i=0;value.getField(str, i);	refValue()+=value;*this=this->refValue();}
template<> inline void MFNode::addValFromStr(const String& str){M_Node value;		int i=0;value.getField(str, i);	refValue()+=value;*this=this->refValue();}
template<> inline void MFRotation::addValFromStr(const String& str){M_Rotation value;	int i=0;value.getField(str, i);	refValue()+=value;*this=this->refValue();}
template<> inline void MFString::addValFromStr(const String& str){M_String value;	int i=0;value.getField(str, i);	refValue()+=value;*this=this->refValue();}
template<> inline void MFTime::addValFromStr(const String& str){throw this;}
template<> inline void MFVec2f::addValFromStr(const String& str){M_Vec2f value;	int i=0;value.getField(str, i);	refValue()+=value;*this=this->refValue();}
template<> inline void MFVec3f::addValFromStr(const String& str){M_Vec3f value;	int i=0;value.getField(str, i);	refValue()+=value;*this=this->refValue();}

template<> inline void SFBool::addValSameAsLast(){throw this;}
inline void SFColor::addValSameAsLast(){throw this;}
template<> inline void SFFloat::addValSameAsLast(){throw this;}
template<> inline void SFImage::addValSameAsLast(){throw this;}
template<> inline void SFInt32::addValSameAsLast(){throw this;}
inline void SFNode::addValSameAsLast(){throw this;}
inline void SFRotation::addValSameAsLast(){throw this;}
inline void SFString::addValSameAsLast(){throw this;}
template<> inline void SFTime::addValSameAsLast(){throw this;}
inline void SFVec2f::addValSameAsLast(){throw this;}
inline void SFVec3f::addValSameAsLast(){throw this;}

template<> inline void MFBool::addValSameAsLast(){	refValue()+=this->refValue()[length-1];*this=this->refValue();}
template<> inline void MFColor::addValSameAsLast(){	refValue()+=this->refValue()[length-1];*this=this->refValue();}
template<> inline void MFFloat::addValSameAsLast(){	refValue()+=this->refValue()[length-1];*this=this->refValue();}
template<> inline void MFImage::addValSameAsLast(){throw this;}
template<> inline void MFInt32::addValSameAsLast(){	refValue()+=this->refValue()[length-1];*this=this->refValue();}
template<> inline void MFNode::addValSameAsLast(){		refValue()+=this->refValue()[length-1];*this=this->refValue();}
template<> inline void MFRotation::addValSameAsLast(){	refValue()+=this->refValue()[length-1];*this=this->refValue();}
template<> inline void MFString::addValSameAsLast(){	refValue()+=this->refValue()[length-1];*this=this->refValue();}
template<> inline void MFTime::addValSameAsLast(){		refValue()+=this->refValue()[length-1];*this=this->refValue();}
template<> inline void MFVec2f::addValSameAsLast(){	refValue()+=this->refValue()[length-1];*this=this->refValue();}
template<> inline void MFVec3f::addValSameAsLast(){	refValue()+=this->refValue()[length-1];*this=this->refValue();}


template<> inline int SFBool::countItems(const String& str, int i){throw this;}
inline int SFColor::countItems(const String& str, int i){throw this;}
template<> inline int SFFloat::countItems(const String& str, int i){throw this;}
template<> inline int SFImage::countItems(const String& str, int i){throw this;}
template<> inline int SFInt32::countItems(const String& str, int i){throw this;}
inline int SFNode::countItems(const String& str, int i){throw this;}
inline int SFRotation::countItems(const String& str, int i){throw this;}
inline int SFString::countItems(const String& str, int i){throw this;}
template<> inline int SFTime::countItems(const String& str, int i){throw this;}
inline int SFVec2f::countItems(const String& str, int i){throw this;}
inline int SFVec3f::countItems(const String& str, int i){throw this;}

template<> inline int MFBool::countItems(const String& str, int i)	{str.stb(i);i++; M_Bool dummy; dummy.getField(str,i); return dummy.length;}
template<> inline int MFColor::countItems(const String& str, int i){str.stb(i);i++; M_Color3f dummy; dummy.getField(str,i); return dummy.length;}
template<> inline int MFFloat::countItems(const String& str, int i){str.stb(i);i++; M_Float dummy; dummy.getField(str,i); return dummy.length;}
template<> inline int MFImage::countItems(const String& str, int i){throw this;}
template<> inline int MFInt32::countItems(const String& str, int i){str.stb(i);i++; M_Int dummy; dummy.getField(str,i); return dummy.length;}
template<> inline int MFNode::countItems(const String& str, int i)	{str.stb(i);i++; M_Node dummy; dummy.getField(str,i); return dummy.length;}
template<> inline int MFRotation::countItems(const String& str, int i){str.stb(i);i++; M_Rotation dummy; dummy.getField(str,i); return dummy.length;}
template<> inline int MFString::countItems(const String& str, int i){str.stb(i);i++; M_String dummy; dummy.getField(str,i); return dummy.length;}
template<> inline int MFTime::countItems(const String& str, int i){throw this;}
template<> inline int MFVec2f::countItems(const String& str, int i){str.stb(i);i++; M_Vec2f dummy; dummy.getField(str,i); return dummy.length;}
template<> inline int MFVec3f::countItems(const String& str, int i){str.stb(i);i++; M_Vec3f dummy; dummy.getField(str,i); return dummy.length;}
/*
const M_Int& M_Int::operator =(const MFInt32& par){
	*this=par.getValue();
	return *this;
}
const M_Vec3f& M_Vec3f::operator =(const MFVec3f& par){
	*this=par.getValue();
	return *this;
}
*/
extern M_ProtoInterface g_lookupProtoInterface;

ProtoInterface* Node::getProtoInterface()const{
	return g_lookupProtoInterface[getType()];
}
const M_Properties& Node::getProperties()const{
	return getProtoInterface()->properties;
}

//inline NodePtr NodePtr::refValue(){return *refValue();};
//inline M_Node refValue(){return *refValue();};
/*
Field* Node::getField(const int i) const{
	return 0;
}
*/

//Rotation Rotation::slerp(const Rotation rot1, const float t)const;
//Adapted from Nick Thompson, Jim Kent and Daniel Woods code
//which is heavily based on code from Ken Shoemake and Gavin Bell

Rotation Rotation::slerp(const Rotation rot1, const float t)const{
	const Rotation& rot0=*this;
	Quaternion qrot1;
	rotationToQuaternion(rot1, qrot1);

	Quaternion qrot0;
	rotationToQuaternion(rot0, qrot0);

	float       rot1q[4];
	double      omega, cosom, sinom;
	double          scalerot0, scalerot1;
	int             i;

	// Calculate the cosine
	cosom = qrot0[0]*qrot1[0] + qrot0[1]*qrot1[1] + qrot0[2]*qrot1[2] + qrot0[3]*qrot1[3];

	// adjust signs if necessary
	if (cosom < 0.0 ){
		cosom = -cosom;
		for (int j = 0; j < 4; j++ ){
			rot1q[j] = -qrot1[j];
		}
	}
	else{
		for (int j = 0; j < 4; j++ ){
			rot1q[j] = qrot1[j];
		}
	}
	// calculate interpolating coeffs
	if ((1.0 - cosom) > 0.00001 ) {
		// standard case
		omega = acos(cosom);
		sinom = sin(omega);
		scalerot0 = sin((1.0 - t) * omega) / sinom;
		scalerot1 = sin(t * omega) / sinom;
	}
	else{
		// rot0 and rot1 very close - just do linear interp.
		scalerot0 = 1.0 - t;
		scalerot1 = t;
	}
	// build the new quarternion
	Quaternion qrot;
	for (i = 0; i <4; i++){
		qrot[i] = (float)(scalerot0 * qrot0[i] + scalerot1 * rot1q[i]);
	}
	Rotation	rot;
#ifdef DIRECT3D
	quaternionToRotation(qrot, rot);
#endif
	return rot;
}

void OrientationInterpolator::do_set_fraction(){
	M_Float& key = this->key.refValue();
	M_Rotation& keyValue = this->keyValue.refValue();
	float frac = set_fraction.refValue();
	int i = 0;
	if(key.length < 2)goto here;
	for(; i < key.length - 1; i++){
		if(key[i] > frac){
			i--;
			break;
		}
	}
	if(i < 0){
here:
		if(key.length > 0){
			value_changed.sendOutRoute(keyValue[0]);
		}
		return;
	}

	if(i == key.length - 1) i = key.length-2;
	float diff = key[i + 1] - key[i];

	float interp = 1;
	if(diff > .0000001){
		interp = 1.0 / diff * (frac - key[i]);
	}
	if(interp < 0){
		interp = -interp;
	}
	//////////////////////////////
	value_changed.sendOutRoute(keyValue[i].slerp(keyValue[i + 1], interp));
}

void PositionInterpolator::do_set_fraction(){
	M_Float& key = this->key.refValue();
	M_Vec3f& keyValue = this->keyValue.refValue();
	float frac = set_fraction.refValue();
	int i = 0;
	if(key.length < 2)goto here;
	for(; i < key.length-1; i++){
		if(key[i] > frac){
			i--;
			break;
		}
	}
	if(i < 0){
here:
		if(key.length > 0){
			value_changed.sendOutRoute(keyValue[0]);
		}
		return;
	}
	if(i == key.length-1)i = key.length - 2;
	float diff = key[i + 1] - key[i];

	float interp = 1;
	if(diff > .0000001){
		interp = 1.0 / diff * (frac - key[i]);
	}
	if(interp < 0){
		interp = -interp;
	}

	//////////////////////////////
	value_changed.sendOutRoute(keyValue[i] + (keyValue[i + 1] - keyValue[i]) * interp);
}

void ScalarInterpolator::do_set_fraction(){
	M_Float& _key=key.refValue();
	M_Float& _keyValue=keyValue.refValue();
	float frac=set_fraction.refValue();
	int i=0;
	if(_key.length<2)goto here;
	for(; i<_key.length-1; i++){
		if(_key[i]>frac){
			i--;
			break;
		}
	}
	if(i<0){
here:
		if(_key.length>0){
			value_changed.sendOutRoute(_keyValue[0]);
		}
		return;
	}
	if(i==_key.length-1)i=_key.length-2;
	float diff=_key[i+1]-_key[i];

	float interp=1;
	if(diff>.0000001){
		interp=1.0/diff*(frac-_key[i]);
	}
	if(interp<0){
		interp=-interp;
	}

	//////////////////////////////
	value_changed.sendOutRoute(_keyValue[i]+(_keyValue[i+1]-_keyValue[i])*interp);
}

void CameraSelector::do_set_fraction(){
	M_Float& key = this->key.refValue();
	M_Node& keyValue = this->keyValue.refValue();
	float frac = set_fraction.refValue();
	int i = 0;
	if(key.length < 2)goto here;
	for(; i < key.length-1; i++){
		if(key[i] > frac){
			i--;
			break;
		}
	}
	if(i < 0){
here:
		if(key.length > 0){
			dynamic_cast<Viewpoint*>(keyValue[0].p)->live = true;
		}
		return;
	}
	if(i == key.length-1)i = key.length - 2;
	float diff = key[i + 1] - key[i];

	float interp = 1;
	if(diff > .0000001){
		interp = 1.0 / diff * (frac - key[i]);
	}
	if(interp < 0){
		interp = -interp;
	}

	//////////////////////////////
	//value_changed.sendOutRoute(keyValue[i] + (keyValue[i + 1] - keyValue[i]) * interp);
	dynamic_cast<Viewpoint*>(keyValue[i].p)->live = true;
}

void ColorInterpolator::do_set_fraction(){
	M_Float& _key=key.refValue();
	M_Color3f& _keyValue=keyValue.refValue();
	float frac=set_fraction.refValue();
	int i=0;
	if(_key.length<2)goto here;
	for(; i<_key.length-1; i++){
		if(_key[i]>frac){
			i--;
			break;
		}
	}
	if(i<0){
here:
		if(_key.length>0){
			value_changed.sendOutRoute(_keyValue[0]);
		}
		return;
	}
	if(i==_key.length-1)i=_key.length-2;
	float diff=_key[i+1]-_key[i];

	float interp=1.0/diff*(frac-_key[i]);
	if(interp<0){
		interp=-interp;
	}

	//////////////////////////////
	value_changed.sendOutRoute(_keyValue[i]+(_keyValue[i+1]-_keyValue[i])*interp);
}

void CPMorpher::do_frame(Art* part){
	M_Float& key = this->key.refValue();
	M_Vec3f& keyValue = this->keyValue.refValue();
	if(!key.length)return;

//	for(int k = controlPoint.length; --k >= 0;){
//	}
	int portionLen = cpIndex.length;

	float frac = frame;
	int i = 0;
	if(key.length < 2)goto here;
	for(; i<key.length-1; i++){
		if(key[i] > frac){
			i--;
			break;
		}
	}
	if(i < 0){
here:
		if(key.length > 0){
			//value_changed.sendOutRoute(keyValue[0]);
		}
		return;
	}
	if(i == key.length - 1){
		i = key.length - 2;
	}
	float diff = key[i + 1] - key[i];

	float interp = 1.0 / diff * (frac - key[i]);
	if(interp < 0){
		interp =- interp;
	}

	const M_ControlPoint& controlPoint = part->controlPoint;

	for(int j = 0; j < portionLen; j++){
		int h = i * portionLen + j;
		Vec3f result = keyValue[h] + (keyValue[(i + 1) * portionLen + j] - keyValue[h]) * interp;
		const ControlPoint& cp = controlPoint[cpIndex[j]];
		NurbsPatch* patch = cp.p;
		patch->point.refValue()[cp.i] += result;
	}
}

void CoordinateInterpolator::do_set_fraction(){
	M_Float& key = this->key.refValue();
	M_Vec3f& keyValue = this->keyValue.refValue();
	if(!key.length)return;

	int portionLen = keyValue.length / key.length;
	result.length = portionLen;

	float frac = set_fraction.refValue();
	int i = 0;
	if(key.length < 2)goto here;
	for(; i<key.length-1; i++){
		if(key[i] > frac){
			i--;
			break;
		}
	}
	if(i < 0){
here:
		if(key.length > 0){
			value_changed.sendOutRoute(keyValue[0]);
		}
		return;
	}
	if(i == key.length - 1){
		i = key.length - 2;
	}
	float diff = key[i + 1] - key[i];

	float interp = 1.0 / diff * (frac - key[i]);
	if(interp < 0){
		interp =- interp;
	}


	for(int j = 0; j < portionLen; j++){
		int h = i * portionLen + j;
		result[j] = keyValue[h] + (keyValue[(i + 1) * portionLen + j] - keyValue[h]) * interp;
	}

	//////////////////////////////
	value_changed.sendOutRoute(result);
}
void NormalInterpolator::do_set_fraction(){
	M_Float& _key=key.refValue();
	M_Vec3f& _keyValue=keyValue.refValue();

	int portionLen=_keyValue.length/_key.length;
	result.length=portionLen;

	float frac=set_fraction.refValue();
	int i=0;
	if(_key.length<2)goto here;
	for(; i<_key.length-1; i++){
		if(_key[i]>frac){
			i--;
			break;
		}
	}
	if(i<0){
here:
		if(_key.length>0){
			value_changed.sendOutRoute(_keyValue[0]);
		}
		return;
	}
	if(i==_key.length-1)i=_key.length-2;
	float diff=_key[i+1]-_key[i];

	float interp=1.0/diff*(frac-_key[i]);
	if(interp<0){
		interp=-interp;
	}


	for(int j=0; j<portionLen; j++){
		int h=i*portionLen+j;
		result[j]=(_keyValue[h]+(_keyValue[(i+1)*portionLen+j]-_keyValue[h])*interp).normalize();
	}

	//////////////////////////////
	value_changed.sendOutRoute(result);
}

void morpheCoords(float interp, const Art* a, const Art* b, Art* dest){
	const M_Vec3f& a_coord = a->coord;
	const M_Vec3f& b_coord = b->coord;
	M_Vec3f& dest_coord = dest->coord.refValue();
	for(int i = a_coord.length; --i>=0;){
		dest_coord[i] = a_coord[i]+(b_coord[i]-a_coord[i])*interp;
	}
	M_SArt& m_dest = dest->part.refValue();
	const M_SArt& m_a = a->part;
	const M_SArt& m_b = b->part;
	for(int k = 0; k < m_dest.length; k++){
		morpheCoords(interp, m_a[k], m_b[k], m_dest[k]);
	}
}

void ModelSequence::do_frame(){
	if(key.length != model.length){
		return;
	}
	int i=0;
	if(key.length<2)goto here;
	for(; i<key.length-1; i++){
		if(key[i]>frame){
			i--;
			break;
		}
	}
	if(i<0){
here:
		if(key.length>0){
			//value_changed.sendOutRoute(keyValue[0]);
		}
		return;
	}
	if(i==key.length-1)i=key.length-2;
	float diff=key[i+1]-key[i];

	float interp=1;
	if(diff>.0000001){
		interp=1.0/diff*(frame-key[i]);
	}
	if(interp<0){
		interp=-interp;
	}

	M_SArt& m_dest = parent->smls->skeleton.refValue();
	const M_SArt& m_a = dynamic_cast<Seamless*>(model[i].p)->skeleton;
	const M_SArt& m_b = dynamic_cast<Seamless*>(model[i+1].p)->skeleton;
	for(int k = 0; k < m_dest.length; k++){
		morpheCoords(interp, m_a[k], m_b[k], m_dest[k]);
	}
	parent->smls->updateAllVertexInBuffers();

	//////////////////////////////
	//value_changed.sendOutRoute(_keyValue[i]+(_keyValue[i+1]-_keyValue[i])*interp);
}

#ifdef RENDER
void renderMorphPart(const Art& part, const M_Vec3f& srcCoordA, const M_Vec3f& srcCoordB, float interp);
#endif

Rotation RotAccelerator::accumRotation(int targetIndex, float time){

	M_Float& accelerate=this->accelerate.refValue();
	M_Float& attack=this->attack.refValue();
	M_Float& sustain=this->sustain.refValue();
	M_Float& decay=this->decay.refValue();
	M_Float& delay=this->delay.refValue();
	M_Vec3f& axis=this->axis.refValue();
	Rotation init=this->init;

	Rotation accum=init;
	float remTime=time;
	for(int i=0; i<delay.length; i++){

		if(i==targetIndex)continue;

		float time=remTime-delay[i];
		if(time<=0){
			continue;
		}
		float _accelerate=accelerate[i];
		float _attack=attack[i];
		float _sustain=sustain[i];
		float _decay=decay[i];
		if(time<_attack){
			float angle=_accelerate*.5*(time*time);
			accum=accum*Rotation(axis[i].normalize(), angle);
			continue;
		}
		time-=_attack;
		if(time<_sustain){
			float attackDist=_accelerate*.5*(_attack*_attack);
			float attackVelocity=_accelerate*_attack;
			float angle=attackDist+attackVelocity*time;
			accum=accum*Rotation(axis[i].normalize(), angle);
			continue;
		}
		time-=_sustain;
		if(time>_decay){
			time=0;
		}
		else {
			time=_decay-time;
		}
		float attackDist=_accelerate*.5*(_attack*_attack);
		float attackVelocity=_accelerate*_attack;	//exactly same as attackDist?

		float decayAccel=attackVelocity/_decay;
		float sustainDist=attackDist+attackVelocity*_sustain;
		float decayDist=decayAccel*0.5*(_decay*_decay);
		float decayOffset=sustainDist+decayDist;

		float angle=decayOffset-decayAccel*0.5*(time*time);

		accum=accum*Rotation(axis[i].normalize(), angle);
	}
	return accum;
}

void RotAccelerator::do_set_fraction(){
	float fracMult=this->fracMult;

	M_Float& accelerate=this->accelerate.refValue();
	M_Float& attack=this->attack.refValue();
	M_Float& sustain=this->sustain.refValue();
	M_Float& decay=this->decay.refValue();
	M_Float& delay=this->delay.refValue();
	M_Vec3f& axis=this->axis.refValue();

	Rotation accum=init;
	float remTime=set_fraction.refValue()*fracMult;
	for(int i=0; i<delay.length; i++){
		float time=remTime-delay[i];
		if(time<=0){
			continue;
		}
		float _accelerate=accelerate[i];
		float _attack=attack[i];
		float _sustain=sustain[i];
		float _decay=decay[i];
		if(time<_attack){
			float angle=_accelerate*.5*(time*time);
			accum=accum*Rotation(axis[i].normalize(), angle);
			continue;
		}
		time-=_attack;
		if(time<_sustain){
			float attackDist=_accelerate*.5*(_attack*_attack);
			float attackVelocity=_accelerate*_attack;
			float angle=attackDist+attackVelocity*time;
			accum=accum*Rotation(axis[i].normalize(), angle);
			continue;
		}
		time-=_sustain;
		if(time>_decay){
			time=0;
		}
		else {
			time=_decay-time;
		}
		float attackDist=_accelerate*.5*(_attack*_attack);
		float attackVelocity=_accelerate*_attack;	//exactly same as attackDist?

		float decayAccel=attackVelocity/_decay;
		float sustainDist=attackDist+attackVelocity*_sustain;
		float decayDist=decayAccel*0.5*(_decay*_decay);
		float decayOffset=sustainDist+decayDist;

		float angle=decayOffset-decayAccel*0.5*(time*time);

		accum=accum*Rotation(axis[i].normalize(), angle);
	}

	//////////////////////////////
	value_changed.sendOutRoute(accum);
	fraction_changed.sendOutRoute(set_fraction);
}
void PosAccelerator::do_set_fraction(){
	float fracMult=this->fracMult;

	M_Float& accelerate=this->accelerate.refValue();
	M_Float& attack=this->attack.refValue();
	M_Float& sustain=this->sustain.refValue();
	M_Float& decay=this->decay.refValue();
	M_Float& delay=this->delay.refValue();
	M_Vec3f& direction=this->direction.refValue();

	Vec3f accum=init;
	float remTime=set_fraction.refValue()*fracMult;
	for(int i=0; i<delay.length; i++){
		float time=remTime-delay[i];
		if(time<=0){
			continue;
		}
		float _accelerate=accelerate[i];
		float _attack=attack[i];
		float _sustain=sustain[i];
		float _decay=decay[i];
		if(time<_attack){
			float angle=_accelerate*.5*(time*time);
			accum+=direction[i]*angle;
			continue;
		}
		time-=_attack;
		if(time<_sustain){
			float attackDist=_accelerate*.5*(_attack*_attack);
			float attackVelocity=_accelerate*_attack;
			float angle=attackDist+attackVelocity*time;
			accum+=direction[i]*angle;
			continue;
		}
		time-=_sustain;
		if(time>_decay){
			time=0;
		}
		else {
			time=_decay-time;
		}
		float attackDist=_accelerate*.5*(_attack*_attack);
		float attackVelocity=_accelerate*_attack;	//exactly same as attackDist?

		float decayAccel=attackVelocity/_decay;
		float sustainDist=attackDist+attackVelocity*_sustain;
		float decayDist=decayAccel*0.5*(_decay*_decay);
		float decayOffset=sustainDist+decayDist;

		float angle=decayOffset-decayAccel*0.5*(time*time);

		accum+=direction[i]*angle;
	}

	//////////////////////////////
	value_changed.sendOutRoute(accum);
	fraction_changed.sendOutRoute(set_fraction);
}
void FloatAccelerator::Set_fraction::operator=(const float& par){
	refValue()=par;

	float fracMult=home.fracMult;

	M_Float& accelerate=home.accelerate.refValue();
	M_Float& attack=home.attack.refValue();
	M_Float& sustain=home.sustain.refValue();
	M_Float& decay=home.decay.refValue();
	M_Float& delay=home.delay.refValue();

	float accum=home.init;
	float remTime=par*fracMult;
	for(int i=0; i<delay.length; i++){
		float time=remTime-delay[i];
		if(time<=0){
			continue;
		}
		float _accelerate=accelerate[i];
		float _attack=attack[i];
		float _sustain=sustain[i];
		float _decay=decay[i];
		if(time<_attack){
			float angle=_accelerate*.5*(time*time);
//			accum+=direction[i]*angle;
			accum+=angle;
			continue;
		}
		time-=_attack;
		if(time<_sustain){
			float attackDist=_accelerate*.5*(_attack*_attack);
			float attackVelocity=_accelerate*_attack;
			float angle=attackDist+attackVelocity*time;
//			accum+=direction[i]*angle;
			accum+=angle;
			continue;
		}
		time-=_sustain;
		if(time>_decay){
			time=0;
		}
		else {
			time=_decay-time;
		}
		float attackDist=_accelerate*.5*(_attack*_attack);
		float attackVelocity=_accelerate*_attack;	//exactly same as attackDist?

		float decayAccel=attackVelocity/_decay;
		float sustainDist=attackDist+attackVelocity*_sustain;
		float decayDist=decayAccel*0.5*(_decay*_decay);
		float decayOffset=sustainDist+decayDist;

		float angle=decayOffset-decayAccel*0.5*(time*time);

//		accum+=direction[i]*angle;
		accum+=angle;
	}

	//////////////////////////////
	home.value_changed.sendOutRoute(accum);
	home.fraction_changed.sendOutRoute(par);
}

void AnimByAnkle::Position::operator=(const Vec3f& par){
	refValue()=par;
	Vec3f val=par;
	if(hip==0)return;
	const SFVec3f hipTranslation=hip->translation;
	SFRotation& hipRotation=hip.refValue()->rotation;
	if(knee==0)return;
	SFRotation& kneeRotation=knee.refValue()->rotation;
	if(ankle==0)return;
	const SFVec3f ankleTranslation=ankle->translation;
	SFRotation& ankleRotation=ankle.refValue()->rotation;

	Vec3f location;
	if(pelvis==0)return;
	const Vec3f& pelvisTranslation=pelvis->translation;
	const Rotation& pelvisRotation=pelvis->rotation;

	location=hipTranslation;
	val=translationMatrix(-pelvisTranslation)*rotationMatrix(-pelvisRotation)*par;

	const float upperarmLen=fabs(ankleTranslation.y);
	float xDist=location.x-val.x;
	float yDist=fabs(location.y-val.y);
	float zDist=location.z-val.z;
	float distance=sqrt(xDist*xDist+yDist*yDist+zDist*zDist);
	float halfKneeAng=acos(distance*.5/upperarmLen);

	float xyDist=sqrt(xDist*xDist+yDist*yDist);
	float zHipAng=-asin(xDist/xyDist);
	float xAng=acos(zDist/sqrt(xyDist*xyDist+zDist*zDist));
	float xHipAng=-halfKneeAng-xAng+PI*.5;
	float xAnkleAng=-halfKneeAng+xAng-PI*.5;

	hipRotation=Rotation(1,0,0,xHipAng)*Rotation(0,0,1,zHipAng);
	kneeRotation=Rotation(1,0,0,halfKneeAng*2);
	if(outAnkle){
		ankleRotation=-pelvisRotation*Rotation(0,0,-1,zHipAng)*Rotation(1,0,0,xAnkleAng);
	}
}
void AnimByWrist::Position::operator=(const Vec3f& par){
	refValue()=par;
	ArtPtr sholder = this->sholder;
	ArtPtr elbow = this->elbow;
	ArtPtr wrist = this->wrist;
	const float upperarmLen=fabs(wrist->translation.refValue().x);
	const Vec3f& location=Vec3f();
	float xDist=location.x-par.x;
	float yDist=fabs(location.y-par.y);
	float zDist=location.z-par.z;
	float distance=sqrt(xDist*xDist+yDist*yDist+zDist*zDist);
	float halfKneeAng=acos(distance*.5/upperarmLen);

	float xyDist=sqrt(xDist*xDist+yDist*yDist);
	float zHipAng=-asin(xDist/xyDist);
	float xAng=acos(zDist/sqrt(xyDist*xyDist+zDist*zDist));
	float xHipAng=-halfKneeAng-xAng+PI*.5;

	sholder->rotation=Rotation(1,0,0,xHipAng)*Rotation(0,0,1,zHipAng);
	wrist->rotation=Rotation(1,0,0,halfKneeAng*2);
}
void RotFilter::InRot::operator=(const Rotation& par){
	refValue()=par;
	rf.outRot.sendOutRoute(Rotation((rf.axis*par.getAxis().normalize()),rf.angle*par.angle));
}
void RotMult::InA::operator=(const Rotation& par){
	refValue()=par;
	home.catenation.sendOutRoute(par*home.inB);
}
void RotMult::InB::operator=(const Rotation& par){
	refValue()=par;
	home.catenation.sendOutRoute(home.inA.getValue()*par);
}
void HAnimHumanoid::Skeleton::operator=(const M_Node& par){
	refValue()=par;
	home.processFields();
}
void HAnimJoint::Progenitor::operator=(const M_Node& par){
	refValue()=par;
	home.humanoid->processFields();
}
void HAnimJoint::Displacers::operator=(const M_Node& par){
	refValue()=par;
	home.humanoid->processFields();
}
void ColorEffect::copyColorData(Effect* srcEffect){
	ColorEffect* srcColorEffect=dynamic_cast<ColorEffect*>(srcEffect);
	color.refValue()+=srcColorEffect->color;
}
void MonoColorEffect::copyColorData(Effect* srcEffect){
	MonoColorEffect* srcColorEffect=dynamic_cast<MonoColorEffect*>(srcEffect);
	color.refValue()+=srcColorEffect->color;
}
void TextureEffect::copyColorData(Effect* srcEffect){
	TextureEffect* srcTextureEffect=dynamic_cast<TextureEffect*>(srcEffect);
	texCoord.refValue()+=srcTextureEffect->texCoord;
}
void Art::addIndexUp(int amount){
	for(int i=triangle.length*3; --i>=0;){
		triangle.refTexIndex(i)+=amount;
	}
	for(int k=part.length; --k>=0;){
		part.refValue()[k]->addIndexUp(amount);
	}
}
inline void copyEffectData( Art* partSrc, Effect* effectDest, bool& firstTime){
	static int destOldLen;
	//Effect* effectDest=dynamic_cast<Effect*>(effectDst);
	if(firstTime){
		firstTime=false;
		destOldLen=effectDest->getEffectLength();
		Effect* workEffect=partSrc->workEffect;
		if(workEffect){
			effectDest->copyColorData(workEffect);
		}
	}
	M_SArt& m_part=partSrc->part.refValue();
	for(int k=m_part.length; --k>=0;){
		m_part[k]->addIndexUp( destOldLen);
	}
}
inline void getSmlsFromM_APart(M_SArt& m_part, Seamless* smlsDst){
	bool firstTime=true;
	for(int k=m_part.length; --k>=0;){
		Art* partSrc=m_part[k];
		Seamless* smlsSrc=partSrc->smls;
		if(smlsSrc!=0 && smlsDst!=0 && smlsSrc!=smlsDst){//we must re refernece all the color values
			copyEffectData( partSrc, smlsDst->effect, firstTime);
		}
	}
}
void Seamless::Skeleton::operator=(const M_Node& par){
	refValue()=*(M_SArt*)&par;
	getSmlsFromM_APart(refValue(), &home);
	home.process();
}
void Seamless::Skeleton::operator+=(const NodePtr& par){
	refValue()+=dynamic_cast<Art*>(par.p);
	home.process();
}
void Seamless::Skeleton::operator+=(Node* par){
	refValue()+=dynamic_cast<Art*>(par);
	home.process();
}
void Art::PartChild::operator=(const M_Node& par){
	refValue()=*(M_SArt*)&par;
	Seamless* smls=home.smls;
	getSmlsFromM_APart(refValue(), smls);
	if(smls!=0){
		smls->process();
	}
}
void Art::PartChild::operator+=(const NodePtr& par){
	refValue()+=dynamic_cast<Art*>(par.p);
	Seamless* smls=home.smls;
	if(smls!=0){
		smls->process();
	}
}
void Art::PartChild::operator+=(Art* par){
	refValue()+=par;
	Seamless* smls=home.smls;
	if(smls!=0){
		smls->process();
	}
}
void Art::MFMorpher::operator=(const M_Node& par){
	refValue()=*(M_Morpher*)&par;
	Seamless* smls=home.smls;
	smls->process();
}
void Art::MFMorpher::operator+=(const NodePtr& par){
	refValue()+=dynamic_cast<Morpher*>(par.p);
	Seamless* smls=home.smls;
	smls->process();
}
S_Route& Node::getRoute(const int index){
	int i=0;
	int r=0;
	int indexCntr=0;
	for(i=0; i < getFLen(); i++){
		M_Route& route=getField(i)->refRoute();
		for(r=0; r<route.length; r++){
			if(indexCntr==index){
				return route[r];
			}
			indexCntr++;
		}
	}
	throw this;
	//return *(S_Route*)0;
}
void Node::deleteRoute(const int index){
	int i=0;
	int r=0;
	int j=0;
	int indexCntr=0;
	for(i=0; i < getFLen(); i++){
		M_Route& route=getField(i)->refRoute();
		for(r=0; r<route.length; r++){
			if(indexCntr==index){
				for(j=r; j<route.length-1;j++){
					route[j]=route[j+1];
				}
				route.length=route.length-1;
				return;
			}
			indexCntr++;
		}
	}
	throw this;
	//return *(S_Route*)0;
}
inline void removeAllRelevantParts(M_Art& partList){
	for(int k=0,j=0; k<partList.length; k++){
		partList[k]->relevantPart.length=0;
	}
}

void convertGeneratorTo028( SurfaceGenerator* surface){
	if(surface==0)return;
	M_Node& generator=surface->generator.refValue();
	for(int g=generator.length; --g>=0;){
		if(generator[g]->getType()==SINETUG){
			SineTug* sineTug=dynamic_cast<SineTug*>(generator[g].getRawPtr());
			M_Float& decay=sineTug->decay.refValue();
			for(int i=decay.length; --i>=0; ){
				decay[i]=0;
			}
		}
	}
}
void Seamless::convertTo028(){
	M_Build& build=this->build.refValue();
	M_Vec3f& coord=refPoint();
	if(&coord!=0&&coord.length!=0){
		removeAllRelevantParts(partList);
		M_Color3f& color=refColor();
		M_Int& coordIndex=refCoordIndex();
		for(int k=0,j=0; k<partList.length; k++){
			Art* part=partList[k];
			int coordLen=part->toCoord-part->fromCoord;
			int polyLen=part->toPoly-part->fromPoly;
			M_Vec3f& partCoord=part->coord.refValue();
			partCoord.length=coordLen;
			if(&color!=0){
				M_Color3f& partColor=part->color;
				partColor.length=coordLen;
				for(int i=0; i<coordLen; i++){
					partCoord[i]=coord[j];
					partColor[i]=color[j++];
				}
			}
			else{
				for(int i=0; i<coordLen; i++){
					partCoord[i]=coord[j++];
				}
			}
			M_Triangle& triangle=part->triangle;
			triangle.length=polyLen;
			for(int i=0; i<polyLen; i++){
				int v=coordIndex[(part->fromPoly+i)*4];
				if(v>=part->fromCoord && v<part->toCoord){
					triangle[i].ia=v-part->fromCoord;
					triangle[i].pa=part;
					addRelevantPart(part->relevantPart, part);
				}
				else {
					Art* outsidePart=getPartFromVertex(v, part);
					triangle[i].ia=v-outsidePart->fromCoord;
					triangle[i].pa=outsidePart;
					addRelevantPart(outsidePart->relevantPart, part);
				}

				v=coordIndex[(part->fromPoly+i)*4+1];
				if(v>=part->fromCoord && v<part->toCoord){
					triangle[i].ib=v-part->fromCoord;
					triangle[i].pb=part;
					addRelevantPart(part->relevantPart, part);
				}
				else {
					Art* outsidePart=getPartFromVertex(v, part);
					triangle[i].ib=v-outsidePart->fromCoord;
					triangle[i].pb=outsidePart;
					addRelevantPart(outsidePart->relevantPart, part);
				}

				v=coordIndex[(part->fromPoly+i)*4+2];
				if(v>=part->fromCoord && v<part->toCoord){
					triangle[i].ic=v-part->fromCoord;
					triangle[i].pc=part;
					addRelevantPart(part->relevantPart, part);
				}
				else {
					Art* outsidePart=getPartFromVertex(v, part);
					triangle[i].ic=v-outsidePart->fromCoord;
					triangle[i].pc=outsidePart;
					addRelevantPart(outsidePart->relevantPart, part);
				}
			}
		}
	}
	for(int b=build.length; --b>=0;){
		BranchJoin* branchJoin=dynamic_cast<BranchJoin*>(build[b].getRawPtr());
		if(branchJoin!=0){
			convertGeneratorTo028(branchJoin->surface);
		}
		if(build[b]->getType()==STEM){
			Stem* stem=dynamic_cast<Stem*>(build[b].p);
			M_Int& bands=stem->bands.refValue();
			for(int i=bands.length; --i>=0;){
				bands[i]-=1;
			}
			stem->uWrap.refValue()=false;
			Rotation& orientation=stem->orientation.refValue();
			if(orientation.getAxis()==Vec3f(1, 0, 0)&&fabs(orientation.angle-1.5708)<.001){
				orientation=Rotation(1, 0, 0, PI*.5);
				stem->insideOut=true;
			}
			convertGeneratorTo028( dynamic_cast<SurfaceGenerator*>(stem->surface.p));
		}
		else if(build[b]->getType()==JOINVERTEX){
			JoinVertex* joinVertex=dynamic_cast<JoinVertex*>(build[b].getRawPtr());
			M_Int& fromVertex=joinVertex->srcCoord.refValue();
			M_Int& toVertex=joinVertex->dstCoord.refValue();
			for(int i=fromVertex.length; --i>=0; ){
				for(int j=i; --j>=0; ){
					if(fromVertex[j]<=fromVertex[i]){
						fromVertex[i]++;
					}
					if(joinVertex->source.refValue()==0 || joinVertex->source.refValue()==joinVertex->part.refValue()){
						if(fromVertex[j]<=toVertex[i]){
							toVertex[i]++;
						}
					}
				}
			}
		}
		else if(build[b]->getType()==DELPOLY){
			DelPoly* delPoly=dynamic_cast<DelPoly*>(build[b].getRawPtr());
			M_Int& poly=delPoly->poly.refValue();
			for(int i=poly.length; --i>=0; ){
				for(int j=i; --j>=0; ){
					if(poly[j]<=poly[i]){
						poly[i]++;
					}
				}
			}
			bubbleSort(poly);
		}
	}
}
inline void setupPolyFromTo( Seamless* smls){
	M_Art& partList=smls->partList;
	int k=0;
	int u=0;
	for(k=0; k<partList.length; k++){
		Art* part=partList[k];
		part->fromPoly=u;
		part->toPoly=u+part->triangle.length;
	}
}
inline void transferVertexToLeft028(Art*& source, Art*& dest, int& srcFromVertex, int& srcToVertex, int numOfVertices){
	if(numOfVertices==0)return;

	M_Vec3f& dstCoord=dest->coord.refValue();
	M_Vec3f& srcCoord=source->coord.refValue();

	for(int k=source->relevantPart.length;--k>=0;){
		Art* part=source->relevantPart[k];
		M_Triangle& triangle=part->triangle;
		int o=dest->coord.length;
		int len=triangle.length*3;
		for(int i=0; i<len; i++ ){
			Art*& part= triangle.refPart(i);
			int& index= triangle.refIndex(i);
			if(part==source){
				if(index>=srcFromVertex && index<srcToVertex){
					part=dest;
					index+=o-srcFromVertex;
				}
				else if(index>=srcToVertex){
					index-=numOfVertices;
				}
			}
		}
	}
	int d=dstCoord.length;
	dstCoord.length += numOfVertices;
	int s = srcFromVertex;

	if(source->smls->bnspe){
		M_Vec2f& dstUvCoord = dest->uvCoord.refValue();
		M_Vec2f& srcUvCoord = source->uvCoord.refValue();

		M_NurbsPatch& dstPatchOwner = dest->patchOwner;
		M_NurbsPatch& srcPatchOwner = source->patchOwner;

		dstUvCoord.length += numOfVertices;
		dstPatchOwner.length += numOfVertices;

		for(int i = numOfVertices; --i>=0; s++,d++){
			dstCoord[d] = srcCoord[s];
			dstUvCoord[d] = srcUvCoord[s];
			dstPatchOwner[d] = srcPatchOwner[s];
		}
		for(int d = srcFromVertex, s = srcToVertex; s<srcCoord.length; s++,d++){
			srcUvCoord[d] = srcUvCoord[s];
			srcUvCoord[d] = srcUvCoord[s];
			srcPatchOwner[d] = srcPatchOwner[s];
		}
		srcCoord.length -= numOfVertices;
		srcUvCoord.length -= numOfVertices;
		srcPatchOwner.length -= numOfVertices;
	}
	else{
		for(int i=numOfVertices; --i>=0; s++,d++){
			dstCoord[d] = srcCoord[s];
		}
		for(int d = srcFromVertex, s = srcToVertex; s<srcCoord.length; s++,d++){
			srcCoord[d] = srcCoord[s];
		}
		srcCoord.length -= numOfVertices;
	}

	transferTrianglesToLeft(source, dest);

}
inline void transferVertexToRight028(Art*& source, Art*& dest, int& srcFromVertex, int& srcToVertex, int numOfVertices){
	if(numOfVertices==0)return;

	M_Vec3f& dstCoord=dest->coord.refValue();

	M_Vec3f& srcCoord=source->coord.refValue();

	for(int k=dest->relevantPart.length;--k>=0;){
		Art* part=dest->relevantPart[k];
		M_Triangle& triangle=part->triangle;
		int len=triangle.length*3;
		for(int i=0; i<len; i++ ){
			if(triangle.refPart(i)==dest){
				triangle.refIndex(i)+=numOfVertices;
			}
		}
	}
	for(int k=source->relevantPart.length;--k>=0;){
		Art* part=source->relevantPart[k];
		M_Triangle& triangle=part->triangle;
		int len=triangle.length*3;
		for(int i=0; i<len; i++ ){
			Art*& part= triangle.refPart(i);
			if(part==source){
				int& index= triangle.refIndex(i);
				if(index>=srcFromVertex && index<srcToVertex){
					part=dest;
					index-=srcFromVertex;
				}
				else if(index>=srcToVertex){
					index-=numOfVertices;
				}
			}
		}
	}

	int s=dstCoord.length;
	dstCoord.length+=numOfVertices;
	int d=dstCoord.length;
	while(--s>=0){
		--d;
		dstCoord[d]=dstCoord[s];
	}
	d=0;
	s=srcFromVertex;
	for(int i=numOfVertices; --i>=0; s++,d++){
		dstCoord[d]=srcCoord[s];
	}
	for(int d=srcFromVertex, s=srcToVertex; s<srcCoord.length; s++,d++){
		srcCoord[d]=srcCoord[s];
	}
	srcCoord.length-=numOfVertices;

	transferTrianglesToLeft(source, dest);

}
M_Int TransferBCyl::reOrderSrcLookup;
M_Int TransferBCyl::reOrderSrcLookup2;

inline void TransferBCyl::commonTransferVertexByIndex(Art* source, Art* dest, int& oldSrcCoordLen, M_Vec3f& srcCoord, int& indicesLen){
	for(int i=0, last=0; i<oldSrcCoordLen; i++){
		int rem=reOrderSrcLookup2[i];
		reOrderSrcLookup2[i]=last;
		srcCoord[last]=srcCoord[i];
		last+=rem;
	}
	srcCoord.length-=indicesLen;
	for(int k=source->relevantPart.length;--k>=0;){
		Art* part=source->relevantPart[k];
		M_Triangle& triangle=part->triangle;
		int len=triangle.length*3;
		for(int i=0; i<len; i++ ){
			Art*& part= triangle.refPart(i);
			if(part==source){
				int& index= triangle.refIndex(i);
				if(reOrderSrcLookup[index]!=-1){
					part=dest;
					index=reOrderSrcLookup[index];
				}
				else {
					index=reOrderSrcLookup2[index];
					srcCoord[index]=srcCoord[index];
				}
			}
		}
	}
	transferTrianglesToLeft(source, dest);
}
inline void TransferBCyl::transferVertexAppend(Art* source, Art* dest, M_Int& vertex){
	int indicesLen=vertex.length;
	if(indicesLen==0)return;

	M_Vec3f& dstCoord=dest->coord.refValue();
	M_Vec3f& srcCoord=source->coord.refValue();
	int oldDstCoordLen=dstCoord.length;
	int oldSrcCoordLen=srcCoord.length;
	dstCoord.length+=indicesLen;
	for(int i=indicesLen; --i>=0; ){
		dstCoord[oldDstCoordLen+i]=srcCoord[vertex[i]];
	}
	reOrderSrcLookup2.length=reOrderSrcLookup.length=srcCoord.length;
	for(int i=reOrderSrcLookup.length; --i>=0;){
		reOrderSrcLookup[i]=-1;
		reOrderSrcLookup2[i]=1;
	}
	for(int i=vertex.length; --i>=0;){
		reOrderSrcLookup[vertex[i]]=oldDstCoordLen+i;
		reOrderSrcLookup2[vertex[i]]=0;
	}
	commonTransferVertexByIndex(source, dest, oldSrcCoordLen, srcCoord, indicesLen);
}
inline void TransferBCyl::transferVertexInsert(Art* source, Art* dest, M_Int& vertex){
	int indicesLen=vertex.length;
	if(indicesLen==0)return;

	M_Vec3f& dstCoord=dest->coord.refValue();
	M_Vec3f& srcCoord=source->coord.refValue();
	int oldDstCoordLen=dstCoord.length;
	int oldSrcCoordLen=srcCoord.length;
	dstCoord.length+=indicesLen;
	for(int i=oldDstCoordLen, k=dstCoord.length; --i>=0; ){//make gap for dest
		dstCoord[--k]=dstCoord[i];
	}
	for(int k=dest->relevantPart.length;--k>=0;){
		Art* part=dest->relevantPart[k];
		M_Triangle& triangle=part->triangle;
		int len=triangle.length*3;
		for(int i=0; i<len; i++ ){
			if(triangle.refPart(i)==dest){
				triangle.refIndex(i)+=indicesLen;
			}
		}
	}
	for(int i=indicesLen; --i>=0; ){//add source to gap
		dstCoord[i]=srcCoord[vertex[i]];
	}
	reOrderSrcLookup2.length=reOrderSrcLookup.length=srcCoord.length;
	for(int i=reOrderSrcLookup.length; --i>=0;){
		reOrderSrcLookup[i]=-1;
		reOrderSrcLookup2[i]=1;
	}
	for(int i=vertex.length; --i>=0;){
		reOrderSrcLookup[vertex[i]]=i;
		reOrderSrcLookup2[vertex[i]]=0;
	}
	commonTransferVertexByIndex(source, dest, oldSrcCoordLen, srcCoord, indicesLen);
	int a=0;
}

inline bool TransferBCyl::bubbleSortIndexByPosition(M_Int& vertex){

	for(int i=vertex.length-1; --i>=0;){
		if(g_tmpVec3f[vertex[i]].y>g_tmpVec3f[vertex[i+1]].y){
			goto here;

			for( ;--i>=0;){
				if (g_tmpVec3f[vertex[i]].y>g_tmpVec3f[vertex[i+1]].y){
here:
					int swap = vertex[i+1];
					vertex[i+1]=vertex[i];
					vertex[i]=swap;
				}
				return true;
			}
		}
	}
	return false;
}
void TransferBCyl::doBuild(){
	Art* src=source.refValue();
	Art* dst=dest.refValue();
	M_Int& vertex=this->vertex;
	if(src==0||dst==0||src==dst)return;

	if(in){
		TransferVertex transferVertex;

		M_Vec3f& srcCoord=src->coord.refValue();
		float sizeX=radius;
		float sizeY=height;
		float msizeX=-sizeX;
		float msizeY=-sizeY;

		Matrix imatrix = translationMatrix(-location)*rotationMatrix(-orientation);

		g_tmpVec3f.length=srcCoord.length;
		vertex.length=0;
		{
			for(int i=g_tmpVec3f.length; --i>=0;){
				g_tmpVec3f[i]=imatrix*srcCoord[i];
			}
		}
		{
			for(int i=0; i<g_tmpVec3f.length; i++){
				Vec3f crd=g_tmpVec3f[i];
				if(	crd.x>=msizeX && crd.x<=sizeX &&
					crd.y>=0 && crd.y<=sizeY &&
					crd.z>=msizeX && crd.z<=sizeX &&
					fabs(sqrt(crd.x*crd.x+crd.z*crd.z))<=sizeX
					){
					vertex+=i;
				}
			}
		}
		if(indexPosition){
			while(bubbleSortIndexByPosition(vertex));
		}
	}
	if(group >= 0){
		M_Int& dstCoordGroup = dst->coordGroup;
		M_Float& dstWeight = dst->weight;
		int oldGroupLength=dstWeight.length;
		if(group >= oldGroupLength){
			dstWeight.length = dstCoordGroup.length = group + 1;
			for(int i=dstWeight.length; --i>=oldGroupLength;){
				dstWeight[i]=1;
			}
		}
		dstCoordGroup[group]=vertex.length;
	}
	if(insert){
		transferVertexInsert(source.refValue(), dest.refValue(), vertex);
	}
	else {
		transferVertexAppend(source.refValue(), dest.refValue(), vertex);
	}
	if(group < 0 && indexPosition && groupingHeight > 0){
		Matrix imatrix = translationMatrix(-location)*rotationMatrix(-orientation);
		M_Vec3f& coord=dst->coord.refValue();
		M_Int& coordGroup = dst->coordGroup.refValue();
		M_Float& weight=dst->weight.refValue();

		coordGroup.length = 0;
		weight.length=0;
		float granule = groupingHeight/(numOfGroups+1);

		float ght = 0;
		float wt = ght/groupingHeight;
		for(int i = 0; i < coord.length; i++){
			float y = (imatrix*coord[i]).y;
			if(y > ght || coordGroup.length == 0){
				ght += granule;
				wt = ght/groupingHeight;
				if(wt >= .999999){
					break;
				}
				weight += wt;
				coordGroup += 1;
			}
			else{
				coordGroup[coordGroup.length-1]++;
			}
		}
	}

/*
	{
		for(int i=0; i<g_tmpIndex.length; i++){
			transferVertex.fromVertex=g_tmpIndex[i];
			transferVertex.toVertex=transferVertex.fromVertex+1;
			transferVertex.source=source;
			transferVertex.dest=dest;
			transferVertex.doBuild();
		}
	}
	*/
}
void TransferVertex::doBuild(){

	Art* src=source.refValue();
	Art* dst=dest.refValue();
	if(src==0||dst==0||src==dst)return;

	int fromVertex=this->fromVertex*qty;
	if(beginSP && fromVertex > 0){
		fromVertex-=qty-1;
	}
	int numOf=this->numOf;
	if(numOf>=0){
		numOf*=qty;
		if(beginSP && fromVertex == 0){
			numOf-=qty-1;
		}
	}
	else{
		numOf = 0x7fffffff;
	}

	int maxNumOf=src->coord.length-fromVertex;
	if(maxNumOf<0){
		return;
	}

	if (numOf>maxNumOf){
		numOf=maxNumOf;
	}
	int toVertex=fromVertex+numOf;

	if(src->plidx<dst->plidx && dst->coord.length){
		transferVertexToRight028( src, dst, fromVertex, toVertex, numOf);
	}
	else{
		transferVertexToLeft028( src, dst, fromVertex, toVertex, numOf);
	}
}
bool catchSelCoord(Seamless* smls);
void TransferOwnership::doBuild(){
	if(g_joinAndTransferActive == false){
		return;
	}

	TransferVertex transferVertex;
	if(Art* src = dynamic_cast<Art*>(source.p)){
		transferVertex.source = src;
		transferVertex.beginSP = beginSP;
		int k = 0;
		for(; k < numOf.length; k++){
			if(k < dest.length){
				if(numOf[k]<0){
					numOf.refValue()[k]=0;
				}
				if(qty[k]<0){
					qty.refValue()[k]=0;
				}
				transferVertex.numOf = numOf[k]*qty[k];
				if(k==0 && beginSP && transferVertex.numOf>0){
					transferVertex.numOf.refValue()-=qty[k] - 1;
				}
				Art* dst = dest[k].p;
				if(src == dst){
					transferVertex.fromVertex.refValue() += transferVertex.numOf;
					transferVertex.beginSP=false;
					continue;
				}
				transferVertex.dest = dst;
				transferVertex.doBuild();
				transferVertex.beginSP=false;
			}
		}
		if(k < dest.length){
			Art* dst = dest[k].p;
			if(src != dst){
				transferVertex.dest = dst;
				transferVertex.numOf = -1;
				transferVertex.beginSP = false; //do we need this line?
				transferVertex.doBuild();
			}
		}
		Seamless* smls = src->smls;
		if(smls->atThisStage==0){
			if(catchSelCoord(smls)){
				smls->atThisStage=this;
			}
		}
	}

}
void SplitTri::doBuild(){

	Art* part=this->part.refValue();
	if(part==0)return;

	M_Int& corner=this->corner.refValue();
	M_Triangle& triangle=part->triangle;
	M_Vec3f& coord=part->coord.refValue();
	M_Vec3f& normal=part->normal.refValue();
	Effect* workEffect=part->workEffect;
	int newColIndex=0;

	if(workEffect->getType()==COLOREFFECT){
		ColorEffect* effect=dynamic_cast<ColorEffect*>(workEffect);
		M_Color3f& color=effect->color.refValue();
		for(int i=0; i<corner.length; i++){

			int cnr=corner[i];
			S_Triangle& tri=triangle[cnr/3];

			int newCoordIndex=coord.length;
			Vec3f newCoord;


			int a=cnr%3;  //abc=triangle we r splitting;
			int b=(a+1)%3;
			int c=(b+1)%3;

			int ia=tri.refIndex(a);
			int ib=tri.refIndex(b);
			int ic=tri.refIndex(c);

			Art* pa=tri.refPart(a);
			Art* pb=tri.refPart(b);
			Art* pc=tri.refPart(c);


			Vec3f vecA=pa->coord.refValue()[ia];
			Vec3f vecB=pb->coord.refValue()[ib];

			newCoord.x = vecA.x - (vecA.x - vecB.x)/2.;
			newCoord.y = vecA.y - (vecA.y - vecB.y)/2.;
			newCoord.z = vecA.z - (vecA.z - vecB.z)/2.;

			int newColIndex=0;
			newColIndex=color.length;

			Color3f colA=color[tri.refTexIndex(a)];
			Color3f colB=color[tri.refTexIndex(b)];

			color+=Color3f(
				colA.r - (colA.r - colB.r)/2.,
				colA.g - (colA.g - colB.g)/2.,
				colA.b - (colA.b - colB.b)/2.
				);

			coord+=newCoord;

			if(normal.length){
				normal+=Vec3f();
			}

			int tb=tri.refTexIndex(b);
			tri.refPart(b)=part;
			tri.refIndex(b)=newCoordIndex;
			tri.refTexIndex(b)=newColIndex;

			triangle+=S_Triangle(	// tri can loose its reference to the memory after this instruction
				part, newCoordIndex,newColIndex,
				pb, ib, tb,
				pc, ic, tri.refTexIndex(c)
			);

		}
	}
	else if(workEffect->getType()==TEXTUREEFFECT){
		TextureEffect* effect=dynamic_cast<TextureEffect*>(workEffect);
		M_Vec2f& texCoord=effect->texCoord.refValue();
		for(int i=0; i<corner.length; i++){

			int cnr=corner[i];
			S_Triangle& tri=triangle[cnr/3];

			int newCoordIndex=coord.length;
			Vec3f newCoord;


			int a=cnr%3;  //abc=triangle we r splitting;
			int b=(a+1)%3;
			int c=(b+1)%3;

			int ia=tri.refIndex(a);
			int ib=tri.refIndex(b);
			int ic=tri.refIndex(c);

			Art* pa=tri.refPart(a);
			Art* pb=tri.refPart(b);
			Art* pc=tri.refPart(c);


			Vec3f vecA=pa->coord.refValue()[ia];
			Vec3f vecB=pb->coord.refValue()[ib];

			newCoord.x = vecA.x - (vecA.x - vecB.x)/2.;
			newCoord.y = vecA.y - (vecA.y - vecB.y)/2.;
			newCoord.z = vecA.z - (vecA.z - vecB.z)/2.;

			int newColIndex=0;
			newColIndex=texCoord.length;

			newColIndex=texCoord.length;
			Vec2f texA=texCoord[tri.refTexIndex(a)];;
			Vec2f texB=texCoord[tri.refTexIndex(b)];;

			texCoord+=Vec2f(
				texA.x - (texA.x - texB.x)/2.,
				texA.y - (texA.y - texB.y)/2.
				);

			coord+=newCoord;

			if(normal.length){
				normal+=Vec3f();
			}

			int tb=tri.refTexIndex(b);
			tri.refPart(b)=part;
			tri.refIndex(b)=newCoordIndex;
			tri.refTexIndex(b)=newColIndex;

			triangle+=S_Triangle(	// tri can loose its reference to the memory after this instruction
				part, newCoordIndex,newColIndex,
				pb, ib, tb,
				pc, ic, tri.refTexIndex(c)
			);

		}
	}
	else throw this;

}
void changeTriOwner(Art* ownerBefore, const int triIndex, Art* ownerAfter){
	ownerAfter->triangle+=ownerBefore->triangle[triIndex];
	ownerBefore->triangle.deleteElement(triIndex);
	addRelevantPart(ownerAfter->relevantPart, ownerAfter);
}
void SwapTriDivide::doBuild(){

	Art* part=this->part.refValue();
	if(part==0)return;

	M_Int& corner=this->corner.refValue();
	M_Art& partList=part->smls->partList;

	for(int i=0; i<corner.length; ){
		int tri0=corner[i++];
		int tri1=corner[i++];

		Art* p0=getPartFromRelIndex(part, tri0, partList);
		Art* p1=getPartFromRelIndex(part, tri1, partList);

		M_Triangle& triangle0=p0->triangle;
		M_Triangle& triangle1=p1->triangle;

		int cnr0=tri0%3;
		int cnr1=tri1%3;

		tri0-=cnr0;
		tri1-=cnr1;

		int a0=tri0+cnr0;
		int a1=tri1+cnr1;

		int b0=tri0+(cnr0+2)%3;
		int b1=tri1+(cnr1+2)%3;


		const int elementTriIdx0=tri0/3;
		const int elementTriIdx1=tri1/3;

		Art* ownerBefore0 = triangle0[elementTriIdx0].getOwner();
		Art* ownerBefore1 = triangle1[elementTriIdx1].getOwner();

		triangle0.refPart(b0)=triangle1.refPart(a1);
		triangle0.refIndex(b0)=triangle1.refIndex(a1);
		triangle0.refTexIndex(b0)=triangle1.refTexIndex(a1);

		triangle1.refPart(b1)=triangle0.refPart(a0);
		triangle1.refIndex(b1)=triangle0.refIndex(a0);
		triangle1.refTexIndex(b1)=triangle0.refTexIndex(a0);

		Art* ownerAfter0 = triangle0[elementTriIdx0].getOwner();
		Art* ownerAfter1 = triangle1[elementTriIdx1].getOwner();

		if(elementTriIdx1 > elementTriIdx0){//do highest triangle in index order first to prevent posibility of messing up tri below
			if(ownerBefore1 != ownerAfter1){
				changeTriOwner(ownerBefore1, elementTriIdx1, ownerAfter1);
			}
			if(ownerBefore0 != ownerAfter0){
				changeTriOwner(ownerBefore0, elementTriIdx0, ownerAfter0);
			}
		}
		else{
			if(ownerBefore0 != ownerAfter0){
				changeTriOwner(ownerBefore0, elementTriIdx0, ownerAfter0);
			}
			if(ownerBefore1 != ownerAfter1){
				changeTriOwner(ownerBefore1, elementTriIdx1, ownerAfter1);
			}
		}
	}
}
void PlaceTexture::doBuild(){

	M_SArt& part=this->part.refValue();
	for(int k=0; k<part.length; k++){
		Art* prt=part[k];
		TextureEffect* effect=dynamic_cast<TextureEffect*>(prt->workEffect);
		if(effect==0)continue;
		M_Vec2f& texCoord=effect->texCoord.refValue();
	//	Vec2f translation=this->translation;
		Matrix3 matrix=scaleMatrix(scale)*rotationMatrix(rotation)*translationMatrix(translation);

		for(int i=prt->fromEffectCoord; i<prt->toEffectCoord; i++){
			texCoord[i]=matrix*texCoord[i];
		}
	}
}
void CopyToMorph::doBuild(){

	Art* part=this->part.refValue();
	if(part==0)return;
	Morpher* morpher=dynamic_cast<Morpher*>(this->morpher.p);
	if(morpher==0)return;
	M_Vec3f& srcCoord=part->coord;
	M_Vec3f& dstCoord=morpher->displacement;
	dstCoord.length=srcCoord.length;
	for(int i=srcCoord.length; --i>=0;){
		dstCoord[i]=srcCoord[i];
	}
}
void RelateMorpher::doBuild(){
	if(part==0)return;
	const M_Morpher& morpher=part->morpher;
	const M_Vec3f& coord=part->coord;
	for (int k=morpher.length; --k>=0;){
		M_Vec3f& displacement=morpher[k]->displacement.refValue();
		if(displacement.length==coord.length){
			for(int i=coord.length; --i>=0;){
				displacement[i]-=coord[i];
			}
		}
	}
}
void TapeMeasure::doBuild(){
	Art* part=this->part.refValue();

	if(part==0)return;
	M_Vec3f& coord=part->coord.refValue();
	int baseVertex=this->baseVertex.refValue();
	if(coord.length-baseVertex<1)return;
	int staves=this->staves.refValue()+1;
	if(staves+baseVertex>coord.length){
		staves=coord.length-baseVertex;
	}
	float measurement=this->measurement.refValue();
	measurement=0;
	Vec3f prevCoord=coord[0];
	for(int i=baseVertex+1; i<staves; i++){
		Vec3f curCoord=coord[i];
		float xDiff=curCoord.x-prevCoord.x;
		float yDiff=curCoord.y-prevCoord.y;
		float zDiff=curCoord.z-prevCoord.z;
		measurement+=sqrt(xDiff*xDiff+yDiff*yDiff+zDiff*zDiff);
		prevCoord=curCoord;
	}
	this->measurement=measurement;
}
int Node::getFieldIndex(const Field* field)const{
	for(int i = getFLen(); --i>=0;){
		const Field*guessField=getField(i);
		if(field == guessField){
			return i;
		}
	}
	throw;
}
int Node::getFieldIndex(const String& name){
	const M_Properties& properties=getProperties();
	for(int i=properties.length;--i>=0;){
		if(properties[i].name==name){
			return i;
		}
	}
	throw this;
}
/*
inline void Seeder::createDestNodes(M_Node& dest){

}
void Seeder::doBuild(){
	return;
}
*/
inline void ColorEffect::optimize(M_Art& partList){
	int targetColorIndex;
	Color3f targetColor;

	for(int k=0; k<partList.length; k++){
		Art* targetPart=partList[k];
		const int coordLen = targetPart->coord.length;

		for(int targetIndex=0; targetIndex<coordLen; targetIndex++){
			bool flag=false;
			for(int k=0; k<partList.length; k++){
				Art* part=partList[k];
				M_Triangle& triangle=part->triangle;
				for(int i=0; i<triangle.length*3; i++){
					if(triangle.refPart(i)==targetPart && triangle.refIndex(i)==targetIndex){
						if(flag){
							if(targetColor==color[triangle.refTexIndex(i)]){
								triangle.refTexIndex(i)=targetColorIndex;
							}
						}
						else{
							flag=true;
							targetColorIndex=triangle.refTexIndex(i);
							targetColor=color[targetColorIndex];
						}
					}
				}
			}
		}
	}
}
inline void MonoColorEffect::optimize(M_Art& partList){
	int targetColorIndex;
	Color3f targetColor;

	for(int k=0; k<partList.length; k++){
		Art* targetPart=partList[k];
		const int coordLen = targetPart->coord.length;

		for(int targetIndex=0; targetIndex<coordLen; targetIndex++){
			bool flag=false;
			for(int k=0; k<partList.length; k++){
				Art* part=partList[k];
				M_Triangle& triangle=part->triangle;
				for(int i=0; i<triangle.length*3; i++){
					if(triangle.refPart(i)==targetPart && triangle.refIndex(i)==targetIndex){
						if(flag){
							if(targetColor==color[triangle.refTexIndex(i)]){
								triangle.refTexIndex(i)=targetColorIndex;
							}
						}
						else{
							flag=true;
							targetColorIndex=triangle.refTexIndex(i);
							targetColor=color[targetColorIndex];
						}
					}
				}
			}
		}
	}
}
inline void TextureEffect::optimize(M_Art& partList){
	int targetTexCoordIndex;
	Vec2f targetTexCoord;

	for(int k=0; k<partList.length; k++){
		Art* targetPart=partList[k];
		const int coordLen = targetPart->coord.length;

		for(int targetIndex=0; targetIndex<coordLen; targetIndex++){
			bool flag=false;
			for(int k=0; k<partList.length; k++){
				Art* part=partList[k];
				M_Triangle& triangle=part->triangle;
				for(int i=0; i<triangle.length*3; i++){
					if(triangle.refPart(i)==targetPart && triangle.refIndex(i)==targetIndex){
						if(flag){
							if(targetTexCoord==texCoord[triangle.refTexIndex(i)]){
								triangle.refTexIndex(i)=targetTexCoordIndex;
							}
						}
						else{
							flag=true;
							targetTexCoordIndex=triangle.refTexIndex(i);
							targetTexCoord=texCoord[targetTexCoordIndex];
						}
					}
				}
			}
		}
	}
}
char nibleToHex(char par){
	char result;
	if(par>=0&&par<=9){
		result=par+0x30;
	}
	else{
		result=par-0xa+'a';
	}
	return result;
}

void makeNodeNameFromAddr(String& str, const Node* par){
	__int64 parI = reinterpret_cast<__int64>(par);
	for(int i = 16; --i >= 0; parI <<= 4){
		str += nibleToHex((parI&0xf000000000000000) >> 15 * 4);
	}
}

void Color3f::toStrCpp(String& str)const{
	str+="Color3f(";
	str+=r;
	str+=',';
	str+=g;
	str+=',';
	str+=b;
	str+=')';
}
void Rotation::toStrCpp(String& str)const{
	str+="Rotation(";
	str+=x;
	str+=',';
	str+=y;
	str+=',';
	str+=z;
	str+=',';
	str+=angle;
	str+=')';
}
void String::toStrCpp(String& str)const{
	str+="String(\"";
	str+=*this;
	str+="\")";
}
template<>
void Vec2f::toStrCpp(String& str)const{
	str+="Vec2f(";
	str+=x;
	str+=',';
	str+=y;
	str+=')';
}
template<>
void Vec3f::toStrCpp(String& str)const{
	str+="Vec3f(";
	str+=x;
	str+=',';
	str+=y;
	str+=',';
	str+=z;
	str+=')';
}
template<>
void SFBool::toStrCpp(String& str)const{
	if(refValue()){
		str+="true";
	}
	else{
		str+="false";
	}
}
void SFColor::toStrCpp(String& str)const{
	refValue().toStrCpp(str);
}
template<>
void SFFloat::toStrCpp(String& str)const{
	str+=refValue();
}
template<>
void SFImage::toStrCpp(String& str)const{
	throw;
}
template<>
void SFInt32::toStrCpp(String& str)const{
	str+=refValue();
}
void SFNode::toStrCpp(String& str)const{
	if(refValue()==0){
		str+=0;
	}
	else{
		throw;
	}
}
void SFRotation::toStrCpp(String& str)const{
	str+="Rotation(";
	Rotation val=refValue();
	str+=val.x;
	str+=',';
	str+=val.y;
	str+=',';
	str+=val.z;
	str+=',';
	str+=val.angle;
	str+=')';
}
void SFString::toStrCpp(String& str)const{
	str+='\"';
	str+=refValue();
	str+='\"';
}
template<>
void SFTime::toStrCpp(String& str)const{
	str+=refValue();
}
void SFVec2f::toStrCpp(String& str)const{
	refValue().toStrCpp(str);
}
void SFVec3f::toStrCpp(String& str)const{
	refValue().toStrCpp(str);
}
template<>
void MFBool::toStrCpp(String& str)const{
	str+="M_Bool(";
	for(int i=0; i<length; i++){
		str+=(*this)[i];
		if(i<length-1){
			str+=',';
		}
	}
	str+=')';
}
template<>
void MFColor::toStrCpp(String& str)const{
	str+="M_Color3f(";
	for(int i=0; i<length; i++){
		(*this)[i].toStrCpp(str);
		if(i<length-1){
			str+=',';
		}
	}
	str+=')';
}
template<>
void MFFloat::toStrCpp(String& str)const{
	str+="M_Float(";
	for(int i=0; i<length; i++){
		str+=(*this)[i];
		if(i<length-1){
			str+=',';
		}
	}
	str+=')';
}
template<>
void MFImage::toStrCpp(String&)const{
}
template<>
void MFInt32::toStrCpp(String& str)const{
	str+="M_Int(";
	for(int i=0; i<length; i++){
		str+=(*this)[i];
		if(i<length-1){
			str+=',';
		}
	}
	str+=')';
}
template<>
void MFNode::toStrCpp(String& str)const{
	str+="M_Node(";
	str+=')';
}
template<>
void MFRotation::toStrCpp(String& str)const{
	str+="M_Rotation(";
	for(int i=0; i<length; i++){
		(*this)[i].toStrCpp(str);
		if(i<length-1){
			str+=',';
		}
	}
	str+=')';
}
template<>
void MFString::toStrCpp(String& str)const{
	str+="M_String(";
	for(int i=0; i<length; i++){
		(*this)[i].toStrCpp(str);
		if(i<length-1){
			str+=',';
		}
	}
	str+=')';
}
template<>
void MFTime::toStrCpp(String& str)const{
	str+="M_Double(";
	for(int i=0; i<length; i++){
		str+=(*this)[i];
		if(i<length-1){
			str+=',';
		}
	}
	str+=')';
}
template<>
void MFVec2f::toStrCpp(String& str)const{
	str+="M_Vec2f(";
	for(int i=0; i<length; i++){
		(*this)[i].toStrCpp(str);
		if(i<length-1){
			str+=',';
		}
	}
	str+=')';
}
template<>
void MFVec3f::toStrCpp(String& str)const{
	str+="M_Vec3f(";
	for(int i=0; i<length; i++){
		(*this)[i].toStrCpp(str);
		if(i<length-1){
			str+=',';
		}
	}
	str+=')';
}
/*
void genNodeCode(int nodeIdx, Node* protoInterface, M_Properties& prop, String& str){
	int f;
	str+="class ";
	str+=nodeLookUpStr[nodeIdx];
	str+=" : public Node, public MField {\n\
	public:\n\
	int getType()const{return ";
	str+=nodeLookUpStr[nodeIdx].upperCase();
	str+=";}\n";

	str+="\tint getFLen()const{return length;};\n";
	str+="\tField* getField(const int i){if((unsigned)i>(unsigned)length)throw this; return pointer[i];}\n";
	str+="\tint getFieldGroupIndex(const int &i){return index[i];}\n";
	str+="\tvoid setFieldGroupIndex(const int &i, const int &val){index[i]=val;}\n\n";

	for(f=0; f<prop.length; f++){
		Field* field=protoInterface->getField(f);
		switch(field->getType()){
			break;
		case SFBOOL:
			str+="\tSFBool ";
			break;
		case SFCOLOR:
			str+="\tSFColor ";
			break;
		case SFFLOAT:
			str+="\tSFFloat ";
			break;
		case SFIMAGE:
			str+="\tSFImage ";
			break;
		case SFINT32:
			str+="\tSFInt32 ";
			break;
		case SFNODE:
			str+="\tSFNode ";
			break;
		case SFROTATION:
			str+="\tSFRotation ";
			break;
		case SFSTRING:
			str+="\tSFString ";
			break;
		case SFTIME:
			str+="\tSFTime ";
			break;
		case SFVEC2F:
			str+="\tSFVec2f ";
			break;
		case SFVEC3F:
			str+="\tSFVec3f ";
			break;
		case MFBOOL:
			str+="\tMFBool ";
			break;
		case MFCOLOR:
			str+="\tMFColor ";
			break;
		case MFFLOAT:
			str+="\tMFFloat ";
			break;
		case MFIMAGE:
			str+="\tMFImage ";
			break;
		case MFINT32:
			str+="\tMFInt32 ";
			break;
		case MFNODE:
			str+="\tMFNode ";
			break;
		case MFROTATION:
			str+="\tMFRotation ";
			break;
		case MFSTRING:
			str+="\tMFString ";
			break;
		case MFDOUBLE:
			str+="\tMFTime ";
			break;
		case MFVEC2F:
			str+="\tMFVec2f ";
			break;
		case MFVEC3F:
			str+="\tMFVec3f ";
			break;
		default:
			throw;
		}
		str+=prop[f].name;
		str+=";\n";
	}
	str+="\n\tField* fieldArray[";
	str+=prop.length;
	str+="];\n\n\t";

	str+=nodeLookUpStr[nodeIdx];

//	bool readyForComma;
	str+="():\n\t\tMField(\n\t\t\tfieldArray,\n";
	for(f=0; f<prop.length; f++){
		if(f>0)str+=",\n";
		str+="\t\t\t&";
		str+=prop[f].name;
	}
	str+="\n\t\t),\n";
	str+="\n";
	for(f=0; f<prop.length; f++){
		Field* field=protoInterface->getField(f);
		str+="\t\t";
		str+=prop[f].name;
		str+='(';
		static String typeStr;
		switch(field->getType()){
		case SFBOOL:
		case SFINT32:
		case SFFLOAT:
		case SFTIME:
			str+=protoInterface->getField(f)->toStrCpp();
			str+=',';
			break;
		case SFNODE:
			str+="0,";
			break;
		default:
			typeStr=field->getTypeStr();
			typeStr[1]='_';
			str+=typeStr;
			str+='(';
			str+=protoInterface->getField(f)->toStrCpp();
			str+="),";
		}

		if		(!field->isWritable()&&!field->isReadable()){
			str+="initializeOnly)";
		}
		else if	(!field->isWritable()&&field->isReadable()){
			str+="outputOnly)";
		}
		else if	(field->isWritable()&&!field->isReadable()){
			str+="inputOnly)";
		}
		else if	(field->isWritable()&&field->isReadable()){
			str+="inputOutput)";
		}
		if(f<prop.length-1){
			str+=',';
		}
		str+='\n';
	}
	str+="\n\t{\n";
	str+="\t}\n};\n";
}
void genNodeCodeCpp(M_ProtoInterface& mnode){
	String str;
	for(int k=0; k<mnode.length; k++){
		ProtoInterfacePtr proto=mnode[k];
		genNodeCode(k, mnode[k], proto->properties, str);
	}
	str.write("code.cpp");
}
*/
////////////////////////////////////////////////////////////////////////////////////////

template<>
void SFBool::toCPPString(String& str, const String& nodeName, Field *protoField, bool newList){
	if(refValue()!=dynamic_cast<SFBool*>(protoField)->refValue()){
		str+=nodeName;
		str+='=';
		toStrCpp(str);
	}
}
void SFColor::toCPPString(String& str, const String& nodeName, Field *protoField, bool newList){
	if(refValue()!=dynamic_cast<SFColor*>(protoField)->refValue()){
		str+=nodeName;
		toStrCpp(str);
	}
}
template<>
void SFFloat::toCPPString(String& str, const String& nodeName, Field *protoField, bool newList){
	if(refValue()!=dynamic_cast<SFFloat*>(protoField)->refValue()){
		str+=nodeName;
		str+='=';
		toStrCpp(str);
	}
}
template<>
void SFImage::toCPPString(String& str, const String& nodeName, Field *protoField, bool newList){
	throw this;
}
template<>
void SFInt32::toCPPString(String& str, const String& nodeName, Field *protoField, bool newList){
	if(refValue()!=dynamic_cast<SFInt32*>(protoField)->refValue()){
		str+=nodeName;
		str+='=';
		toStrCpp(str);
	}
}
void SFNode::toCPPString(String& str, const String& nodeName, Field *protoField, bool newList){
	if(refValue()!=dynamic_cast<SFNode*>(protoField)->refValue()){
		refValue()->toCPPString(str, newList);
		str+=nodeName;
		str+='=';
		toStrCpp(str);
	}
}
void SFRotation::toCPPString(String& str, const String& nodeName, Field *protoField, bool newList){
	if(refValue()!=dynamic_cast<SFRotation*>(protoField)->refValue()){
		str+=nodeName;
		str+='=';
		toStrCpp(str);
	}
}
void SFString::toCPPString(String& str, const String& nodeName, Field *protoField, bool newList){
	if(refValue()!=dynamic_cast<SFString*>(protoField)->refValue()){
		str+=nodeName;
		str+='=';
		toStrCpp(str);
	}
}
template<>
void SFTime::toCPPString(String& str, const String& nodeName, Field *protoField, bool newList){
	if(refValue()!=dynamic_cast<SFTime*>(protoField)->refValue()){
		str+=nodeName;
		str+='=';
		toStrCpp(str);
	}
}
void SFVec2f::toCPPString(String& str, const String& nodeName, Field *protoField, bool newList){
	if(refValue()!=dynamic_cast<SFVec2f*>(protoField)->refValue()){
		str+=nodeName;
		str+='=';
		toStrCpp(str);
	}
}
void SFVec3f::toCPPString(String& str, const String& nodeName, Field *protoField, bool newList){
	if(refValue()!=dynamic_cast<SFVec3f*>(protoField)->refValue()){
		str+=nodeName;
		str+='=';
		toStrCpp(str);
	}
}
template<>
void MFBool::toCPPString(String& str, const String& nodeName, Field *protoField, bool newList){
	if(refValue()!=dynamic_cast<MFBool*>(protoField)->refValue()){
		str+=nodeName;
		str+='=';
		toStrCpp(str);
	}
}
template<>
void MFColor::toCPPString(String& str, const String& nodeName, Field *protoField, bool newList){
	if(refValue()!=dynamic_cast<MFColor*>(protoField)->refValue()){
		str+=nodeName;
		str+="=M_Color3f(";
		toStrCpp(str);
		str+=')';
	}
}
template<>
void MFFloat::toCPPString(String& str, const String& nodeName, Field *protoField, bool newList){
	if(refValue()!=dynamic_cast<MFFloat*>(protoField)->refValue()){
		str+=nodeName;
		str+="=M_Float(";
		toStrCpp(str);
		str+=')';
	}
}
template<>
void MFImage::toCPPString(String& str, const String& nodeName, Field *protoField, bool newList){
	throw this;
}
template<>
void MFInt32::toCPPString(String& str, const String& nodeName, Field *protoField, bool newList){
	if(refValue()!=dynamic_cast<MFInt32*>(protoField)->refValue()){
		str+=nodeName;
		str+="=M_Int(";
		toStrCpp(str);
		str+=')';
	}
}
template<>
void MFNode::toCPPString(String& str, const String& nodeName, Field *protoField, bool newList){
	if(refValue()!=dynamic_cast<MFNode*>(protoField)->refValue()){
		for(int k=0; k<length; k++){
			NodePtr node=this->refValue()[k];
			node->toCPPString(str, newList);

			str+=nodeName;
			str+="+=";
			if(node->name!=""){
				str+=node->name;
			}
			else {
				str+='n';
				makeNodeNameFromAddr(str, node);
			}
			if(k<length-1){
				str+=";\n";
			}
		}
	}
}
template<>
void MFRotation::toCPPString(String& str, const String& nodeName, Field *protoField, bool newList){
	if(refValue()!=dynamic_cast<MFRotation*>(protoField)->refValue()){
		str+=nodeName;
		str+='=';
		toStrCpp(str);
	}
}
template<>
void MFString::toCPPString(String& str, const String& nodeName, Field *protoField, bool newList){
	if(refValue()!=dynamic_cast<MFString*>(protoField)->refValue()){
		str+=nodeName;
		str+='=';
		toStrCpp(str);
	}
}
template<>
void MFTime::toCPPString(String& str, const String& nodeName, Field *protoField, bool newList){
	if(refValue()!=dynamic_cast<MFTime*>(protoField)->refValue()){
		str+=nodeName;
		str+='=';
		toStrCpp(str);
	}
}
template<>
void MFVec2f::toCPPString(String& str, const String& nodeName, Field *protoField, bool newList){
	if(refValue()!=dynamic_cast<MFVec2f*>(protoField)->refValue()){
		str+=nodeName;
		str+='=';
		toStrCpp(str);
	}
}
template<>
void MFVec3f::toCPPString(String& str, const String& nodeName, Field *protoField, bool newList){
	if(refValue()!=dynamic_cast<MFVec3f*>(protoField)->refValue()){
		str+=nodeName;
		str+='=';
		toStrCpp(str);
	}
}

#ifndef SEAMLESS3DGUI

void CCLathe::Master::operator=(const bool& par){
	refValue()=par;
}
void CCLathe::Curve::operator=(const M_Float& par){
	refValue()=par;
}
void CCLathe::Curve::resize(int newLen){
	refValue().length=newLen;
}
void Morpher::Weight::operator =(const float& interp){
}
void HAnimDisplacer::Weight::operator =(const float& interp){
}
void TapeMeasure::Measurement::operator=(const float& par){
	refValue()=par;
}
void Settings::SaveGZipped::operator=(const bool& par){}
void Settings::IncDecRes::operator=(const float& par){}

#endif

void Color3f::toPOVRayString(String& str)const{
	str+="rgb<";
	str+=r;
	str+=',';
	str+=g;
	str+=',';
	str+=b;
	str+='>';
}

void Color3f::toPOVRayString(String& str, float f)const{
	str+="rgbf<";
	str+=r;
	str+=',';
	str+=g;
	str+=',';
	str+=b;
	str+=',';
	str+=f;
	str+='>';
}

void Color3f::toPOVRayString(String& str, const String& finish)const{
	str+="texture{pigment{";
	toPOVRayString(str);
	str+="}finish{";
	str+=finish;
	str+="}}";
//	str+=">}}";
}
void Color3f::toPOVRayString(String& str, const ColorEffect* effect)const{
	str+="texture{pigment{";
	if(effect->transparency == 0){
		toPOVRayString(str);
	}
	else{
		toPOVRayString(str, effect->transparency);
	}
	str+="}finish{";

	if(effect->name==""){
		str+='e';
		makeNodeNameFromAddr(str, effect);
	}
	else{
		str+=effect->name;
	}
	str+="}}";
}
template<>
void Vec2f::toPOVRayString(String& str, const Node* node)const{
	str+="texture{";
	if(node->name==""){
		str+='e';
		makeNodeNameFromAddr(str, node);
	}
	else{
		str+=node->name;
	}
	str+="}";
}
template<>
void Vec2f::toPOVRayString(String& str)const{
	str+='<';
	str+=x;
	str+=',';
	str+=y;
	str+='>';
}
template<>
void Vec3f::toPOVRayString(String& str)const{
	str+='<';
	str+=x;
	str+=',';
	str+=y;
	str+=',';
	str+=-z;
	str+='>';
}
void IndexedFaceSet::toPOVRayString(String& str){
	M_Vec3f& point=dynamic_cast<Coordinate*>(coord.p)->point;
	M_Vec3f& vector=dynamic_cast<Normal*>(normal.p)->vector;

	str+="vertex_vectors {\n";
	str+=point.length;
	str+="\n";
	{
		for(int i=0; i<point.length; i++){
			point[i].toPOVRayString(str);
			str+=' ';
		}
	}
	str+="\n}normal_vectors {\n";
	str+=vector.length;
	str+="\n";
	{
		for(int i=0; i<vector.length; i++){
			vector[i].toPOVRayString(str);
			str+=' ';
		}
	}
	if(TextureCoordinate* texCoord =dynamic_cast<TextureCoordinate*>(this->texCoord.p)){
		M_Vec2f& point2f=texCoord->point;
		str+="\n}uv_vectors {\n";
		str+=point2f.length;
		str+="\n";
		{
			for(int i=0; i<point2f.length; i++){
				point2f[i].toPOVRayString(str);
				str+=' ';
			}
		}
	}
	if(Color* colorNode =dynamic_cast<Color*>(this->color.p)){
		M_Color3f& color=colorNode->color;
		str+="\n}texture_list {\n";
		str+=color.length;
		str+="\n";
		{
			for(int i=0; i<color.length; i++){
				color[i].toPOVRayString(str, String("myFinish"));
				str+=' ';
			}
		}
		str+="\n}\nface_indices {\n";
		str+=coordIndex.length/4;
		str+='\n';
		for(int i=0; i<coordIndex.length; ){
			str+='<';
			str+=coordIndex[i];
			str+=',';
			str+=coordIndex[i+2];
			str+=',';
			str+=coordIndex[i+1];
			str+=">, ";

			str+=coordIndex[i];
			str+=',';
			str+=coordIndex[i+2];
			str+=',';
			str+=coordIndex[i+1];
			str+=", ";

			i+=4;
		}
	}
	else{
		str+="\n}\nface_indices {\n";
		str+=coordIndex.length/4;
		str+='\n';
		if(colorIndex.refValue().length){
			for(int i=0; i<coordIndex.length; ){
				str+='<';
				str+=coordIndex[i];
				str+=',';
				str+=coordIndex[i+2];
				str+=',';
				str+=coordIndex[i+1];
				str+=">, ";

				str+=colorIndex[i];
				str+=',';
				str+=colorIndex[i+2];
				str+=',';
				str+=colorIndex[i+1];
				str+=", ";

				i+=4;
			}
		}
		else{
			for(int i=0; i<coordIndex.length; ){
				str+='<';
				str+=coordIndex[i];
				str+=',';
				str+=coordIndex[i+2];
				str+=',';
				str+=coordIndex[i+1];
				str+="> ";
				i+=4;
			}
		}
	}
	str+="\n}\n";
}
void Shape::toPOVRayString(String& str){
	str+="mesh2{\n";
	dynamic_cast<IndexedFaceSet*>(geometry.p)->toPOVRayString(str);
	str+="texture {\n\
pigment {color White}\n\
finish { ambient .1 diffuse 0.5 phong .5 }\n\
}\n}\n";
}

void addMonoCShader(GeomRender& mesh);

void IndexedFaceSet::convertToGeomSource(GeomRender& mesh, M_Int& maskIndex)const{
	Appearance* appearance = mesh.appearance;
	Coordinate* ifsCoord=dynamic_cast<Coordinate*>(coord.p);
	if(ifsCoord==0)return;
	mesh.reset();

	M_Int ifsCoordIndex=coordIndex;
	if(convex){
		convertIFSIndexToTriIndex(ifsCoordIndex);
	}
	else{
		convertConcaveIFSIndexToTriIndex(ifsCoord->point, ifsCoordIndex);
	}

	mesh.coord.refValue()=ifsCoord->point;

	M_Int& srcCoordIndex = mesh.coordIndex;
	{
		srcCoordIndex.length=(ifsCoordIndex.length+1)/4*3; //*** keep *** length+1) to ensure in case of absence of -1 at end of list does not cause not enough items
		int j=0;
		for(int i=0; i<ifsCoordIndex.length; i++){
			if(ifsCoordIndex[i]!=-1){
				srcCoordIndex[j++]=ifsCoordIndex[i];
			}
		}
		srcCoordIndex.length=j;
	}

	if(color){
		M_Color3f& ifsColor=dynamic_cast<Color*>(color.p)->color;
		M_Int ifsColorIndex=colorIndex;
		convertIFSIndexToTriIndex(ifsColorIndex);
		if(ifsColorIndex.length>0){
			mesh.convertFromColorIndex(srcCoordIndex, ifsColorIndex, ifsColor );
		}
		else{
			mesh.color.refValue()=ifsColor;
		}
	}
	else if(texCoord && appearance && appearance->texture){
		M_Vec2f& ifsTexCoord=dynamic_cast<TextureCoordinate*>(texCoord.p)->point;
		M_Int ifsTexCoordIndex=texCoordIndex;
		convertIFSIndexToTriIndex(ifsTexCoordIndex);
		if(ifsTexCoordIndex.length>0){
			mesh.convertFromTexCoordIndex(srcCoordIndex, ifsTexCoordIndex, ifsTexCoord );
		}
		else{
			mesh.texCoord.refValue()=ifsTexCoord;
		}
	}
	if (Normal* ifsNormal = dynamic_cast<Normal*>(normal.p)) {
		mesh.normal = ifsNormal->vector;
	}
	else{
		mesh.genNormals();
	}
	mesh.ensureIntegrity();
	Effect* effect=0;
	Seamless smls;

	if(color){
		effect=new ColorEffect;
	}
	else if(appearance && appearance->texture){
		effect=new TextureEffect;
	}
	else{
		MonoColorEffect* monoColorEffect = new MonoColorEffect;
		effect = monoColorEffect;
		if (appearance) {
			if (Material* material = dynamic_cast<Material*>(appearance->material.p)) {
				monoColorEffect->diffuse = material->diffuseColor;

				if (renderProcessed == false || mesh.shader == 0) {
					addMonoCShader(mesh);
				}
				return;

			}
		}
	}
	Part* part=new Part;
	smls.skeleton+=part;
	smls.effect=effect;
//	if(mesh.normal.length==0){
//		mesh.genNormals();
//	}
	effect->convertGeomSourceToPart(part, mesh);
	if(ccw==false){
		CopyPart copyPart;
		copyPart.source+=part;
		copyPart.dest+=part;
		copyPart.insideOut=true;
		copyPart.doBuild();
	}
	if(solid==false){
		CopyPart copyPart;
		copyPart.source+=part;
		copyPart.dest+=part;
		copyPart.insideOut=true;
		copyPart.add=true;
		copyPart.doBuild();
	}
	if(appearance && appearance->texture && texCoord==0){//generate texture map?
		part->setDefaultVrmlTextureMap();
	}
	part->extractGeometryGS(mesh, maskIndex);
}
void IndexedTriangleSet::convertToGeomSource(GeomRender& mesh, M_Int& maskIndex)const{
	Appearance* appearance = mesh.appearance;
	Coordinate* ifsCoord=dynamic_cast<Coordinate*>(coord.p);
	if(ifsCoord==0)return;
	mesh.reset();

	mesh.coord.refValue()=ifsCoord->point;

	mesh.coordIndex=index;

	if(color!=0){
		M_Color3f& ifsColor=dynamic_cast<Color*>(color.p)->color;
		mesh.color.refValue()=ifsColor;
	}
	if(texCoord!=0){
		M_Vec2f& ifsTexCoord=dynamic_cast<TextureCoordinate*>(texCoord.p)->point;
		mesh.texCoord.refValue()=ifsTexCoord;
	}
	if(Normal* ifsNormal=dynamic_cast<Normal*>(normal.p)){
		mesh.normal=ifsNormal->vector;
	}
	else{
		mesh.genNormals();
	}
	mesh.ensureIntegrity();
	Effect* effect=0;
	Seamless smls;

	if(color){
		effect=new ColorEffect;
	}
	else if(appearance->texture){
		effect=new TextureEffect;
	}
	else{
		effect=new MonoColorEffect;
	}
	Part* part=new Part;
	smls.skeleton+=part;
	smls.effect=effect;
//	if(mesh.normal.length==0){
//		mesh.genNormals();
//	}
	effect->convertGeomSourceToPart(part, mesh);
	if(ccw==false){
		CopyPart copyPart;
		copyPart.source+=part;
		copyPart.dest+=part;
		copyPart.insideOut=true;
		copyPart.doBuild();
	}
	if(solid==false){
		CopyPart copyPart;
		copyPart.source+=part;
		copyPart.dest+=part;
		copyPart.insideOut=true;
		copyPart.add=true;
		copyPart.doBuild();
	}
	if(appearance!=0 && appearance->texture!=0 && texCoord==0){//generate texture map?
		part->setDefaultVrmlTextureMap();
	}
	part->extractGeometryGS(mesh, maskIndex);
}
void ElevationGrid::convertToGeomSource(GeomRender& mesh, M_Int& maskIndex)const{
	mesh.reset();
}
void PointSet::convertToGeomSource(GeomRender& mesh, M_Int& maskIndex)const{
	mesh.reset();
}
void Polyline2D::convertToGeomSource(GeomRender& mesh, M_Int& maskIndex)const{
	mesh.reset();
}
void IndexedLineSet::convertToGeomSource(GeomRender& mesh, M_Int& maskIndex)const{
	mesh.reset();
}
void Text::convertToGeomSource(GeomRender& mesh, M_Int& maskIndex)const{
	mesh.reset();
}
void TimeSensor::Loop::operator=(const bool& par){
	refValue()=par;
	//FreezeTime::updateScene=true;
}

void Anim::quantizeSlider(){
	//AnimTimeLine& tl = *(timeLine);
	//AnimSlider& as = *(animSlider);

	//float a = static_cast<float>(as.dstPos.x+(as.bSize.x/2)-tl.dstPos.x-tl.lineOffset);
	//a = a / tl.lineDist;
	if(period < .000001){
		slider = 0;
	}
	else if(play == false){
		int frames = stepsPerSecond * period + .5;
		slider = quantize(slider, 1.0 / frames);
	}
}

void Anim::StepsPerSecond::operator=(const int& par){
	if(par<1){
		refValue()=1;
	}
	else{
		refValue()=par;
	}
	home.quantizeSlider();
}

void togglePlaySound(bool state, short int* ibuf, int bufLen);

void Anim::Play::operator=(const bool& par){
	refValue() = par;
	home.pause.refValue() = !par;
	if(par){
		//home.startTime = g_time - home.playFrom * home.period;
		home.startTime = g_time - home.playFrom.refValue() * home.period;
		if(home.zoom < home.period){
			home.displayFrom = home.playFrom;
		}
		if(home.soundTrack.length){
			if(Composition* composition = dynamic_cast<Composition*>(home.soundTrack.refValue()[0].p)){
				togglePlaySound(par, composition->ibuf, composition->bufLen);
			}
		}
	}
	else{
		home.quantizeSlider();
		if(home.soundTrack.length){
			if(Composition* composition = dynamic_cast<Composition*>(home.soundTrack.refValue()[0].p)){
				togglePlaySound(par, composition->ibuf, composition->bufLen);
			}
		}
	}
}

void Anim::Pause::operator=(const bool& par){
	refValue() = par;
	home.play.refValue() = !par;
	if(par){
		home.quantizeSlider();
		if(home.soundTrack.length){
			if(Composition* composition = dynamic_cast<Composition*>(home.soundTrack.refValue()[0].p)){
				togglePlaySound(par, composition->ibuf, composition->bufLen);
			}
		}
	}
	else{
		home.startTime = g_time - home.slider.refValue() * home.period;
		if(home.soundTrack.length){
			if(Composition* composition = dynamic_cast<Composition*>(home.soundTrack.refValue()[0].p)){
				togglePlaySound(par, composition->ibuf, composition->bufLen);
			}
		}
	}
}

String g_loadStr;

void genSkinPovTransforms(Art* part, Matrix parentMatrix, Matrix parentRotMatrix, int& smi, M_Matrix& m_matrix, M_Matrix& m_rotMatrix){
	Rotation rotation=part->rotation;
	Vec3f scale=part->scale;

	Matrix matrix = translationMatrix(part->translation+part->absolutePivotPoint) * parentMatrix;

	M_Int& coordGroup=part->coordGroup.refValue();
	M_Float& weight=part->weight.refValue();

	for(int cg=0;cg<coordGroup.length; cg++){
		m_matrix+=translationMatrix(-part->absolutePivotPoint)*scaleMatrix(scale)*rotationMatrix(Rotation(rotation.x,rotation.y,rotation.z, rotation.angle*weight[cg]))*matrix;
		m_rotMatrix+=rotationMatrix(Rotation(rotation.x,rotation.y,rotation.z, rotation.angle*weight[cg]))*parentRotMatrix;
		smi++;
	}

	matrix = translationMatrix(-part->absolutePivotPoint)*scaleMatrix(scale) * rotationMatrix(rotation) * matrix;
	parentRotMatrix = rotationMatrix(rotation)*parentRotMatrix;

	m_matrix+=matrix;
	m_rotMatrix+=parentRotMatrix;
	smi++;
	M_SArt& children=part->part.refValue();
	for(int k=0;k<children.length;k++){
		genSkinPovTransforms(children[k], matrix, parentRotMatrix, smi, m_matrix, m_rotMatrix);
	}
}

void genSkinPovTransformsCP(Art* part, Matrix parentMatrix, Matrix parentRotMatrix, int& smi, M_Matrix& m_matrix, M_Matrix& m_rotMatrix){

	M_Int& coordGroup=part->coordGroup.refValue();

	for(int cg = 0; cg < coordGroup.length; cg++){
		m_matrix+=parentMatrix;
		m_rotMatrix+=parentRotMatrix;
		smi++;
	}

	m_matrix+=parentMatrix;
	m_rotMatrix+=parentRotMatrix;
	smi++;
	M_SArt& children=part->part.refValue();
	for(int k=0;k<children.length;k++){
		genSkinPovTransformsCP(children[k], parentMatrix, parentRotMatrix, smi, m_matrix, m_rotMatrix);
	}
}

void Art::outPovRayCoord(String& str, Matrix matrix)const{

}
void Art::outPovRayNormal(String& str, Matrix rotMatrix)const{
	for(int  e=0, i=0; i<vCoordGroup.length; i++){
		Rotation rot=rotation;
		if(i<weight.length){
			rot.angle*=weight[i];
		}
		Matrix rotMat=rotationMatrix(rot)*rotMatrix;
		for(int j=0; j<vCoordGroup[i]; j++, e++){
			(rotMat*normal[scratchC[e]]).toPOVRayString(str);
		}
	}
	rotMatrix=rotationMatrix(rotation)*rotMatrix;
	for(int k=0; k<part.length; k++){
		if(Art* partNode=dynamic_cast<Art*>(part[k].p)){
			partNode->outPovRayCoord(str, rotMatrix);
		}
	}
}
void MonoColorEffect::outPovValues(String& str)const{
	str+="=finish {";
	str+="diffuse ";
	str+=sumUpColor(diffuse);
	str+="}\n";
}
void ColorEffect::outPovValues(String& str)const{
	str+="=finish{";
	str+="ambient ";
	ambient.refValue().toPOVRayString(str);
	str+=" diffuse ";
	str+=sumUpColor(diffuse);
	str+=" brilliance ";
	str+=brilliance;
	str+=" specular ";
	str+=sumUpColor(specular);
	str+=" phong ";
	str+=phong;
	str+=" phong_size ";
	str+=phong_size;
	str+=" roughness ";
	str+=roughness;
	str+=" crand ";
	str+=crand;
	if(conserve_energy){
		str+=" conserve_energy";
	}
	if(reflection!=Color3f(0,0,0)){
		str+=" reflection{";
		reflection.refValue().toPOVRayString(str);
		str+=" falloff ";
		str+=falloff;
		if(fresnel){
			str+=" fresnel";
		}
		str+=" exponent ";
		str+=exponent;
		str+=" metallic ";
		str+=metallic;
		str+="} ";
	}
	if(iridescence){
		str+=" irid{";
		str+=iridescence;
		str+=" thickness ";
		str+=thickness;
		str+=" turbulence ";
		str+=turbulence;
		str+="}";
	}
	str+="}\n";
}
void TextureEffect::outPovValues(String& str)const{
	str+="=texture{uv_mapping pigment{image_map{png \"";
	if(url.refValue().length){
		str+=url.refValue();
	}
	str+="\" map_type 0 interpolate 2}}finish{diffuse 1 ";
	str+="ambient ";
	ambient.refValue().toPOVRayString(str);
	str+=" diffuse ";
	str+=sumUpColor(diffuse);
	str+=" brilliance ";
	str+=brilliance;
	str+=" specular ";
	str+=sumUpColor(specular);
	str+=" phong ";
	str+=phong;
	str+=" phong_size ";
	str+=phong_size;
	str+=" roughness ";
	str+=roughness;
	str+=" crand ";
	str+=crand;
	if(conserve_energy){
		str+=" conserve_energy";
	}
	if(reflection!=Color3f(0,0,0)){
		str+=" reflection{";
		reflection.refValue().toPOVRayString(str);
		str+=" falloff ";
		str+=falloff;
		if(fresnel){
			str+=" fresnel";
		}
		str+=" exponent ";
		str+=exponent;
		str+=" metallic ";
		str+=metallic;
		str+="} ";
	}
	if(iridescence){
		str+="irid{";
		str+=iridescence;
		str+=" thickness ";
		str+=thickness;
		str+=" turbulence ";
		str+=turbulence;
		str+="}";
	}
	str+="}}\n";
}

void MonoColorEffect::outPovRay(const M_Art& partList, String& str, int vertexCount, M_Int& coordIndex, const int triLenTotal)const{
}
void TextureEffect::outPovRay(const M_Art& partList, String& str, int vertexCount, M_Int& coordIndex, const int triLenTotal)const{

	str+="\n}uv_vectors {\n";
	str+=vertexCount;
	str+="\n";
	for(int k=0; k<partList.length; k++){
		Art* part=partList[k];
		M_Vec2f& texCoord=part->texCoord;
		for(int i=0; i<texCoord.length; i++){
			texCoord[i].toPOVRayString(str);
			str+=' ';
		}
	}
/*
	str+="\n}\nface_indices {\n";
	str+=coordIndex.length/3;
	str+='\n';
	for(int i=0; i<coordIndex.length; ){
		str+='<';
		str+=coordIndex[i];
		str+=',';
		str+=coordIndex[i+2];
		str+=',';
		str+=coordIndex[i+1];
		str+=">, ";
		i+=4;
	}
	str+="\n}\n";
*/
	//////////////
	static String extraColors;
	extraColors="";
	static M_Int colorCoordIndex;
	colorCoordIndex.length=coordIndex.length;

	for(int t=2, k=0; k<partList.length; k++){
		Art* part=partList[k];
		M_Triangle& triangle=part->triangle;
		int triLen=triangle.length;
		if(part->show){
			for(int tr=0, i=triLen; --i>=0; ){
				colorCoordIndex[t]=coordIndex[t];
				if(triangle[tr].pc->workEffect!=part->workEffect){
					texCoord[triangle[tr].tc].toPOVRayString(extraColors, part->workEffect);
					colorCoordIndex[t]=vertexCount;
					vertexCount++;
				}
				--t;
				colorCoordIndex[t]=coordIndex[t];
				if(triangle[tr].pb->workEffect!=part->workEffect){
					texCoord[triangle[tr].tc].toPOVRayString(extraColors, part->workEffect);
					colorCoordIndex[t]=vertexCount;
					vertexCount++;
				}
				--t;
				colorCoordIndex[t]=coordIndex[t];
				if(triangle[tr].pa->workEffect!=part->workEffect){
					texCoord[triangle[tr].tc].toPOVRayString(extraColors, part->workEffect);
					colorCoordIndex[t]=vertexCount;
					vertexCount++;
				}
				t+=5;
				tr++;
			}
		}
		else{
			t+=triLen*3;
		}
	}
	str+="\n}texture_list {\n";
	str+=vertexCount;
	str+="\n";
	static String finish;
	if(name==""){
		finish='e';
		makeNodeNameFromAddr(finish, this);
	}
	else{
		finish=name;
	}

	for(int k=0;k<partList.length;k++){

		Art* part=partList[k];
		const M_Vec2f& texCoord=part->texCoord;

		for(int i=0;i<part->texCoord.length;i++){
			texCoord[i].toPOVRayString(str, part->workEffect);
		}
	}
	str+=extraColors;

	str+="\n}\nface_indices {\n";
	str+=triLenTotal;
	str+='\n';

	for(int t=2, k=0; k<partList.length; k++){
		Art* part=partList[k];
		M_Triangle& triangle=part->triangle;
		int triLen=triangle.length;
		if(part->show){
			for(int tr=0, i=triLen; --i>=0; ){
				str+='<';
				str+=coordIndex[t];
				str+=',';
				str+=coordIndex[--t];
				str+=',';
				str+=coordIndex[--t];
				str+=">,";

				t+=2;

				str+=colorCoordIndex[t];
				--t;
				str+=',';
				str+=colorCoordIndex[t];
				--t;
				str+=',';
				str+=colorCoordIndex[t];
				str+=',';
				t+=5;
				tr++;
			}
		}
		else{
			t+=triLen*3;
		}
	}
}

void ColorEffect::outPovRay(const M_Art& partList, String& str, int vertexCount, M_Int& coordIndex, const int triLenTotal)const{
	static String extraColors;
	extraColors="";
	static M_Int colorCoordIndex;
	colorCoordIndex.length = coordIndex.length;

	for(int t=2, k=0; k<partList.length; k++){
		Art* part=partList[k];
		M_Triangle& triangle=part->triangle;
		int triLen=triangle.length;
		if(part->show){
			ColorEffect* colorEffect = dynamic_cast<ColorEffect*>(part->workEffect);
			for(int tr=0, i=triLen; --i>=0; ){
				colorCoordIndex[t] = coordIndex[t];
				if(triangle[tr].pc->workEffect!=part->workEffect){
					color[triangle[tr].tc].toPOVRayString(extraColors, colorEffect);
					colorCoordIndex[t] = vertexCount;
					vertexCount++;
				}
				--t;
				colorCoordIndex[t] = coordIndex[t];
				if(triangle[tr].pb->workEffect!=part->workEffect){
					color[triangle[tr].tc].toPOVRayString(extraColors, colorEffect);
					colorCoordIndex[t] = vertexCount;
					vertexCount++;
				}
				--t;
				colorCoordIndex[t] = coordIndex[t];
				if(triangle[tr].pa->workEffect!=part->workEffect){
					color[triangle[tr].tc].toPOVRayString(extraColors, colorEffect);
					colorCoordIndex[t] = vertexCount;
					vertexCount++;
				}
				t+=5;
				tr++;
			}
		}
		else{
			t+=triLen*3;
		}
	}
	str+="\n}texture_list {\n";
	str+=vertexCount;
	str+="\n";
	static String finish;
	if(name==""){
		finish='e';
		makeNodeNameFromAddr(finish, this);
	}
	else{
		finish=name;
	}

	for(int k=0;k<partList.length;k++){

		Art* part=partList[k];
		ColorEffect* colorEffect = dynamic_cast<ColorEffect*>(part->workEffect);
		const M_Color3f& color=part->color;

		for(int i=0;i<part->color.length;i++){
			color[i].toPOVRayString(str, colorEffect);
		}
	}
	str+=extraColors;

	str+="\n}\nface_indices {\n";
	str+=triLenTotal;
	str+='\n';

	for(int t=2, k=0; k<partList.length; k++){
		Art* part=partList[k];
		M_Triangle& triangle=part->triangle;
		int triLen=triangle.length;
		if(part->show){
			for(int tr=0, i=triLen; --i>=0; ){
				str+='<';
				str+=coordIndex[t];
				str+=',';
				str+=coordIndex[--t];
				str+=',';
				str+=coordIndex[--t];
				str+=">,";

				t+=2;

				str+=colorCoordIndex[t];
				--t;
				str+=',';
				str+=colorCoordIndex[t];
				--t;
				str+=',';
				str+=colorCoordIndex[t];
				str+=',';
				t+=5;
				tr++;
			}
		}
		else{
			t+= triLen * 3;
		}
	}
}

inline void addEffectToList(M_<Effect*>& effectList, Effect* effect, String& str){
	for(int i=effectList.length; --i>=0;){
		if(effect==effectList[i]){
			return;
		}
	}
	effectList+=effect;

	str+="#declare ";
	if(effect->name==""){
		str+='e';
		makeNodeNameFromAddr(str, effect);
	}
	else{
		str+=effect->name;
	}
	effect->outPovValues(str);

}
void ModelMorpher::outPovRay(String& str, M_<Effect*>& effectList, Matrix matrix, Matrix rotMatrix){
	smls->outPovRay(str, effectList, matrix, rotMatrix);
}
void Seamless::outPovRay(String& str, M_<Effect*>& effectList, Matrix matrix, Matrix rotMatrix){//much the same as render triangles
	if(effect==0)return;

	matrix = scaleMatrix(scale) * rotationMatrix(rotation) * translationMatrix(translation) * matrix;
	rotMatrix = rotationMatrix(rotation) * rotMatrix;

	addEffectToList(effectList, dynamic_cast<Effect*>(effect.p), str);
	for(int k=0; k<partList.length; k++){
		if(Effect* effect=dynamic_cast<Effect*>(partList[k]->effect.p)){
			addEffectToList(effectList, effect, str);
		}
	}
	str.requestMoreIfNeeded(2000000000);

	const M_SArt& skeleton=this->skeleton;

	static M_Matrix m_matrix;
	static M_Matrix m_rotMatrix;

	m_matrix.length=0;
	m_rotMatrix.length=0;

	if(workAnimCP){
		for(int i=0, smi=0; i<skeleton.length; i++){
			genSkinPovTransformsCP(skeleton[i], matrix, rotMatrix, smi, m_matrix, m_rotMatrix);
		}
	}
	else{
		for(int i=0, smi=0; i<skeleton.length; i++){
			genSkinPovTransforms(skeleton[i], matrix, rotMatrix, smi, m_matrix, m_rotMatrix);
		}
	}


/////////////////////////
	static Mesh* mesh= new Mesh;

	if(partList.length==0 || partList[0]->workEffect==0)return;
	int j=0;

	M_Int& coordIndex=mesh->coordIndex.refValue();

	const_cast<Seamless*>(this)->catTranslationMatrix();
	mesh->genTransformedNormals(partList);
	int coordLen=mesh->coord.length;
	catMatrix();

	addExtraCoordsRenderForGeomSource(coordIndex);

	int indexCount=coordIndex.length*3;

	if(!indexCount){
		return;
	}
	{
		coordLen = 0;
		for(int k=0; k<partList.length; k++){
			coordLen+=partList[k]->coord.length;
		}
	}
	int triLenTotal = 0;
	for(int t=0, k=partList.length; --k>=0;){
		Art* part=partList[k];
		if(part->show){
			triLenTotal+=part->triangle.length;
		}
	}
	if(triLenTotal == 0){
		return;
	}

	if(shadow == false){
		str+="object{\n";
	}
	str+="mesh2{\nvertex_vectors {\n";
	str+=coordLen;
	str+="\n";

	for(int j = 0, smi = 0,k = 0; k < partList.length; k++){

		Art* part = partList[k];

		part->renderFromCoord = j;

		const M_Vec3f& coord = part->coord;
		M_Int& coordGroup = part->coordGroup.refValue();
		M_Int& matrixIndex = part->matrixIndex;

		int i = 0;

		for(int cg = 0;cg<coordGroup.length; cg++){
			for(int c = coordGroup[cg]; --c >= 0 && i < part->coord.length; i++, j++){
				(m_matrix[matrixIndex[i]] * coord[i]).toPOVRayString(str);
			}
			smi++;
		}
		for(;i<part->coord.length; i++, j++){
			(m_matrix[matrixIndex[i]] * coord[i]).toPOVRayString(str);
		}
		smi++;
	}

	str+="\n}normal_vectors {\n";
	str+=coordLen;
	str+="\n";

	for(int j = 0, smi = 0, k = 0; k < partList.length; k++){

		Art* part = partList[k];

		part->renderFromCoord = j;

		const M_Vec3f& normal = part->normal;
		M_Int& coordGroup = part->coordGroup.refValue();
		M_Int& matrixIndex = part->matrixIndex;

		int i = 0;

		for(int cg = 0; cg < coordGroup.length; cg++){
			for(int c = coordGroup[cg]; --c >= 0 && i < part->coord.length; i++, j++){
				(m_rotMatrix[matrixIndex[i]] * normal[i]).toPOVRayString(str);
			}
			smi++;
		}
		for(;i<part->coord.length;i++, j++){
			(m_rotMatrix[matrixIndex[i]]*normal[i]).toPOVRayString(str);
		}
		smi++;
	}
	setUpInternalEffectPtrs2();
	effect->outPovRay(partList, str, coordLen, coordIndex, triLenTotal);
	setUpInternalEffectPtrs();

	str+="\n}\n}\n";
	if(shadow == false){
		str+="no_shadow\n}\n";
	}

	resetCoordLen();
}
void PovLight::outPovRay(String& str, M_<Effect*>& effectList, Matrix matrix, Matrix rotMatrix){
	str+="light_source{";
	//str+=<0,0,-100>
	location.refValue().toPOVRayString(str);
	str+="color ";
	color.refValue().toPOVRayString(str);
	if(spotlight){
		str+=" spotlight";

		str+=" radius degrees(";
		str+=radius;
		str+=") falloff degrees(";
		str+=falloff;
		str+=") tightness ";
		str+=tightness;
		str+=" point_at";
		point_at.refValue().toPOVRayString(str);;
	}
	if(shadowless){
		str+=" shadowless ";
	}
	str+=" fade_distance ";
	str+=fade_distance;
	str+=" fade_power ";
	str+=fade_power;
	str+="}\n";
}
void PovFog::outPovRay(String& str, M_<Effect*>& effectList, Matrix matrix, Matrix rotMatrix){
	str+="fog{";
	str+="lambda ";
	str+=lambda;
	str+=" fog_type ";
	str+=fog_type;
	str+=" fog_offset ";
	str+=fog_offset;
	str+=" fog_alt ";
	str+=fog_alt;
	str+=" octaves ";
	str+=octaves;
	str+=" omega ";
	str+=omega;
	str+=" turbulence ";
	turbulence.refValue().toPOVRayString(str);
	str+=" turb_depth ";
	str+=turb_depth;
	str+=" up ";
	up.refValue().toPOVRayString(str);
	str+="}\n";
}
void Transform::outPovRay(String& str, M_<Effect*>& effectList, Matrix matrix, Matrix rotMatrix){
	const M_Node& children = this->children.refValue();
	for(int k=0; k<children.length; k++){
		if(PovRay* povRay=dynamic_cast<PovRay*>(children[k].p)){
			povRay->outPovRay(str, effectList, translationMatrix(-center) * rotationMatrix(-scaleOrientation) * scaleMatrix(scale) * rotationMatrix(scaleOrientation) * rotationMatrix(rotation) * translationMatrix(center) * translationMatrix(translation) * matrix,
				rotationMatrix(-scaleOrientation) * rotationMatrix(scaleOrientation) * rotationMatrix(rotation) * rotMatrix
			);
		}
	}
}
void Group::outPovRay(String& str, M_<Effect*>& effectList, Matrix matrix, Matrix rotMatrix){
	const M_Node& children = this->children.refValue();
	for(int k=0; k<children.length; k++){
		if(PovRay* povRay=dynamic_cast<PovRay*>(children[k].p)){
			povRay->outPovRay(str, effectList, matrix, rotMatrix);
		}
	}
}
void LOD::outPovRay(String& str, M_<Effect*>& effectList, Matrix matrix, Matrix rotMatrix){
	const M_Node& children = this->children.refValue();
	for(int k=0; k<children.length; k++){
		if(PovRay* povRay=dynamic_cast<PovRay*>(children[k].p)){
			povRay->outPovRay(str, effectList, matrix, rotMatrix);
		}
	}
}
/*
void HAnimHumanoid::outPovRay(String& str, M_<Effect*>& effectList, Matrix matrix, Matrix rotMatrix){
	const M_Node& skeleton = this->skeleton.refValue();
	for(int k=0; k<skeleton.length; k++){
		if(PovRay* povRay=dynamic_cast<PovRay*>(skeleton[k].p)){
			povRay->outPovRay(str, effectList, translationMatrix(-center) * rotationMatrix(-scaleOrientation) * scaleMatrix(scale) * rotationMatrix(scaleOrientation) * rotationMatrix(rotation) * translationMatrix(center) * translationMatrix(translation) * matrix,
				rotationMatrix(-scaleOrientation) * rotationMatrix(scaleOrientation) * rotationMatrix(rotation) * rotMatrix
			);
		}
	}
}
*/
void Switch::outPovRay(String& str, M_<Effect*>& effectList, Matrix matrix, Matrix rotMatrix){
	const M_Node& children = this->children.refValue();
	for(int k=0; k<children.length; k++){
		if(PovRay* povRay=dynamic_cast<PovRay*>(children[k].p)){
			povRay->outPovRay(str, effectList, matrix, rotMatrix);
		}
	}
}
void Collision::outPovRay(String& str, M_<Effect*>& effectList, Matrix matrix, Matrix rotMatrix){
	const M_Node& children = this->children.refValue();
	for(int k=0; k<children.length; k++){
		if(PovRay* povRay=dynamic_cast<PovRay*>(children[k].p)){
			povRay->outPovRay(str, effectList, matrix, rotMatrix);
		}
	}
}
void Anchor::outPovRay(String& str, M_<Effect*>& effectList, Matrix matrix, Matrix rotMatrix){
	const M_Node& children = this->children.refValue();
	for(int k=0; k<children.length; k++){
		if(PovRay* povRay=dynamic_cast<PovRay*>(children[k].p)){
			povRay->outPovRay(str, effectList, matrix, rotMatrix);
		}
	}
}
void Billboard::outPovRay(String& str, M_<Effect*>& effectList, Matrix matrix, Matrix rotMatrix){
	const M_Node& children = this->children.refValue();
	for(int k=0; k<children.length; k++){
		if(PovRay* povRay=dynamic_cast<PovRay*>(children[k].p)){
			povRay->outPovRay(str, effectList, matrix, rotMatrix);
		}
	}
}
void ProtoDeclare::outPovRay(String& str, M_<Effect*>& effectList, Matrix matrix, Matrix rotMatrix){
	M_Node& children=protoBody->children.refValue();
	for(int k=0;k<children.length;k++){
		if(PovRay* povRay=dynamic_cast<PovRay*>(children[k].p)){
			povRay->outPovRay(str, effectList, matrix, rotMatrix);
		}
	}
}
void Scene::outPovRay(String& str, M_<Effect*>& effectList, Matrix matrix, Matrix rotMatrix){
	const M_Node& children = this->children.refValue();

	//matrix = getViewMatrix();
	//for(int i = 0; i<4; i++){
	//	for(int j = 0; j<4; j++){
	//		float a = matrix.m[i][j];
	//		str+=a;
	//		str+=",";
	//	}
	//}
	//rotMatrix=rotationMatrix(rotation);

	for(int k=0; k<children.length; k++){
		if(PovRay* povRay=dynamic_cast<PovRay*>(children[k].p)){
			povRay->outPovRay(str, effectList, matrix, rotMatrix);
		}
	}
	str+="\n#declare vlocation=";
	Vec3f location = -g_scene->translation * g_scene->getZoom() + (-g_scene->rotation * Vec3f(0, 0, g_viewDist));
	location.toPOVRayString(str);

	str+=";\n#declare vlook_at=";
	Vec3f look_at = -g_scene->translation * g_scene->getZoom();
	look_at.toPOVRayString(str);

	str+=";\n#declare vsky=";
	Vec3f sky = -g_scene->rotation * Vec3f(0,1,0);
	sky.toPOVRayString(str);

	str+=";\n#declare vscale=";
	str += 1 / g_scene->getZoom();

	str+=";\n#declare zangle=";

	float a = g_scene->fieldOfView / 2;
	float y = sin(a);
	float z = cos(a);
	float x = y * screenRatio();

	float angle = atan2(x, z) * 2;

	angle = angle * (180.0/PI);
	str += angle;
	str += ";\n";
}

void CPMorpher::morph(float v, Art& part){
	frame = v;
	Seamless* smls = part.smls;
	smls->cpMorpherPart += CPMorpherPart(this, &part);
	smls->morphFlag = true;
	smls->animCPBuildFlag = true;
}

void PosCompensator::comp(float v, Art& ankle){
	frac = v;
	Seamless* smls = ankle.smls;
	this->ankle = &ankle;
	this->pelvis = smls->partList[0];

	M_Vec2f& moment = this->moment.refValue();
	for(int i = 0; i < moment.length; i++){
		if(v >= moment[i].x && v < moment[i].y){
			smls->posCompensator = this;
			momentIdx = i;
		}
	}
}

Node* SphereBuild::HSFPart::operator=(Node* par){
	SFNode::refValue()=par;
	home.processFields();
	return par;
};

Node* GroupBuild::HSFPart::operator=(Node* par){
	SFNode::refValue()=par;
	home.processFields();
	return par;
};

void Art::doBuild(NurbsPatch* patch){
	M_Vec3f& coord = this->coord.refValue();
	int coordMark = coord.length;

	patch->doBuild();

	//if(smls->workAnimCP){
	//	for(int i = coord.length; --i>=coordMark;){
	//		coord[i] *= matrix;
	//	}
	//}
}
void Art::doBuild(M_<NurbsPatchPtr>& m_patch){
	int patchLen = m_patch.length;

	if(smls->workAnimCP){
		M_Vec3f& coord = this->coord.refValue();

		for(int k = 0; k < patchLen; k++){
			int coordMark = coord.length;
			m_patch[k]->doBuild();
			for(int i = coord.length; --i>=coordMark;){
				coord[i] *= matrix;
			}
		}
	}
	else{
		for(int k = 0; k < patchLen; k++){
			m_patch[k]->doBuild();
		}
	}
}

void SphereBuild::doBuild(){
	patch->doBuild();

	if(descendants){
		Matrix matrix = scaleMatrix(scale) * rotationMatrix(rotation) * translationMatrix(translation);
		staticTransformDecendentParts(dynamic_cast<Art*>(part.p), matrix);
	}
}

void CylinderBuild::TopHeight::operator=(const float& par){
	if(refValue() == 0 && par > 0 && home.topBands == 1){
		home.topBands.refValue()=4;
	}
	if(refValue() > 0 && par == 0){
		home.topBands.refValue()=1;
	}
	refValue()=par;
	home.processFields();
};
void CylinderBuild::BaseHeight::operator=(const float& par){
	if(refValue() == 0 && par > 0 && home.baseBands == 1){
		home.baseBands.refValue()=4;
	}
	if(refValue() > 0 && par == 0){
		home.baseBands.refValue()=1;
	}
	refValue()=par;
	home.processFields();
};
Node* CylinderBuild::HSFPart::operator=(Node* par){
	SFNode::refValue()=par;
	home.processFields();
	return par;
};
void CylinderBuild::doBuild(){
	patch->doBuild();

	if(descendants){
		Matrix matrix = scaleMatrix(scale) * rotationMatrix(rotation) * translationMatrix(translation);
		staticTransformDecendentParts(dynamic_cast<Art*>(part.p), matrix);
	}
}

Node* ConeBuild::HSFPart::operator=(Node* par){
	SFNode::refValue()=par;
	home.processFields();
	return par;
};
void ConeBuild::doBuild(){
	if(Art* part=dynamic_cast<Art*>(this->part.p)){

		part->doBuild(patch);

		if(descendants){
			Matrix matrix = scaleMatrix(scale) * rotationMatrix(rotation) * translationMatrix(translation);
			staticTransformDecendentParts(part, matrix);
		}
	}
}
Node* TorusBuild::HSFPart::operator=(Node* par){
	SFNode::refValue()=par;
	home.processFields();
	return par;
};
void TorusBuild::doBuild(){
	//return;
	if(Art* part = dynamic_cast<Art*>(this->part.p)){
		int h = part->coord.length;
		patch->doBuild();
		//if(textureMapper.length == 0 || dynamic_cast<Stem*>(textureMapper[0].p) == 0){
		//	joinVertex->in.refValue() = false;
		//	joinVertex->srcCoord.refValue().length = 0;
		//	joinVertex->dstCoord.refValue().length = 0;
		//	joinVertex->localIn(part, h, joinVertex->srcCoord.refValue(), joinVertex->dstCoord.refValue());
		//	joinVertex->doBuild();
		//}

		if(descendants){
			Matrix matrix = scaleMatrix(scale) * rotationMatrix(rotation) * translationMatrix(translation);
			staticTransformDecendentParts(part, matrix);
		}
	}
	else{
		patch->doBuild();
	}
}

Node* BoxBuild::HSFPart::operator=(Node* par){
	SFNode::refValue()=par;
	home.processFields();
	return par;
};

void BoxBuild::doBuild(){
	for(int k = 0; k < 6; k++){
		m_patch[k]->doBuild();
	}
	if(descendants){
		Matrix matrix = scaleMatrix(scale) * rotationMatrix(rotation) * translationMatrix(translation);
		staticTransformDecendentParts(dynamic_cast<Art*>(part.p), matrix);
	}
}


void RectangleBuild::doBuild(){
	patch->doBuild();

	if(descendants){
		Matrix matrix = scaleMatrix(scale) * rotationMatrix(rotation) * translationMatrix(translation);
		staticTransformDecendentParts(dynamic_cast<Art*>(part.p), matrix);
	}
}

Node* RectangleBuild::HSFPart::operator=(Node* par){
	SFNode::refValue()=par;
	home.processFields();
	return par;
};
Node* BezierLatheBuild::HSFPart::operator=(Node* par){
	SFNode::refValue()=par;
	home.processFields();
	return par;
};
void BezierLatheBuild::doBuild(){
	if(Art* part=dynamic_cast<Art*>(this->part.p)){
		if(point.refValue().length==0){
			return;
		}

		part->doBuild(m_patch);

		if(m_patch.length > 1){
			joinVertex->doBuild();
			joinVertex->in.refValue() = false;
		}
		if(descendants){
			Matrix matrix = scaleMatrix(scale) * rotationMatrix(rotation) * translationMatrix(translation);
			staticTransformDecendentParts(part, matrix);
		}
	}
}
void ECLathe::getBandAndYDist(M_Int& bands, M_Float& bandDist, M_Float& yDistance)const{
	commonGetSegAndBandDist(bands, distance, bandDist);
	yDistance=distance;
}
void ECSyncBend::getBandAndYDist(M_Int& bands, M_Float& bandDist, M_Float& yDistance)const{
	commonGetSegAndBandDist(bands, distance, bandDist);
	yDistance=distance;
}


/*
void BezierLathe::getBandAndYDist(const M_Int& bands, M_Float& bandDist, M_Float& yDistance)const{
	if(bands.length){

		bandDist.length=0;
		bandDist+=0;
		int from=0; //flatHeight from
		int to=numOfSteps; //flatHeight to

		for(int g=0; g<bands.length; g++){
			int numOfBands=bands[g];
			float td=tangentDiff[to-1];
			if(td < .002){//do linear bands
				float step=segDistance[g]/numOfBands;
				float h=flatHeight[from];
				for(int k=0; k<numOfBands-1; k++){
					h+=step;
					bandDist+=h;
				}
			}
			else{//do non linear bands
				float step=td/numOfBands;
				float stepAccum=step;

				int tdl=to-2;
				for(int b=0, k=from; b < numOfBands-1; b++){

					while(k < tdl && tangentDiff[k] < stepAccum)k++;

					if(tangentDiff[k] > stepAccum){
						k--;
					}

					float lastAngleBit = stepAccum - tangentDiff[k];
					float lastTangentDiff = tangentDiff[k+1] - tangentDiff[k];
					float ratio = lastAngleBit/lastTangentDiff;
					bandDist += flatHeight[k] + (flatHeight[k+1] - flatHeight[k]) * ratio;
					stepAccum += step;
				}
			}

			bandDist+=flatHeight[to-1];
			from=to;
			to+=numOfSteps;
		}
	}
	yDistance=segDistance;
}
*/

//M_Vec2f Bezier::normA;
//M_Vec2f Bezier::normB;
//M_Vec2f Bezier::normC;
//M_Vec2f Bezier::normD;

//Intersection point of two lines (2 dimensions) formula from http://local.wasp.uwa.edu.au/~pbourke/geometry/lineline2d Written by Paul Bourke
Vec2f lineIntersect(Vec2f p1, Vec2f p2, Vec2f p3, Vec2f p4){
	float ua=((p4.x-p3.x)*(p1.y-p3.y)-(p4.y-p3.y)*(p1.x-p3.x)) / ((p4.y-p3.y)*(p2.x-p1.x)-(p4.x-p3.x)*(p2.y-p1.y));
	return Vec2f(p1.x + ua * (p2.x-p1.x), p1.y + ua * (p2.y-p1.y));
}
M_Vec2f Bezier::pnt;
M_Vec3f Bezier3f::pnt;


bool twoLineSegmentsIntersect(Vec2f AS, Vec2f AE, Vec2f BS, Vec2f BE){

	Vec2f DP, QA, QB;

	DP.x = BS.x - AS.x + BE.x - AE.x;
	DP.y = BS.y - AS.y + BE.y - AE.y;
	QA.x = AE.x - AS.x;
	QA.y = AE.y - AS.y;
	QB.x = BE.x - BS.x;
	QB.y = BE.y - BS.y;

	float d  = QA.y * QB.x - QB.y * QA.x ;
	float la = QB.x * DP.y - QB.y * DP.x ;
	float lb = QA.x * DP.y - QA.y * DP.x ;

	if(abs(la) <= abs(d) && abs(lb) <= abs(d)){
		return true;
	}
	return false;
}

bool doesLineSegmentIntersectRect(Vec2f a, Vec2f b){
	//top edge
	if(twoLineSegmentsIntersect(a, b, Vec2f(MultiSelect::topLeft.x, MultiSelect::topLeft.y), Vec2f(MultiSelect::bottomRight.x, MultiSelect::topLeft.y))){
		return true;
	}
	//right edge
	if(twoLineSegmentsIntersect(a, b, Vec2f(MultiSelect::bottomRight.x, MultiSelect::topLeft.y), Vec2f(MultiSelect::bottomRight.x, MultiSelect::bottomRight.y))){
		return true;
	}
	//bottom edge
	if(twoLineSegmentsIntersect(a, b, Vec2f(MultiSelect::topLeft.x, MultiSelect::bottomRight.y), Vec2f(MultiSelect::bottomRight.x, MultiSelect::bottomRight.y))){
		return true;
	}
	//left edge
	if(twoLineSegmentsIntersect(a, b, Vec2f(MultiSelect::topLeft.x, MultiSelect::topLeft.y), Vec2f(MultiSelect::topLeft.x, MultiSelect::bottomRight.y))){
		return true;
	}
	return false;
}
bool triIntersectsRect(Vec2f a, Vec2f b, Vec2f c){
	if(doesLineSegmentIntersectRect(a, b)){
		return true;
	}
	if(doesLineSegmentIntersectRect(b, c)){
		return true;
	}
	if(doesLineSegmentIntersectRect(c, a)){
		return true;
	}
	return false;
}
bool isRectCornerInTriangle(const Vec2f* tri){
	if(isTriangleHit(Vec2f(MultiSelect::topLeft.x, MultiSelect::topLeft.y), tri)){
		return true;
	}
	if(isTriangleHit(Vec2f(MultiSelect::bottomRight.x, MultiSelect::bottomRight.y), tri)){
		return true;
	}
	return false;
}
void BezierLathe::getBandAndYDist(M_Int& bands, M_Float& bandDist, M_Float& yDistance)const{

	bandDist.length=0;
	bands.length=yDistance.length=nopfs.length;
	float yd=0;
	for(int k=0; k<nopfs.length; k++){
		if(bands[k]==0){
			bands[k]=16;
		}
		yDistance[k]=1;
		float step=1.0/bands[k];
		bandDist+=yd;
		float h=yd;
		for(int j=1; j<bands[k]; j++){
			h+=step;
			bandDist+=h;
		}
		yd+=1;
	}
	bandDist+=yd;
}
void BezierTape::getBandAndYDist(M_Int& bands, M_Float& bandDist, M_Float& yDistance)const{
	if(bands.length){

		bandDist.length=0;
		bandDist+=0;
		int from=0; //flatHeight from
		int to=numOfSteps; //flatHeight to

		for(int p = 0, s = 0; s < nopfs.length; p+=nopfs[s++]-1){

			pnt.length = nopfs[s];

			int numOfBands=bands[s];
			float step=1.0/numOfBands;
			float stepAccum=step;

			for(int b=1, k=from; b < numOfBands; b++){

				Vec2f vec = genCurve3(numOfSteps, p, s, stepAccum, pnt);
				Vec2f nrm = genCurve3(numOfSteps, p, s, stepAccum+.001, pnt) - vec;
				nrm = vec + Vec2f(nrm.y, -nrm.x);

				int i = stepAccum*numOfSteps;
				i+=from;

				Vec2f intersect = lineIntersect(vec, nrm, line[i], line[i+1]);

				Vec2f lineDiff = line[i+1]-line[i];

				float ratio = (lineDiff.y > lineDiff.x) ?
					fabsf((line[i].y-intersect.y)/lineDiff.y):
					fabsf((line[i].x-intersect.x)/lineDiff.x);

				bandDist += flatHeight[i-1] + (flatHeight[i] - flatHeight[i-1]) * ratio;
				stepAccum += step;
			}

			bandDist+=flatHeight[to-1];
			from=to;
			to+=numOfSteps;
		}
	}
	yDistance=segDistance;
}
void BezierScale::generate(M_Vec3f& coord, const int h){
	if(!point.length)return;
	if(!flatHeight.length)return;

	float bezArcLen=flatHeight[flatHeight.length-1];

	float fNegX=1/negX;
	float fPosX=1/posX;
	float fNegZ=1/negZ;
	float fPosZ=1/posZ;

	for(int i=0; i<coord.length; i++){
		Vec3f s_crd=gen_imatrix*coord[i];

		int idx = s_crd.y/bezArcLen*flatHeight.length;

		if(idx>=heightIndex.length){
			idx=heightIndex.length-1;
		}
		else if(idx<0){
			idx=0;
		}
		idx=heightIndex[idx];

		while(idx>0 && s_crd.y < flatHeight[idx-1]){
			idx--;
		}
		float flatDiff=flatHeight[idx];
		float coordDiff=s_crd.y;

		if(idx>0){
			flatDiff-=flatHeight[idx-1];
			coordDiff-=flatHeight[idx-1];
		}
		float ratio=coordDiff/flatDiff;


		float rad = line[idx].x+(line[idx+1].x-line[idx].x)*ratio;

//
		if(bNegX==true && s_crd.x<0){
			s_crd.x*=rad*fNegX;
			if(bPosX==true && s_crd.x>0){
jumpPosX:
				s_crd.x*=rad*fPosX;
			}
			if(bNegZ==true && s_crd.z<0){
jumpNegZ:
				s_crd.z*=rad*fNegZ;
			}
			if(bNegZ==true && s_crd.z>0){
jumpPosZ:
				s_crd.z*=rad*fPosZ;
			}
			coord[i]=gen_matrix*s_crd;
		}
		else if(bPosX==true && s_crd.x>0){
			goto jumpPosX;
		}
		else if(bNegZ==true && s_crd.z<0){
			goto jumpNegZ;
		}
		else if(bPosZ==true && s_crd.z>0){
			goto jumpPosZ;
		}
	}
}
inline float Bezier::getCircumScaleVal(const float& scaleVal, const float& bezArcLen)const{
	int idx = scaleVal/bezArcLen*flatHeight.length;

	if(idx>=heightIndex.length){
		idx=heightIndex.length-1;
	}
	else if(idx<0){
		idx=0;
	}
	idx=heightIndex[idx];

	while(idx>0 && scaleVal < flatHeight[idx-1]){
		idx--;
	}
	float flatDiff=flatHeight[idx];
	float coordDiff=scaleVal;

	if(idx>0){
		flatDiff-=flatHeight[idx-1];
		coordDiff-=flatHeight[idx-1];
	}
	float ratio=coordDiff/flatDiff;


	float x = line[idx].x+(line[idx+1].x-line[idx].x)*ratio;
	float y = line[idx].y+(line[idx+1].y-line[idx].y)*ratio;
	return sqrt(x*x+y*y);
}
void BezierCircumScale::generate(M_Vec3f& coord, const int h){
	if(!point.length)return;
	if(!flatHeight.length)return;

	M_Vec2f& point=this->point.refValue();

	float ccHeight=point[point.length-1].y/(PI*2);
	float tmpHeight=point[point.length-1].y;
	float bezArcLen=flatHeight[flatHeight.length-1];
	coord.multMatrixLeft(imatrix);

	float level=1.0/(this->level-centerOffset);

////////////////////////////////////////////////////////

	for(int i=0;i<coord.length;i++){
		Vec3f &svec=coord[i];
		float polarPos=sqrt(svec.x*svec.x+svec.z*svec.z)-centerOffset;
		if(abs(polarPos) > .0001){
			float polarAngle=Vec2f(svec.x, svec.z).getAngle3();

			float r = level * (getCircumScaleVal(polarAngle, PI*2)-centerOffset);
			//float r=1;

			svec.z=r*(polarPos)*-sin(polarAngle+PI*.5);
			svec.z+=centerOffset*-sin(polarAngle+PI*.5);

			svec.x=r*(polarPos)*cos(polarAngle+PI*.5);
			svec.x+=centerOffset*cos(polarAngle+PI*.5);
		}
	}
	coord.multMatrixLeft(matrix);
}
float pan(float a, float b, float v){
	return (a * v + b * (1.0 - v));
}
void BezierShift::generate(M_Vec3f& coord, const int h){
	if(!point.length)return;
	if(!flatHeight.length)return;

	float bezArcLen=flatHeight[flatHeight.length-1];

	for(int i=0; i<coord.length; i++){
		Vec3f s_crd=gen_imatrix*coord[i];

		int idx = s_crd.y/bezArcLen*flatHeight.length;

		if(idx>=heightIndex.length){
			idx=heightIndex.length-1;
		}
		else if(idx<0){
			idx=0;
		}
		idx=heightIndex[idx];

		while(idx>0 && s_crd.y < flatHeight[idx-1]){
			idx--;
		}
		float flatDiff=flatHeight[idx];
		float coordDiff=s_crd.y;

		if(idx>0){
			flatDiff-=flatHeight[idx-1];
			coordDiff-=flatHeight[idx-1];
		}
		float ratio=coordDiff/flatDiff;


		float rad = line[idx].x+(line[idx+1].x-line[idx].x)*ratio;

		if(s_crd.x<0){
			if(s_crd.x<-negDecay)continue;
			float sineCurve=cos((fabs(s_crd.x)/negDecay)*PI)*.5+.5;
			s_crd.x+=rad*sineCurve;
		}
		else{
			if(s_crd.x>posDecay)continue;
			float sineCurve=cos((fabs(s_crd.x)/posDecay)*PI)*.5+.5;
			s_crd.x+=rad*sineCurve;
		}
		coord[i]=gen_matrix*s_crd;
	}
}
void BezierTwist::generate(M_Vec3f& coord, const int h){
	if(!point.length)return;
	if(!flatHeight.length)return;

	float bezArcLen=flatHeight[flatHeight.length-1];

	for(int i=0; i<coord.length; i++){
		Vec3f s_crd=imatrix*coord[i];

		int idx = s_crd.y/bezArcLen*flatHeight.length;

		if(idx>=heightIndex.length){
			idx=heightIndex.length-1;
		}
		else if(idx<0){
			idx=0;
		}
		idx=heightIndex[idx];

		while(idx>0 && s_crd.y < flatHeight[idx-1]){
			idx--;
		}
		float flatDiff=flatHeight[idx];
		float coordDiff=s_crd.y;

		if(idx>0){
			flatDiff-=flatHeight[idx-1];
			coordDiff-=flatHeight[idx-1];
		}
		float ratio=coordDiff/flatDiff;

		float rad = line[idx].x+(line[idx+1].x-line[idx].x)*ratio;

		float cosAngle=cos(PI*2*rad);
		float sinAngle=sin(PI*2*rad);
		float newZ=s_crd.z*cosAngle-s_crd.x*sinAngle;
		float newX=s_crd.z*sinAngle+s_crd.x*cosAngle;
		s_crd.z=newZ;
		s_crd.x=newX;

		coord[i]=matrix*s_crd;
	}
}
void BezierTape::generate(M_Vec3f& coord, const int h){
	if(!flatHeight.length)return;

	float bezArcLen=flatHeight[flatHeight.length-1];
	float step=bezArcLen/flatHeight.length;

	if(g_projectDirection!=PROJECT_FORWARD){
		for(int i=coord.length; --i>=0;){
			Vec3f svec = coord[i];

			float polarDistance=sqrt(svec.x*svec.x+svec.z*svec.z);//-point[0].x; //point[0].x is same as a CCLathe's baseRad
			float polarAngle = acos(svec.z/polarDistance);
			if(svec.x < 0){
				//polarAngle = PI*2 - polarAngle;
				polarAngle = -polarAngle;
			}


			int idx = polarDistance/bezArcLen*flatHeight.length;

			if(idx>=heightIndex.length){
				idx=heightIndex.length-1;
			}
			else if(idx<0){
				idx=0;
			}
			idx=heightIndex[idx];

			while(idx>0 && polarDistance < flatHeight[idx-1]){
				idx--;
			}
			float flatDiff=flatHeight[idx];
			float coordDiff=polarDistance;

			if(idx>0){
				flatDiff-=flatHeight[idx-1];
				coordDiff-=flatHeight[idx-1];
			}
			float ratio=coordDiff/flatDiff;


			svec.y = line[idx].y+(line[idx+1].y-line[idx].y)*ratio;
			float rad = line[idx].x+(line[idx+1].x-line[idx].x)*ratio;

			svec.z =-sin(polarAngle-PI*.5)*rad;
			svec.x = cos(polarAngle-PI*.5)*rad;
			coord[i] = svec;
		}
	}
	else{
		for(int i=coord.length; --i>=0;){
			Vec3f svec = coord[i];

			int idx = svec.y/bezArcLen*flatHeight.length;


			if(idx>=heightIndex.length){
				idx=heightIndex.length-1;
			}
			else if(idx<0){
				idx=0;
			}
			idx=heightIndex[idx];

			while(idx>0 && svec.y < flatHeight[idx-1]){
				idx--;
			}
			float flatDiff=flatHeight[idx];
			float coordDiff=svec.y;

			if(idx>0){
				flatDiff-=flatHeight[idx-1];
				coordDiff-=flatHeight[idx-1];
			}
			float ratio=coordDiff/flatDiff;


			svec.y = line[idx].y+(line[idx+1].y-line[idx].y)*ratio;
			float rad = line[idx].x+(line[idx+1].x-line[idx].x)*ratio;

			svec.z = rad;
			//svec.x = cos(svec.x-PI*.5)*rad;
			coord[i] = svec;
		}
	}
}
void BezierBend::generate(M_Vec3f& coord, const int h){
	if(!flatHeight.length)return;

	float bezArcLen=flatHeight[flatHeight.length-1];
	float step=bezArcLen/flatHeight.length;

	Vec3f firstPoint;
	if(point.length){
		Vec3f v2 = point[0];
		firstPoint = Vec3f(0, -v2.y, 0);
	}

	for(int i=coord.length; --i>=0;){
		Vec3f s_crd = gen_imatrix * coord[i];

		int idx = s_crd.y/bezArcLen*flatHeight.length;

		if(idx>=heightIndex.length){
			idx=heightIndex.length-1;
		}
		else if(idx<0){
			idx=0;
		}
		idx=heightIndex[idx];

		while(idx>0 && s_crd.y < flatHeight[idx-1]){
			idx--;
		}
		float flatDiff=flatHeight[idx];
		float coordDiff=s_crd.y;

		if(idx>0){
			flatDiff-=flatHeight[idx-1];
			coordDiff-=flatHeight[idx-1];
		}
		float ratio=coordDiff/flatDiff;

		s_crd.y = line[idx].y+(line[idx+1].y-line[idx].y)*ratio;
		float radX = line[idx].x+(line[idx+1].x-line[idx].x)*ratio;
		float radZ = line[idx].z+(line[idx+1].z-line[idx].z)*ratio;

		Rotation rot = Rotation(Vec3f(0,1,0), line[idx+1] - line[idx]);
		Vec3f v = rot * Vec3f(s_crd.x, 0, s_crd.z);

		s_crd.x = radX;
		s_crd.z = radZ;
		coord[i] = matrix*(v + s_crd);
	}
}
void BezierLathe::inverseSurface(M_Node& generator, BranchJoinNegY& branchJoinNegY){
	int i,j;
	static BezierLathePtr inverse=new BezierLathe;


	M_Vec2f& inversePoint=inverse->point;
	M_Bool& inverseFix=inverse->fix;
	M_Int& inverseNopfs=inverse->nopfs;

	inverseFix.length=inversePoint.length=point.length;

	inverseNopfs.length=nopfs.length;


	for( i=0,j=point.length-1; i<point.length;i++,j--){
		inversePoint[i]=point[j];
		inverseFix[i]=fix[j];
	}
	for( i=0,j=nopfs.length-1; i<nopfs.length;i++,j--){
		inverseNopfs[i]=nopfs[j];
	}
	inverse->processFields();

	BezierLathePtr store=this;	//store pointer is only for stopping this from getting deleted in the next line
	generator[0]=inverse;
	branchJoinNegY.genJoinCurve();
	generator[0]=this;
}
void BezierTape::inverseSurface(M_Node& generator, BranchJoinNegY& branchJoinNegY){
	int i,j;
	static BezierLathePtr inverse=new BezierLathe;


	M_Vec2f& inversePoint=inverse->point;
	M_Bool& inverseFix=inverse->fix;

	inverseFix.length=inversePoint.length=point.length;


	for( i=0,j=point.length-1; i<point.length;i++,j--){
		inversePoint[i]=point[j];
		inverseFix[i]=fix[j];
	}
	inverse->processFields();

	BezierTapePtr store=this;	//store pointer is only for stopping this from getting deleted in the next line
	generator[0]=inverse;
	branchJoinNegY.genJoinCurve();
	generator[0]=this;
}

Vec2f BezierLathe::calcSegmentXY(float t, M_Vec2f& point, int segDim, int bp){
	static M_Vec2f pnt;

	pnt.length=segDim;

	for(int i = 0; i<segDim; i++){
		pnt[i] = point[bp+i];
	}

	for(int j = segDim; --j >= 0;){
		for(int i = 0; i < j; i++){
			float it = 1.0 - t;
			pnt[i] = Vec2f(it, it) * pnt[i] + Vec2f(t, t) * pnt[i+1];
		}
	}
	return pnt[0];
}

void BezierLathe::generate(M_Vec3f& coord, const int h){
//	if(!flatHeight.length)return;

//	float bezArcLen=flatHeight[flatHeight.length-1];
//	float step=bezArcLen/flatHeight.length;
//
	if(g_projectDirection!=PROJECT_FORWARD){

		for(int i=coord.length; --i>=0;){
			Vec3f svec = coord[i];

			float polarDistance=sqrt(svec.x*svec.x+svec.z*svec.z);//-point[0].x; //point[0].x is same as a CCLathe's baseRad
			if(polarDistance==0){
				coord[i].y=0;
				continue;
			}

			float polarAngle = acos(svec.z/polarDistance);
			if(svec.x < 0){
				//polarAngle = PI*2 - polarAngle;
				polarAngle = -polarAngle;
			}

			int b = polarDistance;
			float t = polarDistance-b;
			if(b >= nopfs.length){
				b = nopfs.length-1;
				t = 1;
			}
			Vec2f v = calcSegmentXY(t, point, nopfs[b], m_bp[b]);

			svec.y = v.y;

			//svec.y = line[idx].y+(line[idx+1].y-line[idx].y)*ratio;
			//v.x = line[idx].x+(line[idx+1].x-line[idx].x)*ratio;

			svec.z =-sin(polarAngle-PI*.5)*v.x;
			svec.x = cos(polarAngle-PI*.5)*v.x;
			coord[i] = svec;

/*
			if(polarDistance==0){
				coord[i].y=line[0].y;
				continue;
			}
			float polarAngle = acos(svec.z/polarDistance);
			if(svec.x < 0){
				//polarAngle = PI*2 - polarAngle;
				polarAngle = -polarAngle;
			}


			int idx = polarDistance/bezArcLen*flatHeight.length;

			if(idx>=heightIndex.length){
				idx=heightIndex.length-1;
			}
			else if(idx<0){
				idx=0;
			}
			idx=heightIndex[idx];

			while(idx>0 && polarDistance < flatHeight[idx-1]){
				idx--;
			}
			float flatDiff=flatHeight[idx];
			float coordDiff=polarDistance;

			if(idx>0){
				flatDiff-=flatHeight[idx-1];
				coordDiff-=flatHeight[idx-1];
			}
			float ratio=coordDiff/flatDiff;


			svec.y = line[idx].y+(line[idx+1].y-line[idx].y)*ratio;
			float rad = line[idx].x+(line[idx+1].x-line[idx].x)*ratio;

			svec.z =-sin(polarAngle-PI*.5)*rad;
			svec.x = cos(polarAngle-PI*.5)*rad;
			coord[i] = svec;
			*/
		}
	}
	else{
		for(int i=coord.length; --i>=0;){
			Vec3f svec = coord[i];
/*
			int idx = svec.y/bezArcLen*flatHeight.length;


			if(idx>=heightIndex.length){
				idx=heightIndex.length-1;
			}
			else if(idx<0){
				idx=0;
			}
			idx=heightIndex[idx];

			while(idx>0 && svec.y < flatHeight[idx-1]){
				idx--;
			}
			float flatDiff=flatHeight[idx];
			float coordDiff=svec.y;

			if(idx>0){
				flatDiff-=flatHeight[idx-1];
				coordDiff-=flatHeight[idx-1];
			}
			float ratio=coordDiff/flatDiff;
*/
			int b = svec.y;
			float t = svec.y-b;
			if(b >= nopfs.length){
				b = nopfs.length-1;
				t = 1;
			}
			Vec2f v = calcSegmentXY(t, point, nopfs[b], m_bp[b]);

			svec.y = v.y;

			//svec.y = line[idx].y+(line[idx+1].y-line[idx].y)*ratio;
			//v.x = line[idx].x+(line[idx+1].x-line[idx].x)*ratio;

			svec.z =-sin(svec.x-PI*.5)*v.x;
			svec.x = cos(svec.x-PI*.5)*v.x;
			coord[i] = svec;
		}
	}
}
float BezierLathe::getDistance(int index)const{
	float distance=0;
	for(int i=segDistance.length; --i>=0;){
		distance += segDistance[i];
	}
	return distance;
}
float BezierTape::getDistance(int index)const{
	float distance=0;
	for(int i=segDistance.length; --i>=0;){
		distance += segDistance[i];
	}
	return distance;
}


const Properties& Node::getProp(const int index)const{
	return getProperties()[index];
}
inline void accumList(M_Art& accumOthers, const Art* part){
	if(dynamic_cast<const Part*>(part)){
		for(int r=accumOthers.length; --r>=0;){
			if(accumOthers[r]==part)return;
		}
		accumOthers+=const_cast<Art*>(part);
	}
}
bool commonIllegalImmediateRef(const M_SArt& m_art, const Art* part, M_Art& accumOthers){
	for(int k=m_art.length; --k>=0;){
		if(m_art[k]==part){

			for(int k=m_art.length; --k>=0;){
				const Art* part=m_art[k].p;
				accumList(accumOthers, m_art[k].p);
			}
			return true;
		}
	}
	return false;
}
bool commonIllegalImmediateRef(const Node* part, const Art* p){
	if(part){
		if(part==p){
			return true;
		}
	}
	return false;
}
bool RingBuild::illegalImmediateRef(const Art* p, M_Art& accumOthers)const{
	return commonIllegalImmediateRef(part, p, accumOthers);
}
bool Polymap::illegalImmediateRef(const Art* p, M_Art& accumOthers)const{
	return commonIllegalImmediateRef(part, p);
}
bool NurbsPatch::illegalImmediateRef(const Art* p, M_Art& accumOthers)const{
	return false;
}
bool GroupBuild::illegalImmediateRef(const Art* p, M_Art& accumOthers)const{
	return false;
}
bool TransferOwnership::illegalImmediateRef(const Art* p, M_Art& accumOthers)const{
	if(commonIllegalImmediateRef(source, p) || commonIllegalImmediateRef(dest, p, accumOthers)){
		return true;
	}
	return false;
}
bool JoinVertex::illegalImmediateRef(const Art* p, M_Art& accumOthers)const{
	if(commonIllegalImmediateRef(source, p) || commonIllegalImmediateRef(part, p)){
		return true;
	}
	return false;
}
bool JoinCloseVertex::illegalImmediateRef(const Art* p, M_Art& accumOthers)const{
	if(commonIllegalImmediateRef(source, p, accumOthers) || commonIllegalImmediateRef(part, p, accumOthers)){
		return true;
	}
	return false;
}
bool EvolveCoord::illegalImmediateRef(const Art* p, M_Art& accumOthers)const{
	if(commonIllegalImmediateRef(source, p, accumOthers) || commonIllegalImmediateRef(part, p, accumOthers)){
		return true;
	}
	return false;
}
bool TapeMeasure::illegalImmediateRef(const Art* p, M_Art& accumOthers)const{
	return false;
}
bool PolyToNurbs::illegalImmediateRef(const Art* p, M_Art& accumOthers)const{
	return false; //need something here
}
bool Subdivision::illegalImmediateRef(const Art* p, M_Art& accumOthers)const{
	return commonIllegalImmediateRef(part, p);
}
bool DelSphere::illegalImmediateRef(const Art* p, M_Art& accumOthers)const{
	return commonIllegalImmediateRef(part, p, accumOthers);
}
bool DelPlane::illegalImmediateRef(const Art* p, M_Art& accumOthers)const{
	return commonIllegalImmediateRef(part, p, accumOthers);
}
bool DelPoly::illegalImmediateRef(const Art* p, M_Art& accumOthers)const{
	return commonIllegalImmediateRef(part, p);
}
bool GenPoly::illegalImmediateRef(const Art* p, M_Art& accumOthers)const{
	return commonIllegalImmediateRef(part, p);
}
bool SplitTri::illegalImmediateRef(const Art* p, M_Art& accumOthers)const{
	return commonIllegalImmediateRef(part, p);
}
bool SwapTriDivide::illegalImmediateRef(const Art* p, M_Art& accumOthers)const{
	return commonIllegalImmediateRef(part, p);
}
bool ProjectOntoTri::illegalImmediateRef(const Art* p, M_Art& accumOthers)const{
	return commonIllegalImmediateRef(part, p, accumOthers);
}
bool Place::illegalImmediateRef(const Art* p, M_Art& accumOthers)const{
	if(staticTransform){
		return commonIllegalImmediateRef(part, p, accumOthers);
	}
	else return false;
}
bool PlaceTexture::illegalImmediateRef(const Art* p, M_Art& accumOthers)const{
	return commonIllegalImmediateRef(part, p, accumOthers);
}
bool BranchJoin::illegalImmediateRef(const Art* p, M_Art& accumOthers)const{
	return commonIllegalImmediateRef(part, p, accumOthers);
}
bool SphereBuild::illegalImmediateRef(const Art* p, M_Art& accumOthers)const{
	return commonIllegalImmediateRef(part, p);
}
bool CylinderBuild::illegalImmediateRef(const Art* p, M_Art& accumOthers)const{
	return commonIllegalImmediateRef(part, p);
}
bool ConeBuild::illegalImmediateRef(const Art* p, M_Art& accumOthers)const{
	return commonIllegalImmediateRef(part, p);
}
bool TorusBuild::illegalImmediateRef(const Art* p, M_Art& accumOthers)const{
	return commonIllegalImmediateRef(part, p);
}
bool BoxBuild::illegalImmediateRef(const Art* p, M_Art& accumOthers)const{
	return commonIllegalImmediateRef(part, p);
}
bool RectangleBuild::illegalImmediateRef(const Art* p, M_Art& accumOthers)const{
	return commonIllegalImmediateRef(part, p);
}
bool BezierLatheBuild::illegalImmediateRef(const Art* p, M_Art& accumOthers)const{
	return commonIllegalImmediateRef(part, p);
}
bool NurbsLathe::illegalImmediateRef(const Art* p, M_Art& accumOthers)const{
	return commonIllegalImmediateRef(part, p);
}
bool BendyLathe::illegalImmediateRef(const Art* p, M_Art& accumOthers)const{
	return commonIllegalImmediateRef(part, p);
}
bool CenterJoint::illegalImmediateRef(const Art* p, M_Art& accumOthers)const{
	return commonIllegalImmediateRef(part, p);
}
bool TransferVertex::illegalImmediateRef(const Art* p, M_Art& accumOthers)const{
	if(commonIllegalImmediateRef(source, p) || commonIllegalImmediateRef(dest, p)){
		return true;
	}
	return false;
}
bool TransferBCyl::illegalImmediateRef(const Art* p, M_Art& accumOthers)const{
	if(commonIllegalImmediateRef(source, p) || commonIllegalImmediateRef(dest, p)){
		return true;
	}
	return false;
}
void RingBuild::doBuild(){
	const M_Int& ownership=this->ownership.refValue();
	M_SArt& m_part=this->part.refValue();
	if(part.length > 0){
		int partIndex=0;
		int ownershipIdx=0;
		int ownershipCntr=-1;
		if(ownership.length){
			ownershipCntr=ownership[0];
		}
		Art* prevPart=dynamic_cast<Art*>(m_part[0].p);
		Art* part=prevPart;

		int lastIndex=scaleX.length-1;

		bool beginSP = scaleX[0].x == 0 && scaleX[0].y == 0 && scaleZ[0].x == 0 && scaleZ[0].y == 0 ? true : false;
		bool endSP = scaleX[lastIndex].x == 0 && scaleX[lastIndex].y == 0 && scaleZ[lastIndex].x == 0 && scaleZ[lastIndex].y == 0 ? true : false;

		//Matrix matrix;
		int markC = 0;
		Matrix baseMatrix = scaleMatrix(scale.refValue()) * rotationMatrix(rotation.refValue()) * translationMatrix(translation.refValue());
		for(int k=0; k<scaleX.length; k++){
			bool lastRing = k == scaleX.length - 1;
			Vec2f scaleX = this->scaleX.refValue()[k];
			Vec2f scaleZ = this->scaleZ.refValue()[k];
			Matrix matrix = rotationMatrix(orientation.refValue()[k]) * translationMatrix(location.refValue()[k]) * baseMatrix;
			M_Vec3f& coord = part->coord;
			int idx=coord.length;
			ColorEffect* colorEffect=dynamic_cast<ColorEffect*>(part->workEffect);
			M_Color3f& color = colorEffect->color.refValue();
			markC = color.length;
			//Color3f newCol=vertexRing->color;
			if((beginSP && k == 0) || (endSP && lastRing)){
				coord += matrix * Vec3f(0,0,0);
				color.length++;
			}
			else{
				coord.length+=staves;
				for(int s=0; s<staves; s++, idx++){
					float angle = static_cast<float>(s)/static_cast<float>(staves)*PI*2+PI*.5;
					Vec3f s_crd = Vec3f(cos(angle), 0, -sin(angle));
					if(s_crd.x<=0){
						s_crd.x*=scaleX.x;
					}
					else{
						s_crd.x*=scaleX.y;
					}
					if(s_crd.z<=0){
						s_crd.z*=scaleZ.x;
					}
					else{
						s_crd.z*=scaleZ.y;
					}
					coord[idx] = matrix * s_crd;
				}
				color.length+=staves;
//					for(int c = markC; c < color.length; c++){
//						color[c] = newCol;
//					}
			}
			if(k>0){
				M_Triangle& triangle = prevPart->plidx < part->plidx ? prevPart->triangle : part->triangle;
				int samePartOffset = 0;
//				int triOffset=part->triangle.length;
//				int prevTriOffset=prevPart->triangle.length;
				if(prevPart == part){
					samePartOffset = staves;
				}
				if(beginSP && (k == 1)){
					genIndicesJoinEndsBeginSP(prevPart, part, triangle, staves.refValue(), markC);
				}
				else if(endSP && lastRing){
					genIndicesJoinEndsEndSP(prevPart, part, triangle, staves.refValue(), markC);
				}
				else{
					genIndicesJoinEnds(prevPart, part, triangle, staves.refValue(), markC);
				}
//				if(insideOut){
//					prevPart->plidx < part->plidx ? prevPart->insideOut(prevTriOffset) : part->insideOut(triOffset);
//				}
			}
			prevPart=part;
			addRelevantPart(part->relevantPart, part);
			addRelevantPart(part->relevantPart, prevPart);
			if(ownershipIdx < ownership.length){
				if(--ownershipCntr==0){
					if(++ownershipIdx < m_part.length){
						part = m_part[ownershipIdx].p;
					}
					if(ownershipIdx < ownership.length){
						ownershipCntr = ownership[ownershipIdx];
					}
				}
			}
		}
		part->smls->resetRelevantParts();
		//part->smls->verifyIntegrity();
	}
}
void Art::CoordGroup::operator=(const M_Int& par){
	refValue()=par;
	if(Art* mirrorPart = home.mirrorPart){
		mirrorPart->coordGroup.refValue() = par;
	}
	home.smls->convertToRenderFormat();
}
void Art::Weight::operator=(const M_Float& par){
	refValue()=par;
	home.smls->convertToRenderFormat();
	Art* sisterPart = 0;
	if(Art* mirrorPart = home.mirrorPart){
		mirrorPart->weight.refValue() = par;
	}
	home.smls->convertToRenderFormat();
}

Vec3f BezierPatch::calculateAltV(float t,int row, int numOfU, int numOfV, int bp){//do V first
	const M_Vec3f& point = this->point;

	static M_Vec3f pnt;

	pnt.length=numOfV;

	for(int i=0; i<numOfV; i++){
		pnt[i] = point[i*numOfU+row+bp];
	}
	for (int j = numOfV; --j >= 0;){
		for (int i = 0; i < j; i++){
			float it=1-t;
			pnt[i] = Vec3f(it, it, it) * pnt[i] + Vec3f(t, t, t) * pnt[i + 1];
		}
	}
	return pnt[0];
}


Vec3f BezierPatch::calculateUEdge02(float t, int row, int idx, int numOfU, int bp){
	const M_Vec3f& point = this->point;

	static M_Vec3f pnt;

	pnt.length=numOfU;

	float m = 1.0f / (numOfU-1) * idx;
	float m2 = 1.0f - m;
	int len=idx;
	int len2=numOfU-len;
	len++;
	int g=len;
	int r = row*numOfU+bp;

	if(t < m){
		t = t * (1.0f/m);
		for(int i=0; i<len; i++){
			pnt[i] = point[r+i];
		}
	}
	else{
		g=len2;
		t = (t - m) * (1.0f/m2);
		for(int i=0; i<len2; i++){
			pnt[i] = point[r+i+idx];
		}
	}
	for(int j = g; --j >= 0;){
		for(int i = 0; i < j; i++){
			float it=1-t;
			pnt[i] = Vec3f(it, it, it) * pnt[i] + Vec3f(t, t, t) * pnt[i + 1];
		}
	}
	return pnt[0];
}

Vec3f BezierPatch::calculateU(float t, int row, int numOfU, float& rw, int bp){
	const M_Vec3f& point = this->point;
	const M_Float& weight = this->weight;

	static M_Vec3f pnt;
	static M_Float w;

	w.length=pnt.length=numOfU;

	int re=row*numOfU+bp;
	for(int i=0; i<numOfU; i++){
		pnt[i] = point[re+i];
		w[i] = weight[re+i];
	}
	for (int j = numOfU; --j >= 0;){
		for (int i = 0; i < j; i++){
			float it = 1 - t;
			pnt[i] = Vec3f(it, it, it) * w[i] * pnt[i] + Vec3f(t, t, t) * w[i+1] * pnt[i+1];
			w[i] = it * w[i] + t * w[i+1];
			pnt[i] /= w[i];
		}
	}
	rw = w[0];
	return pnt[0];
}
Vec3f BezierPatch::calculateV(float t, M_Vec3f& pnt, M_Float& w, int bp){

	for(int j = pnt.length; --j >= 0;){
		for(int i = 0; i < j; i++){
			float it = 1 - t;
			pnt[i] = Vec3f(it, it, it) * w[i] * pnt[i] + Vec3f(t, t, t) * w[i+1] * pnt[i+1];
			w[i] = it * w[i] + t * w[i+1];
			pnt[i] /= w[i];
		}
	}
	return pnt[0];
}
/*
Vec3f BezierPatch::calculateU(float t, int row, int numOfU, int bp){
	static M_Vec3f pnt;
	static M_Float w;

	w.length = pnt.length = numOfU;

	int re=row*numOfU+bp;
	for(int i=0; i<numOfU; i++){
		pnt[i] = point[re+i];
		w[i] = weight[re+i];
	}
	for (int j = numOfU; --j >= 0;){
		for (int i = 0; i < j; i++){
			float it=1-t;
			pnt[i] = Vec3f(it, it, it) * w[i] * pnt[i] + Vec3f(t, t, t) * w[i+1] * pnt[i+1];
		}
	}
	return pnt[0];
}
Vec3f BezierPatch::calculateV(float t, M_Vec3f& pnt, M_Float& w, int bp){
	for(int j = pnt.length; --j >= 0;){
		for(int i = 0; i < j; i++){
			float it=1-t;
			pnt[i] = Vec3f(it, it, it) * w[i] * pnt[i] + Vec3f(t, t, t) * w[i+1] * pnt[i+1];
		}
	}
	return pnt[0];
}
*/
//2, 3
//1, 2

void BezierPatch::generate(M_Vec3f& coord, const int h){

	static M_Vec3f pnt;
	static M_Float w;

	for(int i=coord.length; --i>=0;){
		Vec3f t = coord[i];
		int zx = t.x;
		int zy = t.y;

		if(zx>=zoneDim.x){
			zx = zoneDim.x-1;
			t.x = zoneDim.x;
		}
		if(zy>=zoneDim.y){
			zy = zoneDim.y-1;
			t.y = zoneDim.y;
		}

		t.x-=zx;
		t.y-=zy;
		//if(t.x == 0 && zx!=0){
		//	t.x = 1;
		//}
		//if(t.y == 0 && zy!=0){
		//	t.y = 1;
		//}

		int idx = zoneLookUp[zy*zoneDim.x+zx];
		if(idx < 0){
			if(t.x == 0){
				zx--;
				t.x = 1;
				idx = zoneLookUp[zy*zoneDim.x+zx];
				if(idx < 0){
					if(t.y == 1){
						zy--;
						t.y = 1;
					}
					idx = zoneLookUp[zy*zoneDim.x+zx];
				}
			}
			else if(t.y == 0){
				zy--;
				t.y = 1;
				idx = zoneLookUp[zy*zoneDim.x+zx];
			}
		}

		int bp = m_bp[idx];
		int numOfU = dim[idx].x;
		int numOfV = dim[idx].y;

		w.length = pnt.length = numOfV;


		for(int k=0; k<numOfV; k++){
			pnt[k] = calculateU(t.x, k, numOfU, w[k], bp); //create a horizontal bezier line (to create control points) for t.x
		}

		coord[i] = calculateV(t.y, pnt, w, bp);
	}
}
/*
void BezierPatch::simGenerate(M_Vec3f& coord){

	static M_Vec3f temp;
	temp.length=numOfV;

	for(int i=coord.length; --i>=0;){

		Vec3f t = coord[i];

		for(int k=0; k<numOfV; k++){
			temp[k] = calculateU(t.x, k); //create a vertical bezier line (to create control points) for t.x
		}

		coord[i] = calculateV(t.y, temp);
	}
}
*/
void BezierPatch::simGenerateExtra(M_Vec3f& coord, int numOfU, int numOfV, int bp){

	static M_Vec3f temp;
	static M_Float w;

	for(int i=coord.length; --i>=0;){
		Vec3f t = coord[i];
		int zx = t.x;
		int zy = t.y;

		int idx = zoneLookUp[zy*zoneDim.y+zx];
		int bp = m_bp[idx];
		int numOfU = dim[idx].x;
		int numOfV = dim[idx].y;

		temp.length=numOfV;

		t.x-=zx;
		t.y-=zy;
/*
		if(fixIdx.length==1){
			if(fixIdx.refValue()[0]<numOfU){

				temp[0] = calculateUEdge02(t.x, 0, fixIdx.refValue()[0], numOfU, bp); //create first control point

				for(int k=1; k<numOfV; k++){
					temp[k] = calculateU(t.x, k, numOfU, bp); //create a vertical bezier line (to create control points) for t.x
				}

				coord[i] = calculateV(t.y, temp, bp);
			}
			else{

				for(int k=0; k<numOfV-1; k++){
					temp[k] = calculateU(t.x, k, numOfU, bp); //create a vertical bezier line (to create control points) for t.x
				}

				temp[numOfV-1] = calculateUEdge02(t.x, numOfV-1, fixIdx.refValue()[0]-numOfU*(numOfV-1), numOfU, bp); //create first control point

				coord[i] = calculateV(t.y, temp, bp);
			}
		}
		else*/
		{
//			for(int k=0; k<numOfV; k++){
//				temp[k] = calculateU(t.x, k, numOfU, bp); //create a vertical bezier line (to create control points) for t.x
//			}

//			coord[i] = calculateV(t.y, temp, w, bp);

/*

			temp[0] = calculateUEdge02(t.x, 0, fixIdx.refValue()[0], numOfU, bp); //create first control point

			for(int k=0; k<numOfV; k++){
				temp[k] = calculateU(t.x, k, numOfU, bp); //create a vertical bezier line (to create control points) for t.x
			}

			temp[numOfV-1] = calculateUEdge02(t.x, numOfV-1, fixIdx.refValue()[1]-numOfU*(numOfV-1), numOfU, bp); //create first control point

			coord[i] = calculateV(t.y, temp, bp);
*/
		}
	}
}
/*
void BezierPatch::generate(M_Vec3f& coord){
//	static Stem stem;

	M_Vec2f& dim = this->dim.refValue();
//	M_Node& source = this->source.refValue();

//	M_Node& task=this->task;
//	for(int i=0; i<task.length; i++){
//		dynamic_cast<PatchTask*>(task[i].p)->doTask(point.refValue(), numOfU, numOfV);
//	}
//	stem.part.refValue().length=1;

//	stem.xDistance=1;
//	stem.yDistance=1;
//	stem.uWrap=false;

	for(int k=0, bp=0; k<dim.length; k++){
		if(k<this->part.refValue().length){
			if(Art* part=dynamic_cast<Art*>(this->part[k].p)){
				simGenerate(coord.refValue());
			}

//			int noU=dim[k].x;
//			int noV=dim[k].y;
//			if(k<source.length && dynamic_cast<Art*>(source[k].p)){
//				stem.source.refValue().length=1;
//				stem.source.refValue()[0]=dynamic_cast<Art*>(source[k].p);
//				stem.slide=Vec2f(0,0);
//			}
//			else{
//				stem.slide=Vec2f(.5,0);
//				stem.source.refValue().length=0;
//			}
//			stem.staves.refValue()=staves[k];
//			stem.bands.refValue().length=1;
//			stem.bands.refValue()[0]=bands[k];

//			stem.part.refValue()[0]=part;

			//checkSinglePoint(0, numOfU[k], point, stem.beginSP.refValue());
			//checkSinglePoint(numOfV[k]-1, numOfU[k], point, stem.endSP.refValue());

//			stem.doBuild();

			//if(fixIdx.refValue().length==0){
			//	simGenerate(part->coord.refValue(), noU, noV, bp);
			//}
			//else{
			//	simGenerateExtra(part->coord.refValue(), noU, noV, bp);
			//}
			//bp+=noU*noV;
		}
	}
}
*/
int BezierPatch::getBasePoint(int par)const{
	int bp=0;
	for(int k=0; k<par; k++){
		bp+=dim[k].x*dim[k].y;
	}
	return bp;
}

int getEdge2(int i, int numOfU, int numOfV){
	return i*numOfU+numOfU-1;
}
/*
void PatchLink::doTask(M_Vec3f& pointA, int uPointsA, int vPointsA){
	if(BezierPatch* patchB = dynamic_cast<BezierPatch*>(this->patchB.p)){
		M_Vec3f& pointB = patchB->point;
		M_Int& hPointA = this->hPointA;
		M_Int& sPointA = this->sPointA;
		M_Int& hPointB = this->hPointB;
		M_Int& sPointB = this->sPointB;

		for(int i=0; i<hPointA.length; i++){
			pointA[hPointA[i]]=pointB[hPointB[i]];
			pointA[sPointA[i]]=pointB[hPointB[i]]-pointB[sPointB[i]]+pointB[hPointB[i]];
		}
	}
	else if(NurbsPatch* patchB = dynamic_cast<NurbsPatch*>(this->patchB.p)){
		M_Vec3f& pointB = patchB->point;
		M_Int& hPointA = this->hPointA;
		M_Int& sPointA = this->sPointA;
		M_Int& hPointB = this->hPointB;
		M_Int& sPointB = this->sPointB;

		for(int i=0; i<hPointA.length; i++){
			pointA[hPointA[i]]=pointB[hPointB[i]];
			pointA[sPointA[i]]=pointB[hPointB[i]]-pointB[sPointB[i]]+pointB[hPointB[i]];
		}
	}
}
*/
inline void setComponents(Vec3f& dstVec, Vec3f srcVec, const bool& x, const bool& y, const bool& z){
	if(x){
		dstVec.x=srcVec.x;
	}
	if(y){
		dstVec.y=srcVec.y;
	}
	if(z){
		dstVec.z=srcVec.z;
	}
}
#define GNURBS_MAX_ORDER 16


int FindSpan (int n, int p, float u, const float *U);
void BasisFuns ( int i, float u, int p, const float *U, float *Nf );
bool SurfacePoint(
				  float u, float v, int uOrder, int vOrder, int uDimension, int vDimension,
				  const float *uKnot, const float * vKnot,
				  const Vec3f *P,const float *W,
				  Vec3f *dest // result
				  );
void CurvePoint (int numOf, int degree, const float *knot, const Vec3f* point, const float *weight, const float t, Vec3f& outPoint);

Vec3f curvePoint(const int numOf, const int degree, const M_Float& knot, const M_Vec3f& point, const M_Float& weight, const float t){
	//Compute point on rational B-spline curve
	//Input: n,p,U,P,w, u

	// n: dimension (number of points)
	// p: degree (3)
	// U: knot vector
	// P: control Points
	// w: weights
	// u: fraction

	//Output: Cx, Cy, Cz

	float Nf[GNURBS_MAX_ORDER];
	float Cw=0, Cwx=0, Cwy=0, Cwz=0;

	int span = FindSpan (numOf, degree, t, knot.getArray());
	BasisFuns (span, t, degree, knot.getArray(), Nf);

	if (weight.length == 0) Cw= 1.0f;

	for(int i=0; i<=degree; i++){
		Cwx = Cwx +  Nf[i] * point[(span-degree+i)].x;
		Cwy = Cwy +  Nf[i] * point[(span-degree+i)].y;
		Cwz = Cwz +  Nf[i] * point[(span-degree+i)].z;
		if(weight.length) Cw  = Cw  +  Nf[i]  * weight[span-degree+i];
	}

	Cw = 1.0f/Cw;
	return Vec3f(Cwx*Cw, Cwy*Cw, Cwz*Cw);
}
Vec2f curvePoint2d(int dim, int degree, const float *knot, const Vec2f *pnt, const float *w, const float t);

void FuseSurface::storeV(M_Vec3f& coord){
	int count = baseCoord.length;
	if(count >= coord.length){
		count = coord.length;
	}
	vPan.length=count;
	for(int i=0; i<count; i++){
		float f = coord[i].y;
		if(f>0 && f <=1){
			vPan[i] = f;
		}
	}
}

/*

void JoinSurface::calcEdge(M_Vec3f& coord, NurbsPatch& branchPatch){
	M_Vec2f& zeroIn = this->zeroIn.refValue();
	const int dim=point.refValue().length>>1;
	NurbsPatch* stemPatch = dynamic_cast<NurbsPatch*>(this->patch.p);

	const Vec2f multiply = this->multiply;

	int count = zeroIn.length>>1;

	Vec2f* p = point.getArray();
	float* w=weight.getArray();

	for(int i=0; i<count; i++){

		float ringT = (maxT()-minT())*coord[i].x + minT();
		Vec2f projectPoint = curvePoint2d(dim-1, degree, knot.getArray(), p, w, ringT);
		Vec2f projectPoint2 = curvePoint2d(dim-1, degree, knot.getArray(), p+dim, w+dim, ringT);
		Vec2f projectDiff = projectPoint2-projectPoint;
		float longDistance=projectDiff.length();
		float smallDistance=.0001;
		Vec2f offset=projectDiff*smallDistance;//(smallDistance*(1.0/longDistance));
		offset=projectDiff*(smallDistance*(1.0/longDistance));
		projectPoint2=projectPoint+offset;
		//smallDistance=offset.length();

		float u = (stemPatch->maxU()-stemPatch->minU())*zeroIn[i].x + stemPatch->minU();
		float v = (stemPatch->maxV()-stemPatch->minV())*zeroIn[i].y + stemPatch->minV();

		float u2 = (stemPatch->maxU()-stemPatch->minU())*zeroIn[i+count].x + stemPatch->minU();
		float v2 = (stemPatch->maxV()-stemPatch->minV())*zeroIn[i+count].y + stemPatch->minV();

		Vec3f curvedEdgePoint = stemPatch->calculate(u, v);
		Vec3f curvedEdgePoint2 = stemPatch->calculate(u2, v2);

		{
			Vec3f t = coord[i];

			float u = (branchPatch.maxU()-branchPatch.minU())*t.x + branchPatch.minU();

			// calculate the parametric v value
			float v = (branchPatch.maxV()-branchPatch.minV())*t.y + branchPatch.minV();

			Vec3f d = curvedEdgePoint-curvedEdgePoint2;
			Vec3f e = curvedEdgePoint - d * (longDistance/smallDistance);
			coord[i]=branchPatch.calculate(u, v, curvedEdgePoint, e);
		}

		Vec2f diff = projectPoint - Vec2f(curvedEdgePoint.x, curvedEdgePoint.y);
		zeroIn[i].x += diff.x * multiply.x;
		zeroIn[i].y += diff.y * multiply.y;

		Vec2f diff2 = projectPoint2 - Vec2f(curvedEdgePoint2.x, curvedEdgePoint2.y);
		zeroIn[i+count].x += diff2.x * multiply.x;
		zeroIn[i+count].y += diff2.y * multiply.y;
	}
}

void NurbsPatch::generate(M_Vec3f& coord){
	int numOfU=dim.x;

	static M_Vec3f temp;
	temp.length=numOfU;

	M_Node& m_joinSurface = joinSurface.refValue();

	if(m_joinSurface.length){
		JoinSurface* joinSurface = dynamic_cast<JoinSurface*>(m_joinSurface[0].p);
		joinSurface->calcEdge(coord, *this);
	}
	else{
		for(int i=coord.length; --i>=0;){

			Vec3f t = coord[i];

			float u = (maxU()-minU())*t.x + minU();

			// calculate the parametric v value
			float v = (maxV()-minV())*t.y + minV();

			coord[i]=calculate(u, v);
		}
	}
}
*/


const float NurbsLathe::initPointData[]={
	.18, 0,		.41, 0,		.55, .35,	.13, .66,	.11, .82,	.36, 1,		.093679, 1
};

const float BendyLathe::initPointData[]={
	0, 0, 0,	0, .1, 0,	0, .3, 0,	0, .5, 0,	0, .7, 0,	0, .9, 0,	0, 1, 0
};
void addFacePoints(const M_Int& coordIndex, const M_Vec3f& point, M_Vec3f& coord){
	for(int i=0; i<coordIndex.length; i+=4){
		coord+=(point[coordIndex[i+0]]+point[coordIndex[i+1]]+point[coordIndex[i+2]]+point[coordIndex[i+3]])*.25;
	}
}
class SubdivEdge{
public:
	int b;
	int mid;
	int next;
	int poly0;
	int poly1;
	SubdivEdge():
		b(-1),
		mid(0),
		next(-1),
		poly0(-1),
		poly1(-1)
	{}
	SubdivEdge(int b, int mid, int poly0):
		b(b),
		mid(mid),
		next(-1),
		poly0(poly0),
		poly1(-1)
	{}
	bool initialised()const{
		return b!=-1;
	}

};
typedef M_<SubdivEdge> M_Edge;

bool getMatchingEdge(const int a, const int b, int& i, const M_Edge& edge){
	i = a;
	while(true){
		if(edge[i].b == b){
			return true;
		}
		i=edge[i].next;
		if(i==-1){
			return false;
		}
	}
}
void setPoly(const int a, const int b, M_Edge& edge, int poly){
	int i = a;
	while(true){
		if(edge[i].b == b){
			if(edge[i].poly0==-1 && edge[i].poly1 != poly){
				edge[i].poly0=poly;
			}
			else if(edge[i].poly1==-1 && edge[i].poly0 != poly){
				edge[i].poly1=poly;
			}
		}
		i=edge[i].next;
		if(i==-1){
			return;
		}
	}
}
int addEdge(int a, int b, const M_Vec3f& point, M_Vec3f& coord, int edgeBase, M_Vec3f& midPoint, M_Edge& edge, Vec3f facePoint, int poly){//returns midpoint index
	int i=0;
	if(getMatchingEdge(a, b, i, edge) || getMatchingEdge(b, a, i, edge)){
		int m = edge[i].mid;

		setPoly(a, b, edge, poly);
		setPoly(b, a, edge, poly);

		coord[m]=(coord[m]+facePoint + point[a] + point[b])*.25;
		midPoint[m-edgeBase]=(point[a] + point[b])*.5;

		return m;
	}
	else{
		int len=edge.length;
		int coordLen=coord.length;

		if(edge[a].initialised()){
			while(edge[a].next != -1){
				a = edge[a].next;
			}
			edge[a].next=len;
			edge+=SubdivEdge(b, coordLen, poly);
			len++;
		}
		else{
			edge[a]=SubdivEdge(b, coordLen, poly);
		}

		if(edge[b].initialised()){
			while(edge[b].next != -1){
				b = edge[b].next;
			}
			edge[b].next=len;
			edge+=SubdivEdge(a, coordLen, poly);
		}
		else{
			edge[b]=SubdivEdge(a, coordLen, poly);
		}

		coord+=facePoint;
		midPoint+=facePoint;

		return coordLen;
	}
}
void addMidPoints(M_Int& newCoordIndex, const M_Int& coordIndex, const M_Vec3f& point, M_Vec3f& coord, M_Vec3f& midPoint, M_Edge& edge){
	int polyLen=coordIndex.length>>2;
	int pointLen=point.length;
	int edgeBase=pointLen+polyLen;

	for(int i=0; i<polyLen; i++){
		int j=i*4;
		int ma=addEdge(coordIndex[j+0], coordIndex[j+1], point, coord, edgeBase, midPoint, edge, coord[pointLen+i], i);
		int mb=addEdge(coordIndex[j+1], coordIndex[j+2], point, coord, edgeBase, midPoint, edge, coord[pointLen+i], i);
		int mc=addEdge(coordIndex[j+2], coordIndex[j+3], point, coord, edgeBase, midPoint, edge, coord[pointLen+i], i);
		int md=addEdge(coordIndex[j+3], coordIndex[j+0], point, coord, edgeBase, midPoint, edge, coord[pointLen+i], i);

		//newCoordIndex+=coordIndex[j+0];
		//newCoordIndex+=ma;
		//newCoordIndex+=pointLen+i; //face point
		//newCoordIndex+=md;

		newCoordIndex+=md;
		newCoordIndex+=coordIndex[j+0];
		newCoordIndex+=ma;
		newCoordIndex+=pointLen+i; //face point

		newCoordIndex+=ma;
		newCoordIndex+=coordIndex[j+1];
		newCoordIndex+=mb;
		newCoordIndex+=pointLen+i; //face point

		newCoordIndex+=mb;
		newCoordIndex+=coordIndex[j+2];
		newCoordIndex+=mc;
		newCoordIndex+=pointLen+i; //face point

		newCoordIndex+=mc;
		newCoordIndex+=coordIndex[j+3];
		newCoordIndex+=md;
		newCoordIndex+=pointLen+i; //face point

		//		coord+=(point[coordIndex[j+0]]+point[coordIndex[j+1]])*.5;
//		coord+=(point[coordIndex[j+1]]+point[coordIndex[j+2]])*.5;
//		coord+=(point[coordIndex[j+2]]+point[coordIndex[j+3]])*.5;
//		coord+=(point[coordIndex[j+3]]+point[coordIndex[j+0]])*.5;
	}
}
void addNewCoordIndex(M_Int& newCoordIndex, const M_Int& coordIndex, const M_Vec3f& point, const M_Vec3f& coord, const M_Edge& edge, const int midBase){
	int polyLen=coordIndex.length;
	int pointLen=point.length;

	for(int i=0; i<polyLen; i++){
		int j=polyLen*4;
		newCoordIndex+=coordIndex[j+0];
		newCoordIndex+=pointLen+j;
		newCoordIndex+=pointLen+midBase+i;
	}
}
void addPolyIndex(int poly, M_Int& facePoints){
	if(poly==-1)return;
	for(int i=facePoints.length; --i>=0;){
		if(facePoints[i]==poly){
			return;
		}
	}
	facePoints+=poly;
}
Vec3f getSuroundingFaceAverage(const M_Vec3f& coord, const int pointLen, const M_Int& facePoints, const int n){
	Vec3f average;
	for(int i=n; --i>=0;){
		average+=coord[facePoints[i]+pointLen];
	}
	return average/n;
}
Vec3f getSuroundingMidPointAverage(int i, const M_Vec3f& midPointCoord, const int edgeBase, const M_Edge& edge){
	Vec3f average;
	if(edge[i].initialised()==false){
		return average;
	}
	int n=0;
	do{
		n++;
		average += midPointCoord[edge[i].mid - edgeBase];
	}while((i=edge[i].next)!=-1);

	return average/n;
}
void repositionOldPoints(const M_Int& coordIndex, const M_Vec3f& point, M_Vec3f& coord, M_Vec3f& midPoint, const M_Edge& edge){
	static M_Int facePoints;

	int polyLen=coordIndex.length>>2;
	int pointLen=point.length;
	int edgeBase=pointLen+polyLen;


	for(int i=0; i<point.length; i++){
		facePoints.length=0;

		int e=i;
		do{
			addPolyIndex(edge[e].poly0, facePoints);
			addPolyIndex(edge[e].poly1, facePoints);
		}while((e=edge[e].next)!=-1);

		int n = facePoints.length;
		Vec3f f = getSuroundingFaceAverage(coord, pointLen, facePoints, n);
		Vec3f m = getSuroundingMidPointAverage(i, midPoint, edgeBase, edge);
		Vec3f p = point[i];
		coord[i]=(f+m*2+p*(n-3))/n;


//		coord[i]=(Q+R*2+S*(n-3))/n;
//		coord[i]=S*(n-3)/n+(Vec3f(1,1,1)/n)*Q+R;
//		coord[i]= v * ((n-2)/n) + R * (1/pow(n,2.0)) + f * (1/pow(n, 2.0));
		//coord[i]=Q;
	}
}
/*
void add4Points(int oa, int ob, int oc, M_Vec3f& point, M_Vec3f& coord){
	coord+=point[oa]+(point[ob]-point[0])*.5;
	coord+=point[0];
	coord+=point[0]+(point[3]-point[0])*.5;
	coord+=(point[0]+point[1]+point[2]+point[3])*.25;
}
*/
void Subdivision::doBuild(){
	static M_Int newCoordIndex;
	static M_Int oldCoordIndex;
	static M_Edge edge;
	static M_Vec3f midPoint;
	static M_Vec3f oldPoint;



	oldCoordIndex=coordIndex;
	oldPoint=point;


	Art* part=this->part.refValue();
	if(part==0)return;


	M_Vec3f& coord = part->coord;

	for(int cycleCtr=0;;){
		newCoordIndex.length=0;
		edge.length=0;
		edge.length=oldPoint.length;
		midPoint.length=0;


		//add4Points(0, 1, 3, point, coord);
		coord=oldPoint;
		addFacePoints(oldCoordIndex, oldPoint, coord);
		addMidPoints(newCoordIndex, oldCoordIndex, oldPoint, coord, midPoint, edge);

		repositionOldPoints(oldCoordIndex, oldPoint, coord, midPoint, edge);

		if(++cycleCtr>=cycle){
			break;
		}
		oldCoordIndex=newCoordIndex;
		oldPoint=coord;
	}

	Effect* effect=part->workEffect;
	int effectLength=effect->getEffectLength();
	if(effect->getType()==COLOREFFECT){
		dynamic_cast<ColorEffect*>(effect)->color.refValue().length+=newCoordIndex.length;
	}
	else if(effect->getType()==TEXTUREEFFECT){
		dynamic_cast<TextureEffect*>(effect)->texCoord.refValue().length+=newCoordIndex.length;
	}
	else throw;

	for(int i=0; i<newCoordIndex.length; i+=4){
		part->triangle+=S_Triangle(part, newCoordIndex[i+0], effectLength+i+0, part, newCoordIndex[i+1], effectLength+i+1, part, newCoordIndex[i+2], effectLength+i+2);
		part->triangle+=S_Triangle(part, newCoordIndex[i+0], effectLength+i+0, part, newCoordIndex[i+2], effectLength+i+2, part, newCoordIndex[i+3], effectLength+i+3);
	}

	addRelevantPart(part->relevantPart, part);
}
int Subdivision::initCoordIndexData[24]={
	0, 1, 3, 2,
	5, 4, 6, 7,
	4, 0, 2, 6,
	1, 5, 7, 3,
	4, 5, 1, 0,
	2, 3, 7, 6
};

void NurbsLathe::dpointsToPoints(){
	M_Vec2f& point = this->point.refValue();
	M_Vec2f& dpoint = this->dpoint.refValue();

	point.length = dpoint.length;

	for(int i = dpoint.length; --i>=0;){
		point[i] = dpoint[i] * weight[i];
	}
}


void NurbsLathe::Weight::operator=(const M_Float& m_par){
	refValue() = m_par;
	int index = home.index[0];
	home.point.refValue()[index]=home.dpoint[index] * home.weight[index];
}

void NurbsLathe::DPoint::operator=(const M_Vec2f& m_par){
	int idx = home.index[0];
	float w = 1;
	if(home.weight.length){
		w = home.weight[idx];
	}
	home.point.refValue()[idx]=m_par[idx]*w;
	home.processFields();
	home.houseKeep();
}

Node* NurbsLathe::HSFPart::operator=(Node* par){
	SFNode::refValue()=par;
	home.processFields();
	return par;
};

/*
void NurbsLathe::setStemValues(Stem* stem){
	stem->sync.refValue() = true;
	stem->beginSP = beginSP;
	stem->endSP = endSP;
	stem->staves.refValue()[0]=staves;
	stem->bands.refValue().length = 0;
	stem->bands.refValue() += bands;
	stem->sync.refValue() = true;
	stem->processFields();
}

void NurbsLathe::processFields(){
	M_Node& externalNode = textureMapper.refValue();
	if(externalNode.length){
		if(dynamic_cast<Stem*>(externalNode[0].p)){
			part.refValue() = 0;
			patch->stem.refValue().length = 0;
			patch->stem.refValue() += externalNode;
		}
		else{
			if(patch->stem.length == 0){
				Stem* stem = new Stem;
				stem->xDistance.refValue() = 1;
				stem->yDistance.refValue() += 1;
				stem->xCentred.refValue() = false;
				stem->bands.refValue().length = 1;
				patch->stem.refValue() += stem;
				if(part){
					stem->part += part.refValue();
				}
			}
			Stem* stem = dynamic_cast<Stem*>(patch->stem[0].p);
			stem->textureMapper.refValue().length=0;
			stem->textureMapper.refValue()+=externalNode;
			setStemValues(stem);
		}
	}
	else{
		if(patch->stem.length == 0){
			Stem* stem = new Stem;
			stem->xDistance.refValue() = 1;
			stem->yDistance.refValue() += 1;
			stem->xCentred.refValue() = false;
			stem->bands.refValue().length = 1;
			patch->stem.refValue() += stem;
		}
		Stem* stem = dynamic_cast<Stem*>(patch->stem[0].p);
		ColorSweep* colorSweep = 0;
		if(stem->textureMapper.length == 0){
			colorSweep = new ColorSweep;
			stem->textureMapper.refValue() += colorSweep;
		}
		else{
			colorSweep = dynamic_cast<ColorSweep*>(stem->textureMapper[0].p);
		}

		colorSweep->color.refValue().length = 0;
		colorSweep->color.refValue() += Color3f(color);
		colorSweep->color.refValue() += Color3f(color);
		colorSweep->straight.refValue() = true;
		colorSweep->distance.refValue().length = 0;
		colorSweep->distance.refValue() += 1;
		colorSweep->processFields();

		stem->part.refValue().length = 0;
		if(part){
			stem->part += part.refValue();
		}
		setStemValues(stem);
	}

////

	matrix = scaleMatrix(scale) * rotationMatrix(rotation) * translationMatrix(translation);
	imatrix = translationMatrix(-translation) * rotationMatrix(-rotation) * scaleMatrix(Vec3f(1,1,1)/scale);

	int dimU = ctrlStaves * 2 + 1;
	int dimV = point.length;

	degree = order - 1;
	patch->order.refValue().y = order;

	M_Vec2f& point = this->point.refValue();
	M_Vec2f& dpoint = this->dpoint.refValue();
	Stem* stem = dynamic_cast<Stem*>(patch->stem[0].p);
	if(patch->beginSP.refValue() = stem->beginSP.refValue() = beginSP){
		point[0].x = 0;
	}
	if(patch->endSP.refValue() = stem->endSP.refValue() = endSP){
		point[point.length - 1].x = 0;
	}

	dpoint.length=dimV;

	for(int i=dpoint.length; --i>=0;){
		dpoint[i] = point[i] / weight[i];
	}
	if(point.length == 0){
		return;
	}
	double sinWeight=cos(PI/ctrlStaves);

	M_Vec3f& patchPoint = patch->point.refValue();
	M_Float& patchWeight = patch->weight.refValue();

	patchPoint.length = dimV*dimU;
	patchWeight.length = dimV*dimU;
	for(int r = 0; r < dimV; r++){
		double a = PI*2/(ctrlStaves*2);
		for(int c = 0; c < dimU; c+=2){
			double b = a*c;
			patchPoint[r*dimU+c] = Vec3f(-sin(b)*point[r].x, point[r].y, -cos(b)*point[r].x);
			patchWeight[r*dimU+c] = 1;
		}
		double div = cos(a)*point[r].x;
		double ratio = 0;
		if(div!=0){
			ratio = -patchPoint[r*dimU].z / div;
		}

		for(int c = 1; c < dimU; c+=2){
			double b = a*c;
			patchPoint[r*dimU+c] = Vec3f(-sin(b)*point[r].x*ratio, point[r].y, -cos(b)*point[r].x*ratio);
			patchWeight[r*dimU+c] = sinWeight;
			patchPoint[r*dimU+c] *= sinWeight;
		}
	}
	for(int i = patchPoint.length; --i>=0;){
		patchPoint[i]=matrix * (patchPoint[i] / patchWeight[i]) * patchWeight[i];
	}
	M_Float& uKnot = patch->uKnot.refValue();
	patch->vKnot.refValue()=knot;

	updateNurbsCurveKnotMid1(dimU, patch->uKnot.refValue());

	patch->dim.refValue()=Vec2f(dimU, dimV);


//	stem->staves.refValue()=staves;
//	stem->bands.refValue()[0]=bands;
//	stem->part.refValue().length=0;
//	stem->part+=part.refValue();
//	if(part.refValue()->isRightMirror()){
//		stem->insideOut.refValue() = true;
//	}


	patch->processFields();

}

void NurbsLathe::mirrorPoints(){
	for(int i = point.length; --i >= 0;){
		point.refValue()[i] = point[i].mirrorX();
	}
}

void NurbsLathe::doBuild(){
	patch->doBuild();

	if(descendants){
		staticTransformDecendentParts(dynamic_cast<Art*>(part.p), matrix);
	}
}
*/
//////////

void NurbsLine::processFields(){
	M_Vec3f& point = this->point.refValue();
	M_Vec3f& dpoint = this->dpoint.refValue();
	M_Float& weight = this->weight.refValue();
	M_Color3f& color = this->color.refValue();
	M_Float& knot = this->uKnot.refValue();

	int dim = point.length;
	this->dim.refValue() = Vec2f(dim, 1);

	if(vKnot.length == 0){
		dpoint = point;
		weight.length = dim;
		color.length = dim;
		for(int w = dim; --w >= 0;){
			weight[w] = 1;
		}

		vKnot.refValue() = M_Float(0, 0, 0, .25, .25, .5, .5, .75, .75, 1, 1, 1);
		//updateNurbsCurveKnotMid0(dim, knot);
	}
	if(uKnot.length == 0){
		updateNurbsCurveKnotMid0(dim, uKnot);
	}
	NurbsPatch::processFields();
}

/*
void NurbsLathe::processFields(){
	matrix = scaleMatrix(scale) * rotationMatrix(rotation) * translationMatrix(translation);
	imatrix = translationMatrix(-translation) * rotationMatrix(-Rotation(rotation)) * scaleMatrix(Vec3f(1,1,1)/scale);

	int dimU = ctrlStaves * 2 + 1;
	int dimV = point.length;// * 2 - 3;

	stem->textureMapper.refValue().length = 0;
	if(textureMapper){
		stem->textureMapper.refValue()+=textureMapper.refValue();
	}
	else{
		colorSweep->color.refValue().length = 0;
		colorSweep->color.refValue() += Color3f(color);
		colorSweep->color.refValue() += Color3f(color);
		colorSweep->straight.refValue() = true;
		colorSweep->distance.refValue().length = 0;
		colorSweep->distance.refValue() += 1;
		stem->textureMapper.refValue() += colorSweep.p;
	}

	degree = order - 1;
	//patch->showAll.refValue() = true;


	M_Vec2f& point = this->point.refValue();
	M_Vec2f& dpoint = this->dpoint.refValue();
	if(patch->beginSP.refValue() = stem->beginSP.refValue() = beginSP){
		point[0].x = 0;
	}
	if(patch->endSP.refValue() = stem->endSP.refValue() = endSP){
		point[point.length - 1].x = 0;
	}

	dpoint.length = point.length;

	for(int i=dpoint.length; --i>=0;){
		dpoint[i] = point[i] / weight[i];
	}
	if(point.length == 0){
		return;
	}
	double sinWeight=cos(PI/ctrlStaves);

	M_Vec3f& patchPoint = patch->point.refValue();
	M_Float& patchWeight = patch->weight.refValue();

	patch->dim.refValue() = Vec2f(dimU, dimV);

	patchPoint.length = dimV * dimU;
	patchWeight.length = dimV * dimU;
	for(int r = 0, r2 = 0; r2 < dimV; r++, r2++){
		double a = PI * 2 / (ctrlStaves * 2);
		for(int c = 0; c < dimU; c += 2){
			double b = a * c;
			if(insideOut){
				b += PI;
			}
			patchPoint[r2 * dimU + c] = Vec3f(-sin(b) * point[r].x, point[r].y, -cos(b) * point[r].x);
			patchWeight[r2 * dimU + c] = 1;
		}
		double div = cos(a)*point[r].x;
		double ratio = 0;
		if(div!=0){
			ratio = -patchPoint[r2 * dimU].z / div;
		}

		for(int c = 1; c < dimU; c+=2){
			double b = a * c;
			patchPoint[r2 * dimU + c] = Vec3f(-sin(b) * point[r].x * ratio, point[r].y, -cos(b)  * point[r].x * ratio);
			patchWeight[r2 * dimU + c] = sinWeight;
			patchPoint[r2 * dimU + c] *= sinWeight;
		}
		if(r2 > 0 && r2 < dimV - 2){
			r2++;
			for(int c = 0; c < dimU; c += 2){
				patchWeight[r2 * dimU + c] = 1;
			}
			for(int c = 1; c < dimU; c+=2){
				patchWeight[r2 * dimU + c] = sinWeight;
			}
		}
	}
	patch->insideOut.refValue() = insideOut;
	patch->mirror.refValue() = mirror;
	if(insideOut){
		patch->setInsideOut();
		patch->setInsideOut3();
	}
	for(int i = patchPoint.length; --i>=0;){
		patchPoint[i] = matrix * (patchPoint[i] / patchWeight[i]) * patchWeight[i];
	}
	M_Float& uKnot = patch->uKnot.refValue();
	patch->vKnot.refValue()=knot;

	updateNurbsCurveKnotMid1(dimU, patch->uKnot.refValue());
	updateNurbsCurveKnotMid1(dimV, patch->vKnot.refValue());



	stem->staves.refValue()[0] = staves;
	stem->bands.refValue()[0] = bands;
	stem->part.refValue().length = 0;
	if(part){
		stem->part+=part.refValue();
	}
//	if(part.refValue()->isRightMirror()){
//		stem->insideOut.refValue() = true;
//	}


	colorSweep->processFields();
	stem->processFields();
	patch->processFields();
}
*/


void NurbsLathe::processFields(){
	matrix = scaleMatrix(scale) * rotationMatrix(rotation) * translationMatrix(translation);
	imatrix = translationMatrix(-translation) * rotationMatrix(-Rotation(rotation)) * scaleMatrix(Vec3f(1,1,1)/scale);


	int uDim = ctrlStaves * 2 + 1;
	int vDim = point.length;

	stem->textureMapper.refValue().length = 0;
	if(textureMapper){
		stem->textureMapper.refValue()+=textureMapper.refValue();
	}
	else{
		colorSweep->color.refValue().length = 0;
		colorSweep->color.refValue() += Color3f(color);
		colorSweep->color.refValue() += Color3f(color);
		colorSweep->straight.refValue() = true;
		colorSweep->distance.refValue().length = 0;
		colorSweep->distance.refValue() += 1;
		stem->textureMapper.refValue() += colorSweep.p;
	}

	degree = order - 1;
	patch->order.refValue().y = order;



	M_Vec2f& point = this->point.refValue();
	M_Vec2f& dpoint = this->dpoint.refValue();
	M_Float& weight = this->weight.refValue();

	if(patch->beginSP.refValue() = stem->beginSP.refValue() = beginSP){
		point[0].x = 0;
	}
	if(patch->endSP.refValue() = stem->endSP.refValue() = endSP){
		point[point.length - 1].x = 0;
	}

	if(NurbsPatch::MID_0 == checkNurbsType(vDim, knot.refValue())){ //convert to standard mid point for old files
		for(int i = vDim - 1; --i >= 2;){
			point.insertElement(i, Vec2f());
			weight.insertElement(i, 1);
		}
		vDim = point.length;
		updateNurbsCurveKnotMid1(vDim, knot.refValue());
	}


	dpoint.length = vDim;

	for(int i=dpoint.length; --i>=0;){
		dpoint[i] = point[i] / weight[i];
	}
	if(point.length == 0){
		return;
	}

	for(int i = 2; i < vDim - 1; i += 2){
		dpoint[i] = (dpoint[i - 1] + dpoint[i + 1]) * .5;
		point[i] = dpoint[i] * weight[i];
	}



	M_Vec3f& patchDPoint = patch->dpoint.refValue();
	M_Float& patchWeight = patch->weight.refValue();

	patch->dim.refValue() = Vec2f(uDim, vDim);

	patchDPoint.length = vDim * uDim;
	patchWeight.length = vDim * uDim;

	for(int r = 0; r < vDim; r++){
		double a = PI * 2 / (ctrlStaves * 2);
		for(int c = 0; c < uDim; c += 2){
			double b = a * c;
			if(insideOut){
				b += PI;
			}
			patchDPoint[r * uDim + c] = Vec3f(-sin(b) * dpoint[r].x, dpoint[r].y, -cos(b) * dpoint[r].x);
			patchWeight[r * uDim + c] = weight[r];
		}
		double sinWeight = cos(PI/ctrlStaves) * weight[r];

		double div = cos(a) * point[r].x;
		double ratio = 0;
		if(div!=0){
			ratio = -patchDPoint[r * uDim].z / div;
		}
		//ratio = 1;
		for(int c = 1; c < uDim; c+=2){
			double b = a * c;
			patchDPoint[r * uDim + c] = Vec3f(-sin(b) * dpoint[r].x * ratio, dpoint[r].y, -cos(b)  * dpoint[r].x * ratio);
			patchWeight[r * uDim + c] = sinWeight;
		}
	}
	for(int i = patchDPoint.length; --i>=0;){
		patchDPoint[i] = matrix * patchDPoint[i];
	}
	patch->dpointsToPoints();

	patch->insideOut.refValue() = insideOut;
	patch->mirror.refValue() = mirror;
	if(insideOut){
		patch->setInsideOut();
		patch->setInsideOut3();
	}

	updateNurbsCurveKnotMid1(uDim, patch->uKnot.refValue());
	updateNurbsCurveKnotMid1(vDim, patch->vKnot.refValue());



	stem->staves.refValue()[0] = staves;
	stem->bands.refValue()[0] = bands;
	stem->part.refValue().length = 0;
	if(part){
		stem->part+=part.refValue();
	}
//	if(part.refValue()->isRightMirror()){
//		stem->insideOut.refValue() = true;
//	}


	colorSweep->processFields();
	stem->processFields();
	patch->processFields();
//	patch->patchToCentralCPs(); //why is this not needed if the lathe was created (as opposed to a lathe existing in an opened file) ?
}

void NurbsLathe::mirrorPoints(){
	for(int i = point.length; --i >= 0;){
		point.refValue()[i] = point[i].mirrorX();
	}
}

void NurbsLine::mirrorPoints(){
	for(int i = point.length; --i >= 0;){
		point.refValue()[i] = point[i].mirrorX();
	}
}

void NurbsLathe::doBuild(){
	if(Art* part=dynamic_cast<Art*>(this->part.p)){
		if(point.refValue().length==0){
			return;
		}

		part->doBuild(patch);

		if(descendants){
			staticTransformDecendentParts(part, matrix);
		}
	}
}

void colorLine(const int dim, const int steps, const M_Color3f& color, M_Color3f& ilsColor, int k){
	
	for(int p = 0; p < dim - 1; p++){
		Color3f clA = color[p];
		Color3f clB = color[p + 1];

		int sp = steps;
		if(p == 0 || p == dim - 2){
			sp = steps >> 1;
		}
		float d = sp + 1;
		for(int i = 0; i < sp; i++){
			float t = static_cast<float>(i) / d;
			ilsColor[k++] = clB * t + clA * (1.0 - t);
		}
	}
	ilsColor[k++] = color[color.length - 1];
}

void fillPolyLineIndex(M_Vec3f& point, int markPoint, M_Int& coordIndex);
void CurvePoint (int dim, int degree, const float *knot, const Vec3f *pnt, const float *w, const float t, Vec3f& outPoint);

void genNurbsLine(const M_Vec3f& dpoint, const M_Vec3f& point, const M_Float& weight, const M_Color3f color, const M_Float& knot, const int idx, float minT, float maxT, const int degree, Shape* shape, const int steps){
	IndexedLineSet* ils = dynamic_cast<IndexedLineSet*>(shape->geometry.p);
	if(!point.length){
		return;
	}
	const float inverseSize = (1 / g_scene->getZoom());

	const int dim = point.length;

	M_Vec3f& coord = dynamic_cast<Coordinate*>(ils->coord.p)->point;
	M_Color3f& ilsColor = dynamic_cast<Color*>(ils->color.p)->color;

	int len = dim == 2 ? 2 : (dim - 2) * steps + 1;
	int markCoord = coord.length;

	coord.length += len;
	ilsColor.length += len;

	const Vec3f* p = point.getArray();
	const float* w = weight.getArray();

	int k = markCoord;

	if(dim == 2){
		coord[k] = dpoint[0];
		ilsColor[k] = color[0];
		k++;
		ilsColor[k] = color[1];
	}
	else{
		int i = len - 1;
		float d = i + 1;
		for(int i = 0; i < len - 1; i++){
			float t = static_cast<float>(i) / d;
			t = (maxT - minT) * t + minT;
			CurvePoint(dim - 1, degree, knot.getArray(), p, w, t, coord[k++]);
		}
		colorLine(dim, steps, color, ilsColor, markCoord);
	}
	coord[k] = dpoint[dpoint.length - 1];


	fillPolyLineIndex(coord, markCoord, ils->coordIndex.refValue());
	ils->coordIndex.refValue() += -1;


	ils->renderProcessed = false;
}

void NurbsLine::doBuild(){
	if(smls){
		if(smls->nurbsLineShape == 0){
			Shape* shape = new Shape;
			smls->nurbsLineShape = shape;
			IndexedLineSet* ils = new IndexedLineSet;
			Appearance* appearance = new Appearance;
			appearance->material = new Material;
			shape->appearance = appearance;
			ils->coord = new Coordinate;
			ils->color = new Color;
			shape->geometry = ils;
		}
	}

	if(Shape* shape = smls->nurbsLineShape){
		genNurbsLine(dpoint, point, weight, color, uKnot, index[0], minU(), maxU(), uDegree, shape, steps);
	}
}

void DrawNurbsLine::displayDrawnLine(){
	if(Shape* shape = smls->nurbsLineShape){
		IndexedLineSet* ils = dynamic_cast<IndexedLineSet*>(shape->geometry.p);
		M_Vec3f& coord = dynamic_cast<Coordinate*>(ils->coord.p)->point;
		M_Color3f& ilsColor = dynamic_cast<Color*>(ils->color.p)->color;

		int markCoord = coord.length;

		int len = g_drawNurbsLine->pixelSketch.length;

		coord += g_drawNurbsLine->pixelSketch;
		ilsColor.length += len;

		for(int c = 0; c < len; c++){
			
			if(colorPressure){
				ilsColor[markCoord + c] = convertPressureToCol(c);
			}
			else{
				ilsColor[markCoord + c] = Color3f(1, .5, .5);
			}
		}

		fillPolyLineIndex(coord, markCoord, ils->coordIndex.refValue());
		ils->coordIndex.refValue() += -1;
	}
}


void BendyLathe::dpointsToPoints(){
	M_Vec3f& point = this->point.refValue();
	M_Vec3f& dpoint = this->dpoint.refValue();

	point.length = dpoint.length;

	for(int i=dpoint.length; --i>=0;){
		point[i] = dpoint[i] * weight[i];
	}
}
void BendyLathe::Weight::operator=(const M_Float& m_par){
	refValue() = m_par;
	int index = home.index[0];
	home.point.refValue()[index]=home.dpoint[index] * home.weight[index];
}
void BendyLathe::DPoint::operator=(const M_Vec3f& m_par){
	int idx = home.index[0];
	float w = 1;
	if(home.weight.length){
		w = home.weight[idx];
	}
	home.point.refValue()[idx]=m_par[idx]*w;
	home.processFields();
	home.houseKeep();
}
Node* BendyLathe::HSFPart::operator=(Node* par){
	SFNode::refValue()=par;
	home.processFields();
	return par;
};

/*
void BendyLathe::setStemValues(Stem* stem){
	stem->sync.refValue() = true;
	stem->staves.refValue()[0] = staves;
	stem->bands.refValue()[0] = bands;
	stem->part.refValue().length = 0;
	stem->part+=part.refValue();
	stem->processFields();
}

void BendyLathe::processFields(){

	matrix = scaleMatrix(scale) * rotationMatrix(rotation) * translationMatrix(translation);

	M_Node& externalNode = textureMapper.refValue();
	if(externalNode.length){
		if(dynamic_cast<Stem*>(externalNode[0].p)){
			part.refValue() = 0;
			patch->stem.refValue().length = 0;
			patch->stem.refValue() += externalNode;
		}
		else{
			Stem* stem = dynamic_cast<Stem*>(patch->stem[0].p);
			stem->textureMapper.refValue().length=0;
			stem->textureMapper.refValue()+=externalNode;
			setStemValues(stem);
		}
	}
	else{
		if(patch->stem.length == 0){
			ColorSweep* colorSweep = new ColorSweep;
			Stem* stem = new Stem;
			stem->xDistance.refValue() = 1;
			stem->yDistance.refValue() += 1;
			stem->xCentred.refValue() = false;
			stem->bands.refValue().length = 1;
			patch->stem.refValue() += stem;
		}
		Stem* stem = dynamic_cast<Stem*>(patch->stem[0].p);
		ColorSweep* colorSweep = 0;
		if(stem->textureMapper.length == 0){
			colorSweep = new ColorSweep;
			stem->textureMapper.refValue() += colorSweep;
		}
		else{
			colorSweep = dynamic_cast<ColorSweep*>(stem->textureMapper[0].p);
		}

		colorSweep->color.refValue().length = 0;
		colorSweep->color.refValue() += Color3f(color);
		colorSweep->color.refValue() += Color3f(color);
		colorSweep->straight.refValue() = true;
		colorSweep->distance.refValue().length = 0;
		colorSweep->distance.refValue() += 1;
		colorSweep->processFields();

		stem->part.refValue().length = 0;
		if(part){
			stem->part += part.refValue();
		}
		setStemValues(stem);
	}

////


	degree = order-1;

	M_Vec3f& point = this->point.refValue();
	M_Vec3f& dpoint = this->dpoint.refValue();

	dpoint.length=point.length;

	for(int i=dpoint.length; --i>=0;){
		dpoint[i] = point[i] / weight[i];
	}
	if(point.length==0){
		return;
	}
	double sinWeight=cos(PI/ctrlStaves);
	int dimU = ctrlStaves*2+1;
	int dimV = radius.length;


	M_Vec3f& patchPoint = patch->point.refValue();
	M_Vec3f& patchDPoint = patch->dpoint.refValue();
	M_Float& patchWeight = patch->weight.refValue();
	const M_Float& radius = this->radius;
	const M_Float& at = this->at;
	ringBase.length = ringRot.length = at.length;


	float* knotPtr = knot.refValue().getArray();
	Vec3f* pointPtr = point.getArray();
	float* weightPtr = weight.refValue().getArray();

	patchWeight.length = patchPoint.length = dimV * dimU;
	Rotation prevRot;
	Vec3f prevPos;
	double a = PI/ctrlStaves;
	double div = cos(a);
	ratio = 1.0 / div;

	float mxt = maxT();
	float mnt = minT();

	for(int r = 0; r < dimV; r++){
		double t = (mxt - mnt) * at[r] + mnt;
		double tm = (mxt - mnt) * (at[r] - .0001) + mnt;
		double tp = (mxt - mnt) * (at[r] + .0001) + mnt;
		Vec3f dest0;
		CurvePoint(point.length-1, degree, knotPtr, pointPtr, weightPtr, t, dest0);
		ringBase[r] = dest0;
		Rotation rot;
		if(at[r]<1){
			Vec3f dest1;
			CurvePoint(point.length-1, degree, knotPtr, pointPtr, weightPtr, tp, dest1);
			rot = Rotation(prevRot*Vec3f(0,1,0), Vec3f(dest1-dest0).normalize());
		}
		else{
			Vec3f dest1;
			CurvePoint(point.length-1, degree, knotPtr, pointPtr, weightPtr, tm, dest1);
			rot = Rotation(prevRot*Vec3f(0,1,0), Vec3f(dest0-dest1).normalize());
		}
		prevRot=prevRot*rot;
		ringRot[r] = prevRot;
		for(int c = 0; c < dimU; c+=1){
			double b = a*c;
			Vec3f pos = prevRot * Vec3f(-sin(b)*radius[r], 0, -cos(b)*radius[r]) + dest0;

			patchPoint[r*dimU+c] = pos;
			patchWeight[r*dimU+c] = 1;
		}
		for(int c = 1; c < dimU; c+=2){
			double b = a*c;
			Vec3f pos = prevRot * Vec3f(-sin(b)*radius[r]*ratio, 0, -cos(b)*radius[r]*ratio) + dest0;

			patchPoint[r*dimU+c] = pos;
			patchWeight[r*dimU+c] = sinWeight;
			patchPoint[r*dimU+c] *= sinWeight;
		}
	}
	for(int i = patchPoint.length; --i>=0;){
		patchPoint[i] = matrix * (patchPoint[i] / patchWeight[i]) * patchWeight[i];
	}
	updateNurbsCurveKnotMid1(dimU, patch->uKnot.refValue());
	updateNurbsCurveKnotMid0(dimV, patch->vKnot.refValue());

	patch->dim.refValue()=Vec2f(dimU, dimV);

	patch->processFields();

	float lowestPoint=0;
	float highestPoint=0;
	for(int v = dimV; --v>=0;){
		for(int u = 1; u < dimU; u+=2){
			int i = v * dimU + u;
			if(patchDPoint[i].y > highestPoint){
				highestPoint = patchDPoint[i].y;
			}
			if(patchDPoint[i].y < lowestPoint){
				lowestPoint = patchDPoint[i].y;
			}
		}
	}

	Stem* stem = dynamic_cast<Stem*>(patch->stem[0].p);
	patch->beginSP.refValue() = stem->beginSP.refValue() = radius[0] == 0;
	patch->endSP.refValue() = stem->endSP.refValue() = radius[dimV-1] == 0;

}
//bendy method
*/

void Strand::setStemValues(Stem* stem, int staves, int bands){
	stem->sync.refValue() = true;
	stem->staves.refValue()[0] = staves;
	stem->bands.refValue()[0] = bands * (point.length - 1);
	stem->part.refValue().length = 0;
	if(staves > 1){
		stem->uWrap.refValue() = true;
		stem->xDistance.refValue() = 1;
	}
	else{
		stem->uWrap.refValue() = false;
		stem->xDistance.refValue() = .5;
	}
	stem->processFields();
}

void BendyLathe::processFields(){
	matrix = scaleMatrix(scale) * rotationMatrix(rotation) * translationMatrix(translation);
	imatrix = translationMatrix(-translation) * rotationMatrix(-Rotation(rotation)) * scaleMatrix(Vec3f(1,1,1) / scale);

	stem->textureMapper.refValue().length=0;
	if(textureMapper){
		stem->textureMapper.refValue()+=textureMapper.refValue();
	}
	else{
		colorSweep->color.refValue().length=0;
		colorSweep->color.refValue()+=Color3f(color);
		colorSweep->color.refValue()+=Color3f(color);
		colorSweep->straight.refValue()=true;
		colorSweep->distance.refValue().length=0;
		colorSweep->distance.refValue()+=1;
		stem->textureMapper.refValue()+=colorSweep.p;
	}

	degree = order-1;

	M_Vec3f& point = this->point.refValue();
	M_Vec3f& dpoint = this->dpoint.refValue();

	dpoint.length=point.length;

	for(int i = dpoint.length; --i>=0;){
		dpoint[i] = point[i] / weight[i];
	}
	if(part==0 || point.length==0){
		return;
	}
	double sinWeight = cos(PI / ctrlStaves);
	int dimU = ctrlStaves * 2 + 1;
	int dimV = radius.length * 2 - 3;

	M_Vec3f& patchDPoint = patch->dpoint.refValue();
	M_Float& patchWeight = patch->weight.refValue();
	const M_Float& radius = this->radius;
	const M_Float& at = this->at;
	ringBase.length = ringRot.length = at.length;


	float* knotPtr = knot.refValue().getArray();
	Vec3f* pointPtr = point.getArray();
	float* weightPtr = weight.refValue().getArray();

	patchWeight.length = patchDPoint.length = dimV * dimU;
	Rotation prevRot;
	Vec3f prevPos;
	double a = PI / ctrlStaves;
	double div = cos(a);
	ratio = 1.0 / div;

	for(int r = 0, r2 = 0; r2 < dimV; r++, r2++){
		double t = (maxT()-minT()) * at[r] + minT();
		double tm = (maxT()-minT()) * (at[r]-.0001) + minT();
		double tp = (maxT()-minT()) * (at[r]+.0001) + minT();
		Vec3f dest0;
		CurvePoint(point.length - 1, degree, knotPtr, pointPtr, weightPtr, t, dest0);
		ringBase[r] = dest0;
		Rotation rot;
		if(at[r] < 1){
			Vec3f dest1;
			CurvePoint(point.length-1, degree, knotPtr, pointPtr, weightPtr, tp, dest1);
			rot = Rotation(prevRot*Vec3f(0, 1, 0), Vec3f(dest1 - dest0).normalize());
		}
		else{
			Vec3f dest1;
			CurvePoint(point.length-1, degree, knotPtr, pointPtr, weightPtr, tm, dest1);
			rot = Rotation(prevRot*Vec3f(0, 1, 0), Vec3f(dest0 - dest1).normalize());
		}
		prevRot=prevRot*rot;
		ringRot[r] = prevRot;
		for(int c = 0; c < dimU; c += 1){
			double b = a * c;
			Vec3f pos = prevRot * Vec3f(-sin(b)*radius[r], 0, -cos(b) * radius[r]) + dest0;
			patchDPoint[r2 * dimU+c] = pos;
			patchWeight[r2 * dimU+c] = 1;
		}
		for(int c = 1; c < dimU; c+=2){
			double b = a*c;
			Vec3f pos = prevRot * Vec3f(-sin(b)*radius[r]*ratio, 0, -cos(b) * radius[r] * ratio) + dest0;
			patchDPoint[r2 * dimU+c] = pos;
			patchWeight[r2 * dimU+c] = sinWeight;
		}
		if(r2 > 0 && r2 < dimV - 2){
			r2++;
			for(int c = 0; c < dimU; c += 2){
				patchWeight[r2 * dimU + c] = 1;
			}
			for(int c = 1; c < dimU; c += 2){
				patchWeight[r2 * dimU + c] = sinWeight;
			}
		}
	}
	for(int i = patchDPoint.length; --i>=0;){
		patchDPoint[i] = matrix * patchDPoint[i];
	}
	patch->dpointsToPoints();

	patch->insideOut.refValue() = insideOut;
	patch->mirror.refValue() = mirror;
	if(insideOut){
		patch->setInsideOut();
		patch->setInsideOut3();
	}

	updateNurbsCurveKnotMid1(dimU, patch->uKnot.refValue());
	updateNurbsCurveKnotMid1(dimV, patch->vKnot.refValue());

	patch->dim.refValue() = Vec2f(dimU, dimV);

	patch->processFields();


	stem->staves.refValue()[0] = staves;
	stem->bands.refValue()[0] = bands;
	stem->part.refValue().length = 0;
	stem->part+=part.refValue();
	patch->beginSP.refValue() = stem->beginSP.refValue() = radius[0] == 0;
	patch->endSP.refValue() = stem->endSP.refValue() = radius[radius.length - 1] == 0;

	colorSweep->processFields();
	stem->processFields();
}

void Strand::generateGeometry(Art* part, float radius, int staves, int bands){//direct curve

	{
		if(geometry->stem.length == 0){
			Stem* stem = new Stem;
			stem->yDistance.refValue() += 1;
			stem->xCentred.refValue() = false;
			stem->bands.refValue().length = 1;
			//stem->uWrap.refValue() = true;
			//stem->endSP.refValue() = true;
			geometry->stem.refValue() += stem;

		}
		Stem* stem = dynamic_cast<Stem*>(geometry->stem[0].p);

		stem->part.refValue().length = 0;

		setStemValues(stem, staves, bands);
		stem->part += part;
	}

////
	int ctrlStaves = 4;

	int degree = order.y - 1;

	M_Vec3f& point = this->point.refValue();
	M_Vec3f& dpoint = this->dpoint.refValue();
	const M_Float& weight = this->weight;

	dpoint.length = point.length;

	for(int i = dpoint.length; --i>=0;){
		dpoint[i] = point[i] / weight[i];
	}
	if(point.length==0){
		return;
	}

	double sinWeight = cos( PI / ctrlStaves);
	int dimU = ctrlStaves * 2 + 1;

	//M_Vec3f& geomPoint = geometry->point.refValue();
	M_Vec3f& geomDPoint = geometry->dpoint.refValue();
	M_Float& geomWeight = geometry->weight.refValue();

	int dimV = point.length;
	float step = 1.0 / (dimV - 2);

	float* knotPtr = this->vKnot.refValue().getArray();
	Vec3f* pointPtr = point.getArray();
	float* weightPtr = this->weight.getArray();

	geomWeight.length = geomDPoint.length = dimV * dimU;


	Rotation prevRot;
	Vec3f prevPos;
	double a = PI/ctrlStaves;
	double div = cos(a);
	double ratio = 1.0 / div;

	float mxv = maxV();
	float mnv = minV();

	float at = 0;

	for(int r = 0; r < dimV - 1; r++){
		if(r == dimV){
			at = 1;
		}
		else if(r > 0){
			at = r * step - step * .5;
		}
		double t = (mxv - mnv) * at + mnv;
		double tm = (mxv - mnv) * (at - .0001) + mnv;
		double tp = (mxv - mnv) * (at + .0001) + mnv;
		Vec3f dest0;
		Vec3f dest1;
		CurvePoint(point.length - 1, degree, knotPtr, pointPtr, weightPtr, t, dest0);
		Rotation rot;
		if(at < 1){
			CurvePoint(point.length - 1, degree, knotPtr, pointPtr, weightPtr, tp, dest1);
			rot = Rotation(prevRot * Vec3f(0,1,0), Vec3f(dest1 - dest0).normalize());
		}
		else{
			CurvePoint(point.length-1, degree, knotPtr, pointPtr, weightPtr, tm, dest1);
			rot = Rotation(prevRot * Vec3f(0,1,0), Vec3f(dest0 - dest1).normalize());
		}
		prevRot = prevRot * rot;

		for(int c = 0; c < dimU; c += 1){
			double b = a * c;
			Vec3f pos = prevRot * Vec3f(-sin(b) * radius, 0, -cos(b) * radius) + point[r];

			geomDPoint[r * dimU + c] = pos;
			geomWeight[r * dimU + c] = 1;
		}
		for(int c = 1; c < dimU; c += 2){
			double b = a * c;
			Vec3f pos = prevRot * Vec3f(-sin(b) * radius * ratio, 0, -cos(b) * radius * ratio) + point[r];

			geomDPoint[r * dimU + c] = pos;
			geomWeight[r * dimU + c] = sinWeight;
			//geomPoint[r*dimU+c] *= sinWeight;
		}
	}
	//for(int c = 1; c < dimU; c += 2){
	Vec3f endPoint = point[point.length - 1];
	int id = 0;
	for(int c = 1; c < dimU; c+=2){
		id = geomDPoint.length - dimU + c - 1;

		geomDPoint[id] = endPoint;
		geomWeight[id] = 1;
		id++;
		geomDPoint[id] = endPoint;
		geomWeight[id] = sinWeight;
	}
	id++;
	geomDPoint[id] = endPoint;
	geomWeight[id] = 1;



	//for(int i = geomPoint.length; --i>=0;){
	//	geomPoint[i] = (geomPoint[i] / geomWeight[i]) * geomWeight[i];
	//}
	geometry->dpointsToPoints();

	updateNurbsCurveKnotMid1(dimU, geometry->uKnot.refValue());
	updateNurbsCurveKnotMid0(dimV, geometry->vKnot.refValue());

	geometry->dim.refValue()=Vec2f(dimU, dimV);

	geometry->processFields();


	Stem* stem = dynamic_cast<Stem*>(geometry->stem[0].p);
	geometry->beginSP.refValue() = false;
	geometry->endSP.refValue() = true;
	geometry->cage.refValue() = false;
	geometry->uJoinEnds.refValue() = true;
}


/*
void Strand::generateGeometry(Art* part, float radius, int staves, int bands){//indirect curve
	static M_Rotation ringRot;
	static M_Vec3f ringBase;
	static M_Float at;

	{
		if(geometry->stem.length == 0){
			Stem* stem = new Stem;
			stem->yDistance.refValue() += 1;
			stem->xCentred.refValue() = false;
			stem->bands.refValue().length = 1;
			//stem->uWrap.refValue() = true;
			//stem->endSP.refValue() = true;
			geometry->stem.refValue() += stem;

		}
		Stem* stem = dynamic_cast<Stem*>(geometry->stem[0].p);

		stem->part.refValue().length = 0;

		setStemValues(stem, staves, bands);
		stem->part += part;
	}

////
	int ctrlStaves = 4;

	int degree = order.y - 1;

	M_Vec3f& point = this->point.refValue();
	M_Vec3f& dpoint = this->dpoint.refValue();
	const M_Float& weight = this->weight;

	dpoint.length = point.length;

	for(int i=dpoint.length; --i>=0;){
		dpoint[i] = point[i] / weight[i];
	}
	if(point.length==0){
		return;
	}

	double sinWeight = cos( PI / ctrlStaves);
	int dimU = ctrlStaves * 2 + 1;

	//M_Vec3f& geomPoint = geometry->point.refValue();
	M_Vec3f& geomDPoint = geometry->dpoint.refValue();
	M_Float& geomWeight = geometry->weight.refValue();

	int atLen = (point.length - 1)  * 2;
	int dimV = atLen;
	at.length = atLen;
	at[0] = 0;
	at[atLen - 1] = 1;
	float step = 1.0 / (atLen - 2);
	float offset = step * .5;
	for(int i = 1; i < atLen - 1; i++){
		at[i] = (i - 1) * step + offset;
	}
	ringBase.length = ringRot.length = at.length;

	float* knotPtr = this->vKnot.refValue().getArray();
	Vec3f* pointPtr = point.getArray();
	float* weightPtr = this->weight.getArray();

	geomWeight.length = geomDPoint.length = dimV * dimU;


	Rotation prevRot;
	Vec3f prevPos;
	double a = PI/ctrlStaves;
	double div = cos(a);
	double ratio = 1.0 / div;

	float mxv = maxV();
	float mnv = minV();

	for(int r = 0; r < dimV - 1; r++){
		double t = (mxv - mnv) * at[r] + mnv;
		double tm = (mxv - mnv) * (at[r] - .0001) + mnv;
		double tp = (mxv - mnv) * (at[r] + .0001) + mnv;
		Vec3f dest0;
		CurvePoint(point.length-1, degree, knotPtr, pointPtr, weightPtr, t, dest0);
		ringBase[r] = dest0;
		Rotation rot;
		if(at[r]<1){
			Vec3f dest1;
			CurvePoint(point.length-1, degree, knotPtr, pointPtr, weightPtr, tp, dest1);
			rot = Rotation(prevRot*Vec3f(0,1,0), Vec3f(dest1-dest0).normalize());
		}
		else{
			Vec3f dest1;
			CurvePoint(point.length-1, degree, knotPtr, pointPtr, weightPtr, tm, dest1);
			rot = Rotation(prevRot*Vec3f(0,1,0), Vec3f(dest0-dest1).normalize());
		}
		prevRot=prevRot*rot;
		ringRot[r] = prevRot;
		for(int c = 0; c < dimU; c += 1){
			double b = a * c;
			Vec3f pos = prevRot * Vec3f(-sin(b) * radius, 0, -cos(b) * radius) + dest0;

			geomDPoint[r*dimU+c] = pos;
			geomWeight[r*dimU+c] = 1;
		}
		for(int c = 1; c < dimU; c+=2){
			double b = a*c;
			Vec3f pos = prevRot * Vec3f(-sin(b) * radius * ratio, 0, -cos(b) * radius * ratio) + dest0;

			geomDPoint[r*dimU+c] = pos;
			geomWeight[r*dimU+c] = sinWeight;
			//geomPoint[r*dimU+c] *= sinWeight;
		}
	}
	//for(int c = 1; c < dimU; c += 2){
	Vec3f endPoint = point[point.length - 1];
	int id = 0;
	for(int c = 1; c < dimU; c+=2){
		id = geomDPoint.length - dimU + c - 1;

		geomDPoint[id] = endPoint;
		geomWeight[id] = 1;
		id++;
		geomDPoint[id] = endPoint;
		geomWeight[id] = sinWeight;
	}
	id++;
	geomDPoint[id] = endPoint;
	geomWeight[id] = 1;



	//for(int i = geomPoint.length; --i>=0;){
	//	geomPoint[i] = (geomPoint[i] / geomWeight[i]) * geomWeight[i];
	//}
	geometry->dpointsToPoints();

	updateNurbsCurveKnotMid1(dimU, geometry->uKnot.refValue());
	updateNurbsCurveKnotMid0(dimV, geometry->vKnot.refValue());

	geometry->dim.refValue()=Vec2f(dimU, dimV);

	geometry->processFields();


	Stem* stem = dynamic_cast<Stem*>(geometry->stem[0].p);
	geometry->beginSP.refValue() = false;
	geometry->endSP.refValue() = true;
	geometry->cage.refValue() = false;
	geometry->uJoinEnds.refValue() = true;
}
*/


/*
//torus method
void Strand::processFields(Art* part, float radius){

	{
		if(geometry->stem.length == 0){
			Stem* stem = new Stem;
			stem->xDistance.refValue() = 1;
			stem->yDistance.refValue() += 1;
			stem->xCentred.refValue() = false;
			stem->bands.refValue().length = 1;
			stem->uWrap.refValue() = true;
			stem->endSP.refValue() = true;
			geometry->stem.refValue() += stem;

		}
		Stem* stem = dynamic_cast<Stem*>(geometry->stem[0].p);

		stem->part.refValue().length = 0;

		setStemValues(stem);
		stem->part += part;
	}

////
	int ctrlStaves = 4;

	int degree = ctrlLine->order.y - 1;

	M_Vec3f& point = ctrlLine->point.refValue();
	M_Vec3f& dpoint = ctrlLine->dpoint.refValue();
	M_Float& weight = ctrlLine->weight.refValue();

	if(point.length==0){
		return;
	}

	float visWeight = cos(PI / ctrlStaves);
	int uDim = ctrlStaves * 2 + 1;

	M_Vec3f& patchPoint = geometry->point.refValue();
	M_Vec3f& patchDPoint = geometry->dpoint.refValue();
	M_Float& patchWeight = geometry->weight.refValue();

	dpoint.length=point.length;

	for(int i=dpoint.length; --i>=0;){
		dpoint[i] = point[i] / weight[i];
	}


	for(int w = patchWeight.length;  --w>=0;){
		patchWeight[w] = 1;
	}

	int vDim = (point.length - 2)  * 2 + 1;

	patchWeight.length = patchPoint.length = patchDPoint.length = vDim * uDim;

	Rotation prevRot;
	int pointIdx = 0;
	int i = 0;
	Rotation rot;
	int nextV = 0;

	double a = PI * 2 /(ctrlStaves * 2);
	double div = cos(a);
	double ratio = 0;
	if(div != 0){
		ratio = 1 / div;
	}
	float rad = radius * ratio;

	for(int v = 0; v < vDim - 1; pointIdx++){
		if(v == 0){
			rot = Rotation(Vec3f(0, 1, 0), Vec3f(point[1] - point[0]).normalize());
			nextV = 1;
		}
		else{
			Rotation rotA = Rotation(prevRot*Vec3f(0, 1, 0), Vec3f(point[pointIdx] - point[pointIdx - 1]).normalize());
			Rotation rotB = Rotation(prevRot*Vec3f(0, 1, 0), Vec3f(point[pointIdx + 1] - point[pointIdx]).normalize());

			rot = rotA.slerp(rotB, .5);
			nextV += 2;
		}
		prevRot = prevRot * rot;

		for(int u = 1; u < uDim - 1; u+=2){
			float angle = u * PI * 2 / (uDim - 1);
			Vec3f localPos = Rotation(0, 1, 0, angle) * Vec3f(0, 0, rad);

			i = v * uDim + u;
			patchDPoint[i] = point[pointIdx] + prevRot * localPos;
			patchWeight[i] = visWeight;
			patchWeight[i + uDim] = 1;
		}
		v = nextV;
	}

	for(int u = 1; u < uDim - 1; u+=2){
		i = patchPoint.length - uDim + u;
		patchDPoint[i] = point[point.length - 1];
		patchWeight[i] = visWeight;
	}

	geometry->dpointsToPoints();

	updateNurbsCurveKnotMid1(uDim, geometry->uKnot.refValue());
	updateNurbsCurveKnotMid1(vDim, geometry->vKnot.refValue());

	geometry->dim.refValue() = Vec2f(uDim, vDim);

	geometry->showAll.refValue() = false;
	geometry->processFields();

	Stem* stem = dynamic_cast<Stem*>(geometry->stem[0].p);
	geometry->beginSP.refValue() = false;
	geometry->endSP.refValue() = true;
	geometry->uJoinEnds.refValue() = true;
	geometry->cage.refValue() = false;
	//geometry->showAll.refValue() = true;


}
*/
void BendyLathe::doBuild(){
	patch->doBuild();

	if(descendants){
		staticTransformDecendentParts(dynamic_cast<Art*>(part.p), matrix);
	}
}

float BendyLathe::zeroinOnVerticalPoint(M_Vec3f& point, const M_Float& weight, const M_Float& knot, float minT, float maxT, float targetHeight){
	Vec3f atVec;
	float lowT = 0;
	float highT = 1;
	int dim=point.length;
	while(true){
		float midT = lowT + (highT-lowT) * .5;

		float t = (maxT-minT)*midT + minT;

		const Vec3f* p = point.getArray();
		const float* w=weight.getArray();
		const float* knotPtr = knot.getArray();

		CurvePoint(dim-1, degree, knotPtr, p, w, t, atVec);

		if(abs(targetHeight - atVec.y) < .000001){
			if(abs(atVec.y) < .000001){
				atVec.y = 0;
			}
			if(abs(atVec.y - 1) < .000001){
				atVec.y = 1;
			}
			return midT;
		}
		if(targetHeight > atVec.y){
			lowT = midT;
		}
		else{
			highT = midT;
		}
	}
}
void Art::calcAbsolutePivotPoint(Vec3f param){
	absolutePivotPoint = param + pivotPoint;
	int len = part.length;
	for(int k =0; k < len; k++){
		part.refValue()[k]->calcAbsolutePivotPoint(absolutePivotPoint);
	}
}
void Seamless::calcAbsolutePivotPoint(){
	int len = skeleton.length;
	for(int k =0; k < len; k++){
		skeleton.refValue()[k]->calcAbsolutePivotPoint(Vec3f());
	}
}

/*
void convertArtToArt(Art* from, Art* to, bool refNodes = true);
Art* createNewTypeOf(Art* part){
	if(dynamic_cast<Part*>(part)){
		return new Part;
	}
	else if(dynamic_cast<ImmediatePart*>(part)){
		return new ImmediatePart;
	}
	else throw;
}
void convertM_SArtToM_SArt(M_SArt& from, M_SArt& to){
	for(int k = 0; k < from.length; k++){
		Art* partFrom = from[k];
		Art* partTo = 0;
		if(k < to.length){
			partTo = to[k];
			if(partFrom == partTo || partFrom->getType() != partTo->getType()){
				partTo = createNewTypeOf(partFrom);
				to[k] = partTo;
			}
		}
		else{
			partTo = createNewTypeOf(partFrom);
			to+=partTo;
		}
		convertArtToArt(partFrom, partTo);
	}
	to.length = from.length;
}
void convertSeamlessToSeamlessLoop(Art* from, Art* to){
	M_SArt& m_partFrom = from->part.refValue();
	M_SArt& m_partTo = to->part.refValue();
	convertM_SArtToM_SArt(m_partFrom, m_partTo);
	for(int k = 0; k < m_partFrom.length; k++){
		convertSeamlessToSeamlessLoop(m_partFrom[k], m_partTo[k]);
	}
}
void convertSeamlessToSeamless(Seamless* from, Seamless* to){
	to->effect.refValue() = from->effect.refValue();
	M_SArt& skeletonFrom = from->skeleton.refValue();
	M_SArt& skeletonTo = to->skeleton.refValue();
	convertM_SArtToM_SArt(skeletonFrom, skeletonTo);
	for(int k = 0; k < skeletonFrom.length; k++){
		convertSeamlessToSeamlessLoop(skeletonFrom[k], skeletonTo[k]);
	}
}
Seamless* copySeamless(Seamless* from){
	Seamless* to = new Seamless;
	convertSeamlessToSeamless(from, to);
	return to;
}
void ModelMorpher::processFields(){
//setParenReferences
	M_Node& sequence = this->sequence.refValue();
	for(int s = sequence.length; --s>=0;){
		dynamic_cast<ModelSequence*>(sequence[s].p)->parent = this;
	}
	if(sequence.length>0){
		ModelSequence* modelSequence = dynamic_cast<ModelSequence*>(sequence[0].p);
		M_Node& model = modelSequence->model.refValue();
		if(model.length > 0){
			Seamless* from = dynamic_cast<Seamless*>(model[0].p);
			convertSeamlessToSeamless(from, smls);
		}
	}
}
*/
void ModelMorpher::processFields(){
	M_Node& sequence = this->sequence.refValue();
	for(int s = sequence.length; --s>=0;){
		dynamic_cast<ModelSequence*>(sequence[s].p)->parent = this;
	}
	if(sequence.length>0){
		ModelSequence* modelSequence = dynamic_cast<ModelSequence*>(sequence[0].p);
		M_Node& model = modelSequence->model.refValue();
		if(model.length > 0){
			Seamless* from = dynamic_cast<Seamless*>(model[0].p);
			if(from->ctrf || justLoaded){
				from->processFields();//stop folowing toString loosing data
			}
			justLoaded=false;

			String str;
			from->toString(str="");
			M_Node m_smls;
			createSmlsFromString(m_smls, str);
			smls = dynamic_cast<Seamless*>(m_smls[0].p);
			smls->processFields();
			smls->processBuild();
			smls->doBuild();
		}
	}
}
void processBuildLoop(M_Build& m_build){
	for(int k=0; k<m_build.length; k++){
		Build* build=m_build[k].p;
		build->processFields();
		if(GroupBuild* groupBuild = dynamic_cast<GroupBuild*>(build)){
			processBuildLoop(groupBuild->build.refValue());
		}
	}
}
void Seamless::processBuild(){
	processBuildLoop(build.refValue());
}
void ModelSequence::MFModel::operator=(const NodePtr& par){
	refValue()=par;
	if(home.parent){
		home.parent->processFields();
	}
}
void ModelSequence::MFModel::operator=(const M_Node& par){
	refValue()=par;
	if(home.parent){
		home.parent->processFields();
	}
}
void ModelSequence::MFModel::resize(int newLen){
	refValue().length=newLen;
	if(home.parent){
		home.parent->processFields();
	}
}
void FuseSurface::NumOf::operator=(int const& par){
	int dummy=0;
	if(home.qty>0){
		if(par == refValue()+1){
			home.incBaseCoordGroup(dummy);
		}
		else if(par == refValue()-1){
			home.decBaseCoordGroup(dummy);
		}
		else{
			refValue()=par;
			home.baseCoord.refValue().length=par*home.qty;
		}
	}
}
void FuseSurface::Qty::operator=(int const& par){
	M_Vec2f& baseCoord = home.baseCoord.refValue();
	if(home.numOf==1){
		if(par == refValue()+1){
			if(baseCoord.length){
				baseCoord+=baseCoord[baseCoord.length-1];
			}
			else{
				baseCoord+=Vec2f(.5, .5);
			}
			refValue()++;
		}
		else if(par == refValue()-1){
			if(refValue()){
				refValue()--;
				baseCoord.length=refValue();
			}
		}
		else{
			refValue()=par;
			baseCoord.length=par*home.numOf;
		}
	}
}
void PolyToNurbs::MimicOriginalIndex::operator=(bool const& val){
	if(Art* originalPart = dynamic_cast<Art*>(home.original.p)){
		originalPart->swapVertex(home.selOriginalVertex,home.selMimicVertex);
	}
}


void setAllIns(FuseSurface* fuseSurface, const bool inState){
	if(fuseSurface->snap){
		fuseSurface->in=inState;
	}
	if(NurbsPatch* patch=dynamic_cast<NurbsPatch*>(fuseSurface->patch.refValue().p)){
		setAllIns(patch, inState);
	}
}
void setAllIns(Stem* stem, const bool inState){
	M_Node& fuse = stem->fuse.refValue();
	for(int k = fuse.length; --k>=0;){
		setAllIns(dynamic_cast<FuseSurface*>(fuse[k].p), inState);
	}
}
void setAllIns(NurbsPatch* patch, const bool inState){
	M_Node& m_stem = patch->stem.refValue();
	for(int k = m_stem.length; --k>=0;){
		setAllIns(dynamic_cast<Stem*>(m_stem[k].p), inState);
	}
}
void setAllIns(M_Build& m_build, const bool inState){
	for(int k = m_build.length; --k>=0;){
		Node* node=m_build[k].p;
		if(JoinVertex* joinVertex=dynamic_cast<JoinVertex*>(node)){
			joinVertex->in=inState;
		}
		else if(TransferBCyl* transferBCyl=dynamic_cast<TransferBCyl*>(node)){
			transferBCyl->in=inState;
		}
		else if(NurbsPatch* patch=dynamic_cast<NurbsPatch*>(node)){
			setAllIns(patch, inState);
		}
		else if(GroupBuild* groupBuild=dynamic_cast<GroupBuild*>(node)){
			setAllIns(groupBuild->build.refValue(), inState);
		}
	}
}




void Seamless::AutoIn::operator=(const bool& par){
	refValue()=par;
	home.masterIn.refValue()=false;
	if(par){
		setAllIns(home.build.refValue(), false);
	}
}
void Subdivision::processFields(){
	const M_Int& coordIndex = this->coordIndex;
	polyIndex.length = 0;
	for(int i = 0; i < coordIndex.length; i+=4){
		polyIndex += PolyIndex(i, 4);
	}
}
void setFog(unsigned int color, float visibilityStart, float visibilityRange, float density, const char* fogType);

void ColorEffect::Diffuse::operator =(const Color3f &par){
	refValue()=par;
	for(int k = home.partList.length; --k>=0;){
		Art* part = home.partList[k];
		part->colorize(par);
		part->smls->ctrf = true;
	}
}
/*
void TextureMapper::doBuild(){
	if(part.length==0)return;

	Seamless* smls = part[0]->smls;

	Matrix parentMatrix = rotationMatrix(-orientation);

	//for(int k=0; k<smls->skeleton.length;k++){
	//	buildSceneMatrices( smls->skeleton.refValue()[k], parentMatrix);
	//}
	static M_Int mask;

	for(int k = 0; k < part.length; k++){
		Art* part = this->part.refValue()[k].p;
		if(TextureEffect* textureEffect = dynamic_cast<TextureEffect*>(part->workEffect)){
			M_Triangle& triangle=part->triangle;
			M_Vec2f& texCoord = textureEffect->texCoord.refValue();
			mask.length = 0;
			mask.length = texCoord.length;
			M_Vec3f& coord = part->coord.refValue();

			if(autoSize3d){
				part->findSetVertexRange(*this);
			}
			float scaleX=(1./select.x)*size2d.x;
			float scaleY=(1./select.y)*size2d.y;

			Matrix matrix = parentMatrix;

			if(select.z == 0){
				for(int i=triangle.length*3;--i>=0;){
					int ti = triangle.refTexIndex(i);
					if(mask[ti] == 0){
						Art* part = triangle.refPart(i);
						int ci = triangle.refIndex(i);
						Vec3f crd = matrix * part->coord[ci];
						texCoord[ti].x = (crd.x - position3d.x) * scaleX + position2d.x;
						texCoord[ti].y = (crd.y - position3d.y) * scaleY + position2d.y;
					}
				}
			}
			else if(isolate == false){
				float depth = select.z;
				float posZ = position3d.z;
				if(depth > position3d.z){
					depth = position3d.z;
					posZ = select.z;
				}
				for(int i=triangle.length;--i>=0;){
					Art* pa = triangle[i].pa;
					Art* pb = triangle[i].pb;
					Art* pc = triangle[i].pc;

					int ia = triangle[i].ia;
					int ib = triangle[i].ib;
					int ic = triangle[i].ic;

					int ta = triangle[i].ta;
					int tb = triangle[i].tb;
					int tc = triangle[i].tc;

					Vec3f crdA = matrix * pa->coord[ia];
					Vec3f crdB = matrix * pb->coord[ib];
					Vec3f crdC = matrix * pc->coord[ic];



					if(	crdA.z >= depth && crdA.z <= posZ &&
						crdB.z >= depth && crdB.z <= posZ &&
						crdC.z >= depth && crdC.z <= posZ){

						texCoord[ta] = Vec2f((crdA.x - position3d.x) * scaleX + position2d.x, (crdA.y - position3d.y) * scaleY + position2d.y);
						texCoord[tb] = Vec2f((crdB.x - position3d.x) * scaleX + position2d.x, (crdB.y - position3d.y) * scaleY + position2d.y);
						texCoord[tc] = Vec2f((crdC.x - position3d.x) * scaleX + position2d.x, (crdC.y - position3d.y) * scaleY + position2d.y);
					}
				}
			}
			else{
				float depth = select.z;
				float posZ = position3d.z;
				if(depth > position3d.z){
					depth = position3d.z;
					posZ = select.z;
				}
				for(int i=triangle.length;--i>=0;){
					Art* pa = triangle[i].pa;
					Art* pb = triangle[i].pb;
					Art* pc = triangle[i].pc;

					int ia = triangle[i].ia;
					int ib = triangle[i].ib;
					int ic = triangle[i].ic;

					int ta = triangle[i].ta;
					int tb = triangle[i].tb;
					int tc = triangle[i].tc;

					Vec3f crdA = matrix * pa->coord[ia];
					Vec3f crdB = matrix * pb->coord[ib];
					Vec3f crdC = matrix * pc->coord[ic];

					if(	crdA.z >= depth && crdA.z <= posZ &&
						crdB.z >= depth && crdB.z <= posZ &&
						crdC.z >= depth && crdC.z <= posZ){

						int len = texCoord.length;
						triangle[i].ta = len;
						triangle[i].tb = len + 1;
						triangle[i].tc = len + 2;

						texCoord += Vec2f((crdA.x - position3d.x) * scaleX + position2d.x, (crdA.y - position3d.y) * scaleY + position2d.y);
						texCoord += Vec2f((crdB.x - position3d.x) * scaleX + position2d.x, (crdB.y - position3d.y) * scaleY + position2d.y);
						texCoord += Vec2f((crdC.x - position3d.x) * scaleX + position2d.x, (crdC.y - position3d.y) * scaleY + position2d.y);
					}
				}
			}
		}
	}
}
*/
bool String::greaterThanCaseInsensitive(const String& par1)const{
	int len = length;
	const char* pa = pointer->rawChar;
	const char* pb = par1.pointer->rawChar;
	if(len > par1.length){
		len = par1.length;
	}
	for(int i = 0; i < len; i++){
		char a = upperCaseChar(pa[i]);
		char b = upperCaseChar(pb[i]);
		if(a == b){
			continue;
		}
		return a > b;
	}
	return false;
}
bool equals(const char* pa, const char* pb, const int lenA, const int lenB){
	if(lenA != lenB){
		return false;
	}
	for(int i = 0; i < lenA; i++){
		if(pa[i] != pb[i]){
			return false;
		}
	}
	return true;
}
bool notEquals(const char* pa, const char* pb, const int lenA, const int lenB){
	if(lenA != lenB){
		return true;
	}
	for(int i = 0; i < lenA; i++){
		if(pa[i] != pb[i]){
			return true;
		}
	}
	return false;
}
bool greaterThan(const char* pa, const char* pb, const int lenA, const int lenB){
	int len = lenA;

	if(lenA > lenB){
		len = lenB;
	}
	for(int i = 0; i < len; i++){
		char a = pa[i];
		char b = pb[i];
		if(a == b){
			continue;
		}
		return a > b;
	}
	return false;
}
bool lessThan(const char* pa, const char* pb, const int lenA, const int lenB){
	int len = lenA;

	if(lenA > lenB){
		len = lenB;
	}
	for(int i = 0; i < len; i++){
		char a = pa[i];
		char b = pb[i];
		if(a == b){
			continue;
		}
		return a < b;
	}
	return false;
}
bool greaterEqual(const char* pa, const char* pb, const int lenA, const int lenB){
	if(greaterThan(pa, pb, lenA, lenB)){
		return true;
	}
	return equals(pa, pb, lenA, lenB);
}
bool lessEqual(const char* pa, const char* pb, const int lenA, const int lenB){
	if(lessThan(pa, pb, lenA, lenB)){
		return true;
	}
	return equals(pa, pb, lenA, lenB);
}

inline Art* getMatchingName(M_Art& partList, const String& matchName){
	for(int k = partList.length; --k >= 0;){
		Art* part = partList[k];
		if(part->name == matchName){
			return part;
		}
	}
	return 0;
}

inline NurbsPatch* getMatchingName(M_<NurbsPatch*>& cpNodeList, const String& matchName){
	for(int k = cpNodeList.length; --k >= 0;){
		NurbsPatch* patch = cpNodeList[k];
		if(patch->name == matchName){
			return patch;
		}
	}
	return 0;
}

Art* getMirrorPartNoName(Art* part){
	if(fabs(part->absolutePivotPoint.x) > .000001){
		Vec3f r_absolutePivotPoint = part->absolutePivotPoint;
		r_absolutePivotPoint.x = -r_absolutePivotPoint.x;

		M_Art& partList = part->smls->partList;
		for(int k = partList.length; --k >= 0;){
			Art* r_part = partList[k];
			if(r_part->mirror &&
				fabs(r_part->absolutePivotPoint.x - r_absolutePivotPoint.x) < .000001 &&
				fabs(r_part->absolutePivotPoint.y - r_absolutePivotPoint.y) < .000001 &&
				fabs(r_part->absolutePivotPoint.z - r_absolutePivotPoint.z) < .000001
				){
				return r_part;
			}
		}
	}
	return 0;
}

bool Art::isLeftRightName(){
	if(name.length > 2){
		if(name[1] < 'a' ){ //if uppercase or underbar
			if(name[0] == 'l'){
				return true;
			}
			else if(name[0] == 'r'){
				return true;
			}
		}
	}
	return false;
}

Build* getMatchingMirrorObject(M_Build& build, const String& mirrorName){
	for(int k = 0; k < build.length; k++){
		Build* bld = build[k];
		if(bld->name == mirrorName){
			return bld;
		}
		else if(GroupBuild* groupBuild = dynamic_cast<GroupBuild*>(bld)){
			if(Build* build = getMatchingMirrorObject(groupBuild->build.refValue(), mirrorName)){
				return build;
			}
		}
	}
	return 0;
}

void BoxBuild::setMirrorObjects(BoxBuild* mirrorContainer){
	if(mirror && mirrorContainer->mirror){
		bool flag = false;
		if(mirror != getPatch()->mirror || this->mirrorContainer != mirrorContainer || mirrorContainer->mirrorContainer != this){
			this->mirrorContainer = mirrorContainer;
			mirrorContainer->mirrorContainer = this;

			for(int k = numOfPatches(); --k>=0;){
				NurbsPatch* patch = m_patch[k];
				NurbsPatch* mpatch = mirrorContainer->m_patch[k];
				patch->mirror.refValue() = true;
				mpatch->mirror.refValue() = true;
				patch->mirrorPatch = mpatch;
				mpatch->mirrorPatch = patch;


				if(patch->stem.length){
					if(Stem* stem = dynamic_cast<Stem*>(patch->stem[0].p)){
						stem->insideOut.refValue() = false;
					}
				}
				patch->insideOut.refValue() = false;

				if(mpatch->stem.length){
					if(Stem* stem = dynamic_cast<Stem*>(mpatch->stem[0].p)){
						stem->insideOut.refValue() = false;
					}
				}
				mpatch->insideOut.refValue() = false;



			}
			processFields();
			mirrorContainer->processFields();
		}
	}
}

void setMirrorObjectsLoop(M_Build& m_build){
	String mirrorName;
	for(int k = 0; k < m_build.length; k++){
		Build* build = m_build[k];
		if(dynamic_cast<NurbsPatch*>(build)){
			continue;
		}
		if(NurbsLathe* lathe = dynamic_cast<NurbsLathe*>(build)){
			lathe->mirrorContainer = 0;
		}
		if(GroupBuild* groupBuild = dynamic_cast<GroupBuild*>(build)){
			setMirrorObjectsLoop(groupBuild->build.refValue());
		}
		else if(build->name.length > 1){
			if(build->name[1] < 'a' ){ //if uppercase or underbar
				if(build->name[0] == 'l'){
					mirrorName = build->name;
					mirrorName[0] = 'r';
					if(Build* mirrorBuild = getMatchingMirrorObject(m_build, mirrorName)){
						if(NurbsLathe* lathe = dynamic_cast<NurbsLathe*>(build)){
							if(NurbsLathe* mirrorContainer = dynamic_cast<NurbsLathe*>(mirrorBuild)){
								if(lathe->mirror && mirrorContainer->mirror){
									bool flag = false;
									if(lathe->mirror != lathe->getPatch()->mirror || lathe->mirrorContainer != mirrorContainer || mirrorContainer->mirrorContainer != lathe){
										lathe->mirrorContainer = mirrorContainer;
										mirrorContainer->mirrorContainer = lathe;
										lathe->getPatch()->mirrorPatch = mirrorContainer->getPatch();
										mirrorContainer->getPatch()->mirrorPatch = lathe->getPatch();
										lathe->processFields();
									}
								}
							}
						}
						else if(BendyLathe* bendy = dynamic_cast<BendyLathe*>(build)){
							if(BendyLathe* mirrorContainer = dynamic_cast<BendyLathe*>(mirrorBuild)){
								if(bendy->mirror && mirrorContainer->mirror){
									bool flag = false;
									if(bendy->mirror != bendy->getPatch()->mirror || bendy->mirrorContainer != mirrorContainer || mirrorContainer->mirrorContainer != bendy){
										bendy->mirrorContainer = mirrorContainer;
										mirrorContainer->mirrorContainer = bendy;
										bendy->getPatch()->mirrorPatch = mirrorContainer->getPatch();
										mirrorContainer->getPatch()->mirrorPatch = bendy->getPatch();
										bendy->processFields();
									}
								}
							}
						}
						else if(SphereBuild* sphere = dynamic_cast<SphereBuild*>(build)){
							if(SphereBuild* mirrorContainer = dynamic_cast<SphereBuild*>(mirrorBuild)){
								if(sphere->mirror && mirrorContainer->mirror){
									bool flag = false;
									if(sphere->mirror != sphere->getPatch()->mirror || sphere->mirrorContainer != mirrorContainer || mirrorContainer->mirrorContainer != sphere){
										sphere->mirrorContainer = mirrorContainer;
										mirrorContainer->mirrorContainer = sphere;
										sphere->getPatch()->mirrorPatch = mirrorContainer->getPatch();
										mirrorContainer->getPatch()->mirrorPatch = sphere->getPatch();
										sphere->processFields();
									}
								}
							}
						}
						else if(CylinderBuild* cylinder = dynamic_cast<CylinderBuild*>(build)){
							if(CylinderBuild* mirrorContainer = dynamic_cast<CylinderBuild*>(mirrorBuild)){
								if(cylinder->mirror && mirrorContainer->mirror){
									bool flag = false;
									if(cylinder->mirror != cylinder->getPatch()->mirror || cylinder->mirrorContainer != mirrorContainer || mirrorContainer->mirrorContainer != cylinder){
										cylinder->mirrorContainer = mirrorContainer;
										mirrorContainer->mirrorContainer = cylinder;
										cylinder->getPatch()->mirrorPatch = mirrorContainer->getPatch();
										mirrorContainer->getPatch()->mirrorPatch = cylinder->getPatch();
										cylinder->processFields();
									}
								}
							}
						}
						else if(ConeBuild* cone = dynamic_cast<ConeBuild*>(build)){
							if(ConeBuild* mirrorContainer = dynamic_cast<ConeBuild*>(mirrorBuild)){
								if(cone->mirror && mirrorContainer->mirror){
									bool flag = false;
									if(cone->mirror != cone->getPatch()->mirror || cone->mirrorContainer != mirrorContainer || mirrorContainer->mirrorContainer != cone){
										cone->mirrorContainer = mirrorContainer;
										mirrorContainer->mirrorContainer = cone;
										cone->getPatch()->mirrorPatch = mirrorContainer->getPatch();
										mirrorContainer->getPatch()->mirrorPatch = cone->getPatch();
										cone->processFields();
									}
								}
							}
						}
						else if(TorusBuild* torus = dynamic_cast<TorusBuild*>(build)){
							if(TorusBuild* mirrorContainer = dynamic_cast<TorusBuild*>(mirrorBuild)){
								if(torus->mirror && mirrorContainer->mirror){
									bool flag = false;
									if(torus->mirror != torus->getPatch()->mirror || torus->mirrorContainer != mirrorContainer || mirrorContainer->mirrorContainer != torus){
										torus->mirrorContainer = mirrorContainer;
										mirrorContainer->mirrorContainer = torus;
										torus->getPatch()->mirrorPatch = mirrorContainer->getPatch();
										mirrorContainer->getPatch()->mirrorPatch = torus->getPatch();
										torus->processFields();
									}
								}
							}
						}
						else if(BoxBuild* box = dynamic_cast<BoxBuild*>(build)){
							if(BoxBuild* mirrorContainer = dynamic_cast<BoxBuild*>(mirrorBuild)){
								box->setMirrorObjects(mirrorContainer);
							}
						}
						else if(RectangleBuild* rectangle = dynamic_cast<RectangleBuild*>(build)){
							if(RectangleBuild* mirrorContainer = dynamic_cast<RectangleBuild*>(mirrorBuild)){
								if(rectangle->mirror && mirrorContainer->mirror){
									bool flag = false;
									if(rectangle->mirror != rectangle->getPatch()->mirror || rectangle->mirrorContainer != mirrorContainer || mirrorContainer->mirrorContainer != rectangle){
										rectangle->mirrorContainer = mirrorContainer;
										mirrorContainer->mirrorContainer = rectangle;
										rectangle->getPatch()->mirrorPatch = mirrorContainer->getPatch();
										mirrorContainer->getPatch()->mirrorPatch = rectangle->getPatch();
										rectangle->processFields();
									}
								}
							}
						}
					}
				}
				else if(build->name[0] == 'r'){
					continue;
				}
			}
		}
	}
}


void Seamless::populateMirrorNodes(){
	cpNodeListAdd();
	flagCtrlPolyPointRefs();

	String mirrorName;

	{
		int len = partList.length;

		for(int k = 0; k < len; k++){
			Art* part = partList[k];
			if(part->mirror){
				if(part->name.length == 0){
					if(Art* mirrorPart = getMirrorPartNoName(part)){
						part->mirrorPart = mirrorPart;
						mirrorPart->mirrorPart = part;
					}
				}

				else if(part->name.length > 1){

					if(part->name[1] < 'a' ){ //if uppercase or underbar
						if(part->name[0] == 'l'){
							mirrorName = part->name;
							mirrorName[0] = 'r';
							if(Art* mirrorPart = getMatchingName(partList, mirrorName)){
								if(mirrorPart->mirror){
									part->mirrorPart = mirrorPart;
									mirrorPart->mirrorPart = part;
								}
							}
						}
						else if(part->name[0] == 'r'){
							continue;
						}
					}
				}
			}
		}
	}
	{
		int len = cpNodeList.length;
		for(int k = 0; k < len; k++){
			NurbsPatch* patch = cpNodeList[k];
			if(patch->mirror){
				if(patch->name.length > 1){

					if(patch->name[1] < 'a' ){ //if uppercase or underbar
						if(patch->name[0] == 'l'){
							mirrorName = patch->name;
							mirrorName[0] = 'r';
							if(NurbsPatch* mirrorPatch = getMatchingName(cpNodeList, mirrorName)){
								if(mirrorPatch->mirror){
									patch->mirrorPatch = mirrorPatch;
									mirrorPatch->mirrorPatch = patch;
									M_Node& m_stem = patch->stem.refValue();
									M_Node& m_mirrorStem = mirrorPatch->stem.refValue();
									int len = m_stem.length < m_mirrorStem.length ? m_stem.length : m_mirrorStem.length;
									for(int i = 0; i < len; i++){
										Stem* stem = dynamic_cast<Stem*>(m_stem[i].p);
										Stem* mirrorStem = dynamic_cast<Stem*>(m_mirrorStem[i].p);
										if(stem->mirror && mirrorStem->mirror){
											stem->mirrorStem = mirrorStem;
											mirrorStem->mirrorStem = stem;
										}
									}
								}
							}
						}
						else if(patch->name[0] == 'r'){
							continue;
						}
					}
				}
			}
		}
	}
	setMirrorObjectsLoop(build.refValue());
}

void Art::Mirror::operator=(const bool& par){
	refValue() = par;
	home.smls->populateMirrorNodes();
}

void NurbsPatch::Mirror::operator=(const bool& par){
	refValue() = par;
	home.smls->populateMirrorNodes();
}



Art* getToe(Art* ankle){
	M_SArt& m_part = ankle->part.refValue();
	if(m_part.length){
		return m_part[0];
	}
	return 0;
}

/*
Vec3d calcPositionAngle(Vec3d base, Art* knee, Vec3d kneePos, Art* ankle, bool& runOutlength){
	Vec3d end = conv3fto3d(ankle->ikPos);
	Vec3d guid = kneePos - base;
	Vec3d diff = end - base;
	Vec3d diff2 = diff;
	Rotation4d hipToAnkleRot = Rotation4d(Vec3d(0, -1, 0), diff);
	guid = -hipToAnkleRot * guid;
	diff2 = -hipToAnkleRot * diff2;
	guid.y = 0;
	Rotation4d twist = Rotation4d(Vec3d(0, 0, 1), guid);

	double a = diff.length();
	double b = knee->pivotPoint.refValue().length();
	double c = ankle->pivotPoint.refValue().length();

	if(b + c < a){
		runOutlength = true;
	}

	double _a = (b * b + c * c - a * a) / (2 * b * c);
	double _b = (a * a + c * c - b * b) / (2 * a * c);
	double _c = (a * a + b * b - c * c) / (2 * a * b);

	if(_a > 1){
		_a = 1;
	}
	if(_b > 1){
		_b = 1;
	}
	if(_c > 1){
		_c = 1;
	}
	if(_a < -1){
		_a = -1;
	}
	if(_b < -1){
		_b = -1;
	}
	if(_c < -1){
		_c = -1;
	}

	double A = acos(_a);
	double B = acos(_b);
	double C = acos(_c);
	double sum = A + B + C;

	kneePos = Vec3d(0, -knee->pivotPoint.refValue().length(), 0);

	kneePos = Rotation4d(1, 0, 0, -C) * kneePos;
	kneePos = twist * kneePos;
	kneePos = hipToAnkleRot * kneePos;
	kneePos = base + kneePos;

	return kneePos;
}
*/

Vec3d calcPositionAngleAndBlend(Vec3d base, Art* knee, Art* ankle, double& hipToAnkleDistance, Vec3d& anklePos, bool& runOutlength){
	Vec3d end = conv3fto3d(ankle->ikPos);
	Vec3d guid = conv3fto3d(knee->ikPos) - base;
	Vec3d diff = end - base;
	Vec3d diff2 = diff;
	Rotation4d hipToAnkleRot = Rotation4d(Vec3d(0, -1, 0), diff);
	guid = -hipToAnkleRot * guid;
	diff2 = -hipToAnkleRot * diff2;
	guid.y = 0;
	Rotation4d twist = Rotation4d(Vec3d(0, 0, 1), guid);

	double a = diff.length();
	double b = knee->pivotPoint.refValue().length();
	double c = ankle->pivotPoint.refValue().length();

	if(b + c < a){
		runOutlength = true;
	}

	double _a = (b * b + c * c - a * a) / (2 * b * c);
	double _b = (a * a + c * c - b * b) / (2 * a * c);
	double _c = (a * a + b * b - c * c) / (2 * a * b);

	if(_a > 1){
		_a = 1;
	}
	if(_b > 1){
		_b = 1;
	}
	if(_c > 1){
		_c = 1;
	}
	if(_a < -1){
		_a = -1;
	}
	if(_b < -1){
		_b = -1;
	}
	if(_c < -1){
		_c = -1;
	}

	double A = acos(_a);
	double B = acos(_b);
	double C = acos(_c);
	double sum = A + B + C;

	Vec3d kneePos = Vec3d(0, -knee->pivotPoint.refValue().length(), 0);
	anklePos = Vec3d(0, -ankle->pivotPoint.refValue().length(), 0);

	double A2 = (PI - A) * (1 - knee->fkb) + knee->fka * knee->fkb;
	anklePos = Rotation4d(1, 0, 0, A2) * anklePos;
	anklePos = anklePos + kneePos;
	hipToAnkleDistance = anklePos.length();
	Rotation4d hipAngle = Rotation4d(anklePos, Vec3d(0,-1,0));
	//double hipAngle = PI - Vec2f(anklePos.z, anklePos.y).getAngle();

	kneePos = hipAngle * kneePos;


	//kneePos = twist * kneePos;
	//kneePos = hipToAnkleRot * kneePos;
	//kneePos = base + kneePos;
	//hipToAnkleDistance = 1;
	anklePos = hipToAnkleRot * Vec3d(0, -hipToAnkleDistance, 0) + base;

	//return kneePos;


//	kneePos = Rotation4d(1, 0, 0, -C) * kneePos;
	kneePos = twist * kneePos;
	kneePos = hipToAnkleRot * kneePos;
	kneePos = base + kneePos;

	return kneePos;
}

Vec3d calcPositionAngle2(Vec3d base, Art* knee, Art* ankle, Art* toe, double& b){
	Vec3d end = conv3fto3d(toe->ikPos);
	Vec3d guid = conv3fto3d(ankle->ikPos) - base;
	Vec3d diff = end - base;
	Vec3d diff2 = diff;
	Rotation4d hipTotoeRot = Rotation4d(Vec3d(0, -1, 0), diff);
	guid = -hipTotoeRot * guid;
	diff2 = -hipTotoeRot * diff2;
	guid.y = 0;
	Rotation4d twist = Rotation4d(Vec3d(0, 0, 1), guid);

	double a = diff.length();
	//double b = knee->pivotPoint.refValue().length() + ankle->pivotPoint.refValue().length();
	b = (conv3fto3d(ankle->ikPos) - base).length();
	double c = toe->pivotPoint.refValue().length();

	double _a = (b * b + c * c - a * a) / (2 * b * c);
	double _b = (a * a + c * c - b * b) / (2 * a * c);
	double _c = (a * a + b * b - c * c) / (2 * a * b);

	if(_a > 1){
		_a = 1;
	}
	if(_b > 1){
		_b = 1;
	}
	if(_c > 1){
		_c = 1;
	}
	if(_a < -1){
		_a = -1;
	}
	if(_b < -1){
		_b = -1;
	}
	if(_c < -1){
		_c = -1;
	}

	double A = acos(_a);
	double B = acos(_b);
	double C = acos(_c);
	double sum = A + B + C;

	Vec3d anklePos = Vec3d(0, -b, 0);

	anklePos = Rotation4d(1, 0, 0, -C) * anklePos;
	anklePos = twist * anklePos;
	anklePos = hipTotoeRot * anklePos;
	anklePos = base + anklePos;

	return anklePos;
}

Vec3d calcPositionAngle3(Vec3d base, Vec3d anklePos, Art* toe, double hipToAnkleDistance){
	Vec3d end = conv3fto3d(toe->ikPos);
	Vec3d guid = anklePos - base;
	Vec3d diff = end - base;
	Vec3d diff2 = diff;
	Rotation4d hipTotoeRot = Rotation4d(Vec3d(0, -1, 0), diff);
	guid = -hipTotoeRot * guid;
	diff2 = -hipTotoeRot * diff2;
	guid.y = 0;
	Rotation4d twist = Rotation4d(Vec3d(0, 0, 1), guid);

	double a = diff.length();
	//double b = knee->pivotPoint.refValue().length() + ankle->pivotPoint.refValue().length();
	double b = (anklePos - base).length();
	double c = toe->pivotPoint.refValue().length();

	double _a = (b * b + c * c - a * a) / (2 * b * c);
	double _b = (a * a + c * c - b * b) / (2 * a * c);
	double _c = (a * a + b * b - c * c) / (2 * a * b);

	if(_a > 1){
		_a = 1;
	}
	if(_b > 1){
		_b = 1;
	}
	if(_c > 1){
		_c = 1;
	}
	if(_a < -1){
		_a = -1;
	}
	if(_b < -1){
		_b = -1;
	}
	if(_c < -1){
		_c = -1;
	}

	double A = acos(_a);
	double B = acos(_b);
	double C = acos(_c);
	double sum = A + B + C;

	anklePos = Vec3d(0, -b, 0);

	anklePos = Rotation4d(1, 0, 0, -C) * anklePos;
	anklePos = twist * anklePos;
	anklePos = hipTotoeRot * anklePos;
	anklePos = base + anklePos;

	return anklePos;
}

void calcPelvis(Art* pelvis){
	pelvis->rotation = pelvis->ikRot;
	pelvis->translation = pelvis->ikPos - pelvis->absolutePivotPoint;
}

void setAllPinedFields(M_Art& partList){//do we need to set any more than the first pinned part ?
	static bool firstTime = true;
	firstTime = true;
	if(firstTime){
		for(int k = 0; k < partList.length; k++){
			Art* part = partList[k];
			if(part->name == "r_shoulder"){
				part->rotation.refValue() = Rotation(-1, 0, 0, .203423);
				part->rotation.refValue() = Rotation(-1, 0, 0, .20342301);
				part->translation.refValue() = Vec3f();
			}
			if(part->ik){
				part->rotation.refValue() = Rotation();
				part->translation.refValue() = Vec3f();
			}
		}
	}
	firstTime = false;
}


void Seamless::calcNIK() {
	/*
	for (int j = 0; j < partList.length; j++) {
		Art* part = partList[j];
		if (part->pin) {
			calcPelvis(part);
		}
	}
	catMatrix();
	*/

	for (int k = 0; k < nikiList.length; k++) {
		NikInterpolator* niki = nikiList[k];
		niki->calcIK(this, 0);
	}
	nikiList.length = 0;
	//houseKeepVisualClue();
}

void Seamless::calcIKLoop(Art* pelvis, Art* hip, Art* knee, Art* ankle){

	if(hip && hip->niki){
		calcNIK();
		return;
	}
	else if(pelvis && hip && knee){
		//if(ankle->ik){
			//if(ankle->pin){//if its the toe
				//calcPelvis(knee);
			//	catMatrix();
			//}
		if(knee->ik && ankle->ik && knee->pin == false){
			//if(pelvis->pin){
			//	calcPelvis(pelvis);
			//	catMatrix();
			//}
			Vec3d hipPos = conv3fto3d(hip->matrix * hip->absolutePivotPoint);
			bool runOutlength = false;
			double hipToAnkleDistance = 0;
			Vec3d anklePos;
			Vec3d kneePos = calcPositionAngleAndBlend(hipPos, knee, ankle, hipToAnkleDistance, anklePos, runOutlength);

			Matrix hipMatrixNoBend = translationMatrix(-hip->absolutePivotPoint) * scaleMatrix(hip->scale) * translationMatrix(hip->absolutePivotPoint + hip->translation) * pelvis->matrix;
			Vec3d kneePos_noBend = conv3fto3d(hipMatrixNoBend * knee->absolutePivotPoint);

			hip->rotation = convRot4dto4f(g_m_accumRot[pelvis->plidx] * Rotation4d(kneePos_noBend - hipPos, kneePos - hipPos) * g_m_iaccumRot[pelvis->plidx]);
			g_m_accumRot[hip->plidx] = convRot4fto4d(hip->rotation) * g_m_accumRot[pelvis->plidx];
			g_m_iaccumRot[hip->plidx] = g_m_iaccumRot[pelvis->plidx] * convRot4fto4d(-hip->rotation);

			catMatrix();


			if(posCompensator && knee == posCompensator->ankle){
				if(dynamic_cast<Art*>(pelvis->parent)){
					pelvis = pelvis->parent;
				}
				Vec3d kneeComp = conv3fto3d(posCompensator->relPos[posCompensator->momentIdx]);
				Vec3d diff = kneePos - kneeComp;
				pelvis->translation = pelvis->translation - conv3dto3f(diff);
				catMatrix();
				kneePos = kneeComp;
			}
			Matrix kneeMatrixNoBend = translationMatrix(-knee->absolutePivotPoint) * scaleMatrix(knee->scale) * translationMatrix(knee->absolutePivotPoint + knee->translation) * hip->matrix;
			Vec3d anklePos_noBend = conv3fto3d(kneeMatrixNoBend * ankle->absolutePivotPoint);
			Rotation4d r = Rotation4d(anklePos_noBend - kneePos, anklePos - kneePos);

			knee->rotation = convRot4dto4f(g_m_accumRot[hip->plidx] * r * g_m_iaccumRot[hip->plidx]);
			g_m_accumRot[knee->plidx] = convRot4fto4d(knee->rotation) * g_m_accumRot[hip->plidx];
			g_m_iaccumRot[knee->plidx] = g_m_iaccumRot[hip->plidx] * convRot4fto4d(-knee->rotation);
			catMatrix();

			//if(ankle->pin){
			//	calcPelvis(ankle);
			//}
			//else{
			//	ankle->rotation = convRot4dto4f(convRot4fto4d(ankle->ikRot) * g_m_iaccumRot[knee->plidx]);
			//}
			ankle->rotation = convRot4dto4f(convRot4fto4d(ankle->ikRot) * g_m_iaccumRot[knee->plidx]);
			catMatrix();


			if(posCompensator && ankle == posCompensator->ankle){
				if(dynamic_cast<Art*>(pelvis->parent)){
					pelvis = pelvis->parent;
				}

				Vec3d ankleComp = conv3fto3d(posCompensator->relPos[posCompensator->momentIdx]);
				Vec3d diff = anklePos - ankleComp;
				pelvis->translation = pelvis->translation - conv3dto3f(diff);
				catMatrix();
				anklePos = ankleComp;
			}
		}
	}
	else if(knee == 0 && ankle->ik && ankle->pin){//no decendent parts (ankle = pelvis without a leg){
		if(knee == 0){
			if(posCompensator == 0 || posCompensator->ankle->niki == 0){
			//if(ankle->name == "pelvis"){
				calcPelvis(ankle);
				catMatrix();
			}
		}
	}
	M_SArt& m_part = ankle->part.refValue();
	pelvis = hip;
	hip = knee;
	knee = ankle;
	for(int k = 0; k < m_part.length; k++){
		ankle = m_part[k];
		calcIKLoop(pelvis, hip, knee, ankle);
	}
}

void Seamless::calcIK(){

	//setAllPinedFields(partList);
	//catMatrix();
	//catIMatrix();

	g_m_iaccumRot.length = g_m_accumRot.length = partList.length;
	catMatrix(g_m_accumRot);
	catIMatrix(g_m_iaccumRot);

	if(posCompensator && posCompensator->ankle->niki){
		posCompensator->ankle->niki->calcIK(this, posCompensator->ankle);
	}

	for(int k = 0; k < skeleton.length; k++){
		Art* ankle = skeleton.refValue()[k];
		calcIKLoop(0, 0, 0, ankle);
	}
	posCompensator = 0;
}


/*
void Seamless::calcIK(){
	g_m_iaccumRot.length = g_m_accumRot.length = partList.length;
	catMatrix(g_m_accumRot);
	catIMatrix(g_m_iaccumRot);
	for(int k = 0; k < partList.length; k++){
		Art* knee = partList[k];
		if(knee->ik){
			if(knee->pin){//if it the toe
				calcPelvis(knee);
				catMatrix();
			}
			else{
				M_SArt& m_part = knee->part.refValue();
				for(int i = 0; i < m_part.length; i++){
					Art* hip = knee->parent;
					if(dynamic_cast<Art*>(hip) == 0){
						k = partList.length;
						break;
					}
					Art* pelvis = hip->parent;
					Art* ankle = m_part[i];
					if(ankle->ik == false || pelvis == 0){
						k = partList.length;
						break;
					}

					//Art* anklePosP = partList[partList.length - 1];
					//Art* kneePosP = partList[partList.length - 2];
					//Art* ankle_guide = partList[partList.length - 3];
					//Art* knee_guide = partList[partList.length - 4];

					Vec3d hipPos = conv3fto3d(hip->matrix * hip->absolutePivotPoint);
					bool runOutlength = false;
					double hipToAnkleDistance = 0;
					Vec3d anklePos;
					Vec3d kneePos = calcPositionAngleAndBlend(hipPos, knee, ankle, hipToAnkleDistance, anklePos, runOutlength);
					//anklePosP->translation = anklePos;
					//kneePosP->translation = kneePos;

					Matrix hipMatrixNoBend = translationMatrix(-hip->absolutePivotPoint) * scaleMatrix(hip->scale) * translationMatrix(hip->absolutePivotPoint + hip->translation) * pelvis->matrix;
					Vec3d kneePos_noBend = conv3fto3d(hipMatrixNoBend * knee->absolutePivotPoint);

					hip->rotation = convRot4dto4f(g_m_accumRot[pelvis->plidx] * Rotation4d(kneePos_noBend - hipPos, kneePos - hipPos) * g_m_iaccumRot[pelvis->plidx]);

					g_m_accumRot[hip->plidx] = convRot4fto4d(hip->rotation) * g_m_accumRot[pelvis->plidx];
					g_m_iaccumRot[hip->plidx] = g_m_iaccumRot[pelvis->plidx] * convRot4fto4d(-hip->rotation);

					catMatrix();

					if(ankle->pin == false){
						Art* toe = getToe(ankle);
						if(toe == 0){
							k = partList.length;
							break;
						}

						//runOutlength = false;
						//if(runOutlength){
						//	knee->rotation = Rotation();
						//	catMatrix();
						//	double b = 0;
						//	anklePos = calcPositionAngle2(hipPos, knee, ankle, toe, b);
						//	Vec3d anklePos_noBend = conv3fto3d(hipMatrixNoBend * ankle->absolutePivotPoint);
						//	Rotation4d r = Rotation4d(anklePos_noBend - hipPos, anklePos - hipPos);
						//	hip->rotation = convRot4dto4f(convRot4fto4d(pelvis->rotation) * r * convRot4fto4d(-pelvis->rotation));
						//	catMatrix();

						//	if(posCompensator && ankle == posCompensator->ankle){
						//		Vec3f anklePos = ankle->matrix * ankle->absolutePivotPoint;
						//		Vec3f diff = anklePos - posCompensator->relPos[posCompensator->momentIdx];
						//		pelvis->translation = pelvis->translation - diff;
						//		catMatrix();
						//	}

						//	Matrix ankleMatrixNoBend = translationMatrix(-ankle->absolutePivotPoint) * scaleMatrix(ankle->scale) * translationMatrix(ankle->absolutePivotPoint + ankle->translation) * hip->matrix;
						//	Vec3d toePos_noBend = conv3fto3d(ankleMatrixNoBend * toe->absolutePivotPoint);
						//	r = Rotation4d(toePos_noBend - anklePos, conv3fto3d(toe->ikPos) - anklePos);
						//	ankle->rotation =  convRot4dto4f(convRot4fto4d(knee->rotation) *  convRot4fto4d(hip->rotation) *  convRot4fto4d(pelvis->rotation) * r *  convRot4fto4d(-pelvis->rotation) *  convRot4fto4d(-hip->rotation) *  convRot4fto4d(-knee->rotation));
						//	catMatrix();
						//}
						//else{

						Matrix kneeMatrixNoBend = translationMatrix(-knee->absolutePivotPoint) * scaleMatrix(knee->scale) * translationMatrix(knee->absolutePivotPoint + knee->translation) * hip->matrix;
						Vec3d anklePos_noBend = conv3fto3d(kneeMatrixNoBend * ankle->absolutePivotPoint);
						Rotation4d r = Rotation4d(anklePos_noBend - kneePos, anklePos - kneePos);
						knee->rotation = convRot4dto4f(g_m_accumRot[hip->plidx] * r * g_m_iaccumRot[hip->plidx]);
						catMatrix();

						Vec3d anklePos2 = calcPositionAngle3(hipPos, anklePos, toe, hipToAnkleDistance);


						r = Rotation4d(anklePos - hipPos, anklePos2 - hipPos);
						hip->rotation = convRot4dto4f(convRot4fto4d(hip->rotation) * r);
						g_m_accumRot[hip->plidx] = convRot4fto4d(hip->rotation) * g_m_accumRot[pelvis->plidx];
						g_m_iaccumRot[hip->plidx] = g_m_iaccumRot[pelvis->plidx] * convRot4fto4d(-hip->rotation);
						g_m_accumRot[knee->plidx] = convRot4fto4d(knee->rotation) * g_m_accumRot[hip->plidx];
						g_m_iaccumRot[knee->plidx] = g_m_iaccumRot[hip->plidx] * convRot4fto4d(-knee->rotation);
						catMatrix();

						Vec3d anklePos = conv3fto3d(ankle->matrix * ankle->absolutePivotPoint);

						if(posCompensator && ankle == posCompensator->ankle){
							Vec3d ankleComp = conv3fto3d(posCompensator->relPos[posCompensator->momentIdx]);
							Vec3d diff = anklePos - ankleComp;
							pelvis->translation = pelvis->translation - conv3dto3f(diff);
							catMatrix();
							anklePos = ankleComp;
						}

						Matrix ankleMatrixNoBend = translationMatrix(-ankle->absolutePivotPoint) * scaleMatrix(ankle->scale) * translationMatrix(ankle->absolutePivotPoint + ankle->translation) * knee->matrix;
						Vec3d toePos_noBend = conv3fto3d(ankleMatrixNoBend * toe->absolutePivotPoint);
						r = Rotation4d(toePos_noBend - anklePos, conv3fto3d(toe->ikPos) - anklePos);
						ankle->rotation = convRot4dto4f(g_m_accumRot[knee->plidx] * r * g_m_iaccumRot[knee->plidx]);
						//catMatrix();
						g_m_iaccumRot[ankle->plidx] = g_m_iaccumRot[knee->plidx] * convRot4fto4d(-ankle->rotation);
						//}
						toe->rotation = convRot4dto4f(convRot4fto4d(toe->ikRot) * g_m_iaccumRot[ankle->plidx]);
						k+=3;
					}
					else{

						if(posCompensator && knee == posCompensator->ankle){
							//catMatrix();
							//kneePos = conv3fto3d(knee->matrix * knee->absolutePivotPoint);
							if(dynamic_cast<Art*>(pelvis->parent)){
								pelvis = pelvis->parent;
							}
							Vec3d kneeComp = conv3fto3d(posCompensator->relPos[posCompensator->momentIdx]);
							Vec3d diff = kneePos - kneeComp;
							pelvis->translation = pelvis->translation - conv3dto3f(diff);
							catMatrix();
							kneePos = kneeComp;
						}
						Matrix kneeMatrixNoBend = translationMatrix(-knee->absolutePivotPoint) * scaleMatrix(knee->scale) * translationMatrix(knee->absolutePivotPoint + knee->translation) * hip->matrix;
						Vec3d anklePos_noBend = conv3fto3d(kneeMatrixNoBend * ankle->absolutePivotPoint);
						Rotation4d r = Rotation4d(anklePos_noBend - kneePos, anklePos - kneePos);

						knee->rotation = convRot4dto4f(g_m_accumRot[hip->plidx] * r * g_m_iaccumRot[hip->plidx]);
						//catMatrix();

						//ankle->rotation = ankle->ikRot * -pelvis->rotation * -hip->rotation * -knee->rotation;
						//catRot = catRot * convRot4fto4d(-rotation);
						g_m_iaccumRot[knee->plidx] = g_m_iaccumRot[hip->plidx] * convRot4fto4d(-knee->rotation);

						ankle->rotation = convRot4dto4f(convRot4fto4d(ankle->ikRot) * g_m_iaccumRot[knee->plidx]);



						if(posCompensator && ankle == posCompensator->ankle){
							if(dynamic_cast<Art*>(pelvis->parent)){
								pelvis = pelvis->parent;
							}

							Vec3d ankleComp = conv3fto3d(posCompensator->relPos[posCompensator->momentIdx]);
							Vec3d diff = anklePos - ankleComp;
							pelvis->translation = pelvis->translation - conv3dto3f(diff);
							catMatrix();
							anklePos = ankleComp;
						}





						k+=2;
					}
				}
			}
		}
	}
	posCompensator = 0;
}
*/

void NurbsTrack::do_set_fraction(){

	float t = frame;
	const Vec3f* p = keyValue.getArray();
	const float* w = weight.getArray();;
	const float* knotPtr = knot.getArray();



	M_Float& key = this->key.refValue();
	M_Vec3f& keyValue = this->keyValue.refValue();
	int i = 0;
	if(key.length < 2)goto here;
	for(; i < key.length-1; i++){
		if(key[i] > frame){
			i--;
			break;
		}
	}
	if(i < 0){
here:
		if(key.length > 0){
			pos.refValue() = keyValue[0];
		}
		return;
	}
	if(i == key.length - 1) i = key.length - 2;
	float diff = key[i + 1] - key[i];

	int seg = i;
	const int dm = dim[seg];
	p += keyValueRedirect[seg];
	if(dm <= 2){
		float interp = 1;
		if(diff > .0000001){
			interp = 1.0 / diff * (frame - key[i]);
			pos.refValue() = p[0] + (p[1] - p[0]) * interp;
		}
		else{
			pos.refValue() = p[0];
		}
		return;
	}

	t = t - key[i];
	t = t / diff;

	if(seg >= keyValueRedirect.length){
		pos.refValue() = keyValue[keyValue.length - 1];
		return;
	}
	int const knotOffset = knotRedirect[seg];
	knotPtr += knotOffset;

	float mint = minT(knotOffset);
	int const endOfKnotSeg = seg < (knotRedirect.length - 1) ? knotRedirect[seg + 1] : static_cast<int>(knot.length);
	float maxt = maxT(endOfKnotSeg);

	t = (maxt - mint) * t + mint;
	CurvePoint(dm - 1, degree, knotPtr, p, w, t, pos.refValue());
}

void NurbsTrack::processFields(){
	int total = keyValue.length;
	keyValueRedirect.length = 0;
	knotRedirect.length = 0;
	knot.length = 0;
	weight.length = total;

	for(int i = total; --i>=0;){
		weight[i] = 1;
	}

	int keyValueOffset = 0;
	for(int k = 0, knotOffset = 0; k < dim.length; k++){
		int dm = dim[k];
		keyValueRedirect += keyValueOffset;
		knotRedirect += knotOffset;
		if(dm > 2){
			updateNurbsCurveKnotMid0(dm, knot, knotOffset);
		}
		knotOffset = knot.length;
		keyValueOffset += dm - 1;
	}
}

void NurbsTrack::take(float slider, Node* node, Field* field){}
void NurbsTrack::deleteSample(float slider){}
void NurbsTrack::connectNode(Node* view, Field* field, String& name, Anim* anim, Function* function){}

void Anim::updateDisplayFrom(){
	if(slider > displayFrom + zoom / period){
		displayFrom = slider - zoom / period;
	}
	else if(slider < displayFrom){
		displayFrom = slider;
	}
}

void SphereBuild::remTransform(){
	remScale = scale;
	remRotation = rotation;
	remTranslation = translation;
}

void CylinderBuild::remTransform(){
	remScale = scale;
	remRotation = rotation;
	remTranslation = translation;
}

void ConeBuild::remTransform(){
	remScale = scale;
	remRotation = rotation;
	remTranslation = translation;
}

void TorusBuild::remTransform(){
	remScale = scale;
	remRotation = rotation;
	remTranslation = translation;
}

void BezierLatheBuild::remTransform(){
	remScale = scale;
	remRotation = rotation;
	remTranslation = translation;
}

void BoxBuild::remTransform(){
	remScale = scale;
	remRotation = rotation;
	remTranslation = translation;
}

void RectangleBuild::remTransform(){
	remScale = scale;
	remRotation = rotation;
	remTranslation = translation;
}

void NurbsLathe::remTransform(){
	remScale = scale;
	remRotation = rotation;
	remTranslation = translation;
}

void BendyLathe::remTransform(){
	remScale = scale;
	remRotation = rotation;
	remTranslation = translation;
}

void Seamless::remObjTransform(){
	remScale = scale;
	remRotation = rotation;
	remTranslation = translation;
}

void BoxBuild::setCage(bool state){
	for(int k = 6; --k>=0;){
		m_patch[k]->cage = state;
	}
}

void BezierLatheBuild::setCage(bool state){
	for(int k = m_patch.length; --k>=0;){
		m_patch[k]->cage = state;
	}
}

bool ControlPoint::compGT(const ControlPoint& cp) const{
	if(p->fromPoint > cp.p->fromPoint){
		return true;
	}
	if(p->fromPoint == cp.p->fromPoint){
		if(i > cp.i){
			return true;
		}
	}
	return false;
}

bool ControlPoint::compEQ(const ControlPoint& cp) const{
	if(p == cp.p && i == cp.i){
		return true;
	}
	return false;
}

//int NurbsPatch::uGetCurveType()const{
//	return checkNurbsType(dim.x, uKnot);
//}

//int NurbsPatch::vGetCurveType()const{
//	return checkNurbsType(dim.y, vKnot);
//}

void synchronizeCells(NurbsPatch* patch, Seamless* smls, Stem* stem){

	int uDim = patch->dim.x;
	int vDim = patch->dim.y;
	stem->cells = Vec2f(uDim / 2, vDim / 2);
	int uCells = stem->cells.x;
	int vCells = stem->cells.y;

	M_Int& staves = stem->staves;
	M_Int& bands = stem->bands;

	int upperUOffset = uDim * (vDim - 1);
	int upperStavesOffset = staves.length - uCells;
	int upperBandsOffset = bands.length - vCells;

	for(int u = 0; u + 2 < uDim; u += 2){
		if(smls->inPolyCenter(patch, u, u + 1, u + 2)){
			staves[u / 2] = 0;
		}
		if(smls->inPolyCenter(patch, upperUOffset + u, upperUOffset + u + 1, upperUOffset + u + 2)){
			staves[upperStavesOffset + u / 2] = 0;
		}
	}
	for(int v = 0; v + 2 < vDim; v += 2){
		if(smls->inPolyCenter(patch, v * uDim, (v + 1) * uDim, (v + 2) * uDim)){
			bands[v / 2] = 0;
		}
		if(smls->inPolyCenter(patch, v * uDim + uDim - 1, (v + 1) * uDim + uDim - 1, (v + 2) * uDim + uDim - 1)){
			bands[upperBandsOffset + v / 2] = 0;
		}
	}
}


void NurbsPatch::synchronizeAllStemCells(){
	M_Node& m_stem = this->stem.refValue();
	if(m_stem.length){
		for(int k = m_stem.length; --k>= 0;){
			if(Stem* stem = dynamic_cast<Stem*>(m_stem[k].p)){
				if(stem->sync){
					synchronizeCells(this, smls, stem);
				}
			}
		}
	}
}

void Seamless::synchronizeAllStemCells(){
	for(int k = this->cpNodeList.length; --k >= 0;){
		cpNodeList[k]->synchronizeAllStemCells();
	}
}


void Seamless::Bnspe::operator=(const bool& par){
	if(par){
		M_Art& partList = home.partList;
		M_<NurbsPatch*>& cpNodeList = home.cpNodeList;
		//if(cpNodeList.length){
		//	NurbsPatch* patch = cpNodeList[0].patch;
		//	for(int k = partList.length; --k>=0;){
		//		Art* part = partList[k];
		//		part->uvCoord.refValue().length = part->coord.length;
		//		M_NurbsPatch& patchOwner = part->patchOwner;
		//		patchOwner.length = part->coord.length;
		//		for(int j = patchOwner.length; --j>=0;){
		//			patchOwner[j] = patch;
		//		}
		//	}
		//}
		for(int k = partList.length;--k >= 0;){
			Art* part = partList[k];
			part->doClearCycle();
		}
		refValue() = par;
		home.resetRelevantParts();

		for(int k = 0; k < cpNodeList.length; k++){
			NurbsPatch* patch = cpNodeList[k];
			patch->uJoinEnds.refValue() = false;
			patch->beginSP.refValue() = false;
			patch->endSP.refValue() = false;
			int uMid = checkNurbsType(patch->dim.x, patch->uKnot);
			int vMid = checkNurbsType(patch->dim.y, patch->vKnot);
			if(NurbsPatch::MID_1 != uMid){
				patch->convert_to_1_u_mid_points();
			}
			if(NurbsPatch::MID_1 != vMid){
				patch->convert_to_1_v_mid_points();
			}
			patch->bnspeBuildPolies();
		}
		home.populateMirrorNodes();
	}
	refValue() = par;
	home.ctrf = true;
}


Vec3f NurbsPatch::generate(Vec2f t){
	Vec3f dest;

	int uDim=dim.x;
	int vDim=dim.y;
	int uOrder=order.x;
	int vOrder=order.y;

	float* uKnotPtr = uKnot.refValue().getArray();
	float* vKnotPtr = vKnot.refValue().getArray();
	Vec3f* pointPtr = point.refValue().getArray();
	float* weightPtr = weight.refValue().getArray();

	Vec3f* topPointPtr = pointPtr+uDim*vDim-uDim;
	float* topWeightPtr = weightPtr+uDim*vDim-uDim;

	Vec3f* r_pointPtr = r_point.getArray();
	float* r_weightPtr = r_weight.getArray();


	if(t.x > 1){
		t.x-=1;
	}
	if(t.x < 0 || t.y < 0 || t.y > 1){
		return dest;
	}
	float u = (maxU()-minU())*t.x + minU();
	float v = (maxV()-minV())*t.y + minV();


	if(t.x == 1){
		CurvePoint(vDim-1, vDegree, vKnotPtr, r_pointPtr, r_weightPtr, v, dest);
	}
	else if(t.y == 1){
		CurvePoint(uDim-1, uDegree, uKnotPtr, topPointPtr, topWeightPtr, u, dest);
	}
	else{
		SurfacePoint(u, v, uOrder, vOrder, uDim, vDim, uKnotPtr, vKnotPtr, pointPtr, weightPtr, &dest);
	}
	return dest;
}


void NurbsPatch::generate3(M_Vec3f& dst, const int dstIdx, Vec2f t){

	int uDim=dim.x;
	int vDim=dim.y;
	int uOrder=order.x;
	int vOrder=order.y;

	Vec3f* coordPtr = dst.getArray();
	float* uKnotPtr = uKnot.refValue().getArray();
	float* vKnotPtr = vKnot.refValue().getArray();
	Vec3f* pointPtr = point.refValue().getArray();
	float* weightPtr = weight.refValue().getArray();

	Vec3f* topPointPtr = pointPtr+uDim*vDim-uDim;
	float* topWeightPtr = weightPtr+uDim*vDim-uDim;

	Vec3f* r_pointPtr = r_point.getArray();
	float* r_weightPtr = r_weight.getArray();


	if(t.x > 1){
		t.x-=1;
	}
	if(t.x < 0 || t.y < 0 || t.y > 1){
		return;
	}
	float u = (maxU()-minU())*t.x + minU();
	float v = (maxV()-minV())*t.y + minV();


	if(t.x == 1){
		CurvePoint(vDim-1, vDegree, vKnotPtr, r_pointPtr, r_weightPtr, v, *(coordPtr + dstIdx));
	}
	else if(t.y == 1){
		CurvePoint(uDim-1, uDegree, uKnotPtr, topPointPtr, topWeightPtr, u, *(coordPtr + dstIdx));
	}
	else{
		SurfacePoint(u, v, uOrder, vOrder, uDim, vDim, uKnotPtr, vKnotPtr, pointPtr, weightPtr, coordPtr + dstIdx);
	}
}

void Hair::doBuild(){

	doLocksUpdateStrands();

	if(Art* part = dynamic_cast<Art*>(this->part.p)){
		Seamless* smls = part->smls;
		M_Strand& m_strand = this->strand.refValue();
		int strandlen = m_strand.length;
		for(int k = 0; k < strandlen; k++){
			Strand* strand = m_strand[k];
			if(smls->bnspe){
				smls->fastDeletePoliesOwnedByPatch(strand);
			}
			if(strand->cage){
				//strand->ctrlLine->processFields();
				//strand->geometry->processFields();
				strand->generateGeometry(part, radius, staves, bands);
				strand->geometry->doBuild();
			}
		}
	}
}

void Hair::doLocksUpdateStrands(){
	int lockListLen = lockList.length;
	for(int k = 0; k < lockListLen; k++){
		Lock* lock = lockList[k];
		M_Vec2f& hole = lock->hole.refValue();
		M_Int& scpi = lock->scpi.refValue();
		M_<Strand*>& scpiLathe = lock->scpiLathe;
		int nHoles = hole.length;
		for(int i = 0; i < nHoles; i++){
			Strand* sd = scpiLathe[i];
			if(sd->cage){
				lock->generate3(sd->point.refValue(), scpi[i], hole[i]);
				//sd->setPointAndCp(scpi[i], sd->point[scpi[i]]);
				int sdI = scpi[i];
				int cpi = sd->cpIndex[sdI];
				sd->smls->cp.refValue()[cpi] = sd->point[sdI];
			}
		}
	}
}


void Hair::convertOldScpiToNew(){
	for(int k = 0; k < numOfHoles.length; k++){
		Lock* lock = lockList[k];
		int nHoles = numOfHoles[k];
		int beginIdx = beginIndex[k];
		M_Vec2f& lockHole = lock->hole.refValue();
		M_Int& lockScpi = lock->scpi.refValue();

		lockHole.length = nHoles;
		lockScpi.length = nHoles;

		for(int i = 0; i < nHoles; i++){
			lockHole[i] = hole[beginIdx + i];
			lockScpi[i] = scpi[beginIdx + i];
		}
	}
}

void Hair::buildLockListLoop(Lock* lock, int &_llidx){
	lockList += lock;
	lock->llidx = _llidx++;
	M_Node& m_lock = lock->lock.refValue();
	for(int k = 0; k < m_lock.length; k++){
		buildLockListLoop(dynamic_cast<Lock*>(m_lock[k].p), _llidx);
	}
}

void Hair::buildLockList(){
	lockList.length = 0;
	int _llidx = 0;
	M_Node& m_lock = lock.refValue();
	for(int k = 0; k < lock.length; k++){
		buildLockListLoop(dynamic_cast<Lock*>(m_lock[k].p), _llidx);
	}
}

void Hair::processFields(){
	buildLockList();

	if(scpi.length){
		convertOldScpiToNew();
		scpi.refValue().length = 0;
		hole.refValue().length = 0;
	}



	M_Strand& m_strand = this->strand.refValue();
	if(postLoadProcess2){
		postLoadProcess2 = false;
		M_<Strand*> contigLookup;
		M_Int contigLookupOffset;

		int totalLen = 0;
		for(int k = 0; k < m_strand.length; k++){
			Strand* sd = m_strand[k];

			int uDim = 1;
			int vDim = sd->point.length;

			sd->hairFrom = totalLen;
			sd->slidx = k;

			contigLookup.length += vDim;
			contigLookupOffset.length += vDim;
			for(int j = vDim; --j>=0;){
				contigLookup[totalLen + j] = sd;
				contigLookupOffset[totalLen + j] = totalLen;
			}

			totalLen += vDim;
		}

		for(int l = 0; l < lockList.length; l++){
			Lock* lock = lockList[l];
			M_Int& scpi = lock->scpi.refValue();
			M_<Strand*>& scpiLathe = lock->scpiLathe;
			scpiLathe.length = scpi.length;
			for(int k = 0; k < scpi.length; k++){

				int longIdx = scpi[k];
				scpiLathe[k] = contigLookup[longIdx];
				int less = scpi[k] - contigLookupOffset[longIdx];
				scpi[k] = less;
			}
		}
	}
	else{
		int totalLen = 0;
		for(int k = 0; k < m_strand.length; k++){
			Strand* sd = m_strand[k];

			int uDim = 1;
			int vDim = sd->point.length;

			sd->hairFrom = totalLen;
			sd->slidx = k;

			totalLen += vDim;
		}
	}

	strands = m_strand.length;

	if(Art* part = dynamic_cast<Art*>(this->part.p)){
		Seamless* smls = part->smls;
		for(int k = 0; k < m_strand.length; k++){
			Strand* strand = m_strand[k];
			strand->geometry->smls = smls;
			strand->generateGeometry(part, radius, staves, bands);
			//strand->geometry->patchToCentralCPs();
			strand->geometry->cpIndex.refValue().length = 0;
		}
	}

	M_Node& externalNode = textureMapper.refValue();
	if(externalNode.length){
		for(int k = 0; k < m_strand.length; k++){
			NurbsPatch* patch = m_strand[k]->geometry;
			if(patch->stem.length){
				if(Stem* stem = dynamic_cast<Stem*>(patch->stem[0].p)){
					stem->textureMapper.refValue().length = 0;
					stem->textureMapper.refValue() += externalNode;
				}
			}
		}
	}
}

void Hair::processToFileFormat(){
	//M_Strand& m_strand = this->strand.refValue();
	//for(int k = m_strand.length; --k>=0;){
	//	Strand* sd = m_strand[k].p;
	//	sd->geometry->plidx = sd->plidx;
	//}

	int lockListLen = lockList.length;
	for(int k = 0; k < lockListLen; k++){
		Lock* lock = lockList[k];
		M_Int& scpi = lock->scpi.refValue();
		M_<Strand*>& scpiLathe = lock->scpiLathe;
		int nHoles = scpi.length;
		for(int i = 0; i < nHoles; i++){
			scpi[i] += scpiLathe[i]->hairFrom;
		}
	}
}

void Hair::processToWorkFormat(){
	int lockListLen = lockList.length;
	for(int k = 0; k < lockListLen; k++){
		Lock* lock = lockList[k];
		M_Int& scpi = lock->scpi.refValue();
		M_<Strand*>& scpiLathe = lock->scpiLathe;
		int nHoles = scpi.length;
		for(int i = 0; i < nHoles; i++){
			scpi[i] -= scpiLathe[i]->hairFrom;
		}
	}
}

int Hair::refreshHairOffset(){ //also getTotoalLen
	M_Strand& m_strand = this->strand.refValue();
	int totalLen = 0;
	for(int k = 0; k < m_strand.length; k++){
		m_strand[k]->hairFrom = totalLen;
		totalLen += m_strand[k]->point.length;
	}
	return totalLen;
}

Node* Hair::HSFPart::operator=(Node* par){
	SFNode::refValue()=par;
	//home.processFields();
	return par;
};


void Hair::deleteStrandPoint(Strand* strand, int idx){
	int lockListLen = lockList.length;
	for(int k = 0; k < lockListLen; k++){
		Lock* lock = lockList[k];

		M_Vec2f& hole = lock->hole.refValue();
		M_Int& scpi = lock->scpi.refValue();
		M_<Strand*>& scpiLathe = lock->scpiLathe;

		int nHoles = hole.length;

		for(int i = 0; i < nHoles; ){
			if(scpiLathe[i] == strand){
				int pointIdx = scpi[i];
				if(pointIdx >= idx){
					if(pointIdx > idx){
						scpi[i] = pointIdx - 1;
					}
					else{
						hole.deleteElement(i);
						scpi.deleteElement(i);
						scpiLathe.deleteElement(i);
						nHoles--;
						continue;
					}
				}
			}
			i++;
		}
	}
}

void Hair::insertStrandPoint(Strand* strand, int idx){
	int lockListLen = lockList.length;
	for(int k = 0; k < lockListLen; k++){
		Lock* lock = lockList[k];
		M_Int& scpi = lock->scpi.refValue();
		M_<Strand*>& scpiLathe = lock->scpiLathe;
		int nHoles = scpi.length;
		for(int i = 0; i < nHoles; i++){
			if(scpiLathe[i] == strand){
				if(scpi[i] > idx){
					scpi[i]++;
				}
			}
		}
	}
}



void Strand::processFields(){
	if(uKnot.length == 0){
		int uDim = 1;
		int vDim = point.length;
		dim.refValue() = Vec2f(uDim, vDim);
		dpoint.refValue() = point;
		M_Float& weight = this->weight.refValue();
		weight.length = vDim;
		for(int w = vDim; --w>=0;){
			weight[w] = 1;
		}

		uKnot.refValue() = M_Float(0, 0, 0, .25, .25, .5, .5, .75, .75, 1, 1, 1);
		updateNurbsCurveKnotMid0(vDim, vKnot.refValue());
	}
	NurbsPatch::processFields();
}

void processFieldsPlus(Node& node){
	node.processFields();
	if(Build* build = dynamic_cast<Build*>(&node)){
		if(dynamic_cast<NurbsPatch*>(&node) == 0){
			if(NurbsPatch* patch = build->getPatch()){
				if(Seamless* smls = patch->smls){
					if(smls->bnspe){
						M_<NurbsPatch*>& cpNodeList = smls->cpNodeList;
						int plidx = patch->plidx;
						for(int k = build->numOfPatches(); --k>=0;){
							cpNodeList[plidx++]->bnspeBuildPolies();
						}
					}
				}
			}
		}
	}
}

template<class T>
void FD<T,Multiple>::connect(Node* node, Node* protoInterface, Field *protoField){
	S_Route( protoInterface, protoInterface->getFieldIndex(protoField), node, node->getFieldIndex(this));
	length.is=true;
}

template<>
bool M_<char>::write(String fileName)const{
	FILE* file = fopen(fileName.c_str(), "w+b");
	if(file==0)return false;
	fwrite( getArray(), length, 1, file);
	fclose(file);
	return true;
}

void Polymap::convertCoordIndexToTriangle(Art* part, int h, int c, int s){
	M_Triangle& triangle = part->triangle;
	const M_Int& rc = this->rc;
	int ia = 0;
	int ib = 0;
	int ic = 0;

	int ta = 0;
	int tb = 0;
	int tc = 0;
	for(int ci=0; ci < coordIndex.length; ){
		
		ia = rcCoord(ci++, ta);
		ib = rcCoord(ci++, tb);
		ic = rcCoord(ci++, tc);

		triangle[s++] = S_Triangle(part,ia+h, ta+c, part,ib+h,tb+c, part,ic+h,tc+c);
	}
}

void TextureEffect::addTextureMapper(Stem& stem){
	stem.uWrap=false;
	stem.textureMapper.refValue()+=new TextureMapper;
}

void TextureEffect::textureMapDisk(Art* part, float radius, float angle){
	Stem stem;
	stem.part+=part;
	TextureMapper* textureMapper=new TextureMapper;
	stem.textureMapper.refValue()+=textureMapper;
	textureMapper->autoSize3d=false;
	textureMapper->size3d=Vec2f(radius*2,radius*2);
	textureMapper->position3d=Vec3f(-radius,radius, 0);
	textureMapper->orientation=Rotation(1,0,0,angle);
	stem.doBuild();
}

void Seamless::convertToCpFormat(){
	createCpFromPatchPoints();
	createSharedPointsFromStitchPolys();
	cleanCps();
}

Color3f ColorPatch::getColor(Vec2f crd){
	int uDim = dim.x;
	int vDim = dim.y;
	int len = vDim * uDim;
    M_Color3f& m_color = this->color.refValue();

	float uSpace = 1.0 / static_cast<float>(uDim - 1);
	float vSpace = 1.0 / static_cast<float>(vDim - 1);

	float uMult = 1.0 / uSpace;
	float vMult = 1.0 / vSpace;

	float uGrid = floor(crd.x / uSpace);
	float vGrid = floor(crd.y / vSpace);
	if(uGrid >= uDim - 1){
		uGrid = uDim - 2;
	}
	if(vGrid >= vDim - 1){
		vGrid = vDim - 2;
	}
	int uGridI = uGrid;
	int vGridI = vGrid;
	Vec2f mcrd = Vec2f((crd.x - uGrid * uSpace) * uMult, (crd.y - vGrid * vSpace) * vMult);

	Color3f swColor = m_color[vGridI * uDim + uGridI];
	Color3f seColor = m_color[vGridI * uDim + uGridI + 1];
	Color3f nwColor = m_color[(vGridI + 1) * uDim + uGridI];
	Color3f neColor = m_color[(vGridI + 1) * uDim + uGridI + 1];

	Color3f sColor = swColor * (1.0 - mcrd.x) + seColor * mcrd.x;
	Color3f nColor = nwColor * (1.0 - mcrd.x) + neColor * mcrd.x;
	Color3f color =  sColor * (1.0 - mcrd.y) + nColor * mcrd.y;

	return color;
}

void ColorPatch::doColorForStem(Art* part, M_Vec3f& coord, const int h, const int texCoordBase){
	int uDim = dim.x;
	int vDim = dim.y;
	int len = vDim * uDim;
    M_Color3f& m_color = this->color.refValue();

	if(len > m_color.length){
		m_color.length = len;
	}


	float uSpace = 1.0 / static_cast<float>(uDim - 1);
	float vSpace = 1.0 / static_cast<float>(vDim - 1);

	float uMult = 1.0 / uSpace;
	float vMult = 1.0 / vSpace;


    ColorEffect* colorEffect = dynamic_cast<ColorEffect*>(part->workEffect);
    if(colorEffect==0)return;
    M_Color3f& eColor=dynamic_cast<ColorEffect*>(part->workEffect)->color;

    {


		int coordLen = coord.length;
	    for(int i = h, j = texCoordBase; i < coordLen; i++, j++){
			Vec3f crd = coord[i];
			float uGrid = floor(crd.x / uSpace);
			float vGrid = floor(crd.y / vSpace);
			if(uGrid >= uDim - 1){
				uGrid = uDim - 2;
			}
			if(vGrid >= vDim - 1){
				vGrid = vDim - 2;
			}
			int uGridI = uGrid;
			int vGridI = vGrid;
			Vec2f mcrd = Vec2f((crd.x - uGrid * uSpace) * uMult, (crd.y - vGrid * vSpace) * vMult);

			Color3f swColor = m_color[vGridI * uDim + uGridI];
			Color3f seColor = m_color[vGridI * uDim + uGridI + 1];
			Color3f nwColor = m_color[(vGridI + 1) * uDim + uGridI];
			Color3f neColor = m_color[(vGridI + 1) * uDim + uGridI + 1];


			Color3f sColor = swColor * (1.0 - mcrd.x) + seColor * mcrd.x;
			Color3f nColor = nwColor * (1.0 - mcrd.x) + neColor * mcrd.x;
			Color3f color =  sColor * (1.0 - mcrd.y) + nColor * mcrd.y;
			eColor[j] = color;
        }
    }
}

